/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Manages app level upgrade process
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_upgrade.h"
#include "crypto/crypto.h"
#include "crypto/src/md5.h"
#include "utilities/upgrade_metadata.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * uploads upgrade data
 * @param message
 */
void app_commands_upgrade_upload_data(STRUCTURED_MESSAGE* message)
{
    UPGRADE_METADATA_LOAD_ID load_id = message->data[0];
    
    UPGRADE_METADATA_UPLOAD_STATUS result = upgrade_metadata_upload_data(load_id, &message->data[1], message->data_length - 1);

    // Add the Id to the message
    if (result != UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS)
    {
        memcpy(message->data, &load_id, sizeof(UPGRADE_METADATA_LOAD_ID));
        memcpy(message->data + sizeof(UPGRADE_METADATA_LOAD_ID), &result, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
        message->data_length = sizeof(UPGRADE_METADATA_LOAD_ID) + sizeof(UPGRADE_METADATA_UPLOAD_STATUS);
        message->response = STATUS_UPDATE;      

        app_commands_send_message(message);
    }
}

/**-----------------------------------------------------------------------------
 * Starts an upgrade upload
 * @param message
 */
void app_commands_upgrade_upload_end(STRUCTURED_MESSAGE* message)
{
    logger_log_debugf(123, "UPGRADE METADATA - Upload completing");

    UPGRADE_METADATA_LOAD_ID load_id = message->data[0];
    
    UPGRADE_METADATA_UPLOAD_STATUS result = upgrade_metadata_upload_end(load_id);

    memcpy(message->data, &load_id, sizeof(UPGRADE_METADATA_LOAD_ID));
    memcpy(message->data + sizeof(UPGRADE_METADATA_LOAD_ID), &result, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
    message->data_length = sizeof(UPGRADE_METADATA_LOAD_ID) + sizeof(UPGRADE_METADATA_UPLOAD_STATUS);
    message->response = STATUS_UPDATE;      
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Starts an upgrade upload
 * @param message
 */
void app_commands_upgrade_upload_start(STRUCTURED_MESSAGE* message)
{
    logger_log_debugf(123, "UPGRADE START");

    uint32_t size;
    uint8_t checksum[MD5_DIGEST_SIZE];
    
    uint8_t major_version = message->data[0];
    uint8_t minor_version = message->data[1];
    memcpy(&size, &message->data[2], 4);
    
    // Extract the checksum from the message
    memcpy(checksum, &message->data[6], MD5_DIGEST_SIZE);
    bool override = (bool)message->data[6 + MD5_DIGEST_SIZE];
    
    logger_log_debugf(123, "UPGRADE INFO ver:%d.%d size:%d ovr:%s  --", 
            major_version, minor_version,
            size, 
            override == 0 ? "FALSE" : "TRUE");
    logger_log_debugf(123, "UPGRADE Checksum %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d --", 
            checksum[0],
            checksum[1],
            checksum[2],
            checksum[3],
            checksum[4],
            checksum[5],
            checksum[6],
            checksum[7],
            checksum[8],
            checksum[9],
            checksum[10],
            checksum[11],
            checksum[12],
            checksum[13],
            checksum[14],
            checksum[15]
            );

    UPGRADE_METADATA_LOAD_ID result = upgrade_metadata_prepare_upload(major_version, 
            minor_version, 
            size, 
            checksum, 
            override);
    
    message->data_length = sizeof(UPGRADE_METADATA_LOAD_ID);
    memcpy(message->data, &result, sizeof(UPGRADE_METADATA_LOAD_ID));
    
    UPGRADE_METADATA_UPLOAD_STATUS status;
    
    if (result > 0)
    {
        status = UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
        memcpy(message->data + message->data_length, &status, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
        message->response = STATUS_UPDATE;      
    }
    else
    {
        status = UPGRADE_METADATA_UPLOAD_STATUS_START_INVALID;
        memcpy(message->data + message->data_length, &status, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
        message->response = COMMAND_NOT_RECOGNISED;
    }
    
    message->data_length += sizeof(UPGRADE_METADATA_UPLOAD_STATUS);
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Callback invoked when a packet of data has been uploaded into the image data 
 * @param request
 */
void app_commands_upgrade_write_callback(UPGRADE_METADATA_LOAD_ID load_id, FLASH_MEMORY_CLIENT_REQUEST* request)
{
    logger_suppress_command_logging(true);
    
    //logger_log_debugf(0, "app_commands_upgrade_write_callback %d --", load_id);
    STRUCTURED_MESSAGE message;
    structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_RS232);
    message.command = APP_COMMANDS_UPGRADE_LOAD_DATA;
    message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
    
    // Add the Id to the message
    UPGRADE_METADATA_UPLOAD_STATUS status = UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
    memcpy(message.data, &load_id, sizeof(UPGRADE_METADATA_LOAD_ID));
    memcpy(message.data + sizeof(UPGRADE_METADATA_LOAD_ID), &status, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
    message.data_length = sizeof(UPGRADE_METADATA_LOAD_ID) + sizeof(UPGRADE_METADATA_UPLOAD_STATUS);
    message.response = STATUS_UPDATE;      
    
    app_commands_send_message(&message);   
    
    logger_suppress_command_logging(false);
}

/* *****************************************************************************
 End of File
 */
