/*******************************************************************************
  MPLAB Harmony Application Source File
  
  Company:
    Microchip Technology Inc.
  
  File Name:
    app.c

  Summary:
    This file contains the source code for the MPLAB Harmony application.

  Description:
    This file contains the source code for the MPLAB Harmony application.  It 
    implements the logic of the application's state machine and it may call 
    API routines of other MPLAB Harmony modules in the system, such as drivers,
    system services, and middleware.  However, it does not call any of the
    system interfaces (such as the "Initialize" and "Tasks" functions) of any of
    the modules in the system or make any assumptions about when those functions
    are called.  That is the responsibility of the configuration-specific system
    files.
 *******************************************************************************/

// DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright (c) 2013-2014 released Microchip Technology Inc.  All rights reserved.

Microchip licenses to you the right to use, modify, copy and distribute
Software only when embedded on a Microchip microcontroller or digital signal
controller that is integrated into your product or third party product
(pursuant to the sublicense terms in the accompanying license agreement).

You should refer to the license agreement accompanying this Software for
additional information regarding your rights and obligations.

SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER
CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR
OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
(INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 *******************************************************************************/
// DOM-IGNORE-END


// *****************************************************************************
// *****************************************************************************
// Section: Included Files 
// *****************************************************************************
// *****************************************************************************

#include "app.h"
#include "app_config.h"
#include "app_context.h"
#include "app_initialise.h"
#include "utilities/logger.h"
#include "utilities/rgb_led.h"
#include "utilities/rs232.h"
#include "utilities/structured_message.h"
#include "utilities/timers.h"
#include "utilities/upgrade_metadata.h"

//#include "framework/driver/spi/static/drv_spi_static.h"

// *****************************************************************************
// *****************************************************************************
// Section: Global Data Definitions
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Functions
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
// *****************************************************************************
// Section: Application Local Functions
// *****************************************************************************
// *****************************************************************************

// Forward declaration
//TODOstatic long check_temperature(APP_ADC_INPUTS input, SYSTEM_FAULT_CODE fault_code);
//TODOstatic void fault_callback(SYSTEM_FAULT* fault);
static void message_logged(LOGGER_MESSAGE* message);
static void structured_message_callback(STRUCTURED_MESSAGE* message);

/**-----------------------------------------------------------------------------
 * 
 */
static void jump_to_application()
{
    // Disable all interrupts
    SYS_INT_Disable();

    // APP_LAUNCH_ADDRESS DEFINED IN app_config at the moment, but may need moving to a 
    // product specific files that also includes configuration bits?
    void (*fptr)(void);
    fptr = (void (*)(void))APP_LAUNCH_ADDRESS;
    fptr();
}

/**-----------------------------------------------------------------------------
 * Handles messages generated by the logger
 * @param message
 */
static void message_logged(LOGGER_MESSAGE* message)
{
// ST60 Display has issues when used in the bootloader as the actual sending of the messages 
// carries on over into the running of the main app - as the baud rates are different 38400/115200
// it messes up the RS232 which then stops working...
//TODO - We are working on making UART1 the RS232 message/Logger port by default - once
// thats done we should be able to remove this.    
//#ifndef APP_CONFIG_ST60_DISPLAY    
    // Convert the logger message to a structured message
    STRUCTURED_MESSAGE output_message;
    output_message.command = APP_COMMANDS_LOGGER_MESSAGE;
    output_message.zone = STRUCTURED_MESSAGE_ZONE_1;
    output_message.response = STATUS_UPDATE;
    output_message.data_only = false;
    
    // wrap up in logger_pack_message()?)
    int len = sizeof(LOGGER_MESSAGE) + strlen(message->message);
    char buffer[len];
    memcpy(&buffer, &message->code, sizeof(LOGGER_CODE));
    memcpy(&buffer[2], &message->type, sizeof(LOGGER_TYPE));
    memcpy(&buffer[4], &message->message[0], min(strlen(message->message), STRUCTURED_MESSAGE_MAX_DATA));
    buffer[len - 1] = 0;
    
    structured_message_set_message_data(&output_message, buffer, len);
    rs232_send_message(&output_message);
    
    // Send to the first TCP/IP client
    output_message.state_data[0] = 0;
    //TODOtcpip_send_structured_message(&output_message);
//#endif    
}

/**-----------------------------------------------------------------------------
 * Works out which model we are from the pins
 */
static void setup_software_model()
{
    SOFTWARE_MODEL = SOFTWARE_MODEL_SA30;
    SOFTWARE_MODEL_STRING = SOFTWARE_VERSION_STRINGS[SOFTWARE_MODEL];
}

/**-----------------------------------------------------------------------------
 * A number of the app states have some standard processing which is wrapped up here
 */
static void standard_processing()
{
    // RGB LEd
    rgb_led_process();

    rs232_process();
    
    flash_memory_process();
    
    // Clear the watch dog
    //TODO watchdog_reset();
}

/**-----------------------------------------------------------------------------
 * Handle structured messages received via rs232
 * @param message - A pointer to a structured message - do NOT store this pointer
 */
static void structured_message_callback(STRUCTURED_MESSAGE* message)
{
    switch (message->source)
    {
        case STRUCTURED_MESSAGE_SOURCE_REMOTE_RAW:
        {
            // Add to input context - this will be processed in a short while as 
            // an STRUCTURED_MESSAGE_SOURCE_INPUTS message
            //TODOinput_context_set_remote_message(message);
            break;
        }
        
        default:
        {
            // Now process the message
            // Translate command based on message source
            // Process message? or store for later processing
            //app_commands_process_message(message);
            
            // Check if the external message indicates activity
            if ((message->source == STRUCTURED_MESSAGE_SOURCE_RS232) ||
                (message->source == STRUCTURED_MESSAGE_SOURCE_TCPIP))
            {
                if (message->is_activity)
                {
                    //TODOinput_context_external_message_processed(message);
                }
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Called during upgrade progress
 */
void upgrade_callback(UPGRADE_METADATA_LOAD_ID load_id, UPGRADE_METADATA_UPGRADE_PROGRESS* progress)
{
    switch (progress->state)
    {
        case UPGRADE_METADATA_UPGRADE_PROGRESS_STATE_ERASING:
        {
            logger_log_infof(LOGGER_CODE_UPGRADE, "APP UPGRADE - Starting Upgrade version %d.%d ", progress->version_major, progress->version_minor);
            logger_log_infof(LOGGER_CODE_UPGRADE, "APP UPGRADE - Erasing Flash ");
            break;
        }
        
        case UPGRADE_METADATA_UPGRADE_PROGRESS_STATE_WRITING:
        {
            //logger_log_infof(LOGGER_CODE_UPGRADE, "APP UPGRADE - Writing program memory %d % complete ", (progress->bytes_written * 100) / progress->size);
            
            if ((progress->bytes_written * 100) / progress->size == 100)
            {
                logger_log_infof(LOGGER_CODE_UPGRADE, "APP UPGRADE - Complete ");
            }
            break;
        }
    }
}

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_Initialize ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Initialize ( void )
{
    // Software version SA30
    setup_software_model();

    /* Place the App state machine in its initial state. */
    app_context_initialise();
}

/******************************************************************************
  Function:
    void APP_Tasks ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Tasks ( void )
{
    /* Check the application's current state. */
    switch ( app_context_get_app_state() )
    {
        /* Application's initial state. */
        case APP_STATE_INIT:
        {
            initialise_rgb_led();
            
            // Initialise the timers
            initialise_timers();
            
            // Initialise RS232 port
            initialise_rs232(structured_message_callback);
                
            // Initialise the logger
            initialise_logger(message_logged);
            
            logger_log_debugf(0, "----------------------------------------------");
            logger_log_debugf(0, "----------------------------------------------");
            logger_log_debugf(0, "----------------------------------------------");
            logger_log_debugf(0, "BOOTLOADER START ");
            logger_log_debugf(0, "----------------------------------------------");
            logger_log_debugf(0, "----------------------------------------------");
            
            // Initialise flash_memory
            initialise_flash_memory();
            logger_log_debugf(0, "APP_STATE_INIT 1 ");
            
            // Initialise upgrade info
            upgrade_metadata_initialise(NULL, upgrade_callback);
            logger_log_debugf(0, "APP_STATE_INIT 2");
            
            rgb_led_process();
 
            // Start a power cycle timer to give time for the Power to fully kick in
            timers_reset(TIMER_INDEX_TEST);
            
            app_context_set_app_state(APP_STATE_CHECK_FOR_UPGRADE);
            //app_context_set_app_state(APP_STATE_JUMP_TO_APP);
            
            break;
        }
   
        case APP_STATE_CHECK_FOR_UPGRADE:
        {
            logger_log_debugf(0, "APP_STATE_CHECK_FOR_UPGRADE");
            standard_processing();
            
            if (upgrade_metadata_is_loaded())
            {
                upgrade_metadata_print_metadata();
                     
                if (upgrade_metadata_get_upgrade_state() == UPGRADE_METADATA_STATE_AVAILABLE_INSTALL)
                {
                    logger_log_debugf(0, "UPGRADE - INSTALL AVAILABLE ");
                    logger_process();
                    
                    rgb_led_flash();
                    app_context_set_app_state(APP_STATE_LOAD_UPGRADE);                  
                }
                else
                {
                    logger_log_debugf(0, "NO UPGRADE - JUMP TO APP ");
                    logger_process();
                    app_context_set_app_state(APP_STATE_JUMP_TO_APP);
                }
            }
            break;
        }
        
        case APP_STATE_LOAD_UPGRADE:
        {
            standard_processing();
            
            UPGRADE_METADATA_UPGRADE_STATUS status = upgrade_metadata_perform_upgrade();
            switch (status)
            {
                case UPGRADE_METADATA_UPGRADE_STATUS_OK:
                {
                    logger_log_debugf(0, "UPGRADE COMPLETE ");
                    // Set pause before app launch
                    timers_reset(TIMER_INDEX_UPGRADE_LAUNCH_APP);
 
#ifdef APP_CONFIG_ST60_DISPLAY 
                    app_context_set_app_state(APP_STATE_JUMP_TO_APP);                    
#else
                    app_context_set_app_state(APP_STATE_RESTART);
#endif                    
                    break;
                }

                case UPGRADE_METADATA_UPGRADE_STATUS_ERROR:
                {
                    logger_log_debugf(0, "UPGRADE ROLLBACK ");
                    app_context_set_app_state(APP_STATE_ROLLBACK_UPGRADE);
                    break;
                }
                
                default:
                    break;
            }
            break;
        }
        
        case APP_STATE_ROLLBACK_UPGRADE:
        {
            // This will need to identify the last successful know version and 
            // set the upgrade to load from that
            logger_log_debugf(0, "APP UPGRADE ROLLBACK ");
            if (upgrade_metadata_prepare_rollback())
            {
                // Set pause before app launch
                timers_reset(TIMER_INDEX_UPGRADE_LAUNCH_APP);
            
                app_context_set_app_state(APP_STATE_LOAD_UPGRADE);
            }
            else
            {
                // We cant roll back so try launching app anyway
                // Set pause before app launch
                timers_reset(TIMER_INDEX_UPGRADE_LAUNCH_APP);
            
                app_context_set_app_state(APP_STATE_JUMP_TO_APP);
            }
            
            break;
        }
            
        case APP_STATE_JUMP_TO_APP:
        {
            standard_processing();
            
            if (timers_has_expired(TIMER_INDEX_UPGRADE_LAUNCH_APP))
            {
                logger_log_infof(0, "-----------------------------------------------------");
                logger_log_infof(0, "Launching Application");
                logger_log_infof(0, "-----------------------------------------------------");
                logger_log_infof(0, "-----------------------------------------------------");
                logger_process();

                jump_to_application();
            }
            
            break;
        }
        
        case APP_STATE_RESTART:
        {
            standard_processing();
            
            if (timers_has_expired(TIMER_INDEX_UPGRADE_LAUNCH_APP))
            {
                // The Restart mode has been added because I was having trouble with 
                // doing a second Upgrade Transfer when running in a session that
                // was started as a result of an Upgrade.  For some reason the Send queue was sticking
                // if we just jumped straight to the app, so we are doing a restart instead
                // and having another run through the bootloader.
                logger_log_infof(0, "-----------------------------------------------------");
                logger_log_infof(0, "Restarting");
                logger_log_infof(0, "-----------------------------------------------------");
                logger_log_infof(0, "-----------------------------------------------------");
                logger_process();
                
                //https://www.microchip.com/forums/m/tm.aspx?m=1005422
                // MUST disable interrupts and dma, as reset timing is critical
                // use the compiler built-in function to disable interrupts
                SYS_INT_StatusGetAndDisable();
                
                SYS_RESET_SoftwareReset();
            }
        }
        
        /* The default state should never be executed. */
        default:
        {
            break;
        }
    }
}

/*******************************************************************************
 End of File
 */
