/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    avatar.h

  @Summary
    Encapsulates communication with the avatar module

  @Description
    Encapsulates communication with the avatar module.  Internally this maintains 
    queues for incoming and outgoing packets.
 */
/* ************************************************************************** */

#ifndef _AVATAR_H    /* Guard against multiple inclusion */
#define _AVATAR_H

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include "system_definitions.h"
#include "utilities/avatar_data.h"
#include "utilities/avatar_message.h"
#include "utilities/avatar_packet.h"
#include "utilities/dirac.h"
#include "utilities/i2c_slave.h"
//#include "driver/i2c/drv_i2c.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

//! Used for defining a length to allocate a short message
#define AVATAR_ABITARY_SHORT_MESSAGE_LENGTH  22

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
 
    /**
     * Avatar processing states
     * Note: These should be in ascending order of progress as their values are check 
     * or things like whether handshaking is complete or not.
     */
    typedef enum
    {
        AVATAR_STATE_INITIALISE,
        AVATAR_STATE_HANDSHAKE,
        AVATAR_STATE_CONFIGURATION,
        AVATAR_STATE_RUNNING,
        AVATAR_STATE_ERROR,
    } AVATAR_STATE;
   
    /**
     * Callback when data is received from the Avatar
     */
    typedef void (*AVATAR_CALLBACK)(AVATAR_MESSAGE* message);
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Used to set up avatar once handshaking is complete etc..
     * @return 
     */
    void avatar_configure();
    
    /**
     * Returns the device specific during initialisation
     * @return 
     */
    AVATAR_DEVICE avatar_get_device();
    
    /**
     * Gets a packet for use with avatar_send_packet
     * Use with avatar_packet_add_data to construct the packet
     * @param message_id - The message type for the packet
     * @param min_size - The minimum size for the packet (including the header) - used to allocate a packet
     * @return 
     */
    AVATAR_PACKET* avatar_get_packet(AVATAR_MESSAGE_ID message_id, uint16_t min_size);
    
    /**
     * Gets the current avatar processing state
     * @return 
     */
    AVATAR_STATE avatar_get_state();
    
    /**
     * Gets the avatar upgrade option
     */
    AVATAR_UPGRADE_OPTION avatar_get_upgrade_option();
    
    /**
     * Performs handshaking process with the avatar.
     * Returns tru when handshaking complete
     */
    bool avatar_handshake();
    
    /**
     * Callback used by I2C to inform when data is available 
     * https://www.microchip.com/forums/m1065552.aspx
     */
    void avatar_i2c_slave_callback(I2C_SLAVE_EVENT* message);
    
    /**
     * Initialises the avatar module
     */
    void avatar_initialise(AVATAR_DEVICE device, 
            AVATAR_CALLBACK message_callback,
            AVATAR_STATE_ID* subscriptions,
            uint8_t subscriptions_len);
    
    /**
     * Determines whether the avatar has outstanding messages/packets to process (send/receive)
     */
    bool avatar_is_active();
    
    /**
     * Determines whether the avatar is in a state to be communicated with
     */
    bool avatar_is_loaded();
      
    /**
     * Set to true while the Avatar is processing a volume message.  Used to stop 
     * volume message being sent back
     */
    bool avatar_is_processing_mute();
      
    /**
     * Set to true while the Avatar is processing a volume message.  Used to stop 
     * volume message being sent back
     */
    bool avatar_is_processing_volume();

    /**
     * Determines whether the avatar is in the running state
     */
    bool avatar_is_running();
    
    /**
     * Adds a packet to the processing queue
     * @param packet
     */
    void avatar_prepare_for_processing(AVATAR_PACKET* packet);

    /**
     * Perform Avatar processing
     * Returns true if any Receive processing has occurred
     */
    bool avatar_process();
    
    /**
     * Resets the avatar queues when going into standby //STANDBY
     */
    void avatar_reset_queues();
    
    /**
     * Queues a packet for sending to the avatar
     * @param packet
     */
    void avatar_send_packet(AVATAR_PACKET* packet);
      
    /**
     * Used by app_commands_process_volume to indicate that it is processing an
     * Avatar mute message
     */
    void avatar_set_processing_mute(bool on);
      
    /**
     * Used by app_commands_process_volume to indicate that it is processing an
     * Avatar volume message
     */
    void avatar_set_processing_volume(bool on);
    
    /**
     * Gets a string representation of the message id.  Note if there is no match this may
     * give unpredictable results
     * @param message_id
     * @return 
     */
    char* avatar_state_id_string(AVATAR_STATE_ID state_id);

    /**
     * Called to reset the Avatar 
     */
    void avatar_stop();
    
    /**
     * Returns whether an upgrade is available
     * @return 
     */
    bool avatar_upgrade_available();
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _AVATAR_H */

/* *****************************************************************************
 End of File
 */
