/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _AVATAR_BROWSING_H    /* Guard against multiple inclusion */
#define _AVATAR_BROWSING_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include "utilities/avatar_data.h"
#include "utilities/avatar_message.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    typedef enum
    {
        AVATAR_BROWSING_STATE_INACTIVE,
        AVATAR_BROWSING_STATE_STARTING,
        AVATAR_BROWSING_STATE_BROWSING,
        AVATAR_BROWSING_STATE_PLAYING,
    } AVATAR_BROWSING_STATE;

    typedef enum
    {
        AVATAR_BROWSING_SKIP_CONTROL_NONE,
        AVATAR_BROWSING_SKIP_CONTROL_FBWD,
        AVATAR_BROWSING_SKIP_CONTROL_BACK,
        AVATAR_BROWSING_SKIP_CONTROL_EXIT,
        AVATAR_BROWSING_SKIP_CONTROL_FORWARD,
        AVATAR_BROWSING_SKIP_CONTROL_FFWD,
    } AVATAR_BROWSING_SKIP_CONTROL;
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
 
    /**
     * Selects the current browse item
     */
    void avatar_browsing_browse_current(bool play_all);
    
    /**
     * Displays th skip controls
     */
    void avatar_browsing_display_skip(AVATAR_BROWSING_SKIP_CONTROL direction);
    
    /**
     * Populates the supplied buffer with a current breadcrumb representation
     * @param buffer
     * @param len
     */
    void avatar_browsing_get_breadcrumb_string(char* buffer, size_t len);
    
    /**
     * Gets a reference to current items 
     * @return 
     */
    AVATAR_BROWSE_ITEM* avatar_browsing_get_current_item();
    
    /**
     * Gets the current browsing level
     * @return 
     */
    uint8_t avatar_browsing_get_current_level();
    
    /**
     * Gets the avatar browsing state that was current when the browsing menu was left
     * This is used to identify when we leave the currently laying screen whether we were browsing beforehand
     * @return 
     */
    AVATAR_BROWSING_STATE avatar_browsing_get_exit_state();
    
    /**
     * Gets a reference to current items 
     * @return 
     */
    AVATAR_BROWSE_ITEM* avatar_browsing_get_items();
 
    /**
     * Gets a reference to the current parent item 
     * @return 
     */
    AVATAR_BROWSE_ITEM* avatar_browsing_get_parent_item();
 
    /**
     * Get the current skip control index (0 = No skip menu)
     * @return 
     */
    AVATAR_BROWSING_SKIP_CONTROL avatar_browsing_get_skip_control();
    
    /**
     * Gets the current avatar browsing state
     * @return 
     */
    AVATAR_BROWSING_STATE avatar_browsing_get_state();
 
    /**
     * Gets a reference to the item to be displayed at the top of the list items 
     * @return 
     */
    AVATAR_BROWSE_ITEM* avatar_browsing_get_top_item();
 
    /**
     * Checks whether we have actually played a track in this browsing session.
     * @return 
     */
    bool avatar_browsing_has_played_in_session();
    
    /**
     * Regular check to see if USB stick has been inserted or UPnP services available
     * @param message
    */
    void avatar_browsing_process();
 
    /**
     * Callback processing for Avatar Browse back request
     * @param message
    */
    bool avatar_browsing_process_browse_back(AVATAR_MESSAGE* message);
 
    /**
     * Callback processing for Avatar Browse Page request
     * @param message
    */
    bool avatar_browsing_process_browse_folder(AVATAR_MESSAGE* message);
 
    /**
     * Callback processing for Avatar Browse Item request
     * @param message
    */
    //bool avatar_browsing_process_browse_item(AVATAR_MESSAGE* message);
 
    /**
     * Callback processing for Avatar Browse Page request
     * @param message
    */
    //bool avatar_browsing_process_browse_page(AVATAR_MESSAGE* message);

    /**
     * Used to restart a previous playing track
     */
    void avatar_browsing_restart_playing();
    
    /**
     * Move to the Browsing item next to the current one i the hierarchy
     */
    void avatar_browsing_select_next();

    /**
     * Move to the Browsing item previous to the current one i the hierarchy
     */
    void avatar_browsing_select_previous();
    
    /**
     * makes a Browse request to the avatar
     * @param 
     */
    void avatar_browsing_send_browse_back(AVATAR_BROWSE_BACK_LEVEL level);
    
    /**
     * makes a Browse request to the avatar
     * @param 
     */
    //void avatar_browsing_send_browse_item(uint16_t item_id, AVATAR_BROWSE_PLAY_OPTION option, uint16_t num_items);
    
    /**
     * makes a Browse request to the avatar
     * @param 
     */
    void avatar_browsing_send_browse_folder(uint16_t folder_id, uint16_t start_item_id, uint16_t num_items);
    
    /**
     * makes a Browse request to the avatar
     * @param 
     */
    //void avatar_browsing_send_browse_page(uint16_t start_item_id, uint16_t num_items);
    
    /**
     * makes a Play Track request to the avatar
     * @param 
     */
    void avatar_browsing_send_play_track(uint16_t track_id);

    /**
     * Used to set the Avatar browsing state into running mode (also sets the last_exit state)
     * @param state
     */
    void avatar_browsing_set_running_state(AVATAR_BROWSING_STATE state);

    /**
     * Set the current skip control index (0 = No skip menu)
     * @return 
     */
    void avatar_browsing_set_skip_control(AVATAR_BROWSING_SKIP_CONTROL skip_control);

    /**
     * Sets up an Avatar browsing session when the NET input becomes current (via the menu)
     */
    void avatar_browsing_start();

    /**
     * Starts browsing the folder hierarchy
     * @param - start_folder_name - specific that name of the top level folder to browse into e.g USB
     */
    void avatar_browsing_start_browsing(char* start_folder_name);

    /**
     * Stops an Avatar browsing session when the NET input becomes current (via the menu)
     */
    void avatar_browsing_stop();

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
