/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    avatar_control.h

  @Summary
    Functions for controlling the Avatar

  @Description
    Functions for controlling the Avatar
 */
/* ************************************************************************** */

#ifndef _AVATAR_CONTROL_H    /* Guard against multiple inclusion */
#define _AVATAR_CONTROL_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/avatar.h"
#include "utilities/avatar_data.h"
#include "utilities/avatar_process_rule.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Requests a factory reset in the avatar
     */
    void avatar_factory_reset();

    /**
     * Send a mute/unmute
     */
    void avatar_mute(AVATAR_MUTE_STATE state);

    /**
     * Send a pause message to the avatar
     */
    void avatar_pause_playback();
 
    /**
     * Makes a request to the Avatar to retrieve the album art for the supplied URL
     */
    void avatar_request_album_art(char* album_art_url);
    
    /**
     * Send a Firm ware version request to the Avatar
     */
    void avatar_request_dirac_curve_names();
    
    /**
     * Send a Firm ware version request to the Avatar
     */
    void avatar_request_firmware_version();
    
    /**
     * Send a Avatar System Info request to the Avatar
     */
    void avatar_request_mime_type();
    
    /**
     * Request the region from the Avatar
     */
    void avatar_request_region();
    
    /**
     * Send a Serial Number request to the Avatar
     */
    void avatar_request_serial_number();

    /**
     * Send a Avatar System Info request to the Avatar
     */
    void avatar_request_system_info();

    /**
     * Send the message infoming the avatar of the current balance value (-128-127)
     */
    void avatar_send_balance(int8_t balance);
    
    /**
     * Sets the Control4 Discovery Active state
     * @param 
     */
    void avatar_send_control4_discovery(AVATAR_CONTROL4_DISCOVERY index);
 
    /**
     * Send the message instructing the Avatar to send a Control4 Identity message
     */
    void avatar_send_control4_identity();

    /**
     * Sends the response to the Crestron configure message
     */
    void avatar_send_crestron_configure_response();
    
    /**
     * Sets the Crestron Discovery Active state
     * @param 
     */
    void avatar_send_crestron_discovery(AVATAR_CRESTRON_DISCOVERY index);
    
    /**
     * Send a message to the avatar indicating the devices upgrade status
     * @param status
     */
    void avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS status);
   
    /**
     * Send a message to the avatar indicating the devices headphone status
     * @param status
     */
    void avatar_send_headphone_status(AVATAR_HEADPHONE_STATUS status);
   
    /**
     * Send a message to the avatar indicating the devices headphone status
     * @param status
     */
    void avatar_send_lip_sync_delay(uint16_t delay);
   
    /**
     * Send a message to the avatar indicating the current maximum volume
     * @param status
     */
    void avatar_send_maximum_volume(uint8_t max_volume);
 
    /**
     * Send the message informing the Avatar of the MQA Decoder level
     */
    void avatar_send_mqa_decoder_level(AVATAR_MQA_DECODER_LEVEL level);

    /**
     * Send a player control message to the avatar
     */
    void avatar_send_player_control(AVATAR_PLAYER_CONTROL command);

    /**
     * Send a player control message to the avatar
     */
    void avatar_send_power_state(AVATAR_POWER_STATE state);

    /**
     * Send a repeat message to the avatar
     */
    void avatar_send_repeat_state(AVATAR_REPEAT_STATE state);

    /**
     * Send a repeat message to the avatar
     */
    void avatar_send_shuffle_state(AVATAR_SHUFFLE_STATE state);

    /**
     * Send a seek message to the avatar time in ms
     */
    void avatar_send_track_seek(uint64_t seek_pos);

    /**
     * Sends data out over the Avatars TCP connection
     */
    void avatar_send_tcp_data(uint8_t client_id, char* data, size_t len);
   
    /**
     * Sets the currently selected Dirac curve - DIRAC_NO_CURVE clears the selection
     * @param index - The index of the curve to be set
     */
    void avatar_set_dirac_curve(uint8_t index);
    
    /**
     * Used to inform the avatar that it should switch its input
     * @param input
     */
    void avatar_set_input(AVATAR_INPUT input);
    
    /**
     * Used to inform the avatar that it should switch its region
     * @param input
     */
    void avatar_set_region(AVATAR_REGION region);
    
    /**
     * Used to inform the Avatar of the current input sample rate
     * @param input
     */
    void avatar_set_sample_rate(AVATAR_SAMPLE_RATE sample_rate, bool force);
    
    /**
     * Used to inform the Avatar of the current input sample rate converted from
     * an integer value
     * @param input
     */
    void avatar_set_sample_rate2(uint32_t sample_rate);
    
    /**
     * Sets the advisory upgrade option
     * @param option
     */
    void avatar_set_upgrade_option(AVATAR_UPGRADE_OPTION option);
    
    /**
     * Used to inform the avatar that it should display the new volume
     * @param input
     */
    void avatar_set_volume(uint8_t volume);
    
    /**
     * Invoked when a volume message has been processed
     * @param rule
     * @param packet
     */
    void avatar_set_volume_complete(AVATAR_PROCESS_RULE* rule, AVATAR_MESSAGE* message);
    
    /**
     * Invokes the Avatar TCP Client Control message
     * @param client_id
     * @param state
     */
    void avatar_tcp_client_control(uint8_t client_id, AVATAR_TCP_CLIENT_CONTROL_STATE state);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
