/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    avatar_message.h

  @Summary
    The Avatar Message can be used to obtain StateID specific data from an Avatar Packet.

  @Description
    An Avatar Message parses an Avatar Packet providing a quick look up for specific
    StateIDs included in the packet.  This is useful where the contents of the 
    received packet are unknown and are repeatedly interrogated for specific StateIDs
 */
/* ************************************************************************** */

#ifndef _AVATAR_MESSAGE_H    /* Guard against multiple inclusion */
#define _AVATAR_MESSAGE_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/avatar_data.h"
#include "utilities/avatar_packet.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

//! The maximum number of entries expected in a avatar_packet
#define AVATAR_MESSAGE_MAX_STATES   27  //TODO Review
    
    /**
     * Message metadata
     */
    typedef struct
    {
        AVATAR_ENCODING encoding;
        
        //! A avatar state Id
        AVATAR_STATE_ID state_id;
        
        //! The offset of the state Id in the packet
        uint16_t offset;
        
        //! The length of the data contained in the current packet. This includes the 
        // 2-bytes of the length part of BIN16 fields ONLY if packet_id = 0
        uint16_t data_length;
        
        //! The total length of the state data.  Differs from data_length if the 
        // data is split over multiple packets.
        uint16_t total_length;
    } AVATAR_MESSAGE_LOOKUP;
    
    /**
     * Pointer to analysed avatar packet data that contains values (Set States, 
     * State Change, Get States Response)
     * Note none of the data should be used directly 
     */
    typedef struct 
    {
        //! Used by the processing cache to determine whether message is currently active
        bool active;

        // The message Id for the message
        AVATAR_MESSAGE_ID message_id;
        
        //! Used by the processing cache to identify the message
        uint8_t frame_id;
        
        //! Lookup table for packet data
        AVATAR_MESSAGE_LOOKUP lookup[AVATAR_MESSAGE_MAX_STATES];
        
        //! The number of states contained in the message
        uint8_t num_states;
        
        //! A pointer to the avatar packet
        AVATAR_PACKET* packet;
        
    } AVATAR_MESSAGE;


    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Process a packet containing data from the Avatar (Set States, Get States Response)
     * @param packet
     * @param message - pointer to an AVATAR_MESSAGE to be populated
     */
    void avatar_message_build(AVATAR_PACKET* packet, AVATAR_MESSAGE* message);

    /**
     * Searches the processing queue for a message associated with the packets frame id
     * @param packet
     * @return A pointer to a message representing the packet. If this is a new frame then
     * a free message will be returned (already built)
     */
    AVATAR_MESSAGE* avatar_message_for_packet(AVATAR_PACKET* packet);
    
    /**
     * Gets the number of state ids in the message
     * @return 
     */
    uint8_t avatar_message_get_num_states();
    
    /**
     * Puts the state data into an array.  For bin 16 fields this will be the data only and omit the
     * length part of the bin 16 field.
     * @param message - The message containing the data
     * @param state_id - The state Id for the data
     * @param encoding - The encoding type of the data
     * @param data - A pointer to an array of uint8_t to hold the data
     * @param len - The length of data available in the array - if the actual data is longer only this amount will be copied.
     * @return the length of the data copied (0 if no data or NIL type).
     */
    uint16_t avatar_message_get_state_data(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id, AVATAR_ENCODING* encoding, uint8_t* data, uint16_t data_len);
 
    /**
     * Gets the state Id at a specific index in the message.  Assumes the message has been built
     * with avatar_message_build
     * @param message
     * @param index
     * @return 
     */
    uint16_t avatar_message_get_state_id(AVATAR_MESSAGE* message, uint8_t index);
    
    /**
     * Gets the total length of the state data reported in the first packet
     * @param message
     * @param index
     * @return 
     */
    uint16_t avatar_message_get_state_total_length(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id);
    
    /**
     * Populates a pointer with the address of the actual data in the packet
     * @param message
     * @param state_id
     * @param encoding
     * @param value - A pointer to a pointer that will be set to point to the data in th packet
     * @return true if state_if found in message
     */
    bool avatar_message_get_state_value(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id, AVATAR_ENCODING* encoding, void** value, uint16_t* len);
    
    /**
     * Determines whether an avatar message contains info on a specific state
     * @param message
     * @param state_id
     * @return 
     */
    bool avatar_message_has_state(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id);

    /**
     * Initialises processing cache
     */
    void avatar_message_initialise_cache();
    
    /**
     * Called once a message packet has been processed.  Will either release the 
     * message opr return it to the cache dependent on whether it is an FEND or not.
     * @param message
     */
    void avatar_message_processed(AVATAR_MESSAGE* message);

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
