/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Handles Avatar message processing for incoming messages
 */
/* ************************************************************************** */

#ifndef _AVATAR_MESSAGE_PROCESSING_H    /* Guard against multiple inclusion */
#define _AVATAR_MESSAGE_PROCESSING_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/avatar.h"
#include "utilities/avatar_packet.h"
#include "utilities/avatar_message.h"
#include "utilities/upgrade_metadata.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

#define AVATAR_UPGRADE_IMAGE_PRODUCT_LEN    20

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
  
    /**
     * Callback when TCP data is received from the Avatar
     * @param client_idx - The TCP client identifier
     * @param data - A pointer to the TCP data
     * @param offset - For multi-part mesages this is the offset of this data in the whole messaage
     * @param data_len - The length of the data pointed at by the data pointer.
     */
    typedef void (*AVATAR_TCP_CALLBACK)(uint8_t client_idx, char* data, uint16_t offset, size_t data_len);

    /**
     * Message called when an upgrade image is about to start processing.  Used to check the version
     * being sent from the avatar against the current version
     */
    typedef bool (*AVATAR_UPLOAD_CHECK_VERSION_CALLBACK)(uint8_t major_version, uint8_t minor_version);

    /**
     * Message called when an upgrade image is about to start processing
     */
    typedef UPGRADE_METADATA_LOAD_ID (*AVATAR_UPLOAD_PREPARE_CALLBACK)(uint8_t major_version, uint8_t minor_version, size_t size, uint8_t* checksum, bool override);

    /**
     * Message called when an upgrade image is uploading data
     */
    typedef UPGRADE_METADATA_UPLOAD_STATUS (*AVATAR_UPLOAD_DATA_CALLBACK)(UPGRADE_METADATA_LOAD_ID load_id, 
        uint8_t* data, size_t 
        data_length,
        UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback);

    /**
     * Message called when an upgrade image has been uploaded
     */
    typedef UPGRADE_METADATA_UPLOAD_STATUS (*AVATAR_UPLOAD_END_CALLBACK)(UPGRADE_METADATA_LOAD_ID load_id,
        UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback, UPGRADE_METADATA_UPGRADE_STATUS status);
    
    /**
     * Defines actions to be used for a given product
     * @return 
     */
    typedef struct
    {
        char product[AVATAR_UPGRADE_IMAGE_PRODUCT_LEN + 1];
        
        AVATAR_UPLOAD_CHECK_VERSION_CALLBACK check_version_callback;
        
        AVATAR_UPLOAD_PREPARE_CALLBACK prepare_callback;
        
        AVATAR_UPLOAD_DATA_CALLBACK data_callback;
        
        AVATAR_UPLOAD_END_CALLBACK end_callback;
    } AVATAR_UPGRADE_IMAGE_ACTIONS;
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Gets the current avatar audio input
     * @return 
     */
    AVATAR_INPUT avatar_message_get_audio_input();

    /**
     * Gets the current avatar audio source
     * @return 
     */
    AVATAR_AUDIO_SOURCE avatar_message_get_audio_source();

    /**
     * Gets the current audio source as a string, adjusted for the Chromecast app name
     * @return 
     */
    void avatar_message_get_audio_source_string(char* buffer, uint8_t len);

    /**
     * Gets the current  Chromecast app name
     * @return 
     */
    void avatar_message_get_chromecast_app_name_string(char* buffer, uint8_t len);
    
    /**
     * Gets the current dirac client status
     */
    AVATAR_DIRAC_CLIENT_STATUS avatar_message_get_dirac_client_status();
    
    /**
     * Returns the current sample rate setting for the Avatar
     * @return 
     */
    AVATAR_SAMPLE_RATE avatar_message_get_display_sample_rate();

    /**
     * Gets the Avatar firmware version
     * @return 
     */
    void avatar_message_get_firmware_version(char* buffer, uint8_t len);

    /**
     * Gets the current  Friendly Name
     * @return 
     */
    void avatar_message_get_friendly_name_string(char* buffer, uint8_t len);

    /**
     * Gets the index in the AVATAR_MIME_TYPE_STRINGS, AVATAR_MIME_TYPE_VALUES, 
     * AVATAR_MIME_TYPE_DISPLAY_STRINGS or the current Mime Type
     * @return 
     */
    uint8_t avatar_message_get_mime_type_index();

    /**
     * Gets the current Mime type
     * @return 
     */
    void avatar_message_get_mime_type_string(char* buffer, uint8_t len);
    
    /**
     * Gets the current region
     * @return 
     */
    AVATAR_REGION avatar_message_get_region();

    /**
     * Gets the current MQA Decoder level
     * @return 
     */
    void avatar_message_get_mqa_decoder_level_string(char* buffer, uint8_t len);
    
    /**
     * Returns the current MQA state for the Avatar
     * @return 
     */
    AVATAR_MQA_STATE avatar_message_get_mqa_state();
    
    /**
     * Returns the current sample rate setting for the Avatar
     * @return 
     */
    AVATAR_SAMPLE_RATE avatar_message_get_sample_rate();

    /**
     * Returns true if the Avatar is currently updating
     * @return 
     */
    bool avatar_message_in_upgrade();
    
    /**
     * Gets the Serial Number
     * @return 
     */
    void avatar_message_get_serial_number_string(char* buffer, uint8_t len);
    
    /**
     * Initialises the avatar message for upgrade actions
     * @param upgrade_actions
     * @param num_upgrade_actions
     */
    void avatar_message_initialise(AVATAR_UPGRADE_IMAGE_ACTIONS* upgrade_actions, uint8_t num_upgrade_actions);
    
    /**
     * Gets the Avatar firmware version
     * @return 
     */
    void avatar_message_get_upgrade_option_string(char* buffer, uint8_t len);

    /**
     * Gets the current Avatar upgrade status
     * @param packet
     */
    AVATAR_UPGRADE_STATUS avatar_message_get_upgrade_status();
    
    /**
     * Application specific Avatar processing
     * @param packet
     */
    void avatar_message_process(AVATAR_MESSAGE* message);
    
    /**
     * Set the callback handler
     * @param callback
     */
    void avatar_message_set_callback(AVATAR_CALLBACK callback);
    
    /**
     * Set the callback handler for TCP messages ONLY
     * @param callback
     */
    void avatar_message_set_tcp_callback(AVATAR_TCP_CALLBACK callback);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
