/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    avatar_packet.h

  @Summary
    Encapsulates the handling of packets of data sent or received from the Avatar

  @Description
    Provides methods for interrogating Avatar data packets, and allocating packets 
    from a pool of packets of varying sizes.
 */
/* ************************************************************************** */

#ifndef _AVATAR_PACKET_H    /* Guard against multiple inclusion */
#define _AVATAR_PACKET_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include "utilities/avatar_data.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

/**
 * The size of the queue to hold avatar packets ready for processing
 * will be <= AVATAR_PACKET_CACHE_SIZE (see avatar_packet.c)
 */
#define AVATAR_PACKET_QUEUE_SIZE 60 //**********TODO PUT BACK 40 //TODO This only need to be 25 for the SA30
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

/**
 * Avatar Control Flags
 */

typedef enum
{
    AVATAR_CONTROL_FLAGS_TYPE_RST   = 0x01,
    AVATAR_CONTROL_FLAGS_TYPE_IRQ   = 0x02,
    AVATAR_CONTROL_FLAGS_TYPE_MSG   = 0x03,
    AVATAR_CONTROL_FLAGS_TYPE_FEND  = 0x20,
    AVATAR_CONTROL_FLAGS_TYPE_NACK  = 0x40,
    AVATAR_CONTROL_FLAGS_TYPE_ACK   = 0x80,
} AVATAR_CONTROL_FLAGS;
    
/**
 * Packet header size 8 bytes (not including optional Payload Data + Payload Data tail)
 * Maximum packet size 65535 bytes
 * Maximum packet payload 65525 bytes (65535 - header size - data tail)
 * Maximum Frame ID 255, increasing until 255. Then reset to 0.
 * Maximum Packet ID 255, increasing until 255. Then reset to 0.
 * Max frame length for sequence of packets, 16774400 bytes. (65525 * 256)
 * Max number of retries on failure to transmit, 2.
 */
#define AVATAR_PACKET_MAX_DATA      65525 + 2 // (+2 for Data Tail)

    /**
     */
    typedef enum
    {
        AVATAR_PACKET_STATE_FREE = 0,
        AVATAR_PACKET_STATE_ALLOCATED,
        AVATAR_PACKET_STATE_LOADING,
        AVATAR_PACKET_STATE_READY_TO_PROCESS,
        AVATAR_PACKET_STATE_PROCESSING,
        AVATAR_PACKET_STATE_READY_TO_SEND,
        AVATAR_PACKET_STATE_SEND_INITIATED,
        AVATAR_PACKET_STATE_SENDING,
    } AVATAR_PACKET_STATE;
    
    /**
     * Avatar message types - these match the stream unlimited message types included
     * in the message header control bits
     */
    typedef enum
    {
        AVATAR_PACKET_TYPE_RESET = AVATAR_CONTROL_FLAGS_TYPE_RST,
        AVATAR_PACKET_TYPE_IRQ = AVATAR_CONTROL_FLAGS_TYPE_IRQ,
        AVATAR_PACKET_TYPE_MSG = AVATAR_CONTROL_FLAGS_TYPE_MSG,
    } AVATAR_PACKET_TYPE;
    
    /**
     * Holds the header for a packet of data that is either sent or received from the avatar
     * Note: This structure is defined in the same order as the data is received/sent
     */
    typedef struct
    {
        uint16_t start_of_packet;

        uint16_t packet_length;

        uint8_t control_flags;

        uint8_t frame_id;

        uint8_t packet_id;

        //! Always set to 0
        uint8_t reserved;
    } AVATAR_PACKET_HEADER;

    /**
     * Holds a packet of data that is either sent or received from the avatar
     * Note: This structure is defined in the same order as the data is received
     */
    typedef struct
    {
        // Determines whether the packet is in use
        AVATAR_PACKET_STATE state;
        
        //! Used in debug to identify the packet
        uint8_t index;
        
        // The data containing the packet header
        // Includes the length of the data
        AVATAR_PACKET_HEADER header;

        // Normally this can be quite small
        // but for BIN16 data it may be quite large especially when receiving Upgrade data
        // where I think it will be full size.
        uint8_t* data;
    } AVATAR_PACKET;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Add a data item to a Tag list packet - this will increment the number of values in the packet, 
     * and the size of the packet header (assuming it was initialised to 8)
     * and add the item as the next value.
     * Note: Bin16 fields need adding in their entirety - meaning this does not
     * convert data into BIN16 format
     * @param packet
     */
    void avatar_packet_add_data(AVATAR_PACKET* packet, uint16_t tag, AVATAR_ENCODING encoding, uint8_t* data, uint16_t data_len);

    /**
     * Add a value to a UINT16 values list packet (e.g. for adding a subscribe item to a 
     * SUBSCRIBE packet)
     * This will increment the number of values in the packet, 
     * and the size of the packet header (assuming it was initialised to 8)
     * and add the item as the next value
     * @param packet
     */
    void avatar_packet_add_value(AVATAR_PACKET* packet, uint16_t value);

    /**
     * Get the message Id from the packet (first 2 bytes)
     * @param packet
     * @return 
     */
    AVATAR_MESSAGE_ID avatar_packet_get_message_id(AVATAR_PACKET* packet);
    
    /**
     * Get the number of values in a packet (byte 3)
     * @param packet
     * @return 
     */
    uint8_t avatar_packet_get_num_values(AVATAR_PACKET* packet);

    /**
     * Get the packet Id from the packet 
     * @param packet
     * @return 
     */
    uint8_t avatar_packet_get_packet_id(AVATAR_PACKET* packet);
    
    /**
     * Get the values at the specified index in a packet. Note this assumes that the message
     * is a list of uint16 values
     * @param packet
     * @return 
     */
    uint16_t avatar_packet_get_value(AVATAR_PACKET* packet, uint8_t index);
    
    /**
     * Returns the total length of the packets data (including tail)
     */
    uint16_t avatar_packet_header_get_data_length(AVATAR_PACKET_HEADER* header);

    /**
     * Returns the total length of the packet
     */
    uint16_t avatar_packet_header_get_length(AVATAR_PACKET_HEADER* header);
    
    /**
     * Gets the message type of the Avatar packet header
     * @param packet
     * @return 
     */
    AVATAR_PACKET_TYPE avatar_packet_header_get_type(AVATAR_PACKET_HEADER* header);

    /**
     * Initialises an empty packet header with start/tail stuff populated
     * @return 
     */
    void avatar_packet_header_initialise(AVATAR_PACKET_HEADER* header);
    
    /**
     * Returns true if a packet is ACK'd
     */
    bool avatar_packet_header_is_ack(AVATAR_PACKET_HEADER* header);

    /**
     * Returns true if a packet is NACK'd
     */
    bool avatar_packet_header_is_nack(AVATAR_PACKET_HEADER* header);
    
    /**
     * Initialise a packet to contain data - sets the Message Id and the Count to 0
     * @param packet
     * @param message_id
     */
    void avatar_packet_initialise_message(AVATAR_PACKET* packet, AVATAR_MESSAGE_ID message_id);

    /**
     * logs out a message containing information about the supplied packet
     * @param packet
     */
    void avatar_packet_print(AVATAR_PACKET* packet, char* prefix);
    
    /**
     * set the message Id from the packet (first 2 bytes)
     * @param packet
     * @return 
     */
    void avatar_packet_set_message_id(AVATAR_PACKET* packet, AVATAR_MESSAGE_ID message_id);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
