/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _AVATAR_PROCESS_RULE_H    /* Guard against multiple inclusion */
#define _AVATAR_PROCESS_RULE_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\avatar_message.h"
#include "utilities\avatar_packet.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

//! Needs to be as big as the number of processes defined
#define AVATAR_PROCESS_RULE_MAX         3

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Callback function template
     */
    typedef struct AVATAR_PROCESS_RULE AVATAR_PROCESS_RULE;
    typedef void (*AVATAR_PROCESS_RULE_CALLBACK)(AVATAR_PROCESS_RULE* rule, AVATAR_MESSAGE* message);

    typedef enum
    {
        AVATAR_PROCESS_RULE_ACTION_PRE,
        AVATAR_PROCESS_RULE_ACTION_POST,
        AVATAR_PROCESS_RULE_ACTION_ACK,
    } AVATAR_PROCESS_RULE_ACTION;
    
    struct AVATAR_PROCESS_RULE
    {
        //! determines whether the process should be considered - this should be checked in 
        //  the callback.  Doing it in the callback allows it to be activated/deactivated as neccessary
        bool active;
        
        //! The Id of the message type to be looked for
        AVATAR_MESSAGE_ID message_id;
        
        //! The Id of the state to be looked for
        AVATAR_STATE_ID state_id;

        //! A callback to be invoked when the message has been processed (actually)
        //  invoked before the message is actioned and also influence by wait_for_ack (below))
        AVATAR_PROCESS_RULE_CALLBACK callback;

        //! If the rule is for a message that requires a response the this will mean the
        //  callback is invoked when the response ACK is received from the Avatar
        //bool wait_for_ack;
        
        //! Determines when the rule check is initiated
        AVATAR_PROCESS_RULE_ACTION action;
    };

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Adds a process rule - checks for the existence of a state within a packet
     * @param message_id -  the type of message the rule applies to
     * @param state_id - The state to check for
     * @param callback - the callback to be invoked if the state exists
     * @param wait_for_ack - if true the callback processing is done when the ack is received from the avatar
     */
    void avatar_process_rule_add(AVATAR_MESSAGE_ID message_id,
                                 AVATAR_STATE_ID state_id,
                                 AVATAR_PROCESS_RULE_CALLBACK callback,
                                 AVATAR_PROCESS_RULE_ACTION action);
    
    /**
     * Deactivates the rule and stops it being considered for new messages
     * @param rule
     * @param packet
     */
    void avatar_process_rule_deactivate(AVATAR_PROCESS_RULE* rule, AVATAR_PACKET* packet);

    /**
     * Checks to see if any rules match the supplied packet, and invokes if found
     * @param match
     * @return - true if processed, false if not
     */
    bool avatar_process_rule_check(AVATAR_MESSAGE* message, AVATAR_PROCESS_RULE_ACTION action);

    /**
     * Compares message input with the Rules to look for a match
     * @param match
     * @return 
     */
    AVATAR_PROCESS_RULE* avatar_process_rule_find(AVATAR_MESSAGE* message);

    /**
     * Processes the active rule if there is one
     * @return 
     */
    bool avatar_process_rule_process_active();
         
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
