/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Provides metadata about the DIRAC implementation.  Can be used by other modules 
    that handle the actual implementation.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _DIRAC_H    /* Guard against multiple inclusion */
#define _DIRAC_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include "app_config.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

/**
 * Defines the maximum length of a DIRAC EQ Curve name 
 */
#ifndef DIRAC_MAX_EQ_CURVE_NAME_LEN
#define DIRAC_MAX_EQ_CURVE_NAME_LEN     40    
#endif    

/**
 * Defines the maximum number of DIRAC curves tha the system can handle
 * This should include one for No Curve
 */
#ifndef DIRAC_MAX_EQ_CURVES
#define DIRAC_MAX_EQ_CURVES             4    
#endif    

/**
 * Used to indicate that there as no currently selected EQ curve.  This is the 
 * index of the DIRAC_EQ_CURVE that represents no curve (defaults to 0)
 */    
#ifndef DIRAC_NO_CURVE
#define DIRAC_NO_CURVE                  0
#endif
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Structure representing a single DIRAC EQ Curve
     */
    typedef struct
    {
        //! An identifer for the curve
        uint8_t index;
        
        //! The name given to the DIRAC Curve
        char name[DIRAC_MAX_EQ_CURVE_NAME_LEN + 1];
        
    } DIRAC_EQ_CURVE;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Add/update a DIRAC Curve at the specified index
     * @param index
     * @param name
     */
    //void dirac_add_curve(uint8_t index, char* name);
    
    /**
     * Configures the module from the settings
     */
    void dirac_configure();
    
    /**
     * Returns the dirac curve at the given index
     * @param input
     * @return 
     */
    DIRAC_EQ_CURVE* dirac_get_curve(uint8_t index);
    
    /**
     * Returns the dirac curve for the given AUDIO_INPUT
     * @param input
     * @return 
     */
    uint8_t dirac_get_curve_for_input(uint8_t input);
    
    /**
     * Gets the name associated with the supplied index, returns NULL if invalid
     * @param index
     * @return 
     */
    char* dirac_get_curve_name(uint8_t index);
    
    /**
     * Gets the info on the next curve after the selected curve 
     * @current_curve - The index of the current curve
     * @return 
     */
    DIRAC_EQ_CURVE* dirac_get_next_curve(uint8_t current_curve);
    
    /**
     * Gets maximum number of available DIRAC curves
     * @return 
     */
    uint8_t dirac_get_num_curves();
    
    /**
     * Gets the info on the previous curve after the selected curve 
     * @current_curve - The index of the current curve
     * @return 
     */
    DIRAC_EQ_CURVE* dirac_get_previous_curve(uint8_t current_curve);

    /**
     * Initialises the DIRAC module - should be called before any curves are added
     */
    void dirac_initialise();
    
    /**
     * Perform dirac processing (getting Curve names)
     */
    void dirac_process();

    /**
     * Sets the Dirac curve as uncalibrated, making it unavailable
     * @param index
     */
    void dirac_remove_curve_calibration(uint8_t index, bool clear_usage);
    
    /**
     * Used to define whether the curve has been through a calibration process
     * @param index
     */
    void dirac_set_curve_calibrated(uint8_t index, char* name);
    
    /**
     * Sets the dirac curve for the given AUDIO_INPUT
     * @param input - The input index
     * @param index - The curve index
     * @return - The index supplied if successful, DIRAC_NO_CURVE if not.
     */
    uint8_t dirac_set_curve_for_input(uint8_t input, uint8_t index);
    
    /**
     * Sets the name associated with the supplied index.  Note: This is used to 
     * update the curve name retrieved from the Avatar.  It will not update the 
     * curve name in Dirac
     * @param index
     * @return 
     */
    void dirac_set_curve_name(uint8_t index, char* name);

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _DIRAC_H */

/* *****************************************************************************
 End of File
 */
