/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
 * Generic header for wrapping flash memory. e.g. a contiguous block of memory that
 * is accessed via a number of pages.
 */
/* ************************************************************************** */

#ifndef _FLASH_MEMORY_H    /* Guard against multiple inclusion */
#define _FLASH_MEMORY_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>
#include "app_config.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif

    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

//! The maximum number of clients    
#ifndef FLASH_MEMORY_MAX_CLIENTS
#define FLASH_MEMORY_MAX_CLIENTS        2
#endif    
    
    /**
     * These should be defined in app_config.h
     */
#ifndef FLASH_MEMORY_PAGE_COUNT
#define FLASH_MEMORY_PAGE_COUNT 1
#endif    
    
#ifndef FLASH_MEMORY_PAGE_SIZE
#define FLASH_MEMORY_PAGE_SIZE 256
#endif    
    
    /**
     * The size (in pages) of the cache to hold page data
     */
#ifndef FLASH_MEMORY_CACHE_SIZE
#define FLASH_MEMORY_CACHE_SIZE 0
#endif    
    
//! Used to identify an invalid client id
#define FLASH_MEMORY_INVALID_CLIENT_ID   -1
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
  
    //! Represents a unique Id identifying each client.
    //  The reason this is fixed rather than allocated dynamically is that in
    //  general we always want the clients to be allocated in the same order
    typedef int8_t FLASH_MEMORY_CLIENT_ID;

    /**
     * Structure containing details of a client request
     */
    typedef struct
    {
        //! The Id of the client
        FLASH_MEMORY_CLIENT_ID client_id;
        
        //! A unique index of the file to read/write
        uint8_t file_index;
        
        //! A pointer to a buffer that is sufficiently large enough to hold the 
        //  data that is being read, or that contains the data being written
        void* data;
        
        //! The offset of the start of the read in the client data
        uint32_t offset;

        //! The length of data to be read
        uint32_t length;
        
        //! READ-ONLY Contains the actual number of bytes read or written by the request
        size_t bytes_read_written;
        
        //! READ-ONLY Indicates whether this is a read or write request.  This can be 
        //  interrogated in the callback
        bool write_request;
        
        // Optional pointer to some associated data object 
        void* additional_info;

    } FLASH_MEMORY_CLIENT_REQUEST;
    
    /**
     * flash_memory_read_page callback function template
     */
    typedef void (*FLASH_MEMORY_CALLBACK)(FLASH_MEMORY_CLIENT_REQUEST* request);


    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * A Dummy callback used for testing flash_memory functionality
     * @param page_data
     */
    void flash_memory_dummy_callback(FLASH_MEMORY_CLIENT_REQUEST* request);
    
    /**
     * Each module using flash memory adds itself as a client
     * @param client_id - The client Id to add (defined in app_config.h)
     * @param size      - The amount of space to allocate (in bytes)
     */
    void flash_memory_add_client(FLASH_MEMORY_CLIENT_ID client_id, char* name, FLASH_MEMORY_CALLBACK callback);
    
    /** 
     * Where multiple reads/writes are being performed flash_memory_close releases
     * any resources previously allocated using flash_memory_open
     */
    void flash_memory_close(FLASH_MEMORY_CLIENT_ID client_id, uint8_t file_index);
    
    /** 
     * Completely clears flash memory
     */
    void flash_memory_erase();
    
    /**
     * Initialise the flash memory system
     */
    void flash_memory_initialise();
    
    /** 
     * Where multiple reads/writes are being performed flash_memory_open requests
     * that the flash resources are retained between calls (rather than open/close 
     * on each call to read/write.
     */
    void flash_memory_open(FLASH_MEMORY_CLIENT_ID client_id, uint8_t file_index);
    
    /**
     * flash memory processing
     */
    void flash_memory_process();
    
    /**
     * Reads contents of flash memory into the supplied buffer
     * @param request - A pointer to a request.  This information is copied and 
     * so does not need retaining
     */
    void flash_memory_read(FLASH_MEMORY_CLIENT_REQUEST* client_request);
    
    /**
     * Can be used to see if a request is read or write
     * @param client_request
     * @return 
     */
    bool flash_memory_request_is_read(FLASH_MEMORY_CLIENT_REQUEST* client_request);
    
    /**
     * Can be used to see if a request is read or write
     * @param client_request
     * @return 
     */
    bool flash_memory_request_is_write(FLASH_MEMORY_CLIENT_REQUEST* client_request);
    
    /** 
     * Sets the size of the specified file.  If the size supplied is too large 
     * it leaves the size alone
     */
    uint32_t flash_memory_set_size(FLASH_MEMORY_CLIENT_ID client_id, uint8_t file_index, uint32_t size);
   
    /** 
     * Gets the size of the specified file
     */
    uint32_t flash_memory_size(FLASH_MEMORY_CLIENT_ID client_id, uint8_t file_index);
    
    /**
     * Reads contents of flash memory into the supplied buffer
     * @param request - A pointer to a request.  This information is copied and 
     * so does not need retaining
     */
    void flash_memory_write(FLASH_MEMORY_CLIENT_REQUEST* client_request);

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _FLASH_MEMORY_H */

/* *****************************************************************************
 End of File
 */
