/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _HDMI_ARC_DATA_H    /* Guard against multiple inclusion */
#define _HDMI_ARC_DATA_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdint.h>
#include <stddef.h>

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Constants                                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define HDMI_ARC_MESSAGE_START_OF_PACKET    0x23
#define HDMI_ARC_MESSAGE_END_OF_PACKET      0x0D

#define HDMI_ARC_MESSAGE_MAX_DATA_LENGTH    256
    
//------------------------------------------------------------------------------
// Internal processing enums
//------------------------------------------------------------------------------

/**
 * Describes the states when parsing incoming messages
 */
typedef enum
{
    //! Indicates that the message has been initialised.  This may be the 
    //  final state where messages are not processed through process buffer
    HDMI_ARC_MESSAGE_STATE_INITIALISED,
    HDMI_ARC_MESSAGE_STATE_FOR_LEN_LSB,
    HDMI_ARC_MESSAGE_STATE_WAIT_LEN_MSB,
    HDMI_ARC_MESSAGE_STATE_WAIT_COMMAND_TYPE,
    HDMI_ARC_MESSAGE_STATE_POPULATING,

    // The message has been completely received and is ready for processing
    HDMI_ARC_MESSAGE_STATE_POPULATED,

} HDMI_ARC_MESSAGE_STATE;

// *****************************************************************************
// *****************************************************************************
// Section: Data Types
// *****************************************************************************
// *****************************************************************************

/**
 */
typedef enum
{
    HDMI_ARC_SYSTEM_READY_STATE_HANDSHAKE_START = 0x00,
    HDMI_ARC_SYSTEM_READY_STATE_HANDSHAKE_END = 0x01,
} HDMI_ARC_SYSTEM_READY_STATE;

/**
 */
typedef enum
{
    HDMI_ARC_MUTE_STATE_MUTED   = 0x00,
    HDMI_ARC_MUTE_STATE_UNMUTED = 0x01,
} HDMI_ARC_MUTE_STATE;

/**
 */
typedef enum
{
    HDMI_ARC_POWER_STATE_OFF    = 0x00,
    HDMI_ARC_POWER_STATE_IDLE   = 0x01,
    HDMI_ARC_POWER_STATE_ON     = 0x02,
} HDMI_ARC_POWER_STATE;

/**
 */
typedef enum
{
    HDMI_ARC_STREAM_STATE_INACTIVE  = 0x00,
    HDMI_ARC_STREAM_STATE_ACTIVE    = 0x01,
} HDMI_ARC_STREAM_STATE;

/**
 */
#define HDMI_ARC_FORMATS \
    X(ARC, ARC)\
    X(EARC, eARC)\

// Now use the input data to define the enum
#define X(a, b) HDMI_ARC_FORMAT_##a,
	typedef enum {HDMI_ARC_FORMATS} HDMI_ARC_FORMAT;
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a, b) #b,
	static char *HDMI_ARC_FORMAT_STRINGS[] = {HDMI_ARC_FORMATS};
#undef X

typedef enum
{
    HDMI_ARC_UPGRADE_START_STATE_UPGRADE_AVAILABLE = 0x00,
    HDMI_ARC_UPGRADE_START_STATE_READY_FOR_DATA = 0x01,
} HDMI_ARC_UPGRADE_START_STATE;

typedef enum
{
    HDMI_ARC_UPGRADE_DATA_STATE_TRANSFER_FAILED = 0x00,
    HDMI_ARC_UPGRADE_DATA_STATE_TRANSFER_SUCCESS = 0x01,
} HDMI_ARC_UPGRADE_DATA_STATE;

typedef enum
{
    HDMI_ARC_UPGRADE_END_STATE_TRANSFER_COMPLETE = 0x00,
    HDMI_ARC_UPGRADE_END_STATE_UPGRADE_COMPLETE = 0x01,
    HDMI_ARC_UPGRADE_END_STATE_UPGRADE_FAILED = 0x02,
} HDMI_ARC_UPGRADE_END_STATE;

/**
 * Avatar Encoding
 */
#define HDMI_ARC_COMMAND_TYPES \
	X(POWER, =0x00, Power)\
	X(VOLUME, =0x01, Volume)\
	X(MUTE, =0x02, Mute)\
	X(ARC_FORMAT, =0x03, ARC Format)\
	X(ARC_STREAM, =0x04, ARC Stream)\
	X(CEC_NAME, =0x05, CEC Name)\
	X(UPGRADE_START, =0x06, Upgrade Start)\
	X(UPGRADE_DATA, =0x07, Upgrade Data)\
	X(UPGRADE_END, =0x08, Upgrade End)\
	X(SYSTEM_READY, =0x09, System Ready)\
	X(ARC_FIRMWARE_VERSION, =0x0A, Firmware Version)\
	X(ARC_TRANSMITTER_FIRMWARE_VERSION, =0x0B, Transmitter Firmware Version)\
	X(UNKNOWN, =0xFF, Unknown)\

// Now use the input data to define the enum
#define X(a, b, c) HDMI_ARC_COMMAND_##a b,
	typedef enum {HDMI_ARC_COMMAND_TYPES} HDMI_ARC_COMMAND_TYPE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *HDMI_ARC_COMMAND_STRINGS[] = { HDMI_ARC_COMMAND_TYPES };
#undef X

//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------

/**
 */
typedef struct
{
    //! Identifies the type of the message
    HDMI_ARC_COMMAND_TYPE command_type;
    
    //! The buffer to hold the data associated with the message
    uint8_t data[HDMI_ARC_MESSAGE_MAX_DATA_LENGTH];
    
    // The length of the entire message including the Start and End of Message 
    // bytes, Message Length, and Command Type bytes. (N+2 bytes). 
    uint16_t length;

    //! Indicates the current state of processing for the message (incoming))
    HDMI_ARC_MESSAGE_STATE processing_state;

    // Maintains an index of the current write position for the data
    // Internal use only
    unsigned char processing_data_index;

} HDMI_ARC_MESSAGE;
    
/**-----------------------------------------------------------------------------
 * Callback function template
 */
typedef void (*HDMI_ARC_MESSAGE_CALLBACK)(HDMI_ARC_MESSAGE* message);

// *****************************************************************************
// *****************************************************************************
// Section: Interface Functions
// *****************************************************************************
// *****************************************************************************

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
