/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    input_control.h

  @Summary
 Handles input management

  @Description
    
 * For initial use wit the SA 20 input_contro, assumes a single analog input control (3-pins) 
 * and an I2C controlled DAC 
 *
 */
/* ************************************************************************** */

#ifndef _INPUT_CONTROL_H    /* Guard against multiple inclusion */
#define _INPUT_CONTROL_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>
#include "app_config.h"
#include "utilities/input_control_avatar.h"
#include "utilities/input_control_defs.h"
#include "utilities/input_control_dac.h"
#include "utilities/input_control_digital.h"
#include "utilities/input_control_mux.h"
#include "utilities/input_control_usb.h"

#ifdef INPUT_CONTROL_ANALOGUE_ENABLED
#include "utilities/input_control_analogue.h"
#endif

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
    typedef enum
    {
        INPUT_CONTROL_SWITCH_STATE_PRE,
        INPUT_CONTROL_SWITCH_STATE_POST,
    } INPUT_CONTROL_SWITCH_STATE;
    
    /**
     * Callback function template
     */
    typedef void (*INPUT_CONTROL_SWITCH_CALLBACK)(INPUT_CONTROL_SWITCH_STATE state);
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Add a definition of an input control to the module
     * Assumes that the indexes used range from 0-max, and can be used to index in the
     * internal array 
     * @param definition - A Pointer to an INPUT_CONTROL_DEFINITION structure which will be copied.
     */
    void input_control_add(INPUT_CONTROL_DEFINITION* definition);
    
    /**
     * Used to determine whether an input signal is present..
     */
    bool input_control_audio_input_present();
    
    /**
     * Gets the display sample rate of the current digital input 19200 if analog
     * @return 
     */
    SAMPLE_RATE input_control_get_display_sample_rate();
    
    /**
     * Returns the number of registered inputs
     * @param forwards - Indicates the direction to move through the inputs
     * @return The Id of the next input
     */
    uint8_t input_control_get_next_input(uint8_t input, bool forwards);
    
    /**
     * Returns the number of registered inputs
     * @return 
     */
    uint8_t input_control_get_num_inputs();

    /**
     * Gets the previous input.  If that is not an enable input then it looks for the
     * first enabled input
     * @return 
     */
    uint8_t input_control_get_previous_input();
    
    /**
     * Gets the index of the current input designated for Processor mode
     * @return 
     */
    uint8_t input_control_get_processor_mode_input();
    
    /**
     * Gets current volume for Processor mode
     * @return 
     */
    uint8_t input_control_get_processor_mode_volume();
    
    /**
     * Gets the sample rate of the current digital input 0 if analog
     * @return 
     */
    SAMPLE_RATE input_control_get_sample_rate();

    /**
     * Handles interrupt e.g.on the DAC GPIO1 which is configured via Register8 to change
     * When the dpll_lock status changes
     * @param channel - The channel the interrupt is being raised on
     */
    void input_control_handle_interrupt(uint8_t channel);

    /**
     * Gets the index of the currently selected input
     * @return 
     */
    uint8_t input_control_get_selected_input();
    
    /**
     * Gets the current type of input control (Analog/Digital)
     * @return 
     */
    INPUT_CONTROL_TYPE input_control_get_type();
    
    /**
     * Determines whether the input control believes that the volume should be muted
     * This is not the same as volume_is_muted as its possible to have the volume
     * muted but input control say its not (e.g. when there is an error on the digital signal)
     * @return 
     */
    bool input_control_in_mute_state();
    
    /**
     * Can be used to determine whether the selected input is the processor mode input
     * @return 
     */
    bool input_control_in_processor_mode();
    
    /**
     * Can be used to determine whether the input control is currently switching inputs
     * @return 
     */
    bool input_control_in_switch();
    
    /**
     * Initialise the input control module
     */
    void input_control_initialise();
    
    /**
     * This actually switches the SPDIF MUX to a channel with no inputs.
     */
    void input_control_mute_inputs();
    
    /**
     * Can be used to determine whether the input was muted prior to a current switch action
     * @return 
     */
    bool input_control_mute_state_pre_switch();
    
    /**
     * Sets a callback that will be used to inform when the input is switching
     */
    void input_control_set_switch_callback(INPUT_CONTROL_SWITCH_CALLBACK callback);
    
    /**
     * Used to display the new volume during a switch operation
     * @return 
     */
    uint8_t input_control_volume_in_switch();
    
    /**
     * Perform input control processing
     */
    void input_control_process();

    /**
    * Reset the input control
    * @param index
    */
    void input_control_reset_input();
    
    /**
     * Use to reset the state to running when waking up from standby - prevents
     * the ignoring of the reset_input..caused by the attempted switch from ARC to 
     * previous input
     */
    //void input_control_reset_state();
    
    /**
     * Select the input control identified by the index
     * @param index
     */
    void input_control_select_input(uint8_t index);
   
    /**
     * Sets the avatar volume
     */
    void input_control_set_avatar_balance(uint8_t volume);
   
    /**
     * Sets the avatar volume
     */
    void input_control_set_avatar_volume(uint8_t volume);
   
    /**
     * Sets the HDMI ARC volume
     */
    void input_control_set_hdmi_arc_volume(uint8_t volume);
    
    /**
     * Sets the index of the current input designated for Processor mode
     * @param index - The index of the input to set for processor mode 
     */
    void input_control_set_processor_mode_input(uint8_t index);
    
    /**
     * Sets current volume for Processor mode
     * @volume - The new processsor mode volume 
     */
    void input_control_set_processor_mode_volume(uint8_t volume);
    
    /**
     * Starts input control processing - performs initialisation based on settings
     */
    void input_control_start();
    
    /**
     * Starts input control configuration process - performs initialisation based on settings
     */
    void input_control_start_configuration();
    
    /**
     * This actually switches the SPDIF MUX to a channel with no inputs.
     */
    void input_control_unmute_inputs(INPUT_CONTROL_TYPE type);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
