/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _INPUT_CONTROL_DAC_H    /* Guard against multiple inclusion */
#define _INPUT_CONTROL_DAC_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>
#include "utilities/input_control_defs.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    //! The maximum number of filter types supported (does not mean we are using all of these)
#define MAXIMUM_FILTER_TYPES        7

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Define the FILTER_TYPE 
     */
    typedef uint8_t FILTER_TYPE;
    
    /**
     * Callback for configuration function
     * @param address
     * @param value
     */
    typedef void (*INPUT_CONTROL_DAC_CONFIGURE_CALLBACK) ();
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
 
    /**
     * Used to define DAC configuration values.
     * Note: This works at a register level rather than the individual settings
     * @param register_ident - the register identifier
     * @param value - th value to be written
     */
    void input_control_add_dac_setting(uint8_t address, uint8_t value);
    
    /**
     * Starts the process of applying settings to the dac
     * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
     */
    void input_control_apply_dac_settings();

    /**
     * Initiates call to DAC to check on the current source type (Audio/Non Audio)
     */
    void input_control_check_dac_reset();

    /**
     * Initiates call to DAC to check on the current source type (Audio/Non Audio)
     * Then calculates the Sample Rate if applicable
     */
    void input_control_check_source_type();
    
    /**
     * Configures the digital module
     * @return 
     */
    bool input_control_dac_configure();
    
    /**
     * True when the DAC has been configured
     * @return 
     */
    bool input_control_dac_configured();

    /**
     * Used to change the state of the DAC reset switch
     * Switched ON during power up, OFF the rest of the time
     * @param on
     */
    void input_control_dac_reset(bool on);
    
    /**
     * 
     * @param channel
     */
    void input_control_dac_set_callbacks(INPUT_CONTROL_DAC_CONFIGURE_CALLBACK configure_callback);
    
    /**
     * Defines the address of the register used to setup the custom filter.  It assumes that
     * 4 registers will be used.
     * @param address
     */
    void input_control_define_custom_filter_address(REGISTER_BITS_DEFINITION addresses[]);

    /**
     * Defines the address of the register used to manage the sample_rate
     * @param address
     */
    void input_control_define_custom_filter_enable_address(REGISTER_BITS_DEFINITION address);
    
    /**
     * Defines the address of the register used to manage the sample_rate
     * @param address
     */
    void input_control_define_dac_reset_check_address(REGISTER_BITS_DEFINITION address);
    
    /**
     * Defines the address of the register used to manage the sample_rate
     * @param address
     */
    void input_control_define_dac_sample_rate_address(REGISTER_BITS_DEFINITION addresses[], size_t num);
    
    /**
     * Defines the address of the register used to manage the volume
     * @param address - expects a 2 element array
     */
    void input_control_define_dac_volume_address(REGISTER_BITS_DEFINITION addresses[]);
        
    /**
     * Defines the address of the register used to manage the current filter
     * @param address
     */
    void input_control_define_filter_address(REGISTER_BITS_DEFINITION address);
    
    /**
     * Defines the address of the register used to manage the current input
     * @param address
     */
    void input_control_define_source_type_address(REGISTER_BITS_DEFINITION address);
    
    /**
     * Initialise the input control module
     * @return - True if successfully initialised
     */
    bool input_control_dac_initialise();

    /**
     * Returns the current filter
     */
    FILTER_TYPE input_control_get_filter();

    /**
     * Returns a string representation of the current filter
     * @return 
     */
    void input_control_get_filter_string(char* buffer, size_t size);
    
    /**
     * Returns the number of registered filters
     * @return 
     */
    uint8_t input_control_get_num_filters();
    
    /**
     * Gets the sample rate of the current digital input 0 if analog
     * @return 
     */
    SAMPLE_RATE input_control_get_dac_sample_rate();

    /**
     * Handles interrupt on the DAC GPIO1 which is configured via Register8 to change
     * When the dpll_lock status changes
     * @param definition
     */
    void input_control_handle_dac_interrupt();
    
    /**
     * Mutes / Unmutes the DAC
     * @param on
     */
    void input_control_mute_dac(bool on);
    
    /**
     * Sets the Data volume registers
     * @param volume1
     * @param volume2
     */
    void input_control_set_dac_volume(uint8_t volume1, uint8_t volume2);
    
    /**
     * Sets the current filter
     */
    void input_control_set_custom_filter();
    
    /**
     * Sets the current filter
     */
    void input_control_set_filter(FILTER_TYPE filter);
    
    /**
     * Called after settings initialisation to setup the filter stored in the settings
     */
    void input_control_setup_filter();
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
