/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _INPUT_CONTROL_DIGITAL_H    /* Guard against multiple inclusion */
#define _INPUT_CONTROL_DIGITAL_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>
#include "utilities/input_control_defs.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Callback for configuration function
     * @param address
     * @param value
     */
    typedef void (*INPUT_CONTROL_DIGITAL_CONFIGURE_CALLBACK) ();
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
 
    /**
     * Used to define SPDIF MUX configuration values.
     * Note: This works at a register level rather than the individual settings
     * @param register_ident - the register identifier
     * @param value - th value to be written
     */
    void input_control_add_digital_setting(uint8_t address, uint8_t value);

    /**
     * Starts the process of applying settings to the SPDIF Mux
     * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
     */
    void input_control_apply_digital_setting(INPUT_CONTROL_I2C_REQUEST* setting);
    
    /**
     * Starts the process of applying settings to the dac
     * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
     */
    void input_control_apply_digital_settings();
 
    /**
     * Initiates call to SPDIF RX to get sample rate
     */
    void input_control_calculate_digital_sample_rate();
    
    /**
     * Defines the address of the register used to report on errors
     * @param address
     */
    void input_control_define_digital_error_address(REGISTER_BITS_DEFINITION address);
    
    /**
     * Defines the address of the register used to manage the current input
     * @param address
     */
    void input_control_define_digital_switch_address(REGISTER_BITS_DEFINITION address);
    
    /**
     * Defines the address of the register used to manage the sample_rate calculation
     * @param address
     */
    void input_control_define_digital_sample_rate_address(REGISTER_BITS_DEFINITION address);
    
    /**
     * Configures the digital module
     * @return 
     */
    bool input_control_digital_configure();
    
    /**
     * Initialise the input control module
     * @return - True if successfully initialised
     */
    bool input_control_digital_initialise();
    
    /**
     * 
     */
    void input_control_digital_process();
    
    /**
     * 
     * @param channel
     */
    void input_control_digital_set_callbacks(INPUT_CONTROL_DIGITAL_CONFIGURE_CALLBACK configure_callback);
    
    /**
     * Gets the sample rate of the current SPDIF MUX input 
     * @return 
     */
    SAMPLE_RATE input_control_get_digital_sample_rate();

    /**
     * Returns value indication whether the digital signal is invalid at present
     * @return 
     */
    bool input_control_digital_invalid_signal();
    
    /**
     * Handles interrupt on the SPDIF RX Interrupt pins
     * @param definition
     */
    void input_control_handle_digital_interrupt(uint8_t channel);

    /**
     * Used to change the state of the SPDIF MUX reset switch
     * Switched OFF during power up, ON the rest of the time
     * @param on
     */
    void input_control_digital_reset(bool on);

    /**
     * Used to set the Spdif Rx mux Master/Slave mode
     * @param reset
     * @param master
     */
    void input_control_digital_set_master_slave(bool master);

    /**
     * Used to set the SPdif Rx Power state
     * @param on
     */
    void input_control_digital_set_power(bool on);

    /**
     * Switch to a specific input on the digital input control
     * @param definition - the definition of the control to switch to
     */
    void input_control_switch_digital(INPUT_CONTROL_DEFINITION* definition);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
