/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    menu.h

  @Summary
    Brief description of the file.

  @Description
 * Maintains navigation around an array of MENU_ITEMS which should be declared outside this 
 * code (e.g. in application specific code
 */
/* ************************************************************************** */

#ifndef _MENU_H    /* Guard against multiple inclusion */
#define _MENU_H

#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include "app_config.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    /**
     * This can be used as the parameter to menu_get_edit_pos num_fields such that 
     * there is no adjustment for the number of field - used in draw functions that
     * are not changing the edit position.
     * 
     */
#define MENU_IGNORE_NUM_FIELDS  -1
    
    /**
     * The maximum length that data can be stored form comparison when fields are being edited
     */
#define MENU_EDIT_INFO_DATA_MAX 16
    
// Represents a calue for a NULL Menu ID or an unassigned option selection    
#define MENU_DISPLAY_DATA_EMPTY_ID  0xFFFF    
    
//! used in menu_set_current_item to indicate that the switch should be to the 
//  supplied item rather than another menu item supplied by the previous menus
//  exit procedure
#define MENU_USE_SUPPLIED_MENU_ID   0xFF    

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    //! Setup MENU_ID type
    typedef uint8_t MENU_ID;
    
    /**
     * A structure associated with a MENU_ITEM that can be used to store edit state 
     * information.
     * Note: The reset flag is set to true whenever the MENU_ITEM is returned by
     * one of the menu navigation operations.  If using during editing then the bespoke
     * code should set this to false if it need to detect that the edit s a re-entry
     */
    typedef struct
    {
        //! Set to true if the MENU_ITEM has been returned from one of the navigation commands
        bool reset;
        
        //! Used to maintain the current filed being edited in multi field values (IP Address, Network Friendly Name))
        uint8_t edit_pos;
        
        //! Used to indicate whether the value has been changed during an edit
        bool changed;
        
        //! Structure to hold bespoke info - might need making bigger or smaller
        // depending on how its ultimately used
        char data[MENU_EDIT_INFO_DATA_MAX];
    } MENU_EDIT_INFO;
    
    /**
     * Data used by the display message functionality
     */
    typedef struct 
    {
        //! Allows a type of menu to be set - for use with the display_message utility
        uint8_t screen_type;

        //! This is used to determine whether an Option is displaying its details for edit/view
        // or if it is being displayed in an Option list.
        // These require the edit function to be set up as edit_on_entry = False.
        // When this happens the user needs to select to edit the option, when leaving
        // we need to know whether the detail was being displayed - this is OK for
        // editable options but display only (e.g. Software Version) do not have edit functions
        // and so we cannot tell whether we are exiting edit or returning to the mani list..
        bool in_display;
        
        //! Identifies the number of options/items to be displayed on a screen
        //  if 0 then DISPLAY_MESSAGE_MAX_ROWS is used.  Usually this will be a
        //  multiple of the number of available rows
        uint8_t page_size;
        
        //! Option strings to be used when displaying as set of options
        char** option_strings;
                
        //! The number of options available. This is either the number of child menu items ,
        //  or the number of option values for an option 
        uint8_t option_count;
        
        //! For option lists this will store the menu_id of the first item on the current page
        //  For options it is the ordinal position of the first option on the current page
        //  A value of MENU_DISPLAY_DATA_EMPTY_ID (0xFFFF) is used to indicate a NULL value
        uint16_t page_start_id;
        
        //! If specified this can be used to indicate to the display MCU which ICON should be 
        //  associated with this menu item.
        uint8_t icon_id;
        
        //! A function for use with the display_message utility
        //  MENU_ID, buffer to populate, DISPLAY_TYPE (as owner, as child))
        //  Return - the selected DataID
        uint16_t (*build_screen_data)(MENU_ID, uint8_t*, uint8_t);
    } MENU_ITEM_DISPLAY_DATA;
    
    /**
     * MENU_ITEM structure - contains a definition of a single menu item
     */
    typedef struct MENU_ITEM
    {
        //! The identifier of the menu - MUST be the same as its position in the 
        //  array of menu items
        MENU_ID menu_id; 
        
        // Then parent of this menu item
        struct MENU_ITEM* parent;
        
        // The default child item that is displayed when this menu level is entered
        struct MENU_ITEM* default_item;
        
        // The first child in the list of children
        struct MENU_ITEM* first_child;
        
        // The last child in the list of children
        struct MENU_ITEM* last_child;
        
        struct MENU_ITEM* previous_sibling;
        
        struct MENU_ITEM* next_sibling;
        
        //! Data used by the display_message functionality
        MENU_ITEM_DISPLAY_DATA display_data;
        
        // Change - if true the move-left right on a child which is at the end will not 
        // go to the first child, false it will loop round
        bool loop_back_children;
        
        //! A function to be invoked when the menu has a bespoke drawing 
        void (*draw_function)(MENU_ID);
        
        //! A function to be invoked once an edit of the menu item has been completed
        // NOTE: This would be responsible for any state change
        bool (*exit_edit_function)(MENU_ID);
        
        //! A function that will return a display friendly version of current value
        // for the menu being displayed/edited
        void (*display_function)(char*, size_t);
        
        //! A function to be invoked when an input has changed a value while the 
        //  menu is being displayed
        // @return - true if the function has handled the message, false if not
        bool (*edit_function)(void* func);
        
        //! If true edit mode is entered when the menu item is selected rather 
        //  than waiting for enter to be pressed
        bool edit_on_entry;
        
        //! If true then when in edit mode navigation messages are sent to the edit function
        // before being actioned.  If the edit_function returns true then it is assumed that
        // it has handled the navigation and further navigation gan be ignored
        bool edit_wants_navigation;
        
        //! A function to be invoked when a menu is entered for the first time
        void (*enter_function)(MENU_ID);
        
        //! A function to be invoked when a menu is exited in menu_set_current_item.
        //  If a value other than MENU_USE_SUPPLIED_MENU_ID is returned then it 
        //  will override the menu supplied to the function.
        MENU_ID (*exit_function)(MENU_ID);
        
        //! A function to be invoked from a menu item when the OK button is pressed
        // NOTE: This would be responsible for any state change
        void (*process_function)(MENU_ID);
        
        //! Structure to hold bespoke information during editing - it is the responsibility
        // of the code using it to initialise and maintain the data[]
        MENU_EDIT_INFO edit_info;
        
    } MENU_ITEM;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Add a child menu to the supplied menu
     * @param menu - The parent menu
     * @param child - The child menu
     * @param default_item - Sets this item as the default item when navigating to this branch
     */
    void menu_add_child(MENU_ITEM* parent, 
            MENU_ITEM* child, 
            bool default_item);
    
    /**
     * Adds a function that will return a character string representation of 
     * the current menu item.  The function must take a single 
     * @param menu
     * @param func
     */
    void menu_add_display_function(MENU_ITEM* menu, void (*func)(char* buffer, size_t size));
    
    /**
     * Adds a function that can be used by the code that process menu commands to
     * respond to an input change that may require the data maintained by the menu item
     * to be updated/changed
     * @param menu - The menu to move from
     * @param func - a function pointer to a function that takes a single argument
     * which is the MENU_ID
     */
    void menu_add_draw_function(MENU_ITEM* menu, void (*func)(MENU_ID));
    
    /**
     * Adds a function that can be used by the code that process menu commands to
     * respond to an input change that may require the data maintained by the menu item
     * to be updated/changed
     * @param menu - The menu to move from
     * @param func - a function pointer to a function that takes a single pointer argument
     * which will be to a structured message (APP_COMMANDS_MENU_VALUE_CHANGE)
     * @param on_entry - If yes the item will enter Edit mode as soon as its selected, rather
     * than requiring OK to be pressed
     * @param wants_navigation - Navigation messages will be passed to the edit function, 
     * allowing it to use them for multi character editing
     */
    void menu_add_edit_function(MENU_ITEM* menu, bool (*func)(void*), bool on_entry, bool wants_navigation);
    
    /**
     * Adds a function that will be called when the menu exits from edit state
     * By default the menu will return to its non-edit display, this allows the code
     * to move to different code (e.g back to volume display or to another screen)
     * @param menu
     * @param func
     * @return - true if the function has done something, false if not
     */
    void menu_add_edit_exit_function(MENU_ITEM* menu, bool (*func)(MENU_ID));

    /** 
     * Adds enter and exit functions to the menu item -  NULL = no action
     */
    void menu_add_enter_exit_functions(MENU_ITEM* menu, void (*enter_func)(MENU_ID), MENU_ID (*exit_func)(MENU_ID));

    /**
     * Adds a function that can be invoked when the OK button is pressed when in 
     * MENU mode - this is an alternative to menu_add_edit_function
     * @param menu - The menu to move from
     * @param func - a function pointer to a function that takes a single pointer argument
     * which will the MENU_ID
     */
    void menu_add_process_function(MENU_ITEM* menu, void (*func)(MENU_ID));
    
    /**
    * Adds a function that can be used by the code to define ScreenData for use with the display_message utility
    * @param menu - The menu the function is attached to
    * @param screen_type - The type of screen CURRENTLY_PLAYING, OPTION_LIST, OPTION etc
    * @param func -  pointer to a function used to build the screen and its data
    * @param page_size - The maximum number of items to be displayed (OPTION_LIST/OPTIONS only) 
    *                  - if 0 then defaults to DISPLAY_MESSAGE_MAX_ROWS
    * @param option_strings - if supplied provides an array of strings representing 
    *                       - the option values (OPTION only) - MUST match the values 
    *                       - returned by the menu_add_display_function function
    * @param option_count - indicates the size of the options_strings array
    * @icon_id - identifies an icon that can be associated with the menu item
    */
    void menu_define_screen_data(MENU_ITEM* menu, 
            uint8_t screen_type, 
            uint16_t (*func)(MENU_ID, uint8_t*, uint8_t),
            uint8_t page_size,
            char** option_strings,
            uint8_t option_count,
            uint8_t icon_id);

    /**
     * The Menu Item can be used to store the current field in a multi field edit 
     * procedure (IP Address, Friendly Name)
     * It uses the first position of the edit_info structure
     * @param menu
     * @param num_fields - The number of fields that can be edited - note this may 
     * change throughout the lifetime of the field (e.g. when editing Network Friendly Name)
     * @return 
     */
    uint8_t menu_get_edit_pos(MENU_ITEM* menu, int num_fields);
    
    /**
     * Gets the currently selected menu item
     * @return 
     */
    MENU_ITEM* menu_get_current_item();
    
    /**
     * Get a reference to a MENU_ITEM base on its index
     */
    MENU_ITEM* menu_get_item(MENU_ID index);
    
    /**
     * Get a reference to the MENU_ITEMS parent, returns NULL for top level menus
     */
    MENU_ITEM* menu_get_parent(MENU_ITEM* menu);
    
    /**
     * Gets the previously selected menu item
     * @return 
     */
    MENU_ITEM* menu_get_previous_item();
    
    /**
     * Get a reference to a MENU_ITEM's title
     */
    char* menu_get_title(MENU_ID index);
    
    /**
     * Returns true if the current menu item has children
     */
    bool menu_has_children(MENU_ITEM* menu);
    
    /**
     * Initialises the menus data which is a pointer to an array of menu items
     * @param items
     * @param count
     */
    void menu_initialise(MENU_ITEM* items, char** titles, size_t count);
    
    /**
     * Initialises the supplied menu item
     * @param item
     * @param loop_back_children
     */
    void menu_initialise_item(MENU_ITEM* menu, MENU_ID menu_id, bool loop_back_children);
    
    /**
     * Returns whether the current MenuItem represents an editable value that 
     * is set to its original (unedited) value
     * @param menu
     * @return 
     */
    bool menu_is_editable_and_original(MENU_ITEM* menu);
    
    /**
     * Return the active or fist child of the menu item
     * @param menu - The menu to move from
      * @return 
     */
    MENU_ITEM* menu_move_down(MENU_ITEM* menu);
    
    /**
     * Return the next sibling of the menu item
     * @param menu
     * @return 
     */
    MENU_ITEM* menu_move_left(MENU_ITEM* menu);
    
    /**
     * Return the next logical item in the menu tree.  It will try and move down, if 
     * there is no child it will move right, again if no sibling it will move up
     * @param menu
     * @return 
     */
    MENU_ITEM* menu_move_next(MENU_ITEM* menu);
    
    /**
     * Return the previous sibling of the menu item
     * @param menu
     * @return 
     */
    MENU_ITEM* menu_move_right(MENU_ITEM* menu);
    
    /**
     * Returns the parent item
     * @param menu
     * @return 
     */
    MENU_ITEM* menu_move_up(MENU_ITEM* menu);
    
    /**
     * Changes the current menu, executing exit/enter functions if defined
     * @param menu_id
     */
    void menu_set_current_item(MENU_ID menu_id);
    
    /**
     * Used to set the current ordinal edit position in a multi field edit - to be used
     * in conjunction with menu_get_edit_pos
     * @param menu
     * @param edit_pos - the new edit position (note thsi may be < 0 or > num_fields - 1 and will be adjusted
     * @param num_fields - The number of fields that can be edited - note this may 
     * change throughout the lifetime of the field (e.g. when editing Network Friendly Name)
     */
    void menu_set_edit_pos(MENU_ITEM* menu, int8_t edit_pos, int num_fields);
    
    /**
     * To be used in edit functions to initialise the change state.  This is used
     * to monitor whether the current edit value is a change or is the original value
     * @param value
     * @param size
     */
    void menu_try_initialise_change_state(void* value, size_t size);

    /**
     * Used in conjunction with menu_try_initialise_change_state to determine
     * whether an edit is altering the original value of a menu item or returning it
     * to its original state
     * @param value
     * @param size
     */
    void menu_update_change_state(void* value, size_t size);

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
