/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    remote.h

  @Summary
 Remote control handling

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _REMOTE_H    /* Guard against multiple inclusion */
#define _REMOTE_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdint.h>
#include <stdbool.h>
#include "app_config.h"
#include "utilities/structured_message.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */
    
#ifndef BIT
#define BIT(N)                      (1 << (N))
#endif
#define SET_BIT_NR(VAL, BIT_N)		((VAL) |= (BIT(BIT_N)))
#define CLR_BIT_NR(VAL, BIT_N)		((VAL) &= ~ (BIT(BIT_N)))
#define BIT_NR_IS_SET(VAL, BIT_N)	((((VAL) & (BIT(BIT_N))) > 0) ? 1 : 0)

#define SET_BITS(VAL, BITS)         ((VAL) |= (BITS))
#define CLR_BITS(VAL, BITS)         ((VAL) &= ~ (BITS))
#define BITS_ARE_SET(VAL, BITS)     ((((VAL) & (BITS)) == (BITS)) ? 1 : 0)
    
#define RC5_FIELD_MASK			0x40
#define RC5_CMD_MASK			0x3F
#define RC5_ADD_MASK			0x1F

#define RC5_COMMAND(MES)	((MES) & (RC5_CMD_MASK)) | ((((MES) >> 6) & (RC5_FIELD_MASK)) ^ RC5_FIELD_MASK)
#define RC5_ADDRESS(MES)	(((MES) >> 6) & (RC5_ADD_MASK))

#ifndef REMOTE_SYSTEM_CODE_DEFAULT    
    /**
     * The address of the system used for IR remote messages
     */
    #define REMOTE_SYSTEM_CODE_DEFAULT   16
#endif

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    typedef unsigned char REMOTE_SYSTEM_CODE;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Clear down all outstanding messages - used when entering standby
     */
    void remote_clear();

    /**
     * Gets the current status of the remote system code (16/19)
     * @return 
     */
    REMOTE_SYSTEM_CODE remote_get_system_code();

    /**
     * Gets the remote system code (16/19) from the message
     * @return 
     */
    REMOTE_SYSTEM_CODE remote_get_system_code_from_message(STRUCTURED_MESSAGE* message);

    /**
     * Gets the current status of the remote system code (16/19) as a string
     * @return 
     */
    void remote_get_system_code_string(char* buffer, size_t size);
    
    /**
     * This detects changes on the pins.  To use this enable Interrupts in the 
     * Harmony configuration - that should generate a system_interrupt.c file.
     * Then add this code into system_interrupt.c
     * e.g.
     * #include "utilities/remote.h"
     *
     * void __ISR(_CHANGE_NOTICE_VECTOR, ipl1AUTO) _IntHandlerChangeNotification(void)
     * {
     *     remote_handle_interrupt();
     *     PLIB_INT_SourceFlagClear(INT_ID_0,INT_SOURCE_CHANGE_NOTICE);
     * }
     */
    void remote_handle_interrupt();
    
    /**
     * Initialise the remote 
     */
    void remote_initialise(STRUCTURED_MESSAGE_CALLBACK callback);
    
    // Called before a sleep to ensure the remote picks up the first interrupt command
    void remote_pause();
    
    /**
     * Process remote messages 
     */
    void remote_process();
    
    /**
     * Sets the remote system code
     * @param code
     */
    void remote_set_system_code(REMOTE_SYSTEM_CODE code);
    
    /**
     * Return true if the message has the same system code as the current system code
     * @param message
     * @return 
     */
    bool remote_use_message(STRUCTURED_MESSAGE* message);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
