/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    settings.h

  @Summary
 * Reads and stores settings to/from flash memory via I2C

  @Description
 * Maintains settings stored in flash. When a setting is changed a timer is 
 * invoked that means that the setting will be saved automatically if it has not 
 * been explicitly saved in the meantime.
 * 
 * The save operation saves whatever to current state of the settings are.
 * 
 * Note: It relies on a timer called TIMER_INDEX_SETTINGS setting up in 
 * system_setup.h
 */
/* ************************************************************************** */

#ifndef _SETTINGS_H    /* Guard against multiple inclusion */
#define _SETTINGS_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include "system/common/sys_module.h"
#include "app_config.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

/**
 * Define the number of settings to be managed - defined in system_setup.h
 */
#ifndef SETTINGS_NUM_SETTINGS
#define SETTINGS_NUM_SETTINGS           4
#endif

/**
 * Constant identifying an invalid setting index request
 */
#define SETTINGS_INVALID_INDEX          -1
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Initialise Callback function template
     */
    typedef void (*SETTINGS_INITIALISE_CALLBACK)();

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Adds a setting for the given index
     * @param index - the index to referto for the setting
     * @param size - the size (in bytes) of the setting
     */
    void settings_add(uint8_t index, size_t size);
    
    /**
    * Used to see if the settings are being saved/loaded
    * Note: Only returns true during normal running
    */
    bool settings_available();
   
    /**
    * Get the specified setting
    * @param index - The index of the setting
    * @param output - A pointer to storage to receive the setting value
     */
    void settings_get(uint8_t index, void* output);
    
    /**
     * Initialises the resources used for the settings
     * 
     * @returns true if settings have been read from flash, false if there are 
     * no settings in flash - allows the code to set defaults if this is the 
     * first time.  The method 
     */
    void settings_initialise(SETTINGS_INITIALISE_CALLBACK callback);
    
    /**
     * Used to check whether the settings have been initialised.
     * Note during setup this need to be called repeatedly in the main application 
     * loop until settings have been loaded
     * @return - True if the settings have been initialised, false if not.
     */
    bool settings_initialised();
    
    /**
     * Checks whether settings are ready to save and performs save and reset
     */
    void settings_process();

    /**
     * Saves the current state of the settings (if changed)
     */
    void settings_save();
    
    /**
     * Set the specified setting
     * @param index - The index of the setting
     * @param - a pointer to the value to be stored - note this must be the same size 
     * @return 
     */
    int settings_set(uint8_t index, void* value);

    /**
     * Use to temporarily suppress/un-suppress the automatic triggering of a save.
     * Note if a save has already been triggered then it will still go ahead, and any
     * changes made if this is subsequently set before the save is complete will 
     * also be saved
     * This was intended to prevent a save during the loading of settings where 
     * the loaded settings may be used to set properties such as volume/balance etc
     * using methods that may inturn usually trigger another save.
     * @param suppress - true - an automatic delayed save will not be triggered
     * by any calls to settings_set(), false - calls to settings_set(0 will trigger 
     * an automatic delayed save
     */
    void settings_suppress_save(bool suppress);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _SETTINGS_H */

/* *****************************************************************************
 End of File
 */
