/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _TFT_CONTROL_H    /* Guard against multiple inclusion */
#define _TFT_CONTROL_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

/* This section lists the other files that are included in this file.
 */
#include <stdint.h>
#include "utilities/display_message_data.h"
#include "utilities/display_message_display.h"
#include "utilities/font_manager.h"
#include "utilities/tft_display_data.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    typedef enum 
    {
        TFT_CONTROL_TYPE_TEXT,
        TFT_CONTROL_TYPE_PROGRESS,
        TFT_CONTROL_TYPE_IMAGE,
        TFT_CONTROL_TYPE_RECTANGLE,
        TFT_CONTROL_TYPE_GOOGLE_VOICE,
    } TFT_CONTROL_TYPE;
    
    typedef enum
    {
        TFT_CONTROL_ALIGNMENT_TOP_LEFT,
        TFT_CONTROL_ALIGNMENT_CENTER,
        TFT_CONTROL_ALIGNMENT_TOP_RIGHT,
    } TFT_CONTROL_ALIGNMENT;
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     */
    typedef struct 
    {
        // The ordinal position i the group of the active control
        uint8_t active_index;
        
        //! Determines whether a scroll group is active or not
        uint8_t control_count;
        
        uint8_t pause_count; // The number of seconds elapsed in the current pause
        
        // The number of seconds that scrolling should be paused in-between cycles
        uint8_t pause_max;
    } TFT_CONTROL_SCROLL_GROUP;
    
    /**
     */
    struct TFT_CONTROL
    {
        DISPLAY_MESSAGE_DATA_ID data_id;

        uint8_t attribute_id;    
        
        //! Defines the bounds of the control
        BOUNDS bounds;
        
        // determines whether to display the control or not
        bool visible;
        
        //! the type of control - set i the tft_control_define_nnn_properties functions
        TFT_CONTROL_TYPE type;
        
        union control_data
        {
            struct text
            {
                TFT_CONTROL_ALIGNMENT alignment;
                FONT_ID font_id; 
                OFFSET offset;
                COLOUR_INDEX foreground;
                COLOUR_INDEX background;
                
                //! Scroll group identifier - 0 indicates no scroll group
                uint8_t scroll_group;
                uint16_t scroll_offset;
                uint16_t scroll_max;
            } text;
            
            struct image
            {
                TFT_CONTROL_ALIGNMENT alignment;
                bool being_retrieved;
            } image;
            
            struct progress
            {
                uint8_t progress;
                uint8_t prev_progress;
                COLOUR_INDEX foreground; 
                COLOUR_INDEX background;
            } progress;
            
            struct rectangle
            {
                COLOUR_INDEX foreground; 
                COLOUR_INDEX background;
            } rectangle;
            
            struct google_voice
            {
                COLOUR_INDEX foreground;         
                uint16_t cycle_length;
                uint16_t cycle_position;
            } google_voice;
            
        } control_data;
        
        void (*draw_function)(struct TFT_CONTROL*);
    };
    
    typedef struct TFT_CONTROL TFT_CONTROL;
    
    /**
     * Callback function template
     */
    typedef void (*TFT_CONTROL_DRAW_FUNCTION)(TFT_CONTROL* ctrl);

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * gets the current number of active controls
     * @param data_id
     * @param attribute_id
     * @return 
     */
    uint8_t tft_control_active_controls();
    
    /**
     * Allocates a new TFT control
     * @return 
     */
    TFT_CONTROL* tft_control_allocate(DISPLAY_MESSAGE_DATA_ID data_id,
                                      uint8_t attribute_id,
                                      TFT_CONTROL_DRAW_FUNCTION draw_function);
    
    /**
     * Sets the area defined for the control to the background colour
     * @param redraw_immediate
     */
    void tft_control_clear(TFT_CONTROL* ctrl, bool redraw_immediate);
    
    /**
     * Used when displaying the screen to clear down/free control data ready for controls to be
     * defined.
     */
    void tft_control_clear_controls();
 
    /**
     * 
     * @param ctrl
     * @param top
     * @param left
     * @param bottom
     * @param right
     * @return 
     */
    void tft_control_define_bounds(TFT_CONTROL* ctrl, uint16_t top, uint16_t left, uint16_t bottom, uint16_t right);

    /**
     */
    void tft_control_define_google_voice_properties(TFT_CONTROL* ctrl, COLOUR_INDEX foreground_colour);

    /**
     */
    void tft_control_define_image_properties(TFT_CONTROL* ctrl, TFT_CONTROL_ALIGNMENT alignment);

    /**
     */
    void tft_control_define_progress_properties(TFT_CONTROL* ctrl,
            COLOUR_INDEX foreground_colour, 
            COLOUR_INDEX background_colour);

    /**
     */
    void tft_control_define_rectangle_properties(TFT_CONTROL* ctrl,
            COLOUR_INDEX foreground_colour, 
            COLOUR_INDEX background_colour);
 
    /**
     * 
     * @param ctrl
     * @param top
     * @param left
     * @param bottom
     * @param right
     * @return 
     */
    void tft_control_define_rect(TFT_CONTROL* ctrl, uint16_t top, uint16_t left, uint16_t height, uint16_t width);

    /**
     */
    void tft_control_define_scroll_info(TFT_CONTROL* ctrl, 
            uint8_t scroll_group);
    
    /**
     */
    void tft_control_define_text_colours(TFT_CONTROL* ctrl, 
            COLOUR_INDEX foreground_colour, 
            COLOUR_INDEX background_colour);

    /**
     */
    void tft_control_define_text_properties(TFT_CONTROL* ctrl, 
            TFT_CONTROL_ALIGNMENT alignment, 
            FONT_ID font_id, 
            OFFSET offset,
            COLOUR_INDEX foreground_colour, 
            COLOUR_INDEX background_colour);
    
    /**
     * Invokes the controls draw function
     * @param 
     */
    void tft_control_draw(TFT_CONTROL* ctrl, bool force);

    /**
     * attempts to draw all allocated controls
     */
    void tft_control_draw_all(bool force);

    /**
     * 
     * @param bounds
     */
    void tft_control_draw_progress(TFT_CONTROL* ctrl);

    /**
     * 
     * @param bounds
     */
    void tft_control_draw_text(TFT_CONTROL* ctrl);
    
    /**
     * 
     * @param data_id
     * @param attribute_id
     * @return 
     */
    TFT_CONTROL* tft_control_get(DISPLAY_MESSAGE_DATA_ID data_id, uint8_t attribute_id);
    
    /**
     * Gets the first control matching the data id - for use with Option Lists
     * @param data_id
     * @param attribute_id
     * @return 
     */
    TFT_CONTROL* tft_control_get_by_data_id(DISPLAY_MESSAGE_DATA_ID data_id);
    
    /**
     * Gets the control at the specified index - requires the called to know what order controls are i the screen
     * @param data_id
     * @param attribute_id
     * @return 
     */
    TFT_CONTROL* tft_control_get_at(uint8_t index);
    
    /**
     * Gets whether the control is visible or not
     */
    bool tft_control_get_visibility(TFT_CONTROL* ctrl);

    /**
     * Can be used to determine whether any controls are currently allocated
     * @return 
     */
    bool tft_control_has_controls();
    
    /**
     */
    void tft_control_initialise();

    /**
     * Used to reset all active controls to the state where they have just been created
     */
    void tft_control_reset_controls();
    
    /**
     * 
     * @param data_id
     * @param attribute_id
     * @param draw_function
     * @return 
     */
    void tft_control_scroll_group_initialise();
    
    /**
     * Used to review control groups
     */
    void tft_control_scroll_group_review();
    
    /**
     */
    void tft_control_set_visibility(TFT_CONTROL* ctrl, bool visible);

    /**
     * Used to update the progress value of a cntrol
     */
    void tft_control_update_progress_value(TFT_CONTROL* ctrl, uint8_t progress);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
