/* ************************************************************************** */
/** Descriptive File Name

  @Company
 * Arcam

  @File Name
    timers.h

  @Summary
    Manages a number of timers each of a whole number of milliseconds

  @Description
    Manages a number of timers each of a whole number of milliseconds.
 * Timers are added using the timers_add_timer method which should be called after
 * timers_initialise().  Note a separate timer instance needs setting up in the Harmony
 * configuration tool for each timer added
 * 
 * I had issues using instance 0 as when initialising during APP_Tasks it appears 
 * that the module is not in the right state?
 */
/* ************************************************************************** */

#ifndef _TIMERS_H    /* Guard against multiple inclusion */
#define _TIMERS_H

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include "system/common/sys_module.h"
#include "app_config.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */
    
    //! TIMERS_NUM_TIMERS
    /**
     * Defines the number of timers to be managed - These should be defined in 
     * system_setup.h
     * 
     * A timer can be initialise in one of 3 states:
     *  NONE -      The timer is not active but is in an expired state meaning that a 
     *              call to timers_has_expired will return true.  Intended for use where
     *              the timer does nothing until it is checked and reset.
     *  INACTIVE -  The timer is not active and not expired.  timers_has_expired will 
     *              return false.  The timer will need a reset before doing anything
     *  IMMEDIATE - The timer will start counting down as soon as it is initialised.
     * 
     * Suspending the timer does not stop it counting down but calls to timers_reset 
     * will be ignored.
     */

//! Uses TIMER_ITEMS defined in app_config.h
// Now use the input data to define the enum
#define X(a, b, c, d) TIMER_INDEX_##a,
	typedef enum {TIMER_ITEMS TIMER_INDEX_MAX} TIMER_INDEX;
#undef X

// Now use the input data to define the enum
#define X(a, b, c, d) TIMER_TIMEOUT_MS_##a,
	typedef enum {TIMER_ITEMS} TIMER_TIMEOUT;
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d) b,
    static int TIMER_TIMEOUT_VALUES[] = { TIMER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d) c,
    static int TIMER_TIMEOUT_MODES[] = { TIMER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d) d,
    static int TIMER_TIMEOUT_HANDLES[] = { TIMER_ITEMS };
#undef X

/**
 * Defines the number of timers to be managed
 */
#define TIMERS_NUM_TIMERS TIMER_INDEX_MAX

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    // Method triggered when a timer has expired
    typedef void (*TIMERS_EXPIRED_EVENT)(unsigned int timer_idx);
    
    /*  Used to determine whether the timer is invoked immediately or waits for 
     * the first reset.
     */
    typedef enum
    {
        //! The timer is left in an expired state, meaning a timers_has_expired will return true
        TIMERS_TIMEOUT_RESET_NONE = 0,            

       //! The timer is left in a reset state but inactive, meaning a timers_has_expired will return false
        TIMERS_TIMEOUT_RESET_INACTIVE = 1,          //TIMERS_RESET_INACTIVE from app_config.h        

        //! The timer is immediatly counting down
        TIMERS_TIMEOUT_RESET_IMMEDIATE = 2,         //TIMERS_RESET_IMMEDIATE from app_config.h 
     } TIMERS_TIMEOUT_RESET;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Activates a suspended timer - when suspended it will never expire
     * @param timer_idx
     */
    void timers_activate(unsigned int timer_idx);
    
    /**
     * Adds an event to the specified timer tat is triggered whenever the timer 
     * expires this is done as part of the interrupt event so should be kept as 
     * brief as possible
     * @param timer_idx
     * @param event
     */
    void timers_add_expired_event(unsigned int timer_idx, TIMERS_EXPIRED_EVENT event);
    
    /**
     * Add a timer to the timers module
     * @param module_index - The module index e.g. DRV_TMR_INDEX_0
     * @param frequency - the frequency of the time (in Hz)
     * @return - An index to reference the timer by when adding timer instances (see
     * timers_set_timeout)
     */
    short timers_add_timer(const SYS_MODULE_INDEX module_index, int frequency);
    
     /**
      * Delay a process by a given number of milliseconds
      * Note: Uses the core timer
      */
     void timers_delay_ms(uint16_t milliseconds);
    
     /**
      * Delay a process by a given number of 100 nanoseconds
      * Note: Uses the core timer
      */
     void timers_delay_100ns(uint16_t nanoseconds);
     
     /**
      * Delay a process by a given number of microseconds
      * Note: Uses the core timer
      */
     void timers_delay_us(uint16_t microseconds);

     /**
      * Forces the current timer to put itself into the expired state
      * @param timer_idx - The timer index
      */
     void timers_expire(unsigned int timer_idx);
     
    /**
     * Gets the number of cycles since the last reset
     * Note : there is no range checking on the index supplied
     * @param timer_idx - The timer index
     * @return - the number of complete cycles
     */
     unsigned int timers_get_elapsed_count(unsigned int timer_idx);
     
    /**
     * Gets the number of cycles left until last reset
     * Note : there is no range checking on the index supplied
     * @param timer_idx - The timer index
     * @return - the number of complete cycles
     */
    unsigned int timers_get_remaining_count(unsigned int timer_idx);
     
    /**
     * Gets the current timeout value for a given timer
     * @param timer_idx
     * @return 
     */
    unsigned int timers_get_timeout(unsigned int timer_idx);      
    
    //! has_timer_expired
    /**
     * Checks if the timer has expired.  Note this returns true whether the timer is active or not
     * use timers_is_active in conjunction with this to determine whether this processes
     * and active timer or not.
     * @param timer_idx
     * @return 
     */
    bool timers_has_expired(unsigned int timer_idx);
  
    //! initialise_timers
    /**
     * Initialise the timer functionality 
     * This should be called directly during initialisation when  a dedicated 
     * timer is not being used, otherwise call the initialise_managed_timers function
     */
    void timers_initialise(void);
    
    /**
     * Used to test whether a timer is currently active.  
     * @param timer_idx
     * @return 
     */
    bool timers_is_active(unsigned int timer_idx);
    
    //! reset_timer
    /**
     * resets the timeout of a timer in ms if the timer is active
     * Note: Will cause a reset
     * Note: timer_idx must be < TIMERS_NUM_TIMERS
     */
    void timers_reset(unsigned int timer_idx);
    
    /**
     * Suspends the timer - when suspended it will never expire
     * @param timer_idx
     */
    void timers_suspend(unsigned int timer_idx);
    
    /**
     * Set the timeout of a timer in ms
     * Note: Will cause a reset
     * Note: timer_idx must be < TIMERS_NUM_TIMERS
     * 
     * @param module_context - the context returned by timers_add_timer
     * @param timer_idx - An identifer for the TIMER
     * @param timeout - the timeout frequency in units based on the module frequency
     * @param reset - Indicates the start state of the timer
     */
    void timers_set_timeout(short module_context, unsigned int timer_idx, unsigned int timeout, TIMERS_TIMEOUT_RESET reset);
    
    //! set_timer
    /**
     * Set the timeout of a timer in ms
     * Note: Will cause a reset
     * Note: timer_idx must be < TIMERS_NUM_TIMERS
     * 
     * @param timer_idx - An identifer for the TIMER
     * @param timeout - the timeout frequency in units based on the module frequency
     * @param reset - Indicates the start state of the timer
     */
    void timers_update_timeout(unsigned int timer_idx, unsigned int timeout, TIMERS_TIMEOUT_RESET reset);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
