/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    ak4104_spdif_tx.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\ak4104_spdif_tx.h"
#include "app_config.h"
#include "utilities\timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Disable Chip Select pin
 */
static void ak4104_spdif_tx_disable_transfer()
{
    PLIB_PORTS_PinSet(AK4104_SPDIF_TX_CS_PORTS_ID, 
                      AK4104_SPDIF_TX_CS_PORT_CHANNEL, 
                      AK4104_SPDIF_TX_CS_PIN);
}

/**-----------------------------------------------------------------------------
 * Enable CS pin
 */
static void ak4104_spdif_tx_enable_transfer()
{
    PLIB_PORTS_PinClear(AK4104_SPDIF_TX_CS_PORTS_ID, 
                        AK4104_SPDIF_TX_CS_PORT_CHANNEL, 
                        AK4104_SPDIF_TX_CS_PIN);
}

/**-----------------------------------------------------------------------------
 * Send SPI data (register settings)
 * @param on
 */
static void ak4104_spdif_tx_send_register_data(uint8_t reg, uint8_t data)
{
 	uint16_t value = 0xE000;  // 1110000000000000 
    
    // Add register address (assumes address is <= 5 bits))
    value |= (reg << 8);
    
    // Add data
    value |= data;
    
    //logger_log_debugf(LOGGER_CODE_AK4104_SPDIF_TX, "AK4104 Audio - Setting register reg:%d data:%d %d --", reg, data, value);    
 	
	// enable transfer
    ak4104_spdif_tx_enable_transfer();
    
  	int pos;
	for (pos = 15; pos >= 0; pos--)
	{
		// set data from msb
        PLIB_PORTS_PinWrite(AK4104_SPDIF_TX_MOSI_PORTS_ID, 
                            AK4104_SPDIF_TX_MOSI_PORT_CHANNEL, 
                            AK4104_SPDIF_TX_MOSI_PIN,
                            ((value >> pos) & 1) == 1);
		
		// set clock low
        PLIB_PORTS_PinClear(AK4104_SPDIF_TX_SCLK_PORTS_ID, 
                            AK4104_SPDIF_TX_SCLK_PORT_CHANNEL, 
                            AK4104_SPDIF_TX_SCLK_PIN);
		
		timers_delay_us(1);
		
		// set clock high
        PLIB_PORTS_PinSet(AK4104_SPDIF_TX_SCLK_PORTS_ID, 
                          AK4104_SPDIF_TX_SCLK_PORT_CHANNEL, 
                          AK4104_SPDIF_TX_SCLK_PIN);
		
		timers_delay_us(1);
	}

	// terminate transfer
	ak4104_spdif_tx_disable_transfer();
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Switch Avatar Audio feed
 */
void ak4104_spdif_tx_audio(bool on)
{
#if AK4104_SPDIF_TX_PINS_ENABLED
    logger_log_debugf(LOGGER_CODE_AK4104_SPDIF_TX, "AK4104 Audio - Switching %s --", on == true ? "On" : "Off");
    
    PLIB_PORTS_PinWrite(AK4104_SPDIF_TX_AUDIO_PORTS_ID, 
                        AK4104_SPDIF_TX_AUDIO_CHANNEL, 
                        AK4104_SPDIF_TX_AUDIO_PIN,
                        !on); // Note Audio on is low

#endif        
}

/**-----------------------------------------------------------------------------
 * Initialise the module
 * Note - do after power_initialise as we assume that the module has power
 */
void ak4104_spdif_tx_initialise()
{
    logger_log_debugf(LOGGER_CODE_AK4104_SPDIF_TX, "AK4104 Initialise --");
    
    // Take out of reset
    ak4104_spdif_tx_reset(AK4104_SPDIF_TX_SWITCH_ON);
    
    // Register settings
    ak4104_spdif_tx_send_register_data(0x02, 0x04);
    ak4104_spdif_tx_send_register_data(0x04, 0x00);

    // Enable audio (probably async)
    ak4104_spdif_tx_audio(AK4104_SPDIF_TX_SWITCH_ON);
}

/**-----------------------------------------------------------------------------
 * Mutes the module
 * Note its only bit 0 that does the mute if we change the default value, in initlaise
 * then the value (0x80) passed must be different
 */
void ak4104_spdif_tx_mute(bool on)
{
    //logger_log_debugf(LOGGER_CODE_AK4104_SPDIF_TX, "AK4104 Mute - Switching %s --", on == true ? "On" : "Off");
    
    if (on)
    {
        // Mute the output
        ak4104_spdif_tx_send_register_data(0x03, 0x80);
    }
    else
    {
        // Unmute
        ak4104_spdif_tx_send_register_data(0x03, 0x81);
    }
}

/**-----------------------------------------------------------------------------
 * Put the module in reset
 */
void ak4104_spdif_tx_reset(bool on)
{
#if AK4104_SPDIF_TX_PINS_ENABLED
    logger_log_debugf(LOGGER_CODE_AK4104_SPDIF_TX, "AK4104 Reset - Switching %s --", on == true ? "On" : "Off");
    
    PLIB_PORTS_PinWrite(AK4104_SPDIF_TX_RESET_PORTS_ID, 
                        AK4104_SPDIF_TX_RESET_CHANNEL, 
                        AK4104_SPDIF_TX_RESET_PIN,
                        on); 

#endif        
    
}

/* *****************************************************************************
 End of File
 */
