/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/avatar_message.h"
#include <string.h>
#include <stdlib.h>
#include "app_config.h"
#include "utilities/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! A cache of AVATAR_MESSAGES for processing.  
    // This was added for handling multi-packet messages.
    AVATAR_MESSAGE processing_cache[AVATAR_PACKET_QUEUE_SIZE];
    
} AVATAR_MESSAGE_DATA;

static AVATAR_MESSAGE_DATA avatar_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the index of the supplied state in the message lookup info , -1 if not found
 * @param message
 * @param state_id
 * @return 
 */
static int8_t avatar_message_state_idx(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id)
{
    int8_t result = -1;
    
    int x;
    for (x = 0; x < message->num_states; x++)
    {
        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR 0x%04X 0x%04X --", message->lookup[x].state_id , state_id);
        if (message->lookup[x].state_id == state_id)
        {
            result = x;
            break;
        }
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Process a packet containing data from the Avatar (Set States, Get States Response)
 * @param packet
 * @param message - pointer to an AVATAR_MESSAGE to be populated
 */
void avatar_message_build(AVATAR_PACKET* packet, AVATAR_MESSAGE* message)
{
    uint16_t data_length = avatar_packet_header_get_data_length(&packet->header);
    uint16_t offset = 0;

    // Remove tail
    data_length -= 2;
    
    // Set the packet
    message->packet = (void*)packet;

    /* Were going to make an assumption about the data in message here, otherwise 
     * this will get very complicated!!
     * We will assume that where there are multiple packets the data contained in 
     * the packets after the first only contain data for the last field in the message
     * (this will only occur where data > 4096bytes) which will be a BIN16
     * Otherwise we are into trying to work out what info has been processed so 
     * far and what hasnt..
     */
    if (packet->header.packet_id == 0)
    {
        // The avatar is requesting data this will be in a list of uint16 values
        // Get number of values
        message->message_id = avatar_packet_get_message_id(packet);
        message->num_states = avatar_packet_get_num_values(packet);

        // Set the offset after the message type (2-bytes) and number of states (1-byte))
        offset = 3;
        
        // Start finding the position of the states in the message
        int x;
        for (x = 0; x < message->num_states; x++)
        {
            // Set the ordinal position in the data of the start of the message data
            message->lookup[x].offset = offset;

            // Process each value AVATAR_STATE_ID
            uint16_t state_id;
            memcpy(&state_id, packet->data + offset, sizeof(uint16_t));
            offset += sizeof(uint16_t);

            message->lookup[x].state_id = state_id;

            if ((message->message_id == AVATAR_MESSAGE_ID_SET_STATES) ||
                (message->message_id == AVATAR_MESSAGE_ID_STATE_CHANGE) ||
                (message->message_id == AVATAR_MESSAGE_ID_GET_STATES_RESPONSE) ||
                (message->message_id == AVATAR_MESSAGE_ID_SET_STATES_RESPONSE) ||
                (message->message_id == AVATAR_MESSAGE_ID_TCPIP_TRANSFER) ||
                (message->message_id == AVATAR_MESSAGE_ID_TCPIP_TRANSFER_RESPONSE))
            {
                uint8_t encoding = *(packet->data + offset);
                offset += sizeof(uint8_t);

                message->lookup[x].encoding = encoding;

                // Calculate the offset of the next entry
                switch((AVATAR_ENCODING)encoding)
                {
                    case AVATAR_ENCODING_UINT8:
                    {
                        message->lookup[x].data_length = sizeof(uint8_t);
                        break;
                    }
                    case AVATAR_ENCODING_UINT16:
                    {
                        message->lookup[x].data_length = sizeof(uint16_t);
                        break;
                    }
                    case AVATAR_ENCODING_UINT32:
                    {
                        message->lookup[x].data_length = sizeof(uint32_t);
                        break;
                    }
                    case AVATAR_ENCODING_UINT64:
                    {
                        message->lookup[x].data_length = sizeof(uint64_t);
                        break;
                    }
                    case AVATAR_ENCODING_INT8:
                    {
                        message->lookup[x].data_length = sizeof(int8_t);
                        break;
                    }
                    case AVATAR_ENCODING_INT16:
                    {
                        message->lookup[x].data_length = sizeof(int16_t);
                        break;
                    }
                    case AVATAR_ENCODING_INT32:
                    {
                        message->lookup[x].data_length = sizeof(int32_t);
                        break;
                    }
                    case AVATAR_ENCODING_INT64:
                    {
                        message->lookup[x].data_length = sizeof(int64_t);
                        break;
                    }        
                    case AVATAR_ENCODING_NIL:
                    {
                        message->lookup[x].data_length = 0;
                        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet Data NIL %d [%s] NIL --", x, avatar_state_id_string(state_id));
                        break;
                    }        
                    case AVATAR_ENCODING_BIN16:
                    {
                        memcpy(&message->lookup[x].total_length, packet->data + offset, sizeof(uint16_t));
                        
                        // Here len represents the total size of the BIN16 data
                        // even if that spreads over a number of packets
                        // So restrict it to fit only in this message
                        message->lookup[x].data_length = min(message->lookup[x].total_length + sizeof(uint16_t), data_length - offset);
                        
                        //logger_log_debugf(0, "AVATAR DATA %d %d %d %d --", message->lookup[x].total_length, data_length, offset, message->lookup[x].data_length);

                        // If this BIN16 field is actually spread over more than one 
                        // packet then it doesnt matter what offset will be set to as this will be the last
                        // state in the message anyway...
                        //offset += len + sizeof(uint16_t);
                        break;
                    }        
                }
                
                offset += message->lookup[x].data_length;
            }       
        }
    }
    else
    {
        // The message now only contains the final state info, so we will move the
        // final lookup to be the first.  By doing this the first lookup will
        // be used for the satate id when searching
        if (message->num_states > 1)
        {
            memcpy(&message->lookup[0], &message->lookup[message->num_states - 1], sizeof(AVATAR_MESSAGE_LOOKUP));
        }
        
        message->lookup[0].offset = offset;
        message->lookup[0].data_length = data_length;

        // Adjust data_length if last packet
        // NOTE!!! There is an assumption here that the tail will not be spread over 2 packets!!!
        // if for any reason this is a problem then we will need to store the BIN16 field length
        // when packet_id is 0 and use that to determine whether data_length includes any of the tail...
        // BUG 3735 - Subtle one this.... the album art for the track in question 
        // (Agnes Obel, Dorian, via webclient / QoBuz) hapens to take 2 packest with 
        // the second packet of data being 1 byte long.  When parsing the second packt 
        // in avatar_message_build I was subtracting the length of the tail twice 
        // making the expected data lengt for the fiels 65,535 so no surpise that 
        // it fell over....Now removed - you would think this might mess up thr 
        // album art but so far it does not apear to....
        //if ((message->packet->header.control_flags & AVATAR_CONTROL_FLAGS_TYPE_FEND) == AVATAR_CONTROL_FLAGS_TYPE_FEND)
        //{
            //logger_log_debugf(0, "AVATAR ALBUM ART END MESSAGE %d %d %d --", message->num_states, offset, data_length);        
            //message->lookup[0].data_length = data_length - 2;
        //}
    }
}


/**-----------------------------------------------------------------------------
 * Searches the processing queue for a message associated with the packets frame id
 * @param packet
 * @return A pointer to a message representing the packet. If this is a new frame then
 * a free message will be returned (already built)
 */
AVATAR_MESSAGE* avatar_message_for_packet(AVATAR_PACKET* packet)
{
    AVATAR_MESSAGE* result = NULL;

    int x;
    for (x = 0; x < AVATAR_PACKET_QUEUE_SIZE; x++)
    {
        if ((avatar_data.processing_cache[x].active) &&
            (packet->header.frame_id == avatar_data.processing_cache[x].frame_id))
        {
            result = &avatar_data.processing_cache[x];
            break;
        }
        else if (result == NULL)
        {
            // Nominally allocate the next free AVATAR_MESSAGE as the result,
            // unless we find a match
            result = &avatar_data.processing_cache[x];
        }
    }
    
    if (result != NULL)
    {
        avatar_message_build(packet, result);
        
        result->active = true;
        result->frame_id = packet->header.frame_id;
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Unable to obtain AVATAR_MESSAGE for frame --", packet->header.frame_id);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the number of state ids in the message
 * @return 
 */
uint8_t avatar_message_get_num_states(AVATAR_MESSAGE* message)
{
    return message->num_states;
}
    
/**-----------------------------------------------------------------------------
 * Puts the state data into an array.  For bin 16 fields this will be the data 
 * only and omit the length part of the bin 16 field.
 * @param message - The message containing the data
 * @param state_id - The state Id for the data
 * @param encoding - The encoding type of the data
 * @param data - A pointer to an array of uint8_t to hold the data
 * @param len - The length of data available in the array - if the actual data is longer only this amount will be copied.
 * @return the length of the data copied (0 if no data or NIL type).
 */
uint16_t avatar_message_get_state_data(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id, AVATAR_ENCODING* encoding, uint8_t* data, uint16_t data_len)
{
    void* packet_data_ptr;
    uint16_t len = 0;
    
    if (avatar_message_get_state_value(message, 
            state_id, 
            encoding, 
            &packet_data_ptr, 
            &len))
    {
        // Adjust the length for BIN16
        switch(*encoding)
        {
            case AVATAR_ENCODING_BIN16:
            {
                // Check whether this is the first packet in the frame
                if (message->packet->header.packet_id == 0)
                {
                    packet_data_ptr += 2;
                    
                    // Check len > 2 - Alan (29/10/220 21:36) reported that the Avatar has sent a packet len = 1 
                    // which causes the length to be set to 0XFFFF!!!
                    if (len >= 2)
                    {
                        len -= 2;
                    }
                    else
                    {
                        logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Invalid BIN16 length detected %d --", len);
                        len = 0;
                    }
                }
                break;
            }
            
            default:
            {
                break;
            }
        }

        // Cope for when data_len is not specified, in that case have a go with len
        data_len = data_len > 0 ? data_len : len;
        
        memset(data, 0, data_len);
        memcpy(data, packet_data_ptr, min(len, data_len));
    }
    
    return len;
}
 
/**-----------------------------------------------------------------------------
 * Gets the state Id at a specific index in the message.  Assumes the message has been built
 * with avatar_message_build
 * @param message
 * @param index
 * @return 
 */
uint16_t avatar_message_get_state_id(AVATAR_MESSAGE* message, uint8_t index)
{
    return message->lookup[index].state_id;
}
     
/**-----------------------------------------------------------------------------
 * Gets the total length of the state data reported in the first packet
 * @param message
 * @param index
 * @return 
 */
uint16_t avatar_message_get_state_total_length(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id)
{
    uint16_t result = 0;
    
    int8_t idx = avatar_message_state_idx(message, state_id);
    if (idx >= 0)
    {
        result = message->lookup[idx].total_length;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Populates a pointer with the address of the actual data in the packet
 * @param message
 * @param state_id
 * @param encoding
 * @param value - A pointer to a pointer that will be set to point to the data in th packet
 * @return true if state_if found in message
 */
bool avatar_message_get_state_value(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id, AVATAR_ENCODING* encoding, void** value, uint16_t* len)
{
    bool result = false;
    
    int8_t idx = avatar_message_state_idx(message, state_id);
    if (idx >= 0)
    {
        uint16_t offset = message->lookup[idx].offset;
        
        // Setup the encoding
        if (encoding != NULL)
        {
            *encoding = message->lookup[idx].encoding;
        }
        
        AVATAR_PACKET* packet = (AVATAR_PACKET*)message->packet;
        int x = idx;
        
        if (value != NULL)
        {   
            // If this is the first packet for the frame offset will always be > 0
            // Otherwise we are assuming it is the final BIN16 field in the message
            // so the data is directly onto the message payload (no BIN16 length))
            if (packet->header.packet_id == 0)
            {
                // Offset points to the start of the state info which is the state_id followed by the encoding
                // so we need to skip over those
                *value = (void*)(packet->data + offset + sizeof(uint16_t) + sizeof(uint8_t));
            }
            else
            {
                *value = (void*)(packet->data);
            }
        }
        
        if (len != NULL)
        {
            *len = message->lookup[idx].data_length;
        }
        
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Determines whether an avatar message contains info on a specific state
 * @param message
 * @param state_id
 * @return 
 */
bool avatar_message_has_state(AVATAR_MESSAGE* message, AVATAR_STATE_ID state_id)
{
    bool result = false;
    
    if (avatar_message_state_idx(message, state_id) >= 0)
    {
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Initialises processing cache
 */
void avatar_message_initialise_cache()
{
    int x;
    for (x = 0; x < AVATAR_PACKET_QUEUE_SIZE; x++)
    {
        avatar_data.processing_cache[x].active = false;
    }
}
    
/**-----------------------------------------------------------------------------
 * Called once a message packet has been processed.  Will either release the 
 * message opr return it to the cache dependent on whether it is an FEND or not.
 * @param message
 */
void avatar_message_processed(AVATAR_MESSAGE* message)
{
    // Check for FEND and release if processing of frame is complete
    if ((message->packet->header.control_flags & AVATAR_CONTROL_FLAGS_TYPE_FEND) == AVATAR_CONTROL_FLAGS_TYPE_FEND)
    {
        message->active = false;
    }
}

/* *****************************************************************************
 End of File
 */
