/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/avatar_packet.h"
#include <string.h>
#include "app_config.h"
#include "utilities/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//!  Defines for Avatar packet header/trailer
#define AVATAR_PACKET_START             0xFEFF
#define AVATAR_PACKET_TAIL              0x0000

//! BIT positions for the package control flags
#define AVATAR_PACKET_ACK_BIT           8
#define AVATAR_PACKET_NACK_BIT          7   

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets a string representation of the message id.  Note if there is no match this may
 * give unpredictable results
 * @param message_id
 * @return 
 */
static char* avatar_packet_message_id_string(AVATAR_MESSAGE_ID message_id)
{
    const char empty_string[1] = "";
    char* result = (char*)&empty_string[0];
    int8_t idx = -1;
    
    // Find the index of the current value
    int x;
    for (x = 0; x < sizeof(AVATAR_MESSAGE_ID_STRINGS)/sizeof(AVATAR_MESSAGE_ID_STRINGS[0]); x++)
    {
        if (message_id == AVATAR_MESSAGE_ID_VALUES[x])
        {
            result = AVATAR_MESSAGE_ID_STRINGS[x];
            break;
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Set the number of values in a packet (byte 3)
 * @param packet
 * @return 
 */
static void avatar_packet_set_num_values(AVATAR_PACKET* packet, uint8_t num_values)
{
    packet->data[2] = num_values;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Add a data item to a packet - this will increment the number of values in the packet, 
 * and the size of the packet header
 * and add the item as the next value
 * Assumes this was initialised using avatar_packet_initialise_message()
 * Note for BIN16 data the data needs to include the length of the data in the first 2 bytes
 * (the length is the length of the string, so the total data length is string length + 2)
 * e.g. 0x05, 0x00, 'H', 'e', 'l', 'l', 'o'
 * @param packet
 */
void avatar_packet_add_data(AVATAR_PACKET* packet, uint16_t tag, AVATAR_ENCODING encoding, uint8_t* data, uint16_t data_len)
{
    // Increment the number of values in the data
    uint8_t num_values = avatar_packet_get_num_values(packet);
    num_values++;
    avatar_packet_set_num_values(packet, num_values);
    
    // Get the current length of the data
    AVATAR_PACKET_HEADER* header = &packet->header;
    uint16_t length = avatar_packet_header_get_length(header);
    
    // Find the current offset of the data in the packet data
    uint16_t offset = length - sizeof(AVATAR_PACKET_HEADER);
    
    // If there are existing values then a Tail will have been added - overwrite this and re-add at the end
    if (num_values > 1)
    {
        offset -= sizeof(uint16_t);
    }
    
    // Add the StateID/Tag
    memcpy(&packet->data[offset], &tag, sizeof(uint16_t));
    offset += sizeof(uint16_t);
    
    if (data != NULL)
    {
        // Set the encoding
        packet->data[offset] = (uint8_t)encoding;
        offset += sizeof(uint8_t);

        // Set the data
        memcpy(&packet->data[offset], data, data_len);
        offset += data_len;
    }
    
    // Write tail
    packet->data[offset] = 0x00;
    offset++;
    packet->data[offset] = 0x00;
    offset++;
    
    // Set the header length
    header->packet_length = (offset + sizeof(AVATAR_PACKET_HEADER));
}

/**-----------------------------------------------------------------------------
 * Add a subscribe item to a SUBSCRIBE packet - this will increment the number of values in the packet, 
 * and the size of the packet header (assuming it was initialised to 8)
 * and add the item as the next value
 * @param packet
 */
void avatar_packet_add_value(AVATAR_PACKET* packet, uint16_t value)
{
    // Increment the number of values in the data
    uint8_t num_values = avatar_packet_get_num_values(packet);
    num_values++;
    avatar_packet_set_num_values(packet, num_values);
    
    // Get the current length of the data
    AVATAR_PACKET_HEADER* header = &packet->header;
    uint16_t length = avatar_packet_header_get_length(header);
    
    // Find the current offset of the data in the packet data
    uint16_t offset = length - sizeof(AVATAR_PACKET_HEADER);
    
    // If there are existing values then a Tail will have been added - overwrite this and re-add at the end
    if (num_values > 1)
    {
        offset -= sizeof(uint16_t);
    }
    
    // Add the StateID/Tag
    memcpy(&packet->data[offset], &value, sizeof(uint16_t));
    offset += sizeof(uint16_t);
    
    // Write tail
    packet->data[offset] = 0x00;
    offset++;
    packet->data[offset] = 0x00;
    offset++;
    
    // Set the header length
    header->packet_length = (offset + sizeof(AVATAR_PACKET_HEADER));
}

/**-----------------------------------------------------------------------------
 * Get the message Id from the packet (first 2 bytes)
 * @param packet
 * @return 
 */
AVATAR_MESSAGE_ID avatar_packet_get_message_id(AVATAR_PACKET* packet)
{
    uint16_t result;

    memcpy(&result, packet->data, sizeof(uint16_t));
    
    return (AVATAR_MESSAGE_ID)result;
}

/**-----------------------------------------------------------------------------
 * Get the number of values in a packet (byte 3)
 * @param packet
 * @return 
 */
uint8_t avatar_packet_get_num_values(AVATAR_PACKET* packet)
{
    return packet->data[2];
}   

/**-----------------------------------------------------------------------------
 * Get the packet Id from the packet 
 * @param packet
 * @return 
 */
uint8_t avatar_packet_get_packet_id(AVATAR_PACKET* packet)
{
    return packet->header.packet_id;
}
    
/**-----------------------------------------------------------------------------
 * Get the values at the specified index in a UINT16 Value List packet.  
 * @param packet
 * @return 
 */
uint16_t avatar_packet_get_value(AVATAR_PACKET* packet, uint8_t index)
{
    uint16_t result;
    
    memcpy(&result, &packet->data[3 + (index * sizeof(uint16_t))], sizeof(uint16_t));
    
    return result;
}
     
/**-----------------------------------------------------------------------------
 * Returns the total length of the packets data (including tail)
 */
uint16_t avatar_packet_header_get_data_length(AVATAR_PACKET_HEADER* header)
{
    return avatar_packet_header_get_length(header) - sizeof(AVATAR_PACKET_HEADER);
}

/**-----------------------------------------------------------------------------
 * Returns the total length of the packet
 */
uint16_t avatar_packet_header_get_length(AVATAR_PACKET_HEADER* header)
{
    return header->packet_length;
}
    
/**-----------------------------------------------------------------------------
 * Gets the message type of the Avatar packet header
 * @param packet
 * @return 
 */
AVATAR_PACKET_TYPE avatar_packet_header_get_type(AVATAR_PACKET_HEADER* header)
{
    return (AVATAR_PACKET_TYPE) header->control_flags & 0x03;
}
     
/**-----------------------------------------------------------------------------
 * Initialises an empty packet header with start/tail stuff populated
 * @return 
 */
void avatar_packet_header_initialise(AVATAR_PACKET_HEADER* header)
{
    header->start_of_packet = AVATAR_PACKET_START;
    header->packet_length = 8;
    header->frame_id = 0;
    header->packet_id = 0;
    header->control_flags = 0x00;
    header->reserved = 0x00;
}
    
/**-----------------------------------------------------------------------------
 * Initialise a packet to contain data - sets the Message Id and the Count to 0
 * Sets the header length to 11 (header + message_id + count)
 * @param packet
 * @param message_id
 */
void avatar_packet_initialise_message(AVATAR_PACKET* packet, AVATAR_MESSAGE_ID message_id)
{
    avatar_packet_set_message_id(packet, message_id);
    avatar_packet_set_num_values(packet, 0);  
    
    packet->header.packet_length = sizeof(AVATAR_PACKET_HEADER) + sizeof(uint16_t) + sizeof(uint8_t);
}

/**-----------------------------------------------------------------------------
 * logs out a message containing information about the supplied packet
 * @param packet
 */
void avatar_packet_print(AVATAR_PACKET* packet, char* prefix)
{
    if ((logger_get_level() >= LOGGER_TYPE_DEBUG) && (packet->header.packet_id == 0))
    {
        uint16_t message_id;
        memcpy(&message_id, &packet->data[0], sizeof(uint16_t));

        uint8_t num_values = avatar_packet_get_num_values(packet);
        uint16_t offset = 3;

        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE, "AVATAR Packet %d %s [%s] (0x%04X) %d -- ", 
                packet->header.frame_id,
                prefix,
                avatar_packet_message_id_string(message_id),
                message_id,
                num_values);   

        int x;
        for (x = 0; x < num_values; x++)
        {
            // Process each value AVATAR_STATE_ID
            uint16_t state_id;
            memcpy(&state_id, packet->data + offset, sizeof(uint16_t));
            offset += sizeof(uint16_t);

            if ((message_id == AVATAR_MESSAGE_ID_SET_STATES) ||
                (message_id == AVATAR_MESSAGE_ID_STATE_CHANGE) ||
                (message_id == AVATAR_MESSAGE_ID_GET_STATES_RESPONSE))
            {
                uint8_t encoding = *(packet->data + offset);
                offset += sizeof(uint8_t);

                // Calculate the offset of the next entry
                switch((AVATAR_ENCODING)encoding)
                {
                    case AVATAR_ENCODING_UINT8:
                    {
                        uint8_t value;
                        memcpy(&value, packet->data + offset, sizeof(uint8_t));
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) UINT8 0x%02X --", avatar_state_id_string(state_id), state_id, value);
                        offset += sizeof(uint8_t);
                        break;
                    }
                    case AVATAR_ENCODING_UINT16:
                    {
                        uint16_t value;
                        memcpy(&value, packet->data + offset, sizeof(uint16_t));
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) UINT16 0x%02X --", avatar_state_id_string(state_id), state_id, value);
                        offset += sizeof(uint16_t);
                        break;
                    }
                    case AVATAR_ENCODING_UINT32:
                    {
                        uint32_t value;
                        memcpy(&value, packet->data + offset, sizeof(uint32_t));
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) UINT32 0x%04X --", avatar_state_id_string(state_id), state_id, value);
                        offset += sizeof(uint32_t);
                        break;
                    }
                    case AVATAR_ENCODING_UINT64:
                    {
                        uint64_t value;
                        memcpy(&value, packet->data + offset, sizeof(uint64_t));
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) UINT64 0x%08X --", avatar_state_id_string(state_id), state_id, value);
                        offset += sizeof(uint64_t);
                        break;
                    }
                    case AVATAR_ENCODING_INT8:
                    {
                        int8_t value;
                        memcpy(&value, packet->data + offset, sizeof(int8_t));
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) INT8 0x%02X --", avatar_state_id_string(state_id), state_id, value);
                        offset += sizeof(int8_t);
                        break;
                    }
                    case AVATAR_ENCODING_INT16:
                    {
                        int16_t value;
                        memcpy(&value, packet->data + offset, sizeof(int16_t));
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) INT16 0x%02X --", avatar_state_id_string(state_id), state_id, value);
                        offset += sizeof(int16_t);
                        break;
                    }
                    case AVATAR_ENCODING_INT32:
                    {
                        int32_t value;
                        memcpy(&value, packet->data + offset, sizeof(int32_t));
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) INT32 0x%04X --", avatar_state_id_string(state_id), state_id, value);
                        offset += sizeof(int32_t);
                        break;
                    }
                    case AVATAR_ENCODING_INT64:
                    {
                        int64_t value;
                        memcpy(&value, packet->data + offset, sizeof(int64_t));
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) INT64 0x%08X --", avatar_state_id_string(state_id), state_id, value);
                        offset += sizeof(int64_t);
                        break;
                    }        
                    case AVATAR_ENCODING_NIL:
                    {
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) NIL NIL --", avatar_state_id_string(state_id), state_id);
                        break;
                    }        
                    case AVATAR_ENCODING_BIN16:
                    {
                        uint16_t len;
                        memcpy(&len, packet->data + offset, sizeof(uint16_t));
                        char data[len + 1];
                        memset(data, 0, len + 1);
                        memcpy(data, packet->data + offset + sizeof(uint16_t), len);

                        if (len < 60)
                        {
                            logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) BIN16 %s --", avatar_state_id_string(state_id), state_id, data);
                        }
                        else
                        {
                            logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) BIN16 BIG(%d) --", avatar_state_id_string(state_id), state_id, len);
                        }
                        offset += len + sizeof(uint16_t);
                        break;
                    }        
                }
            }   
            else if ((message_id == AVATAR_MESSAGE_ID_TCPIP_TRANSFER) || (message_id == AVATAR_MESSAGE_ID_TCPIP_TRANSFER_RESPONSE))
            {
                // We are assuming only one State per message for TCP messages
                uint8_t encoding = *(packet->data + offset);
                offset += sizeof(uint8_t);

                switch (state_id)
                {
                    case AVATAR_STATE_ID_TCP_CLIENT_CONNECTION:
                    {
                        // Read a UINT16
                        uint8_t client_id = *(packet->data + offset);
                        offset += sizeof(uint8_t);

                        uint8_t connected = *(packet->data + offset); 
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [TCP Client Connection] (0x0000) TCP %d Client ID %d %s --", encoding, client_id, connected == 0x00 ? "Disconnected" : "Connected");
                        break;
                    }
                    case AVATAR_STATE_ID_TCP_MESSAGE:
                    {
                        uint16_t len;
                        memcpy(&len, packet->data + offset, sizeof(uint16_t));
                        offset += sizeof(uint16_t);

                        uint8_t client_id = *(packet->data + offset);
                        offset += sizeof(uint8_t);

                        char data[len]; // No need for +1 as that was the client Id
                        memset(data, 0, len);
                        memcpy(data, packet->data + offset, len);
                        offset += len;
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [TCP Message] (0x0001) TCP %d Client ID %d %s --", len, client_id, data);
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [TCP Message] 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X --", data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7] );
                        break;
                    }
                    case AVATAR_STATE_ID_TCP_CLIENT_CONTROL:
                    {
                        // Read a UINT16
                        uint8_t client_id = *(packet->data + offset);
                        offset += sizeof(uint8_t);

                        uint8_t connected = *(packet->data + offset); 
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [TCP Client Control] (0x0002) TCP Client ID %d %s --", client_id, connected == 0x00 ? "Disconnect" : "Undefined");
                        break;
                    }
                    default:
                    {
                        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [Unknown] (0x%04X) --", state_id);
                        break;
                    }
                }
            }
            else
            {
                logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE_DATA, "     [%s] (0x%04X) --", avatar_state_id_string(state_id), state_id);
            }
        }
    }
    else
    {
        logger_log_debugf(LOGGER_CODE_AVATAR_MESSAGE, "AVATAR Packet next packet ID %d -- ", 
                packet->header.packet_id);   
    }
}

/**-----------------------------------------------------------------------------
 * Returns true if a packet is ACK'd
 */
bool avatar_packet_header_is_ack(AVATAR_PACKET_HEADER* header)
{
    return (header->control_flags & (1 << AVATAR_PACKET_ACK_BIT - 1)) != 0;
}

/**-----------------------------------------------------------------------------
 * Returns true if a packet is NACK'd
 */
bool avatar_packet_header_is_nack(AVATAR_PACKET_HEADER* header)
{
    return (header->control_flags & (1 << AVATAR_PACKET_NACK_BIT - 1)) != 0;   
}
    
/**-----------------------------------------------------------------------------
 * Set the message Id from the packet (first 2 bytes)
 * @param packet
 * @return 
 */
void avatar_packet_set_message_id(AVATAR_PACKET* packet, AVATAR_MESSAGE_ID message_id)
{
    memcpy(&packet->data[0], &message_id, sizeof(uint16_t));
}

/* *****************************************************************************
 End of File
 */
