/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
   base64 Encode / Decode based on the code from https://nachtimwald.com/2017/11/18/base64-encode-and-decode-in-c/
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

/* This section lists the other files that are included in this file.
 */

#include "utilities/base64.h"
#include <string.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Encoding table
const char base64_chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

//! Decoding table
int base64_invs[] = { 62, -1, -1, -1, 63, 52, 53, 54, 55, 56, 57, 58,
                      59, 60, 61, -1, -1, -1, -1, -1, -1, -1, 0, 1, 2, 3, 4, 5,
                      6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
                      21, 22, 23, 24, 25, -1, -1, -1, -1, -1, -1, 26, 27, 28,
                      29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42,
                      43, 44, 45, 46, 47, 48, 49, 50, 51 };

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * This function is used to generate the base64_invs table above - for reference only
 */
static void base64_generate_decode_table()
{
	int    inv[80];
	size_t x;

	memset(inv, -1, sizeof(inv));
	for (x = 0; x < sizeof(base64_chars)-1; x++) 
    {
		inv[base64_chars[x]-43] = x;
	}
}

/**-----------------------------------------------------------------------------
 * 
 * @param c
 * @return 
 */
int base64_is_valid_char(char c)
{
	if (c >= '0' && c <= '9')
		return 1;
	if (c >= 'A' && c <= 'Z')
		return 1;
	if (c >= 'a' && c <= 'z')
		return 1;
	if (c == '+' || c == '/' || c == '=')
		return 1;
	return 0;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * 
 * @param in
 * @param out
 * @param outlen
 * @return 
 */
bool base64_decode(const char *in, unsigned char *out, size_t outlen)
{
    bool result = true;
	size_t len;
	size_t x;
	size_t y;
	int    v;

	if (in != NULL && out != NULL)
    {
        len = strlen(in);
        if (outlen < base64_decoded_size(in) || len % 4 != 0)
        {
            result = false;
        }

        for (x = 0; x < len; x++) 
        {
            if (!base64_is_valid_char(in[x])) 
            {
                result = false;
            }
        }

        if (result)
        {
            for (x = 0, y = 0; x < len; x += 4, y += 3) 
            {
                v = base64_invs[in[x]-43];
                v = (v << 6) | base64_invs[in[x + 1]-43];
                v = in[x + 2]=='=' ? v << 6 : (v << 6) | base64_invs[in[x + 2]-43];
                v = in[x + 3]=='=' ? v << 6 : (v << 6) | base64_invs[in[x + 3]-43];

                out[y] = (v >> 16) & 0xFF;

                // Sort out non-padded end characters
                if (in[x + 2] != '=')
                {
                    out[y + 1] = (v >> 8) & 0xFF;
                }

                if (in[x + 3] != '=')
                {
                    out[y + 2] = v & 0xFF;
                }
            }
        }
    }
    else
    {
        result = false;
    }

	return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param in
 * @return 
 */
size_t base64_decoded_size(const char *in)
{
	size_t len;
	size_t result = 0;
	size_t x;

	if (in != NULL)
    {
        len = strlen(in);
        result = (len / 4) * 3;

        for (x = len; x-- > 0; ) 
        {
            if (in[x] == '=') 
            {
                result--;
            } 
            else 
            {
                break;
            }
        }
    }

	return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param in
 * @param len
 * @param out - a buffer to hold the base 64 encoded output must be encoded_size + 1
 * @return 
 */
void base64_encode(const unsigned char* in, size_t len, char* out)
{
	size_t  encoded_len;
	size_t  x;
	size_t  y;
	size_t  v;

	if (in != NULL && len > 0)
    {
        encoded_len = base64_encoded_size(len);
        out[encoded_len] = '\0';

        for (x = 0, y = 0; x < len; x += 3, y += 4) 
        {
            v = in[x];
            v = x + 1 < len ? v << 8 | in[x + 1] : v << 8;
            v = x + 2 < len ? v << 8 | in[x + 2] : v << 8;

            out[y]   = base64_chars[(v >> 18) & 0x3F];
            out[y + 1] = base64_chars[(v >> 12) & 0x3F];
            
            // Add padding
            if (x + 1 < len) 
            {
                out[y + 2] = base64_chars[(v >> 6) & 0x3F];
            } 
            else 
            {
                out[y + 2] = '=';
            }

            if (x + 2 < len) 
            {
                out[y + 3] = base64_chars[v & 0x3F];
            } 
            else 
            {
                out[y + 3] = '=';
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Gets the size of the base 64 encoded string
 */
size_t base64_encoded_size(size_t inlen)
{
	size_t result;

	result = inlen;
	if (inlen % 3 != 0)
    {
		result += 3 - (inlen % 3);
    }
	result /= 3;
	result *= 4;

	return result;
}

// *****************************************************************************


/* *****************************************************************************
 End of File
 */
