/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    buttons.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/buttons.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct 
{
    bool            changed;
    int             debounce_counter;
    BUTTON_STATE    state;
    uint8_t         port;
    uint8_t         channel;
    uint8_t         pin;
    uint8_t         max_count;
} BUTTON_DATA;

typedef struct
{
    //! Determines if the change state need clearing on the next processing cycle
    bool clear_change_state;
    
    //! An array of buttons
    BUTTON_DATA button[BUTTONS_MAX];
} BUTTONS_DATA;

static BUTTONS_DATA buttons_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
   
/**-----------------------------------------------------------------------------
 * Defines a new button
 */
void buttons_add_button(uint8_t index, uint8_t port, uint8_t channel, uint8_t pin, uint8_t max_count)
{
    if (index < BUTTONS_MAX)
    {
        buttons_data.button[index].changed = false;
        buttons_data.button[index].state = BUTTON_STATE_OFF;
        buttons_data.button[index].debounce_counter = 0;
        buttons_data.button[index].port = port;
        buttons_data.button[index].channel = channel;
        buttons_data.button[index].pin = pin;
        buttons_data.button[index].max_count = max_count;
    }
}
    
/**
 * Returns true is the selected button/switch has been moved
 * @param index
 * @return 
 */
bool buttons_changed(uint8_t index)
{
    bool result = false;
    
    if (index < BUTTONS_MAX)
    {
        if (buttons_data.button[index].changed)
        {
            result = true;
        }
    }
    
    return result;
}
    
/**
 * Gets the state of the supplied button
 */
BUTTON_STATE buttons_get_state(uint8_t index)
{
    BUTTON_STATE result = BUTTON_STATE_UNDEFINED;
    
    if (index < BUTTONS_MAX)
    {
        result = buttons_data.button[index].state;
    }
    
    return result;
}
    
/**
 * Returns true is the selected button has been pushed in the current processing cycle
 * @param index
 * @return 
 */
bool buttons_get_pushed(uint8_t index)
{
    bool result = false;
    
    if (index < BUTTONS_MAX)
    {
        if ((buttons_data.button[index].changed) &&
            (buttons_data.button[index].state == BUTTON_STATE_ON))
        {
            result = true;
        }
    }
    
    return result;
}
    
/**
 * Returns true is the selected button is currently signalling as being down
 * @param index
 * @return 
 */
bool buttons_get_raw_value(uint8_t index)
{
    // Read the pin
    return PLIB_PORTS_PinGet(buttons_data.button[index].port, 
                             buttons_data.button[index].channel, 
                             buttons_data.button[index].pin);
}

/**
 * Returns true is the selected button has been released in the current processing cycle
 * @param index
 * @return 
 */
bool buttons_get_released(uint8_t index)
{
    bool result = false;
    
    if (index < BUTTONS_MAX)
    {
        if ((buttons_data.button[index].changed) &&
            (buttons_data.button[index].state == BUTTON_STATE_OFF))
        {
            result = true;
        }
    }
    
    return result;
}

/**
 * Initialises the buttons module
 * Note this should be called BEFORE buttons_add_button not after
 */
void buttons_initialise()
{
    buttons_data.clear_change_state = true;
    
    int x;
    for (x = 0; x < BUTTONS_MAX; x++)
    {
        buttons_data.button[x].state = BUTTON_STATE_UNDEFINED;
        buttons_data.button[x].debounce_counter = 0;
    }
}

/**-----------------------------------------------------------------------------
 * Processes buttons
 */
void buttons_process()
{
    int x;
    // Clear changed state
    if (buttons_data.clear_change_state)
    {
        for (x = 0; x < BUTTONS_MAX; x++)
        {
            if (buttons_data.button[x].state != BUTTON_STATE_UNDEFINED)
            {
                buttons_data.button[x].changed = false;
            }
        }
    }
            
    // Check buttons
    if (timers_has_expired(TIMER_INDEX_BUTTON_DEBOUNCE))
    {
        for (x = 0; x < BUTTONS_MAX; x++)
        {
            if (buttons_data.button[x].state != BUTTON_STATE_UNDEFINED)
            {
                // Read the pin
                bool state = buttons_get_raw_value(x);

                if (buttons_data.button[x].state != state)
                {
                    buttons_data.button[x].debounce_counter++;
                    if (buttons_data.button[x].debounce_counter > buttons_data.button[x].max_count)
                    {
                        // The button has changed state
                        buttons_data.button[x].state = !buttons_data.button[x].state;
                        buttons_data.button[x].debounce_counter = 0;

                        buttons_data.clear_change_state = true;
                        buttons_data.button[x].changed = true;
                        //logger_log_debugf(237, "BUTTONS buttons_process new state %d %d %d --", x, buttons_data.button[x].state, state); 
                    }
                }
                else
                {
                    buttons_data.button[x].debounce_counter = 0;
                }                
            }
        }
        
        timers_reset(TIMER_INDEX_BUTTON_DEBOUNCE);
    }
}

/**-----------------------------------------------------------------------------
 * Sets the sets of all buttons
 */
void buttons_reset()
{
    int x;
    for (x = 0; x < BUTTONS_MAX; x++)
    {
        buttons_data.button[x].state = buttons_get_raw_value(x);
        buttons_data.button[x].changed = false;
        //logger_log_debugf(237, "BUTTONS buttons_reset state %d %d --", x, buttons_data.button[x].state); 
    }
}

/* *****************************************************************************
 End of File
 */
