/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/dirac.h"
#include <stdio.h>
#include <string.h>
#include "utilities/avatar_control.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! The number of attempts we will make when loading the unit to get the DIRAC Curve names
//  As we cannot tell whether there are no curves defined or whether the avatar 
//  has not yet sent them, we put a limit on the number of times we will try 
//  requesting them, after that we assume there are no curves defined.
//  Normally we expect it to return after 2 or 3 attempts
#define DIRAC_MAX_CURVE_NAME_REQUEST    10
/**
 * Structure for holding information on DIRAC Curves
 */
typedef struct
{
    //! DIRAC EQ Cure information
    DIRAC_EQ_CURVE eq_curves[DIRAC_MAX_EQ_CURVES];

    //! The number of available DIRAC curves)
    //uint8_t num_curves;
    
    uint8_t num_curve_name_requests;
    
} DIRAC_INFO;

static DIRAC_INFO dirac_data = {
//  .num_curves = 1,  
  .num_curve_name_requests = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Gets whether the curve is enabled or not
 * @param index
 * @return 
 */
static bool dirac_is_curve_enabled(uint8_t index)
{
    bool result = true;
    if (index > 0)
    {
        uint8_t dirac_eq_curve_enabled_settings[DIRAC_MAX_EQ_CURVES];
        settings_get(SETTING_DIRAC_CURVE_ENABLED, dirac_eq_curve_enabled_settings);
        result = (bool)dirac_eq_curve_enabled_settings[index];
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Configures the module from the settings
 */
void dirac_configure()
{
    logger_log_debugf(0, "DIRAC Requesting Curve names --");
    avatar_request_dirac_curve_names();
}
    
/**-----------------------------------------------------------------------------
 * Returns the dirac curve at the given index
 * @param input
 * @return 
 */
DIRAC_EQ_CURVE* dirac_get_curve(uint8_t index)
{
    DIRAC_EQ_CURVE* result = NULL;
    
    if (index < DIRAC_MAX_EQ_CURVES)
    {
        result = &dirac_data.eq_curves[index];
    }
    
    return result;
}   
    
/**-----------------------------------------------------------------------------
 * Returns the dirac curve for the given AUDIO_INPUT
 * @param input
 * @return 
 */
uint8_t dirac_get_curve_for_input(uint8_t input)
{
    uint8_t result = DIRAC_NO_CURVE;

    if (input < INPUT_CONTROL_NUM_SETTINGS)
    {
        uint8_t dirac_eq_curve_settings[INPUT_CONTROL_NUM_SETTINGS];
        settings_get(SETTING_DIRAC_CURVE, dirac_eq_curve_settings);
        result = dirac_eq_curve_settings[input];
    }
    
    return result;
}   
    
/**-----------------------------------------------------------------------------
 * Gets the name associated with the supplied index, returns NULL if invalid
 * @param index
 * @return 
 */
char* dirac_get_curve_name(uint8_t index)
{
    char* result = NULL;
    //logger_log_debugf(0, "INPUT dirac_get_curve_name %d %s --", index, dirac_data.eq_curves[index].name[0]);
    if ((index < DIRAC_MAX_EQ_CURVES) && (dirac_data.eq_curves[index].name[0] != 0))
    {
        result = dirac_data.eq_curves[index].name;
    }
    return result;
}   
    
/**-----------------------------------------------------------------------------
 * Gets the info on the next curve after the selected curve 
 * @return 
 */
DIRAC_EQ_CURVE* dirac_get_next_curve(uint8_t current_curve)
{
    DIRAC_EQ_CURVE* result = NULL;
    uint8_t index = current_curve;//dirac_data.selected_curve;

    int x;
    for (x = 0; x < DIRAC_MAX_EQ_CURVES; x++)
    {
        // Move the index on
        index++;
        if (index >= DIRAC_MAX_EQ_CURVES)
        {
            index = 0;
        }
        
        // Check if this curve is in use
        //if (dirac_data.eq_curves[index].name[0] != 0)
        if (dirac_is_curve_enabled(index))
        {
            result = &dirac_data.eq_curves[index];
            break;
        }
    }

    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets maximum number of available DIRAC curves
 * @return 
 */
uint8_t dirac_get_num_curves()
{
    uint8_t result = 0;
    
    uint8_t dirac_eq_curve_enabled_settings[DIRAC_MAX_EQ_CURVES];
    settings_get(SETTING_DIRAC_CURVE_ENABLED, dirac_eq_curve_enabled_settings);

    int x;
    for (x = 0; x < DIRAC_MAX_EQ_CURVES; x++)
    {
        if (dirac_eq_curve_enabled_settings[x] != 0)
        {
            result++;
        }
    }
    
    return result;
}
       
/**-----------------------------------------------------------------------------
 * Gets the info on the next curve after the selected curve 
 * @return 
 */
DIRAC_EQ_CURVE* dirac_get_previous_curve(uint8_t current_curve)
{
    DIRAC_EQ_CURVE* result = NULL;
    int8_t index = (int8_t)current_curve;//dirac_data.selected_curve;
    
    int x;
    for (x = 0; x < DIRAC_MAX_EQ_CURVES; x++)
    {
        // Move the index on
        index--;
        if (index < 0)
        {
            index = DIRAC_MAX_EQ_CURVES - 1;
        }
        
        // Check if this curve is in use
        //if (dirac_data.eq_curves[index].name[0] != 0)
        if (dirac_is_curve_enabled(index))
        {
            result = &dirac_data.eq_curves[index];
            break;
        }
    }

    return result;
}
    
/**-----------------------------------------------------------------------------
 * Initialises the DIRAC module - should be called before any curves are added
 */
void dirac_initialise()
{
    // Set the enabled curves to false except for No Curve
    uint8_t dirac_eq_curve_enabled_settings[DIRAC_MAX_EQ_CURVES];
    memset(dirac_eq_curve_enabled_settings, 0, sizeof(uint8_t) * DIRAC_MAX_EQ_CURVES);
    dirac_eq_curve_enabled_settings[0] = true;  
    settings_set(SETTING_DIRAC_CURVE_ENABLED, dirac_eq_curve_enabled_settings);
    
    int x;
    for (x = 0; x < DIRAC_MAX_EQ_CURVES; x++)
    {
        dirac_data.eq_curves[x].index = x;
        
        // Set the name to a NULL string to indicate that its not assigned
        memset(dirac_data.eq_curves[x].name, 0, DIRAC_MAX_EQ_CURVE_NAME_LEN);

        logger_log_debugf(0, "DIRAC INIT CURVE %d --", x);
    }
    
    // Set up no curve name
    sprintf(dirac_data.eq_curves[DIRAC_NO_CURVE].name, "%s", "No Curve");
}
    
/**-----------------------------------------------------------------------------
 * Perform dirac processing (getting Curve names)
 */
void dirac_process()
{
    //logger_log_debugf(0, "DIRAC REQUESTING CURVE NAMES %d %d %d --", timers_has_expired(TIMER_INDEX_DIRAC_CURVE_NAME_CHECK), dirac_get_num_curves(), dirac_get_num_curves());

    if (timers_has_expired(TIMER_INDEX_DIRAC_CURVE_NAME_CHECK) && 
        (dirac_get_num_curves() == 1) &&
        dirac_data.num_curve_name_requests < DIRAC_MAX_CURVE_NAME_REQUEST)
    {
        logger_log_debugf(LOGGER_CODE_DIRAC, "DIRAC Requesting Curve Names --");
        //logger_log_debugf(0, "DIRAC REQUESTING CURVE NAMES --");
        avatar_request_dirac_curve_names();
        
        dirac_data.num_curve_name_requests++;
        
        if (dirac_data.num_curve_name_requests == DIRAC_MAX_CURVE_NAME_REQUEST)
        {
            logger_log_infof(LOGGER_CODE_DIRAC, "DIRAC Assuming there are no Dirac curves --");
            
            // Set input curve selections all to No Curve
            uint8_t dirac_eq_curve_settings[INPUT_CONTROL_NUM_SETTINGS];
            settings_get(SETTING_DIRAC_CURVE, dirac_eq_curve_settings);
            memset(dirac_eq_curve_settings, DIRAC_NO_CURVE, INPUT_CONTROL_NUM_SETTINGS);
            settings_set(SETTING_DIRAC_CURVE, dirac_eq_curve_settings);
        }
        else
        {
            timers_reset(TIMER_INDEX_DIRAC_CURVE_NAME_CHECK);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Sets the Dirac curve as uncalibrated, making it unavailable
 * @param index
 */
void dirac_remove_curve_calibration(uint8_t index, bool clear_usage)
{
    if (index < DIRAC_MAX_EQ_CURVES)
    {
        uint8_t dirac_eq_curve_enabled_settings[DIRAC_MAX_EQ_CURVES];
        settings_get(SETTING_DIRAC_CURVE_ENABLED, dirac_eq_curve_enabled_settings);

        logger_log_debugf(0, "DIRAC Removing Curve %d --", index);

        // Set the curve to enabled
        dirac_eq_curve_enabled_settings[index] = (uint8_t)false;
        settings_set(SETTING_DIRAC_CURVE_ENABLED, dirac_eq_curve_enabled_settings);

        if (clear_usage)
        {
            // If any inputs current use this curve then set them back to no curve
            uint8_t dirac_eq_curve_settings[INPUT_CONTROL_NUM_SETTINGS];
            settings_get(SETTING_DIRAC_CURVE, dirac_eq_curve_settings);
            int x;
            for (x = 0; x < INPUT_CONTROL_NUM_SETTINGS; x++)
            {
                if (dirac_eq_curve_settings[x] == index)
                {
                    dirac_eq_curve_settings[x] = DIRAC_NO_CURVE; // No curve
                }
            }

            settings_set(SETTING_DIRAC_CURVE, dirac_eq_curve_settings);
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Used to define whether the curve has been through a calibration process
 * @param index
 */
void dirac_set_curve_calibrated(uint8_t index, char* name)
{
    // Protect for empty names as sometimes, as sometimes it appears to return empty names
    if (strlen(name) > 0)
    {
        uint8_t dirac_eq_curve_enabled_settings[DIRAC_MAX_EQ_CURVES];
        settings_get(SETTING_DIRAC_CURVE_ENABLED, dirac_eq_curve_enabled_settings);

        logger_log_debugf(0, "DIRAC Adding Curve %s --", name);

        // Set the curve to enabled
        dirac_eq_curve_enabled_settings[index] = (uint8_t)true;

        // Set the Curve name
        dirac_set_curve_name(index, name);

        settings_set(SETTING_DIRAC_CURVE_ENABLED, dirac_eq_curve_enabled_settings);
    }
}   
   
/**-----------------------------------------------------------------------------
 * Sets the dirac curve for the given AUDIO_INPUT
 * @param input - The input index
 * @param index - The curve index
 * @return 
 */
uint8_t dirac_set_curve_for_input(uint8_t input, uint8_t index)
{
    uint8_t result = DIRAC_NO_CURVE;
    
    if (index < DIRAC_MAX_EQ_CURVES)
    {
        uint8_t dirac_eq_curve_settings[INPUT_CONTROL_NUM_SETTINGS];
        settings_get(SETTING_DIRAC_CURVE, dirac_eq_curve_settings);
        dirac_eq_curve_settings[input] = index;
        settings_set(SETTING_DIRAC_CURVE, dirac_eq_curve_settings);
        result = index;//true;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Sets the name associated with the supplied index
 * @param index
 * @return 
 */
void dirac_set_curve_name(uint8_t index, char* name)
{
    if (index < DIRAC_MAX_EQ_CURVES)
    {
        if (strlen(name) > 0)
        {
            memset(dirac_data.eq_curves[index].name, 0, DIRAC_MAX_EQ_CURVE_NAME_LEN);
            memcpy(dirac_data.eq_curves[index].name, name, min(strlen(name), DIRAC_MAX_EQ_CURVE_NAME_LEN));
            dirac_data.eq_curves[index].name[DIRAC_MAX_EQ_CURVE_NAME_LEN] = 0;
        }
        else
        {
            // Just incase the avatar dosn't give us a name
            sprintf(dirac_data.eq_curves[index].name, "EQ %d", index);
        }
    }
}   
    
/* *****************************************************************************
 End of File
 */
