/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    display.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/display.h"
#include <stdio.h>
#include <string.h>
#include "app_config.h"
#include "utilities/timers.h"

#if OLED_VFD_ENABLED
#include "oled_vfd_emulation.h"
#endif

#if VFD_ENABLED
#include "utilities/vfd.h"
#endif


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

// States for display scrolling
typedef enum
{
    DISPLAY_SCROLL_STATE_OFF,
    DISPLAY_SCROLL_STATE_PAUSE,
    DISPLAY_SCROLL_STATE_ROLLOVER,
} DISPLAY_SCROLL_STATE;

// Scrolling info
typedef struct
{
    DISPLAY_SCROLL_STATE scroll_state;
    
    DISPLAY_LINE display_line;
    
    //! The number of times it has scrolled
    uint8_t repeats;
    
    // The current line being displayed
    int scroll_current_line;
    
    // When the user chooses to move the line on , the new line becomes the focussed line.
    // 1-time scrolling will stop on this line
    int scroll_focus_line;
   
    // The current line being displayed
    int scroll_num_lines;
    
    //! The start position on the line
    int scroll_start_at;
    
    //! The 
    int scroll_width;
    
    //! The time in ms that it pauses on the current line before the next scroll
    uint16_t scroll_pause_ms;
    
    //! The time it takes for a rollover
    uint16_t scroll_rollover_ms;
    
    int scroll_rollover_offset;
    
    // The text to be displayed
    char text[DISPLAY_MAX_SCROLL_TEXT_LEN];
    
} DISPLAY_SCROLL_INFO;

typedef struct
{
    // The current brightness of the display
    DISPLAY_BRIGHTNESS brightness;
    
    // Buffer to hold the display lines
    char display_lines[DISPLAY_LINE_COUNT][DISPLAY_LINE_LEN];

    // Scrolling information
    DISPLAY_SCROLL_INFO scroll_info;
    
    // Handles alternative for UTF-8 characters.
    // UTF-8 special characters are 2byte.  The first byte is either C2 or C3 and
    // the second is in the range 80-BF
    // In this map the first 64 elements represent the C2 range and the second
    // the C3 range
    // The value in the element is the VFD character e.g 0x41 = 'A'
    // Refer to https://www.utf8-chartable.de/ for the UTF-08 character set
    char utf_8_character_map[128];
    char utf_8_gp_character_map[128]; // general punctuation

} DISPLAY_DATA;

static DISPLAY_DATA display_data =
{
  .scroll_info.scroll_focus_line = 0,  
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * If Scrolling is paused this causes the display to scroll a single line
*/    
static void display_rollover_line()
{
    display_data.scroll_info.scroll_rollover_offset = 0;
    int rollover_timeout = display_data.scroll_info.scroll_rollover_ms / display_get_character_height(); // Assuming vertical scrolling for the min
    timers_update_timeout(TIMER_INDEX_DISPLAY_SCROLL, rollover_timeout , TIMERS_TIMEOUT_RESET_IMMEDIATE);
    display_data.scroll_info.scroll_state = DISPLAY_SCROLL_STATE_ROLLOVER;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**
 * Clears the content of the line
 */
void display_append_text(DISPLAY_LINE line, int at, char* fmt, ...)
{
    char buffer[DISPLAY_LINE_LEN];
    memset(buffer, ' ', DISPLAY_LINE_LEN);
    
    va_list args;
    va_start(args, fmt);
    int len = vsprintf(buffer, fmt, args);
    va_end(args);
    
    strncpy(display_data.display_lines[line] + at, buffer, len);// strlen(buffer));
}
    
/**
 * Clears the content of the line
 */
void display_append_text_right(DISPLAY_LINE line, char* fmt, ...)
{
    char buffer[DISPLAY_LINE_LEN];
    memset(buffer, ' ', DISPLAY_LINE_LEN);
    
    va_list args;
    va_start(args, fmt);
    int len = vsprintf(buffer, fmt, args);
    va_end(args);
    
    //strncpy(display_data.display_lines[line] + DISPLAY_LINE_LEN - strlen(buffer), buffer, strlen(buffer));
    strncpy(display_data.display_lines[line] + DISPLAY_LINE_LEN - len, buffer, len);
}

/**
 * Clear the display
 *
 * @param immediate_refresh - Forces an immediate screen refresh
 */
void display_clear(bool immediate_refresh)
{
#if OLED_VFD_ENABLED
    oled_vfd_clear();
#endif
    
#if VFD_ENABLED
    vfd_clear();
#endif
    
    // Clear the buffer
    memset(display_data.display_lines, ' ', DISPLAY_LINE_COUNT * DISPLAY_LINE_LEN);
    
    if (immediate_refresh)
    {
        display_refresh();
    }
}
    
/**
 * Clears the content of the line
 */
void display_clear_line(DISPLAY_LINE line)
{
    //strncpy(display_data.display_lines[line], BLANK16, strlen(BLANK16));
    memset(display_data.display_lines[line], ' ', DISPLAY_LINE_LEN);
}
    
/**
 * Returns the current display brightness
 * @return 
 */
DISPLAY_BRIGHTNESS display_get_brightness()
{
    return display_data.brightness;
}
  
/**
 * Gets the height of a character in pixels
 * @return 
 */
size_t display_get_character_height()
{
    size_t result = 7;  // Default for VFD
#if OLED_VFD_ENABLED
    //TODO: Implement
#endif
    
#if VFD_ENABLED
    result = vfd_get_character_height();
#endif    
    
    return result;
}

/**
 * Gets the width of a character in pixels
 * @return 
 */
size_t display_get_character_width()
{
    size_t result = 5;  // Default for VFD
#if OLED_VFD_ENABLED
    //TODO: Implement
#endif
    
#if VFD_ENABLED
    result = vfd_get_character_width();
#endif    
    
    return result;    
}
   
/**
 * Copies the current line contents into a buffer
 * @param line - the line to copy
 * @param buffer - a pointer to a buffer to hold the output
 * @param size - the size of the output buffer
 * @return the number of bytes copied
 */
size_t display_get_line(DISPLAY_LINE line, char* buffer, size_t size)
{
    size_t result = min(size, DISPLAY_LINE_LEN);
    
    memcpy(buffer, display_data.display_lines[line], result);
    
#if VFD_ENABLED
    
    // Convert characters representing special character slot positions back into
    // the original special character.  This makes the assumption that the special 
    // characters are in the same ASCII range as the main characters and are
    // replacing the characters in the ASCII character set
    int x;
    for (x = 0; x < result; x++)
    {
        buffer[x] = vfd_get_special_character_in_slot((uint8_t)buffer[x]);
    }
#endif
            
    return result;
}
     
/**
 * Returns the length of the display in characters
 * @return 
 */
size_t display_get_line_length()
{
    size_t result = 0;
    
#if OLED_VFD_ENABLED
    result = oled_vfd_get_line_length();
#endif
    
#if VFD_ENABLED
    result = vfd_get_line_length();
#endif    
    
    return result;
}

/**
 * Gets a pointer to the scroll text used for display - MESS with this at your own peril!!!!
 * @return 
 */
char* display_get_scroll_text()
{
    return display_data.scroll_info.text;
}

/**
 * Allocates a special character and returns its character code
 */
char display_get_special_character(DISPLAY_LINE line, DISPLAY_SPECIAL_CHARACTERS character)
{
    char result = character;
#if OLED_VFD_ENABLED
    //TODO: To Implement
#endif
    
#if VFD_ENABLED
    result =  vfd_get_special_character(line, (char*)&character, 1, 1, 0, 0, 0);
#endif
    
    return result;
}

/**
 * Highlights the character at the specified ordinal position in the line
 * How the character is highlighted depends on the underlying display
 * @param line - the line to copy
 * @param at -  the ordinal position of the character in the line
 */
void display_highlight_character(DISPLAY_LINE line, int at, bool highlight_spaces)
{
#if OLED_VFD_ENABLED
    //TODO Implement
#endif
    
#if VFD_ENABLED
    char highlight = vfd_get_highlight_character(line, &display_data.display_lines[line][at], highlight_spaces);
    if (highlight != 0)
    {
        display_data.display_lines[line][at] = highlight;
    }
#endif
    
}
    
/**
 * Highlights the character at the specified ordinal position in the line relative to the right hand side
 * e.g. for a 16 char display an at value of 2 will highlight character 14
 * How the character is highlighted depends on the underlying display
 * @param line - the line to copy
 * @param at -  the ordinal position of the character in the line
 */
void display_highlight_character_right(DISPLAY_LINE line, int at, bool highlight_spaces)
{
    display_highlight_character(line, DISPLAY_LINE_LEN - at - 1, highlight_spaces);
}

//! Initialise the Display
/** 
 * Initialise the Display
 */
void display_initialise()
{
    // Initialise dodgy_character_map - done to display Motorhead properly
    memset(display_data.utf_8_character_map, ' ', 128);
    
    // C2 range index is value - 0x80 e.g C289 = 9
    // C3 range index is value - 0x80 + 0x40 e.g. C389 = 73
    // C2
    // No C2 matches
    
    // C3
    display_data.utf_8_character_map[0x80 - 0x80 + 0x40] = (uint8_t)'A';//0xC1; // A Grave
    display_data.utf_8_character_map[0x81 - 0x80 + 0x40] = (uint8_t)'A';//0xC0; // A Acute
    display_data.utf_8_character_map[0x82 - 0x80 + 0x40] = (uint8_t)'A';//0xD0; // A Circumflex
    display_data.utf_8_character_map[0x83 - 0x80 + 0x40] = (uint8_t)'A';//0xE0; // A Tilde
    display_data.utf_8_character_map[0x84 - 0x80 + 0x40] = (uint8_t)'A';//0xD1; // A Dieresis
    display_data.utf_8_character_map[0x85 - 0x80 + 0x40] = (uint8_t)'A';//0xE1; // A Ring
    display_data.utf_8_character_map[0x86 - 0x80 + 0x40] = 0xE2; // AE
    display_data.utf_8_character_map[0x87 - 0x80 + 0x40] = (uint8_t)'C';//0x8B; // C Cedila
    display_data.utf_8_character_map[0x88 - 0x80 + 0x40] = (uint8_t)'E';//0xC3; // E Grave
    display_data.utf_8_character_map[0x89 - 0x80 + 0x40] = (uint8_t)'E';//0xC4; // E Acute
    display_data.utf_8_character_map[0x8A - 0x80 + 0x40] = (uint8_t)'E';//0xD2; // E Circumflex
    display_data.utf_8_character_map[0x8B - 0x80 + 0x40] = (uint8_t)'E';//0xD3; // E Dieresis
    display_data.utf_8_character_map[0x8C - 0x80 + 0x40] = (uint8_t)'I';//0xC5; // I Grave
    display_data.utf_8_character_map[0x8D - 0x80 + 0x40] = (uint8_t)'I';//0xC4; // I Acute
    display_data.utf_8_character_map[0x8E - 0x80 + 0x40] = (uint8_t)'I';//0xD4; // I Circumflex
    display_data.utf_8_character_map[0x8F - 0x80 + 0x40] = (uint8_t)'I';//0xD5; // I Dieresis
    display_data.utf_8_character_map[0x90 - 0x80 + 0x40] = 0xCE; // ETH
    display_data.utf_8_character_map[0x91 - 0x80 + 0x40] = (uint8_t)'N';//0x??; // N Tilde
    display_data.utf_8_character_map[0x92 - 0x80 + 0x40] = (uint8_t)'O';//0xC6; // O Grave
    display_data.utf_8_character_map[0x93 - 0x80 + 0x40] = (uint8_t)'O';//0xC7; // O Acute
    display_data.utf_8_character_map[0x94 - 0x80 + 0x40] = (uint8_t)'O';//0xD6; // O Circumflex
    display_data.utf_8_character_map[0x95 - 0x80 + 0x40] = (uint8_t)'O';//0xE6; // O Tilde
    display_data.utf_8_character_map[0x96 - 0x80 + 0x40] = (uint8_t)'O';//0xD7; // O Dieresis
    display_data.utf_8_character_map[0x97 - 0x80 + 0x40] = (uint8_t)'x';// Multiplication
    display_data.utf_8_character_map[0x98 - 0x80 + 0x40] = (uint8_t)'O';//0xF7; // O Stroke
    display_data.utf_8_character_map[0x99 - 0x80 + 0x40] = (uint8_t)'U';//0xC9; // U Grave
    display_data.utf_8_character_map[0x9A - 0x80 + 0x40] = (uint8_t)'U';//0xC8; // U Acute
    display_data.utf_8_character_map[0x9B - 0x80 + 0x40] = (uint8_t)'U';//0xD8; // U Circumflex
    display_data.utf_8_character_map[0x9C - 0x80 + 0x40] = (uint8_t)'U';//0xD9; // U Dieresis
    display_data.utf_8_character_map[0x9D - 0x80 + 0x40] = (uint8_t)'Y';//0xE5; // Y Acute
    display_data.utf_8_character_map[0x9E - 0x80 + 0x40] = (uint8_t)'D';//0xF8; // Thorn
    display_data.utf_8_character_map[0x9F - 0x80 + 0x40] = (uint8_t)'B';//0x8D; // Sharp S
  
    display_data.utf_8_character_map[0xA0 - 0x80 + 0x40] = (uint8_t)'a';//0x81; // A Grave
    display_data.utf_8_character_map[0xA1 - 0x80 + 0x40] = (uint8_t)'a';//0x80; // A Acute
    display_data.utf_8_character_map[0xA2 - 0x80 + 0x40] = (uint8_t)'a';//0x80; // A Circumflex
    display_data.utf_8_character_map[0xA3 - 0x80 + 0x40] = (uint8_t)'a';//0xF0; // A Tilde
    display_data.utf_8_character_map[0xA4 - 0x80 + 0x40] = (uint8_t)'a';//0x91; // A Dieresis
    display_data.utf_8_character_map[0xA5 - 0x80 + 0x40] = (uint8_t)'a';//0xF1; // A Ring
    display_data.utf_8_character_map[0xA6 - 0x80 + 0x40] = 0xE2; // ae
    display_data.utf_8_character_map[0xA7 - 0x80 + 0x40] = (uint8_t)'c';//0x9B; // C Cedila
    display_data.utf_8_character_map[0xA8 - 0x80 + 0x40] = (uint8_t)'e';//0x83; // E Grave
    display_data.utf_8_character_map[0xA9 - 0x80 + 0x40] = (uint8_t)'e';//0x82; // E Acute
    display_data.utf_8_character_map[0xAA - 0x80 + 0x40] = (uint8_t)'e';//0x92; // E Circumflex
    display_data.utf_8_character_map[0xAB - 0x80 + 0x40] = (uint8_t)'e';//0x93; // E Dieresis
    display_data.utf_8_character_map[0xAC - 0x80 + 0x40] = (uint8_t)'i';//0x85; // I Grave
    display_data.utf_8_character_map[0xAD - 0x80 + 0x40] = (uint8_t)'i';//0x84; // I Acute
    display_data.utf_8_character_map[0xAE - 0x80 + 0x40] = (uint8_t)'i';//0x94; // I Circumflex
    display_data.utf_8_character_map[0xAF - 0x80 + 0x40] = (uint8_t)'i';//0x95; // I Dieresis
    display_data.utf_8_character_map[0xB0 - 0x80 + 0x40] = 0xCE; // ETH
    display_data.utf_8_character_map[0xB1 - 0x80 + 0x40] = (uint8_t)'n';//0x??; // N Tilde
    display_data.utf_8_character_map[0xB2 - 0x80 + 0x40] = (uint8_t)'o';//0x87; // O Grave
    display_data.utf_8_character_map[0xB3 - 0x80 + 0x40] = (uint8_t)'o';//0x86; // O Acute
    display_data.utf_8_character_map[0xB4 - 0x80 + 0x40] = (uint8_t)'o';//0x96; // O Circumflex
    display_data.utf_8_character_map[0xB5 - 0x80 + 0x40] = (uint8_t)'o';//0xA7; // O Tilde
    display_data.utf_8_character_map[0xB6 - 0x80 + 0x40] = 0xD7; // O Dieresis
    display_data.utf_8_character_map[0xB7 - 0x80 + 0x40] = (uint8_t)'/';//0xBA;// Division
    display_data.utf_8_character_map[0xB8 - 0x80 + 0x40] = (uint8_t)'o';//0xF7; // O Stroke
    display_data.utf_8_character_map[0xB9 - 0x80 + 0x40] = (uint8_t)'u';//0x89; // U Grave
    display_data.utf_8_character_map[0xBA - 0x80 + 0x40] = (uint8_t)'u';//0x88; // U Acute
    display_data.utf_8_character_map[0xBB - 0x80 + 0x40] = (uint8_t)'u';//0x98; // U Circumflex
    display_data.utf_8_character_map[0xBC - 0x80 + 0x40] = (uint8_t)'u';//0x99; // U Dieresis
    display_data.utf_8_character_map[0xBD - 0x80 + 0x40] = (uint8_t)'y';//0xF5; // Y Acute
    display_data.utf_8_character_map[0xBE - 0x80 + 0x40] = (uint8_t)'D';//0xF8; // Thorn
    display_data.utf_8_character_map[0xBF - 0x80 + 0x40] = (uint8_t)'y';//0x??; // y Dieresis
    
    
    // for E2 series; 3 characters e.g. E2 80 99 is apostrophe
    memset(display_data.utf_8_gp_character_map, ' ', 128);
    
    // we detect E2 and decode 2nd and 3rd characters as ((<2nd> - 80) * 64) + (<3rd> - 80)
    
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x80 - 0x80)] = (uint8_t)' '; // 0x20; // EN QUAD
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x81 - 0x80)] = (uint8_t)' '; // 0x20; // EM QUAD
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x82 - 0x80)] = (uint8_t)' '; // 0x20; // EN SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x83 - 0x80)] = (uint8_t)' '; // 0x20; // EM SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x84 - 0x80)] = (uint8_t)' '; // 0x20; // THREE-PER-EM SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x85 - 0x80)] = (uint8_t)' '; // 0x20; // FOUR-PER-EM SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x86 - 0x80)] = (uint8_t)' '; // 0x20; // SIX-PER-EM SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x87 - 0x80)] = (uint8_t)' '; // 0x20; // FIGURE SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x88 - 0x80)] = (uint8_t)' '; // 0x20; // PUNCTUATION SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x89 - 0x80)] = (uint8_t)' '; // 0x20; // THIN SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x8a - 0x80)] = (uint8_t)' '; // 0x20; // HAIR SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x8b - 0x80)] = (uint8_t)' '; // 0x20; // ZERO WIDTH SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x8c - 0x80)] = (uint8_t)' '; // 0x20; // ZERO WIDTH NON-JOINER
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x8d - 0x80)] = (uint8_t)' '; // 0x20; // ZERO WIDTH JOINER
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x8e - 0x80)] = (uint8_t)' '; // 0x20; // LEFT-TO-RIGHT MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x8f - 0x80)] = (uint8_t)' '; // 0x20; // RIGHT-TO-LEFT MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x90 - 0x80)] = (uint8_t)'-'; // 0x2d; // HYPHEN
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x91 - 0x80)] = (uint8_t)'-'; // 0x2d; // NON-BREAKING HYPHEN
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x92 - 0x80)] = (uint8_t)'-'; // 0x2d; // FIGURE DASH
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x93 - 0x80)] = (uint8_t)'-'; // 0x2d; // EN DASH
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x94 - 0x80)] = (uint8_t)'-'; // 0x2d; // EM DASH
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x95 - 0x80)] = (uint8_t)'-'; // 0x2d; // HORIZONTAL BAR
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x96 - 0x80)] = 0x60; //  DOUBLE VERTICAL LINE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x97 - 0x80)] = (uint8_t)'='; // 0x3d; // DOUBLE LOW LINE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x98 - 0x80)] = 0x27;  // LEFT SINGLE QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x99 - 0x80)] = 0x27;  // RIGHT SINGLE QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x9a - 0x80)] = 0x27;  // SINGLE LOW-9 QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x9b - 0x80)] = 0x27;  // SINGLE HIGH-REVERSED-9 QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x9c - 0x80)] = 0x12;  // LEFT DOUBLE QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x9d - 0x80)] = 0x13;  // RIGHT DOUBLE QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x9e - 0x80)] = (uint8_t)' '; // 0x20; // DOUBLE LOW-9 QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0x9f - 0x80)] = 0x12;  // DOUBLE HIGH-REVERSED-9 QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa0 - 0x80)] = (uint8_t)' '; // 0x20; // DAGGER
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa1 - 0x80)] = (uint8_t)' '; // 0x20; // DOUBLE DAGGER
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa2 - 0x80)] = (uint8_t)' '; // 0x20; // BULLET
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa3 - 0x80)] = (uint8_t)' '; // 0x20; // TRIANGULAR BULLET
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa4 - 0x80)] = (uint8_t)' '; // 0x20; // ONE DOT LEADER
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa5 - 0x80)] = (uint8_t)' '; // 0x20; // TWO DOT LEADER
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa6 - 0x80)] = (uint8_t)' '; // 0x20; // HORIZONTAL ELLIPSIS
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa7 - 0x80)] = (uint8_t)' '; // 0x20; // HYPHENATION POINT
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa8 - 0x80)] = (uint8_t)' '; // 0x20; // LINE SEPARATOR
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xa9 - 0x80)] = (uint8_t)' '; // 0x20; // PARAGRAPH SEPARATOR
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xaa - 0x80)] = (uint8_t)' '; // 0x20; // LEFT-TO-RIGHT EMBEDDING
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xab - 0x80)] = (uint8_t)' '; // 0x20; // RIGHT-TO-LEFT EMBEDDING
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xac - 0x80)] = (uint8_t)' '; // 0x20; // POP DIRECTIONAL FORMATTING
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xad - 0x80)] = (uint8_t)' '; // 0x20; // LEFT-TO-RIGHT OVERRIDE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xae - 0x80)] = (uint8_t)' '; // 0x20; // RIGHT-TO-LEFT OVERRIDE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xaf - 0x80)] = (uint8_t)' '; // 0x20; // NARROW NO-BREAK SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb0 - 0x80)] = (uint8_t)' '; // 0x20; // PER MILLE SIGN
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb1 - 0x80)] = (uint8_t)' '; // 0x20; // PER TEN THOUSAND SIGN
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb2 - 0x80)] = 0x27; // PRIME
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb3 - 0x80)] = 0x22;  // DOUBLE PRIME
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb4 - 0x80)] = (uint8_t)' '; // 0x20d; // TRIPLE PRIME
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb5 - 0x80)] = 0x27; // REVERSED PRIME
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb6 - 0x80)] = 0x22;  // REVERSED DOUBLE PRIME
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb7 - 0x80)] = (uint8_t)' '; // 0x20d; // REVERSED TRIPLE PRIME
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb8 - 0x80)] = (uint8_t)' '; // 0x20; // CARET
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xb9 - 0x80)] = 0x3c; // SINGLE LEFT-POINTING ANGLE QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xba - 0x80)] = 0x3e; // SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xbb - 0x80)] = (uint8_t)' '; // 0x20; // REFERENCE MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xbc - 0x80)] = (uint8_t)'!'; // 0x21; // DOUBLE EXCLAMATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xbd - 0x80)] = (uint8_t)' '; // 0x20; // INTERROBANG
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xbe - 0x80)] = 0x7e; // OVERLINE
    display_data.utf_8_gp_character_map[((0x80 - 0x80) * 64) + (0xbf - 0x80)] = (uint8_t)' '; // 0x20; // UNDERTIE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x80 - 0x80)] = (uint8_t)' '; // 0x20; // CHARACTER TIE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x81 - 0x80)] = (uint8_t)' '; // 0x20; // CARET INSERTION POINT
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x82 - 0x80)] = (uint8_t)' '; // 0x20; // ASTERISM
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x83 - 0x80)] = (uint8_t)' '; // 0x20; // HYPHEN BULLET
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x84 - 0x80)] = (uint8_t)'/'; // 0x2d; // FRACTION SLASH
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x85 - 0x80)] = (uint8_t)'['; // 0x5b; // LEFT SQUARE BRACKET WITH QUILL
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x86 - 0x80)] = (uint8_t)']'; // 0x5d; // RIGHT SQUARE BRACKET WITH QUILL
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x87 - 0x80)] = (uint8_t)'!'; // 0x21; // DOUBLE QUESTION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x88 - 0x80)] = (uint8_t)' '; // 0x20; // QUESTION EXCLAMATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x89 - 0x80)] = (uint8_t)' '; // 0x20; // EXCLAMATION QUESTION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x8a - 0x80)] = (uint8_t)'7'; // 0x37; // TIRONIAN SIGN ET
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x8b - 0x80)] = (uint8_t)' '; // 0x20; // REVERSED PILCROW SIGN
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x8c - 0x80)] = (uint8_t)' '; // 0x20; // BLACK LEFTWARDS BULLET
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x8d - 0x80)] = (uint8_t)' '; // 0x20; // BLACK RIGHTWARDS BULLET
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x8e - 0x80)] = (uint8_t)' '; // 0x20; // LOW ASTERISK
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x8f - 0x80)] = (uint8_t)' '; // 0x20; // REVERSED SEMICOLON
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x90 - 0x80)] = (uint8_t)' '; // 0x20; // CLOSE UP
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x91 - 0x80)] = (uint8_t)' '; // 0x20; // TWO ASTERISKS ALIGNED VERTICALLY
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x92 - 0x80)] = (uint8_t)' '; // 0x20; // COMMERCIAL MINUS SIGN
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x93 - 0x80)] = (uint8_t)' '; // 0x20; // SWUNG DASH
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x94 - 0x80)] = (uint8_t)' '; // 0x20; // INVERTED UNDERTIE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x95 - 0x80)] = (uint8_t)' '; // 0x20; // FLOWER PUNCTUATION MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x96 - 0x80)] = (uint8_t)' '; // 0x20; // THREE DOT PUNCTUATION
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x97 - 0x80)] = (uint8_t)' '; // 0x20; // QUADRUPLE PRIME
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x98 - 0x80)] = (uint8_t)' '; // 0x20; // FOUR DOT PUNCTUATION
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x99 - 0x80)] = (uint8_t)' '; // 0x20; // FIVE DOT PUNCTUATION
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x9a - 0x80)] = (uint8_t)' '; // 0x20; // TWO DOT PUNCTUATION
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x9b - 0x80)] = (uint8_t)' '; // 0x20; // FOUR DOT MARK
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x9c - 0x80)] = (uint8_t)' '; // 0x20; // DOTTED CROSS
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x9d - 0x80)] = (uint8_t)' '; // 0x20; // TRICOLON
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x9e - 0x80)] = (uint8_t)' '; // 0x20; // VERTICAL FOUR DOTS
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0x9f - 0x80)] = (uint8_t)' '; // 0x20; // MEDIUM MATHEMATICAL SPACE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa0 - 0x80)] = (uint8_t)' '; // 0x20; // WORD JOINER
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa1 - 0x80)] = (uint8_t)' '; // 0x20; // FUNCTION APPLICATION
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa2 - 0x80)] = (uint8_t)' '; // 0x20; // INVISIBLE TIMES
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa3 - 0x80)] = (uint8_t)' '; // 0x20; // INVISIBLE SEPARATOR
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa4 - 0x80)] = (uint8_t)' '; // 0x20; // INVISIBLE PLUS
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa5 - 0x80)] = (uint8_t)' '; // 0x20; // 
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa6 - 0x80)] = (uint8_t)' '; // 0x20; // LEFT-TO-RIGHT ISOLATE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa7 - 0x80)] = (uint8_t)' '; // 0x20; // RIGHT-TO-LEFT ISOLATE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa8 - 0x80)] = (uint8_t)' '; // 0x20; // FIRST STRONG ISOLATE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xa9 - 0x80)] = (uint8_t)' '; // 0x20; // POP DIRECTIONAL ISOLATE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xaa - 0x80)] = (uint8_t)' '; // 0x20; // INHIBIT SYMMETRIC SWAPPING
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xab - 0x80)] = (uint8_t)' '; // 0x20; // ACTIVATE SYMMETRIC SWAPPING
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xac - 0x80)] = (uint8_t)' '; // 0x20; // INHIBIT ARABIC FORM SHAPING
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xad - 0x80)] = (uint8_t)' '; // 0x20; // ACTIVATE ARABIC FORM SHAPING
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xae - 0x80)] = (uint8_t)' '; // 0x20; // NATIONAL DIGIT SHAPES
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xaf - 0x80)] = (uint8_t)' '; // 0x20; // NOMINAL DIGIT SHAPES
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb0 - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT ZERO
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb1 - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT LATIN SMALL LETTER I
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb2 - 0x80)] = (uint8_t)' '; // 0x20; // 
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb3 - 0x80)] = (uint8_t)' '; // 0x20; // 
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb4 - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT FOUR
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb5 - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT FIVE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb6 - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT SIX
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb7 - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT SEVEN
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb8 - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT EIGHT
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xb9 - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT NINE
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xba - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT PLUS SIGN
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xbb - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT MINUS
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xbc - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT EQUALS SIGN
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xbd - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT LEFT PARENTHESIS
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xbe - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT RIGHT PARENTHESIS
    display_data.utf_8_gp_character_map[((0x80 - 0x81) * 64) + (0xbf - 0x80)] = (uint8_t)' '; // 0x20; // SUPERSCRIPT LATIN SMALL LETTER N

     
#if OLED_VFD_ENABLED
    oled_vfd_initialise();
#endif
    
#if VFD_ENABLED
    vfd_initialise();
#endif
    
    display_scroll_cancel();
    
    display_clear(true);
    
    display_set_brightness(DISPLAY_BRIGHTNESS_FULL);
}
      
/**
 * Refreshes the screen with the current pixel buffer
 */
void display_refresh()
{
#if OLED_VFD_ENABLED
    oled_vfd_refresh();
#endif
    
#if VFD_ENABLED
    vfd_refresh();
#endif
    
    //vfd_signal_test();
    //vfd_set_all_lights(0x03);
}
    
/**
 * Used to reset the display brightness even if it hasn't changed
 */
void display_reset_brightness()
{
    switch(display_data.brightness)
    {
        case DISPLAY_BRIGHTNESS_OFF:
        {
#if OLED_VFD_ENABLED
            oled_vfd_set_brightness(OLED_VFD_BRIGHTNESS_OFF);
#endif                
                
#if VFD_ENABLED
            vfd_set_brightness(DISPLAY_BRIGHTNESS_OFF);
#endif
            break;
        }
        case DISPLAY_BRIGHTNESS_DIMMED:
        {
#if OLED_VFD_ENABLED
            oled_vfd_set_brightness(OLED_VFD_BRIGHTNESS_DIMMED);
#endif                
                
#if VFD_ENABLED
            vfd_set_brightness(DISPLAY_BRIGHTNESS_DIMMED);
#endif
            break;
        }
        case DISPLAY_BRIGHTNESS_FULL:
        {
#if OLED_VFD_ENABLED
            oled_vfd_set_brightness(OLED_VFD_BRIGHTNESS_FULL);
#endif                
                
#if VFD_ENABLED
            vfd_set_brightness(DISPLAY_BRIGHTNESS_FULL);
#endif
            break;
        }
    }
 }

/**-----------------------------------------------------------------------------
 * Cancel any scrolling
 */    
void display_scroll_cancel()
{
    display_data.scroll_info.scroll_state = DISPLAY_SCROLL_STATE_OFF;
}
    
/**-----------------------------------------------------------------------------
 * If Scrolling is paused this causes the display to scroll a single line
*/    
void display_scroll_line()
{
    if (display_data.scroll_info.scroll_state == DISPLAY_SCROLL_STATE_PAUSE)
    {
        display_data.scroll_info.scroll_focus_line++;
        if (display_data.scroll_info.scroll_focus_line >= display_data.scroll_info.scroll_num_lines)
        {
            display_data.scroll_info.scroll_focus_line = 0;
        }
        display_rollover_line();
    }
}

/**-----------------------------------------------------------------------------
 * Process any scrolling actions
 */
void display_scroll_process()
{
    switch(display_data.scroll_info.scroll_state)
    {
        case DISPLAY_SCROLL_STATE_OFF:
        {
            // Either display a blank line or the last line that was paused on?
            break;
        }

        case DISPLAY_SCROLL_STATE_PAUSE:
        {
            if (timers_has_expired(TIMER_INDEX_DISPLAY_SCROLL))
            {
                //logger_log_debugf(89, "DISPLAY_SCROLL_STATE_PAUSE -1");
                // Move to the rollover state
                // Were going to scroll the current line up
                display_rollover_line();
                
                // This line means that the scrolling is only done once, then it
                // stops on the first line.  If we don't need this then we can remove 
                // display_data.scroll_info.repeats
                if (display_data.scroll_info.repeats > 0)
                {
                    // Scroll to the focus line
                    if (display_data.scroll_info.scroll_current_line == display_data.scroll_info.scroll_focus_line)
                    {
                        display_data.scroll_info.scroll_state = DISPLAY_SCROLL_STATE_PAUSE;
                    }
                }
            }
            else
            {
//                logger_log_debugf(89, "DISPLAY_SCROLL_STATE_PAUSE - Display %d %d %d %d ", display_data.scroll_info.scroll_start_at,
//                        display_data.scroll_info.scroll_num_lines, display_data.scroll_info.scroll_width,
//                        (display_data.scroll_info.scroll_current_line * display_data.scroll_info.scroll_width));
//                logger_log_debugf(89, "DISPLAY_SCROLL_STATE_PAUSE - %d %d ", 
//                        (display_data.scroll_info.scroll_current_line * display_data.scroll_info.scroll_width),
//                        (uint8_t)(display_data.scroll_info.scroll_current_line * display_data.scroll_info.scroll_width));
                
                // display the current line
                int x;
                for (x = 0; x < display_data.scroll_info.scroll_width; x++)
                {
                    display_special_character_offset(display_data.scroll_info.display_line, 
                                                     display_data.scroll_info.scroll_start_at + x, 
                                                     display_data.scroll_info.text,
                                                     display_data.scroll_info.scroll_num_lines,
                                                     display_data.scroll_info.scroll_width,
                                                     (display_data.scroll_info.scroll_current_line * display_data.scroll_info.scroll_width) + x,
                                                     0,
                                                     0);
                }
                
                display_write_line(display_data.scroll_info.display_line, false);

            }
            break;
        }

        case DISPLAY_SCROLL_STATE_ROLLOVER:
        {
            if (display_data.scroll_info.scroll_num_lines > 1)
            {
                if (timers_has_expired(TIMER_INDEX_DISPLAY_SCROLL))
                {
                    display_data.scroll_info.scroll_rollover_offset--;

    //                logger_log_debugf(89, "DISPLAY_SCROLL_STATE_ROLLOVER - %d %d ", display_data.scroll_info.scroll_rollover_offset,
    //                        (display_get_character_height() * -1));
    //                logger_log_debugf(89, "DISPLAY_SCROLL_STATE_ROLLOVER - %d %d ", 
    //                        (display_data.scroll_info.scroll_current_line * display_data.scroll_info.scroll_width),
    //                        (uint8_t)(display_data.scroll_info.scroll_current_line * display_data.scroll_info.scroll_width));

                    if (display_data.scroll_info.scroll_rollover_offset > (display_get_character_height() * -1))
                    {
                        // If we haven't reached the end of the rollover the update the offset and redisplay
                        // draw the current line with an offset
                        // display the current line
                        int x;
                        for (x = 0; x < display_data.scroll_info.scroll_width; x++)
                        {
                            display_special_character_offset(display_data.scroll_info.display_line, 
                                                             display_data.scroll_info.scroll_start_at + x, 
                                                             display_data.scroll_info.text,
                                                             display_data.scroll_info.scroll_num_lines,
                                                             display_data.scroll_info.scroll_width,
                                                             (display_data.scroll_info.scroll_current_line * display_data.scroll_info.scroll_width) + x,
                                                             0,
                                                             display_data.scroll_info.scroll_rollover_offset);
                        }

                        display_write_line(display_data.scroll_info.display_line, false);

                        timers_reset(TIMER_INDEX_DISPLAY_SCROLL);
                    }
                    else
                    {
                        // If we have the move back to the pause state
                        //int curr_line = display_data.scroll_info.scroll_current_line;

                        // Move the line number on
                        display_data.scroll_info.scroll_current_line++;
                        if (display_data.scroll_info.scroll_current_line >= display_data.scroll_info.scroll_num_lines)
                        {
                            display_data.scroll_info.scroll_current_line = 0;
                            display_data.scroll_info.repeats++;
                        }
    //                    else
    //                    {
    //                        timers_update_timeout(TIMER_INDEX_DISPLAY_SCROLL, display_data.scroll_info.scroll_pause_ms , TIMERS_TIMEOUT_RESET_IMMEDIATE);
    //                        display_data.scroll_info.scroll_state = DISPLAY_SCROLL_STATE_PAUSE;
    //                    }
                        //logger_log_debugf(89, "ROLLOVER END %d %d %d", curr_line, display_data.scroll_info.scroll_current_line, display_data.scroll_info.scroll_num_lines);


                        // Move to pause state
                        timers_update_timeout(TIMER_INDEX_DISPLAY_SCROLL, display_data.scroll_info.scroll_pause_ms , TIMERS_TIMEOUT_RESET_IMMEDIATE);
                        display_data.scroll_info.scroll_state = DISPLAY_SCROLL_STATE_PAUSE;
                    }
                }
            }
            break;
        }
    }
}

/**
 * Sets up the text block for scrolling and starts the timer
 * @param line
 * @param start_at
 * @param width
 * @param text_block
 * @param pause_ms
 * @param rollover_ms
 * @param split_words
 */
void display_scroll_text_block(DISPLAY_LINE line, int start_at, int width, char* text_block, uint16_t pause_ms, uint16_t rollover_ms, bool split_words, uint8_t scroll_focus_line)
{
    if (text_block != NULL)
    {
        display_data.scroll_info.scroll_focus_line = scroll_focus_line;
        
        display_data.scroll_info.display_line = line;
        display_data.scroll_info.scroll_start_at = start_at;
        display_data.scroll_info.scroll_current_line = display_data.scroll_info.scroll_focus_line;
        display_data.scroll_info.scroll_width = width;
        display_data.scroll_info.scroll_pause_ms = pause_ms;
        display_data.scroll_info.scroll_rollover_ms = rollover_ms;
        display_data.scroll_info.scroll_state = DISPLAY_SCROLL_STATE_PAUSE;
        display_data.scroll_info.scroll_rollover_offset = 0;
        display_data.scroll_info.repeats = 0;
        
        // Split the text block into lines
        memset(display_data.scroll_info.text, ' ', DISPLAY_MAX_SCROLL_TEXT_LEN);
        
        uint16_t text_len = min(strlen(text_block), DISPLAY_MAX_SCROLL_TEXT_LEN - 1);
        if (!split_words)
        {
            // Assumes no special characters
            text_len = min(strlen(text_block), DISPLAY_MAX_SCROLL_TEXT_LEN - 1);
            
            // Just copy the text block into our block
            memcpy(display_data.scroll_info.text, text_block, text_len);
            
            // Add .. to end if text too long
            if (strlen(text_block) > DISPLAY_MAX_SCROLL_TEXT_LEN - 1)
            {
                display_data.scroll_info.text[DISPLAY_MAX_SCROLL_TEXT_LEN - 2] = '.';
                display_data.scroll_info.text[DISPLAY_MAX_SCROLL_TEXT_LEN - 3] = '.';
            }
        }
        else
        {
            // Split the line
            text_len = 0;//min(strlen(text_block), DISPLAY_MAX_SCROLL_TEXT_LEN - 1);
            
            //int character_idx = 0;    // position of the character in the text;
            int word_start_idx = -1;    // Index in text_block of the start char of the current word
            int line_idx = 0;           // The position that text will be written to on the current line    
            int line_len = width;       // The length of a single line of scrollable text
            bool force_line_break = false; // Set to true to force a line break
            bool write_more_dots = false;   // True is the text is too long for the buffer
            //bool utf_8_skip_character = false;
            int utf_8_skip_character = 0;
            
            int x;
            int write_pos = 0;
            for (x = 0; x < strlen(text_block); x++)
            {
                if (utf_8_skip_character <= 0)
                {
                    char character = text_block[x];
                    
                    text_block[write_pos] = character;

                    // Am I at a space
                    if (character == ' ')
                    {
                        // Yes:
                        // Is there a word on the go at the moment (word_start_idx != -1)
                        if (word_start_idx != -1)
                        {
                            // Can I write the word into the space available on the line
                            int word_len = write_pos - word_start_idx;

                            // Check we're not blowing the buffer
                            if (text_len + word_len >=  DISPLAY_MAX_SCROLL_TEXT_LEN)
                            {
                                write_more_dots = true;
                                break;
                            }

                            //logger_log_debugf(89, "WRITING WORD %d %d %d %d --", word_start_idx, word_len, text_len, line_idx);

                            // Is there enough space for the word on the line
                            if (line_len >= line_idx + word_len)
                            {
                                if ((line_idx > 0) && force_line_break)  
                                {
                                    text_len += line_len - line_idx;
                                    line_idx = 0;

                                    // reset force_line_break
                                    force_line_break = false;
                                }

                                // Yes - add the word to the text block
                                int y;
                                for (y = word_start_idx; y < write_pos; y++)
                                {
                                    display_data.scroll_info.text[text_len] = text_block[y];
                                    text_len++;
                                    //line_idx++;
                                    line_idx = (++line_idx % line_len);
                                    //logger_log_debugf(89, "LINE_IDX 2 set to %d %d --", text_len, line_idx);
                                }
                            }
                            else
                            {
                                // No - how long is the word
                                if ((word_len <= line_len) ||
                                    (line_idx >= line_len - 3)) // Note: 3 is abitary at the moment...    
                                {
                                    // <= line length
                                    // OR > line length but close to the end of the line
                                    // move to start of next line
                                    // OR > line length but we are close enough to the end of the line to start writing a new line
                                    // e.g if theres only a couple of characters itll be easier to read if we display the
                                    // start of the word rather than the word with a couple of characters cut off..
                                    //logger_log_debugf(89, "LINE START %d %d %d %d --", word_len, line_idx, line_len, text_len);

                                    text_len += line_len - line_idx;
                                    line_idx = 0;

                                    // reset force_line_break
                                    force_line_break = false;
                                }

                                // write word
                                int y;
                                for (y = word_start_idx; y < write_pos; y++)
                                {
                                    display_data.scroll_info.text[text_len] = text_block[y];
                                    text_len++;
                                    line_idx = (++line_idx % line_len);
                                }
                            }

                            // Try and add a trailing space - not this may be beyond the end of the line but will be corrected below
                            //logger_log_debugf(89, "ADDING SPACE %d %d %d --", line_idx, line_len, text_len);
                            //logger_log_debugf(89, "AFTER WRITING WORD %d %d --", line_idx, line_len);
                            if (line_idx > 0)
                            {
                                text_len++;
                                line_idx = (++line_idx % line_len);
                            }
                            //logger_log_debugf(89, "LINE_IDX set to %d %d --", text_len, line_idx);
                        }
                        else
                        {
                            // Do nothing
                        }

                        // Clear the current word
                        word_start_idx = -1;
                    }
                    else if (character == DISPLAY_NEWLINE_CHARACTER)
                    {
                        // Yes - add whatever we've got to the text block
                        int y;
                        for (y = word_start_idx; y < write_pos; y++)
                        {
                            display_data.scroll_info.text[text_len] = text_block[y];
                            text_len++;
                            line_idx = (++line_idx % line_len);
                        }

                        // Add trailing spaces to force the new line
                        while (line_idx > 0)
                        {
                            text_len++;
                            line_idx = (++line_idx % line_len);
                        }

                        // Clear the current word
                        word_start_idx = -1;
                    }
                    else
                    {
                        // Check for significant whitespace
                        if (character == DISPLAY_WHITESPACE_CHARACTER)
                        {
                            text_block[write_pos] = ' ';
                        }

                        // Check for UTF-8 characters used by the Avatar
                        if ((uint8_t)character > 127)
                        {
                            // Assume this is a UTF-8 special character and convert
                            uint8_t data = text_block[x + 1];
                            uint8_t data2 = text_block[x + 2];
                            
                            //text_block[x] = ' ';
                            if (character == (char)0xC2)
                            {
                                text_block[write_pos] = display_data.utf_8_character_map[(uint8_t)data - 0x80];
                                
                                // Skip over next 1 character
                                utf_8_skip_character = 1;
                            }
                            else if (character == (char)0xC3)
                            {
                                text_block[write_pos] = display_data.utf_8_character_map[(uint8_t)data - 0x80 + 0x40];
                                //logger_log_debugf(89, "SPECIAL CHARACTER %d %d --", (uint8_t)data, (uint8_t)text_block[write_pos]);
                                
                                // Skip over next 1 character
                                utf_8_skip_character = 1;
                            }
                            else if (character == (char)0xE2) // general punctuation
                            {
                                if ((data >= 0x80) && (data <= 0x81) & (data2 >= 80) & (data2 <= 0xbf))
                                {
                                    text_block[write_pos] =
                                        display_data.utf_8_gp_character_map[(uint8_t)(((data - 0x80) * 64) + (data2 - 0x80))];
                                    
                                    //logger_log_debugf(89, "SPECIAL CHARACTER %d %d %d --", (uint8_t)data, (uint8_t)data2, (uint8_t)text_block[write_pos]);
                                }
                                else
                                {
                                    text_block[write_pos] = ' ';
                                }
                                
                                // Skip over next 2 characters
                                utf_8_skip_character = 2;
                            }
                            else
                            {
                                // Skip over next 1 character
                                utf_8_skip_character = 1;
                                text_block[write_pos] = ' ';
                            }
 
                            
                            
                            // Copy the rest of the text block to cut out the extra
                            // utf-8 byte
//                            int y;
//                            for (y = x + 2; y < strlen(text_block); y++)
//                            {
//                                text_block[y - 1] = text_block[y];
//                            }
                            
//                            text_block[strlen(text_block) - 1] = 0;
                        }

                        // No:
                        // Are we in a word?
                        if (word_start_idx == -1)
                        {
                            // No: so set this as the start of the word
                            word_start_idx = write_pos;//x;
                            //logger_log_debugf(89, "WORD STARTS AT %d %c --", x, text_block[x]);

                            force_line_break = false;
                        }
                        else if (x == strlen(text_block) - 1)
                        {
                            //logger_log_debugf(89, "WRITING WORD %c %d %d %d %d %d --", text_block[word_start_idx], word_start_idx, x - word_start_idx, text_len, line_idx, text_len);

                            //logger_log_debugf(89, "END WORD %d %d %d %d %d", text_len, x, word_start_idx, DISPLAY_MAX_SCROLL_TEXT_LEN, text_len + x - word_start_idx + 1);
                            if (text_len + write_pos - word_start_idx + 1 <  DISPLAY_MAX_SCROLL_TEXT_LEN)
                            {
                                //logger_log_debugf(89, "WRITING END WORD");
                                // Write the final word
                                int y;
                                for (y = word_start_idx; y <= write_pos && text_len < DISPLAY_MAX_SCROLL_TEXT_LEN; y++)
                                {
                                    display_data.scroll_info.text[text_len] = text_block[y];
                                    text_len++;
                                    line_idx = (++line_idx % line_len);
                                }
                            }
                            else
                            {
                                //logger_log_debugf(89, "WRITING DOTS");
                                write_more_dots = true;
                            }
                        }

                         // Line break if word contains a ":"
                        if (character == ':')
                        {
                            force_line_break = true;
                        }
                    }

                    if (text_len >= DISPLAY_MAX_SCROLL_TEXT_LEN - 1)
                    {
                        write_more_dots = true;
                        break;
                    }
                    
                    // Increment the write position
                    write_pos++;
                }
                else // utf_8_skip_character >= 0
                {
                    utf_8_skip_character--;
                }
            }
                 
            // Append .. if not all of the line is displayed due to it being larger
            // than the available buffer
            if (write_more_dots) 
            {
                //logger_log_debugf(89, "WRITING DOTS %d %d %d", text_len, DISPLAY_MAX_SCROLL_TEXT_LEN, line_idx);
                // Check whether were actually filling a whole line
                if (text_len < DISPLAY_MAX_SCROLL_TEXT_LEN - 3)
                {
                    if (line_idx == 0)
                    {
                        // Move to before the ' '
                        text_len--;
                        text_len--;
                    }
                    display_data.scroll_info.text[text_len] = '.';
                    display_data.scroll_info.text[text_len + 1] = '.';
                }
                else
                {
                    display_data.scroll_info.text[DISPLAY_MAX_SCROLL_TEXT_LEN - 2] = '.';
                    display_data.scroll_info.text[DISPLAY_MAX_SCROLL_TEXT_LEN - 3] = '.';
                }
            }
        }
             
        //logger_log_debugf(89, "TEXT LEN: %d ", text_len);
        // Work out number of lines used
        display_data.scroll_info.scroll_num_lines = text_len / display_data.scroll_info.scroll_width;
        //logger_log_debugf(89, "NUM LINES: %d ", display_data.scroll_info.scroll_num_lines);
        if (text_len % display_data.scroll_info.scroll_width > 0)
        {
            display_data.scroll_info.scroll_num_lines++;
            //logger_log_debugf(89, "NUM LINES2: %d %d ", display_data.scroll_info.scroll_num_lines, text_len % display_data.scroll_info.scroll_width);
        }
       
        // Check the current line is in range
        if (display_data.scroll_info.scroll_current_line >= display_data.scroll_info.scroll_num_lines)
        {
            display_data.scroll_info.scroll_current_line = 0;
            display_data.scroll_info.scroll_focus_line = 0;
        }
        
        logger_log_debugf(89, "DISPLAY: scrolling text added len:%d lines:%d", text_len, display_data.scroll_info.scroll_num_lines);
        timers_update_timeout(TIMER_INDEX_DISPLAY_SCROLL, display_data.scroll_info.scroll_pause_ms , TIMERS_TIMEOUT_RESET_IMMEDIATE);
    }
    else
    {
        display_scroll_cancel();
    }
}
    
/**
 * Used to set the display brightness
 * @param brightness - A value from the DISPLAY_BRIGHTNESS enum
 */
void display_set_brightness(DISPLAY_BRIGHTNESS brightness)
{
    if (display_data.brightness != brightness)
    {
        display_data.brightness = brightness;
        
        display_reset_brightness();
    }
 }
    
/**
 * Displays a special character - Note: The values in this list match the special keys in vfd.h
 */
void display_special_character(DISPLAY_LINE line, int at, DISPLAY_SPECIAL_CHARACTERS character)
{
#if OLED_VFD_ENABLED
    display_data.display_lines[line][at] = character;
#endif
    
#if VFD_ENABLED
    display_data.display_lines[line][at] = display_get_special_character(line, character);
#endif
}
    
/**
 * Displays a special character that is offset 
 * Note: The values in this list match the special keys in display_constants.h
 * 
 */
void display_special_character_offset(DISPLAY_LINE line, int at, char* buffer, uint8_t buffer_height, uint8_t buffer_width,  uint16_t index, int offset_h, int offset_v)
{
#if OLED_VFD_ENABLED
    display_data.display_lines[line][at] = character;
#endif
    
#if VFD_ENABLED
    display_data.display_lines[line][at] = vfd_get_special_character(line, buffer, buffer_height, buffer_width, index, offset_h, offset_v);
#endif
}
    
// Clears the content of the line and adds the specified text left justified
void display_start_text(DISPLAY_LINE line, char* text)
{
    display_clear_line(line);
    strncpy(display_data.display_lines[line], text, strlen(text));
}

/**
 * Display text on the bottom line of a 2 line display
 * @param text - the text to be written
 * @param immediate_refresh - Forces an immediate screen refresh
 */
void display_write_line(DISPLAY_LINE line, bool immediate_refresh)
{
    switch (line)
    {
        case DISPLAY_LINE_1:
        {
#if OLED_VFD_ENABLED
            oled_vfd_add_top_string(display_data.display_lines[DISPLAY_LINE_1]);
#endif            
                
#if VFD_ENABLED
            vfd_write_line(DISPLAY_LINE_1, display_data.display_lines[DISPLAY_LINE_1]);
#endif
            break;
        }

        case DISPLAY_LINE_2:
        {
#if OLED_VFD_ENABLED
            oled_vfd_add_bottom_string(display_data.display_lines[DISPLAY_LINE_2]);
#endif            
                
#if VFD_ENABLED
            vfd_write_line(DISPLAY_LINE_2, display_data.display_lines[DISPLAY_LINE_2]);
#endif
            break;
        }
        
        default:
            break;
    }
   
    if (immediate_refresh)
    {
        display_refresh();
    }
}

/**
 * Handles the display of UTF characters
 * @param line
 * @param immediate_refresh
 */
void display_write_line_inc_utf(DISPLAY_LINE line, bool immediate_refresh)
{
    int utf_8_skip_character = 0;
            
    int x;
    int write_pos = 0;
    
    if ((int)line < DISPLAY_LINE_COUNT)
    {
        for (x = 0; x < strlen(display_data.display_lines[line]); x++)
        {
            if (utf_8_skip_character <= 0)
            {
                char character = display_data.display_lines[line][x];

                display_data.display_lines[line][write_pos] = character;
                
                // Check for UTF-8 characters used by the Avatar
                if ((uint8_t)character > 127)
                {
                    // Assume this is a UTF-8 special character and convert
                    uint8_t data = display_data.display_lines[line][x + 1];
                    uint8_t data2 = display_data.display_lines[line][x + 2];

                    //text_block[x] = ' ';
                    if (character == (char)0xC2)
                    {
                        display_data.display_lines[line][write_pos] = display_data.utf_8_character_map[(uint8_t)data - 0x80];
                        // Skip over next character
                        utf_8_skip_character = 1;
                    }
                    else if (character == (char)0xC3)
                    {
                        display_data.display_lines[line][write_pos] = display_data.utf_8_character_map[(uint8_t)data - 0x80 + 0x40];
                        //logger_log_debugf(89, "SPECIAL CHARACTER %d %d --", (uint8_t)data, (uint8_t)text_block[write_pos]);
                        // Skip over next character
                        utf_8_skip_character = 1;
                    }
                    else if (character == (char)0xE2)
                    {
                        // Skip over next 2 characters
                        utf_8_skip_character = 2;
                        
                        if ((data >= 0x80) && (data <= 0x81) & (data2 >= 80) & (data2 <= 0xbf))
                        {
                            display_data.display_lines[line][write_pos] 
                                    = display_data.utf_8_gp_character_map[(uint8_t)(((data - 0x80) * 64) + (data2 - 0x80))];
                            
                        }
                        else
                        {
                            display_data.display_lines[line][write_pos] = ' ';
                        }
                    }
                    else
                    {
                        // Skip over next character
                        utf_8_skip_character = 1;
                        display_data.display_lines[line][write_pos] = ' ';
                    }

                    // Copy the rest of the text block to cut out the extra
                    // utf-8 byte
//                            int y;
//                            for (y = x + 2; y < strlen(text_block); y++)
//                            {
//                                text_block[y - 1] = text_block[y];
//                            }

//                            text_block[strlen(text_block) - 1] = 0;
                }
                
                write_pos++;
            }
            else // >1
            {
                utf_8_skip_character--; // count off skipped character(s)
            }
        }
    }

    display_write_line(line, immediate_refresh);
}

/* *****************************************************************************
 End of File
 */
