/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    display_message.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/display_message.h"
#include "app_config.h"
#include "utilities/logger.h"
#include "utilities/rs232_display.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#ifndef MENU_DISPLAY_DATA_EMPTY_ID   
// Represents a value for a NULL Menu ID or an unassigned option selection    
#define MENU_DISPLAY_DATA_EMPTY_ID  0xFFFF    
#endif

#define DISPLAY_MESSAGE_MESSAGE_CACHE_SIZE  5

typedef struct
{
    //! An incrementing Frame Id that cycles round
    uint8_t frame_id;
    uint32_t prev_processing_index;
    
    
    uint8_t message_cache[DISPLAY_MESSAGE_MESSAGE_CACHE_SIZE][DISPLAY_MESSAGE_UPLOAD_DATA_BUFFER_SIZE];
//+++++    uint8_t message_cache[DISPLAY_MESSAGE_MESSAGE_CACHE_SIZE][DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE];
    uint8_t message_cache_write_idx;
} DISPLAY_MESSAGE_DATA;

static DISPLAY_MESSAGE_DATA display_message_data =
{
    .frame_id = 0,
    .prev_processing_index = 0,
    .message_cache_write_idx = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Initialises a buffer that will be a display message.  Note the buffer should be 
 * large enough to hold the data
 * @param buffer
 */
void display_message_initialise_message(uint8_t* buffer, DISPLAY_MESSAGE_TYPE message_type)
{
    // Use bytes 0-1 as a uint16 holding the length 
    buffer[0] = DISPLAY_MESSAGE_START_OF_PACKET_LSB;
    buffer[1] = DISPLAY_MESSAGE_START_OF_PACKET_MSB;
    buffer[2] = 9;                                      // Message Length LSB
    buffer[3] = 0;                                      // Message Length MSB
    
    display_message_set_frame_and_packet_id(buffer, display_message_data.frame_id++, 0);

    buffer[6] = DISPLAY_MESSAGE_CONTROL_FLAGS_END_OF_MESSAGE; // Control Flags
    buffer[7] = message_type;       // Message Type
    buffer[8] = 0;  // Number of items                  // Number of Items
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Initialises a buffer that will be a screen_data message.  Note the buffer should be 
 * large enough to hold the data
 * @param buffer        The buffer to be populated  
 * @param data_id       The ID of the data item 
 * @param attribute_id  Allows for more than one screen data record per data value (i.e. Folder Name, Folder Image)
 * @param data_id       The type data item (OptionList, PreDefined value)
 * @param encoding      The encoding used for the data (Boolean, UINT8, BIN16)
 * @param data          The encoded data - Note: if the encoding is BIN16 - the data will be CONVERTED into BIN16, so no need to do it beforehand
 * @param data_len      The length of the encoded data
 */
void display_message_add_screen_data(uint8_t* screen_data, 
        DISPLAY_MESSAGE_DATA_ID data_id,
        uint8_t attribute_id,
        DISPLAY_MESSAGE_DATA_ENCODING data_encoding, 
        uint8_t* data, uint16_t data_len)
{
    // Use bytes 0-1 as a uint16 holding the length 
    uint16_t len_used;
    memcpy(&len_used, &screen_data[2], sizeof(uint16_t));

    // Update the number of items
    screen_data[8] = screen_data[8] + 1;
    
    // Add the Data ID
    memcpy(&screen_data[len_used], &data_id, sizeof(DISPLAY_MESSAGE_DATA_ID));
    len_used += sizeof(DISPLAY_MESSAGE_DATA_ID);
    
    // Add the attribute id
    screen_data[len_used] = (uint8_t)attribute_id;
    len_used++;
    
    // Add the encoding
    screen_data[len_used] = (uint8_t)data_encoding;
    len_used++;
    
    if (data_encoding == DISPLAY_MESSAGE_DATA_ENCODING_BIN16)
    {
        // Write the data length to implement BIN16 format
        memcpy(&screen_data[len_used], &data_len, sizeof(uint16_t));
        len_used += sizeof(uint16_t);
    }

    // Debugging
    switch (data_encoding)
    {
        case DISPLAY_MESSAGE_DATA_ENCODING_NONE:
        {
            logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA_VALUE, "DISPLAY MESSAGE Data (NONE) %d %d --", data_id, attribute_id);
            break;
        }
        case DISPLAY_MESSAGE_DATA_ENCODING_BOOLEAN:
        {
            logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA_VALUE, "DISPLAY MESSAGE Data (BOOL) %d %d %s --", data_id, attribute_id, data[0] == 0 ? "False" : "True");
            break;
        }
        case DISPLAY_MESSAGE_DATA_ENCODING_UINT8:
        {
            logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA_VALUE, "DISPLAY MESSAGE Data (UINT8) %d %d %d --", data_id, attribute_id, data[0]);
            break;
        }
        case DISPLAY_MESSAGE_DATA_ENCODING_UINT16:
        {
            uint16_t val;
            memcpy(&val, data, sizeof(uint16_t));
            logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA_VALUE, "DISPLAY MESSAGE Data (UINT16) %d %d %d --", data_id, attribute_id, val);
            break;
        }
        case DISPLAY_MESSAGE_DATA_ENCODING_UINT32:
        {
            uint32_t val;
            memcpy(&val, data, sizeof(uint32_t));
            logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA_VALUE, "DISPLAY MESSAGE Data (UINT32) %d %d %d --", data_id, attribute_id, val);
            break;
        }
        case DISPLAY_MESSAGE_DATA_ENCODING_UINT64:
        {
            uint64_t val;
            memcpy(&val, data, sizeof(uint64_t));
            logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA_VALUE, "DISPLAY MESSAGE Data (UINT64) %d %d %d --", data_id, attribute_id, val);
            break;
        }
        case DISPLAY_MESSAGE_DATA_ENCODING_BIN16:
        {
            logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA_VALUE, "DISPLAY MESSAGE Data (BIN16) %d %d %s --", data_id, attribute_id, data);
            break;
        }
        
    }
    
    // Copy the data
    memcpy(&screen_data[len_used], data, data_len);
    len_used += data_len;
    
    // Update the length
    memcpy(&screen_data[2], &len_used, sizeof(uint16_t));
}

/**-----------------------------------------------------------------------------
 * Works out what the first option to be displayed should be based on the page_size 
 * and currently selected option.  This works on the ordinal position of the item in
 * the option list.
 * @param menu_item
 * @param selected_idx
 * @param page_size
 * @return 
 */
uint8_t display_message_calculate_top_option_idx(uint16_t page_start_idx, uint16_t selected_idx, uint8_t page_size, uint8_t option_count)
{
    // Now calculate the top item
    uint8_t top_idx = 0;

    if (selected_idx != MENU_DISPLAY_DATA_EMPTY_ID)//> -1)
    {
        // Get the id of the item currently displayed at the top of the screen
        //uint16_t page_start_idx = menu_item->display_data.page_start_id;
        if (page_start_idx == MENU_DISPLAY_DATA_EMPTY_ID)
        {
            page_start_idx = 0;
        }
        if (page_size <= DISPLAY_MESSAGE_MAX_ROWS)
        {
            // Scrolling by item (e.g when we scroll we only move the display by 1 item)
            // This is best suited to a single column of options where we only display
            // a subset of the options

            // If the top item can stay the same and still display the selected option then great
            // otherwise put the selected option at the top, while maintaining a full screen
            if (selected_idx < page_start_idx)
            {
               // Moving Up
                top_idx = selected_idx;
            }
            else if (selected_idx >= page_start_idx + page_size)
            {
               // Moving Down
                top_idx = selected_idx - page_size + 1;

                // Adjust for the end of the page, so that we try and have a full page
                if ((top_idx + page_size) > option_count)
                {
                    top_idx = option_count - page_size;
                }
            }
            else
            {
                // Dont move
                top_idx = page_start_idx;
            }
        }
        else
        {
            // Scrolling by page (when the display change a complete new page is displayed)
            if (selected_idx < page_start_idx)
            {
                // Moving Up
                if (page_start_idx >= page_size)
                {
                    top_idx = page_start_idx - page_size;
                }
                else
                {
                    top_idx = 0;
                }
            }
            else if (selected_idx >= page_start_idx + page_size)
            {
                // Moving Down
                top_idx = page_start_idx + page_size;
                if (top_idx > option_count)
                {
                    top_idx = (top_idx / page_size) * page_size;
                }
            }
            else
            {
                // Dont move
                top_idx = page_start_idx;
            }                    
        }
    }

    return top_idx;
}
   
/**-----------------------------------------------------------------------------
 * Sets up a DISPLAY_MESSAGE_DATA_MAP for the supplied display message
 * Is used to speed up ordinal access of Screen Data 
 */
void display_message_data_map_build(DISPLAY_MESSAGE* message, DISPLAY_MESSAGE_DATA_MAP* map)
{
    if (message->message_type == DISPLAY_MESSAGE_TYPE_SCREEN_DATA)
    {
        // Get the number of items
        map->num_items = message->data[0];
        
        // Loop through the items and get their data_id, attribute_id, encoding, and 
        // offset of the data and length of the data
        uint8_t item_idx;
        uint16_t offset = 1;
        for (item_idx = 0; item_idx < map->num_items; item_idx++)
        {
            memcpy(&map->items[item_idx].data_id, &message->data[offset], sizeof(DISPLAY_MESSAGE_DATA_ID));
            offset += sizeof(DISPLAY_MESSAGE_DATA_ID);

            map->items[item_idx].attribute_id = message->data[offset];
            offset += sizeof(uint8_t);

            map->items[item_idx].encoding = (DISPLAY_MESSAGE_DATA_ENCODING)message->data[offset];
            offset += sizeof(uint8_t);

            switch (map->items[item_idx].encoding)
            {
                case DISPLAY_MESSAGE_DATA_ENCODING_NONE:
                {
                    map->items[item_idx].length = 0;
                    break;
                }
                case DISPLAY_MESSAGE_DATA_ENCODING_BOOLEAN:
                {
                    map->items[item_idx].length = sizeof(bool);
                    break;
                }
                case DISPLAY_MESSAGE_DATA_ENCODING_UINT8:
                {
                    map->items[item_idx].length = sizeof(uint8_t);
                    break;
                }
                case DISPLAY_MESSAGE_DATA_ENCODING_UINT16:
                {
                    map->items[item_idx].length = sizeof(uint16_t);
                    break;
                }
                case DISPLAY_MESSAGE_DATA_ENCODING_UINT32:
                {
                    map->items[item_idx].length = sizeof(uint32_t);
                    break;
                }
                case DISPLAY_MESSAGE_DATA_ENCODING_UINT64:
                {
                    map->items[item_idx].length = sizeof(uint64_t);
                    break;
                }
                case DISPLAY_MESSAGE_DATA_ENCODING_BIN16:
                {
                    memcpy(&map->items[item_idx].length, &message->data[offset], sizeof(uint16_t));
                    break;
                }         
            }

            // Move the offset on
            if (map->items[item_idx].encoding != DISPLAY_MESSAGE_DATA_ENCODING_BIN16)
            {
                map->items[item_idx].data = &message->data[offset];
                offset += map->items[item_idx].length;
            }
            else
            {
                // Add BIN16 length offset
                map->items[item_idx].data = &message->data[offset + 2]; // Points at the start of the data
                offset += (map->items[item_idx].length + 2);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Finalises a buffer that will be a screen_data message.  Note the buffer should be 
 * large enough to hold the data
 * @param buffer
 */
uint16_t display_message_finalise_and_send_screen_data(uint8_t* buffer)
{
    // Use bytes 0-1 as a uint16 holding the length 
    uint16_t len_used;
    memcpy(&len_used, &buffer[2], sizeof(uint16_t));

    // Add the tail
    buffer[len_used] = DISPLAY_MESSAGE_END_OF_PAYLOAD_LSB;
    buffer[len_used + 1] = DISPLAY_MESSAGE_END_OF_PAYLOAD_MSB;
    
    // Update the length
    len_used += 2;
    memcpy(&buffer[2], &len_used, sizeof(uint16_t));
    
    // Only send if there is data
    uint8_t no_of_items = buffer[8]; 
    if (no_of_items > 0)
    {
        logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA, "DISPLAY MESSAGE Sending Screen Data %d/%d typ:%d items:%d len:%d --", buffer[4], buffer[5], buffer[7], buffer[8], len_used);

        if (buffer[7] == 5)
        {
            logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA, "ALBUM ART Data %d %d %d %d - %d %d %d %d --",
                    buffer[len_used - 8],
                    buffer[len_used - 7],
                    buffer[len_used - 6],
                    buffer[len_used - 5],
                    buffer[len_used - 4],
                    buffer[len_used - 3],
                    buffer[len_used - 2],
                    buffer[len_used - 1]);
                    
        }
        
        // Check if were adding to the message cache for resending 
        if (len_used <= DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE)
        {
            // Only add screen messages not upgrade
            memcpy(&display_message_data.message_cache[display_message_data.message_cache_write_idx][0], buffer, len_used);
            display_message_data.message_cache_write_idx++;
            display_message_data.message_cache_write_idx %= DISPLAY_MESSAGE_MESSAGE_CACHE_SIZE;
        }

        rs232_display_send_buffer(buffer, len_used);
    }
    else
    {
        logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA, "DISPLAY MESSAGE NOT Sending Screen Data typ:%d items:%d len:%d --", buffer[7], buffer[8], len_used);

    }
    
    return len_used;
}
    
/**-----------------------------------------------------------------------------
 * Puts the display data into an array.  For bin 16 fields this will be the data 
 * only and omit the length part of the bin 16 field.
 * @param message - The message containing the data
 * @param data_id - The data Id for the data
 * @param attribute_id - The attribute Id for the data
 * @param encoding - The encoding type of the data
 * @param data - A pointer to an array of uint8_t to hold the data
 * @param len - The length of data available in the array - if the actual data is longer only this amount will be copied.
 * @return the length of the data copied (0 if no data or NIL type).
 */
uint16_t display_message_get_data(DISPLAY_MESSAGE* message, 
        DISPLAY_MESSAGE_DATA_ID data_id, 
        uint8_t attribute_id, 
        DISPLAY_MESSAGE_DATA_ENCODING* encoding, 
        uint8_t* data, 
        uint16_t data_len)
{
    void* data_ptr;
    uint16_t len = 0;
    
    if (display_message_get_value(message, 
            data_id, 
            attribute_id,
            encoding, 
            &data_ptr, 
            &len))
    {
        // Adjust the length for BIN16
        switch(*encoding)
        {
            case DISPLAY_MESSAGE_DATA_ENCODING_BIN16:
            {
                data_ptr += 2;
                len -= 2;
                break;
            }
            
            default:
            {
                break;
            }
        }
        
        memset(data, 0, data_len);
        memcpy(data, data_ptr, min(len, data_len));
    }

    return len;
}
     
/**-----------------------------------------------------------------------------
 * Gets the next frame Id
 * @return 
 */
uint8_t display_message_get_next_frame_id()
{
    return display_message_data.frame_id++;
}
    
/**-----------------------------------------------------------------------------
 * Populates a pointer with the address of the actual data in the message
 * @param message
 * @param data_id
 * @param attribute_id
 * @param encoding
 * @param value - A pointer to a pointer that will be set to point to the data in th packet
 * @return true if state_if found in message
 */
bool display_message_get_value(DISPLAY_MESSAGE* message, 
        DISPLAY_MESSAGE_DATA_ID data_id, 
        uint8_t attribute_id,
        DISPLAY_MESSAGE_DATA_ENCODING* encoding, 
        void** value, uint16_t* len)
{
    bool result = false;
    
    // Find the number of data items in the message
    uint8_t num_items = message->data[0];
 
    // Loop through the items to see if this is the data_id/attribute we want
    uint8_t item_idx;
    uint16_t offset = 1;
    for (item_idx = 0; item_idx < num_items; item_idx++)
    {
        DISPLAY_MESSAGE_DATA_ID item_id;
        memcpy(&item_id, &message->data[offset], sizeof(DISPLAY_MESSAGE_DATA_ID));
        offset += sizeof(DISPLAY_MESSAGE_DATA_ID);
        
        uint8_t item_attr_id = message->data[offset];
        offset += sizeof(uint8_t);
        
        *encoding = (DISPLAY_MESSAGE_DATA_ENCODING)message->data[offset];
        offset += sizeof(uint8_t);
        
        switch (*encoding)
        {
            case DISPLAY_MESSAGE_DATA_ENCODING_NONE:
            {
                *len = 0;
                break;
            }
            case DISPLAY_MESSAGE_DATA_ENCODING_BOOLEAN:
            {
                *len = sizeof(uint8_t);
                break;
            }
            case DISPLAY_MESSAGE_DATA_ENCODING_UINT8:
            {
                *len = sizeof(uint8_t);
                break;
            }
            case DISPLAY_MESSAGE_DATA_ENCODING_UINT16:
            {
                *len = sizeof(uint16_t);
                break;
            }
            case DISPLAY_MESSAGE_DATA_ENCODING_UINT32:
            {
                *len = sizeof(uint32_t);
                break;
            }
            case DISPLAY_MESSAGE_DATA_ENCODING_UINT64:
            {
                *len = sizeof(uint64_t);
                break;
            }
            case DISPLAY_MESSAGE_DATA_ENCODING_BIN16:
            {
                uint16_t bin16_data_len;
                memcpy(&bin16_data_len, &message->data[offset], sizeof(uint16_t));
                bin16_data_len += 2; // To include the BIN16 length bit
                *len = bin16_data_len;
                break;
            }         
        }
         
        // Now check if this is the data we want
        if ((item_id == data_id) && (item_attr_id == attribute_id))
        {
            *value = (void*)(message->data + offset);
            result = true;
            break;
        }
        else
        {
            // Move the offset on
            offset += *len;
        }
    }
    
    if (!result)
    {
        *len = 0;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether the display message has had errors when being transferred
 * @param message
 * @return 
 */
bool display_message_has_errors(DISPLAY_MESSAGE* message)
{
    return message->processing_state == DISPLAY_MESSAGE_STATE_ERRORS;
}

/**-----------------------------------------------------------------------------
 * Initialises the supplied Display Message readying it for populating
 * @param message
 */
void display_message_initialise(DISPLAY_MESSAGE* message)
{
    message->message_type = DISPLAY_MESSAGE_TYPE_UNKNOWN;
    message->processing_state = DISPLAY_MESSAGE_STATE_INITIALISED;
    message->processing_data_index = 0;
    memset(message->data, 0, DISPLAY_MESSAGE_MAX_DATA_LENGTH);
}

/**-----------------------------------------------------------------------------
 * Initialises a buffer that will be a screen_data message.  Note the buffer should be 
 * large enough to hold the data
 * @param buffer
 */
void display_message_initialise_screen_data(uint8_t* screen_data)
{
    display_message_initialise_message(screen_data, DISPLAY_MESSAGE_TYPE_SCREEN_DATA);
}

/**-----------------------------------------------------------------------------
 * Initialises a buffer that will be a data_transfer message.  Note the buffer should be 
 * large enough to hold the data
 * @param buffer
 */
void display_message_initialise_data_transfer_data(uint8_t* screen_data, uint8_t packet_id)
{
    display_message_initialise_message(screen_data, DISPLAY_MESSAGE_TYPE_DATA_TRANSFER);
    
    // Set packet Id
    screen_data[5] = packet_id;
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether the display message has been fully populated
 * @param message
 * @return 
 */
bool display_message_is_complete(DISPLAY_MESSAGE* message)
{
    return message->processing_state == DISPLAY_MESSAGE_STATE_POPULATED;
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether the display message is currently being populated
 * @param message
 * @return 
 */
bool display_message_is_processing(DISPLAY_MESSAGE* message)
{
    return message->processing_state != DISPLAY_MESSAGE_STATE_INITIALISED;
}

/**-----------------------------------------------------------------------------
 * Process a byte of data read from the RX channel
 * @param data_byte
 */
//uint8_t count = 0;
void display_message_process_byte(DISPLAY_MESSAGE* message, unsigned char data_byte)
{
//THIS COCKS UP THE UPGRADE!!! WHICH CAN CONTAIN 0xFF    if (data_byte == DISPLAY_MESSAGE_START_OF_PACKET_LSB)
//    {
//        message->processing_state = DISPLAY_MESSAGE_STATE_INITIALISED;
//    }

    switch (message->processing_state)
    {
        case DISPLAY_MESSAGE_STATE_INITIALISED:
        {
            if (data_byte == DISPLAY_MESSAGE_START_OF_PACKET_LSB)
            {
                //logger_log_debugf(0, "DISPLAY ALBUM ART display_message_process_byte A %d --", display_message_data.prev_processing_index);
                display_message_data.prev_processing_index = 0;
                
                //logger_log_debugf(0, "DISPLAY display_message_process_byte F/P Start Of Message %d --", data_byte);
                message->processing_state = DISPLAY_MESSAGE_STATE_WAIT_FOR_START_MSB;
            }
            break;
        }
        
        case DISPLAY_MESSAGE_STATE_WAIT_FOR_START_MSB:
        {
            //logger_log_debugf(0, "DISPLAY display_message_process_byte A %d --", data_byte);
            if (data_byte == DISPLAY_MESSAGE_START_OF_PACKET_MSB)
            {
                message->processing_state = DISPLAY_MESSAGE_STATE_FOR_LEN_LSB;
            }
            else
            {
                display_message_initialise(message);
            }
            
            break;
        }
        
        case DISPLAY_MESSAGE_STATE_FOR_LEN_LSB:
        {
            //logger_log_debugf(0, "DISPLAY display_message_process_byte B %d --", data_byte);
            message->length = (uint8_t)data_byte;
            message->processing_state = DISPLAY_MESSAGE_STATE_WAIT_LEN_MSB;
            break;
        }
        
        case DISPLAY_MESSAGE_STATE_WAIT_LEN_MSB:
        {
            message->length += ((uint16_t)data_byte * 256);
            message->processing_state = DISPLAY_MESSAGE_STATE_WAIT_FRAME_ID;
            //logger_log_debugf(0, "DISPLAY ALBUM ART display_message_process_byte %d  message_len %d --", data_byte, message->length);
            break;
        }
        
        case DISPLAY_MESSAGE_STATE_WAIT_FRAME_ID:
        {
            message->frame_id = (uint8_t)data_byte;
            message->processing_state = DISPLAY_MESSAGE_STATE_WAIT_PACKET_ID;
            break;
        }
        
        case DISPLAY_MESSAGE_STATE_WAIT_PACKET_ID:
        {
            message->packet_id = (uint8_t)data_byte;
            message->processing_state = DISPLAY_MESSAGE_STATE_WAIT_CONTROL_FLAGS;
            //logger_log_debugf(0, "DISPLAY F/P display_message_process_byte F/P %d/%d %d --", message->frame_id, message->packet_id, message->length);

            break;
        }
        
        case DISPLAY_MESSAGE_STATE_WAIT_CONTROL_FLAGS:
        {
            message->control_flags = (uint8_t)data_byte;
            message->processing_state = DISPLAY_MESSAGE_STATE_WAIT_MESSAGE_TYPE;
            break;
        }
        
        case DISPLAY_MESSAGE_STATE_WAIT_MESSAGE_TYPE:
        {
            message->message_type = (DISPLAY_MESSAGE_TYPE)data_byte;
            message->processing_data_index = 0;
            message->processing_state = DISPLAY_MESSAGE_STATE_POPULATING;
//count++;            
            //logger_log_debugf(0, "DISPLAY display_message_process_byte D2 %d %d --", data_byte, message->length);
            break;
        }
        
        case DISPLAY_MESSAGE_STATE_POPULATING:
        {
            //logger_log_debugf(0, "DISPLAY display_message_process_byte E F/P %d %d %d --", message->processing_data_index, data_byte, message->length - 8);
            
            message->data[message->processing_data_index] = data_byte;
            message->processing_data_index++;
            
            display_message_data.prev_processing_index = message->processing_data_index;
             
//            if ((count == 17) && (message->processing_data_index % 7 == 0))
//            {
//                logger_log_debugf(0, "DISPLAY display_message_process_byte EX %d %d %d --", message->processing_data_index, data_byte, message->length - 8);
//                logger_process();
//            }
//            if ((message->processing_data_index > 2040) || (message->processing_data_index % 500 == 0))
//            {
//                logger_log_debugf(0, "DISPLAY display_message_process_byte ESS %d %d %d --", message->processing_data_index, data_byte, message->length);
//            }
            
            if (message->processing_data_index >= message->length - 8)
            {
//                logger_log_debugf(0, "DISPLAY display_message_process_byte E %d %d %d %d %d --", message->processing_data_index, data_byte, message->data[message->processing_data_index - 1], message->data[message->processing_data_index - 2], message->length - 8);
                if ((message->data[message->processing_data_index - 2] == DISPLAY_MESSAGE_END_OF_PAYLOAD_LSB) &&
                    (message->data[message->processing_data_index - 1] == DISPLAY_MESSAGE_END_OF_PAYLOAD_MSB))
                {
                    //logger_log_debugf(0, "DISPLAY ALBUM ART display_message_process_byte P F/P %d %d %d --", message->processing_data_index, data_byte, message->length - 8);
                    message->processing_state = DISPLAY_MESSAGE_STATE_POPULATED;
                }
                else
                {
                    if (message->processing_data_index == message->length - 8)
                    {
                        logger_log_debugf(0, "DISPLAY ALBUM ART F/P display_message_process_byte - Invalid end characters %d/%d %d %d --", 
                                message->frame_id,
                                message->packet_id,
                                message->message_type,
                                message->length);
//                        logger_log_debugf(0, "DISPLAY ALBUM ART %d %d %d %d - %d %d %d %d --", 
//                                message->data[0], 
//                                message->data[1], 
//                                message->data[2], 
//                                message->data[3], 
//                                message->data[message->processing_data_index - 4], 
//                                message->data[message->processing_data_index - 3], 
//                                message->data[message->processing_data_index - 2], 
//                                message->data[message->processing_data_index - 1]);
                        
                        //TODO Remove display_message_initialise(message);
                    }
                    
                    message->processing_state = DISPLAY_MESSAGE_STATE_ERRORS; 
                }
                
//                if ((message->message_type == 5) && (message->processing_data_index == message->length - 8))
//                {
//                    logger_log_debugf(0, "DISPLAY ALBUM ART Data %d %d %d %d - %d %d %d %d --", 
//                            message->data[message->processing_data_index - 8], 
//                            message->data[message->processing_data_index - 7], 
//                            message->data[message->processing_data_index - 6], 
//                            message->data[message->processing_data_index - 5], 
//                            message->data[message->processing_data_index - 4], 
//                            message->data[message->processing_data_index - 3], 
//                            message->data[message->processing_data_index - 2], 
//                            message->data[message->processing_data_index - 1]);
//                }
            }
            break;
        }

        // The message has been completely received and is ready for processing
        case DISPLAY_MESSAGE_STATE_POPULATED:
        {
            break;
        }

        // The message has been completely received but has failed
        case DISPLAY_MESSAGE_STATE_ERRORS:
        {
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Checks the message cache for a matching message and resends if found
 * @param frame_id
 * @param packet_id
 */
void display_message_resend_from_cache(uint8_t frame_id, uint8_t packet_id)
{
    bool found = false;
    // Try and find the frame/packt matching the request
    int x;
    for (x = 0; x < DISPLAY_MESSAGE_MESSAGE_CACHE_SIZE; x++)
    {
        if ((display_message_data.message_cache[x][4] == frame_id) && (display_message_data.message_cache[x][5] == packet_id))
        {
            display_message_resend_screen_data(&display_message_data.message_cache[x][0]);
            found = true;
            break;
        }
    }
    
    if (!found)
    {
        logger_log_errorf(0, "ALBUM ART F/P Unable to find in cache %d/%d --", frame_id, packet_id);
    }
}

/**-----------------------------------------------------------------------------
 * resends the screen data buffer without adding the trailing EOF markers
 * for use with a backup of a buffer sent using display_message_finalise_and_send_screen_data
 */
void display_message_resend_screen_data(uint8_t* buffer)
{
    // Use bytes 0-1 as a uint16 holding the length 
    uint16_t len_used;
    memcpy(&len_used, &buffer[2], sizeof(uint16_t));
    
    // Only send if there is data
    uint8_t no_of_items = buffer[8]; 
    if (no_of_items > 0)
    {
        logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA, "DISPLAY MESSAGE F/P Resending Screen Data %d/%d typ:%d items:%d len:%d --", buffer[4], buffer[5], buffer[7], buffer[8], len_used);

        rs232_display_send_buffer(buffer, len_used);
    }
    else
    {
        logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_DATA, "DISPLAY MESSAGE F/P NOT Resending Screen Data typ:%d items:%d len:%d --", buffer[7], buffer[8], len_used);

    }
}

/**-----------------------------------------------------------------------------
 * Sends a single item data request - intended for use from Display MCU to Main MCU
 * @param data - A pointer to the data to be sent
 * @param data_len - The length of the data
 */
void display_message_send_data_request(DISPLAY_MESSAGE_DATA_ID data_id,
        uint8_t attribute_id,
        DISPLAY_MESSAGE_DATA_ENCODING data_encoding,
        uint8_t* data, 
        size_t data_len)
{
    // Declare a screen data buffer
    uint8_t screen_data[DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE];
    display_message_initialise_message(screen_data, DISPLAY_MESSAGE_TYPE_DATA_REQUEST);

    // Populate the screen data
    display_message_add_screen_data(screen_data, 
        data_id,
        attribute_id,
        data_encoding,
        (uint8_t*)data,
        data_len);

    display_message_finalise_and_send_screen_data(screen_data);
}

/**-----------------------------------------------------------------------------
 * Sends a single item data request - intended for use from Display MCU to Main MCU
 * @param data - A pointer to the data to be sent
 * @param data_len - The length of the data
 */
void display_message_send_display_control(DISPLAY_MESSAGE_DATA_ID data_id,
        uint8_t attribute_id,
        DISPLAY_MESSAGE_DATA_ENCODING data_encoding,
        uint8_t* data, 
        size_t data_len)
{
    // Declare a screen data buffer
    uint8_t screen_data[DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE];
    display_message_initialise_message(screen_data, DISPLAY_MESSAGE_TYPE_DISPLAY_CONTROL);

    // Populate the screen data
    display_message_add_screen_data(screen_data, 
        data_id,
        attribute_id,
        data_encoding,
        (uint8_t*)data,
        data_len);

    display_message_finalise_and_send_screen_data(screen_data);
}

/**-----------------------------------------------------------------------------
 * Sends a Log message - intended for use from Display MCU to Main MCU
 * @param message - A pointer to the message to be sent
 * @param message_len -  The length of the message
 */
void display_message_send_log_message(LOGGER_MESSAGE* message)
{
    // Declare a screen data buffer
    uint8_t screen_data[DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE];
    display_message_initialise_message(screen_data, DISPLAY_MESSAGE_TYPE_LOG_MESSAGE);

    // Add the message type DEBUG, INFO etc
    uint8_t type = message->type;
    display_message_add_screen_data(screen_data, 
        0x0001, // Message
        DISPLAY_MESSAGE_ATTRIBUTE_LOGGER_TYPE, 
        DISPLAY_MESSAGE_DATA_ENCODING_UINT8,
        (uint8_t*)&type,
        sizeof(uint8_t));

    // Add the message code 
    uint8_t code = message->code;
    display_message_add_screen_data(screen_data, 
        0x0001, // Message
        DISPLAY_MESSAGE_ATTRIBUTE_LOGGER_CODE, 
        DISPLAY_MESSAGE_DATA_ENCODING_UINT8,
        (uint8_t*)&code,
        sizeof(uint8_t));

    // Add the message
    display_message_add_screen_data(screen_data, 
        0x0001, // Message
        DISPLAY_MESSAGE_ATTRIBUTE_LOGGER_MESSAGE, 
        DISPLAY_MESSAGE_DATA_ENCODING_BIN16,
        (uint8_t*)message->message,
        strlen(message->message));

    // Replaced by code below to stop logging display_message_finalise_and_send_screen_data(screen_data);
   // Use bytes 0-1 as a uint16 holding the length 
    uint16_t len_used;
    memcpy(&len_used, &screen_data[2], sizeof(uint16_t));

    // Add the tail
    screen_data[len_used] = DISPLAY_MESSAGE_END_OF_PAYLOAD_LSB;
    screen_data[len_used + 1] = DISPLAY_MESSAGE_END_OF_PAYLOAD_MSB;
    
    // Update the length
    len_used += 2;
    memcpy(&screen_data[2], &len_used, sizeof(uint16_t));
    
    rs232_display_send_buffer(screen_data, len_used);
 }

///**-----------------------------------------------------------------------------
// * Sends Upgrade Data from Host->Display
// * @param total_size -  The total size of the Upgrade data
// * @param offset -  The offset in the total data of the data included in this message
// * @param upgrade_data - The data to be transferred
// * @param data_length -  The size of the data to be transferred
// */
//void display_message_send_upgrade_data(uint32_t total_size,
//        uint32_t offset,
//        uint8_t* upgrade_data,
//        uint32_t data_length)
//{
//    // Declare a screen data buffer
//    uint8_t screen_data[DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE];
//    display_message_initialise_message(screen_data, DISPLAY_MESSAGE_TYPE_LOG_MESSAGE);
//
//     display_message_add_screen_data(screen_data, 
//        DISPLAY_MESSAGE_UPGRADE_DATA_ID_SIZE, 
//        DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT, 
//        DISPLAY_MESSAGE_DATA_ENCODING_UINT32,
//        (uint8_t*)&total_size,
//        sizeof(uint32_t));
//
//    display_message_add_screen_data(screen_data, 
//        DISPLAY_MESSAGE_UPGRADE_DATA_ID_OFFSET,
//        DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT, 
//        DISPLAY_MESSAGE_DATA_ENCODING_UINT32,
//        (uint8_t*)&offset,
//        sizeof(uint32_t));
//
//    // Add the message
//    display_message_add_screen_data(screen_data, 
//        DISPLAY_MESSAGE_UPGRADE_DATA_ID_DATA,
//        DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT, 
//        DISPLAY_MESSAGE_DATA_ENCODING_BIN16,
//        (uint8_t*)upgrade_data,
//        data_length);
//
//    display_message_finalise_and_send_screen_data(screen_data);
//}

/**-----------------------------------------------------------------------------
 * Sends an Upgrade Status message either between Host->Display or Display->Host
 * @param status - The status to be transferred
 */
void display_message_send_upgrade_status(DISPLAY_MESSAGE_UPGRADE_STATUS status)
{
    // Declare a screen data buffer
    uint8_t screen_data[DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE];
    display_message_initialise_message(screen_data, DISPLAY_MESSAGE_TYPE_UPGRADE_STATUS);

    // convert enum (uint32) to uint8_t)
    uint8_t stat = status;
    display_message_add_screen_data(screen_data, 
        DISPLAY_MESSAGE_UPGRADE_DATA_ID_STATUS, // Message
        DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT, // Default attribute
        DISPLAY_MESSAGE_DATA_ENCODING_UINT8,
        (uint8_t*)&stat,
        sizeof(uint8_t));

    display_message_finalise_and_send_screen_data(screen_data);
}
    
/**-----------------------------------------------------------------------------
 * Sets the control, flags on a display message buffer
 * @param screen_data
 * @param control_flags
 */
void display_message_set_control_flags(uint8_t* screen_data, DISPLAY_MESSAGE_CONTROL_FLAGS control_flags)
{
    screen_data[6] = control_flags;
}
    
/**-----------------------------------------------------------------------------
 * Can be used to fix the frame and packet Id on a message
 * Note: Does not stop the internal frame id being upated.
 * @param message
 * @param frame_id
 * @param packet_id
 */
void display_message_set_frame_and_packet_id(uint8_t* screen_data, uint8_t frame_id, uint8_t packet_id)
{
    screen_data[4] = frame_id;        // Frame ID
    screen_data[5] = packet_id;       // Packet ID
}

/* *****************************************************************************
 End of File
 */
