/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    display_message.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/display_message_controller.h"
#include "utilities/display_message.h"
#include "app_config.h"
#include "utilities/logger.h"
#include "utilities/power.h"
#include "utilities/rs232_display.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    // The current type of navigation
    DISPLAY_MESSAGE_SCREEN_NAVIGATION current_navigation;
    
    uint8_t handshake_state;
} DISPLAY_MESSAGE_DATA;

static DISPLAY_MESSAGE_DATA display_message_data = {
    .handshake_state = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations
static uint16_t display_message_populate_screen_data(MENU_ID menu_id, MENU_ID selected_id);
static void display_message_send_display_screen(DISPLAY_MESSAGE_SCREEN_TYPE screen_type, DISPLAY_MESSAGE_SCREEN_ID screen_id, char* title);

/**-----------------------------------------------------------------------------
 * Sets the supplied item as the currently selected item in the list
 * Checks whether it is currently being displayed, and if so returns the menu_id
 * of the item,  otherwise rebuilds the screen 
 * @param menu_item
 */
static uint16_t display_message_option_list_select_item(MENU_ITEM* menu_item)
{
    uint16_t result = menu_item->menu_id;
    MENU_ITEM* parent_item = menu_get_parent(menu_item);
    MENU_ITEM* child_item = menu_get_item(parent_item->display_data.page_start_id);
    
    // First see if the current item is being displayed in the list by looping
    // from the top menu item being displayed upto the current row count
    bool found = false;
    int x;
    for (x = 0; x < display_message_get_page_size(parent_item); x++) 
    {
        //logger_log_debugf(0, "DISPLAY MESSAGE display_message_option_list_select_item %s %s ", menu_get_title(child_item->menu_id), menu_get_title(menu_item->menu_id));

        if (child_item->menu_id == menu_item->menu_id)
        {
            found = true;
            break;
        }

        child_item = child_item->next_sibling;    
        if (child_item == NULL)
        {
            break;
        }
    }
    
    // If we cant find the item then we need to redraw the screen
    if (!found)
    {
        // Get the page size for the parent menu
        uint8_t page_size = display_message_get_page_size(parent_item);

        // First get the ordinal position of the selected menu item in its parents 
        // list of items and the menu item representing the page start (top)
        // Also get the total number of options in the menu
        uint16_t selected_idx = 0;
        uint16_t page_start_idx = 0;
        uint16_t option_count = 0;
        
        // Get the ordinal position of the selected menu item
        child_item = parent_item->first_child;
        while (child_item != NULL)
        {
            if (child_item == menu_item)
            {
                selected_idx = option_count;
            }

            if (child_item->menu_id == parent_item->display_data.page_start_id)
            {
                page_start_idx = option_count;
            }

            child_item = child_item->next_sibling;
            option_count++;
        }

        // Calculate the ordinal position of the top menun item
        uint8_t top_idx = display_message_calculate_top_option_idx(page_start_idx, selected_idx, page_size, option_count);
        
        // Convert back to the menu item and set page_start_id
        int x;
        child_item = parent_item->first_child;
        for (x = 0; x <= top_idx; x++)
        {
            if (x == top_idx)
            {
                parent_item->display_data.page_start_id = child_item->menu_id;
                break;
            }
    
            child_item = child_item->next_sibling;
        }
        
        // Now rebuild the Option List
        // Refresh the screen data for the new page items
    logger_log_debugf(0, "DISPLAY MESSAGE display_message_option_list_select_item %d --", parent_item->menu_id);
        display_message_populate_screen_data(parent_item->menu_id, menu_item->menu_id);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param menu_id
 */
static uint16_t display_message_populate_screen_data(MENU_ID menu_id, MENU_ID selected_id)
{
    logger_log_debugf(0, "DISPLAY MESSAGE Populate Screen %d %s %d --", menu_id, app_menu_titles[menu_id], selected_id);
    uint16_t result = selected_id;
    
    MENU_ITEM* menu_item = menu_get_item(menu_id);

    // Declare a screen data buffer
    uint8_t screen_data[DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE];
    display_message_initialise_screen_data(screen_data);

    // Populate the screen data
    if (menu_item->display_data.build_screen_data != NULL)
    {
        result = menu_item->display_data.build_screen_data(menu_item->menu_id, screen_data, DISPLAY_MESSAGE_DISPLAY_TYPE_PARENT);
        
        bool selected = true;
        if (selected_id != menu_id)
        {
            result = selected_id;
            
            // Ignore the result and use the selected Id
            display_message_add_screen_data(screen_data, 
                     selected_id,
                     DISPLAY_MESSAGE_ATTRIBUTE_SELECTED, // Attribute ID
                     DISPLAY_MESSAGE_DATA_ENCODING_BOOLEAN,
                     (uint8_t*)&selected,
                     sizeof(bool));
        }
        else if (result != menu_id)
        {
            display_message_add_screen_data(screen_data, 
                    result,
                    DISPLAY_MESSAGE_ATTRIBUTE_SELECTED, // Attribute ID
                    DISPLAY_MESSAGE_DATA_ENCODING_BOOLEAN,
                    (uint8_t*)&selected,
                    sizeof(bool));
        }
    }

    if (result != DISPLAY_MESSAGE_DATA_ID_INVALID)
    {
        display_message_finalise_and_send_screen_data(screen_data);
    }
    else
    {
        logger_log_debugf(0, "DISPLAY MESSAGE Aborted Sending Screen %d %s %d --", menu_id, app_menu_titles[menu_id], selected_id);
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param menu_id
 */
static uint16_t display_message_display_and_populate(MENU_ID menu_id, MENU_ID selected_id)
{
    uint16_t result = menu_id;
    
    MENU_ITEM* menu_item = menu_get_item(menu_id);
    
    display_message_send_display_screen(
            (DISPLAY_MESSAGE_SCREEN_TYPE)menu_item->display_data.screen_type, 
            (DISPLAY_MESSAGE_SCREEN_ID)menu_item->menu_id, 
            app_menu_titles[menu_item->menu_id]);

    result = display_message_populate_screen_data(menu_id, selected_id);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Sends a DisplayScreen message
 * @param type
 * @param id
 */
static void display_message_send_display_screen(DISPLAY_MESSAGE_SCREEN_TYPE screen_type, DISPLAY_MESSAGE_SCREEN_ID screen_id, char* title)
{
    uint16_t title_len = strlen(title);
    uint8_t len = 10 + 2 + title_len + 2;  // Header + BIN16 len + text + Tail
    uint8_t buffer[len];
    //logger_log_debugf(0, "DISPLAY MESSAGE Display Screen Len %d %d --", len, title_len);
    
    buffer[0] = DISPLAY_MESSAGE_START_OF_PACKET_LSB;
    buffer[1] = DISPLAY_MESSAGE_START_OF_PACKET_MSB;
    buffer[2] = len;
    buffer[3] = 0;
    
    display_message_set_frame_and_packet_id(buffer, display_message_get_next_frame_id(), 0);

    buffer[6] = DISPLAY_MESSAGE_CONTROL_FLAGS_END_OF_MESSAGE; // Control Flags
    buffer[7] = DISPLAY_MESSAGE_TYPE_DISPLAY_SCREEN;
    buffer[8] = screen_type;
    buffer[9] = screen_id;
    
    uint8_t offset = 10;

    // Add length  of title to make a BIN16 field
    memcpy(&buffer[offset], &title_len, sizeof(uint16_t));
    offset += sizeof(uint16_t);
     
    // Add title
    memcpy(&buffer[offset], title, title_len);
    offset += title_len;

    buffer[offset] = DISPLAY_MESSAGE_END_OF_PAYLOAD_LSB;
    offset++;
    buffer[offset] = DISPLAY_MESSAGE_END_OF_PAYLOAD_MSB;
    
//    logger_log_debugf(0, "DISPLAY %d %d %d %d %d %d %d %d %d %d %d %d --",
//            buffer[0],
//           buffer[1],
//           buffer[2],
//           buffer[3],
//           buffer[4],
//           buffer[5],
//           buffer[6],
//           buffer[7],
//           buffer[8],
//            buffer[9],
//            buffer[10],
//           buffer[11],
//           buffer[12]
//  );
    
    logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_DISPLAY_SCREEN, "DISPLAY MESSAGE Displaying Screen %d %d %s --", screen_type, screen_id, title);

    rs232_display_send_buffer(buffer, len);
    
    display_message_initiate_handshake();
}

/**-----------------------------------------------------------------------------
 * This will set the page_start_id on the parent to the menu item that will be 
 * at the top of the page when a menu item is to be displayed.  The menu items Id should be used as the Select action id
 */
static void display_message_set_menu_page_start(MENU_ITEM* menu_item)
{
    MENU_ITEM* parent_item = menu_item->parent;

    // We have to work out what the top menu item is that needs to be displayed
    uint8_t count = 0;
    uint8_t selected_idx = 0xFF;
    uint8_t top_idx = 0;
    uint8_t page_size = max(parent_item->display_data.page_size, DISPLAY_MESSAGE_MAX_ROWS);
    
    // Find the ordinal position of the menu in its parent list
    MENU_ITEM* child_item = parent_item->first_child;
    MENU_ITEM* top_item = menu_item; 
    while(child_item != NULL)
    {
        if (child_item->menu_id == menu_item->menu_id)
        {
            selected_idx = count;
        }

        child_item = child_item->next_sibling;
        count++;
    }

    top_idx = (selected_idx / page_size) * page_size;

    if (top_idx + page_size > count)
    {
        top_idx = count - page_size;
    }

    // Work out the top item ensuring that a whole page is displayed
    child_item = parent_item->first_child;
    count = 0;
    while(child_item != NULL)
    {
        if (count == top_idx)
        {
            top_item = child_item;
            break;
        }
        child_item = child_item->next_sibling;
        count++;
    }

    // Select the page that contains the current item
    parent_item->display_data.page_start_id = top_item->menu_id;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used as a default for menu_add_build_screen_data_function that generates the
 * screen data based on the type of menu item (Option List, Edit)
 * @param menu_id
 */
void display_message_build_screen_data(MENU_ID menu_id)
{
    logger_log_debugf(0, "DISPLAY MESSAGE Build Screen %d %s --", menu_id, app_menu_titles[menu_id]);
    logger_process();
    //For Option Lists and Edit screens add app_commands_get_xxx_string() function 
    //where bespoke drawing has been used.
    //  This will replace the current app_commands_draw_xxx() function where that 
    //  has been used. (14 instances in SA30) 
    //  Augment app_commands_menu_value_change() to generate a ScreenData message.

    MENU_ITEM* menu_item = menu_get_item(menu_id);
    
    // Declare Screen Data
    
    // Check whether there is a bespoke function (for Currently Playing)
    if (menu_item->display_data.build_screen_data != NULL)
    {
        //SELECT - APP_COMMAND_MENU_NAVIGATION_OK
        //display_message_display_and_populate(menu_item->menu_id);

        uint16_t selected_id = display_message_populate_screen_data(menu_item->menu_id, menu_item->menu_id);

        if (selected_id != DISPLAY_MESSAGE_DATA_ID_INVALID)
        {
            // Send a Screen Action using the screen id as the selected item
            // The screen should select its default item
            display_message_send_screen_action(
                    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT,
                    selected_id);
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Checks the state of handshaking with the Display
 */
void display_message_check_handshake()
{
    if (( display_message_data.handshake_state == 1) && (timers_has_expired(TIMER_INDEX_DISPLAY_HANDSHAKE)))
    {
        logger_log_debugf(0, "DISPLAY HANDSHAKE Checking Handshake For Display Switch Off --");
        
        power_disable_by_index(POWER_INDEX_DISPLAY_PSU);

        timers_update_timeout(TIMER_INDEX_DISPLAY_HANDSHAKE, 1000, TIMERS_TIMEOUT_RESET_IMMEDIATE);
        
        display_message_data.handshake_state = 2;
    }
    else if (( display_message_data.handshake_state == 2) && (timers_has_expired(TIMER_INDEX_DISPLAY_HANDSHAKE)))
    {
        logger_log_debugf(0, "DISPLAY HANDSHAKE Checking Handshake For Display Restart --");
        
        power_enable_by_index(POWER_INDEX_DISPLAY_PSU);
        
        display_message_data.handshake_state = 0;
    }
}

/**-----------------------------------------------------------------------------
 * Called when the host receives a message from the display
 */
void display_message_complete_handshake()
{
    if (display_message_data.handshake_state <= 2)
    {
        logger_log_debugf(0, "DISPLAY HANDSHAKE Complete --");
        display_message_data.handshake_state = 3;
    }
}

/**-----------------------------------------------------------------------------
 * Generates the appropriate display_messages based on the current item, and 
 * navigation type
 * @param menu
 * @param navigation_type
 */
void display_message_display_screen(MENU_ID menu_id, DISPLAY_MESSAGE_SCREEN_NAVIGATION navigation_type)
{
    //logger_log_debugf(0, "DISPLAY MESSAGE Display Screen %d %d %s --", menu_id, navigation_type, app_menu_titles[menu_id]);
    logger_log_debugf(0, "DISPLAY MESSAGE Display Screen %d %d %s --", navigation_type, menu_id, app_menu_titles[menu_id]);
    logger_process();
    
    // Set the current navigation type
    display_message_data.current_navigation = navigation_type;
    
    MENU_ITEM* menu_item = menu_get_item(menu_id);
    MENU_ITEM* parent_item = menu_get_parent(menu_item);
    //MENU_ITEM* current_item = menu_get_current_item();

    //display_message_send_display_screen((DISPLAY_MESSAGE_SCREEN_TYPE)menu_item->menu_type, (DISPLAY_MESSAGE_SCREEN_ID)menu_item->menu_id, app_menu_titles[menu_item->menu_id]);
    
    switch (navigation_type)
    {
        case DISPLAY_MESSAGE_SCREEN_NAVIGATION_NONE:
        {
            // This shouldn't be called
            break;
        }
        
        case DISPLAY_MESSAGE_SCREEN_NAVIGATION_DIRECT:
        {
            //DIRECT - APP_COMMAND_MENU_NAVIGATION_NONE
            // We will always display the screen as if it is the parent

            if ((menu_item->display_data.screen_type == DISPLAY_MESSAGE_SCREEN_TYPE_OPTION_LIST) ||
                (menu_item->display_data.screen_type == DISPLAY_MESSAGE_SCREEN_TYPE_OPTION_LIST_WITH_DATA))
            {
                uint16_t selected_id = menu_id;
                
                if (parent_item != NULL)
                {
                    display_message_set_menu_page_start(menu_item);

                    display_message_display_and_populate(parent_item->menu_id, selected_id);
                }
                else
                {
                    selected_id = display_message_display_and_populate(menu_item->menu_id, selected_id);
                }
                
                // This will have selected the default item under the current menu
                // For the input selection it would be good to select the current input by default

                // Send a Screen Action using the screen id as the selected item
                // The screen should select its default item
                display_message_send_screen_action(
                        DISPLAY_MESSAGE_SCREEN_ACTION_SELECT,
                        selected_id);//menu_id);
            }
            else
            {
                uint16_t selected_id = display_message_display_and_populate(menu_item->menu_id, menu_id);

                // Send a Screen Action using the screen id as the selected item
                // The screen should select its default item
                display_message_send_screen_action(
                        DISPLAY_MESSAGE_SCREEN_ACTION_SELECT,
                        selected_id);
            }

            break;
        }
        
        case DISPLAY_MESSAGE_SCREEN_NAVIGATION_BACK:
        {
            if (parent_item != NULL)
            {
                uint16_t selected_id = display_message_display_and_populate(parent_item->menu_id, menu_id);

                // Send a Screen Action using the screen id as the selected item
                // The screen should select its default item
                display_message_send_screen_action(
                        DISPLAY_MESSAGE_SCREEN_ACTION_SELECT,
                        selected_id);
            }
            
            break;
        }
        
        // Display a menu item in its parents list of children with the menu item selected
        case DISPLAY_MESSAGE_SCREEN_NAVIGATION_SELECT:
        {
            display_message_set_menu_page_start(menu_item);
                
            // Check current Menu edit state and either draw the Parent (if Menu) or the item (if MENU EDIT))
            display_message_display_and_populate(parent_item->menu_id, menu_id);

            // Send a Screen Action using the screen id as the selected item
            // The screen should select its default item
            display_message_send_screen_action(
                    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT,
                    menu_id);

            break;
        }
        
        case DISPLAY_MESSAGE_SCREEN_NAVIGATION_SELECT_EDIT:
        {
            //SELECT - APP_COMMAND_MENU_NAVIGATION_OK
            //menu_item->display_data.page_start = NULL;
            menu_item->display_data.page_start_id = -1;
            //parent_item->display_data.option_string_start = 0;
            
            uint16_t selected_id = display_message_display_and_populate(menu_item->menu_id, menu_id);

            // Send a Screen Action using the screen id as the selected item
            // The screen should select its default item
            display_message_send_screen_action(
                    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT,
                    selected_id);

            break;
        }
        
        case DISPLAY_MESSAGE_SCREEN_NAVIGATION_NEXT:
        {
            //NEXT/PREVIOUS - APP_COMMAND_MENU_NAVIGATION_LEFT/RIGHT
            logger_log_debugf(0, "DISPLAY MESSAGE NEXT Screen Type %d ", menu_item->display_data.screen_type);

            uint16_t selected_id = menu_item->menu_id;
            if ((parent_item != NULL) && (parent_item->default_item != NULL))
            {

                selected_id = parent_item->default_item->menu_id;
            }
            
            // Check if we need to regenerate the screen data for scrolling
            switch (menu_item->display_data.screen_type)
            {
                case DISPLAY_MESSAGE_SCREEN_TYPE_OPTION_LIST:
                case DISPLAY_MESSAGE_SCREEN_TYPE_OPTION_LIST_WITH_DATA:
                {
                    logger_log_debugf(0, "DISPLAY MESSAGE start_move_next OL %s ", menu_get_title(menu_item->menu_id));
                    if (parent_item != NULL)
                    {
                        selected_id = display_message_option_list_select_item(menu_item);
                    }
                    break;
                }

                case DISPLAY_MESSAGE_SCREEN_TYPE_OPTION:
                {
                    logger_log_debugf(0, "DISPLAY MESSAGE start_move_next O %s ", menu_get_title(menu_item->menu_id));
                    selected_id = display_message_option_list_select_item(menu_item);
                    //display_message_option_page_start_move_next(menu_item);
                        break;
                }

                case DISPLAY_MESSAGE_SCREEN_TYPE_MESSAGE:
                {
                    logger_log_debugf(0, "DISPLAY MESSAGE start_move_next M %s ", menu_get_title(menu_item->menu_id));
                    selected_id = display_message_option_list_select_item(menu_item);
                    //display_message_option_page_start_move_next(menu_item);
                        break;
                }

                case DISPLAY_MESSAGE_SCREEN_TYPE_UNKNOWN:
                {
                    logger_log_debugf(0, "DISPLAY MESSAGE start_move_next U %s ", menu_get_title(menu_item->menu_id));
                    selected_id = display_message_option_list_select_item(menu_item);
                    break;
                }
                
                default:
                    break;
            }
            logger_log_debugf(0, "DISPLAY MESSAGE start_move_next sel_id %d --", selected_id);
            logger_process();
            
            // Send a Screen Action using the selected id as the selected item
            display_message_send_screen_action(
                    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT,
                    selected_id);

            break;
        }
        
        case DISPLAY_MESSAGE_SCREEN_NAVIGATION_PREVIOUS:
        {
            //NEXT/PREVIOUS - APP_COMMAND_MENU_NAVIGATION_LEFT/RIGHT           
            uint16_t selected_id = menu_item->menu_id;
            if ((parent_item != NULL) && (parent_item->default_item != NULL))
            {
                selected_id = parent_item->default_item->menu_id;
            }
            
            // Check if we need to regenerate the screen data for scrolling
            switch (menu_item->display_data.screen_type)
            {
                case DISPLAY_MESSAGE_SCREEN_TYPE_OPTION_LIST:
                case DISPLAY_MESSAGE_SCREEN_TYPE_OPTION_LIST_WITH_DATA:
                {
                    logger_log_debugf(0, "DISPLAY MESSAGE start_move_previou OL %s ", menu_get_title(menu_item->menu_id));
                    selected_id = display_message_option_list_select_item(menu_item);
                    break;
                }

                case DISPLAY_MESSAGE_SCREEN_TYPE_OPTION:
                {
                    logger_log_debugf(0, "DISPLAY MESSAGE start_move_previou O %s ", menu_get_title(menu_item->menu_id));
                    selected_id = display_message_option_list_select_item(menu_item);
                    break;
                }

                case DISPLAY_MESSAGE_SCREEN_TYPE_MESSAGE:
                {
                    logger_log_debugf(0, "DISPLAY MESSAGE start_move_previou M %s ", menu_get_title(menu_item->menu_id));
                    selected_id = display_message_option_list_select_item(menu_item);
                    break;
                }

                case DISPLAY_MESSAGE_SCREEN_TYPE_UNKNOWN:
                {
                    logger_log_debugf(0, "DISPLAY MESSAGE start_move_previou U %s ", menu_get_title(menu_item->menu_id));
                    selected_id = display_message_option_list_select_item(menu_item);
                    break;
                }
                
                default:
                    break;
            }
            
            logger_log_debugf(0, "DISPLAY MESSAGE start_move_previou sel_id %d --", selected_id);
            // Send a Screen Action using the screen id as the selected item
            // The screen should select its default item
            display_message_send_screen_action(
                    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT,
                    selected_id);

            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Can be used in menu quit functions to determine how the menu is being navigated away from
 * @return 
 */
DISPLAY_MESSAGE_SCREEN_NAVIGATION display_message_get_current_navigation_type()
{
    return display_message_data.current_navigation;
}

/**-----------------------------------------------------------------------------
 * Gets the page size of the MENU_ITEM (set by menu_define_screen_data)
 * @param menu_item
 * @return 
 */
uint8_t display_message_get_page_size(MENU_ITEM* menu_item)
{
    uint8_t result = menu_item->display_data.page_size;
    
    if (result == 0)
    {
       result = DISPLAY_MESSAGE_MAX_ROWS;
    }
    
    return  result;
}
   
/**-----------------------------------------------------------------------------
 * Called whenever a message is sent to initiate handshaking if neccessary
 */
void display_message_initiate_handshake()
{
    if (display_message_data.handshake_state == 0)
    {
        logger_log_debugf(0, "DISPLAY HANDSHAKE Starting --");
        timers_reset(TIMER_INDEX_DISPLAY_HANDSHAKE);
        display_message_data.handshake_state = 1;
    }
}

/**-----------------------------------------------------------------------------
 * Sends a Screen Action message
 * @param type
 * @param id
 */
void display_message_send_screen_action(
        DISPLAY_MESSAGE_SCREEN_ACTION action, 
        DISPLAY_MESSAGE_DATA_ID screen_data_id)
{
    uint8_t len = 13;
    uint8_t buffer[len];
    
    buffer[0] = DISPLAY_MESSAGE_START_OF_PACKET_LSB;
    buffer[1] = DISPLAY_MESSAGE_START_OF_PACKET_MSB;
    buffer[2] = len;
    buffer[3] = 0;
    
    display_message_set_frame_and_packet_id(buffer, display_message_get_next_frame_id(), 0);

    buffer[6] = DISPLAY_MESSAGE_CONTROL_FLAGS_END_OF_MESSAGE; // Control Flags
    buffer[7] = DISPLAY_MESSAGE_TYPE_SCREEN_ACTION;
    buffer[8] = action;
    memcpy(&buffer[9], &screen_data_id, sizeof(DISPLAY_MESSAGE_DATA_ID));
    buffer[11] = DISPLAY_MESSAGE_END_OF_PAYLOAD_LSB;
    buffer[12] = DISPLAY_MESSAGE_END_OF_PAYLOAD_MSB;
    
    logger_log_debugf(LOGGER_CODE_DISPLAY_MESSAGE_SCREEN_ACTION, "DISPLAY MESSAGE Sending Screen Action %d %d --", action, screen_data_id);

    rs232_display_send_buffer(buffer, len);
    
}

/* *****************************************************************************
 End of File
 */
