/* ************************************************************************** */
/** Descriptive File Name

  @Company
 ARCAM

  @File Name
    display_test.c

  @Summary
    Brief description of the file.

  @Description
 Runs the display test sequence
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/display_test.h"
#include "app_context.h"
#include "app_config.h"
#include "utilities/display.h"
#include "utilities/logger.h"
#include "utilities/stateful_process.h"
#include "utilities/timers.h"

#include <string.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef enum
{
    DISPLAY_TEST_STATE_START = 0,
    DISPLAY_TEST_STATE_SHOW_MAKE = 1,
    DISPLAY_TEST_STATE_WAIT_MAKE = 2,
    DISPLAY_TEST_STATE_SHOW_MODEL = 3,
    DISPLAY_TEST_STATE_WAIT_MODEL = 4,
    DISPLAY_TEST_STATE_MODEL_END = 5,
    DISPLAY_TEST_STATE_FULL_BLOCKS = 6,
    DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_LR = 7,
    DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_LR_END = 8,
    DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_RL = 9,
    DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_RL_END = 10,
    DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_LR = 11,
    DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_LR_END = 12,
    DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_RL = 13,
    DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_RL_END = 14,
    DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_LR = 15,
    DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_LR_END = 16,
    DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_RL = 17,
    DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_RL_END = 18,
    DISPLAY_TEST_STATE_SCROLL_BLOCKS_END = 19,
    DISPLAY_TEST_STATE_LOGO_WAIT = 20,
    DISPLAY_TEST_STATE_EXIT = 21,
    DISPLAY_TEST_STATE_ENDED = 22,
} DISPLAY_TEST_STATE;

typedef struct
{
    // The mode in which the test is running
    DISPLAY_TEST_MODE mode;
    
    // The current processing point in a phase
    int pos;
    
} DISPLAY_TEST_DATA;

static DISPLAY_TEST_DATA display_test_data;

static volatile bool redraw = false;  // only calculate screen display if something changed

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 */
static void display_test_display_full_line(DISPLAY_LINE line, DISPLAY_SPECIAL_CHARACTERS character)
{
    int line_len = display_get_line_length();
    int char_width = display_get_character_width();
    
    int x;
    for (x = 0; x < line_len; x++)
    {
        if (x < display_test_data.pos / char_width)
        {
            display_special_character(line, x, character);
            redraw = true;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Scrolls the suppplied character from left to right on the specified line
 */
static display_test_left_right(DISPLAY_LINE line, DISPLAY_SPECIAL_CHARACTERS character)
{
    int line_len = display_get_line_length();
    int char_width = display_get_character_width();

    if (display_test_data.pos <= line_len * char_width) // 5 = pixel width in a block
    {
        int x;
        for (x = 0; x < line_len; x++)
        {
            if (x < display_test_data.pos / char_width)
            {
                display_special_character(line, x, character);
                redraw = true;
            }
            else if (x == display_test_data.pos / char_width)
            {
                uint8_t diff = (display_test_data.pos % char_width);
                //display_special_character(line, x, character);
                display_special_character_offset(line, x, (char*)&character, 1, 1, 0, diff - char_width, 0);
                redraw = true;
                break;
            }
            else
            {
                break;
            }
        }

        display_write_line(line, false);
    }
}

/**-----------------------------------------------------------------------------
 * Display the MAKE/ Model logo with colons moving from the inside to the edges
 * Note: This works specifically with a Colon.  Other characters may require changes
 * to the steps, diff and offset variables.
 */
static void display_test_logo_wait()
{
    int line_len = display_get_line_length();
    int char_width = display_get_character_width();
    uint8_t steps = char_width + 2;
    uint8_t max_pos = steps * char_width;
    
    DISPLAY_SPECIAL_CHARACTERS character;
    
    switch (PRODUCT_MAKE)
    {
        case MAKE_ARCAM:
             character = DISPLAY_SPECIAL_CHARACTERS_COLON;
            break;
        case MAKE_JBL:
            character = DISPLAY_SPECIAL_CHARACTERS_EXCLAM;
            break;
    }
    
    display_test_data.pos %= max_pos;
    
    display_clear_line(DISPLAY_LINE_1);
    display_clear_line(DISPLAY_LINE_2);
   
    // Display MAKE / Model
    
    switch (PRODUCT_MAKE)
    {
        case MAKE_ARCAM:
            display_append_text(DISPLAY_LINE_1, (line_len - 5) / 2, APP_STRINGS[APP_STRINGS_ENUM_ARCAM]);
            break;
        case MAKE_JBL:
            display_append_text(DISPLAY_LINE_1, (line_len - 3) / 2, APP_STRINGS[APP_STRINGS_ENUM_JBL]);
            break;
    }
            
    
    display_append_text(DISPLAY_LINE_2, (line_len - strlen(SOFTWARE_MODEL_STRING)) / 2, "%s", SOFTWARE_MODEL_STRING);
   
    // Calculate offsets
    uint8_t diff = (display_test_data.pos % steps);
    
    uint8_t offset_r = 11 + (display_test_data.pos / steps);
    if (offset_r >= line_len)
    {
        offset_r = 11;
        display_test_data.pos += steps;
    }
    
    uint8_t offset_l = 4 - (display_test_data.pos / steps);
    if (offset_l >= line_len) // Roll over 0 - -1 results in 0xFF
    {
        offset_l = 4;
    }
    
    // Display colon moving to the left
    display_special_character_offset(DISPLAY_LINE_1, offset_l, (char*)&character, 1, 1, 0, 3 - diff, 0);
    
    // Display colon moving to the right
    display_special_character_offset(DISPLAY_LINE_1, offset_r, (char*)&character, 1, 1, 0, diff - 2, 0);
    if ((line_len < 15) && (diff == steps - 1))
    {
        // Start next position if the current one is about to be hidden
        display_special_character_offset(DISPLAY_LINE_1, offset_r + 1, (char*)&character, 1, 1, 0, -2, 0);
    }
    
    display_write_line(DISPLAY_LINE_1, false);

    display_special_character_offset(DISPLAY_LINE_2, offset_l, (char*)&character, 1, 1, 0, 3 - diff, 0);

    display_special_character_offset(DISPLAY_LINE_2, offset_r, (char*)&character, 1, 1, 0, diff - 2, 0);
    if ((line_len < 15) && (diff == steps - 1))
    {
        display_special_character_offset(DISPLAY_LINE_2, offset_r + 1, (char*)&character, 1, 1, 0, -2, 0);
    }
    
    display_write_line(DISPLAY_LINE_2, false);
    
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR LOGO WAIT %d %d %d -- ", display_test_data.pos, 11 + (display_test_data.pos / char_width), diff);
    redraw = true;
}

/**-----------------------------------------------------------------------------
 */
static display_test_right_left(DISPLAY_LINE line, DISPLAY_SPECIAL_CHARACTERS character)
{
    int line_len = display_get_line_length();
    int char_width = display_get_character_width();

    if (display_test_data.pos <= line_len * char_width) // 5 = pixel width in a block
    {
        int x;
        for (x = 0; x < line_len; x++)
        {
            if (x < display_test_data.pos / char_width)
            {
                display_special_character(line, line_len - 1 - x, character);
            }
            else if (x == display_test_data.pos / char_width)
            {
                uint8_t diff = (display_test_data.pos % char_width);
                display_special_character_offset(line, line_len - 1 - x, (char*)&character, 1, 1, 0, char_width - diff, 0);
                redraw = true;
                break;
            }
            else
            {
                break;
            }
        }

        display_write_line(line, false);
    }
}



/**-----------------------------------------------------------------------------
 * Reset pos and move the state on
 * @param timeout
 */
static void display_test_move_state(int timeout)
{
    display_test_data.pos = 0;
    timers_update_timeout(TIMER_INDEX_DISPLAY_TEST, timeout, TIMERS_TIMEOUT_RESET_IMMEDIATE);
    stateful_process_get_info()->state++;
    redraw = true;
}

/**-----------------------------------------------------------------------------
 * Check if timer has expired and moves on pos.  If pos == max_pos it then moves 
 * onto the next state, setting the timeout
 */
static display_test_test_wait(int max_pos, int timeout)
{
        // Draw Top line progress phase
    if (timers_has_expired(TIMER_INDEX_DISPLAY_TEST))
    {
        if (display_test_data.pos < max_pos) 
        {
            display_test_data.pos++;
            timers_reset(TIMER_INDEX_DISPLAY_TEST);
            redraw = true;
        }
        else
        {
            display_test_move_state(timeout);
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
 
/**-----------------------------------------------------------------------------
 * Draw the bespoke Display Test display
 */
void display_test_draw(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();
    int line_len = display_get_line_length();
    int line_height = display_get_character_height(); 
    int char_width = display_get_character_width();

    // Draw the top line
        // Clear the display
    display_clear(false);
    
    switch(stateful_process_get_info()->state)
    {
        case DISPLAY_TEST_STATE_START:
        {
            break;
        }
        
        case DISPLAY_TEST_STATE_SHOW_MAKE:
        {
            if (display_test_data.pos <= line_height - 1) 
            {
                int start_idx = (line_len - 5) / 2;
                int offset = (-1 * line_height) + display_test_data.pos + 1;
                
                uint8_t text_height = 2;
                uint8_t text_width = 5;
                                
                char text[10] = { DISPLAY_SPECIAL_CHARACTERS_A,
                                 DISPLAY_SPECIAL_CHARACTERS_R,
                                 DISPLAY_SPECIAL_CHARACTERS_C,
                                 DISPLAY_SPECIAL_CHARACTERS_A,
                                 DISPLAY_SPECIAL_CHARACTERS_M,
                                 DISPLAY_SPECIAL_CHARACTERS_SPACE,
                                 DISPLAY_SPECIAL_CHARACTERS_SPACE,
                                 DISPLAY_SPECIAL_CHARACTERS_SPACE,
                                 DISPLAY_SPECIAL_CHARACTERS_SPACE,
                                 DISPLAY_SPECIAL_CHARACTERS_SPACE,};
                
                switch (PRODUCT_MAKE)
                {
                    case MAKE_JBL:
                    text[0] = DISPLAY_SPECIAL_CHARACTERS_J;
                    text[1] = DISPLAY_SPECIAL_CHARACTERS_B;
                    text[2] = DISPLAY_SPECIAL_CHARACTERS_L;
                    text[3] = DISPLAY_SPECIAL_CHARACTERS_SPACE;
                    text[4] = DISPLAY_SPECIAL_CHARACTERS_SPACE;
                    text[5] = DISPLAY_SPECIAL_CHARACTERS_SPACE;
                    text[6] = DISPLAY_SPECIAL_CHARACTERS_SPACE;
                    text[7] = DISPLAY_SPECIAL_CHARACTERS_SPACE;
                    text[8] = DISPLAY_SPECIAL_CHARACTERS_SPACE;
                    text[9] = DISPLAY_SPECIAL_CHARACTERS_SPACE;
                    
                    start_idx = 6; // adjust JBL position
                }
                
                int offset_h = 0;
                display_special_character_offset(DISPLAY_LINE_1, start_idx, text, text_height, text_width, 0, offset_h, offset);
                display_special_character_offset(DISPLAY_LINE_1, start_idx + 1, text, text_height, text_width, 1, offset_h, offset);
                display_special_character_offset(DISPLAY_LINE_1, start_idx + 2, text, text_height, text_width, 2, offset_h, offset);
                display_special_character_offset(DISPLAY_LINE_1, start_idx + 3, text, text_height, text_width, 3, offset_h, offset);
                display_special_character_offset(DISPLAY_LINE_1, start_idx + 4, text, text_height, text_width, 4, offset_h, offset);
                
                
//                display_special_character_offset(DISPLAY_LINE_1, start_idx, &DISPLAY_SPECIAL_CHARACTERS_A, 1, 1, 0, 0, offset);
//                display_special_character_offset(DISPLAY_LINE_1, start_idx + 1, &DISPLAY_SPECIAL_CHARACTERS_R, 1, 1, 0, 0, offset);
//                display_special_character_offset(DISPLAY_LINE_1, start_idx + 2, &DISPLAY_SPECIAL_CHARACTERS_C, 1, 1, 0, 0, offset);
//                display_special_character_offset(DISPLAY_LINE_1, start_idx + 3, &DISPLAY_SPECIAL_CHARACTERS_A, 1, 1, 0, 0, offset);
//                display_special_character_offset(DISPLAY_LINE_1, start_idx + 4, &DISPLAY_SPECIAL_CHARACTERS_M, 1, 1, 0, 0, offset);

                display_write_line(DISPLAY_LINE_1, false);
            }
            
            display_clear_line(DISPLAY_LINE_2);
            display_write_line(DISPLAY_LINE_2, false);

            break;
        }
        
        case DISPLAY_TEST_STATE_SHOW_MODEL:
        {
            if (display_test_data.pos <= line_height - 1) 
            {
                uint8_t text_width = strlen(SOFTWARE_MODEL_STRING);
                int start_idx = (line_len - text_width) / 2;
                int offset = line_height - display_test_data.pos - 1;
                
                char text[text_width * 2];
                uint8_t text_height = 2;
                
                memset(text, DISPLAY_SPECIAL_CHARACTERS_SPACE, text_width * 2);
                memcpy(text + text_width, SOFTWARE_MODEL_STRING, text_width);
                
                int offset_h = 0;
                int x;
                for (x = 0; x < text_width; x++)
                {
                    display_special_character_offset(DISPLAY_LINE_2, start_idx + x, text, text_height, text_width, text_width + x, offset_h, offset);
                }

                display_write_line(DISPLAY_LINE_2, false);
            }
            
            break;
        }
        
        case DISPLAY_TEST_STATE_WAIT_MAKE:
        case DISPLAY_TEST_STATE_WAIT_MODEL:
        case DISPLAY_TEST_STATE_MODEL_END:
        {
            break;
        }
        
        case DISPLAY_TEST_STATE_FULL_BLOCKS:
        {
            int line_len = display_get_line_length();

            int x;
            for (x = 0; x < line_len; x++)
            {
                display_special_character(DISPLAY_LINE_1, x, DISPLAY_SPECIAL_CHARACTERS_BLOCK);
                display_special_character(DISPLAY_LINE_2, x, DISPLAY_SPECIAL_CHARACTERS_BLOCK);
            }

            display_write_line(DISPLAY_LINE_1, false);
            display_write_line(DISPLAY_LINE_2, false);

            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_LR:
        {
            
            // 
            display_test_left_right(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK);
            
            display_clear_line(DISPLAY_LINE_2);
            display_write_line(DISPLAY_LINE_2, false);

            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_LR_END:
        {
            display_test_display_full_line(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_RL:
        {
            display_test_right_left(DISPLAY_LINE_2, DISPLAY_SPECIAL_CHARACTERS_BLOCK);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_RL_END:
        {
            display_test_display_full_line(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_LR:
        {
            display_test_left_right(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_1);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_LR_END:
        {
            display_test_display_full_line(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_1);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_RL:
        {
            display_test_right_left(DISPLAY_LINE_2, DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_1);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_RL_END:
        {
            display_test_display_full_line(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_1);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_LR:
        {
            display_test_left_right(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_2);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_LR_END:
        {
            display_test_display_full_line(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_2);
            break;
        }
       
        case DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_RL:
        {
            display_test_right_left(DISPLAY_LINE_2, DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_2);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_RL_END:
        {
            display_test_display_full_line(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_2);
            break;
        }
        
        case DISPLAY_TEST_STATE_LOGO_WAIT:
        {
            display_test_logo_wait();
            break;
        }
        
        default:
            break;
    }
}

/**-----------------------------------------------------------------------------
 * Invoked when the display test ends in test mode
 */
void display_test_end()
{
    logger_log_testf(LOGGER_CODE_TEST_DISPLAY_TEST, "Complete --");
 }

/**-----------------------------------------------------------------------------
 * Runs the display test sequence
 * To be  used as an input to app_commands_start_process
 */
void display_test_run(MENU_ID menu_id)
{
   // Get the state from the Menu items PROCESS_INFO
    switch (stateful_process_get_info()->state)
    {
        case DISPLAY_TEST_STATE_START:
        {
            logger_log_info(LOGGER_CODE_FACTORY_RESET, "DISPLAY TEST: Start --");
            display_test_data.pos = 0;
            
            timers_reset(TIMER_INDEX_DISPLAY_TEST_TIMEOUT);
            
            // Set scroll time to 100 ms
            timers_update_timeout(TIMER_INDEX_DISPLAY_TEST, 100, TIMERS_TIMEOUT_RESET_IMMEDIATE);

            //NOTE: Dont reset the mode during a process....until you've finished
            //app_context_set_mode(APP_CONTEXT_MODE_MENU);
#ifdef DISPLAY_MESSAGE_ACTIVE            
            app_context_set_menu(APP_COMMAND_MENU_DISPLAY_TESTING, DISPLAY_MESSAGE_SCREEN_NAVIGATION_DIRECT);
#else            
            app_context_set_menu(APP_COMMAND_MENU_DISPLAY_TESTING);
#endif            
            
            stateful_process_get_info()->state = 1;
            
            break;
        }

        case DISPLAY_TEST_STATE_SHOW_MAKE:
        case DISPLAY_TEST_STATE_SHOW_MODEL:
        {
            // Scroll ARCAM / Model and pause for 2 secs at the end of
            // each scrolling stage
            display_test_test_wait(display_get_character_height(), 2000);
            break;
        }

        case DISPLAY_TEST_STATE_WAIT_MAKE:
        {
            // Check that the ARCAM display section has completed then 
            // reset scrolling time to 100ms and move on
            if (timers_has_expired(TIMER_INDEX_DISPLAY_TEST))
            {
                display_test_move_state(100);
            }
            break;
        }

        // Pauses at end of changes
        case DISPLAY_TEST_STATE_WAIT_MODEL:
        case DISPLAY_TEST_STATE_FULL_BLOCKS:
        case DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_LR_END:
        case DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_RL_END:
        case DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_LR_END:
        case DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_RL_END:
        case DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_LR_END:
        case DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_RL_END:
        {
            // Reset pos and set off bottom line
            if (timers_has_expired(TIMER_INDEX_DISPLAY_TEST))
            {
                display_test_move_state(10);
            }
            
            break;
        }

        case DISPLAY_TEST_STATE_MODEL_END:
        {
            // End of ARCAM/Model display
            switch (display_test_data.mode)
            {
                case DISPLAY_TEST_MODE_FULL:
                {
                    display_test_test_wait(display_get_line_length() * 5, 4000);
                    break;
                }
                
                case DISPLAY_TEST_MODE_LOGO:
                {
                    display_test_data.pos = 0;
                    
                    timers_update_timeout(TIMER_INDEX_DISPLAY_TEST, 30, TIMERS_TIMEOUT_RESET_IMMEDIATE);
                    stateful_process_get_info()->state = DISPLAY_TEST_STATE_LOGO_WAIT;
                    break;
                }
            }
            break;
        }

        // Change processing while blocks are scrolling across the screen
        case DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_LR:
        case DISPLAY_TEST_STATE_SCROLL_FULL_BLOCKS_RL:
        case DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_LR:
        case DISPLAY_TEST_STATE_SCROLL_GREY1_BLOCKS_RL:
        case DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_LR:
        case DISPLAY_TEST_STATE_SCROLL_GREY2_BLOCKS_RL:
        {
            // Wait for scrolling to finish then pause for 2 seconds
            display_test_test_wait(display_get_line_length() * 5, 2000);
            break;
        }
        
        case DISPLAY_TEST_STATE_SCROLL_BLOCKS_END:
        {
            if (timers_has_expired(TIMER_INDEX_DISPLAY_TEST))
            {
                stateful_process_get_info()->state = DISPLAY_TEST_STATE_EXIT;
            }
            break;
        }
        
        case DISPLAY_TEST_STATE_LOGO_WAIT:
        {
            // Reset pos and set off bottom line
            if (timers_has_expired(TIMER_INDEX_DISPLAY_TEST))
            {
                if (app_context_is_running())
                {
                    logger_log_debugf(0, "DISPLAY TEST Exiting application running");
                    stateful_process_get_info()->state = DISPLAY_TEST_STATE_EXIT;
                }
                else
                {
                    display_test_data.pos++;
                    redraw = true;
                    timers_reset(TIMER_INDEX_DISPLAY_TEST);
                }
            }
            else if (timers_has_expired(TIMER_INDEX_DISPLAY_TEST_TIMEOUT))
            {
                // Timing out because the device hasn't loaded properly
                // Note this is suspended when the Avatar is upgrading
                logger_log_debugf(0, "DISPLAY TEST Exiting has timed out");
                stateful_process_get_info()->state = DISPLAY_TEST_STATE_EXIT;
            }
            break;
        }
        
        case DISPLAY_TEST_STATE_EXIT:
        {
            // We've waited for the power up so now going back to the splash screen
            // Changing the mode finishes the process
            app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
#ifdef DISPLAY_MESSAGE_ACTIVE            
            app_context_set_menu(app_commands_get_default_screen(), DISPLAY_MESSAGE_SCREEN_NAVIGATION_DIRECT);
#else
            app_context_set_menu(APP_COMMAND_MENU_DEFAULT);
#endif
            
            stateful_process_end();
            
            // Reset the mode so it does a full test from the menus
            display_test_set_mode(DISPLAY_TEST_MODE_FULL);

            stateful_process_get_info()->state = DISPLAY_TEST_STATE_ENDED;
            
            logger_log_info(0,"DISPLAY TEST: End --");
            break;
        }
       
        case DISPLAY_TEST_STATE_ENDED:
        {
            // Do nothing waiting to end
            break;
        }
    }
    
    // Make sure we don't dim while displaying the test
    timers_reset(TIMER_INDEX_BRIGHTNESS);   
}

/**
 * Sets the mode for the display test - Note this is reset to DISPLAY_TEST_MODE_FULL
 * at the end of each run
 * @param mode
 */
void display_test_set_mode(DISPLAY_TEST_MODE mode)
{
    display_test_data.mode = mode;
}


 void set_redraw(void)
 {
     redraw = true;
 }
 
 bool get_redraw(void)
 {
     bool ret_val = false;
     
     if (redraw)
     {
         ret_val = true;
         redraw = false; // clear flag
     }
     return ret_val;
 }

 
 bool display_test_ended(void)
 {
     return (stateful_process_get_info()->state == DISPLAY_TEST_STATE_ENDED);
 }
 
/* *****************************************************************************
 End of File
 */
