/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    encoder.c

  @Summary
 Encapsulates rotary encoder functionality

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/encoder.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

 #define ENCODER_QUEUE_LEN	64

    /*
     * Core encoder data used internally
     */
    typedef struct ENCODER_DATA
    {
        volatile bool in_update;
        volatile uint8_t queue[ENCODER_QUEUE_LEN];
        volatile uint8_t queue_start;
        volatile uint8_t queue_end;

        volatile bool read_busy;

        ENCODER_STATE encoder;
        bool rotary_moved;
        bool wrap;

        uint8_t ucLastPhase;

        uint8_t ucLastRotReading;

        volatile bool a_value;
        volatile bool b_value;
    } ENCODER_DATA;
    
static ENCODER_DATA encoder_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Clear the rotary buffer
 */
static void clear_buffer(void)
{
	int x;
	
	while (encoder_data.in_update)
	{
		Nop();
	}
	encoder_data.in_update = true;
	
	for (x = 0; x < ENCODER_QUEUE_LEN; x++)
	{
		encoder_data.queue[x] = 0xFF;
	}
	
	encoder_data.in_update = false;
	encoder_data.queue_start = 0;
	encoder_data.queue_end = 0;
	
	encoder_data.ucLastPhase = 0;
}

/**-----------------------------------------------------------------------------
 * Review encode data to detect changes
 */
static void handle_sample(void)
{
	static uint8_t ucPulseCount = 10; // a value big enough to decrement by 4
	uint8_t ucPhase;
	volatile uint8_t ucPattern;

    // Get the phase
	ucPhase = encoder_data.queue[encoder_data.queue_start];
	encoder_data.queue[encoder_data.queue_start] = 0xFF; // remove the value from the buffer
	
    // Increment next start and wrap if at end
	if ( ++encoder_data.queue_start >= ENCODER_QUEUE_LEN )
	{
		encoder_data.queue_start = 0;
	}

	// examine bit pattern of last and current phases
	ucPattern = (encoder_data.ucLastPhase << 2) | ucPhase;
	
	switch (ucPattern)
	{
        case 2:     // 0010
        case 11:    // 1011
        case 13:    // 1101
        case 4:     // 0100
            // valid sequence for increment
            //if (++ucPulseCount == 14) // all 4 states detected
            if ((++ucPulseCount >= 12) && (ucPattern == 11)) // 2 states detected and at middle of indent
            {
                ucPulseCount = 10;
                if (encoder_data.encoder.current_value < encoder_data.encoder.max_value)
                {
                    encoder_data.encoder.current_value++;
                    encoder_data.rotary_moved = true;
                }
 
                encoder_data.encoder.delta++;
            }
            break;
            
        case 1:     // 0001
        case 7:     // 0111
        case 14:    // 1110
        case 8:     // 1000
            // valid sequence for decrement
            //if (--ucPulseCount == 6) // all 4 states detected
            if ((--ucPulseCount <= 8) && (ucPattern == 7)) // 2 states detected and at middle of indent
            {
                ucPulseCount = 10;
                if (encoder_data.encoder.current_value > encoder_data.encoder.min_value)
                {
                    encoder_data.encoder.current_value--;
                    encoder_data.rotary_moved = true;
                }

                encoder_data.encoder.delta--;
            }
            break;
            
        case 3:
        case 15:
            // stable sequence after missing a pulse
            // allow through if we've counted 2 transitions already
            if (ucPulseCount > 11)
            {
                if (encoder_data.encoder.current_value < encoder_data.encoder.max_value)
                {
                    encoder_data.encoder.current_value++;
                    encoder_data.rotary_moved = true;
                }

                encoder_data.encoder.delta++;
            }
            if (ucPulseCount < 9)
            {
                if (encoder_data.encoder.current_value > encoder_data.encoder.min_value)
                {
                    encoder_data.encoder.current_value--;
                    encoder_data.rotary_moved = true;
                }

                encoder_data.encoder.delta--;
            }
            ucPulseCount = 10;
            break;
            
        default:
            break;
	}
    
	encoder_data.ucLastPhase = ucPhase;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Copies current encoder state into supplied structure
 * Adapted from rotary.c get_rotary()
 * @param state
 * @return - an indication of whether the rotary has changed or not
 */
bool encoder_get_state(ENCODER_STATE* state)
{
	bool rotary_moved = encoder_data.rotary_moved;
	encoder_data.rotary_moved = false;
    
    if (encoder_data.wrap)
    {
        // handle wrap       
        if (encoder_data.encoder.current_value == encoder_data.encoder.min_value)
        {
            encoder_data.encoder.current_value = encoder_data.encoder.max_value - 1;
        }

        if (encoder_data.encoder.current_value == encoder_data.encoder.max_value)
        {
            encoder_data.encoder.current_value = encoder_data.encoder.min_value + 1;
        }
        
        // remove offsets 
        state->min_value = encoder_data.encoder.min_value;
        state->current_value = encoder_data.encoder.current_value - 1;
        state->max_value = encoder_data.encoder.max_value - 2;
    }
    else
    {
        state->min_value = encoder_data.encoder.min_value;
        state->current_value = encoder_data.encoder.current_value;
        state->max_value = encoder_data.encoder.max_value;
    }
    
    state->moved = rotary_moved;
    state->delta = encoder_data.encoder.delta;
    
	return rotary_moved;
}

/**-----------------------------------------------------------------------------
 * Get the current wrap state
 * @return 
 */
bool encoder_get_wrap()
{
    return encoder_data.wrap;
}

/**-----------------------------------------------------------------------------
 * Initialise the encoder
 */
void encoder_initialise()
{
    clear_buffer();
    	
	encoder_data.read_busy = false;
	
	encoder_data.a_value = PLIB_PORTS_PinGet(ENCODER_A_PORTS_ID,
                                             ENCODER_A_PORT_CHANNEL, ENCODER_A_PIN);;
	encoder_data.b_value = PLIB_PORTS_PinGet(ENCODER_B_PORTS_ID,
                                             ENCODER_B_PORT_CHANNEL, ENCODER_B_PIN);;
	
	encoder_data.encoder.current_value = 0;
	encoder_data.encoder.max_value = 100;
	encoder_data.encoder.min_value = 0;
	encoder_data.rotary_moved = false;
}
 
/**-----------------------------------------------------------------------------
* Perform encoder processing
 * 
* @return - true if samples processed, false if not
*/
bool encoder_process(void)
{
    encoder_data.encoder.delta = 0;
    
	// is rotary encoder being updated by the interrupt?
	if (!encoder_data.in_update)
	{
		encoder_data.in_update = true;
		
		while (encoder_data.queue_end != encoder_data.queue_start)
		{
			handle_sample();
		}
		
		encoder_data.in_update = false;
	}
}

/**-----------------------------------------------------------------------------
 * Adapted from rotary.c get_rotary()
 * @param state
 * @return - an indication of whether the rotary has changed or not
 */
bool encoder_process_state(ENCODER_STATE* state)
{
    bool result = false;
    
    // Timer setting for encoder
    //     X(ENCODER, 5, TIMERS_RESET_IMMEDIATE, TIMERS_HANDLE_100US)
    if (timers_has_expired(TIMER_INDEX_ENCODER))
    {
        encoder_read();
 
        // Update the encoder state
        encoder_process();

        // Update the current input context - Get the current encoder values
        encoder_get_state(state);
        
        timers_reset(TIMER_INDEX_ENCODER); 
        
        result = true;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Read the current encoder state from the ports
 */
void encoder_read(void)
{
	while (encoder_data.read_busy)
	{
		Nop();
	}
    
	encoder_data.read_busy = true;
	
	encoder_data.a_value = PLIB_PORTS_PinGet(ENCODER_A_PORTS_ID,
                                             ENCODER_A_PORT_CHANNEL, ENCODER_A_PIN);

	encoder_data.b_value = PLIB_PORTS_PinGet(ENCODER_B_PORTS_ID,
                                             ENCODER_B_PORT_CHANNEL, ENCODER_B_PIN);
	
    uint8_t current_reading;
	current_reading = (encoder_data.a_value) ? 1 : 0;
	current_reading <<= 1;
	
	current_reading |= (encoder_data.b_value) ? 1 : 0;
	

	if (current_reading != encoder_data.ucLastRotReading)
	{
		if (!encoder_data.in_update)
		{
			// semaphore during processing of encoder
			encoder_data.in_update = true;
											
			encoder_data.queue[encoder_data.queue_end] = current_reading;
			encoder_data.ucLastRotReading = current_reading;

			if (++encoder_data.queue_end >= ENCODER_QUEUE_LEN)
			{
				encoder_data.queue_end = 0; // wrap buffer end to buffer start
			}

			encoder_data.in_update = false;
		}
	}
	
	encoder_data.read_busy = false;
}

/**-----------------------------------------------------------------------------
 * Update the encoder state
 * @param state
 * @param wrap
 */
void encoder_set(ENCODER_STATE state, bool wrap)
{
    encoder_data.wrap = wrap;
    
    if (encoder_data.wrap)
    {
        // Add extra value at top and bottom of range.
        // Low limit may be 0, so shift top limit up by 2 and
        // the reading up by 1.
        
        encoder_data.encoder.min_value = state.min_value;
        encoder_data.encoder.current_value = state.current_value + 1;
        encoder_data.encoder.max_value = state.max_value + 2;
    }
    else
    {
        encoder_data.encoder = state;
    }
}

/* *****************************************************************************
 End of File
 */
