/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    encoder_multi.c

  @Summary
 Encapsulates rotary encoder functionality

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/encoder_multi.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define ENCODER_QUEUE_LEN	64

/*
 * Core encoder data used internally
 */
typedef struct ENCODER_DATA
{
    volatile bool in_update;
    volatile uint8_t queue[ENCODER_QUEUE_LEN];
    volatile uint8_t queue_start;
    volatile uint8_t queue_end;

    volatile bool read_busy;

    ENCODER_STATE encoder;
    bool rotary_moved;
    bool wrap;

    uint8_t ucLastPhase;

    uint8_t ucLastRotReading;

    volatile bool a_value;
    volatile bool b_value;

    ENCODER_PIN_DEFS encoder_pin_defs;
} ENCODER_DATA;
    
static ENCODER_DATA encoder_data[APP_ENCODERS_ENCODERS_MAX];
//static int actual_nr_encoders = APP_ENCODERS_ENCODERS_MAX; // default


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Clear the rotary buffer
 */
static void clear_buffer(int encoder_id)
{
	int x;
	
	while (encoder_data[encoder_id].in_update)
	{
		Nop();
	}
	encoder_data[encoder_id].in_update = true;
	
	for (x = 0; x < ENCODER_QUEUE_LEN; x++)
	{
		encoder_data[encoder_id].queue[x] = 0xFF;
	}
	
	encoder_data[encoder_id].in_update = false;
	encoder_data[encoder_id].queue_start = 0;
	encoder_data[encoder_id].queue_end = 0;
	
	encoder_data[encoder_id].ucLastPhase = 0;
}

/**-----------------------------------------------------------------------------
 * Review encode data to detect changes
 */
static void handle_sample(int encoder_id)
{
	static uint8_t ucPulseCount = 10; // a value big enough to decrement by 4
	uint8_t ucPhase;
	volatile uint8_t ucPattern;

    // Get the phase
    ucPhase = encoder_data[encoder_id].queue[encoder_data[encoder_id].queue_start];
    encoder_data[encoder_id].queue[encoder_data[encoder_id].queue_start] = 0xFF; // remove the value from the buffer
	
    // Increment next start and wrap if at end
	if ( ++encoder_data[encoder_id].queue_start >= ENCODER_QUEUE_LEN )
	{
		encoder_data[encoder_id].queue_start = 0;
	}

	// examine bit pattern of last and current phases
	ucPattern = (encoder_data[encoder_id].ucLastPhase << 2) | ucPhase;
	
	switch (ucPattern)
	{
        case 2:     // 0010
        case 11:    // 1011
        case 13:    // 1101
        case 4:     // 0100
            // valid sequence for increment
            //if (++ucPulseCount == 14) // all 4 states detected
            if ((++ucPulseCount >= 12) && (ucPattern == 11)) // 2 states detected and at middle of indent
            {
                ucPulseCount = 10;
                if (encoder_data[encoder_id].encoder.current_value < encoder_data[encoder_id].encoder.max_value)
                {
                    encoder_data[encoder_id].encoder.current_value++;
                    encoder_data[encoder_id].rotary_moved = true;
                }
 
                encoder_data[encoder_id].encoder.delta++;
            }
            break;
            
        case 1:     // 0001
        case 7:     // 0111
        case 14:    // 1110
        case 8:     // 1000
            // valid sequence for decrement
            //if (--ucPulseCount == 6) // all 4 states detected
            if ((--ucPulseCount <= 8) && (ucPattern == 7)) // 2 states detected and at middle of indent
            {
                ucPulseCount = 10;
                if (encoder_data[encoder_id].encoder.current_value > encoder_data[encoder_id].encoder.min_value)
                {
                    encoder_data[encoder_id].encoder.current_value--;
                    encoder_data[encoder_id].rotary_moved = true;
                }

                encoder_data[encoder_id].encoder.delta--;
            }
            break;
            
        case 3:
        case 15:
            // stable sequence after missing a pulse
            // allow through if we've counted 2 transitions already
            if (ucPulseCount > 11)
            {
                if (encoder_data[encoder_id].encoder.current_value < encoder_data[encoder_id].encoder.max_value)
                {
                    encoder_data[encoder_id].encoder.current_value++;
                    encoder_data[encoder_id].rotary_moved = true;
                }

                encoder_data[encoder_id].encoder.delta++;
            }
            if (ucPulseCount < 9)
            {
                if (encoder_data[encoder_id].encoder.current_value > encoder_data[encoder_id].encoder.min_value)
                {
                    encoder_data[encoder_id].encoder.current_value--;
                    encoder_data[encoder_id].rotary_moved = true;
                }

                encoder_data[encoder_id].encoder.delta--;
            }
            ucPulseCount = 10;
            break;
            
        default:
            break;
	}
    
	encoder_data[encoder_id].ucLastPhase = ucPhase;
}


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Copies current encoder state into supplied structure
 * Adapted from rotary.c get_rotary()
 * @param state
 * @return - an indication of whether the rotary has changed or not
 */
bool encoder_get_state(ENCODER_STATE* state, int encoder_id)
{
	bool rotary_moved = encoder_data[encoder_id].rotary_moved;
	encoder_data[encoder_id].rotary_moved = false;
    
    if (encoder_data[encoder_id].wrap)
    {
        // handle wrap       
        if (encoder_data[encoder_id].encoder.current_value == encoder_data[encoder_id].encoder.min_value)
        {
            encoder_data[encoder_id].encoder.current_value = encoder_data[encoder_id].encoder.max_value - 1;
        }

        if (encoder_data[encoder_id].encoder.current_value == encoder_data[encoder_id].encoder.max_value)
        {
            encoder_data[encoder_id].encoder.current_value = encoder_data[encoder_id].encoder.min_value + 1;
        }
        
        // remove offsets 
        state->min_value = encoder_data[encoder_id].encoder.min_value;
        state->current_value = encoder_data[encoder_id].encoder.current_value - 1;
        state->max_value = encoder_data[encoder_id].encoder.max_value - 2;
    }
    else
    {
        state->min_value = encoder_data[encoder_id].encoder.min_value;
        state->current_value = encoder_data[encoder_id].encoder.current_value;
        state->max_value = encoder_data[encoder_id].encoder.max_value;
    }
    
    state->moved = rotary_moved;
    state->delta = encoder_data[encoder_id].encoder.delta;
    
	return rotary_moved;
}


/**-----------------------------------------------------------------------------
 * Get the current wrap state
 * @return 
 */
bool encoder_get_wrap(int encoder_id)
{
    return encoder_data[encoder_id].wrap;
}



/**-----------------------------------------------------------------------------
 * Initialise the encoder
 */
void encoder_initialise(ENCODER_PIN_DEFS encoder_pin_defs[], int nr_encoders)
{
    int encoder_nr;
    //actual_nr_encoders = nr_encoders;
    
    
//    test_encoder0.current_value = 0;
//    test_encoder0.max_value = 0;
//    test_encoder0.min_value = 0;
//    test_encoder0.moved = false;
//    test_encoder0.delta = 0;
//    
//    test_encoder1.current_value = 0;
//    test_encoder1.max_value = 0;
//    test_encoder1.min_value = 0;
//    test_encoder1.moved = false;
//    test_encoder1.delta = 0;
            
    
    
    for (encoder_nr = ENCODER_ID_MAIN ; encoder_nr < nr_encoders; encoder_nr++)
    {
        clear_buffer(encoder_nr);
    
    	
        encoder_data[encoder_nr].read_busy = false;
        
        encoder_data[encoder_nr].encoder_pin_defs.pin_A.index = encoder_pin_defs[encoder_nr].pin_A.index;
        encoder_data[encoder_nr].encoder_pin_defs.pin_A.channel = encoder_pin_defs[encoder_nr].pin_A.channel;
        encoder_data[encoder_nr].encoder_pin_defs.pin_A.bitPos = encoder_pin_defs[encoder_nr].pin_A.bitPos;
        
        encoder_data[encoder_nr].encoder_pin_defs.pin_B.index = encoder_pin_defs[encoder_nr].pin_B.index;
        encoder_data[encoder_nr].encoder_pin_defs.pin_B.channel = encoder_pin_defs[encoder_nr].pin_B.channel;
        encoder_data[encoder_nr].encoder_pin_defs.pin_B.bitPos = encoder_pin_defs[encoder_nr].pin_B.bitPos;
        
        encoder_data[encoder_nr].a_value = PLIB_PORTS_PinGet(
                encoder_data[encoder_nr].encoder_pin_defs.pin_A.index,
                encoder_data[encoder_nr].encoder_pin_defs.pin_A.channel,
                encoder_data[encoder_nr].encoder_pin_defs.pin_A.bitPos);
        
        encoder_data[encoder_nr].a_value = PLIB_PORTS_PinGet(
                encoder_data[encoder_nr].encoder_pin_defs.pin_B.index,
                encoder_data[encoder_nr].encoder_pin_defs.pin_B.channel,
                encoder_data[encoder_nr].encoder_pin_defs.pin_B.bitPos);

        encoder_data[encoder_nr].encoder.current_value = 0;
        encoder_data[encoder_nr].encoder.max_value = 100;
        encoder_data[encoder_nr].encoder.min_value = 0;
        encoder_data[encoder_nr].rotary_moved = false;
    }
}
 
/**-----------------------------------------------------------------------------
* Perform encoder processing
 * 
* @return - true if samples processed, false if not
*/
bool encoder_process(int encoder_id)
{
    encoder_data[encoder_id].encoder.delta = 0;
    
	// is rotary encoder being updated by the interrupt?
	if (!encoder_data[encoder_id].in_update)
	{
		encoder_data[encoder_id].in_update = true;
		
		while (encoder_data[encoder_id].queue_end != encoder_data[encoder_id].queue_start)
		{
			handle_sample(encoder_id);
		}
		
		encoder_data[encoder_id].in_update = false;
	}
}


/**-----------------------------------------------------------------------------
 * Adapted from rotary.c get_rotary()
 * @param state
 * @return - an indication of whether the rotary has changed or not
 */
bool encoder_process_state(ENCODER_STATE state[APP_ENCODERS_ENCODERS_MAX], uint8_t actual_nr_encoders)
{
    bool result = false;
    int encoder_nr;
    
    // Timer setting for encoder
    //     X(ENCODER, 5, TIMERS_RESET_IMMEDIATE, TIMERS_HANDLE_100US)
    if (timers_has_expired(TIMER_INDEX_ENCODER))
    {
        for (encoder_nr = ENCODER_ID_MAIN; encoder_nr < actual_nr_encoders; encoder_nr++)
        {
            encoder_read(encoder_nr);
        }
 
        // Update the encoder states
        for (encoder_nr = ENCODER_ID_MAIN; encoder_nr < actual_nr_encoders; encoder_nr++)
        {
            encoder_process(encoder_nr);
        }
        
        
        // Update the current input context - Get the current encoder values
        
                
        for (encoder_nr = ENCODER_ID_MAIN; encoder_nr < actual_nr_encoders; encoder_nr++)
        {
            encoder_get_state(&state[encoder_nr], encoder_nr);
        }
        
        timers_reset(TIMER_INDEX_ENCODER); 
        
        result = true;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Read the current encoder state from the ports
 */
void encoder_read(int encoder_id)
{
	while (encoder_data[encoder_id].read_busy)
	{
		Nop();
	}
    
	encoder_data[encoder_id].read_busy = true;
    
//    
//    if (encoder_id == 1)
//    {
//        a_val =  PLIB_PORTS_PinGet(ENCODER_A_PORTS_ID, ENCODER_A_PORT_CHANNEL, ENCODER_A_PIN); 
//        b_val =  PLIB_PORTS_PinGet(ENCODER_B_PORTS_ID, ENCODER_B_PORT_CHANNEL, ENCODER_B_PIN); 
//        c_val =  PLIB_PORTS_PinGet(INPUT_ENCODER_A_PORTS_ID, INPUT_ENCODER_A_PORT_CHANNEL, INPUT_ENCODER_A_PIN); 
//        d_val =  PLIB_PORTS_PinGet(INPUT_ENCODER_B_PORTS_ID, INPUT_ENCODER_B_PORT_CHANNEL, INPUT_ENCODER_B_PIN); 
//        
//        int log_val = 0;
//        log_val |= a_val ? 1 : 0;
//        log_val |= b_val ? 2 : 0;
//        log_val |= c_val ? 4 : 0;
//        log_val |= d_val ? 8 : 0;
//        
//        
//        
////        if ((a_val != last_a_val) ||
////            (b_val != last_b_val) ||
////            (c_val != last_c_val) ||
////            (d_val != last_d_val))
//            
//        // test port c    
//        
//        if (d_val != last_d_val)    
//        {
//            // something changed - update log
//            
//            if (encoder_log_index < 1000)
//            {
//                encoder_log[encoder_log_index] = log_val;
//                encoder_log_index++;
//            }
//        }
//                
//                
//        
//        last_a_val = a_val;
//        last_b_val = b_val;
//        last_c_val = c_val;
//        last_d_val = d_val;
//        
//        
//    }
//    
    encoder_data[encoder_id].a_value = PLIB_PORTS_PinGet(
                encoder_data[encoder_id].encoder_pin_defs.pin_A.index,
                encoder_data[encoder_id].encoder_pin_defs.pin_A.channel,
                encoder_data[encoder_id].encoder_pin_defs.pin_A.bitPos);
        
    encoder_data[encoder_id].b_value = PLIB_PORTS_PinGet(
                encoder_data[encoder_id].encoder_pin_defs.pin_B.index,
                encoder_data[encoder_id].encoder_pin_defs.pin_B.channel,
                encoder_data[encoder_id].encoder_pin_defs.pin_B.bitPos);
	
		
    uint8_t current_reading;
	current_reading = (encoder_data[encoder_id].a_value) ? 1 : 0;
	current_reading <<= 1;
	
	current_reading |= (encoder_data[encoder_id].b_value) ? 1 : 0;
	

	if (current_reading != encoder_data[encoder_id].ucLastRotReading)
	{
		if (!encoder_data[encoder_id].in_update)
		{
			// semaphore during processing of encoder
			encoder_data[encoder_id].in_update = true;
											
			encoder_data[encoder_id].queue[encoder_data[encoder_id].queue_end] = current_reading;
			encoder_data[encoder_id].ucLastRotReading = current_reading;

			if (++encoder_data[encoder_id].queue_end >= ENCODER_QUEUE_LEN)
			{
				encoder_data[encoder_id].queue_end = 0; // wrap buffer end to buffer start
			}

			encoder_data[encoder_id].in_update = false;
		}
	}
	
	encoder_data[encoder_id].read_busy = false;
}


/**-----------------------------------------------------------------------------
 * Update the encoder state
 * @param state
 * @param wrap
 */
void encoder_set(ENCODER_STATE state, int encoder_id, bool wrap)
{
    encoder_data[encoder_id].wrap = wrap;
    
    if (encoder_data[encoder_id].wrap)
    {
        // Add extra value at top and bottom of range.
        // Low limit may be 0, so shift top limit up by 2 and
        // the reading up by 1.
        
        encoder_data[encoder_id].encoder.min_value = state.min_value;
        encoder_data[encoder_id].encoder.current_value = state.current_value + 1;
        encoder_data[encoder_id].encoder.max_value = state.max_value + 2;
    }
    else
    {
        encoder_data[encoder_id].encoder = state;
    }
}


/* *****************************************************************************
 End of File
 */
