/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
 PLIB version of I2C Slave
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\i2c_slave.h"
#include "app_config.h"
#include <stdbool.h>
#include "utilities\logger.h"

// Only include if I2C_SLAVE_TYPE_PLIB set in app_config.h
#if I2C_SLAVE_TYPE == I2C_SLAVE_TYPE_PLIB

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#ifndef I2C_SLAVE_NUM_MODULES
#define I2C_SLAVE_NUM_MODULES   1
#endif

#ifndef I2C_SLAVE_READ_BUFFER_SIZE
#define I2C_SLAVE_READ_BUFFER_SIZE   20
#endif

#ifndef I2C_SLAVE_WRITE_BUFFER_SIZE
#define I2C_SLAVE_WRITE_BUFFER_SIZE   20
#endif

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Handles a stop condition
 * @param module
 */
static void i2c_slave_handle_stop(I2C_SLAVE_RUNTIME_MODULE* module)
{
    // Check for message stops - Note this is called for every stop sent over the I2C bus so
    // we need to check that we are interested in  it
    if (PLIB_I2C_StopWasDetected(module->module_info.module_id))
    {
        //PLIB_I2C_SlaveClockHold(module->module_info.module_id);
        
        // Get ready to process the data we've read
        switch(module->state)
        {
            case I2C_SLAVE_STATE_RECEIVING:
            {
                module->state = I2C_SLAVE_STATE_RECEIVE_END;
                
                // Invoke the callback
                if (module->module_info.callback != NULL)
                {
                    I2C_SLAVE_EVENT event = 
                    {
                        .state = module->state,
                    };

                    module->module_info.callback(&event);
                }
                
                break;
            }
            
            case I2C_SLAVE_STATE_SENDING:
            {
                module->state = I2C_SLAVE_STATE_SEND_END;
                
                // Invoke the callback
                if (module->module_info.callback != NULL)
                {
                    I2C_SLAVE_EVENT event = 
                    {
                        .state = module->state,
                    };

                    module->module_info.callback(&event);
                }
                
                break;
            }
            
            default:
            {
                // Ignore it - its not for us..
                break;
            }
        }
        
        //PLIB_I2C_SlaveClockRelease(module->module_info.module_id);
        
        // Set the stare back to IDLE
        module->state = I2C_SLAVE_STATE_IDLE;
    }
}

/**-----------------------------------------------------------------------------
 */
static void i2c_slave_transmit_byte(I2C_SLAVE_RUNTIME_MODULE* module)
{
    if (PLIB_I2C_TransmitterByteWasAcknowledged(module->module_info.module_id)) 
    {
        module->state = I2C_SLAVE_STATE_SENDING;
        
        // Is there a byte to send - do the callback
        if (module->module_info.callback != NULL)
        {
            I2C_SLAVE_EVENT event = 
            {
                .data = NULL,
                .state = module->state,
            };

            module->module_info.callback(&event);

            if (event.data != NULL) 
            {
                // write next byte in buffer
                PLIB_I2C_TransmitterByteSend(module->module_info.module_id, *event.data);
            } 
            else
            {
                PLIB_I2C_TransmitterByteSend(module->module_info.module_id, 0);
            }
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Initialises the I2C Slave module
 */
void i2c_slave_initialise()
{
    logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE Initialise --");
    
    i2c_slave_initialise_data();

    // Set Desired baud rate
    //PLIB_I2C_BaudRateSet(module->module_id, 100000000, 100000);

    // Set the appropriate slave address (slave only)
    // Clear and enable interrupts before enabling the I2C module.
    /* Initialise slave interrupt */
    PLIB_INT_SourceFlagClear(INT_ID_0, INT_SOURCE_I2C_2_SLAVE);
    PLIB_INT_SourceEnable(INT_ID_0, INT_SOURCE_I2C_2_SLAVE);
    PLIB_INT_VectorPrioritySet(INT_ID_0, INT_VECTOR_I2C2_SLAVE, INT_PRIORITY_LEVEL1);
    PLIB_INT_VectorSubPrioritySet(INT_ID_0, INT_VECTOR_I2C2_SLAVE, INT_SUBPRIORITY_LEVEL0);

    /* Initialise fault interrupt */
    PLIB_INT_SourceFlagClear(INT_ID_0, INT_SOURCE_I2C_2_BUS);
    PLIB_INT_SourceEnable(INT_ID_0, INT_SOURCE_I2C_2_BUS);
    PLIB_INT_VectorPrioritySet(INT_ID_0, INT_VECTOR_I2C2_BUS, INT_PRIORITY_LEVEL1);
    PLIB_INT_VectorSubPrioritySet(INT_ID_0, INT_VECTOR_I2C2_BUS, INT_SUBPRIORITY_LEVEL0);
}
    
/**-----------------------------------------------------------------------------
 * Starts processing on the module
 */
DRV_HANDLE i2c_slave_open(uint8_t index)
{
    DRV_HANDLE result = DRV_HANDLE_INVALID;
    
    if (index < I2C_SLAVE_NUM_MODULES)
    {
        I2C_SLAVE_RUNTIME_MODULE* module = i2c_slave_get_module(index);

        if ((module != NULL) && (module->in_use))
        {
            // Configure General I2C Options
            I2C2CON = 0;
                   
            // This is commented out because we are trying to intercept both 0x38 and 0x05 messages
            // Put back now with ARCAM Avatar version which no longer use the 0x05 address
            PLIB_I2C_SlaveAddress7BitSet(module->module_info.module_id, module->module_info.address_7bit);
            PLIB_I2C_SlaveMask7BitSet(module->module_info.module_id, 0XFF);

            PLIB_I2C_GeneralCallDisable(module->module_info.module_id);
            
            // This was commented out because we are trying to intercept the 0x05 messages from the Avatar
            // Put back now with ARCAM Avatar version which no longer use the 0x05 address
            PLIB_I2C_ReservedAddressProtectEnable(module->module_info.module_id);
            
            //PLIB_I2C_SlaveClockStretchingDisable(module->module_info.module_id);
            PLIB_I2C_SlaveClockStretchingEnable(module->module_info.module_id);
            PLIB_I2C_SlaveAddressHoldDisable(module->module_info.module_id);
            PLIB_I2C_SlaveInterruptOnStartEnable(module->module_info.module_id);
            PLIB_I2C_SlaveInterruptOnStopEnable(module->module_info.module_id);
            
            // Enable the module
            PLIB_I2C_Enable(module->module_info.module_id);
            
            result = (DRV_HANDLE)index;
            
            logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE Open - Module %d --", index);
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_I2C, "I2C SLAVE Open - ERROR Module not added %d --", index);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_I2C, "I2C SLAVE Open - ERROR Module out of range %d --", index);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Processes I2C Slave interrupt messages.
 */
void i2c_slave_process_interrupt(uint8_t index)
{
    I2C_SLAVE_RUNTIME_MODULE* module = i2c_slave_get_module(index);
        
    // Check if there is data to read
    if (PLIB_I2C_ReceivedByteIsAvailable(module->module_info.module_id)) 
    {
        // Check to see if the byte is data or an address
        if (PLIB_I2C_SlaveDataIsDetected(module->module_info.module_id))
        {
            // Check whether we are sending or receiving data
            // Its requesting a read so we need to be sending data
            if (PLIB_I2C_SlaveReadIsRequested(module->module_info.module_id))
            {
                // This never appears to be called
                // Write Buffer from Slave to Master
                // if R_W set write the rest of the buffer to master
                i2c_slave_transmit_byte(module);
            } 
            else 
            {
                module->state = I2C_SLAVE_STATE_RECEIVING;
                
                // Were going to read data from the MAster
                char data = PLIB_I2C_ReceivedByteGet(module->module_info.module_id);
                
                if (module->module_info.callback != NULL)
                {
                    I2C_SLAVE_EVENT event = 
                    {
                        .data = &data,
                        .state = module->state,
                    };

                    module->module_info.callback(&event);
                }
            
                PLIB_I2C_SlaveClockRelease(module->module_info.module_id);
            }
        } 
        else 
        {
            // The data is an address so read it to clear the buffer
            char data = PLIB_I2C_ReceivedByteGet(module->module_info.module_id); 
            //logger_log_errorf(LOGGER_CODE_I2C, "I2C SLAVE Address 0X%02X ", data);  
            
            // Check whether were to send or receive data
            if (PLIB_I2C_SlaveReadIsRequested(module->module_info.module_id))
            {
                // Were sending data
                module->state = I2C_SLAVE_STATE_SEND_START;
            } 
            else 
            {
                // We are receiving data
                module->state = I2C_SLAVE_STATE_RECEIVE_START;
            }
            
            // Invoke the callback
            if (module->module_info.callback != NULL)
            {
                I2C_SLAVE_EVENT event = 
                {
                    .data = &data,
                    .state = module->state,
                };
                
                module->module_info.callback(&event);
            }

            if (module->state == I2C_SLAVE_STATE_SEND_START)
            {
                module->state = I2C_SLAVE_STATE_SENDING;
                
                // Send the first byte
                i2c_slave_transmit_byte(module);
            }
            
            PLIB_I2C_SlaveClockRelease(module->module_info.module_id);
        }
    } 
    else 
    {
        // There's no data to be read so see if we have anything to send to the Master
        // Data or an address
        // Data - check whether we are sending or receiving
        if (PLIB_I2C_SlaveReadIsRequested(module->module_info.module_id))
        {
            // Were sending so transmit whatever data we have left
            i2c_slave_transmit_byte(module);
        }
            
        // Adding this seemed to clear the block on ACKS - not yet sure if it is needed
        PLIB_I2C_SlaveClockRelease(module->module_info.module_id);
    }
         
    // Check for message stops - Note this is called for every stop sent over the I2C bus so
    // we need to check that we are interested in  it
    i2c_slave_handle_stop(module);

    // Clear the interrupts
    PLIB_INT_SourceFlagClear(INT_ID_0, INT_SOURCE_I2C_2_SLAVE);
}

#endif //I2C_SLAVE_TYPE

/* *****************************************************************************
 End of File
 */
