/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/i2c_slave.h"
#include <stdbool.h>
#include "app_config.h"
#include "utilities\logger.h"
#include "utilities\timers.h"

// Only include if I2C_SLAVE_TYPE_BIT_BASH set in app_config.h
#if I2C_SLAVE_TYPE == I2C_SLAVE_TYPE_BIT_BASH

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

// Forward declarations
static void i2c_slave_bit_bash_initialise_data(I2C_SLAVE_RUNTIME_MODULE* module);
static bool i2c_slave_bit_bash_read_scl(I2C_SLAVE_RUNTIME_MODULE* module);
static bool i2c_slave_bit_bash_read_sda(I2C_SLAVE_RUNTIME_MODULE* module);
static void i2c_slave_bit_bash_write_sda(I2C_SLAVE_RUNTIME_MODULE* module, bool data);

/**-----------------------------------------------------------------------------
 * Initialises the module data ready for reading/writing
 * @param module
 */
static void i2c_slave_bit_bash_initialise_data(I2C_SLAVE_RUNTIME_MODULE* module)
{
    module->bit_pos = 0;
    module->data = 0;
}

/**-----------------------------------------------------------------------------
 * Returns true if whole byte has been read (clock == 9)
 * @param module
 */
static bool i2c_slave_bit_bash_read_data(I2C_SLAVE_RUNTIME_MODULE* module)
{
    bool result = false;

    if (module->scl_change && module->scl_state)
    {
        // Rising edge of clock
        if (module->bit_pos < 8)
        {
            // Shift data by 1
            module->data = module->data << 1;

            // add the current value if the sda line 
            if (module->sda_state)
            {
                module->data += 1;
            }
        }
    }
    else if (module->scl_change && !module->scl_state)
    {
        // Falling edge of clock
        if (module->bit_pos == 8)
        {
            // Send ACK
            //Pull SDA Line low
            i2c_slave_bit_bash_write_sda(module, false);
        }
        else if (module->bit_pos == 9)
        {
            //Pull SDA Line High
            i2c_slave_bit_bash_write_sda(module, true);

            // Byte read is complete so set result to true so the data can be processed
            result = true;
        }
    }

    // increment the bit pos
    module->bit_pos++;

    return result;
}

/**-----------------------------------------------------------------------------
 * Reads the clock pin
 * @return 
 */
static bool i2c_slave_bit_bash_read_scl(I2C_SLAVE_RUNTIME_MODULE* module)
{
    bool result = false;
    
    if (module != NULL)
    {
        result = PLIB_PORTS_PinGet(
                module->module_info.bit_bash.scl_pin.port,
                module->module_info.bit_bash.scl_pin.channel,
                module->module_info.bit_bash.scl_pin.pin); 
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Reads the data pin
 * @return 
 */
static bool i2c_slave_bit_bash_read_sda(I2C_SLAVE_RUNTIME_MODULE* module)
{
    bool result = false;
    
    if (module != NULL)
    {
        result = PLIB_PORTS_PinGet(
                module->module_info.bit_bash.sda_pin.port,
                module->module_info.bit_bash.sda_pin.channel,
                module->module_info.bit_bash.sda_pin.pin); 
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Returns true if there has been a change on the clock line, then updates the clock state for the module
 * @param 
 * @return 
 */
static bool i2c_slave_bit_bash_review_scl(I2C_SLAVE_RUNTIME_MODULE* module)
{
    bool result = false;
    if (module->scl_state != i2c_slave_bit_bash_read_scl(module))
    {
        result = true;
        module->scl_state = !module->scl_state;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns true if there has been a change on the clock line, then updates the clock state for the module
 * @param 
 * @return 
 */
static bool i2c_slave_bit_bash_review_sda(I2C_SLAVE_RUNTIME_MODULE* module)
{
    bool result = false;
    if (module->sda_state != i2c_slave_bit_bash_read_sda(module))
    {
        result = true;
        module->sda_state = !module->sda_state;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns true if whole byte has been read (clock == 9)
 * Note: Should only be invoked on a falling clock edge
 * @param module
 */
static bool i2c_slave_bit_bash_write_data(I2C_SLAVE_RUNTIME_MODULE* module)
{
    bool result = false;
    
    if (module->scl_change && !module->scl_state)
    {
        if (module->bit_pos < 8)
        {
            bool data = module->data & (1 << (7 - module->bit_pos));
            if (module->sda_state != data)
            {
                i2c_slave_bit_bash_write_sda(module, data);
            }
        }

        module->bit_pos++;
    }
    else if (module->scl_change && !module->scl_state)
    {
        // Falling edge of clock
        if (module->bit_pos == 8)
        {
            // Send ACK
            //Pull SDA Line low
            i2c_slave_bit_bash_write_sda(module, false);
        }
        else if (module->bit_pos == 9)
        {
            //Pull SDA Line High - not needed as were moving to the next byte of data
            i2c_slave_bit_bash_write_sda(module, true);

            // Byte read is complete so set result to true so the data can be processed
            result = true;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param data
 */
static void i2c_slave_bit_bash_write_sda(I2C_SLAVE_RUNTIME_MODULE* module, bool data)
{
    if (module != NULL)
    {
//        PLIB_PORTS_PinWrite(
//                module->module_info.bit_bash.sda_pin.port,
//                module->module_info.bit_bash.sda_pin.channel,
//                module->module_info.bit_bash.sda_pin.pin,
//                data); 
        
        if (data)
        {
            // Set SDA to input to pull it high
            PLIB_PORTS_PinDirectionInputSet(module->module_info.bit_bash.sda_pin.port, 
                                            module->module_info.bit_bash.sda_pin.channel, 
                                            module->module_info.bit_bash.sda_pin.pin);

        }
        else
        {
            // Set SDA to output
            PLIB_PORTS_PinDirectionOutputSet(module->module_info.bit_bash.sda_pin.port, 
                                             module->module_info.bit_bash.sda_pin.channel, 
                                             module->module_info.bit_bash.sda_pin.pin);
            
            // Set SDA Low
            PLIB_PORTS_PinClear(module->module_info.bit_bash.sda_pin.port, 
                                module->module_info.bit_bash.sda_pin.channel, 
                                module->module_info.bit_bash.sda_pin.pin);     
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Initialises the I2C Slave module
 */
void i2c_slave_initialise()
{
    logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE BIT BASH Initialise --");
    
    i2c_slave_initialise_data();
}

/**-----------------------------------------------------------------------------
 * Starts processing on the module
 */
DRV_HANDLE i2c_slave_open(uint8_t index)
{
    DRV_HANDLE result = DRV_HANDLE_INVALID;
    
    if (index < I2C_SLAVE_NUM_MODULES)
    {
        I2C_SLAVE_RUNTIME_MODULE* module = i2c_slave_get_module(index);

        if ((module != NULL) && (module->in_use))
        {
//            // Configure General I2C Options
//            I2C2CON = 0;
//                   
//            // This is commented out because we are trying to intercept both 0x38 and 0x05 messages
//            // Put back now with ARCAM Avatar version which no longer use the 0x05 address
//            PLIB_I2C_SlaveAddress7BitSet(module->module_info.module_id, module->module_info.address_7bit);
//            PLIB_I2C_SlaveMask7BitSet(module->module_info.module_id, 0XFF);
//
//            PLIB_I2C_GeneralCallDisable(module->module_info.module_id);
//            
//            // This was commented out because we are trying to intercept the 0x05 messages from the Avatar
//            // Put back now with ARCAM Avatar version which no longer use the 0x05 address
//            PLIB_I2C_ReservedAddressProtectEnable(module->module_info.module_id);
//            
//            //PLIB_I2C_SlaveClockStretchingDisable(module->module_info.module_id);
//            PLIB_I2C_SlaveClockStretchingEnable(module->module_info.module_id);
//            PLIB_I2C_SlaveAddressHoldDisable(module->module_info.module_id);
//            PLIB_I2C_SlaveInterruptOnStartEnable(module->module_info.module_id);
//            PLIB_I2C_SlaveInterruptOnStopEnable(module->module_info.module_id);
//            
//            // Enable the module
//            PLIB_I2C_Enable(module->module_info.module_id);

    
            logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE Input - SCL %d %d %d %d --", index,module->module_info.bit_bash.scl_pin.port, 
                                            module->module_info.bit_bash.scl_pin.channel, 
                                            module->module_info.bit_bash.scl_pin.pin);
            logger_process();
            
            // Set SCL to input
            PLIB_PORTS_PinDirectionInputSet(module->module_info.bit_bash.scl_pin.port, 
                                            module->module_info.bit_bash.scl_pin.channel, 
                                            module->module_info.bit_bash.scl_pin.pin);

//            logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE CN - SCL %d %d %d %d --", index,module->module_info.bit_bash.scl_pin.port, 
//                                            module->module_info.bit_bash.scl_pin.channel, 
//                                            module->module_info.bit_bash.scl_pin.pin);
//            logger_process();
//
//            PLIB_PORTS_PinChangeNoticePerPortEnable(module->module_info.bit_bash.scl_pin.port, 
//                                            module->module_info.bit_bash.scl_pin.channel, 
//                                            module->module_info.bit_bash.scl_pin.pin);

            logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE Input - SDA %d %d %d %d --", index, module->module_info.bit_bash.sda_pin.port, 
                                            module->module_info.bit_bash.sda_pin.channel, 
                                            module->module_info.bit_bash.sda_pin.pin);
            logger_process();
            // Set SDA to input
            PLIB_PORTS_PinDirectionInputSet(module->module_info.bit_bash.sda_pin.port, 
                                            module->module_info.bit_bash.sda_pin.channel, 
                                            module->module_info.bit_bash.sda_pin.pin);

            logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE Open Drain - SDA %d %d %d %d --", index, module->module_info.bit_bash.sda_pin.port, 
                                            module->module_info.bit_bash.sda_pin.channel, 
                                            module->module_info.bit_bash.sda_pin.pin);
            logger_process();
            
            PLIB_PORTS_PinOpenDrainEnable(module->module_info.bit_bash.sda_pin.port, 
                                            module->module_info.bit_bash.sda_pin.channel, 
                                            module->module_info.bit_bash.sda_pin.pin);

//            logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE CN - SDA %d %d %d %d --", index, module->module_info.bit_bash.sda_pin.port, 
//                                            module->module_info.bit_bash.sda_pin.channel, 
//                                            module->module_info.bit_bash.sda_pin.pin);
//            logger_process();
//            
//            PLIB_PORTS_PinChangeNoticePerPortEnable(module->module_info.bit_bash.sda_pin.port, 
//                                            module->module_info.bit_bash.sda_pin.channel, 
//                                            module->module_info.bit_bash.sda_pin.pin);
            
            module->open = true;
            
            result = (DRV_HANDLE)index;
            
            logger_log_infof(LOGGER_CODE_I2C, "I2C SLAVE Open - Module %d --", index);
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_I2C, "I2C SLAVE Open - ERROR Module not added %d --", index);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_I2C, "I2C SLAVE Open - ERROR Module out of range %d --", index);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Processes I2C Slave interrupt messages.
 */
uint16_t test = 0;
void i2c_slave_process_interrupt(uint8_t index)
{
    if (index < I2C_SLAVE_NUM_MODULES)
    {
        I2C_SLAVE_RUNTIME_MODULE* module = i2c_slave_get_module(index);

        if ((module != NULL) && module->in_use && module->open)
        {
            // Update the change states
            module->scl_change = i2c_slave_bit_bash_review_scl(module);
            module->sda_change = i2c_slave_bit_bash_review_sda(module);

if (module->scl_change || module->sda_change)
{
logger_log_debugf(LOGGER_CODE_I2C, "I2C SLAVE i2c_slave_process_interrupt %d %d %d %d %d -- ", test++,  module->scl_change, module->scl_state, module->sda_change, module->sda_state);
//logger_process();
}

            if (!module->scl_change && !module->sda_change)
            {
                // Do nothing
            }
            else if (module->sda_change && !module->sda_state && module->scl_state)
            {
                // Start bit detected
logger_log_debugf(LOGGER_CODE_I2C, "I2C SLAVE BB - Start bit detected -- ");
                i2c_slave_bit_bash_initialise_data(module);
                
                module->state = I2C_SLAVE_STATE_READ_ADDRESS;
            }
            else if (module->sda_change && module->sda_state && module->scl_state)
            {
logger_log_debugf(LOGGER_CODE_I2C, "I2C SLAVE BB - Stop bit detected -- ");
                // Stop bit detected..
                switch (module->state)
                {
                    case I2C_SLAVE_STATE_RECEIVING:
                    {
                        module->state = I2C_SLAVE_STATE_RECEIVE_END;
                        
                        i2c_slave_invoke_callback(module);
                        
                        break;
                    }

                    case I2C_SLAVE_STATE_SENDING:
                    {
                        module->state = I2C_SLAVE_STATE_SEND_END;
                        
                        i2c_slave_invoke_callback(module);

                        break;
                    }

                    default:
                    {
                        break;
                    }
                }
                
                module->state = I2C_SLAVE_STATE_IDLE;
            }
            else
            {
                switch (module->state)
                {
                    case I2C_SLAVE_STATE_READ_ADDRESS:
                    {
                        // Note: i2c_slave_bit_bash_read_data is reading data even if it returns false...
                        if (i2c_slave_bit_bash_read_data(module))
                        {
                            // Time to process the data
                            // Check if we are interested in this address and whether it is a read or a write
                            if (module->module_info.address_7bit == (module->data >> 1))
                            {
                                // Yes we are interested
                                i2c_slave_bit_bash_initialise_data(module);
                                
                                //  Check read or write
                                if (module->data & 0x01)
                                {
    logger_log_debugf(LOGGER_CODE_I2C, "I2C SLAVE BB - Read Message %d -- ", module->data);
                                    // This is a read message
                                    module->state = I2C_SLAVE_STATE_SEND_START;

                                    // Invoke the callback
                                    i2c_slave_invoke_callback(module);

                                    module->state = I2C_SLAVE_STATE_SENDING;

                                    // Invoke the callback - this should populate the data with the first byte
                                    i2c_slave_invoke_callback(module);
                                    
                                    // At this point we are on the falling edge of the 9th clock pulse
                                    // Set up the sda line for the first bit of the first byte
                                    i2c_slave_bit_bash_write_data(module);
                                }
                                else
                                {
    logger_log_debugf(LOGGER_CODE_I2C, "I2C SLAVE BB - Write Message %d -- ", module->data);
                                    // This is a write message
                                    module->state = I2C_SLAVE_STATE_RECEIVE_START;

                                    i2c_slave_invoke_callback(module);

                                    module->state = I2C_SLAVE_STATE_RECEIVING;
                                }
                            }
                            else
                            {
                                // We're not interested in this address, wait for the stop bit
                                module->state = I2C_SLAVE_STATE_IDLE;
                            }
                        }
                        
                        break;
                    }
                
                    case I2C_SLAVE_STATE_RECEIVING:
                    {
                        if (i2c_slave_bit_bash_read_data(module))
                        {
                            i2c_slave_invoke_callback(module);
                        }
                        
                        break;
                    }
                
                    case I2C_SLAVE_STATE_SENDING:
                    {
                        // use module data and the bit pos to set the SDA line
                        if (i2c_slave_bit_bash_write_data(module))
                        {   
                            // Populate the data with the next byte
                            i2c_slave_invoke_callback(module);
                        }        
                        
                        break;
                    }
                
                    default:
                        break;
                }
            }
        }
    }   
}

#endif // I2C_SLAVE_TYPE

/* *****************************************************************************
 End of File
 */
