/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_config.h"
#include "utilities/i2c.h"
#include "utilities/input_control_analogue.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */
/**
 * Global data
 */
typedef struct
{
    //! I2C Client Id for identifying the module on the I2C bus
    I2C_CLIENT_ID i2c_client_id;
    
    //! An array of setting values 
    INPUT_CONTROL_I2C_REQUEST adc_settings[INPUT_CONTROL_NUM_ADC_SETTING];
    
    //! Keeps track of the number of settings that have been added 
    //  will be < INPUT_CONTROL_NUM_ADC_SETTING
    uint8_t adc_setting_count;
    
    //! Keeps track of the number of settings that have been applied
    //  will be <= adc_setting_count
    uint8_t adc_applied_setting_count;
    
    //! switch to determine whether settings have been applied (actually the process started))
    bool adc_settings_applied;
    
    //TODO - Remove when Pins working OK
    PHONO_INPUT phono_input;
    
    //! Callback to apply configuration for the ADC
    INPUT_CONTROL_ANALOGUE_CONFIGURE_CALLBACK configure_callback;
    
} INPUT_CONTROL_ANALOGUE_DATA;

static INPUT_CONTROL_ANALOGUE_DATA input_control_analogue_data =
{
    .configure_callback = NULL,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * 
 * @param event
 * @param context
 */
static void input_control_adc_callback(DRV_I2C_BUFFER_EVENT event,
                                           I2C_REQUEST_TYPE request_type,
                                           uintptr_t context)
{
    //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Analogue Callback %d -- ", event);

    switch(event)
    {
        case DRV_I2C_BUFFER_EVENT_COMPLETE:
        {
            if (context != NULL)
            {
                // Look at the context
                INPUT_CONTROL_I2C_REQUEST* request = (INPUT_CONTROL_I2C_REQUEST*)context;
                request->actioned = true;

                //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL ADC Callback addr:%x val:%x", request->address, request->value);

                // Invoke a callback proc
                if (request->callback != NULL)
                {
                    INPUT_CONTROL_I2C_REQUEST_CALLBACK callback = (INPUT_CONTROL_I2C_REQUEST_CALLBACK)request->callback;
                    callback(request);
                }
            }
            
            break;
        }
        
        case DRV_I2C_BUFFER_EVENT_ERROR:
        {
            if (context != NULL)
            {
                INPUT_CONTROL_I2C_REQUEST* request = (INPUT_CONTROL_I2C_REQUEST*)context;
                logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Analogue Callback Error addr:%x val:%x", request->address, request->value);
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Analogue Callback Error unknown context");
            }
            break;
        }
        
        default:
        {
            //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "Input Control Callback Other %d %d", (int)event, settings_data.state);
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Called when a setting has been applied
 * @param request
 */
static void input_control_adc_setting_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    input_control_analogue_data.adc_applied_setting_count++;
//    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "Input Control input_control_dac_setting_callback %d/%d ", 
//            input_control_data.dac_applied_setting_count,
//            input_control_data.dac_setting_count);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Sets pins for analogue direct on/off
 * @param on
 */
void input_control_adc_reset(bool on)
{
    // Check whether Analogue direct is in use
#if INPUT_CONTROL_PINS_ENABLED    
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL ADC Reset - Switching %s ", on == true ? "On" : "Off");

    PLIB_PORTS_PinWrite(INPUT_CONTROL_ADC_RESET_PORTS_ID, 
                    INPUT_CONTROL_ADC_RESET_PORT_CHANNEL, 
                    INPUT_CONTROL_ADC_RESET_PIN,
                    on);    
#endif
        
    if (on)
    {
        // Setup for a resend of the settings
        input_control_analogue_data.adc_settings_applied = false;
    }
}

/**-----------------------------------------------------------------------------
 * Used to define ADC Configuration values.
 * Note: This works at a register level rather than the individual settings
 * @param register_ident - the register identifier
 * @param value - th value to be written
 */
void input_control_add_adc_setting(uint8_t register_ident, uint8_t value)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Add ADC Setting addr: 0x%02X value: 0x%02X ", register_ident, value);
    
    INPUT_CONTROL_I2C_REQUEST* setting = &input_control_analogue_data.adc_settings[input_control_analogue_data.adc_setting_count];
    setting->address = register_ident;
    setting->value = value;
    setting->callback = input_control_adc_setting_callback;
    
    input_control_analogue_data.adc_setting_count++;
    
    //TOD For some reason when trying to log all of these this is causing the micro to keel over later - no idea why at the moment
    // So only logging the first few
    if (input_control_analogue_data.adc_setting_count < 6)
    {
        //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Add ADC Setting addr: 0x%02X value: 0x%02X ", setting->address, setting->value);
    }
}

/**-----------------------------------------------------------------------------
 * Starts the process of applying settings to the adc
 * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
 */
void input_control_apply_adc_settings()
{
    if (!input_control_analogue_data.adc_settings_applied)
    {
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Applying ADC Settings ");

        // Reset the ADC
        //input_control_adc_reset(INPUT_CONTROL_SWITCH_OFF);
        
        input_control_analogue_data.adc_applied_setting_count = 0;

        // Make all settings un-applied
        int x;
        for (x = 0; x < input_control_analogue_data.adc_setting_count; x++)
        {
            INPUT_CONTROL_I2C_REQUEST* setting = &input_control_analogue_data.adc_settings[x];

#if INPUT_CONTROL_I2C_ADC_ENABLED   

            //Trigger a register update
            // Note makes the assumption that the address/value are the first two
            // elements in the INPUT_CONTROL_DAC_SETTING structure
            if (i2c_transmit(input_control_analogue_data.i2c_client_id, 
                             setting,  
                             2 * sizeof(uint8_t), 
                             setting))
            {
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Apply ADC settings error");
            }
#endif      
        }  
        
        input_control_analogue_data.adc_settings_applied = true;
    }
}
    
/**-----------------------------------------------------------------------------
 * Configures the analogue module
 * @return 
 */
bool input_control_analogue_configure()
{
    if (input_control_analogue_data.configure_callback != NULL)
    {
        input_control_analogue_data.configure_callback();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Analogue Configuration callback not specified --");
    }
}
    
/**-----------------------------------------------------------------------------
 * True when the ADC has been configured
 * @return 
 */
bool input_control_analogue_configured()
{
    bool result = false;
    
    if (input_control_analogue_data.adc_setting_count > 0)
    {
        result = input_control_analogue_data.adc_applied_setting_count == input_control_analogue_data.adc_setting_count;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Initialise the input control module
 */
bool input_control_analogue_initialise()
{
    bool result = true;
    
    i2c_initialise();

    input_control_analogue_data.adc_setting_count = 0;
    input_control_analogue_data.adc_settings_applied = false;
    
    // read input_enabled, and analogue direct from settings
    uint8_t analogue_direct_settings[INPUT_CONTROL_NUM_SETTINGS];
    uint8_t input_enabled_settings[INPUT_CONTROL_NUM_SETTINGS];
    
    settings_get(SETTING_ANALOGUE_DIRECT, analogue_direct_settings);
    settings_get(SETTING_INPUT_ENABLED, input_enabled_settings);
    
    int x;
    for (x = 0; x < INPUT_CONTROL_NUM_SETTINGS; x++)
    {
        INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(x);
        definition->index = 0;
        definition->analogue_direct = analogue_direct_settings[x];
        definition->enabled = input_enabled_settings[x];
        
        logger_log_debugf(0, "INPUT CONTROL %d %s %s %s -- ", x, AUDIO_INPUT_STRINGS[x], analogue_direct_settings[x] == 0 ? "DIGITAL" : "ANALOGUE", input_enabled_settings[x] == 0 ? "DISABLED" : "ENABLED");
    }
    
    settings_get(SETTING_MM_MC, &input_control_analogue_data.phono_input);

#if INPUT_CONTROL_I2C_ADC_ENABLED   
     
    input_control_analogue_data.i2c_client_id = i2c_open(INPUT_CONTROL_ADC_I2C_MODULE, 
            INPUT_CONTROL_ADC_I2C_READ_ADDRESS,
            INPUT_CONTROL_ADC_I2C_WRITE_ADDRESS,
            input_control_adc_callback);
    
    if (I2C_INVALID_CLIENT_ID != input_control_analogue_data.i2c_client_id)
    {
        // Initialise the DAC
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL ADC I2C Open");
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL ADC I2C Open Error");
        result = false;
    }    
   
#endif
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Sets callbacks for the input_control_analogue module
 * @param channel
 */
void input_control_analogue_set_callbacks(INPUT_CONTROL_ANALOGUE_CONFIGURE_CALLBACK configure_callback)
{
    input_control_analogue_data.configure_callback = configure_callback;
}

/**-----------------------------------------------------------------------------
 * Sets analogue direct mod on/off
 * @param on
 */
bool input_control_get_analogue_direct()
{
    bool result = false;
    
#if INPUT_CONTROL_PINS_ENABLED    
    result = !PLIB_PORTS_PinGet(INPUT_CONTROL_ANALOGUE_BYPASS_PORTS_ID, 
                INPUT_CONTROL_ANALOGUE_BYPASS_PORT_CHANNEL, 
                INPUT_CONTROL_ANALOGUE_BYPASS_PIN);
        
    //logger_log_debugf(0, "INPUT CONTROL ANALOG DIRECT GET  %d ", result);

#endif
        
    return result;
}   
    
/**-----------------------------------------------------------------------------
 * Sets analogue direct setting for a given input channel
 * @param on
 */
bool input_control_get_input_analogue_direct(uint8_t index)
{
    return input_control_get_definition(index)->analogue_direct;
}

/**-----------------------------------------------------------------------------
 * Gets the number of enabled inputs
 * @return 
 */
static uint8_t input_control_get_num_enabled()
{
    uint8_t result = 0;
    
    // Note: Start from 1 as we ignore the None input
    int x;
    for (x = 1; x < INPUT_CONTROL_NUM_SETTINGS; x++)
    {
        INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(x);
        if (definition->enabled)
        {
            result++;
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the phono input to be MM or MC
 */
PHONO_INPUT input_control_get_phono_input()
{
    return min(input_control_analogue_data.phono_input, PHONO_INPUT_MAX - 1);
}

/**-----------------------------------------------------------------------------
 * Sets pins for analogue direct on/off
 * @param on
 */
void input_control_set_analogue_direct(bool on)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Setting Analogue Direct %s ", on == true ? "On" : "Off");

    // Check whether Analogue direct is in use
#if INPUT_CONTROL_PINS_ENABLED    
    PLIB_PORTS_PinWrite(INPUT_CONTROL_ANALOGUE_BYPASS_PORTS_ID, 
                    INPUT_CONTROL_ANALOGUE_BYPASS_PORT_CHANNEL, 
                    INPUT_CONTROL_ANALOGUE_BYPASS_PIN,
                    !on);    
#endif
}

/**-----------------------------------------------------------------------------
 * Sets analogue direct setting for a given input channel
 * @param on
 */
void input_control_set_input_analogue_direct(uint8_t index, bool on)
{
    INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(index);
    definition->analogue_direct = on;
    
    // Settings
    logger_log_debugf(0, "INPUT CONTROL Analogue Direct %d %d %s --", definition->index, on, AUDIO_INPUT_STRINGS[definition->index]);
    uint8_t analogue_direct_settings[INPUT_CONTROL_NUM_SETTINGS];
    settings_get(SETTING_ANALOGUE_DIRECT, analogue_direct_settings);
    analogue_direct_settings[definition->index] = on;
    settings_set(SETTING_ANALOGUE_DIRECT, analogue_direct_settings);
}

/**-----------------------------------------------------------------------------
 * Sets enabled setting for a given input channel
 * @param on
 */
void input_control_set_input_enabled(uint8_t index, bool on)
{
    // Check that were not setting them all off
    if (on || (input_control_get_num_enabled() > 1))
    {
        INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(index);
        definition->enabled = on;

        // Settings
        uint8_t input_enabled_settings[INPUT_CONTROL_NUM_SETTINGS];
        settings_get(SETTING_INPUT_ENABLED, input_enabled_settings);

        logger_log_debugf(0, "INPUT CONTROL input_control_set_input_enabled %d %d %d %d --", definition->index, index, on, input_enabled_settings[definition->index]);
        input_enabled_settings[definition->index] = on;
        settings_set(SETTING_INPUT_ENABLED, input_enabled_settings);
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Attempt to set available inputs to 0 ");
    }
}   
    
/**-----------------------------------------------------------------------------
 * Sets the phono input to be MM or MC
 * @param type
 */
void input_control_set_phono_input(PHONO_INPUT type)
{
    input_control_analogue_data.phono_input = type;
    
    bool switch0 = false;
    
    switch(type)
    {
        case PHONO_INPUT_MM:
            switch0 = false;
            break;
            
        case PHONO_INPUT_MC:
            switch0 = true;
            break;
            
        default:
            break;
    }

    PLIB_PORTS_PinWrite(INPUT_CONTROL_MM_MC_SELECT_ID, 
                INPUT_CONTROL_MM_MC_SELECT_PORT_CHANNEL, 
                INPUT_CONTROL_MM_MC_SELECT_PIN,
                switch0);  
    
    settings_set(SETTING_MM_MC, &input_control_analogue_data.phono_input);
}
    
/**-----------------------------------------------------------------------------
 * Switch to a specific input on the analog input control
 * @param definition - the definition of the control to switch to
 */
void input_control_switch_analog(INPUT_CONTROL_DEFINITION* definition)
{
    bool switch_0 = definition->value & 0x01;
    bool switch_1 = definition->value & 0x02;
    bool switch_2 = definition->value & 0x04;
    
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Switching Analogue Mux to %s (%d %d %d) --", AUDIO_INPUT_STRINGS[definition->index], switch_0, switch_1, switch_2);
    
    // Reapply the settings (just in case);
    input_control_apply_adc_settings();
    
#if INPUT_CONTROL_PINS_ENABLED
 
    PLIB_PORTS_PinWrite(INPUT_CONTROL_SELECT_0_PORTS_ID, 
                    INPUT_CONTROL_SELECT_0_PORT_CHANNEL, 
                    INPUT_CONTROL_SELECT_0_PIN,
                    switch_0);
 
    PLIB_PORTS_PinWrite(INPUT_CONTROL_SELECT_1_PORTS_ID, 
                    INPUT_CONTROL_SELECT_1_PORT_CHANNEL, 
                    INPUT_CONTROL_SELECT_1_PIN,
                    switch_1);
 
    PLIB_PORTS_PinWrite(INPUT_CONTROL_SELECT_2_PORTS_ID, 
                    INPUT_CONTROL_SELECT_2_PORT_CHANNEL, 
                    INPUT_CONTROL_SELECT_2_PIN,
                    switch_2);

    input_control_set_analogue_direct(definition->analogue_direct);
#endif    
}

/* *****************************************************************************
 End of File
 */
