/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/input_control_digital.h"
#include "app_config.h"
#include "system_definitions.h"
#include "utilities/avatar_data.h"
#include "utilities/hdmi_arc.h"
#include "utilities/i2c.h"
#include "utilities/input_control.h"
#include "utilities/input_control_defs.h"
#include "utilities/logger.h"
#include "utilities/mutes.h"
#include "utilities/settings.h"
#include "utilities/timers.h"
#include "utilities/volume.h"
#include <math.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**
 * Converts AKMN4113 Sample rates (Register 0x08 bits 4-7) to our internal sample rates
 */
#define SAMPLE_RATE_CONVERSIONS \
	X(0, SAMPLE_RATE_44100)\
	X(1, SAMPLE_RATE_NOTSUPPORTED)\
    X(2, SAMPLE_RATE_48000)\
   	X(3, SAMPLE_RATE_32000)\
    X(4, SAMPLE_RATE_NOTSUPPORTED)\
   	X(5, SAMPLE_RATE_NOTSUPPORTED)\
	X(6, SAMPLE_RATE_NOTSUPPORTED)\
    X(7, SAMPLE_RATE_NOTSUPPORTED)\
    X(8, SAMPLE_RATE_88200)\
    X(9, SAMPLE_RATE_NOTSUPPORTED)\
    X(10, SAMPLE_RATE_96000)\
    X(11, SAMPLE_RATE_NOTSUPPORTED)\
    X(12, SAMPLE_RATE_176400)\
    X(13, SAMPLE_RATE_NOTSUPPORTED)\
    X(14, SAMPLE_RATE_192000)\
    X(15, SAMPLE_RATE_NOTSUPPORTED)\
    
// Redefine the macro and use to create an array of int
#define X(a, b) b,
	static SAMPLE_RATE SAMPLE_RATE_CONVERSION_VALUES[] = { SAMPLE_RATE_CONVERSIONS };
#undef X

/**
 * Global data
 */
typedef struct
{
    //! I2C Client Id for identifying the module on the I2C bus
    I2C_CLIENT_ID i2c_client_id;
    
    //! The request structure used to gather data for any errors
    INPUT_CONTROL_I2C_REQUEST error_request;
   
    //! Keeps track of the state of the MUX reset.  
    bool in_reset;
    
    //! An array of setting values 
    INPUT_CONTROL_I2C_REQUEST mux_settings[INPUT_CONTROL_NUM_DIGITAL_SETTING];
    
    INPUT_CONTROL_I2C_REQUEST setting_request;
    
    //! Keeps track of the number of settings that have been added 
    //  will be < INPUT_CONTROL_NUM_DAC_SETTING
    uint8_t mux_setting_count;
    
    //! Keeps track of the number of settings that have been applied
    //  will be <= dac_setting_count
    uint8_t mux_applied_setting_count;
    
    //! switch to determine whether settings have been applied (actually the process started))
    bool mux_settings_applied;
    
    //! Set to true if error checking decides there are errors that need to be muted.
    bool invalid_signal;
    
    //! The last calculated digital sample rate
    SAMPLE_RATE sample_rate;
    
    //! The steady digital sample rate
    SAMPLE_RATE steady_sample_rate;
    
    //! The request structure used to gather data for the sample rate calculation
    INPUT_CONTROL_I2C_REQUEST sample_rate_request;
    
    //! The request structure used to switch inputs
    INPUT_CONTROL_I2C_REQUEST switch_request;
    
    // Required to ensure only one sample_rate request is active at a time
    bool calculating_sample_rate;
    
    // True when the SPDIF Rx is detected as being in error
    bool in_error;

    // 20190607 - Added to try and reduce the number of interrupts - not sure it
    // is doing anything
    bool in_interrupt;
    
    // 20190607 - Removed after cocking about with interrupts going into massive loops
    // Now handled by input_control_digital_process polling
    //bool check_errors;
    
    // Set to true if the SPDIF RX has just occurred.  It appears that the SPDIF RX
    // returns a NOT SUPPORTED sample rate when queried just after a switch
    bool post_switch;
    
    //! Callback to apply configuration for the SPDIF MUX
    INPUT_CONTROL_DIGITAL_CONFIGURE_CALLBACK configure_callback;
    
    bool needs_mute;
    bool needs_unmute;
    
} INPUT_CONTROL_DIGITAL_DATA;

static INPUT_CONTROL_DIGITAL_DATA input_control_digital_data =
{
    .in_reset = false,
    .invalid_signal = false,
    .error_request.read_state =  INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE,
    .calculating_sample_rate = false,
    .in_error = false,
    .in_interrupt = false,
    //.check_errors = false,
    .post_switch = false,
    .configure_callback = NULL,
    .needs_mute = false,
    .needs_unmute = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward Definitions
static bool input_control_spdif_rx_is_reset();
static void input_control_get_spdif_rx_error();
static void input_control_digital_mute_input();
static void input_control_set_spdif_rx_sample_rate(SAMPLE_RATE sample_rate);
static void input_control_switch_digital_callback(INPUT_CONTROL_I2C_REQUEST* request);

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * A separate call is made for each register read for the address
 * @param request
 */
static void input_control_calculate_digital_sample_rate_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX sample_rate_callback %d %d ---", request->index, request->read_state);
     
    switch (request->read_state)
    {
        case INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS:
        {
            //logger_log_info(LOGGER_CODE_INPUT_CONTROL, "Input Control INITIATE RECEIVE");
            request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA;
            
            break;
        }
        
        case INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA:
        {
            // Get the current value of the register
            if (input_control_digital_data.sample_rate != SAMPLE_RATE_CONVERSION_VALUES[request->value >> 4])
            {
                input_control_set_spdif_rx_sample_rate(SAMPLE_RATE_CONVERSION_VALUES[request->value >> 4]);
            
                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate 0x%02X 0x%02X %s --", request->value, request->value >> 4, SAMPLE_RATE_STRINGS[input_control_digital_data.sample_rate]);
                
                input_control_digital_data.invalid_signal = false;
                
                // Check for errors and trigger an error check if found
                if (request->value & 0x0F > 0x00)
                {
                    logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate errors found 0x%02X --", request->value - (request->value >> 4));
                    
                    // Set internal mute flag
                    input_control_digital_data.invalid_signal = true;
                    
                    // Check if the digital input is currently active
                    if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
                    {
                        logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate errors MUTING OUTPUTS ");
                        
                        // Mute all inputs
                        input_control_digital_mute_input();
                    }
                    
                    input_control_get_spdif_rx_error();
                }
                else
                {
                    logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate Un-muting Inputs --");
                    
                    input_control_digital_data.needs_unmute = true;
                    input_control_unmute_inputs(input_control_get_current_input_type()); // TODO Moved to Avatar Set States Response callback
                    // BUG 3825 
//                    output_control_unmute_outputs();
//                    input_control_mute_dac(false);
                }
            }
            
            // Inform the Avatar of the new sample rate
            if (input_control_digital_data.post_switch && (input_control_digital_data.sample_rate == SAMPLE_RATE_NOTSUPPORTED))
            {
                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL AVATAR Sample Rate suppressed --");
            }
            else
            {
                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL AVATAR Setting Sample Rate %d --", SAMPLE_RATE_VALUES[input_control_digital_data.sample_rate]);
                avatar_set_sample_rate2(SAMPLE_RATE_VALUES[input_control_digital_data.sample_rate]);

                input_control_digital_data.post_switch = false;
            }
            
            break;
        }
    }
    
    input_control_digital_data.calculating_sample_rate = false;
}

/**-----------------------------------------------------------------------------
 * Mutes the SPDIF MUX output when an invalid signal is detected
 */
static void input_control_digital_mute_input()
{
    input_control_mute_inputs();
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * @param request
 */
static void input_control_get_spdif_rx_error_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX input_control_error_callback %d %d --", request->index, request->read_state);
     
    switch (request->read_state)
    {
        case INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS:
        {
            //logger_log_info(LOGGER_CODE_INPUT_CONTROL, "Input Control INITIATE RECEIVE");
            request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA;
            
            break;
        }
        
        case INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA:
        {
            // Split up request.value which is the state of register 0x07 to determine 
            // the type of error..
            //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR %d --", request->value);
            bool retry = false;
            
            // Report on the error
            bool par = false;
            if ((request->value & 0x01) == 0x01)
            {
                par = true;
            }

            bool audion = false;
            if ((request->value & 0x02) == 0x02)
            {
                audion = true;
            }

            bool stc = false;
            if ((request->value & 0x04) == 0x04)
            {
                stc = true;
            }

            bool v = false;
            if ((request->value & 0x08) == 0x08)
            {
                //v = true;
                // See if we have channel status checking switch on (engineering menu))
                settings_get(SETTING_CHECK_CHANNEL_STATUS, &v);
                if (!v)
                {
                    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR Ignoring Channel Status error --");
                }
            }

            bool unlock = false;
            if ((request->value & 0x10) == 0x10)
            {
                unlock = true;
            }

            //CInt
            //if ((request->value & 0x20) == 0x20)
            //{
            //}

            // Auto
            bool auto_detect = false;
            if ((request->value & 0x40) == 0x40)
            {
                auto_detect = true;
            }

            //QINT
            //if ((request->value & 0x80) == 0x80)
            //{
            //}

//            logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR 0x%02X [par:%d aud:%d stc:%d v:%d, unlk:%d auto:%d] --",
//                    request->value, par, audion, stc, v, unlock, auto_detect);
                
            // First check that its not just a sample rate change (shouldn't be but...)
            if (par || audion || v || unlock || auto_detect)
            {
                if (par || audion || v || auto_detect)
                {
                    // Unlock errors can be ignored fo the moment - this is when thre is no signal
                    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR 0x%02X [par:%d aud:%d stc:%d v:%d, unlk:%d auto:%d] --",
                        request->value, par, audion, stc, v, unlock, auto_detect);
                }
    
                if (!unlock)
                {
                    input_control_set_spdif_rx_sample_rate(SAMPLE_RATE_NOTVALID);
                }
                else
                {
                    input_control_set_spdif_rx_sample_rate(SAMPLE_RATE_NOINPUT);
                }
                
                // Set the internal mute flag
                input_control_digital_data.invalid_signal = true;
                
                if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
                {
                    // Commented out as this error occurs when there is no input AND IT BCAM IRRITATION - JUST WORRIED THAT WE WILL NOW MISS IT.....
                    //logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Errors MUTING OUTPUTS --");
                    input_control_digital_data.needs_mute = true;
                    
                    input_control_digital_mute_input();
                    
                        // BUG 3825 Changed to output_control_mute_digital_outputs as Pete did not want th analogue outputs muting
//                    output_control_mute_digital_outputs();
//                    input_control_mute_dac(true);
                    
                    // Have another go
                    retry = true;
                    
                    input_control_digital_data.in_error = true;
                }
            }
            else if (stc)
            {
                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR Sample Rate ONLY --");
                
                // No errors so check if we currently think were muted and unmute
                input_control_digital_data.invalid_signal = false;
                
                // If were active then unmute outputs
                if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
                {
                    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Callback UNMUTING INPUTS ");

                    input_control_calculate_digital_sample_rate();
                         
                    input_control_digital_data.in_error = false;
                }                
            }
            else
            {
                // In theory there are no errors , sometimes the inputs have been muted
                /// by the interrupt even when it reports no erros so unmute
                if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
                {
                    //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX Callback NO ERRORS BUT UNMUTING INPUTS ANYWAY --");
                    
                    // Unmute the inputs
                    // input_control_unmute_inputs(INPUT_CONTROL_TYPE_DIGITAL); 

                    // Check for an invalid sample rate and retry if found TODO - Test
                    SAMPLE_RATE sample_rate = input_control_get_digital_sample_rate();
                    if (SAMPLE_RATE_VALUES[sample_rate] == 0)
                    {
                        input_control_calculate_digital_sample_rate();
                    }
                    else
                    {
                        // Unmute the inputs
                        input_control_unmute_inputs(INPUT_CONTROL_TYPE_DIGITAL); 
                    }
                }                
            }

            // Set the read state to inactive
            request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE;
            
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Searches the dac settings to see if we have one that matches the supplied address
 * @return 
 */
static INPUT_CONTROL_I2C_REQUEST* input_control_find_definition_for_address(uint8_t address)
{
    INPUT_CONTROL_I2C_REQUEST* result = NULL;
            
    // Find mux setting corresponding to the mux_reset address
    int x;
    for (x = 0; x < input_control_digital_data.mux_setting_count; x++)
    {
        INPUT_CONTROL_I2C_REQUEST* setting = &input_control_digital_data.mux_settings[x];
        if (setting->address == address)
        {
            result = setting;
            break;
        }
    }

    return result;
}
 
/**-----------------------------------------------------------------------------
 * Initiates call to DAC to get sample rate
 */
static void input_control_get_spdif_rx_error()
{
    if (!input_control_in_switch()) // Added to try and lower the number of errors during a switch
    {
        //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL input_control_get_spdif_rx_error %d %d --",
        //        input_control_spdif_rx_is_reset(), input_control_digital_data.error_request.read_state == INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE);

        if ((!input_control_spdif_rx_is_reset()) &&
            (input_control_digital_data.error_request.read_state == INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE))
        {
            // Note: address should have been defined
            input_control_digital_data.error_request.actioned = false;
            input_control_digital_data.error_request.value = 0x00;
            input_control_digital_data.error_request.read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;

    #if INPUT_CONTROL_I2C_DIGITAL_ENABLED 

            // Trigger a request for the source type (Audio/Non Audio))
            if (i2c_read(input_control_digital_data.i2c_client_id, 
                         &input_control_digital_data.error_request.address,  
                         sizeof(uint8_t), 
                         &input_control_digital_data.error_request.value,  
                         sizeof(uint8_t), 
                         &input_control_digital_data.error_request,
                         true))
            {
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL calculate mux sample rate error");
            }
    #endif  
        }
        else if (input_control_digital_data.error_request.read_state != INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE)
        {
//            logger_log_debugf(0, "INPUT CONTROL Get Error suppressed --");
        }
    }
}

/**-----------------------------------------------------------------------------
 * Called when a setting has been applied
 * @param request
 */
static void input_control_mux_setting_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    input_control_digital_data.mux_applied_setting_count--;
    
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL input_control_mux_setting_callback %d/%d --", 
    //        input_control_digital_data.mux_applied_setting_count,
    //        input_control_digital_data.mux_setting_count);

    if (input_control_digital_data.mux_applied_setting_count == 0)//input_control_digital_data.mux_setting_count)
    {
        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL COMPLETE input_control_mux_setting_callback %d/%d --", 
                input_control_digital_data.mux_applied_setting_count,
                input_control_digital_data.mux_setting_count);
        
        // Switch on SPDIF Interrupts now that the unit is configured
        // If they are not set on before then we get a load of interrupts while initialising
        // for Ports RF0/RF1
        PLIB_PORTS_ChannelChangeNoticeEnable(PORTS_ID_0, PORT_CHANNEL_F, SYS_PORT_F_CNEN || 0x03);
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param event
 * @param context
 */
static void input_control_digital_callback(DRV_I2C_BUFFER_EVENT event,
                                           I2C_REQUEST_TYPE request_type,
                                           uintptr_t context)
{
    //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Digital MUX Callback %d -- ", event);

    switch(event)
    {
        case DRV_I2C_BUFFER_EVENT_COMPLETE:
        {
            if (context != NULL)
            {
                // Look at the context
                INPUT_CONTROL_I2C_REQUEST* request = (INPUT_CONTROL_I2C_REQUEST*)context;
                request->actioned = true;

                //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX Callback addr:%02x val:0x%02x", request->address, request->value);

                // Invoke a callback proc
                if (request->callback != NULL)
                {
                    INPUT_CONTROL_I2C_REQUEST_CALLBACK callback = (INPUT_CONTROL_I2C_REQUEST_CALLBACK)request->callback;
                    callback(request);
                }               
            }
            
            break;
        }
        
        case DRV_I2C_BUFFER_EVENT_ERROR:
        {
            if (context != NULL)
            {
                INPUT_CONTROL_I2C_REQUEST* request = (INPUT_CONTROL_I2C_REQUEST*)context;
                logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Digital Callback Error addr:%02x val:0x%02x", request->address, request->value);
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Digital Callback Error unknown context");
            }
            break;
        }
        
        default:
        {
            //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "Input Control Callback Other %d %d", (int)event, settings_data.state);
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Used to set the sample rate - including handling for sample rate smoothing so that
 * we don't show sudden transient changes
 * @param register_ident
 * @param value
 */
static void input_control_set_spdif_rx_sample_rate(SAMPLE_RATE sample_rate)
{
    // Check if its changed, if not set the new sample rate and start the timer
    if (input_control_digital_data.sample_rate != sample_rate)
    {
        input_control_digital_data.sample_rate = sample_rate;
    
        timers_reset(TIMER_INDEX_INPUT_CONTROL_SAMPLE_RATE_SMOOTHING);
    }
}
    
/**-----------------------------------------------------------------------------
 * Used to get the state of the DAC reset switch
 * @param on
 */
static bool input_control_spdif_rx_is_reset()
{
    return input_control_digital_data.in_reset;
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * @param request
 */
static void input_control_switch_digital_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX CALLBACK ");
    
    // Set the post switch flag to suppress reporting of NOT SUPPORTED sample rate
    // in input_control_calculate_digital_sample_rate if we have just
    // switched the SPDIX RX Mux
    input_control_digital_data.post_switch = true;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used to define SPDIF MUX configuration values.
 * Note: This works at a register level rather than the individual settings
 * @param register_ident - the register identifier
 * @param value - th value to be written
 */
void input_control_add_digital_setting(uint8_t register_ident, uint8_t value)
{
    INPUT_CONTROL_I2C_REQUEST* setting = &input_control_digital_data.mux_settings[input_control_digital_data.mux_setting_count];
    setting->address = register_ident;
    setting->value = value;
    setting->callback = input_control_mux_setting_callback;
    
    input_control_digital_data.mux_setting_count++;
    
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Add SPDIF MUX Setting addr: 0x%02X value: 0x%02X ", setting->address, setting->value);
}

/**-----------------------------------------------------------------------------
 * Starts the process of applying settings to the SPDIF Mux
 * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
 */
void input_control_apply_digital_setting(INPUT_CONTROL_I2C_REQUEST* setting)
{

#if INPUT_CONTROL_I2C_DIGITAL_ENABLED   

    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Applying SPDIF RX Setting 0x%2X 0x%2X --", setting->address, setting->value);
    //Trigger a register update
    // Note makes the assumption that the address/value are the first two
    // elements in the INPUT_CONTROL_DAC_SETTING structure
    if (i2c_transmit(input_control_digital_data.i2c_client_id, 
                     setting,  
                     2 * sizeof(uint8_t), 
                     setting))
    {
        input_control_digital_data.mux_applied_setting_count++;
        PLIB_PORTS_ChannelChangeNoticeDisable(PORTS_ID_0, PORT_CHANNEL_F, SYS_PORT_F_CNEN || 0x03);
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Apply SPDIF MUX settings error");
    }
#endif      

}

/**-----------------------------------------------------------------------------
 * Starts the process of applying settings to the SPDIF Mux
 * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
 */
void input_control_apply_digital_settings()
{
    if (!input_control_digital_data.mux_settings_applied)
    {
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Applying SPDIF RX Settings --");
        
        // Make all settings unapplied
        int x;
        for (x = 0; x < input_control_digital_data.mux_setting_count; x++)
        {
            INPUT_CONTROL_I2C_REQUEST* setting = &input_control_digital_data.mux_settings[x];

#if INPUT_CONTROL_I2C_DIGITAL_ENABLED   

            input_control_apply_digital_setting(setting);
#endif      
        }  
        
        input_control_digital_data.mux_settings_applied = true;
    }
}
 
/**-----------------------------------------------------------------------------
 * Initiates call to DAC to get sample rate
 */
void input_control_calculate_digital_sample_rate()
{
    // Check we're not currently doing this - running multiple instances causes the 
    // unit to keel over during startup if s digital signal is present.
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL input_control_calculate_digital_sample_rate A --");
    if (!input_control_digital_data.calculating_sample_rate)
    {
        input_control_digital_data.calculating_sample_rate = true;
        
        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL input_control_calculate_digital_sample_rate B --");
        if (!input_control_spdif_rx_is_reset())
        {
            // Note: address should have been defined
            input_control_digital_data.sample_rate_request.actioned = false;
            input_control_digital_data.sample_rate_request.value = 0x00;
            input_control_digital_data.sample_rate_request.read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;

#if INPUT_CONTROL_I2C_DIGITAL_ENABLED 

            // Trigger a request for the source type (Audio/Non Audio))
            if (i2c_read(input_control_digital_data.i2c_client_id, 
                         &input_control_digital_data.sample_rate_request.address,  
                         sizeof(uint8_t), 
                         &input_control_digital_data.sample_rate_request.value,  
                         sizeof(uint8_t), 
                         &input_control_digital_data.sample_rate_request,
                         true))
            {
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL calculate mux sample rate error");
            }
#endif  
        }
    }
}
   
/**-----------------------------------------------------------------------------
 * Defines the address of the register used to manage the current input
 * @param address
 */
void input_control_define_digital_error_address(REGISTER_BITS_DEFINITION address)
{
    input_control_digital_data.error_request.address = address.address;
    input_control_digital_data.error_request.bit_low = address.bit_low;
    input_control_digital_data.error_request.bit_high = address.bit_high;
}
   
/**-----------------------------------------------------------------------------
 * Defines the address of the register used to manage the current input
 * @param address
 */
void input_control_define_digital_switch_address(REGISTER_BITS_DEFINITION address)
{
    input_control_digital_data.switch_request.address = address.address;
    input_control_digital_data.switch_request.bit_low = address.bit_low;
    input_control_digital_data.switch_request.bit_high = address.bit_high;
}
    
/**-----------------------------------------------------------------------------
 * Defines the address of the register used to manage the sample_rate calculation
 * @param address
 */
void input_control_define_digital_sample_rate_address(REGISTER_BITS_DEFINITION address)
{
    input_control_digital_data.sample_rate_request.address = address.address;
    input_control_digital_data.sample_rate_request.bit_low = address.bit_low;
    input_control_digital_data.sample_rate_request.bit_high = address.bit_high;
}
    
/**-----------------------------------------------------------------------------
 * Configures the digital module
 * @return 
 */
bool input_control_digital_configure()
{
    if (input_control_digital_data.configure_callback != NULL)
    {
        input_control_digital_data.configure_callback();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Digital Configuration callback not specified --");
    }
}

/**-----------------------------------------------------------------------------
 * Initialise the input control module
 */
bool input_control_digital_initialise()
{
    bool result = true;
    
    input_control_digital_data.mux_setting_count = 0;
    input_control_digital_data.mux_settings_applied = false;
    input_control_set_spdif_rx_sample_rate(SAMPLE_RATE_0);
    input_control_digital_data.steady_sample_rate = SAMPLE_RATE_0;
    
    input_control_digital_data.error_request.callback = input_control_get_spdif_rx_error_callback;
    input_control_digital_data.switch_request.callback = input_control_switch_digital_callback;
    input_control_digital_data.sample_rate_request.callback = input_control_calculate_digital_sample_rate_callback;
    
    i2c_initialise();

#if INPUT_CONTROL_I2C_DIGITAL_ENABLED   
     
    input_control_digital_data.i2c_client_id = i2c_open(INPUT_CONTROL_DIGITAL_I2C_MODULE, 
            INPUT_CONTROL_DIGITAL_I2C_READ_ADDRESS,
            INPUT_CONTROL_DIGITAL_I2C_WRITE_ADDRESS,
            input_control_digital_callback);
    
    if (I2C_INVALID_CLIENT_ID != input_control_digital_data.i2c_client_id)
    {
        // Initialise the DAC
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX I2C Open");
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX I2C Open Error");
        result = false;
    }    
   
#endif
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns value indication whether the digital signal is invalid at present
 * @return 
 */
bool input_control_digital_invalid_signal()
{
    return input_control_digital_data.invalid_signal;
}
     
/**-----------------------------------------------------------------------------
 * 
 */
void input_control_digital_process()
{
    if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
    {
        if (timers_has_expired(TIMER_INDEX_SPDIF_RX_POLL))
        {
            input_control_get_spdif_rx_error();

            timers_reset(TIMER_INDEX_SPDIF_RX_POLL);
        }
    }
    
//    if (input_control_digital_data.needs_mute)
//    {
//        //input_control_mute_inputs(); // TODO Moved to Avatar Set States Response callback
//        output_control_mute_outputs();
//        input_control_mute_dac(true);
//        
//        input_control_digital_data.needs_mute = false;
//    }
//    
//    if (input_control_digital_data.needs_unmute)
//    {
//        //input_control_unmute_inputs(input_control_get_current_input_type()); // TODO Moved to Avatar Set States Response callback
//        output_control_unmute_outputs();
//        input_control_mute_dac(false);
//        
//        input_control_digital_data.needs_unmute = false;
//    }
}
    
/**-----------------------------------------------------------------------------
 * Sets callbacks for the input_control_digital module
 * @param channel
 */
void input_control_digital_set_callbacks(INPUT_CONTROL_DIGITAL_CONFIGURE_CALLBACK configure_callback)
{
    input_control_digital_data.configure_callback = configure_callback;
}

/**-----------------------------------------------------------------------------
 * Gets the smoothed sample rate of the current SPDIF MUX input 
 * @return 
 */
SAMPLE_RATE input_control_get_digital_sample_rate()
{
    if (timers_has_expired(TIMER_INDEX_INPUT_CONTROL_SAMPLE_RATE_SMOOTHING))
    {
        // Copy sample_rate to steady_sample_rate
        input_control_digital_data.steady_sample_rate = input_control_digital_data.sample_rate;
    }
    
    return input_control_digital_data.steady_sample_rate; 
}

/**-----------------------------------------------------------------------------
 * Handles interrupt on the SPDIF RX Interrupt pins
 * Note while we did get some interrupts raised when the pins weren't set as CN pins in Harmony
 * it only started working properly when they were set (when trying to detect the 
 * AV signal being removed!!)
 * @param definition
 */
void input_control_handle_digital_interrupt(uint8_t channel)
{
#if INPUT_CONTROL_PINS_ENABLED

    if (!input_control_digital_data.in_interrupt)
    {
        input_control_digital_data.in_interrupt = true;

        bool error_check_reqd = true;

        if (channel == INPUT_CONTROL_SPDIF_RX_INT0_PORT_CHANNEL)
        {
            //logger_log_debugf(0, "INPUT CONTROL INTERRUPT --");
            // Read the port to clear the interrupt
            uint32_t value = PLIB_PORTS_Read(INPUT_CONTROL_SPDIF_RX_INT0_ID,
                                             INPUT_CONTROL_SPDIF_RX_INT0_PORT_CHANNEL);

            // Check it was the sample rate pin that triggered the interrupt
            bool sample_rate_change = (value & (1 << INPUT_CONTROL_SPDIF_RX_INT0_PIN)) != 0;
            if (sample_rate_change)
            {
                //volume_mute();
                // Mute the SPDIF MUX output as fast as possible
                if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
                {
                    input_control_mute_inputs();
                }

                input_control_calculate_digital_sample_rate();

                // Note the sample rate calculation will trigger a check for errors
                error_check_reqd = false;
            }

            bool error_change = (value & (1 << INPUT_CONTROL_SPDIF_RX_INT1_PIN)) != 0;
            if (error_change)
            {
                // Mute the SPDIF MUX output as fast as possible
                if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
                {
                    input_control_mute_inputs();
                }
            }
            
            // Disable Interrupts and move to polling as this goes into a loop
            //PLIB_PORTS_ChannelChangeNoticeDisable(PORTS_ID_0, PORT_CHANNEL_F, SYS_PORT_F_CNEN || 0x03);

        }
        else if (channel == INPUT_CONTROL_SPDIF_RX_INT1_PORT_CHANNEL)
        {
            //logger_log_debugf(0, "INPUT CONTROL INTERRUPT ERROR --");
            // Read the other interrupt port to clear it.
            // Note this would clear the interrupts for any other peripherals 
            // monitoring on this port.
            bool error_detected = PLIB_PORTS_PinGet(INPUT_CONTROL_SPDIF_RX_INT1_ID, 
                                                    INPUT_CONTROL_SPDIF_RX_INT1_PORT_CHANNEL, 
                                                    INPUT_CONTROL_SPDIF_RX_INT1_PIN);

            bool error_change = (error_detected & (1 << INPUT_CONTROL_SPDIF_RX_INT1_PIN)) != 0;
            if (error_change)
            {
                // Mute the SPDIF MUX output as fast as possible
                if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
                {
                    input_control_mute_inputs();
                }
            }
        }

        input_control_digital_data.in_interrupt = false;
    }

#endif        

}

/**-----------------------------------------------------------------------------
 * Used to set the Spdif Rx Master/Slave mode
 * @param reset
 * @param master
 */
void input_control_digital_set_master_slave(bool master)
{
    logger_log_debugf(0, "INPUT CONTROL Set mode %s --", master ? "Master" : "Slave");
    INPUT_CONTROL_I2C_REQUEST* setting = input_control_find_definition_for_address(0X01);
    if (setting != NULL)
    {
        if (!master)
        {
            // Slave mode used for Analogue and Avatar 0x6A 0110 1010
            setting->value &= ~(1 << 4);    // Set bit 4 OFF
            setting->value |= 1 << 5;       // Set bit 5 ON
        }
        else
        {
            // Normal mode 0x5A  0101 1010
            setting->value |= 1 << 4;       // Set bit 4 ON
            setting->value &= ~(1 << 5);    // Switch bit 5 OFF
        }

        input_control_apply_digital_setting(setting);
    }
}

/**-----------------------------------------------------------------------------
 * Used to set the SPdif Rx Power state
 * @param on
 */
void input_control_digital_set_power(bool on)
{
    if (on)
    {
        INPUT_CONTROL_I2C_REQUEST* setting = input_control_find_definition_for_address(0X00);
        setting->value = 0x23;
        input_control_apply_digital_setting(setting);
        
        input_control_digital_data.mux_settings_applied = false;
        input_control_apply_digital_settings();
    }
    else
    {
        INPUT_CONTROL_I2C_REQUEST* setting = input_control_find_definition_for_address(0X00);
        setting->value = 0x00;
        input_control_apply_digital_setting(setting);
    }
}

/**-----------------------------------------------------------------------------
 * Used to change the state of the SPDIF MUX reset switch
 * Switched OFF during power up, ON the rest of the time
 * @param on - true - take the mux out of reset, false - puit it into reset
 */
void input_control_digital_reset(bool on)
{
#if INPUT_CONTROL_PINS_ENABLED
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Reset - Switching %s ", on == true ? "On" : "Off");
    
    PLIB_PORTS_PinWrite(INPUT_CONTROL_SPDIF_RX_RESET_PORTS_ID, 
                        INPUT_CONTROL_SPDIF_RX_RESET_PORT_CHANNEL, 
                        INPUT_CONTROL_SPDIF_RX_RESET_PIN,
                        on); // Note pin is low

#endif        
    
    input_control_digital_data.in_reset = !on;
    
    if (on)
    {
        // Setup for a resend of the settings
        input_control_digital_data.mux_settings_applied = false;
    }
}

/**-----------------------------------------------------------------------------
 * Switch to a specific input on the digital input control
 * @param definition - the definition of the control to switch to
 */
void input_control_switch_digital(INPUT_CONTROL_DEFINITION* definition)
{
    uint8_t new_value = definition->value << input_control_digital_data.switch_request.bit_low;
    
    // Check there isnt a default for this register, if there is then OR is 
    // (assumes default is only in the lower bits)..
    INPUT_CONTROL_I2C_REQUEST* setting = input_control_find_definition_for_address(input_control_digital_data.switch_request.address);
    if (setting != NULL)
    {
        new_value |= setting->value;
    }

    // Check if we have had an invalid signal and if so mute outputs before switching
    // to stop noise..
    if (input_control_digital_data.invalid_signal)
    {
        logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX Switch Digital errors MUTING OUTPUTS ");

        // Mute all outputs
        input_control_digital_mute_input();
    }
    else
    {
    }

    // Reapply the settings (just in case);
    input_control_apply_digital_settings();
    
    // Removed as was not catching analogue to digital switches
    //if (new_value != input_control_data.switch_request.value)
    //{
        // Note: address should have been defined
        input_control_digital_data.switch_request.actioned = false;
        input_control_digital_data.switch_request.value = new_value;

        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX %d %d %d ", definition->value, input_control_digital_data.switch_request.bit_low, input_control_digital_data.switch_request.value);

#if INPUT_CONTROL_I2C_DIGITAL_ENABLED 
    
    // Trigger a request for the dpll status
        if (i2c_transmit(input_control_digital_data.i2c_client_id, 
                         &input_control_digital_data.switch_request,  
                         2, 
                         &input_control_digital_data.switch_request))
        {
        }
        else
        {
            logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL switch input SPDIF MUX error");
        }

#endif    
    //}
 }

/* *****************************************************************************
 End of File
 */
