/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    menu.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/menu.h"
#include <stdio.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! The currently selected menu item
    MENU_ITEM* current_item;
    
    //! A pointer to an array of menu items
    MENU_ITEM* items;
    
    //! Pointer to the titles associated with the menu
    char** titles;

    //! The number of items in the array
    size_t item_count;

    //! The previously selected item before the current item.  can be used with menu_get_previous
    // in scenarios where the Menu Item doesnt have a parent.
    MENU_ITEM* previous_item;
    
    /**
     * A Menu item that is the root of all items added to the menu
     */
    MENU_ITEM root_item;

    /**
     * Set to true if the root item has been initialised
     */
    bool root_initialised;
    
} MENU_DATA;

static MENU_DATA menu_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**-----------------------------------------------------------------------------
 * Resets the menu items edit state
 * @param menu
 */
static void menu_reset_edit_state(MENU_ITEM* menu)
{
    menu->edit_info.reset = true;
    menu->edit_info.edit_pos = 0;
    menu->edit_info.changed = false;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Add a child menu to the supplied menu
 * @param menu - The parent menu
 * @param child - The child menu
 */
void menu_add_child(MENU_ITEM* parent, MENU_ITEM* child, bool default_item)
{
    MENU_ITEM* parent_item = parent;
    MENU_ITEM* last_child = NULL;
    
    if (child->parent == NULL)
    {
        if (parent_item == NULL)
        {
            // Check that we have setup the root item
            if (!menu_data.root_initialised)
            {
                menu_data.root_item.first_child = NULL;
                menu_data.root_item.default_item = NULL;
                menu_data.root_item.last_child = NULL;
                menu_data.root_item.next_sibling = NULL;
                menu_data.root_item.previous_sibling = NULL;
                menu_data.root_item.loop_back_children = false;
                menu_data.root_initialised = true;
            }

            parent_item = &menu_data.root_item;
        }

        // Get current state
       last_child = parent_item->last_child;

       // Sort out parent
       parent_item->last_child = child;
       if (parent_item->first_child == NULL)
       {
           parent_item->first_child = child;
           parent_item->default_item = child;
       }

       if (default_item)
       {
           parent_item->default_item = child;
       }

       // Sort out last child
       if (last_child != NULL)
       {
           last_child->next_sibling = child;
       }

       // Sort out child
       child->parent = parent_item;
       child->previous_sibling = last_child;
       child->next_sibling = NULL;
       //child->default_item = NULL;

       menu_reset_edit_state(child);
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_MENU, "MENU attempting to add %s to %s - already a child of %s ", 
                app_menu_titles[child->menu_id],
                parent == NULL ? "TOP" : app_menu_titles[parent->menu_id],
                app_menu_titles[child->parent->menu_id]);
    }
}
   
/**-----------------------------------------------------------------------------
 * Adds a function that will return a character string representation of 
 * the current menu item.
 * @param menu
 * @param func
 */
void menu_add_display_function(MENU_ITEM* menu, void (*func)(char*, size_t))
{
    menu->display_function = func;
}
    
/**-----------------------------------------------------------------------------
 * Adds a function that can be used by the code that process menu commands to
 * respond to an input change that may require the data maintained by the menu item
 * to be updated/changed
 * @param menu - The menu to move from
 * @param func - a function pointer to a function that takes a single argument
 * which is the MENU_ID
 */
void menu_add_draw_function(MENU_ITEM* menu, void (*func)(MENU_ID))
{
    menu->draw_function = func;    
}

/**-----------------------------------------------------------------------------
 * Adds a function that can be used by the code that process menu commands to
 * respond to an input change that may require the data maintained by the menu item
 * to be updated/changed
 * @param func - a function pointer to a function that takes a single pointer argument
 * @param on_entry - If yes the item will enter Edit mode as soon as its selected, rather
 * than requiring OK to be pressed
 */
void menu_add_edit_function(MENU_ITEM* menu, bool (*func)(void*), bool on_entry, bool wants_navigation)
{
    menu->edit_function = func;
    menu->edit_on_entry = on_entry;
    menu->edit_wants_navigation = wants_navigation;
}
    
/**-----------------------------------------------------------------------------
 * Adds a function that will be called when the menu exits from edit state
 * By default the menu will return to its non-edit display, this allows the code
 * to move to different code (e.g back to volume display or to another screen)
 * @param menu
 * @param func
 */
void menu_add_edit_exit_function(MENU_ITEM* menu, bool (*func)(MENU_ID))
{
    menu->exit_edit_function = func;    
}
  
/**-----------------------------------------------------------------------------
 * Adds enter and exit functions to the menu item -  NULL = no action
 */
void menu_add_enter_exit_functions(MENU_ITEM* menu, void (*enter_func)(MENU_ID), MENU_ID (*exit_func)(MENU_ID))
{
    menu->enter_function = enter_func;
    menu->exit_function = exit_func;
}

/**-----------------------------------------------------------------------------
 * Adds a function that can be invoked when the OK button is pressed when in 
 * MENU mode - this is an alternative to menu_add_edit_function
 * @param menu - The menu to move from
 * @param func - a function pointer to a function that takes a single pointer argument
 * which will the MENU_ID
 */
void menu_add_process_function(MENU_ITEM* menu, void (*func)(MENU_ID))
{
    menu->process_function = func;    
}
    
/**-----------------------------------------------------------------------------
 * Adds a function that can be used by the code to define ScreenData for use with the display_message utility
 * @param menu - The menu the function is attached to
 * @param screen_type - The type of screen CURRENTLY_PLAYING, OPTION_LIST, OPTION etc
 * @param func -  pointer to a function used to build the screen and its data
 * @param page_size - The maximum number of items to be displayed (OPTION_LIST/OPTIONS only) 
 *                  - if 0 then defaults to DISPLAY_MESSAGE_MAX_ROWS
 * @param option_strings - if supplied provides an array of strings representing 
 *                       - the option values (OPTION only) - MUST match the values 
 *                       - returned by the menu_add_display_function function
 * @param option_count - indicates the size of the options_strings array
 * @icon_id - identifies an icon that can be associated with the menu item
 */
void menu_define_screen_data(MENU_ITEM* menu, 
        uint8_t screen_type, 
        uint16_t (*func)(MENU_ID, uint8_t*, uint8_t),
        uint8_t page_size,
        char** option_strings,
        uint8_t option_count,
        uint8_t icon_id)
{
    // Overrides any setting used by menu_add_edit_function!!
    // Required for navigation to work correctly
    menu->edit_on_entry = false; 
    menu->display_data.screen_type = screen_type;
    menu->display_data.build_screen_data = func;
    menu->display_data.page_size = page_size;
    menu->display_data.option_strings = option_strings;
    menu->display_data.option_count = option_count;
    menu->display_data.icon_id = icon_id;
}

/**-----------------------------------------------------------------------------
 * The Menu Item can be used to store the current field in a multi field edit 
 * procedure (IP Address, Friendly Name)
 * It uses the first position of the edit_info structure
 * @param menu
 * @param num_fields - the number of fields for editing - used to adjust the position
 * if the number of fields has decreased.
 * @return 
 */
uint8_t menu_get_edit_pos(MENU_ITEM* menu, int num_fields)
{
    // Work out which element of the name we are editing
    int result = 0;
    if (!menu->edit_info.reset)
    {
        result = menu->edit_info.edit_pos;
        if ((num_fields != MENU_IGNORE_NUM_FIELDS) && (result > num_fields - 1))
        {
            result = 0;
            menu->edit_info.edit_pos = result;
        }
    }
    else
    {
        menu->edit_info.edit_pos = result;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
* Gets the currently selected menu item
* @return 
*/
MENU_ITEM* menu_get_current_item()
{
    return menu_data.current_item;
}
    
/**-----------------------------------------------------------------------------
 * Get a reference to a MENU_ITEM base on its index
 */
MENU_ITEM* menu_get_item(MENU_ID index)
{
    MENU_ITEM* result = NULL;
    
    if (index < menu_data.item_count)
    {
        result = &menu_data.items[index];
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Get a reference to the MENU_ITEMS parent, returns NULL for top level menus
 */
MENU_ITEM* menu_get_parent(MENU_ITEM* menu)
{
    MENU_ITEM* result = menu->parent;
     
    if (result == &menu_data.root_item)
    {
        result = NULL;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the previously selected menu item
 * @return 
 */
MENU_ITEM* menu_get_previous_item()
{
    return menu_data.previous_item;
}
    
/**-----------------------------------------------------------------------------
 * Get a reference to a MENU_ITEM's title
 */
char* menu_get_title(MENU_ID index)
{
    char* result = NULL;
     
    if (index < menu_data.item_count)
    {
        result = menu_data.titles[index];
    }
    
    return result;
    
}
    
/**-----------------------------------------------------------------------------
 * Returns true if the current menu item has children
 */
bool menu_has_children(MENU_ITEM* menu)
{
    return menu->first_child != NULL;
}
    
/**-----------------------------------------------------------------------------
 * Initialises the menus data which is a pointer to an array of menu items
 * @param items
 * @param count
 */
void menu_initialise(MENU_ITEM* items, char** titles, size_t count)
{
    menu_data.current_item = NULL;
    menu_data.items = items;
    menu_data.titles = titles;
    menu_data.item_count = count;
}
    
/**-----------------------------------------------------------------------------
 * Initialises the supplied menu item
 * @param item
 * @param loop_back_children
 */
void menu_initialise_item(MENU_ITEM* menu, MENU_ID menu_id, bool loop_back_children)
{
    menu->menu_id = menu_id;
    menu->first_child = NULL;
    menu->default_item = NULL;
    menu->last_child = NULL;
    menu->next_sibling = NULL;
    menu->previous_sibling = NULL;
    menu->loop_back_children = loop_back_children;
    menu_reset_edit_state(menu);
    menu->display_data.screen_type = 0; // DISPLAY_MESSAGE_SCREEN_TYPE_UNKNOWN
    menu->display_data.in_display = false;
    menu->display_data.build_screen_data = NULL;
    menu->display_data.page_size = 0;
    menu->display_data.option_strings = NULL;
    menu->display_data.option_count = 0;
    //menu->display_data.option_string_start = 0;
    //menu->display_data.page_start = NULL;
    menu->display_data.page_start_id = MENU_DISPLAY_DATA_EMPTY_ID;
    menu->draw_function = NULL;
    menu->edit_function = NULL;
    menu->edit_on_entry = false;
    menu->exit_edit_function = NULL;
    menu->process_function = NULL;
    menu->enter_function = NULL;
    menu->exit_function = NULL;
}
    
/**-----------------------------------------------------------------------------
 * Returns whether the current MenuItem represents an editable value that 
 * is set to its original (unedited) value
 * @param menu
 * @return 
 */
bool menu_is_editable_and_original(MENU_ITEM* menu)
{
    bool result = false;
    if ((menu->edit_function != NULL) && (!menu->edit_info.changed))
    {
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param menu - The menu to move from
 * @param reset - true - returns the first child, false returns the last active
 * @return 
 */
MENU_ITEM* menu_move_down(MENU_ITEM* menu)
{
    MENU_ITEM* result = menu->default_item;

    if (result == NULL)
    {
        result = menu->first_child;
    }
    
    if (result != NULL)
    {
        menu_reset_edit_state(result);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Return the next sibling of the menu item
 * @param menu
 * @return 
 */
MENU_ITEM* menu_move_left(MENU_ITEM* menu)
{
    MENU_ITEM* result = menu->previous_sibling;
    
    if (result == NULL)
    {
        if (menu->parent != NULL)
        {
            // Check whether parent menu loops round
            if (menu->parent->loop_back_children) 
            {
                result = menu->parent->last_child;
            }
        }
    }
    
    if (result != NULL)
    {
        menu_reset_edit_state(result);
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Return the next logical item in the menu tree.  It will try and move down, if 
 * there is no child it will move right, again if no sibling it will move up
 * @param menu
 * @return 
 */
MENU_ITEM* menu_move_next(MENU_ITEM* menu)
{
    MENU_ITEM* result = menu_move_down(menu);
    if (result == NULL)
    {
        result = menu_move_right(menu);
        if ((result == NULL) || (result == menu))
        {
            result = menu_move_up(menu);
        }
    }
    
    return result;
}    

/**-----------------------------------------------------------------------------
 * Return the previous sibling of the menu item
 * @param menu
 * @return 
 */
MENU_ITEM* menu_move_right(MENU_ITEM* menu)
{
    MENU_ITEM* result = menu->next_sibling;
    if (result == NULL)
    {
        if (menu->parent != NULL)
        {
            // Check whether parent menu loops round
            if (menu->parent->loop_back_children) 
            {
                result = menu->parent->first_child;
            }
            else
            {
                // Go back up to parent menu
                if (menu->parent != &menu_data.root_item)
                {
                    result = menu->parent;
                }
            }
        }
    }
     
    if (result != NULL)
    {
        menu_reset_edit_state(result);
    }
   
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Returns the parent item
 * @param menu
 * @return 
 */
MENU_ITEM* menu_move_up(MENU_ITEM* menu)
{
    MENU_ITEM* result = menu->parent;

    if (result != NULL)
    {
        menu_reset_edit_state(result);
    }
    
    if (result == &menu_data.root_item)
    {
        result = NULL;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Changes the current menu, executing exit/enter functions if defined
 * @param menu_id
 */
void menu_set_current_item(MENU_ID menu_id)
{
    MENU_ITEM* current = menu_get_current_item();

    if ((current == NULL) || (current->menu_id != menu_id))
    {
        menu_data.previous_item = current;
        
        if (current != NULL)
        {
            if (current->exit_function != NULL)
            {
                 // To prevent recursion if the exit function calls app_context_set_menu()
                // remove the exit function temporarily from the menu item before executing
                // it, so that this code will not be entered)
                MENU_ID (*exit_function)(MENU_ID) = current->exit_function;
                current->exit_function = NULL;
                
                // Execute the function
                //current->exit_function(current->menu_id);
                MENU_ID override_id = exit_function(current->menu_id);
                if (override_id != MENU_USE_SUPPLIED_MENU_ID)
                {
                    logger_log_debugf(0, "MENU Overriding new menu %s with menu %s --", menu_get_title(menu_id), menu_get_title(override_id));
                    menu_id = override_id;
                }
                
                // Put the function back onto the menu item
                current->exit_function = exit_function;
            }
        }

        menu_data.current_item = menu_get_item(menu_id);
        if (menu_data.current_item != NULL)
        {
            menu_reset_edit_state(menu_data.current_item);

            if (menu_data.current_item->enter_function != NULL)
            {
                menu_data.current_item->enter_function(menu_id);
            }
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Used to set the current ordinal edit position in a multi field edit - to be used
 * in conjunction with menu_get_edit_pos
 * @param menu
 * @param edit_pos - the new value
     * @param num_fields - The number of fields that can be edited - note this may 
     * change throughout the lifetime of the field (e.g. when editing Network Friendly Name)
 */
void menu_set_edit_pos(MENU_ITEM* menu, int8_t edit_pos, int num_fields)
{
    menu->edit_info.reset = false;
    
    if (edit_pos > num_fields - 1)
    {
        edit_pos = 0;
    }
    else if (edit_pos < 0)
    {
        edit_pos = num_fields - 1;
    }

    menu->edit_info.edit_pos = edit_pos;   
}
    
/**-----------------------------------------------------------------------------
 * To be used in edit functions to initialise the change state.  This is used
 * to monitor whether the current edit value is a change or is the original value
 * @param value
 * @param size
 */
void menu_try_initialise_change_state(void* value, size_t size)
{
    MENU_ITEM* menu = menu_get_current_item(); 
    
    if ((menu != NULL) && (menu->edit_info.reset))
    {
        // Sore the current value
        menu->edit_info.reset = false;
        menu->edit_info.changed = false;
        if (size <= MENU_EDIT_INFO_DATA_MAX)
        {
            memcpy(&menu->edit_info.data[0], value, size);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Used in conjunction with menu_try_initialise_change_state to determine
 * whether an edit is altering the original value of a menu item or returning it
 * to its original state
 * @param value
 * @param size
 */
void menu_update_change_state(void* value, size_t size)
{
    MENU_ITEM* menu = menu_get_current_item(); 
    
    if (menu != NULL)
    {
        if (memcmp(menu->edit_info.data, value, size) != 0)
        {
            menu->edit_info.changed = true;
            memcpy(&menu->edit_info.data[0], value, size);
        }
        else
        {
            menu->edit_info.changed = false;
        }
    }    
}

/* *****************************************************************************
 End of File
 */
