/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    output_control.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/output_control.h"
#include "app_config.h"
#include "utilities/dirac.h"
#include "utilities/input_control_defs.h"
#include "utilities/mutes.h"
#include "utilities/power.h"
#include "utilities/structured_message.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    // A callback for when system faults are detected
    SYSTEM_FAULT_CALLBACK fault_callback;
    
    //! Records whether the headphones are connected
    bool headphones_connected;
    
    //! Records the current headphone override status
    bool headphone_override;
    
    //! We allow 20ms of DC Offset before triggering a fault.  This is used to 
    //  determive whether we are currently in a fault state or not
    bool dc_offset_in_wait;
    
    //! We allow 20ms of Short circuit before triggering a fault.  This is used to 
    //  determive whether we are currently in a fault state or not
    bool short_circuit_in_wait;
    
} OUTPUT_CONTROL_DATA;

static OUTPUT_CONTROL_DATA output_control_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**
 * DC Offset fault processing
 */
static void output_control_process_dc_offset()
{
    if (output_control_dc_offset_detect())
    {
        if (timers_has_expired(TIMER_INDEX_DC_OFFSET))
        {
            if (!output_control_data.dc_offset_in_wait)
            {
                logger_log_infof(0, "DC OFFSET DETECT: Detected - waiting 20ms");
                output_control_data.dc_offset_in_wait = true;
                timers_reset(TIMER_INDEX_DC_OFFSET);
            }
            else
            {
                // Trigger a fault
                output_control_data.dc_offset_in_wait = false;

                logger_log_infof(0, "DC OFFSET DETECT: Triggering");
                if (output_control_data.fault_callback != NULL)
                {
                    SYSTEM_FAULT fault;
                    fault.code = SYSTEM_FAULT_CODE_DC_OFFSET;
                    strcpy(fault.description, system_fault_codes[SYSTEM_FAULT_CODE_DC_OFFSET]);

                    output_control_data.fault_callback(&fault);
                }
            }
        }
    }
    else
    {
        output_control_data.dc_offset_in_wait = false;
        timers_expire(TIMER_INDEX_DC_OFFSET);
    }
}

/**
 * Headphone connection processing
 */
static void output_control_process_headphone_connection()
{
    bool connected = output_control_headphones_connected(); // Note this updates output_control_data.headphones_connected
    bool change = output_control_data.headphones_connected != connected;

    //logger_log_debugf(211, "HEADPHONES %d %d %d  --", change, connected, last_pass_connection);

    if (change)
    {
        //logger_log_debugf(211, "HEADPHONES %d %d %d  --", change, output_control_data.headphones_connected);
        // Check whether we need to mute the speakers if the headphones are plugged in
        if (!output_control_get_headphone_override() && connected)
        {
            output_control_mute_outputs(false);
            
            logger_log_testf(LOGGER_CODE_TEST_HEADPHONES_CONNECTED, "");
            
            avatar_send_headphone_status(AVATAR_HEADPHONE_STATUS_CONNECTED);
            
            // Switch off Dirac
            avatar_set_dirac_curve(DIRAC_NO_CURVE);
        }
        else if (!connected)
        {
            // Headphones have been disconnected, so unmute the speakers etc..
            output_control_unmute_outputs();
            
            logger_log_testf(LOGGER_CODE_TEST_HEADPHONES_DISCONNECTED, "");
            
            avatar_send_headphone_status(AVATAR_HEADPHONE_STATUS_DISCONNECTED);
            
            INPUT_CONTROL_DEFINITION* def = input_control_get_current_definition();
            if (def != NULL)
            {
                avatar_set_dirac_curve(dirac_get_curve_for_input(def->index));
            }
        }

        output_control_data.headphones_connected = connected;

        // Send unsolicited message
        STRUCTURED_MESSAGE message;
        structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
        message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
        message.command = APP_COMMANDS_HEADPHONES;
        message.data[0] = APP_COMMAND_VALUE_REQUEST_STATUS;
        message.data_length = 1;

        app_commands_headphones(&message);
    }
}

/**
 * DC Offset fault processing
 */
static void output_control_process_short_circuit()
{
    if (output_control_short_circuit_detect())
    {
        if (timers_has_expired(TIMER_INDEX_SHORT_CIRCUIT))
        {
            if (!output_control_data.short_circuit_in_wait)
            {
                logger_log_infof(0, "SHORT CIRCUIT DETECT: Detected - waiting %dms", TIMER_TIMEOUT_VALUES[TIMER_INDEX_SHORT_CIRCUIT]);
                output_control_data.short_circuit_in_wait = true;
                timers_reset(TIMER_INDEX_SHORT_CIRCUIT);
            }
            else
            {
                // Trigger a fault
                output_control_data.short_circuit_in_wait = false;

                logger_log_infof(0, "SHORT CIRCUIT DETECT: Triggering");
                if (output_control_data.fault_callback != NULL)
                {
                    SYSTEM_FAULT fault;
                    fault.code = SYSTEM_FAULT_CODE_SHORT_CIRCUIT;
                    strcpy(fault.description, system_fault_codes[SYSTEM_FAULT_CODE_SHORT_CIRCUIT]);

                    output_control_data.fault_callback(&fault);
               }
            }
        }
    }
    else
    {
        output_control_data.short_circuit_in_wait = false;
        timers_expire(TIMER_INDEX_SHORT_CIRCUIT);
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Use to determine whether there is a DC offset fault
 * @return 
 */
bool output_control_dc_offset_detect()
{
    bool result = false;
        
#if OUTPUT_CONTROL_PINS_ENABLED   
        
    result = PLIB_PORTS_PinGet(OUTPUT_DC_OFFSET_PORTS_ID,
                               OUTPUT_DC_OFFSET_PORT_CHANNEL, 
                               OUTPUT_DC_OFFSET_PIN);
 
#endif
    
    //logger_log_debugf(0, "DC OFFSET: %d --", result);
   
    return result;
}
    
/**
 * Gets the headphone override status
 * @return - true if speakers are unmuted when the headphones are plugged in
 */
bool output_control_get_headphone_override()
{
    return output_control_data.headphone_override;
}
    
/**
 * Determines whether the headphones are currently connected or not
 */
bool output_control_headphones_connected()
{
    bool result = false;
        
#if OUTPUT_CONTROL_PINS_ENABLED   
        
    result = PLIB_PORTS_PinGet(OUTPUT_HEADPHONES_PORTS_ID,
                               OUTPUT_HEADPHONES_PORT_CHANNEL, 
                               OUTPUT_HEADPHONES_PIN);
 
#endif
   
    //output_control_data.headphones_connected = result;
    
    return result;
}
    
/**
 * Initialise the output control module
 */
void output_control_initialise(SYSTEM_FAULT_CALLBACK fault_callback)
{
    output_control_data.headphones_connected = output_control_headphones_connected();
    output_control_data.fault_callback = fault_callback;
    
    output_control_data.dc_offset_in_wait = false;
    output_control_data.short_circuit_in_wait = false;
        
    //Read the volume/balance from the settings
    settings_get(SETTING_HEADPHONE_OVERRIDE, &output_control_data.headphone_override);
}
    
/**
 */
void output_control_mute_analogue_outputs()
{
    
}

/**
  */
void output_control_mute_digital_outputs()
{
    
}
    
/**
 * Mutes Speaker, Headphone, Preamp, and Power AMp outputs
 * @param include_power_amp - True for SA10, false SA20
 */
void output_control_mute_outputs(bool include_headphones)
{
    mutes_mute(MUTE_INDEX_SPEAKER);
    mutes_mute(MUTE_INDEX_PRE_OUT);
    
    if (include_headphones)
    {
        mutes_mute(MUTE_INDEX_VOLUME);
        mutes_mute(MUTE_INDEX_HEADPHONE);
    }
}   
    
/**
 * Perform regular output control tasks
 */
void output_control_process()
{
    output_control_process_headphone_connection();
   
    if (power_get_state() == POWER_STATE_ON)
    {
        // See Pete email 28/4/2020 15:07
#ifdef USE_SHORT_CIRCUIT_SENSE_SETTING     
        bool short_circuit_sense;
        settings_get(SETTING_SHORT_CIRCUIT_SENSE, &short_circuit_sense);
        if (short_circuit_sense && !output_control_headphones_connected())
        {
            //logger_log_debugf(0, "SHORT CIRCUIT - CHECKING --");
            output_control_process_dc_offset();
            output_control_process_short_circuit();
        }
        else
        {
            //logger_log_debugf(0, "SHORT CIRCUIT - NOT CHECKING --");
        }
#else
        output_control_process_dc_offset();
        output_control_process_short_circuit();
#endif        
    }
}
     
/**
 * Sets the headphone override status
 */
void output_control_set_headphone_override(bool override)
{
    if (output_control_data.headphone_override != override)
    {
        output_control_data.headphone_override = override;
                
        settings_set(SETTING_HEADPHONE_OVERRIDE, &output_control_data.headphone_override);

        // Check whether we need to mute the speakers if the headphones are plugged in
        // while we are editing this value
        bool connected = output_control_headphones_connected(); 
        if (!output_control_get_headphone_override() && connected)
        {
            output_control_mute_outputs(false);
            
            avatar_send_headphone_status(AVATAR_HEADPHONE_STATUS_CONNECTED);
        }
        else //if (!connected)
        {
            // Headphones have been disconnected, so unmute the speakers etc..
            output_control_unmute_outputs();
            
            avatar_send_headphone_status(AVATAR_HEADPHONE_STATUS_DISCONNECTED);
       }
    }
}

/**
 * Use to determine whether there is a short circuit fault
 * @return 
 */
bool output_control_short_circuit_detect()
{
    bool result = false;
        
#if OUTPUT_CONTROL_PINS_ENABLED   
        
    result = PLIB_PORTS_PinGet(OUTPUT_SHORT_CIRCUIT_PORTS_ID,
                               OUTPUT_SHORT_CIRCUIT_PORT_CHANNEL, 
                               OUTPUT_SHORT_CIRCUIT_PIN) ? false : true;
 
#endif
    
    //logger_log_debugf(0, "SHORT CIRCUIT: %d --", result);
   
    return result;
}
    
/**
 * Unmutes 
 */
void output_control_unmute_analogue_outputs()
{
   logger_log_debugf(0, "OUTPUT CONTROL unmute_outputs() --");
    mutes_unmute(MUTE_INDEX_HEADPHONE);
    mutes_unmute(MUTE_INDEX_VOLUME);

    if (!output_control_headphones_connected() || output_control_get_headphone_override())    
    {
        mutes_unmute(MUTE_INDEX_SPEAKER);
        mutes_unmute(MUTE_INDEX_PRE_OUT);
        //mutes_unmute(MUTE_INDEX_VOLUME);
    }
}

/**
 * Unmutes 
 */
void output_control_unmute_digital_outputs()
{
    
}
    
/**
 * Mutes Speaker, Headphone, Preamp, and Power AMp outputs
 * @param include_power_amp - True for SA10, false SA20
 */
void output_control_unmute_outputs()
{
    output_control_unmute_analogue_outputs();
    output_control_unmute_digital_outputs();
}   

/* *****************************************************************************
 End of File
 */
