/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    remote.c

  @Summary
 Remote control handling

  @Description
 * http://www.picprojects.net/rc5_decoder/
 * see rc5.h for commented code
 * 
 * https://clearwater.com.au/code/rc5
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/remote.h"
#include "utilities/timers.h"
#include "utilities/logger.h"
#include <stdio.h>
#include <string.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Pulse widths in 100us; tolerance 1/3 of short pulse
 */
#define PULSE_WIDTH_SHORT_MIN   7
#define PULSE_WIDTH_SHORT_MAX   11
#define PULSE_WIDTH_LONG_MIN    16
#define PULSE_WIDTH_LONG_MAX    20

#define RC5_SEQUENCE_POOL_SIZE  4

typedef struct
{
    //! Indicates whether this is a complete message
    volatile bool complete;
    
    uint8_t address;
    
    uint8_t command;
    
    //! The field bit is used to determine whether the command range is 0-63 or 64-127
    bool field_bit;
    
    //! Indicates whether this is a repeat message
    bool repeat;
    
    //! The current bit position in the sequence
    volatile uint8_t bit_pos;
    
    //! Determines whether the current interrupt is at a bit start or a bit transition
    volatile bool in_transition;
} RC5_SEQUENCE;

/**
 * Remote global data
 */
typedef struct
{
    volatile uint8_t rc5_sequence_pool_read_idx;
    volatile uint8_t rc5_sequence_pool_write_idx;
    RC5_SEQUENCE rc5_sequence_pool[RC5_SEQUENCE_POOL_SIZE];
    
    //! Keeps track of the last toggle value - used to determine whether this is a repeat or not.
    volatile bool prev_toggle;
    
    // A callback to handle commands
    STRUCTURED_MESSAGE_CALLBACK callback;
    
    // The system code (16/19) used with the remote)
    REMOTE_SYSTEM_CODE system_code;
} REMOTE_DATA;

static REMOTE_DATA remote_data =
{
    .prev_toggle = false,   // toggle appears to start at 1.
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the current status of the remote system code (16/19)
 * @return 
 */
REMOTE_SYSTEM_CODE remote_get_system_code()
{
    return remote_data.system_code;
}

/**-----------------------------------------------------------------------------
 * Gets the remote system code (16/19) from the message
 * @return 
 */
REMOTE_SYSTEM_CODE remote_get_system_code_from_message(STRUCTURED_MESSAGE* message)
{
    REMOTE_SYSTEM_CODE result;
    
    memcpy(&result, message->data, sizeof(REMOTE_SYSTEM_CODE));
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the current status of the remote system code (16/19) as a string
 * @return 
 */
void remote_get_system_code_string(char* buffer, size_t size)
{
    sprintf(buffer, "%02d", remote_data.system_code);
}

/**-----------------------------------------------------------------------------
 * 
 * @param buffer
 * @param size
 */
static void remote_initialise_rc5_sequence(RC5_SEQUENCE* sequence)
{
    sequence->complete = false;
    sequence->address = 0;
    sequence->command = 0;
    sequence->bit_pos = 0;
    sequence->in_transition = true;
    sequence->repeat = false;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Clear down all outstanding messages - used when entering standby
 */
void remote_clear()
{
    remote_data.rc5_sequence_pool_read_idx = 0x00;
    remote_data.rc5_sequence_pool_write_idx = 0x00;
    
    int x;
    for (x = 0; x < RC5_SEQUENCE_POOL_SIZE; x++)
    {
        remote_initialise_rc5_sequence(&remote_data.rc5_sequence_pool[x]);
    }
}

void remote_handle_interrupt()
{
    // read the pin
    bool ir_value = PLIB_PORTS_PinGet(IR_PORTS_ID,
                                      IR_PORT_CHANNEL,   
                                      IR_PIN );

    
    RC5_SEQUENCE* sequence = &remote_data.rc5_sequence_pool[remote_data.rc5_sequence_pool_write_idx];

    // Workout how many 100us have elapsed since last cycle
    int diff = timers_get_elapsed_count(TIMER_INDEX_REMOTE_CLOCK) % TIMER_TIMEOUT_REMOTE_CLOCK;
    timers_reset(TIMER_INDEX_REMOTE_CLOCK); 
    
    // Is this a new sequence.  Bit pos will be 0 and the pin value = high
    if ((!ir_value) &&
        ((sequence->bit_pos == 0) || (diff > PULSE_WIDTH_LONG_MAX)))
	{
        //logger_log_debugf(0, "REMOTE NEW SEQUENCE %d %d %d --", !ir_value, sequence->bit_pos, diff);
        // Initialise the sequence
        remote_initialise_rc5_sequence(sequence);
    }
    else
    {
        // We're in a sequence
        // Check if were within the range of a half pulse
		if ((diff > PULSE_WIDTH_SHORT_MIN) && (diff < PULSE_WIDTH_SHORT_MAX))
		{
			// time advanced by a half cycle - toggle transition flag
			sequence->in_transition = !sequence->in_transition;
		}
		else if ((diff > PULSE_WIDTH_LONG_MIN) && (diff < PULSE_WIDTH_LONG_MAX))
		{
            // Its been a whole pulse since the last interrupt, no need to do anything
        }
		else 
		{
            // This is an invalid sequence - re-initialise to start reading
            // Initialise the sequence
            remote_initialise_rc5_sequence(sequence);
        }
    }

    //logger_log_debugf(0, "REMOTE MESS %d %d %d %d --", ir_value, diff, sequence->in_transition, sequence->bit_pos);
        
    // Now set the bit value
    if (sequence->in_transition)
    {
        if (sequence->bit_pos == 0)
        {
            // This should be low
            if (!ir_value)
            {
                // Good start
            }
            else
            {
                // Abandon message
                // This is an invalid sequence - re-initialise to start reading
                // Initialise the sequence
                remote_initialise_rc5_sequence(sequence);
            }
        }
        else if (sequence->bit_pos == 1)
        {
            // This is the field bit which determines whether the command range is 0-63 0r 64-127
            sequence->field_bit = ir_value;
        }
        else if (sequence->bit_pos == 2)
        {
            sequence->repeat = true;
            if (remote_data.prev_toggle != ir_value)
            {
                sequence->repeat = false;
            }
            else
            {
                if (timers_has_expired(TIMER_INDEX_REMOTE_RESET_REPEAT))
                {
                    sequence->repeat = false;
                }
                else
                {
                   //logger_log_debugf(0, "REMOTE IS REPEAT %d %d %d --", !ir_value, sequence->bit_pos, diff);
                }
             }
            
            timers_reset(TIMER_INDEX_REMOTE_RESET_REPEAT);
            remote_data.prev_toggle = ir_value;
        }
        else if ((sequence->bit_pos >= 3) && (sequence->bit_pos <= 7))
        {
            // Device address
            uint8_t shift = sequence->bit_pos - 3;
            sequence->address <<= 1;
            sequence->address |= (ir_value ? 0 : 1);
        }
        else if ((sequence->bit_pos >= 8) && (sequence->bit_pos <= 13))
        {
            // Command 
            uint8_t shift = sequence->bit_pos - 8;
            sequence->command <<= 1;
            sequence->command |= (ir_value ? 0 : 1);
        }

        sequence->bit_pos++;

        if (sequence->bit_pos == 14)
        {
            // Apply the field bit - bump the command from 0-63 to 64-127
            if (sequence->field_bit)
            {
                sequence->command += 0x40;
            }
            
            // Reset the repeat timer so that the next few repeats are ignored if they come too fast
            // This is to allow for single presses that actually generate a couple of repeats
            if (!sequence->repeat)
            {
                timers_reset(TIMER_INDEX_REMOTE_REPEAT);
            }
            
            //logger_log_debugf(0, "REMOTE FINISHED %d %d %d %d --", remote_data.rc5_sequence_pool_write_idx, sequence->repeat, sequence->address, sequence->command);

            sequence->complete = true;
            
            // Move write index on
            remote_data.rc5_sequence_pool_write_idx++;
            remote_data.rc5_sequence_pool_write_idx %= RC5_SEQUENCE_POOL_SIZE;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Initialise the remote 
 */
void remote_initialise(STRUCTURED_MESSAGE_CALLBACK callback)
{
    remote_data.callback = callback;
    settings_get(SETTING_REMOTE_SYSTEM_CODE, &remote_data.system_code);
    
    remote_clear();
}

/**-----------------------------------------------------------------------------
 * Called before a sleep to ensure the remote picks up the first interrupt command
 */
void remote_pause()
{
    //TODO  No longer required
}

/**-----------------------------------------------------------------------------
 * Process remote messages 
 */
void remote_process()
{
    // First attempt - process 1 message per cycle
    RC5_SEQUENCE* sequence = &remote_data.rc5_sequence_pool[remote_data.rc5_sequence_pool_read_idx];
    if (sequence->complete)
    {
        // Only process the message if it is a. not a repeat, or b. if it is a repeat and happens
        // after an initial button press.
        if (!sequence->repeat || timers_has_expired(TIMER_INDEX_REMOTE_REPEAT))
        {
            // We were checking that the remote message matched to system_code for the amp
            // however we want to display all IR codes in the Engineering menu so
            // let them past here and filter in app_context processing
            // Check if the IR command was intended for this system by checking the 
            // address (defined in system_setup.h)
            // Check if we have a callback defined
            if (remote_data.callback != NULL)
            {
                // Create a structured message
                STRUCTURED_MESSAGE message;
                message.source = STRUCTURED_MESSAGE_SOURCE_REMOTE_RAW;
                message.zone = STRUCTURED_MESSAGE_ZONE_1;
                message.command = sequence->command;
                message.response = STATUS_UPDATE;
                message.data_length = 0;    // Gets set below
                message.repeat = sequence->repeat;

                structured_message_set_message_data(&message, &sequence->address, sizeof(REMOTE_SYSTEM_CODE));

                // Pass the message onto a callback method for validation and processing
                remote_data.callback(&message);
            }
        }
        
        remote_initialise_rc5_sequence(sequence);
        
        remote_data.rc5_sequence_pool_read_idx++;
        remote_data.rc5_sequence_pool_read_idx %= RC5_SEQUENCE_POOL_SIZE;
    }
}
    
/**-----------------------------------------------------------------------------
 * Sets the remote system code
 * @param code
 */
void remote_set_system_code(REMOTE_SYSTEM_CODE code)
{
    if ((remote_data.system_code != code) && 
       ((code == 16) || (code == 19) || (code == 21) || (code == 25)))
    {
        remote_data.system_code = code;
        settings_set(SETTING_REMOTE_SYSTEM_CODE, &remote_data.system_code);
        
        logger_log_infof(LOGGER_CODE_REMOTE_SYSTEM_CODE, "REMOTE: System Code changed %d", remote_data.system_code);
    }
}
    
/**-----------------------------------------------------------------------------
 * Return true if the message has the same system code as the current system code
 * @param message
 * @return 
 */
bool remote_use_message(STRUCTURED_MESSAGE* message)
{
    bool result = remote_data.system_code == remote_get_system_code_from_message(message);
//        logger_log_warningf(LOGGER_CODE_REMOTE_SYSTEM_CODE, "REMOTE: System Code received expected:%d received:%d --", remote_data.system_code, remote_get_system_code_from_message(message));
    if (!result)
    {
        logger_log_warningf(LOGGER_CODE_REMOTE_SYSTEM_CODE, "REMOTE: Unknown System Code received expected:%d received:%d --", remote_data.system_code, remote_get_system_code_from_message(message));
    }
    
    return result;
}

/* *****************************************************************************
 End of File
 */
