/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    rgb_led.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/rgb_led.h"
#include <stdbool.h>
#include "utilities/timers.h"
#include "utilities/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define RGB_LED_DUTY_CYCLE_STEPS     32

typedef void (*RGB_SET_METHOD)(bool on);
    
typedef struct
{
    uint8_t r;
    uint8_t g;
    uint8_t b;
} RGB_LED_COLOR_VALUE;

//! Color values - note these are indexed by the RGB_LED_COLOUR enum
// NOTE: These may be reset in initialise_rgb_led as they may differ between devices
static RGB_LED_COLOR_VALUE rgb_color_values[] = {
    {0, 0, 0},      // OFF
    {31, 0, 0},     // RED
    {0, 31, 0},     // GREEN
    {0, 0, 31},     // BLUE
    {31, 5, 0},     // ORANGE
    {31, 0, 31},    // PURPLE
    {0, 0, 31},    // WHITE {31, 16, 10}
};

typedef struct
{
    // The current colour combination
    RGB_LED_COLOUR colour;
    
    // Determines whether the led flashes or not
    bool flash;
    
    // Holds current state of LED's on/off
    bool flash_on;
    
    //! Current position in the PWM duty cycle
    uint8_t duty_cycle_pos;
    
    //! The step number for the 3 RGB colours in the PWM duty cycle that it switches from
    // on to off
    uint8_t pwm_switch_off[3];
    
    //! the switch that needs checking in the current cycle - this is intended to 
    // minimise the amount of checking that is done.
    uint8_t current_switch;
    
    RGB_SET_METHOD pwm_switch_off_methods[3];
    
    //! Used to determine whether we need PWM to generate the desired colour
    bool needs_pwm;
} RGB_LED_DATA;

static RGB_LED_DATA rgb_led_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * 
 * @param on
 */
static void set_blue(bool on)
{
    PLIB_PORTS_PinWrite(RGB_LED_BLUE_PORTS_ID, 
                        RGB_LED_BLUE_PORT_CHANNEL, 
                        RGB_LED_BLUE_PIN,
                        on);
}

/**-----------------------------------------------------------------------------
 * 
 * @param on
 */
static void set_green(bool on)
{
    PLIB_PORTS_PinWrite(RGB_LED_GREEN_PORTS_ID, 
                        RGB_LED_GREEN_PORT_CHANNEL, 
                        RGB_LED_GREEN_PIN,
                        on);
}

/**-----------------------------------------------------------------------------
 * 
 * @param on
 */
static void set_red(bool on)
{
    PLIB_PORTS_PinWrite(RGB_LED_RED_PORTS_ID, 
                        RGB_LED_RED_PORT_CHANNEL, 
                        RGB_LED_RED_PIN,
                        on);
}

/**-----------------------------------------------------------------------------
 * 
 * @param value
 */
static void set_colours(RGB_LED_COLOUR value)
{
    RGB_LED_COLOR_VALUE colour = rgb_color_values[value];

    rgb_led_set_colour(colour.r, colour.g, colour.b);
}

/**
 * tests whether the colour requires PWM
 * @param colour
 */
static bool test_pwm(uint8_t red, uint8_t green, uint8_t blue)
{
    bool result = false;
    
    // Work out whether we need PWM or not
    if (((red > 0) && (red < (RGB_LED_DUTY_CYCLE_STEPS - 1))) ||
        ((green > 0) && (green < (RGB_LED_DUTY_CYCLE_STEPS - 1))) ||
        ((blue > 0) && (blue < (RGB_LED_DUTY_CYCLE_STEPS - 1))))
    {
        result = true;
        //logger_log_debugf(245, "RGB: Needs PWM for %d %d %d ", red, green, blue);
    }
    else
    {
        //logger_log_debugf(245, "RGB: Does NOT Need PWM for %d %d %d ", red, green, blue);
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Can be used to adjust the default value settings for a colour - used for
 * example to match the White colour with the CDS50
 * @param colour
 * @param red
 * @param green
 * @param blue
 */
void rgb_led_adjust_colour(RGB_LED_COLOUR colour, uint8_t red, uint8_t green, uint8_t blue)
{
    rgb_color_values[colour].r = red;
    rgb_color_values[colour].g = green;
    rgb_color_values[colour].b = blue;
}
    
/**-----------------------------------------------------------------------------
 * Sets the RGB colour of the led
 * @param red
 * @param green
 * @param blue
 */
void rgb_led_set_colour(uint8_t red, uint8_t green, uint8_t blue)
{
    // Adjust for duty cycle resolution
//    red = RGB_LED_DUTY_CYCLE_STEPS * red / 256;
//    green = RGB_LED_DUTY_CYCLE_STEPS * green / 256;
//    blue = RGB_LED_DUTY_CYCLE_STEPS * blue / 256;
    
    //logger_log_debugf(0, "RGB LED:A r:%d g:%d b:%d --", red, green, blue);

    rgb_led_data.needs_pwm = test_pwm(red, green, blue);
    
    // work out the order on which the colours need switching off, the higher the 
    // RGB value the longer the colour will be on for.
    if (red > green)
    {
        if (blue > red)
        {
            //logger_log_debugf(0, "RGB:A r:%d g:%d b:%d --", red, green, blue);
            rgb_led_data.pwm_switch_off[0] = blue;
            rgb_led_data.pwm_switch_off[1] = red;
            rgb_led_data.pwm_switch_off[2] = green;
            rgb_led_data.pwm_switch_off_methods[0] = set_blue;
            rgb_led_data.pwm_switch_off_methods[1] = set_red;
            rgb_led_data.pwm_switch_off_methods[2] = set_green;
        }
        else
        {
            if (blue > green)
            {
                //logger_log_debugf(0, "RGB:B r:%d g:%d b:%d --", red, green, blue);
                rgb_led_data.pwm_switch_off[0] = red;
                rgb_led_data.pwm_switch_off[1] = blue;
                rgb_led_data.pwm_switch_off[2] = green;
                rgb_led_data.pwm_switch_off_methods[0] = set_red;
                rgb_led_data.pwm_switch_off_methods[1] = set_blue;
                rgb_led_data.pwm_switch_off_methods[2] = set_green;
            }
            else
            {
                //logger_log_debugf(0, "RGB:C r:%d g:%d b:%d --", red, green, blue);
                rgb_led_data.pwm_switch_off[0] = red;
                rgb_led_data.pwm_switch_off[1] = green;
                rgb_led_data.pwm_switch_off[2] = blue;
                rgb_led_data.pwm_switch_off_methods[0] = set_red;
                rgb_led_data.pwm_switch_off_methods[1] = set_green;
                rgb_led_data.pwm_switch_off_methods[2] = set_blue;
            }
        }
    }
    else
    {
        if (blue > green)
        {
            //logger_log_debugf(0, "RGB:D r:%d g:%d b:%d --", red, green, blue);
            rgb_led_data.pwm_switch_off[0] = blue;
            rgb_led_data.pwm_switch_off[1] = green;
            rgb_led_data.pwm_switch_off[2] = red;
            rgb_led_data.pwm_switch_off_methods[0] = set_blue;
            rgb_led_data.pwm_switch_off_methods[1] = set_green;
            rgb_led_data.pwm_switch_off_methods[2] = set_red;
       }
        else
        {
            if (blue > red)
            {
                //logger_log_debugf(0, "RGB:E r:%d g:%d b:%d --", red, green, blue);
                rgb_led_data.pwm_switch_off[0] = green;
                rgb_led_data.pwm_switch_off[1] = blue;
                rgb_led_data.pwm_switch_off[2] = red;
                rgb_led_data.pwm_switch_off_methods[0] = set_green;
                rgb_led_data.pwm_switch_off_methods[1] = set_blue;
                rgb_led_data.pwm_switch_off_methods[2] = set_red;
            }
            else
            {
                //logger_log_debugf(0, "RGB:F r:%d g:%d b:%d --", red, green, blue);
                rgb_led_data.pwm_switch_off[0] = green;
                rgb_led_data.pwm_switch_off[1] = red;
                rgb_led_data.pwm_switch_off[2] = blue;
                rgb_led_data.pwm_switch_off_methods[0] = set_green;
                rgb_led_data.pwm_switch_off_methods[1] = set_red;
                rgb_led_data.pwm_switch_off_methods[2] = set_blue;
            }
        }   
    }
    
    rgb_led_data.duty_cycle_pos = 0;
    
    rgb_led_process_interrupt();
}

/**-----------------------------------------------------------------------------
 * Flash the led
 */
void rgb_led_flash()
{
    // Check if we're already flashing
    if (!rgb_led_data.flash)
    {
        logger_log_debugf(0, "RGB LED Flash --");
        rgb_led_data.flash = true;
        rgb_led_data.flash_on = true;
    
        timers_reset(TIMER_INDEX_RGB_LED);
    }
}

/**-----------------------------------------------------------------------------
 * Gets the current RGB LED colour
 * @return 
 */
RGB_LED_COLOUR rgb_led_get_colour()
{
    return rgb_led_data.colour;
}
    
/**-----------------------------------------------------------------------------
 * Returns whether the led is flashing or not
 */
bool rgb_led_get_flash()
{
    return rgb_led_data.flash;
}
    
/**-----------------------------------------------------------------------------
 * Initialises the RGB LEd module
 */
void rgb_led_initialise()
{
    rgb_led_data.duty_cycle_pos = 0;
    rgb_led_set(RGB_LED_COLOUR_OFF);
}
    
/**-----------------------------------------------------------------------------
 * Set the colors of the led - can be a combination of Red-Green-Blue
 * @param value - an Or'd combination of RGB_LED_COLOR constants
 */
void rgb_led_set(RGB_LED_COLOUR value)
{
    if (rgb_led_data.colour != value)
    {
        rgb_led_data.colour = value;
    
        //logger_log_debugf(0, "RGB LED Setting colour %d --", value);
        // Reset the colour if not already flashing (and hence resetting)
        set_colours(rgb_led_data.colour);
    }
}

/**-----------------------------------------------------------------------------
 * Flash the led
 */
void rgb_led_process()
{
    if (rgb_led_data.flash)
    {
        if (timers_has_expired(TIMER_INDEX_RGB_LED))
        {
            if (rgb_led_data.flash_on)
            {
                set_colours(RGB_LED_COLOUR_OFF);
            }
            else
            {
                set_colours(rgb_led_data.colour);
            }
            
            rgb_led_data.flash_on = !rgb_led_data.flash_on;
            
            timers_reset(TIMER_INDEX_RGB_LED);
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Used as an interrupt event timer - added to timers initialisation as
 * timers_add_expired_event(TIMER_INDEX_RGB_PWM, rgb_led_process_interrupt);
 */
void rgb_led_process_interrupt()
{
    int x;
    for (x = 0; x < 3; x++)
    {
        if (rgb_led_data.duty_cycle_pos == rgb_led_data.pwm_switch_off[x])
        {
            rgb_led_data.pwm_switch_off_methods[x](false);
        }
        else if (rgb_led_data.duty_cycle_pos == 0)
        {
            rgb_led_data.pwm_switch_off_methods[x](true);
        }
    }

    if (rgb_led_data.needs_pwm)
    {
        rgb_led_data.duty_cycle_pos++;
        rgb_led_data.duty_cycle_pos = rgb_led_data.duty_cycle_pos % RGB_LED_DUTY_CYCLE_STEPS;
    
        timers_reset(TIMER_INDEX_RGB_PWM);
    }
}

/**-----------------------------------------------------------------------------
 * Stop the LED flashing
 */
void rgb_led_flash_stop()
{
    logger_log_debugf(0, "RGB LED Flash Stop %d --", rgb_led_data.colour);
    rgb_led_data.flash = false;
    rgb_led_data.flash_on = false;
    set_colours(rgb_led_data.colour);
}

/* *****************************************************************************
 End of File
 */
