/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    tcpip.c

  @Summary
 TCP/IP management

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/tcpip.h"
#include "utilities/logger.h"
#include "utilities/timers.h"
#include "app_config.h"
#include "peripheral/adc/plib_adc.h"
#include <sys/errno.h>
#include <time.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Maximum number of client connections note to work with the UDP stack this number 
//  needs to be a maximum of the Harmony TCP config "Maximum Number Of Sockets Supported" - the number of UDP sockets + 1
#ifndef TCPIP_MAX_CLIENTS
#define TCPIP_MAX_CLIENTS       4 //MAX_BSD_SOCKETS - 4    
#endif

//! The size of the buffer used to hold outgoing data
#define TCPIP_SEND_BUFFER_SIZE  120 // Size of structured message + some....

//! Port number for TCPIP Server to listen on
#define TCPIP_SERVER_PORT       50000


typedef enum
{
    TCPIP_STATE_WAIT_INIT,
	/* Application's state machine's initial state. */
    TCPIP_STATE_WAIT_FOR_IP, 
    TCPIP_STATE_GET_MAC_ADDRESS,
    TCPIP_STATE_BSD_INIT,
    TCPIP_STATE_BSD_CREATE_SOCKET,
    TCPIP_STATE_BSD_BIND,
    TCPIP_STATE_BSD_LISTEN,
    TCPIP_STATE_BSD_OPERATION,
    TCPIP_STATE_CABLE_DISCONNECTED,
    TCPIP_STATE_ERROR,

} TCPIP_STATE;

typedef struct
{
    //! A reference to the actual socket
    SOCKET client_socket;
    
    //! The address of the client that made the connection
    struct sockaddr_in address;
    
} TCPIP_CLIENT_CONNECTION;
/**
 * Global TCP/IP data
 */
typedef struct
{
    // Client sockets
    TCPIP_CLIENT_CONNECTION client_connections[TCPIP_MAX_CLIENTS];
    
    // Holds the current message read from the client socket
    STRUCTURED_MESSAGE  current_message[TCPIP_MAX_CLIENTS];
    
    // Indicates whether to use dhcp or a static ip address
    bool dhcp_enabled;
    
    // True once TCPIP_Initialise has been run
    bool initialised;
    
    // The MAC address of the network card
    TCPIP_MAC_ADDR mac_address;
            
    // A friendly name for the unit on the network
    char network_name[TCPIP_NETWORK_NAME_MAX_LEN];

    // The IP Address currently used.
    IPV4_ADDR runtime_address;
    
    // Records the running DHCP state prior to any changes so that it can be 
    // reviewed when calling tcpip_start_connection
    bool prev_dhcp_state;
    
    // Records the running Static IP state prior to any changes so that it can be 
    // reviewed when calling tcpip_start_connection
    IPV4_ADDR prev_static_ip_address;
    
    //! A circular buffer to hold outgoing data
    char send_buffer[TCPIP_SEND_BUFFER_SIZE];
    
    //! The amount of outstanding data to be send (in bytes)
    int send_buffer_len;
    
    //! The current start offset in the send buffer
    int send_buffer_offset;
    
    // Server socket
    SOCKET server_socket;
    
    // The static IP Address if set
    IPV4_ADDR static_address;

    // A pointer to the object returned from TCPIP_STACK_Initialize
    // As this is available through system_definitions.h its passed its generally set to sysObj.tcpip
    SYS_MODULE_OBJ* tcpip_module;
    
    // The current state of the TCP/IP connection
    TCPIP_STATE state;
        
    // Callback function supplied to tcpip_initialise - called when a message is 
    // received
    STRUCTURED_MESSAGE_CALLBACK callback;

    //! Callback invoked when a new client connection is established
    TCP_CONNECTED_CALLBACK connected_callback;
    
    // A buffer used to receive data from a socket - will be overwritten regularly
    unsigned char receive_buffer[STRUCTURED_MESSAGE_MAX_SIZE];

} TCPIP_DATA;

static TCPIP_DATA tcpip_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

// Forward declarations
static void tcpip_receive(int client_idx);

/**
 * Check whether the IP address is valid
 * @param ip_address
 * @return 
 */
static bool tcpip_address_is_valid(IP_ADDR* ip_address)
{
    bool result = false;
    
           // Check whether this has changed?
    if (ip_address->v[0] != 0 && ip_address->v[0] != 169) // Wait for a Valid IP
    {
        result = true;
    }

    return result;
}
/**
 * Bind the server socket
 */
static bool tcpip_bind()
{
    bool result = true;
    
    // Bind socket to a local port
    struct sockaddr_in address;
    int addrlen = sizeof(struct sockaddr_in);
    address.sin_port = TCPIP_SERVER_PORT;
    address.sin_addr.S_un.S_addr = IP_ADDR_ANY;

    if (bind(tcpip_data.server_socket, (struct sockaddr*)&address, addrlen) == SOCKET_ERROR)
    {
        logger_log_error(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Unable to bind server socket");

        result = false;       
    }
    {
        logger_log_info(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Server Socket bound");
    }
    
    return result;
}

/**
 * Setup client sockets
 */
static void tcpip_clear_client_sockets()
{
    logger_log_info(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Clear Sockets");
    
    int x;
    for(x = 0; x < TCPIP_MAX_CLIENTS; x++)
    {
        // closesocket not required - we would need to check if it was connected first
        //closesocket(tcpip_data.client_connections[x].client_socket);
        tcpip_data.client_connections[x].client_socket = INVALID_SOCKET;
    }
}

/*
 * Close the Server and Client sockets prior to changing the machines IP address
 */
static void tcpip_close_connection()
{
    logger_log_info(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Closing connection");
    
    logger_log_debugf(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Closing server socket");
    if (tcpip_data.server_socket != INVALID_SOCKET)
    {
		closesocket(tcpip_data.server_socket);
		tcpip_data.server_socket = INVALID_SOCKET;
    }
    
    int x;
    for(x = 0; x < TCPIP_MAX_CLIENTS; x++)
    {
        logger_log_debugf(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Closing client socket %d --", x);
        if (tcpip_data.client_connections[x].client_socket != INVALID_SOCKET)
        {
			closesocket(tcpip_data.client_connections[x].client_socket);
			tcpip_data.client_connections[x].client_socket = INVALID_SOCKET;
		}
    }
}

/**
 * Setup client sockets
 */
static bool tcpip_create_server_socket()
{
    bool result = true;
    
    // Create a socket for this server to listen and accept connections on
    tcpip_data.server_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(tcpip_data.server_socket == INVALID_SOCKET)
    {
        logger_log_error(LOGGER_CODE_TCPIP_SOCKET_ERROR, "TCPIP Unable to create server socket");
        
        result = false;
    }
    else
    {
        logger_log_info(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Server Socket created");
    }
    
    return result;
}

/**
 * Gets the IP address from DHCP
 * return true if a valid IP Address has been obtained
 */
static bool tcpip_dhcp_get_ip()
{
    bool result = false;
    int count = TCPIP_STACK_NumberOfNetworksGet();  // There should only be one..
    IPV4_ADDR ip_address;
    
    //logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP DHCP networks %d", count);    
        
    int x;
    for (x = 0; x < count; x++)
    {
        TCPIP_NET_HANDLE handle = TCPIP_STACK_IndexToNet(x);
 
        //logger_log_info(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Test Enabling DHCP");
        if (!TCPIP_DHCP_IsEnabled(handle))
        {
            logger_log_info(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Enabling DHCP");
            TCPIP_DHCP_Enable(handle);
        }
        
        // Wait for DHCP to be up and running
        if (TCPIP_DHCP_IsBound(handle))
        {
            //logger_log_info(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP TCPIP_DHCP_IsBound");
            
            // Get the IP address allocated
            ip_address.Val = TCPIP_STACK_NetAddress(handle);

            // Check whether this has changed?
            if (tcpip_address_is_valid(&ip_address)) // Wait for a Valid IP
            {
                tcpip_data.runtime_address.Val = ip_address.Val;

                char buffer[20];
                TCPIP_Helper_IPAddressToString(&tcpip_data.runtime_address, buffer, 20);
                logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP DHCP IP %s", buffer);    

                if (tcpip_data.network_name[0] != '\0')
                {
                    logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP Setting Network Name %s", tcpip_data.network_name);    
                    TCPIP_STACK_NetBiosNameSet(handle, tcpip_data.network_name);
                }
                else
                {
                    // allow host to obtain ping-able host name
                    strncpy(tcpip_data.network_name, TCPIP_STACK_NetBIOSName(handle), TCPIP_NETWORK_NAME_MAX_LEN);

                    logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP BIOS Network Name %s", tcpip_data.network_name);    
                }

                result = true;
            }
        }
    }
    
    return result;
}

/**
 * Gets the IP address from DHCP
 * return true if a valid IP Address has been obtained
 */
static bool tcpip_get_ip_mac_address()
{
    bool result = false;
    
    int count = TCPIP_STACK_NumberOfNetworksGet();  // There should only be one..
    IPV4_ADDR ip_address;
    
    //logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP DHCP networks %d", count);    
    
    int x;
    for (x = 0; x < count; x++)
    {
        TCPIP_NET_HANDLE handle = TCPIP_STACK_IndexToNet(x);

        // Check whether this has changed?
        if (tcpip_address_is_valid(&tcpip_data.runtime_address)) // Wait for a Valid IP
        {
            TCPIP_MAC_ADDR* mac_address;
            mac_address = (TCPIP_MAC_ADDR*)TCPIP_STACK_NetAddressMac(handle);

            tcpip_data.mac_address.v[0] = mac_address->v[0];
            tcpip_data.mac_address.v[1] = mac_address->v[1];
            tcpip_data.mac_address.v[2] = mac_address->v[2];
            tcpip_data.mac_address.v[3] = mac_address->v[3];
            tcpip_data.mac_address.v[4] = mac_address->v[4];
            tcpip_data.mac_address.v[5] = mac_address->v[5];

            char buffer[20];
            TCPIP_Helper_MACAddressToString(&tcpip_data.mac_address, buffer, 20);
            logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP MAC Address %s", buffer);    
            
            result = true;
        }
    }
    
    return result;
}

/**
 * Setup static IP address
 */
static void tcpip_initialise_static_ip()
{
    int count = TCPIP_STACK_NumberOfNetworksGet();
    IPV4_ADDR mask;
 
    // Initialise the mask (assume it will be the same for all networks)
    mask.v[0] = 255;
    mask.v[1] = 255;
    mask.v[2] = 255;
    mask.v[3] = 0;
    
    int x;
    for(x = 0; x < count; x++)
    {
        TCPIP_NET_HANDLE handle = TCPIP_STACK_IndexToNet(x);
        
        tcpip_data.runtime_address.Val = tcpip_data.static_address.Val;
        
        logger_log_infof(LOGGER_CODE_TCPIP_SET_STATIC_IP, "TCPIP: Setting Static IP %d.%d.%d.%d", 
                tcpip_data.runtime_address.v[0],
                tcpip_data.runtime_address.v[1],
                tcpip_data.runtime_address.v[2],
                tcpip_data.runtime_address.v[3]);

        if (TCPIP_DHCP_IsEnabled(handle))
        {
            logger_log_info(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Disabling DHCP");
            TCPIP_DHCP_Disable(handle);
        }

        TCPIP_STACK_NetAddressSet(handle, &tcpip_data.runtime_address, &mask, true);
   }
}
    
/**
 * Initialises a structured message
 * @param message
 */
static void tcpip_initialise_structured_message(STRUCTURED_MESSAGE* message)
{
    structured_message_initialise(message, STRUCTURED_MESSAGE_SOURCE_TCPIP);
    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
};

/**
 * Read incoming data and generate message
 */
static void tcpip_process_buffers()
{
    int length;
    struct sockaddr_in address;
    int addrlen = sizeof(struct sockaddr_in);
    int error = 0;
    
//    char buffer[20];
//    TCPIP_Helper_IPAddressToString(&tcpip_data.runtime_address, buffer, 20);
//    logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP reading %s", buffer);    
    
    int x;
    for (x = 0; x < TCPIP_MAX_CLIENTS; x++)
    {
        bool new_connection = false;
        
        // Accept any pending connection requests, assuming we have a place to store the socket descriptor
        if(tcpip_data.client_connections[x].client_socket == INVALID_SOCKET)
        {
            //logger_log_infof(LOGGER_CODE_TCPIP_READ_ERROR, "TCPIP Attempting connect %d", x);
            tcpip_data.client_connections[x].client_socket = accept(tcpip_data.server_socket, (struct sockaddr*)&address, &addrlen);
            new_connection = true;
        }

        // If this socket is not connected then no need to process anything
        if(tcpip_data.client_connections[x].client_socket == INVALID_SOCKET)
            continue;

        if (new_connection)
        {
            tcpip_data.client_connections[x].address = address;
            
            logger_log_infof(LOGGER_CODE_TCPIP_READ_ERROR, "TCPIP Socket Connected %d %d.%d.%d.%d:%d  ", x,
                    tcpip_data.client_connections[x].address.sin_addr.S_un.S_un_b.s_b1,
                    tcpip_data.client_connections[x].address.sin_addr.S_un.S_un_b.s_b2,
                    tcpip_data.client_connections[x].address.sin_addr.S_un.S_un_b.s_b3,
                    tcpip_data.client_connections[x].address.sin_addr.S_un.S_un_b.s_b4,
                    tcpip_data.client_connections[x].address.sin_port
                    );
            
            if (tcpip_data.connected_callback != NULL)
            {
                tcpip_data.connected_callback(true, &address, &addrlen);
            }
        }
        
        // Check if there is anything coming in
        tcpip_receive(x);
    }
}

/**
 * Receive incoming data and turn into messages
 */
static void tcpip_receive(int client_idx)
{
    if (tcpip_data.client_connections[client_idx].client_socket != INVALID_SOCKET)
    {
        // See if there is any data to be retrieved into the global work buffer
        int count = recv(tcpip_data.client_connections[client_idx].client_socket, 
                tcpip_data.receive_buffer, 
                sizeof(tcpip_data.receive_buffer), 
                0);

        if(count > 0)
        {
//            logger_log_debugf(0, "TCPIP receive %x %x %x %x %x %x - %d %d", tcpip_data.receive_buffer[0], 
//                    tcpip_data.receive_buffer[1], 
//                    tcpip_data.receive_buffer[2], 
//                    tcpip_data.receive_buffer[3], 
//                    tcpip_data.receive_buffer[4], 
//                    tcpip_data.receive_buffer[5],
//                    tcpip_data.current_message[client_idx].processing_state,
//                    tcpip_data.current_message[client_idx].command);
            // Added a brief pause as occasionally it does not get the data in non-blocking mode
            //timers_delay_us(10);
            structured_message_process_buffer(&tcpip_data.current_message[client_idx], tcpip_data.receive_buffer, count);
        }
        else if (count == SOCKET_ERROR)
        {
            // Clear the socket down
            if (errno == ENOTCONN)
            {
                closesocket(tcpip_data.client_connections[client_idx].client_socket);

                logger_log_infof(LOGGER_CODE_TCPIP_READ_ERROR, "TCPIP Socket Disconnected %d %d.%d.%d.%d:%d  ", client_idx,
                    tcpip_data.client_connections[client_idx].address.sin_addr.S_un.S_un_b.s_b1,
                    tcpip_data.client_connections[client_idx].address.sin_addr.S_un.S_un_b.s_b2,
                    tcpip_data.client_connections[client_idx].address.sin_addr.S_un.S_un_b.s_b3,
                    tcpip_data.client_connections[client_idx].address.sin_addr.S_un.S_un_b.s_b4,
                    tcpip_data.client_connections[client_idx].address.sin_port
                    );

                tcpip_data.client_connections[client_idx].client_socket = INVALID_SOCKET;
                //logger_log_infof(LOGGER_CODE_TCPIP_READ_ERROR, "TCPIP Socket Disconnected %d ", client_idx);
            }
        }
        else if (count < 0)
        {
            // Clear the socket down
            closesocket(tcpip_data.client_connections[client_idx].client_socket);
            logger_log_infof(LOGGER_CODE_TCPIP_READ_ERROR, "TCPIP Socket Unexpected error %d %d %d  %d.%d.%d.%d:%d", client_idx, count, errno,
                tcpip_data.client_connections[client_idx].address.sin_addr.S_un.S_un_b.s_b1,
                tcpip_data.client_connections[client_idx].address.sin_addr.S_un.S_un_b.s_b2,
                tcpip_data.client_connections[client_idx].address.sin_addr.S_un.S_un_b.s_b3,
                tcpip_data.client_connections[client_idx].address.sin_addr.S_un.S_un_b.s_b4,
                tcpip_data.client_connections[client_idx].address.sin_port
                );
            tcpip_data.client_connections[client_idx].client_socket = INVALID_SOCKET;
        }

        // If we've received a message  then lets process it
        if (tcpip_data.current_message[client_idx].processing_state == STRUCTURED_MESSAGE_STATE_PERFORM_ACTION)
        {

            // Pass the message onto a callback method for validation and processing
            if (tcpip_data.callback != NULL)
            {
                tcpip_data.current_message[client_idx].state_data[0] = client_idx;

                logger_log_infof(0, "TCPIP Received Message");
                tcpip_data.callback(&tcpip_data.current_message[client_idx]);
            }

            // Re initialise the message
            tcpip_initialise_structured_message(&tcpip_data.current_message[client_idx]);
        }

        // If we need to send a reply then lets get that started
        if (tcpip_data.current_message[client_idx].processing_state == STRUCTURED_MESSAGE_STATE_SEND_REPLY)
        {
            tcpip_data.current_message[client_idx].state_data[0] = client_idx;

            tcpip_send_structured_message(&tcpip_data.current_message[client_idx]);
        }
    }
}

/**
 * Event used to detect when the network cable is inserted/removed
 */   
static void tcpip_register_event(TCPIP_NET_HANDLE hNet, TCPIP_EVENT event, const void* fParam)
{
    //TCPIP_EV_CONN_ESTABLISHED|TCPIP_EV_CONN_LOST
    switch(event)
    {
        case TCPIP_EV_CONN_ESTABLISHED:
        {
            logger_log_debugf(0, "TCPIP Network Cable Connected");
            tcpip_data.state = TCPIP_STATE_BSD_OPERATION;

            if (tcpip_data.connected_callback != NULL)
            {
                tcpip_data.connected_callback(true, NULL, NULL);
            }

            break;
        }
        
        case TCPIP_EV_CONN_LOST:
        {
            logger_log_debugf(0, "TCPIP Network Cable Disconnected");
            tcpip_data.state = TCPIP_STATE_CABLE_DISCONNECTED;
            
            if (tcpip_data.connected_callback != NULL)
            {
                tcpip_data.connected_callback(false, NULL, NULL);
            }

            break;
        }
    }
}

/**
 * Setup static Ip address
 */
static bool tcpip_update_static_ip()
{
    bool result = true;

    tcpip_data.runtime_address.Val = tcpip_data.static_address.Val;
    
    int count = TCPIP_STACK_NumberOfNetworksGet();

    int x;
    for (x = 0; x < count; x++)
    {
        TCPIP_NET_HANDLE handle = TCPIP_STACK_IndexToNet(x);

        TCPIP_STACK_NetAddressSet(handle, &tcpip_data.runtime_address, NULL, 0);

        if (tcpip_data.network_name[0] != '\0')
        {
            TCPIP_STACK_NetBiosNameSet(handle, tcpip_data.network_name);
        }
        
        result = true;
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/** 
 * Close all open TCPIP connections
 */
void tcpip_close()
{
    tcpip_close_connection();
}

/** 
 * Performs steps to connect TCP/IP to a network
 */
void tcpip_connect()
{
    // We should have read the settings by now and know whether we are doing DHCP or Static IP
    // If we are already connected then disconnect
    if (!tcpip_data.initialised)
    {
        return;
    }
    tcpip_data.initialised = true;

    switch(tcpip_data.state)
    {
        case TCPIP_STATE_WAIT_INIT:
        {
            SYS_STATUS status = TCPIP_STACK_Status(*tcpip_data.tcpip_module);
            if (status == SYS_STATUS_READY)
            {
                logger_log_info(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP Closing connection");
                
                tcpip_close_connection();
                
                logger_log_info(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP SYS_STATUS_READY");
                
                // If using static IP then set it
                if (!tcpip_data.dhcp_enabled)
                {
                    tcpip_initialise_static_ip();
                }

                tcpip_data.state = TCPIP_STATE_WAIT_FOR_IP;
            }
            else
            {
                if (status != SYS_STATUS_ERROR)
                {
                    logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP NOT YET READY STATE %d --", status);
                }
            }
            break;
        }

        case TCPIP_STATE_WAIT_FOR_IP:
        {
            if (!tcpip_data.dhcp_enabled)
            {
                // Static IP
                // NOT SURE WHY THIS IS NEEDED GIVEN STEP 1 UNLESS WE RETURN TO THIS POINT
                // WHEN WE CHANGE STATIC IP ADDRESS
                logger_log_info(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP Update Static IP");
                if (tcpip_update_static_ip())
                {
                    tcpip_data.state = TCPIP_STATE_GET_MAC_ADDRESS;
                }          
            }
            else
            {
                // DHCP
                //logger_log_info(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP get DHCP IP");
                if (tcpip_dhcp_get_ip())
                {
                    tcpip_data.state = TCPIP_STATE_GET_MAC_ADDRESS;
                }
            }
            
            break;
        }        
        
        case TCPIP_STATE_GET_MAC_ADDRESS:
        {
            if (tcpip_get_ip_mac_address())
            {
                tcpip_data.state = TCPIP_STATE_BSD_INIT;
            }
            
            break;
        }
        
        case TCPIP_STATE_BSD_INIT:
        {
            tcpip_clear_client_sockets();

            tcpip_data.state = TCPIP_STATE_BSD_CREATE_SOCKET;

            break;
        }
        
        case TCPIP_STATE_BSD_CREATE_SOCKET:
        {
            // This will loop if there is an error! 
            if (tcpip_create_server_socket())
            {
                tcpip_data.state = TCPIP_STATE_BSD_BIND;
            }
            
            break;
        }
        
        case TCPIP_STATE_BSD_BIND:
        {
            if (tcpip_bind())
            {
                tcpip_data.state = TCPIP_STATE_BSD_LISTEN;   
            }
            
            break;
        }
        
        case TCPIP_STATE_BSD_LISTEN:
        {
            if(listen(tcpip_data.server_socket, TCPIP_MAX_CLIENTS) == 0)
            {
                logger_log_info(LOGGER_CODE_TCPIP_SOCKET_LISTENING, "TCPIP Listening");

                // Register an event handler to detect cable connect/disconnect
                //! Note this needs "Enable User Notification" to be on in harmony for some reason!
                TCPIP_STACK_HandlerRegister(TCPIP_STACK_NetDefaultGet(), 
                        TCPIP_EV_CONN_ESTABLISHED | TCPIP_EV_CONN_LOST,
                        tcpip_register_event, 0);
             
                // Invoke the connected callback
                if (tcpip_data.connected_callback != NULL)
                {
                    tcpip_data.connected_callback(true, NULL, NULL);
                }
                
                tcpip_data.state = TCPIP_STATE_BSD_OPERATION; // Handled in tcpip_process())
            }

            break;
        }
        
        default:
            break;
    }
}
    
/**
 * Returns whether DHCP is enabled or not
 * @param enable
 */
bool tcpip_get_dhcp_enabled()
{
    return tcpip_data.dhcp_enabled > 0;
}
    
/**
 * Gets a displayable version if the IP Address
 * @param buffer
 */
void tcpip_get_ip_address_string(char* buffer, size_t size)
{
    if (!TCPIP_Helper_IPAddressToString(&tcpip_data.runtime_address, buffer, size))
    {
        sprintf(buffer, "%d.%d.%d.%d", 
                tcpip_data.runtime_address.v[0],
                tcpip_data.runtime_address.v[1],
                tcpip_data.runtime_address.v[2],
                tcpip_data.runtime_address.v[3]);
    }
}
    
/**
 * Returns a string representation of the current MAC address
 * @return 
 */
void tcpip_get_mac_address_string(char* buffer, size_t size)
{
    // If its not been initialised then do it here
    if ((tcpip_data.mac_address.v[0] == 0) &&
        (tcpip_data.mac_address.v[1] == 0) &&
        (tcpip_data.mac_address.v[2] == 0) &&
        (tcpip_data.mac_address.v[3] == 0) &&
        (tcpip_data.mac_address.v[4] == 0) &&
        (tcpip_data.mac_address.v[5] == 0))
    {
        //PLIB_ETH_MACGetAddress(0, tcpip_data.mac_address.v);
        TCPIP_NET_HANDLE hNet = TCPIP_STACK_NetHandleGet("PIC32INT"); 
        memcpy(&tcpip_data.mac_address, TCPIP_STACK_NetAddressMac(hNet), sizeof(TCPIP_MAC_ADDR)); 
    }
         
    if (!TCPIP_Helper_MACAddressToString(&tcpip_data.mac_address, buffer, size))
    {
        sprintf(buffer, "%2.2x%2.2x%2.2x%2.2x%2.2x%2.2x", 
            tcpip_data.mac_address.v[0],
            tcpip_data.mac_address.v[1],
            tcpip_data.mac_address.v[2],
            tcpip_data.mac_address.v[3],
            tcpip_data.mac_address.v[4],
            tcpip_data.mac_address.v[5]);
    }
}
    
/**
 * Returns a string representation of the current MAC address
 * @return 
 */
void tcpip_get_network_name_string(char* buffer, size_t size)
{
    snprintf(buffer, min(size, TCPIP_NETWORK_NAME_MAX_LEN), "%s", tcpip_data.network_name);
}

/**
 * Gets the static IP Address
 */
IP_ADDR tcpip_get_runtime_ip_address()
{
    return tcpip_data.runtime_address;
}

/**
 * Gets the static IP Address
 */
IP_ADDR tcpip_get_static_ip_address()
{
    return tcpip_data.static_address;
}

/**
 * Initialise the tcpip module
 */
void tcpip_initialise(SYS_MODULE_OBJ* tcpip_module, 
        STRUCTURED_MESSAGE_CALLBACK callback,
        TCP_CONNECTED_CALLBACK connected_callback)
{
    //tcpip_data.dhcp_enabled = true;
    tcpip_data.tcpip_module = tcpip_module;
    tcpip_data.callback = callback;
    tcpip_data.connected_callback = connected_callback;
    tcpip_data.state = TCPIP_STATE_WAIT_INIT;
    
    settings_get(SETTING_TCPIP_DHCP_ENABLED, &tcpip_data.dhcp_enabled);
    settings_get(SETTING_TCPIP_STATIC_IP_ADDRESS, &tcpip_data.static_address);
    settings_get(SETTING_TCPIP_NETWORK_NAME, tcpip_data.network_name);

    // Turn on dhcp if no static ip address available.
    if (tcpip_data.static_address.Val == 0)
    {
        tcpip_data.dhcp_enabled = true;
    }
    
    tcpip_data.prev_dhcp_state = tcpip_data.dhcp_enabled;
    tcpip_data.prev_static_ip_address.Val = tcpip_data.static_address.Val;
     
    logger_log_infof(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP DHCP %s", tcpip_data.dhcp_enabled ? APP_STRINGS[APP_STRINGS_ENUM_ON] : APP_STRINGS[APP_STRINGS_ENUM_OFF] );
    logger_log_infof(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Static IP %d.%d.%d.%d", 
                tcpip_data.static_address.v[0],
                tcpip_data.static_address.v[1],
                tcpip_data.static_address.v[2],
                tcpip_data.static_address.v[3]);
    logger_log_infof(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Name %s", tcpip_data.network_name);
    
    tcpip_data.send_buffer_offset = 0;
    tcpip_data.send_buffer_len = 0;
    
    int x;
    for (x = 0; x < TCPIP_MAX_CLIENTS; x++)
    {
        // Initialise the message
        tcpip_initialise_structured_message(&tcpip_data.current_message[x]);
    }
    
    // Get the mac address upfront (assumes ETHERNET_MODULE_ID = 0), otherwise its not 
    // initalised until we setup the TCPIP stack
    // NOTE: This needs the network PSU to be enabled! (Generally done in APP_INIT - power_enable_by_index(POWER_INDEX_NETWORK_PSU))
//    PLIB_ETH_MACGetAddress(0, tcpip_data.mac_address.v);
    char buffer[20];
    TCPIP_NET_HANDLE hNet = TCPIP_STACK_NetHandleGet("PIC32INT"); 
    memcpy(&tcpip_data.mac_address, TCPIP_STACK_NetAddressMac(hNet), sizeof(TCPIP_MAC_ADDR)); 
    TCPIP_Helper_MACAddressToString(&tcpip_data.mac_address, buffer, 20);
    logger_log_infof(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP ETH MAC Address %s", buffer);  
    
    tcpip_data.initialised = true;
}

/** 
 * Determines whether TCP/IP is connected to a network
 */
bool tcpip_is_connected()
{
    return tcpip_data.state == TCPIP_STATE_BSD_OPERATION;
}

/**
 * Perform TCP/IP processing
 */
void tcpip_process()
{
   // Check whether we are in a DHCP/StaticIP change
    if (timers_has_expired(TIMER_INDEX_TCPIP_CHANGE))
    {
        tcpip_review_connection();
    }
    
     // Now review current state
    switch(tcpip_data.state)
    {
        case TCPIP_STATE_BSD_OPERATION:
        {
//            if (TCPIP_STACK_NetIsLinked(TCPIP_STACK_NetDefaultGet()))
//            {
//                logger_log_debugf(0, "CONTROL4 Network Cable In");
//            }
//            else
//            {
//                 logger_log_debugf(0, "CONTROL4 Network Cable In");
//            }
    
            tcpip_process_buffers();
            break;
        }
        
        default:
            tcpip_connect();
            break;
    }    
}

/*
 * Reviews the TCPIP connection process
 */
void tcpip_review_connection()
{
    if ((tcpip_data.dhcp_enabled != tcpip_data.prev_dhcp_state) ||
        (tcpip_data.static_address.Val != tcpip_data.prev_static_ip_address.Val))
    {
        logger_log_info(LOGGER_CODE_TCPIP_IP_ADDRESS, "TCPIP Starting connection");
        tcpip_data.state = TCPIP_STATE_WAIT_INIT;
        
        tcpip_data.prev_dhcp_state = tcpip_data.dhcp_enabled;
        tcpip_data.prev_static_ip_address.Val = tcpip_data.static_address.Val;
    }
}
/**
 * Send a bufer of data to the specified client
 * @param client_idx - The index of the client socket
 * @param buffer - A pointer to the buffer of data to be sent
 * @param len - The length of the buffer
 */
void tcpip_send_buffer(uint8_t client_idx, char* buffer, int len)
{
    // Allocate a buffer to hold the message
    SOCKET client_socket;

    // Find the socket that originated it
    client_socket = tcpip_data.client_connections[client_idx].client_socket;
    
    if (client_socket != INVALID_SOCKET)
    {
        // Send the buffer
        int total = 0;
        do
        {
            int count = send(client_socket, 
                             &buffer[total], 
                             len - total, 0);

            if (count == SOCKET_ERROR)
            {
                //logger_log_errorf(0, "TCPIP: send error for client %d", client_idx);
                tcpip_data.client_connections[client_idx].client_socket = INVALID_SOCKET;
                break;
            }

            total += count;
        } while(total < len);
    }
}

/**
 * Send a bufer of data to the specified client
 * @param client_idx - The index of the client socket
 * @param buffer - A pointer to the buffer of data to be sent
 * @param len - The length of the buffer
 */
void tcpip_send_buffer_to(uint8_t client_idx, char* buffer, int len)
{
    // Allocate a buffer to hold the message
    SOCKET client_socket;

    // Find the socket that originated it
    client_socket = tcpip_data.client_connections[client_idx].client_socket;
    
    if (client_socket != INVALID_SOCKET)
    {
        // Send the buffer
        int total = 0;
        do
        {
            int count = send(client_socket, 
                             &buffer[total], 
                             len - total, 0);

            if (count == SOCKET_ERROR)
            {
                logger_log_errorf(0, "TCPIP: send error for client %d", client_idx);
                break;
            }

            total += count;
        } while(total < len);
    }
    else
    {
        logger_log_errorf(0, "TCPIP: attempt to send to closed socket %d", client_idx);
    }
}

/**
 * Send a structured message over TCPIP
 * @param message
 */
void tcpip_send_structured_message(STRUCTURED_MESSAGE* message)
{
    char buffer[STRUCTURED_MESSAGE_MAX_SIZE];
    
    // Fill the buffer with the message
    int len = structured_message_write_response_buffer(message, buffer, STRUCTURED_MESSAGE_MAX_SIZE);
    
    if (message->state_data[0] != TCPIP_UNSOLICITED_CLIENT_CONNECTION)
    {
        tcpip_send_buffer(message->state_data[0], buffer, len);
    }
    else
    {
        // Unsolicited messages are broadcast to all active clients
        int x;
        for (x = 0; x < TCPIP_MAX_CLIENTS; x++)
        {
//TODO - Review this is causing problems with Control4/UDP sendto            tcpip_send_buffer(x, buffer, len);
        }
    }
}

/**
 * Control whether DHCP is enabled or not
 * @param enable
 */
void tcpip_set_dhcp_enabled(bool enable)
{
    if (tcpip_data.dhcp_enabled != enable)
    {
        // Store the previous state
        tcpip_data.prev_dhcp_state = tcpip_data.dhcp_enabled;
        
        tcpip_data.dhcp_enabled = enable;
        
        settings_set(SETTING_TCPIP_DHCP_ENABLED, &tcpip_data.dhcp_enabled);
        
        // Trigger a reconnect (Note DHCP timeout is quicker than static IP)
        // This is designed so that if the user sits in the DHCP menu it will 
        // automatically reset the connection after 4 seconds
        timers_update_timeout(TIMER_INDEX_TCPIP_CHANGE, TIMER_TCPIP_DHCP_CHANGE_TIMEOUT_MS, TIMERS_TIMEOUT_RESET_IMMEDIATE);
    }
}
 
/**
 * Sets the network name for the device
 * @param enable
 */
void tcpip_set_network_name(char* name, size_t size)
{
    if (strncmp(tcpip_data.network_name, name, min(size, TCPIP_NETWORK_NAME_MAX_LEN)) != 0)
    {
        memset(tcpip_data.network_name, 0, TCPIP_NETWORK_NAME_MAX_LEN);
        memcpy(tcpip_data.network_name, name, min(size, TCPIP_NETWORK_NAME_MAX_LEN));
        
        settings_set(SETTING_TCPIP_NETWORK_NAME, &tcpip_data.network_name);
    }
}

/**
 * Sets the static IP Address
 */
void tcpip_set_static_ip_address(IP_ADDR* ip_address)
{
    if ((tcpip_data.static_address.Val != ip_address->Val) ||
        (tcpip_get_dhcp_enabled()))
    {
        // Store the previous state
        tcpip_data.prev_static_ip_address.Val = tcpip_data.static_address.Val;
        
        // Check the IP address is not 0.0.0.0 - if so set DHCP on
        if (ip_address->Val != 0)
        {
            tcpip_data.static_address.Val = ip_address->Val;

            char buffer[15];
            sprintf(buffer, "%d.%d.%d.%d", 
                tcpip_data.static_address.v[0],
                tcpip_data.static_address.v[1],
                tcpip_data.static_address.v[2],
                tcpip_data.static_address.v[3]);
            logger_log_infof(0, "TCPIP Setting static IP %s", buffer);

            settings_set(SETTING_TCPIP_STATIC_IP_ADDRESS, &tcpip_data.static_address);
 
            if (tcpip_get_dhcp_enabled())
            {
                tcpip_set_dhcp_enabled(false);
            }
            
            // Trigger a reconnect (Note DHCP timeout is quicker than static IP)
            // This is only relevant if the user is sitting in the Static IP menu and 
            // taking time to set the IP address in which case the current IP address 
            // will be used for a reset after about 20 seconds inactivity in the menu.
            timers_update_timeout(TIMER_INDEX_TCPIP_CHANGE, TIMER_TCPIP_STATIC_IP_CHANGE_TIMEOUT_MS, TIMERS_TIMEOUT_RESET_IMMEDIATE);
        }
        else
        {
            // Note: The actual value of the static IP is not cleared but is saved
            // so if DHCP is switched off again it will revert to the previous static IP.
            tcpip_set_dhcp_enabled(true);
        }
    }
}

/* *****************************************************************************
 End of File
 */
