/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    tft_control.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

/* This section lists the other files that are included in this file.
 */

#include "utilities/tft_control.h"
#include "app_config.h"
#include "utilities/display_message_display.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define TFT_CONTROL_MAX_CONTROLS        DISPLAY_MESSAGE_SCREEN_DATA_MAX_DATA_ITEMS
#define TFT_CONTROL_MAX_SCROLL_GROUPS   2

typedef struct
{
    //! Holds the current number of active controls
    uint8_t active_control_count;
    
    TFT_CONTROL controls[TFT_CONTROL_MAX_CONTROLS];
    
    TFT_CONTROL_SCROLL_GROUP scroll_groups[TFT_CONTROL_MAX_SCROLL_GROUPS];
} TFT_CONTROL_DATA;


static TFT_CONTROL_DATA tft_control_data = {
    .active_control_count = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * 
 * @param data_id
 * @param attribute_id
 * @param draw_function
 * @return 
 */
static void tft_control_scroll_group_add(uint8_t scroll_group, TFT_CONTROL* ctrl)
{
    if ((scroll_group > 0) && (scroll_group <= TFT_CONTROL_MAX_SCROLL_GROUPS))
    {
        TFT_CONTROL_SCROLL_GROUP* group = &tft_control_data.scroll_groups[scroll_group - 1];
        group->control_count++;
        ctrl->control_data.text.scroll_group = scroll_group;
        
        //logger_log_debugf(0, "TFT CONTROL ADD %d %d %d --", scroll_group, group->control_count, ctrl->control_data.text.scroll_group);
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param data_id
 * @param attribute_id
 * @param draw_function
 * @return 
 */
static void tft_control_scroll_group_remove(uint8_t scroll_group, TFT_CONTROL* ctrl)
{
    if ((scroll_group > 0) && (scroll_group <= TFT_CONTROL_MAX_SCROLL_GROUPS))
    {
        TFT_CONTROL_SCROLL_GROUP* group = &tft_control_data.scroll_groups[scroll_group - 1];
        if (group->control_count > 0)
        {
            group->control_count--;
        }
        ctrl->control_data.text.scroll_group = 0;
        
        //logger_log_debugf(0, "TFT CONTROL REMOVE %d %d %d --", scroll_group, group->control_count, ctrl->control_data.text.scroll_group);
    }    
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * gets the current number of active controls
 * @param data_id
 * @param attribute_id
 * @return 
 */
uint8_t tft_control_active_controls()
{
    return tft_control_data.active_control_count;
}

/**-----------------------------------------------------------------------------
 * Allocates a new TFT control
 * @return 
 */
TFT_CONTROL* tft_control_allocate(DISPLAY_MESSAGE_DATA_ID data_id,
                                  uint8_t attribute_id,
                                  TFT_CONTROL_DRAW_FUNCTION draw_function)
{
    TFT_CONTROL* result = NULL;
    
    if (tft_control_data.active_control_count < TFT_CONTROL_MAX_CONTROLS)
    {
        //logger_log_debugf(LOGGER_CODE_TFT_CONTROL, "TFT CONTROL Allocating control %d %d at %d --", data_id, attribute_id, tft_control_data.active_control_count);
        result = &tft_control_data.controls[tft_control_data.active_control_count];
        tft_control_data.active_control_count++;
        
        // Set up the control data
        result->data_id = data_id;
        result->attribute_id = attribute_id;
        result->draw_function = draw_function;
        result->visible = true;
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_TFT_CONTROL, "TFT CONTROL Unable to allocate control --");
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Sets the area defined for the control to the background colour
 * @param redraw_immediate
 */
void tft_control_clear(TFT_CONTROL* ctrl, bool redraw_immediate)
{
    PIXEL pixels[tft_display_bounds_area(&ctrl->bounds)];

    // Clear the background - assumes that there are no overlapping controls
    tft_display_draw_rectangle(pixels, &ctrl->bounds, COLOUR_INDEX_IMAGE_BACKGROUND );

    tft_display_write_buffer(&ctrl->bounds, pixels, true);

    if (redraw_immediate)
    {
        tft_control_draw(ctrl, true);
    }
}

/**-----------------------------------------------------------------------------
 * Used when displaying the screen to clear down/free control data ready for controls to be
 * defined.
 */
void tft_control_clear_controls()
{
    logger_log_debugf(LOGGER_CODE_TFT_CONTROL, "TFT CONTROL Clearing controls --");
    
    tft_control_data.active_control_count = 0;
    
//    tft_control_scroll_group_initialise();
}
 
/**-----------------------------------------------------------------------------
 * 
 * @param ctrl
 * @param top
 * @param left
 * @param bottom
 * @param right
 * @return 
 */
void tft_control_define_bounds(TFT_CONTROL* ctrl, uint16_t top, uint16_t left, uint16_t bottom, uint16_t right)
{
    if (ctrl != NULL)
    {
        ctrl->bounds.top = top;
        ctrl->bounds.left = left;
        ctrl->bounds.bottom = bottom;
        ctrl->bounds.right = right;
    }
}

/**-----------------------------------------------------------------------------
 */
void tft_control_define_google_voice_properties(TFT_CONTROL* ctrl, 
        COLOUR_INDEX foreground_colour)
{
    if (ctrl != NULL)
    {
        ctrl->type = TFT_CONTROL_TYPE_GOOGLE_VOICE;
        ctrl->control_data.google_voice.foreground = foreground_colour;
        ctrl->control_data.google_voice.cycle_length = 120;
        ctrl->control_data.google_voice.cycle_position = 0;
    }        
}
 
/**-----------------------------------------------------------------------------
 */
void tft_control_define_image_properties(TFT_CONTROL* ctrl, TFT_CONTROL_ALIGNMENT alignment)
{
    if (ctrl != NULL)
    {
        ctrl->type = TFT_CONTROL_TYPE_IMAGE;
        ctrl->control_data.image.alignment = alignment;
        ctrl->control_data.image.being_retrieved = false;
    }    
}

/**-----------------------------------------------------------------------------
 */
void tft_control_define_progress_properties(TFT_CONTROL* ctrl,
        COLOUR_INDEX foreground_colour, 
        COLOUR_INDEX background_colour)
{
    if (ctrl != NULL)
    {
        ctrl->type = TFT_CONTROL_TYPE_PROGRESS;
        ctrl->control_data.progress.progress = 0;
        ctrl->control_data.progress.prev_progress = 0;
        ctrl->control_data.progress.foreground = foreground_colour;
        ctrl->control_data.progress.background = background_colour;
    }    
}

/**-----------------------------------------------------------------------------
 * 
 * @param ctrl
 * @param top
 * @param left
 * @param bottom
 * @param right
 * @return 
 */
void tft_control_define_rect(TFT_CONTROL* ctrl, uint16_t top, uint16_t left, uint16_t height, uint16_t width)
{
    if (ctrl != NULL)
    {
        ctrl->bounds.top = top;
        ctrl->bounds.left = left;
        ctrl->bounds.bottom = top + height - 1;
        ctrl->bounds.right = left + width - 1;
    }
}

/**-----------------------------------------------------------------------------
 */
void tft_control_define_rectangle_properties(TFT_CONTROL* ctrl,
        COLOUR_INDEX foreground_colour, 
        COLOUR_INDEX background_colour)
{
    if (ctrl != NULL)
    {
        ctrl->type = TFT_CONTROL_TYPE_RECTANGLE;
        ctrl->control_data.rectangle.foreground = foreground_colour;
        ctrl->control_data.rectangle.background = background_colour;
    }    
}

/**-----------------------------------------------------------------------------
 * Defines properties associated with the text control
 * @param data_id
 * @param attribute_id
 * @return 
 */
void tft_control_define_text_colours(TFT_CONTROL* ctrl, 
        COLOUR_INDEX foreground_colour, 
        COLOUR_INDEX background_colour)
{
    if (ctrl != NULL)
    {
        ctrl->control_data.text.foreground = foreground_colour;
        ctrl->control_data.text.background = background_colour;
    }
}

/**-----------------------------------------------------------------------------
 * Defines scroll info for text controls 
 * Note: This must be called after tft_control_define_text_properties
 */
void tft_control_define_scroll_info(TFT_CONTROL* ctrl, 
        uint8_t scroll_group)
{
    //if ((ctrl != NULL) && (ctrl->control_data.text.scroll_group != scroll_group))
    if (ctrl != NULL)
    {
        // Remove from scroll group
        tft_control_scroll_group_remove(ctrl->control_data.text.scroll_group, ctrl);
        
        // Add to new scroll group
        tft_control_scroll_group_add(scroll_group, ctrl);
    }    
}

/**-----------------------------------------------------------------------------
 * Defines properties associated with the text control
 * @param data_id
 * @param attribute_id
 * @return 
 */
void tft_control_define_text_properties(TFT_CONTROL* ctrl, 
        TFT_CONTROL_ALIGNMENT alignment, 
        FONT_ID font_id, 
        OFFSET offset,
        COLOUR_INDEX foreground_colour, 
        COLOUR_INDEX background_colour)
{
    if (ctrl != NULL)
    {
        //logger_log_debugf(0, "TFT CONTROL tft_control_define_text_properties --");
        ctrl->type = TFT_CONTROL_TYPE_TEXT;
        ctrl->control_data.text.alignment = alignment;
        ctrl->control_data.text.font_id = font_id;
        ctrl->control_data.text.offset = offset;
        ctrl->control_data.text.foreground = foreground_colour;
        ctrl->control_data.text.background = background_colour;
        
        // Remove from scroll group
        ctrl->control_data.text.scroll_offset = 0;
        ctrl->control_data.text.scroll_max = 0;
        tft_control_scroll_group_remove(ctrl->control_data.text.scroll_group, ctrl);
    }
}

/**-----------------------------------------------------------------------------
 * Invokes the controls draw function
 * @param 
 */
void tft_control_draw(TFT_CONTROL* ctrl, bool force)
{
    if ((ctrl != NULL) && (ctrl->draw_function != NULL) && (ctrl->visible))
    {
        // Check if it has changed by looking at the data item
        // Get the latest representation of the data item - that contains the old value
        // compare it with the new value
        // decide how to draw it...
        // Get a data item
        DISPLAY_MESSAGE_SCREEN_DATA_ITEM* data_item = display_message_screen_data_find_item(&display_message_metadata_current()->screen_data, ctrl->data_id, false);

        if (data_item != NULL)
        {
            //logger_log_debugf(0, "DISPLAY FOUND DATA ITEM %d %s --", ctrl->data_id, 
            //        data_item->title);

            if (force || (data_item->has_changed))
            {
                ctrl->draw_function(ctrl);
            }
        }
        else
        {
            //logger_log_debugf(0, "DISPLAY DID NOT FIND DATA ITEM %d --", ctrl->data_id);

            // No data item so draw anyway
            ctrl->draw_function(ctrl);
        }
    }
}

/**-----------------------------------------------------------------------------
 * attempts to draw all allocated controls
 */
void tft_control_draw_all(bool force)
{
    int x;
    for (x = 0; x < tft_control_data.active_control_count; x++)
    {
        tft_control_draw(&tft_control_data.controls[x], force);
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param bounds
 */
void tft_control_draw_progress(TFT_CONTROL* ctrl)
{
//logger_log_debugf(0, "DRAW PROGRESS %d %d --", ctrl->control_data.progress.prev_progress, ctrl->control_data.progress.progress);
    if ((ctrl->control_data.progress.prev_progress == 0) && (ctrl->control_data.progress.prev_progress == ctrl->control_data.progress.progress))
    {
        if (ctrl->visible)
        {
//logger_log_debugf(0, "DRAW PROGRESS Clear Background %d %d --", ctrl->control_data.progress.prev_progress, ctrl->control_data.progress.progress);
            PIXEL pixels_bkgd[tft_display_bounds_area(&ctrl->bounds)];

            // Clear the background - change to a draw rectangle //TODO Adjust for highlighted text
            tft_display_draw_rectangle(pixels_bkgd, &ctrl->bounds, ctrl->control_data.progress.background );

            tft_display_write_buffer(&ctrl->bounds, pixels_bkgd, true);
        }
    }
    else
    {
        // Work out the progress change
        // progress is always 0-99 so we can use an int8
        int8_t delta = ctrl->control_data.progress.progress - ctrl->control_data.progress.prev_progress;

        if (delta != 0)
        {
            // Caclulate the bounds of the area that has changed
            BOUNDS bounds;
            bounds.top = ctrl->bounds.top;
            bounds.left = ctrl->bounds.left + (((int16_t)min(ctrl->control_data.progress.progress, ctrl->control_data.progress.prev_progress) * tft_display_bounds_width(&ctrl->bounds)) / 100);
            bounds.bottom = ctrl->bounds.bottom;
            bounds.right = ctrl->bounds.left + (((int16_t)max(ctrl->control_data.progress.progress, ctrl->control_data.progress.prev_progress) * tft_display_bounds_width(&ctrl->bounds)) / 100);

            // Allocate pixels to draw the progress 
            PIXEL pixels[tft_display_bounds_area(&bounds)];

            int x;
            for (x = 0; x < tft_display_bounds_area(&bounds); x++)
            {
                if (delta > 0)
                {
                    colour_set(&pixels[x], ctrl->control_data.progress.foreground);
                }
                else
                {
                    colour_set(&pixels[x], ctrl->control_data.progress.background);
                }
            }

            tft_display_write_buffer(&bounds, pixels, true);

            ctrl->control_data.progress.prev_progress = ctrl->control_data.progress.progress;    
        }
    }
}

/**-----------------------------------------------------------------------------
 * 
 */
void tft_control_draw_text(TFT_CONTROL* ctrl)
{
     // Get the data item fo the image
    DISPLAY_MESSAGE_SCREEN_DATA_ITEM* data_item = display_message_screen_data_find_item(&display_message_metadata_current()->screen_data, ctrl->data_id, true);
    if (data_item != NULL)
    {
        // get the value from the data item
        // get the text and draw t...
        PIXEL pixels[tft_display_bounds_area(&ctrl->bounds)];
        
        // defines the ordinal index in the text where ellipses should be placed 0x0000 means no ellipses
        uint16_t ellipse_idx = 0x0000;

        // Clear the background - change to a draw rectangle //TODO Adjust for highlighted text
        tft_display_draw_rectangle(pixels, &ctrl->bounds, ctrl->control_data.text.background );

//        logger_log_debugf(0, "TEXT Data Item %d Attr %d --", data_item->data_id, ctrl->attribute_id);
//        logger_process();
        char* text = NULL;
        switch(ctrl->attribute_id)
        {
            case DISPLAY_MESSAGE_ATTRIBUTE_TITLE:
            {
                text = data_item->title;
                break;
            }
            
            case DISPLAY_MESSAGE_ATTRIBUTE_DATA:
            {
                text = data_item->value;
                break;
            }
            
            default:
            {
                // Not sure what to do here if its needs at all - perhaps needs some other kind of callback?
                break;
            }
        }
        
        if (text != NULL)
        {
            // Get a unicode representation of the string
            uint16_t unicode_text[strlen(text) + 1];
            memset(unicode_text, 0, strlen(text) + 1);

            font_manager_utf8_string_to_unicode(text, unicode_text);
            
            // Control offset allows for the font display to be offset mainly to 
            // handle fonts like NewsCycl that have a load of white space above the 
            // characters
            int x;
            int16_t top_offset = ctrl->control_data.text.offset.top;    // May need adjusting for padding inside control
            int16_t left_offset = ctrl->control_data.text.offset.left;
            
            FONT_ID font_id = ctrl->control_data.text.font_id;
            uint8_t character_spacing = 0; // The number of pixels separating characters

            uint16_t text_width = tft_display_get_unicode_text_width(font_id, unicode_text);
            uint16_t control_width = tft_display_bounds_width(&ctrl->bounds);
            
            //logger_log_debugf(0, "TEXT is %d %d %d %s %d %d %d %d %s --", data_item->data_id, strlen(text), text[40], data_item->title, font_manager_get_unicode_text_length(unicode_text), text_width, control_width, ctrl->control_data.text.scroll_group, text);
            //logger_process();
            //logger_log_debugf(0, "TEXT is %d %d %d %d %d %s --", data_item->data_id, strlen(text), font_manager_get_unicode_text_length(unicode_text), tft_display_get_text_width(font_id, text), text_width, text);
            //logger_process();
            
            // Store the length of the characters to be written
            uint16_t text_length = font_manager_get_unicode_text_length(unicode_text);//strlen(text);
            uint16_t loop_length = text_length;
            
            // Define the gap between the two instances of the text whe scrolling
            uint16_t scroll_text_separator = (tft_display_bounds_width(&ctrl->bounds) / 4);

            // adjust top offset to align text (top, center, bottom - needs addint to TFT_CONTROL definition)
            switch (ctrl->control_data.text.alignment)
            {
                case TFT_CONTROL_ALIGNMENT_TOP_LEFT:
                {
                    break;
                }

                case TFT_CONTROL_ALIGNMENT_CENTER:
                {
                    // use the conrol width  and the text width to calculate the
                    // top/left offset
                    left_offset = (control_width - text_width) / 2;
                     
                    break;
                }         
        
                case TFT_CONTROL_ALIGNMENT_TOP_RIGHT:
                {
                    left_offset = control_width - text_width;
                    break;
                }
                
                default:
                    break;
            }
            
            // adjust left offset for padding/ alignment etc...
            
            // Adjust for scrolling / truncation
            if (text_width > control_width)
            {
                //logger_log_debugf(0, "TFT CONTROL Scroll %d %d %d --", ctrl->control_data.text.scroll_group, text_width, scroll_text_separator);
                // Check if this is part of a scroll group
                if (ctrl->control_data.text.scroll_group > 0)
                {
                    // Update the width of the scrolling text
                    ctrl->control_data.text.scroll_max = text_width + scroll_text_separator;
                    
                    // Sort out scrolling
                    // Get a reference to the scroll group
                    // see if this is the current control in the scroll group 
                    left_offset -= ctrl->control_data.text.scroll_offset;
                    
                    loop_length = text_length * 2;
                }
                else
                {
                    // Sort out ellipses
                    // First work out the pixels required to hold 2 ellipses
                    UNICODE_CHAR ellipse = 0x002E;
                    uint8_t ellipse_width = font_manager_get_unicode_character_width(font_id, &ellipse); //".");
                    uint16_t available_text_width = control_width - left_offset - (ellipse_width * 2);
                    
                    // Get the index of the character that would get cut off
                    //ellipse_idx = tft_display_get_character_at_pos(font_id, text, available_text_width - 1);
                    ellipse_idx = tft_display_get_unicode_character_at_pos(font_id, unicode_text, available_text_width - 1);
                    
                    // Adjust the text length 
                    loop_length = ellipse_idx + 2; // For 2 ellipses
                }
            }

            for (x = 0; x < loop_length; x++)
            {
                //char character = text[x % text_length];
                UNICODE_CHAR character = unicode_text[x % text_length];
                
                //uint8_t character_width = font_manager_get_character_width(font_id, &character) + character_spacing;
                uint8_t character_width = font_manager_get_unicode_character_width(font_id, &character) + character_spacing;
                
                // Only draw what we need
                if ((left_offset + character_width - 1 > 0) && (left_offset <= ctrl->bounds.right))
                {
                    // Check when using ellipses (does not apply to scrolling))
                    if ((ellipse_idx != 0x0000) && (ellipse_idx <= x))
                    {
                        character = '.';
                    }

                    if (x != data_item->edit_pos)
                    {
                        tft_display_draw_unicode_character(pixels, 
                                &ctrl->bounds, 
                                top_offset,
                                left_offset,
                                font_id, 
                                &character,
                                ctrl->control_data.text.foreground,
                                ctrl->control_data.text.background);
                    }
                    else
                    {
                        tft_display_draw_unicode_character(pixels, 
                                &ctrl->bounds, 
                                top_offset,
                                left_offset,
                                font_id, 
                                &character,
                                ctrl->control_data.text.background,
                                ctrl->control_data.text.foreground);
                    }
                }
                                        
                // Adjust the left offset
                left_offset += character_width;
                
                if ((x > 0) && ((x + 1) % text_length == 0))
                {
                    // Add a separator to the width
                    left_offset += scroll_text_separator;
                }
                
                // Quit if we've drawn to the bounds
                if (left_offset > ctrl->bounds.right)
                {
                    break;
                }
            }
        }
        else
        {
            logger_log_debugf(0, "TEXT is NULL %d --", ctrl->data_id);
            logger_process();
        }
        
        //logger_log_debugf(0, "TEXT BOUNDS %d %d %d %d --", ctrl->bounds.top, ctrl->bounds.left, ctrl->bounds.bottom, ctrl->bounds.right);
        tft_display_write_buffer(&ctrl->bounds, pixels, true);
    }
    else
    {
        logger_log_debugf(0, "TEXT Data Item Not found %d --", ctrl->data_id);
        logger_process();
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param data_id
 * @param attribute_id
 * @return 
 */
TFT_CONTROL* tft_control_get(DISPLAY_MESSAGE_DATA_ID data_id, uint8_t attribute_id)
{
    TFT_CONTROL* result = NULL;
    
    //logger_log_debugf(0, "TFT GET %d %d %d --", data_id, attribute_id, tft_control_data.active_control_count);
    int x;
    for (x = 0; x < tft_control_data.active_control_count; x++)
    {
        TFT_CONTROL* ctrl = &tft_control_data.controls[x];
        if ((ctrl->data_id == data_id) && (ctrl->attribute_id == attribute_id))
        {
            //logger_log_debugf(0, "TFT GET a FOUND %d %d --", ctrl->data_id, ctrl->attribute_id);
            result = &tft_control_data.controls[x];//ctrl;
            break;
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the first control matching the data id - for use with Option Lists where only
 * once control represents a data Id
 * @param data_id
 * @param attribute_id
 * @return 
 */
TFT_CONTROL* tft_control_get_by_data_id(DISPLAY_MESSAGE_DATA_ID data_id)
{
    TFT_CONTROL* result = NULL;
    
    //logger_log_debugf(0, "TFT GET %d %d %d --", data_id, attribute_id, tft_control_data.active_control_count);
    int x;
    for (x = 0; x < tft_control_data.active_control_count; x++)
    {
        TFT_CONTROL* ctrl = &tft_control_data.controls[x];
        if (ctrl->data_id == data_id)
        {
            //logger_log_debugf(0, "TFT GET a FOUND %d %d --", ctrl->data_id, ctrl->attribute_id);
            result = &tft_control_data.controls[x];//ctrl;
            break;
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the control at the specified index - requires the called to know what order controls are i the screen
 * @param data_id
 * @param attribute_id
 * @return 
 */
TFT_CONTROL* tft_control_get_at(uint8_t index)
{
    TFT_CONTROL* result = NULL;
    
    if (index < tft_control_data.active_control_count)
    {
        result = &tft_control_data.controls[index];
    }
     
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets whether the control is visible or not
 */
bool tft_control_get_visibility(TFT_CONTROL* ctrl)
{
    bool result = false;
    if (ctrl != NULL)
    {
        result = ctrl->visible;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether any controls are currently allocated
 * @return 
 */
bool tft_control_has_controls()
{
    return tft_control_data.active_control_count > 0 ? true : false;
}
   
/**-----------------------------------------------------------------------------
 */
void tft_control_initialise()
{
    tft_control_scroll_group_initialise();
}

/**-----------------------------------------------------------------------------
 * Used to reset all active controls to the state where they have just been created
 */
void tft_control_reset_controls()
{
    int x;
    for (x = 0; x < tft_control_data.active_control_count; x++)
    {
        TFT_CONTROL* ctrl = &tft_control_data.controls[x];
        ctrl->visible = false;
        
        switch (ctrl->type)
        {
            case TFT_CONTROL_TYPE_TEXT:
            {
                break;
            }
            
            TFT_CONTROL_TYPE_PROGRESS:
            {
                ctrl->control_data.progress.progress = 0;
                ctrl->control_data.progress.prev_progress = 0;
                break;
            }
            
            TFT_CONTROL_TYPE_IMAGE:
            {
                ctrl->control_data.image.being_retrieved = false;
                break;
            }
            
            default:
                break;
            
        }
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param data_id
 * @param attribute_id
 * @param draw_function
 * @return 
 */
void tft_control_scroll_group_initialise()
{
    int x;
    for (x = 0; x < TFT_CONTROL_MAX_SCROLL_GROUPS; x++)
    {
        TFT_CONTROL_SCROLL_GROUP* group = &tft_control_data.scroll_groups[x];
        group->active_index = 0;
        group->control_count = 0;
        group->pause_count = 0;
        group->pause_max = 4; // 3 seconds TODO make configurable
    }
}
    
/**-----------------------------------------------------------------------------
 * Used to review control groups
 */
void tft_control_scroll_group_review()
{
    // Check timer
    bool increment_pause = false;
    if (timers_has_expired(TIMER_INDEX_TFT_CONTROL_TIMER))
    {
        increment_pause = true;
        timers_reset(TIMER_INDEX_TFT_CONTROL_TIMER);
    }
    
    uint8_t x;
    for (x = 0; x < TFT_CONTROL_MAX_SCROLL_GROUPS; x++)
    {
        TFT_CONTROL_SCROLL_GROUP* group = &tft_control_data.scroll_groups[x];
        if (group->control_count > 0)
        {
            //logger_log_debugf(0, "TFT CONTROL Reviewing Scroll Group %d %d --", x, group->control_count);
            // get the active control
            TFT_CONTROL* active_control = NULL;
            uint8_t y;
            uint8_t count = 0;
            for (y = 0; y < tft_control_data.active_control_count; y++)
            {
                TFT_CONTROL* ctrl = &tft_control_data.controls[y];
                if ((ctrl->type == TFT_CONTROL_TYPE_TEXT) && (ctrl->control_data.text.scroll_group == x + 1))
                {
            //logger_log_debugf(0, "TFT CONTROL Reviewing Scroll Group A %d %d --", y, ctrl->control_data.text.scroll_group);
                    if (group->active_index == count)
                    {
            //logger_log_debugf(0, "TFT CONTROL Reviewing Scroll Group B %d %d --", y, ctrl->control_data.text.scroll_group);
                        active_control = ctrl;
                        break;
                    }
                    count++;
                }
            }
            
            if (active_control != NULL)
            {
                // Tell the control to scroll
                tft_control_draw(active_control, true);

                if (group->control_count == 1)
                {
                    // Check pausing
                    if (increment_pause)
                    {
                        if (active_control->control_data.text.scroll_offset == 0)
                        {
                            group->pause_count++;
                            
                            if (group->pause_count == group->pause_max)
                            {
                                // Increment scroll pos and start scrolling
                                active_control->control_data.text.scroll_offset++;
                            }
                        }
                        else
                        {
                            active_control->control_data.text.scroll_offset++;
                        }
                    }
                    else
                    {
                        if (active_control->control_data.text.scroll_offset > 0)
                        {
                            active_control->control_data.text.scroll_offset++;
                        }
                    }
                }
                else
                {
                    // Move the control scroll offset on
                    // this could be used to control the speed of scrolling
                    active_control->control_data.text.scroll_offset++;
                }
                
                //logger_log_debugf(0, "TFT CONTROL Reviewing Scroll Group C %d %d --", y, active_control->control_data.text.scroll_offset, active_control->control_data.text.scroll_max);
                
                // Get the text length to determine whether we have reached the end of the scrolling
                if (active_control->control_data.text.scroll_offset > active_control->control_data.text.scroll_max)
                {
                    // Reset the control scroll pos
                    active_control->control_data.text.scroll_offset = 0;
                    
                    // Move to the next control
                    group->active_index++;
                    group->active_index %= group->control_count;

                    if (group->control_count == 1)
                    {
                        // reset the pause count
                        group->pause_count = 0;
                    }
                }
            }
        }
    }
}
   
/**-----------------------------------------------------------------------------
 */
void tft_control_set_visibility(TFT_CONTROL* ctrl, bool visible)
{
    if (ctrl != NULL)
    {
        ctrl->visible = visible;
    }
}

/**-----------------------------------------------------------------------------
 * Used to update the progress value of a cntrol
 */
void tft_control_update_progress_value(TFT_CONTROL* ctrl, uint8_t progress)
{
    if (ctrl != NULL)
    {
        ctrl->control_data.progress.progress = min(progress, 100);
    }    
}

/* *****************************************************************************
 End of File
 */
