/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/trigger.h"
#include "utilities/logger.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#ifndef TRIGGER_PINS_ENABLED
#define TRIGGER_PINS_ENABLED 0
#endif

typedef struct
{
    TRIGGER_IN_STATE trigger_in_state;
} TRIGGER_DATA;

static TRIGGER_DATA trigger_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**
 * Gets the state of the trigger in pin
 */
bool trigger_in_get()
{
    bool result = false;
    
#if TRIGGER_PINS_ENABLED    
    result = PLIB_PORTS_PinGet(TRIGGER_IN_PORTS_ID, 
                               TRIGGER_IN_PORT_CHANNEL, 
                               TRIGGER_IN_PIN);
#endif
    
    return result;
}
     
/**
 * Gets the calculated state of the trigger in pin
 */
TRIGGER_IN_STATE trigger_in_get_state()
{
    return trigger_data.trigger_in_state;
}
    
/**
 * Sets the state of the trigger in pin (used for debugging)
 */
void trigger_in_set_state(TRIGGER_IN_STATE state)
{
    trigger_data.trigger_in_state = state;
}

/**
 * Initiallse the trigger module
 */
void trigger_initialise()
{
    trigger_data.trigger_in_state = TRIGGER_IN_STATE_UNDEFINED;
}
    
/**
 * Perform trigger based processing
 * @return 
 */
void trigger_process()
{
    bool trigger_on = trigger_in_get();
    TRIGGER_IN_STATE new_state = trigger_on ? TRIGGER_IN_STATE_ON : TRIGGER_IN_STATE_OFF;
    
    if (trigger_data.trigger_in_state == new_state)
    {
        // reset the  timer
        timers_reset(TIMER_INDEX_TRIGGER);
    }
    else if (timers_has_expired(TIMER_INDEX_TRIGGER))
    {
        // Check if the timer has expired (so its been different for the duration of the timer)
        if (trigger_data.trigger_in_state != TRIGGER_IN_STATE_UNDEFINED)
        {
            logger_log_infof(LOGGER_CODE_TRIGGER_SET, "TRIGGER: In A set %s", trigger_on ? "TRIGGER_IN_STATE_ON" : "TRIGGER_IN_STATE_OFF");
            trigger_data.trigger_in_state = new_state;
        }
        else if (new_state == TRIGGER_IN_STATE_ON)
        {
            logger_log_infof(LOGGER_CODE_TRIGGER_SET, "TRIGGER: In B set %s", trigger_on ? "TRIGGER_IN_STATE_ON" : "TRIGGER_IN_STATE_OFF");
            trigger_data.trigger_in_state = new_state;
        }
        else
        {
            //New state is OFF so pin may not have been set, lets ignore for the moment
        }
    }
    else
    {
        // The state has changed but the timer hasn't expired so lets wait
    }
}

/**
 * Gets the state of the trigger out pin
 */
bool trigger_out_get()
{
    bool result = false;
    
#if TRIGGER_PINS_ENABLED    
    result = PLIB_PORTS_PinGet(TRIGGER_OUT_PORTS_ID, 
                               TRIGGER_OUT_PORT_CHANNEL, 
                               TRIGGER_OUT_PIN);
#endif    
    
    return result;   
}

/**
 * Sets the trigger out pin high
 */
void trigger_out_set(bool on)
{
    logger_log_infof(LOGGER_CODE_TRIGGER_SET, "TRIGGER: Out set %s", on == true ? "On" : "Off");
    
#if TRIGGER_PINS_ENABLED    
    PLIB_PORTS_PinWrite(TRIGGER_OUT_PORTS_ID, 
                        TRIGGER_OUT_PORT_CHANNEL, 
                        TRIGGER_OUT_PIN,
                        !on);    
#endif    
}
    
/* *****************************************************************************
 End of File
 */
