/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/upgrade_metadata.h"
#include <string.h>
#include "crypto/crypto.h"
#include "crypto/src/md5.h"
#include "peripheral/nvm/plib_nvm.h"
#include "peripheral/int/plib_int.h"
#include "app_config.h"
#include "utilities/flash_memory.h"

#ifdef UPGRADE_METADATA_INCLUDE_UPGRADE_FUNCTIONALITY
#include "utilities/upgrade.h"
#endif

#include "bootloader/src/nvm.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//#define UPGRADE_METADATA_STATUS_LOADED          0
//#define UPGRADE_METADATA_STATUS_LOADING         1

//! Define a default value for UPGRADE_IMAGE_NUM_IMAGES
#ifndef UPGRADE_IMAGE_NUM_IMAGES
#define UPGRADE_IMAGE_NUM_IMAGES                2    
#endif

/**
 * A generic Id that can be used to identify the upgrade module when a number of
 * modules may be using the same storage
 */
#ifndef UPGRADE_METADATA_CLIENT_ID
#define UPGRADE_METADATA_CLIENT_ID              0
#endif

#define IMAGE_STATE_OPTIONS \
    X(UNINITIALISED)\
    X(UPLOADING)\
    X(UPLOADED)\
    X(LOADING)\
    X(LOADED)\
    X(ERROR)\

// Now use the input data to define the enum
#define X(a) IMAGE_STATE_##a,
	typedef enum {IMAGE_STATE_OPTIONS} IMAGE_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a) #a,
	static char *IMAGE_STATE_STRINGS[] = {IMAGE_STATE_OPTIONS};
#undef X
    
//typedef enum
//{
//    //! The image has never been initialised / used
//    IMAGE_STATE_UNINITIALISED,
//            
//    //! The image has been uploaded and is ready to be loaded into program memory
//    IMAGE_STATE_UPLOADING,
//            
//    //! The image has been uploaded and is ready to be loaded into program memory
//    IMAGE_STATE_UPLOADED,
//            
//    //! The image is being loaded into program memory
//    IMAGE_STATE_LOADING,
//            
//    //! The image has been successfully loaded into program memory
//    IMAGE_STATE_LOADED,
//            
//    //! The image has been unsuccessfully loaded into program memory
//    IMAGE_STATE_ERROR,
//            
//} IMAGE_STATE;

/**
 * Defines an upgrade image
 */
typedef struct
{
    UPGRADE_METADATA_LOAD_ID load_id;
    
    //! An MD5 Checksum of the data
    uint8_t checksum[MD5_DIGEST_SIZE];
    
    //! The size of the metadata
    uint32_t size;
    
    //! Version info
    uint8_t version_major;
    uint8_t version_minor;

    //! The status of the image data 
    IMAGE_STATE state;
    
    //! callback invoked when upload data has been written
    UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback;
    
} IMAGE_METADATA;

/**
 * Maintains upgrade info
 */
typedef struct
{
    UPGRADE_METADATA_LOAD_ID current_load_id;
    
    //! The index of the currently loaded image (0..UPGRADE_IMAGE_NUM_IMAGES-1)
    uint8_t current_idx;

    //! The index of the image_data to be used as an upgrade - this is the index of the Image 
    // that had the last load 
    uint8_t upgrade_idx;
    
    //! The overall upgrade state
    UPGRADE_METADATA_STATE state;

    //! An array of image data structures containing data on individual image info
    IMAGE_METADATA image_metadata[UPGRADE_IMAGE_NUM_IMAGES];
} UPGRADE_METADATA;

#define UPLOAD_BUFFER_SIZE  8192

/**
 * Maintains upgrade info
 */
typedef struct
{
    //! MD5 context structure used in MD5 checksum generation
    CRYPT_MD5_CTX md5_context;
        
    //!! used during upload to keep track of the bytes written in the current load
    uint32_t bytes_written;
    
    uint16_t curr_write_pos;
    uint8_t write_buffer[UPLOAD_BUFFER_SIZE];
    
} UPLOAD_RUNTIME_DATA;

/**
 * 
 */
typedef enum
{
    UPGRADE_RUNTIME_STATE_INIT,
    UPGRADE_RUNTIME_STATE_ERASE_FLASH,
    UPGRADE_RUNTIME_STATE_OPEN_IMAGE,
    UPGRADE_RUNTIME_STATE_LOAD_BLOCK,
    UPGRADE_RUNTIME_STATE_LOADING_BLOCK,
    UPGRADE_RUNTIME_STATE_CLOSE_IMAGE,
    UPGRADE_RUNTIME_STATE_COMPLETE,
    UPGRADE_RUNTIME_STATE_ERROR,
            
} UPGRADE_RUNTIME_STATE;

//! The length of an upgrade data block
#define UPGRADE_RUNTIME_DATA_LEN  512

/**
 * Data used during uploads
 */
typedef struct
{
    //! memory used to store image data read from flash which is then written to program memory
    char data[UPGRADE_RUNTIME_DATA_LEN];
    
    //! The current offset of the data to be read in the image data
    uint32_t offset;
    
    UPGRADE_RUNTIME_STATE state;
} UPGRADE_RUNTIME_DATA;

/**
 */
typedef enum
{
    UPGRADE_DATA_STATE_LOADING,
    UPGRADE_DATA_STATE_LOADED,
    UPGRADE_DATA_STATE_UPLOADING,
    UPGRADE_DATA_STATE_END_UPLOAD,
    UPGRADE_DATA_STATE_UPGRADING,
} UPGRADE_DATA_STATE;

typedef struct
{
    //! Indicates whether metadata is loaded from flash
    UPGRADE_DATA_STATE state;
    
    //! The metadata associated with the upgrade process stored in flash memory
    UPGRADE_METADATA metadata;
    
    UPLOAD_RUNTIME_DATA upload_data;
    
    UPGRADE_METADATA_UPLOAD_CALLBACK upload_callback;
    
    UPGRADE_METADATA_UPGRADE_CALLBACK upgrade_callback;
    
    //! Data used during uploads
    UPGRADE_RUNTIME_DATA upgrade_data;
} UPGRADE_DATA;

static UPGRADE_DATA upgrade_data =
{
    .state = UPGRADE_DATA_STATE_LOADING,
    .metadata.current_load_id = 0,
    .metadata.current_idx = 0,
    .metadata.upgrade_idx = 0,        
    .metadata.state = UPGRADE_METADATA_STATE_NONE,
    .upload_data.curr_write_pos = 0,
    .upload_callback = NULL,
    .upgrade_callback = NULL,
    .upgrade_data.state = UPGRADE_RUNTIME_STATE_INIT,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * A Callback used for flash_memory functionality
 * @param page_data
 */
void upgrade_metadata_callback(FLASH_MEMORY_CLIENT_REQUEST* request)
{
    //logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Flash Memory Callback client: %d ", request->client_id);
    
    switch(upgrade_data.state)
    {
        case UPGRADE_DATA_STATE_LOADING:
        {
            //logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE_METADATA: Loading Callback ");

            //upgrade_metadata_print_metadata();
            
//            // Check model 
//            if (upgrade_data.metadata.image_metadata[upgrade_data.metadata.current_idx].state != IMAGE_STATE_UNINITIALISED)     
//            {
//                logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE_METADATA: Upgrade info loaded - current %d.%d vs %d.%d ",
//                        upgrade_data.metadata.image_metadata[upgrade_data.metadata.current_idx].version_major,
//                        upgrade_data.metadata.image_metadata[upgrade_data.metadata.current_idx].version_minor,
//                        SOFTWARE_VERSION_MAJOR,
//                        SOFTWARE_VERSION_MINOR);
//            }
//            else
//            {
//                logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE_METADATA: Upgrade info NOT loaded");
//            }
            
            upgrade_data.state = UPGRADE_DATA_STATE_LOADED;
            break;
        }
        
        case UPGRADE_DATA_STATE_LOADED:
        {
            logger_log_info(LOGGER_CODE_UPGRADE, "UPGRADE_METADATA: Loaded Callback????? ");
            break;
        }
        
        case UPGRADE_DATA_STATE_UPLOADING:
        {
            // Update bytes written if we are writing the data (not the metadata)
            if (request->file_index > 0)
            {
                upgrade_data.upload_data.bytes_written += request->length;

                logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Bytes written %d at offset %d total %d (%d) --",
                        request->length,
                        request->offset,
                        upgrade_data.upload_data.bytes_written,
                        request->file_index);

                IMAGE_METADATA* image_data = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx];
                if (image_data->local_callback != NULL)
                {
                    image_data->local_callback(image_data->load_id, image_data->size, upgrade_data.upload_data.bytes_written, UPGRADE_METADATA_UPGRADE_STATUS_OK);
                }

                // Check for a general callback - used to send messages
                if (upgrade_data.upload_callback != NULL)
                {
                    upgrade_data.upload_callback(image_data->load_id, request);
                }
            }
            else
            {
                logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Metadata Bytes written %d --",
                        request->length);
            }

            break;
        }
        
        case UPGRADE_DATA_STATE_END_UPLOAD:
        {
            logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA UPGRADE_DATA_STATE_END_UPLOAD --");

            IMAGE_METADATA* image_data = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx];
            if (image_data->local_callback != NULL)
            {
                image_data->local_callback(image_data->load_id, image_data->size, upgrade_data.upload_data.bytes_written, UPGRADE_METADATA_UPGRADE_STATUS_OK);

                // Clear the callback
                image_data->local_callback = NULL;
            }
            
            upgrade_data.state = UPGRADE_DATA_STATE_LOADED;
            break;
        }
        
        case UPGRADE_DATA_STATE_UPGRADING:
        {
#ifdef UPGRADE_METADATA_INCLUDE_UPGRADE_FUNCTIONALITY

            IMAGE_METADATA* image_data = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx];
            //image_data->size; // The size of the upgrade file
                    
//            logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA read:%d len:%d total:%d size:%d ",
//                request->bytes_read_written, // Contains the actual number of bytes read
//                request->length, // Should equal UPGRADE_RUNTIME_DATA_LEN
//                upgrade_data.upgrade_data.offset + request->bytes_read_written,
//                image_data->size);

            // Write the data read from the upgrade data into program flash
            if ((request->bytes_read_written > 0) && 
                (upgrade_write_buffer(request->data, request->bytes_read_written) != UPGRADE_WRITE_ERROR))
            {
                // Invoke a callback here to enable UI updates (LED, VFD))
                // Need total size, written size so we can calculate percentage
                if (upgrade_data.upgrade_callback != NULL)
                {
                    UPGRADE_METADATA_UPGRADE_PROGRESS progress;
                    progress.load_id = image_data->load_id;
                    progress.state = UPGRADE_METADATA_UPGRADE_PROGRESS_STATE_WRITING;
                    progress.bytes_written = upgrade_data.upgrade_data.offset + request->bytes_read_written;
                    progress.size = image_data->size;
                    progress.version_major = image_data->version_major;
                    progress.version_minor = image_data->version_minor;

                    upgrade_data.upgrade_callback(image_data->load_id, &progress);

                    //logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE_METADATA Upgrading %d complete ", ((upgrade_data.upgrade_data.offset + request->bytes_read_written) *100) / image_data->size);
                }

                upgrade_data.upgrade_data.offset += request->bytes_read_written;

                if (upgrade_data.upgrade_data.offset < image_data->size)
                {
                    // More data to read
                    upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_LOAD_BLOCK;
                }
                else
                {
                    // All data read
                    upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_CLOSE_IMAGE;
                }
            }
            else
            {
                logger_log_errorf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Upgrade failed ");
                upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_ERROR;
            }
#endif
            break;
        }

    }
}

/**-----------------------------------------------------------------------------
 * Initiates reading the settings from the Flash memory
 * Note: If a setting has been updated it will be overwritten (open to review!!)
 */
static void upgrade_metadata_load()
{
    if (upgrade_data.state == UPGRADE_DATA_STATE_LOADING)
    {
        logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Loading ");

        // MAke a flash request 
        FLASH_MEMORY_CLIENT_REQUEST request;
        request.client_id = UPGRADE_METADATA_CLIENT_ID;
        request.file_index = 0;
        request.data = (void*)&upgrade_data.metadata;
        request.offset = 0;
        request.length = sizeof(UPGRADE_METADATA);

        flash_memory_read(&request);
    }
}

/**-----------------------------------------------------------------------------
 * Save the metadata
 */
static void upgrade_metadata_save_metadata()
{
    // Save the metadata
    FLASH_MEMORY_CLIENT_REQUEST request;
    request.client_id = UPGRADE_METADATA_CLIENT_ID;
    request.file_index = 0;
    request.data = (void*)&upgrade_data.metadata;
    request.offset = 0;
    request.length = sizeof(UPGRADE_METADATA);

    flash_memory_write(&request);
}

/**-----------------------------------------------------------------------------
 * Gets a pointer to the IMAGE_METADATA to be used for the upgrade
 * @return 
 */
static IMAGE_METADATA* upgrade_metadata_upgrade_info()
{
    IMAGE_METADATA* result = NULL;
    
    if ((upgrade_data.metadata.upgrade_idx > 0) && (upgrade_data.metadata.upgrade_idx <= UPGRADE_IMAGE_NUM_IMAGES))
    {
        result = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx - 1];
    }

    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Called after a successful upgrade. Sets the metadata up for a successful upgrade
 */
void upgrade_metadata_complete()
{
    upgrade_data.metadata.current_idx = upgrade_data.metadata.upgrade_idx;
    upgrade_data.metadata.image_metadata[upgrade_data.metadata.current_idx].state = IMAGE_STATE_LOADED;

    upgrade_data.metadata.state = UPGRADE_METADATA_STATE_NONE;
    upgrade_data.metadata.upgrade_idx = 0;
}

/**-----------------------------------------------------------------------------
 * Called after image data loaded into flash.  Update the upgrade_idx, sets state to loaded
 */
void upgrade_metadata_complete_load(size_t image_size)
{
    // Update the metadata status
    upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx].state = IMAGE_STATE_UPLOADED;
    upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx].size = image_size;

    upgrade_data.metadata.state = UPGRADE_METADATA_STATE_AVAILABLE_INSTALL;
}

/**-----------------------------------------------------------------------------
 * Get the current upgrade state
 * @return 
 */
UPGRADE_METADATA_STATE upgrade_metadata_get_upgrade_state()
{
    return upgrade_data.metadata.state;
}

/**-----------------------------------------------------------------------------
 * Initialise the internal data structures.  Assumes the pointer to the metadata
 * comes from data ready from flash memory
 * @param metadata
 */    
void upgrade_metadata_initialise(UPGRADE_METADATA_UPLOAD_CALLBACK upload_callback,
        UPGRADE_METADATA_UPGRADE_CALLBACK upgrade_callback)
{
    logger_log_infof(LOGGER_CODE_INITIALISATION, "UPGRADE METADATA - Initialise - %d images --", UPGRADE_IMAGE_NUM_IMAGES);
    
    upgrade_data.upload_callback = upload_callback;
    upgrade_data.upgrade_callback = upgrade_callback;
    
    int x;
    for (x = 0; x < UPGRADE_IMAGE_NUM_IMAGES; x++)
    {
        upgrade_data.metadata.image_metadata[x].size = 0;
        upgrade_data.metadata.image_metadata[x].state = IMAGE_STATE_UNINITIALISED;
    }
    
    flash_memory_add_client(FLASH_MEMORY_CLIENT_UPGRADE, UPGRADE_STORAGE_NAME, &upgrade_metadata_callback);
    
    upgrade_metadata_load();
}

/**-----------------------------------------------------------------------------
 * Test whether the meta data has been loaded
 * @return 
 */
bool upgrade_metadata_is_loaded()
{
    return upgrade_data.state != UPGRADE_DATA_STATE_LOADING;
}

/**-----------------------------------------------------------------------------
 * Returns true if an upgrade is required.
 * @return 
 */    
bool upgrade_metadata_needs_upgrade()
{
    bool result = false;
    
    IMAGE_METADATA* metadata = upgrade_metadata_upgrade_info();
    
    if ((metadata != NULL) && (metadata->state == IMAGE_STATE_LOADED))
    {
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs an upgrade using the most recent available upload data
 * @return 
 */
UPGRADE_METADATA_UPGRADE_STATUS upgrade_metadata_perform_upgrade()
{
#ifdef UPGRADE_METADATA_INCLUDE_UPGRADE_FUNCTIONALITY

    UPGRADE_METADATA_UPGRADE_STATUS result = UPGRADE_METADATA_UPGRADE_STATUS_IN_PROGRESS;

    //logger_log_info(LOGGER_CODE_INITIALISATION, "UPGRADE METADATA - Performing Upgrade ");
    switch (upgrade_data.upgrade_data.state)
    {
        case UPGRADE_RUNTIME_STATE_INIT:
        {
            // Check the image is in an uploaded state
            bool proceed = true;
            if ((upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx].state == IMAGE_STATE_UPLOADED) ||
                (upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx].state == IMAGE_STATE_LOADED))
            {
                logger_log_infof(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Upgrading from image %d ", upgrade_data.metadata.upgrade_idx);
                
                // Check that the upgrade has data - normally it will as we do check the checksum, but during testing
                // we are not saving the data
                if (upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx].size == 0)
                {
                    logger_log_errorf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Cannot upgrade from empty image %d ", upgrade_data.metadata.upgrade_idx);
                    proceed = false;
                }
            }
            else if (upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx].state == IMAGE_STATE_ERROR)
            {
                logger_log_errorf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Cannot upgrade from corrupt image %d ", upgrade_data.metadata.upgrade_idx);
                proceed = false;
            }
            else
            {
                logger_log_warningf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Upgrading from image %d in invalid state %s ", 
                        upgrade_data.metadata.upgrade_idx,
                        IMAGE_STATE_STRINGS[upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx].state]);
                proceed = false;
            }
            
            if (proceed)
            {
                if (upgrade_can_proceed())
                {
                    upgrade_data.state = UPGRADE_DATA_STATE_UPGRADING;
                    upgrade_data.metadata.state = UPGRADE_METADATA_STATE_IN_PROGRESS;
                    upgrade_data.upgrade_data.offset = 0;
                    upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_ERASE_FLASH;
                }
                else
                {
                    logger_log_errorf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - NVM not ready for upgrade "); 
                    upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_ERROR;
                }
            }
            else
            {
                result = UPGRADE_METADATA_UPGRADE_STATUS_ERROR;
            }
            
            break;
        }
        
        case UPGRADE_RUNTIME_STATE_ERASE_FLASH:
        {
            logger_log_infof(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Erasing Flash ");
            
            if (upgrade_data.upgrade_callback != NULL)
            {
                IMAGE_METADATA* image_data = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx];
                
                UPGRADE_METADATA_UPGRADE_PROGRESS progress;
                progress.load_id = image_data->load_id;
                progress.state = UPGRADE_METADATA_UPGRADE_PROGRESS_STATE_ERASING;
                progress.bytes_written = 0;
                progress.size = image_data->size;
                progress.version_major = image_data->version_major;
                progress.version_minor = image_data->version_minor;

                upgrade_data.upgrade_callback(image_data->load_id, &progress);
            }
                
            upgrade_initialise();
            
            upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_OPEN_IMAGE;
            break;
        }

        case UPGRADE_RUNTIME_STATE_OPEN_IMAGE:
        {
            logger_log_infof(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Opening Image %d ", upgrade_data.metadata.upgrade_idx);
            flash_memory_open(UPGRADE_METADATA_CLIENT_ID, upgrade_data.metadata.upgrade_idx + 1);
            
            upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_LOAD_BLOCK;
            break;
        }
        
        case UPGRADE_RUNTIME_STATE_LOAD_BLOCK:
        {
            //logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Loading at offset %d ", upgrade_data.upgrade_data.offset);

            // No read/write in progress so read the next block
            // MAke a flash request 
            FLASH_MEMORY_CLIENT_REQUEST request;
            request.client_id = UPGRADE_METADATA_CLIENT_ID;
            request.file_index = upgrade_data.metadata.upgrade_idx + 1;
            request.data = (void*)&upgrade_data.upgrade_data.data;
            request.offset = upgrade_data.upgrade_data.offset;
            request.length = UPGRADE_RUNTIME_DATA_LEN;

            flash_memory_read(&request);
            
            upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_LOADING_BLOCK;
            break;
            
            break;
        }

        case UPGRADE_RUNTIME_STATE_LOADING_BLOCK:
        {
            // Do nothing here - we are in the process of upgrading a single block
        }
        
        case UPGRADE_RUNTIME_STATE_CLOSE_IMAGE:
        {
            logger_log_infof(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Closing Image %d ", upgrade_data.metadata.upgrade_idx);
            flash_memory_close(UPGRADE_METADATA_CLIENT_ID, upgrade_data.metadata.upgrade_idx + 1);
            
            upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_COMPLETE;
            break;
        }       
       
        case UPGRADE_RUNTIME_STATE_COMPLETE:
        {
            // Set Image state to IMAGE_STATE_LOADED
            upgrade_metadata_complete();
                        
            upgrade_data.state = UPGRADE_DATA_STATE_LOADED;
            
            upgrade_metadata_save_metadata();
            
            result = UPGRADE_METADATA_UPGRADE_STATUS_OK;
            break;
        }
        
        case UPGRADE_RUNTIME_STATE_ERROR:
        {
            //??upgrade_data.metadata.state = UPGRADE_METADATA_STATE_NONE;
            upgrade_data.state = UPGRADE_DATA_STATE_LOADED;
            
            IMAGE_METADATA* image_data = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx];
            image_data->state = IMAGE_STATE_ERROR;
            
            upgrade_metadata_save_metadata();
            
            result = UPGRADE_METADATA_UPGRADE_STATUS_ERROR;
            break;
        }
        
        default:
            break;
    }
    
    return result;

#endif
}

/**-----------------------------------------------------------------------------
 * Called after a failed upgrade. Sets up the metadata to attempt a restore of the current version
 */
bool upgrade_metadata_prepare_rollback()
{
   bool result = true;
    
    if (upgrade_data.metadata.state != UPGRADE_METADATA_STATE_IN_ROLLBACK)
    {
        logger_log_infof(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Preparing Rollback to image %d ", upgrade_data.metadata.current_idx);

        upgrade_data.metadata.upgrade_idx = upgrade_data.metadata.current_idx;
        upgrade_data.upgrade_data.state = UPGRADE_RUNTIME_STATE_INIT;

        upgrade_data.metadata.state = UPGRADE_METADATA_STATE_IN_ROLLBACK;
    }
    else
    {
        result = false;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Called before loading image data into flash.  Update the upgrade_idx, sets state to loading
 * @param major_version
 * @param minor_version
 * @param size
 * @param checksum - Pointer to a 16byte array holding the checksum
 * @param override
 * @return 
 */
UPGRADE_METADATA_LOAD_ID upgrade_metadata_prepare_upload(uint8_t major_version, uint8_t minor_version, size_t size, uint8_t* checksum, bool override)
{
   logger_log_info(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Preparing Load --");

   uint8_t upgrade_idx = 0;
    
    // First check whether the current image has been initialised
    if (upgrade_data.metadata.image_metadata[upgrade_data.metadata.current_idx].state != IMAGE_STATE_LOADED)
    {
        upgrade_idx = upgrade_data.metadata.current_idx;
        
        // Debugging
        if (upgrade_data.metadata.image_metadata[upgrade_data.metadata.current_idx].state == IMAGE_STATE_UNINITIALISED)
        {
            logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Loading into uninitialised image %d --", upgrade_idx);
        }
        else if (upgrade_data.metadata.image_metadata[upgrade_data.metadata.current_idx].state == IMAGE_STATE_UPLOADED)
        {
            logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Overwriting uploaded image %d --", upgrade_idx);
        }
        else if (upgrade_data.metadata.image_metadata[upgrade_data.metadata.current_idx].state == IMAGE_STATE_LOADING)
        {
            logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Overwriting loading image %d !!!!! --", upgrade_idx);
        }
    }
    else
    {
        // set upgrade_idx = current_idx + 1
        upgrade_idx = upgrade_data.metadata.current_idx;
        upgrade_idx++;
        if (upgrade_idx >= UPGRADE_IMAGE_NUM_IMAGES)
        {
            upgrade_idx = 0;
        }
        
        logger_log_infof(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Loading into image %d --", upgrade_idx);
    }

    // Check state of upgrade image data
    bool proceed = true;
    if ((upgrade_data.metadata.image_metadata[upgrade_idx].state == IMAGE_STATE_LOADED) && !override)
    {
        uint8_t current_major = upgrade_data.metadata.image_metadata[upgrade_idx].version_major;
        uint8_t current_minor = upgrade_data.metadata.image_metadata[upgrade_idx].version_minor;
        
        // Check the version
        if (current_major == major_version)
        {
            if (current_minor > minor_version)
            {
                proceed = false;
            }
            else if (current_minor == minor_version)
            {
                if (memcmp(upgrade_data.metadata.image_metadata[upgrade_idx].checksum, checksum, MD5_DIGEST_SIZE) != 0)
                {
                    logger_log_warningf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA: Overwriting current image - checksums differ --");
                }
                else
                {
                    proceed = false;
                }
            }
        }
        else if (current_major > major_version)
        {
            proceed = false;
        }
        
        if (!proceed)
        {
            logger_log_errorf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA: Attempt to upload a duplicate or older version %d.%d vs %d.%d --",
                    major_version, minor_version, current_major, current_minor);
        }
    }
    
    if (proceed)
    {
        logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - Loading into image %d --", upgrade_idx);
        
        // Start the upload
        // Update the metadata status
        upgrade_data.metadata.upgrade_idx = upgrade_idx;
        upgrade_data.metadata.current_load_id++;
        
        upgrade_data.upload_data.bytes_written = 0;
        
        // Check for overrun
        if (upgrade_data.metadata.current_load_id == 0)
        {
            upgrade_data.metadata.current_load_id = 1;
        }

        IMAGE_METADATA* data = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx];
        
        if (memcmp(data->checksum, checksum, MD5_DIGEST_SIZE) == 0)
        {
            logger_log_warningf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA: Overwriting image with same checksum --");
        }

        data->load_id = upgrade_data.metadata.current_load_id;
        data->state = IMAGE_STATE_UPLOADING;
        data->size = size;
        data->version_major = major_version;
        data->version_minor = minor_version;
        memcpy(data->checksum, checksum, MD5_DIGEST_SIZE);
        
        CRYPT_MD5_Initialize(&upgrade_data.upload_data.md5_context);

        logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA - SIZE %d --", sizeof(UPGRADE_METADATA));

        flash_memory_open(UPGRADE_METADATA_CLIENT_ID, upgrade_data.metadata.upgrade_idx + 1);
        
        // Save the metadata
        upgrade_metadata_save_metadata();
    }
    else
    {
        upgrade_data.metadata.current_load_id = 0;
    }
    
    return upgrade_data.metadata.current_load_id;
}

/**-----------------------------------------------------------------------------
 * Prints the metadata as a series of logger_log_info messages
 */
void upgrade_metadata_print_metadata()
{
    logger_log_infof(LOGGER_CODE_UPGRADE, "------------------------------------------------------------");

    int x;
    for (x = UPGRADE_IMAGE_NUM_IMAGES - 1; x >= 0; x--)
    {
        IMAGE_METADATA* image = &upgrade_data.metadata.image_metadata[x];

        logger_log_debugf(LOGGER_CODE_UPGRADE, "-    checksum: %02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x --", 
            image->checksum[0],
            image->checksum[1],
            image->checksum[2],
            image->checksum[3],
            image->checksum[4],
            image->checksum[5],
            image->checksum[6],
            image->checksum[7],
            image->checksum[8],
            image->checksum[9],
            image->checksum[10],
            image->checksum[11],
            image->checksum[12],
            image->checksum[13],
            image->checksum[14],
            image->checksum[15]
            );
        logger_log_infof(LOGGER_CODE_UPGRADE, "-    load index: %d bytes --", 
            image->load_id);
        logger_log_infof(LOGGER_CODE_UPGRADE, "-    size: %d bytes --", 
            image->size);
        logger_log_infof(LOGGER_CODE_UPGRADE, "-    state: %s --", 
            IMAGE_STATE_STRINGS[min(image->state, IMAGE_STATE_ERROR)]);
        logger_log_infof(LOGGER_CODE_UPGRADE, "-    version: %d.%d --", 
            image->version_major, image->version_minor);
        
        if (x == upgrade_data.metadata.current_idx)
        {
            logger_log_infof(LOGGER_CODE_UPGRADE, "-  IMAGE %d (Current) ------------ ", x);
        }
        else
        {
            logger_log_infof(LOGGER_CODE_UPGRADE, "-  IMAGE %d ---------------------- ", x);
        }
    }
    
    logger_log_infof(LOGGER_CODE_UPGRADE, "-  state: %s --", 
            UPGRADE_METADATA_STATE_STRINGS[upgrade_data.metadata.state]);
    logger_log_infof(LOGGER_CODE_UPGRADE, "-  current load id: %d --", 
            upgrade_data.metadata.current_load_id);
    logger_log_infof(LOGGER_CODE_UPGRADE, "-  upgrade image index: %d --", 
            upgrade_data.metadata.upgrade_idx);
    logger_log_infof(LOGGER_CODE_UPGRADE, "-  loaded image index: %d --", 
            upgrade_data.metadata.current_idx);
    
    logger_log_infof(LOGGER_CODE_UPGRADE, "UPGRADE METADATA -------------- ");
    logger_log_infof(LOGGER_CODE_UPGRADE, "------------------------------------------------------------ ");
}

/**-----------------------------------------------------------------------------
 * Uploads image data into the storage.
 * It is assumed that successive calls to this will append data onto the end of the file
 * @param load_id
 * @param data
 * @param data_length
 */
UPGRADE_METADATA_UPLOAD_STATUS upgrade_metadata_upload_data(UPGRADE_METADATA_LOAD_ID load_id, 
        uint8_t* data, 
        size_t data_length,
        UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback)
{
    UPGRADE_METADATA_UPLOAD_STATUS result = UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
    
    IMAGE_METADATA* image_data = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx];
    image_data->local_callback = local_callback;

//logger_log_debugf(LOGGER_CODE_UPGRADE, "DEVICE_UPGRADE_STATUS_CHECKSUM_ERROR D %d %d --", image_data->load_id, load_id);
   
    if (image_data->load_id == load_id)
    {
        logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Upload Data %d --", data_length);
    
        // Update the checksum
        CRYPT_MD5_DataAdd(&upgrade_data.upload_data.md5_context, (char*)data, data_length);

        // Set the offset and state
        uint32_t offset = upgrade_data.upload_data.bytes_written;
        
        upgrade_data.state = UPGRADE_DATA_STATE_UPLOADING;
        
        // Copy the data into the write_buffer
        // If we look like were going to go past the end of the buffer then reset to the beginning
        // This means that the buffer must be sufficiently large so that there
        // will be space at the beginning when we get to the end.
        if (upgrade_data.upload_data.curr_write_pos + data_length > UPLOAD_BUFFER_SIZE)
        {
            upgrade_data.upload_data.curr_write_pos = 0;
        }
        
        memcpy(&upgrade_data.upload_data.write_buffer[upgrade_data.upload_data.curr_write_pos], data, data_length);

        // Get the image file and append data to it......
        // Save the image data
        FLASH_MEMORY_CLIENT_REQUEST request;
        request.client_id = UPGRADE_METADATA_CLIENT_ID;
        request.file_index = upgrade_data.metadata.upgrade_idx + 1;
        //request.data = (void*)data;
        request.data = (void*)&upgrade_data.upload_data.write_buffer[upgrade_data.upload_data.curr_write_pos];
        request.offset = offset;
        request.length = data_length;

        logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Upload Data %d %d %d --", upgrade_data.upload_data.curr_write_pos, offset, data_length);

        // Update circular buffer write position
        upgrade_data.upload_data.curr_write_pos += data_length;

        flash_memory_write(&request);
    }
    else
    {
        logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Load Id Mismatch %d %d --", image_data->load_id, load_id);
        result = UPGRADE_METADATA_UPLOAD_STATUS_LOAD_ID_MISMATCH;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Completes the Uploads image data into the storage
 * @param load_id
 * @param data
 * @param data_length
 */
UPGRADE_METADATA_UPLOAD_STATUS upgrade_metadata_upload_end(UPGRADE_METADATA_LOAD_ID load_id,
        UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback, UPGRADE_METADATA_UPGRADE_STATUS status)
{
    //TODO Handle UPGRADE_METADATA_UPGRADE_STATUS_ERROR state
    UPGRADE_METADATA_UPLOAD_STATUS result = UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
    
    upgrade_data.state = UPGRADE_DATA_STATE_END_UPLOAD;
   
    IMAGE_METADATA* image_data = &upgrade_data.metadata.image_metadata[upgrade_data.metadata.upgrade_idx];
    image_data->local_callback = local_callback;

    // Complete and compare checksums and data_length?
    uint8_t checksum[MD5_DIGEST_SIZE];
    CRYPT_MD5_Finalize(&upgrade_data.upload_data.md5_context, checksum);

    flash_memory_set_size(UPGRADE_METADATA_CLIENT_ID, 
            upgrade_data.metadata.upgrade_idx + 1,
            upgrade_data.upload_data.bytes_written);

    flash_memory_close(UPGRADE_METADATA_CLIENT_ID, upgrade_data.metadata.upgrade_idx + 1);

    if (memcmp(image_data->checksum, checksum, MD5_DIGEST_SIZE) == 0)
    {
        upgrade_metadata_complete_load(upgrade_data.upload_data.bytes_written);
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA: Checksum mismatch - image marked as ERROR ");
        image_data->state = IMAGE_STATE_ERROR;
        image_data->size = upgrade_data.upload_data.bytes_written;
        upgrade_data.metadata.state = UPGRADE_METADATA_STATE_UPLOAD_ERRORS;
        result = UPGRADE_METADATA_UPLOAD_STATUS_CHECKSUM_MISMATCH;
    }
    
    // Save the metadata
    upgrade_metadata_save_metadata();

    // Output upload statistics and checksums
    logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE METADATA Bytes written %d file size % d ",
            upgrade_data.upload_data.bytes_written,
            flash_memory_size(UPGRADE_METADATA_CLIENT_ID, upgrade_data.metadata.upgrade_idx + 1));

    logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE Checksum - Orig %02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x --", 
        image_data->checksum[0],
        image_data->checksum[1],
        image_data->checksum[2],
        image_data->checksum[3],
        image_data->checksum[4],
        image_data->checksum[5],
        image_data->checksum[6],
        image_data->checksum[7],
        image_data->checksum[8],
        image_data->checksum[9],
        image_data->checksum[10],
        image_data->checksum[11],
        image_data->checksum[12],
        image_data->checksum[13],
        image_data->checksum[14],
        image_data->checksum[15]
        );
 
    logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE Checksum - Tran %02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x --", 
        checksum[0],
        checksum[1],
        checksum[2],
        checksum[3],
        checksum[4],
        checksum[5],
        checksum[6],
        checksum[7],
        checksum[8],
        checksum[9],
        checksum[10],
        checksum[11],
        checksum[12],
        checksum[13],
        checksum[14],
        checksum[15]
        );

//        logger_process();
//        if (local_callback != NULL)
//        {
//     logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE Executing END local callback A --");
//     logger_process();
//          local_callback(image_data->load_id, image_data->size, upgrade_data.upload_data.bytes_written, UPGRADE_METADATA_UPGRADE_STATUS_OK);
//     logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE Executing END local callback B --");
//     logger_process();
//
//            // Clear the callback
//            image_data->local_callback = NULL;
//        }
    
    logger_process();
    return result;
}

/* *****************************************************************************
 End of File
 */
