/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    upnp.c

  @Summary
    Implements UPnP discover functionality.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/upnp.h"
#include "utilities/logger.h"
#include "utilities/settings.h"
#include "app_config.h"
#include "utilities/tcpip.h"
#include "utilities/timers.h"
#include "utilities/udp.h"
#include "utilities/uuid.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define UPNP_DISCOVERY_FILE  "upnpdiscovery.xml"
#define UPNP_MAX_AGE_SECS   TIMER_TIMEOUT_UPNP // Shoud be a minumim of 1800 
#define UPNP_DISCOVERY_URI  "239.255.255.250:1900"    
#define UPNP_SSDP_DISCOVER  "ssdp:discover"   

//#define UPNP_ARCAM_AMPLIFIER_URN "urn:arcam-co-uk:device:Amplifier:1"

typedef struct
{
    //! Indicates whether the initial notify message has been sent
    bool notify_message_sent;
    
    //! Keeps a record of how many notify message have been sent in the current batch
    uint8_t batch_index;
    
    //! Holds the device description used for the service description
    char device_description[UPNP_DESCRIPTION_MAX_LENGTH];
    
    //! Holds the device URN used for the service description
    char device_urn[UPNP_URN_MAX_LENGTH];
    
} UPNP_DATA;

static UPNP_DATA upnp_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declaration
static void upnp_send_notify_message();
static void upnp_send_search_response_message(struct sockaddr_in* address, int* address_len, char* st_buffer);

/**
 * Callback for udp messages 
 */
static bool upnp_callback(int port, struct sockaddr_in* address, int* address_len, char* message, int len)
{
    bool result = false;
    bool send_response = true;

    char st_buffer[100];
    
//    logger_log_debugf(0, "UPNP: -----------------------------");
//  
//    memcpy(st_buffer, message, 19);
//    logger_log_debugf(0, "UPNP: %s --", st_buffer);
    
    // Were looking for a M-SEARCH message
    // If found check its from HOST 239.255.255.250:1900
    // and that MAN = "ssdp:discover"
    // and that ST includes rootDevice or the UUID of this
    if (strncmp(message, "M-SEARCH * HTTP/1.1", 19) == 0)
    {
        char buffer[UDP_MESSAGE_MAX_SIZE];
        memcpy(buffer, message, UDP_MESSAGE_MAX_SIZE);
        
        char* ptr_start = buffer;
        char* ptr_end = strstr(ptr_start, "\r\n");
        int ptr_len = 0;
        int count = 0;
        
        while (ptr_end != NULL)
        {
            ptr_len = ptr_end - ptr_start;
            if (ptr_len == 0)
            {
                // We've reached the final line
                break;
            }
            
            // Check we dont process over the end of the buffer if it is malformed
            count += ptr_len + 2;
            if (count >= UDP_MESSAGE_MAX_SIZE)
            {
                break;
            }
        
            // null terminate the string before the current pos
            *ptr_end = 0;
            
            // Determine the line type
            // Check the line is valid
            // Bomb out if not
            
            //logger_log_debugf(0, "%s --", ptr_start);
            
            if (strncmp(ptr_start, "HOST:", 5) == 0)
            {
                // Check URL - looking for failure
                if (strstr(ptr_start, UPNP_DISCOVERY_URI) == NULL)
                {
                    send_response = false;

                    break;
                }
            }
            else if (strncmp(ptr_start, "MAN:", 4) == 0)
            {
                if (strstr(ptr_start, """ssdp:discover""") == NULL)
                {
                    send_response = false;
                    //logger_log_debugf(0, "UPNP: NO MATCH");
                    break;
                }
            }
            else if (strncmp(ptr_start, "ST:", 3) == 0)
            {
                if (logger_get_level() >= LOGGER_TYPE_DEBUG)
                {
                    char match[80];
                    memset(match, 0, 80);
                    memcpy(match, ptr_start + 3, ptr_end - ptr_start);

                    logger_log_debugf(235, "UPNP: request %s --", match);
                    logger_log_debugf(235, "UPNP: actual %s --", upnp_data.device_urn);
                }
                
                if ((strstr(ptr_start, "ssdp:all") == NULL) &&
                    (strstr(ptr_start, "upnp:rootdevice") == NULL) &&
                    (strstr(ptr_start, upnp_data.device_urn) == NULL))    
                {
                    UUID uuid;
                    char uuid_buffer[38];
                    settings_get(SETTING_MACHINE_UUID, &uuid);
                    uuid_to_string(&uuid, uuid_buffer);
                    
                    if (strstr(ptr_start, uuid_buffer) == NULL)   
                    {
                        send_response = false;
                        //logger_log_debugf(0, "UPNP: NO MATCH");
                        break;
                    }
                }
               
                // Copy the whole ST: line
                memcpy(st_buffer, ptr_start, ptr_end - ptr_start + 1);
            }
            
            // Move onto the next line
            ptr_start = ptr_end + 2;
            
            ptr_end = strstr(ptr_start, "\r\n");
        }
    }
    else
    {
        send_response = false;
    }
    
    if (send_response)
    {
        upnp_send_search_response_message(address, address_len, st_buffer);   
    }
    else
    {
        //logger_log_debugf(0, "UPNP: failed match");
    }
    
    // When using the TP-Link Wi-Fi Router the initial notify was being missed 
    // as it appears the network was not ready.  Given we get loads of UDP messages coming in
    // lets send another notify the first time we get one as at least we know the 
    // network is there
    if (!upnp_data.notify_message_sent)
    {
        logger_log_debugf(235, "UPNP: Sending notify message on first message received");
        
        timers_expire(TIMER_INDEX_UPNP);
   
        upnp_data.notify_message_sent = true;
    }
    
    return result;
}

/**
 * 
 */
static void upnp_send_notify_message()
{
    // Get the IP Address
    char ip_address[16]; // Include trailing 0
    tcpip_get_ip_address_string(ip_address, 16);
    
    // Get the UUID
    UUID uuid;
    char uuid_buffer[38];
    settings_get(SETTING_MACHINE_UUID, &uuid);
    uuid_to_string(&uuid, uuid_buffer);
    
    uint16_t counter;
    settings_get(SETTING_UPNP_COUNTER, &counter);
    
    char config_buffer[10];
    upnp_get_configuration_number_string(config_buffer);  
    
    char buffer[UDP_MESSAGE_MAX_SIZE];
    const char cr = '\r';
    const char lf = '\n';
    
    const char notify_message[UDP_MESSAGE_MAX_SIZE] = "NOTIFY * HTTP/1.1%c%c\
HOST:%s%c%c\
CACHE-CONTROL:max-age = %d%c%c\
LOCATION:http://%s/%s%c%c\
NT:upnp:rootdevice%c%c\
NTS:ssdp:alive%c%c\
SERVER:microchip/pic32 UPnP/1.1 %s/%d.%d%c%c\
USN:uuid:%s::upnp:rootdevice%c%c\
BOOTID.UPNP.ORG:%d%c%c\
CONFIG.UPNP.ORG:%s%c%c%c%c";
    
    sprintf(buffer, notify_message, 
            cr, lf, 
            UPNP_DISCOVERY_URI, cr, lf,
            UPNP_MAX_AGE_SECS, cr, lf,
            ip_address, UPNP_DISCOVERY_FILE, cr, lf,
            cr, lf,
            cr, lf,
            SOFTWARE_MODEL_STRING, SOFTWARE_VERSION_MAJOR, SOFTWARE_VERSION_MINOR, cr, lf,
            uuid_buffer, cr, lf,
            counter, cr, lf,
            config_buffer,
            cr, lf,
            cr, lf);
    
    // Send the buffer via UDP
    IP_ADDR address;
    TCPIP_Helper_StringToIPAddress("239.255.255.250", &address);
    //TCPIP_Helper_StringToIPAddress("192.168.0.1", &address);
    
    udp_sendto(&address, 1900, buffer, strlen(buffer));
}

/**
 * 
 */
static void upnp_send_search_response_message(struct sockaddr_in* address, int* address_len, char* st_buffer)
{
    // Get the IP Address
    char ip_address[16]; // Include trailing 0
    tcpip_get_ip_address_string(ip_address, 16);
    
    // Get the UUID
    UUID uuid;
    char uuid_buffer[38];
    settings_get(SETTING_MACHINE_UUID, &uuid);
    uuid_to_string(&uuid, uuid_buffer);
    
    uint16_t counter;
    settings_get(SETTING_UPNP_COUNTER, &counter);
     
    char config_buffer[10];
    upnp_get_configuration_number_string(config_buffer);
   
    char buffer[UDP_MESSAGE_MAX_SIZE];
    const char cr = '\r';
    const char lf = '\n';
    
    const char response_message[UDP_MESSAGE_MAX_SIZE] = "HTTP/1.1 200 OK%c%c\
CACHE-CONTROL:max-age = %d%c%c\
EXT:%c%c\
LOCATION:http://%s/%s%c%c\
SERVER:microchip/pic32 UPnP/1.1 %s/%d.%d%c%c\
%s%c%c\
USN:uuid:%s::upnp:rootdevice%c%c\
BOOTID.UPNP.ORG:%d%c%c\
CONFIG.UPNP.ORG:%s%c%c%c%c";
    
    sprintf(buffer, response_message, 
            cr, lf, 
            UPNP_MAX_AGE_SECS, cr, lf,
            cr, lf, 
            ip_address, UPNP_DISCOVERY_FILE, cr, lf,
            SOFTWARE_MODEL_STRING, SOFTWARE_VERSION_MAJOR, SOFTWARE_VERSION_MINOR, cr, lf,
            st_buffer, cr, lf, 
            uuid_buffer, cr, lf,
            counter, cr, lf,
            config_buffer,
            cr, lf,
            cr, lf);

    // Send the buffer via UDP
    IP_ADDR ip4_address;
    ip4_address.Val = address->sin_addr.s_addr;
   
    udp_sendto(&ip4_address, address->sin_port, buffer, strlen(buffer));
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Generates a UPnP configuration number from the software version
 * @param buffer
 */
void upnp_get_configuration_number_string(char* buffer)
{
    sprintf(buffer, "%d", (SOFTWARE_VERSION_MAJOR * 256) + SOFTWARE_VERSION_MINOR);
}

/**
 * Gets the device URN specified in upnp_initialise
 * @return 
 */
void upnp_get_device_description_string(char* buffer, size_t len)
{
    memcpy(buffer, upnp_data.device_description, min(len, UPNP_DESCRIPTION_MAX_LENGTH));
}

/**
 * Gets the device URN specified in upnp_initialise
 * @return 
 */
void upnp_get_device_urn_string(char* buffer, size_t len)
{
    memcpy(buffer, upnp_data.device_urn, min(len, UPNP_URN_MAX_LENGTH));
}
    
/*
 * Initialise UPnP functionality - must be called after TCPIP/UDP initialisation
 */
void upnp_initialise()
{
    // Initialise variables
    upnp_data.notify_message_sent = false;
    upnp_data.batch_index = 0;
    
    // Copy Device URN
    memset(upnp_data.device_urn, 0, UPNP_URN_MAX_LENGTH);
    sprintf(upnp_data.device_urn, "urn:arcam-co-uk:device:%s:%d", ARCAM_AMPLIFIER_DESCRIPTION, SOFTWARE_VERSION_MAJOR);
    if (strlen(upnp_data.device_urn) >= UPNP_URN_MAX_LENGTH)
    {
        logger_log_warningf(LOGGER_CODE_UPNP_URN_EXCEEDS_MAX_LENGTH, "UPNP urn exceeds max length %d %d ", strlen(upnp_data.device_urn), UPNP_URN_MAX_LENGTH);
    }
    
    // Copy Device description
    if (strlen(ARCAM_AMPLIFIER_DESCRIPTION) >= UPNP_DESCRIPTION_MAX_LENGTH)
    {
        logger_log_warningf(LOGGER_CODE_UPNP_DESCRIPTION_EXCEEDS_MAX_LENGTH, "UPNP description exceeds max length %d %d ", strlen(ARCAM_AMPLIFIER_DESCRIPTION), UPNP_DESCRIPTION_MAX_LENGTH);
    }
    memset(upnp_data.device_description, 0, UPNP_DESCRIPTION_MAX_LENGTH);
    memcpy(upnp_data.device_description, ARCAM_AMPLIFIER_DESCRIPTION, min(strlen(ARCAM_AMPLIFIER_DESCRIPTION), UPNP_DESCRIPTION_MAX_LENGTH - 1));

    // Increment the counter
    uint16_t counter;
    settings_get(SETTING_UPNP_COUNTER, &counter);
    counter++;
    settings_set(SETTING_UPNP_COUNTER, &counter);

    logger_log_debugf(234, "UPNP: urn=%s desc=%s counter=%d ", upnp_data.device_urn, upnp_data.device_description, counter);
    
    // Register a callback with UDP
    udp_add_callback(upnp_callback);
}

/**
 * Processes UPnP actions
 */
void upnp_process()
{
    if (tcpip_is_connected() && udp_is_connected())
    {
        if (timers_has_expired(TIMER_INDEX_UPNP))
        {
            logger_log_debugf(0, "UPNP: sending notify message");
            
            upnp_send_notify_message();
            
            upnp_data.batch_index++;
            if (upnp_data.batch_index < 3)
            {
                // 200 ms gap between batch mesages
                timers_update_timeout(TIMER_INDEX_UPNP, 200, TIMERS_TIMEOUT_RESET_IMMEDIATE);
            }
            else
            {
                upnp_data.batch_index = 0;
                
                timers_update_timeout(TIMER_INDEX_UPNP, TIMER_UPNP_NOTIFY_TIMEOUT_MS, TIMERS_TIMEOUT_RESET_IMMEDIATE);
           }
           //timers_reset(TIMER_INDEX_UPNP);
        }
    }
    // Check if udp is ready - register a callback then if so broadcast discovery
    // Check if we need to re-broadcast discovery
}

/* *****************************************************************************
 End of File
 */
