/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    uuid.c

  @Summary
 Manages UUID generation (using Version 5 UUIDs)

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/uuid.h"
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "utilities/logger.h"
#include "crypto/crypto.h"
#include "crypto/src/sha.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define MAX_UUID_NAMESPACE  37
typedef struct
{
    bool initialised;
    
    char namespace_buffer[MAX_UUID_NAMESPACE];
    
    //! An empty UUID used for comparison
    UUID empty_uuid;
    
} UUID_DATA;

static UUID_DATA uuid_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Sets the supplied uuid to 00000000-0000-0000-0000-000000000000
 * @param uuid
 */
void uuid_empty(UUID* uuid)
{
    memset(uuid, 0, sizeof(UUID));
}
    
/**
 * Can be used to test whetger a UUID is empty (00000000-0000-0000-0000-000000000000)
 * @param uuid
 * @return - true is empty , false if not
 */
bool uuid_is_empty(UUID* uuid)
{
    return memcmp(uuid, &uuid_data.empty_uuid, sizeof(UUID)) == 0;
}

/*
 * Initialises the UUID module with a given namespace
 */
void uuid_initialise(const char* namespace_string, size_t ns_len)
{
    size_t len = min(ns_len, MAX_UUID_NAMESPACE - 1);
    memcpy(uuid_data.namespace_buffer, namespace_string, len);
    uuid_data.namespace_buffer[len] = 0;
    
    uuid_data.initialised = true;
}
    
/*
 * Returns true is uuid_initialise has been called
 */
bool uuid_is_initialised()
{
    return uuid_data.initialised;
}

/**
 * Generates a new UUID and places it in the supplied UUID variable
 * @param uuid
 */
void uuid_generate(UUID* uuid, char* name, size_t name_len)
{
    //logger_log_debugf(0, "UUID: %s %s --", uuid_data.namespace_buffer, name);

    uuid_empty(&uuid_data.empty_uuid);
    
    CRYPT_SHA_CTX sha;
    
    const int buffer_len = 1024;
    uint8_t buffer[buffer_len];
    uint8_t shasum[SHA_DIGEST_SIZE];

    memset(buffer, 0, buffer_len);
    strcpy(buffer, uuid_data.namespace_buffer);
    strcat(buffer, name);
    
    CRYPT_SHA_Initialize(&sha);
    CRYPT_SHA_DataSizeSet(&sha, sizeof(buffer));
    CRYPT_SHA_DataAdd(&sha, buffer, sizeof(buffer));
    CRYPT_SHA_Finalize(&sha, shasum);
    
    memcpy(uuid, shasum, 16);
 
//    logger_log_debugf(0, "UUID: %8.8x-%4.4x-%4.4x-%4.4x-%4.4x%8.8x --", 
//            uuid->data1,
//            uuid->data2,
//            uuid->data3,
//            uuid->data4,
//            uuid->data5,
//            uuid->data6);
}

/**
 * Generates a string version of the UUID in the format XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
 * The buffer must be at least 36 chars long
 */
void uuid_to_string(UUID* uuid, char* buffer)
{
    sprintf(buffer, "%8.8x-%4.4x-%4.4x-%4.4x-%4.4x%8.8x", 
          uuid->data1,
          uuid->data2,
          uuid->data3,
          uuid->data4,
          uuid->data5,
          uuid->data6);  
}
    
/* *****************************************************************************
 End of File
 */
