/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    volume.c

  @Summary
 Manages volume control

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/volume.h"
#include "utilities/mutes.h"
#include "utilities/logger.h"
#include "app_config.h"
#include "utilities/timers.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */
//
///**
// * Volume mapping 
// * 
// * We needed to cut 4dB in software from the output to match with other amplifiers
// * The 5dB steps at the bottom end have therefore been cut to 4dB (step increment of 8 x 0.5dB)
// * 
// * Copied from A29 audio_settings.c
// */
//static uint8_t volume_mapping[] =
//{
//    1,	// 0
//    26,	// 1
//    34,	// 2
//    42,	// 3
//    50,	// 4
//    58,	// 5
//    66,	// 6
//    74,	// 7
//    82,	// 8
//    90,	// 9
//    98,	// 10
//    102,	// 11
//    106,	// 12
//    110,	// 13
//    114,	// 14
//    118,	// 15
//    122,	// 16
//    126,	// 17
//    130,	// 18
//    134,	// 19
//    138,	// 20
//    140,	// 21
//    142,	// 22
//    144,	// 23
//    146,	// 24
//    148,	// 25
//    150,	// 26
//    152,	// 27
//    154,	// 28
//    156,	// 29
//    158,	// 30
//    160,	// 31
//    162,	// 32
//    164,	// 33
//    166,	// 34
//    168,	// 35
//    170,	// 36
//    172,	// 37
//    174,	// 38
//    176,	// 39
//    178,	// 40
//    180,	// 41
//    182,	// 42
//    184,	// 43
//    186,	// 44
//    188,	// 45
//    190,	// 46
//    192,	// 47
//    194,	// 48
//    196,	// 49
//    198,	// 50
//    199,	// 51
//    200,	// 52
//    201,	// 53
//    202,	// 54
//    203,	// 55
//    204,	// 56
//    205,	// 57
//    206,	// 58
//    207,	// 59
//    208,	// 60
//    209,	// 61
//    210,	// 62
//    211,	// 63
//    212,	// 64
//    213,	// 65
//    214,	// 66
//    215,	// 67
//    216,	// 68
//    217,	// 69
//    218,	// 70
//    219,	// 71
//    220,	// 72
//    221,	// 73
//    222,	// 74
//    223,	// 75
//    224,	// 76
//    225,	// 77
//    226,	// 78
//    227,	// 79
//    228,	// 80
//    229,	// 81
//    230,	// 82
//    231,	// 83
//    232,	// 84
//    233,	// 85
//    234,	// 86
//    235,	// 87
//    236,	// 88
//    237,	// 89
//    238,	// 90
//    239,	// 91
//    240,	// 92
//    241,	// 93
//    242,	// 94
//    243,	// 95
//    244,	// 96
//    245,	// 97
//    246,	// 98
//    247,	// 99
//};

/**
 * Global data structure to hold volume data.
 */
typedef struct
{
    //! A value to automatically attenuate the signal whenever volume_set is called
    // Then value is a number of 0dB steps
    int8_t attenuation;
    
    //! The current balance
    uint8_t balance;

    //! The current volume
    uint8_t volume;

    // The current ramp volume
    // Set to volume when not ramping
    uint8_t ramp_volume;
    
    // True if the volume is fixed at 99
    bool fixed_volume;
    
    //! Indicates that the volume has been explicitly muted rather than maumally set
    // to zero (which also mutes it))
    bool muted;
    
    //! Determines whether the max on volume should be checked
    bool check_max_on_volume;
    
    //! Used to determine which value is returned by volume_get_max_volume
    VOLUME_INPUT current_input;
    
} VOLUME_DATA;

static VOLUME_DATA volume_data =
{
    .check_max_on_volume = true,
    .current_input = VOLUME_INPUT_DEFAULT,
    .fixed_volume = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Start a ramp between volumes
 */
static void volume_ramp_start()
{
    // Set the starting value - this is the same as the current value so may need to increment first
    volume_set_volume(volume_data.ramp_volume);
    
    timers_reset(TIMER_INDEX_VOLUME_RAMP);
    timers_reset(TIMER_INDEX_BRIGHTNESS);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Sets a number of 0.5dB steps by which the volume is automatically attenuated by
 * No no change is made until the next volume_set
 */
void volume_attenuate(int8_t by)
{
    volume_data.attenuation = by;
    
    logger_log_infof(LOGGER_CODE_VOLUME_CHANGE, "VOLUME attenuation set at %ddB ", (by / 2));
}

/**-----------------------------------------------------------------------------
 * Sets a number of 0.5dB steps by which the volume is automatically boosted by
 * No no change is made until the next volume_set
 */
void volume_boost(int8_t by)
{
    volume_data.attenuation = by * -1;
    
    logger_log_infof(LOGGER_CODE_VOLUME_CHANGE, "VOLUME boost set at %ddB ", (by / 2));
}
   
/**-----------------------------------------------------------------------------
 * Sets it so that the max on volume is checked when the volume is next set
 */
void volume_check_max_on_volume()
{
    volume_data.check_max_on_volume = true;
}

/**-----------------------------------------------------------------------------
 * Simulates a fade in from silence
 */
void volume_fade_in()
{
    logger_log_infof(LOGGER_CODE_VOLUME_CHANGE, "VOLUME Fade In Start --");
   // volume_mute();
    //volume_unmute();
    uint8_t volume = volume_get();
    volume_set(0, false);
    volume_unmute();
    volume_set(volume, false);
    volume_data.muted = false;
    logger_log_infof(LOGGER_CODE_VOLUME_CHANGE, "VOLUME Fade In End --");
}
    
/**-----------------------------------------------------------------------------
 * Gets the absolute volume 
 * @return - the absolute volume
 */
uint8_t volume_get()
{
    return volume_data.volume;
}

/**-----------------------------------------------------------------------------
 * Get whether fixed volume is in force.
 * @return 
 */
bool volume_get_fixed_volume()
{
    return volume_data.fixed_volume;
}
    
/**-----------------------------------------------------------------------------
 * Gets the absolute volume level from the settings
 * @return - the absolute volume
 */
uint8_t volume_get_stored()
{
    uint8_t result;
    settings_get(SETTING_CURRENT_VOLUME, &result);
    logger_log_debugf(0, "VOLUME getting stored volume %d %d %d --", result, volume_data.check_max_on_volume, settings_initialised());
    
    if (volume_data.check_max_on_volume && settings_initialised())
    {
        result = min(result, volume_get_max_on_volume());
        volume_data.check_max_on_volume = false;
        //logger_log_debugf(0, "VOLUME Check Max On Volume result %d --", result);
        
        // Make sure the stored volume is now the smaller of max on and stored as
        // it is read again after this
        settings_set(SETTING_CURRENT_VOLUME, &result);
    }
    return result;
}

/**-----------------------------------------------------------------------------
 * The RS232 commands can adjust balance relative to center e.g. a value 0f -6
 * changes the balance to L06 - this does the calculation to convert the relative 
 * value to  an absolute balance
 * @param relative_balance
 * @return 
 */
uint8_t volume_get_adjusted_balance(int relative_balance)
{
    return VOLUME_DEFAULT_BALANCE + relative_balance;
}

/**-----------------------------------------------------------------------------
 * Gets the current volume level
 * @return - the current volume
 */
uint8_t volume_get_current()
{
    return volume_data.ramp_volume;
}

/**-----------------------------------------------------------------------------
 * Gets the current balance
 * @return - the current balance
 */
uint8_t volume_get_balance()
{
    return volume_data.balance;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the current balance value
 * @return 
 */
 void volume_get_balance_string(char* buffer, size_t size)
{
    uint8_t balance = volume_get_balance();   
    
    if (balance < VOLUME_DEFAULT_BALANCE)
    {
        sprintf(buffer, "L%02d", abs(balance - VOLUME_DEFAULT_BALANCE));
    }
    else if (balance > VOLUME_DEFAULT_BALANCE)
    {
        sprintf(buffer, "R%02d", abs(balance - VOLUME_DEFAULT_BALANCE));
    }
    else
    {
        sprintf(buffer, "%d ", abs(balance - VOLUME_DEFAULT_BALANCE));
    }
}

/**-----------------------------------------------------------------------------
 * Gets the maximum volume after power-up
 * @return 
 */
uint8_t volume_get_max_on_volume()
{
    uint8_t result;
    settings_get(SETTING_MAX_ON_VOLUME, &result);
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the maximum net volume allowed - note this is always < maximum_volume
 * @return 
 */
uint8_t volume_get_maximum_net_volume()
{
    uint8_t result;
    settings_get(SETTING_MAXIMUM_STREAMING_VOLUME, &result);
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the maximum volume 
 * @return 
 */
uint8_t volume_get_maximum_volume()
{
    uint8_t result;
    
    switch (volume_data.current_input)
    {
        case VOLUME_INPUT_STREAMER:
            settings_get(SETTING_MAXIMUM_STREAMING_VOLUME, &result);
            break;
            
        default:
            settings_get(SETTING_MAXIMUM_VOLUME, &result);
            break;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * The RS232 commands can adjust balance relative to center e.g. a value 0f -6
 * changes the balance to L06 - this does the calculation to convert the absolute 
 * value to  a relative balance
 * @param relative_balance
 * @return 
 */
int volume_get_relative_balance(uint8_t balance)
{
    return balance - VOLUME_DEFAULT_BALANCE;
}

/**-----------------------------------------------------------------------------
 * Initialise volume control management
 */
void volume_initialise()
{
    volume_data.volume = VOLUME_DEFAULT_VOLUME;
    volume_data.ramp_volume = VOLUME_DEFAULT_VOLUME;
    volume_data.balance = VOLUME_DEFAULT_BALANCE;
    volume_data.muted = false;

    //Read the balance from the settings
    settings_suppress_save(true);

#ifndef BALANCE_CONTROL_IS_POTENTIOMETER
    
    uint8_t balance;
    
    settings_get(SETTING_CURRENT_BALANCE, &balance);
    volume_set_balance(balance, false);
    
#endif
    
    
    volume_set_volume(min(volume_get(), volume_get_max_on_volume()));

#ifdef VOLUME_HAS_SETTING_FIXED_VOLUME
    bool fixed_volume;
    settings_get(SETTING_FIXED_VOLUME, &fixed_volume);
    volume_set_fixed_volume(fixed_volume);
#endif

    settings_suppress_save(false);
    
    volume_hw_initialise();

//
//#if VOLUME_PINS_ENABLED   
//
//    // Setup the pins
//    disable_transfer();
//
//    PLIB_PORTS_PinSet(VOLUME_SCLK_PORTS_ID, 
//                      VOLUME_SCLK_PORT_CHANNEL, 
//                      VOLUME_SCLK_PIN);
//    PLIB_PORTS_PinClear(VOLUME_MOSI_PORTS_ID, 
//                        VOLUME_MOSI_PORT_CHANNEL, 
//                        VOLUME_MOSI_PIN);
//
//#endif
}

/**-----------------------------------------------------------------------------
 * Check whether the volume is currently ramping
 * @return 
 */
bool volume_in_ramp()
{
    return volume_data.ramp_volume != volume_data.volume;
}

/**-----------------------------------------------------------------------------
 * Check whether the volume is muted
 */
bool volume_is_muted()
{
//    MUTES_MUTE_STATE state = mutes_state(MUTES_INDEX_VOLUME);
//    return state == MUTES_MUTE_STATE_ON;
    
    return volume_data.muted == true;
}

/**-----------------------------------------------------------------------------
 * Can be used to check whether the volume is within the valid range
 */
bool volume_is_valid(uint8_t value)
{
    return ((value >= 0) && (value <= VOLUME_MAX));
}
    
/**-----------------------------------------------------------------------------
 * Mute the volume
 */
void volume_mute()
{
    mutes_mute(MUTE_INDEX_VOLUME);

    volume_data.muted = true;
    
    volume_hw_mute();
    
    //set AK4101 volume to 0 so that when we unmute we don't get a blip before it starts ramping in
    // Log volume muted
    logger_log_info(LOGGER_CODE_VOLUME_MUTE, "VOLUME Muted --");
        
    // BUG 3762 
    bool mute_on_startup = true;
    logger_log_infof(LOGGER_CODE_INITIALISATION, "APP: Storing MUTE_ON_STARTUP %d --", mute_on_startup);
    settings_set(SETTING_MUTE_ON_STARTUP, &mute_on_startup);
    settings_save();
}

/**-----------------------------------------------------------------------------
 * Process volume for volume ramping
 */
void volume_process()
{
    if (timers_has_expired(TIMER_INDEX_VOLUME_RAMP))
    {
        if (volume_data.ramp_volume != volume_data.volume)
        {
            if (volume_data.ramp_volume < volume_data.volume)
            {
                // Move onto the next ramp position
                volume_data.ramp_volume++;
            }
            else
            {
                volume_data.ramp_volume--;
            }
        
            volume_set_volume(volume_data.ramp_volume);
        
            timers_reset(TIMER_INDEX_VOLUME_RAMP);
            
            if (volume_data.ramp_volume == volume_data.volume)
            {
                // Log the volume change
                if (!volume_is_muted() && (volume_data.volume == 0))
                {
                    mutes_mute(MUTE_INDEX_VOLUME);
                }

                //logger_log_infof(LOGGER_CODE_VOLUME_CHANGE, "VOLUME %d", (int)volume_get_current());      
            }
            
            timers_reset(TIMER_INDEX_BRIGHTNESS);
        }
    }
    
    // Give the volume a periodic kick just incase the volume control has collapsed
    if (timers_has_expired(TIMER_INDEX_VOLUME_KEEP_ALIVE))
    {
        if (volume_data.ramp_volume == volume_data.volume)
        {
            //logger_log_debugf(241, "VOLUME Keep Alive %d --", volume_data.volume);
            volume_hw_keep_alive(volume_data.ramp_volume);
            //volume_set_volume(volume_data.ramp_volume);
        }
        
        timers_reset(TIMER_INDEX_VOLUME_KEEP_ALIVE);
    }
}

/**-----------------------------------------------------------------------------
 * Sets the current volume level
 * 
 * @param value the new volume (in the range 0-99)
 */
void volume_set(uint8_t value, bool store)
{
    if (!volume_data.fixed_volume)
    {
        logger_log_infof(LOGGER_CODE_VOLUME_CHANGE, "VOLUME Set to %d from %d --", value, volume_data.volume);

        if (volume_is_valid(value))
        {
            if (volume_data.volume != value)
            {
                 // Store the previous value in ramp_volume as this is where the ramp will 
                // start from
                volume_data.ramp_volume = volume_data.volume;

                // Set the volume
                // ISSUE-20171211-1 added attenuation subtraction
                // Add check for MAX volume
                volume_data.volume = min(value, volume_get_maximum_volume());

                logger_log_command(APP_COMMANDS_VOLUME, &value, 1);
                //logger_log_debugf(0, "VOLUME %d %d  --", volume_data.ramp_volume, volume_data.volume);            

                if (volume_data.ramp_volume == 0)
                {
                    // BUG-20170927-1 - force muted state to ensure unmute() does something
                    volume_data.muted = true;
                    volume_unmute();
                }
                else 
                {
                    if (volume_data.volume == 0)
                    {
                        //volume_mute();
                    }
                    else if (volume_is_muted())
                    {
                        volume_unmute();
                    }
                }

                // We will ramp up to the new volume
                volume_ramp_start();
            }

            // Update settings?
            if (store)
            {
                settings_set(SETTING_CURRENT_VOLUME, &volume_data.volume);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Sets the current balance
 * 
 * @param value the new balance (in the range 0-99/100?)
 */
void volume_set_balance(uint8_t value, bool store)
{
    // Make sure value is within limits
    value = min(max(value, VOLUME_DEFAULT_BALANCE - 12), VOLUME_DEFAULT_BALANCE + 12);
    
    if (volume_data.balance != value)
    {
        // Set the volume
        volume_data.balance = value;
        
        //logger_log_command(APP_COMMANDS_BALANCE, &value, 1);

        //Set the value on the dac
        volume_set_volume(volume_data.ramp_volume);

        // Update settings?
        if (store)
        {
      
            
#ifndef BALANCE_CONTROL_IS_POTENTIOMETER
            
            // for a potentiometer balance control, the setting is the physical position
            
            settings_set(SETTING_CURRENT_BALANCE, &volume_data.balance);
#endif
        }

        logger_log_infof(LOGGER_CODE_BALANCE_CHANGE, "VOLUME Balance %d --", (int)volume_data.balance);
    }
}
    
/**-----------------------------------------------------------------------------
 * Sets the current input type - this affects which values is returned by volume_get_max_volume
 * @param input
 */
void volume_set_current_input(VOLUME_INPUT input)
{
    logger_log_debugf(LOGGER_CODE_VOLUME, "VOLUME setting current input %d --", input);
    volume_data.current_input = input;
}
    
/**-----------------------------------------------------------------------------
 * Sets the volume to a fixed 99. volume_set_volume is ignored..
 * @param on
 */
void volume_set_fixed_volume(bool on)
{    
    logger_log_debugf(LOGGER_CODE_VOLUME_CHANGE, "VOLUME Setting fixed volume %s --", on ? "ON" : "OFF");
    
    if (on)
    {
        // First set the volume - this assumes that fixed-volume is false, but 
        // if it is on then the volume should be 99 anyway!
        volume_set(VOLUME_FIXED_MAX, true);
    
        // Now set the fixed volume flag
        volume_data.fixed_volume = true;
    }
    else
    {
        // unset fixed volume
        volume_data.fixed_volume = false;
        
        logger_log_debugf(LOGGER_CODE_VOLUME_CHANGE, "VOLUME Setting fixed volume %d --", volume_get_stored());
        // Reset the volume - note this will force it to go to maximum volume if specified
        volume_set(volume_get_stored(), true); // BUG 3885 - set to current volume not max volume
        //volume_set(VOLUME_DEFAULT_VOLUME, true);
    }
 
#ifdef VOLUME_HAS_SETTING_FIXED_VOLUME
    settings_set(SETTING_FIXED_VOLUME, &on);
    
    // BUG 3841
    logger_log_command(APP_COMMANDS_FIXED_VOLUME, (unsigned char*)&volume_data.fixed_volume, 1);
#endif
}
     
/**-----------------------------------------------------------------------------
 * Sets the current volume level without ramping up from the current level
 * 
 * @param value the new volume (in the range 0-99)
 * @param store - true stores value in settings, false (processor mode) - does not store the volume
 */
void volume_set_immediate(uint8_t value, bool store)
{
    if (!volume_data.fixed_volume)
    {
        logger_log_infof(LOGGER_CODE_VOLUME_CHANGE, "VOLUME Immediate Set to %d from %d --", value, volume_data.volume);

        if (volume_is_valid(value) &&(volume_data.volume != value))
        {
            volume_data.volume = value;
            volume_data.ramp_volume = value;

            volume_set_volume(volume_data.volume);
            
            if (volume_data.volume == 0)
            {
                mutes_mute(MUTE_INDEX_VOLUME);
            }
            else if (volume_is_muted())
            {
                volume_data.muted = false;
                mutes_unmute(MUTE_INDEX_VOLUME);
            }

            // Update settings?
            if (store)
            {
                settings_set(SETTING_CURRENT_VOLUME, &volume_data.volume);
            }
        }
    }
}
   
/**-----------------------------------------------------------------------------
 * Sets the maximum value for when the unit is switched on
 * 
 * @param value the new balance (in the range 0-??)
 */
void volume_set_max_on_volume(uint8_t value)
{
    settings_set(SETTING_MAX_ON_VOLUME, &value);
}
    
/**-----------------------------------------------------------------------------
 * Sets the maximum volume that can be set
 * 
 * @param value the new balance (in the range 0-??)
 */
uint8_t volume_set_maximum_volume(uint8_t value, VOLUME_INPUT input)
{
    uint8_t result = value;
    uint8_t curr_vol;
    
    switch (input)
    {
        case VOLUME_INPUT_DEFAULT:
            settings_set(SETTING_MAXIMUM_VOLUME, &result);
            
            // Ensure NET volume is always < max volume
            settings_get(SETTING_MAXIMUM_STREAMING_VOLUME, &curr_vol);
            if (curr_vol > result)
            {
                settings_set(SETTING_MAXIMUM_STREAMING_VOLUME, &result);
            }
            
            break;
            
        case VOLUME_INPUT_STREAMER:
            // Ensure NET volume is always < max volume
            settings_get(SETTING_MAXIMUM_VOLUME, &curr_vol);
            result = min(result, curr_vol);
            
            settings_set(SETTING_MAXIMUM_STREAMING_VOLUME, &result);
            break;
        default:
            break;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Sets the actual volume control value using volume mapping
 * Note: This should only be called by volume.c or volume_hw_xxx.c
 */
void volume_set_volume(uint8_t value)
{
    uint8_t left;
    uint8_t right;

    // Sort out balance
    // Adapted from A29 set_volume())
    // To maintain compatibility with previous products, only cut, don't boost on balance
    // This helpfully cures the overflow when balance is adjusted at volume setting of 99.
    // Minimum non zero volume of rotary = 1 corresponds to a setting of 26 which allows us
    // to reduce it by the maximum of 12dB or setting 24, which doesn't cause underflow.
    int16_t mapped_value = (int16_t)volume_mapping[value] - (int16_t)volume_data.attenuation;
    
//    logger_log_debugf(0, "VOLUME %d %d %d %d --", value, mapped_value, volume_mapping[value], volume_data.attenuation);
    mapped_value = min(mapped_value, (uint16_t)volume_mapping[99]);
//    logger_log_debugf(0, "VOLUME %d %d %d %d --", value, mapped_value, volume_mapping[value], volume_data.attenuation);
    mapped_value = max(mapped_value, (uint16_t)1);
//    logger_log_debugf(0, "VOLUME %d %d %d %d --", value, mapped_value, volume_mapping[value], volume_data.attenuation);
    left = (uint8_t)mapped_value;    //min(max((int8_t)volume_mapping[value] - volume_data.attenuation, 1), volume_mapping[99]);
    right = left;

    if (value > 0)
    {
        // correct MSB for right balance
        if (volume_data.balance < VOLUME_DEFAULT_BALANCE)
        {
            // right cut Note ensure balance adjustment does not take right < 0 as we get underflow resulting in a POP!!!
            right -= min((VOLUME_DEFAULT_BALANCE - volume_data.balance) * 2, right);
        }
        else if (volume_data.balance > VOLUME_DEFAULT_BALANCE)
        {
            // left cut: Note ensure balance adjustment does not take left < 0 as we get underflow resulting in a POP!!!
            left -= min((volume_data.balance - VOLUME_DEFAULT_BALANCE) * 2, left);
        }
    }

    volume_hw_set_volume(left, right);
}
    
/**-----------------------------------------------------------------------------
 * Starts volume processing
 */
void volume_start()
{ 
    logger_log_info(LOGGER_CODE_VOLUME_UNMUTE, "VOLUME Starting --");
    
    volume_mute();
    
    //Read the volume/balance from the settings
    settings_suppress_save(true);

    uint8_t balance;
    settings_get(SETTING_CURRENT_BALANCE, &balance);
    volume_set_balance(balance, false);
    
    uint8_t volume;
    uint8_t max_on_volume;
    settings_get(SETTING_CURRENT_VOLUME, &volume);
    settings_get(SETTING_MAX_ON_VOLUME, &max_on_volume);
    volume_set(min(max_on_volume, volume), false); // No need to store
    
    settings_suppress_save(false);
    
    logger_log_info(LOGGER_CODE_VOLUME_UNMUTE, "VOLUME Started --");
}

/**-----------------------------------------------------------------------------
 * Unmutes the volume
 */
void volume_unmute()
{
    if (volume_is_muted())//BUG-20170927-1 && (volume_get() > 0))
    {
        volume_data.muted = false;
        volume_data.ramp_volume = 0;
 
        volume_set_volume(0);
        
        mutes_unmute(MUTE_INDEX_VOLUME);

        // We will ramp up to the new volume
        volume_ramp_start();

        // Log volume unmuted
        logger_log_info(LOGGER_CODE_VOLUME_UNMUTE, "VOLUME Unmuted --");
       
        // BUG 3762 
        bool mute_on_startup = false;
        logger_log_infof(LOGGER_CODE_INITIALISATION, "APP: Storing MUTE_ON_STARTUP %d --", mute_on_startup);
        settings_set(SETTING_MUTE_ON_STARTUP, &mute_on_startup);
        settings_save();
    }
}


/* *****************************************************************************
 End of File
 */
