//-------------------------------------------------------------
// app_config.h
//-------------------------------------------------------------

//! Defines the number of state machines available in the application
#define STATE_MACHINE_NUMBER_OF_MACHINES	5


//-------------------------------------------------------------
// example_module.c
//-------------------------------------------------------------

typedef enum
{
	PROCESS_INFO_STATE_INITIALISE,
	PROCESS_INFO_STATE_RUN,
	PROCESS_INFO_STATE_MAX,
} PROCESS_INFO_STATES;

typedef struct
{
	STATE_MACHINE_STATE_INFO process_info[2];
	
	STATE_MACHINE_ID process_id;	
	
	uint8_t process_data;
}
EXAMPLE_MODULE_DATA;

static EXAMPLE_MODULE_DATA example_module_data = 
{
};

/**-----------------------------------------------------------------------------
 *
 */
void example_module_initialise()
{
	STATE_MACHINE_STATE_INFO* process_info = example_module_data.process_info[0];
	process_info->state_id = PROCESS_INFO_STATE_INITIALISE;
	process_info->callback = example_module_initialise_process;
	process_info->data = NULL;
	
	process_info = example_module_data.process_info[1];
	process_info->state_id = PROCESS_INFO_STATE_RUN;
	process_info->callback = example_module_run_process;
	process_info->data = (void*)&example_module_data.process_data;
	
	example_module_data.process_id = state_machine_add_machine(example_module_data.process_info, PROCESS_INFO_STATE_MAX, PROCESS_INFO_STATE_INITIALISE, true);
}

/**-----------------------------------------------------------------------------
 *
 */
void example_module_initialise_process(STATE_MACHINE_ID machine_id, STATE_ID state_id, void* data)
{
	// Initialise the process
	
	// Move to the next state
	state_machine_move_state(example_module_data.process_id or machine_id, PROCESS_INFO_STATE_RUN);
}

/**-----------------------------------------------------------------------------
 *
 */
void example_module_run_process(STATE_MACHINE_ID machine_id, STATE_ID state_id, void* data)
{
	// Run the process
	(uint8_t)*data = 210;
}

//-------------------------------------------------------------
// app.c
//-------------------------------------------------------------

void APP_Tasks ( void )
{
	switch()
	{
		case APP_STATE_RUNNING:
        {
			state_machine_process();
			
			break;
		}

	}
}

//-------------------------------------------------------------
// state_machine.h
//-------------------------------------------------------------

//! Identifies an invalid state machine id
#define STATE_MACHINE_ID_INVALID	0XFF

/**
 * An identifier used to identify the state machine ( allocated by the add functionand always used as machine->id )
 */
typedef uint8_t STATE_MACHINE_ID;

/**
 * An identifier used to identify a state 
 */
typedef uint8_t STATE_ID;

/**
 *
 */
typedef void (*STATE_MACHINE_STATE_CALLBACK)(STATE_MACHINE_ID machine_id, STATE_ID state_id, void* data);

/**
 * Structure defining a single state 
 */
typedef struct
{
	//! An idenifier for a state (will usually be an enum value)
	// This shoudl go from 0 - num_stats - 1 to avoid having to map 
	// a state to a state index
	STATE_ID state_id;
	
	//! The callback to be invoked when the stat is processed
	STATE_MACHINE_STATE_CALLBACK callback;
	
	//! Data that is apssed into the callback function
	void* data;
	
	TIMER_INDEX timer_idx;
	
} STATE_MACHINE_STATE_INFO;

/**
 * Structure defining a state machine
 */
typedef struct
{
	//! The id assigned to this state machine
	STATE_MACHINE_ID id;
	
	//! Pointer to an array of STATE_MACHINE_STATE_INFO structs
	// These are maintained by the module using the state machine
	STATE_MACHINE_STATE_INFO* states;
	
	//! Indicates whther the stat machine is active or not
	bool active;
	
	//! The index in the state array of the current state.
	uint8_t current_state;
	
	//! The number of states in this state machine
	uint8_t num_states;
} STATE_MACHINE_INFO;	

//-------------------------------------------------------------
// state_machine.c
//-------------------------------------------------------------

typedef struct
{
	//! An array of stat machines info pointers
	STATE_MACHINE_INFO machines[STATE_MACHINE_NUMBER_OF_MACHINES];
	
	//! Used to keep count of the number of machines acually added.
	uint8_t num_machines;
} STATE_MACHINE_DATA;

static STATE_MACHINE_DATA state_machine_data =
{
	.num_machines = 0,
};

/**-----------------------------------------------------------------------------
 *
 */
void state_machine_activate_machine(STATE_MACHINE_ID id)
{
	if (id < STATE_MACHINE_NUMBER_OF_MACHINES}
	{
		STATE_MACHINE_INFO* machine = &state_machine_data.machines[id];
		machine->active = true;
	}
}

/**-----------------------------------------------------------------------------
 * Add a State Machine
 */
STATE_MACHINE_ID state_machine_add_machine(STATE_MACHINE_STATE_INFO* states, uint8_t num_states, uint8_t current_state, bool active)
{
	STATE_MACHINE_ID result = STATE_MACHINE_ID_INVALID;
	
	// Find the next available slot
	if (state_machine_data.num_machines < STATE_MACHINE_NUMBER_OF_MACHINES)
	{
		STATE_MACHINE_INFO* machine = &state_machine_data.machines[state_machine_data.num_machines];
		machine->id = state_machine_data.num_machines;
		machine->states = states;
		machine->num_states = num_stats;
		machine->current_state = current_state;
		machine->active = active;
		
		result = machine->id;
		
		state_machine_data.num_machines++;
	}
	
	return result;
}

/**-----------------------------------------------------------------------------
 * Change stat in a state machine
 */
void state_machine_move_state(STATE_MACHINE_ID id, STATE_ID state_id)
{
	if (id < STATE_MACHINE_NUMBER_OF_MACHINES}
	{
		STATE_MACHINE_INFO* machine = state_machine_data.machines[id];
		machine->current_state = state_id;
	}
}

/**-----------------------------------------------------------------------------
 * Invoke the state machine processing
 */
void state_machine_process()
{
	uint8_t x;
	for (x = 0; x < STATE_MACHINE_NUMBER_OF_MACHINES; x++)
	{
		STATE_MACHINE_INFO* machine = state_machine_data.machines[x];
		if (machine->active)
		{
			STATE_MACHINE_STATE_INFO* state = machine->states[machine->current_state];
			if (state->callback != NULL)
			{
				if (state->timed_idx != TIMER_INDEX_NULL)
				{
					if (timers_has_expired(state->timer_idx))
					{
						state->callback(machine->id, state->state_id, state->data);
					}
				}
				else
				{
					state->callback(machine->id, state->state_id, state->data);
				}
			}
		}
	}
}

/**-----------------------------------------------------------------------------
 *
 */
void state_machine_suspend_machine(STATE_MACHINE_ID id)
{
	if (id < STATE_MACHINE_NUMBER_OF_MACHINES}
	{
		STATE_MACHINE_INFO* machine = state_machine_data.machines[id];
		machine->active = false;
	}
}

	

