/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    adc_defs.h

  @Summary
    Public Definitions for the adc module

  @Description
    Note: This is designed to work with PIC32MZ chips which do not require Harmony ADC Calibration
    if using a PIC32 MX the calibration will need adding
 */
/* ************************************************************************** */

#ifndef ADC_DEFS_H    /* Guard against multiple inclusion */
#define ADC_DEFS_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stddef.h>
#include "system/ports/sys_ports.h"
#include "driver/adc/drv_adc_static.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


// *****************************************************************************
// *****************************************************************************
// Section: Constants                                                         
// *****************************************************************************
// *****************************************************************************

/**
 * Identifies an invalid Input Id
 */
#define ADC_INPUT_ID_INVALID    0xFF
    
/**
 * Value returned when a request to one of the get messages is made when the 
 * input type does not support the request
 */
#define ADC_INVALID_REQUEST     0xFF
    
/** 
 * Values returned from a request to get temperature has been made but the data is not yet available
 */    
#define ADC_INVALID_INPUT       0xFE    

// *****************************************************************************
// *****************************************************************************
// Section: Data Types
// *****************************************************************************
// *****************************************************************************
 
// Alias the ADC Input ID
typedef ADCHS_AN_INPUT_ID ADC_INPUT_ID;

// State of the ADC Input
typedef enum
{
    ADC_INPUT_STATE_INACTIVE = false,
    ADC_INPUT_STATE_ACTIVE = true,
} ADC_INPUT_STATE;
        
//------------------------------------------------------------------------------
// Buttons
//------------------------------------------------------------------------------
/*
 * Enum of available buttons - Currently supports upto 9 but can be extended
 */
typedef enum 
{
    ADC_BUTTON_NONE,
    ADC_BUTTON_1,
    ADC_BUTTON_2,
    ADC_BUTTON_3,
    ADC_BUTTON_4,
    ADC_BUTTON_5,
    ADC_BUTTON_6,
    ADC_BUTTON_7,
    ADC_BUTTON_8,
    ADC_BUTTON_9,
    ADC_BUTTON_ERROR = -1,
} ADC_BUTTON;  

// Defines a single button on a ADC Input
typedef struct
{
    //! The lower value (in mV) that defines the range
    uint32_t range_low;

    //! The upper value (in mV) that defines the range
    uint32_t range_high;

    //! The switch identifier to be returned
    ADC_BUTTON button;

} ADC_BUTTON_DEFINITION;

/**
 * Store data on the currently selected button
 */
typedef struct
{
    //! The Input Id that this button info is associated with
    ADC_INPUT_ID input;
    
    // The switch whose details are being recorded
    ADC_BUTTON button;

    // The time the switch was depressed for
    uint32_t start_ticks;

    // The time the button was/has been depressed
    uint32_t duration_ticks;

} ADC_BUTTON_INFO;

//------------------------------------------------------------------------------
// Pots
//------------------------------------------------------------------------------

#define ADC_POT_INVALID_STEP    0xFFFF

//------------------------------------------------------------------------------
// General
//------------------------------------------------------------------------------

//! Define a single ADC Input
typedef struct
{
    //! Determines whether the ADC input is currently active.  If not then 
    //  it is not read or updated
    ADC_INPUT_STATE active;
        
    // The specific ADC Input
    ADC_INPUT_ID input;
    
    // The last read value of the input
    uint32_t current_value;
    
    //! Callback for when data has become available wil be cast to a ADC_DATA_AVAILABLE_CALLBACK
    void* data_available_callback;
    
    //! Called when an input scan ends
    void* scan_end_callback;
    
    //! Called when an input scan starts
    void* scan_start_callback;
    
} ADC_INPUT_DEFINITION;

//! Data Available callback
typedef void (*ADC_DATA_AVAILABLE_CALLBACK)(ADC_INPUT_DEFINITION* input_def);

//! Called when a scan ends after Inputs have been tested
typedef void (*ADC_INPUT_SCAN_END_CALLBACK)();

//! Called when a scan starts before any Inputs have ben tested
typedef void (*ADC_INPUT_SCAN_START_CALLBACK)();

//------------------------------------------------------------------------------
// Events
//------------------------------------------------------------------------------

/**
 * a: Enum name b: Event Module Description
 */
#define ADC_EVENT_ITEMS \
    X(POT_DATA_AVAILABLE, ADC Pot Data Available)\

//! Uses EVENT_MODULE_ITEMS defined in app_config.h
// Define enum
#define X(a, b) ADC_EVENT_##a,
	typedef enum {ADC_EVENT_ITEMS} ADC_EVENT;
#undef X
    
// Create array of Event module names
#define X(a, b) #b,
	static char *ADC_EVENT_NAMES[] = { ADC_EVENT_ITEMS };
#undef X

//! ADC Pot data available event
typedef struct
{
    // The specific ADC Input
    ADC_INPUT_ID input;
    
    // The current step value for the Pot
    uint16_t current_step;
    
    // Indicates whether the step has changed in the last scan
    bool has_changed;
    
    //! Determines whether the pot has had its first reading taken
    bool initialised;
    
    //! The number of values to be read from the pot.  This will usually be an un-even number so it can be centered
    uint16_t step_count;
    
} ADC_POT_DATA_AVAILABLE_EVENT_DATA;

// *****************************************************************************
// *****************************************************************************
// Section: Interface Functions
// *****************************************************************************
// *****************************************************************************


    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* ADC_DEFS_H */

/* *****************************************************************************
 End of File
 */
