/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Enums and string lists used in Avatar messages.

  @Description
    Enums and string lists used in Avatar messages.
 */
/* ************************************************************************** */

#ifndef _AVATAR_DATA_H    /* Guard against multiple inclusion */
#define _AVATAR_DATA_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

// The size of data that can be sent in response to an IRQ request
#define AVATAR_IRQ_BLOCK_SIZE   64 
    
// The size of an AVATAR TCP Packet
#define AVATAR_TCP_PACKET_SIZE  512    

// The length of the Album Art URL (Actually a SHA256 hash so 32 bytes)
#define AVATAR_ALBUM_ART_URL_MAX_LENGTH  32 

// The maximum size of album art data in a single Avatar packet
#define AVATAR_ALBUM_ART_MAX_PACKET_SIZE    4096 

    
/**
 * Enum for types of Avatar input
 */
typedef enum
{
    AVATAR_INPUT_INTERNAL = 0x00,
    AVATAR_INPUT_EXTERNAL = 0x01,
    AVATAR_INPUT_BYPASS = 0x02,
} AVATAR_INPUT;

/**
 * Avatar Encoding
 */
#define AVATAR_ENCODINGS \
	X(UINT8, =0xCC, UINT8)\
	X(UINT16, =0xCD, UINT16)\
	X(UINT32, =0xCE, UINT32)\
	X(UINT64, =0xCF, UINT64)\
	X(INT8, =0xD0, INT8)\
	X(INT16, =0xD1, INT16)\
	X(INT32, =0xD2, INT32)\
	X(INT64, =0xD3, INT64)\
	X(NIL, =0xC0, NIL)\
	X(BIN16, =0xC5, BIN16)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_ENCODING_##a b,
	typedef enum {AVATAR_ENCODINGS} AVATAR_ENCODING;
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c) AVATAR_ENCODING_##a,
    static int AVATAR_ENCODING_VALUES[] = { AVATAR_ENCODINGS };
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_ENCODING_STRINGS[] = { AVATAR_ENCODINGS };
#undef X

/**
 * Avatar Message Ids
 */
#define AVATAR_MESSAGE_IDS \
	X(SET_STATES, =0x0001, Set States)\
	X(GET_STATES, =0x0002, Get States)\
	X(SUBSCRIBE, =0x0003, Subscribe)\
	X(STATE_CHANGE, =0x0004, State Change)\
	X(UNSUBSCRIBE, =0x0005, Unsubscribe)\
	X(INPUT_EVENT, =0x0006, Input Event)\
	X(TCPIP_TRANSFER, =0x0008, TCPIP Transfer)\
	X(SET_STATES_RESPONSE, =0x8001, Set States Response)\
	X(GET_STATES_RESPONSE, =0x8002, Get States Response)\
	X(SUBSCRIBE_RESPONSE, =0x8003, Subscribe Response)\
	X(TCPIP_TRANSFER_RESPONSE, =0x8008, TCPIP Transfer)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_MESSAGE_ID_##a b,
	typedef enum {AVATAR_MESSAGE_IDS} AVATAR_MESSAGE_ID;
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c) AVATAR_MESSAGE_ID_##a,
    static int AVATAR_MESSAGE_ID_VALUES[] = { AVATAR_MESSAGE_IDS };
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_MESSAGE_ID_STRINGS[] = { AVATAR_MESSAGE_IDS };
#undef X

/**
 * Avatar State Ids
 */
#define AVATAR_STATE_IDS \
	X(FIRMWARE_VERSION, =0x0001, Firmware Version, AVATAR_ENCODING_BIN16)\
	X(POWER_STATE, =0x0003, Power State, AVATAR_ENCODING_UINT8)\
	X(PLAYER_CONTROL, =0x0004, Player Control, AVATAR_ENCODING_UINT8)\
	X(REPEAT, =0x0005, Repeat, AVATAR_ENCODING_UINT8)\
	X(SHUFFLE, =0x0006, Shuffle, AVATAR_ENCODING_UINT8)\
	X(MUTE, =0x0007, Mute, AVATAR_ENCODING_UINT8)\
	X(SEEK_TRACK_TIME, =0x0008, Seek Track Time, AVATAR_ENCODING_UINT64)\
	X(VOLUME, =0x0009, Volume, AVATAR_ENCODING_UINT8)\
	X(TRACK_TIME, =0x000A, Track Time, AVATAR_ENCODING_UINT64)\
	X(TRACK_LENGTH, =0x0014, Track Length, AVATAR_ENCODING_UINT64)\
	X(PLAYER_STATUS, =0x000B, Player Status, AVATAR_ENCODING_UINT8)\
	X(TRACK_NAME, =0x000C, Track Name, AVATAR_ENCODING_BIN16)\
	X(ALBUM_NAME, =0x000D, Album Name, AVATAR_ENCODING_BIN16)\
	X(ARTIST_NAME, =0x000F, Artist Name, AVATAR_ENCODING_BIN16)\
	X(MIME_TYPE, =0x0010, MIME Type, AVATAR_ENCODING_BIN16)\
	X(AUDIO_SOURCE, =0x0015, Audio Source, AVATAR_ENCODING_UINT8)\
	X(BALANCE, =0x0019, Balance, AVATAR_ENCODING_UINT8)\
	X(NETWORK_STATE, =0x001E, Network State, AVATAR_ENCODING_UINT8)\
	X(SYSTEM_UP_STATE, =0x0020, System Up State, AVATAR_ENCODING_UINT8)\
	X(GOOGLE_LED, =0x0106, Set Google LED, AVATAR_ENCODING_BIN16)\
	X(IP_ADDRESS, =0x0301, IP Address, AVATAR_ENCODING_UINT32)\
	X(ETHERNET_MAC_ADDRESS, =0x0302, Ethernet MAC Address, AVATAR_ENCODING_BIN16)\
	X(HOST_NAME, =0x0303, Host Name, AVATAR_ENCODING_BIN16)\
	X(CHROMECAST_APP_NAME, =0x0304, Chromecast App Name, AVATAR_ENCODING_BIN16)\
	X(AUDIO_INPUT, =0x0305, Audio Input, AVATAR_ENCODING_UINT8)\
	X(DIRAC_CALIBRATION, =0x0306, Dirac Calibration, AVATAR_ENCODING_UINT8)\
	X(ALBUM_ART_URL, =0x0307, Album Art URL, AVATAR_ENCODING_BIN16)\
	X(ALBUM_ART, =0x0308, Album Art, AVATAR_ENCODING_BIN16)\
	X(SAMPLE_RATE, =0x0309, Sample Rate, AVATAR_ENCODING_UINT8)\
	X(CRESTRON_CONFIGURE, =0x030A, Crestron Configure, AVATAR_ENCODING_UINT8)\
	X(UPGRADE_STATUS, =0x030B, Upgrade Status, AVATAR_ENCODING_UINT8)\
	X(UPGRADE_DATA_SIZE, =0x030C, Upgrade Data Size, AVATAR_ENCODING_UINT32)\
	X(UPGRADE_DATA_OFFSET, =0x030D, Upgrade Data Offset, AVATAR_ENCODING_UINT32)\
	X(UPGRADE_DATA, =0x030E, Upgrade Data, AVATAR_ENCODING_BIN16)\
	X(SSID, =0x030F, SSID, AVATAR_ENCODING_BIN16)\
	X(WIFI_MAC_ADDRESS, =0x0310, WiFi MAC Address, AVATAR_ENCODING_BIN16)\
	X(FRIENDLY_NAME, =0x0311, Friendly Name, AVATAR_ENCODING_BIN16)\
	X(MQA_STATE, =0x0312, MQA State, AVATAR_ENCODING_UINT8)\
	X(REGION, =0x0313, Region, AVATAR_ENCODING_UINT8)\
	X(SERIAL_NUMBER, =0x0314, Serial Number, AVATAR_ENCODING_BIN16)\
	X(MAC_ADDRESS, =0x0315, MAC Address, AVATAR_ENCODING_BIN16)\
	X(DIRAC_CLIENT_STATUS, =0x0319, Dirac Client Status, AVATAR_ENCODING_UINT8)\
	X(HEADPHONE_STATUS, =0x0320, Headphone Status, AVATAR_ENCODING_UINT8)\
	X(DISPLAY_SAMPLE_RATE, =0x0329, Display Sample Rate, AVATAR_ENCODING_UINT8)\
	X(HOST_FIRMWARE_VERSION, =0xF001, Host Firmware Version, AVATAR_ENCODING_BIN16)\
	X(HOST_FIRMWARE_STATE, =0xF002, Host Firmware State, AVATAR_ENCODING_UINT8)\
	X(DEVICE_SYSTEM_INFO, =0xF300, Device System Info, AVATAR_ENCODING_UINT16)\
	X(FACTORY_RESET, =0xF301, Factory Reset, AVATAR_ENCODING_NIL)\
	X(BROWSE_FOLDER, =0xF302, Browse Folder, AVATAR_ENCODING_BIN16)\
	X(PLAY_TRACK, =0xF303, Play Track, AVATAR_ENCODING_UINT16)\
	X(BROWSE_BACK, =0xF304, Browse Back, AVATAR_ENCODING_UINT64)\
	X(CONTROL4_DISCOVERY, =0xF305, Control4 Discovery, AVATAR_ENCODING_UINT8)\
	X(CONTROL4_IDENTIFY, =0xF306, Control4 Identify, AVATAR_ENCODING_UINT8)\
	X(DIRAC_CURVE_NAMES, =0xF307, Dirac Curve Names, AVATAR_ENCODING_BIN16)\
	X(DIRAC_SELECT_CURVE, =0xF308, Dirac Select Curve, AVATAR_ENCODING_UINT8)\
	X(SET_SAMPLE_RATE, =0xF309, Sample Rate, AVATAR_ENCODING_UINT8)\
	X(UPGRADE_OPTION, =0xF30A, Upgrade Option, AVATAR_ENCODING_UINT8)\
	X(DEVICE_UPGRADE_STATUS, =0xF30B, Device Upgrade Status, AVATAR_ENCODING_UINT8)\
	X(CRESTRON_DISCOVERY, =0xF30C, Crestron Discovery, AVATAR_ENCODING_UINT8)\
	X(MQA_DECODER_LEVEL, =0xF30F, MQA Decoder Level, AVATAR_ENCODING_UINT8)\
	X(LIP_SYNC_DELAY, =0xF310, Lip Sync Delay, AVATAR_ENCODING_UINT16)\
	X(MAXIMUM_VOLUME, =0xF311, Maximum Volume, AVATAR_ENCODING_UINT8)\
	X(TCP_CLIENT_CONNECTION, =0x0000, TCP Client Connection, AVATAR_ENCODING_UINT8)\
	X(TCP_MESSAGE, =0x0001, TCP Message, AVATAR_ENCODING_BIN16)\
	X(TCP_CLIENT_CONTROL, =0x0002, TCP Client Control, AVATAR_ENCODING_UINT16)\
    
// Now use the input data to define the enum
#define X(a, b, c, d) AVATAR_STATE_ID_##a b,
	typedef enum {AVATAR_STATE_IDS} AVATAR_STATE_ID;
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d) AVATAR_STATE_ID_##a,
    static int AVATAR_STATE_ID_VALUES[] = { AVATAR_STATE_IDS };
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c, d) #c,
	static char *AVATAR_STATE_ID_STRINGS[] = { AVATAR_STATE_IDS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d) d,
    static int AVATAR_STATE_ID_ENCODING[] = { AVATAR_STATE_IDS };
#undef X
    
/**
 * Avatar Key Events
 */
#define AVATAR_KEY_EVENTS \
	X(RELEASE, =0x00, Release)\
	X(PRESS, =0x01, Press)\
	X(REPEAT, =0x02, Repeat)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_KEY_EVENT_##a b,
	typedef enum {AVATAR_KEY_EVENTS} AVATAR_KEY_EVENT;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_KEY_EVENT_STRINGS[] = { AVATAR_KEY_EVENTS };
#undef X
    
/**
 * Avatar Key Events
 */
#define AVATAR_REPEAT_STATES \
	X(NORMAL, =0x00, Normal)\
	X(ONE, =0x01, One)\
	X(ALL, =0x02, All)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_REPEAT_STATE_##a b,
	typedef enum {AVATAR_REPEAT_STATES AVATAR_REPEAT_STATE_MAX} AVATAR_REPEAT_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_REPEAT_STATE_STRINGS[] = { AVATAR_REPEAT_STATES };
#undef X
    
/**
 * Avatar Key Events
 */
#define AVATAR_SHUFFLE_STATES \
	X(OFF, =0x00, Off)\
	X(ON, =0x01, On)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_SHUFFLE_STATE_##a b,
	typedef enum {AVATAR_SHUFFLE_STATES} AVATAR_SHUFFLE_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_SHUFFLE_STATE_STRINGS[] = { AVATAR_SHUFFLE_STATES };
#undef X
    
/**
 * Avatar Player Control 0x0003
 */
#define AVATAR_POWER_STATES \
	X(ONLINE, 10, Online)\
	X(STANDBY, 20, Standby)\
	X(OFFLINE, 30, Offline)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_POWER_STATE_##a,
	typedef enum {AVATAR_POWER_STATES} AVATAR_POWER_STATE;
#undef X
    
// an enum value.    
#define X(a, b, c) b,
	static int AVATAR_POWER_STATE_VALUES[] = { AVATAR_POWER_STATES };
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_POWER_STATE_STRINGS[] = { AVATAR_POWER_STATES };
#undef X
       
/**
 * Avatar Player Control 0x0004
 */
#define AVATAR_PLAYER_CONTROL_STATES \
	X(PAUSE, =0x00, Pause)\
	X(STOP, =0x01, Stop)\
	X(NEXT, =0x02, Next)\
	X(PREVIOUS, =0x03, Previous)\
	X(LIKE, =0x04, Like)\
	X(DISLIKE, =0x05, Dislike)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_PLAYER_CONTROL_##a b,
	typedef enum {AVATAR_PLAYER_CONTROL_STATES} AVATAR_PLAYER_CONTROL;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_PLAYER_CONTROL_STRINGS[] = { AVATAR_PLAYER_CONTROL_STATES };
#undef X

/**
 * Avatar Player Statuses 0x000B
 */
#define AVATAR_PLAYER_STATUSES \
	X(STOPPED, =0x00, Stopped)\
	X(CHANGING, =0x01, Changing)\
	X(PLAYING, =0x02, Playing)\
	X(PAUSED, =0x03, Paused)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_PLAYER_STATUS_##a b,
	typedef enum {AVATAR_PLAYER_STATUSES} AVATAR_PLAYER_STATUS;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_PLAYER_STATUS_STRINGS[] = { AVATAR_PLAYER_STATUSES };
#undef X

/**
 * Avatar Audio Encoding
 */    
 #define AVATAR_AUDIO_ENCODINGS \
	X(UNDEFINED, =0x0A, Undefined)\
	X(MP3, =0x00, MP3)\
	X(WAV, =0x01, Wav)\
	X(WMA, =0x02, Wma)\
	X(FLAC, =0x03, Flac)\
	X(ALAC, =0x04, Alac)\
	X(MQA, =0x05, Mqa)\


// Now use the input data to define the enum
#define X(a, b, c) AVATAR_AUDIO_ENCODING_##a b,
	typedef enum {AVATAR_AUDIO_ENCODINGS} AVATAR_AUDIO_ENCODING;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_AUDIO_ENCODING_STRINGS[] = { AVATAR_AUDIO_ENCODINGS };
#undef X
   
    
/**
 * Avatar Audio Sources 0x0015
 * Note  If these ar changed then the DISPLAY_MESSAGE_AUDIO_SOURCES define also need changing in display_message_data.h
 */
#define AVATAR_AUDIO_SOURCES \
	X(UNKNOWN, =0x00, Unknown)\
	X(AIRABLE, =0x01, Airable)\
	X(MULTIROOM, =0x02, Multiroom)\
	X(FILES, =0x03, Files)\
	X(USB, =0x04, USB)\
	X(IPOD, =0x05, Ipod)\
	X(VTUNER, =0x06, VTuner)\
	X(RHAPSODY, =0x07, Rhapsody)\
	X(SIRIUS, =0x08, Sirius)\
	X(TUNEIN, =0x09, TuneIn)\
	X(UPNP, =0x0A, Streaming)\
	X(QPLAY, =0x0B, QPlay)\
	X(BLUETOOTH, =0x0C, Bluetooth)\
	X(AIRPLAY, =0x0D, AirPlay)\
	X(CDDA, =0x0E, CD-DA)\
	X(SPOTIFY, =0x0F, Spotify)\
	X(GOOGLECAST, =0x10, Casting)\
	X(AIRABLERADIOS, =0x11, Airable Radio)\
	X(AIRABLEPODCASTS, =0x12, Airable Podcasts)\
	X(NAPSTER, =0x13, Napster)\
	X(QOBUZZ, =0x14, Qobuz)\
	X(DEEZER, =0x15, Deezer)\
	X(TIDAL, =0x16, Tidal)\
	X(ROON, =0x17, Roon Ready)\
	X(AUX, =0x18, AUX)\
    
// Now use the input data to define the enum
#define X(a, b, c) AVATAR_AUDIO_SOURCE_##a b,
	typedef enum {AVATAR_AUDIO_SOURCES AVATAR_AUDIO_SOURCE_MAX} AVATAR_AUDIO_SOURCE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_AUDIO_SOURCE_STRINGS[] = { AVATAR_AUDIO_SOURCES };
#undef X

/**
 * Avatar Crestron Configure statuses 0x030A
 */
#define AVATAR_CRESTRON_CONFIGURATION_STATUSES \
	X(START, =0x00, Start Configuration)\
	X(END, =0x01, End Configuration)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_CRESTRON_CONFIGURATION_STATUS_##a b,
	typedef enum {AVATAR_CRESTRON_CONFIGURATION_STATUSES} AVATAR_CRESTRON_CONFIGURATION_STATUS;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_CRESTRON_CONFIGURATION_STATUS_STRINGS[] = { AVATAR_CRESTRON_CONFIGURATION_STATUSES };
#undef X

/**
 * Avatar Dirac Calibration statuses 0x0306
 */
#define AVATAR_DIRAC_CALIBRATION_STATUSES \
	X(STARTED, =0x00, Calibration Started)\
	X(STOPPED, =0x01, Calibration Stopped)\
	X(UPDATED, =0x02, Calibration Updated)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_DIRAC_CALIBRATION_STATUS_##a b,
	typedef enum {AVATAR_DIRAC_CALIBRATION_STATUSES} AVATAR_DIRAC_CALIBRATION_STATUS;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_DIRAC_CALIBRATION_STATUS_STRINGS[] = { AVATAR_DIRAC_CALIBRATION_STATUSES };
#undef X

/**
 * Avatar Dirac Calibration statuses 0x0306
 */
#define AVATAR_DIRAC_CLIENT_STATUSES \
	X(DISCONNECTED, =0x00, Disconnected)\
	X(CONNECTED, =0x01, Connected)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_DIRAC_CLIENT_STATUS_##a b,
	typedef enum {AVATAR_DIRAC_CLIENT_STATUSES} AVATAR_DIRAC_CLIENT_STATUS;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_DIRAC_CLIENT_STATUS_STRINGS[] = { AVATAR_DIRAC_CLIENT_STATUSES };
#undef X

/**
 * Avatar Mute statuses 0x0007
 */
#define AVATAR_MUTE_STATES \
	X(UNMUTE, =0x00, Unmute)\
	X(MUTE, =0x01, Mute)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_MUTE_STATE_##a b,
	typedef enum {AVATAR_MUTE_STATES} AVATAR_MUTE_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_MUTE_STATE_STRINGS[] = { AVATAR_MUTE_STATES };
#undef X

/**
 * Avatar Player Statuses 0x000B
 */
#define AVATAR_HEADPHONE_STATUSES \
	X(DISCONNECTED, =0x00, Disconnected)\
	X(CONNECTED, =0x01, Connected)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_HEADPHONE_STATUS_##a b,
	typedef enum {AVATAR_HEADPHONE_STATUSES} AVATAR_HEADPHONE_STATUS;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_HEADPHONE_STATUS_STRINGS[] = { AVATAR_HEADPHONE_STATUSES };
#undef X

/**
 * Avatar TCP Client Connected 0x0000
 */
#define AVATAR_TCP_CLIENT_CONNECTED_STATES \
	X(DISCONNECTED, =0x00, Disconnected)\
	X(CONNECTED, =0x01, Connected)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_TCP_CLIENT_CONNECTED_STATE_##a b,
	typedef enum {AVATAR_TCP_CLIENT_CONNECTED_STATES} AVATAR_TCP_CLIENT_CONNECTED_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_TCP_CLIENT_CONNECTED_STATE_STRINGS[] = { AVATAR_TCP_CLIENT_CONNECTED_STATES };
#undef X

/**
 * Avatar TCP Client Connected 0x0002
 */
#define AVATAR_TCP_CLIENT_CONTROL_STATES \
	X(DISCONNECT, =0x00, Disconnect)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_TCP_CLIENT_CONTROL_STATE_##a b,
	typedef enum {AVATAR_TCP_CLIENT_CONTROL_STATES} AVATAR_TCP_CLIENT_CONTROL_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_TCP_CLIENT_CONTROL_STATE_STRINGS[] = { AVATAR_TCP_CLIENT_CONTROL_STATES };
#undef X

/**
 * Avatar TCP Transfer State - response to TCP Message and TCP Client Control
 */
#define AVATAR_TCP_TRANSFER_STATES \
	X(SUCCESS, =0x00, Success)\
	X(INVALID_MESSAGE, =0x01, Invalid Message)\
	X(INVALID_CONNECTION, =0x02, Invalid Connection)\
	X(TRANSFER_FAILED, =0x03, Transfer Failed)\
	X(INVALID_COMMAND, =0x04, Invalid Command)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_TCP_TRANSFER_STATE_##a b,
	typedef enum {AVATAR_TCP_TRANSFER_STATES} AVATAR_TCP_TRANSFER_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_TCP_TRANSFER_STATE_STRINGS[] = { AVATAR_TCP_TRANSFER_STATES };
#undef X

/**
 * Avatar Player Statuses 0x000B
 */
#define AVATAR_UPGRADE_OPTIONS \
	X(OTA_DISABLED, =0x00, Disabled)\
	X(OTA_ENABLED, =0x01, Enabled)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_UPGRADE_OPTION_##a b,
	typedef enum {AVATAR_UPGRADE_OPTIONS AVATAR_UPGRADE_OPTION_MAX} AVATAR_UPGRADE_OPTION;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_UPGRADE_OPTION_STRINGS[] = { AVATAR_UPGRADE_OPTIONS };
#undef X

/**
 * Avatar Player Statuses 0x000B
 */
#define AVATAR_UPGRADE_STATUSES \
	X(AVATAR_UPGRADE_AVAILABLE, =0x00, Avatar Upgrade Available)\
	X(DUAL_UPGRADE_AVAILABLE, =0x01, Avatar and Device Upgrade Available)\
	X(DEVICE_UPGRADE_AVAILABLE, =0x02, Device Upgrade Available)\
	X(AVATAR_UPGRADE_START, =0x03, Start Device Upgrade)\
	X(AVATAR_UPGRADE_IN_PROGRESS, =0x04, Avatar Upgrade In Progress)\
	X(AVATAR_UPGRADE_COMPLETE, =0x05, Avatar Upgrade Complete)\
	X(AVATAR_UPGRADE_FAIL, =0x06, Avatar Upgrade Fail)\
	X(DEVICE_UPGRADE_TRANSFER_COMPLETE, =0x07, Device Upgrade Transfer Complete)\

// Note  DEVICE_UPGRADE_TRANSFER_COMPLETE is retained because while its not in the spec
//       its used internally to trigger a transgfer complete message
    
// Now use the input data to define the enum
#define X(a, b, c) AVATAR_UPGRADE_STATUS_##a b,
	typedef enum {AVATAR_UPGRADE_STATUSES} AVATAR_UPGRADE_STATUS;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_UPGRADE_STATUS_STRINGS[] = { AVATAR_UPGRADE_STATUSES };
#undef X
   
/**
 * Avatar Control4 Discovery Active (for message Id 0xF305)
 */
#define AVATAR_CONTROL4_DISCOVERY_ITEMS \
	X(INACTIVE, =0x00, Inactive)\
	X(ACTIVE, =0x01, Active)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_CONTROL4_DISCOVERY_##a b,
	typedef enum {AVATAR_CONTROL4_DISCOVERY_ITEMS} AVATAR_CONTROL4_DISCOVERY;
#undef X   
    
/**
 * Avatar Player Statuses 0xF30A
 */
#define DEVICE_UPGRADE_STATUSES \
	X(READY_TO_RECEIVE, =0x00, Ready to Receive Data)\
	X(DATA_RECEIVED, =0x01, Data Received)\
	X(TRANSFER_SUCCESSFUL, =0x02, Transfer Successful)\
	X(CHECKSUM_ERROR, =0x03, Checksum Error)\
	X(UPGRADE_STARTING, =0x04, Upgrade Starting)\
	X(UPGRADE_COMPLETE, =0x05, Upgrade Complete)\
	X(UPGRADE_FAIL, =0x06, Upgrade Fail)\
	X(START_AVATAR_UPGRADE, =0x07, Start Avatar Upgrade)\
	X(START_USB_UPGRADE, =0x08, Start Avatar USB Upgrade)\

// Now use the input data to define the enum
#define X(a, b, c) DEVICE_UPGRADE_STATUS_##a b,
	typedef enum {DEVICE_UPGRADE_STATUSES} DEVICE_UPGRADE_STATUS;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *DEVICE_UPGRADE_STATUS_STRINGS[] = { DEVICE_UPGRADE_STATUSES };
#undef X
   
/**
 * Avatar Crestron Discovery Active (for message Id 0xF30C)
 */
#define AVATAR_CRESTRON_DISCOVERY_ITEMS \
	X(INACTIVE, =0x00, Inactive)\
	X(ACTIVE, =0x01, Active)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_CRESTRON_DISCOVERY_##a b,
	typedef enum {AVATAR_CRESTRON_DISCOVERY_ITEMS} AVATAR_CRESTRON_DISCOVERY;
#undef X

/**
 * AVATAR Sample rates used in HDMIAudioFormat 0x0210
 */    
   
// a. Sample Rate, b. String representation, c. Index for get_sample_rate RS232 message    
// d. The sample rate as an integer
    // Note: At the moment these are kept in sync with the SAMPLE_RATE enum
#define AVATAR_SAMPLE_RATES \
	X(0, No Signal, 0x07)\
	X(32000, 32kHz, 0x00)\
    X(44100, 44.1kHz, 0x01)\
   	X(48000, 48kHz, 0x02)\
    X(88200, 88.2kHz, 0x03)\
   	X(96000, 96kHz, 0x04)\
	X(176400, 176.4kHz, 0x05)\
    X(192000, 192kHz, 0x06)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_SAMPLE_RATE_##a,
	typedef enum {AVATAR_SAMPLE_RATES AVATAR_SAMPLE_RATE_MAX} AVATAR_SAMPLE_RATE;
#undef X
    
// Redefine the macro and use to create an array of int
#define X(a, b, c) a,
	static int AVATAR_SAMPLE_RATE_VALUES[] = { AVATAR_SAMPLE_RATES };
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a, b, c) #b,
	static char *AVATAR_SAMPLE_RATE_STRINGS[] = { AVATAR_SAMPLE_RATES };
#undef X
     
// Redefine the macro and use to create an array of int
#define X(a, b, c) c,
	static uint8_t AVATAR_SAMPLE_RATE_RS232[] = { AVATAR_SAMPLE_RATES };
#undef X
   
/**
 * Device System Info 0XF300
 */    
   
// a. Model b. Avatar Code c. description
#define AVATAR_DEVICES \
	X(ARCAM_SOLO_UNO, 0x00A0, ARCAM Solo Uno)\
	X(ARCAM_SA30, 0x00A1, ARCAM SA30)\
	X(ARCAM_AVR10, 0x00A2, ARCAM AVR10)\
	X(ARCAM_AVR20, 0x00A3, ARCAM AVR20)\
	X(ARCAM_AVR30, 0x00A4, ARCAM AVR30)\
	X(ARCAM_AV40, 0x00A5, ARCAM AV40)\
	X(NOTUSED, 0x00A6, Not Used)\
	X(JBL_SDR35, 0x00A7, JBL Synthesis SDR-35)\
	X(JBL_SDP55, 0x00A8, JBL Synthesis SDP-55)\
	X(ARCAM_ST60, 0x00A9, ARCAM ST60)\
	X(AUDIO_CONTROL_MAESTRO_X7, 0x00AA, Audio Control Maestro X7)\
	X(AUDIO_CONTROL_MAESTRO_X9, 0x00AB, Audio Control Maestro X9)\
	X(AUDIO_CONTROL_CONCERT_XR4, 0x00AC, Audio Control Concert XR-4)\
	X(AUDIO_CONTROL_CONCERT_XR6, 0x00AD, Audio Control Concert XR-6)\
	X(AUDIO_CONTROL_CONCERT_XR8, 0x00AE, Audio Control Concert XR-8)\
	X(MARK_LEVINSON_NO5707, 0x00AF, Mark Levinson No5707)\
	X(JBL_HDI_ACTIVE, 0x00B0, JBL HDi active)\
	X(JBL_SA750, 0x00B1, JBL SA750)\
	X(MARK_LEVINSON_NOXXX, 0x00AF, Mark Levinson NoXXXX)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_DEVICE_##a,
	typedef enum {AVATAR_DEVICES AVATAR_DEVICE_MAX} AVATAR_DEVICE;
#undef X
    
// Redefine the macro and use to create an array of int
#define X(a, b, c) b,
	static uint16_t AVATAR_DEVICE_VALUES[] = { AVATAR_DEVICES };
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a, b, c) #c,
	static char *AVATAR_DEVICE_STRINGS[] = { AVATAR_DEVICES };
#undef X
    
/**
 * Avatar MQA State (0x0312)
 */
#define MQA_STATES \
	X(NO_MQA, =0x00, No MQA)\
	X(MQA, =0x01, MQA)\
	X(MQA_STUDIO, =0x02, MQA.)\
	X(MQA_ORFS, =0x03, MQA ORFS)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_MQA_STATE_##a b,
	typedef enum {MQA_STATES} AVATAR_MQA_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_MQA_STATE_STRINGS[] = { MQA_STATES };
#undef X
    
/**
 * Avatar MQA State (0x0312)
 */
#define MQA_DECODER_LEVELS \
	X(FULL, =0x00, Digital Off)\
	X(CORE, =0x01, MQA Core)\
    X(AUTHENTICATING, =0x02, MQA Bitstream)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_MQA_DECODER_LEVEL_##a b,
	typedef enum {MQA_DECODER_LEVELS AVATAR_MQA_DECODER_LEVEL_MAX} AVATAR_MQA_DECODER_LEVEL;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_MQA_DECODER_LEVEL_STRINGS[] = { MQA_DECODER_LEVELS };
#undef X
   
// a. Model b. Avatar Code c. description
#define MIME_TYPES \
	X(audio/unknown, 0x00, Unknown)\
	X(audio/mpeg, 0x01, MP3)\
	X(audio/x-ms-wma, 0x02, WMA)\
	X(audio/ogg, 0x03, Ogg Vorbis)\
	X(audio/flac, 0x04, FLAC)\
	X(audio/x-wav, 0x05, WAV)\
	X(audio/x-aiff, 0x06, AIFF)\
	X(audio/x-pn-realaudio, 0x07, Real Audio)\
	X(audio/x-mpegurl, 0x08, MPEG URL)\
	X(audio/x-scpls, 0x09, SCLPS)\
	X(application/vnd.ms-wpl, 0x0A, WPL)\
	X(audio/mp4, 0x0B, MP4)\
	X(audio/x-dsd, 0x0C, DSD)\
	X(audio/opus, 0x0D, OPUS)\
	X(audio/x-siriusxm-2, 0x0E, Sirius)\

// Now use the input data to define the enum
#define X(a, b, c) #a,
	static char *AVATAR_MIME_TYPE_STRINGS[] = { MIME_TYPES };
#undef X
    
// Redefine the macro and use to create an array of int
#define X(a, b, c) b,
	static uint8_t AVATAR_MIME_TYPE_VALUES[] = { MIME_TYPES };
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a, b, c) #c,
	static char *AVATAR_MIME_TYPE_DISPLAY_STRINGS[] = { MIME_TYPES };
#undef X

    
/**
 * Avatar Regions 0x0313
 */    
 #define AVATAR_REGIONS \
    X (UNSUPPORTED, =0x00, Unsupported)\
    X (US, =0x01, US)\
    X (CA, =0x02, CA)\
    X (TW, =0x03, TW)\
    X (SG, =0x04, SG)\
    X (DE, =0x05, DE)\
    X (KR, =0x06, KR)\
    X (VN, =0x07, VN)\
    X (MY, =0x08, MY)\
    X (PH, =0x09, PH)\
    X (NZ, =0x0A, NZ)\
    X (AU, =0x0B, AU)\
    X (AE, =0x0C, AE)\
    X (ZA, =0x0D, ZA)\
    X (HR, =0x0E, HR)\
    X (JP, =0x0F, JP)\
    X (RU, =0x10, RU)\
    X (IN, =0x11, IN)\
    X (CN, =0x12, CN)\
    X (ID, =0x13, ID)\
    X (IL, =0x14, IL)\
    X (BR, =0x15, BR)\

// Now use the input data to define the enum
#define X(a, b, c) AVATAR_REGION_##a b,
	typedef enum {AVATAR_REGIONS AVATAR_REGION_MAX} AVATAR_REGION;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_REGION_STRINGS[] = { AVATAR_REGIONS };
#undef X
    
//------------------------------------------------------------------------------
// Browsing
//------------------------------------------------------------------------------    
    
#define AVATAR_BROWSE_PARENT_FOLDER_ID    0 
    
//! The maximum length of a browse item name/description   
#define AVATAR_BROWSE_ITEM_MAX_LENGTH   40    

/**
 * Avatar Browse Item types
 */
#define BROWSE_ITEM_TYPES \
	X(FOLDER, =0x00, Folder)\
	X(TRACK, =0x01, Track)\
 
// Now use the input data to define the enum
#define X(a, b, c) AVATAR_BROWSE_ITEM_TYPE_##a b,
	typedef enum {BROWSE_ITEM_TYPES AVATAR_BROWSE_ITEM_TYPE_MAX} AVATAR_BROWSE_ITEM_TYPE;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_BROWSE_ITEM_TYPE_STRINGS[] = { BROWSE_ITEM_TYPES };
#undef X
    
typedef struct
{
    bool in_use;
    uint16_t item_id;
    AVATAR_BROWSE_ITEM_TYPE type;
    char name[AVATAR_BROWSE_ITEM_MAX_LENGTH];
} AVATAR_BROWSE_ITEM;

/**
 * Avatar Browse Item
 */
#define BROWSE_PLAY_OPTIONS \
	X(DEFAULT, =0x00, Default)\
	X(ALWAYS_PLAY, =0x01, Always Play)\
 
// Now use the input data to define the enum
#define X(a, b, c) AVATAR_BROWSE_PLAY_OPTION_##a b,
	typedef enum {BROWSE_PLAY_OPTIONS AVATAR_BROWSE_PLAY_OPTION_MAX} AVATAR_BROWSE_PLAY_OPTION;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_BROWSE_PLAY_OPTION_STRINGS[] = { BROWSE_PLAY_OPTIONS };
#undef X
   
/**
 * Avatar Browse Back
 */
#define BROWSE_BACK_LEVELS \
	X(ROOT, =0x00, Root Folder)\
	X(UP_ONE, =0x01, Up One)\
 
// Now use the input data to define the enum
#define X(a, b, c) AVATAR_BROWSE_BACK_LEVEL_##a b,
	typedef enum {BROWSE_BACK_LEVELS AVATAR_BROWSE_BACK_LEVEL_MAX} AVATAR_BROWSE_BACK_LEVEL;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *AVATAR_BROWSE_BACK_LEVEL_STRINGS[] = { BROWSE_BACK_LEVELS };
#undef X
   
/**
 * Avatar Browsing Command data options
 */
#define BROWSE_COMMAND_OPTIONS \
	X(BROWSE_SELECT, Browse Select)\
	X(BROWSE_NEXT, Browse Next)\
	X(BROWSE_PREVIOUS, Browse Previous)\
	X(BROWSE_BACK, Browse Back)\
	X(BROWSE_TOP, Browse Top)\
	X(BROWSE_SKIP_UP, Browse Skip Up)\
	X(BROWSE_SKIP_DOWN, Browse Skip Down)\
	X(BROWSE_PLAY, Browse Play)\
 
// Now use the input data to define the enum
#define X(a, b) BROWSE_COMMAND_OPTION_##a,
	typedef enum {BROWSE_COMMAND_OPTIONS BROWSE_COMMAND_OPTION_MAX} BROWSE_COMMAND_OPTION;
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b) #b,
	static char *BROWSE_COMMAND_OPTION_STRINGS[] = { BROWSE_COMMAND_OPTIONS };
#undef X
    
    
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
