/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    display.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _DISPLAY_H    /* Guard against multiple inclusion */
#define _DISPLAY_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */
#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdlib.h>

#include "utilities/display/display_defs.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

#define DISPLAY_LINE_COUNT      2

// The max length of a display line
#define DISPLAY_LINE_LEN        16

// Define the maximum size for the buffer to hold display text - defaults to 10 lines
// plus space for a NULL terminator
#ifndef DISPLAY_MAX_SCROLL_TEXT_LEN
#define DISPLAY_MAX_SCROLL_TEXT_LEN (DISPLAY_LINE_LEN * 30) + 1
#endif

//! Characters for use with scrolling     
#define DISPLAY_NEWLINE_CHARACTER       10  // Forces a newline
#define DISPLAY_WHITESPACE_CHARACTER    31  // Preserves whitespace      
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Appends to the line
     */
    void display_append_text(DISPLAY_LINE line, int at, char* fmt, ...);
    
    /**
     * Appends write justified text to the line
     */
    void display_append_text_right(DISPLAY_LINE line, char* fmt, ...);
    
    /**
     * Clear the display
     *
     * @param immediate_refresh - Forces an immediate screen refresh
    */
    void display_clear(bool immediate_refresh);

    /**
     * Clears the content of the line
     */
    void display_clear_line(DISPLAY_LINE line);
    
    /**
     * Returns the current display brightness
     * @return 
     */
    DISPLAY_BRIGHTNESS display_get_brightness();

    /**
     * Gets the height of a character in pixels
     * @return 
     */
    size_t display_get_character_height();

    /**
     * Gets the width of a character in pixels
     * @return 
     */
    size_t display_get_character_width();
    
    /**
     * Copies the current line contents into a buffer
     * @param line - the line to copy
     * @param buffer - a pointer to a buffer to hold the output
     * @param size - the size of the output buffer
     * @return the number of bytes copied
     */
    size_t display_get_line(DISPLAY_LINE line, char* buffer, size_t size);
    
    /**
     * Returns the length of the display in characters
     * @return 
     */
    size_t display_get_line_length();
    
    /**
     * Allocates a special character and returns its character code
     */
    char display_get_special_character(DISPLAY_LINE line, DISPLAY_SPECIAL_CHARACTERS character);
    
    /**
     * Highlights the character at the specified ordinal position in the line
     * How the character is highlighted depends on the underlying display
     * @param line - the line to copy
     * @param at -  the ordinal position of the character in the line
     * @param highlight_spaces - If spaces are present the this determines whether they should be highlighted
     */
    void display_highlight_character(DISPLAY_LINE line, int at, bool highlight_spaces);
    
    /**
     * Highlights the character at the specified ordinal position in the line relative to the right hand side
     * e.g. for a 16 char display an at value of 2 will highlight character 14
     * How the character is highlighted depends on the underlying display
     * @param line - the line to copy
     * @param at -  the ordinal position of the character in the line
     * @param highlight_spaces - If spaces are present the this determines whether they should be highlighted
     */
    void display_highlight_character_right(DISPLAY_LINE line, int at, bool highlight_spaces);
    
    //! Initialise the Display
    /** 
     * Initialise the Display
     */
    void display_initialise();
      
    /**
     * Refreshes the screen with the current pixel buffer
     */
    void display_refresh();
    
    /**
     * Used to reset the display brightness even if it hasnt changed
     */
    void display_reset_brightness();
    
    /**
     * Cancel any scrolling
    */    
    void display_scroll_cancel();
    
    /**
     * If Scrolling is paused this causes the display to scroll a single line
    */    
    void display_scroll_line();
    
    /**
     * Process scrolling function
     */
    void display_scroll_process();
    
    /**
     * 
     * @param line
     * @param start_at
     * @param width
     * @param text_block - A NULL terminated block of text
     * @param pause_ms
     * @param rollover_ms
     * @param split_words
     */
    void display_scroll_text_block(DISPLAY_LINE line, int start_at, int width, char* text_block, uint16_t pause_ms, uint16_t rollover_ms, bool split_words, uint8_t scroll_focus_line);
    
    
    /**
     * Used to set the display brightness
     * @param brightness - A value from the DISPLAY_BRIGHTNESS enum
     */
    void display_set_brightness(DISPLAY_BRIGHTNESS brightness);
    
    /**
     * Displays a special character - Note: The values in this list match the special keys in display_defs.h
     */
    void display_special_character(DISPLAY_LINE line, int at, DISPLAY_SPECIAL_CHARACTERS character);
    
    /**
     * Displays a special character that is offset 
     * Note: The values in this list match the special keys in display_defs.h
     * 
     */
    void display_special_character_offset(DISPLAY_LINE line, int at, char* buffer, uint8_t buffer_height, uint8_t buffer_width,  uint16_t index, int offset_h, int offset_v);
    
    /**
     * Clears the content of the line and adds the specified text left justified
     */
    void display_start_text(DISPLAY_LINE line, char* text);
    
    /**
     * Display text on the bottom line of a 2 line display
     * @param text - the text to be written
     */
    void display_write_line(DISPLAY_LINE line, bool immediate_refresh);
    
    /**
     * Handles the display of UTF characters
     * @param line
     * @param immediate_refresh
     */
    void display_write_line_inc_utf(DISPLAY_LINE line, bool immediate_refresh);
    
    /**
     * Display text on the top line of a 2 line display
     * @param text - the text to be written
     */
    //void display_write_top_line(char* text, bool immediate_refresh);

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
