/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _DISPLAY_MESSAGE_H    /* Guard against multiple inclusion */
#define _DISPLAY_MESSAGE_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/display_message_data.h"
#include <stdbool.h>
#include "utilities/logger/logger.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    //------------------------------------------------------------------------------
    // MCU Data Types/Enums
    // These data types are used by the MCU when construction Display Messages
    //------------------------------------------------------------------------------

    /**
     * Enum used with the build_screen_data callback to identify what type of display 
     * the menu item is being used for
     */
    typedef enum
    {
        DISPLAY_MESSAGE_DISPLAY_TYPE_PARENT = 0x00, 
        DISPLAY_MESSAGE_DISPLAY_TYPE_CHILD = 0x01,
    } DISPLAY_MESSAGE_DISPLAY_TYPE;

    /**
     * Defines an enumeration that identifies how the screen is being navigated to
     * as the behaviour depends on it.
     */
    typedef enum
    {
        DISPLAY_MESSAGE_SCREEN_NAVIGATION_NONE,
        DISPLAY_MESSAGE_SCREEN_NAVIGATION_DIRECT,
        DISPLAY_MESSAGE_SCREEN_NAVIGATION_BACK,
        DISPLAY_MESSAGE_SCREEN_NAVIGATION_SELECT,
        DISPLAY_MESSAGE_SCREEN_NAVIGATION_SELECT_EDIT,
        DISPLAY_MESSAGE_SCREEN_NAVIGATION_NEXT,
        DISPLAY_MESSAGE_SCREEN_NAVIGATION_PREVIOUS,
    } DISPLAY_MESSAGE_SCREEN_NAVIGATION;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Initialises a buffer that will be a screen_data message.  Note the buffer should be 
     * large enough to hold the data
     * @param buffer    The buffer to be populated  
     * @param data_id   The ID of the data item 
     * @param data_id   The type data item (OptionList, PreDefined value)
     * @param encoding  The encoding used for the data (Boolean, UINT8, BIN16)
     * @param data      The encoded data - Note: if the encoding is BIN16 - the data will be CONVERTED into BIN16, so no need to do it beforehand
     * @param data_len  The length of the encoded data
     */
    void display_message_add_screen_data(uint8_t* screen_data, 
            DISPLAY_MESSAGE_DATA_ID data_id,
            uint8_t attribute_id,
            DISPLAY_MESSAGE_DATA_ENCODING data_encoding, 
            uint8_t* data, uint16_t data_len);

//    /**
//     * Used as a default for menu_add_build_screen_data_function that generates the
//     * screen data based on the type of menu item (Option List, Edit)
//     * @param menu_id
//     * @param screen_data
//     */
//    void display_message_build_screen_data(MENU_ID menu_id);

    /**
     * Works out what the first option to be displayed should be based on the page_size and currently selected option
     * @param menu_item
     * @param selected_idx
     * @param page_size
     * @return 
     */
    uint8_t display_message_calculate_top_option_idx(uint16_t page_start_idx, uint16_t selected_idx, uint8_t page_size, uint8_t option_count);
    
//    /**
//     * Generates the appropriate display_messages based on the current item, and 
//     * navigation type
//     * @param menu
//     * @param navigation_type
//     */
//    void display_message_display_screen(MENU_ID menu_id, DISPLAY_MESSAGE_SCREEN_NAVIGATION navigation_type);

    /**
     * Finalises a buffer that will be a screen_data message.  Note the buffer should be 
     * large enough to hold the data
     * @param buffer
     */
    uint16_t display_message_finalise_and_send_screen_data(uint8_t* buffer);
    
    /**
     * Sets up a DISPLAY_MESSAGE_DATA_MAP for the supplied display message
     * Is used to speed up ordinal access of Screen Data 
     */
    void display_message_data_map_build(DISPLAY_MESSAGE* message, DISPLAY_MESSAGE_DATA_MAP* map);
    
    /**
     * Puts the display data into an array.  For bin 16 fields this will be the data 
     * only and omit the length part of the bin 16 field.
     * @param message - The message containing the data
     * @param data_id - The data Id for the data
     * @param attribute_id - The attribute Id for the data
     * @param encoding - The encoding type of the data
     * @param data - A pointer to an array of uint8_t to hold the data
     * @param len - The length of data available in the array - if the actual data is longer only this amount will be copied.
     * @return the length of the data copied (0 if no data or NIL type).
     */
    uint16_t display_message_get_data(DISPLAY_MESSAGE* message, 
            DISPLAY_MESSAGE_DATA_ID data_id, 
            uint8_t attribute_id, 
            DISPLAY_MESSAGE_DATA_ENCODING* encoding, 
            uint8_t* data, 
            uint16_t data_len);

//    /**
//     * Gets the page size of the MENU_ITEM (set by menu_define_screen_data)
//     * @param menu_item
//     * @return 
//     */
//    uint8_t display_message_get_page_size(MENU_ITEM* menu_item);
    
    /**
     * Gets the next frame Id
     * @return 
     */
    uint8_t display_message_get_next_frame_id();
    
    /**
     * Populates a pointer with the address of the actual data in the message
     * @param message
     * @param data_id
     * @param attribute_id
     * @param encoding
     * @param value - A pointer to a pointer that will be set to point to the data in th packet
     * @return true if state_if found in message
     */
    bool display_message_get_value(DISPLAY_MESSAGE* message, 
            DISPLAY_MESSAGE_DATA_ID data_id, 
            uint8_t attribute_id,
            DISPLAY_MESSAGE_DATA_ENCODING* encoding, 
            void** value, 
            uint16_t* len);

    /**
     * Can be used to determine whether the display message has had errors when being transferred
     * @param message
     * @return 
     */
    bool display_message_has_errors(DISPLAY_MESSAGE* message);

    /**
     * Initialises the supplied Display Message readying it for populating
     * @param message
     */
    void display_message_initialise(DISPLAY_MESSAGE* message);

    /**
     * Initialises a buffer that will be a screen_data message.  Note the buffer should be 
     * large enough to hold the data
     * @param buffer
     */
    void display_message_initialise_screen_data(uint8_t* buffer);

    /**
     * Initialises a buffer that will be a data_transfer message.  Note the buffer should be 
     * large enough to hold the data
     * @param buffer
     */
    void display_message_initialise_data_transfer_data(uint8_t* buffer, uint8_t packet_id);

    /**
     * Can be used to determine whether the display message has been fully populated
     * @param message
     * @return 
     */
    bool display_message_is_complete(DISPLAY_MESSAGE* message);

    /**
     * Can be used to determine whether the display message is currently being populated
     * @param message
     * @return 
     */
    bool display_message_is_processing(DISPLAY_MESSAGE* message);
    
    /**
     * Process a byte of data read from the RX channel
     * @param data_byte
     */
    void display_message_process_byte(DISPLAY_MESSAGE* message, unsigned char data_byte);

    /**
     * Checks the message cache for a matching message and resends if found
     * @param frame_id
     * @param packet_id
     */
    void display_message_resend_from_cache(uint8_t frame_id, uint8_t packet_id);
    
    /**
     * resends the screen data buffer without adding the trailing EOF markers
     * for use with a backup of a buffer sent using display_message_finalise_and_send_screen_data
     */
    void display_message_resend_screen_data(uint8_t* buffer);
    
    /**
     * Sends a single item data request - intended for use from Display MCU to Main MCU
     * @param data - A pointer to the data to be sent
     * @param data_len - The length of the data
     */
    void display_message_send_data_request(DISPLAY_MESSAGE_DATA_ID data_id,
        uint8_t attribute_id,
        DISPLAY_MESSAGE_DATA_ENCODING data_encoding,
        uint8_t* data, 
        size_t data_len);

    /**
     * Sends a single item display control message - intended for use from Main MCU to Display MCU
     * @param data - A pointer to the data to be sent
     * @param data_len - The length of the data
     */
    void display_message_send_display_control(DISPLAY_MESSAGE_DATA_ID data_id,
        uint8_t attribute_id,
        DISPLAY_MESSAGE_DATA_ENCODING data_encoding,
        uint8_t* data, 
        size_t data_len);

    /**
     * Sends a Log message - intended for use from Display MCU to Main MCU
     * @param message - A pointer to the message to be sent
     * @param message_len -  The length of the message
     */
    void display_message_send_log_message(LOGGER_MESSAGE* message);

    /**
     * Sends Upgrade Data from Host->Display
     * @param total_size -  The total size of the Upgrade data
     * @param offset -  The offset in the total data of the data included in this message
     * @param upgrade_data - The data to be transferred
     * @param data_length -  The size of the data to be transferred
     */
//    void display_message_send_upgrade_data(uint32_t total_size,
//            uint32_t offset,
//            uint8_t* upgrade_data,
//            uint32_t data_length);

    /**
     * Sends an Upgrade Status message either between Host->Display or Display->Host
     * @param status - The status to be transferred
     */
    void display_message_send_upgrade_status(DISPLAY_MESSAGE_UPGRADE_STATUS status);
    
    /**
     * Sets the control, flags on a display message buffer
     * @param screen_data
     * @param control_flags
     */
    void display_message_set_control_flags(uint8_t* screen_data, DISPLAY_MESSAGE_CONTROL_FLAGS control_flags);
    
    /**
     * Can be used to fix the frame and packet Id on a message
     * Note: Does not stop the internal frame id being upated.
     * @param message
     * @param frame_id
     * @param packet_id
     */
    void display_message_set_frame_and_packet_id(uint8_t* screen_data, uint8_t frame_id, uint8_t packet_id);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _DISPLAY_MESSAGE_H */

/* *****************************************************************************
 End of File
 */
