/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    display_message_data.h

  @Summary
    Header file containing structs and enums designed to be shared between the 
    Main and Display MCUs.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _DISPLAY_MESSAGE_DATA_H    /* Guard against multiple inclusion */
#define _DISPLAY_MESSAGE_DATA_H

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Constants                                                         */
/* ************************************************************************** */
/* ************************************************************************** */

// *****************************************************************************
// *****************************************************************************
// Section: Data Types
// *****************************************************************************
// *****************************************************************************

//------------------------------------------------------------------------------
// Message based Data Types/Enums
// These data types are used in the messages sent between the MCU and the Display MCU    
//------------------------------------------------------------------------------

// Defines the number of rows of data that can be displayed on a screen
// This is used by OPTION_LISTS and OPTIONS to work out how best to send pages of
// data for display    
#define DISPLAY_MESSAGE_MAX_ROWS   3    

//! The size of the default screen data buffer - must be large enough to hold
//  all the data for the biggest screen....
#define DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE     1000
#define DISPLAY_MESSAGE_UPLOAD_DATA_BUFFER_SIZE     4200 // 4096 + 22 + other fields - think it is actually 4134 max

//! Defines the maximum size of data in an Upgrade Data packet (not the full size of the packet)    
#define DISPLAY_MESSAGE_UPLOAD_DATA_MAX_PACKET_SIZE    2048//2048//490 //(512 - other data in the message)

/**
 * Message constants (blatently ripped off the Hostlink protocol)
 */
#define DISPLAY_MESSAGE_START_OF_PACKET_MSB     0xFE
#define DISPLAY_MESSAGE_START_OF_PACKET_LSB     0xFF
#define DISPLAY_MESSAGE_END_OF_PAYLOAD_MSB      0x00
#define DISPLAY_MESSAGE_END_OF_PAYLOAD_LSB      0x00

//! The size of the data buffer - at the moment this need to be bug enough to hold a 
// album art URL and Upgrade data (4096 + 35(8+1+8+8+4+2+data) = 4129)
#define DISPLAY_MESSAGE_MAX_DATA_LENGTH         DISPLAY_MESSAGE_UPLOAD_DATA_BUFFER_SIZE//4140    //256 DISPLAY_MESSAGE_UPLOAD_DATA_MAX_PACKET_SIZE+22+8

//! The maximum length used for Album/Artist/Track names    
#define DISPLAY_MESSAGE_MAX_INFO_LENGTH         100    //AVATAR_PLAYING_INFO_MAX_LENGTH
    
#define DISPLAY_MESSAGE_ALBUM_ART_URL_LENGTH    32
    
#define DISPLAY_MESSAGE_DATA_TRANSFER_PACKET_SIZE   3072 //2048 //512 //+++++
    
/**
 * Define an enumeration of Message Types
 */
typedef enum
{
    DISPLAY_MESSAGE_TYPE_UNKNOWN = 0x00,
    DISPLAY_MESSAGE_TYPE_DISPLAY_SCREEN = 0x01,
    DISPLAY_MESSAGE_TYPE_SCREEN_DATA = 0x02,
    DISPLAY_MESSAGE_TYPE_SCREEN_ACTION = 0x03,
    DISPLAY_MESSAGE_TYPE_DATA_REQUEST = 0x04,
    DISPLAY_MESSAGE_TYPE_DATA_TRANSFER = 0x05,
    DISPLAY_MESSAGE_TYPE_LOG_MESSAGE = 0x06,
    DISPLAY_MESSAGE_TYPE_UPGRADE_STATUS = 0x07,
    DISPLAY_MESSAGE_TYPE_UPGRADE_DATA = 0x08,
    DISPLAY_MESSAGE_TYPE_DISPLAY_CONTROL = 0x09,
    DISPLAY_MESSAGE_TYPE_DISPLAY_VERSION_INFO = 0x0A,
    DISPLAY_MESSAGE_TYPE_TRANSFER_ERROR = 0x0B,

} DISPLAY_MESSAGE_TYPE;

/**
 * Data encoding - defines the format of the data in a ScreenData message
 */
typedef enum
{
    DISPLAY_MESSAGE_DATA_ENCODING_NONE = 0x00,
    DISPLAY_MESSAGE_DATA_ENCODING_BOOLEAN = 0x01,
    DISPLAY_MESSAGE_DATA_ENCODING_UINT8 = 0x02,
    DISPLAY_MESSAGE_DATA_ENCODING_UINT16 = 0x03,
    DISPLAY_MESSAGE_DATA_ENCODING_UINT32 = 0x04,
    DISPLAY_MESSAGE_DATA_ENCODING_UINT64 = 0x05,
    DISPLAY_MESSAGE_DATA_ENCODING_BIN16 = 0x06,
} DISPLAY_MESSAGE_DATA_ENCODING;

/**
 * Flags used in the Display message Control Flags byte
 */
typedef enum 
{
    DISPLAY_MESSAGE_CONTROL_FLAGS_NONE = 0x00,
    DISPLAY_MESSAGE_CONTROL_FLAGS_END_OF_MESSAGE = 0x01,
} DISPLAY_MESSAGE_CONTROL_FLAGS;
    
/**
 * DISPLAY_MESSAGE_DATA_ID definition
 */    
typedef uint16_t DISPLAY_MESSAGE_DATA_ID;

/**
 * Represents an invalid DataID
 */
#define DISPLAY_MESSAGE_DATA_ID_INVALID 0xFFFF

/**
 * DISPLAY_MESSAGE_SCREEN_ID definition
 */    
typedef uint8_t DISPLAY_MESSAGE_SCREEN_ID;

/**
 * Define an enumeration of Screen Types
 */
#define DISPLAY_MESSAGE_SCREEN_TYPES \
	X(UNKNOWN, =0x00, Unknown)\
	X(SPLASH, =0x01, Splash)\
	X(CURRENTLY_PLAYING, =0x02, Currently Playing)\
	X(PROGRESS, =0x03, Progress)\
	X(OPTION_LIST, =0x04, Option List)\
	X(OPTION_LIST_WITH_DATA, =0x05, Option List With Data)\
	X(OPTION, =0x06, Option)\
	X(UPGRADE, =0x07, Upgrade)\
	X(MESSAGE, =0x08, Message)\
	X(BROWSING, =0x09, Browsing)\
	X(DISPLAY_TEST, =0x0A, Display Test)\

// Now use the input data to define the enum
#define X(a, b, c) DISPLAY_MESSAGE_SCREEN_TYPE_##a b,
	typedef enum {DISPLAY_MESSAGE_SCREEN_TYPES} DISPLAY_MESSAGE_SCREEN_TYPE;
#undef X
     
#define X(a, b, c) #c,
	static char *DISPLAY_MESSAGE_SCREEN_TYPE_STRINGS[] = { DISPLAY_MESSAGE_SCREEN_TYPES };
#undef X

/**
 * Define an enumeration of Screen Actions
 */
typedef enum
{
    DISPLAY_MESSAGE_SCREEN_ACTION_UNKNOWN       = 0x00,
    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT        = 0x01,
    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT_NEXT   = 0x02,
} DISPLAY_MESSAGE_SCREEN_ACTION;

//------------------------------------------------------------------------------
// Application specific Enums
//------------------------------------------------------------------------------

/**
 * Defines application specific Attribute Ids (also context specific)
 */
typedef enum
{
    //! Identifies that the data is the default for the current DataId 
    //  This will usually be the Title, or Value
    DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT = 0x00,
 
    // Option Lists
    DISPLAY_MESSAGE_ATTRIBUTE_TITLE         = DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT,
    DISPLAY_MESSAGE_ATTRIBUTE_DATA          = 0x01,
    DISPLAY_MESSAGE_ATTRIBUTE_ICON          = 0x02,
    DISPLAY_MESSAGE_ATTRIBUTE_SELECTED      = 0x03,

    // Splash Screen
    DISPLAY_MESSAGE_ATTRIBUTE_SPLASH_MESSAGE    = DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT,
    DISPLAY_MESSAGE_ATTRIBUTE_SPLASH_PROGRESS   = 0x01,
    DISPLAY_MESSAGE_ATTRIBUTE_SPLASH_BUILDTYPE  = 0x00,
    DISPLAY_MESSAGE_ATTRIBUTE_SPLASH_BUILDVERSION = 0x01,
            
    // Edit position on Input Name menu
    DISPLAY_MESSAGE_ATTRIBUTE_OPTION_EDIT_POSITION = 0x04,

    // Currently Playing
    //! Indicates the suggested type of display for the Currently Playing screen
    // 0x00 - Simple screen, 0x01 - Full(Avatar) Screen))
    //see DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_INPUT_TYPE_VALUE
    DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_INPUT_NAME = DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT,    
    DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_INPUT_TYPE = 0x01,    
    DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_PROVIDER = 0x02,    // The Avatar provider type - Spotify/Tidal etc...TODO - Make an Enum
    DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_TRACK_TIME_INT = 0x03,    
            
    //No5707 Specific
    DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_GOOGLE_VOICE_STATE = 0x02,    // The Avatar provider type - Spotify/Tidal etc...TODO - Make an Enum

    // Volume
    DISPLAY_MESSAGE_ATTRIBUTE_PROGRESS_VALUE = DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT,
    DISPLAY_MESSAGE_ATTRIBUTE_PROGRESS_MIN = 0x01,
    DISPLAY_MESSAGE_ATTRIBUTE_PROGRESS_MAX = 0x02,
    DISPLAY_MESSAGE_ATTRIBUTE_PROGRESS_TEXT = DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT,
            
    DISPLAY_MESSAGE_ATTRIBUTE_LOGGER_TYPE = DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT,
    DISPLAY_MESSAGE_ATTRIBUTE_LOGGER_CODE = 0x01,
    DISPLAY_MESSAGE_ATTRIBUTE_LOGGER_MESSAGE = 0x02,

    DISPLAY_MESSAGE_ATTRIBUTE_ALBUM_ART_REQUEST = DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT,
    DISPLAY_MESSAGE_ATTRIBUTE_ALBUM_ART_DATA_REQUEST = 0x01,
} DISPLAY_MESSAGE_ATTRIBUTE_ID;

/**
 * Used to indicate the type of Currently Playing screen that should be displayed
 */
typedef enum
{
    DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_INPUT_TYPE_SIMPLE = 0x00,
    DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_INPUT_TYPE_FULL = 0x01,
} DISPLAY_MESSAGE_ATTRIBUTE_CURRENTLY_PLAYING_INPUT_TYPE_VALUE;

/**
 * Defines identifiers for icons that can be associated with some screen data
 * a.  Id for the enum value
 * b.  A Fixed identifier for the ICON that can be used to reference an ICON cache
 * c.  A description of the ICON
 */
#define DISPLAY_MESSAGE_ICON_TYPES \
	X(NONE, 0x00, Unknown)\
	X(AUDIO_SOURCES, 0x01, Audio Sources)\
	X(OPTICAL, 0x02, Optical)\
	X(COAXIAL, 0x03, Coaxial)\
	X(BLUETOOTH, 0x04, Bluetooth)\
	X(IPOD, 0x05, IPod)\
	X(AUDIO_SOURCE_SPOTIFY, 0x10, Spotify)\
	X(AUDIO_SOURCE_TIDAL, 0x11, Tidal)\

// Now use the input data to define the enum
#define X(a, b, c) DISPLAY_MESSAGE_ICON_##a,
	typedef enum {DISPLAY_MESSAGE_ICON_TYPES} DISPLAY_MESSAGE_ICON_ID;
#undef X

// Maps to a fixed ICON value that can be used to index icons in the display
#define X(a, b, c) b,
	static uint8_t DISPLAY_MESSAGE_ICON_VALUES[] = { DISPLAY_MESSAGE_ICON_TYPES };
#undef X
     
#define X(a, b, c) #c,
	static char *DISPLAY_MESSAGE_ICON_STRINGS[] = { DISPLAY_MESSAGE_ICON_TYPES };
#undef X

/**
 * Fixed data ids (used across multiple screens) note this is also used with Input providers
 */    
#define DISPLAY_MESSAGE_DATA_ID_MUTE        0x00F0 

/**
 * Display Message Data IDs for the Currently Playing screen
 */
typedef enum
{
    DISPLAY_MESSAGE_PROGRESS_DATA_ID_VALUE      = 0x0001,   
    DISPLAY_MESSAGE_PROGRESS_DATA_ID_BALANCE    = 0x0002,   
    DISPLAY_MESSAGE_PROGRESS_DATA_ID_TEXTUAL    = 0x0003,   
} DISPLAY_MESSAGE_PROGRESS_DATA_ID;

/**
 * Display Message Data IDs for the Currently Playing screen
 */
typedef enum
{
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_INPUT_TYPE    = 0x0000,   // Identifies the types of data that will be transferred
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_INPUT         = 0x0001,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_ARTIST        = 0x0002,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_ALBUM         = 0x0003,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_TRACK         = 0x0004,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_TRACK_LENGTH  = 0x0005,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_TRACK_TIME    = 0x0006,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_ALBUM_ART_URL = 0x0007,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_ALBUM_ART     = 0x0008,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_FORMAT        = 0x0009,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_SAMPLE_RATE   = 0x000A,
            
    // No5707 specific data Ids       
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_BLUETOOTH     = 0x000B,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_WIFI          = 0x000C,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_VOLUME        = 0x000D,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_MQA           = 0x000E,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_GOOGLE_VOICE  = 0x000F,
            
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_MESSAGE       = 0x0010,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_PLAYER_STATUS = 0x0011,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_REPEAT        = 0x0012,
    DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_SHUFFLE       = 0x0013,
} DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID;

/**
 * Display Message Data IDs for the Splash screen
 */
typedef enum
{
    DISPLAY_MESSAGE_SPLASH_SCREEN_DATA_ID_LOGO              = 0x0000,
    DISPLAY_MESSAGE_SPLASH_SCREEN_DATA_ID_PROGRESS          = 0x0001,
    DISPLAY_MESSAGE_SPLASH_SCREEN_DATA_ID_BUILD             = 0x0002,
} DISPLAY_MESSAGE_SPLASH_SCREEN_DATA_ID;

/**
 * Display Message Data IDs for Upgrade
 */
typedef enum
{
    DISPLAY_MESSAGE_UPGRADE_DATA_ID_STATUS          = 0x0001,
    DISPLAY_MESSAGE_UPGRADE_DATA_ID_SIZE            = 0x0002,
    DISPLAY_MESSAGE_UPGRADE_DATA_ID_OFFSET          = 0x0003,
    DISPLAY_MESSAGE_UPGRADE_DATA_ID_DATA_LENGTH     = 0x0004,
    DISPLAY_MESSAGE_UPGRADE_DATA_ID_DATA            = 0x0005,
    DISPLAY_MESSAGE_UPGRADE_DATA_ID_UPGRADE_TYPE    = 0x0006,
} DISPLAY_MESSAGE_UPGRADE_DATA_ID;

/**
 * Display Version Info data Ids
 */
typedef enum
{
    DISPLAY_MESSAGE_VERSION_DATAID_DISPLAY_MAJOR    = 0x0001,
    DISPLAY_MESSAGE_VERSION_DATAID_DISPLAY_MINOR    = 0x0002,
    DISPLAY_MESSAGE_VERSION_DATAID_IMAGES_MAJOR     = 0x0003,
    DISPLAY_MESSAGE_VERSION_DATAID_IMAGES_MINOR     = 0x0004,
} DISPLAY_MESSAGE_VERSION_DATAID;

/**
 * Display Version Info data Ids
 */
typedef enum
{
    DISPLAY_MESSAGE_BROWSING_DATAID_NOITEMS                 = 0x8000,
    DISPLAY_MESSAGE_BROWSING_DATAID_BREADCRUMB              = 0xFF00,
    DISPLAY_MESSAGE_BROWSING_DATAID_BREADCRUMB_UNDERLINE    = 0xFF01,
    DISPLAY_MESSAGE_BROWSING_DATAID_SKIP_CONTROLS           = 0xFF02,   // Skip controls ascend after thsi e.g. 0x8200 - None 0x8201 -BBWD
} DISPLAY_MESSAGE_BROWSING_DATAID;

/**
 * Used to identify the type of upgrade
 */
typedef enum
{
    //! Identifies the upgrade data as a Firmware upgrade 
    DISPLAY_MESSAGE_UPGRADE_TYPE_FIRMWARE   = 0x01,
            
    //! Identifies the upgrade data as an Image upgrade
    DISPLAY_MESSAGE_UPGRADE_TYPE_IMAGES     = 0x02,
} DISPLAY_MESSAGE_UPGRADE_TYPE;

/**
 * Display Message Data IDs for the Display Control
 */
typedef enum
{
    DISPLAY_MESSAGE_DISPLAY_CONTROL_DATA_ID_BRIGHTNESS    = 0x0001,
    DISPLAY_MESSAGE_DISPLAY_CONTROL_DATA_ID_LOGGER        = 0x0002,
} DISPLAY_MESSAGE_DISPLAY_CONTOL_DATA_ID;

/**
 * Used to indicate the status of the upgrade - transferred both ways between Host 
 * and Display
 */
typedef enum
{
    //! Host->Display - Indicates that upgrade data is ready to be transferred
    //TODO REMOVE - NOT USED DISPLAY_MESSAGE_UPGRADE_STATUS_UPGRADE_AVAILABLE        = 0x01,

    //! Host->Display - Indicates that all the upgrade data has been transferred
    //TODO REMOVE DISPLAY_MESSAGE_UPGRADE_STATUS_UPGRADE_COMPLETE         = 0x02,

    //! Display->Host - Indicates that the Display is ready t start receiving Upgrade Data messages
    //TODO REMOVE - NOT USED DISPLAY_MESSAGE_UPGRADE_STATUS_READY_TO_RECEIVE_DATA    = 0xF1,

    //! Display->Host - Indicates that the Display has received the upgrade data packet
    DISPLAY_MESSAGE_UPGRADE_STATUS_DATA_RECEIVED            = 0xF2,

    //! Display->Host - Indicates that the Display has received all the upgrade data
    DISPLAY_MESSAGE_UPGRADE_STATUS_UPLOAD_COMPLETE          = 0xF3,

    //! Display->Host - Indicates that the Display has received all the upgrade data but there are errors
    DISPLAY_MESSAGE_UPGRADE_STATUS_UPLOAD_ERRORS            = 0xF4,

    //! Display->Host - Indicates that the Display has no yet received a packet and it should be resent
    DISPLAY_MESSAGE_UPGRADE_STATUS_DATA_RETRY               = 0xF5,
} DISPLAY_MESSAGE_UPGRADE_STATUS;

//------------------------------------------------------------------------------
// RS232 Processing and Display Message data type
//------------------------------------------------------------------------------

/**
 * Describes the states when parsing incoming messages
 */
typedef enum
{
    //! Indicates that the message has been initialised.  This may be the 
    //  final state where messages are not processed through process buffer
    DISPLAY_MESSAGE_STATE_INITIALISED,
    DISPLAY_MESSAGE_STATE_WAIT_FOR_START_MSB,
    DISPLAY_MESSAGE_STATE_FOR_LEN_LSB,
    DISPLAY_MESSAGE_STATE_WAIT_LEN_MSB,
    DISPLAY_MESSAGE_STATE_WAIT_FRAME_ID,
    DISPLAY_MESSAGE_STATE_WAIT_PACKET_ID,
    DISPLAY_MESSAGE_STATE_WAIT_CONTROL_FLAGS,
    DISPLAY_MESSAGE_STATE_WAIT_MESSAGE_TYPE,
    DISPLAY_MESSAGE_STATE_POPULATING,

    // The message has been completely received and is ready for processing
    DISPLAY_MESSAGE_STATE_POPULATED,
    DISPLAY_MESSAGE_STATE_ERRORS,

} DISPLAY_MESSAGE_STATE;

    
/**
 * This is based on the AVATAR AUDIO SOURCES list but extended for service like Google Play etc
 * BUT NOTE - IT IS NOT THE SAME.....
 */
#define DISPLAY_MESSAGE_AUDIO_SOURCES \
	X(UNKNOWN, 0x00, Unknown)\
    X(AIRABLE, 0x01, Airable)\
	X(MULTIROOM, 0x02, Multiroom)\
	X(FILES, 0x03, Files)\
	X(USB, 0x04, USB)\
	X(IPOD, 0x05, Ipod)\
	X(VTUNER, 0x06, VTuner)\
	X(RHAPSODY, 0x07, Rhapsody)\
	X(SIRIUS, 0x08, Sirius)\
	X(TUNEIN, 0x09, TuneIn)\
	X(UPNP, 0x0A, Streaming)\
	X(QPLAY, 0x0B, QPlay)\
	X(BLUETOOTH, 0x0C, Bluetooth)\
	X(AIRPLAY, 0x0D, AirPlay)\
	X(CDDA, 0x0E, CD-DA)\
	X(SPOTIFY, 0x0F, Spotify)\
	X(GOOGLECAST, 0x10, Spotify)\
	X(AIRABLERADIOS, 0x11, Airable Radio)\
	X(AIRABLEPODCASTS, 0x12, Airable Podcasts)\
	X(NAPSTER, 0x13, Napster)\
	X(QOBUZZ, 0x14, Qobuz)\
	X(DEEZER, 0x15, Deezer)\
	X(TIDAL, 0x16, Tidal)\
	X(ROON, 0x17, Roon Ready)\
	X(AUX, 0x18, AUX)\
	X(GOOGLEPLAYMUSIC, 0x20, Google Play Music)\
	X(AMAZON, 0x21, Amazon Music)\
	X(YOUTUBEMUSIC, 0x22, YouTube Music)\
	X(SOUNDCLOUD, 0x23, SoundCloud)\
	X(BBCSOUNDS, 0x24, BBC Sounds)\
	X(MQA, 0x25, MQA)\
	X(MQASTUDIO, 0x26, MQA Studio)\
    
// Now use the input data to define the enum
#define X(a, b, c) DISPLAY_MESSAGE_AUDIO_SOURCE_##a,
	typedef enum {DISPLAY_MESSAGE_AUDIO_SOURCES DISPLAY_MESSAGE_AUDIO_SOURCE_MAX} DISPLAY_MESSAGE_AUDIO_SOURCE;
#undef X
    
#define X(a, b, c) b,
	static int DISPLAY_MESSAGE_AUDIO_SOURCE_IDENTIFIERS[] = { DISPLAY_MESSAGE_AUDIO_SOURCES };
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *DISPLAY_MESSAGE_AUDIO_SOURCE_STRINGS[] = { DISPLAY_MESSAGE_AUDIO_SOURCES };
#undef X

/**
 * Image Identifiers
 * Note: This is a super set of the Audio Sources enumeration and includes Mute/Splash screen images
 */
#define STATIC_IMAGES \
    X(NONE, 0x80, None)\
    X(SPLASH, 0x81, Splash)\
	X(MUTE, 0x82, Mute)\
	X(MUTE_LARGE, 0x83, Mute)\
	X(COAXIAL, 0x84, Coaxial)\
	X(OPTICAL, 0x85, Optical)\
	X(AUDIO_SOURCES, 0x86, Audio Sources)\
	X(BLUETOOTH_ICON, 0x87, Bluetooth Icon)\
	X(WIFI, 0x88, Wifi Icon)\
	X(FOLDER, 0x89, Folder)\
	X(TRACK, 0x8A, Track)\
	X(PLAY, 0x8B, Play)\
	X(PAUSE, 0x8C, Pause)\
	X(REPEAT, 0x8D, Repeat)\
	X(REPEAT_ONE, 0x8E, Repeat-One)\
	X(SHUFFLE, 0x8F, Shuffle)\
    
// Now use the input data to define the enum
#define X(a, b, c) IMAGE_IDENTIFIER_##a,
	typedef enum {DISPLAY_MESSAGE_AUDIO_SOURCES STATIC_IMAGES IMAGE_IDENTIFIER_MAX} IMAGE_IDENTIFIER;
#undef X
    
#define X(a, b, c) b,
	static int IMAGE_IDENTIFIERS[] = { DISPLAY_MESSAGE_AUDIO_SOURCES STATIC_IMAGES };
#undef X
    
// Redefine the macro and use to create an array of strings - note these need
// app_commands_get_array_index_of() to calculate the index of the string in the array from
// an enum value.    
#define X(a, b, c) #c,
	static char *IMAGE_IDENTIFIER_STRINGS[] = { DISPLAY_MESSAGE_AUDIO_SOURCES STATIC_IMAGES };
#undef X
    
/**
 * DisplayMessage DataType
 */
typedef struct
{
    // The length of the message
    uint16_t length;
    
    // Used to identify a set of data messages
    uint8_t frame_id;
    
    // Used to identify a packet within a set of data messages
    uint8_t packet_id;
    
    // Used to identify attributes of the message
    uint8_t control_flags;
    
    //! Identifies the type of the message
    DISPLAY_MESSAGE_TYPE message_type;
    
    //! The buffer to hold the data associated with the message
    uint8_t data[DISPLAY_MESSAGE_MAX_DATA_LENGTH];
    
    //! Internal variables - should not be touched by code outside structured_message.c
    // The state of the message (used by process_buffer to determine how far through
    // the message construction is)
    DISPLAY_MESSAGE_STATE processing_state;

    // Maintains an index of the current write position for the data
    // Internal use only
    uint16_t processing_data_index;

} DISPLAY_MESSAGE;

/**
 * An individual data items map
 */
typedef struct
{
    DISPLAY_MESSAGE_DATA_ID data_id;
    
    uint8_t attribute_id;
    
    DISPLAY_MESSAGE_DATA_ENCODING encoding;
    
    uint16_t length;
    
    uint8_t* data;
} DISPLAY_MESSAGE_DATA_ITEM_MAP;

/**
 * Defines a structure that can be used with the display_message_data_ functions 
 * to get data from a Display Message based on ordinal position
 */    
typedef struct
{
    uint8_t num_items;
    
    DISPLAY_MESSAGE_DATA_ITEM_MAP items[20];    //TODO Define the maximum number of items we can have
} DISPLAY_MESSAGE_DATA_MAP;

/**-----------------------------------------------------------------------------
 * Callback function template
 */
typedef void (*DISPLAY_MESSAGE_CALLBACK)(DISPLAY_MESSAGE* message, bool complete);

// *****************************************************************************
// *****************************************************************************
// Section: Interface Functions
// *****************************************************************************
// *****************************************************************************

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
