/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
 Contains typedefs for data used by the Display MCU

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _DISPLAY_MESSAGE_DISPLAY_H    /* Guard against multiple inclusion */
#define _DISPLAY_MESSAGE_DISPLAY_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\display_message.h"
#include "utilities\display_message_data.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
/* ************************************************************************** */
/* Section: Constants                                                         */
/* ************************************************************************** */
/* ************************************************************************** */

// *****************************************************************************
// *****************************************************************************
// Section: Data Types
// *****************************************************************************
// *****************************************************************************

//! The maximum number of data items you can have on a screen at a time    
#define DISPLAY_MESSAGE_SCREEN_DATA_MAX_DATA_ITEMS  20//16  
    
#define DISPLAY_MESSAGE_DISPLAY_TITLE_LENGTH        100
#define DISPLAY_MESSAGE_DISPLAY_VALUE_LENGTH        100

/**
 * Defines a structure to hold an individual data item for display on the screen
 */
typedef struct
{
    //! Indicates whether the data item is currently in use
    bool in_use;
    
    // The ID of the data item
    DISPLAY_MESSAGE_DATA_ID data_id;
    
    // The ID of the icon (from the ICON values list)
    IMAGE_IDENTIFIER image_id;
    
    // For option lists indicates whether the item is selected
    bool selected;
    
    // Indicates whether it has changed since the last time it was refreshed
    bool has_changed;
            
    // The Title/Name of the Option
    char title[DISPLAY_MESSAGE_DISPLAY_TITLE_LENGTH];
    
    // The Value associated with the option
    char value[DISPLAY_MESSAGE_DISPLAY_VALUE_LENGTH];
    
    // Used for text controls for storing the edit position
    int8_t edit_pos;
    
} DISPLAY_MESSAGE_SCREEN_DATA_ITEM;

/**
 * Defines a structure to hold the current state of the data on the current screen
 */
typedef struct
{
    //! The number of allocated items
    uint8_t num_items;
    
    uint8_t num_header_controls;
    
    //! An array of screen data items.  For Option Lists these will be in the order that
    // the options should be displayed.
    DISPLAY_MESSAGE_SCREEN_DATA_ITEM data[DISPLAY_MESSAGE_SCREEN_DATA_MAX_DATA_ITEMS];
} DISPLAY_MESSAGE_SCREEN_DATA;

/**
 * Defines a structure to hold information about the current screen that is being 
 * displayed
 */
typedef struct
{
    //! The Id of the currently selected data item
    DISPLAY_MESSAGE_DATA_ID currently_selected_id;
    
    //! The Id of the screen associated with the metadata.
    DISPLAY_MESSAGE_DATA_ID screen_id;
    
    //! The title of the current screen
    char screen_title[DISPLAY_MESSAGE_DISPLAY_TITLE_LENGTH];
    
    //! The type of the current screen
    DISPLAY_MESSAGE_SCREEN_TYPE screen_type;
    
    //! The data associated with the current screen
    DISPLAY_MESSAGE_SCREEN_DATA screen_data;
    
} DISPLAY_MESSAGE_SCREEN_METADATA;

// *****************************************************************************
// *****************************************************************************
// Section: Interface Functions
// *****************************************************************************
// *****************************************************************************
    
    /**
     * Can be used to get the action from a a Display Message that represents a
     * Screen Action message
     * @param message
     * @return - The action, DISPLAY_MESSAGE_SCREEN_ACTION_UNKNOWN if invalid message
     */
    DISPLAY_MESSAGE_SCREEN_ACTION display_message_action_get_action(DISPLAY_MESSAGE* message);
    
     /**
     * Can be used to get the Data Id from a a Display Message that represents a
     * Screen Action message
     * @param message
     * @return - The data Id, DISPLAY_MESSAGE_DATA_ID_INVALID if invalid message
     */
    DISPLAY_MESSAGE_DATA_ID display_message_action_get_data_id(DISPLAY_MESSAGE* message);
    
    /**
     * Gets the number of data items present in a Screen Data message
     * @param message
     * @return 
     */
    uint8_t display_message_data_get_num_items(DISPLAY_MESSAGE* message);
    
    /**
     * Returns a pointer to a struct holding information about the current screen 
     */
    DISPLAY_MESSAGE_SCREEN_METADATA* display_message_metadata_current();

    /**
     * Updates the current screen metadata from a DisplayScreen message
     * @param display_message
     */
    void display_message_metadata_update_from_screen(DISPLAY_MESSAGE* display_message);

    /**
     * Can be used to get the Screen Id from a Display Message that represents a
     * Display Screen message
     * @param message
     * @return - The Screen Id, DISPLAY_MESSAGE_DATA_ID_INVALID if invalid message
     */
    DISPLAY_MESSAGE_DATA_ID display_message_screen_get_id(DISPLAY_MESSAGE* message);

    /**
     * Can be used to get the title from a from a Display Message that represents a
     * Display Screen message
     * @param message
     * @param title - A buffer to hold the title
     * @param len - The size of the buffer
     */
    void display_message_screen_get_title(DISPLAY_MESSAGE* message, char* title, size_t len);

    /**
     * Can be used to obtain the Screen Type from a Display Message that represents a
     * Display Screen message
     * @param message
     * @return - The screen type, DISPLAY_MESSAGE_SCREEN_TYPE_UNKNOWN if invalid message.
     */
    DISPLAY_MESSAGE_SCREEN_TYPE display_message_screen_get_type(DISPLAY_MESSAGE* message);

    /**
     * Removes all the data items for the screen data 
     */
    void display_message_screen_data_clear(DISPLAY_MESSAGE_SCREEN_DATA* data);


    /**
     * Attempts to find a data_item based on the data_id in a screen data struct
     * Ii assumes that data_items are allocated in ascending order, so that as soon 
     * as it finds one that is not in use it assumes it has parsed all available items
     */
    DISPLAY_MESSAGE_SCREEN_DATA_ITEM* display_message_screen_data_find_item(
            DISPLAY_MESSAGE_SCREEN_DATA* data, 
            DISPLAY_MESSAGE_DATA_ID data_id,
            bool allocate);

    /**
     *  Clears down the Screen data struct and re-builds from the current Screen Data DISPLAY_MESSAGE
     */
    void display_message_screen_data_replace(DISPLAY_MESSAGE_SCREEN_DATA* data, DISPLAY_MESSAGE* message);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
