/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _EVENTS_H    /* Guard against multiple inclusion */
#define _EVENTS_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_config.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Constants                                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Uses EVENT_MODULE_ITEMS defined in app_config.h
// Define enum
#define X(a, b) EVENT_MODULE_##a,
	typedef enum {EVENT_MODULE_ITEMS EVENT_MODULE_MAX} EVENT_MODULE;
#undef X
    
// Create array of Event module names
#define X(a, b) #b,
	static char *EVENT_MODULE_NAMES[] = { EVENT_MODULE_ITEMS };
#undef X
   
//! Define the number of modules using events
#define EVENT_MODULES_NUM_MODULES	EVENT_MODULE_MAX    
    
//! Typedef for the module id
typedef EVENT_MODULE EVENT_MODULE_ID;

//! An identifier used for individual events.  For a given module
// this will be bin the range 0 -> n-1. As it is used to directly reference the 
// EVENT in the EVENTS list of event by ordinal position.
typedef uint8_t EVENT_TYPE;

//! Describes an event
typedef struct
{
	EVENT_MODULE_ID module_id;

	// A unique identifier for the event type in the set of events used for the curent module.  These should range 0->n-1.
	EVENT_TYPE type;

	// Any data that is context specific to this event being raise
	void* context_data;

	//!  Any data that is associated with the subscription
	void* subscription_data;
} EVENT;

//! Template for an Event callback
// Note it taks a pointer to an EVENT structure and rturns a void* pointr which can be cast 
// depending on the event type.
typedef void (*EVENTS_CALLBACK)(EVENT* event);

//! Describes the info needed fo a callback
typedef struct
{
	// The callback to be executed when the event is raised
	EVENTS_CALLBACK callback;

	// Subscription specific data to be passed when the event is raised.
	void* data;  
} EVENT_CALLBACK_INFO;

//! Describes a subscription to an event
typedef struct
{
	// Defines the type of event the subscription is for
	EVENT_TYPE type;

	//! Identifs whether the event is persistnt or transient.  If transient
	// it will be removed once its callback has been executed.
	bool persistent;

	// The info needed for the callback to be executed when the event is raised
	EVENT_CALLBACK_INFO callback_info;

	//! Used internally to determin whether he subsription is actually in use
	bool in_use;
} EVENT_SUBSCRIPTION;

//! Container for a modules events.
typedef struct
{
	EVENT_MODULE_ID module_id;

	EVENT_SUBSCRIPTION* subscriptions;

	uint8_t num_subscriptions;
} EVENTS;

// *****************************************************************************
// *****************************************************************************
// Section: Data Types
// *****************************************************************************
// *****************************************************************************


// *****************************************************************************
// *****************************************************************************
// Section: Interface Functions
// *****************************************************************************
// *****************************************************************************

/**
 * This is called by the module when invoking one of its events.
 data is data that is relevant to the event bing raised at this point in time
 */
void events_invoke(EVENT_MODULE_ID module_id, 
        EVENT_TYPE event_type, 
        void* context_data, 
        EVENT_CALLBACK_INFO* callback_info);

/**
 * This is called by the module when invoking one of its events.
 data is data that is relevant to the event bing raised at this point in time
 */
void events_raise(EVENT_MODULE_ID module_id, 
        EVENT_TYPE event_type, 
        void* context_data);

/**
 * Register the modules event list in the modules EVENTS structure
 */
void events_register(EVENT_MODULE_ID module_id, 
        EVENT_SUBSCRIPTION* subscription_list, 
        uint8_t num_subscriptions);

/**
 * Subscribe to an event
 */
void events_subscribe(EVENT_MODULE_ID module_id, 
        EVENT_TYPE type,
        EVENTS_CALLBACK callback,
        void* data,
        bool persistent);

/**
 * Unsubscribe from an event -  note it only makes a comparison on event type and callback, so
 * if there ar more than one subscriotion with differing data they will bot be removed
 */
void events_unsubscribe(EVENT_MODULE_ID module_id, 
        EVENT_SUBSCRIPTION* event_subscription);


    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
