/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    font_manager.h

  @Summary
 Used to manage access to Monotype fonts generated using BitFontCreator
 * The files are expected to be Antialiased 2pp, Unicode and exported in Microchip C format
 * The generated c file is then added to the project and the FONT_FLASH lines at the top commented out

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _FONT_MANAGER_H    /* Guard against multiple inclusion */
#define _FONT_MANAGER_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include "app_config.h"

    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    typedef uint16_t UNICODE_CHAR;
    typedef int FONT_ID;
    
    typedef struct
    {
        // The Unicode representation of the first character
        UNICODE_CHAR start;
        UNICODE_CHAR end;
        
        // The offset of the start character in the whole list of characters - 
        // calculated when adding the font
        uint16_t offset;
    } FONT_RANGE;  
   
//! Uses FONT_ITEMS defined in device_configuration.h
// Now use the input data to define the enum
#define X(a) FONT_INDEX_##a,
	typedef enum {FONT_ITEMS FONT_INDEX_MAX} FONT_INDEX;
#undef X

    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Gets a pointer to the data associated with the character in the font set
     * @param font_id
     * @return 
     */
    const char* font_manager_get_character_data(FONT_ID font_id, char* character);
    const char* font_manager_get_unicode_character_data(FONT_ID font_id, UNICODE_CHAR* character);

    /**
     * Gets the width of the character in the font set
     * @param font_id
     * @return 
     */
    uint8_t font_manager_get_character_width(FONT_ID font_id, char* character);
    uint8_t font_manager_get_unicode_character_width(FONT_ID font_id, UNICODE_CHAR* character);
    
    /**
     * Gets the height of the font
     * @param font_id
     * @return 
     */
    uint8_t font_manager_get_font_height(FONT_ID font_id);

    /**
     * Gets the width of the monotype of the font
     * @param font_id
     * @return 
     */
    int font_manager_get_font_average_width(FONT_ID font_id);

    /**
     * Gets the width of the monotype of the font
     * @param font_id
     * @return 
     */
    int font_manager_get_font_max_width(FONT_ID font_id);

    /**
     * Gets the width of the monotype of the font
     * @param font_id
     * @return 
     */
    int font_manager_get_font_width(FONT_ID font_id);

    /**
     * Gets the number of pixels represented by a single byte
     * @param font_id
     * @return 
     */
    uint8_t font_manager_get_pixels_per_byte(FONT_ID font_id);

    /**
     * Gets the pixels colour at the given index based on whether its a monochrome or anti-aliased font
     * @param data
     * @param pixel_idx
     * @param pixels_per_byte
     * @return 
     */
    uint8_t font_manager_get_pixel_saturation(FONT_ID font_id, char* data, uint8_t pixel_idx);

    /**
     * 
     * @param data
     * @param pixel_idx
     * @return 
     */
    uint8_t font_manager_get_pixel_saturation_1bit(char* data, uint8_t bit_idx);

    /**
     * 
     * @param data
     * @param pixel_idx
     * @return 
     */
    uint8_t font_manager_get_pixel_saturation_2bit(char* data, uint8_t bit_idx);

    /**
     * 
     * @param data
     * @param pixel_idx
     * @return 
     */
    uint8_t font_manager_get_pixel_saturation_4bit(char* data, uint8_t bit_idx);
    
    /**
     * Returns the length of a null terminated unicode string 
     * @param text
     * @return 
     */
    uint16_t font_manager_get_unicode_text_length(UNICODE_CHAR* text);
    
    /**
     * Initialise the font manager
     */
    void font_manager_initialise();
    
    /**
     * Used to register a font with a specific font id in the system.  Font ids must be unique to each font
     * to be used.
     * @param font_id
     * @param font_data a pointer to the character array generated by BitFontCreator
     * @param pixels_per_byte   The number of pixels a single byte of data represents - Aliased 2pp = 4, Monochrome = 8
     */
    void font_manager_register_font(FONT_ID font_id, const char* font_data, int pixels_per_byte, FONT_RANGE* ranges, uint8_t font_range_count);

    /**
     * Converts a utf-8 character string into an array of uint16_t unicode characters
     * @param uint8_str
     * @param unicode_str - assumes this is large enough to hold a 2-byte unicode 
     * representation of the string. can be declared by using uint16_t unicode_str[strlen(utf8_str)]
     */
    void font_manager_utf8_string_to_unicode(char* utf8_str, uint16_t* unicode_str);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
