/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    i2c.h

  @Summary
    Manages I2C communication.

  @Description
    As modules can share the same I2C connection this module abstracts the I2C layer.
 */
/* ************************************************************************** */

#ifndef _I2C_H    /* Guard against multiple inclusion */
#define _I2C_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdint.h>
#include "driver/i2c/drv_i2c.h"
#include "utilities/i2c_bit_bash.h"
 

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */
    
//! Used to identify an invalid client id
#define I2C_INVALID_CLIENT_ID   -1
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
    //! Defines an id value for a I2c client - used to identify the client
    typedef int8_t I2C_CLIENT_ID;

    /**
     * Request types
     */
    typedef enum
    {
        I2C_REQUEST_TYPE_RECEIVE,
        I2C_REQUEST_TYPE_TRANSMIT,
        I2C_REQUEST_TYPE_TRANSMIT_RECEIVE,
    } I2C_REQUEST_TYPE;

    /**
     * Callback function template
     */
    typedef void (*I2C_CALLBACK)(DRV_I2C_BUFFER_EVENT event,
                                I2C_REQUEST_TYPE request_type,
                                uintptr_t context);
    
    /**
     * Module type - determines how I2C messages are processed for a given module
     */
    typedef enum
    {
        I2C_MODULE_TYPE_DEFAULT,
        I2C_MODULE_TYPE_HARMONY,
        I2C_MODULE_TYPE_BIT_BASH,
    } I2C_MODULE_TYPE;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    /**
     * Initalise I2C
     */
    void i2c_initialise();

    /**
     * Opens the specified I2C module - if already open the open handle is returned
     * @param index
     * @param callback
     * @return 
     */
    I2C_CLIENT_ID i2c_open(const SYS_MODULE_INDEX index, 
            I2C_SLAVE_ADDRESS_VALUE read_address,
            I2C_SLAVE_ADDRESS_VALUE write_address,
            I2C_CALLBACK callback);

    /**
     * Perform I2C request processing
     */
    void i2c_process();
    
    /**
     * Read data from the write address.  This works by placing a transmit followed 
     * by a receive on the queue in sequence ensuring that no other requests are
     * executed in-between
     * Note: The callback will be invoked on completion of both the transmit and the 
     * receive so will need checking for
     * @param id
     * @param address
     * @param address_size
     * @param buffer
     * @param buffer_size
     * @param callbackContext - Not used
     * @param include_intermediate_stopbit - true if a stop bit is required between the transmit address and the receive request
     * @return 
     */
    bool i2c_read(const I2C_CLIENT_ID id,
            void * address,
            size_t address_size,
            void * buffer, 
            size_t buffer_size, 
            void * callbackContext,
            bool include_intermediate_stopbit);
    
    /**
     * Read data from an I2C address
     * @param index - The module index identifying the peripheral to read from
     * @param buffer
     * @param size
     * @param callbackContext - Not used
     * @return 
     */
    bool i2c_receive(const I2C_CLIENT_ID id,
                void * buffer, 
                size_t size, 
                void * callbackContext);
    
    /** 
     * Used to setup the type of processing for the module
     * This should be called after i2c_initialise and before and i2c_open calls 
     * are made on the module.
     */
    void i2c_set_module_type(const SYS_MODULE_INDEX index,
                             I2C_MODULE_TYPE module_type);
    
    /**
     * Transmit data to the write address
     * @param id
     * @param buffer
     * @param size
     * @param callbackContext - Not used
     * @return 
     */
    bool i2c_transmit(const I2C_CLIENT_ID id,
                void * buffer, 
                size_t size, 
                void * callbackContext);
    
    /**
     * Transmit data to the write address the receives data back
     * @param id
     * @param write_buffer
     * @param write_size
     * @param read_buffer
     * @param read_size
     * @param callbackContext - Not used
     * @return 
     */
    bool i2c_transmit_receive(const I2C_CLIENT_ID id,
                void * write_buffer, 
                size_t write_size, 
                void * read_buffer, 
                size_t read_size, 
                void * callbackContext);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
