/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
 * The Image Caches is a single file on disk that matches the structure of the upgrade 
 * file (but converted into Binary format to save space)
 */
/* ************************************************************************** */

#ifndef _IMAGE_CACHE_H    /* Guard against multiple inclusion */
#define _IMAGE_CACHE_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */


#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include "utilities/flash_memory.h"
#include "utilities/tft_control.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    /**
     * The idea of this is for use with both the IMAGE_CACHE_REQUEST and ALBUM_ART_DATA structures
     * to enable them both to use the same image cache routines for displaying images
     */
    typedef struct 
    {
        // The offset in the jpeg decoder when reading from the buffer
        uint16_t buffer_offset;

        // Indicates that the image buffer is ready to decode
        bool ready_to_process;

        // The size of the data in the buffer
        size_t size;

        // Pointer to something passed to retrieve_image for reference 
        TFT_CONTROL* tft_control;
    
        // A pointer to the coded image buffer contained in the owning structure
        uint8_t* buffer_ptr;	
        
        // Indicates that the request associated with this is currently being serviced
        bool in_use;

    } IMAGE_CACHE_CODED_IMAGE_HEADER;

    /**
     * A callback in addition to the one above which is supplied in the upgrade_metadata_upload_data
     * call and set to NULL once processed
     */
    typedef void (*IMAGE_CACHE_INITIALISED_CALLBACK)();

    /**
     * A callback in addition to the one above which is supplied in the upgrade_metadata_upload_data
     * call and set to NULL once processed
     */
    typedef void (*IMAGE_CACHE_UPGRADE_CALLBACK)(uint32_t total_size, uint32_t bytes_written);

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Gets the height of the supplied image (0 if not found)
     * @param image_id
     * @return 
     */
    uint32_t image_cache_get_image_height(uint8_t image_id);

    /**
     * Gets the width of the supplied image (0 if not found)
     * @param image_id
     * @return 
     */
    uint32_t image_cache_get_image_width(uint8_t image_id);
    
    /**
     * Gets the major version number
     */
    uint8_t image_cache_get_major_version();
    
    /**
     * Gets the minor version number
     */
    uint8_t image_cache_get_minor_version();
    
    /**
     * Can be used to check whether the cache has the requested image
     * @param image_id
     * @return 
     */
    bool image_cache_has_image(uint8_t image_id);
    
    /**
     * Initialises the Image cache module
     */
    void image_cache_initialise(IMAGE_CACHE_INITIALISED_CALLBACK callback);
    
    /**
     * prepares the image cache to receive upgrade data
     */
    void image_cache_prepare_upgrade(uint8_t major_version, uint8_t minor_version, size_t size, uint8_t* checksum);
    
    /**
     */
    void image_cache_process();

    /**
     * 
     * @param request
     */
    bool image_cache_process_image(IMAGE_CACHE_CODED_IMAGE_HEADER* coded_image_header);
    
    /**
     * 
     * @param image_id
     */
    void image_cache_retrieve_image(uint8_t image_id,
                                    TFT_CONTROL* tft_control);
    
    /**
     * Takes upgrade data, decodes it and puts it into the image cache
     */
    void image_cache_unpack_upgrade_data(uint8_t* data, 
                                        size_t data_length,
                                        IMAGE_CACHE_UPGRADE_CALLBACK upgrade_callback);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
