/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _INPUT_CONTROL_DEFS_H    /* Guard against multiple inclusion */
#define _INPUT_CONTROL_DEFS_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
// a. Sample Rate, b. String representation, c. Index for get_sample_rate RS232 message    
// d. The sample rate as an integer
#define SAMPLE_RATES \
	X(0, No Signal, 0x08, 0)\
	X(32000, 32kHz, 0x00, 32000)\
    X(44100, 44.1kHz, 0x01, 44100)\
   	X(48000, 48kHz, 0x02, 48000)\
    X(88200, 88.2kHz, 0x03, 88200)\
   	X(96000, 96kHz, 0x04, 96000)\
	X(176400, 176.4kHz, 0x05, 176400)\
    X(192000, 192kHz, 0x06, 192000)\
    X(NOTSUPPORTED, Not Supported, 0x08, 0)\
    X(NOTVALID, Not Valid, 0x08, 0)\
    X(NOINPUT, No Input, 0x08, 0)\

// Now use the input data to define the enum
#define X(a, b, c, d) SAMPLE_RATE_##a,
	typedef enum {SAMPLE_RATES SAMPLE_RATE_MAX} SAMPLE_RATE;
#undef X
    
// Redefine the macro and use to create an array of int
#define X(a, b, c, d) d,
	static int SAMPLE_RATE_VALUES[] = { SAMPLE_RATES };
#undef X
    
// Redefine the macro to return an array of result values for the GETSTATUS message
#define X(a, b, c, d) c,
	static int SAMPLE_RATE_GETSTATUS_VALUES[] = { SAMPLE_RATES };
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a, b, c, d) #b,
	static char *SAMPLE_RATE_STRINGS[] = { SAMPLE_RATES };
#undef X

/*
 * These are used with the reset functions to make clearer that we are switching on/off the device
 */
#define INPUT_CONTROL_SWITCH_ON     true        // Taking out of Reset
#define INPUT_CONTROL_SWITCH_OFF    false       // Putting into Reset
    
typedef enum
{
    INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE,
    INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS,
    INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA,
} INPUT_CONTROL_I2C_REQUEST_STATE;

/**
 * Information on a DAC request
 * These may be used to set/request register values
 * Designed to work with the ES9038/9016
 * This is needed to store information during the async call
 */
typedef struct
{
    // NOTE: The first two members address/value MUST always be at the top of the structure
    //! The address of the register to be updated
    uint8_t address;
    
    //! The value to be written to the register
    uint8_t value;
    
    //! The ordinal index of the lowest bit that is significant
    uint8_t bit_low;
    
    //! The ordinal index of the highest bit that is significant    
    uint8_t bit_high;
    
    //! Set to true once it has been sent to the DAV and a success response has been received
    bool actioned;
    
    //! Pointer to a INPUT_CONTROL_I2C_REQUEST_CALLBACK that will be invoked when
    // The request has been processed
    void* callback;
    
    //! Reading data is a 2 phase process (send address/get data) this is used to identify
    // which part of the process we are in (used in the callback)
    INPUT_CONTROL_I2C_REQUEST_STATE read_state;
    
    //! Where the request is part of a multiple phase request (sample_rate) then 
    // this stores the index in the address)
    uint8_t index;
} INPUT_CONTROL_I2C_REQUEST;
    
/**
 * Callback function template
 */
typedef bool (*INPUT_CONTROL_I2C_REQUEST_CALLBACK)(INPUT_CONTROL_I2C_REQUEST* request);

    
    /**
     * Input Control TYpes
     */
#define INPUT_CONTROL_TYPES \
	X(NONE, None)\
	X(ANALOGUE, Analogue)\
	X(DIGITAL, Digital)\
	X(AVATAR, Avatar)\

// Now use the input data to define the enum
#define X(a, b) INPUT_CONTROL_TYPE_##a,
	typedef enum {INPUT_CONTROL_TYPES INPUT_CONTROL_TYPE_MAX} INPUT_CONTROL_TYPE;
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a, b) #b,
	static char *INPUT_CONTROL_TYPE_STRINGS[] = { INPUT_CONTROL_TYPES };
#undef X
    
typedef struct
{
    //! The index representing the control
    uint8_t index;

    //! A bitmap that indicates whether the switch should be on/off when selected
    uint8_t value;
    
    //! The type of input
    INPUT_CONTROL_TYPE type;
    
    //! True if this input uses Analogue Direct rather than routing through the Avatar
//    bool analogue_direct;
    
    //! Determines whether the input is selectable - used to reduce the number of 
    //  options if not being used.
    bool enabled;
    
} INPUT_CONTROL_DEFINITION;
    
/** 
 * Structure that can be used to define a range of bits in a specific register
 */
typedef struct
{
    //! The address of the register
    uint8_t address;

    //! The ordinal position of the lowest bit
    uint8_t bit_low;

    //! The ordinal position of the highest bit
    uint8_t bit_high;

} REGISTER_BITS_DEFINITION;
 
/**
 * Summary information used when previewing input selection - saves on 
 * making multiple calls.  Should be treated as read-only.
 */
typedef struct
{
    //! The preview is currently active
    bool active;
    
    // The input being previewed
    uint8_t input;
    
    // The type of input ANALOGUE/DIGITAL
    INPUT_CONTROL_TYPE type;
    
} INPUT_CONTROL_PREVIEW_INFO;
   
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Deactivate the preview mode
     */
    void input_control_deactivate_preview();

    /**
     * Gets a pointer to the current definition
     * @param index
     * @return 
     */
    INPUT_CONTROL_DEFINITION* input_control_get_current_definition();

    /**
     * Gets the type of the currently selected input (Analogue, Digital, Avatar etc..)
     * @return 
     */
    INPUT_CONTROL_TYPE input_control_get_current_input_type();

    /**
     * Gets a pointer to the supplied definition
     * @param index
     * @return 
     */
    INPUT_CONTROL_DEFINITION* input_control_get_definition(uint8_t index);

    /**
     * Gets a pointer to the new definition
     * @param index
     * @return 
     */
    INPUT_CONTROL_DEFINITION* input_control_get_new_definition();

    /**
     * Gets a pointer to the new definition
     * @param index
     * @return 
     */
    INPUT_CONTROL_DEFINITION* input_control_get_next_definition(uint8_t index, bool forwards);

    /**
     * Gets the index of the currently selected input
     * @return 
     */
    INPUT_CONTROL_PREVIEW_INFO* input_control_get_preview_info();

    /**
     * Gets a pointer to the previous definition
     * @param index
     * @return 
     */
    INPUT_CONTROL_DEFINITION* input_control_get_previous_definition();
 
    /**
     * Sets the current definition by index
     * @param index
     * @return 
     */
    void input_control_set_current_definition(uint8_t index);

    /**
     * Sets the new definition by index
     * @param index
     * @return 
     */
    void input_control_set_new_definition(uint8_t index);

    /**
     * Updates preview info with the supplied input
     * @param input - the input to be set as the preview input
     */
    void input_control_set_preview(uint8_t input);

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
