/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _INPUT_CONTROL_I2C_H    /* Guard against multiple inclusion */
#define _INPUT_CONTROL_I2C_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>
#include "utilities/i2c.h"
#include "utilities/input_control_defs.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
   
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Used to define an Input Control I2C setting
     * Note: This works at a register level rather than the individual settings
     */
    void input_control_i2c_add_setting(INPUT_CONTROL_I2C_REQUEST* setting, uint8_t address, uint8_t value, void* callback);

    /**
     * Starts the process of applying settings to the SPDIF Mux
     */
    void input_control_i2c_apply_setting(I2C_CLIENT_ID i2c_client_id, INPUT_CONTROL_I2C_REQUEST* setting);

    /**
     * Starts the process of applying settings to an Inpt Control settings array
     * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
     * @param setting - a pointer to the first setting in the array
     * @param count - the number of settings in the aray
     */
    void input_control_i2c_apply_settings(I2C_CLIENT_ID i2c_client_id, INPUT_CONTROL_I2C_REQUEST* setting, uint8_t count);

    /**
     * Generic handler for Input Control I2C request callbacks
     * @param event
     * @param context
     */
    void input_control_i2c_callback(DRV_I2C_BUFFER_EVENT event,
                                    I2C_REQUEST_TYPE request_type,
                                    uintptr_t context);
    
    /**
     * For use in I2C read callbacks.  This checks the request state, setting the state to 
     * INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA if it is INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS
     * (returning false) an returning true if state is INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA
     * The idea of this is to remove the need for a switch statement in the callback -
     * with this you just need an if (input_control_i2c_callback_has_data()) then 
     * handle the data.
     */
    bool input_control_i2c_callback_has_data(INPUT_CONTROL_I2C_REQUEST* request);
    
    /**
     * MUST be called at the end of the callback processing where input_control_i2c_callback_has_data
     * is in use..
     */
    bool input_control_i2c_callback_release(INPUT_CONTROL_I2C_REQUEST* request);
    
    /**
     * Defines an I2C register used for a specific task
     * @param address
     */
    void input_control_i2c_define_register(INPUT_CONTROL_I2C_REQUEST* register_request, 
                                           REGISTER_BITS_DEFINITION definition);

    /**
     * Searches the settings to see if we have one that matches the supplied address
     * @param definition - pointer to the first definition in the array
     * @param count - the number of definitions in the array
     * @param address - the address to match
     * @return 
     */
    INPUT_CONTROL_I2C_REQUEST* input_control_i2c_definition_for_address(INPUT_CONTROL_I2C_REQUEST* definitions, 
                                            uint8_t count, 
                                            uint8_t address);

    /**
     * Starts the process of reading a setting from the SPDIF Mux
     */
    void input_control_i2c_read_setting(I2C_CLIENT_ID i2c_client_id, 
                                            INPUT_CONTROL_I2C_REQUEST* setting);

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
