/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
 Provides logging functionality to the utility libraries

  @Description
 * The logger provides the ability for the application to provide a callback that 
 * is invoked when a logger message is called.  Ideally it will implement a number 
 * of logging levels that can be controlled
 * 
 * v2.1 - Added logger_override_async
 */
/* ************************************************************************** */

#ifndef _LOGGER_H    /* Guard against multiple inclusion */
#define _LOGGER_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    /**
     * LOGGER_ENABLED - should be defined in system_setup.h
     */
//#define LOGGER_ENABLED 1
    
    /**
     * Constant can be used when we don't want to or don't have a specific code
     */
#define LOGGER_CODE_UNDEFINED 0
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

    /**
     * A unique code used to define a particular message
     */
    typedef int LOGGER_CODE;

#define LOGGER_TYPE_ITEMS \
    X(NONE, "None")\
    X(ERROR, "Error")\
    X(WARNING, "Warning")\
    X(INFO, "Info")\
    X(TEST, "Test")\
    X(DEBUG, "Debug")\
    X(ALL, "All")\
    X(COMMAND, "Command")\
  
    //! Define a menu types enumeration
    typedef enum 
    {
#define X(type, title) LOGGER_TYPE_##type,
	LOGGER_TYPE_ITEMS
#undef X
    } LOGGER_TYPE;
    
    //! Declare a menu names array - This is passed to menu_initialise()
#define X(type, title) title,
	static char *LOGGER_TYPE_STRINGS[] = { LOGGER_TYPE_ITEMS };
#undef X
   
//    typedef enum
//    {
//        LOGGER_TYPE_NONE,
//        LOGGER_TYPE_ERROR,
//        LOGGER_TYPE_WARNING,
//        LOGGER_TYPE_INFO,
//        LOGGER_TYPE_TEST,
//        LOGGER_TYPE_DEBUG,
//        LOGGER_TYPE_ALL,
//    } LOGGER_TYPE;
    
    /**
     * Structure holding a Logger message , includes the text of the message,
     * message level
     */
    typedef struct
    {
        // A unique code for the message
        LOGGER_CODE code;
        
        // The type log message - set by the logger functions
        LOGGER_TYPE type;
        
        // The message text
        char* message;
        
        //! The length of the message (optional))
        size_t length;
        
    } LOGGER_MESSAGE;
    
    /**
     * Callback function template
     */
    typedef void (*LOGGER_CALLBACK)(LOGGER_MESSAGE* message);
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Returns the level of logging
     * e.g LOGGER_TYPE_WARNING include LOGGER_TYPE_WARNING, and LOGGER_TYPE_ERROR
     */
    LOGGER_TYPE logger_get_level();

    /**
     * Initialises the Logger
     */
    void logger_initialise();
    
    /**
     * Used to generate a RS232 command message - note callback needs to handle LOGGER_TYPE_COMMAND
     * and generate the message.
     * This has been added so that unsolicited command messages can be generated anywhere in 
     * the code, though in general they will be a fallout of the standard menu edit commands
     * which send a response. Note these messages are suppressed during the app_commands_process_message
     * function to try and avoid duplication.
     * @param command - The associated command command code from app_config.g APP_COMMANDS
     * @param data - The data to be put into the message
     * @param data_length - The length of the data
     */
    void logger_log_command(unsigned char command, unsigned char* data, size_t data_length);
    
    /**
     * Log an debug message
     * @param message
     */
    inline void logger_log_debug(LOGGER_CODE code, char* message);

    /**
    * Formatted logger debug message
    * @param code The logger code to be sent with the message
    * @param fmt - the sprintf like format string
    * @param ... A list of variables used to populate the string;
    */
    void logger_log_debugf(LOGGER_CODE code, char* fmt, ...);

    /**
     * Log an error message
     * @param message
     */
    inline void logger_log_error(LOGGER_CODE code, char* message);

    /**
    * Formatted logger error message
    * @param code The logger code to be sent with the message
    * @param fmt - the sprintf like format string
    * @param ... A list of variables used to populate the string;
    */
    inline void logger_log_errorf(LOGGER_CODE code, char* fmt, ...);
        
    /**
     * Log an information message
     * @param message
     */
    inline void logger_log_info(LOGGER_CODE code, char* message);

    /**
    * Formatted logger info message
    * @param code The logger code to be sent with the message
    * @param fmt - the sprintf like format string
    * @param ... A list of variables used to populate the string;
    */
    inline void logger_log_infof(LOGGER_CODE code, char* fmt, ...);
   
    /**
     * Used for debug purposes
     */
    void logger_log_nothing();
    
    /**
    * Formatted logger test message
    * @param code The logger code to be sent with the message
    * @param fmt - the sprintf like format string
    * @param ... A list of variables used to populate the string;
    */
    inline void logger_log_testf(LOGGER_CODE code, char* fmt, ...);
        
    /**
     * Log a warning message
     * @param message
     */
    inline void logger_log_warning(LOGGER_CODE code, char* message);

    /**
    * Formatted logger info message
    * @param code The logger code to be sent with the message
    * @param fmt - the sprintf like format string
    * @param ... A list of variables used to populate the string;
    */
    inline void logger_log_warningf(LOGGER_CODE code, char* fmt, ...);

    /**
     * Overrides async mode for logger_log_debug
     * @param override
     */
    void logger_override_async(bool override);
    
    /**
     * Processes the logger queue when LOGGER_PROCESS_ASYNC is defined
     */
    void logger_process();
   
    /**
     * Sets the application callback for logger messages
     * @param callback
     */
    void logger_set_callback(LOGGER_CALLBACK callback);

    /**
     * Sets the level of logging
     * e.g LOGGER_TYPE_WARNING include LOGGER_TYPE_WARNING, and LOGGER_TYPE_ERROR
     * @param level The maximum logging level
     */
    void logger_set_level(LOGGER_TYPE level);
    
    /**
     * Temporarily suppresses logging.
     * Note: Unsetting assumes that a set has been called before
     * Note: If nested the first call will win
     * @param suppress
     */
    void logger_suppress_logging(bool suppress);
    
    /**
     * Suppresses the logging of Command logs messages
     * @param suppress
     */
    void logger_suppress_command_logging(bool suppress);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _LOGGER_H */

/* *****************************************************************************
 End of File
 */
