/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    power.h

  @Summary
    Manages simple power control functions.

  @Description
 The ports used for power control must be defined in ports_setup.h
 */
/* ************************************************************************** */

#ifndef _POWER_H    /* Guard against multiple inclusion */
#define _POWER_H

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_config.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif

    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

#define POWER_AUTO_TIMEOUT_INCREMENT    30  // minutes
#define POWER_AUTO_TIMEOUT_MAX          POWER_AUTO_TIMEOUT_INCREMENT * 8   // 4 hours
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

// Now use the input data to define the enum
#define X(a, b, c, d, e) POWER_INDEX_##a,
	typedef enum {POWER_ITEMS POWER_INDEX_MAX} POWER_INDEX;
#undef X

    typedef enum 
    {
        POWER_STATE_OFF,
        POWER_STATE_ON,
    } POWER_STATE;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * Disables PSU's based on their Flags
     */
    void power_disable();

    /**
     * Disables an individual PSU by Index
     */
    void power_disable_by_index(POWER_INDEX index);
 
    /**
     * Enables both the PSU and Digital PSU
     */
    void power_enable();

    /**
     * Enables an individual PSU by Index
     */
    void power_enable_by_index(POWER_INDEX index);   
    
    /**
     * Used to determine whether AC is present
     * @return 
     */
    bool power_get_ac_present();
    
    /**
     * Gets the auto shutdown setting
     * @return 
     */
    bool power_get_auto_shutdown();
     
    /**
     * Sets the auto shutdown timeout setting (in mins)
     * @return 
     */
    uint16_t power_get_auto_shutdown_timeout();

    /**
     * Gets the On/Off state of network standby
     * @return 
     */
    bool power_get_network_active_in_standby();    

    /**
     * Gets the On/Off stane of rs232 standby
     * @return 
     */
    bool power_get_rs232_active_in_standby();    
    
    /**
     * Returns the current Power state
     * @return 
     */
    POWER_STATE power_get_state();
    
    /**
     * Initialises power options from settings
     */
    void power_initialise(bool ac_detect_pin_high);
    
    /**
     * Resets the auto shutdown timer - used by heartbeat
     */
    void power_reset_auto_shutdown_timer();
    
    /**
     * Gets the auto shutdown setting
     * @return 
     */
    void power_set_auto_shutdown_timeout(uint16_t timeout);
 
    /**
     * Sets the On/Off state of network standby.  So if true then standby mode will
     * be ACTIVE, otherwise it will be IDLE unless the RS232 option allows SLEEP
     * @param on_off - true - On , false - off 
     */
    void power_set_network_active_in_standby(bool on_off);    

    /**
     * Sets the On/Off state of rs232 standby
     * @param on_off - true - On , false - off 
     */
    void power_set_rs232_active_in_standby(bool on_off);    
    
    /**
     * Clears the soft start enable switch - needs to be called 0.5s+ after 
     * power is applied
     */
    void power_soft_start_clear();
    
    /**
     * Enables the soft start pin
     */
    void power_soft_start_enable();
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
