/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _SPDIF_RX_DEFS_H    /* Guard against multiple inclusion */
#define _SPDIF_RX_DEFS_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stddef.h>
#include "utilities/input_control_defs.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Constants                                                         */
/* ************************************************************************** */
/* ************************************************************************** */


// *****************************************************************************
// *****************************************************************************
// Section: Data Types
// *****************************************************************************
// *****************************************************************************

//------------------------------------------------------------------------------
// ADC Mode enum
//------------------------------------------------------------------------------
typedef enum
{
    SPDIF_RX_ADC_MODE_POWER_ON_MASTER,
    SPDIF_RX_ADC_MODE_POWER_OFF_MASTER,
    SPDIF_RX_ADC_MODE_POWER_ON_SLAVE,
    SPDIF_RX_ADC_MODE_POWER_OFF_SLAVE,
} SPDIF_RX_ADC_MODE;

//------------------------------------------------------------------------------
// SPDIF RX Hardware Events
// Note: These are implemented by bot the spdif_rx and spdif_rx_hw levels
//------------------------------------------------------------------------------

/**
 * a: Enum name b: Event Module Description
 */
#define SPDIF_RX_EVENT_ITEMS \
    X(ERROR, SPDIF Rx Error)\
    X(SAMPLE_RATE, SPDIF Rx Sample Rate)\
    X(SWITCH, SPDIF Rx Switch)\

//! Uses EVENT_MODULE_ITEMS defined in app_config.h
// Define enum
#define X(a, b) SPDIF_RX_EVENT_##a,
	typedef enum {SPDIF_RX_EVENT_ITEMS} SPDIF_RX_EVENT;
#undef X
    
// Create array of Event module names
#define X(a, b) #b,
	static char *SPDIF_RX_EVENT_NAMES[] = { SPDIF_RX_EVENT_ITEMS };
#undef X

//! Structure for passing info on SPDTFC Rx errors
typedef struct
{
    //! Indicates that we identified the error
    bool error_identified;
    
    //! Indicates that there has ben a sample rate change
    bool sample_rate_change;
    
    //! Indicats whether the PLL is locked 
    bool pll_locked;
} SPDIF_RX_HW_ERROR_EVENT_DATA;
       
//! Structure for passing Sample rate event data
typedef struct
{
    //! The sample rate returned
    SAMPLE_RATE sample_rate;
    
    //! Indicates that we identified the error
    bool error_identified;
    
    // It appears that the SPDIF RX returns a NOT SUPPORTED sample rate when 
    // queried just after a switch, so this will be marked as not valid.  The input_contol
    // module still mutes/unmutes but if false it will not inform the avatar
    bool is_valid;
} SPDIF_RX_SAMPLE_RATE_EVENT_DATA;
       
//! Structure for passing Switch event data
typedef struct
{
    // The register value of he new input
    uint8_t input;
} SPDIF_RX_SWITCH_EVENT_DATA;
        
//------------------------------------------------------------------------------
// SPDIF RX Hardware Functions
//------------------------------------------------------------------------------

/**
 * Template for the ADC Mode function - used by the hardware layer
 */
typedef void (*SPDIF_RX_ADC_MODE_FUNCTION)(SPDIF_RX_ADC_MODE mode);     

/**
 * Template for the configure function - used by the hardware layer
 */
typedef void (*SPDIF_RX_CONFIGURE_FUNCTION)();     

/**
 * Template for the initialise function - used by the hardware layer
 */
typedef bool (*SPDIF_RX_INITIALISE_FUNCTION)();     

/**
 * Template for the initialise function - used by the hardware layer
 */
typedef void (*SPDIF_RX_INTERRUPT_FUNCTION)(uint8_t channel);     

/**
 * Template for the process function - used by the hardware layer
 */
typedef void (*SPDIF_RX_PROCESS_FUNCTION)();     

/**
 * Template for the reset function - used by the hardware layer
 */
typedef void (*SPDIF_RX_RESET_FUNCTION)(bool on);     

/**
 * Template for the refresh sample rate function - used by the hardware layer
 * Return true if sample rate request has been made
 */
typedef bool (*SPDIF_RX_SAMPLE_RATE_FUNCTION)();     

/**
 * Template for the switch function - used by the hardware layer
 */
typedef void (*SPDIF_RX_SWITCH_FUNCTION)(INPUT_CONTROL_DEFINITION* definition);    

//! Struct used by the application to decide which SPDIF_RX to use
// This is defined by th spdif_rx_hw_nnn_define_callbacks method which is called in 
// app_initialise depends on which spdif_rx reciever is being used.
typedef struct
{
    //! ADC Mode function
    SPDIF_RX_ADC_MODE_FUNCTION adc_mode_function;
    
    //! Used to configure the spdif_rx module
    SPDIF_RX_CONFIGURE_FUNCTION configure_function;
    
    //! handles the module initialisation
    SPDIF_RX_INITIALISE_FUNCTION initialise_function;
    
    //! Handles interrupt processing
    SPDIF_RX_INTERRUPT_FUNCTION interrupt_function;
    
    //! Used for regular processing.  Note this does not need a check for whether
    // the input is digital or the POLL timer has expired
    SPDIF_RX_PROCESS_FUNCTION process_function;
    
    //! Used to put the spdif_rx in or out of reset
    SPDIF_RX_RESET_FUNCTION reset_function;
    
    //! Used to trigger a sample rate calculation
    SPDIF_RX_SAMPLE_RATE_FUNCTION sample_rate_function;
    
    //! Used to switch inputs
    SPDIF_RX_SWITCH_FUNCTION switch_function;
    
} SPDIF_RX_FUNCTION_DATA;

// *****************************************************************************
// *****************************************************************************
// Section: Interface Functions
// *****************************************************************************
// *****************************************************************************


    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
