/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    stateful_process.h

  @Summary
    Used to managed a multi state process

  @Description
 Currently handles a single process but could be extended to handle multiple 
 * simulataneous processes if required
 */
/* ************************************************************************** */

#ifndef _STATEFUL_PROCESS_H    /* Guard against multiple inclusion */
#define _STATEFUL_PROCESS_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stdint.h>

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
    /**
     * A structure that can be used to maintain state during multiple calls to
     * a process invoked by a menu item (see menu_add_process_function())
     */
    typedef struct
    {
        // A placeholder for state - initialised to 0 on first call
        uint8_t state;
        
        //! In general a reference to the function that was invoked at the start
        // of the process - this was the process_function associated with the menu 
        // item at the time.  It is possible to change the function if necessary 
        // during processing
        void (*process_function)();
        
        //! function called when stateful_process_end is called
        void (*end_function)();

    } PROCESS_INFO;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Invokes the function to be called at the end of the process - should be called in 
     * the code that implements a process
     */
    void stateful_process_end();
    
    /**
     * Gets a pointer to a common processing info block that can be used to store 
     * state details when in APP_CONTEXT_MODE_PROCESS mode
     * Note: Originally this was put onto individual menu items which had the problem
     * that when the menu was changed during processing the process info needed copying
     * It may be necessary to go back to that model if ever we need to run two 
     * processes simultaneously
     */
    PROCESS_INFO* stateful_process_get_info();
    
    /**
     * Initialises the process info for the start of a process
     * @param process_function
     */
    void stateful_process_initialise_info(void (*process_function)(), void (*end_function)());

    /**
     * Invokes the process function
     */
    void stateful_process_invoke();
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
