/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    structured_message.h

  @Summary
    Header for the common structured message used to handle rs232, IR and Ethernet.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _STRUCTURED_MESSAGE_H    /* Guard against multiple inclusion */
#define _STRUCTURED_MESSAGE_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <string.h>
#include <stdlib.h>
#include "app_config.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Constants                                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! The maximum length for serial data in a message
#define STRUCTURED_MESSAGE_MAX_DATA             255

//! The maximum buffer size required for a message
#define STRUCTURED_MESSAGE_MAX_SIZE             STRUCTURED_MESSAGE_MAX_DATA + 6
    
//! The command id for an undefined message (note - this needs tying up with 
//  the commands in app_config.h    
#define STRUCTURED_MESSAGE_COMMAND_UNDEFINED    0xFF
    
#define STRUCTURED_MESSAGE_COMMAND_AMX_DUET     APP_COMMANDS_AMX_DUET    
    
#define STRUCTURED_MESSAGE_AMX_DUET_RESPONSE    "AMXB<Device-SDKClass=Amplifier><Device-Make=%s><Device-Model=%s><Device-Revision=%d.%d.0>\r"    

// *****************************************************************************
// *****************************************************************************
// Section: Data Types
// *****************************************************************************
// *****************************************************************************

    typedef enum
    {
        STRUCTURED_MESSAGE_SOURCE_RS232,
                
        // The message was generated from the remote before being added to 
        // the input context
        STRUCTURED_MESSAGE_SOURCE_REMOTE_RAW,
        STRUCTURED_MESSAGE_SOURCE_TCPIP,
        // Can be used to check whether the message originated externally
        STRUCTURED_MESSAGE_SOURCE_EXTERNAL_MAX,
                
        // The message was generated by input processing
        STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT,  // The input hasn't yet been defined
        STRUCTURED_MESSAGE_SOURCE_BUTTONS,        
        STRUCTURED_MESSAGE_SOURCE_ENCODER,
        STRUCTURED_MESSAGE_SOURCE_REMOTE,
        STRUCTURED_MESSAGE_SOURCE_BALANCE_POT,
                
        // The message has been generated by a process
        STRUCTURED_MESSAGE_SOURCE_PROCESS,
    } STRUCTURED_MESSAGE_SOURCE;

    /** 
     * Messages are categorised as performing certain types of action.  In general
     * this involves changing numeric values either with an absolute value or a
     * incremental delta, setting specific data (i.e. names), or requesting the 
     * current status
     */
    typedef enum
    {
        STRUCTURED_MESSAGE_ACTION_UNDEFINED,
        STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE,     // Indicates that the value will set the the value
        STRUCTURED_MESSAGE_ACTION_SET_INCREMENT,    // Indicates that the value will be incremented by 1
        STRUCTURED_MESSAGE_ACTION_SET_DECREMENT,    // Indicates that the value will be decremented by 1
        STRUCTURED_MESSAGE_ACTION_SET_DATA,         // Indicates that the message contains data
        STRUCTURED_MESSAGE_ACTION_GET_STATUS,       // Indicates that the message is retrieving the current status
    } STRUCTURED_MESSAGE_ACTION;

    /**
     * Enumeration for the Structured message zone
     */
    typedef enum
    {
        STRUCTURED_MESSAGE_ZONE_1   = 0x01,
        STRUCTURED_MESSAGE_ZONE_2   = 0x02,
    } STRUCTURED_MESSAGE_ZONE;

    /**
     * The states a message can be in while decoding
     */
    typedef enum
    {
        STATUS_UPDATE           = 0x00,
        ZONE_INVALID            = 0x82,
        COMMAND_NOT_RECOGNISED  = 0x83,
        PARAM_NOT_RECOGNISED    = 0x84,
        COMMAND_INVALID_NOW     = 0x85,
        INVALID_DATA_LENGTH     = 0x86,
    }  STRUCTURED_MESSAGE_RESPONSE;

    /**
     * Describes the states when parsing incoming messages
     */
    typedef enum
    {
        //! Indicates that the message has been initialised.  This may be the 
        //  final state where messages are not processed through process buffer
        STRUCTURED_MESSAGE_STATE_INITIALISED,
//        STRUCTURED_MESSAGE_STATE_WAIT_FOR_START,
        STRUCTURED_MESSAGE_STATE_WAIT_FOR_ZONE,
        STRUCTURED_MESSAGE_STATE_WAIT_FOR_CMD,
        STRUCTURED_MESSAGE_STATE_WAIT_FOR_LENGTH,
        STRUCTURED_MESSAGE_STATE_WAIT_DATA,
        STRUCTURED_MESSAGE_STATE_WAIT_FOR_END,

        // The message has been completely received and is ready for processing
        STRUCTURED_MESSAGE_STATE_PERFORM_ACTION,

        // A reply needs sending to the sender
        STRUCTURED_MESSAGE_STATE_SEND_REPLY,

        STRUCTURED_MESSAGE_STATE_WAIT_FOR_M,
        STRUCTURED_MESSAGE_STATE_WAIT_FOR_X,
        STRUCTURED_MESSAGE_STATE_WAIT_FOR_CR,

    } STRUCTURED_MESSAGE_STATE;

    /**
     * Structure encapsulating a standard message that we transmit/receive via RS232 
     * or Ethernet
     * STX|Zone|Command|Data Length|Data|ET
     * or
     * AMX\r
     */
    typedef struct
    {
        //IMPORTANT - the first 4 data items MUST be the first in the structure
        // as a reference to the structure is used to read from various streams
        // The message zone Zone1|Zone2
        STRUCTURED_MESSAGE_ZONE zone;

        // The command to be executed
        unsigned char command;

        // A response sent back to client to indicate the state of the message 
        // (in terms of how valid it is)
        STRUCTURED_MESSAGE_RESPONSE response;

        // The length of the following data
        unsigned char data_length;

        // Message specific data
        unsigned char data[STRUCTURED_MESSAGE_MAX_DATA];

        // Identifies message as only requiring the data to be sent, zone,command etc can be ignored
        bool data_only;
        
        // Determines whether the message is a repeat message (used with IR))
        bool repeat;

        // Identifies the source of the message
        STRUCTURED_MESSAGE_SOURCE source;
        
        // The type of action expected from the message
        STRUCTURED_MESSAGE_ACTION action;
        
        // Indicates whether the message represents system activity.  This defaults 
        // to true but can be set to false if we do not want the message to be 
        // regarded as being actively using the system (see auto timeout querying)
        bool is_activity;

        //! Internal variables - should not be touched by code outside structured_message.c
        // The state of the message (used by process_buffer to determine how far through
        // the message construction is)
        STRUCTURED_MESSAGE_STATE processing_state;
                        
        // Maintains an index of the current write position for the data
        // Internal use only
        unsigned char processing_data_index;
        
        //! A buffer of data that can be used to add extra info about the message 
        // used by TCPIP to identify the socket that the message was received on 
        // Manage completely by the specific code
        unsigned char state_data[4];
    } STRUCTURED_MESSAGE; 
    
    /**
     * Callback function template
     */
    typedef void (*STRUCTURED_MESSAGE_CALLBACK)(STRUCTURED_MESSAGE* message);

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
    * Initialises a structured message
    * @param message
    */
   void structured_message_initialise(STRUCTURED_MESSAGE* message, STRUCTURED_MESSAGE_SOURCE source);

   /**
    * Copies the details from one structured message into another
    * @param from - the message to copy
    * @param to - the message whose data will be overwritten by the first message
    */
   void structured_message_copy(STRUCTURED_MESSAGE* from, STRUCTURED_MESSAGE* to);
   
   /**
    * Returns the length of the structured message including headers, commands, footer
    */
   int structured_message_get_length(STRUCTURED_MESSAGE* message);

    /**
     * Can be used to determine whether the structured message is currently being populated
     * @param message
     * @return 
     */
    bool structured_message_is_processing(STRUCTURED_MESSAGE* message);
   
   /**
    * Process a buffer of data - used to construct structured messages from streams of data
    * This takes a structured message a and depending on its current state adds data 
    * into the message.  It is intended for use with controllers that produce streams of data
    * in the standard format for our internal messaging structure.
    * @param message   - The message the data is to be added to 
    * @param buffer - The message that the data is added to 
    * @param size  - The size of the buffer (amount of data to be processed)
    */
   void structured_message_process_buffer(STRUCTURED_MESSAGE* message, unsigned char* buffer, size_t size);

   /**
    * Process a byte of data read from the RX channel
    * @param data_byte
    * @return true if the byte was processed
    */
   bool structured_message_process_byte(STRUCTURED_MESSAGE* message, unsigned char data_byte);
    
    /**
     * Set the data on a structured message (saves memcpys etc...)
     */
    void structured_message_set_message_data(STRUCTURED_MESSAGE* message, char* data, size_t data_length);

   /**
    * Fills the supplied buffer with a structured message ready for transmission 
    * (including STX and ET start/stop bytes).
    * Note: The buffer MUST be sufficiently large enough to hold the length of the data + 6
    * @param message - the message to be written
    * @param buffer - the buffer to be written to
    * @param size - the actual size of the buffer
    * @return - the number of bytes written to the buffer
    */
   int structured_message_write_response_buffer(STRUCTURED_MESSAGE* message, unsigned char* buffer, size_t size);

   /**
    * Fills the supplied buffer with a structured message ready for transmission 
    * (including STX and ET start/stop bytes).
    * Note: The buffer MUST be sufficiently large enough to hold the length of the data + 6
    * @param message - the message to be written
    * @param buffer - the buffer to be written to
    * @param size - the actual size of the buffer
    * @return - the number of bytes written to the buffer
    */
   int structured_message_write_request_buffer(STRUCTURED_MESSAGE* message, unsigned char* buffer, size_t size);

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
