/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    tcpip.h

  @Summary
    Handles tcpip functionality.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _TCPIP_H    /* Guard against multiple inclusion */
#define _TCPIP_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <string.h>
#include "tcpip/tcpip.h"

#include "utilities/structured_message.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */
 
//! Maximum name length for the network friendly name
#ifndef TCPIP_NETWORK_NAME_MAX_LEN
#define TCPIP_NETWORK_NAME_MAX_LEN  12 // Includes \0
#endif
    
#define TCPIP_SSID_MAX_LEN  80 // Includes \0

//! A pseudo index used to identify that the TCPIP message to be sent is unsolicited
//  the app can then decide whether to send to all clients, or we can implement 
// a method allowing a TCP client to elect whether to receive unsolicited messages
#define TCPIP_UNSOLICITED_CLIENT_CONNECTION     0xFF

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************

     /**
      * Connection Callback function template
      * Called when a network connection is established or dropped
      * The address is passed if known, otherwise NULL
     */
    typedef void (*TCP_CONNECTED_CALLBACK)(bool connected, struct sockaddr_in* address, int* address_len);

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif
    
    /**
     * Close all open TCPIP connections
     */
    void tcpip_close();

    /** 
     * Performs steps to connect TCP/IP to a network
     */
    void tcpip_connect();
    
    /**
     * Returns whether DHCP is enabled or not
     * @param enable
     */
    bool tcpip_get_dhcp_enabled();
    
    /**
     * Gets a displayable version if the IP Address
     * @param buffer
     */
    void tcpip_get_ip_address_string(char* buffer, size_t size);
    
    /**
     * Returns a the current Ethernet MAC address
     * @return 
     */
    TCPIP_MAC_ADDR tcpip_get_mac_address();
    
    /**
     * Returns a string representation of the current Ethernet MAC address
     * @return 
     */
    void tcpip_get_mac_address_string(char* buffer, size_t size);  
    
    /**
     * Returns a string representation of the current friendly name
     * @return 
     */
    void tcpip_get_network_name_string(char* buffer, size_t size);
    
    /**
     * Gets the runtime IP Address
     */
    IP_ADDR tcpip_get_runtime_ip_address();
    
    /**
     * Returns a string representation of the current SSID
     * @return 
     */
    void tcpip_get_ssid_string(char* buffer, size_t size);

    /**
     * Gets the static IP Address
     */
    IP_ADDR tcpip_get_static_ip_address();
    
    /**
     * Returns a the current WiFi MAC address
     * @return 
     */
    TCPIP_MAC_ADDR tcpip_get_wifi_mac_address();
    
    /**
     * Returns a string representation of the current WiFi MAC address
     * @return 
     */
    void tcpip_get_wifi_mac_address_string(char* buffer, size_t size);  
    
    /**
     * Initialise the tcpip module
     * @tcpip_module - set this to be system_definitions.h sysObj.tcpip
     */
    void tcpip_initialise(SYS_MODULE_OBJ* tcpip_module, 
            STRUCTURED_MESSAGE_CALLBACK callback,
            TCP_CONNECTED_CALLBACK connected_callback);

    /** 
     * Determines whether TCP/IP is connected to a network
     */
    bool tcpip_is_connected();
    
    /**
     * Perform TCP/IP processing
     */
    void tcpip_process();

    /*
     * Reviews the TCPIP connection process.
     * Used to check whether DHCP/StaticIP settings have changed and to initiate a 
     * new connection
     */
    void tcpip_review_connection();

    /**
     * Send a bufer of data to the specified client
     * @param client_idx - The index of the client socket
     * @param buffer - A pointer to the buffer of data to be sent
     * @param len - The length of the buffer
     */
    void tcpip_send_buffer(uint8_t client_idx, char* buffer, int len);
    
    /**
     * Send a structured message over TCPIP
     * @param message
     */
    void tcpip_send_structured_message(STRUCTURED_MESSAGE* message); 
    
    /**
     * Control whether DHCP is enabled or not
     * @param enable
     */
    void tcpip_set_dhcp_enabled(bool enable);
 
    /**
     * Sets the network name for the device
     * @param enable
     */
    void tcpip_set_network_name(char* name, size_t size);
 
    /**
     * Sets the network name for the device
     * @param enable
     */
    void tcpip_set_ssid(char* ssid, size_t size);

    /**
     * Sets the static IP Address
     * Note: does not attempt a reconnect if changed..This needs doing in other code?
     */
    void tcpip_set_static_ip_address(IP_ADDR* ip_address);

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
