/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _UPGRADE_METADATA_H    /* Guard against multiple inclusion */
#define _UPGRADE_METADATA_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include "utilities/flash_memory.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Constants                                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define UPGRADE_METADATA_STATE_OPTIONS \
    X(NONE)\
    X(AVAILABLE_INSTALL)\
    X(IN_PROGRESS)\
    X(IN_ROLLBACK)\
    X(UPLOAD_ERRORS)\

// Now use the input data to define the enum
#define X(a) UPGRADE_METADATA_STATE_##a,
	typedef enum {UPGRADE_METADATA_STATE_OPTIONS} UPGRADE_METADATA_STATE;
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a) #a,
	static char *UPGRADE_METADATA_STATE_STRINGS[] = {UPGRADE_METADATA_STATE_OPTIONS};
#undef X

///**
// * Upgrade State
// */
//typedef enum
//{
//    //! No upgrade available at the moment
//    UPGRADE_STATE_NONE,
//            
//    //! Upgrade available but do not automatically install
//    //UPGRADE_STATE_AVAILABLE_NO_INSTALL,
//            
//    //! Upgrade available - automatically install at next boot
//    UPGRADE_STATE_AVAILABLE_INSTALL,
//            
//    //! Upgrade is in progress
//    UPGRADE_STATE_IN_PROGRESS,
//} UPGRADE_STATE;

//! An identifier for individual upgrades
typedef uint8_t UPGRADE_METADATA_LOAD_ID;

/**
 * Enum representing statues during image upload into flash
 */
typedef enum
{
    UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS  = 0X0,
    UPGRADE_METADATA_UPLOAD_STATUS_START_INVALID = 0X1,             // Start conditions invalid
    UPGRADE_METADATA_UPLOAD_STATUS_LOAD_ID_MISMATCH = 0x2,
    UPGRADE_METADATA_UPLOAD_STATUS_CHECKSUM_MISMATCH  = 0X3,
            
} UPGRADE_METADATA_UPLOAD_STATUS;
    
/**
 * Upgrade status enum
 */
typedef enum
{
    UPGRADE_METADATA_UPGRADE_STATUS_IN_PROGRESS,
    UPGRADE_METADATA_UPGRADE_STATUS_OK,
    UPGRADE_METADATA_UPGRADE_STATUS_ERROR,
    UPGRADE_METADATA_UPGRADE_STATUS_RETRY,
} UPGRADE_METADATA_UPGRADE_STATUS;
   
/**
 * flash_memory_read_page callback function template
 */
typedef void (*UPGRADE_METADATA_UPLOAD_CALLBACK)(UPGRADE_METADATA_LOAD_ID load_id, FLASH_MEMORY_CLIENT_REQUEST* request);

/**
 * A callback in addition to the one above which is supplied in the upgrade_metadata_upload_data
 * call and set to NULL once processed
 */
typedef void (*UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK)(UPGRADE_METADATA_LOAD_ID load_id, uint32_t total_size, uint32_t bytes_written, UPGRADE_METADATA_UPGRADE_STATUS status);

/**
 * Upgrade states enum
 */
typedef enum
{
    //! The upgrade is erasing the flash
    UPGRADE_METADATA_UPGRADE_PROGRESS_STATE_ERASING,
    //! The upgrade is writing program memory
    UPGRADE_METADATA_UPGRADE_PROGRESS_STATE_WRITING,
} UPGRADE_METADATA_UPGRADE_PROGRESS_STATE;

/**
 * structure passed to UPGRADE_METADATA_UPGRADE_CALLBACK
 */
typedef struct
{
    UPGRADE_METADATA_LOAD_ID load_id;
    UPGRADE_METADATA_UPGRADE_PROGRESS_STATE state;
    uint32_t bytes_written;
    uint32_t size;
    uint8_t version_major;
    uint8_t version_minor;
} UPGRADE_METADATA_UPGRADE_PROGRESS;

//! callback template to monitor upgrade progress
typedef void (*UPGRADE_METADATA_UPGRADE_CALLBACK)(UPGRADE_METADATA_LOAD_ID load_id, UPGRADE_METADATA_UPGRADE_PROGRESS* progress);

// *****************************************************************************
// *****************************************************************************
// Section: Interface Functions
// *****************************************************************************
// *****************************************************************************

/**
 * Called after a successful upgrade. Sets the metadata up for a successful upgrade
 */
void upgrade_metadata_complete();

/**
 * Called after image data loaded into flash.  Update the upgrade_idx, sets state to loaded
 */
void upgrade_metadata_complete_load();

/**
 * Get the current upgrade state
 * @return 
 */
UPGRADE_METADATA_STATE upgrade_metadata_get_upgrade_state();

/**
 * Initialise the internal data structures.  Assumes the pointer to the metadata
 * comes from data ready from flash memory
 * @param write_callback - A callback invoked when image data is written
 */    
void upgrade_metadata_initialise(UPGRADE_METADATA_UPLOAD_CALLBACK upload_callback,
                                UPGRADE_METADATA_UPGRADE_CALLBACK upgrade_callback);

/**
 * Test whether the meta data has been loaded
 * @return 
 */
bool upgrade_metadata_is_loaded();

/**
 * Returns true if an upgrade is required.
 * @return 
 */    
bool upgrade_metadata_needs_upgrade();

/**
 * Performs an upgrade using the most recent available upload data
 * Note: UPGRADE_METADATA_INCLUDE_UPGRADE_FUNCTIONALITY must be defined and
 * upgrade.h/c included in the project for this functionality to work.
 * @return 
 */
UPGRADE_METADATA_UPGRADE_STATUS upgrade_metadata_perform_upgrade();

/**
 * Called before loading image data into flash.  Update the upgrade_idx, sets state to loading
 * @param major_version
 * @param minor_version
 * @param size
 * @param checksum - Pointer to a 16byte array holding the checksum
 * @param override
 * @return 
 */
UPGRADE_METADATA_LOAD_ID upgrade_metadata_prepare_upload(uint8_t major_version, uint8_t minor_version, size_t size, uint8_t* checksum, bool override);

/**
 * Called after a failed upgrade. Sets up the metadata to attempt a restore of the current version
 */
bool upgrade_metadata_prepare_rollback();

/**
 * Prints the metadata as a series of logger_log_info messages
 */
void upgrade_metadata_print_metadata();

/**
 * Uploads image data into the storage
 * @param load_id
 * @param data
 * @param data_length
 * @param local_callback - pointer to a function to be called when the write associated
 * with this request completes
 */
UPGRADE_METADATA_UPLOAD_STATUS upgrade_metadata_upload_data(UPGRADE_METADATA_LOAD_ID load_id, 
        uint8_t* data, 
        size_t data_length,
        UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback);

/**
 * Completes the Uploads image data into the storage
 * @param load_id
 * @param data
 * @param data_length
 */
UPGRADE_METADATA_UPLOAD_STATUS upgrade_metadata_upload_end(UPGRADE_METADATA_LOAD_ID load_id,
        UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback, UPGRADE_METADATA_UPGRADE_STATUS status);

/**
 * Gets a pointer to the IMAGE_METADATA to be used for the upgrade
 * @return 
 */
//IMAGE_METADATA* upgrade_metadata_upgrade_info();

    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
