/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    adc.c

  @Summary
    Public Interface to the adc module

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/adc/adc.h"
#include "app_config.h"
#include "utilities/adc/private/adc_data.h"
#include "utilities/adc/adc_defs.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */



/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Required to allocate an input for the ADC module.  This should be called once for
 * each input.
 */
ADC_INPUT_DEFINITION* adc_add_input(ADC_INPUT_ID input)
{
    ADC_INPUT_DEFINITION* def = adc_data_allocate_input(input);
}

/**-----------------------------------------------------------------------------
 * Gets the raw value of the supplied input.
 * @param input_index - The index of the input to be reported on
 * @return - The raw value read from the pin
 * it returns 0
 */
uint32_t adc_get_raw_value(ADC_INPUT_ID input)
{
    uint32_t result = ADC_INVALID_REQUEST;
   
    if (input < ADC_INPUT_COUNT)
    {
        ADC_INPUT_DEFINITION* def = adc_data_get_input_definition(input);
        result = def->current_value; 
        logger_log_debugf(LOGGER_CODE_DEBUG, "ADC: %d --", def->current_value);
    }
    
    return result;    
}

/**-----------------------------------------------------------------------------
 *  Initialises the adc module
 * @return 
 */
void adc_initialise()
{
    adc_data_initialise();
    
    // Opens the static ADC Driver
    // Note we need to open ALL the channel instances that are being used.
    // This is setup in the Harmony ADC configuration
    DRV_ADC0_Open();
    DRV_ADC1_Open();
    DRV_ADC2_Open();
    DRV_ADC3_Open();
    DRV_ADC4_Open();
    DRV_ADC5_Open();

    // Input selection
    //DRV_ADC_NegativeInputSelect(DRV_ADC_MUX_A, DRV_ADC_INPUT_NEGATIVE_VREF_MINUS);
    //DRV_ADC_PositiveInputSelect(DRV_ADC_MUX_A, DRV_ADC_POSITIVE_INPUT_NUM);

    // Enable the scan input on MUX A
    PLIB_ADC_MuxAInputScanEnable(DRV_ADC_ID_1);
    
    // Start the ADC
    DRV_ADC_Start();
}

/**-----------------------------------------------------------------------------
 * Performs a cycle of adc processing
 * @return 
 */
bool adc_process()
{
    bool result = false;
    bool has_stopped = false;
    
    int x;

    for (x = 0; x < ADC_INPUT_COUNT; x++)
    {
        ADC_INPUT_DEFINITION* def = adc_data_get_input_definition_by_index(x);
        if (def->input != ADC_INPUT_ID_INVALID)
        {
            if (def->active)
            {
                if (DRV_ADC_SamplesAvailable(def->input))
                {
                    //logger_log_debugf(0, "ADC Samples available %d --", x);
                    // Stop the ADC while we process it
                    if (!has_stopped)
                    {
                        DRV_ADC_Stop();
                        has_stopped = true;

                        // Call Scan start on all definitions.  Note this may result in the same 
                        // scan function running more than once
                        int y;
                        for (y = 0; y < ADC_INPUT_COUNT; y++)
                        {
                            ADC_INPUT_DEFINITION* def = adc_data_get_input_definition_by_index(y);
                            if (def->input != ADC_INPUT_ID_INVALID)
                            {
                                if (def->scan_start_callback != NULL)
                                {
                                    ((ADC_INPUT_SCAN_START_CALLBACK)def->scan_start_callback)();
                                }
                            }
                        }
                    }

                    def->current_value = DRV_ADC_SamplesRead(def->input);
                    //logger_log_debugf(0, "ADC Samples Read %d %d --", x, def->current_value);
                   
                    // Initiate the callback if it is required
                    if (def->data_available_callback != NULL)
                    {
                        ((ADC_DATA_AVAILABLE_CALLBACK)def->data_available_callback)(def);
                    }
                }
            }
        }
    }

    if (has_stopped)
    {
        // Restart the ADC
        DRV_ADC_Start();
        
        // Call Scan start on all definitions.  Note this may result in the same 
        // scan function running more than once
        int x;
        for (x = 0; x < ADC_INPUT_COUNT; x++)
        {
            ADC_INPUT_DEFINITION* def = adc_data_get_input_definition_by_index(x);
            if (def->input != ADC_INPUT_ID_INVALID)
            {
                if (def->scan_end_callback != NULL)
                {
                    ((ADC_INPUT_SCAN_END_CALLBACK)def->scan_end_callback)();
                }
            }
        }

        // Set result
        result = true;
    }
    else
    {
    }
    
    return result;
}
   
/**-----------------------------------------------------------------------------
 * Gets Whether the input is currently active or not 
 * @return - Whether the input is currently active or not 
 */
void adc_set_input_active(ADC_INPUT_ID input, bool active)
{
    ADC_INPUT_DEFINITION* def = adc_data_get_input_definition(input);
    if (def != NULL)
    {
        def->active = active;
        logger_log_infof(LOGGER_CODE_ADC, "ADC: Input %d set to %s ", input, def->active ? "Active" : "Inactive");
    }
}

/* *****************************************************************************
 End of File
 */
