/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    adc.c

  @Summary
    Public Interface to the adc module

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/adc/adc.h"
#include "app_config.h"
#include "utilities/adc/adc_defs.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    ADC_INPUT_ID input;
    
    ADC_INPUT_DEFINITION* input_def;
    
    ADC_BUTTON_DEFINITION input_buttons[ADC_MAX_BUTTON_COUNT];
    
    ADC_BUTTON_INFO current_state;

} ADC_BUTTON_INPUT_DEFINITION;

typedef struct
{
    ADC_BUTTON_INPUT_DEFINITION inputs[ADC_BUTTON_INPUT_COUNT];
    
    // A pointer to the ADC_BUTTON_INFO that is the most current
    ADC_BUTTON_INFO* current_button;
    
    //! True if we are currently in adc processing - used to ensure that the star/end scan
    // functions are only processed once..
    bool in_processing;
} ADC_DATA;

static ADC_DATA adc_data = 
{
    .in_processing = false,
    .current_button = NULL,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward Declarations
static ADC_BUTTON_INPUT_DEFINITION* adc_button_get_definition(ADC_INPUT_ID input);

/**-----------------------------------------------------------------------------
 * Used to find an ADC_BUTTON_DEFINITION.  If not found it returns the next unused 
 * ADC_BUTTON_DEFINITION
 * @param input
 * @return 
 */
static ADC_BUTTON_INPUT_DEFINITION* adc_button_allocate(ADC_INPUT_ID input)
{
    ADC_BUTTON_INPUT_DEFINITION* result = NULL;
    ADC_BUTTON_INPUT_DEFINITION* next_free_result = NULL;
    
    int x;
    for (x = 0; x < ADC_BUTTON_INPUT_COUNT; x++)
    {
        if (adc_data.inputs[x].input == input)
        {
            result = &adc_data.inputs[x];
            break;
        }
        else if (adc_data.inputs[x].input == ADC_INPUT_ID_INVALID)
        {
            if (next_free_result == NULL)
            {
                next_free_result = &adc_data.inputs[x];
            }
        }
    }
    
    // Allocate the next free result
    if (result == NULL)
    {
        result = next_free_result;
        result->input = input;
    }
    
    if (result != NULL)
    {
        // Set current state input
        result->current_state.input = input;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Callback for the ADC data callback
 * @param input_def
 */
static void adc_button_data_callback(ADC_INPUT_DEFINITION* input_def)
{
    ADC_BUTTON_INPUT_DEFINITION* def = adc_button_get_definition(input_def->input);

    // Convert to mV
    // 10 bits = 1024 steps (0 to 1023) 
    // Result_mV = sample_value * (Reference voltage in mV / 1023)
    uint64_t mV = adc_calculate_mV(input_def);            
    
    uint32_t ticks = SYS_TMR_TickCountGet();
    bool found = false;

//    if (mV < 3300)
//    {
//        logger_log_debugf(0, "ADC Button Samples Read %d %d %d %d --", mV, def->input, input_def->input, input_def->current_value);
//    }
    
    // First check whether the current input voltage is within rang of any of the buttons
    uint16_t mV16 = mV;
    int x;
    for (x = 0; x < ADC_MAX_BUTTON_COUNT; x++)
    {
        if ((mV16 >= def->input_buttons[x].range_low) && (mV16 <= def->input_buttons[x].range_high))
        {
            //logger_log_debugf(0, "ADC Button Samples match %d %d %d %d %d --", mV16, x, def->input, def->input_buttons[x].range_low, def->input_buttons[x].range_high);
            // Update the current state for the input
            if (def->current_state.button == def->input_buttons[x].button)
            {
                // The current button is still being held down, so update the ticks for it..
                def->current_state.duration_ticks = ticks - def->current_state.start_ticks;
            }
            else
            {
                // A New button is being held down, so start the ticks for it
                def->current_state.button = def->input_buttons[x].button;
                def->current_state.start_ticks = ticks;
                def->current_state.duration_ticks = 0;
            }
            
            adc_data.current_button = &def->current_state;
            
            found = true;
            break;
        }
    }

    // Check if we have a match
    if (!found)
    {
        // If not then test for an error
        if (mV > ADC_BUTTON_ERROR_MV)
        {
            def->current_state.button = ADC_BUTTON_ERROR;
            def->current_state.start_ticks = 0;
            def->current_state.duration_ticks = 0;
        }
        else
        {
            // There's no error, so check if the current button has timed out
            // which takes 1 second) and if so set the current button to ADC_BUTTON_NONE
            if ((def->current_state.button != ADC_BUTTON_NONE) &&
               ((def->current_state.duration_ticks > (SYS_TMR_TickCounterFrequencyGet() * 1))))
            {
                def->current_state.button = ADC_BUTTON_NONE;
                def->current_state.start_ticks = 0;
                def->current_state.duration_ticks = 0;
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback for the ADC scan end callback
 * @param input_def
 */
static void adc_button_scan_end_callback()
{
    if (adc_data.in_processing)
    {
        adc_data.in_processing = false;
    }
}

/**-----------------------------------------------------------------------------
 * Callback for the ADC scan start callback
 * @param input_def
 */
static void adc_button_scan_start_callback()
{
    adc_data.in_processing = true;
}

/**-----------------------------------------------------------------------------
 * Gets a reference to the supplied input  definition
 * @param input
 * @return 
 */
static ADC_BUTTON_INPUT_DEFINITION* adc_button_get_definition(ADC_INPUT_ID input)
{
    ADC_BUTTON_INPUT_DEFINITION* result = NULL;
    
    int x;
    for (x = 0; x < ADC_BUTTON_INPUT_COUNT; x++)
    {
        if (adc_data.inputs[x].input == input)
        {
            result = &adc_data.inputs[x];
            break;
        }
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Required to allocate an input for the ADC module.  This should be called once for
 * each input.
 */
void adc_button_add_input(ADC_INPUT_ID input, ADC_BUTTON* buttons, ADC_INPUT_STATE state)
{
    ADC_BUTTON_INPUT_DEFINITION* def = adc_button_allocate(input);
    if ((def != NULL) && (def->input_def == NULL))
    {
        int x;
        for (x = 0; x < ADC_MAX_BUTTON_COUNT; x++)
        {
            def->input_buttons[x].button = buttons[x];
            def->input_buttons[x].range_low = ADC_BUTTON_DEFAULTS[x].range_low;
            def->input_buttons[x].range_high = ADC_BUTTON_DEFAULTS[x].range_high;
        }
        
        // Associate with the ADC Input
        ADC_INPUT_DEFINITION* input_def = adc_add_input(input);
        if (input_def != NULL)
        {
            def->input_def = input_def;
            input_def->active = state;
            input_def->data_available_callback = adc_button_data_callback;
            input_def->scan_end_callback = adc_button_scan_end_callback;
            input_def->scan_start_callback = adc_button_scan_start_callback;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Returns an ADC_BUTTON_INFO for the current selected  button
 * @return 
 */
ADC_BUTTON_INFO* adc_button_get_current()
{
    return adc_data.current_button;
}

/**-----------------------------------------------------------------------------
 * Returns an ADC_BUTTON_INFO for each input supporting buttons (ADC_BUTTON_INPUT_COUNT)
 * @return 
 */
ADC_BUTTON* adc_button_get_current_buttons()
{
    static ADC_BUTTON result[ADC_BUTTON_INPUT_COUNT];
    
    int x;
    for (x = 0; x < ADC_BUTTON_INPUT_COUNT; x++)
    {
        result[x] = adc_data.inputs[x].current_state.button;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Initialises the ADC Buttons
 */
void adc_button_initialise()
{
    int x;
    for (x = 0; x < ADC_BUTTON_INPUT_COUNT; x++)
    {
        adc_data.inputs[x].input = ADC_INPUT_ID_INVALID;
        adc_data.inputs[x].input_def = NULL;
    }
        
    adc_data.current_button = NULL;
}

/* *****************************************************************************
 End of File
 */
