/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    adc.c

  @Summary
    Public Interface to the adc module

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/adc/adc_pot.h"
#include "utilities/adc/adc.h"
#include <math.h>
#include "app_config.h"
#include "utilities/adc/adc_defs.h"
#include "utilities/events.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    ADC_INPUT_ID input;
    
    ADC_INPUT_DEFINITION* input_def;
    
    // The current step value for the Pot
    uint16_t current_step;
    
    // Indicates whether the step has changed in the last scan
    bool has_changed;
    
    //! Determines whether the pot has had its first reading taken
    bool initialised;
    
    //! The number of values to be read from the pot.  This will usually be an un-even number so it can be centered
    uint16_t step_count;
} ADC_POT_DEFINITION;

typedef struct
{
    ADC_POT_DEFINITION inputs[ADC_POT_INPUT_COUNT];
    
	//! An Array of event subscription structures.  These ara managed by the events module when
	// invoking events.  There MUST be enough declared to handle all the concurrent event 
	// subscriptions for the module.  
	EVENT_SUBSCRIPTION event_subscriptions[ADC_POT_MAX_EVENT_SUBSCRIPIONS];    
} ADC_DATA;

static ADC_DATA adc_data = 
{
    
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward Declarations
static ADC_POT_DEFINITION* adc_pot_get_definition(ADC_INPUT_ID input);

/**-----------------------------------------------------------------------------
 * Used to find an ADC_BUTTON_DEFINITION.  If not found it returns the next unused 
 * ADC_BUTTON_DEFINITION
 * @param input
 * @return 
 */
static ADC_POT_DEFINITION* adc_pot_allocate(ADC_INPUT_ID input)
{
    ADC_POT_DEFINITION* result = NULL;
    ADC_POT_DEFINITION* next_free_result = NULL;
    
    int x;
    for (x = 0; x < ADC_POT_INPUT_COUNT; x++)
    {
        if (adc_data.inputs[x].input == input)
        {
            result = &adc_data.inputs[x];
            break;
        }
        else if (adc_data.inputs[x].input == ADC_INPUT_ID_INVALID)
        {
            if (next_free_result == NULL)
            {
                next_free_result = &adc_data.inputs[x];
            }
        }
    }
    
    // Allocate the next free result
    if (result == NULL)
    {
        result = next_free_result;
        result->input = input;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Calculates the current step based on the raw ADC value (in the range 0 - ADC_DATA_RESOLUTION
 * @param raw_value
 * @return 
 */
static uint16_t adc_pot_calculate_step(ADC_POT_DEFINITION* def, uint32_t raw_value)
{
    return ((raw_value * def->step_count) / (ADC_DATA_RESOLUTION + 1));
}

/**-----------------------------------------------------------------------------
 * Callback for the ADC data callback
 * @param input_def
 */
static void adc_pot_data_callback(ADC_INPUT_DEFINITION* input_def)
{
    ADC_POT_DEFINITION* def = adc_pot_get_definition(input_def->input);
        
    ADC_POT_DATA_AVAILABLE_EVENT_DATA event_data;
    event_data.input = def->input;
    event_data.step_count = def->step_count;
    event_data.initialised = true;

    uint32_t raw_value = input_def->current_value;
    
    if (!def->initialised)
    {
        def->initialised = true;
        
        def->current_step = adc_pot_calculate_step(def, raw_value);
//
        event_data.initialised = false;

        def->has_changed = true;
    }
    else // Normal reading
    {
        // Calculate the new balance in case we've moved more than one zone
        // 
        uint16_t new_step = adc_pot_calculate_step(def, raw_value);
        if (new_step != def->current_step)
        {
            // If we have only moved one step either way then check that we have made
            // a significant move rather than it being just the pot flapping around
            // To do this we check we split the total range (ADC_DATA_RESOLUTION)
            // into however many steps there are the check that the current raw value is 
            // within the middle 25% of that range, or to the ends if we are at 0/Range end
            uint16_t step_width = ADC_DATA_RESOLUTION / def->step_count;

            if (new_step > def->current_step)
            {
                // Only move to the new step if it is high than the low value of
                // the next step
                uint16_t low_value = (new_step * step_width) + ((3 * step_width) / 8);
                if (raw_value < low_value)
                {
                    // Its not higher so set it back to the current step
                    new_step = def->current_step;
                }
            }
            else
            {
                // Only move to the new step if it is lower than the low value of
                // the next step
                uint16_t high_value = (new_step * step_width) + ((5 * step_width) / 8);
                if (raw_value > high_value)
                {
                    // Its not lower so set it back to the current step
                    new_step = def->current_step;
                }
            }
                
            if (def->current_step = new_step)
            {
                // Update with whatever the new_step is set at
                def->current_step = new_step;
                def->has_changed = true;
            }
        }
    }   
    
    event_data.current_step = def->current_step;
    event_data.has_changed = def->has_changed;
    events_raise(EVENT_MODULE_ADC_POT, ADC_EVENT_POT_DATA_AVAILABLE, (void*)&event_data);
}

/**-----------------------------------------------------------------------------
 * Callback for the ADC scan start callback
 * @param input_def
 */
static void adc_pot_scan_start_callback()
{
    int x;
    for (x = 0; x < ADC_POT_INPUT_COUNT; x++)
    {
        adc_data.inputs[x].has_changed = false;
    }
}

/**-----------------------------------------------------------------------------
 * Gets a reference to the supplied input  definition
 * @param input
 * @return 
 */
static ADC_POT_DEFINITION* adc_pot_get_definition(ADC_INPUT_ID input)
{
    ADC_POT_DEFINITION* result = NULL;
    
    int x;
    for (x = 0; x < ADC_POT_INPUT_COUNT; x++)
    {
        if (adc_data.inputs[x].input == input)
        {
            result = &adc_data.inputs[x];
            break;
        }
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Required to allocate an input for the ADC module.  This should be called once for
 * each input.
 */
void adc_pot_add_input(ADC_INPUT_ID input, uint16_t step_count, ADC_INPUT_STATE state)
{
    ADC_POT_DEFINITION* def = adc_pot_allocate(input);
    if ((def != NULL) && (def->input_def == NULL))
    {
        def->step_count = step_count;
        
        ADC_INPUT_DEFINITION* input_def = adc_add_input(input);
        if (input_def != NULL)
        {
            def->input_def = input_def;
            input_def->active = state;
            input_def->data_available_callback = adc_pot_data_callback;
            input_def->scan_start_callback = adc_pot_scan_start_callback;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Returns the current step value for the pot
 * @param input
 * @return 
 */
uint16_t adc_pot_get_current_step(ADC_INPUT_ID input)
{
    uint16_t result = ADC_POT_INVALID_STEP;
    
    ADC_POT_DEFINITION* def = adc_pot_get_definition(input);
    if (def != NULL)
    {
        result = def->current_step;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the offset of the current step from the calculated center of the range
 * @param input
 * @return 
 */
int16_t adc_pot_get_offset_from_center(ADC_INPUT_ID input)
{
    int16_t result = 0;
    
    ADC_POT_DEFINITION* def = adc_pot_get_definition(input);
    if (def != NULL)
    {
        if (def->step_count % 2 == 1)
        {
            result = def->current_step - ((def->step_count / 2) + 1);
        }
        else
        {
            result = def->current_step - (def->step_count / 2);
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns true if the pot has changed in the last scan
 * @param input
 * @return 
 */
bool adc_pot_has_changed(ADC_INPUT_ID input)
{
    bool result = false;

    ADC_POT_DEFINITION* def = adc_pot_get_definition(input);
    if (def != NULL)
    {
        result = def->has_changed;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Initialises the ADC Buttons
 */
void adc_pot_initialise()
{
    int x;
    for (x = 0; x < ADC_POT_INPUT_COUNT; x++)
    {
        adc_data.inputs[x].input = ADC_INPUT_ID_INVALID;
        adc_data.inputs[x].input_def = NULL;
        adc_data.inputs[x].current_step = ADC_POT_INVALID_STEP;
        adc_data.inputs[x].initialised = false;
    }

	// Register the events that the spdif_rx module supports in the EVENTS structure, 
	events_register(EVENT_MODULE_ADC_POT, adc_data.event_subscriptions, ADC_POT_MAX_EVENT_SUBSCRIPIONS);
}

/* *****************************************************************************
 End of File
 */
