/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    adc.c

  @Summary
    Public Interface to the adc module

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/adc/adc.h"
#include <math.h>
#include "app_config.h"
#include "utilities/adc/adc_defs.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    ADC_INPUT_ID input;
    
    ADC_INPUT_DEFINITION* input_def;
} ADC_TEMPERATURE_DEFINITION;

typedef struct
{
    ADC_TEMPERATURE_DEFINITION inputs[ADC_TEMPERATURE_INPUT_COUNT];
} ADC_DATA;

static ADC_DATA adc_data = 
{
    
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used to find an ADC_BUTTON_DEFINITION.  If not found it returns the next unused 
 * ADC_BUTTON_DEFINITION
 * @param input
 * @return 
 */
static ADC_TEMPERATURE_DEFINITION* adc_temperature_allocate(ADC_INPUT_ID input)
{
    ADC_TEMPERATURE_DEFINITION* result = NULL;
    ADC_TEMPERATURE_DEFINITION* next_free_result = NULL;
    
    int x;
    for (x = 0; x < ADC_TEMPERATURE_INPUT_COUNT; x++)
    {
        if (adc_data.inputs[x].input == input)
        {
            result = &adc_data.inputs[x];
            break;
        }
        else if (adc_data.inputs[x].input == ADC_INPUT_ID_INVALID)
        {
            if (next_free_result == NULL)
            {
                next_free_result = &adc_data.inputs[x];
            }
        }
    }
    
    // Allocate the next free result
    if (result == NULL)
    {
        result = next_free_result;
        result->input = input;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets a reference to the supplied input  definition
 * @param input
 * @return 
 */
static ADC_TEMPERATURE_DEFINITION* adc_temperature_get_definition(ADC_INPUT_ID input)
{
    ADC_TEMPERATURE_DEFINITION* result = NULL;
    
    int x;
    for (x = 0; x < ADC_TEMPERATURE_INPUT_COUNT; x++)
    {
        if (adc_data.inputs[x].input == input)
        {
            result = &adc_data.inputs[x];
            break;
        }
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Required to allocate an input for the ADC module.  This should be called once for
 * each input.
 */
void adc_temperature_add_input(ADC_INPUT_ID input, ADC_INPUT_STATE state)
{
    ADC_TEMPERATURE_DEFINITION* def = adc_temperature_allocate(input);
    if ((def != NULL) && (def->input_def == NULL))
    {
        ADC_INPUT_DEFINITION* input_def = adc_add_input(input);
        if (input_def != NULL)
        {
            def->input_def = input_def;
            input_def->active = state;
            input_def->data_available_callback = NULL;
        }
    }
}
   
/**-----------------------------------------------------------------------------
 * Gets the calculated temperature on the supplied input.
 * @param input_index - The index of the input to be reported on
 * @return - The calculated temperature to 1 decimal place e.g. 45.1degrees will be returned as 451.  
 *           If the input != ADC_SCAN_INPUT_MODE_TEMPERATURE it returns 0
 */
int16_t adc_temperature_get(ADC_INPUT_ID input)
{
    int16_t result = ADC_INVALID_REQUEST;
    
    ADC_TEMPERATURE_DEFINITION* def = adc_temperature_get_definition(input);
    if (def != NULL)
    {
        // Get the latest value - not we are not subtracting the offset, as it 
        // appears that by the time we do the calibration its already a full-running mode.
        if (DRV_ADC_SamplesAvailable(input))
        {
            def->input_def->current_value = DRV_ADC_SamplesRead(input);
        }
        
        if (def->input_def->current_value != 0)
        {
            // Calculate the delta reading
            float delta = (float)def->input_def->current_value / (float)ADC_DATA_RESOLUTION;
            
            // Check the delta is within sensor ranges - we were getting overtemps
            // because the delta was in the order of 0.00005
            if ((delta < (float)0.9) && (delta > (float)0.1))
            {
                // Adjust the temperature for a linear temperature reading 
                // LMT89 datasheet Figure 1.
                //float temperature = (delta * (float)ADC_TEMP_RANGE * -1.0f) + (float)ADC_TEMP_OFFSET;

                // Calculate the temperature
                // LMT89 datasheet transfer function http://www.ti.com/lit/ds/symlink/lmt89.pdf
                float voltage = delta * (float)3.3;
                float temperature = (float)-1481.96 + (float)sqrt(2196200.0 + (((float)1.8639 - voltage) / (float)0.00000388));

                // Convert by factor of 10
                result = (int16_t)(temperature * 10.0);

                if (result > (APP_MAX_WORKING_TEMPERATURE * 10.0))
                {
                    logger_log_debugf(0, "TEMP MAX EXCEEDED %d %d %2.4f %2.4f --", input, result, temperature, delta);
                }
            }
            else
            {
                result = ADC_INVALID_INPUT;
            }
        }
        else
        {
            result = ADC_INVALID_INPUT;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Initialises the ADC Buttons
 */
void adc_temperature_initialise()
{
    int x;
    for (x = 0; x < ADC_TEMPERATURE_INPUT_COUNT; x++)
    {
        adc_data.inputs[x].input = ADC_INPUT_ID_INVALID;
        adc_data.inputs[x].input_def = NULL;
    }
}

/* *****************************************************************************
 End of File
 */
