/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    adc.c

  @Summary
    Public Interface to the adc module

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/adc/adc.h"
#include "app_config.h"
#include "utilities/adc/adc_defs.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    ADC_INPUT_ID input;
    
    ADC_INPUT_DEFINITION* input_def;
} ADC_VOLTAGE_DEFINITION;

typedef struct
{
    ADC_VOLTAGE_DEFINITION inputs[ADC_VOLTAGE_INPUT_COUNT];
} ADC_DATA;

static ADC_DATA adc_data = 
{
    
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used to find an ADC_BUTTON_DEFINITION.  If not found it returns the next unused 
 * ADC_BUTTON_DEFINITION
 * @param input
 * @return 
 */
static ADC_VOLTAGE_DEFINITION* adc_voltage_allocate(ADC_INPUT_ID input)
{
    ADC_VOLTAGE_DEFINITION* result = NULL;
    ADC_VOLTAGE_DEFINITION* next_free_result = NULL;
    
    int x;
    for (x = 0; x < ADC_VOLTAGE_INPUT_COUNT; x++)
    {
        if (adc_data.inputs[x].input == input)
        {
            result = &adc_data.inputs[x];
            break;
        }
        else if (adc_data.inputs[x].input == ADC_INPUT_ID_INVALID)
        {
            if (next_free_result == NULL)
            {
                next_free_result = &adc_data.inputs[x];
            }
        }
    }
    
    // Allocate the next free result
    if (result == NULL)
    {
        result = next_free_result;
        result->input = input;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets a reference to the supplied input  definition
 * @param input
 * @return 
 */
static ADC_VOLTAGE_DEFINITION* adc_voltage_get_definition(ADC_INPUT_ID input)
{
    ADC_VOLTAGE_DEFINITION* result = NULL;
    
    int x;
    for (x = 0; x < ADC_VOLTAGE_INPUT_COUNT; x++)
    {
        if (adc_data.inputs[x].input == input)
        {
            result = &adc_data.inputs[x];
            break;
        }
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Required to allocate an input for the ADC module.  This should be called once for
 * each input.
 */
void adc_voltage_add_input(ADC_INPUT_ID input, ADC_INPUT_STATE state)
{
    ADC_VOLTAGE_DEFINITION* def = adc_voltage_allocate(input);
    if ((def != NULL) && (def->input_def == NULL))
    {
        ADC_INPUT_DEFINITION* input_def = adc_add_input(input);
        if (input_def != NULL)
        {
            def->input_def = input_def;
            input_def->active = state;
            input_def->data_available_callback = NULL;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Gets the calculated voltage on the supplied input in mV
 * @param input_index - The index of the input to be reported on
 * @return - The calculated voltage in mV.  
 * 
 * Note : This will return values for all input types..
 */
uint32_t adc_voltage_get(ADC_INPUT_ID input)
{
    uint32_t result = ADC_INVALID_REQUEST;
   
    ADC_VOLTAGE_DEFINITION* def = adc_voltage_get_definition(input);
    if (def != NULL)
    {
        // Get the latest value - not we are not subtracting the offset, as it 
        // appears that by the time we do the calibration its already a full-running mode.
        if (DRV_ADC_SamplesAvailable(input))
        {
            def->input_def->current_value = DRV_ADC_SamplesRead(input);
        }
        
        // Convert to mV
        // 10 bits = 1024 steps (0 to 1023) 
        // Result_mV = sample_value * (Reference voltage in mV / 1023)
        //logger_log_debugf(12, "ADC: %d %d --", adc_data.adc_values[input_index], (adc_data.adc_values[input_index] * 3300) / 1023);
        result = adc_calculate_mV(def->input_def);//((float)def->input_def->current_value * (float)3300.0) / (float)ADC_DATA_RESOLUTION;    
    }

    return result;    
}

/**-----------------------------------------------------------------------------
 * Initialises the ADC Buttons
 */
void adc_voltage_initialise()
{
    int x;
    for (x = 0; x < ADC_VOLTAGE_INPUT_COUNT; x++)
    {
        adc_data.inputs[x].input = ADC_INPUT_ID_INVALID;
        adc_data.inputs[x].input_def = NULL;
    }
}

/* *****************************************************************************
 End of File
 */
