/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/album_art.h"
#include <stdint.h>
#include "app_config.h"
#include "utilities/avatar_data.h"
#include "utilities/base64.h"
#include "utilities/display_message.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define ALBUM_ART_MAX_URL_LENGTH            AVATAR_ALBUM_ART_URL_MAX_LENGTH

//! Maximum supported size for Album Art (112 pixels square X 4 (ARGB))
#define ALBUM_ART_MAX_SIZE                  50176 //48400

//! The packet size for album art packets sent to the display
#define ALBUM_ART_DATA_PACKET_SIZE          DISPLAY_MESSAGE_DATA_TRANSFER_PACKET_SIZE

typedef struct
{
    // Storage for the album art data
    uint8_t data[ALBUM_ART_MAX_SIZE];
    
    char album_art_url[ALBUM_ART_MAX_URL_LENGTH];

    //! The length of the data (when complete, 0 otherwise)
    uint16_t data_length;
    
    //! Indicated whether a load is underway
    bool data_load_complete;
    
    //! The offset of the current position for reading data
    uint16_t data_read_offset;
    
    //! The offset of the current position for writing data
    uint16_t data_write_offset;

    // The current packet Id of transferred packets
    uint8_t packet_id;
    
    //! True if we are preparing the first packet, or if the Display has informed
    //  us it is ready for the next packet.
    bool ready_to_receive;
    
    //! The url has been set at least once
    bool url_initialised;
    
} ALBUM_ART_DATA;

static ALBUM_ART_DATA album_art_data = {
    .data_length = 0,
    .data_load_complete = true,
    .packet_id = 0,
    .ready_to_receive = true,
    .url_initialised = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */
// Forward declarations
static void album_art_send_data();

/**-----------------------------------------------------------------------------
 * Determines whether there is data to be transferred to the display
 */
static bool album_art_data_to_transfer()
{
    bool result = false;
    
    if (album_art_data.ready_to_receive)
    {
//        logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Checking for data to transfer --");
//        logger_process();

        // Is there data to send 
        if (album_art_data.data_write_offset != 0)
        {
            //logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Data ready to transfer to Display A %d %d --", album_art_data.data_write_offset, album_art_data.data_length);
            //logger_process();
            
            if (album_art_data.data_write_offset - album_art_data.data_read_offset >= ALBUM_ART_DATA_PACKET_SIZE)
            {
                // We can send data
                album_art_send_data();
                
                album_art_data.data_read_offset += ALBUM_ART_DATA_PACKET_SIZE;
                
                logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Transferring Full Size Data Packet %d %d %d --", album_art_data.packet_id, album_art_data.data_write_offset, album_art_data.data_read_offset);
                
                // Increment the packet Id
                album_art_data.packet_id++;
                
                // Set read to receive to false (will be set back when Display responds)
                album_art_data.ready_to_receive = false;
            }
        }
        else if (album_art_data.data_length - album_art_data.data_read_offset >= ALBUM_ART_DATA_PACKET_SIZE)
        {
            // We can send data
            album_art_send_data();

            album_art_data.data_read_offset += ALBUM_ART_DATA_PACKET_SIZE;

            logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Transferring Data Packet %d %d %d --", album_art_data.packet_id, album_art_data.data_write_offset, album_art_data.data_read_offset);

            // Increment the packet Id
            album_art_data.packet_id++;

            // Set read to receive to false (will be set back when Display responds)
            album_art_data.ready_to_receive = false;
        }
        else if (album_art_data.data_length != 0)
        {
            logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Transferring Last Data Packet %d %d %d --", album_art_data.packet_id, album_art_data.data_length, album_art_data.data_read_offset);

            // We can send data
            album_art_send_data();

            // Check we have sent all the packets, and if not then send the last one
            album_art_data.data_read_offset = 0;//album_art_data.data_length;
            album_art_data.data_length = 0;

            // Increment the packet Id
            album_art_data.packet_id++;

            // Set read to receive to false (will be set back when Display responds)
            album_art_data.ready_to_receive = false;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Send the next ScreenData packet to the Display
 */
static void album_art_send_data()
{
    uint8_t screen_data[DISPLAY_MESSAGE_UPLOAD_DATA_BUFFER_SIZE];
    display_message_initialise_data_transfer_data(screen_data, album_art_data.packet_id);
   
    //logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Sending Data to the Display %d/%d --", album_art_data.data_length, album_art_data.data_length);
    //logger_process();

    uint16_t len = 0;
    if ((album_art_data.data_length != 0) &&
        (album_art_data.data_length - album_art_data.data_read_offset <= ALBUM_ART_DATA_PACKET_SIZE))
    {
        len = album_art_data.data_length - album_art_data.data_read_offset;
        
        logger_log_debugf(LOGGER_CODE_AVATAR, "ALBUM ART Sending Last Packet of Data to Display id:%d len:%d --", album_art_data.packet_id, len);
        logger_process();
        display_message_set_control_flags(screen_data, DISPLAY_MESSAGE_CONTROL_FLAGS_END_OF_MESSAGE); 
    }
    else
    {
        len = ALBUM_ART_DATA_PACKET_SIZE;//album_art_data.data_write_offset - album_art_data.data_read_offset;
        
        //logger_log_debugf(LOGGER_CODE_AVATAR, "ALBUM ART Sending Data Packet to Display id:%d len:%d --", album_art_data.packet_id, len, album_art_data.data_length);
        //logger_process();
        display_message_set_control_flags(screen_data, DISPLAY_MESSAGE_CONTROL_FLAGS_NONE);
    }

    // Copy from the main buffer into a temporary buffer, this is because it 
    // appears we cant assume that the main buffer is in contiguous memory space
    
    uint16_t encoded_len = base64_encoded_size(ALBUM_ART_MAX_URL_LENGTH);
    char album_art_url[encoded_len + 1];
    
    base64_encode(&album_art_data.album_art_url[0], ALBUM_ART_MAX_URL_LENGTH, album_art_url);
    
    // Album Art URL
    display_message_add_screen_data(screen_data, 
        DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_ALBUM_ART_URL,
        DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT, 
        DISPLAY_MESSAGE_DATA_ENCODING_BIN16,
        (uint8_t*)album_art_url,
        encoded_len);

    // Copy from the main buffer into a temporary buffer, this is because it 
    // appears we cant assume that the main buffer is in contiguous memory space
    
    encoded_len = base64_encoded_size(len);
    char album_art[encoded_len + 1];
    
    base64_encode(&album_art_data.data[album_art_data.data_read_offset], len, album_art);

    len = encoded_len;

    // Populate the screen data
    display_message_add_screen_data(screen_data, 
        DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_ALBUM_ART, // Album Art
        DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT, // Default attribute
        DISPLAY_MESSAGE_DATA_ENCODING_BIN16,
        album_art,
        len);

    // We can use this as data_transfer and screen data are both in the same format
    display_message_finalise_and_send_screen_data(screen_data);
    
    logger_log_debugf(LOGGER_CODE_AVATAR, "ALBUM ART Sent to display id:%d len:%d --", album_art_data.packet_id, len);
    logger_process();
}

/**-----------------------------------------------------------------------------
 * Send the next ScreenData packet to the Display
 */
static void album_art_send_url()
{
    uint8_t screen_data[DISPLAY_MESSAGE_SCREEN_DATA_BUFFER_SIZE];
    display_message_initialise_screen_data(screen_data);
    
    uint16_t encoded_len = base64_encoded_size(ALBUM_ART_MAX_URL_LENGTH);
    char album_art_url[encoded_len + 1];
    
    base64_encode(&album_art_data.album_art_url[0], ALBUM_ART_MAX_URL_LENGTH, album_art_url);

    //logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Sending URL to Display --");
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR ALBUM ART %X %X %X %X - %X %X %X %X --", album_art_data.album_art_url[0], album_art_data.album_art_url[1], album_art_data.album_art_url[2], album_art_data.album_art_url[3], album_art_url[0], album_art_url[1], album_art_url[2], album_art_url[3]);
    //logger_process();
    
    // Album Art URL
    display_message_add_screen_data(screen_data, 
        DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_ALBUM_ART_URL,
        DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT, 
        DISPLAY_MESSAGE_DATA_ENCODING_BIN16,
        (uint8_t*)album_art_url,
        encoded_len);
    
//    // Album Art URL
//    display_message_add_screen_data(screen_data, 
//        DISPLAY_MESSAGE_CURRENTLY_PLAYING_DATA_ID_ALBUM_ART_URL,
//        DISPLAY_MESSAGE_ATTRIBUTE_DEFAULT, 
//        DISPLAY_MESSAGE_DATA_ENCODING_BIN16,
//        (uint8_t*)album_art_data.album_art_url,
//        ALBUM_ART_MAX_URL_LENGTH);

    // We can use this as data_transfer and screen data are both in the same format
    display_message_finalise_and_send_screen_data(screen_data);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used to move the write pointer on by however many bytes have been written to the album art data
 */    
void album_art_advance_write_pointer(uint16_t len, bool complete)
{
    album_art_data.data_write_offset += len;
    
    // Set the load complete marker
    album_art_data.data_load_complete = complete;
    
    if (complete)
    {
        //logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Write is Complete %d --", album_art_data.data_write_offset);
        album_art_data.data_length = album_art_data.data_write_offset;
        album_art_data.data_write_offset = 0;
    }
    
    //logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Write Offset set to %d --", album_art_data.data_write_offset);
}

/**-----------------------------------------------------------------------------
 * Resets the album art URL
 */
void album_art_clear_url()
{
    memset(album_art_data.album_art_url, 0, ALBUM_ART_MAX_URL_LENGTH);
}
    
/**-----------------------------------------------------------------------------
 * Returns a pointer to the current write position
 */    
uint8_t* album_art_get_write_pointer()
{
    return &album_art_data.data[album_art_data.data_write_offset];
}

/**-----------------------------------------------------------------------------
 * Used to check if the album art module is currently dealing with the supplied URL
 */    
bool album_art_has_url(char* url)
{
    return memcmp(url, album_art_data.album_art_url, ALBUM_ART_MAX_URL_LENGTH) == 0;
}

/**-----------------------------------------------------------------------------
 * Sets up the album art module to start receiving new album art.  If album art is 
 * currently being populated then a cancellation will be triggered.
 */
bool album_art_prepare_for_url(char* url, size_t len)
{
    bool result = false;
    
    if (!album_art_has_url(url) || ((album_art_data.data_write_offset == 0) && !album_art_data.data_load_complete))
//    if (album_art_data.data_write_offset == 0) TODO see BUG 3829....
    {
        //logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Prepare URL --");
        
        // Raise a cancellation message (unless album art is not initialised wit the first URL)
        // Removed as this does nothing and the avatar will send a complt Album art once started.
        //if (album_art_data.url_initialised && !album_art_data.data_load_complete)
        //{
        //    // Raise a cancellation message
        //   logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Prepare Raise Cancellation message --");
        //}
        
        // Initialise storage
        memset(album_art_data.data, 0, ALBUM_ART_MAX_SIZE); // TODO - If we cant assume memory is contiguous then this will cause problems??
        album_art_data.data_read_offset = 0;
        album_art_data.data_write_offset = 0;
        album_art_data.data_length = 0;
        album_art_data.packet_id = 0;
        album_art_data.ready_to_receive = true;
        
        // Copy the URL
        memset(album_art_data.album_art_url, 0, ALBUM_ART_MAX_URL_LENGTH);
        memcpy(album_art_data.album_art_url, url, min(len, ALBUM_ART_MAX_URL_LENGTH));
        album_art_data.data_load_complete = false;
        album_art_data.url_initialised = true;
        
        album_art_send_url();
        
//        logger_log_debugf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Prepare Complete --");
//        logger_process();
        
        result = true;
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_ALBUM_ART, "ALBUM ART Prepare for URL aborted - URL is current --");
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Processes album art that has been transferred from the avatar, sends it onto the display
 */
void album_art_process()
{
    if (album_art_data_to_transfer())
    {
        
    }
}

/**-----------------------------------------------------------------------------
 * Sets the Ready To Receive flag on the album art
 */
void album_art_set_ready_to_receive()
{
    album_art_data.ready_to_receive = true;
}

/* *****************************************************************************
 End of File
 */
