/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    app_commands_utilities.c

  @Summary
    Brief description of the file.

  @Description
    Utility commands used in the app_command_ modules.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/app_commands_utilities.h"
#include "utilities/input_control_defs.h"
#include "utilities/rs232.h"
//TODO#include "utilities/tcpip.h"

#ifdef DISPLAY_MESSAGE_ACTIVE            
#include "utilities/display_message.h"
#endif       

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Sets up the message to display the default screen
 * @param message
 * @return 
 */
void app_commands_display_default_screen(STRUCTURED_MESSAGE* message)
{
    //logger_log_debugf(0, "APP Displaying Default Screen --");
    bool app_running = app_context_is_running();

    // required to stop MENU Option display during startup
    if (app_running)
    {
        app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);

        MENU_ID screen_id = app_commands_get_default_screen();
         
#ifdef DISPLAY_MESSAGE_ACTIVE            
        app_context_set_menu(screen_id, DISPLAY_MESSAGE_SCREEN_NAVIGATION_DIRECT);
#else
        app_context_set_menu(screen_id);
#endif
    }
 }
 
/**-----------------------------------------------------------------------------
 * Get the index of a value in an array of int[]
 * @param value
 * @param array
 * @param array_len
 * @return 
 */
int8_t app_commands_get_array_index_of(int value, int* array, int array_len)
{
    int8_t result = -1;
    
    // Find the index of the current value
    int x;
    for (x = 0; x < array_len; x++)
    {
        if (value == array[x])
        {
            result = x;
            break;
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Sets up the message to get the default screen
 * @param message
 * @return 
 */
MENU_ID app_commands_get_default_screen()
{
    MENU_ID result = APP_COMMAND_MENU_DEFAULT;
    
    uint8_t index = input_control_get_current_definition()->index;
    INPUT_CONTROL_PREVIEW_INFO* preview = input_control_get_preview_info();
    if (preview->active)
    {
        index = preview->input;
    }

    result = AUDIO_INPUT_MENUS[index];

    return result;
 }

/**-----------------------------------------------------------------------------
 * Calculates the next position in an array based on the direction and array length
 * @param idx
 * Note: This version is making an assumption that direction is +/- 1 and not a larger number 
 * when wrapping round,
 * @param direction
 * @param array_len
 */
void app_commands_get_next_array_pos(int8_t* idx, int8_t direction, int array_len)
{
    *idx += direction;
    if (*idx >= (int8_t)array_len)
    {
       *idx = 0;
    }
    else if (*idx < 0)
    {
       *idx = (int8_t)array_len - 1;
    }
}

/**-----------------------------------------------------------------------------
 * Generate a structure message that can be used to generate a message.
 * @param value
 * @param array
 * @param array_len
 * @return 
 */    
 void app_commands_invoke_message(uint8_t command, STRUCTURED_MESSAGE_CALLBACK status_proc, uint8_t action)
 {
    STRUCTURED_MESSAGE message;
    structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
    message.command = command;
    message.data_length = 1;   
    message.data[0] = action;
    status_proc(&message); 
 }

/**
 * Provides a standard set of logic to update the message action parameter 
 * based upon the value of the message.  This can be used in edit functions to save 
 * having to interrogate the message, but can be ignored if neccessary?
 * Assumes data[0] contains the value of a command message
 * So if the message is absolute and the value is 0xF0 it becomes GetStatus
 * If it is 0xF1 it becomes increment
 * If it is 0xF2 it becomes decrement
 * @param message
 */

/**-----------------------------------------------------------------------------
 * Performs standard processing on a pointer to a value
 * @param message       - the message
 * @param current_value - the current value to be updated
 * @param value         - the value used for update
 * @param increment     - the value to increment/decrement by
 */
void app_commands_process_message_value(STRUCTURED_MESSAGE* message, 
        int16_t* current_value, 
        int16_t* value,
        int16_t increment)
{
    switch(message->action)
    {
        case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
        {
            *current_value = *value;
            break;
        }
        case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
        {
            *current_value += increment;
            break;
        }
        case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
        {
            *current_value -= increment;
            break;
        }
        
        case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
        {
            message->data[0] = *current_value;
            message->response = STATUS_UPDATE;
            
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Performs standard processing on a pointer to a value based on values in an int array
 * @param message       - the message
 * @param current_value - the current value to be updated
 * @param value         - the value used for update
 * @param array         - a pointer to an int[] array
 * @param array_len     - the number of elements in  the array
 */
void app_commands_process_message_value_array(STRUCTURED_MESSAGE* message, 
        int16_t* current_value, 
        int16_t* value,
        int* array,
        int array_len)
{
    switch(message->action)
    {
        case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
        {
            *current_value = *value;
            break;
        }
        case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
        {
            int8_t idx = app_commands_get_array_index_of(*current_value, array, array_len);
            if (idx >= 0)
            {
                // Increment it and adjust if we are over the end
                app_commands_get_next_array_pos(&idx, 1, array_len);
            
                // set the current_value
                *current_value = array[idx];
            }
            
            break;
        }
        
        case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
        {
            int8_t idx = app_commands_get_array_index_of(*current_value, array, array_len);
            if (idx >= 0)
            {
                // Increment it and adjust if we are over the end
                app_commands_get_next_array_pos(&idx, -1, array_len);
            
                // set the current_value
                *current_value = array[idx];
            }

            break;
        }
        
        case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
        {
            message->data[0] = *current_value;
            message->response = STATUS_UPDATE;
            
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Boolean version of process_message_value.  Includes range checking
 * @param message
 * @param current_value
 * @param value
 * @param increment
 * @return true if processed, false if not
 */
bool app_commands_process_message_value_bool(STRUCTURED_MESSAGE* message, 
        int8_t* current_value, 
        int8_t* value,
        bool send_message)
{
    bool result = true;

    message->response = STATUS_UPDATE;
    
    switch(message->action)
    {
        case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
        {
            if ((*value >= 0) && (*value <= 1))
            {
                *current_value = *value;
            }
            else
            {
                message->response = PARAM_NOT_RECOGNISED;
            }

            break;
        }
        case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
        case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
       {
            *current_value = !*current_value;
            break;
        }

        case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
        {
            break;
        }
        
        default:
            result = false;
            break;
    }

    if (send_message && result)
    {
        message->data[0] = *current_value;
        app_commands_send_message(message);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Generate a structure message that can be used to generate a request status message.
 * @param value
 * @param array
 * @param array_len
 * @return 
 */    
void app_commands_request_status(uint8_t command, STRUCTURED_MESSAGE_CALLBACK status_proc)
{
    app_commands_invoke_message(command, status_proc, APP_COMMAND_VALUE_REQUEST_STATUS);
}

/**-----------------------------------------------------------------------------
 * Uses data[0] from the message to identify whether the message is a standard REques Status/Increment/Decrement 
 * type of message of pure data.
 * @param message
 */
void app_commands_review_message_action(STRUCTURED_MESSAGE* message)
{
    switch(message->action)
    {
        case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
        {
            uint8_t value = (uint8_t)message->data[0];
            switch(value)
            {
                case APP_COMMAND_VALUE_REQUEST_STATUS:
                    message->action = STRUCTURED_MESSAGE_ACTION_GET_STATUS;
                    break;
                    
                case APP_COMMAND_VALUE_INCREMENT:
                    message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                    break;
                    
                case APP_COMMAND_VALUE_DECREMENT:
                    message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                    break;
            }
            break;
        }
        
        default:
            // Leave as is
            break;
    }
}

/**-----------------------------------------------------------------------------
 * Sends a response message to either the RS232 or Ethernet port dependent on the 
 * message source
 * @param message
 */
void app_commands_send_message(STRUCTURED_MESSAGE* message)
{
    switch (message->source)
    {
        case STRUCTURED_MESSAGE_SOURCE_RS232:
            rs232_send_message(message);
            break;

        case STRUCTURED_MESSAGE_SOURCE_TCPIP:
            //rs232_send_message(message);
            tcpip_send_structured_message(message);
            break;
            
        default:
        {
            // Dont send messages for menu changes that do not have an external command
            // Note - This needs setting in the edit method manually
            if (message->command != APP_COMMANDS_MENU_VALUE_CHANGE)
            {
                // Unsolicited messages
                rs232_send_message(message);
                message->state_data[0] = 0;
                tcpip_send_structured_message(message);
            }
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Use to kick off a stateful process
 */
void app_commands_start_process(void (*process_function)(MENU_ID), void (*end_function)())
{
    // Invoke a process if set.
    app_context_set_mode(APP_CONTEXT_MODE_PROCESS);
    stateful_process_initialise_info(process_function, end_function);
    stateful_process_invoke();   
}

/* *****************************************************************************
 End of File
 */
