/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    avatar.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/avatar.h"
#include <stdio.h>
#include "utilities/avatar_control.h"
#include "utilities/avatar_packet.h"
#include "utilities/avatar_message_processing.h"
#include "utilities/avatar_packet_data.h"
#include "utilities/avatar_process_rule.h"
#include "utilities/logger/logger.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Can be use to output the size of data allocated for messages.  Comment out when not needed
 */
//#define AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE

/**
 * A queue of packets received from the Avatar to be processed
 */
typedef struct
{
    // Gives a name to the queue for debugging
    char name[10];
    
    // Stores the size of the queue - MUST be the same as the #define used to allocate the queue
    uint8_t size;
    
    // The queue holding pointers to the packets to be processed
    AVATAR_PACKET* queue[AVATAR_PACKET_QUEUE_SIZE];
            
    // The current packet ready to be processed
    volatile uint16_t read_pos;
    
    // Then position in the queue to add the next packet to be processed
    volatile uint16_t write_pos;
    
} AVATAR_PACKET_QUEUE;

/**
 * Avatar data
 */
typedef struct
{
    // I2C handle
    DRV_HANDLE i2c_handle;
    
    // The type of device being used e.g. SA30
    AVATAR_DEVICE device;

    // Current processing state
    AVATAR_STATE state;

    // Static holder for current data
    AVATAR_PACKET_HEADER current_header;
    
    // A pointer to the currently used packet
    AVATAR_PACKET* current_packet;
    
    // A pointer to the packet currently being sent
    AVATAR_PACKET* current_send_packet;
    
    // Pointer to the storage for message data (header/or packet data))
    uint8_t* data_ptr;
    
    // The offset in the data to take account of the header in the message
    // This is subtracted from the message_pos when working out where to read/write data from.
    uint8_t data_offset;
    
    // Current position in the message
    uint16_t message_pos;
    
    //! Frame ID - holds a running count of the frame id to be added to messages
    //  sent to the avatar
    uint8_t frame_id;    
    
    //! Set to true when an Avatar Mute state change is being processed.
    bool is_processing_mute;
   
    //! Set to true when an Avatar Volume state change is being processed.
    bool is_processing_volume;
    
    // The queue of packets ready for processing
    AVATAR_PACKET_QUEUE processing_queue;
    
    // The queue of packets ready for sending
    AVATAR_PACKET_QUEUE send_queue;
    
    // A pointer to an array of AVATAR_STATE_ID to be subscribed to
    AVATAR_STATE_ID* subscriptions;
    
    // The number of state id subscriptions
    uint8_t subscriptions_len;
    
//    //! 
//    AVATAR_PROCESSING_RULES processing_rules;
    
#ifdef AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE
    uint16_t allocated_message_size;
#endif    
    
    // Indicates a packet is being sent - this is included because during the transfer of the upgrade
    // files the I2C_SLAVE_STATE_SEND_END state was not being triggered (although you could see it in the 
    // SAELEA).  This is required to move the send queue on - so logic has been added to the MSG start 
    // to sort the send queue out of it thinks it is currently sending..
    volatile bool sending_packet;
} AVATAR_DATA;

static AVATAR_DATA avatar_data =
{
    .state = AVATAR_STATE_INITIALISE,
    .frame_id  = 1,
    .is_processing_volume = false,
    .processing_queue.read_pos = 0,
    .processing_queue.write_pos = 0,
    .send_queue.read_pos = 0,
    .send_queue.write_pos = 0,
//    .processing_rules.active_rule = NULL,
    .subscriptions = NULL,
    .subscriptions_len = 0,
    
#ifdef AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE
    .allocated_message_size = 0,
#endif    
    
    .sending_packet = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations
static void avatar_end_send_sequence();
static void avatar_handshake_complete(AVATAR_PROCESS_RULE* rule, AVATAR_MESSAGE* message);
static void avatar_queue_add_packet(AVATAR_PACKET_QUEUE* queue, AVATAR_PACKET* packet);
static void avatar_queue_advance_read(AVATAR_PACKET_QUEUE* queue);
static AVATAR_PACKET* avatar_queue_get_packet(AVATAR_PACKET_QUEUE* queue);
static void avatar_prepare_send();
static void avatar_setup_ack();
static void avatar_setup_nack();
static void avatar_start_send_sequence();
static void avatar_system_ready_complete(AVATAR_PROCESS_RULE* rule, AVATAR_MESSAGE* message);
void avatar_set_state(AVATAR_STATE state);

/**-----------------------------------------------------------------------------
 * Invoked when the handshaking is deemed to be complete
 * @param rule
 * @param packet
 */
static void avatar_handshake_complete(AVATAR_PROCESS_RULE* rule, AVATAR_MESSAGE* message)
{
    if (rule->active)
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave HANDSHAKE IS COMPLETE ");
        avatar_set_state(AVATAR_STATE_CONFIGURATION);
        
        avatar_configure();
    }
}

/**-----------------------------------------------------------------------------
 * Callback used by I2C to inform when data is available 
 * As far as I understand it when we are working in slave mode you set the IRQ
 * pin to let the master know you want to talk to it, then I'm assuming that
 * Harmony magic means that this method is called to let you send or receive the 
 * data.
 * https://www.microchip.com/forums/m1065552.aspx
 * See drv_i2c.h
 */
void avatar_i2c_slave_callback(I2C_SLAVE_EVENT* event)
{
//    logger_log_debugf(0, "AVATAR Callback %d --", *event->data);
//    logger_process();
    switch (event->state)
    {
        case I2C_SLAVE_STATE_RECEIVE_START:
        {
            // Check state - were doing it here to minimise the number of times this is called
            // as we expect the avatar to send us something before we have to reply.
            // Note this will probably have already been called earlier during Initialisation
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave RECEIVE START %d addr: 0x%02X  --", event->state, *event->data);
            if (avatar_data.state == AVATAR_STATE_INITIALISE)
            {
                avatar_set_state(AVATAR_STATE_HANDSHAKE);
            }
            
            avatar_data.data_ptr = (uint8_t*)&avatar_data.current_header;
            avatar_data.message_pos = 0;
            avatar_data.data_offset = 0;
            break;
        }
        
        case I2C_SLAVE_STATE_SEND_START:
        {
//            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave SEND START %d addr: 0x%02X  --", event->state, *event->data);
//            logger_process();
            avatar_data.data_ptr = (uint8_t*)&avatar_data.current_header;
            avatar_data.message_pos = 0;
            avatar_data.data_offset = 0;
            break;
        }
        
        case I2C_SLAVE_STATE_RECEIVING:
        {
            // We will read the header first then  allocate a packet to get then rest
            // data_pos represents the position in the message rather than in the data
            // Set the data at the current position 
            if (avatar_data.data_ptr != NULL)
            {
                //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR RECEIVING %d addr: 0x%02X  --", event->state, *event->data);

                *(avatar_data.data_ptr + avatar_data.message_pos - avatar_data.data_offset) = *event->data;
            
                if (++avatar_data.message_pos == sizeof(AVATAR_PACKET_HEADER))
                {
//                    if (avatar_packet_header_get_data_length(&avatar_data.current_header) > 2096)
//                    {
//                        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR RECEIVING %d addr: 0x%02X %d  --", event->state, *event->data, avatar_packet_header_get_data_length(&avatar_data.current_header));
//                    }
                    // Check the message type and data length.  If the message is an IRQ
                    // message then we want to send pre prepared data however the length of data to allocate is in the 
                    // first 4 bytes of the payload, so we do need data
                    // also if its an ACK/NACK then we don't need to allocate data
                    if (avatar_packet_header_get_data_length(&avatar_data.current_header) > 0)
                    {
#ifdef AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE
                        avatar_data.allocated_message_size += avatar_packet_header_get_length(&avatar_data.current_header);
                        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Message Size Receive %d %d --", avatar_data.allocated_message_size, avatar_packet_header_get_length(&avatar_data.current_header));
#endif   
                        // Allocate a packet
                        // Note there is an issue somewhere (not yet found) where if a packet cannot be allocated
                        //      then the whole lot keels over instead of being ignored - required more investigation
                        //      as at the moment I cant see why its failing.
                        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet Allocating Request --");
                        avatar_data.current_packet = avatar_packet_allocate(&avatar_data.current_header, true);

                        if (avatar_data.current_packet != NULL)
                        {
                            avatar_data.current_packet->state = AVATAR_PACKET_STATE_LOADING;

                            // Set the data_ptr
                            avatar_data.data_ptr = (uint8_t*)avatar_data.current_packet->data;
                            avatar_data.data_offset = sizeof(AVATAR_PACKET_HEADER);
                        }
                        else
                        {
                            //logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Error allocating packet ");

                            // Just keep overwriting the header - this message is knackered!
                            avatar_data.data_ptr = (uint8_t*)&avatar_data.current_header;
                            avatar_data.message_pos = 0;
                            avatar_data.data_offset = 0;
                        }
                    }
                    else
                    {
                        avatar_data.data_ptr = NULL;
                        avatar_data.current_packet = NULL;
                    }
                }
            }
            break;
        }
        
        case I2C_SLAVE_STATE_SENDING:
        {
            event->data = (avatar_data.data_ptr + avatar_data.message_pos - avatar_data.data_offset);

//            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave callback SEND %d %d %d 0x%02X %s --", event->state, avatar_data.message_pos, avatar_data.data_offset, *event->data,
//                    avatar_data.current_packet != NULL ? "DATA" : "NULL");
            
            // Check if we have data to send or not
            if (avatar_data.current_packet != NULL)
            {
                // There is data to send, check if we have sent all the header yet
                if (++avatar_data.message_pos == sizeof(AVATAR_PACKET_HEADER))
                {
                    // We have finished sending the header so set up for the next read 
                    // from the data packet
                    avatar_data.data_ptr = (uint8_t*)avatar_data.current_packet->data;
                    avatar_data.data_offset = sizeof(AVATAR_PACKET_HEADER);
                }
                else if ((avatar_data.message_pos >= sizeof(AVATAR_PACKET_HEADER)) && (avatar_data.message_pos > avatar_packet_header_get_length(&avatar_data.current_header)))
                {
                    // We've read all the data from the data packet so just send a dummy byte
                    // For convenience this is the first byte of the SOP (0xFF)
                    event->data = (uint8_t*)&avatar_data.current_header.start_of_packet;
                }
            }
            else
            {
                // There was no data to send (this was an ACK/NACK))
                if (avatar_data.message_pos <= sizeof(AVATAR_PACKET_HEADER))
                {
                    // Move the position of the header on 
                    avatar_data.message_pos++;
                }
                else
                {
                    // We've read all the data from the header so just send a dummy byte
                    // For convenience this is the first byte of the SOP (0xFF)
                    event->data = avatar_data.data_ptr;
                }
            }

//            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave callback SEND %d 0x%02X %s --", avatar_data.message_pos, *event->data,
//                    avatar_data.current_packet != NULL ? "DATA" : "NULL");
            
            break;
        }
        
        case I2C_SLAVE_STATE_RECEIVE_END:
        {
            // Check type of message received
            switch (avatar_packet_header_get_type(&avatar_data.current_header))
            {
                case AVATAR_PACKET_TYPE_RESET:
                {
                    //TODO Avatar - work out what needs doing for a reset..
                    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave RESET Request ");

                    // Setup ACK to be returned to the avatar
                    avatar_setup_ack();
                    
                    break;
                }
                
                case AVATAR_PACKET_TYPE_IRQ:
                {
                    // This is the avatar responding to us requesting to send data
                    // we need to copy the first packet in the send queue to be sent
                    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave IRQ Request --");
                    
                    // Get out how much data the Avatar is ready to receive
                    uint32_t available_length = 0;
                    memcpy(&available_length, avatar_data.current_packet->data, sizeof(uint32_t));
                    
                    uint16_t required_length = 0;
                    AVATAR_PACKET* packet = avatar_queue_get_packet(&avatar_data.send_queue);
                    if (packet != NULL)
                    {
                        required_length = packet->header.packet_length;
                    }

                    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave IRQ Request %d %d --", available_length, required_length);
                    
                    avatar_end_send_sequence();
                    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR AVATAR_PACKET_TYPE_IRQ");
                    
                    // Free the IRQ message packet before we start reassigning current_packet
#ifdef AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE
                    avatar_data.allocated_message_size -= avatar_packet_header_get_length(&avatar_data.current_packet->header);
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Message Size at IRQ free %d %d --", avatar_data.allocated_message_size, avatar_packet_header_get_length(&avatar_data.current_packet->header));
#endif
                    
                    // The message is OK
                    if (required_length <= available_length)
                    {
                        avatar_packet_free(avatar_data.current_packet);
                        
                        //TODO Tried this to cope with repeating IRQ request loop but removed again as it didnt fix itxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
                        //uint16_t zero = 0;
                        //memcpy(avatar_data.current_packet->data, &zero, sizeof(uint16_t));
                    
                        // With the fix for sending messages when a MSG is received
                        // we can then get a knock-on effect of IRQ messages that
                        // don't want to send any data
                        if (required_length > 0)
                        {
                            avatar_prepare_send();
                        }
                        else
                        {
                            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave IRQ Request Request had length 0!! --");
                            // BUG 3386 Set IRQ request length to zero - stops looping when in standby
                            memcpy(avatar_data.current_packet->data, &required_length, sizeof(uint16_t));
                        }
                    }
                    else
                    {
                        // BUG 3576 - It looks like when we re-use this packet it never gets freed down the line.
                        //  By freeing it here it is in theory open to being re-allocated, however as the IRQ sequence
                        //  is executed before any other messages I think we should be safe
                        avatar_packet_free(avatar_data.current_packet);
                        
                        // Reuse the current packet (which is an IRQ message) to
                        // request more data space.  Things like length etc will
                        // already be set
                        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave IRQ Request Request more data %d --", required_length);
                        memcpy(avatar_data.current_packet->data, &required_length, sizeof(uint16_t));
                    }

                    break;
                }
                
                case AVATAR_PACKET_TYPE_MSG:
                {
                    //TODO - NEED TO WORK OUT HOW GENERIC CODE CAN ACCESS FIXED structs like I2C2CONbits.
                    bool overflow_detected = false;
                    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave MSG %d %d %d rbf:%d i2cov:%d --", avatar_data.current_header.packet_id, avatar_data.current_header.frame_id, avatar_data.current_header.packet_length, I2C2STATbits.RBF, I2C2STATbits.I2COV);
                    if (I2C2STATbits.I2COV == 1)
                    {
                        overflow_detected = true;
                        I2C2STATbits.I2COV = 0;
                        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave MSG Request i2cov cleared:%d --", I2C2STATbits.RBF, I2C2STATbits.I2COV);
                    }
                    //logger_process();

                    // We have had problems where the the SEND_END Interrupt has been getting lost
                    // Not sure if iot was die to timing, but seems to have been 
                    // helped by not sending messages before the avatar is ready
//TODO-SEND                    if (avatar_data.sending_packet)
//                    {
//                        avatar_data.sending_packet = false;
//
//                        AVATAR_PACKET* packet = avatar_queue_get_packet(&avatar_data.send_queue);
//                        logger_log_debugf(LOGGER_CODE_AVATAR + 4, "AVATAR Slave MSG Request Fix Send %d --", packet->header.frame_id);
//                        
//                        avatar_packet_free(avatar_data.current_packet);
//                        // Check its not an IRQ packet requesting more data for the current
//                        // packet in the send queue
//                        if (avatar_packet_header_get_type(&avatar_data.current_header) != AVATAR_PACKET_TYPE_IRQ)
//                        {
//                            avatar_queue_advance_read(&avatar_data.send_queue);
//                        }
//                    }
//                    else
//                    {
//                        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave MSG Request A --");
//                    }
                    
                    // If ACK - increment frame_id
                    //TODO sounds like we dont increment frame Id until we get an ack back
                    // So if the  message is an ACK then increment the frame Id , otherwise resend the message upto 3 times
                    
                    if (avatar_packet_header_is_ack(&avatar_data.current_header))
                    {
                        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave MSG is ACK --");
                        // Check if this is an ACK to a send packet - if it is then tidyup 
                        if (avatar_data.current_send_packet != NULL)
                        {
                            avatar_data.sending_packet = false;
            
                            // Free the send packet if there is one.....
                            avatar_packet_free(avatar_data.current_send_packet);

                            // advance the send queue
                            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave callback SEND END %d %d --", avatar_data.current_packet->index, avatar_data.current_packet->header.frame_id);
#ifdef AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE
                            avatar_data.allocated_message_size -= avatar_packet_header_get_length(&avatar_data.current_packet->header);
                            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Message Size at free %d %d --", avatar_data.allocated_message_size, avatar_packet_header_get_length(&avatar_data.current_packet->header));
#endif
                
                            avatar_queue_advance_read(&avatar_data.send_queue);
                            
                            avatar_data.current_send_packet = NULL;
                        }

                        // See if we need to process a rule callback
                        avatar_process_rule_process_active();

                        //TODO Increment frame Id
                    }
                    else if (avatar_packet_header_is_nack(&avatar_data.current_header))
                    {
                        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave MSG is NACK --");
                        // TODO resend message....
                    }
                    // Workout correct response
                    else if ((avatar_packet_header_get_data_length(&avatar_data.current_header) > 0) &&
                        (avatar_data.current_packet != NULL))
                    {
                        if (!overflow_detected)
                        {
                            // Put the packet onto the processing queue
                            avatar_prepare_for_processing(avatar_data.current_packet);

                            // Setup ACK to be returned to the avatar
                            avatar_setup_ack();
                        }
                        else
                        {
                            // Free the packet
                            avatar_packet_free(avatar_data.current_packet);
                            
                            // Set up a NACK
                            avatar_setup_nack();
                        }
                    }

                    break;
                }
                
                default:
                    break;
            }
            
            break;
        }
        
        case I2C_SLAVE_STATE_SEND_END:
        {
//TODO-SEND           //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave callback SEND END %s --", avatar_data.current_packet == NULL ? "ACK" : "RESPONSE"); 
////            if (avatar_data.current_packet != NULL)
////            {
////                logger_log_debugf(LOGGER_CODE_AVATAR + 4, "AVATAR Slave callback SEND END %d %d %d --", avatar_data.current_packet->state, avatar_data.current_packet->index, avatar_data.current_packet->header.frame_id);
////            }
//            avatar_data.sending_packet = false;
//            
//            // Free the send packet if there is one.....
//            avatar_packet_free(avatar_data.current_packet);
//            
//            // release the IRQ
//            //avatar_end_send_sequence();
//            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave callback SEND END %d --", avatar_data.current_packet != NULL ? avatar_data.current_packet->header.frame_id : -1);
//            
//            // advance the send queue
//            if (avatar_data.current_packet != NULL)
//            {
//                // Check its not an IRQ packet requesting more data for the current
//                // packet in the send queue
//                if (avatar_packet_header_get_type(&avatar_data.current_header) != AVATAR_PACKET_TYPE_IRQ)
//                {
//                    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave callback SEND END %d %d --", avatar_data.current_packet->index, avatar_data.current_packet->header.frame_id);
//#ifdef AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE
//                    avatar_data.allocated_message_size -= avatar_packet_header_get_length(&avatar_data.current_packet->header);
//                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Message Size at free %d %d --", avatar_data.allocated_message_size, avatar_packet_header_get_length(&avatar_data.current_packet->header));
//#endif
//                
//                    avatar_queue_advance_read(&avatar_data.send_queue);
//                }
//            }
            
            break;
        }
        
        default:
        {
            break;
        }
    }

    return;
}

/**-----------------------------------------------------------------------------
 * Called to clear the interrupt pin -  to be called once the Avatar has responded 
 * to the communication request raised by avatar_interrupt_trigger()
 */
static void avatar_end_send_sequence()
{
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave Ending Send sequence --");
    PLIB_PORTS_PinSet(AVATAR_COMMUNICATION_PORTS_ID, 
                      AVATAR_COMMUNICATION_PORT_CHANNEL, 
                      AVATAR_COMMUNICATION_PIN);    
}

/**-----------------------------------------------------------------------------
 * Puts the packet at the head of the send queue ready for transmission
 */
static void avatar_prepare_send()
{
    AVATAR_PACKET* packet = avatar_queue_get_packet(&avatar_data.send_queue);
    if (packet != NULL)
    {
        memcpy(&avatar_data.current_header, &packet->header, sizeof(AVATAR_PACKET_HEADER));

        avatar_data.data_ptr = (uint8_t*)&avatar_data.current_header;
        avatar_data.message_pos = 0;
        avatar_data.data_offset = 0;

        avatar_data.current_packet = packet;
        
        avatar_data.current_send_packet = packet;
    }  
}

/**-----------------------------------------------------------------------------
 * Message specific processing of avatar packets
 * @param packet
 */
static void avatar_process_packet(AVATAR_PACKET* packet)
{
    // First get the type of message
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave PROCESSING Start ----------------------------------------------"); 
    //logger_process();
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave PROCESSING 0x%04X %s --", message_id, avatar_message_id_string(message_id));
    
    // Check whether we already know about the FrameID of this packet
    // Yes - Retrieve the Message from the message cache
    // No - setup a new message   
    
    // Check if we need to hang onto this Message
    // Control Flags FEND - NO - Discard
    //                      YES - Save
    if (timers_has_expired(TIMER_INDEX_AVATAR_VOLUME_CHANGE)) // Suppress volume messages
    {
        avatar_packet_print(packet, "Received");
    }
    
    AVATAR_MESSAGE* message = avatar_message_for_packet(packet);

    if (message != NULL)
    {
        avatar_process_rule_check(message, AVATAR_PROCESS_RULE_ACTION_PRE);

        avatar_message_process(message);
        
        avatar_process_rule_check(message, AVATAR_PROCESS_RULE_ACTION_POST);
        
        // Add message back to processing queue
        avatar_message_processed(message);
    }
}

/**-----------------------------------------------------------------------------
 * Initialises the avatar process stack note this can not be any bigger than AVATAR_PROCESS_RULE_MAX
 */
static void avatar_process_rules_initialise()
{
    // ARCAM Avatar implementation
    avatar_process_rule_add(AVATAR_MESSAGE_ID_STATE_CHANGE, AVATAR_STATE_ID_SYSTEM_UP_STATE, avatar_system_ready_complete, AVATAR_PROCESS_RULE_ACTION_PRE);  // System Ready State Change
    avatar_process_rule_add(AVATAR_MESSAGE_ID_GET_STATES_RESPONSE, AVATAR_STATE_ID_FIRMWARE_VERSION, avatar_handshake_complete, AVATAR_PROCESS_RULE_ACTION_PRE);  // Avatar System Info Get States Response (requires 3 bits of info, assume Firmware version first (0x0001)))
//    avatar_process_rule_add(AVATAR_MESSAGE_ID_SET_STATES_RESPONSE, AVATAR_STATE_ID_VOLUME, avatar_set_volume_complete, AVATAR_PROCESS_RULE_ACTION_POST);  // Avatar System Info Get States Response (requires 3 bits of info, assume Firmware version first (0x0001)))
}

/**-----------------------------------------------------------------------------
 * Adds a packet to the specified queue
 * @param queue
 * @param packet
 */
static void avatar_queue_add_packet(AVATAR_PACKET_QUEUE* queue, AVATAR_PACKET* packet)
{
//    if (&avatar_data.processing_queue == queue)
//    {
//    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet %d Add to Queue %s at %d --", packet->index, queue->name, queue->write_pos);
//    }

    // Add to the queue
    queue->queue[queue->write_pos] = packet;
    
    // and increment the write position
    queue->write_pos++;
    queue->write_pos %= queue->size;
    
    if (queue->write_pos == queue->read_pos)
    {
        logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Queue overflow %s %d %d --", queue->name, queue->write_pos, queue->read_pos);
        
        // Reset write pos so it keeps failing to highlight the problem
        if (queue->write_pos > 0)
        {
            queue->write_pos--;
        }
        else
        {
            queue->write_pos = queue->size - 1;
        }
    }
    else
    {
        uint16_t free_slots = ((queue->read_pos + queue->size) - queue->write_pos) % queue->size;
        if (free_slots * 5 < queue->size)
        {
            logger_log_warningf(LOGGER_CODE_AVATAR, "AVATAR Queue approaching capacity %s %d/%d --", queue->name, queue->size - free_slots, queue->size);
        }
        else
        {
            //logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Queue add packet capacity %s %d/%d %d %d --", queue->name, queue->size - free_slots, queue->size, queue->write_pos, queue->read_pos);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Moves the queue to the next read position
 * @param queue
 */
static void avatar_queue_advance_read(AVATAR_PACKET_QUEUE* queue)
{
    queue->read_pos++;
    queue->read_pos %= queue->size;
    
//    if (&avatar_data.processing_queue == queue)
//    {
//        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Queue advance_read %s w: %d r: %d --", queue->name, queue->write_pos, queue->read_pos);
//    }
}

/**-----------------------------------------------------------------------------
 * Gets the current packet from the queue
 * @param queue
 * @return 
 */
static AVATAR_PACKET* avatar_queue_get_packet(AVATAR_PACKET_QUEUE* queue)
{
    uint16_t read_pos = queue->read_pos;
    
//    if ((&avatar_data.send_queue == queue) && (queue->queue[queue->read_pos] != NULL))
//    {
//        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR avatar_queue_get_packet %d w: %d r: %d --", queue, queue->write_pos, queue->read_pos);
//    }
    
    return queue->queue[read_pos]; 
}

/**-----------------------------------------------------------------------------
 * Initialises the supplied queue
 * @param queue
 */
static void avatar_queue_initialise(AVATAR_PACKET_QUEUE* queue, char* name)
{
    memcpy(queue->name, name, sizeof(queue->name));
    
    queue->size = sizeof(queue->queue) / sizeof (AVATAR_PACKET*);
    
    int x;
    for (x = 0; x < queue->size; x++)
    {
        queue->queue[x] = NULL;
    }
}
      
/**-----------------------------------------------------------------------------
 * Used by app_commands_process_volume to indicate that it is processing an
 * Avatar mute message
 */
void avatar_set_processing_mute(bool on)
{
    avatar_data.is_processing_mute = on;
}
     
/**-----------------------------------------------------------------------------
 * Used by app_commands_process_volume to indicate that it is processing an
 * Avatar volume message
 */
void avatar_set_processing_volume(bool on)
{
    avatar_data.is_processing_volume = on;
}

/**-----------------------------------------------------------------------------
 * Sets up the data for a standard ACK
 */
static void avatar_setup_ack()
{
    avatar_data.data_ptr = (uint8_t*)&avatar_data.current_header;
    avatar_data.message_pos = 0;
    avatar_data.data_offset = 0;
    
    // Queue up an acknowledgement
    avatar_packet_header_initialise(&avatar_data.current_header);
    avatar_data.current_header.packet_length = 0x0008;              
    avatar_data.current_header.control_flags = 0xA3;   // ACK, Frame end, MSG
    avatar_data.current_header.frame_id = avatar_data.frame_id++;   
    avatar_data.current_header.packet_id = 0x00;   
    
    avatar_data.current_packet = NULL;
}

/**-----------------------------------------------------------------------------
 * Sets up the data for a standard NACK
 */
static void avatar_setup_nack()
{
    avatar_data.data_ptr = (uint8_t*)&avatar_data.current_header;
    avatar_data.message_pos = 0;
    avatar_data.data_offset = 0;
    
    // Queue up an acknowledgement
    avatar_packet_header_initialise(&avatar_data.current_header);
    avatar_data.current_header.packet_length = 0x0008;              
    avatar_data.current_header.control_flags = 0x63;   // NACK, Frame end, MSG
    avatar_data.current_header.frame_id = avatar_data.frame_id++;   
    avatar_data.current_header.packet_id = 0x00;   
    
    avatar_data.current_packet = NULL;
}

/**-----------------------------------------------------------------------------
 * Called to reset the Avatar 
 */
static void avatar_start()
{
    PLIB_PORTS_PinSet(AVATAR_RESET_PORTS_ID, 
                      AVATAR_RESET_PORT_CHANNEL, 
                      AVATAR_RESET_PIN);    
}

/**-----------------------------------------------------------------------------
 * Called to inform the Avatar (master) that we wish to communicate with it.  The 
 * Avatar should respond with a start of read message
 */
static void avatar_start_send_sequence()
{
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave Starting Send sequence --");
    //logger_process();
    
//    PLIB_PORTS_PinDirectionOutputSet(AVATAR_COMMUNICATION_PORTS_ID, 
//                        AVATAR_COMMUNICATION_PORT_CHANNEL, 
//                        AVATAR_COMMUNICATION_PIN);
    
    PLIB_PORTS_PinClear(AVATAR_COMMUNICATION_PORTS_ID, 
                        AVATAR_COMMUNICATION_PORT_CHANNEL, 
                        AVATAR_COMMUNICATION_PIN);    
}

/**-----------------------------------------------------------------------------
 * Invoked when the handshaking is deemed to be complete
 * @param rule
 * @param packet
 */
static void avatar_system_ready_complete(AVATAR_PROCESS_RULE* rule, AVATAR_MESSAGE* message)
{
    if (rule->active)
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave Requesting System Info --");
        avatar_request_system_info();
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used to set up avatar once handshaking is complete etc..
 * @return 
 */
void avatar_configure()
{
    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Configure --");
    
#if APP_FEATURE_DIRAC == APP_FEATURE_ON
    // Configure dirac
    dirac_configure();
#endif
    
    if (avatar_data.subscriptions != NULL)
    {
        // Subscribe to messages
        AVATAR_PACKET* packet = avatar_get_packet(AVATAR_MESSAGE_ID_SUBSCRIBE, 40);
        if (packet != NULL)
        {
            int x;
            for (x = 0; x < avatar_data.subscriptions_len; x++)
            {
                avatar_packet_add_value(packet, avatar_data.subscriptions[x]);
            }

            avatar_send_packet(packet);
        }
    }
    
    // Move into Rule where Subscribe response says config is complete...
    // avatar_data.state = AVATAR_STATE_RUNNING; - Now done in subscribe response
    
    bool active;
    settings_get(SETTING_CONTROL4_SDDP_ACTIVE, &active);
    avatar_send_control4_discovery((AVATAR_CONTROL4_DISCOVERY)active);
    //TODO REMOVE - Crestron 29/01/2020settings_get(SETTING_CRESTRON_DISCOVERY_ACTIVE, &active);
    //TODO REMOVE - Crestron 29/01/2020avatar_send_crestron_discovery((AVATAR_CRESTRON_DISCOVERY)active);
    
    uint8_t mqa_decoder_level;
    settings_get(SETTING_MQA_DECODER_LEVEL, (uint8_t*)&mqa_decoder_level);
    avatar_send_mqa_decoder_level((AVATAR_MQA_DECODER_LEVEL) mqa_decoder_level);

    // Request the serial number
    avatar_request_serial_number();
    
    // Request the region
    avatar_request_region();
}
    
/**-----------------------------------------------------------------------------
 * Returns the device specific during initialisation
 * @return 
 */
AVATAR_DEVICE avatar_get_device()
{
    return avatar_data.device;
}
    
/**-----------------------------------------------------------------------------
 * Gets a packet for use with avatar_send_packet
 * @param min_size the minimum size of the packet used to allocate data
 * @return 
 */
AVATAR_PACKET* avatar_get_packet(AVATAR_MESSAGE_ID message_id, uint16_t min_size)
{
    AVATAR_PACKET* result = NULL;

    // Setup a header
    AVATAR_PACKET_HEADER header;
    avatar_packet_header_initialise(&header);

    // Set control flags (FEND, MSG, No ACK/NACK)
    header.control_flags = AVATAR_CONTROL_FLAGS_TYPE_FEND | AVATAR_CONTROL_FLAGS_TYPE_MSG;
    header.frame_id = avatar_data.frame_id++; 
    header.packet_id = 0;
    
    // Set length (this is needed so it knows how much data to allocate)
    header.packet_length = min_size;
                
    // Allocate a packet
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet Allocating Response --");
    result = avatar_packet_allocate(&header, false);

    if (result != NULL)
    {
        // Populate packet data
        avatar_packet_initialise_message(result, message_id);
    }
    
    return result;
}
/**-----------------------------------------------------------------------------
 * Gets the current avatar processing state
 * @return 
 */
AVATAR_STATE avatar_get_state()
{
    return avatar_data.state;
}

/**-----------------------------------------------------------------------------
 * Gets the avatar upgrade option
 */
AVATAR_UPGRADE_OPTION avatar_get_upgrade_option()
{
    AVATAR_UPGRADE_OPTION result;
    
    settings_get(SETTING_AVATAR_UPGRADE_OPTION, &result);
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Performs handshaking process with the avatar.
 * Returns true when handshaking complete
 */
bool avatar_handshake()
{
    bool result = avatar_data.state > AVATAR_STATE_HANDSHAKE;

#if AVATAR_I2C_ENABLED   

    if (avatar_data.state == AVATAR_STATE_INITIALISE)
    {
        avatar_set_state(AVATAR_STATE_HANDSHAKE);
    }

#endif
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Initialises the avatar module
 */
void avatar_initialise(AVATAR_DEVICE device, 
            AVATAR_CALLBACK message_callback,
            AVATAR_STATE_ID* subscriptions,
            uint8_t subscriptions_len)
{
    avatar_data.device = device;
    avatar_data.subscriptions = subscriptions;
    avatar_data.subscriptions_len = subscriptions_len;

    avatar_stop();
    
    // Set the IRQ pin set to its initial condition.
    avatar_end_send_sequence();
    
    // Initialise the packet module
    avatar_packet_initialise();

    // Initialises the Message module
    avatar_message_initialise_cache();    
    
    // Initialise rules
    avatar_process_rules_initialise();
    
    // Set the callback handler
    avatar_message_set_callback(message_callback);
    
    // Initialise processing queue
    avatar_queue_initialise(&avatar_data.processing_queue, "RECEIVE");
    avatar_queue_initialise(&avatar_data.send_queue, "SEND");
    
#if APP_FEATURE_DIRAC == APP_FEATURE_ON
    // Dirac Initialise
    dirac_initialise();
#endif
    
#if AVATAR_I2C_ENABLED   
    
    avatar_data.i2c_handle = i2c_slave_open(AVATAR_I2C_MODULE);
    if (DRV_HANDLE_INVALID != avatar_data.i2c_handle)
    {
        //DRV_I2C_SlaveCallbackSet( avatar_data.i2c_handle, avatar_i2c_slave_callback, 0);
        
        avatar_start();
        
        logger_log_info(LOGGER_CODE_AVATAR, "AVATAR: I2C Open --");
        
        // Start handshaking
        avatar_set_state(AVATAR_STATE_HANDSHAKE);
    }
    else
    {
        avatar_set_state(AVATAR_STATE_ERROR);
        logger_log_error(LOGGER_CODE_AVATAR, "AVATAR: I2C Open --");
    }
   
#endif
}
    
/**-----------------------------------------------------------------------------
 * Determines whether the avatar has outstanding messages/packets to process (send/receive)
 */
bool avatar_is_active()
{
    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR %d %d %d %d --", avatar_data.processing_queue.read_pos, avatar_data.processing_queue.write_pos, avatar_data.send_queue.read_pos, avatar_data.send_queue.write_pos);
    return ((avatar_data.processing_queue.read_pos != avatar_data.processing_queue.write_pos) ||
            (avatar_data.send_queue.read_pos != avatar_data.send_queue.write_pos));
            
}   
    
/**-----------------------------------------------------------------------------
 * Determines whether the avatar is in a state to be communicated with
 */
bool avatar_is_loaded()
{
    bool result = true;

#if AVATAR_I2C_ENABLED   
   
    result = (bool)PLIB_PORTS_PinGet(AVATAR_READY_PORTS_ID, 
                                     AVATAR_READY_PORT_CHANNEL, 
                                     AVATAR_READY_PIN);    

#endif
    
    return result;
}
      
/**-----------------------------------------------------------------------------
 * Set to true while the Avatar is processing a volume message.  Used to stop 
 * mute message being sent back
 */
bool avatar_is_processing_mute()
{
    return avatar_data.is_processing_mute;
}
      
/**-----------------------------------------------------------------------------
 * Set to true while the Avatar is processing a volume message.  Used to stop 
 * volume message being sent back
 */
bool avatar_is_processing_volume()
{
    return avatar_data.is_processing_volume;
}
     
/**-----------------------------------------------------------------------------
 * Determines whether the avatar has outstanding messages/packets to process (send/receive)
 */
bool avatar_is_running()
{
    return avatar_get_state() == AVATAR_STATE_RUNNING;
}   

/**-----------------------------------------------------------------------------
 * Adds the packet to the processing queue
 * @param packet
 */
void avatar_prepare_for_processing(AVATAR_PACKET* packet)
{
    // Set the packet state to make sure it doesn't get re-allocated
    packet->state = AVATAR_PACKET_STATE_READY_TO_PROCESS;
                        
    // Add to the queue
    avatar_queue_add_packet(&avatar_data.processing_queue, packet);
}
   
/**-----------------------------------------------------------------------------
 * Perform Avatar processing
 */
bool avatar_process()
{
    bool result = false;
    
    // Work out if there is anything to process - probably need frame number to decide on order
    switch (avatar_data.state)
    {
        case AVATAR_STATE_INITIALISE:
        {
            // Were just waiting for the Avatar to come on line
            break;
        }
        
        case AVATAR_STATE_HANDSHAKE:
        case AVATAR_STATE_CONFIGURATION:
        case AVATAR_STATE_RUNNING:
        {
            // We've now had at least one message so start processing.
            // See if we've anything to send
            AVATAR_PACKET* packet = avatar_queue_get_packet(&avatar_data.send_queue);

//            if (avatar_data.sending_packet)
//            {
//                logger_log_debugf(LOGGER_CODE_AVATAR + 4, "AVATAR Packet TRYING TO SEND %d %d --", 
//                        packet != NULL ? packet->header.frame_id : -1,
//                        packet != NULL ? packet->state : -1);
//                //avatar_data.sending_packet = false;;
//             }

            if ((packet != NULL) && (packet->state == AVATAR_PACKET_STATE_READY_TO_SEND))
            {
                packet->state = AVATAR_PACKET_STATE_SENDING;
                
                if (timers_has_expired(TIMER_INDEX_AVATAR_VOLUME_CHANGE)) // Suppress volume messages
                {
                    avatar_packet_print(packet, "Sending");
                }
                
                avatar_start_send_sequence();
                // Set sending_packet to true - this is included because during the transfer of the upgrade
                // files the I2C SEND END state was not being triggered (although you could see it in the 
                // SAELEA)
                avatar_data.sending_packet = true;
                
                //REMOVED as we are only using the result to check for messages coming in..result = true;
            }
           
            // Get the current packet for processing in the queue
            packet = avatar_queue_get_packet(&avatar_data.processing_queue);
            
            if ((packet != NULL) && (packet->state == AVATAR_PACKET_STATE_READY_TO_PROCESS))
            {
                packet->state = AVATAR_PACKET_STATE_PROCESSING;

                avatar_process_packet(packet);
                
                // Note we refresh the packet pointer as it appears that sometimes this
                // can be set to NULL (assuming this is due to interrupts!) meaning
                // that the packet was not being freed
                packet = avatar_queue_get_packet(&avatar_data.processing_queue);
#ifdef AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE
                avatar_data.allocated_message_size -= avatar_packet_header_get_length(&packet->header);
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Message Size at free %d %d --", avatar_data.allocated_message_size, avatar_packet_header_get_length(&packet->header));
#endif
                // Free the packet for reuse
                avatar_packet_free(packet);
            
                avatar_queue_advance_read(&avatar_data.processing_queue);
                
                result = true;
            }
            
            if (avatar_data.state == AVATAR_STATE_RUNNING)
            {
#if APP_FEATURE_DIRAC == APP_FEATURE_ON
                dirac_process();
#endif                
            }
            
            break;
        }
        
        case AVATAR_STATE_ERROR:
        {
            //logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR ERROR! ");
            break;
        }

        default:
            logger_log_warningf(LOGGER_CODE_AVATAR, "AVATAR Invalid state (%d) encountered in avatar_process() --", avatar_data.state);
            break;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Resets the avatar queues when going into standby //STANDBY
 */
void avatar_reset_queues()
{
    // Set the IRQ pin set to its initial condition.
    //avatar_end_send_sequence();
    
    // Initialise the packet module
    avatar_packet_initialise();

    // Initialises the Message module
    avatar_message_initialise_cache();    
    
    // Initialise processing queue
    avatar_queue_initialise(&avatar_data.processing_queue, "RECEIVE");
    avatar_queue_initialise(&avatar_data.send_queue, "SEND");
    
    avatar_data.processing_queue.read_pos = 0;
    avatar_data.processing_queue.write_pos = 0;
    avatar_data.send_queue.read_pos = 0;
    avatar_data.send_queue.write_pos = 0;
}

/**-----------------------------------------------------------------------------
 * Queues a packet for sending to the avatar
 * @param packet
 */
void avatar_send_packet(AVATAR_PACKET* packet)
{
    // Set packet state
    packet->state = AVATAR_PACKET_STATE_READY_TO_SEND;
    
#ifdef AVATAR_PACKET_ANALYSE_ALLOCATED_SPACE
    avatar_data.allocated_message_size += avatar_packet_header_get_length(&packet->header);
    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Message Size Send %d %d --", avatar_data.allocated_message_size, avatar_packet_header_get_length(&packet->header));
#endif    

    // Add to send queue 
    //logger_log_debugf(LOGGER_CODE_AVATAR + 4, "AVATAR avatar_send_packet frame: %d --", packet->header.frame_id);
    avatar_queue_add_packet(&avatar_data.send_queue, packet);
}

/**-----------------------------------------------------------------------------
 * Sets the internal state - should only be used internally within the avatar module
 * @param state
 */
void avatar_set_state(AVATAR_STATE state)
{
    avatar_data.state = state;
    
    if (state == AVATAR_STATE_RUNNING)
    {
        app_context_module_running(APP_MODULE_AVATAR);
    }
}

/**-----------------------------------------------------------------------------
 * Gets a string representation of the message id.  Note if there is no match this may
 * give unpredictable results
 * @param message_id
 * @return 
 */
char* avatar_state_id_string(AVATAR_STATE_ID state_id)
{
    char empty_string[8] = "";
    sprintf(empty_string, "0x%04X", state_id);
    char* result = (char*)&empty_string[0];
    int8_t idx = -1;
    
    // Find the index of the current value
    int x;
    for (x = 0; x < sizeof(AVATAR_STATE_ID_STRINGS)/sizeof(AVATAR_STATE_ID_STRINGS[0]); x++)
    {
        if (state_id == AVATAR_STATE_ID_VALUES[x])
        {
            result = AVATAR_STATE_ID_STRINGS[x];
            break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Called to reset the Avatar 
 */
void avatar_stop()
{
    PLIB_PORTS_PinClear(AVATAR_RESET_PORTS_ID, 
                        AVATAR_RESET_PORT_CHANNEL, 
                        AVATAR_RESET_PIN);    
}
     
/**-----------------------------------------------------------------------------
 * Returns whether an upgrade is available
 * @return 
 */
bool avatar_upgrade_available()
{
    bool result = false;
    
    if (avatar_message_get_upgrade_status() <= AVATAR_UPGRADE_STATUS_DEVICE_UPGRADE_AVAILABLE)
    {
        result = true;
    }

    return result;
}
    
/**-----------------------------------------------------------------------------
 * 
 */
void avatar_wakeup()
{
    avatar_set_state(AVATAR_STATE_HANDSHAKE);
    app_context_module_reset(APP_MODULE_AVATAR);
    avatar_message_set_sample_rate(AVATAR_SAMPLE_RATE_0); // Force a reset of the sample rate
}
/* *****************************************************************************
 End of File
 */
