/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\avatar_browsing.h"
#include "app_config.h"
#include "utilities\avatar.h"
#include "utilities\display_message.h"
#include "utilities\display_message_data.h"
#include "utilities/menu.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! The ID that is given to the first item at any level in the hierarchy
#define AVATAR_BROWSING_START_ID    1

//! The number of results to retrieve when browsing
#define AVATAR_BROWSING_GET_COUNT   AVATAR_BROWSE_RESULTS_MAX * 3

//! Note size 10 is abitary - currently it just needs to hold "USB" - START_FOLDER_MAX_NAME_LEN?
#define AVATAR_BROWSING_START_FOLDER_MAX_NAME_LEN   10

/**
 * Local data
 */
typedef struct
{
//    AVATAR_BROWSE_ITEM parent_item;
//    AVATAR_BROWSE_ITEM browse_items[AVATAR_BROWSE_RESULTS_MAX];
//    uint16_t current_item_id;
    
    AVATAR_BROWSING_STATE state;
    
    //! Records the last avtive browsing state
    AVATAR_BROWSING_STATE exit_state;
    
    uint8_t current_level; 
    AVATAR_BROWSE_ITEM levels[AVATAR_BROWSE_LEVELS];
    
    AVATAR_BROWSE_ITEM browse_items[AVATAR_BROWSING_GET_COUNT];
    
    // This is used to store the Id of the current item while browsing down into a new folder
    uint16_t current_item_id;
    uint16_t top_item_id;
    
    // This is ser when the current items for the folder are returned
    AVATAR_BROWSE_ITEM* current_item;
    AVATAR_BROWSE_ITEM* top_item;
    
    //! The index of the current skip control (0 = no skip controls))
    AVATAR_BROWSING_SKIP_CONTROL skip_control;
    
    //! Indicates whether a track has been played in the current browsing session
    bool has_played_in_session;
    
    // Stores the name of the start folder if specified to avatar_start_browsing
    char start_folder[AVATAR_BROWSING_START_FOLDER_MAX_NAME_LEN];
    
} AVATAR_BROWSING_DATA;

static AVATAR_BROWSING_DATA avatar_browsing_data =
{
    .current_level = 0,
    .current_item = NULL,
    .top_item = NULL,
    .state = AVATAR_BROWSING_STATE_INACTIVE,
    .skip_control = AVATAR_BROWSING_SKIP_CONTROL_NONE,
    .has_played_in_session = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

static void avatar_browsing_update_browse_item(AVATAR_BROWSE_ITEM* item, uint16_t item_id, AVATAR_BROWSE_ITEM_TYPE type, uint8_t* name);

/**-----------------------------------------------------------------------------
 * Clears the stored browsing data
 */
static void avatar_browsing_clear_browsing_data()
{
    int x;
    for (x = 0; x < AVATAR_BROWSING_GET_COUNT; x++)
    {
        AVATAR_BROWSE_ITEM* browse_item = &avatar_browsing_data.browse_items[x];
        memset(browse_item, 0, sizeof(AVATAR_BROWSE_ITEM));
        browse_item->in_use = false;
    }
}
 
/**-----------------------------------------------------------------------------
 * Attempts to locate the browse item matching the supplied id
 * @param item_id
 * @return 
 */
static AVATAR_BROWSE_ITEM* avatar_browsing_item_for_id(uint16_t item_id)
{
    AVATAR_BROWSE_ITEM* result = NULL;
    
    // results should be in ascending order so we should be able to calculate 
    // the ordinal pos just by subtracting the first id from the required one.
    int16_t ordinal_idx = item_id - avatar_browsing_data.browse_items[0].item_id;
    if ((ordinal_idx >= 0) && (ordinal_idx < AVATAR_BROWSING_GET_COUNT))
    {
        result = &avatar_browsing_data.browse_items[ordinal_idx];
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @return 
 */
static uint8_t avatar_browsing_active_browse_items()
{
    uint8_t result = 0;
    int x;
    for (x = 0; x < AVATAR_BROWSING_GET_COUNT; x++)
    {
        if (avatar_browsing_data.browse_items[x].in_use)
        {
            result++;
        }
        else
        {
            break;
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Browsing Results
 * @param message
*/
static bool avatar_browsing_process_folder_results(uint8_t* data)
{
    uint8_t num_items;
    uint8_t start_item = 0;     // The item index where the first item in the data is to be written
    uint16_t offset = sizeof(uint16_t);

    memcpy(&num_items, &data[0], sizeof(uint16_t));
        
    if ((num_items == 1) && (strncmp("Attaching new device", &data[offset + sizeof(uint16_t) + sizeof(uint8_t)], 20) == 0))
    {
        num_items = 0;
    }

    // We have reached the end so retain enough items to fill the whole array
    if (num_items > 0)
    {
        // Get the Id of the first item returned
        uint16_t first_item_id;
        memcpy(&first_item_id, &data[offset], sizeof(uint16_t));
        
        // Check if this is <> the first and last items currently retrieved
        // assume there is never a gap in ids between the items returned and the items current 
        if (num_items < AVATAR_BROWSING_GET_COUNT)
        {
            uint8_t active_item_count = avatar_browsing_active_browse_items();
            if (active_item_count > 0)
            {
                logger_log_debugf(0, "BROWSING Moving Items %d %d %d --", first_item_id, num_items, active_item_count); 
                
                // Check if the first result item is < the first active item, if
                // so move the current items down
                if (first_item_id < avatar_browsing_data.browse_items[0].item_id)
                {
                    // Move the items down to make space for the new items (note
                    // if the returned items overlap the current items we will
                    // only move down enough to accommodate the items that are new)
                    uint8_t num_to_insert = avatar_browsing_data.browse_items[0].item_id - first_item_id;
                    int x;
                    for (x = AVATAR_BROWSING_GET_COUNT - num_to_insert - 1; x >= 0; x--)
                    {
                        logger_log_debugf(0, "BROWSING Copying Down  %d to %d  --", x, x + num_to_insert);  
                        memcpy(&avatar_browsing_data.browse_items[x + num_to_insert], &avatar_browsing_data.browse_items[x], sizeof(AVATAR_BROWSE_ITEM));
                    }

                    // Make sure the unpolulatd ones are set to in_ues = false for when we page back a numbr rof pages
                    for (x = 0; x < num_to_insert; x++)
                    {
                        AVATAR_BROWSE_ITEM* browse_item = &avatar_browsing_data.browse_items[x];
                        browse_item->in_use = false;
                    }
                    
                    start_item = 0;
                }
                else if (first_item_id == avatar_browsing_data.browse_items[active_item_count - 1].item_id + 1)
                {
                    // make space for the new items at the bottom of the list
                    // If they overlap then only move enough
                    uint8_t total_items = first_item_id + num_items - avatar_browsing_data.browse_items[0].item_id;
                    uint8_t num_to_move = total_items - AVATAR_BROWSING_GET_COUNT;
                    
                    int x;
                    for (x = num_to_move; x < AVATAR_BROWSING_GET_COUNT; x++)
                    {
                        logger_log_debugf(0, "BROWSING Copying Up  %d to %d  --", x, x - num_to_move);  
                        memcpy(&avatar_browsing_data.browse_items[x - num_to_move], &avatar_browsing_data.browse_items[x], sizeof(AVATAR_BROWSE_ITEM));
                    }
                    
                    start_item = first_item_id - avatar_browsing_data.browse_items[0].item_id;
                }
                else if (first_item_id != avatar_browsing_data.browse_items[0].item_id)
                {
                    // Wee starting a new range so clear down the items - I still think this is needed but was messing up browsing previous after FWD>FWD>BWD then scrolling previous
                    int x;
                    for (x = 0; x < AVATAR_BROWSING_GET_COUNT; x++)
                    {
                        AVATAR_BROWSE_ITEM* browse_item = &avatar_browsing_data.browse_items[x];
                        if (browse_item->item_id < first_item_id)
                        {
                            browse_item->in_use = false;
                        }
                    }
                    start_item = 0;
                }
            }
        }
    }
    else
    {
        // Set that we have reached the end of the folder???????
    }
    
    //logger_log_debugf(0, "BROWSING %d %d %d %d %d %d %d %d %d --", data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7],  data[8]);
    
    logger_log_debugf(0, "BROWSING Num Items %d Start At %d --", num_items, start_item);    
    logger_process();
    
    // The Avatar will send empty data items when it reaches the end of the data rather than 0 (to be raised as a call))
    uint16_t first_id;
    memcpy(&first_id, &data[offset], sizeof(uint16_t));
    
    if (first_id > 0)
    {
        // Extract the folder info
        int x;
        for (x = 0; x < AVATAR_BROWSING_GET_COUNT; x++)
        {
            AVATAR_BROWSE_ITEM* browse_item = &avatar_browsing_data.browse_items[x];

            if ((x >= start_item) && (x < start_item + num_items))
            {
                logger_log_debugf(0, "BROWSING Populating  %d %d  --", x, start_item);    
                memcpy(&browse_item->item_id, &data[offset], sizeof(uint16_t));
                browse_item->type = (AVATAR_BROWSE_ITEM_TYPE) data[offset + sizeof(uint16_t)];
                memcpy(browse_item->name, &data[offset + sizeof(uint16_t) + sizeof(uint8_t)], AVATAR_BROWSE_ITEM_MAX_LENGTH);

                browse_item->in_use = true;

                offset += (sizeof(uint16_t) + sizeof(uint8_t) + AVATAR_BROWSE_ITEM_MAX_LENGTH);
            }

            logger_log_debugf(0, "BROWSING %d %d %s --", browse_item->item_id, browse_item->type, browse_item->name);
        }
    }
    else
    {
        logger_log_debugf(0, "BROWSING Avatar sent empty data --");
    }
    //logger_log_debugf(0, "BROWSING CURRENT B %d --", avatar_browsing_data.current_item->item_id);
    
    // Set current item and top item
    avatar_browsing_data.current_item = avatar_browsing_item_for_id(avatar_browsing_data.current_item_id);
    avatar_browsing_data.top_item = avatar_browsing_item_for_id(avatar_browsing_data.top_item_id);
    logger_log_debugf(0, "BROWSING CURRENT %d --", avatar_browsing_data.current_item_id);
    logger_log_debugf(0, "BROWSING TOP %d --", avatar_browsing_data.top_item_id);

    int8_t active_item_count = avatar_browsing_active_browse_items();
    int8_t max_top_item = max(active_item_count - AVATAR_BROWSE_RESULTS_MAX, 0);
    logger_log_debugf(0, "BROWSING ACTIVE ITEM COUNT %d --", active_item_count);
    logger_log_debugf(0, "BROWSING MAX TOP %d --", max_top_item);

    
    if (avatar_browsing_data.current_item == NULL)
    {
        if (avatar_browsing_data.current_item_id < avatar_browsing_data.browse_items[0].item_id)
        {
            avatar_browsing_data.current_item = &avatar_browsing_data.browse_items[0];
        }
        else if (avatar_browsing_data.current_item_id > avatar_browsing_data.browse_items[active_item_count - 1].item_id)
        {
            avatar_browsing_data.current_item = &avatar_browsing_data.browse_items[active_item_count - 1];
        }
    }
    
    if (avatar_browsing_data.top_item == NULL)
    {
        if (avatar_browsing_data.top_item_id < avatar_browsing_data.browse_items[0].item_id)
        {
            avatar_browsing_data.top_item = &avatar_browsing_data.browse_items[0];
        }
        else if (avatar_browsing_data.top_item_id > avatar_browsing_data.browse_items[max_top_item].item_id)
        {
            avatar_browsing_data.top_item = &avatar_browsing_data.browse_items[max_top_item];
        }
        else if (avatar_browsing_data.top_item_id > avatar_browsing_data.current_item_id)
        {
            avatar_browsing_data.top_item = avatar_browsing_data.current_item;
        }
        else if (avatar_browsing_data.top_item_id <= avatar_browsing_data.current_item_id - AVATAR_BROWSE_RESULTS_MAX)
        {
            avatar_browsing_data.top_item = avatar_browsing_data.current_item;
        }
    }

    avatar_browsing_data.current_item_id = avatar_browsing_data.current_item->item_id;
    avatar_browsing_data.top_item_id = avatar_browsing_data.top_item->item_id;
    
    // Now make sure the top item displays a whole page
    if (avatar_browsing_data.top_item_id < AVATAR_BROWSING_START_ID)
    {
        avatar_browsing_data.top_item = &avatar_browsing_data.browse_items[0];
        avatar_browsing_data.top_item_id = avatar_browsing_data.top_item->item_id;
    }
    else
    {
        if (avatar_browsing_data.top_item_id > avatar_browsing_data.browse_items[max_top_item].item_id)
        {
            avatar_browsing_data.top_item = &avatar_browsing_data.browse_items[max_top_item];
            avatar_browsing_data.top_item_id = avatar_browsing_data.top_item->item_id;
        }
    }
    
    logger_log_debugf(0, "BROWSING CURRENT %d --", avatar_browsing_data.current_item->item_id);
    logger_log_debugf(0, "BROWSING TOP %d --", avatar_browsing_data.top_item->item_id);

//    avatar_browsing_data.current_item_id = avatar_browsing_data.current_item->item_id;
//    avatar_browsing_data.top_item_id = avatar_browsing_data.top_item->item_id;
    
    // Having done all this check if we want to go to a start folder
    bool update_display = true;
    if (strlen(avatar_browsing_data.start_folder) != 0)
    {
        logger_log_debugf(0, "BROWSING TO START FOLDER %s --", avatar_browsing_data.start_folder);
        
        // Find the folder in the current list that matches this name and browse to it
        // Assumes that he name can be found in the results, given we're just looking for USB/UpNP we should be OK
        int x;
        for (x = 0; x < AVATAR_BROWSING_GET_COUNT; x++)
        {
            AVATAR_BROWSE_ITEM* browse_item = &avatar_browsing_data.browse_items[x];
            
            if (strcmp(browse_item->name, avatar_browsing_data.start_folder) == 0)
            {
                logger_log_debugf(0, "BROWSING TO START FOLDER FOUND %s --", avatar_browsing_data.start_folder);
                
                avatar_browsing_data.current_item_id = AVATAR_BROWSING_START_ID;
                avatar_browsing_data.top_item_id = AVATAR_BROWSING_START_ID;

                avatar_browsing_send_browse_folder(browse_item->item_id, AVATAR_BROWSING_START_ID, AVATAR_BROWSING_GET_COUNT);
            
                update_display = false;
                break;
            }
        }
        
        memset(avatar_browsing_data.start_folder, 0, AVATAR_BROWSING_START_FOLDER_MAX_NAME_LEN);
    }

    if (update_display)
    {
#ifdef DISPLAY_MESSAGE_ACTIVE
    
    display_message_build_screen_data(APP_COMMAND_MENU_MEDIA_BROWSING);
    
#endif    
    }
    
    timers_reset(TIMER_INDEX_BROWSING_SOURCE_CHECK);
}

/**-----------------------------------------------------------------------------
 * Updates a browse item with the data supplied
 * @param item
 * @param id
 * @param name
 */
static void avatar_browsing_update_browse_item(AVATAR_BROWSE_ITEM* item, uint16_t item_id, AVATAR_BROWSE_ITEM_TYPE type, uint8_t* name)
{
    if (item != NULL)
    {
        item->item_id = item_id;
        item->type = type;
        memset(item->name, 0, AVATAR_BROWSE_ITEM_MAX_LENGTH);
        memcpy(item->name, name, AVATAR_BROWSE_ITEM_MAX_LENGTH);
    }
}
/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
 
/**-----------------------------------------------------------------------------
 * Selects the current browse item
 */
void avatar_browsing_browse_current(bool play_all)
{
    AVATAR_BROWSING_SKIP_CONTROL skip_control = avatar_browsing_get_skip_control();
    if (skip_control == AVATAR_BROWSING_SKIP_CONTROL_NONE)
    {
        logger_log_debugf(0, "AVATAR avatar_browsing_browse_current --");
        logger_process();

        AVATAR_BROWSE_ITEM* current_item = avatar_browsing_data.current_item;//avatar_browsing_item_for_id(avatar_browsing_data.current_item_id);

        if (current_item != NULL)
        {
            // Check what the current item is
            switch(current_item->type)
            {
                case AVATAR_BROWSE_ITEM_TYPE_FOLDER:
                {
                    // make the current item  the parent and request its children 
                    if (play_all)
                    {
                        // Play all the tracks in the folder
                        avatar_browsing_set_running_state(AVATAR_BROWSING_STATE_PLAYING);
                        logger_log_debugf(0, "AVATAR BR A --");
                        // tell the avatar to play the track
    //TODO Implement PlayFolder                    avatar_browsing_send_browse_item(avatar_browsing_data.current_item->item_id,
    //                            AVATAR_BROWSE_PLAY_OPTION_ALWAYS_PLAY,
    //                            0);
                    }
                    else
                    {
                        logger_log_debugf(0, "AVATAR BROWSE INTO FOLDER --");

                        avatar_browsing_data.current_item_id = AVATAR_BROWSING_START_ID;
                        avatar_browsing_data.top_item_id = AVATAR_BROWSING_START_ID;

                        avatar_browsing_send_browse_folder(avatar_browsing_data.current_item->item_id,
                                AVATAR_BROWSING_START_ID,
                                AVATAR_BROWSING_GET_COUNT);                    
                    }
                    break;
                }

                case AVATAR_BROWSE_ITEM_TYPE_TRACK:
                {
                    avatar_browsing_set_running_state(AVATAR_BROWSING_STATE_PLAYING);

                    // tell the avatar to play the track
                    logger_log_debugf(0, "AVATAR BROWSE PLAY TRACK --");
                    avatar_browsing_send_play_track(avatar_browsing_data.current_item->item_id);
                    break;
                }

                default:
                    break;
            }
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_AVATAR, "BROWSING Current Item not available %d --", avatar_browsing_data.current_item->item_id);
        }
    }
    else
    {
        switch (avatar_browsing_get_skip_control())
        {
            case AVATAR_BROWSING_SKIP_CONTROL_FBWD:
            {
               logger_log_debugf(0, "AVATAR BROWSE FAST BACK 1 --");
                
                //avatar_browsing_set_skip_control(AVATAR_BROWSING_SKIP_CONTROL_NONE);

                avatar_browsing_data.current_item_id = max((int16_t)avatar_browsing_data.current_item_id - (AVATAR_BROWSE_RESULTS_MAX *5), AVATAR_BROWSING_START_ID);
                avatar_browsing_data.top_item_id = max((int16_t)avatar_browsing_data.top_item_id - (AVATAR_BROWSE_RESULTS_MAX * 5), AVATAR_BROWSING_START_ID);

                avatar_browsing_send_browse_folder(AVATAR_BROWSE_PARENT_FOLDER_ID,
                        max((int16_t)avatar_browsing_data.top_item_id - AVATAR_BROWSE_RESULTS_MAX, AVATAR_BROWSING_START_ID),
                        AVATAR_BROWSING_GET_COUNT);                    
                break;
            }
            
            case AVATAR_BROWSING_SKIP_CONTROL_BACK:
            {
               logger_log_debugf(0, "AVATAR BROWSE BACK 1 --");
                
                //avatar_browsing_set_skip_control(AVATAR_BROWSING_SKIP_CONTROL_NONE);

                avatar_browsing_data.current_item_id = max((int16_t)avatar_browsing_data.current_item_id - AVATAR_BROWSE_RESULTS_MAX, AVATAR_BROWSING_START_ID);
                avatar_browsing_data.top_item_id = max((int16_t)avatar_browsing_data.top_item_id - AVATAR_BROWSE_RESULTS_MAX, AVATAR_BROWSING_START_ID);

                avatar_browsing_send_browse_folder(AVATAR_BROWSE_PARENT_FOLDER_ID,
                        max((int16_t)avatar_browsing_data.top_item_id - AVATAR_BROWSE_RESULTS_MAX, AVATAR_BROWSING_START_ID),
                        AVATAR_BROWSING_GET_COUNT);                    
                break;
            }
            
            case AVATAR_BROWSING_SKIP_CONTROL_EXIT:
            {
                logger_log_debugf(0, "AVATAR BROWSE EXIT SKIP --");
                
                avatar_browsing_set_skip_control(AVATAR_BROWSING_SKIP_CONTROL_NONE);

                avatar_browsing_send_browse_folder(AVATAR_BROWSE_PARENT_FOLDER_ID,
                        max((int16_t)avatar_browsing_data.top_item_id - AVATAR_BROWSE_RESULTS_MAX, 1),
                        AVATAR_BROWSING_GET_COUNT);                    
                break;
            }
            
            case AVATAR_BROWSING_SKIP_CONTROL_FORWARD:
            {
               logger_log_debugf(0, "AVATAR BROWSE FORWARD 3 --");
                
                //avatar_browsing_set_skip_control(AVATAR_BROWSING_SKIP_CONTROL_NONE);

                avatar_browsing_data.current_item_id += AVATAR_BROWSE_RESULTS_MAX,
                avatar_browsing_data.top_item_id += AVATAR_BROWSE_RESULTS_MAX,

                avatar_browsing_send_browse_folder(AVATAR_BROWSE_PARENT_FOLDER_ID,
                        avatar_browsing_data.top_item_id - AVATAR_BROWSE_RESULTS_MAX,
                        AVATAR_BROWSING_GET_COUNT);                    
                break;
            }
            
            case AVATAR_BROWSING_SKIP_CONTROL_FFWD:
            {
               logger_log_debugf(0, "AVATAR BROWSE FAST FORWARD 4 --");
                
                //avatar_browsing_set_skip_control(AVATAR_BROWSING_SKIP_CONTROL_NONE);

                avatar_browsing_data.current_item_id += (AVATAR_BROWSE_RESULTS_MAX * 5),
                avatar_browsing_data.top_item_id += (AVATAR_BROWSE_RESULTS_MAX * 5),
                        
               logger_log_debugf(0, "AVATAR BROWSE FAST FORWARD 4 %d %d --", avatar_browsing_data.current_item_id, avatar_browsing_data.top_item_id);

                avatar_browsing_send_browse_folder(AVATAR_BROWSE_PARENT_FOLDER_ID,
                        avatar_browsing_data.top_item_id - AVATAR_BROWSE_RESULTS_MAX,
                        AVATAR_BROWSING_GET_COUNT);                    
                break;
            }
        }
    }
}
      
/**-----------------------------------------------------------------------------
 * Displays th skip controls
 */
void avatar_browsing_display_skip(AVATAR_BROWSING_SKIP_CONTROL direction)
{
    if (avatar_browsing_active_browse_items() >= AVATAR_BROWSE_RESULTS_MAX)
    {
        avatar_browsing_data.skip_control = direction;

#ifdef DISPLAY_MESSAGE_ACTIVE
    
        display_message_build_screen_data(APP_COMMAND_MENU_MEDIA_BROWSING);

        display_message_send_screen_action(
                    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT, 
                    DISPLAY_MESSAGE_BROWSING_DATAID_SKIP_CONTROLS + avatar_browsing_data.skip_control);    
#endif    
        
    }
}
   
/**-----------------------------------------------------------------------------
 * Populates the supplied buffer with a current breadcrumb representation
 * @param buffer
 * @param len
 */
void avatar_browsing_get_breadcrumb_string(char* buffer, size_t len)
{
    memset(buffer, 0, len);
    
    uint8_t offset = 0;
    
    int x;
    for (x = 0; x <= avatar_browsing_data.current_level; x++)
    {
        memcpy(&buffer[offset], &avatar_browsing_data.levels[x].name, strlen(avatar_browsing_data.levels[x].name));
        offset += strlen(avatar_browsing_data.levels[x].name);
        buffer[offset] = '>';
        offset++;
    }
}
    
/**-----------------------------------------------------------------------------
 * Gets the avatar browsing state that was current when the browsing menu was left
 * This is used to identify when we leave the currently laying screen whether we were browsing beforehand
 * @return 
 */
AVATAR_BROWSING_STATE avatar_browsing_get_exit_state()
{
    return avatar_browsing_data.exit_state;
}
 
/**-----------------------------------------------------------------------------
 * Gets a reference to current items 
 * @return 
 */
AVATAR_BROWSE_ITEM* avatar_browsing_get_items()
{
    return &avatar_browsing_data.browse_items[0];
}
 
/**-----------------------------------------------------------------------------
 * Gets a reference to current items 
 * @return 
 */
AVATAR_BROWSE_ITEM* avatar_browsing_get_current_item()
{
    return avatar_browsing_data.current_item;
}
    
/**-----------------------------------------------------------------------------
 * Gets the current browsing level
 * @return 
 */
uint8_t avatar_browsing_get_current_level()
{
    return avatar_browsing_data.current_level;
}
 
/**-----------------------------------------------------------------------------
 * Gets a reference to the current parent item 
 * @return 
 */
AVATAR_BROWSE_ITEM* avatar_browsing_get_parent_item()
{
    return &avatar_browsing_data.levels[avatar_browsing_data.current_level];
}
 
/**-----------------------------------------------------------------------------
 * Get the current skip control index (0 = No skip menu)
 * @return 
 */
AVATAR_BROWSING_SKIP_CONTROL avatar_browsing_get_skip_control()
{
    return avatar_browsing_data.skip_control;
}
    
/**-----------------------------------------------------------------------------
 * Gets the current avatar browsing state
 * @return 
 */
AVATAR_BROWSING_STATE avatar_browsing_get_state()
{
    return avatar_browsing_data.state;
}
 
/**-----------------------------------------------------------------------------
 * Gets a reference to the item to be displayed at the top of the list items 
 * @return 
 */
AVATAR_BROWSE_ITEM* avatar_browsing_get_top_item()
{
    return avatar_browsing_data.top_item;
}
 
/**-----------------------------------------------------------------------------
 * Checks whether we have actually played a track in this browsing session.
 * @return 
 */
bool avatar_browsing_has_played_in_session()
{
    return avatar_browsing_data.has_played_in_session;
}
 
/**-----------------------------------------------------------------------------
 * Regular check to see if USB stick has been inserted or UPnP services available
 * @param message
*/
void avatar_browsing_process()
{
    if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
    {
//logger_log_debugf(0, "BROWSING Looking for USB/UPnP sources AA %d %d --", avatar_browsing_data.current_level, avatar_browsing_active_browse_items());
        if ((avatar_browsing_data.current_level == 0) && (avatar_browsing_active_browse_items() == 0))
        {
            if (timers_has_expired(TIMER_INDEX_BROWSING_SOURCE_CHECK))
            {
                logger_log_debugf(0, "BROWSING Looking for USB/UPnP sources --");
        
                avatar_browsing_send_browse_back(AVATAR_BROWSE_BACK_LEVEL_ROOT);

                avatar_browsing_data.current_item_id = AVATAR_BROWSING_START_ID;
                avatar_browsing_data.top_item_id = AVATAR_BROWSING_START_ID;
        
                avatar_browsing_send_browse_folder(AVATAR_BROWSE_PARENT_FOLDER_ID, AVATAR_BROWSING_START_ID, AVATAR_BROWSE_RESULTS_MAX * 3);
            }
        }
    }
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Host Name
 * @param message
*/
bool avatar_browsing_process_browse_back(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    //uint16_t data;
    uint8_t data[AVATAR_BROWSE_ITEM_MAX_LENGTH];
           
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_BROWSE_BACK, 
            &encoding, 
            (uint8_t*)data, 
            AVATAR_BROWSE_ITEM_MAX_LENGTH))
    {
        avatar_browsing_clear_browsing_data();
        
        if (avatar_browsing_data.current_level > 0)
        {
            // get the item that was selectd in the curent level when browsing into the folder
            AVATAR_BROWSE_ITEM* prev_level = &avatar_browsing_data.levels[avatar_browsing_data.current_level];
            
            avatar_browsing_data.current_level--;
            
            AVATAR_BROWSE_ITEM* level = &avatar_browsing_data.levels[avatar_browsing_data.current_level];
            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR BROWSE BACK TO %d %s %d --", level->item_id, level->name);
            
            uint16_t current_id = prev_level->item_id;
            
            uint16_t start_item_id = current_id;
            if (start_item_id > (AVATAR_BROWSING_GET_COUNT / 2))
            {
                start_item_id -= (AVATAR_BROWSING_GET_COUNT / 2); 
            }
            else
            {
                start_item_id = AVATAR_BROWSING_START_ID;
            }
                    
            avatar_browsing_data.current_item_id = current_id;
            avatar_browsing_data.top_item_id = current_id;

            avatar_browsing_send_browse_folder(avatar_browsing_data.current_item->item_id,
                    start_item_id,
                    AVATAR_BROWSING_GET_COUNT);                    

//            avatar_browsing_browse_current(false);
        }
        
        result = true;
    }

    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Host Name
 * @param message
*/
bool avatar_browsing_process_browse_folder(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    uint16_t len = 2 + (AVATAR_BROWSING_GET_COUNT * (2 + 1 + AVATAR_BROWSE_ITEM_MAX_LENGTH));  // Item Id, Type, description     
    char data[len];
           
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_BROWSE_FOLDER, 
            &encoding, 
            (uint8_t*)data, 
            len))
    {
        uint16_t num_items;
        
        memcpy(&num_items, &data[0], sizeof(uint16_t));
        
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR BROWSE FOLDER %d --", num_items);
       
        avatar_browsing_process_folder_results(data);
        
        result = true;
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Used to restart a previous playing track
 */
void avatar_browsing_restart_playing()
{
    if (avatar_browsing_has_played_in_session())
    {
        avatar_browsing_set_running_state(AVATAR_BROWSING_STATE_PLAYING);

        app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
        app_context_set_menu(APP_COMMAND_MENU_DEFAULT, DISPLAY_MESSAGE_SCREEN_NAVIGATION_DIRECT);
    }
}

/**-----------------------------------------------------------------------------
 * Move to the Browsing item next to the current one i the hierarchy
 */
void avatar_browsing_select_next()
{
    AVATAR_BROWSING_SKIP_CONTROL skip_control = avatar_browsing_get_skip_control();
    if (skip_control == AVATAR_BROWSING_SKIP_CONTROL_NONE)
    {
                logger_log_debugf(0, "BROWSING NEXT B %d %d --", avatar_browsing_data.current_item_id, avatar_browsing_data.top_item_id);
                logger_process();
        if (avatar_browsing_data.current_item != &avatar_browsing_data.browse_items[AVATAR_BROWSING_GET_COUNT - 1])
        {
                logger_log_debugf(0, "BROWSING NEXT C %d %d --", avatar_browsing_data.current_item_id, avatar_browsing_data.top_item_id);
                logger_process();
            // Move the current item on
            avatar_browsing_data.current_item++;

            if (avatar_browsing_data.current_item->in_use)
            {
                // Sort out current item id
                avatar_browsing_data.current_item_id = avatar_browsing_data.current_item->item_id;
                logger_log_debugf(0, "BROWSING NEXT %d %d --", avatar_browsing_data.current_item_id, avatar_browsing_data.top_item_id);
                logger_process();

                if (avatar_browsing_data.current_item_id >= avatar_browsing_data.top_item_id + AVATAR_BROWSE_RESULTS_MAX)
                {
                    avatar_browsing_data.top_item_id = avatar_browsing_data.current_item_id - AVATAR_BROWSE_RESULTS_MAX + 1;

                    avatar_browsing_data.top_item = avatar_browsing_item_for_id(avatar_browsing_data.top_item_id);
                    logger_log_debugf(0, "BROWSING NEXT A %d %d --", avatar_browsing_data.current_item_id, avatar_browsing_data.top_item_id);
                    logger_process();

#ifdef DISPLAY_MESSAGE_ACTIVE

                    display_message_build_screen_data(APP_COMMAND_MENU_MEDIA_BROWSING);

#endif    
                }

                logger_log_debugf(0, "BROWSING NEXT %s %s --", avatar_browsing_data.top_item->name, avatar_browsing_data.current_item->name);
                logger_process();

                //  is this the last item in the current list
                if (avatar_browsing_data.current_item == &avatar_browsing_data.browse_items[AVATAR_BROWSING_GET_COUNT - AVATAR_BROWSE_RESULTS_MAX])
                {
                    // Initiate the call to the avatar to return a set that should contain the new current item.
                    avatar_browsing_send_browse_folder(0, avatar_browsing_data.browse_items[AVATAR_BROWSING_GET_COUNT - 1].item_id + 1, AVATAR_BROWSE_RESULTS_MAX);
                }
                else
                {

#ifdef DISPLAY_MESSAGE_ACTIVE

                    display_message_send_screen_action(
                                DISPLAY_MESSAGE_SCREEN_ACTION_SELECT, 
                                0x8000 + avatar_browsing_data.current_item->item_id);    
#endif    
                }
            }
            else
            {
                // Put current item back
                avatar_browsing_data.current_item--;
            }
        }
// DEBUGGING FOR 3840        else
//        {
//// Initiate the call to the avatar to return a set that should contain the new current item.
//logger_log_debugf(0, "BROWSING NEXT LAST DITCH ATTEMPT %d --", avatar_browsing_data.browse_items[AVATAR_BROWSING_GET_COUNT - 1].item_id + 1);
//logger_process();
//avatar_browsing_send_browse_folder(0, avatar_browsing_data.browse_items[AVATAR_BROWSING_GET_COUNT - 1].item_id + 1, AVATAR_BROWSE_RESULTS_MAX);
//            
//        }
    }
    else
    {
        logger_log_debugf(0, "BROWSING NEXT SKIP CONTROL %d --", skip_control);
        logger_process();
        skip_control++;
        if (skip_control > AVATAR_BROWSING_SKIP_CONTROL_FFWD)
        {
            skip_control = AVATAR_BROWSING_SKIP_CONTROL_FBWD;
        }
        logger_log_debugf(0, "BROWSING NEXT SKIP CONTROL A %d --", skip_control);
        logger_process();
        
#ifdef DISPLAY_MESSAGE_ACTIVE

        display_message_send_screen_action(
                    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT, 
                    DISPLAY_MESSAGE_BROWSING_DATAID_SKIP_CONTROLS + skip_control);    
#endif    
        
        avatar_browsing_set_skip_control(skip_control);

    }
}

/**-----------------------------------------------------------------------------
 * Move to the Browsing item previous to the current one i the hierarchy
 */
void avatar_browsing_select_previous()
{
    AVATAR_BROWSING_SKIP_CONTROL skip_control = avatar_browsing_get_skip_control();
    if (skip_control == AVATAR_BROWSING_SKIP_CONTROL_NONE)
    {
        if (avatar_browsing_data.current_item->item_id > AVATAR_BROWSING_START_ID) 
        {
            // Move the current item on
            //avatar_browsing_data.current_item--;

            // Sort out current item id
            avatar_browsing_data.current_item_id = avatar_browsing_data.current_item->item_id;

            // Now make it the one before
            avatar_browsing_data.current_item_id--;
            avatar_browsing_data.current_item = avatar_browsing_item_for_id(avatar_browsing_data.current_item_id);

            logger_log_debugf(0, "BROWSING PREV %d %d --", avatar_browsing_data.current_item_id, avatar_browsing_data.top_item_id);
            logger_process();

            if (avatar_browsing_data.current_item_id < avatar_browsing_data.top_item_id)
            {
                avatar_browsing_data.top_item_id = avatar_browsing_data.current_item_id;

                avatar_browsing_data.top_item = avatar_browsing_item_for_id(avatar_browsing_data.top_item_id);
//            logger_log_debugf(0, "BROWSING PREV A %d %d --", avatar_browsing_data.current_item_id, avatar_browsing_data.top_item_id);
//            logger_process();

        #ifdef DISPLAY_MESSAGE_ACTIVE

                display_message_build_screen_data(APP_COMMAND_MENU_MEDIA_BROWSING);

        #endif    
            }

            //logger_log_debugf(0, "BROWSING PREV %s %s --", avatar_browsing_data.top_item->name, avatar_browsing_data.current_item->name);
            //logger_process();

            //  is this getting within AVATAR_BROWSE_RESULTS_MAX of the top item item in the current list
            if ((avatar_browsing_data.top_item == NULL) || (avatar_browsing_data.current_item == &avatar_browsing_data.browse_items[AVATAR_BROWSE_RESULTS_MAX - 1]))
            {
                // Initiate the call to the avatar to return a set that should contain the new current item.
                uint16_t start_id = avatar_browsing_data.browse_items[0].item_id;
                if (start_id > AVATAR_BROWSE_RESULTS_MAX)
                {
                    start_id -= AVATAR_BROWSE_RESULTS_MAX; 
                }
                else
                {
                    start_id = AVATAR_BROWSING_START_ID;
                }
                
                avatar_browsing_send_browse_folder(0, start_id, AVATAR_BROWSE_RESULTS_MAX);
            }
            else
            {
#ifdef DISPLAY_MESSAGE_ACTIVE

                display_message_send_screen_action(
                            DISPLAY_MESSAGE_SCREEN_ACTION_SELECT, 
                            0x8000 + avatar_browsing_data.current_item->item_id);    
#endif    
            }
        }
    }
    else
    {
        logger_log_debugf(0, "BROWSING NEXT SKIP CONTROL %d --", skip_control);
        logger_process();
        skip_control--;
        if (skip_control == AVATAR_BROWSING_SKIP_CONTROL_NONE)
        {
            skip_control = AVATAR_BROWSING_SKIP_CONTROL_FFWD;
        }
        logger_log_debugf(0, "BROWSING NEXT SKIP CONTROL A %d --", skip_control);
        logger_process();
        
#ifdef DISPLAY_MESSAGE_ACTIVE

        display_message_send_screen_action(
                    DISPLAY_MESSAGE_SCREEN_ACTION_SELECT, 
                    DISPLAY_MESSAGE_BROWSING_DATAID_SKIP_CONTROLS + skip_control);    
#endif    
        
        avatar_browsing_set_skip_control(skip_control);
    }
}
    
/**-----------------------------------------------------------------------------
 * makes a Browse request to the avatar
 * @param 
 */
void avatar_browsing_send_browse_back(AVATAR_BROWSE_BACK_LEVEL level)
{
    AVATAR_PACKET* packet = avatar_get_packet(AVATAR_MESSAGE_ID_SET_STATES, AVATAR_ABITARY_SHORT_MESSAGE_LENGTH);
    if (packet != NULL)
    {
        if ((level == AVATAR_BROWSE_BACK_LEVEL_ROOT) &&
            (avatar_browsing_data.current_level > 1))
        {
            // Set current level t 1 so when the results ar received it will be set to 0
            avatar_browsing_data.current_level = 1;
        }

        uint8_t data = level;

        avatar_packet_add_data(packet,
                AVATAR_STATE_ID_BROWSE_BACK,
                AVATAR_ENCODING_UINT8,
                &data,
                sizeof(uint8_t));
        
        // Queue the packet for sending
        avatar_send_packet(packet);
        
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Sending Browse Back %d --", level);
    }
}
    
/**-----------------------------------------------------------------------------
 * makes a Browse request to the avatar, this returns results that may or maynot
 * contain the current item
 * @param 
 */
void avatar_browsing_send_browse_folder(uint16_t folder_id, uint16_t start_item_id, uint16_t num_items)
{
    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Sending Browse Folder %d %d %d --", folder_id, start_item_id, num_items);
    logger_process();
    
    AVATAR_PACKET* packet = avatar_get_packet(AVATAR_MESSAGE_ID_SET_STATES, AVATAR_ABITARY_SHORT_MESSAGE_LENGTH);
    if (packet != NULL)
    {
        uint8_t data[8];
        uint16_t length = 3 * sizeof(uint16_t);
        data[0] = length % 256;
        data[1] = length / 256;
        data[2] = folder_id % 256;
        data[3] = folder_id / 256;
        data[4] = start_item_id % 256;
        data[5] = start_item_id / 256;
        data[6] = num_items % 256;
        data[7] = num_items / 256;
        
        if (folder_id != AVATAR_BROWSE_PARENT_FOLDER_ID)
        {
            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Sending Browse Folder A %d %d %d --", folder_id, start_item_id, num_items);
            logger_process();
    
            // Increase the level
            avatar_browsing_data.current_level++;

            AVATAR_BROWSE_ITEM* current_item = avatar_browsing_item_for_id(folder_id);
            
            // Browse the folder
            memcpy(&avatar_browsing_data.levels[avatar_browsing_data.current_level], current_item, sizeof(AVATAR_BROWSE_ITEM));
            
            avatar_browsing_clear_browsing_data();
        }

        avatar_packet_add_data(packet,
                AVATAR_STATE_ID_BROWSE_FOLDER,
                AVATAR_ENCODING_BIN16,
                data,
                8);
        
//        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Sending Browse Folder B %d %d %d --", folder_id, start_item_id, num_items);
//        logger_process();
        
        // Queue the packet for sending
        avatar_send_packet(packet);
    
        timers_reset(TIMER_INDEX_BROWSING_SOURCE_CHECK);
    }
}
    
/**-----------------------------------------------------------------------------
 * makes a Play Track request to the avatar
 * @param 
 */
void avatar_browsing_send_play_track(uint16_t track_id)
{
    AVATAR_PACKET* packet = avatar_get_packet(AVATAR_MESSAGE_ID_SET_STATES, AVATAR_ABITARY_SHORT_MESSAGE_LENGTH);
    if (packet != NULL)
    {
        uint8_t data[sizeof(uint16_t)];
        data[0] = track_id % 256;
        data[1] = track_id / 256;

        avatar_packet_add_data(packet,
                AVATAR_STATE_ID_PLAY_TRACK,
                AVATAR_ENCODING_UINT16,
                data,
                sizeof(uint16_t));
        
        // Queue the packet for sending
        avatar_send_packet(packet);
     
        // Set marker that we have played a track
        avatar_browsing_data.has_played_in_session = true;
        
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Sending Play Track %d %d --", track_id);
    }
}

/**-----------------------------------------------------------------------------
 * Used to set the Avatar browsing state into running mode (also sets the last_exit state)
 * @param state
 */
void avatar_browsing_set_running_state(AVATAR_BROWSING_STATE state)
{
    avatar_browsing_data.state = state;
    avatar_browsing_data.exit_state = avatar_browsing_data.state;
    
    if (state == AVATAR_BROWSING_STATE_INACTIVE)
    {
        avatar_send_shuffle_state(AVATAR_SHUFFLE_STATE_OFF);
        avatar_send_repeat_state(AVATAR_REPEAT_STATE_NORMAL);
    }
}

/**-----------------------------------------------------------------------------
 * Set the current skip control index (0 = No skip menu)
 * @return 
 */
void avatar_browsing_set_skip_control(AVATAR_BROWSING_SKIP_CONTROL skip_control)
{
    avatar_browsing_data.skip_control = skip_control;
}

/**-----------------------------------------------------------------------------
 * Sets up an Avatar browsing session when the NET input becomes current (via the menu)
 */
void avatar_browsing_start()
{
    avatar_browsing_data.state = AVATAR_BROWSING_STATE_STARTING;
    
    avatar_browsing_clear_browsing_data();

    avatar_browsing_data.skip_control = AVATAR_BROWSING_SKIP_CONTROL_NONE;
    
    avatar_browsing_data.current_item = &avatar_browsing_data.browse_items[0];
    avatar_browsing_data.top_item = &avatar_browsing_data.browse_items[0];

    avatar_send_shuffle_state(AVATAR_SHUFFLE_STATE_OFF);
    avatar_send_repeat_state(AVATAR_REPEAT_STATE_NORMAL);
    
#ifdef DISPLAY_MESSAGE_ACTIVE
    display_message_build_screen_data(APP_COMMAND_MENU_MEDIA_BROWSING);
#endif
}

/**-----------------------------------------------------------------------------
 * Starts browsing the folder hierarchy
 */
void avatar_browsing_start_browsing(char* start_folder_name)
{
    if (avatar_browsing_data.state < AVATAR_BROWSING_STATE_BROWSING)
    {
        logger_log_debugf(0, "AVATAR Starting Browsing --");
        logger_process();
//
//        // Make sure the avatar is stopped
//        avatar_send_player_control(AVATAR_PLAYER_CONTROL_STOP); 

        // Get exit state before setting the new state
        AVATAR_BROWSING_STATE exit_state = avatar_browsing_get_exit_state();
        
        avatar_browsing_set_running_state(AVATAR_BROWSING_STATE_BROWSING);
        avatar_browsing_data.skip_control = AVATAR_BROWSING_SKIP_CONTROL_NONE;
        
        //  If we wern't going to play something when we last exited browsing then start from the top
        if (exit_state != AVATAR_BROWSING_STATE_PLAYING)
        {
            // Check the current input - if its USB then go to the USB input
            // Copy the start folder name
            memset(avatar_browsing_data.start_folder, 0, AVATAR_BROWSING_START_FOLDER_MAX_NAME_LEN);
            if ((start_folder_name != NULL) && (strlen(start_folder_name) < AVATAR_BROWSING_START_FOLDER_MAX_NAME_LEN))
            {
                memcpy(avatar_browsing_data.start_folder, start_folder_name, strlen(start_folder_name));
            }
            
            // Make sure the avatar is stopped
            avatar_send_player_control(AVATAR_PLAYER_CONTROL_STOP); 
            
            avatar_browsing_data.has_played_in_session = false;
            avatar_browsing_data.current_level = 0;
            
            avatar_browsing_update_browse_item(&avatar_browsing_data.levels[avatar_browsing_data.current_level],
                    AVATAR_BROWSE_PARENT_FOLDER_ID,
                    AVATAR_BROWSE_ITEM_TYPE_FOLDER,
                    "ROOT");

            avatar_browsing_send_browse_back(AVATAR_BROWSE_BACK_LEVEL_ROOT);

            avatar_browsing_data.current_item_id = AVATAR_BROWSING_START_ID;
            avatar_browsing_data.top_item_id = AVATAR_BROWSING_START_ID;

            avatar_browsing_send_browse_folder(AVATAR_BROWSE_PARENT_FOLDER_ID, AVATAR_BROWSING_START_ID, AVATAR_BROWSING_GET_COUNT);
        }
        else
        {
            // Redisplay the current directory
            avatar_browsing_send_browse_folder(AVATAR_BROWSE_PARENT_FOLDER_ID,
                   max((int16_t)avatar_browsing_data.top_item_id - AVATAR_BROWSE_RESULTS_MAX, 1),
                   AVATAR_BROWSING_GET_COUNT);                    
       }
    }
}

/**-----------------------------------------------------------------------------
 * Sets up an Avatar browsing session when the NET input becomes current (via the menu)
 */
void avatar_browsing_stop()
{
    avatar_browsing_data.state = AVATAR_BROWSING_STATE_INACTIVE;
}
    
/* *****************************************************************************
 End of File
 */
