/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/avatar_message_processing.h"
#include <stdbool.h>
#include "app_config.h"
#include "crypto/crypto.h"
#include "crypto/src/md5.h"
#include "utilities/avatar.h"
#include "utilities/avatar_browsing.h"
#include "utilities/input_control_defs.h"
#include "utilities/upgrade_metadata.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define AVATAR_FIRMWARE_VERSION_LEN         40
#define AVATAR_MIME_TYPE_LEN                40

#define AVATAR_CHROMECAST_APP_NAME_LEN      16
#define AVATAR_FRIENDLY_NAME_LEN            24

#define AVATAR_SERIAL_NUMBER_LEN            20

//! This must be >= the maximum number of upgrade images contained in an upgrade file
#ifndef AVATAR_UPGRADE_IMAGE_NUM_IMAGES
#define AVATAR_UPGRADE_IMAGE_NUM_IMAGES     2
#endif

typedef bool (*AVATAR_PROCESSING_METHOD)(AVATAR_MESSAGE* message);

typedef struct
{
    //! Holds the ordinal position that this image is in the current upgrade data
    uint8_t index;
    
    //! The name of the product that the upgrade image is associated with - this MUST
    // equal the MODEL_NAME for the current device (21 includes a null terminator))
    char product[AVATAR_UPGRADE_IMAGE_PRODUCT_LEN + 1];
    
    //! The size of the upgrade image (including 22 bytes of meta-data)
    uint32_t size;

    // !The size of the actual upgrade image (without meta-data))
    uint32_t image_data_size;
    
    // Get the Major, Minor versions, Image Size, and Checksum
    uint8_t major_version;

    uint8_t minor_version;

    uint8_t checksum[MD5_DIGEST_SIZE];
        
    AVATAR_UPLOAD_CHECK_VERSION_CALLBACK check_version_callback;
    
    AVATAR_UPLOAD_PREPARE_CALLBACK prepare_callback;

    AVATAR_UPLOAD_DATA_CALLBACK data_callback;

    AVATAR_UPLOAD_END_CALLBACK end_callback;
    
    UPGRADE_METADATA_LOAD_ID load_id;
    
    //! Set to true when the upgrade supplied is a newer version than the current one
    bool needs_upgrade;
    
    //! Keeps track of the bytes written during the upgrade process.
    uint32_t bytes_written;
    
    //! Indicates that the upgrade is complete and the end_callback has been invoked
    bool complete;
    
} AVATAR_UPGRADE_IMAGE_INFO;

typedef struct
{
    //! The upgrade image currently being processed
    //AVATAR_UPGRADE_IMAGE_INFO* current_image;
    
    //! Records the version of the upgrade software - this will be 1.0 for
    //  the original upgrade version.
    uint8_t upgrade_major_version;
    uint8_t upgrade_minor_version;
    
    //! Stores the runtime index of the upgrade image being uploaded.
    uint8_t current_idx;
          
    //! The number of images that have been defined and that we have found in the header
    uint8_t active_image_count;
          
    //! The number of images to be processed
    uint8_t processed_image_count;
    
    //! Avatar Upgrade Status
    AVATAR_UPGRADE_STATUS upgrade_status;
    
   //! Holds the image meta-data from the header of the upgrade file.
    // Currently set to 2 as thats enough for the ST60
     AVATAR_UPGRADE_IMAGE_INFO image_info[AVATAR_UPGRADE_IMAGE_NUM_IMAGES];
         
    //! The number of upgrade images in the Upgrade data file
    uint8_t number_of_images;
    
    //! Maps the image_info to their position on the actual upgrade data
    // So image_map[0] refers to the first image in the upgrade data, NOT necessarily
    // the first image in upgrade_info.image_info
    uint8_t image_map[AVATAR_UPGRADE_IMAGE_NUM_IMAGES];
    
    //! Used to keep track of the number of current data write that are going on to ensure we
    // only send one DATA_RECIEVED message per packet(when we are switching between images)
    uint8_t need_data_process_count;
    
    //! Holds header data where the header straddles 2 data packets
    // The first byte is the amount of data stored
    // the size (25 is abitary but bigger than the current header size (22))
    uint8_t cached_header_info[25];

    //! Set to true while an upgrade is being performed.
    bool in_upgrade;
} AVATAR_UPGRADE_INFO;

/**
 * Local data
 */
typedef struct
{
    AVATAR_CALLBACK callback;
    
    //! The current audio source of the avatar as reported by the Current Audio Source state 0x0015
    AVATAR_AUDIO_SOURCE audio_source;
    
    //! Stores the chromecast app name returned from the 0x0304 request
    char chromecast_app_name[AVATAR_CHROMECAST_APP_NAME_LEN + 1];
    
    //! Stores the firmware version
    char firmware_version[AVATAR_FIRMWARE_VERSION_LEN];
    
    //! Stores the friendly name  app name returned from the 0x0304 request
    // Note this is in UTF-8 format, so apostrophies etc will take more than 1 byte
    char friendly_name[AVATAR_FRIENDLY_NAME_LEN];
    
    //! Stores the mime-type;
    char mime_type[AVATAR_MIME_TYPE_LEN];
    
    //! The current AVatar region
    AVATAR_REGION region;
    
    //! Store the serial number
    char serial_number[AVATAR_SERIAL_NUMBER_LEN];
    
    //! Callback used to process Avatar TCP data
    AVATAR_TCP_CALLBACK tcp_callback;
    
    // The last know Avatar input set using
    AVATAR_INPUT current_input;
    
    AVATAR_MQA_STATE mqa_state;
    
    //! This is set when the Subscribe Response is received. After this a State Change
    //  message is sent by the Avatar.  The State Change processing looks at this to 
    //  move to the RUNNING state after it has been processed.
    volatile bool ready_to_run;
    
    AVATAR_SAMPLE_RATE display_sample_rate;
    
    AVATAR_SAMPLE_RATE sample_rate;
    
    //! Holds the image meta-data from the header of the upgrade file.
    // Currently set to 2 as thats enough for the ST60
    AVATAR_UPGRADE_INFO upgrade_info;
    
    AVATAR_DIRAC_CLIENT_STATUS dirac_client_status;
    
} AVATAR_MESSAGE_DATA;

static AVATAR_MESSAGE_DATA avatar_message_data =
{
    .audio_source = AVATAR_AUDIO_SOURCE_UNKNOWN,
    .sample_rate = AVATAR_SAMPLE_RATE_0,
    .serial_number[0] = 0,
    .tcp_callback = NULL,
    .upgrade_info.upgrade_major_version = 1,
    .upgrade_info.upgrade_minor_version = 0,
    .upgrade_info.in_upgrade = false,
    .upgrade_info.upgrade_status = AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_COMPLETE,
    .upgrade_info.need_data_process_count = 0,
    .upgrade_info.cached_header_info[0] = 0,
    .current_input = AVATAR_INPUT_EXTERNAL,
    .mqa_state = AVATAR_MQA_STATE_NO_MQA,
    .mime_type[0] = 0,
    .ready_to_run = false,
    .region = AVATAR_REGION_UNSUPPORTED,
    .dirac_client_status = AVATAR_DIRAC_CLIENT_STATUS_DISCONNECTED,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations
static bool avatar_message_process_firmware_version(AVATAR_MESSAGE* message);
static bool avatar_message_process_friendly_name(AVATAR_MESSAGE* message);
static bool avatar_message_process_host_name(AVATAR_MESSAGE* message);
static bool avatar_message_process_mac_address(AVATAR_MESSAGE* message);
static void avatar_message_process_packet(AVATAR_MESSAGE* message);
static bool avatar_message_process_region(AVATAR_MESSAGE* message);
static bool avatar_message_process_sample_rate(AVATAR_MESSAGE* message);
static bool avatar_message_process_serial_number(AVATAR_MESSAGE* message);
static bool avatar_message_process_ssid(AVATAR_MESSAGE* message);
static void avatar_message_process_tcp_connected(AVATAR_MESSAGE* message);
static void avatar_message_process_tcp_message(AVATAR_MESSAGE* message);
static bool avatar_message_process_upgrade_data(AVATAR_MESSAGE* message);
static uint32_t avatar_message_process_upgrade_data_packet(uint8_t* data, uint32_t data_length);
static uint8_t avatar_message_process_upgrade_header(uint8_t* header_data);
static uint8_t avatar_message_process_upgrade_image_header(AVATAR_UPGRADE_IMAGE_INFO* image_info, uint8_t* header_data);
static void avatar_message_process_upload_data_callback(UPGRADE_METADATA_LOAD_ID load_id, uint32_t total_size, uint32_t bytes_written, UPGRADE_METADATA_UPGRADE_STATUS status);
static void avatar_message_process_upload_end_callback(UPGRADE_METADATA_LOAD_ID load_id, uint32_t total_size, uint32_t bytes_written, UPGRADE_METADATA_UPGRADE_STATUS status);
static bool avatar_message_process_upgrade_status(AVATAR_MESSAGE* message);
static bool avatar_message_process_wifi_mac_address(AVATAR_MESSAGE* message);

/**-----------------------------------------------------------------------------
 * Generates the internal Transfer complete message
 */
static void avatar_message_generate_transfer_complete()
{
    AVATAR_PACKET* packet = avatar_get_packet(AVATAR_MESSAGE_ID_SET_STATES, AVATAR_ABITARY_SHORT_MESSAGE_LENGTH);
    if (packet != NULL)
    {
        uint8_t data;
        data = AVATAR_UPGRADE_STATUS_DEVICE_UPGRADE_TRANSFER_COMPLETE;

        avatar_packet_add_data(packet,
                AVATAR_STATE_ID_UPGRADE_STATUS,
                AVATAR_ENCODING_UINT8,
                &data,
                1);

        AVATAR_MESSAGE message;
        avatar_message_build(packet, &message);

        // Pass upto the app_command level
        avatar_message_process_packet(&message);
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param message
 * @param method
 * @param processed
 * @return 
 */
static bool avatar_message_process_message(AVATAR_MESSAGE* message, AVATAR_PROCESSING_METHOD method, uint8_t* processed)
{
    if (*processed < avatar_message_get_num_states(message))
    {
        if (method(message))
        {
            *processed++;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Check for an Audio Source change 0x0015
 * @param message
 */
static bool avatar_message_check_audio_source(AVATAR_MESSAGE* message)
{
    bool result = false;
    
       // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    void* data;
    uint16_t len;
    
    // Player status processing
    if (avatar_message_get_state_value(message, 
            AVATAR_STATE_ID_AUDIO_SOURCE, 
            &encoding, 
            &data, 
            &len))
    {
        memcpy(&avatar_message_data.audio_source, data, len);
 
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SELECT AUDIO_SOURCE %d --", avatar_message_data.audio_source);


//        // Now check whether the Audio source is google cast
//        if (avatar_message_data.audio_source == AVATAR_AUDIO_SOURCE_GOOGLECAST)
//        {
//            // Use message 0x0304 Chromecast App NAme to get the name of the streaming service
//            avatar_request_chromecast_app_name();
//        }
        
        // TODO Check if the audio Source is USB - if so then we might want to switch to the NET menu rather than the 
        // AVATAR menu??
        result = true;
    }
    
    return result;
}


/**-----------------------------------------------------------------------------
 * Callback processing for Dirac Client Status
 * @param message
*/
static void avatar_message_process_dirac_client_status(AVATAR_MESSAGE* message)
{
#if APP_FEATURE_DIRAC == APP_FEATURE_ON
   AVATAR_ENCODING encoding;
    uint16_t len;
    uint8_t dirac_client_status;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_DIRAC_CLIENT_STATUS, 
            &encoding, 
            (uint8_t*)&dirac_client_status, 
            sizeof(uint8_t)))
    {
        logger_log_debugf(0, "DIRAC CLIENT STATUS %d --", dirac_client_status);
        avatar_message_data.dirac_client_status = (AVATAR_DIRAC_CLIENT_STATUS)dirac_client_status;
    }
#endif
}

/**-----------------------------------------------------------------------------
 * Callback processing for Dirac Curve names request
 * @param message
*/
static bool avatar_message_process_dirac_curve_names(AVATAR_MESSAGE* message)
{
    bool result = false;
#if APP_FEATURE_DIRAC == APP_FEATURE_ON
    
    // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    uint16_t len = DIRAC_MAX_EQ_CURVES * DIRAC_MAX_EQ_CURVE_NAME_LEN + 1;;
    uint16_t actual_len;
    uint8_t data[len];
            
    actual_len = avatar_message_get_state_data(message, 
                    AVATAR_STATE_ID_DIRAC_CURVE_NAMES, 
                    &encoding, 
                    (uint8_t*)data, 
                    len);
    
    if (actual_len > 0)
    {
        // Get the number of curves_returned (ignore data length in bytes 1 and 2))
        uint8_t num_curves = data[2];

        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR DIRAC CURVE NAMES %d --", num_curves);
        
        // Keep track of active curves
        bool active_curves[DIRAC_MAX_EQ_CURVES];
        memset(active_curves, 0, DIRAC_MAX_EQ_CURVES);  // Set all to false;
        
        // Only remove curves if curve names are supplied, it has a habit of sending a load of 
        // empty curve names if it hasnt initialised properly
        bool reconcile_curves = true;
        
        // Reset all curve enabled flags to false.
        uint8_t dirac_eq_curve_enabled_settings[DIRAC_MAX_EQ_CURVES];
        memset(dirac_eq_curve_enabled_settings, 0, sizeof(uint8_t) * DIRAC_MAX_EQ_CURVES);
        dirac_eq_curve_enabled_settings[0] = true;  
        settings_set(SETTING_DIRAC_CURVE_ENABLED, dirac_eq_curve_enabled_settings);
 
        int x;
        for(x = 0; x < min(num_curves, DIRAC_MAX_EQ_CURVES); x++)
        {
            // Check - This was implemented against v0.7 of the spec and assumes 
            // fixed name lengths and curve Ids matching the ordinal positions of the curves
            char curve_name[DIRAC_MAX_EQ_CURVE_NAME_LEN + 1];
            uint8_t curve_id;
            uint8_t* curve_ptr = data + 3 + (x * (DIRAC_MAX_EQ_CURVE_NAME_LEN + 1)); // Pointer to the start of this curves info
            curve_id = *curve_ptr;
            
            // Only process this if the curve_id is > 0.  The Avatar will send
            // a message saying there are curves but not include any info (curve_id/name)
            // These are generally followed by message that do include the curve names)
            memcpy(curve_name, curve_ptr + 1, DIRAC_MAX_EQ_CURVE_NAME_LEN);
            curve_name[DIRAC_MAX_EQ_CURVE_NAME_LEN] = 0;

            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR DIRAC CURVE NAME %d %d %s --", curve_id, strlen(curve_name),curve_name);

            if (curve_id > 0)
            {
            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR DIRAC CURVE NAME Setting Calibrated --");
                dirac_set_curve_calibrated(curve_id, curve_name);

                active_curves[curve_id] = true;
            }
            else
            {
                // Dont reconcile the curves as we have one of the dodgy messages
                reconcile_curves = false;
            }
        }
        
        if ((num_curves > 0) && reconcile_curves)
        {
            // Now remove any curves that are no longer active (ignore No Curve)
            for(x = 1; x < DIRAC_MAX_EQ_CURVES; x++)
            {
                if (!active_curves[x])
                {
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR DIRAC REMOVE CURVE %d --", x);
                    dirac_remove_curve_calibration(x, true);
                }
            }
        }
        
        result = true;
    }
#endif    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for The Citation Select Audio Source message (0x0209)
 * We are currently using this to determine whether input is internal (0x08-Googlecast) or
 * external (0x00-AUX)
 * @param message
*/
static bool avatar_message_process_album_art(AVATAR_MESSAGE* message)
{
    bool result = false;
    // This will generally be handled in the app_command layer (see ST60))
//    AVATAR_ENCODING encoding;
//    uint8_t album_art[500];
//    uint16_t len;
//
//    len = 500;
//
//    if (avatar_message_get_state_data(message, 
//            AVATAR_STATE_ID_ALBUM_ART, 
//            &encoding, 
//            (uint8_t*)&album_art, 
//            len))
//    {
//        // Album Art has been received - send it 
//        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR ALBUM ART %d --", message->packet->header.packet_id);
//        result = true;    
//    }
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for The Citation Select Audio Source message (0x0209)
 * We are currently using this to determine whether input is internal (0x08-Googlecast) or
 * external (0x00-AUX)
 * @param message
*/
static bool avatar_message_process_audio_input(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    if (avatar_get_state() == AVATAR_STATE_RUNNING)
    {
        AVATAR_ENCODING encoding;
        uint8_t audio_input;
        uint16_t len;

        len = 40;

        if (avatar_message_get_state_data(message, 
                AVATAR_STATE_ID_AUDIO_INPUT, 
                &encoding, 
                (uint8_t*)&audio_input, 
                len))
        {
            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SELECT AUDIO_INPUT %d --", audio_input);

            // Input is internal
            avatar_message_data.current_input = audio_input;
        }
        
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Callback processing for Chromecast App Name
 * @param message
*/
static bool avatar_message_process_chromecast_app_name(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
            
    if (avatar_message_get_state_data(message, 
        AVATAR_STATE_ID_CHROMECAST_APP_NAME, 
        &encoding, 
        (uint8_t*)avatar_message_data.chromecast_app_name, 
        AVATAR_CHROMECAST_APP_NAME_LEN))
    {
        avatar_message_data.chromecast_app_name[AVATAR_CHROMECAST_APP_NAME_LEN] = 0;
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Firmware Version
 * @param message
*/
static bool avatar_message_process_device_system_info_request(AVATAR_MESSAGE* message, AVATAR_PACKET* response_packet)
{
    bool result = false;
    
    if (avatar_message_has_state(message, AVATAR_STATE_ID_DEVICE_SYSTEM_INFO))
    {
        // Setup the UINT16 datatype
        uint8_t data[2];
        
        memcpy(data, &AVATAR_DEVICE_VALUES[avatar_get_device()], sizeof(uint16_t));
 
        avatar_packet_add_data(response_packet,
                AVATAR_STATE_ID_DEVICE_SYSTEM_INFO,
                AVATAR_ENCODING_UINT16,
                data,
                sizeof(uint16_t));
        
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Callback processing for Dirac Select Curve notification
 * @param message
*/
static bool avatar_message_process_dirac_select_curve(AVATAR_MESSAGE* message)
{
    bool result = false;
#if APP_FEATURE_DIRAC == APP_FEATURE_ON
    
    AVATAR_ENCODING encoding;
    uint8_t data;
    uint16_t len;
              
    len = 40;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_DIRAC_SELECT_CURVE, 
            &encoding, 
            (uint8_t*)&data, 
            len))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR DIRAC SELECETED CURVE %s --", data);
        
        result = true;
    }
#endif    
    return result;
}

/**-----------------------------------------------------------------------------
 * Callback processing for Sample Rate message (0x0329)
 * @param message
*/
static bool avatar_message_process_display_sample_rate(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    uint8_t sample_rate;
             
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_DISPLAY_SAMPLE_RATE, 
            &encoding, 
            (uint8_t*)&sample_rate, 
            sizeof(uint8_t)))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR DISPLAY SAMPLE RATE %d --", sample_rate);
        
        avatar_message_data.display_sample_rate = sample_rate;
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Firmware Version
 * @param message
*/
static bool avatar_message_process_firmware_version(AVATAR_MESSAGE* message)
{
    bool result = false;
    AVATAR_ENCODING encoding;
    uint16_t len;
              
    len = 40;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_FIRMWARE_VERSION, 
            &encoding, 
            (uint8_t*)avatar_message_data.firmware_version, 
            len))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR FIRMWARE VERSION %s --", avatar_message_data.firmware_version);
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Firmware Version
 * @param message
*/
static bool avatar_message_process_firmware_version_request(AVATAR_MESSAGE* message, AVATAR_PACKET* response_packet)
{
    bool result = false;
   
    if (avatar_message_has_state(message, AVATAR_STATE_ID_HOST_FIRMWARE_VERSION))
    {
        // We store the last upgrade version just incase we want to upgrade by
        // another way (e.g. the programmer), otherwise the Avatar will keep
        // trying to upgrade.
        // Note this requires a tie up between the version number in the Avatars
        // /lib/firmware/versions.json, and the version specified when packaging
        // the SA30 software in the packager.  (versions.json responds to the 
        // version number returned here, Packager version is used to compare to 
        // see whether an upgrade is required by the SA30 upgrade software)
        uint8_t major_version = SOFTWARE_VERSION_MAJOR;
        uint8_t minor_version = SOFTWARE_VERSION_MINOR;
        
        //TODO - Removed as we are no longer trying to return the previous successful 
        //       version.  This was not stopping the Avatar trying to upgrade because
        //       once the (non) upgrade is complete the device reboots and it kicks 
        //       off the upgrade again...
        //settings_get(SETTING_UPGRADE_SOFTWARE_VERSION_MAJOR, &major_version);
        //settings_get(SETTING_UPGRADE_SOFTWARE_VERSION_MINOR, &minor_version);
        //if (major_version == 0)
        //{
        //    major_version = SOFTWARE_VERSION_MAJOR;
        //    settings_set(SETTING_UPGRADE_SOFTWARE_VERSION_MAJOR, &major_version);
        //}
        //if (minor_version == 0)
        //{
        //    minor_version = SOFTWARE_VERSION_MINOR;
        //    settings_set(SETTING_UPGRADE_SOFTWARE_VERSION_MINOR, &minor_version);
        //}
        
        char version[10];
        memset(version, 0, 10);
        sprintf(version, "%d.%d", major_version, minor_version);
        
        int len = strlen(version);
        
        uint8_t data[12];
        data[0] = len;  //Assumes version < 256 though length actually <= 12...
        data[1] = 0x00;                    
        
        int x;
        for (x = 0; x < len; x++)
        {
            data[x + 2] = version[x];
        }

        avatar_packet_add_data(response_packet,
               AVATAR_STATE_ID_HOST_FIRMWARE_VERSION,
               AVATAR_ENCODING_BIN16,
               data,
               len + 2);

        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Host Name
 * @param message
*/
static bool avatar_message_process_friendly_name(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
           
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_FRIENDLY_NAME, 
            &encoding, 
            (uint8_t*)avatar_message_data.friendly_name, 
            AVATAR_FRIENDLY_NAME_LEN))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR FRIENDLY NAME %s --", (char*)avatar_message_data.friendly_name);
        
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Process the Get States message
 * @param packet
 */
static void avatar_message_process_get_states(AVATAR_MESSAGE* message)
{
    // The avatar is requesting data this will be in a list of uint16 values
    // Get number of values
    uint8_t num_values = avatar_packet_get_num_values(message->packet);

    bool processed = false;

    AVATAR_PACKET* response_packet = avatar_get_packet(AVATAR_MESSAGE_ID_GET_STATES_RESPONSE, num_values * AVATAR_ABITARY_SHORT_MESSAGE_LENGTH);
    if (response_packet != NULL)
    {
        processed = avatar_message_process_firmware_version_request(message, response_packet) ? true : processed;

        processed = avatar_message_process_device_system_info_request(message, response_packet) ? true : processed;
        
        if (processed)
        {
            // Queue the packet for sending
            avatar_send_packet(response_packet);
        }
        else
        {
            avatar_packet_free(response_packet);
        }
        
        avatar_message_process_packet(message);
    }
}

/**-----------------------------------------------------------------------------
 * Process the Get States Response message
 * @param packet
 */
static void avatar_message_process_get_states_response(AVATAR_MESSAGE* message)
{
        // Firmware version processing
    avatar_message_process_firmware_version(message);

    avatar_message_process_dirac_client_status(message);
    
    avatar_message_process_dirac_curve_names(message);

//    avatar_message_process_chromecast_app_name(message);

    avatar_message_process_mac_address(message);

    avatar_message_process_wifi_mac_address(message);

    avatar_message_process_friendly_name(message);

    avatar_message_process_host_name(message);
    
    avatar_message_process_sample_rate(message);
    
    avatar_message_process_display_sample_rate(message);
    
    avatar_message_process_packet(message);
    
    avatar_message_process_album_art(message);
    
    avatar_message_process_serial_number(message);

    avatar_message_process_region(message);
    
//    uint8_t processed = 0;
//    
//    // Firmware version processing
//    avatar_message_process_message(message, avatar_message_process_firmware_version, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_dirac_curve_names, &processed);
//
////    avatar_message_process_message(message, avatar_message_process_chromecast_app_name, &processed);
//
//    avatar_message_process_message(message, avatar_message_process_mac_address, &processed);
//
//    avatar_message_process_message(message, avatar_message_process_wifi_mac_address, &processed);
//
//    avatar_message_process_message(message, avatar_message_process_friendly_name, &processed);
//
//    avatar_message_process_message(message, avatar_message_process_host_name, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_sample_rate, &processed);
//     
//    avatar_message_process_message(message, avatar_message_process_serial_number, &processed);
//   
//    avatar_message_process_packet(message);
//    
//    avatar_message_process_message(message, avatar_message_process_album_art, &processed);
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Host Name
 * @param message
*/
static bool avatar_message_process_host_name(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    uint8_t len = 80;   // For some reason if this is set to 20 it causes the system to crash even though it does not appear to be wrong!!!!
    char host_name[len];
           
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_HOST_NAME, 
            &encoding, 
            (uint8_t*)host_name, 
            len))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR HOST NAME %s", (char*)host_name);

        tcpip_set_network_name((char*)host_name, len);
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar IP Address
 * @param message
*/
static bool avatar_message_process_ip_address(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    uint8_t ip_address[4];
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_IP_ADDRESS, 
            &encoding, 
            (uint8_t*)ip_address, 
            sizeof(uint32_t)))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR IP ADDRESS %d.%d.%d.%d --", ip_address[3], ip_address[2], ip_address[1], ip_address[0]);
        
        tcpip_avatar_set_ip_address((uint32_t*)ip_address);
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Mac Address
 * @param message
*/
static bool avatar_message_process_mac_address(AVATAR_MESSAGE* message)
{
    bool result = false;
    AVATAR_ENCODING encoding;
    uint8_t data[6];
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_ETHERNET_MAC_ADDRESS, 
            &encoding, 
            (uint8_t*)data, 
            6))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR MAC ADDRESS %02X:%02X:%02X:%02X:%02X:%02X",
            data[0],
            data[1],
            data[2],
            data[3],
            data[4],
            data[5]);

        tcpip_avatar_set_mac_address(data);
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar MQA State
 * @param message
*/
static bool avatar_message_process_mime_type(AVATAR_MESSAGE* message)
{
    bool result = false;
    AVATAR_ENCODING encoding;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_MIME_TYPE, 
            &encoding, 
            (uint8_t*)avatar_message_data.mime_type, 
            AVATAR_MIME_TYPE_LEN))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR MIME TYPE %s --",
            avatar_message_data.mime_type);
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar MQA State
 * @param message
*/
static bool avatar_message_process_mqa_state(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    uint8_t state;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_MQA_STATE, 
            &encoding, 
            (uint8_t*)&state, 
            6))
    {
        bool change = false;
        if (avatar_message_data.mqa_state != (AVATAR_MQA_STATE)state)
        {
            change = true;
        }
        
        avatar_message_data.mqa_state = (AVATAR_MQA_STATE)state;
        
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR MQA STATE %d %s --",
            avatar_message_data.mqa_state,
            AVATAR_MQA_STATE_STRINGS[avatar_message_data.mqa_state]);
        
        if (change)
        {
            if (avatar_message_get_mqa_state() != AVATAR_MQA_STATE_NO_MQA)
            {
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR MQA Setting Filter Linear Phase Fast --");
                input_control_set_custom_filter();
            }
            else
            {
                // restore the selected filter
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR MQA Restoring Filter %s --", DAC_FILTER_STRINGS[input_control_get_filter()]);
                input_control_setup_filter();
            }
        }
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Firmware Version
 * @param message
*/
//static bool avatar_message_process_mcu_hw_info_request(AVATAR_MESSAGE* message, AVATAR_PACKET* response_packet)
//{
//    bool result = false;
//    
//    if (avatar_message_has_state(message, AVATAR_STATE_ID_MCU_HW_INFO))
//    {
//        // Setup the UINT16 datatype
//        uint8_t data[2];
////        data[0] = 0x00; // Citation Bar
//        data[0] = 0x01; // Citation Tower
//        data[1] = 0x00;
//
//        avatar_packet_add_data(response_packet,
//                AVATAR_STATE_ID_MCU_HW_INFO,
//                AVATAR_ENCODING_UINT16,
//                data,
//                2);
//        
//        result = true;
//    }
//    
//    return result;
//}

/**-----------------------------------------------------------------------------
 * Process a packet containing data from the Avatar (Set States, Get States Response)
 * @param packet
 * @param message - pointer to an AVATAR_MESSAGE to be populated
 */
static void avatar_message_process_packet(AVATAR_MESSAGE* message)
{
//    avatar_message_build(packet, message);
    
    // Invoke the callback to handle any other processing
    if (avatar_message_data.callback)
    {
        avatar_message_data.callback(message);
    }
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for the Serial Number message (0x0314)
 * @param message
*/
static bool avatar_message_process_region(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
            
    uint8_t region;
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_REGION, 
            &encoding, 
            (uint8_t*)&region, 
            sizeof(uint8_t)))
    {
        avatar_message_data.region = (AVATAR_REGION)region;
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR REGION %s --", AVATAR_REGION_STRINGS[avatar_message_data.region]);
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Sample Rate message (0x0309)
 * @param message
*/
static bool avatar_message_process_sample_rate(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    uint8_t sample_rate;
             
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_SAMPLE_RATE, 
            &encoding, 
            (uint8_t*)&sample_rate, 
            sizeof(uint8_t)))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SAMPLE RATE %d --", sample_rate);
        
        avatar_message_data.sample_rate = sample_rate;
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for the Serial Number message (0x0314)
 * @param message
*/
static bool avatar_message_process_serial_number(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_SERIAL_NUMBER, 
            &encoding, 
            (uint8_t*)avatar_message_data.serial_number, 
            AVATAR_SERIAL_NUMBER_LEN))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SERIAL NUMBER %s --", avatar_message_data.serial_number);
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for The Citation Select Audio Source message (0x0209)
 * We are currently using this to determine whether input is internal (0x08-Googlecast) or
 * external (0x00-AUX)
 * @param message
*/
//static void avatar_message_process_select_audio_source(AVATAR_MESSAGE* message)
//{
//    AVATAR_ENCODING encoding;
//    uint8_t audio_source;
//    uint16_t len;
//              
//    len = 40;
//            
//    if (avatar_message_get_state_data(message, 
//            AVATAR_STATE_ID_SELECT_AUDIO_SOURCE, 
//            &encoding, 
//            (uint8_t*)&audio_source, 
//            len))
//    {
//        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SELECT AUDIO_SOURCE %d --", audio_source);
//        
//        if ((audio_source != AVATAR_AUDIO_SOURCE_UNKNOWN) &&
//            (audio_source != AVATAR_AUDIO_SOURCE_AUX))
//        {
//            // Input is internal
//            avatar_message_data.current_input = AVATAR_INPUT_INTERNAL;
//        }
//        else
//        {
//            // Input is external
//            avatar_message_data.current_input = AVATAR_INPUT_EXTERNAL;
//        }
//    }
//}

/**-----------------------------------------------------------------------------
 * Process the Set States message
 * @param packet
 */
static void avatar_message_process_set_states(AVATAR_MESSAGE* message)
{
    avatar_message_process_sample_rate(message);
    
    avatar_message_process_display_sample_rate(message);

    avatar_message_process_ip_address(message);
    
    avatar_message_process_ssid(message);
    
    avatar_message_process_upgrade_status(message);
    
    avatar_message_process_upgrade_data(message);
    
    avatar_message_process_packet(message);

//    //TODO Removeavatar_message_process_select_audio_source(message);
//    uint8_t processed = 0;
//
////    avatar_message_process_message(message, avatar_message_process_audio_input, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_sample_rate, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_ip_address, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_ssid, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_upgrade_status, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_upgrade_data, &processed);
//    
//    avatar_message_process_packet(message);
}

/**-----------------------------------------------------------------------------
 * Process the Set States message
 * @param packet
 */
static void avatar_message_process_set_states_response(AVATAR_MESSAGE* message)
{
//#ifdef DISPLAY_MESSAGE_ACTIVE            
    avatar_browsing_process_browse_folder(message);
    avatar_browsing_process_browse_back(message);
//#endif
    
    avatar_message_process_packet(message);
}

/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Host Name
 * @param message
*/
static bool avatar_message_process_ssid(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    uint8_t len = 80;   // For some reason if this is set to 20 it causes the system to crash even though it does not appear to be wrong!!!!
    char ssid[len];
           
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_SSID, 
            &encoding, 
            (uint8_t*)ssid, 
            len))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SSID %s --", (char*)ssid);

        tcpip_set_ssid((char*)ssid, len);
        
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Process the State Change message
 * @param packet
 */
static void avatar_message_process_state_change(AVATAR_MESSAGE* message)
{
    avatar_message_process_audio_input(message);   

    avatar_message_check_audio_source(message);
    
    avatar_message_process_dirac_select_curve(message);

    avatar_message_process_ip_address(message);

    avatar_message_process_friendly_name(message);

    avatar_message_process_mime_type(message);    
    
    avatar_message_process_ssid(message);

    avatar_message_process_chromecast_app_name(message);
    
    avatar_message_process_album_art(message);
    
    avatar_message_process_mqa_state(message);

    avatar_message_process_dirac_client_status(message);

    avatar_message_process_packet(message);

    avatar_message_process_dirac_curve_names(message);
    
    avatar_message_process_sample_rate(message);
    
    avatar_message_process_display_sample_rate(message);
    
    // Check whether we're ready to move to the RUNNING state
    if (avatar_message_data.ready_to_run)
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Moving to Running State --");
        logger_process();
        avatar_set_state(AVATAR_STATE_RUNNING);
        
        
        // Set the Avatar sample rate here - this is required when setting region 
        // which causes a fresh handshake, otherwise the Avatar output is muted.
        SAMPLE_RATE sample_rate = input_control_get_sample_rate();
//        if (sample_rate == SAMPLE_RATE_0)
//        {
//            sample_rate = SAMPLE_RATE_192000;
//        }
       
        if (SAMPLE_RATE_VALUES[sample_rate] == 0)
        {
            sample_rate = SAMPLE_RATE_192000;
        }
        
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Moving to Running State A --");
        logger_process();
        avatar_set_sample_rate2(SAMPLE_RATE_VALUES[sample_rate]);
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Moving to Running State B --");
        logger_process();
        
        avatar_message_data.ready_to_run = false;
    }

//    uint8_t processed = 0;
//
//    //TODO Look for Audio Source and Sample rate changes
//    avatar_message_process_message(message, avatar_message_process_audio_input, &processed); 
//
//    avatar_message_process_message(message, avatar_message_check_audio_source, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_dirac_select_curve, &processed);
//
//    avatar_message_process_message(message, avatar_message_process_ip_address, &processed);
//
//    avatar_message_process_message(message, avatar_message_process_friendly_name, &processed);
//
//    avatar_message_process_message(message, avatar_message_process_mime_type, &processed);  
//    
//    avatar_message_process_message(message, avatar_message_process_ssid, &processed);  
//
//    avatar_message_process_message(message, avatar_message_process_chromecast_app_name, &processed); 
//    
//    avatar_message_process_message(message, avatar_message_process_album_art, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_mqa_state, &processed);
//
//    avatar_message_process_packet(message);
//
//    avatar_message_process_message(message, avatar_message_process_dirac_curve_names, &processed);
//    
//    avatar_message_process_message(message, avatar_message_process_sample_rate, &processed);
//    
//    // Check whether we're ready to move to the RUNNING state
//    if (avatar_message_data.ready_to_run)
//    {
//        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Moving to Running State --");
//        avatar_set_state(AVATAR_STATE_RUNNING);
//        avatar_message_data.ready_to_run = false;
//    }
////    else
////    {
////        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Moving to Running State NOT ------------------------------------- --");
////    }
}

/**-----------------------------------------------------------------------------
 * Process the Get States Response message
 * @param packet
 */
static void avatar_message_process_subscribe_response(AVATAR_MESSAGE* message)
{
    if (avatar_get_state() == AVATAR_STATE_CONFIGURATION)
    {
        // Setting state to RUNNING has been moved to the end of the processing
        // of the next State Change message which is the one that passes the 
        // current state of the Avatar.
        //avatar_set_state(AVATAR_STATE_RUNNING);
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Setting Ready To Receive --");
        avatar_message_data.ready_to_run = true;
    }
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar TCP Connected Message
 * @param message
*/
static void avatar_message_process_tcp_connected(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint8_t data[2];

    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_TCP_CLIENT_CONNECTION, 
            &encoding, 
            (uint8_t*)data, 
            2))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR TCP %s --", (AVATAR_TCP_CLIENT_CONNECTED_STATE)data[1] == AVATAR_TCP_CLIENT_CONNECTED_STATE_CONNECTED ? "CONNECTED" : "DISCONNECTED");

        tcpip_avatar_set_client_connected(data[0], 
                (AVATAR_TCP_CLIENT_CONNECTED_STATE)data[1] == AVATAR_TCP_CLIENT_CONNECTED_STATE_CONNECTED ? true : false);
    }
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar TCP Message
 * @param message
*/
static void avatar_message_process_tcp_message(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint16_t len = AVATAR_TCP_PACKET_SIZE;
    uint8_t data[len];
            
    uint16_t data_len = avatar_message_get_state_data(message, 
                                                    AVATAR_STATE_ID_TCP_MESSAGE, 
                                                    &encoding, 
                                                    (uint8_t*)data, 
                                                    len);
    
    if (data_len > 0)
    {
        //logger_log_debugf(0LOGGER_CODE_AVATAR, "AVATAR TCP MESSAGE %d --", data_len);
        //logger_log_debugf(LOGGER_CODE_AVATAR, "     [TCP Message] 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X --", data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7] );
        
        if (avatar_message_data.tcp_callback != NULL)
        {
            uint8_t client_id = data[0];
            uint16_t offset = avatar_packet_get_packet_id(message->packet) * AVATAR_TCP_PACKET_SIZE;

            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR TCP MESSAGE %d %d %d --", client_id, data_len, offset);
            
            avatar_message_data.tcp_callback(client_id, &data[1], offset, data_len - 1);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Process the TCPIP Transfer message
 * @param packet
 */
static void avatar_message_process_tcp_transfer(AVATAR_MESSAGE* message)
{
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR avatar_message_process_tcp_transfer --");
    avatar_message_process_tcp_connected(message);
 
    avatar_message_process_tcp_message(message);
    
    avatar_message_process_packet(message);
}

/**-----------------------------------------------------------------------------
 * Process the TCP Transfer Response message
 * @param packet
 */
static void avatar_message_process_tcp_transfer_response(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint8_t transfer_state;
    uint16_t len;
              
    len = sizeof(uint8_t);
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_TCP_MESSAGE, 
            &encoding, 
            (uint8_t*)&transfer_state, 
            len))
    {
        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR TCP MESSAGE RESPONSE %d --", transfer_state);
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR TCP MESSAGE RESPONSE %s --", AVATAR_TCP_TRANSFER_STATE_STRINGS[transfer_state]);
    }
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_TCP_CLIENT_CONTROL, 
            &encoding, 
            (uint8_t*)&transfer_state, 
            len))
    {
        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR TCP CLIENT CONTROL RESPONSE %d --", transfer_state);
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR TCP CLIENT CONTROL RESPONSE %s --", AVATAR_TCP_TRANSFER_STATE_STRINGS[transfer_state]);
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Upgrade Data request
 * @param message
*/
static bool avatar_message_process_upgrade_data(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    uint16_t actual_len;
    
    uint32_t upgrade_data_size;
    uint32_t offset = UINT32_MAX;
    void* data;
            
    // Get the total data size to be transferred
    actual_len = avatar_message_get_state_data(message, 
                    AVATAR_STATE_ID_UPGRADE_DATA_SIZE, 
                    &encoding, 
                    (uint8_t*)&upgrade_data_size, 
                    sizeof(uint32_t));
    
    if (actual_len > 0)
    {
        // This should give us the total size of the upgrade data.  Note the 
        // size of the data in this message is inside the BIN16 field.
        // We don't actually need this (I think..))
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Upgrade Data - Total Size %d --", upgrade_data_size);       
    }
            
    // Get the Offset of this data packet in the total data
    actual_len = avatar_message_get_state_data(message, 
                    AVATAR_STATE_ID_UPGRADE_DATA_OFFSET, 
                    &encoding, 
                    (uint8_t*)&offset, 
                    sizeof(uint32_t));
    
    if (actual_len > 0)
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Upgrade Data - Offset %d --", offset);       
    }
            
    // Get the actual data - use get_state_value to save allocating the data again
    if (avatar_message_get_state_value(message, 
            AVATAR_STATE_ID_UPGRADE_DATA, 
            &encoding, 
            &data, 
            &actual_len))
    {
        /* We need the upgrade data before being able to do anything with the 
         * version, checksums etc.
         * 
         * If the offset is 0 then we are receiving the the first packet of the 
         * upgrade image which will include the upgrade image info (versions,
         * size, checksum, then the data)
         */
        uint32_t data_offset = 0;
        
        // Set data received count to 1 - we are processing at least 1 image;
        avatar_message_data.upgrade_info.need_data_process_count = 1;

        // At this point data[data_offset] is pointing at the start of the BIN16 field
        // where the first 2 bytes are the length of the following data, so to start
        // move the offset past the data length
        data_offset += sizeof(uint16_t);
        
        uint32_t data_length = actual_len - data_offset;
        
        if (offset == 0)
        {
            // Initialise the number of processed images
            // Note each image to be processed must have an end callback defined that
            // increments the number of processed images....
            avatar_message_data.upgrade_info.processed_image_count = 0;
            avatar_message_data.upgrade_info.active_image_count = 0;    // defined while parsing the header
            
            // Note: We're assuming that actual_len is easily big enough to contain 
            // the version header..
            //avatar_message_data.upgrade_info.current_image = NULL;
            
            data_offset += avatar_message_process_upgrade_header(data + data_offset);
            
            // Check for a failure
            if (data_offset == 0)
            {
                // Reading the header has failed
            }
            else
            {
                // Read first image
                // Note at the moment this is assuming that we managed to read the data for 
                // all the images in the upload file, meaning that the image_map is 
                // good...
                avatar_message_data.upgrade_info.current_idx = 0;
                AVATAR_UPGRADE_IMAGE_INFO* image_info = &avatar_message_data.upgrade_info.image_info[avatar_message_data.upgrade_info.image_map[avatar_message_data.upgrade_info.current_idx]];

                data_offset += avatar_message_process_upgrade_image_header(image_info, data + data_offset);
                if ((image_info != NULL) && (image_info->needs_upgrade) && (image_info->prepare_callback != NULL))
                {
                    image_info->load_id = image_info->prepare_callback(
                                                image_info->major_version, 
                                                image_info->minor_version, 
                                                image_info->image_data_size, 
                                                (uint8_t*)image_info->checksum, 
                                                true); // Always override upgrade
                }

                data_length = actual_len - data_offset;
            }
        }
        else
        {
            AVATAR_UPGRADE_IMAGE_INFO* image_info = &avatar_message_data.upgrade_info.image_info[avatar_message_data.upgrade_info.image_map[avatar_message_data.upgrade_info.current_idx]];
            // Check if the offset is at a boundary between two upgrade images
            // header_size + size of current image
            
            // Calculate the header size
            uint32_t current_image_end_offset = 0;
            if (avatar_message_data.upgrade_info.upgrade_major_version == 1)
            {
                // V1 Upgrade version
                current_image_end_offset = sizeof(uint8_t) + 
                    (min(avatar_message_data.upgrade_info.number_of_images, 
                        AVATAR_UPGRADE_IMAGE_NUM_IMAGES) * 
                        (AVATAR_UPGRADE_IMAGE_PRODUCT_LEN + sizeof(uint32_t))
                    );
            }
            else
            {
                // V2 > upgrade version
                current_image_end_offset = (4 * sizeof(uint8_t)) + 
                    (min(avatar_message_data.upgrade_info.number_of_images, 
                        AVATAR_UPGRADE_IMAGE_NUM_IMAGES) * 
                        (AVATAR_UPGRADE_IMAGE_PRODUCT_LEN + (2 * sizeof(uint8_t)) + sizeof(uint32_t))
                    );
            }
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SWITCHING A %d --", current_image_end_offset);
            
            // Add the offsets of all the images that have been processed so far 
            // including this one
            int x;
            for (x = 0; x <= avatar_message_data.upgrade_info.current_idx; x++)
            {
                AVATAR_UPGRADE_IMAGE_INFO* processed_image_info = &avatar_message_data.upgrade_info.image_info[avatar_message_data.upgrade_info.image_map[x]];
                current_image_end_offset += processed_image_info->size;
                //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SWITCHING B %d %d --", current_image_end_offset, processed_image_info->size);
            }

//            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SWITCHING C %d %d %d %d %d --", 
//                    current_image_end_offset, 
//                    offset, 
//                    offset + actual_len - data_offset, 
//                    actual_len, 
//                    data_offset);
            
            // Check if the offset is between 
            if ((current_image_end_offset >= offset) && 
                (current_image_end_offset < offset + actual_len - data_offset))
            {
                // Are there more images to process
                if (avatar_message_data.upgrade_info.current_idx + 1 < avatar_message_data.upgrade_info.number_of_images)
                {
                    // Push up the process count to 2 because we have 2 data uploads for this packet
                    avatar_message_data.upgrade_info.need_data_process_count = 2;
                    
                    // So process the end of the current image
                    data_offset += avatar_message_process_upgrade_data_packet((data + data_offset), current_image_end_offset - offset);
                   
                    // Then start the upgrade of the next image
                    avatar_message_data.upgrade_info.current_idx++;
                     
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SWITCHING *********************************** --");
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SWITCHING *********************************** --");
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SWITCHING TO NEW UPGRADE IMAGE %d %d %d %d --", 
                            offset, 
                            current_image_end_offset, 
                            offset + data_length,
                            data_length);
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SWITCHING *********************************** --");
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR SWITCHING *********************************** --");
                    logger_process();
                
                    // Do we have enough data left to contain a header
                    // If so we can process it fine, otherwise store it and wait till
                    // the next packet to process it
                    uint8_t header_len = (2 * sizeof(uint8_t)) + sizeof(uint32_t) + MD5_DIGEST_SIZE;
                    
                    // Get remaining data length
                    data_length -= (current_image_end_offset - offset);
                    
                    if (data_length >= header_len)
                    {
                        if (avatar_message_data.upgrade_info.cached_header_info[0] == 0)
                        {
                            // Copy the header into the cached_header_info
                            // NOTE!!! We are relying on header_len being correct, so
                            // if more is added to the header then it will need 
                            // changing above.
                            avatar_message_data.upgrade_info.cached_header_info[0] = header_len;
                            memcpy(&avatar_message_data.upgrade_info.cached_header_info[1], (uint8_t*)data + data_offset, header_len);
                        }
                        else
                        {
                            // Get the amount of header data to read
                            header_len -= avatar_message_data.upgrade_info.cached_header_info[0];
                            
                            // Copy the rest of the data into the cached_header_info
                            memcpy(&avatar_message_data.upgrade_info.cached_header_info[avatar_message_data.upgrade_info.cached_header_info[0] + 1], (uint8_t*)data + data_offset, header_len);
                        }

                        // Note: This has no guarantee that we will read the whole header??????
                        AVATAR_UPGRADE_IMAGE_INFO* image_info = &avatar_message_data.upgrade_info.image_info[avatar_message_data.upgrade_info.image_map[avatar_message_data.upgrade_info.current_idx]];

                        // Process the header info
                        //uint8_t header_len = avatar_message_process_upgrade_image_header(image_info, ((uint8_t*)data + data_offset));
                        header_len = avatar_message_process_upgrade_image_header(image_info, ((uint8_t*)&avatar_message_data.upgrade_info.cached_header_info[1]));
                        data_offset += header_len;

                        // remove the header length?????
                        data_length -= header_len;

                        if ((image_info != NULL) && (image_info->needs_upgrade) && (image_info->prepare_callback != NULL))
                        {
                            image_info->load_id = image_info->prepare_callback(
                                                        image_info->major_version, 
                                                        image_info->minor_version, 
                                                        image_info->image_data_size, 
                                                        (uint8_t*)image_info->checksum, 
                                                        true); // Always override upgrade
                        }
                        else if (avatar_message_data.upgrade_info.current_idx + 1 >= avatar_message_data.upgrade_info.number_of_images)
                        {
                            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR ABANDONING UPGRADE %d %d --", avatar_message_data.upgrade_info.current_idx, avatar_message_data.upgrade_info.number_of_images);
                            // We are abandoning the upgrade as this is the last image and it
                            // doesn't need an upgrade
                            // Sending the message is handled below
                            data_offset = 0;
                        }
                        else
                        {
                        }
                        
                        avatar_message_data.upgrade_info.cached_header_info[0] = 0;
                    }
                    else
                    {
                        // Store the header data that is available
                        avatar_message_data.upgrade_info.cached_header_info[0] = data_length;
                        memcpy(&avatar_message_data.upgrade_info.cached_header_info[1], (uint8_t*)data, data_length);
                    }
                }
            }
        }

        // Check we're not aborting
        if (data_offset > 0)
        {
            // Now save the data in the message
            //UPGRADE_METADATA_UPLOAD_STATUS result =UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR upgrade_metadata_upload_data CALLBACK PRE --");
            if (data_length > 0)
            {
                data_offset += avatar_message_process_upgrade_data_packet((data + data_offset), data_length);
            }
        }
        else
        {
            AVATAR_UPGRADE_IMAGE_INFO* image_info = &avatar_message_data.upgrade_info.image_info[avatar_message_data.upgrade_info.image_map[avatar_message_data.upgrade_info.current_idx]];
            
            if ((image_info != NULL) && !image_info->needs_upgrade)
            {
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR ENDING UPGRADE --");
                
                // The upgrade is complete as there are no more device to upgrade
                avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_TRANSFER_SUCCESSFUL);
                avatar_message_generate_transfer_complete();
            }
            else
            {
                logger_log_debugf(LOGGER_CODE_UPGRADE, "DEVICE_UPGRADE_STATUS_CHECKSUM_ERROR A - NO UPGRADE REQD --");
                // The upgrade has failed because the header cannot be processed
                avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_CHECKSUM_ERROR);
            }
        }
        
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Process the data from a single packet
 * @param data
 * @param len
 */
static uint32_t avatar_message_process_upgrade_data_packet(uint8_t* data, uint32_t data_length)
{
    uint32_t result = data_length;
    
    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR DATA PACKET DATA LENGTH %d --", data_length);
    logger_process();
    
    AVATAR_UPGRADE_IMAGE_INFO* image_info = &avatar_message_data.upgrade_info.image_info[avatar_message_data.upgrade_info.image_map[avatar_message_data.upgrade_info.current_idx]];
    if ((image_info != NULL) && (image_info->data_callback != NULL))
    {
        if (image_info->needs_upgrade)
        {
            // Call data callback
            UPGRADE_METADATA_UPLOAD_STATUS result = image_info->data_callback(
                                                        image_info->load_id, 
                                                        data, 
                                                        data_length, //actual_len - data_offset,
                                                        avatar_message_process_upload_data_callback);
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR upgrade_metadata_upload_data POST --");

            // Send a data_received message
            if (result == UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS)
            {
                // Handled by avatar_message_process_upload_data_callback

                // Update the bytes written - Note this is pre-emptive as the data 
                // hasn't actually been written yet but as at the time of writing this
                // is only used where the upgrade is not taking place (see below) it 
                // doesn't matter
                image_info->bytes_written += data_length;
            }
            else
            {
                logger_log_debugf(LOGGER_CODE_UPGRADE, "DEVICE_UPGRADE_STATUS_CHECKSUM_ERROR B - UPGRADE ERR %d %d --", image_info->load_id, result);
                avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_CHECKSUM_ERROR);
            }
        }
        else
        {
            // We have determined that we don't need to upgrade so work out whether 
            // we have processed all the upgrade data or whether there is still more to do
            image_info->bytes_written += data_length;
            
            logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade IGNORING Image byte count %d/%d --", image_info->bytes_written, image_info->image_data_size);
            
            if (image_info->bytes_written >= image_info->image_data_size)
            {
                avatar_message_process_upload_end_callback(image_info->load_id, image_info->image_data_size, image_info->bytes_written, UPGRADE_METADATA_UPGRADE_STATUS_OK);
            }
            else
            {
                avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_DATA_RECEIVED);
            }
        }
    }
    else
    {
        avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_DATA_RECEIVED);
    }
    
    return data_length;
}

/**-----------------------------------------------------------------------------
 * Updates imnage_info with header info
 * @param image_info
 * @param header_data
 */
static uint8_t avatar_message_process_upgrade_header(uint8_t* header_data)
{
    uint8_t data_offset = 0;
    
    // Number of images
    avatar_message_data.upgrade_info.number_of_images = *header_data;
    
    if (avatar_message_data.upgrade_info.number_of_images > 0)
    {
        data_offset++;
        
        // Check for new file version 
        // At the moment we only need to know whether its the new version or not
        // So ignoring the upgrade major/minor version
        bool new_format_upgrade = false;
        if (avatar_message_data.upgrade_info.number_of_images == 0xFE)
        {
            avatar_message_data.upgrade_info.upgrade_major_version = *(header_data + data_offset);
            data_offset++;
            
            avatar_message_data.upgrade_info.upgrade_minor_version = *(header_data + data_offset);
            data_offset++;
         
            logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info - New style upgrade v%d.%d --", avatar_message_data.upgrade_info.upgrade_major_version, avatar_message_data.upgrade_info.upgrade_minor_version);
            
            avatar_message_data.upgrade_info.number_of_images = *(header_data + data_offset);
            
            data_offset++;
            
            new_format_upgrade = true;
        }
        
        // Check upgrade format

        logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info - Number of Images %d --", avatar_message_data.upgrade_info.number_of_images);

        // Set all Image Info to complete = false;
        int x;
        for (x = 0; x < AVATAR_UPGRADE_IMAGE_NUM_IMAGES; x++)
        {
            avatar_message_data.upgrade_info.image_info[x].complete = false;
        }
 
        // Map the upgrade data to the upgrade_info.image_info structure
        //int x;
        for (x = 0; x < avatar_message_data.upgrade_info.number_of_images; x++)
        {
            //logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info - Processing Image %d --", x);
            avatar_message_data.upgrade_info.image_map[x] = 0xFF;   // Set as an unreachable number to initialise

            // Find the IMAGE_INFO for the product
            // Note upgrade_info product should have been filled out during initialisation
            AVATAR_UPGRADE_IMAGE_INFO* image_info = NULL;
            int y;
            for (y = 0; y < AVATAR_UPGRADE_IMAGE_NUM_IMAGES; y++)
            {
                //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Considering product *%s* %s --", avatar_message_data.upgrade_info.image_info[y].product, header_data + data_offset);

                if (strncmp(avatar_message_data.upgrade_info.image_info[y].product, header_data + data_offset, AVATAR_UPGRADE_IMAGE_PRODUCT_LEN) == 0)
                {
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Found matching product %s --", avatar_message_data.upgrade_info.image_info[y].product);
                    image_info = &avatar_message_data.upgrade_info.image_info[y];

                    // Set the map position
                    avatar_message_data.upgrade_info.image_map[x] = y;

                    avatar_message_data.upgrade_info.active_image_count++;
                    break;
                }
                else if (avatar_message_data.upgrade_info.image_info[y].product[0] == ' ')
                {
                    char product[AVATAR_UPGRADE_IMAGE_PRODUCT_LEN + 1];
                    memset(product, 0, AVATAR_UPGRADE_IMAGE_PRODUCT_LEN + 1);
                    memcpy(product, header_data + data_offset, AVATAR_UPGRADE_IMAGE_PRODUCT_LEN);
                    logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info - Image %s NOT Defined --", product);

                   // Check if the image_info.product is spaces.  If so we've
                    // run off the end of defined images and are making this one 
                    // a temporary holder
                    // Note we assume a product name NEVER starts with a space!!
                    image_info = &avatar_message_data.upgrade_info.image_info[y];

                    // Set the map position
                    avatar_message_data.upgrade_info.image_map[x] = y;
                    break;
                }
            }

            if (image_info != NULL)
            {
                // Image name
                data_offset += AVATAR_UPGRADE_IMAGE_PRODUCT_LEN;

                if (new_format_upgrade)
                {
                    image_info->major_version = *(header_data + data_offset);
                    data_offset++;

                    image_info->minor_version = *(header_data + data_offset);
                    data_offset++;
                }
                
                memcpy((uint8_t*)&image_info->size, (header_data + data_offset), sizeof(uint32_t));
                data_offset += sizeof(uint32_t);

                if (new_format_upgrade)
                {
                    logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info - Image %s (%d) %d/%d --", image_info->product, image_info->size, image_info->major_version, image_info->minor_version);
                }
                else
                {
                    logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info - Image %s (%d) --", image_info->product, image_info->size);
                }
            }
            else
            {
                // Theres no space in the array to hold any info about this upgrade
                // and that will cause problems later on as we do not know its length
                // So stop the upgrade now.....
                char product[AVATAR_UPGRADE_IMAGE_PRODUCT_LEN + 1];
                memset(product, 0, AVATAR_UPGRADE_IMAGE_PRODUCT_LEN + 1);
                memcpy(product, header_data + data_offset, AVATAR_UPGRADE_IMAGE_PRODUCT_LEN);
                logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info - Image %s Ignored NO SPACE --", product);

                data_offset = 0; // Indicates a failure
                break;
            }
        }
    }
    
    return data_offset;
}

/**-----------------------------------------------------------------------------
 * Updates imnage_info with header info
 * @param image_info
 * @param header_data
 */
static uint8_t avatar_message_process_upgrade_image_header(AVATAR_UPGRADE_IMAGE_INFO* image_info, uint8_t* header_data)
{
    uint8_t data_offset = 0;
   
    // Get the Major, Minor versions, Image Size, and Checksum
    image_info->major_version = *((uint8_t*)(header_data + data_offset));
    data_offset++;

    image_info->minor_version = *((uint8_t*)(header_data + data_offset));
    data_offset++;

    memcpy((uint8_t*)&image_info->image_data_size, (header_data + data_offset), sizeof(uint32_t));
    data_offset += sizeof(uint32_t);

    memcpy((uint8_t*)&image_info->checksum, (header_data + data_offset), MD5_DIGEST_SIZE);
    data_offset += MD5_DIGEST_SIZE;

    // Set whether the image needs upgrading
    image_info->needs_upgrade = true;
    if ((image_info != NULL) && (image_info->check_version_callback != NULL))
    {
        image_info->needs_upgrade = image_info->check_version_callback(image_info->major_version, image_info->minor_version);
    }
    
    logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info %s v%d:%02d %d %02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x --", 
            image_info->product, 
            image_info->major_version,
            image_info->minor_version,
            image_info->image_data_size,
            image_info->checksum[0],
            image_info->checksum[1],
            image_info->checksum[2],
            image_info->checksum[3],
            image_info->checksum[4],
            image_info->checksum[5],
            image_info->checksum[6],
            image_info->checksum[7],
            image_info->checksum[8],
            image_info->checksum[9],
            image_info->checksum[10],
            image_info->checksum[11],
            image_info->checksum[12],
            image_info->checksum[13],
            image_info->checksum[14],
            image_info->checksum[15]
            );
    
    if (!image_info->needs_upgrade)
    {
        logger_log_infof(LOGGER_CODE_AVATAR, "AVATAR Upgrade Info %s SKIPPING IMAGE --", image_info->product);
    }
    logger_process();
    return data_offset;
}

/**-----------------------------------------------------------------------------
 * Callback invoked when upload data has been successfully transferred
 * @param total_size
 * @param bytes_written
 */
static void avatar_message_process_upload_data_callback(UPGRADE_METADATA_LOAD_ID load_id, uint32_t total_size, uint32_t bytes_written, UPGRADE_METADATA_UPGRADE_STATUS status)
{
    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR avatar_message_process_upload_data_callback %d %d %d %d --", load_id, bytes_written, total_size, status);
    // Check whether we are at the end of the data transfer.  Were going to 
    // assume here that the packet where the offset + actual_length - BIN16 length == upgrade_data_size
    // and that is the last packet sent.
    switch(status)
    {
        case UPGRADE_METADATA_UPGRADE_STATUS_ERROR:
        {
                // Find the image info for the load_id
                AVATAR_UPGRADE_IMAGE_INFO* image_info = NULL;
                int x;
                for (x = 0; x < AVATAR_UPGRADE_IMAGE_NUM_IMAGES; x++)
                {
                    if(avatar_message_data.upgrade_info.image_info[x].load_id == load_id)
                    {
                        image_info = &avatar_message_data.upgrade_info.image_info[x];
                        break;
                    }
                }

                if ((image_info != NULL) && (image_info->end_callback != NULL))
                {
                    image_info->end_callback(image_info->load_id,
                            avatar_message_process_upload_end_callback,
                            status);
                }
                //avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_UPGRADE_FAIL);
            break;
        }

        case UPGRADE_METADATA_UPGRADE_STATUS_RETRY:
        {
        
        }
        
        default:
        {
            if (bytes_written >= total_size)
            {
                // Find the image info for the load_id
                AVATAR_UPGRADE_IMAGE_INFO* image_info = NULL;
                int x;
                for (x = 0; x < AVATAR_UPGRADE_IMAGE_NUM_IMAGES; x++)
                {
                    if(avatar_message_data.upgrade_info.image_info[x].load_id == load_id)
                    {
                        image_info = &avatar_message_data.upgrade_info.image_info[x];
                        break;
                    }
                }

                if ((image_info != NULL) && (image_info->end_callback != NULL))
                {
                    image_info->end_callback(image_info->load_id,
                            avatar_message_process_upload_end_callback,
                            status);
                }
            }
            else
            {
                // Decrement data received count and check if we can send message
                avatar_message_data.upgrade_info.need_data_process_count--;
                if (avatar_message_data.upgrade_info.need_data_process_count == 0)
                {
                    avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_DATA_RECEIVED);
                }
            }
            
            break;
        }
    }
}


/**-----------------------------------------------------------------------------
 * Callback invoked when all upload data has been successfully transferred
 * @param total_size - total size of the upgrade data
 * @param bytes_written - 
 */
static void avatar_message_process_upload_end_callback(UPGRADE_METADATA_LOAD_ID load_id, uint32_t total_size, uint32_t bytes_written, UPGRADE_METADATA_UPGRADE_STATUS status)
{
    avatar_message_data.upgrade_info.processed_image_count++;

    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR avatar_message_process_upload_end_callback %d %d %d %d --", 
            load_id,
            avatar_message_data.upgrade_info.processed_image_count, 
            avatar_message_data.upgrade_info.active_image_count, 
            status);
    logger_process();
    
    switch (status)
    {
        case UPGRADE_METADATA_UPGRADE_STATUS_ERROR:
        {
            avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_UPGRADE_FAIL);
            
            // Process Internally
            AVATAR_PACKET* packet = avatar_get_packet(AVATAR_MESSAGE_ID_SET_STATES, AVATAR_ABITARY_SHORT_MESSAGE_LENGTH);
            if (packet != NULL)
            {
                uint8_t data;
                data = AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_FAIL;

                avatar_packet_add_data(packet,
                        AVATAR_STATE_ID_UPGRADE_STATUS,
                        AVATAR_ENCODING_UINT8,
                        &data,
                        1);

                AVATAR_MESSAGE message;
                avatar_message_build(packet, &message);

                // Pass upto the app_command level
                avatar_message_process_packet(&message);
            }
            break;
        }
        
        default:
        {
            // Find the image info for the current callback
            AVATAR_UPGRADE_IMAGE_INFO* image_info = NULL;
            int x;
            for (x = 0; x < AVATAR_UPGRADE_IMAGE_NUM_IMAGES; x++)
            {
                if(avatar_message_data.upgrade_info.image_info[x].load_id == load_id)
                {
                    image_info = &avatar_message_data.upgrade_info.image_info[x];
                    break;
                }
            }
              
            // set it to complete
            if (image_info != NULL)
            {
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR setting upgrade image %d to complete --", load_id);
                image_info->complete = true;
            }
            
//            // Check whether the complete count - active image count
//            int complete_count = 0;
//            for (x = 0; x < AVATAR_UPGRADE_IMAGE_NUM_IMAGES; x++)
//            {
//                if (avatar_message_data.upgrade_info.image_info[x].complete)
//                {
//                    complete_count++;
//                }
//            }
            
            logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Processed count is %d wants %d --", avatar_message_data.upgrade_info.processed_image_count, avatar_message_data.upgrade_info.active_image_count);
           
            // We only want to invoke this if all the upgrade images have been processed.
//            if (complete_count >= avatar_message_data.upgrade_info.active_image_count)
            if (avatar_message_data.upgrade_info.processed_image_count == avatar_message_data.upgrade_info.active_image_count)
            {
                uint8_t data;
                
                // Check here whether the upload has been successful.  At the moment we're doing this 
                // purely on the success/fail of the MCU upgrade.  This should probably be changed to 
                // check that all uploads completed OK though not sure how to do that at the moment!
                if ((upgrade_metadata_get_upgrade_state() == UPGRADE_METADATA_STATE_AVAILABLE_INSTALL) ||
                    (upgrade_metadata_get_upgrade_state() == UPGRADE_METADATA_STATE_NONE))
                {
                    AVATAR_UPGRADE_IMAGE_INFO* image_info = NULL;
                    int x;
                    for (x = 0; x < AVATAR_UPGRADE_IMAGE_NUM_IMAGES; x++)
                    {
                        if(avatar_message_data.upgrade_info.image_info[x].load_id == load_id)
                        {
                            image_info = &avatar_message_data.upgrade_info.image_info[x];
                            break;
                        }
                    }

                    // This is assuming the SA30 upgrade is always last!!!
                    if (image_info != NULL)
                    {
                        settings_set(SETTING_UPGRADE_SOFTWARE_VERSION_MAJOR, &image_info->major_version);
                        settings_set(SETTING_UPGRADE_SOFTWARE_VERSION_MINOR, &image_info->minor_version);
                        settings_save();
                    }

                    avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_TRANSFER_SUCCESSFUL);
                }
                else
                {
                    logger_log_debugf(LOGGER_CODE_UPGRADE, "DEVICE_UPGRADE_STATUS_CHECKSUM_ERROR C - UPGRADE STATE %d %d --", upgrade_metadata_get_upgrade_state());
                    avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_CHECKSUM_ERROR);
                }
                
                // Simulate a 0x030B Avatar Upgrade Status - 0x07 Upgrade Transfer Complete
                // We were expecting the Avatar to send this message back but have decided its not needed
                // so reusing the message to trigger the restart
                // This is handled by both avatar_message_process_upgrade_status and app_commands_process_upgrade_status
                avatar_message_generate_transfer_complete();
//                AVATAR_PACKET* packet = avatar_get_packet(AVATAR_MESSAGE_ID_SET_STATES, AVATAR_ABITARY_SHORT_MESSAGE_LENGTH);
//                if (packet != NULL)
//                {
//                    uint8_t data;
//                    data = AVATAR_UPGRADE_STATUS_DEVICE_UPGRADE_TRANSFER_COMPLETE;
//
//                    avatar_packet_add_data(packet,
//                            AVATAR_STATE_ID_UPGRADE_STATUS,
//                            AVATAR_ENCODING_UINT8,
//                            &data,
//                            1);
//
//                    AVATAR_MESSAGE message;
//                    avatar_message_build(packet, &message);
//
//                    // Pass upto the app_command level
//                    avatar_message_process_packet(&message);
//                }
            }
            else
            {
                // Decrement data received count and check if we can send message
                avatar_message_data.upgrade_info.need_data_process_count--;
                if (avatar_message_data.upgrade_info.need_data_process_count == 0)
                {
                    avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_DATA_RECEIVED);
                }
            }
            
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Upgrade Status request
 * @param message
*/
static bool avatar_message_process_upgrade_status(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    uint16_t actual_len;
    uint8_t data;
            
    actual_len = avatar_message_get_state_data(message, 
                    AVATAR_STATE_ID_UPGRADE_STATUS, 
                    &encoding, 
                    (uint8_t*)&data, 
                    sizeof(uint8_t));
    
    if (actual_len > 0)
    {
        //logger_log_debugf(0, "AVATAR Setting Upgrade Status %d %s --", data, AVATAR_UPGRADE_STATUS_STRINGS[data]);
        avatar_message_data.upgrade_info.upgrade_status = (AVATAR_UPGRADE_STATUS)data;
        switch (avatar_message_data.upgrade_info.upgrade_status)
        {
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_AVAILABLE:
            case AVATAR_UPGRADE_STATUS_DUAL_UPGRADE_AVAILABLE:
            {
                // Not sure we need to do anything here - if an Device upgrade is
                // available we will get AVATAR_UPGRADE_STATUS_DEVICE_UPGRADE_AVAILABLE
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_DEVICE_UPGRADE_AVAILABLE:
            {
                // TODO - store what type of upgrade is available
                // Removed - this has been removed from the spec and shouldn't be here anyway avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_START_AVATAR_UPGRADE);
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_START:
            {
                avatar_message_data.upgrade_info.in_upgrade = true;
                // If the upgrade includes a device upgrade then send ready to receive
                // (it will do because we have received this message))
                // Send a receiving message indicating to the avatar that it is OK to start sending data
                avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_READY_TO_RECEIVE);
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_IN_PROGRESS:
            {
                avatar_message_data.upgrade_info.in_upgrade = true;
                // Indicates that the AVATAR is doing an internal upgrade
                // Display handled in app_commands_avatar
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_COMPLETE:
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_FAIL:
            {
                avatar_message_data.upgrade_info.in_upgrade = false;
                // Reports on the success/fail of the AVatar internal upgrade
                // Display screen saying avatar upgrade is complete
                // No Action
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_DEVICE_UPGRADE_TRANSFER_COMPLETE:
            {
                avatar_message_data.upgrade_info.in_upgrade = false;
                
                // Check the transfer was successful and send the Avatar a message saying we're starting the upgrade
                if ((upgrade_metadata_get_upgrade_state() == UPGRADE_METADATA_STATE_AVAILABLE_INSTALL) ||
                    (upgrade_metadata_get_upgrade_state() == UPGRADE_METADATA_STATE_NONE))
                {
                    avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_UPGRADE_STARTING);
                }
                break;
            }
        }
        
        result = true;
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Callback processing for Avatar Mac Address
 * @param message
*/
static bool avatar_message_process_wifi_mac_address(AVATAR_MESSAGE* message)
{
    bool result = false;
    
    AVATAR_ENCODING encoding;
    uint8_t data[6];
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_WIFI_MAC_ADDRESS, 
            &encoding, 
            (uint8_t*)data, 
            6))
    {
        logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR WIFI MAC ADDRESS %02X:%02X:%02X:%02X:%02X:%02X",
            data[0],
            data[1],
            data[2],
            data[3],
            data[4],
            data[5]);

         tcpip_avatar_set_wifi_mac_address(data);
         
         result = true;
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the current avatar audio input
 * @return 
 */
AVATAR_INPUT avatar_message_get_audio_input()
{
    return avatar_message_data.current_input;
}

/**-----------------------------------------------------------------------------
 * Gets the current avatar audio source
 * @return 
 */
AVATAR_AUDIO_SOURCE avatar_message_get_audio_source()
{
    return avatar_message_data.audio_source;
}

/**-----------------------------------------------------------------------------
 * Gets the current audio source as a string, adjusted for the Chromecast app name
 * @return 
 */
void avatar_message_get_audio_source_string(char* buffer, uint8_t len)
{
    // Clear the buffer
    memset(buffer, 0, len);
    
    // Get the source as a string
    AVATAR_AUDIO_SOURCE audio_source = avatar_message_get_audio_source();
    if ((audio_source != AVATAR_AUDIO_SOURCE_GOOGLECAST) || (avatar_message_data.chromecast_app_name[0] == 0))
    {
        memcpy(buffer, AVATAR_AUDIO_SOURCE_STRINGS[audio_source], min(strlen(AVATAR_AUDIO_SOURCE_STRINGS[audio_source]), len));
    }
    else
    {
        memcpy(buffer, avatar_message_data.chromecast_app_name, min(strlen(avatar_message_data.chromecast_app_name), len));
    }
}

/**-----------------------------------------------------------------------------
 * Gets the current  Chromecast app name
 * @return 
 */
void avatar_message_get_chromecast_app_name_string(char* buffer, uint8_t len)
{
    memcpy(buffer, avatar_message_data.chromecast_app_name, min(strlen(avatar_message_data.chromecast_app_name), len));
}
    
/**-----------------------------------------------------------------------------
 * Gets the current dirac client status
 */
AVATAR_DIRAC_CLIENT_STATUS avatar_message_get_dirac_client_status()
{
    return avatar_message_data.dirac_client_status;
}

/**-----------------------------------------------------------------------------
 * Returns the current sample rate setting for the Avatar
 * @return 
 */
AVATAR_SAMPLE_RATE avatar_message_get_display_sample_rate()
{
    AVATAR_SAMPLE_RATE result = avatar_message_data.display_sample_rate;
    // BUG 3624 - use the real sample rate if there is no display sample rate available
    if (result == AVATAR_SAMPLE_RATE_0)
    {
        SAMPLE_RATE sample_rate = spdif_rx_get_sample_rate();
        if (sample_rate < AVATAR_SAMPLE_RATE_MAX) // Note the SAMPLE_RATE and AVATAR_SAMPLE_RATE match upto 192000
        {
            result = sample_rate;
        }
    }
    else
    {
        SAMPLE_RATE sample_rate = spdif_rx_get_sample_rate();
        if ((sample_rate = SAMPLE_RATE_0) || (sample_rate >= AVATAR_SAMPLE_RATE_MAX))
        {
            result = AVATAR_SAMPLE_RATE_0;
        }
    }
    return result;
}   

/**-----------------------------------------------------------------------------
 * Gets the Avatar firmware version
 * @return 
 */
void avatar_message_get_firmware_version(char* buffer, uint8_t len)
{
    if (buffer != NULL)
    {
        memcpy(buffer, avatar_message_data.firmware_version, min(AVATAR_FIRMWARE_VERSION_LEN, len));
    }
}

/**-----------------------------------------------------------------------------
 * Gets the current  Friendly Name
 * @return 
 */
void avatar_message_get_friendly_name_string(char* buffer, uint8_t len)
{
    memset(buffer, 0, len);
    memcpy(buffer, avatar_message_data.friendly_name, min(strlen(avatar_message_data.friendly_name), len));
    buffer[len - 1] = 0;
}

/**-----------------------------------------------------------------------------
 * Gets the index in the AVATAR_MIME_TYPE_STRINGS, AVATAR_MIME_TYPE_VALUES, 
 * AVATAR_MIME_TYPE_DISPLAY_STRINGS or the current Mime Type
 * @return 
 */
uint8_t avatar_message_get_mime_type_index()
{
    uint8_t result = 0; // Unknown
    uint8_t count = sizeof(AVATAR_MIME_TYPE_VALUES) / sizeof(uint8_t);
    int x;
    for (x = 0; x < count; x++)
    {
        if (strcmp(avatar_message_data.mime_type, AVATAR_MIME_TYPE_STRINGS[x]) == 0)
        {
            result = x;
            break;
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the current  Friendly Name
 * @return 
 */
void avatar_message_get_mime_type_string(char* buffer, uint8_t len)
{
    memcpy(buffer, avatar_message_data.mime_type, min(strlen(avatar_message_data.mime_type), len));
}

/**-----------------------------------------------------------------------------
 * Gets the current MQA Decoder level
 * @return 
 */
void avatar_message_get_mqa_decoder_level_string(char* buffer, uint8_t len)
{
    uint8_t value;
    settings_get(SETTING_MQA_DECODER_LEVEL, &value);
    
    if (value < AVATAR_MQA_DECODER_LEVEL_MAX)
    {
        memset(buffer, 0, len);
        memcpy(buffer, AVATAR_MQA_DECODER_LEVEL_STRINGS[value], min(strlen(AVATAR_MQA_DECODER_LEVEL_STRINGS[value]), len));
    }
    else
    {
        sprintf(buffer, "Unknown MQA Decoder Level");
    }   
}

/**-----------------------------------------------------------------------------
 * Returns the current sample rate setting for the Avatar
 * @return 
 */
AVATAR_MQA_STATE avatar_message_get_mqa_state()
{
    return avatar_message_data.mqa_state;
}   
     
/**-----------------------------------------------------------------------------
 * Gets the current region
 * @return 
 */
AVATAR_REGION avatar_message_get_region()
{
    return avatar_message_data.region;
}

/**-----------------------------------------------------------------------------
 * Returns the current sample rate setting for the Avatar
 * @return 
 */
AVATAR_SAMPLE_RATE avatar_message_get_sample_rate()
{
    return avatar_message_data.sample_rate;
}   

/**-----------------------------------------------------------------------------
 * Gets the Serial Number
 * @return 
 */
void avatar_message_get_serial_number_string(char* buffer, uint8_t len)
{
    memset(buffer, 0, len);
    memcpy(buffer, avatar_message_data.serial_number, min(strlen(avatar_message_data.serial_number), len));
}

/**-----------------------------------------------------------------------------
 * Gets the Avatar upgrade option string
 * @return 
 */
void avatar_message_get_upgrade_option_string(char* buffer, uint8_t len)
{
    if (buffer != NULL)
    {
        memset(buffer, 0, len);
        AVATAR_UPGRADE_OPTION option = avatar_get_upgrade_option();
        if (option < AVATAR_UPGRADE_OPTION_MAX)
        {
            memcpy(buffer, AVATAR_UPGRADE_OPTION_STRINGS[option], min(strlen(AVATAR_UPGRADE_OPTION_STRINGS[option]), len));
        }
        else
        {
            memcpy(buffer, "Uninitialised", 13);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Gets the current Avatar upgrade status
 * @param packet
 */
AVATAR_UPGRADE_STATUS avatar_message_get_upgrade_status()
{
    return avatar_message_data.upgrade_info.upgrade_status;
}
 
/**
 * Returns true if the Avatar is currently updating
 * @return 
 */
bool avatar_message_in_upgrade()
{
    return avatar_message_data.upgrade_info.in_upgrade;
}
   
/**-----------------------------------------------------------------------------
 * Initialises the avatar message for upgrade actions
 * @param upgrade_actions
 * @param num_upgrade_actions
 */
void avatar_message_initialise(AVATAR_UPGRADE_IMAGE_ACTIONS* upgrade_actions, 
        uint8_t num_upgrade_actions)
{
    // Assumes that the number of upgrade actions < AVATAR_UPGRADE_IMAGE_NUM_IMAGES
    int x;
    for (x = 0; x < num_upgrade_actions; x++)
    {
        AVATAR_UPGRADE_IMAGE_INFO* upgrade_info = &avatar_message_data.upgrade_info.image_info[x];
        memset(upgrade_info->product, ' ', AVATAR_UPGRADE_IMAGE_PRODUCT_LEN);    // Populate with spaces
        upgrade_info->product[AVATAR_UPGRADE_IMAGE_PRODUCT_LEN] = 0;             // Add NULL terminator
        memcpy(upgrade_info->product, upgrade_actions[x].product, strlen(upgrade_actions[x].product));  // Copy the product name
        
        upgrade_info->needs_upgrade = true;
        upgrade_info->check_version_callback = upgrade_actions[x].check_version_callback;
        upgrade_info->prepare_callback = upgrade_actions[x].prepare_callback;
        upgrade_info->data_callback = upgrade_actions[x].data_callback;
        upgrade_info->end_callback = upgrade_actions[x].end_callback;
    }
    
    // Initialise the spare image locations
    for (x = num_upgrade_actions; x < AVATAR_UPGRADE_IMAGE_NUM_IMAGES; x++)
    {
        AVATAR_UPGRADE_IMAGE_INFO* upgrade_info = &avatar_message_data.upgrade_info.image_info[x];
        memset(upgrade_info->product, ' ', AVATAR_UPGRADE_IMAGE_PRODUCT_LEN);    // Populate with spaces
        upgrade_info->product[AVATAR_UPGRADE_IMAGE_PRODUCT_LEN] = 0;             // Add NULL terminator
        
        upgrade_info->prepare_callback = NULL;
        upgrade_info->data_callback = NULL;
        upgrade_info->end_callback = NULL;
    }
}

/**-----------------------------------------------------------------------------
 * Application specific Avatar processing
 * @param packet
 */
void avatar_message_process(AVATAR_MESSAGE* message)
{
    // First get the type of message
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR avatar_message_process %d %d --", message->message_id, message->frame_id);
    switch(message->message_id)
    {
        case AVATAR_MESSAGE_ID_SET_STATES:
        {
            // The Avatar is sending us info
            avatar_message_process_set_states(message);
            break;
        }
        
        case AVATAR_MESSAGE_ID_GET_STATES:
        {
            // Respond to a GetStates message
            avatar_message_process_get_states(message);
            break;
        }
        
        case AVATAR_MESSAGE_ID_SUBSCRIBE:
        {
            // The avatar wants to subscribe to something we're doing????
            //TODO - Send ACK/NACK
            break;
        }
        
        case AVATAR_MESSAGE_ID_STATE_CHANGE:
        {
            // The avatar is sending state change info
            // Read the state change info
            avatar_message_process_state_change(message);
            break;
        }
        
        case AVATAR_MESSAGE_ID_UNSUBSCRIBE:
        {
            // The avatar is ...
            
            break;
        }
        
        case AVATAR_MESSAGE_ID_TCPIP_TRANSFER:
        {
            avatar_message_process_tcp_transfer(message);
            break;
        }
        
        case AVATAR_MESSAGE_ID_SET_STATES_RESPONSE:
        {
            avatar_message_process_set_states_response(message);
            break;
        }
        
        case AVATAR_MESSAGE_ID_GET_STATES_RESPONSE:
        {
            // AVATAR response to an information request
            avatar_message_process_get_states_response(message);
            break;
        }
        
        case AVATAR_MESSAGE_ID_SUBSCRIBE_RESPONSE:
        {
            // This is telling us that the Avatar has acknowledged that it can 
            // subscribe to our messages
            avatar_message_process_subscribe_response(message);
            break;
        }
        
        case AVATAR_MESSAGE_ID_TCPIP_TRANSFER_RESPONSE:
        {
            //TODO - Add error handling for TCPIP messages
            avatar_message_process_tcp_transfer_response(message);
            break;
        }
        
        default:
        {
            logger_log_warningf(LOGGER_CODE_AVATAR, "AVATAR Slave PROCESSING Unknown Message Type 0x%04X --", message->message_id);
            break;
        }
    }
    
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave PROCESSING End ----------------------------------------------");    
}
    
/**-----------------------------------------------------------------------------
 * Set the callback handler
 * @param callback
 */
void avatar_message_set_callback(AVATAR_CALLBACK callback)
{
    avatar_message_data.callback = callback;
}

/**-----------------------------------------------------------------------------
 * Used internally to keep a record of the current region for use with the 
 * Region menu
 */
void avatar_message_set_region(AVATAR_REGION region)
{
    avatar_message_data.region = region;
}

/**-----------------------------------------------------------------------------
 * Used internally to keep a record of the current sample rate for use with the 
 * avatar_set_sample_rate function
 */
void avatar_message_set_sample_rate(AVATAR_SAMPLE_RATE sample_rate)
{
    avatar_message_data.sample_rate = sample_rate;
}

/**-----------------------------------------------------------------------------
 * Set the callback handler for TCP message data ONLY
 * @param callback
 */
void avatar_message_set_tcp_callback(AVATAR_TCP_CALLBACK callback)
{
    avatar_message_data.tcp_callback = callback;
}

/* *****************************************************************************
 End of File
 */
