/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\avatar_packet_data.h"
#include "app_config.h"
#include "utilities\avatar_packet.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Defines for the size an number of the packets in the Avatar packet pool.
 * Currently the pool contains 3 sizes of packet Small/Medium/Large.
 * Note: If the desired pool does not have a free packet, the system will attempt 
 * to allocate a packet from a larger pool.
 */
#define AVATAR_PACKET_DATA_LEN_SMALL    40              // Small packet size    
#define AVATAR_PACKET_DATA_NUM_SMALL    25              // The number of available Small packets
#define AVATAR_PACKET_DATA_LEN_MEDIUM   200             // Medium packet size
#define AVATAR_PACKET_DATA_NUM_MEDIUM   5               // The number of available Medium packets
#define AVATAR_PACKET_DATA_LEN_LARGE    4196            // Maximum packet size
#define AVATAR_PACKET_DATA_NUM_LARGE    50//15              // The number of available Maximum packets //TODO 23/08-Was 4 changed to 15 to handle Album Art - Previously...This only need to be 4 at the moment for the SA30 (though what about upgrade???))
#define AVATAR_PACKET_CACHE_SIZE        AVATAR_PACKET_DATA_NUM_SMALL + AVATAR_PACKET_DATA_NUM_MEDIUM + AVATAR_PACKET_DATA_NUM_LARGE // The number of available PACKETS

/**
 * Can be used to analyse the usage of the Free data pools in the avatar_packet_find_free_data
 * function.  Comment out when not needed as it makes the code more efficient
 */
//#define AVATAR_PACKET_ANALYSE_FREE_DATA

/**
 * An overlay for the Small, Medium and Large packet data - used for identifying 
 * the next free data
 */
typedef struct
{
    //! Determines whether this is currently in use
    bool in_use;
    
    // This equates to the first byte in the data array
    uint8_t data;
   
} AVATAR_PACKET_DATA_TEMPLATE;

/**
 * Small packet data
 */
typedef struct
{
    //! Determines whether this is currently in use
    bool in_use;
    
    uint8_t data[AVATAR_PACKET_DATA_LEN_SMALL];
   
} AVATAR_PACKET_DATA_SMALL;

/**
 * Small packet data
 */
typedef struct
{
    //! Determines whether this is currently in use
    bool in_use;
    
    uint8_t data[AVATAR_PACKET_DATA_LEN_MEDIUM];
   
} AVATAR_PACKET_DATA_MEDIUM;

/**
 * Small packet data
 */
typedef struct
{
    //! Determines whether this is currently in use
    bool in_use;
    
    uint8_t data[AVATAR_PACKET_DATA_LEN_LARGE];
   
} AVATAR_PACKET_DATA_LARGE;

/**
 * Storage cache
 */
typedef struct
{
    //! The available AVATAR_REQUESTS
    AVATAR_PACKET cache[AVATAR_PACKET_CACHE_SIZE];
    
    //! Small data cache
    AVATAR_PACKET_DATA_SMALL small[AVATAR_PACKET_DATA_NUM_SMALL];
    
    //!  Medium data cache
    AVATAR_PACKET_DATA_MEDIUM medium[AVATAR_PACKET_DATA_NUM_MEDIUM];
    
    //!  Large data cache
    AVATAR_PACKET_DATA_LARGE large[AVATAR_PACKET_DATA_NUM_LARGE];
} AVATAR_PACKET_STORAGE;

typedef struct
{
    //! The data storage cache
    AVATAR_PACKET_STORAGE storage;
    
} AVATAR_PACKET_DATA;

static AVATAR_PACKET_DATA avatar_packet_data =
{
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * The AVATAR_PACKET_DATA type point to either a AVATAR_PACKET_DATA_SMALL/MEDIUM or LARGE
 * @param data - points to the AVATAR_PACKET_DATA storage head e.g. &avatar_package_data.storage.small
 * @param data_size -  The Size of a single storage item e.g. sizeof(AVATAR_PACKET_DATA_SMALL)
 * @param num_item - The number of items in the storage array e.g. AVATAR_PACKET_DATA_NUM_SMALL
 */
static uint8_t* avatar_packet_find_free_data(AVATAR_PACKET_DATA_TEMPLATE* data, uint16_t data_size, uint16_t num_items)
{
    uint8_t* result = NULL;

#ifdef AVATAR_PACKET_ANALYSE_FREE_DATA    
    uint16_t in_use_count = 0;
#endif    
    
    int x;
    for (x = 0; x < num_items; x++)
    {
        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR data %d ", data);
        if (!data->in_use)
        {
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Using data %d ", x);
            // Initialise the data (data_size - size of the in_use boolean))
            memset(&data->data, 0, data_size - sizeof(bool));
            
            if (result == NULL)
            {
                result = &data->data;
                data->in_use = true;
            }
            
#ifndef AVATAR_PACKET_ANALYSE_FREE_DATA            
            break;
#endif            
        }
#ifdef AVATAR_PACKET_ANALYSE_FREE_DATA
        else
        {
            in_use_count++;
        }
#endif        
        // Convert data pointer to uint8_t* for pointer arithmatic
        uint8_t* data_ptr = (uint8_t*)data;
        data_ptr += data_size;
        data = (AVATAR_PACKET_DATA_TEMPLATE*)data_ptr;
    }

#ifdef AVATAR_PACKET_ANALYSE_FREE_DATA
    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet Pool %d %d/%d --", data_size, in_use_count, num_items);
#endif    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Allocates a packet from the pool based on the amount of data required..
 * Note: If there are insufficient packets we get the code falling over despite 
 */
AVATAR_PACKET* avatar_packet_allocate(AVATAR_PACKET_HEADER* header, bool receive)
{
    AVATAR_PACKET* result = NULL;

    // Find the next free packet
    int x;
    for (x = 0; x < AVATAR_PACKET_CACHE_SIZE; x++)
    {
        if (avatar_packet_data.storage.cache[x].state == AVATAR_PACKET_STATE_FREE )
        {
            result = &avatar_packet_data.storage.cache[x];
            result->state = AVATAR_PACKET_STATE_ALLOCATED;
            break;
        }
    }

    if (result != NULL)
    {
        // Copy the header into the packet
        memcpy(&result->header, header, sizeof(AVATAR_PACKET_HEADER));
        
        result->data = NULL;
        
        // Get length
        uint16_t length = avatar_packet_header_get_data_length(header);
        uint8_t pool = 0;

        // Find spare entry in pool to accommodate length
        if (length <= AVATAR_PACKET_DATA_LEN_SMALL)
        {
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Considering SMALL");
            pool = 1;
            result->data = avatar_packet_find_free_data((AVATAR_PACKET_DATA_TEMPLATE*)&avatar_packet_data.storage.small, sizeof(AVATAR_PACKET_DATA_SMALL), AVATAR_PACKET_DATA_NUM_SMALL);

            //if (result->data == NULL)
            //{
            //    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet %d Allocate SMALL Pool FULL ", result->index);
            //}
        }
        
        if ((result->data == NULL) && (length <= AVATAR_PACKET_DATA_LEN_MEDIUM))
        {
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Considering MEDIUM");
            pool = 2;
            result->data = avatar_packet_find_free_data((AVATAR_PACKET_DATA_TEMPLATE*)&avatar_packet_data.storage.medium, sizeof(AVATAR_PACKET_DATA_MEDIUM), AVATAR_PACKET_DATA_NUM_MEDIUM);

            //if (result->data == NULL)
            //{
            //    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet %d Allocate MEDIUM Pool FULL ", result->index);
            //}
        }
        
        if ((result->data == NULL) && (length <= AVATAR_PACKET_DATA_LEN_LARGE))
        {
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Considering LARGE");
            pool = 3;
            result->data = avatar_packet_find_free_data((AVATAR_PACKET_DATA_TEMPLATE*)&avatar_packet_data.storage.large, sizeof(AVATAR_PACKET_DATA_LARGE), AVATAR_PACKET_DATA_NUM_LARGE);

            //if (result->data == NULL)
            //{
            //    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet %d Allocate LARGE Pool FULL ", result->index);
            //}
        }

        if (result->data != NULL)
        {
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet %d Allocated len:%d from pool:%s for %s ", result->index, length, pool == 1 ? "SMALL" : (pool == 2 ? "MEDIUM" : "LARGE"), receive ? "RECEIVE" : "SEND");
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet Alloc %d %d --", result->header.frame_id, result->index);
        }
        else
        {
            // Return the packet to the queue
            logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Packet Error allocating packet of length %d ", length);
            
            avatar_packet_free(result);
            result = NULL;
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Packet Error allocating packet from Cache ");
        
        //for (x = 0; x < AVATAR_PACKET_CACHE_SIZE; x++)
        //{
        //    logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Packet Error %d %d --", x, avatar_packet_data.storage.cache[x].state);
        //    avatar_packet_print(&avatar_packet_data.storage.cache[x], "Error");
        //    logger_process();
        //}
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Frees a packet from the pool
 */
AVATAR_PACKET* avatar_packet_free(AVATAR_PACKET* packet)
{
    if (packet != NULL)
    {
        if (packet->data != NULL)
        {
            // Use ptr majik to set in_use pointer
            bool* in_use_ptr = (bool*)packet->data - sizeof(bool);
            *in_use_ptr = false;
        }

        // Mark the packet as free
        packet->state = AVATAR_PACKET_STATE_FREE;
        
        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Packet Alloc %d %d Freed ", packet->header.frame_id, packet->index);
    }
}

/**-----------------------------------------------------------------------------
 * Initialise the avatar_packet module
 */
void avatar_packet_initialise()
{
    // Just initialise all storage to 0 will set all in_uses to false
    memset(&avatar_packet_data.storage, 0, sizeof(AVATAR_PACKET_STORAGE));
    
    int x;
    for (x = 0; x < AVATAR_PACKET_CACHE_SIZE; x++)
    {
        avatar_packet_data.storage.cache[x].index = x;
        avatar_packet_data.storage.cache[x].state = AVATAR_PACKET_STATE_FREE;
    }
    
    for (x = 0; x < AVATAR_PACKET_DATA_NUM_SMALL; x++)
    {
        avatar_packet_data.storage.small[x].in_use = false;
    }
    
    for (x = 0; x < AVATAR_PACKET_DATA_NUM_MEDIUM; x++)
    {
        avatar_packet_data.storage.medium[x].in_use = false;
    }
    
    for (x = 0; x < AVATAR_PACKET_DATA_NUM_LARGE; x++)
    {
        avatar_packet_data.storage.large[x].in_use = false;
    }
}

/* *****************************************************************************
 End of File
 */
