/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\avatar_process_rule.h"
#include "app_config.h"
#include "utilities\avatar_data.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    AVATAR_PROCESS_RULE* active_rule;

    //! The current number of rules added
    uint8_t num_rules;
    
    //AVATAR_PROCESSING_RULES rules;

    AVATAR_PROCESS_RULE rules[AVATAR_PROCESS_RULE_MAX];
} AVATAR_PROCESSING_RULE_DATA;

static AVATAR_PROCESSING_RULE_DATA avatar_processing_rule_data = 
{
    .num_rules = 0,
    .active_rule = NULL,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Adds a process rule - checks for the existence of a state within a packet
 * @param message_id -  the type of message the rule applies to
 * @param state_id - The state to check for
 * @param callback - the callback to be invoked if the state exists
 * @param wait_for_ack - if true the callback processing is done when the ack is received from the avatar
 */
void avatar_process_rule_add(AVATAR_MESSAGE_ID message_id,
                             AVATAR_STATE_ID state_id,
                             AVATAR_PROCESS_RULE_CALLBACK callback,
                             AVATAR_PROCESS_RULE_ACTION action)
{
    if (avatar_processing_rule_data.num_rules < AVATAR_PROCESS_RULE_MAX)
    {
        AVATAR_PROCESS_RULE* rule = &avatar_processing_rule_data.rules[avatar_processing_rule_data.num_rules];

        rule->active = true;
        rule->message_id = message_id;
        rule->state_id = state_id;
        rule->callback = callback;
        rule->action = action;

        avatar_processing_rule_data.num_rules++;
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Number of Processing Rules exceeded %d --", avatar_processing_rule_data.num_rules);
    }
}

/**-----------------------------------------------------------------------------
 * Deactivates the rule and stops it being considered for new messages
 * @param rule
 * @param packet
 */
void avatar_process_rule_deactivate(AVATAR_PROCESS_RULE* rule, AVATAR_PACKET* packet)
{
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave DEACTIVATING RULE ");
    rule->active = false;
}

/**-----------------------------------------------------------------------------
 * Checks to see if any rules match the supplied packet, and invokes if found
 * @param match
 * @return - true if processed, false if not
 */
bool avatar_process_rule_check(AVATAR_MESSAGE* message, AVATAR_PROCESS_RULE_ACTION action)
{
    bool result = false;
    
    if (avatar_message_get_num_states(message) <= 6)
    {
        AVATAR_PROCESS_RULE* rule = avatar_process_rule_find(message);
        if (rule != NULL)
        {
            //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Slave Found Rule");

            if (rule->callback != NULL)
            {
                if (rule->action == action)
                {
                    avatar_processing_rule_data.active_rule = NULL;

                    rule->callback(rule, message);
                }
                else if (rule->action == AVATAR_PROCESS_RULE_ACTION_ACK)
                {
                    avatar_processing_rule_data.active_rule = rule;
                }
            }

            result = true;
        }  
    }
     
    return result;
}

/**-----------------------------------------------------------------------------
 * Compares message input with the Rules to look for a match
 * @param match
 * @return 
 */
AVATAR_PROCESS_RULE* avatar_process_rule_find(AVATAR_MESSAGE* message)
{
    AVATAR_PROCESS_RULE* result = NULL;
    
    // Get Message ID from the packet
    
    AVATAR_MESSAGE_ID message_id = avatar_packet_get_message_id(message->packet);
    
    int x;
    for (x = 0; x < AVATAR_PROCESS_RULE_MAX; x++)
    {
        AVATAR_PROCESS_RULE* rule = &avatar_processing_rule_data.rules[x];
        
        // Check message Id
        if (rule->message_id == message_id)
        {
            if (avatar_message_has_state(message, rule->state_id))
            {
                result = rule;
                break;
            }
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Processes the active rule if there is one
 * @return 
 */
bool avatar_process_rule_process_active()
{
    if ((avatar_processing_rule_data.active_rule != NULL) && 
       (avatar_processing_rule_data.active_rule->callback != NULL))
    {
       avatar_processing_rule_data.active_rule->callback(avatar_processing_rule_data.active_rule, NULL);
       avatar_processing_rule_data.active_rule = NULL;
    } 
}

/* *****************************************************************************
 End of File
 */
