/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/crestron_discovery.h"
#include "tcpip/tcpip.h"
#include "app_config.h"
#include "utilities/logger.h"
#include "utilities/tcpip.h"
#include "utilities/udp.h"
#include "utilities/uuid.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define CRESTRON_DISCOVERY_ADDRESS                  "239.255.255.250"    
#define CRESTRON_DISCOVERY_PORT                     41794
#define CRESTRON_DISCOVERY_FIRMWARE_NAME_LEN        128
#define CRESTRON_DISCOVERY_RESPONSE_PACKET_DATA_LEN sizeof(uint16_t) + CRESTRON_DISCOVERY_HOST_NAME_LEN + CRESTRON_DISCOVERY_FIRMWARE_NAME_LEN
#define CRESTRON_DISCOVERY_EMPTY_ID                 0x00

/**
 * The ordinal position of different fields in a message
 */
typedef enum
{
    CRESTRON_MESSAGE_IDX_TYPE           = 0,
    CRESTRON_MESSAGE_IDX_TRANSACTION    = 1,
    CRESTRON_MESSAGE_IDX_FLAGS          = 2,
    CRESTRON_MESSAGE_IDX_DATA_LENGTH    = 4,
    CRESTRON_MESSAGE_IDX_DATA           = 6,
            
} CRESTRON_MESSAGE_IDX;

/**
 */
typedef enum
{
    CRESTRON_PACKET_TYPE_AUTODISCOVERY_QUERY    = 0x14,
    CRESTRON_PACKET_TYPE_AUTODISCOVERY_RESPONSE = 0x15,
    CRESTRON_PACKET_TYPE_HOSTNAME_QUERY         = 0x16,
    CRESTRON_PACKET_TYPE_HOSTNAME_RESPONSE      = 0x17,
    CRESTRON_PACKET_TYPE_HOSTNAME_SET           = 0x18,
    CRESTRON_PACKET_TYPE_HOSTNAME_ACKNOWLEDGE   = 0x19,
} CRESTRON_PACKET_TYPE;

/**
 * Flags received in the query message
 */
typedef enum
{
    CRESTRON_QUERY_FLAG_NONE                    = 0x0000,
    CRESTRON_QUERY_FLAG_RESPOND                 = 0x0001,
    CRESTRON_QUERY_FLAG_SEND_IP_TABLE           = 0x0002,
    CRESTRON_QUERY_FLAG_CLEAR_IP_TABLE          = 0x0004,
    CRESTRON_QUERY_FLAG_SET_IP_ENTRY            = 0x0008,
    CRESTRON_QUERY_FLAG_START_LIGHT_AND_POLL    = 0x0010,
    CRESTRON_QUERY_FLAG_STOP_LIGHT_AND_POLL     = 0x0020,
} CRESTRON_QUERY_FLAG;

/**
 * Flags sent in the response message
 */
typedef enum
{
    CRESTRON_RESPONSE_FLAG_MATCH_FOUND              = 0x0001,
    CRESTRON_RESPONSE_FLAG_CONFIGURATION_REQUESTED  = 0x0002,   // Configuration Requested. Could be sent broadcast by an unconfigured device when it starts up.
    CRESTRON_RESPONSE_FLAG_CONFIGURATION_BLOCKED    = 0x0004,   // Configuration Blocked. To be sent if the device has manually configured not to response to Autodiscovery configuration packets.
    CRESTRON_RESPONSE_FLAG_LIGHT_AND_POLL_STARTED   = 0x0008,
    CRESTRON_RESPONSE_FLAG_LIGHT_AND_POLL_STOPPED   = 0x0010,
    CRESTRON_RESPONSE_FLAG_SETUP_BUTTON_PRESSED     = 0x0020,
    CRESTRON_RESPONSE_FLAG_CONFUGURATION_SUCCESSFUL = 0x0040,
    CRESTRON_RESPONSE_FLAG_IP_TABLE_CLEARED         = 0x0080,
} CRESTRON_RESPONSE_FLAG;

/**
 * Processing states
 */
typedef enum
{
    CRESTRON_STATE_INIT,
    CRESTRON_STATE_RUNNING,
    CRESTRON_STATE_SEND_IP_TABLE,
    CRESTRON_STATE_START_LIGHT_AND_POLL,
    CRESTRON_STATE_SET_IP_ENTRY,
    CRESTRON_STATE_STOP_LIGHT_AND_POLL,
    CRESTRON_STATE_SEND_HOSTNAME_RESPONSE,
    CRESTRON_STATE_CLEAR_IP_TABLE,
    CRESTRON_STATE_HOSTNAME_SET,
} CRESTRON_STATE;
    
/**
 * Runtime data
 */
typedef struct
{
    //! Determines whether the module is active or not
    bool active;
    
    //! The callback used to indicate configuration start/end
    CRESTRON_DISCOVERY_CONFIGURATION_CALLBACK configuration_callback;
    
    //! Lookup table for host names
    CRESTRON_DISCOVERY_HOST_TABLE host_table;
    
    //! The address of the device that sent the message
    IP_ADDR message_address;
    
    //! The hostname of the message sender to be included in the lookup table
    char message_hostname[CRESTRON_DISCOVERY_HOST_NAME_LEN];
    
    //! The Id passed in by the current incoming message
    uint16_t message_id;
    
    //! The current processing state
    CRESTRON_STATE state;
    
    //! The running transaction number for messages sent
    uint8_t transaction_number;
} CRESTRON_DATA;

static CRESTRON_DATA crestron_data = {
    .transaction_number = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations
static CRESTRON_QUERY_FLAG crestron_discovery_has_flag(uint16_t flags, CRESTRON_QUERY_FLAG flag);
static void crestron_discovery_send_message(uint8_t message_type, uint16_t flags,
        char* data, uint16_t data_length,
        IP_ADDR* address,
        uint16_t port);
static void crestron_discovery_send_response_packet(uint16_t response_flag, uint16_t id);

/**-----------------------------------------------------------------------------
 * Callback for udp messages 
 */
static bool crestron_discovery_callback(int port, struct sockaddr_in* address, int* address_len, char* message, int len)
{
    bool result = false;

    if (port == CRESTRON_DISCOVERY_PORT)
    {
        bool send_response = false;
        
        // Parse message and set state accordingly
        // We will only get 0x14 (QUERY)
        if (message[CRESTRON_MESSAGE_IDX_TYPE] == CRESTRON_PACKET_TYPE_AUTODISCOVERY_QUERY)
        {
            // Get the flags
            uint16_t flags;
            memcpy(&flags, &message[CRESTRON_MESSAGE_IDX_FLAGS], 2);
            
            // Check we've got a CRESTRON_QUERY_FLAG_RESPOND
            if (crestron_discovery_has_flag(flags, CRESTRON_QUERY_FLAG_RESPOND))
            {
                // Copy the data from the message (note not copying port as that is fixed..)
                crestron_data.message_address.Val = address->sin_addr.S_un.S_addr;
                memcpy(&crestron_data.message_id, &message[CRESTRON_MESSAGE_IDX_DATA], sizeof(uint16_t));
                memcpy(crestron_data.message_hostname, &message[CRESTRON_MESSAGE_IDX_DATA + sizeof(uint16_t)], CRESTRON_DISCOVERY_HOST_NAME_LEN);

                logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Received Auto Discovery Query from %d:%s ", crestron_data.message_id, crestron_data.message_hostname);
                
                if (crestron_discovery_has_flag(flags, CRESTRON_QUERY_FLAG_SEND_IP_TABLE))
                {
                    crestron_data.state = CRESTRON_STATE_SEND_IP_TABLE;
                }
                else if (crestron_discovery_has_flag(flags, CRESTRON_QUERY_FLAG_START_LIGHT_AND_POLL))
                {
                    crestron_data.state = CRESTRON_STATE_START_LIGHT_AND_POLL;
                }
                else if (crestron_discovery_has_flag(flags, CRESTRON_QUERY_FLAG_SET_IP_ENTRY))
                {
                    crestron_data.state = CRESTRON_STATE_SET_IP_ENTRY;
                }
                else if (crestron_discovery_has_flag(flags, CRESTRON_QUERY_FLAG_STOP_LIGHT_AND_POLL))
                {
                    crestron_data.state = CRESTRON_STATE_STOP_LIGHT_AND_POLL;
                }
                else if (crestron_discovery_has_flag(flags, CRESTRON_QUERY_FLAG_CLEAR_IP_TABLE))
                {
                    crestron_data.state = CRESTRON_STATE_CLEAR_IP_TABLE;
                }
            }
        }
        else if (message[CRESTRON_MESSAGE_IDX_TYPE] == CRESTRON_PACKET_TYPE_HOSTNAME_QUERY)
        {
            memcpy(crestron_data.message_hostname, &message[CRESTRON_MESSAGE_IDX_DATA], CRESTRON_DISCOVERY_HOST_NAME_LEN);

            logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Received Hostname Query from %s ", crestron_data.message_hostname);
            
            crestron_data.state = CRESTRON_STATE_SEND_HOSTNAME_RESPONSE;
        }
        else if (message[CRESTRON_MESSAGE_IDX_TYPE] == CRESTRON_PACKET_TYPE_HOSTNAME_SET)
        {
            memcpy(crestron_data.message_hostname, &message[CRESTRON_MESSAGE_IDX_DATA], CRESTRON_DISCOVERY_HOST_NAME_LEN);

            logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Received Hostname Set from %s ", crestron_data.message_hostname);
            
            crestron_data.state = CRESTRON_STATE_HOSTNAME_SET;
        }

        if (send_response)
        {
            result = true;
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Clears the IP Table and sends a response
 */
void crestron_discovery_clear_ip_table()
{
    logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Clearing Host Lookup Table");
    
    memset(&crestron_data.host_table, 0, sizeof(CRESTRON_DISCOVERY_HOST_TABLE));

    crestron_discovery_send_response_packet(
            CRESTRON_RESPONSE_FLAG_IP_TABLE_CLEARED, 
            CRESTRON_DISCOVERY_EMPTY_ID);     
}

/**-----------------------------------------------------------------------------
 * Copies the firmware name into the supplied buffer e.g. SA20-1.10
 * @param buffer
 * @return firmware name length
 */
static void crestron_discovery_get_firmwarename(char* buffer)
{
    sprintf(buffer, "%s-%d.%d", SOFTWARE_MODEL_STRING, SOFTWARE_VERSION_MAJOR, SOFTWARE_VERSION_MINOR);
}

/**-----------------------------------------------------------------------------
 * Copies the hostname into the supplied buffer.
 * Note the tcpip network name is used as the host name.  This may put limits on 
 * the length of host name that can be specified which may in turn cause problems 
 * with the set message.
 * 
 * @param buffer
 * @return hostname length
 */
static uint16_t crestron_discovery_get_hostname(char* buffer)
{
    tcpip_get_network_name_string(buffer, CRESTRON_DISCOVERY_HOST_NAME_LEN);
    
//    uint16_t result = 0;
//    
//    if (strlen(crestron_data.hostname) == 0)
//    {
//        // Setup the hostname buffer
//        memcpy(crestron_data.hostname, SOFTWARE_MODEL_STRING, strlen(SOFTWARE_MODEL_STRING));
//        
//        // Append the MAC address
//        tcpip_get_mac_address_string(&crestron_data.hostname[strlen(SOFTWARE_MODEL_STRING)], 13);
//    }
//    
//    memcpy(buffer, crestron_data.hostname, strlen(crestron_data.hostname));
    
    return strlen(buffer);
}

/**-----------------------------------------------------------------------------
 * Returns the current transaction number and increments it.
 * @return 
 */
static uint8_t crestron_discovery_get_next_transaction_number()
{
    return crestron_data.transaction_number++;    
}

/**-----------------------------------------------------------------------------
 * Determines whether the 
 */
static CRESTRON_QUERY_FLAG crestron_discovery_has_flag(uint16_t flags, CRESTRON_QUERY_FLAG flag)
{
    return (flags & flag) == flag ? flag : CRESTRON_QUERY_FLAG_NONE;
}

/**-----------------------------------------------------------------------------
 * Gets the id matching the hostname from the lookup table
 * @param hostname
 * @return 
 */
static uint16_t crestron_discovery_id_for_hostname(char* hostname)
{
    uint16_t result = CRESTRON_DISCOVERY_EMPTY_ID;
    
    // Generate MD5 Checksum
    uint8_t checksum[MD5_DIGEST_SIZE];
    CRYPT_MD5_CTX md5_context;
    CRYPT_MD5_Initialize(&md5_context);
    CRYPT_MD5_DataAdd(&md5_context, (char*)hostname, strlen(hostname));
    CRYPT_MD5_Finalize(&md5_context, checksum);
    
    int x;
    for (x = 0; x < CRESTRON_DISCOVERY_MAX_HOSTS; x++)
    {
        if (crestron_data.host_table.controllers[x].hostname_checksum == checksum)
        {
            logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Hostname match %d:%s ", crestron_data.host_table.controllers[x].id, hostname);
    
            result = crestron_data.host_table.controllers[x].id;
            break;
        }
    }
    return result;
}
   
/**-----------------------------------------------------------------------------
 * Sends the startup auto discovery message
 */
static void crestron_discovery_send_hostname_response()
{
    char data[17];
    uint16_t data_length = 0;
    
    // IPAddress
    IP_ADDR ip_address = tcpip_get_runtime_ip_address();            
    memcpy(data, ip_address.v, 4);
    data_length += 4;

    // Hostname (12 chars + /0)
    data_length += crestron_discovery_get_hostname(&data[data_length]);

    // Set up send address
    IP_ADDR address;
    TCPIP_Helper_StringToIPAddress(CRESTRON_DISCOVERY_ADDRESS, &address);

    crestron_discovery_send_message(
            CRESTRON_PACKET_TYPE_HOSTNAME_RESPONSE, 
            0,
            data,
            data_length,
            &address,
            CRESTRON_DISCOVERY_PORT);
}

/**-----------------------------------------------------------------------------
 * Sends a UDP message
 */
static void crestron_discovery_send_message(uint8_t message_type, uint16_t flags,
        char* data, uint16_t data_length,
        IP_ADDR* address,
        uint16_t port)
{
    // Initialise message data buffer
    char message_data[UDP_MESSAGE_MAX_SIZE];
    memset(message_data, 0, UDP_MESSAGE_MAX_SIZE);
        
    message_data[CRESTRON_MESSAGE_IDX_TYPE] = message_type;
    message_data[CRESTRON_MESSAGE_IDX_TRANSACTION] = crestron_discovery_get_next_transaction_number();

    // Flags
    memcpy(&message_data[CRESTRON_MESSAGE_IDX_FLAGS], &flags, 2);
    
    // Data length
    memcpy(&message_data[CRESTRON_MESSAGE_IDX_DATA_LENGTH], &data_length, 2);

    if (data_length > 0)
    {
        // Data
        memcpy(&message_data[CRESTRON_MESSAGE_IDX_DATA], data, data_length);
    }

    logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Sending message %d with flags %d to %s:%d ", message_type, flags, CRESTRON_DISCOVERY_ADDRESS, port);
    
    // Send the buffer via UDP
    udp_sendto(address, port, message_data, 6 + data_length);
}

/**-----------------------------------------------------------------------------
 * Sends an autodiscovery response packet 
 */
static void crestron_discovery_send_response_packet(uint16_t response_flag, uint16_t id)
{
    char data[CRESTRON_DISCOVERY_RESPONSE_PACKET_DATA_LEN];
    memset(data, 0, CRESTRON_DISCOVERY_RESPONSE_PACKET_DATA_LEN);
    
    // Copy Id
    memcpy(&data[0], &id, sizeof(uint16_t));
    
    // Copy Hostname
    crestron_discovery_get_hostname(&data[sizeof(uint16_t)]);
    
    // Copy Firmware name
    crestron_discovery_get_firmwarename(&data[sizeof(uint16_t) + CRESTRON_DISCOVERY_HOST_NAME_LEN]);
    
    crestron_discovery_send_message(
            CRESTRON_PACKET_TYPE_AUTODISCOVERY_RESPONSE,
            response_flag,
            data,
            CRESTRON_DISCOVERY_RESPONSE_PACKET_DATA_LEN,
            &crestron_data.message_address,
            CRESTRON_DISCOVERY_PORT);
}

/**-----------------------------------------------------------------------------
 * Adds or updates a controller lookup based on the current message details
 */
static void crestron_discovery_set_hostname_from_message()
{
    //TODO Note this will truncate if > TCPIP_NETWORK_NAME_MAX_LEN so may no be workable
    tcpip_set_network_name(crestron_data.message_hostname, strlen(crestron_data.message_hostname));

    crestron_discovery_send_message(
            CRESTRON_PACKET_TYPE_HOSTNAME_ACKNOWLEDGE,
            CRESTRON_QUERY_FLAG_NONE,
            NULL,
            0,
            &crestron_data.message_address,
            CRESTRON_DISCOVERY_PORT);
}

/**-----------------------------------------------------------------------------
 * Adds or updates a controller lookup based on the current message details
 */
static void crestron_discovery_set_ip_entry_from_message()
{
    // Try and find an existing entry for the host name
    // If not found then add in first available position
    int empty_idx = -1;
    bool found = false;
    
    // Generate MD5 Checksum
    uint8_t checksum[MD5_DIGEST_SIZE];
    CRYPT_MD5_CTX md5_context;
    CRYPT_MD5_Initialize(&md5_context);
    CRYPT_MD5_DataAdd(&md5_context, (char*)crestron_data.message_hostname, strlen(crestron_data.message_hostname));
    CRYPT_MD5_Finalize(&md5_context, checksum);
    
    int x;
    for (x = 0; x < CRESTRON_DISCOVERY_MAX_HOSTS; x++)
    {
        if (crestron_data.host_table.controllers[x].hostname_checksum == checksum)
        {
            logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Setting IP Entry %d:%s at %d ", crestron_data.message_id, crestron_data.message_hostname, x);
            
            crestron_data.host_table.controllers[x].id = crestron_data.message_id;
    
            settings_set(SETTING_CRESTRON_DISCOVERY_LOOKUP_TABLE, &crestron_data.host_table);

            found = true;
            break;
        }
        else if ((empty_idx == -1) && (crestron_data.host_table.controllers[x].id = CRESTRON_DISCOVERY_EMPTY_ID))
        {
            empty_idx = x;
        }
    }
    
    if (!found)
    {
        if (empty_idx != -1)
        {
            logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Setting IP Entry %d:%s at %d ", crestron_data.message_id, crestron_data.message_hostname, empty_idx);
            
            // Copy the new entry into the table
            crestron_data.host_table.controllers[empty_idx].id = crestron_data.message_id;
            memcpy(crestron_data.host_table.controllers[empty_idx].hostname_checksum, &checksum, MD5_DIGEST_SIZE);
    
            settings_set(SETTING_CRESTRON_DISCOVERY_LOOKUP_TABLE, &crestron_data.host_table);
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_CRESTRON, "CRESTRON Unable to save host lookup %d:%s - no spare entries ", crestron_data.message_id, crestron_data.message_hostname);
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**-----------------------------------------------------------------------------
 * Returns whether Control 4 SDDP messaging is active or not
 * @return 
 */
bool crestron_discovery_get_active()
{
    return crestron_data.active;
}

/**-----------------------------------------------------------------------------
 * Initialises the crestron_discovery module
 */
void crestron_discovery_initialise(CRESTRON_DISCOVERY_CONFIGURATION_CALLBACK configuration_callback)
{
    logger_log_infof(LOGGER_CODE_CRESTRON, "Initialising Crestron Discovery ");
    
    settings_get(SETTING_CRESTRON_DISCOVERY_ACTIVE, &crestron_data.active);
    logger_log_debugf(LOGGER_CODE_CRESTRON, "CRESTRON Active : %s ", crestron_data.active ? "Yes" : "No");
    
    crestron_data.configuration_callback = configuration_callback;
    
    // Initialise the hostname
    memset(crestron_data.message_hostname, 0, CRESTRON_DISCOVERY_HOST_NAME_LEN);
    crestron_data.message_id = 0x00;
    
    settings_get(SETTING_CRESTRON_DISCOVERY_LOOKUP_TABLE, &crestron_data.host_table);
    
    // Register a callback with UDP
    udp_add_callback(crestron_discovery_callback);
}
     
/**-----------------------------------------------------------------------------
 * Performs Crestron processing
 */
void crestron_discovery_process()
{
    if (crestron_discovery_get_active())
    {
        // We need to manage the state for the configuration process
        switch(crestron_data.state)
        {
            case CRESTRON_STATE_INIT:
            {
                if (tcpip_is_connected())
                {
                    crestron_discovery_send_startup();

                    crestron_data.state = CRESTRON_STATE_RUNNING;
                }

                break;
            }

            case CRESTRON_STATE_RUNNING:
            {
                // Waiting for an external message
                break;
            }

            case CRESTRON_STATE_SEND_IP_TABLE:
            {
                // Check whether the message hostname exists in the lookup table
                uint16_t id = crestron_discovery_id_for_hostname(crestron_data.message_hostname);

                crestron_discovery_send_response_packet(CRESTRON_RESPONSE_FLAG_MATCH_FOUND, id);    

                crestron_data.state = CRESTRON_STATE_RUNNING;

                break;
            }

            case CRESTRON_STATE_START_LIGHT_AND_POLL:
            {
                // invoke callback to trigger polling
                if (crestron_data.configuration_callback != NULL)
                {
                    crestron_data.configuration_callback(
                            CRESTRON_DISCOVERY_CONFIGURATION_STATE_START,
                            crestron_data.message_hostname);
                }

                crestron_discovery_send_response_packet(
                        CRESTRON_RESPONSE_FLAG_LIGHT_AND_POLL_STARTED, 
                        CRESTRON_DISCOVERY_EMPTY_ID);    

                crestron_data.state = CRESTRON_STATE_RUNNING;

                break;
            }

            case CRESTRON_STATE_SET_IP_ENTRY:
            {
                // Store the IP entry
                crestron_discovery_set_ip_entry_from_message();

                crestron_discovery_send_response_packet(
                        CRESTRON_RESPONSE_FLAG_CONFUGURATION_SUCCESSFUL, 
                        CRESTRON_DISCOVERY_EMPTY_ID);    

                crestron_data.state = CRESTRON_STATE_RUNNING;

                break;
            }

            case CRESTRON_STATE_STOP_LIGHT_AND_POLL:
            {
                // Invoke callback to stop polling
                if (crestron_data.configuration_callback != NULL)
                {
                    crestron_data.configuration_callback(
                            CRESTRON_DISCOVERY_CONFIGURATION_STATE_STOP,
                            crestron_data.message_hostname);
                }

                crestron_discovery_send_response_packet(
                        CRESTRON_RESPONSE_FLAG_LIGHT_AND_POLL_STOPPED, 
                        CRESTRON_DISCOVERY_EMPTY_ID);    

                crestron_data.state = CRESTRON_STATE_RUNNING;

                break;
            }

            case CRESTRON_STATE_SEND_HOSTNAME_RESPONSE:
            {
                crestron_discovery_send_hostname_response();

                crestron_data.state = CRESTRON_STATE_RUNNING;

                break;
            }

            case CRESTRON_STATE_CLEAR_IP_TABLE:
            {
                crestron_discovery_clear_ip_table();

                crestron_data.state = CRESTRON_STATE_RUNNING;

                break;
            }

            case CRESTRON_STATE_HOSTNAME_SET:
            {
                crestron_discovery_set_hostname_from_message();

                crestron_data.state = CRESTRON_STATE_RUNNING;

                break;
            }
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Sends the button_pressed auto discovery message
 */
void crestron_discovery_send_button_pressed()
{
    if (crestron_discovery_get_active())
    {
        crestron_discovery_send_response_packet(
                CRESTRON_RESPONSE_FLAG_SETUP_BUTTON_PRESSED, 
                CRESTRON_DISCOVERY_EMPTY_ID);    
    }
}
   
/**-----------------------------------------------------------------------------
 * Sends the startup auto discovery message
 */
void crestron_discovery_send_startup()
{
    if (crestron_discovery_get_active())
    {
        crestron_discovery_send_hostname_response();
    }
}
    
/**-----------------------------------------------------------------------------
 * Sets whether Crestron discovery messaging is active or not
 * @param active - true - messaging is active, false - messaging is off
 */
void crestron_discovery_set_active(bool active)
{
    crestron_data.active = active;
    crestron_data.state = CRESTRON_STATE_RUNNING;
    
    settings_set(SETTING_CRESTRON_DISCOVERY_ACTIVE, &crestron_data.active);
}
    
/* *****************************************************************************
 End of File
 */
