/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\display_message_display.h"
#include "app_config.h"
#include <stdlib.h>
#include <string.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

static DISPLAY_MESSAGE_SCREEN_METADATA current_screen_metadata =
{
    .screen_id = DISPLAY_MESSAGE_DATA_ID_INVALID,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
     
/**
 * Can be used to get the action from a a Display Message that represents a
 * Screen Action message
 * @param message
 * @return - The action, DISPLAY_MESSAGE_SCREEN_ACTION_UNKNOWN if invalid message
 */
DISPLAY_MESSAGE_SCREEN_ACTION display_message_action_get_action(DISPLAY_MESSAGE* message)
{
    DISPLAY_MESSAGE_SCREEN_ACTION result = DISPLAY_MESSAGE_SCREEN_ACTION_UNKNOWN;

    if (message->message_type == DISPLAY_MESSAGE_TYPE_SCREEN_ACTION)
    {
        // Note this converts a uint8_t to a unit16_t
        result = (DISPLAY_MESSAGE_SCREEN_ACTION)message->data[0];
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
* Can be used to get the Data Id from a a Display Message that represents a
* Screen Action message
* @param message
* @return - The data Id, DISPLAY_MESSAGE_DATA_ID_INVALID if invalid message
*/
DISPLAY_MESSAGE_DATA_ID display_message_action_get_data_id(DISPLAY_MESSAGE* message)
{
    DISPLAY_MESSAGE_DATA_ID result = DISPLAY_MESSAGE_DATA_ID_INVALID;

    if (message->message_type == DISPLAY_MESSAGE_TYPE_SCREEN_ACTION)
    {
        // Note this converts a uint8_t to a unit16_t
        result = (DISPLAY_MESSAGE_DATA_ID)message->data[1];
        result += (message->data[2] * 256);
    }
    
    return result;
}   
    
/**-----------------------------------------------------------------------------
 * Gets the number of data items present in a Screen Data message
 * @param message
 * @return 
 */
uint8_t display_message_data_get_num_items(DISPLAY_MESSAGE* message)
{
    uint8_t result = 0;
    
    if (message->message_type == DISPLAY_MESSAGE_TYPE_SCREEN_DATA)
    {
        // Note this converts a uint8_t to a unit16_t
        result = (uint8_t)message->data[0];
     }
    
    return result;
}
     
/**-----------------------------------------------------------------------------
 * Returns a pointer to a struct holding information about the current screen 
 */
DISPLAY_MESSAGE_SCREEN_METADATA* display_message_metadata_current()
{
    return &current_screen_metadata;
}

/**-----------------------------------------------------------------------------
 * Updates the current screen metadata from a DisplayScreen message
 * @param display_message
 */
void display_message_metadata_update_from_screen(DISPLAY_MESSAGE* display_message)
{
    if (display_message->message_type == DISPLAY_MESSAGE_TYPE_DISPLAY_SCREEN)
    {
        // Set the meta-data screen id
        current_screen_metadata.screen_id = display_message_screen_get_id(display_message);
        current_screen_metadata.screen_type = display_message_screen_get_type(display_message);
        current_screen_metadata.currently_selected_id = DISPLAY_MESSAGE_DATA_ID_INVALID;
        display_message_screen_get_title(display_message, current_screen_metadata.screen_title, DISPLAY_MESSAGE_DISPLAY_TITLE_LENGTH);
        display_message_screen_data_clear(&current_screen_metadata.screen_data);
    }    
}
    
/**-----------------------------------------------------------------------------
 * Can be used to get the Screen Id from a Display Message that represents a
 * Display Screen message
 * @param message
 * @return - The Screen Id, DISPLAY_MESSAGE_DATA_ID_INVALID if invalid message
 */
DISPLAY_MESSAGE_DATA_ID display_message_screen_get_id(DISPLAY_MESSAGE* message)
{
    DISPLAY_MESSAGE_DATA_ID result = DISPLAY_MESSAGE_DATA_ID_INVALID;
    
    if (message->message_type == DISPLAY_MESSAGE_TYPE_DISPLAY_SCREEN)
    {
        // Note this converts a uint8_t to a unit16_t
        result = (DISPLAY_MESSAGE_DATA_ID)message->data[1];
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Can be used to get the title from a from a Display Message that represents a
 * Display Screen message
 * @param message
 * @param title - A buffer to hold the title
 * @param len - The size of the buffer
 */
void display_message_screen_get_title(DISPLAY_MESSAGE* message, char* title, size_t len)
{
    if (title != NULL)
    {
        // Clear the title buffer
        memset(title, 0, len);

        if (message->message_type == DISPLAY_MESSAGE_TYPE_DISPLAY_SCREEN)
        {
            // Get the length of the title from the BIN16 field
            uint16_t data_len = message->data[2];
            data_len += (message->data[3] * 256);

            // Copy the title
            memcpy(title, &message->data[4], min(data_len, len));
        }
   }
}
    
/**-----------------------------------------------------------------------------
 * Can be used to obtain the Screen Type from a Display Message that represents a
 * Display Screen message
 * @param message
 * @return - The screen type, DISPLAY_MESSAGE_SCREEN_TYPE_UNKNOWN if invalid message.
 */
DISPLAY_MESSAGE_SCREEN_TYPE display_message_screen_get_type(DISPLAY_MESSAGE* message)
{
    DISPLAY_MESSAGE_SCREEN_TYPE result = DISPLAY_MESSAGE_SCREEN_TYPE_UNKNOWN;
    
    if (message->message_type == DISPLAY_MESSAGE_TYPE_DISPLAY_SCREEN)
    {
        // Note this converts a uint8_t to a unit16_t
        result = (DISPLAY_MESSAGE_SCREEN_TYPE)message->data[0];
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Removes all the data items for the screen data 
 */
void display_message_screen_data_clear(DISPLAY_MESSAGE_SCREEN_DATA* data)
{
    int x;
    for (x = 0; x < DISPLAY_MESSAGE_SCREEN_DATA_MAX_DATA_ITEMS; x++)
    {
        DISPLAY_MESSAGE_SCREEN_DATA_ITEM* data_item =  &data->data[x];
        
        // Clear the data
        memset((uint8_t*)data_item, 0, sizeof(DISPLAY_MESSAGE_SCREEN_DATA_ITEM));
        
        // Set in_use = false;
        data_item->has_changed = false;
        data_item->in_use = false;
        data_item->image_id = IMAGE_IDENTIFIER_NONE;
        data_item->selected = false;
        data_item->edit_pos = -1;
    }
    
    data->num_items = 0;
}

/**-----------------------------------------------------------------------------
 * Attempts to find a data_item based on the data_id in a screen data struct
 * Ii assumes that data_items are allocated in ascending order, so that as soon 
 * as it finds one that is not in use it assumes it has parsed all available items
 */
DISPLAY_MESSAGE_SCREEN_DATA_ITEM* display_message_screen_data_find_item(
        DISPLAY_MESSAGE_SCREEN_DATA* data, 
        DISPLAY_MESSAGE_DATA_ID data_id,
        bool allocate)
{
    DISPLAY_MESSAGE_SCREEN_DATA_ITEM* result = NULL;
    
    int x;
    for (x = 0; x < DISPLAY_MESSAGE_SCREEN_DATA_MAX_DATA_ITEMS; x++)
    {
        if (data->data[x].in_use)
        {
            if (data->data[x].data_id == data_id)
            {
                result = &data->data[x];
                break;
            }
        }
        else
        {
            // We have reached the end of the items that have been allocated
            if (allocate)
            {
                data->num_items++;
                
                data->data[x].has_changed = false;
                data->data[x].in_use = true;
                data->data[x].data_id = data_id;
                data->data[x].selected = false; // default selected to false as its only ever set to true
                result = &data->data[x];
            }
            
            break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 *  Clears down the Screen data struct and re-builds from the current Screen Data DISPLAY_MESSAGE
 * This is primarily designed for use with Option List/ Option List with Data screens
 */
void display_message_screen_data_replace(DISPLAY_MESSAGE_SCREEN_DATA* data, DISPLAY_MESSAGE* message)
{
    if (message->message_type == DISPLAY_MESSAGE_TYPE_SCREEN_DATA)
    {
//        logger_log_debugf(0, "DISPLAY CHANGED REPLACING SCREEN DATA --");
//        logger_process();
        
        // First copy the screen data
        DISPLAY_MESSAGE_SCREEN_DATA screen_data_old;
        memcpy(&screen_data_old, data, sizeof(DISPLAY_MESSAGE_SCREEN_DATA));
        
        // First clear down any data that is already there
        display_message_screen_data_clear(data);
    
        // Now build the screen data from the data items in the display message
        // Setup the map for ordinal access
        DISPLAY_MESSAGE_DATA_MAP map;
        display_message_data_map_build(message, &map);
    
        int x;
        for (x = 0; x < map.num_items; x++)
        {
            // First find the DISPLAY_MESSAGE_SCREEN_DATA_ITEM for the current 
            // data_id, and if its not already allocated, allocate it
            DISPLAY_MESSAGE_SCREEN_DATA_ITEM* data_item = display_message_screen_data_find_item(data, map.items[x].data_id, true);
            
            // get a reference to the previous version of the screen data
            DISPLAY_MESSAGE_SCREEN_DATA_ITEM* data_item_old = display_message_screen_data_find_item(&screen_data_old, map.items[x].data_id, false);

            if (data_item != NULL)
            {
                logger_log_debugf(0, "DATA %d %d %d %d --", x, map.items[x].length, map.items[x].data_id, map.items[x].attribute_id);
                // Check the attribute type
                //  DISPLAY_MESSAGE_ATTRIBUTE_TITLE = 0x00,
                //  DISPLAY_MESSAGE_ATTRIBUTE_DATA = 0x01,
                //  DISPLAY_MESSAGE_ATTRIBUTE_ICON = 0x02,
                switch (map.items[x].attribute_id)
                {
                    case DISPLAY_MESSAGE_ATTRIBUTE_TITLE:
                    {
                        memset(data_item->title, 0, DISPLAY_MESSAGE_DISPLAY_TITLE_LENGTH);
                        memcpy(data_item->title, map.items[x].data, map.items[x].length);
                        
                        if (data_item_old != NULL)
                        {
                            if (strcmp(data_item->title, data_item_old->title) != 0)
                            {
//if ((data_item->data_id >= 4) && (data_item->data_id <=4))
//{
//logger_log_debugf(LOGGER_CODE_DEBUG, "TITLE CHANGE %d (%d %d %d)  %s  %s --", data_item->data_id, data_item->title[0], data_item->title[1], data_item->title[2], data_item->title, data_item_old->title);
//}
                                data_item->has_changed = true;
                            }
                        }
                        else
                        {
//logger_log_debugf(LOGGER_CODE_DEBUG, "TITLE CHANGE %d  %s %s --", data_item->data_id, data_item->title, data_item_old->title);
                            data_item->has_changed = true;
                        }
                        
                        break;
                    }

                    case DISPLAY_MESSAGE_ATTRIBUTE_DATA:
                    {
                        // There will only be DISPLAY_MESSAGE_ATTRIBUTE_DATA attributes
                        // present when the screen type is DISPLAY_MESSAGE_SCREEN_TYPE_OPTION_LIST_WITH_DATA
                        // The value will always be a string
                        memset(data_item->value, 0, DISPLAY_MESSAGE_DISPLAY_VALUE_LENGTH);
                        memcpy(data_item->value, map.items[x].data, map.items[x].length);
                        
                        if (data_item_old != NULL)
                        {
                            if (strcmp(data_item->value, data_item_old->value) != 0)
                            {
                                data_item->has_changed = true;
                            }
                        }
                        else
                        {
                            data_item->has_changed = true;
                        }
                        break;
                    }

                    case DISPLAY_MESSAGE_ATTRIBUTE_ICON:
                    {
                        data_item->image_id = (IMAGE_IDENTIFIER)*map.items[x].data;
                        
                        if (data_item_old != NULL)
                        {
                            if (data_item->image_id != data_item_old->image_id)
                            {
                                data_item->has_changed = true;
                            }
                        }
                        else
                        {
                            data_item->has_changed = true;
                        }
                        break;
                    }

                    case DISPLAY_MESSAGE_ATTRIBUTE_SELECTED:
                    {
                        data_item->selected = (bool)*map.items[x].data;
                        
                        if (data_item_old != NULL)
                        {
                            if (data_item->selected != data_item_old->selected)
                            {
                                data_item->has_changed = true;
                            }
                        }
                        else
                        {
                            data_item->has_changed = true;
                        }

                        break;
                    }
                    
                    case DISPLAY_MESSAGE_ATTRIBUTE_OPTION_EDIT_POSITION:
                    {
                        data_item->edit_pos = (uint8_t)*map.items[x].data;
                        
                        if (data_item_old != NULL)
                        {
//                            if (data_item->edit_pos != data_item_old->edit_pos)
//                            {
                                data_item->has_changed = true;
//                            }
                        }
                        else
                        {
                            data_item->has_changed = true;
                        }

                        break;
                    }

                    default:
                        break;
                }
            }
            else
            {
                logger_log_errorf(LOGGER_CODE_DISPLAY, "DISPLAY Unable to allocate DISPLAY_MESSAGE_SCREEN_DATA_ITEM for %d --", map.items[x].data_id);
                logger_process();
            }
        }
    }
}
    
/* *****************************************************************************
 End of File
 */
