/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    encoder_multi.c

  @Summary
 Encapsulates rotary encoder functionality

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/encoder_multi.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

// Now use the input data to define the enum
#define X(a, b, c, d, e, f, g) b,
    static int ENCODER_PORT_A[] = { ENCODER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e, f, g) c,
    static int ENCODER_CHANNEL_A[] = { ENCODER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e, f, g) d,
    static int ENCODER_BIT_POS_A[] = { ENCODER_ITEMS };
#undef X   

// Now use the input data to define the enum
#define X(a, b, c, d, e, f, g) e,
    static int ENCODER_PORT_B[] = { ENCODER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e, f, g) f,
    static int ENCODER_CHANNEL_B[] = { ENCODER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e, f, g) g,
    static int ENCODER_BIT_POS_B[] = { ENCODER_ITEMS };
#undef X   

#define ENCODER_QUEUE_LEN	64

/*
 * Core encoder data used internally
 */
typedef struct ENCODER_DATA
{
    volatile bool in_update;
    volatile uint8_t queue[ENCODER_QUEUE_LEN];
    volatile uint8_t queue_start;
    volatile uint8_t queue_end;

    volatile bool read_busy;

    ENCODER_STATE encoder;
    bool rotary_moved;
    bool wrap;

    uint8_t last_phase;

    uint8_t lastRotaryReading;

    volatile bool a_value;
    volatile bool b_value;
} ENCODER_DATA;
    
static ENCODER_DATA encoder_data[ENCODER_INDEX_MAX];

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Clear the rotary buffer
 */
static void encoder_clear_buffer(ENCODER_INDEX index)
{
	int x;
	ENCODER_DATA* encoder = &encoder_data[index];
    
	while (encoder->in_update)
	{
		Nop();
	}
	encoder->in_update = true;
	
	for (x = 0; x < ENCODER_QUEUE_LEN; x++)
	{
		encoder->queue[x] = 0xFF;
	}
	
	encoder->in_update = false;
	encoder->queue_start = 0;
	encoder->queue_end = 0;
	
	encoder->last_phase = 0;
}

/**-----------------------------------------------------------------------------
 * Review encode data to detect changes
 */
static void encoder_handle_sample(ENCODER_INDEX index)
{
	ENCODER_DATA* encoder = &encoder_data[index];
	static uint8_t pulse_count = 10; // a value big enough to decrement by 4
	uint8_t phase;
	volatile uint8_t pattern;

    // Get the phase
    phase = encoder->queue[encoder->queue_start];
    encoder->queue[encoder->queue_start] = 0xFF; // remove the value from the buffer
	
    // Increment next start and wrap if at end
	if ( ++encoder->queue_start >= ENCODER_QUEUE_LEN )
	{
		encoder->queue_start = 0;
	}

	// examine bit pattern of last and current phases
	pattern = (encoder->last_phase << 2) | phase;
	
	switch (pattern)
	{
        case 2:     // 0010
        case 11:    // 1011
        case 13:    // 1101
        case 4:     // 0100
            // valid sequence for increment
            if ((++pulse_count >= 12) && (pattern == 11)) // 2 states detected and at middle of indent
            {
                pulse_count = 10;
                if (encoder->encoder.current_value < encoder->encoder.max_value)
                {
                    encoder->encoder.current_value++;
                    encoder->rotary_moved = true;
                }
 
                encoder->encoder.delta++;
            }
            break;
            
        case 1:     // 0001
        case 7:     // 0111
        case 14:    // 1110
        case 8:     // 1000
            // valid sequence for decrement
            if ((--pulse_count <= 8) && (pattern == 7)) // 2 states detected and at middle of indent
            {
                pulse_count = 10;
                if (encoder->encoder.current_value > encoder->encoder.min_value)
                {
                    encoder->encoder.current_value--;
                    encoder->rotary_moved = true;
                }

                encoder->encoder.delta--;
            }
            break;
            
        case 3:
        case 15:
            // stable sequence after missing a pulse
            // allow through if we've counted 2 transitions already
            if (pulse_count > 11)
            {
                if (encoder->encoder.current_value < encoder->encoder.max_value)
                {
                    encoder->encoder.current_value++;
                    encoder->rotary_moved = true;
                }

                encoder->encoder.delta++;
            }
            if (pulse_count < 9)
            {
                if (encoder->encoder.current_value > encoder->encoder.min_value)
                {
                    encoder->encoder.current_value--;
                    encoder->rotary_moved = true;
                }

                encoder->encoder.delta--;
            }
            
            pulse_count = 10;
            break;
            
        default:
            break;
	}
    
	encoder->last_phase = phase;
}

/**-----------------------------------------------------------------------------
 * 
 * @param index
 */
static void encoder_read_pins(ENCODER_INDEX index)
{
	ENCODER_DATA* encoder = &encoder_data[index];
    
    encoder->a_value = PLIB_PORTS_PinGet(
            ENCODER_PORT_A[index],
            ENCODER_CHANNEL_A[index],
            ENCODER_BIT_POS_A[index]);

    encoder->b_value = PLIB_PORTS_PinGet(
            ENCODER_PORT_B[index],
            ENCODER_CHANNEL_B[index],
            ENCODER_BIT_POS_B[index]);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Disables the change notice ports for all encoders
 */
void encoder_change_notice_disable()
{
    int x;
    for (x = 0; x < ENCODER_INDEX_MAX; x++)
    {
        // Encoder
        PLIB_PORTS_PinChangeNoticePerPortDisable(ENCODER_PORT_A[x],
                                                 ENCODER_CHANNEL_A[x], 
                                                 ENCODER_BIT_POS_A[x]);

        PLIB_PORTS_PinChangeNoticePerPortDisable(ENCODER_PORT_B[x],
                                                 ENCODER_CHANNEL_B[x], 
                                                 ENCODER_BIT_POS_B[x]);
    }
}

/**-----------------------------------------------------------------------------
 * Enables the change notice ports for all encoders
 */

void encoder_change_notice_enable()
{
    int x;
    for (x = 0; x < ENCODER_INDEX_MAX; x++)
    {
        // Encoder
        PLIB_PORTS_PinChangeNoticePerPortEnable(ENCODER_PORT_A[x],
                                                 ENCODER_CHANNEL_A[x], 
                                                 ENCODER_BIT_POS_A[x]);

        PLIB_PORTS_PinChangeNoticePerPortEnable(ENCODER_PORT_B[x],
                                                 ENCODER_CHANNEL_B[x], 
                                                 ENCODER_BIT_POS_B[x]);
    }
}

/**-----------------------------------------------------------------------------
 * Copies current encoder state into supplied structure
 * Adapted from rotary.c get_rotary()
 * @param state
 * @return - an indication of whether the rotary has changed or not
 */
bool encoder_get_state(ENCODER_STATE* state, ENCODER_INDEX index)
{
	ENCODER_DATA* encoder = &encoder_data[index];
    
	bool rotary_moved = encoder->rotary_moved;
	encoder->rotary_moved = false;
    
    if (encoder->wrap)
    {
        // handle wrap       
        if (encoder->encoder.current_value == encoder->encoder.min_value)
        {
            encoder->encoder.current_value = encoder->encoder.max_value - 1;
        }

        if (encoder->encoder.current_value == encoder->encoder.max_value)
        {
            encoder->encoder.current_value = encoder->encoder.min_value + 1;
        }
        
        // remove offsets 
        state->min_value = encoder->encoder.min_value;
        state->current_value = encoder->encoder.current_value - 1;
        state->max_value = encoder->encoder.max_value - 2;
    }
    else
    {
        state->min_value = encoder->encoder.min_value;
        state->current_value = encoder->encoder.current_value;
        state->max_value = encoder->encoder.max_value;
    }
    
    state->moved = rotary_moved;
    state->delta = encoder->encoder.delta;
    
	return rotary_moved;
}


/**-----------------------------------------------------------------------------
 * Get the current wrap state
 * @return 
 */
bool encoder_get_wrap(ENCODER_INDEX index)
{
    return encoder_data[index].wrap;
}

/**-----------------------------------------------------------------------------
 * Initialise the encoder
 */
void encoder_initialise()
{
    int x;
    
    for (x = 0 ; x < ENCODER_INDEX_MAX; x++)
    {
        ENCODER_DATA* encoder = &encoder_data[x];
        
        encoder_clear_buffer(x);
        	
        encoder->read_busy = false;
        
        encoder_read_pins(x);
        
        encoder->encoder.current_value = 0;
        encoder->encoder.max_value = 100;
        encoder->encoder.min_value = 0;
        encoder->rotary_moved = false;
    }
}
 
/**-----------------------------------------------------------------------------
* Perform encoder processing
 * 
* @return - true if samples processed, false if not
*/
bool encoder_process(ENCODER_INDEX index)
{
    ENCODER_DATA* encoder = &encoder_data[index];
    encoder->encoder.delta = 0;
    
	// is rotary encoder being updated by the interrupt?
	if (!encoder->in_update)
	{
		encoder->in_update = true;
		
		while (encoder->queue_end != encoder->queue_start)
		{
			encoder_handle_sample(index);
		}
		
		encoder->in_update = false;
	}
}


/**-----------------------------------------------------------------------------
 * Adapted from rotary.c get_rotary()
 * @param state - pointer to an array of Encoder states ENVODE_INDEX_MAX long
 * @return - an indication of whether the rotary has changed or not
 */
bool encoder_process_state(ENCODER_STATE* state)
{
    bool result = false;
    int x;
    
    // Timer setting for encoder
    //     X(ENCODER, 5, TIMERS_RESET_IMMEDIATE, TIMERS_HANDLE_100US)
    if (timers_has_expired(TIMER_INDEX_ENCODER))
    {
        for (x = 0; x < ENCODER_INDEX_MAX; x++)
        {
            encoder_read(x);
        }
 
        // Update the encoder states
        for (x = 0; x < ENCODER_INDEX_MAX; x++)
        {
            encoder_process(x);
        }       
        
        // Update the current input context - Get the current encoder values       
        for (x = 0; x < ENCODER_INDEX_MAX; x++)
        {
            encoder_get_state(&state[x], x);
        }
        
        timers_reset(TIMER_INDEX_ENCODER); 
        
        result = true;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Read the current encoder state from the ports
 */
void encoder_read(ENCODER_INDEX index)
{
    ENCODER_DATA* encoder = &encoder_data[index];
	while (encoder->read_busy)
	{
		Nop();
	}
    
	encoder->read_busy = true;
    encoder_read_pins(index);
		
    uint8_t current_reading;
	current_reading = (encoder->a_value) ? 1 : 0;
	current_reading <<= 1;
	
	current_reading |= (encoder->b_value) ? 1 : 0;

	if (current_reading != encoder->lastRotaryReading)
	{
		if (!encoder->in_update)
		{
			// semaphore during processing of encoder
			encoder->in_update = true;
											
			encoder->queue[encoder->queue_end] = current_reading;
			encoder->lastRotaryReading = current_reading;

			if (++encoder->queue_end >= ENCODER_QUEUE_LEN)
			{
				encoder->queue_end = 0; // wrap buffer end to buffer start
			}

			encoder->in_update = false;
		}
	}
	
	encoder->read_busy = false;
}

/**-----------------------------------------------------------------------------
 * Update the encoder state
 * @param state
 * @param wrap
 */
void encoder_set(ENCODER_STATE state, ENCODER_INDEX index, bool wrap)
{
    ENCODER_DATA* encoder = &encoder_data[index];
    encoder->wrap = wrap;
    
    if (encoder->wrap)
    {
        // Add extra value at top and bottom of range.
        // Low limit may be 0, so shift top limit up by 2 and
        // the reading up by 1.
        encoder->encoder.min_value = state.min_value;
        encoder->encoder.current_value = state.current_value + 1;
        encoder->encoder.max_value = state.max_value + 2;
    }
    else
    {
        encoder->encoder = state;
    }
}

/* *****************************************************************************
 End of File
 */
