/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/events.h"
#include "app_config.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
	// An array of EVENTS structure one per module
	EVENTS events[EVENT_MODULES_NUM_MODULES];
	
} EVENTS_DATA;

static EVENTS_DATA events_data = {};


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * This is called by the module when invoking one of its events.
 data is data that is relevant to the event bing raised at this point in time
 */
void events_invoke(EVENT_MODULE_ID module_id, EVENT_TYPE event_type, void* context_data, EVENT_CALLBACK_INFO* callback_info)
{
	if (callback_info->callback != NULL)
	{
		// Set up the event
		EVENT event;
		event.module_id = module_id;
		event.type = event_type;
		event.context_data = context_data;
		event.subscription_data = callback_info->data;

		callback_info->callback(&event);
	}
}

/**-----------------------------------------------------------------------------
 * This is called by the module when invoking one of its events.
 data is data that is relevant to the event bing raised at this point in time
 */
void events_raise(EVENT_MODULE_ID module_id, EVENT_TYPE event_type, void* context_data)
{
	if (module_id < EVENT_MODULES_NUM_MODULES)
	{
		EVENTS* events = &events_data.events[module_id];
        //logger_log_debugf(0, "EVENTS RAISE %s %d --", EVENT_MODULE_NAMES[module_id], event_type);            
        //logger_process();

		// Set up the event
		EVENT event;
		event.module_id = module_id;
		event.type = event_type;
		event.context_data = context_data;

		// Find a free subscription slot
		uint8_t x;
		bool found = false;

		for (x = 0; x < events->num_subscriptions; x++)
		{
			EVENT_SUBSCRIPTION* subscription = &events->subscriptions[x];
			if (subscription->in_use && (subscription->type == event_type))
			{
				if (subscription->callback_info.callback != NULL)
				{
					// Updat the event with the current subscription data
					event.subscription_data = subscription->callback_info.data;
					subscription->callback_info.callback(&event);

					// Free if not persistent
					if (!subscription->persistent)
					{
						subscription->in_use = false;
					}
				}

				found = true;
			}
		}

		if (!found)
		{
			logger_log_warningf(LOGGER_CODE_EVENTS, "EVENTS - No subscriptions for event type specified %d/%d --", module_id, event_type);
		}
	}
	else
	{
		logger_log_warningf(LOGGER_CODE_EVENTS, "EVENTS - Invoke Invalid module specified %d --", module_id);
	}
}

/**-----------------------------------------------------------------------------
 * Register the modules event list in the modules EVENTS structure
 */
void events_register(EVENT_MODULE_ID module_id, EVENT_SUBSCRIPTION* subscription_list, uint8_t num_subscriptions)
{
	if (module_id < EVENT_MODULES_NUM_MODULES)
	{
        logger_log_warningf(LOGGER_CODE_EVENTS, "EVENTS - Registering events for %s %d --", EVENT_MODULE_NAMES[module_id], num_subscriptions);
        
		EVENTS*	events = &events_data.events[module_id];
		if (events != NULL)
		{
			events->module_id = module_id;
			events->subscriptions = subscription_list;
			events->num_subscriptions = num_subscriptions;
            
            uint8_t x;
            for (x = 0; x < num_subscriptions; x++)
            {
                EVENT_SUBSCRIPTION* subscription = &events->subscriptions[x];
                subscription->in_use = false;
            }
		}
	}
	else
	{
		logger_log_warningf(LOGGER_CODE_EVENTS, "EVENTS - Register - Invalid module specified %d --", module_id);
	}
}

/**-----------------------------------------------------------------------------
 * Subscribe to an event
 */
void events_subscribe(EVENT_MODULE_ID module_id, 
        EVENT_TYPE type,
        EVENTS_CALLBACK callback,
        void* data,
        bool persistent)
{
	if (module_id < EVENT_MODULES_NUM_MODULES)
	{
		EVENTS*	events = &events_data.events[module_id];
		if (events != NULL)
		{
			// Find a free subscription slot
			uint8_t x;
			bool found = false;

			for (x = 0; x < events->num_subscriptions; x++)
			{
				EVENT_SUBSCRIPTION* subscription = &events->subscriptions[x];
				if (!subscription->in_use)
				{
                    subscription->type = type;
                    subscription->callback_info.callback = callback,
                    subscription->callback_info.data = data,
                    subscription->persistent = persistent;
					subscription->in_use = true;
					found = true;
					break;
				}
			}

			if (!found)
			{
				logger_log_warningf(LOGGER_CODE_EVENTS, "EVENTS - Insufficient space to subscribe to event %s type %d %d available--", EVENT_MODULE_NAMES[module_id], type, events->num_subscriptions);
			}
		}
	}
	else
	{
		logger_log_warningf(LOGGER_CODE_EVENTS, "EVENTS - Invalid module %d --", module_id);
	}
}

/**-----------------------------------------------------------------------------
 * Unsubscribe from an event -  note it only makes a comparison on event type and callback, so
 * if there ar more than one subscriotion with differing data they will bot be removed
 */
void events_unsubscribe(EVENT_MODULE_ID module_id, EVENT_SUBSCRIPTION* event_subscription)
{
	if (module_id < EVENT_MODULES_NUM_MODULES)
	{
		EVENTS*	events = &events_data.events[module_id];
		if (events != NULL)
		{
			// Find a free subscription slot
			uint8_t x;
			bool found = false;
			for (x = 0; x < events->num_subscriptions; x++)
			{
				EVENT_SUBSCRIPTION* subscription = &events->subscriptions[x];
				if (subscription->in_use)
				{
					if ((subscription->type == event_subscription->type) && (subscription->callback_info.callback == event_subscription->callback_info.callback))
					subscription->in_use = false;
					found = true;
				}
			}

			if (!found)
			{
				logger_log_warningf(LOGGER_CODE_EVENTS, "EVENTS - Insufficient space to unsubscribe to event %d/%d --", module_id, event_subscription->type);
			}
		}
	}
	else
	{
		logger_log_warningf(LOGGER_CODE_EVENTS, "EVENTS - Invalid module %d --", module_id);
	}
}

/* *****************************************************************************
 End of File
 */
