/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/flash_memory.h"
#include <stdio.h>
#include "system/fs/sys_fs.h"
#include "app_config.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#ifndef FLASH_MEMORY_DEVICE_NAME
#error "FLASH_MEMORY_DEVICE_NAME MUST be defined in app_config.h"
#endif

#ifndef FLASH_MEMORY_MOUNT_NAME
#error "FLASH_MEMORY_MOUNT_NAME MUST be defined in app_config.h"
#endif

//! The number of files that can be owned by a single client >= 1
#ifndef FLASH_MEMORY_MAX_CLIENT_FILES
#define FLASH_MEMORY_MAX_CLIENT_FILES       5
#endif

#define FLASH_MEMORY_FS_EXISTS_DAT_FILE     "FS_EXISTS.DAT"
#define FLASH_MEMORY_MAX_CLENT_NAME_LEN     20

/**
 * Hold the state of the settings
 */
typedef enum 
{
    FLASH_MEMORY_STATE_MOUNTING,
    FLASH_MEMORY_STATE_CHECK_FS,
    FLASH_MEMORY_STATE_FORMAT,
    FLASH_MEMORY_STATE_INITIALISE_FS,
    FLASH_MEMORY_STATE_READY,

    //  We have either tried to read or write from the settings byte there has 
    // been errors
    FLASH_MEMORY_STATE_ERROR,
} FLASH_MEMORY_STATE;
     
//! Struct containing information about a client
typedef struct
{
    FLASH_MEMORY_CLIENT_ID client_id;
    
    //! A name to be associated with the client
    char name[FLASH_MEMORY_MAX_CLENT_NAME_LEN];

    //! The callback invoked when a page request is completed
    FLASH_MEMORY_CALLBACK callback;
        
    //! Active File Handle (opened by flash_memory_start????)
    SYS_FS_HANDLE file_handle[FLASH_MEMORY_MAX_CLIENT_FILES];

} FLASH_MEMORY_CLIENT;

/**
 */
typedef struct
{
    FLASH_MEMORY_CLIENT_REQUEST client_request;

    //! Indicates whether the page request can be re-used or is still waiting to be serviced.
    bool active;

} FLASH_MEMORY_REQUEST;
    
/**
 */
typedef struct
{
    //! Holds info on the clients using flash memory
    FLASH_MEMORY_CLIENT clients[FLASH_MEMORY_MAX_CLIENTS];
    
    //! Indicates whether the module has been initialised yet as it may be called a number of times
    bool initialised;
    
    //! The where the next request should be added
    uint8_t add_request_idx;
    
     //! The index of the  page currently being processed
    uint8_t current_request_idx;
    
    uint8_t mount_attempts;
     
    //! A queue of read requests
    FLASH_MEMORY_REQUEST requests[FLASH_MEMORY_REQUEST_QUEUE_SIZE];

    FLASH_MEMORY_STATE state;
} FLASH_MEMORY_DATA;

static FLASH_MEMORY_DATA flash_memory_data =
{
    .state = FLASH_MEMORY_STATE_MOUNTING,
    .initialised = false,
    .mount_attempts = 0
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * 
 * @param client_id
 * @param offset        The offset in the client data to start at
 * @param buffer        The buffer to write the data to
 * @param length        The length of the data returned
 */
static void flash_memory_add_request(FLASH_MEMORY_CLIENT_REQUEST* client_request, bool write)
{
    FLASH_MEMORY_CLIENT* client = &flash_memory_data.clients[client_request->client_id]; 
    
    // Decide whether this can be done in a single request or multiple requests
    // Assume single request to start
    FLASH_MEMORY_REQUEST* request = &flash_memory_data.requests[flash_memory_data.add_request_idx];
    
    // Create requests
    request = &flash_memory_data.requests[flash_memory_data.add_request_idx];
    
    // Copy client request
    memcpy(&request->client_request, client_request, sizeof(FLASH_MEMORY_CLIENT_REQUEST));
    request->client_request.bytes_read_written = 0;
    request->client_request.write_request = write;
    request->active = true;             // Make the request active

//    logger_log_debugf(0, "FLASH MEMORY: new request client:%d file:%d r/w:%s offset:%d len:%d --", 
//            request->client_request.client_id, 
//            request->client_request.file_index, 
//            request->client_request.write_request ? "WRITE" : "READ", 
//            request->client_request.offset, 
//            request->client_request.length);
    
    // Move onto the next request
    flash_memory_data.add_request_idx++;
    if (flash_memory_data.add_request_idx >= FLASH_MEMORY_REQUEST_QUEUE_SIZE)
    {
        flash_memory_data.add_request_idx = 0;
    }

    if (flash_memory_data.add_request_idx == flash_memory_data.current_request_idx)
    {
        logger_log_errorf(LOGGER_CODE_FLASH_REQUEST_OVERRUN, "FLASH MEMORY Number of available requests exceeded %d ", FLASH_MEMORY_REQUEST_QUEUE_SIZE);
    }
 }

/**-----------------------------------------------------------------------------
 * Attempt to open the FS_EXISTS.DAT file, and trigger a format if it doesn't
 */
static bool flash_memory_check_fs()
{
    bool result = false;

    logger_log_debugf(0, "FLASH MEMORY Checking File System ");
    
    char file_name[20];
    sprintf(file_name, FLASH_MEMORY_FS_EXISTS_DAT_FILE);
    
    logger_log_debugf(0, "FLASH MEMORY Opening %s --", file_name);
    
    SYS_FS_HANDLE file_handle = SYS_FS_FileOpen(file_name, SYS_FS_FILE_OPEN_READ);

    if(file_handle != SYS_FS_HANDLE_INVALID)
    {
        logger_log_debugf(0, "FLASH MEMORY Opened %s --", file_name);
        result = true;
        
        // Read the file and get version info out of it..
        uint8_t version_info[2];
        
        if (SYS_FS_FileRead(file_handle, (void *)version_info, 2) == 2)
        {
            logger_log_infof(0, "FLASH MEMORY Open %s success - found version %d.%d --", file_name, version_info[0], version_info[1]);
            result = true;
        }
        else
        {
            SYS_FS_ERROR error = SYS_FS_Error();

            // File open failed.
            logger_log_errorf(0, "FLASH MEMORY Open %s failed - error %d ", file_name, error);
        }
        
        // Close the file
        SYS_FS_FileClose(file_handle);
        
        uint32_t total_sectors;
        uint32_t free_sectors;
        if (SYS_FS_DriveSectorGet(FLASH_MEMORY_MOUNT_NAME, &total_sectors, &free_sectors) ==  SYS_FS_RES_SUCCESS)
        {
            //Sector information get operation failed.
            logger_log_infof(0, "FLASH MEMORY Total sectors %d  Free sectors %d ", total_sectors, free_sectors);
        }
    }
    else
    {
        SYS_FS_ERROR error = SYS_FS_Error();
        
        // File open failed.
        logger_log_errorf(0, "FLASH MEMORY Open failed - error %d ", error);
    }

    return result;
}
    
/**-----------------------------------------------------------------------------
 * Attempt to format the file system, and create the FS_EXISTS.DAT file
 */
static bool flash_memory_format_fs()
{
    bool result = false;

    logger_log_infof(0, "FLASH MEMORY Preparing File System ");
    logger_process();

    if (SYS_FS_DriveFormat(FLASH_MEMORY_MOUNT_NAME, SYS_FS_FORMAT_SFD, 0) == SYS_FS_RES_SUCCESS)
    {
        logger_log_infof(0, "FLASH MEMORY Format successful ");
        logger_process();
        result = true;
    }
    else
    {
        logger_log_errorf(0, "FLASH MEMORY Preparing File System FAILED %d ", SYS_FS_Error());
        logger_process();
    }

    return result;
}    

/**-----------------------------------------------------------------------------
 * Create the FS_EXISTS.DAT file
 */
static bool flash_memory_initialise_fs()
{
    bool result = false;

    logger_log_infof(0, "FLASH MEMORY Initialising File System ");
    
    char file_name[20];
    sprintf(file_name, FLASH_MEMORY_FS_EXISTS_DAT_FILE);
    
    logger_log_debugf(0, "FLASH MEMORY Opening %s --", file_name);
    
    SYS_FS_HANDLE file_handle = SYS_FS_FileOpen(file_name, SYS_FS_FILE_OPEN_WRITE);
    
    if(file_handle != SYS_FS_HANDLE_INVALID)
    {
        // Create the FS_EXISTS.DAT file and write version info to it
        uint8_t version_info[2];
        version_info[0] = SOFTWARE_VERSION_MAJOR;
        version_info[1] = SOFTWARE_VERSION_MINOR;
        
        if(SYS_FS_FileWrite(file_handle, (void *)version_info, 2) != -1)
        {
            if (SYS_FS_FileSync(file_handle) == SYS_FS_RES_SUCCESS)
            {
                logger_log_infof(0, "FLASH MEMORY Initialised File System ");
                result = true;
            }
            else
            {
                /* Check the file status */
                logger_log_errorf(0, "FLASH MEMORY Initialise File System FAILED sync %d ", SYS_FS_Error());
            }
        }
        else
        {
            // Error
            logger_log_errorf(0, "FLASH MEMORY Initialise File System FAILED write %d ", SYS_FS_Error());
        }
    }
    else
    {
        logger_log_errorf(0, "FLASH MEMORY Initialising File System FAILED %d ", SYS_FS_Error());
    }

    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * A Dummy callback used for testing flash_memory functionality
 * @param page_data
 */
void flash_memory_dummy_callback(FLASH_MEMORY_CLIENT_REQUEST* request)
{
    logger_log_debugf(0, "FLASH MEMORY Dummy Callback client:%d ", 0);
}

/**-----------------------------------------------------------------------------
 * Each module using flash memory adds itself as a client
 * @return 
 */
void flash_memory_add_client(FLASH_MEMORY_CLIENT_ID client_id, char* name, FLASH_MEMORY_CALLBACK callback)
{
    // The Id is actually an ordinal position in the clients array
    if (client_id < FLASH_MEMORY_MAX_CLIENTS)
    {
        // Check its not already in use
        if (flash_memory_data.clients[client_id].client_id == FLASH_MEMORY_INVALID_CLIENT_ID)
        {
            flash_memory_data.clients[client_id].client_id = client_id;
            memcpy(flash_memory_data.clients[client_id].name, name, min(FLASH_MEMORY_MAX_CLENT_NAME_LEN, strlen(name)));
            flash_memory_data.clients[client_id].callback = callback;
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_FLASH_INITIALISATION, "FLASH MEMORY Invalid client id %d %d %s %d (already in use) ", client_id, flash_memory_data.clients[client_id].client_id, flash_memory_data.clients[client_id].name, FLASH_MEMORY_INVALID_CLIENT_ID);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_FLASH_INITIALISATION, "FLASH MEMORY Invalid client id %d (out of bounds) ", client_id);
    }
}
    
/**-----------------------------------------------------------------------------
 * Where multiple reads/writes are being performed flash_memory_close releases
 * any resources previously allocated using flash_memory_open
 */
void flash_memory_close(FLASH_MEMORY_CLIENT_ID client_id, uint8_t file_index)
{
    // Close the file, set the file handle to -1.
    //logger_log_debugf(0, "FLASH MEMORY: Closing file %d %d --", client_id, file_index);
    if (flash_memory_data.clients[client_id].file_handle[file_index] != SYS_FS_HANDLE_INVALID)
    {
        SYS_FS_FileClose(flash_memory_data.clients[client_id].file_handle[file_index]);
        flash_memory_data.clients[client_id].file_handle[file_index] = SYS_FS_HANDLE_INVALID;
    }
}
    
/** ----------------------------------------------------------------------------
 * Completely clears flash memory
 */
void flash_memory_erase()
{
    logger_log_debugf(0, "FLASH MEMORY Erasing flash");
    logger_process();
    flash_memory_format_fs();
}
    
/**-----------------------------------------------------------------------------
 * Initialise the flash memory system
 */
void flash_memory_initialise()
{
    if (!flash_memory_data.initialised)
    {
        logger_log_debugf(0, "FLASH MEMORY initialising");
        
        //! Initialise clients
        int x;
        for (x = 0; x < FLASH_MEMORY_MAX_CLIENTS; x++)
        {
            flash_memory_data.clients[x].client_id = FLASH_MEMORY_INVALID_CLIENT_ID;
            memset(flash_memory_data.clients[x].name, 0, FLASH_MEMORY_MAX_CLENT_NAME_LEN);
            
            int y;
            for (y = 0; y < FLASH_MEMORY_MAX_CLIENT_FILES; y++)
            {
                flash_memory_data.clients[x].file_handle[y] = SYS_FS_HANDLE_INVALID;
            }
        }

        // Initialise page requests
        for (x = 0; x < FLASH_MEMORY_REQUEST_QUEUE_SIZE; x++)
        {
            flash_memory_data.requests[x].active = false;
        }

        flash_memory_data.current_request_idx = 0;
        flash_memory_data.add_request_idx = 0;
        
        // Set initialised to true so this doesn't get run again
        flash_memory_data.initialised = true;
    }
}
    
/**-----------------------------------------------------------------------------
 * Where multiple reads/writes are being performed flash_memory_open requests
 * that the flash resources are retained between calls (rather than open/close 
 * on each call to read/write.
 */
void flash_memory_open(FLASH_MEMORY_CLIENT_ID client_id, uint8_t file_index)
{
    FLASH_MEMORY_CLIENT* client = &flash_memory_data.clients[client_id]; 

    char file_name[FLASH_MEMORY_MAX_CLENT_NAME_LEN + 1];
    if (file_index == 0)
    {
        sprintf(file_name, "%s.DAT", client->name);
    }
    else
    {
        sprintf(file_name, "%s_%d.DAT", client->name, file_index);
    }

    logger_log_debugf(0, "FLASH MEMORY Opening (%d/%d) %s --", client_id, file_index, file_name);
    logger_process();

    if (client->file_handle[file_index] == SYS_FS_HANDLE_INVALID)
    {
        client->file_handle[file_index] = SYS_FS_FileOpen(file_name, SYS_FS_FILE_OPEN_WRITE_PLUS);

        if(client->file_handle[file_index] == SYS_FS_HANDLE_INVALID)
        {
            // File open failed.
            logger_log_errorf(0, "FLASH MEMORY Open failed for %s --", file_name);
            logger_process();
            logger_log_errorf(0, "FLASH MEMORY Open failed for %s - error %d --", file_name, SYS_FS_Error());
            logger_process();
        }
    }
    else
    {
        logger_log_debugf(0, "FLASH MEMORY %s already open --", file_name);
    }
}

/**-----------------------------------------------------------------------------
 * flash memory processing
 */
void flash_memory_process()
{
    //logger_log_debugf(0, "FLASH MEMORY Process %d ", flash_memory_data.state);    

    // Check is the file system initialised?
    switch (flash_memory_data.state)
    {
        case FLASH_MEMORY_STATE_MOUNTING:
        {
            // Check whether it is mounted or not
            if(SYS_FS_Mount(FLASH_MEMORY_DEVICE_NAME, FLASH_MEMORY_MOUNT_NAME, FAT, 0, NULL) == SYS_FS_RES_SUCCESS)
            {
                logger_log_debugf(0, "FLASH MEMORY Mounted %s as %s ", FLASH_MEMORY_DEVICE_NAME, FLASH_MEMORY_MOUNT_NAME);

                // Default to the root drive
                SYS_FS_CurrentDriveSet(FLASH_MEMORY_MOUNT_NAME);
                
                logger_log_debugf(0, "FLASH MEMORY Current drive set to %s ", FLASH_MEMORY_MOUNT_NAME);

                flash_memory_data.mount_attempts = 0;
                flash_memory_data.state = FLASH_MEMORY_STATE_CHECK_FS;
                
//                logger_log_debugf(0, "FLASH MEMORY FORMATTING -------------------- ");
//                flash_memory_format_fs();
//                logger_log_debugf(0, "FLASH MEMORY FORMATTING END ----------------- ");
//                flash_memory_data.state = 34;
            }
            else
            {
                flash_memory_data.mount_attempts++;
                
                SYS_FS_ERROR error = SYS_FS_Error();
                
                if ((error != SYS_FS_ERROR_NOT_READY) || (flash_memory_data.mount_attempts > 10))
                {
                    // Note when setting up the SA30 (again!!) this error was caused
                    // because I had missed setting the Clock Mode (DRV_SPI_CLOCK_MODE_IDLE_LOW_EDGE_FALL)
                    // and Input Phase (SPI_INPUT_SAMPLING_PHASE_AT_END) in the SPI channel config)
                    // Note - also need File System -> Maximum Simultaneous File Access = 4
                    logger_log_debugf(0, "FLASH MEMORY Mount failed - error %d --", error);
                }
            }

            break;
        }

        case FLASH_MEMORY_STATE_CHECK_FS:
        {
            // Try opening a dummy file
            // If it doesn't exist then format the drive
            // Create the file and write version info to it
            if (flash_memory_check_fs())
            {
                flash_memory_data.state = FLASH_MEMORY_STATE_READY;
            }
            else
            {
                flash_memory_data.state = FLASH_MEMORY_STATE_FORMAT;
            }
            
            // Temporarily enable this if we want to force a format of the drive
            //flash_memory_data.state = FLASH_MEMORY_STATE_FORMAT;
            
            break;
        }

        case FLASH_MEMORY_STATE_FORMAT:
        {
            if (flash_memory_format_fs())
            {
                flash_memory_data.state = FLASH_MEMORY_STATE_INITIALISE_FS;
            }
            
            break;
        }

        case FLASH_MEMORY_STATE_INITIALISE_FS:
        {
            if (flash_memory_initialise_fs())
            {
                logger_log_debugf(0, "FLASH MEMORY Ready -------------------- ");
                flash_memory_data.state = FLASH_MEMORY_STATE_READY;
            }
            
            break;
        }
        
        case FLASH_MEMORY_STATE_READY:
        {
            // Are there any requests to process (only do one per cycle)
            FLASH_MEMORY_REQUEST* request = &flash_memory_data.requests[flash_memory_data.current_request_idx];
            if (request->active)
            {
                logger_log_debugf(0, "FLASH MEMORY Processing %s request id %d / file %d at pos %d offset %d len %d xx --", 
                        request->client_request.write_request ? "WRITE" : "READ",
                        request->client_request.client_id, 
                        request->client_request.file_index, 
                        flash_memory_data.current_request_idx,
                        request->client_request.offset, 
                        request->client_request.length);

                // Check whether we have a valid file handle allocated by flash_memory_open
                FLASH_MEMORY_CLIENT* client = &flash_memory_data.clients[request->client_request.client_id]; 
                        
                // Is the file open?
                bool needs_close = false;
                if (client->file_handle[request->client_request.file_index] == SYS_FS_HANDLE_INVALID)
                {
                    // If not open it
                    flash_memory_open(client->client_id, request->client_request.file_index);
                    needs_close = true;
                }
                
                // Check we've managed to open the file
                if (client->file_handle[request->client_request.file_index] != SYS_FS_HANDLE_INVALID)
                {
                    // Move to the position specified by the client offset
                    int32_t res = SYS_FS_FileSeek(client->file_handle[request->client_request.file_index], request->client_request.offset, SYS_FS_SEEK_SET);
                    logger_log_debugf(0, "FLASH MEMORY Seek at %d moved %d --", request->client_request.offset, res);
                    logger_process();
                    
                    //SYS_INT_PROCESSOR_STATUS status = SYS_INT_StatusGetAndDisable();
                    
                    // Read or Write operation.
                    if (!request->client_request.write_request)
                    {
                        // Try reading the file
                        request->client_request.bytes_read_written = SYS_FS_FileRead(client->file_handle[request->client_request.file_index], request->client_request.data, request->client_request.length);
                    }
                    else
                    {
                        request->client_request.bytes_read_written = SYS_FS_FileWrite(client->file_handle[request->client_request.file_index], request->client_request.data, request->client_request.length);
                    }
                    
                    //SYS_INT_StatusRestore(status);

                    logger_log_debugf(0, "FLASH MEMORY Read/Write %d bytes --", request->client_request.bytes_read_written);
                    
                    // Close the file if necessary
                    if (needs_close)
                    {
                        flash_memory_close(client->client_id, request->client_request.file_index);
                    }

                    // Test for success and invoke the callback
                    if (request->client_request.bytes_read_written >= 0)
                    {
                        client->callback(&request->client_request);
                    }
                    else
                    {
                        logger_log_errorf(0, "FLASH MEMORY Read/Write failed - error %d --", SYS_FS_Error());
                    }

                    // Release the request
                    request->active = false;

                    //! Find next free page request 
                    flash_memory_data.current_request_idx++;
                    if (flash_memory_data.current_request_idx >= FLASH_MEMORY_REQUEST_QUEUE_SIZE)
                    {
                        flash_memory_data.current_request_idx = 0;
                    }
                    
                    //logger_log_debugf(123, "FLASH MEMORY Current Request %d --", flash_memory_data.current_request_idx);
                }
                else
                {
                    logger_log_debugf(123, "FLASH MEMORY Invalid Handle %d --", flash_memory_data.current_request_idx);
                    logger_process();                }
            }
            
            break;
        }
        
        default:
            break;
    }    
}

/**-----------------------------------------------------------------------------
 * 
 * @param client_id
 * @param offset        The offset in the client data to start at
 * @param buffer        The buffer to write the data to
 * @param length        The length of the data returned
 */
void flash_memory_read(FLASH_MEMORY_CLIENT_REQUEST* client_request)
{
    flash_memory_add_request(client_request, false);
}
    
/**-----------------------------------------------------------------------------
 * Can be used to see if a request is read or write
 * @param client_request
 * @return 
 */
bool flash_memory_request_is_read(FLASH_MEMORY_CLIENT_REQUEST* client_request)
{
    return !client_request->write_request;
}

/**-----------------------------------------------------------------------------
 * Can be used to see if a request is read or write
 * @param client_request
 * @return 
 */
bool flash_memory_request_is_write(FLASH_MEMORY_CLIENT_REQUEST* client_request)
{
    return client_request->write_request;
}
  
/**----------------------------------------------------------------------------- 
 * Sets the size of the specified file.  If the size supplied is too large 
 * it leaves the size alone
 */
uint32_t flash_memory_set_size(FLASH_MEMORY_CLIENT_ID client_id, uint8_t file_index, uint32_t size)
{
    FLASH_MEMORY_CLIENT* client = &flash_memory_data.clients[client_id]; 

    char file_name[FLASH_MEMORY_MAX_CLENT_NAME_LEN + 1];
    if (file_index == 0)
    {
        sprintf(file_name, "%s.DAT", client->name);
    }
    else
    {
        sprintf(file_name, "%s_%d.DAT", client->name, file_index);
    }

    bool needs_close = false;
    if (client->file_handle[file_index] == SYS_FS_HANDLE_INVALID)
    {
        flash_memory_open(client_id, file_index);
        needs_close = true;
    }
    
    int32_t res = SYS_FS_FileSeek(client->file_handle[file_index], size, SYS_FS_SEEK_SET);
    if (res > -1)
    {
        SYS_FS_FileTruncate(client->file_handle[file_index]);
        logger_log_debugf(0, "FLASH MEMORY truncated %s to %d bytes --", file_name, flash_memory_size(client_id, file_index));
    }
    else
    {
        logger_log_debugf(0, "FLASH MEMORY %s set size error %d --", file_name, SYS_FS_Error());
    }
//    SYS_FS_RESULT SYS_FS_FileTruncate(
//    SYS_FS_HANDLE handle
    
    if (needs_close)
    {
        flash_memory_close(client_id, file_index);
    }

}
  
/**----------------------------------------------------------------------------- 
 * Gets the size of the specified file
 */
uint32_t flash_memory_size(FLASH_MEMORY_CLIENT_ID client_id, uint8_t file_index)
{
    uint32_t result = 0;
    SYS_FS_FSTAT file_stat;
    FLASH_MEMORY_CLIENT* client = &flash_memory_data.clients[client_id]; 

    char file_name[FLASH_MEMORY_MAX_CLENT_NAME_LEN + 1];
    if (file_index == 0)
    {
        sprintf(file_name, "%s.DAT", client->name);
    }
    else
    {
        sprintf(file_name, "%s_%d.DAT", client->name, file_index);
    }
    
    bool needs_close = false;
    if (client->file_handle[file_index] == SYS_FS_HANDLE_INVALID)
    {
        flash_memory_open(client_id, file_index);
        needs_close = true;
    }
    
    result = SYS_FS_FileSize(client->file_handle[file_index]);

//    if (needs_close)
//    {
//        flash_memory_close(client_id, file_index);
//    }
    
    if (result > -1)
    {
        logger_log_debugf(0, "FLASH MEMORY %s %d bytes --", file_name, result);
    }
    else
    {
        logger_log_debugf(0, "FLASH MEMORY %s size file error %d --", file_name, SYS_FS_FileError(client->file_handle[file_index]));
        logger_log_debugf(0, "FLASH MEMORY %s size error %d --", file_name, SYS_FS_Error());
    }

    if (needs_close)
    {
        flash_memory_close(client_id, file_index);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param client_id
 * @param offset        The offset in the client data to start at
 * @param buffer        The buffer to write the data to
 * @param length        The length of the data returned
 */
void flash_memory_write(FLASH_MEMORY_CLIENT_REQUEST* client_request)
{
    flash_memory_add_request(client_request, true);
 }

/* *****************************************************************************
 End of File
 */
