/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "flash_memory.h"
#include "framework/driver/spi/static/drv_spi_static.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */
     
//! Struct containing information about a client
typedef struct
{
    FLASH_MEMORY_CLIENT_ID client_id;

    // The start page in memory that the client data is stored (starting from 0)
    uint32_t page;

    //! The offset of the data on the first page
    uint32_t offset;

    //! The size of the data
    size_t size;

    //! The callback invoked when a page request is completed
    FLASH_MEMORY_CALLBACK callback;
} FLASH_MEMORY_CLIENT;

/**
 */
typedef struct
{
    FLASH_MEMORY_CLIENT_REQUEST client_request;

    //! The page being requested
    uint16_t page_num;

    //! The offset of the data from the requesting client
    uint32_t client_offset;

    //! A pointer to the buffer to receive the data
    char* data;

    //! The offset of the data in the page
    uint32_t offset;

    //! The length of data to be read from this page
    uint32_t length;

    //! Indicates whether the page request can be re-used or is still waiting to be serviced.
    bool active;

    //! false read request, true - write request
    bool write_request;

    //! Set to true if this request is the last in a batch of requests
    bool last_request;

} FLASH_MEMORY_REQUEST;
    
/**
 */
typedef struct
{
    //! Holds info on the clients using flash memory
    FLASH_MEMORY_CLIENT clients[FLASH_MEMORY_MAX_CLIENTS];
    
    //! Indicates whether the module has been initialised yet as it may be called a number of times
    bool initialised;
    
    //! The where the next request should be added
    uint8_t add_request_idx;
    
     //! The index of the  page currently being processed
    uint8_t current_page_request_idx;
    
    // Handle of spi bus
    DRV_HANDLE spi_handle;
    
    //! A queue of read requests
    FLASH_MEMORY_PAGE_REQUEST requests[FLASH_MEMORY_REQUEST_QUEUE_SIZE];

} FLASH_MEMORY_DATA;

static FLASH_MEMORY_DATA flash_memory_data =
{
    .initialised = false
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */
      
/**
 * Disable Chip Select pin
 */
static void flash_memory_disable_transfer()
{
    PLIB_PORTS_PinSet(PORTS_ID_0, 
                      PORT_CHANNEL_G, 
                      PORTS_BIT_POS_6);
}

/**
 * Disable CS pin
 */
static void flash_memory_enable_transfer()
{
    PLIB_PORTS_PinClear(PORTS_ID_0, 
                        PORT_CHANNEL_G, 
                        PORTS_BIT_POS_6);
}
 
/**
 * Calculate the next free page/offset based on the assumption that the memory is 
 * always contiguous.  That means that any space required for future proofing must be 
 * allocated before any memory is written
 * @param client_id
 * @param size
 */
static bool flash_memory_next_free_memory(FLASH_MEMORY_CLIENT_ID client_id, uint32_t* page_num, uint32_t* offset, size_t space_required)
{
    bool result = true;
    uint32_t bytes_used = 0;
    
    // Find how much space has already been allocated
    int x;
    for (x = 0; x < FLASH_MEMORY_MAX_CLIENTS; x++)
    {
        if (flash_memory_data.clients[x].client_id != FLASH_MEMORY_INVALID_CLIENT_ID)
        {
            bytes_used += flash_memory_data.clients[x].size;
        }
    }
    
    *page_num = bytes_used / flash_memory_get_page_size();
    *offset = bytes_used % flash_memory_get_page_size();
    
    // It might be good here to check whether the request would put the data over 
    // a page boundary when it might not need to be...
    
    // Check we have enough space
    uint32_t available_space = (flash_memory_get_page_size() * flash_memory_get_page_count()) - bytes_used;
    if (available_space < space_required)
    {
        *page_num = flash_memory_get_page_count();
        *offset = flash_memory_get_page_size();
        logger_log_errorf(LOGGER_CODE_FLASH_INITIALISATION, "FLASH_MEMORY Unable to allocate space for client %d reqd:%d avail:%d ", client_id, space_required, available_space);
        result = false;
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * A Dummy callback used for testing flash_memory functionailty
 * @param page_data
 */
void flash_memory_dummy_callback(FLASH_MEMORY_CLIENT_REQUEST* request)
{
    logger_log_debugf(0, "FLASH MEMORY Dummy Callback client:%d ", 0);
}

/**-----------------------------------------------------------------------------
 * 
 * @param event
 * @param bufferHandle
 * @param context
 */
static void flash_memory_callback( DRV_SPI_BUFFER_EVENT event,
                                   DRV_SPI_BUFFER_HANDLE bufferHandle,
                                   uintptr_t context )
{
    FLASH_MEMORY_REQUEST* request = (FLASH_MEMORY_REQUEST*)context;
    
    switch(event)
    {
        case DRV_SPI_BUFFER_EVENT_COMPLETE:
        {
            // Handle the completed buffer.
            // Find the client
            FLASH_MEMORY_CLIENT* client = &flash_memory_data.clients[request->client_request.client_id];
            
            if (client->callback != NULL)
            {
                // Invoke the callback - the data points to the buffer that was passed in but
                // might not yet be completely populated.  The request->last_request
                // should be interrogated to check whether it is complete.
                /*
                 * if (request->last_request)
                 * {
                 *      // Buffer is now fully populated
                 *      // so use the data.
                 * }
                 */
                client->callback(request->client_request);
            }

            // Release the request
            request->active = false;

            flash_memory_data.current_page_request_idx++;
            if (flash_memory_data.current_page_request_idx >= FLASH_MEMORY_REQUEST_QUEUE_SIZE)
            {
                flash_memory_data.current_page_request_idx = 0;
            }
            break;
        }
        
        case DRV_SPI_BUFFER_EVENT_ERROR:
        default:
        {
            // Handle error.
            break;
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Each module using flash memory adds itself as a client
 * @return 
 */
void flash_memory_add_client(FLASH_MEMORY_CLIENT_ID client_id, size_t size, FLASH_MEMORY_CALLBACK callback)
{
    // The Id is actually an ordinal position in the clients array
    if (client_id < FLASH_MEMORY_MAX_CLIENTS)
    {
        // Check its not already in use
        if (flash_memory_data.clients[client_id].client_id == FLASH_MEMORY_INVALID_CLIENT_ID)
        {
            // Calculate the page start number and offset
            if (flash_memory_next_free_memory(client_id,
                    &flash_memory_data.clients[client_id].page, 
                    &flash_memory_data.clients[client_id].offset, 
                    size))
            {
                flash_memory_data.clients[client_id].client_id = client_id;
                flash_memory_data.clients[client_id].size = size;
                flash_memory_data.clients[client_id].callback = callback;
            }
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_FLASH_INITIALISATION, "FLASH_MEMORY Invalid client id %d (already in use) ", client_id);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_FLASH_INITIALISATION, "FLASH_MEMORY Invalid client id %d (out of bounds) ", client_id);
    }
}

/**-----------------------------------------------------------------------------
 * Initialise the flash memory system
 */
void flash_memory_initialise()
{
    if (!flash_memory_data.initialised)
    {
        logger_log_debugf(0, "FLASH MEMORY: initialising");
        
        //! Initialise clients
        int x;
        for (x = 0; x < FLASH_MEMORY_MAX_CLIENTS; x++)
        {
            flash_memory_data.clients[x].client_id = FLASH_MEMORY_INVALID_CLIENT_ID;
            flash_memory_data.clients[x].size = 0;
        }

        // Initialise page requests
        for (x = 0; x < FLASH_MEMORY_REQUEST_QUEUE_SIZE; x++)
        {
            flash_memory_data.requests[x].active = false;
        }

        flash_memory_data.current_page_request_idx = 0;
        flash_memory_data.add_request_idx = 0;

        // Open SPI
//        flash_memory_data.spi_handle = DRV_SPI_Open( DRV_SPI_INDEX_0, DRV_IO_INTENT_READWRITE );
//        if( DRV_HANDLE_INVALID == flash_memory_data.spi_handle )
//        {
//            // Unable to open the driver
//            logger_log_errorf(0, "FLASH MEMORY: Unable to Open SPI BUS");
//        }
//        else
//        {
//            logger_log_errorf(0, "FLASH MEMORY: SPI Open");
//        }
        
        // Set initialised to true so this doesn't get run again
        flash_memory_data.initialised = true;
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param client_id
 * @param offset        The offset in the client data to start at
 * @param buffer        The buffer to write the data to
 * @param length        The length of the data returned
 */
void flash_memory_read(FLASH_MEMORY_CLIENT_ID client_id, uint32_t offset, char* buffer, uint32_t length)
{
    // Generate requests to read the pages that are required.
    uint32_t page;
    uint32_t page_offset;
    
    FLASH_MEMORY_CLIENT* client = &flash_memory_data.clients[client_id]; 
    
    // convert the request into a starting point relative to the start of the flash memory
    uint32_t byte_start = (client->page * flash_memory_get_page_size()) + 
            client->offset +
            offset;
    
    uint32_t byte_end = byte_start + length;
    uint32_t bytes_allocated = 0;
    uint32_t byte_length = 0;
    
    page = byte_start / flash_memory_get_page_size();
    page_offset = byte_start % flash_memory_get_page_size();
    
    byte_length = min(flash_memory_get_page_size() - page_offset, length);
    
    FLASH_MEMORY_PAGE_REQUEST* request = NULL;
    
    while(byte_start < byte_end)
    {
        // Create requests
        request = &flash_memory_data.requests[flash_memory_data.add_request_idx];
        request->client_request.client_id = client_id;
        request->client_request.offset = offset;
        request->client_request.length = length;

        request->data = buffer;
        request->client_offset = offset + bytes_allocated;    // The offset of this request in the client data
        request->page_num = page;
        request->offset = page_offset;      // The offset of the request data in the current page (will be 0 for all but the first page))
        request->length = byte_length;      // The length of the data covered by this request
        request->write_request = false;
        request->active = true;             // Make the request active
        request->last_request = false;
        
        page++;                             // Increment the page
        page_offset = 0;                    // Subsequent offsets will always be 0
        
        byte_start += byte_length;
        bytes_allocated += byte_length;
        byte_length = min(byte_end - byte_start, flash_memory_get_page_size());
        
        // Move onto the next request
        flash_memory_data.add_request_idx++;
        if (flash_memory_data.add_request_idx >= FLASH_MEMORY_REQUEST_QUEUE_SIZE)
        {
            flash_memory_data.add_request_idx = 0;
        }
        
        if (flash_memory_data.add_request_idx == flash_memory_data.current_page_request_idx)
        {
            logger_log_errorf(LOGGER_CODE_FLASH_REQUEST_OVERRUN, "FLASH MEMORY Number of available requests exceeded %d ", FLASH_MEMORY_REQUEST_QUEUE_SIZE);
            break;
        }
    }
    
    // Set the last request marker
    if (request != NULL)
    {
        request->last_request = true;
    }
 }

/**-----------------------------------------------------------------------------
*/
//---------------------------------------------------------------------------------

void flash_memory_jedec_read()
{
    logger_log_debugf(0, "FLASH_MEMORY flash_memory_jedec_read ");
    DRV_SPI_BUFFER_HANDLE bufferHandle;
    uint8_t write_len = 1;
    uint8_t read_len = 6;
    
    char write_buffer[1];
    write_buffer[0] = 0x9F;

    char read_buffer[read_len];
    memset(read_buffer, 4, read_len);

    DRV_SPI_BufferAddWriteRead2(flash_memory_data.spi_handle, 
            write_buffer, write_len, 
            read_buffer, read_len, 
            NULL, //event_handler, //DRV_SPI_BUFFER_EVENT_HANDLER
            NULL, 
            &bufferHandle );
    if(bufferHandle != DRV_SPI_BUFFER_HANDLE_INVALID )
    {
        logger_log_debugf(0, "FLASH_MEMORY read success %d %d %d %d %d %d ", read_buffer[0], read_buffer[1], read_buffer[2], read_buffer[3], read_buffer[4], read_buffer[5]);
    }
    
    logger_log_debugf(0, "FLASH_MEMORY flash_memory_jedec_read end ");  
}
    
/**
 * 
 */
void flash_memory_read_status_register()
{
    logger_log_debugf(0, "FLASH_MEMORY flash_memory_read_status_register ");
    DRV_SPI_BUFFER_HANDLE bufferHandle;
    uint8_t write_len = 1;
    uint8_t read_len = 1;
    
    char write_buffer[1];
    write_buffer[0] = 0x05;

    char read_buffer[read_len];
    memset(read_buffer, 0x04, read_len);

    DRV_SPI_BufferAddWriteRead2(flash_memory_data.spi_handle, 
            write_buffer, write_len, 
            read_buffer, read_len, 
            NULL, //event_handler, //DRV_SPI_BUFFER_EVENT_HANDLER
            NULL, 
            &bufferHandle );
    if(bufferHandle != DRV_SPI_BUFFER_HANDLE_INVALID )
    {
        logger_log_debugf(0, "FLASH_MEMORY read success %d ", read_buffer[0]);
    }
    
    logger_log_debugf(0, "FLASH_MEMORY flash_memory_read_status_register end ");  
}

//---------------------------------------------------------------------------------

/**
 * flash memory processing
 */
void flash_memory_process()
{
    // Are there any requests to process (only do one per cycle)
    FLASH_MEMORY_PAGE_REQUEST* request = &flash_memory_data.requests[flash_memory_data.current_page_request_idx];
    if (request->active)
    {
        logger_log_debugf(0, "FLASH_MEMORY Processing request for page %d at pos %d -- ", request->page_num, flash_memory_data.current_page_request_idx);

        // Do a Write (page number to read etc) Read (get the data) request
        DRV_SPI_BUFFER_HANDLE bufferHandle;
        
        //TODO: We need to work out the address based on the Page/Offset
        // This might be more complicated as the SPI flash we were using had 
        // a variety of block sizes ?????????????????????
        
        //TODO: Read or Write request......
        if (!request->write_request)
        {
            // Read request
            //TODO : This bit will need setting up correctly
            char write_buffer_full[10];
            write_buffer_full[0] = 0x06;    // Write_enable
            write_buffer_full[0] = 0x20;    // Erase Sector
            write_buffer_full[2] = 0x00;    // Address
            write_buffer_full[3] = 0x00;    // Address
            write_buffer_full[4] = 0x00;    // Address
            write_buffer_full[5] = 0x00;    // Address

            bufferHandle = DRV_SPI_BufferAddWriteRead2(flash_memory_data.spi_handle, 
                    write_buffer_full,
                    10,
                    request->data + request->offset, 
                    request->length, 
                    (DRV_SPI_BUFFER_EVENT_HANDLER)flash_memory_callback, 
                    (void*)&request, 
                    &bufferHandle );
        }
        else
        {
            // Read request
            //TODO : This bit will need setting up correctly
            char write_buffer_full[10 + FLASH_MEMORY_PAGE_SIZE];
            write_buffer_full[0] = 0x06;    // Write_enable
            write_buffer_full[0] = 0x20;    // Erase Sector
            write_buffer_full[2] = 0x00;    // Address
            write_buffer_full[3] = 0x00;    // Address
            write_buffer_full[4] = 0x00;    // Address
            write_buffer_full[5] = 0x00;    // Address

            bufferHandle = DRV_SPI_BufferAddWrite2(flash_memory_data.spi_handle, 
                    write_buffer_full,
                    request->length, 
                    (DRV_SPI_BUFFER_EVENT_HANDLER)flash_memory_callback, 
                    (void*)&request, 
                    &bufferHandle );
            
        }
        //????????????????????????????????????????????????????????????????????????????????????????

        //DRV_SPI_BUFFER_HANDLE bufferHandle;

        // JEDEC
        flash_memory_enable_transfer();
        flash_memory_jedec_read();
        flash_memory_disable_transfer();
        
        timers_delay_ms(10);
        
        flash_memory_read_status_register();
        
        char write_buffer_full[10];
        write_buffer_full[0] = 0x06;    // Write_enable
        write_buffer_full[0] = 0x20;    // Erase Sector
        write_buffer_full[2] = 0x00;    // Address
        write_buffer_full[3] = 0x00;    // Address
        write_buffer_full[4] = 0x00;    // Address
        write_buffer_full[5] = 0x00;    // Address
        
        bufferHandle = DRV_SPI_BufferAddWrite2(flash_memory_data.spi_handle, 
                write_buffer_full, 5, NULL, NULL, &bufferHandle );
        if(bufferHandle != DRV_SPI_BUFFER_HANDLE_INVALID )
        {
            logger_log_errorf(0, "FLASH_MEMORY erase success ");
            
            if( DRV_SPI_BUFFER_EVENT_COMPLETE & DRV_SPI_BufferStatus( bufferHandle ) )
            {
                logger_log_errorf(0, "FLASH_MEMORY ERASE COMPLETE ");
            }
            else
            {
                logger_log_errorf(0, "FLASH_MEMORY ERASE NOOOOT COMPLETE ");
            }
        }

        write_buffer_full[0] = 0x06;    // Write_enable
        write_buffer_full[1] = 0xAD;    // Write-Program
        write_buffer_full[2] = 0x00;    // Address
        write_buffer_full[3] = 0x00;    // Address
        write_buffer_full[4] = 0x00;    // Address
        write_buffer_full[5] = 0x00;    // Address
        write_buffer_full[6] = 0x01;
        write_buffer_full[7] = 0x02;
        write_buffer_full[8] = 0x04;    // Write disable
        
        bufferHandle = DRV_SPI_BufferAddWrite2(flash_memory_data.spi_handle, 
                write_buffer_full, 10, NULL, NULL, &bufferHandle );
        if(bufferHandle != DRV_SPI_BUFFER_HANDLE_INVALID )
        {
            logger_log_errorf(0, "FLASH_MEMORY write success ");
            
            if( DRV_SPI_BUFFER_EVENT_COMPLETE & DRV_SPI_BufferStatus( bufferHandle ) )
            {
                logger_log_errorf(0, "FLASH_MEMORY WRITE COMPLETE ");
            }
            else
            {
                logger_log_errorf(0, "FLASH_MEMORY WRITE NOOOOT COMPLETE ");
            }
        }
        
        // Make a request to read data from the details in the request
        char write_buffer[5];
        write_buffer[0] = 0x0B;
        write_buffer[1] = 0x00;
        write_buffer[2] = 0x00;
        write_buffer[3] = 0x00;
        write_buffer[4] = 0x00;
        
        char read_buffer[100];
        memset(read_buffer, 4, 100);
        
        //DRV_SPI_BUFFER_EVENT_HANDLER event_handler = &flash_memory_event_handler;

        logger_log_debugf(0, "FLASH_MEMORY pre read %d %d %d %d %d %d ", read_buffer[0], read_buffer[1], read_buffer[2], read_buffer[3], read_buffer[4], read_buffer[5]);
        bufferHandle = DRV_SPI_BufferAddWrite2(flash_memory_data.spi_handle, 
                write_buffer, 5, NULL, NULL, &bufferHandle );
        
        logger_log_errorf(0, "FLASH_MEMORY WRITE REQUEST WAITING ");
        while (!DRV_SPI_BUFFER_EVENT_COMPLETE & DRV_SPI_BufferStatus( bufferHandle ))
        {
            logger_log_errorf(0, "FLASH_MEMORY WRITE REQUEST WAITING xxxx ");
            Nop();
        }
       if( DRV_SPI_BUFFER_EVENT_COMPLETE & DRV_SPI_BufferStatus( bufferHandle ) )
        {
            logger_log_errorf(0, "FLASH_MEMORY WRITE REQUEST COMPLETE ");
            // All transmitter data has been sent successfully.
        }
        else
        {
            logger_log_errorf(0, "FLASH_MEMORY WRITE REQUEST NOOOOT COMPLETE ");
        }

        
        bufferHandle = DRV_SPI_BufferAddRead2(flash_memory_data.spi_handle, 
         read_buffer, 100, NULL, NULL, &bufferHandle );

//        DRV_SPI_BufferAddWriteRead2(flash_memory_data.spi_handle, 
//                write_buffer, 5, 
//                read_buffer, 100, 
//                NULL, //event_handler, //DRV_SPI_BUFFER_EVENT_HANDLER
//                NULL, 
//                &bufferHandle );
        if(bufferHandle != DRV_SPI_BUFFER_HANDLE_INVALID )
        {
            logger_log_errorf(0, "FLASH_MEMORY read success ");
        }

        
        if( DRV_SPI_BUFFER_EVENT_COMPLETE & DRV_SPI_BufferStatus( bufferHandle ) )
        {
            logger_log_errorf(0, "FLASH_MEMORY COMPLETE ");
            // All transmitter data has been sent successfully.
        }
        else
        {
            logger_log_errorf(0, "FLASH_MEMORY NOOOOT COMPLETE ");
        }
        
        
        
            logger_log_errorf(0, "FLASH_MEMORY post read %d %d %d %d %d %d ", read_buffer[0], read_buffer[1], read_buffer[2], read_buffer[3], read_buffer[4], read_buffer[5]);

        //????????????????????????????????????????????????????????????????????????????????????????
//
//        if (request->callback != NULL)
//        {
//            // If this is synchronous the invoke the callback
//            FLASH_MEMORY_PAGE_DATA data;
//            data.page_num = request->page_num;
//            data.data = NULL; //??????? The data
//            data.context = request->context;
//
//            request->callback(&data);
//        }
        
        // Release the request
        request->active = false;

        //! Find next free page request 
        flash_memory_data.current_page_request_idx++;
        if (flash_memory_data.current_page_request_idx >= FLASH_MEMORY_REQUEST_QUEUE_SIZE)
        {
            flash_memory_data.current_page_request_idx = 0;
        }
    }
}

/* *****************************************************************************
 End of File
 */
