/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    font_manager.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/font_manager.h"
#include <stddef.h>
#include <stdint.h>
#include "utilities/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Define the default unicode character that will be used if the specified character is not available in the current font set
#ifndef FONT_MANAGER_DFAULT_UNICODE_CHARACTER
#define FONT_MANAGER_DFAULT_UNICODE_CHARACTER   0x007F
#endif

//! The maximum number of fonts supported by the font manager
#define FONT_MANAGER_MAX_FONTS  12//Arbitary number

//! The offset in the font data of the height and width
// Note its assuming all characters have the same width
#define FIRST_CHAR_OFFSET   2
#define LAST_CHAR_OFFSET    4
#define HEIGHT_OFFSET   6
#define WIDTH_OFFSET    8

#define PIXEL_SATURATION_BLACK  0x00
#define PIXEL_SATURATION_DARK   0x55
#define PIXEL_SATURATION_LIGHT  0xAA
#define PIXEL_SATURATION_WHITE  0xFF

#define BIT_IS_SET(VAL, N)	(((VAL & (1 << N)) > 0) ? true : false)


static uint8_t SATURATION_MAP[] = 
{
    0x00,    
    0x11,
    0x22,
    0x33,
    0x44,
    0x55,
    0x66,
    0x77,
    0x88,
    0x99,
    0xAA,
    0xBB,
    0xCC,
    0xDD,
    0xEE,
    0xFF,
};
/**
 * Information about a font
 */
typedef struct
{
    // The ascii code of the first character in the font set
    uint8_t first_char;
    uint8_t last_char;
    
    uint8_t avg_char_width;
    uint8_t max_char_width;
    
    FONT_ID font_id;
    
    const char* font_data;
    
    //! Then number of pixels represented by a single byte of character data
    // Monochrome = 8
    // Aliased 2pp = 4
    uint8_t pixels_per_byte;
    
    // The number of font ranges
    uint8_t font_range_count;
    
    FONT_RANGE font_range[FONT_RANGE_MAX_COUNT];
} FONT_INFO;

/**
 * Font Manager data
 */
typedef struct
{
    //! Font information
    FONT_INFO font_info[FONT_MANAGER_MAX_FONTS];
} FONT_MANAGER_DATA;

static FONT_MANAGER_DATA font_manager_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Locate font info based on the ID
 * @param font_id
 * @return 
 */
static FONT_INFO* font_manager_get_font_info(FONT_ID font_id)
{
    FONT_INFO* result = NULL;
    
    int x;
    for (x = 0; x < FONT_MANAGER_MAX_FONTS; x++)
    {
        if (font_manager_data.font_info[x].font_id == font_id)
        {
            result = &font_manager_data.font_info[x];
            break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param data
 * @param pixel_idx
 * @return 
 */
uint8_t font_manager_get_pixel_saturation_1bit(char* data, uint8_t bit_idx)
{
    uint8_t result;
    
    uint8_t num_bits = 1;
    
    result = (*data >> ((((bit_idx + 1) / num_bits) - 1) * num_bits)) & 0x01;
    
    result = SATURATION_MAP[result * 0x0F];
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param data
 * @param pixel_idx
 * @return 
 */
uint8_t font_manager_get_pixel_saturation_2bit(char* data, uint8_t bit_idx)
{
    uint8_t result;
    
    uint8_t num_bits = 2;
    
    result = (*data >> ((((bit_idx + 1) / num_bits) - 1) * num_bits)) & 0x03;
    
    result = SATURATION_MAP[result * 0x05];
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param data
 * @param pixel_idx
 * @return 
 */
uint8_t font_manager_get_pixel_saturation_4bit(char* data, uint8_t bit_idx)
{
    uint8_t result;
    
    uint8_t num_bits = 4;
    
    result = (*data >> ((((bit_idx + 1) / num_bits) - 1) * num_bits)) & 0x0F;
    
    result = SATURATION_MAP[result];
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param font_id
 * @param character
 * @return 
 */
uint16_t font_manager_index_of_character(FONT_ID font_id, UNICODE_CHAR* character)
{
    uint16_t result = 0xFFFF;

    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        uint8_t x;
        for (x = 0; x < info->font_range_count; x++)
        {
            if ((*character >= info->font_range[x].start) && (*character <= info->font_range[x].end))
            {
                result = info->font_range[x].offset + (*character - info->font_range[x].start);
                break;
            }
        }
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the height of the font
 * @param font_id
 * @return 
 */
const char* font_manager_get_unicode_character_data(FONT_ID font_id, UNICODE_CHAR* character)
{
    const char* result = NULL;
    
    // Get the offset of the character from the character table
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        uint16_t index = font_manager_index_of_character(font_id, character);
        
        if ((*character != FONT_MANAGER_DFAULT_UNICODE_CHARACTER) && (index == 0xFFFF))
        {
            logger_log_debugf(0, "FONT MANAGER Replacing Unsupported character %x --", *character);
            UNICODE_CHAR default_char = FONT_MANAGER_DFAULT_UNICODE_CHARACTER;
            index = font_manager_index_of_character(font_id, &default_char);
        }

        if (index != 0xFFFF)
        {
            const char* font_data = info->font_data;
            const char* char_info = font_data + WIDTH_OFFSET + (index * 4);
            uint8_t byte1 = *(char_info + 1);
            uint8_t byte2 = *(char_info + 2);
            uint8_t byte3 = *(char_info + 3);
            uint32_t offset = (byte3 << 16) + (byte2 << 8) + byte1;
            result = font_data + offset;
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the height of the font
 * @param font_id
 * @return 
 */
const char* font_manager_get_character_data(FONT_ID font_id, char* character)
{
    const char* result = NULL;
    
    // Get the offset of the character from the character table
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        if ((*character >= info->first_char) && (*character <= info->last_char))
        {
            const char* font_data = info->font_data;
            const char* char_info = font_data + WIDTH_OFFSET + ((*character - info->first_char) * 4);
            uint8_t byte1 = *(char_info + 1);
            uint8_t byte2 = *(char_info + 2);
            uint8_t byte3 = *(char_info + 3);
            uint32_t offset = (byte3 << 16) + (byte2 << 8) + byte1;
            result = font_data + offset;
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the width of the specified character of the font
 * @param font_id
 * @return 
 */
uint8_t font_manager_get_unicode_character_width(FONT_ID font_id, UNICODE_CHAR* character)
{
    uint8_t result = 0;
    
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        uint16_t index = font_manager_index_of_character(font_id, character);
        
        if ((*character != FONT_MANAGER_DFAULT_UNICODE_CHARACTER) && (index == 0xFFFF))
        {
            UNICODE_CHAR default_char = FONT_MANAGER_DFAULT_UNICODE_CHARACTER;
            index = font_manager_index_of_character(font_id, &default_char);
        }
        
        if (index != 0xFFFF)
        {
            const char* font_data = info->font_data;
            result = *(font_data + WIDTH_OFFSET + (index * 4));
        }
        else
        {
            result = font_manager_get_font_average_width(font_id);
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the width of the specified character of the font
 * @param font_id
 * @return 
 */
uint8_t font_manager_get_character_width(FONT_ID font_id, char* character)
{
    uint8_t result = 0;
    
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        if ((*character >= info->first_char) && (*character <= info->last_char))
        {
            const char* font_data = info->font_data;
            result = *(font_data + WIDTH_OFFSET + ((*character - info->first_char) * 4));
        }
        else
        {
            result = font_manager_get_font_average_width(font_id);
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the height of the font
 * @param font_id
 * @return 
 */
uint8_t font_manager_get_font_height(FONT_ID font_id)
{
    uint8_t result = 0;
    
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        const char* font_data = info->font_data;
        result = *(font_data + HEIGHT_OFFSET);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the average width of the font
 * @param font_id
 * @return 
 */
int font_manager_get_font_average_width(FONT_ID font_id)
{
    int result = 0;
    
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        result = info->avg_char_width;
    }
}

/**-----------------------------------------------------------------------------
 * Gets the maximum  width of the font
 * @param font_id
 * @return 
 */
int font_manager_get_font_max_width(FONT_ID font_id)
{
    int result = 0;
    
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        result = info->max_char_width;
    }
}

/**-----------------------------------------------------------------------------
 * Gets the width of the monotype of the font
 * @param font_id
 * @return 
 */
int font_manager_get_font_width(FONT_ID font_id)
{
    int result = 0;
    
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        // Note uses the width of the first character - this should be OK as its a MONOTYPE font
        const char* font_data = info->font_data;
        uint8_t width = *(font_data + WIDTH_OFFSET);
        result = width;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the pixels colour at the given index based on whether its a monochrome or anti-aliased font
 * @param data
 * @param pixel_idx - identifies which pixel in the data to extract 0 - pixels_per_byte - 1
 * @return 
 */
uint8_t font_manager_get_pixel_saturation(FONT_ID font_id, char* data, uint8_t pixel_idx)
{
    uint8_t result = PIXEL_SATURATION_WHITE;
    uint8_t pixels_per_byte = font_manager_get_pixels_per_byte(font_id);
    
    uint8_t bit_idx = 7 - ((pixel_idx % 8) * (8 / pixels_per_byte));

    //logger_log_debugf(0, "FONT %d - %d %d --", *data, pixel_idx, bit_idx );

    switch (pixels_per_byte)
    {
        case 2:
            // Aliased 4pp
            result = font_manager_get_pixel_saturation_4bit(data, bit_idx);
            break;
            
        case 4:
            // Aliased 2pp
            result = font_manager_get_pixel_saturation_2bit(data, bit_idx);
            break;
            
        case 8:
            result = font_manager_get_pixel_saturation_1bit(data, bit_idx);
            break;        
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the number of pixels represented by a single byte
 * @param font_id
 * @return 
 */
uint8_t font_manager_get_pixels_per_byte(FONT_ID font_id)
{
    uint8_t result = 0;
    
    FONT_INFO* info = font_manager_get_font_info(font_id);
    if (info != NULL)
    {
        result = info->pixels_per_byte;
    }
    
    return result;    
}
    
/**-----------------------------------------------------------------------------
 * 
 * @param text
 * @return 
 */
uint16_t font_manager_get_unicode_text_length(UNICODE_CHAR* text)
{
    uint16_t result = 0;
    while ((text[result] != 0) && (text[result] != 0xFFFF))
    {
        //logger_log_debugf(0, "TEXT Unicode %d --", text[result]);
        result++;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Initialise the font manager
 */
void font_manager_initialise()
{
    int x;
    for (x = 0; x < FONT_MANAGER_MAX_FONTS; x++)
    {
        font_manager_data.font_info[x].font_data = NULL;
    }
}

/**-----------------------------------------------------------------------------
 * Used to register a font with a specific font id in the system.  Font ids must be unique to each font
 * to be used.
 * @param font_id
 * @param font_data a pointer to the character array generated by BitFontCreator
 * @param pixels_per_byte - The number of pixels represented by each byte of data
 */
void font_manager_register_font(FONT_ID font_id, const char* font_data, int pixels_per_byte, FONT_RANGE* ranges, uint8_t font_range_count)
{
    int x;
    for (x = 0; x < FONT_MANAGER_MAX_FONTS; x++)
    {
        if (font_manager_data.font_info[x].font_data == NULL)
        {
            logger_log_infof(0, "FONT Registering Font ID %d  ", font_id);
            
            font_manager_data.font_info[x].font_id = font_id;
            font_manager_data.font_info[x].font_data = font_data;
            font_manager_data.font_info[x].pixels_per_byte = pixels_per_byte;
            
            // Get first character
            font_manager_data.font_info[x].first_char = *(font_data + FIRST_CHAR_OFFSET);
            font_manager_data.font_info[x].last_char = *(font_data + LAST_CHAR_OFFSET);
            
            // Work out the widest character
            uint8_t num_chars = font_manager_data.font_info[x].last_char - font_manager_data.font_info[x].first_char + 1;
            uint16_t avg_char_width = 0;
            font_manager_data.font_info[x].max_char_width = 0;
            int y;
            for (y = 0; y < num_chars; y++)
            {
                uint8_t char_width = *(font_data + WIDTH_OFFSET + (y * 4));
                if (char_width > font_manager_data.font_info[x].max_char_width)
                {
                    font_manager_data.font_info[x].max_char_width = char_width;
                }
                
                avg_char_width += char_width;
            }
            
            font_manager_data.font_info[x].avg_char_width = avg_char_width / num_chars;
            
            // Note: The ranges should b added in the same order as they are defined in the font file
            font_manager_data.font_info[x].font_range_count = font_range_count;
            uint16_t font_range_offset = 0;
            for (y = 0; y < font_range_count; y++)
            {
                font_manager_data.font_info[x].font_range[y].start = ranges[y].start;
                font_manager_data.font_info[x].font_range[y].end = ranges[y].end;
                font_manager_data.font_info[x].font_range[y].offset = font_range_offset;
                
                font_range_offset += (ranges[y].end - ranges[y].start + 1);
            }
            
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Converts a utf-8 character string into an array of uint16_t unicode characters
 * @param uint8_str
 * @param unicode_str - assumes this is large enough to hold a 2-byte unicode 
 * representation of the string. can be declared by using uint16_t unicode_str[strlen(utf8_str)]
 */
void font_manager_utf8_string_to_unicode(char* utf8_str, uint16_t* unicode_str)
{
    uint16_t input_len = strlen(utf8_str);
    
    uint16_t input_idx = 0;
    uint16_t output_idx = 0;
    uint16_t unicode_char = 0;
    
    bool unhandled = false;
    
    while (input_idx < input_len)
    {
        uint8_t utf8 = (uint8_t)utf8_str[input_idx];
        
        if (utf8 <= 0x7F)
        {
            // 1-byte utf-8
            unicode_char = (uint16_t)utf8;
            //logger_log_debugf(0, "MOTORHEAD 1-byte %x %x %x --", unicode_char, utf8, unicode_char);
            input_idx++;
            
            unhandled = false;
        }
        else if (!unhandled)
        {
            if (utf8 <= 0xFFFF)
            {   
                // 2-byte utf-8
                unicode_char = utf8;                  // Get 1st byte
                unicode_char -= 192;                  // Subtract 192
                unicode_char *= 64;                   // Multiply by 64
                input_idx++;

                utf8 = (uint8_t)utf8_str[input_idx];
                unicode_char += ((utf8 & 0xFF) - 0x80);   // Get second byte, subtract 128
                //logger_log_debugf(0, "MOTORHEAD 2-byte %x %x %x --", unicode_char, utf8, unicode_char);
                input_idx++;
            }
            else if (utf8 <= 0xFFFFFF)
            {
                // 3-byte utf8
                utf8 -= 224;
                unicode_char = utf8 * 4096;
                input_idx++;

                utf8 = (uint8_t)utf8_str[input_idx];
                utf8  -= 128;
                unicode_char += utf8 * 64;
                input_idx++;

                utf8 = (uint8_t)utf8_str[input_idx];
                unicode_char += (utf8 - 0x80);   // Get second byte, subtract 128
                input_idx++;
            }
            else
            {
                // Unhandled character
                input_idx++;
                unhandled = true;
            }
        }
        else
        {
            input_idx++;
        }
        
        if (!unhandled)
        {
            unicode_str[output_idx] = unicode_char;
            //logger_log_debugf(0, "MOTORHEAD AA %d %d --", unicode_char, unicode_str[output_idx]);
            output_idx++;
        }
    }
    
    unicode_str[output_idx] = 0;
}

/* *****************************************************************************
 End of File
 */
