/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/hdmi_arc.h"
#include "app_config.h"
#include "utilities/i2c.h"
#include "utilities/hdmi_arc_processing.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! I2C Client Id for identifying the module on the I2C bus
    I2C_CLIENT_ID i2c_client_id;
    
} HDMI_ARC_DATA;

static HDMI_ARC_DATA hdmi_arc_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the current setting deciding whether HDMI arc can be used to control
 * audio volume
 * @return 
 */
HDMI_ARC_CONTROL_STATE hdmi_arc_get_audio_control_setting()
{
    uint8_t status;
    settings_get(SETTING_HDMI_AUDIO_CONTROL, &status);
    
    if (status >= HDMI_ARC_CONTROL_STATE_MAX)
    {
        status = HDMI_ARC_CONTROL_STATE_OFF;
        hdmi_arc_set_audio_control_setting(status);
    }
    
    return (HDMI_ARC_CONTROL_STATE)status;
}

/**-----------------------------------------------------------------------------
 * Gets the current setting deciding whether HDMI arc can be used to control
 * power state
 * @return 
 */
HDMI_ARC_CONTROL_STATE hdmi_arc_get_power_control_setting()
{
    uint8_t status;
    settings_get(SETTING_HDMI_POWER_CONTROL, &status);
    
    if (status >= HDMI_ARC_CONTROL_STATE_MAX)
    {
        status = HDMI_ARC_CONTROL_STATE_OFF;
        hdmi_arc_set_power_control_setting(status);
    }
    
    return (HDMI_ARC_CONTROL_STATE)status;
}

/**-----------------------------------------------------------------------------
 * Initialises the HSDMI ARC module
 */
bool hdmi_arc_initialise(HDMI_ARC_MESSAGE_CALLBACK callback)
{
    bool result = true;
    
    // Set reset high
    hdmi_arc_reset(true);
    
    hdmi_arc_processing_initialise(callback);
    
    return true;
}

/**-----------------------------------------------------------------------------
 * Initialises the supplied HDMI ARC Message readying it for populating
 * @param message
 */
void hdmi_arc_message_initialise(HDMI_ARC_MESSAGE* message)
{
    message->command_type = HDMI_ARC_COMMAND_UNKNOWN;
    message->processing_state = HDMI_ARC_MESSAGE_STATE_INITIALISED;
    message->processing_data_index = 0;
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether the display message has been fully populated
 * @param message
 * @return 
 */
bool hdmi_arc_message_is_complete(HDMI_ARC_MESSAGE* message)
{
    return message->processing_state == HDMI_ARC_MESSAGE_STATE_POPULATED;    
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether the display message is currently being populated
 * @param message
 * @return 
 */
bool hdmi_arc_message_is_processing(HDMI_ARC_MESSAGE* message)
{
    return message->processing_state != HDMI_ARC_MESSAGE_STATE_INITIALISED;
}

/**-----------------------------------------------------------------------------
 * Process a byte of data read from the RX channel
 * @param data_byte
 */
void hdmi_arc_message_process_byte(HDMI_ARC_MESSAGE* message, unsigned char data_byte)
{
    //ogger_log_debugf(0, "HDMI_ARC Process %d --", data_byte);
    switch (message->processing_state)
    {
        case HDMI_ARC_MESSAGE_STATE_INITIALISED:
        {
            if (data_byte == HDMI_ARC_MESSAGE_START_OF_PACKET)
            {
                message->processing_state = HDMI_ARC_MESSAGE_STATE_FOR_LEN_LSB;
                //logger_log_debugf(0, "HDMI_ARC Start of packet");
            }
            break;
        }
        
        case HDMI_ARC_MESSAGE_STATE_FOR_LEN_LSB:
        {
            message->length = (uint8_t)data_byte;
            message->processing_state = HDMI_ARC_MESSAGE_STATE_WAIT_LEN_MSB;
            break;
        }
        
        case HDMI_ARC_MESSAGE_STATE_WAIT_LEN_MSB:
        {
            message->length += ((uint16_t)data_byte * 256);
            message->processing_state = HDMI_ARC_MESSAGE_STATE_WAIT_COMMAND_TYPE;
            //logger_log_debugf(0, "HDMI_ARC length %d --", message->length);
            timers_delay_us(50); // Not sure why this is needed but it does not complete unless its put in..
            break;
        }
        
        case HDMI_ARC_MESSAGE_STATE_WAIT_COMMAND_TYPE:
        {
            message->command_type = (HDMI_ARC_COMMAND_TYPE)data_byte;
            message->processing_data_index = 0;
            message->processing_state = HDMI_ARC_MESSAGE_STATE_POPULATING;
            break;
        }
        
        case HDMI_ARC_MESSAGE_STATE_POPULATING:
        {
            message->data[message->processing_data_index] = data_byte;
            message->processing_data_index++;
            
            if (message->processing_data_index >= message->length - 4)
            {
                if (message->data[message->processing_data_index - 1] == HDMI_ARC_MESSAGE_END_OF_PACKET)
                {
//                    logger_log_debugf(0, "HDMI ARC Message %d %d %d --",
//                            message->length,
//                            message->command_type,
//                            message->data[0]);
//                    logger_process();
                    message->processing_state = HDMI_ARC_MESSAGE_STATE_POPULATED;
                }
                else
                {
                    logger_log_debugf(0, "HDMI_ARC display_message_process_byte - Invalid end characters %d %d --", data_byte, message->data[message->processing_data_index - 1]);
                    logger_process();
                    hdmi_arc_message_initialise(message);
                }
            }
            else
            {
//                    logger_log_debugf(0, "HDMI ARC Message %d %d %d %d --",
//                            message->length,
//                            message->command_type,
//                            data_byte,
//                            message->processing_data_index);
//                    logger_process();
            }
            break;
        }

        // The message has been completely received and is ready for processing
        case HDMI_ARC_MESSAGE_STATE_POPULATED:
        {
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Used to reset the ARC chip
 * When  the pin is high it is being taken out of reset.
 */
void hdmi_arc_reset(bool on)
{
#if HDMI_ARC_ENABLED   
       
    if (on)
    {
        PLIB_PORTS_PinSet(HDMI_ARC_RESET_PORTS_ID, 
                          HDMI_ARC_RESET_PORT_CHANNEL, 
                          HDMI_ARC_RESET_PIN);    
    }
    else
    {
        PLIB_PORTS_PinClear(HDMI_ARC_RESET_PORTS_ID, 
                            HDMI_ARC_RESET_PORT_CHANNEL, 
                            HDMI_ARC_RESET_PIN);
    }
#endif
}

/**-----------------------------------------------------------------------------
 * Sets the current setting deciding whether HDMI arc can be used to control
 * audio volume
 * @return 
 */
void hdmi_arc_set_audio_control_setting(HDMI_ARC_CONTROL_STATE value)
{
    uint8_t status = value;
    settings_set(SETTING_HDMI_AUDIO_CONTROL, &status);
}

/**-----------------------------------------------------------------------------
 * Sets the current setting deciding whether HDMI arc can be used to control
 * power state
 * @return 
 */
void hdmi_arc_set_power_control_setting(HDMI_ARC_CONTROL_STATE value)
{
    uint8_t status = value;
    settings_set(SETTING_HDMI_POWER_CONTROL, &status);
}

/* *****************************************************************************
 End of File
 */
