/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    i2c.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/i2c.h"
#include "app_config.h"
#include <stdbool.h>
#include "utilities/i2c_bit_bash.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Default to allow max of 1 I2C connections can b altered in system_setup.h
#ifndef I2C_MAX_MODULES
#define I2C_MAX_MODULES     1
#endif

#ifndef I2C_MAX_CLIENTS
#define I2C_MAX_CLIENTS     2   // For SA20 we only need 2 (eeprom/dac)
#endif

//! Defines the size of the bus request queue for handing unprocessed requests
// 4 is an abitary number
#ifndef I2C_REQUEST_QUEUE_SIZE
#define I2C_REQUEST_QUEUE_SIZE  4
#endif

//! I2C function templates

/**
 * Callback function template
 */
typedef DRV_HANDLE (*I2C_OPEN_FUNCTION)(const SYS_MODULE_INDEX   index,
                                  const DRV_IO_INTENT      intent);

typedef void (*I2C_SET_EVENT_HANDLER_FUNCTION)(const DRV_HANDLE handle,
                                    const DRV_I2C_BUFFER_EVENT_HANDLER event_handler,
                                    const uintptr_t context);

typedef DRV_I2C_BUFFER_HANDLE (*I2C_RECEIVE_FUNCTION)(DRV_HANDLE handle,
                                    uint16_t device_address,
                                    void* read_buffer,
                                    size_t read_size,
                                    void* callback_context);

typedef DRV_I2C_BUFFER_HANDLE (*I2C_TRANSMIT_FUNCTION)(DRV_HANDLE handle,
                                    uint16_t device_address,
                                    void* write_buffer,
                                    size_t write_size,
                                    void* callback_context);

typedef DRV_I2C_BUFFER_HANDLE (*I2C_TRANSMIT_RECEIVE_FUNCTION)(DRV_HANDLE handle,
                                    uint16_t device_address,
                                    void* write_buffer,
                                    size_t write_size,
                                    void* read_buffer,
                                    size_t read_size,
                                    void* callback_context);

/**
 * Contains data related to an I2C client
 */
typedef struct
{
    //! The id/index of the client in the client array
    I2C_CLIENT_ID id;
    
    //! The I2C Module index associated with the client
    SYS_MODULE_INDEX module; 
    
    //! A call back used to send data to an I2C client
    I2C_CALLBACK callback;
    
    //! I2C Read Address for the client
    I2C_SLAVE_ADDRESS_VALUE read_address;
    
    //! I2C Write addres for the peripheral
    I2C_SLAVE_ADDRESS_VALUE write_address;
    
    //! The handle of the currently processing request
    DRV_I2C_BUFFER_HANDLE buffer_handle;
    
    //! Pointer to an object passed to the original call - this should be implemented in 
    // Harmony but isn't yet...
    uintptr_t context;
} I2C_CLIENT;

/**
 * Encapsulates a request for a receive or a transmit
 */
typedef struct
{
    //! A number uniquely identifying the request
    int id;
    I2C_CLIENT* client;
    char* buffer; 
    size_t size;    
    char* buffer2; 
    size_t size2;    
    uintptr_t context;
    I2C_REQUEST_TYPE type;
    bool active;
} I2C_REQUEST;

/** 
 * Encapsulates information about a module
 */
typedef struct
{
    //! The I2C Module index associated with the client
    SYS_MODULE_INDEX index; 

    // Module handle returned by DRV_I2C_Open
    DRV_HANDLE handle;
    
    //! Determines how the I2C messages are processed by the module (e.g Harmony of Bit Bash))
    I2C_MODULE_TYPE module_type;
    
    //! A circular buffer of outstanding requests that get processed serially
    // The purpose of this is so that requests can be queued for a module by different 
    // parts of the system without having to check first whether the bus is in use
    I2C_REQUEST request_queue[I2C_REQUEST_QUEUE_SIZE];
    
    //! Pointers for managing the request_queue
    uint8_t request_queue_active_pos;
    
    //! Indicates whether a request is currently in progress
    bool request_in_progress;
    
    //! Function pointer for the Open function
    I2C_OPEN_FUNCTION open_function;
    
    //! Function pointer for the Set Event Handler function
    I2C_SET_EVENT_HANDLER_FUNCTION set_event_handler_function;
    
    //! Function pointer for the Receive function
    I2C_RECEIVE_FUNCTION receive_function;
    
    //! Function pointer for the Transmit function
    I2C_TRANSMIT_FUNCTION transmit_function;
    
    //! Function pointer for the Transmit Receive function
    I2C_TRANSMIT_RECEIVE_FUNCTION transmit_receive_function;
    
} I2C_MODULE;

/**
 * Main I2C data structure
 */
typedef struct
{
    //! Client information
    I2C_CLIENT clients[I2C_MAX_CLIENTS];
    
    //! Number of clients added
    size_t client_count;
    
    //! Indicates the module is initialised
    bool initialised;
    
    //! Handles to i2c connections
    I2C_MODULE modules[I2C_MAX_MODULES];
    
    //! The next request Id
    int next_id;
        
} I2C_DATA;

static I2C_DATA i2c_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

// Forward declarations
static void i2c_request_recieve(I2C_REQUEST* request);
static void i2c_request_transmit(I2C_REQUEST* request);
static void i2c_request_transmit_receive(I2C_REQUEST* request);

/**-----------------------------------------------------------------------------
 * Callback invoked when a request completes
 * @param event
 * @param bufferHandle
 * @param context
 */
static void i2c_callback(DRV_I2C_BUFFER_EVENT event,
                              DRV_I2C_BUFFER_HANDLE bufferHandle,
                              uintptr_t context)
{
    // Pass onto the module that initiated it
    int x;
    for (x = 0; x < I2C_MAX_CLIENTS; x++)
    {
        if (i2c_data.clients[x].buffer_handle == bufferHandle)
        //if (i2c_data.clients[x].context == context)
        {
            I2C_REQUEST* request = (I2C_REQUEST*)i2c_data.clients[x].context;
            context = request->context;

            // Make request available again
            request->active = false;

            if (i2c_data.clients[x].callback != NULL)
            {
                //logger_log_debugf(0, "I2C: Initiating Callback ");
                // Initiate the callback
                i2c_data.clients[x].callback(event, request->type, context);
            }
            
            // Clear the buffer handle
            i2c_data.clients[x].buffer_handle = NULL;
            i2c_data.clients[x].context = NULL;

            // Unset the in progress flag
            i2c_data.modules[i2c_data.clients[x].module].request_in_progress = false;

            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Find the I2C Client information based on module index
 * @param index
 * @return  - Pointer to I2C_CLIENT structure or a NULL pointer
 */
static I2C_CLIENT* i2c_find_client(I2C_CLIENT_ID id)
{
    I2C_CLIENT* result = NULL;
            
    int x;
    for (x = 0; x < I2C_MAX_CLIENTS; x++)
    {
        if (i2c_data.clients[x].id == id)
        {
            result = &i2c_data.clients[x];
            break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Allocate and return the next request slot
 * @return 
 */
static I2C_REQUEST* i2c_get_next_request(I2C_MODULE* module)
{
    I2C_REQUEST* result = NULL;
    
    // Look at the current read pos
    // if not active then set , otherwise loop through incrementing and find next non-active
    int x;
    for (x = 0; x < I2C_REQUEST_QUEUE_SIZE; x++)
    {
        uint8_t idx = (module->request_queue_active_pos + x) % I2C_REQUEST_QUEUE_SIZE;
        
        I2C_REQUEST* request = &module->request_queue[idx];
        if (!request->active)
        {
            request->active = true;
            request->id = ++i2c_data.next_id;
            result = request;
            //logger_log_debugf(0, "I2C: allocating next request id:%d at %d ", request->id, idx);
            break;
        }
    }
    
    if (result == NULL)
    {
        logger_log_errorf(LOGGER_CODE_I2C_REQUEST_QUEUE_ERROR, "I2C: Request queue overflow on module %d %d", module->index, module->handle);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Process requests for a single module
 * @param module
 */
static void i2c_process_module_requests(I2C_MODULE* module)
{
    //logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL i2c_process_module_requests A ");
    // Check whether we are currently processing a request
    if (!module->request_in_progress)
    {
        //logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL i2c_process_module_requests B ");
        I2C_REQUEST* request = &module->request_queue[module->request_queue_active_pos];
        
        // Check if the request is active
        if (request->active)
        {
            //logger_log_debugf(0, "I2C: Processing request id:%d from:%d", request->id, module->request_queue_active_pos);
        
            switch (request->type)
            {
                case I2C_REQUEST_TYPE_RECEIVE:
                    //logger_log_debugf(0, "I2C: Processing receive request %d 0X%02x 0X%02x", request->id, (uint8_t)request->buffer[0], (uint8_t)request->buffer[1]);
                    i2c_request_recieve(request);
                    break;

                case I2C_REQUEST_TYPE_TRANSMIT:
                    //logger_log_debugf(0, "I2C: Processing transmit request %d 0X%02x 0X%02x", request->id, (uint8_t)request->buffer[0], (uint8_t)request->buffer[1]);
                    i2c_request_transmit(request);
                    break;

                case I2C_REQUEST_TYPE_TRANSMIT_RECEIVE:
                    //logger_log_debugf(0, "I2C: Processing transmit/receive request %d 0X%02x 0X%02x", request->id, (uint8_t)request->buffer[0], (uint8_t)request->buffer[1]);
                    i2c_request_transmit_receive(request);
                    break;
                    
                default:
                    break;
            }

            module->request_queue_active_pos++;
            if (module->request_queue_active_pos >= I2C_REQUEST_QUEUE_SIZE)
            {
                module->request_queue_active_pos = 0;
            }

            //request->active = false;
            module->request_in_progress = true;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Process outstanding requests
 */
static void i2c_process_requests()
{
    // for each module
    int x;
    for  (x = 0; x < I2C_MAX_MODULES; x++)
    {
        if (i2c_data.modules[x].handle != DRV_HANDLE_INVALID)
        {
            i2c_process_module_requests(&i2c_data.modules[x]);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Generate a receive request from an I2C_REQUEST
 * @param request
 */
static void i2c_request_recieve(I2C_REQUEST* request)
{
    I2C_MODULE* module = &i2c_data.modules[request->client->module];
    
    DRV_I2C_BUFFER_HANDLE handle = module->receive_function(module->handle, 
                            request->client->read_address, 
                            request->buffer, 
                            request->size, 
                            NULL);

   if (handle != NULL)
   {
       request->client->buffer_handle = handle;
       request->client->context = (uintptr_t)request;
   }
   else
   {
       logger_log_error(LOGGER_CODE_I2C_READ_DATA_ERROR, "I2C Receive NULL Handle");
   }
}
   
/**----------------------------------------------------------------------------- 
 * Used to setup the type of processing for the module
 * This should be called after i2c_initialise and before and i2c_open calls 
 * are made on the module.
 */
void i2c_set_module_type(const SYS_MODULE_INDEX index,
                         I2C_MODULE_TYPE module_type)
{
    if (index < I2C_MAX_MODULES)
    {
        I2C_MODULE* module = &i2c_data.modules[index];
        
        if (module->handle == DRV_HANDLE_INVALID)
        {
            logger_log_infof(LOGGER_CODE_I2C, "I2C Setting Module Type - Module %d Type %d ", index, module_type);
            module->module_type = module_type;
            
            switch(module->module_type)
            {
                case I2C_MODULE_TYPE_DEFAULT:
                case I2C_MODULE_TYPE_BIT_BASH:
                {
                    module->open_function = i2c_bit_bash_open;
                    module->set_event_handler_function = i2c_bit_bash_set_event_handler;
                    module->receive_function = i2c_bit_bash_receive;
                    module->transmit_function = i2c_bit_bash_transmit;
                    module->transmit_receive_function = i2c_bit_bash_transmit_receive;
                    break;
                }

                case I2C_MODULE_TYPE_HARMONY:
                {
#ifdef I2C_USE_HARMONY                    
                    module->open_function = DRV_I2C_Open;
                    module->set_event_handler_function = DRV_I2C_BufferEventHandlerSet;
                    module->receive_function = DRV_I2C_Receive;
                    module->transmit_function = DRV_I2C_Transmit;
                    module->transmit_receive_function = DRV_I2C_TransmitThenReceive;
#endif                    
                    break;
                }
            }
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_I2C, "I2C Set Module Type FAIL - Module %d has clients ", index);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_I2C, "I2C Set Module Type FAIL - Module %d out of range ", index);
    }
}
 
/**-----------------------------------------------------------------------------
 *  * Generate a transmit request from an I2C_REQUEST
 * @param request
 */
static void i2c_request_transmit(I2C_REQUEST* request)
{
    I2C_MODULE* module = &i2c_data.modules[request->client->module];
    
    //timers_delay_ms(10);
    DRV_I2C_BUFFER_HANDLE handle = module->transmit_function(module->handle, 
                             request->client->write_address, 
                             request->buffer, 
                             request->size, 
                             request);
    //logger_log_debugf(0, "I2C %d %d %02x %02x %02x --", i2c_data.modules[request->client->module].handle, *((uint32_t*)handle), request->client->write_address, request->buffer[0], request->size);

    if (handle != NULL)
    {
        //logger_log_debugf(0, "INPUT CONTROL BUFFER HANDLE %d --", handle);
        request->client->buffer_handle = handle;
        request->client->context = (uintptr_t)request;
    }
    else
    {
        logger_log_error(LOGGER_CODE_I2C_WRITE_DATA_ERROR, "I2C Transmit NULL Handle");
    }
}

/**-----------------------------------------------------------------------------
 *  * Generate a transmit request from an I2C_REQUEST
 * @param request
 */
static void i2c_request_transmit_receive(I2C_REQUEST* request)
{
    I2C_MODULE* module = &i2c_data.modules[request->client->module];
    
    DRV_I2C_BUFFER_HANDLE handle = module->transmit_receive_function(module->handle, 
                             request->client->write_address, 
                             request->buffer, 
                             request->size, 
                             request->buffer2, 
                             request->size2, 
                             NULL);
  
    if (handle != NULL)
    {
        request->client->buffer_handle = handle;
        request->client->context = (uintptr_t)request;
    }
    else
    {
        logger_log_error(LOGGER_CODE_I2C_WRITE_DATA_ERROR, "I2C Transmit Receive NULL Handle");
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Initialise I2C - called in initialisation of modules using i2c (or at leat one...)
 */
void i2c_initialise()
{
    if (!i2c_data.initialised)
    {
        i2c_data.client_count = 0;
        i2c_data.next_id = 0;
        
        // Initialise BUS handles
        int x;
        for  (x = 0; x < I2C_MAX_MODULES; x++)
        {
            i2c_data.modules[x].index = x;
            i2c_data.modules[x].handle = DRV_HANDLE_INVALID;
            i2c_data.modules[x].request_queue_active_pos = 0;
            i2c_data.modules[x].request_in_progress = false;
            
            i2c_set_module_type(i2c_data.modules[x].index, I2C_MODULE_TYPE_DEFAULT);
            
            int y;
            for (y = 0; y < I2C_REQUEST_QUEUE_SIZE; y++)
            {
                i2c_data.modules[x].request_queue[y].active = false;
            }
        }
        
        // Initialise clients
        for  (x = 0; x < I2C_MAX_CLIENTS; x++)
        {
            i2c_data.clients[x].module = 0;
            i2c_data.clients[x].callback = NULL;
            i2c_data.clients[x].buffer_handle = NULL;
            i2c_data.clients[x].context = NULL;
        }
        
        i2c_data.initialised = true;
    }
}

/**-----------------------------------------------------------------------------
 * Opens the specified I2C module - if already open the open handle is returned
 * @param index
 * @return 
 */
I2C_CLIENT_ID i2c_open(const SYS_MODULE_INDEX index, 
            I2C_SLAVE_ADDRESS_VALUE read_address,
            I2C_SLAVE_ADDRESS_VALUE write_address,
            I2C_CALLBACK callback)
{
    I2C_CLIENT_ID result = I2C_INVALID_CLIENT_ID;

#if I2C_ENABLED   
    I2C_MODULE* module = &i2c_data.modules[index];
         
    if (module->handle == DRV_HANDLE_INVALID)
    {
        module->handle = module->open_function(index, DRV_IO_INTENT_READWRITE);
        if (DRV_HANDLE_INVALID != module->handle)
        {
            module->set_event_handler_function(module->handle, i2c_callback, NULL);
         }
        else
        {
            logger_log_error(LOGGER_CODE_I2C_OPEN_ERROR, "I2C Open Error");
        }    
    }
    
    // Sort out the client - note no range checking at the moment
    if (i2c_data.client_count < I2C_MAX_CLIENTS)
    {
        I2C_CLIENT* client = &i2c_data.clients[i2c_data.client_count];
        client->id = i2c_data.client_count;
        client->module = index;
        client->callback = callback;
        client->read_address = read_address;
        client->write_address = write_address;
        i2c_data.client_count++;
        
        result = client->id;
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_I2C_OPEN_ERROR, "I2C Client count exceeded %d", I2C_MAX_CLIENTS);
    }

#endif
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Perform I2C request processing
 */
void i2c_process()
{
    // This delay has been added as we were experiencing issues with I2C "sticking" when 
    // It attempted to read from the DAC once we had removed the OLED code (which was confusing!)
    // It appears that the time taken bu the OLED to initialise some memory (completely unrelated
    // to I2C) was enough delay to allow I2C to work.  This delay appears to fix the issue
    // when OLED is not present
    i2c_bit_bash_process();
    
    i2c_process_requests();
}
    
/**-----------------------------------------------------------------------------
     * Read data from the write address.  This works by placing a transmit followed 
     * by a receive on the queue in sequence ensuring that no other requests are
     * executed in-between
     * Note: The callback will be invoked on completion of both the transmit and the 
     * receive so will need checking for
     * @param id
     * @param address
     * @param address_size
     * @param buffer
     * @param buffer_size
     * @param callbackContext - Not used
     * @param include_intermediate_stopbit - true if a stop bit is required between the transmit address and the receive request
     * @return  
 */
bool i2c_read(const I2C_CLIENT_ID id,
        void * address,
        size_t address_size,
        void * buffer, 
        size_t buffer_size, 
        void * callbackContext,
        bool include_intermediate_stopbit)
{  
    if (!include_intermediate_stopbit)
    {
        if (!i2c_transmit_receive(id, 
                address,
                address_size,
                buffer,  
                buffer_size,
                callbackContext))
        {
            logger_log_error(LOGGER_CODE_I2C_READ_DATA_ERROR, "I2C transmit receive data error");
        }  
    }
    else
    {
        // Trigger a request for the dpll status
        // First transmit the register address
        if (i2c_transmit(id, 
                         address,  
                         address_size,
                         callbackContext))
        {
            if (!i2c_receive(id, 
                            buffer,  
                            buffer_size,
                            callbackContext))
            {
                logger_log_error(LOGGER_CODE_I2C_READ_DATA_ERROR, "I2C read data error");
            }  
        }
        else
        {
            logger_log_error(LOGGER_CODE_I2C_READ_SEND_ADDRESS_ERROR, "I2C read transmit address error");
        } 
    }
}

/**-----------------------------------------------------------------------------
 * Read data from an I2C address
 * @param index - The module index identifying the peripheral to read from
 * @param buffer
 * @param size
 * @param callbackContext - Not used
 * @return 
 */
bool i2c_receive(const I2C_CLIENT_ID id,
            void * buffer, 
            size_t size, 
            void * callbackContext)
{
    DRV_I2C_BUFFER_HANDLE result = false;
    
    // Find the Client
    I2C_CLIENT* client = i2c_find_client(id);
    if (client != NULL)
    {
        I2C_MODULE* module = &i2c_data.modules[client->module];
    
        //logger_log_debugf(0, "I2C: i2c_receive %d", id);
        I2C_REQUEST* request = i2c_get_next_request(module);
        request->client = client;
        request->buffer = buffer;
        request->size = size;
        request->context = (uintptr_t)callbackContext;
        request->type = I2C_REQUEST_TYPE_RECEIVE;
        
        // Try processing it now
        i2c_process_module_requests(module);
        
        result = true;
    }
    else
    {
        logger_log_error(LOGGER_CODE_I2C_INVALID_HANDLE, "I2C Receive Invalid Client Id");
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Transmit data to the write address
 * @param id
 * @param buffer
 * @param size
 * @param callbackContext
 * @return 
 */
bool i2c_transmit(const I2C_CLIENT_ID id,
            void * buffer, 
            size_t size, 
            void * callbackContext)
{
    bool result = false;
    
    // Find the Client
    I2C_CLIENT* client = i2c_find_client(id);
    if (client != NULL)
    {
        //logger_log_debugf(0, "I2C: i2c_transmit %d", id);
        I2C_REQUEST* request = i2c_get_next_request(&i2c_data.modules[client->module]);
        request->client = client;
        request->buffer = buffer;
        request->size = size;
        request->context = (uintptr_t)callbackContext;
        request->type = I2C_REQUEST_TYPE_TRANSMIT;

        i2c_process_module_requests(&i2c_data.modules[client->module]);
        
        result = true;
    }
    else
    {
        logger_log_error(LOGGER_CODE_I2C_INVALID_HANDLE, "I2C Transmit Invalid Client Id");
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Transmit data to the write address the receives data back
 * @param id
 * @param write_buffer
 * @param write_size
 * @param read_buffer
 * @param read_size
 * @param callbackContext - Not used
 * @return 
 */
bool i2c_transmit_receive(const I2C_CLIENT_ID id,
            void * write_buffer, 
            size_t write_size, 
            void * read_buffer, 
            size_t read_size, 
            void * callbackContext)
{
    bool result = false;
    
    // Find the Client
    I2C_CLIENT* client = i2c_find_client(id);
    if (client != NULL)
    {
        //logger_log_debugf(0, "I2C: i2c_transmit %d", id);
        I2C_REQUEST* request = i2c_get_next_request(&i2c_data.modules[client->module]);
        request->client = client;
        request->buffer = write_buffer;
        request->size = write_size;
        request->buffer2 = read_buffer;
        request->size2 = read_size;
        request->context = (uintptr_t)callbackContext;
        request->type = I2C_REQUEST_TYPE_TRANSMIT_RECEIVE;

        i2c_process_module_requests(&i2c_data.modules[client->module]);
        
        result = true;
    }
    else
    {
        logger_log_error(LOGGER_CODE_I2C_INVALID_HANDLE, "I2C Transmit Invalid Client Id");
    }
    
    return result;    
}

/* *****************************************************************************
 End of File
 */
