/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
 For this to work the pins in Harmony need to be set up Output/Low rather than SDA/SCL

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\i2c_bit_bash.h"
#include "app_config.h"
#include "utilities/logger/logger.h"
#include "utilities\timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! General gap in microseconds for a 1/2 clock pulse and things to happen in general.
#define I2C_BIT_BASH_CLOCK_PULSE        1

//! Time in microseconds delay after a start message
#define I2C_BIT_BASH_START_DELAY        7

//! Pause between sending bytes
#define I2C_BIT_BASH_INTER_BYTE_PAUSE   10

#define I2C_BIT_BASH_REQUEST_QUEUE_SIZE 50

/**
 * Processing states
 */
typedef enum
{
    I2C_BIT_BASH_REQUEST_STATE_WAITING,
    I2C_BIT_BASH_REQUEST_STATE_PROCESS_REQUEST,
    I2C_BIT_BASH_REQUEST_STATE_SEND_START,
    I2C_BIT_BASH_REQUEST_STATE_SEND_ADDRESS,
    I2C_BIT_BASH_REQUEST_STATE_READ_DATA,
    I2C_BIT_BASH_REQUEST_STATE_SEND_DATA,
    I2C_BIT_BASH_REQUEST_STATE_SEND_STOP,
    I2C_BIT_BASH_REQUEST_STATE_COMPLETE,
    I2C_BIT_BASH_REQUEST_STATE_ERROR,
} I2C_BIT_BASH_REQUEST_STATE;

/**
 * The request error if known
 */
typedef enum
{
    I2C_BIT_BASH_REQUEST_ERROR_NONE,
    I2C_BIT_BASH_REQUEST_ERROR_INVALID_ADDRESS,
    I2C_BIT_BASH_REQUEST_ERROR_NACK,
} I2C_BIT_BASH_REQUEST_ERROR;

/**
 * Request Type
 */
typedef enum
{
    I2C_BIT_BASH_REQUEST_TYPE_RECEIVE,
    I2C_BIT_BASH_REQUEST_TYPE_TRANSMIT,
    I2C_BIT_BASH_REQUEST_TYPE_TRANSMIT_AND_RECEIVE,
} I2C_BIT_BASH_REQUEST_TYPE;

/**
 * A single read/write request
 */
typedef struct
{
    //! The module handle
    DRV_HANDLE handle;
            
    //! The device address on the modules bus
    uint8_t device_address;
    
    //! The current position in the data
    size_t data_pos;
    
    //! The buffer to send/receive data
    void* read_buffer;
    
    //! The size of the data to be sent/read
    size_t read_size;
    
    //! The buffer to send/receive data
    void* write_buffer;
    
    //! The size of the data to be sent/read
    size_t write_size;
    
    //! Pointer to some data to be returned in the callback
    void* callback_context;
    
    //! The last known error code if the transfer has failed
    I2C_BIT_BASH_REQUEST_ERROR error;
    
    // The current processing state of the request
    I2C_BIT_BASH_REQUEST_STATE state;
    
    //! The type of request
    I2C_BIT_BASH_REQUEST_TYPE type;
    
    //! General variable for managing state 
    uint8_t processing_pos;
} I2C_BIT_BASH_REQUEST;

/**
 * Used to store module info and runtime info
 * Note:  The first 3 data items MUST match I2C_BIT_BASH_MODULE
 */
typedef struct
{
    //! The module index
    SYS_MODULE_INDEX index;

    //! Pin definition for the SCL pin
    I2C_BIT_BASH_PIN scl_pin;

    //! Pin definition for the SDA pin
    I2C_BIT_BASH_PIN sda_pin;
    
    //! The request the module is currently servicing
    // Note: This is being designed to work with the i2c utility which manages 
    // and schedules the request, so there are no queues maintained here
    I2C_BIT_BASH_REQUEST active_request;

    // The event handler to be called on complete or error
    DRV_I2C_BUFFER_EVENT_HANDLER event_handler;
    
} I2C_BIT_BASH_RUNTIME_MODULE;

/**
 * Runtime data
 */
typedef struct
{
    //! Module definitions for bit_bash specific stuff
    I2C_BIT_BASH_RUNTIME_MODULE modules[I2C_MAX_MODULES];
    
    //! An incrementing number for each time a request is used
    uint32_t request_handle;
} I2C_BIT_BASH_DATA;

static I2C_BIT_BASH_DATA i2c_bit_bash_data =
{
    .request_handle = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations
static void i2c_bit_bash_read_byte(I2C_BIT_BASH_RUNTIME_MODULE* module, uint8_t* data);
static void i2c_bit_bash_send_ack(I2C_BIT_BASH_RUNTIME_MODULE* module);
static void i2c_bit_bash_send_byte(I2C_BIT_BASH_RUNTIME_MODULE* module, uint8_t* data);
static bool i2c_bit_bash_send_clock_pulse(I2C_BIT_BASH_RUNTIME_MODULE* module);
static void i2c_bit_bash_send_nack(I2C_BIT_BASH_RUNTIME_MODULE* module);
static void i2c_bit_bash_send_start(I2C_BIT_BASH_RUNTIME_MODULE* module);
static void i2c_bit_bash_send_stop(I2C_BIT_BASH_RUNTIME_MODULE* module);

/**-----------------------------------------------------------------------------
 * Returns a reference to an open module with the matching handle, NULL if not valid.
 * Makes use of the fact that handle = index if properly initialised
 * @param index
 * @return 
 */
static I2C_BIT_BASH_RUNTIME_MODULE* i2c_bit_bash_check_module(const SYS_MODULE_INDEX index)
{
    I2C_BIT_BASH_RUNTIME_MODULE* result = NULL;
    
    if (index < I2C_MAX_MODULES)
    {
        // Check the module has been defined by i2c_bit_bash_add_module
        if (i2c_bit_bash_data.modules[index].index == index)
        {
            result = &i2c_bit_bash_data.modules[index];
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Check - Module %d not defined %d ", index, i2c_bit_bash_data.modules[index].index);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Check - Index out of range");
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Waits for an ACK from the slave device 
 * @param module
 * @return True if ACK, False if NACK
 */
static bool i2c_bit_bash_get_response(I2C_BIT_BASH_RUNTIME_MODULE* module)
{
    // Set SDA to input
    PLIB_PORTS_PinDirectionInputSet(module->sda_pin.port, 
                                    module->sda_pin.channel, 
                                    module->sda_pin.pin);

    return i2c_bit_bash_send_clock_pulse(module);
}

/**-----------------------------------------------------------------------------
 * 
 * @param module
 */
static void i2c_bit_bash_process_module(I2C_BIT_BASH_RUNTIME_MODULE* module)
{
    switch(module->active_request.state)
    {
        case I2C_BIT_BASH_REQUEST_STATE_PROCESS_REQUEST:
        {
            module->active_request.data_pos = 0;
            module->active_request.processing_pos = 0;
                
            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_START;
            break;
        }
        
        case I2C_BIT_BASH_REQUEST_STATE_SEND_START:
        {
            module->active_request.error = I2C_BIT_BASH_REQUEST_ERROR_NONE;
            
            // Send the start message
            i2c_bit_bash_send_start(module);
            
            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_ADDRESS;
            break;
        }
        
        case I2C_BIT_BASH_REQUEST_STATE_SEND_ADDRESS:
        {
            // Send the device address
            i2c_bit_bash_send_byte(module, 
                    (uint8_t*)&module->active_request.device_address);

            // Wait for ACK
            if (!i2c_bit_bash_get_response(module))
            {
                // Delay between command lines
                timers_delay_us(I2C_BIT_BASH_INTER_BYTE_PAUSE);
                
                switch (module->active_request.type)
                {
                    case I2C_BIT_BASH_REQUEST_TYPE_RECEIVE:
                    {
                        // The address sent will be the read address, so now we just start reading the data
                        module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_READ_DATA;
                        break;
                    }
                    case I2C_BIT_BASH_REQUEST_TYPE_TRANSMIT:
                    {
                        module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_DATA;
                        break;
                    }
                    case I2C_BIT_BASH_REQUEST_TYPE_TRANSMIT_AND_RECEIVE:
                    {
                        if (module->active_request.processing_pos == 0)
                        {
                            // Move to sending the Register position to read
                            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_DATA;
                            module->active_request.processing_pos++;
                        }
                        else
                        {
                            module->active_request.data_pos = 0;
                            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_READ_DATA;
                        }

                        break;
                    }
                }
            }
            else
            {
                module->active_request.error = I2C_BIT_BASH_REQUEST_ERROR_INVALID_ADDRESS;
                module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_STOP;
            }
            
            break;
        }

        case I2C_BIT_BASH_REQUEST_STATE_READ_DATA:
        {
            i2c_bit_bash_read_byte(module, 
                (uint8_t*)(module->active_request.read_buffer + module->active_request.data_pos));
            
            module->active_request.data_pos++;
            
            if (module->active_request.data_pos >= module->active_request.read_size)
            {
                // Transfer finished
                // Send an NACK from the Master to the slave
                i2c_bit_bash_send_nack(module);

                module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_STOP;
            }
            else
            {
                // Send an ACK to the slave
                i2c_bit_bash_send_ack(module);
            }
            break;
        }
        
        case I2C_BIT_BASH_REQUEST_STATE_SEND_DATA:
        {
            // Send Data
            i2c_bit_bash_send_byte(module, 
                    (uint8_t*)(module->active_request.write_buffer + module->active_request.data_pos));

            // Wait for ACK
            if (!i2c_bit_bash_get_response(module))
            {
                module->active_request.data_pos++;

                // Delay between command lines
                timers_delay_us(I2C_BIT_BASH_INTER_BYTE_PAUSE);

                if (module->active_request.data_pos >= module->active_request.write_size)
                {
                    switch (module->active_request.type)
                    {
                        case I2C_BIT_BASH_REQUEST_TYPE_RECEIVE:
                        {
                            break;
                        }
                        
                        case I2C_BIT_BASH_REQUEST_TYPE_TRANSMIT:
                        {
                            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_STOP;
                            break;
                        }
                        
                        case I2C_BIT_BASH_REQUEST_TYPE_TRANSMIT_AND_RECEIVE:
                        {
                            if (module->active_request.processing_pos == 1)
                            {
                                // Set SCL High then restart
                                PLIB_PORTS_PinSet(module->scl_pin.port, 
                                                  module->scl_pin.channel, 
                                                  module->scl_pin.pin);

                                module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_START;

                                // Increment the address so that it becomes the read address
                                module->active_request.device_address++;
                                
                                module->active_request.processing_pos++;
                            }
                            else
                            {
                                module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_READ_DATA;
                            }

                            break;
                        }
                    }
                }
            }
            else
            {
                module->active_request.error = I2C_BIT_BASH_REQUEST_ERROR_NACK;
                module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_SEND_STOP;
            }
            
            break;
        }
        
        case I2C_BIT_BASH_REQUEST_STATE_SEND_STOP:
        {
            i2c_bit_bash_send_stop(module);
            
            if (module->active_request.error == I2C_BIT_BASH_REQUEST_ERROR_NONE)
            {
                // Put back to waiting state
                module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_COMPLETE;
            }
            else
            {
                module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_ERROR;
            }
            break;
        }

        case I2C_BIT_BASH_REQUEST_STATE_COMPLETE:
        {
            //logger_log_debugf(LOGGER_CODE_I2C, "I2C BIT BASH Complete ");
            
            // Note: This needs to be before the callback if we want to initiate other calls from the callback
            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_WAITING;
            
            if (module->event_handler != NULL)
            {
                module->event_handler(DRV_I2C_BUFFER_EVENT_COMPLETE,
                                      (DRV_I2C_BUFFER_HANDLE)&module->active_request,
                                      (uintptr_t)module->active_request.callback_context);
            }
            break;
        }
        
        case I2C_BIT_BASH_REQUEST_STATE_ERROR:
        {
            logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Error %d ", module->active_request.error);
           
            // Note: This needs to be before the callback if we want to initiate other calls from the callback
            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_WAITING;
            
            if (module->event_handler != NULL)
            {
                module->event_handler(DRV_I2C_BUFFER_EVENT_ERROR,
                                      (DRV_I2C_BUFFER_HANDLE)&module->active_request,
                                      (uintptr_t)module->active_request.callback_context);
            }

            break;
        }
        
        default:
        {
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Reads a single byte from the slave
 * @param module
 * @param data
 */
static void i2c_bit_bash_read_byte(I2C_BIT_BASH_RUNTIME_MODULE* module, uint8_t* data)
{
    int pos = 0;
    
    // Makes sure data is empty
    *data = 0;
    
    for (pos = 7; pos >= 0; pos--)
    {
        uint8_t value = (uint8_t)i2c_bit_bash_send_clock_pulse(module);
        
        *data |= (value << pos);
    }
}

/**-----------------------------------------------------------------------------
 * Sends an ACK from the master to the slave
 * @param module
 */
static void i2c_bit_bash_send_ack(I2C_BIT_BASH_RUNTIME_MODULE* module)
{
    // Note - We do not need to set SDA to output before pulling it low.  If
    // we do it stops it going high again (presumably because the Slave holds the line 
    // down?)
    PLIB_PORTS_PinDirectionOutputSet(module->sda_pin.port, 
                                     module->sda_pin.channel, 
                                     module->sda_pin.pin);
            
    // Set SDA Low
    PLIB_PORTS_PinClear(module->sda_pin.port, 
                        module->sda_pin.channel, 
                        module->sda_pin.pin);     // Pull SDA low

    // Clock pulse
    i2c_bit_bash_send_clock_pulse(module);
    
    timers_delay_us(I2C_BIT_BASH_CLOCK_PULSE);

    // Release SDA
    PLIB_PORTS_PinDirectionInputSet(module->sda_pin.port, 
                                    module->sda_pin.channel, 
                                    module->sda_pin.pin);
}

/**-----------------------------------------------------------------------------
 * Generates a single clock pulse
 * Return s the state of the SDA pin during the pulse
 */
static bool i2c_bit_bash_send_clock_pulse(I2C_BIT_BASH_RUNTIME_MODULE* module)
{
    bool result = false;
    
    timers_delay_us(I2C_BIT_BASH_CLOCK_PULSE);
            
    // set clock high
    PLIB_PORTS_PinSet(module->scl_pin.port, 
                      module->scl_pin.channel, 
                      module->scl_pin.pin);

    timers_delay_100ns(15);
    
    // Read the state of the SDA pin
    result = (bool) PLIB_PORTS_PinGet(module->sda_pin.port, 
                                     module->sda_pin.channel, 
                                     module->sda_pin.pin);

    // set clock low
    PLIB_PORTS_PinClear(module->scl_pin.port, 
                        module->scl_pin.channel, 
                        module->scl_pin.pin);

    timers_delay_us(I2C_BIT_BASH_CLOCK_PULSE);   
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Sends a single byte of data from MAster to slave
 * @param module
 * @param data_ptr  - a pointer to the data to be sent
 * @return true if the slave ACKs
 */
static void i2c_bit_bash_send_byte(I2C_BIT_BASH_RUNTIME_MODULE* module, uint8_t* data)
{
    //logger_log_debugf(LOGGER_CODE_I2C, "I2C BIT BASH Byte %d 0X%02X ", module->index, *data);
    
    int pos = 0;
    for (pos = 7; pos >= 0; pos--)
    {
        if (((*data >> pos) & 1) == 1)
        {
            // Set SDA to input to pull it high
            PLIB_PORTS_PinDirectionInputSet(module->sda_pin.port, 
                                            module->sda_pin.channel, 
                                            module->sda_pin.pin);
        }
        else
        {
            // Set SDA to output
            PLIB_PORTS_PinDirectionOutputSet(module->sda_pin.port, 
                                             module->sda_pin.channel, 
                                             module->sda_pin.pin);
            
            // Set SDA Low
            PLIB_PORTS_PinClear(module->sda_pin.port, 
                                module->sda_pin.channel, 
                                module->sda_pin.pin);        
        }

        i2c_bit_bash_send_clock_pulse(module);
    }
}

/**-----------------------------------------------------------------------------
 * Sends an ACK from the master to the slave
 * @param module
 */
static void i2c_bit_bash_send_nack(I2C_BIT_BASH_RUNTIME_MODULE* module)
{
    // Set SDA to output
    PLIB_PORTS_PinDirectionInputSet(module->sda_pin.port, 
                                    module->sda_pin.channel, 
                                    module->sda_pin.pin);
            
    // Clock pulse
    i2c_bit_bash_send_clock_pulse(module);
}

/**-----------------------------------------------------------------------------
 * Generates a start condition
 */
static void i2c_bit_bash_send_start(I2C_BIT_BASH_RUNTIME_MODULE* module)
{
    // Set SDA to output
    PLIB_PORTS_PinDirectionOutputSet(module->sda_pin.port, 
                                     module->sda_pin.channel, 
                                     module->sda_pin.pin);

    // Set SDA Low
    PLIB_PORTS_PinClear(module->sda_pin.port, 
                        module->sda_pin.channel, 
                        module->sda_pin.pin);

    timers_delay_us(I2C_BIT_BASH_CLOCK_PULSE);

    // Set Clock Low
    PLIB_PORTS_PinClear(module->scl_pin.port, 
                        module->scl_pin.channel, 
                        module->scl_pin.pin);

    timers_delay_us(I2C_BIT_BASH_START_DELAY);
}

/**-----------------------------------------------------------------------------
 * Generates a end condition
 */
static void i2c_bit_bash_send_stop(I2C_BIT_BASH_RUNTIME_MODULE* module)
{
    // Delay before next transfer
    timers_delay_us(I2C_BIT_BASH_CLOCK_PULSE);

    // Set SDA to output 
    PLIB_PORTS_PinDirectionOutputSet(module->sda_pin.port, 
                                     module->sda_pin.channel, 
                                     module->sda_pin.pin);

    // Set clock high
    PLIB_PORTS_PinSet(module->scl_pin.port, 
                      module->scl_pin.channel, 
                      module->scl_pin.pin);

    // Delay before next transfer
    timers_delay_us(I2C_BIT_BASH_CLOCK_PULSE);

    // Set SDA to input
    PLIB_PORTS_PinDirectionInputSet(module->sda_pin.port, 
                                    module->sda_pin.channel, 
                                    module->sda_pin.pin);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Define a specific module
 * @param module
 */
void i2c_bit_bash_add_module(I2C_BIT_BASH_MODULE* module)
{
    if (module->index < I2C_MAX_MODULES)
    {
        // Copy the data into the I2C_BIT_BASH_RUNTIME_MODULE modules data.
        memcpy(&i2c_bit_bash_data.modules[module->index], module, sizeof(I2C_BIT_BASH_MODULE));
        
        logger_log_debugf(LOGGER_CODE_I2C, "I2C BIT BASH Added Module %d ", module->index);
    }
}

/**-----------------------------------------------------------------------------
 * Initialise the I2C bit bash system.  Must be called before i2c_bit_bash_add_module
 * (or not after at least)
 */
void i2c_bit_bash_initialise()
{
    int x = 0;
    for (x = 0; x < I2C_MAX_MODULES; x++)
    {
        logger_log_debugf(LOGGER_CODE_I2C, "I2C BIT BASH Initialising Module %d ", x);
        
        // Make the index invalid
        i2c_bit_bash_data.modules[x].index = I2C_MAX_MODULES;
        i2c_bit_bash_data.modules[x].active_request.state = I2C_BIT_BASH_REQUEST_STATE_WAITING;
        i2c_bit_bash_data.modules[x].event_handler = NULL;
    }
}
    
/**-----------------------------------------------------------------------------
 * Open a module
 * Note: DRV_IO_INTENT is ignored at the moment but in for compatibility with 
 * Harmony stuff
 * @return 
 */
DRV_HANDLE i2c_bit_bash_open(const SYS_MODULE_INDEX   index,
                             const DRV_IO_INTENT      intent)
{
    logger_log_debugf(LOGGER_CODE_I2C, "I2C BIT BASH Open - Module %d ", index);

    DRV_HANDLE result = DRV_HANDLE_INVALID;
            
    I2C_BIT_BASH_RUNTIME_MODULE* module = i2c_bit_bash_check_module(index);
    if (module != NULL)
    {
        // Note we use the index for the module handle
        result = module->index;

        // Ensure the SDA/SCL lines are ion the right state
        i2c_bit_bash_send_stop(module);
    }
     
    return result;
}   

/**-----------------------------------------------------------------------------
 * Processes I2C Bit Bash requests
 */
void i2c_bit_bash_process()
{
    int x;
    for (x = 0; x < I2C_MAX_MODULES; x++)
    {
        // Make the index invalid
        i2c_bit_bash_process_module(&i2c_bit_bash_data.modules[x]);
    }
}

/**-----------------------------------------------------------------------------
 * Reads bytes from the slave
 */
DRV_I2C_BUFFER_HANDLE i2c_bit_bash_receive(DRV_HANDLE handle,
                                           uint16_t device_address,
                                           void* read_buffer,
                                           size_t read_size,
                                           void* callback_context)
{
    //logger_log_debugf(LOGGER_CODE_I2C, "I2C BIT BASH Receive - Module %d ", handle);
    
    DRV_I2C_BUFFER_HANDLE result = DRV_I2C_BUFFER_HANDLE_INVALID;
    
    I2C_BIT_BASH_RUNTIME_MODULE* module = i2c_bit_bash_check_module(handle);
    if (module != NULL)
    {
        if (module->active_request.state == I2C_BIT_BASH_REQUEST_STATE_WAITING)
        {
            module->active_request.handle = handle;
            module->active_request.device_address = (uint8_t)device_address;
            module->active_request.read_buffer = read_buffer;
            module->active_request.read_size = read_size;
            module->active_request.callback_context = callback_context;
            module->active_request.type = I2C_BIT_BASH_REQUEST_TYPE_RECEIVE;
            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_PROCESS_REQUEST;
            
            result = (DRV_I2C_BUFFER_HANDLE)&module->active_request;
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Receive - Module %d in use state:%d ", module->index, module->active_request.state);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Receive - Invalid module ");
    }
    
    return result;
}
 
/**-----------------------------------------------------------------------------
 * Sets the event handler to be called when a message is processed
 * @param handle
 * @param eventHandler
 * @param context
 */
void i2c_bit_bash_set_event_handler(const DRV_HANDLE handle,
                                    const DRV_I2C_BUFFER_EVENT_HANDLER event_handler,
                                    const uintptr_t context)
{
    DRV_HANDLE result = DRV_HANDLE_INVALID;
            
    I2C_BIT_BASH_RUNTIME_MODULE* module = i2c_bit_bash_check_module(handle);
    if (module != NULL)
    {
        module->event_handler = event_handler;
    }    
}

/**-----------------------------------------------------------------------------
 * Transmit a block of data to an address
 */
DRV_I2C_BUFFER_HANDLE i2c_bit_bash_transmit(DRV_HANDLE handle,
                                            uint16_t device_address,
                                            void* write_buffer,
                                            size_t write_size,
                                            void* callback_context)
{
    //logger_log_debugf(LOGGER_CODE_I2C, "I2C BIT BASH Transmit - Module %d ", handle);
    
    DRV_I2C_BUFFER_HANDLE result = DRV_I2C_BUFFER_HANDLE_INVALID;
    
    I2C_BIT_BASH_RUNTIME_MODULE* module = i2c_bit_bash_check_module(handle);
    if (module != NULL)
    {
        if (module->active_request.state == I2C_BIT_BASH_REQUEST_STATE_WAITING)
        {
            module->active_request.handle = handle;
            module->active_request.device_address = (uint8_t)device_address;
            module->active_request.write_buffer = write_buffer;
            module->active_request.write_size = write_size;
            module->active_request.callback_context = callback_context;
            module->active_request.type = I2C_BIT_BASH_REQUEST_TYPE_TRANSMIT;
            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_PROCESS_REQUEST;
            
            result = (DRV_I2C_BUFFER_HANDLE)&module->active_request;
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Transmit - Module %d in use state:%d ", module->index, module->active_request.state);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Transmit - Invalid module ");
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Transmit a block of data to an address then reads bytes from the slave
 */
DRV_I2C_BUFFER_HANDLE i2c_bit_bash_transmit_receive(DRV_HANDLE handle,
                                            uint16_t device_address,
                                            void* write_buffer,
                                            size_t write_size,
                                            void* read_buffer,
                                            size_t read_size,
                                            void* callback_context)
{
    //logger_log_debugf(LOGGER_CODE_I2C, "I2C BIT BASH Transmit/Receive - Module %d ", handle);
    
    DRV_I2C_BUFFER_HANDLE result = DRV_I2C_BUFFER_HANDLE_INVALID;
    
    I2C_BIT_BASH_RUNTIME_MODULE* module = i2c_bit_bash_check_module(handle);
    if (module != NULL)
    {
        if (module->active_request.state == I2C_BIT_BASH_REQUEST_STATE_WAITING)
        {
            module->active_request.handle = handle;
            module->active_request.device_address = (uint8_t)device_address;
            module->active_request.read_buffer = read_buffer;
            module->active_request.read_size = read_size;
            module->active_request.write_buffer = write_buffer;
            module->active_request.write_size = write_size;
            module->active_request.callback_context = callback_context;
            module->active_request.type = I2C_BIT_BASH_REQUEST_TYPE_TRANSMIT_AND_RECEIVE;
            module->active_request.state = I2C_BIT_BASH_REQUEST_STATE_PROCESS_REQUEST;
            
            result = (DRV_I2C_BUFFER_HANDLE)&module->active_request;
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Transmit/Receive - Module %d in use state:%d ", module->index, module->active_request.state);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_I2C, "I2C BIT BASH Transmit/Receive - Invalid module ");
    }
    
    return result;
}

/* *****************************************************************************
 End of File
 */
