/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/image_cache.h"
#include "app_config.h"
#include "crypto/crypto.h"
#include "crypto/src/md5.h"
#include "utilities/3rd_party/tjpgd.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Holds the version of the meta data format -
// Version 2 upgrade type (incl major/minor version at start of metadata)
#define IMAGE_CACHE_METADATA_VERSION   0x02

//! Essentially contains th number of bytes that are i the image cache metadata before the number of images (0xFF + IMAGE_CACHE_METADATA_VERSION +  major_version + minor_version)
#define IMAGE_CACHE_METADATA_IMAGE_COUNT_OFFSET 4


//! The size of the buffer to hols upgrade data while it is being written to flash
#define IMAGE_CACHE_BUFFER_SIZE         8192
#define IMAGE_CACHE_IMAGE_BUFFER_SIZE   10000//16384    // Needs review - must be same size as largest image...which can be determined by looking at the logger messages from the cache during initialisation.
#define IMAGE_CACHE_DECODE_BUFFER_SIZE  4096

//! Default for the number of images that can be stored in the cache (<256))
#ifndef IMAGE_CACHE_NUM_IMAGES
#define IMAGE_CACHE_NUM_IMAGES          60
#endif

//! Define the maximum number of simultaneous image requests, and is the limit on
// the number of images than can be displayd on a single screen - Note this MUST be
// smaller than FLASH_MEMORY_REQUEST_QUEUE_SIZE
#define IMAGE_CACHE_MAX_IMAGE_REQUESTS  5       // Note this stopped ths display showing if set to 6?

/**
 * Structure to receive the image data 
 */
typedef struct 
{
    IMAGE_CACHE_CODED_IMAGE_HEADER header;
    
    uint8_t buffer[IMAGE_CACHE_IMAGE_BUFFER_SIZE]  __attribute__ ((aligned(2)));
    
} IMAGE_CACHE_REQUEST;


/**
 * Holds data abou tan individual image in the cache
 */
typedef struct
{
    // The identifier for the image
    uint8_t image_id;
    
    // The Offset of the image in the Images file
    uint32_t offset;
    
    //! The size of the image
    uint32_t size;
    
    //! The size of the image
    uint32_t height;
    
    //! The size of the image
    uint32_t width;
    
} IMAGE_METADATA;

typedef struct
{
    uint8_t upgade_version_indicator;
    
    uint8_t upgrade_version;
    
    uint8_t major_version;
    
    uint8_t minor_version;
    
    // The number of images in the cache
    uint8_t number_of_images;
    
    IMAGE_METADATA image_metadata[IMAGE_CACHE_NUM_IMAGES];
} IMAGE_CACHE_METADATA;

typedef struct
{
    uint16_t header_size;
    
    //! Callback invoked when the image cache is loaded
    IMAGE_CACHE_INITIALISED_CALLBACK initialised_callback;
    
    IMAGE_CACHE_METADATA cache_metadata;
    
   //! MD5 context structure used in MD5 checksum generation
    CRYPT_MD5_CTX md5_context;

    //! The number of bytes written in an upgrade
    uint32_t upgrade_bytes_written;
        
    //! An MD5 Checksum of the data
    uint8_t upgrade_checksum[MD5_DIGEST_SIZE];
    
    IMAGE_CACHE_UPGRADE_CALLBACK upgrade_callback;
    
    //! The total size of the upgrade
    size_t upgrade_size;
    
    uint16_t cache_buffer_write_pos;
    uint8_t cache_buffer[IMAGE_CACHE_BUFFER_SIZE];
    
    // Hold the left over byte from the last packet if it exists
    uint8_t upgrade_trailing_byte;
    bool upgrade_has_trailing_byte;
    
    IMAGE_CACHE_REQUEST image_request[IMAGE_CACHE_MAX_IMAGE_REQUESTS];

    // I tried using this in the IMAGE_CACHE_REQUEST struct but for some reason the pointers were getting mixed up when allocating from it, this seems to work...
    // Its a working buffer for the jpeg decode process.
    unsigned char decode_buffer[IMAGE_CACHE_DECODE_BUFFER_SIZE]  __attribute__ ((aligned(2)));		/* Working buffer */

} IMAGE_CACHE_DATA;

static IMAGE_CACHE_DATA image_cache_data = {
    .cache_metadata.upgade_version_indicator = 0XFF,
    .cache_metadata.upgrade_version = 0x01,
    .cache_metadata.major_version = 0,
    .cache_metadata.minor_version = 0,
    .cache_metadata.number_of_images = 0,
    //.image_request.ready_to_process =false,
    .initialised_callback = NULL,
    .upgrade_bytes_written = 0,
    .upgrade_has_trailing_byte = false,
};


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

static void image_cache_load_metadata(FLASH_MEMORY_CLIENT_REQUEST* request);

/**-----------------------------------------------------------------------------
 * 
 * @param data
 * @param value
 * @param high_byte
 */
static void image_cache_decode_upgrade_data_byte(uint8_t* data, uint8_t* value, bool high_byte)
{
    uint8_t converted_value;

    // Check id data is 0-9
    if ((*data >= 48) && (*data <= 57))
    {
        converted_value = *data - 48;
    }
    // Check for A-F
    else if ((*data >= 65) && (*data <= 70))
    {
        converted_value = *data - 55;
    }
    // Check for a-f
    else if ((*data >= 97) && (*data <= 102))
    {
        converted_value = *data - 87;
    }
    
    if (high_byte)
    {
        *value += converted_value << 4;
    }
    else
    {
        *value += converted_value;
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param image_id
 * @return 
 */
static IMAGE_METADATA* image_cache_find_metadata(uint8_t image_id)
{
    IMAGE_METADATA* result = NULL;
    
    int x;
    for (x = 0; x < image_cache_data.cache_metadata.number_of_images; x++)
    {
        if (image_cache_data.cache_metadata.image_metadata[x].image_id == image_id)
        {
            result = &image_cache_data.cache_metadata.image_metadata[x];
            break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param jd
 * @param buffer
 * @param bytes_to_read
 * @return 
 */
static IMAGE_CACHE_REQUEST* image_cache_get_request()
{
    IMAGE_CACHE_REQUEST* result = NULL;
    
    int x;
    for (x = 0; x < IMAGE_CACHE_MAX_IMAGE_REQUESTS; x++)
    {
        if (!image_cache_data.image_request[x].header.in_use)
        {
            image_cache_data.image_request[x].header.in_use = true;
            result = &image_cache_data.image_request[x];
            break;
        }
    }
    
    if (result == NULL)
    {
        logger_log_debugf(0, "IMACE CACHE Unable to allocate a cache request --");
    }
    
    return result;
}
/**-----------------------------------------------------------------------------
 * 
 * @param request
 */
static uint16_t image_cache_read_jpeg_input (
	JDEC* jd,               /* Decompressor object */
	uint8_t* buffer,        /* Pointer to the read buffer (NULL:skip) */
	uint16_t bytes_to_read	/* Number of bytes to read/skip from input stream */
)
{
    IMAGE_CACHE_CODED_IMAGE_HEADER* header = (IMAGE_CACHE_CODED_IMAGE_HEADER*)jd->device;
    
	uint16_t result = min(bytes_to_read, header->size - header->buffer_offset);
//    logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE jpeg_read_input  %d %d %d %d --", bytes_to_read, result, header->size, header->buffer_offset);
//    logger_process();

    if (buffer != NULL)
    {
        uint16_t x;
        for (x = 0; x < result; x++)
        {
            buffer[x] = header->buffer_ptr[header->buffer_offset + x];
        }
    }
    
    header->buffer_offset += result;
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @param jd
 * @param bitmap
 * @param rect
 * @return 
 */
uint16_t image_cache_write_jpeg_output (
	JDEC* jd,		/* Decompression object */
	void* bitmap,	/* Bitmap data to be output */
	JRECT* rect		/* Rectangular region to output */
)
{
    IMAGE_CACHE_CODED_IMAGE_HEADER* header = (IMAGE_CACHE_CODED_IMAGE_HEADER*)jd->device;
    TFT_CONTROL* tft_control = header->tft_control;
    
//    logger_log_debugf(0, "IMAGE CACHE %d %d %d %d --", rect->left, rect->top, rect->right, rect->bottom);
//    logger_log_debugf(0, "IMAGE CACHE Control %d %d %d %d --", tft_control->bounds.left, tft_control->bounds.top, tft_control->bounds.right, tft_control->bounds.bottom);
//    logger_process();

    uint16_t pixel_count = (rect->bottom - rect->top + 1) * (rect->right - rect->left + 1);
    
    PIXEL pixels[pixel_count];
    
    uint8_t* data = (uint8_t*)bitmap;
    
    int x;
    for (x = 0; x < pixel_count; x++)
    {
        pixels[x].red = *data++;//0x00;
        pixels[x].green = *data++;//0xA5;
        pixels[x].blue = *data++;//0xFF;
    }

    BOUNDS bounds;
    bounds.top = rect->top;
    bounds.left = rect->left;
    bounds.bottom = rect->bottom;
    bounds.right = rect->right;

    // Offset for the controls position
    bounds.top += tft_control->bounds.top;
    bounds.left += tft_control->bounds.left;
    bounds.bottom += tft_control->bounds.top;
    bounds.right += tft_control->bounds.left;

    // Offset for alignment
    switch(tft_control->control_data.image.alignment)
    {
        case TFT_CONTROL_ALIGNMENT_TOP_LEFT:
        {
            break;
        }
        
        case TFT_CONTROL_ALIGNMENT_CENTER:
        {
            uint16_t top_offset = (tft_display_bounds_height(&tft_control->bounds) - jd->height) / 2;
            uint16_t left_offset = (tft_display_bounds_width(&tft_control->bounds) - jd->width) / 2;

            bounds.top += top_offset;
            bounds.left += left_offset;
            bounds.bottom += top_offset;
            bounds.right += left_offset;

            //logger_log_debugf(0, "CENTER %d %d - %d %d - %d %d %d %d --", tft_display_bounds_height(&tft_control->bounds), jd->width, top_offset, left_offset, bounds.top, bounds.left, bounds.bottom, bounds.right);
            break;
        }
        
        default:
            break;
    }
    
//    if ((rect->top == 0) && (rect->left == 0))
//    {
//        tft_control_clear(tft_control, false);
//    }
    
    tft_display_write_buffer(&bounds, pixels, false);
    //logger_log_debugf(0, "IMAGE CACHE Control %d %d %d %d %d %d --", bounds.left, bounds.top, bounds.right, bounds.bottom, jd->width, jd->height);
    
    // This is drawn from top left to bottom right so check whether wh have the last bit
    if ((rect->right == jd->width - 1) && (rect->bottom == jd->height - 1))
    {
        tft_display_invalidate_rect(&tft_control->bounds);
    }
     
    return 1;	/* Continue to decompress */
}

/**-----------------------------------------------------------------------------
 * A Callback used for flash_memory functionality
 * @param page_data
 */
static void image_cache_flash_callback(FLASH_MEMORY_CLIENT_REQUEST* request)
{
    logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Flash Memory Callback client: %d ", request->client_id);
    logger_process();
    if (flash_memory_request_is_read(request))
    {
        if (request->offset == 0)
        {
            image_cache_load_metadata(request);
        }
        else
        {
            IMAGE_CACHE_REQUEST* image_cache_request = (IMAGE_CACHE_REQUEST*)request->additional_info;
            image_cache_request->header.ready_to_process = true;
            image_cache_request->header.size = request->bytes_read_written;
            image_cache_request->header.buffer_ptr = (uint8_t*)&image_cache_request->buffer;
           
            // This will be picked up by image_cache_process
            //image_cache_process_image(request);
        }
    }
    else
    {
        if (request->offset == 0)
        {
            // This should be from writing the major/minor version
        }
        else
        {
            // We have successfully written to the image cache - now request more data/end
            image_cache_data.upgrade_bytes_written += request->length;

            logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Bytes written %d at offset %d total %d (%d) --",
                    request->length,
                    request->offset,
                    image_cache_data.upgrade_bytes_written,
                    image_cache_data.upgrade_size);

            if (image_cache_data.upgrade_bytes_written >= image_cache_data.upgrade_size)
            {
                // If all bytes have been written truncate the file/check the checksum???
                // Complete and compare checksums and data_length?
                uint8_t checksum[MD5_DIGEST_SIZE];
                CRYPT_MD5_Finalize(&image_cache_data.md5_context, checksum);

                flash_memory_set_size(FLASH_MEMORY_CLIENT_IMAGES, 
                        0,
                        image_cache_data.upgrade_bytes_written);

                flash_memory_close(FLASH_MEMORY_CLIENT_IMAGES, 0);

                if (memcmp(image_cache_data.upgrade_checksum, checksum, MD5_DIGEST_SIZE) == 0)
                {
                    // Loaded OK
                }
                else
                {
                    logger_log_errorf(LOGGER_CODE_IMAGES, "IMAGE CACHE Checksum mismatch --");
                    logger_process();
                }

                // Output upload statistics and checksums
                logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Bytes written %d file size % d ",
                        image_cache_data.upgrade_bytes_written,
                        flash_memory_size(FLASH_MEMORY_CLIENT_IMAGES, 0));

                logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE UPGRADE Checksum - Orig %02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x --", 
                    image_cache_data.upgrade_checksum[0],
                    image_cache_data.upgrade_checksum[1],
                    image_cache_data.upgrade_checksum[2],
                    image_cache_data.upgrade_checksum[3],
                    image_cache_data.upgrade_checksum[4],
                    image_cache_data.upgrade_checksum[5],
                    image_cache_data.upgrade_checksum[6],
                    image_cache_data.upgrade_checksum[7],
                    image_cache_data.upgrade_checksum[8],
                    image_cache_data.upgrade_checksum[9],
                    image_cache_data.upgrade_checksum[10],
                    image_cache_data.upgrade_checksum[11],
                    image_cache_data.upgrade_checksum[12],
                    image_cache_data.upgrade_checksum[13],
                    image_cache_data.upgrade_checksum[14],
                    image_cache_data.upgrade_checksum[15]
                    );

                logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE UPGRADE Checksum - Tran %02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x --", 
                    checksum[0],
                    checksum[1],
                    checksum[2],
                    checksum[3],
                    checksum[4],
                    checksum[5],
                    checksum[6],
                    checksum[7],
                    checksum[8],
                    checksum[9],
                    checksum[10],
                    checksum[11],
                    checksum[12],
                    checksum[13],
                    checksum[14],
                    checksum[15]
                    );
                logger_process();  
            }

            if (image_cache_data.upgrade_callback != NULL)
            {
                image_cache_data.upgrade_callback(image_cache_data.upgrade_size, image_cache_data.upgrade_bytes_written);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param request
 */
static uint16_t image_cache_get_header_size()
{
    return image_cache_data.header_size;
//    return 1 + (image_cache_data.cache_metadata.number_of_images * (sizeof(uint8_t) + sizeof(uint32_t) + sizeof(uint32_t) + sizeof(uint32_t) + sizeof(uint32_t)));
}

/**-----------------------------------------------------------------------------
 * 
 * @param request
 */
static void image_cache_load_metadata(FLASH_MEMORY_CLIENT_REQUEST* request)
{
    uint8_t* data = (uint8_t*)request->data;
    uint8_t header_offset = 1;
    
    if (data[0] == 0xFF)
    {
        if (data[1] == IMAGE_CACHE_METADATA_VERSION)
        {
            // Version 0x02 handling
            image_cache_data.cache_metadata.upgrade_version = data[1];
            image_cache_data.cache_metadata.major_version = data[2];
            image_cache_data.cache_metadata.minor_version = data[3];
            image_cache_data.cache_metadata.number_of_images = data[4];
            header_offset += IMAGE_CACHE_METADATA_IMAGE_COUNT_OFFSET;
        }
        else
        {
            // We would put other version handling here e.g. 0x03....
        }
    }
    else
    {
        image_cache_data.cache_metadata.major_version = IMAGE_VERSION_MAJOR;
        image_cache_data.cache_metadata.minor_version = IMAGE_VERSION_MINOR;
        image_cache_data.cache_metadata.number_of_images = data[0];
    }

    // We are reading the metadata from the cache_buffer because the upgrade 
    // has all the fields packed, whereas the IMAGE_CACHE_METADATA is a stuct 
    // that includes padding - https://stackoverflow.com/questions/119123/why-isnt-sizeof-for-a-struct-equal-to-the-sum-of-sizeof-of-each-member
    logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE %d/%d Number of Images %d --", 
            image_cache_data.cache_metadata.major_version,
            image_cache_data.cache_metadata.minor_version,
            image_cache_data.cache_metadata.number_of_images);
    int x;
    for (x = 0; x < image_cache_data.cache_metadata.number_of_images; x++)
    {
        uint32_t image_offset = header_offset + (x * (sizeof(uint8_t) + sizeof(uint32_t) + sizeof(uint32_t) + sizeof(uint32_t) + sizeof(uint32_t)));

        image_cache_data.cache_metadata.image_metadata[x].image_id = data[image_offset];
                
//                logger_log_debugf(0, "IMAGE CACHE %X %X %X %X %X %X %X %X %X --",
//                        data[image_offset],
//                        data[image_offset + sizeof(uint8_t)],
//                        data[image_offset + sizeof(uint8_t) + 1],
//                        data[image_offset + sizeof(uint8_t) + 2],
//                        data[image_offset + sizeof(uint8_t) + 3],
//                        data[image_offset + sizeof(uint8_t) + 4],
//                        data[image_offset + sizeof(uint8_t) + 5],
//                        data[image_offset + sizeof(uint8_t) + 6],
//                        data[image_offset + sizeof(uint8_t) + 7]
//                        );
                
        memcpy(&image_cache_data.cache_metadata.image_metadata[x].size, &data[image_offset + sizeof(uint8_t)], sizeof(uint32_t));
        memcpy(&image_cache_data.cache_metadata.image_metadata[x].offset, &data[image_offset + sizeof(uint8_t) + sizeof(uint32_t)], sizeof(uint32_t));
        memcpy(&image_cache_data.cache_metadata.image_metadata[x].height, &data[image_offset + sizeof(uint8_t) + sizeof(uint32_t) + sizeof(uint32_t)], sizeof(uint32_t));
        memcpy(&image_cache_data.cache_metadata.image_metadata[x].width, &data[image_offset + sizeof(uint8_t) + sizeof(uint32_t) + sizeof(uint32_t) + sizeof(uint32_t)], sizeof(uint32_t));

        logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Image %X %d %d %d %d --", 
                image_cache_data.cache_metadata.image_metadata[x].image_id, 
                image_cache_data.cache_metadata.image_metadata[x].offset,
                image_cache_data.cache_metadata.image_metadata[x].size,
                image_cache_data.cache_metadata.image_metadata[x].height,
                image_cache_data.cache_metadata.image_metadata[x].width);
    }

    image_cache_data.header_size = header_offset + 
                                   (image_cache_data.cache_metadata.number_of_images * (sizeof(uint8_t) + sizeof(uint32_t) + sizeof(uint32_t) + sizeof(uint32_t) + sizeof(uint32_t)));
    
    // Invoke the initialised callback
    if (image_cache_data.initialised_callback != NULL)
    {
        image_cache_data.initialised_callback();
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the height of the supplied image (0 if not found)
 * @param image_id
 * @return 
 */
uint32_t image_cache_get_image_height(uint8_t image_id)
{
    uint32_t result = 0;
    
    IMAGE_METADATA* metadata = image_cache_find_metadata(image_id);
    if (metadata != NULL)
    {
        result = metadata->height;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the width of the supplied image (0 if not found)
 * @param image_id
 * @return 
 */
uint32_t image_cache_get_image_width(uint8_t image_id)
{
    uint32_t result = 0;
    
    IMAGE_METADATA* metadata = image_cache_find_metadata(image_id);
    if (metadata != NULL)
    {
        result = metadata->width;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the major version number
 */
uint8_t image_cache_get_major_version()
{
    return image_cache_data.cache_metadata.major_version;
}

/**-----------------------------------------------------------------------------
 * Gets the minor version number
 */
uint8_t image_cache_get_minor_version()
{
    return image_cache_data.cache_metadata.minor_version;
}

/**-----------------------------------------------------------------------------
 * Can be used to check whether the cache has the requested image
 * @param image_id
 * @return 
 */
bool image_cache_has_image(uint8_t image_id)
{
    bool result = false;
    
    IMAGE_METADATA* metadata = image_cache_find_metadata(image_id);
    if (metadata != NULL)
    {
        result = true;
    }
        
    return result;
}

/**-----------------------------------------------------------------------------
 * Initialises the Image cache module
 */
void image_cache_initialise(IMAGE_CACHE_INITIALISED_CALLBACK initialised_callback)
{
    logger_log_infof(LOGGER_CODE_IMAGES, "IMAGE CACHE - Initialising --");
    flash_memory_add_client(FLASH_MEMORY_CLIENT_IMAGES, IMAGES_STORAGE_NAME, &image_cache_flash_callback);
    
    image_cache_data.initialised_callback = initialised_callback;
    
    int x;
    for (x = 0; x < IMAGE_CACHE_MAX_IMAGE_REQUESTS; x++)
    {
        image_cache_data.image_request[x].header.in_use = false;
    }
    
    // Read the configuration index data
    
    // Get the image file and append data to it......
    // Save the image data
    FLASH_MEMORY_CLIENT_REQUEST request;
    request.client_id = FLASH_MEMORY_CLIENT_IMAGES;
    request.file_index = 0;
    request.data = (void*)&image_cache_data.cache_buffer; // Use the cache_buffer rather than the actual metadata as there are padding issues
    request.offset = 0;
    request.length = sizeof(IMAGE_CACHE_METADATA);

    flash_memory_read(&request);
}

/**-----------------------------------------------------------------------------
 * prepares the image cache to receive upgrade data
 */
void image_cache_prepare_upgrade(uint8_t major_version, uint8_t minor_version, size_t size, uint8_t* checksum)
{
    logger_log_info(LOGGER_CODE_IMAGES, "IMAGE CACHE - Preparing Upgrade --");
        
    image_cache_data.upgrade_bytes_written = 0;
    image_cache_data.upgrade_size = size / 2; // The actual upgrade size will be half the size supplied when decoded
    
    image_cache_data.upgrade_has_trailing_byte = false;
        
    // Store the checksum
    memcpy(&image_cache_data.upgrade_checksum, checksum, MD5_DIGEST_SIZE);
    
    CRYPT_MD5_Initialize(&image_cache_data.md5_context);

    flash_memory_open(FLASH_MEMORY_CLIENT_IMAGES, 0);
    
    // Write the Major/Minor version into the Image file header
    if (image_cache_data.cache_buffer_write_pos + IMAGE_CACHE_METADATA_IMAGE_COUNT_OFFSET > IMAGE_CACHE_BUFFER_SIZE)
    {
        image_cache_data.cache_buffer_write_pos = 0;
    }
    image_cache_data.cache_buffer[image_cache_data.cache_buffer_write_pos] = 0XFF;          // Indicates that this now stores the upgrade version
    image_cache_data.cache_buffer[image_cache_data.cache_buffer_write_pos + 1] = IMAGE_CACHE_METADATA_VERSION;      // Version 2 upgrade type (incl major/minor version at stat of metadata)
    image_cache_data.cache_buffer[image_cache_data.cache_buffer_write_pos + 2] = major_version;
    image_cache_data.cache_buffer[image_cache_data.cache_buffer_write_pos + 3] = minor_version;
    
    // Get the image file and append data to it......
    // Save the image data
    FLASH_MEMORY_CLIENT_REQUEST request;
    request.client_id = FLASH_MEMORY_CLIENT_IMAGES;
    request.file_index = 0;
    request.data = (void*)&image_cache_data.cache_buffer[image_cache_data.cache_buffer_write_pos];
    request.offset = 0;
    request.length = IMAGE_CACHE_METADATA_IMAGE_COUNT_OFFSET;

    logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Writing Version Info %d %d %d --", image_cache_data.cache_buffer_write_pos, major_version, minor_version);

    // Update circular buffer write position
    image_cache_data.cache_buffer_write_pos += IMAGE_CACHE_METADATA_IMAGE_COUNT_OFFSET;

    flash_memory_write(&request);

}
    
/**-----------------------------------------------------------------------------
 */
void image_cache_process()
{
    int x;
    for (x = 0; x < IMAGE_CACHE_MAX_IMAGE_REQUESTS; x++)
    {
        // Might need to do it so that we only process one at a time
        if (image_cache_data.image_request[x].header.ready_to_process)
        {
            image_cache_data.image_request[x].header.ready_to_process = false;
            image_cache_process_image(&image_cache_data.image_request[x].header);
        }
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param request
 */
bool image_cache_process_image(IMAGE_CACHE_CODED_IMAGE_HEADER* coded_image_header)
{
    bool result = false;
    
    JDEC jd;
    JRESULT jd_result;
    
    logger_log_debugf(0, "IMAGE CACHE Processing Image %d --", coded_image_header->buffer_ptr);
    //logger_log_debugf(0, "IMAGE CACHE %X %X %X %X --", image_cache_data.decode_buffer, image_cache_request->buffer[822], image_cache_request->buffer[823], image_cache_request->buffer[824], image_cache_request->buffer[825]);
    logger_process();        
    
    // reset the buffer offset
    coded_image_header->buffer_offset = 0;

    if (coded_image_header->tft_control->visible)
    {
        jd_result = jd_prepare(&jd, image_cache_read_jpeg_input, image_cache_data.decode_buffer, IMAGE_CACHE_DECODE_BUFFER_SIZE, coded_image_header);
        if (jd_result == JDR_OK) 
        {
            logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Prepared w:%d h:%d --", jd.width, jd.height);
            logger_process();   
             
            // Start to decompress the JPEG file
            // Note Scale = 0 as we are not doing any scaling
            jd_result = jd_decomp(&jd, image_cache_write_jpeg_output, 0);	/* Start decompression */
            if (jd_result == JDR_OK)
            {
                result = true;
            }
            else
            {
                logger_log_errorf(LOGGER_CODE_IMAGES, "IMAGE CACHE Decode FAILED %d --", result);
                logger_process();
            }
        } 
        else 
        {
            logger_log_errorf(LOGGER_CODE_IMAGES, "IMAGE CACHE Prepare FAILED %d --", result);
            logger_process();
        }
    }
    else
    {
        tft_control_clear(coded_image_header->tft_control, false);
    }

    coded_image_header->in_use = false;
        
    coded_image_header->tft_control->control_data.image.being_retrieved = false;
    
    return result;
}
   
/**-----------------------------------------------------------------------------
 * 
 * @param image_id
 */
void image_cache_retrieve_image(uint8_t image_id,
                                TFT_CONTROL* tft_control)
{
    // get the location of the image
    IMAGE_METADATA* metadata = image_cache_find_metadata(image_id);
    
    if (metadata != NULL)
    {
        logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Retrieving image 0X%X %d %d --", image_id, metadata->offset, metadata->size);
        
        //TODO probably need more that one of these for simultaneous caches requests - e.g multiple images
        IMAGE_CACHE_REQUEST* image_request = image_cache_get_request(); //&image_cache_data.image_request[0];
        if (image_request != NULL)
        {
            image_request->header.buffer_offset = 0; // Initialise the buffer offset
            image_request->header.ready_to_process = false;
            image_request->header.tft_control = tft_control;

            FLASH_MEMORY_CLIENT_REQUEST request;
            request.client_id = FLASH_MEMORY_CLIENT_IMAGES;
            request.file_index = 0;
            request.data = (void*)image_request->buffer;
            request.offset = image_cache_get_header_size() + metadata->offset;
            request.length = metadata->size;
            request.additional_info = image_request;

            flash_memory_read(&request);
        }
    }
    else
    {
        logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Unable to locate image 0X%X --", image_id);
        tft_control->control_data.image.being_retrieved = false;
    }
}

/**-----------------------------------------------------------------------------
 * Takes upgrade data, decodes it and puts it into the image cache
 */
void image_cache_unpack_upgrade_data(uint8_t* data, 
        size_t data_length,
        IMAGE_CACHE_UPGRADE_CALLBACK upgrade_callback)
{
    // Write to flash
    image_cache_data.upgrade_callback = upgrade_callback;

    logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Upload Data %d --", data_length);
    
    // Update the checksum - note this is done on the un-decoded data
    CRYPT_MD5_DataAdd(&image_cache_data.md5_context, (char*)data, data_length);

    // Set the offset and state
    uint32_t offset = IMAGE_CACHE_METADATA_IMAGE_COUNT_OFFSET + image_cache_data.upgrade_bytes_written;
        
    // Copy the data into the write_buffer
    // If we look like were going to go past the end of the buffer then reset to the beginning
    // This means that the buffer must be sufficiently large so that there
    // will be space at the beginning when we get to the end.
    if (image_cache_data.cache_buffer_write_pos + data_length > IMAGE_CACHE_BUFFER_SIZE)
    {
        image_cache_data.cache_buffer_write_pos = 0;
    }

    // Decode the data into the cache_buffer
    // Note the encoded data is a hex representation of the actual jpeg data, hence 
    // it occupies 2bytes for a single byte of data
    uint32_t decode_offset = 0;
    uint8_t decoded_value = 0x00;
    size_t loop_start = 0;
    uint32_t processed_count = 0;
      
    // First check if there is a byte trailing from the last packet of data
    if (image_cache_data.upgrade_has_trailing_byte)
    {
        logger_log_debugf(0, "IMAGE CACHE Processing trailing byte --");
        
        // add the trailing byte conversion to the decode_value
        image_cache_decode_upgrade_data_byte(&image_cache_data.upgrade_trailing_byte, &decoded_value, true);
        // Take first byte from data and add
        image_cache_decode_upgrade_data_byte(&data[0], &decoded_value, false);
        
        image_cache_data.cache_buffer[image_cache_data.cache_buffer_write_pos] = decoded_value;
        
        loop_start = 1;
        processed_count++;
    }
    
    size_t x;
    for (x = loop_start; x < data_length; x += 2)
    {
        decoded_value = 0x00;
        
        if (x < (data_length - 1))
        {
            image_cache_decode_upgrade_data_byte(&data[x], &decoded_value, true);
            image_cache_decode_upgrade_data_byte(&data[x + 1], &decoded_value, false);
        
            image_cache_data.cache_buffer[image_cache_data.cache_buffer_write_pos + processed_count] = decoded_value;
            
            if (x < 5)
            {
                logger_log_debugf(0, "IMAGE CACHE decode %d %d %d --", data[x], data[x+1], decoded_value);
            }
            
            processed_count++;
        }
        else
        {
            logger_log_debugf(0, "IMAGE CACHE Upgrade trailing byte found --");
            image_cache_data.upgrade_has_trailing_byte = true;
            image_cache_data.upgrade_trailing_byte = data[x];
            break;
        }
    }

    // Get the image file and append data to it......
    // Save the image data
    FLASH_MEMORY_CLIENT_REQUEST request;
    request.client_id = FLASH_MEMORY_CLIENT_IMAGES;
    request.file_index = 0;
    request.data = (void*)&image_cache_data.cache_buffer[image_cache_data.cache_buffer_write_pos];
    request.offset = offset; // +2 is the major/minor version info
    request.length = processed_count;

    logger_log_debugf(LOGGER_CODE_IMAGES, "IMAGE CACHE Upload Data %d %d %d --", image_cache_data.cache_buffer_write_pos, offset, processed_count);

    // Update circular buffer write position
    image_cache_data.cache_buffer_write_pos += processed_count;

    flash_memory_write(&request);
}


/* *****************************************************************************
 End of File
 */
