/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/input_adc/input_adc.h"
#include "app_config.h"
#include "utilities/events.h"
#include "utilities/input_adc/private/input_adc_data.h"
#include "utilities/input_adc/input_adc_defs.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */


/**-----------------------------------------------------------------------------
 * Event raised when there is an Sample Rate change detected at the Hardware level
 */
static void input_adc_settings_applied_callback(EVENT* event)
{
    INPUT_ADC_SETTINGS_APPLIED_EVENT_DATA* event_data = (INPUT_ADC_SETTINGS_APPLIED_EVENT_DATA*)event->context_data;

    logger_log_debugf(0, "INPUT ADC SETTINGS APPLIED CALLBACK %d --", event_data->applied);            
    logger_process();
    
    input_adc_data_set_configured(event_data->applied);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Configures the SPDIF RX module.  Note this includes a callback to the 
 * application to define which hardware implementation will be used
 */
void input_adc_configure()
{
    // Invoke the hardware layer configuration method
    INPUT_ADC_DATA* data = input_adc_data_get();
    if (data->functions.configure_function != NULL)
    {
        data->functions.configure_function();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_INPUT_ADC, "INPUT ADC Configuration function not specified --");
    }    
}

/**-----------------------------------------------------------------------------
 * Used to define the functions for the particular adc.  Called in app_initialise
 * and is passed the hw version of define functions
 * @param function_data
 */
void input_adc_define_functions(INPUT_ADC_FUNCTION_DATA* function_data)
{
    INPUT_ADC_DATA* data = input_adc_data_get();
    
    // Copy the function pointers to the data module
    memcpy(&(data->functions), function_data, sizeof(INPUT_ADC_FUNCTION_DATA));    
}

/**-----------------------------------------------------------------------------
 * Performs cyclic INPUT ADC processing
 */
bool input_adc_initialise()
{
    bool result = false;
    
    // Invoke the hardware layer configuration method
    INPUT_ADC_DATA* data = input_adc_data_get();
    
    if (data->functions.initialise_function != NULL)
    {
        result = data->functions.initialise_function();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_SPDIF_RX, "INPUT_ADC Initialise callback not specified --");
    }    
    
    // Subscribe to the events raise in the HW layer.  Note the Event Module ID should
    // work for either HW type 
	events_subscribe(EVENT_MODULE_INPUT_ADC_HW, 
                     INPUT_ADC_EVENT_SETTINGS_APPLIED, 
					 input_adc_settings_applied_callback, 
					 (void*)NULL, 
					 true);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Set to true when all settings are applied
 */
bool input_adc_is_configured()
{
    return input_adc_data_is_configured();
}

/**-----------------------------------------------------------------------------
 * Used to reapply the ADC settings
 */
void input_adc_reapply_settings()
{
    // Invoke the hardware layer configuration method
    INPUT_ADC_DATA* data = input_adc_data_get();
    if (data->functions.reapply_settings_function != NULL)
    {
        data->functions.reapply_settings_function();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_INPUT_ADC, "INPUT ADC Reapply Settings function not specified --");
    }        
}

/**-----------------------------------------------------------------------------
 * Sets pins for the adc
 * @param on
 */
void input_adc_reset(bool on)
{
    // Invoke the hardware layer configuration method
    INPUT_ADC_DATA* data = input_adc_data_get();
    if (data->functions.reset_function != NULL)
    {
        data->functions.reset_function(on);
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_INPUT_ADC, "INPUT ADC Reset callback not specified --");
    }        
}

/* *****************************************************************************
 End of File
 */
