/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/input_adc/input_adc_hw_ak5552.h"
#include "app_config.h"
#include "utilities/events.h"
#include "utilities/i2c.h"
#include "utilities/input_control_i2c.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define INPUT_ADC_AK5552_NUM_SETTINGS               8

//! Defines the maximum number o event subscriptions that can be made to the events 
// in this module - this is set to 3 one for each event as its only the input adc module 
// accesses them
#define INPUT_ADC_AK5552_HW_MAX_EVENT_SUBSCRIPIONS  2

typedef struct
{
    //! I2C Client Id for identifying the module on the I2C bus
    I2C_CLIENT_ID i2c_client_id;

    //! Register setting definitions
    INPUT_CONTROL_I2C_REQUEST settings[INPUT_ADC_AK5552_NUM_SETTINGS];
    
    INPUT_ADC_FUNCTION_DATA functions;
    
    //! used to monitor the number of settings that have been applied in apply_settings
    uint8_t active_setting_count;
    
    //! Indicates whether the settings have been applied
    bool settings_applied;
    
	//! An Array of event subscription structures.  These ara managed by the events module when
	// invoking events.  There MUST be enough declared to handle all the concurrent event 
	// subscriptions for the module.  
    // In this case it will be 1 for each event as its only th spdif_rx module subscribing
	EVENT_SUBSCRIPTION event_subscriptions[INPUT_ADC_AK5552_HW_MAX_EVENT_SUBSCRIPIONS];    
    
} INPUT_ADC_AK5552_DATA;

static INPUT_ADC_AK5552_DATA input_adc_data = {
    .active_setting_count = 0,
    .settings_applied = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Applies the Register settings
 */
static void input_adc_hw_ak5552_apply_settings()
{
    if (!input_adc_data.settings_applied)
    {
        logger_log_info(LOGGER_CODE_INITIALISATION, "INPUT ADC AK5552 Applying Settings --");
        logger_process();
        
        INPUT_ADC_SETTINGS_APPLIED_EVENT_DATA event_data;
        event_data.applied = false;
        events_raise(EVENT_MODULE_INPUT_ADC_HW, INPUT_ADC_EVENT_SETTINGS_APPLIED, (void*)&event_data);
        
        input_adc_data.active_setting_count += INPUT_ADC_AK5552_NUM_SETTINGS;
        
        input_control_i2c_apply_settings(input_adc_data.i2c_client_id, input_adc_data.settings, INPUT_ADC_AK5552_NUM_SETTINGS);
    }
}

/**-----------------------------------------------------------------------------
 * Called when a setting has been applied
 * @param request
 */
static void input_adc_hw_ak5552_setting_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    input_adc_data.active_setting_count--;
    
    logger_log_infof(LOGGER_CODE_INPUT_ADC, "INPUT ADC AK5552 Settings applied %d %d %d --", 
            request->address,
            request->value,
            input_adc_data.active_setting_count);

    if (input_adc_data.active_setting_count == 0)
    {
        logger_log_infof(LOGGER_CODE_INPUT_ADC, "INPUT ADC AK5552 All Settings applied --");
        logger_process();
        
        input_adc_data.settings_applied = true;
       
        // Raise the event
        INPUT_ADC_SETTINGS_APPLIED_EVENT_DATA event_data;
        event_data.applied = true;
        events_raise(EVENT_MODULE_INPUT_ADC_HW, INPUT_ADC_EVENT_SETTINGS_APPLIED, (void*)&event_data);
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Configure the ak5552
 */
void input_adc_hw_ak5552_configure()
{
    logger_log_info(LOGGER_CODE_INPUT_ADC, "INPUT ADC AK5552 Configuration --");
    logger_process();

    input_control_i2c_add_setting(&input_adc_data.settings[0], 0X00, 0xFF, (void*)input_adc_hw_ak5552_setting_callback);
    input_control_i2c_add_setting(&input_adc_data.settings[1], 0X01, 0x01, (void*)input_adc_hw_ak5552_setting_callback);  
    input_control_i2c_add_setting(&input_adc_data.settings[2], 0X02, 0x03, (void*)input_adc_hw_ak5552_setting_callback);  
    input_control_i2c_add_setting(&input_adc_data.settings[3], 0X03, 0x00, (void*)input_adc_hw_ak5552_setting_callback);  
    input_control_i2c_add_setting(&input_adc_data.settings[4], 0X04, 0x00, (void*)input_adc_hw_ak5552_setting_callback);  
    input_control_i2c_add_setting(&input_adc_data.settings[5], 0X05, 0x00, (void*)input_adc_hw_ak5552_setting_callback);  
    input_control_i2c_add_setting(&input_adc_data.settings[6], 0X06, 0x00, (void*)input_adc_hw_ak5552_setting_callback);  
    input_control_i2c_add_setting(&input_adc_data.settings[7], 0X06, 0x00, (void*)input_adc_hw_ak5552_setting_callback);  

    input_adc_hw_ak5552_apply_settings();    
    logger_log_info(LOGGER_CODE_INPUT_ADC, "INPUT ADC AH5552 Configuration End --");
    logger_process();
}

/**-----------------------------------------------------------------------------
 * Defines the INPUT ADC callbacks used when using the AK5552.  To be used as
 * the parameter to input_adc_define_callbacks() function
 * @return 
 */    
INPUT_ADC_FUNCTION_DATA* input_adc_hw_ak5552_define_functions()
{
    // Set up the functions here
    input_adc_data.functions.configure_function = input_adc_hw_ak5552_configure;
    input_adc_data.functions.initialise_function = input_adc_hw_ak5552_initialise;
    input_adc_data.functions.reapply_settings_function = input_adc_hw_ak5552_reapply_settings;
    input_adc_data.functions.reset_function = input_adc_hw_ak5552_reset;
            
    return &input_adc_data.functions;
}
 
/**-----------------------------------------------------------------------------
 * Initialise the AK5552
 */
bool input_adc_hw_ak5552_initialise()
{
    bool result = true;

	// Register the events in the EVENTS structure, 
	events_register(EVENT_MODULE_INPUT_ADC_HW, input_adc_data.event_subscriptions, INPUT_ADC_AK5552_HW_MAX_EVENT_SUBSCRIPIONS);
    
    i2c_initialise();
     
    input_adc_data.i2c_client_id = i2c_open(INPUT_ADC_AK5552_I2C_MODULE, 
            INPUT_ADC_AK5552_I2C_READ_ADDRESS,
            INPUT_ADC_AK5552_I2C_WRITE_ADDRESS,
            input_control_i2c_callback);
    
    if (I2C_INVALID_CLIENT_ID != input_adc_data.i2c_client_id)
    {
        logger_log_info(LOGGER_CODE_INPUT_ADC, "INPUT ADC AK5552 I2C Open");
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_ADC, "INPUT ADC AK5552 I2C Open Error");
        result = false;
    }    
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Applies the settings to the AK4113
 */
void input_adc_hw_ak5552_reapply_settings()
{
    logger_log_debugf(LOGGER_CODE_INPUT_ADC, "INPUT ADC AK5552 Reapply Settings ");

    // Setup for a resend of the settings
    input_adc_data.settings_applied = false;
    input_adc_hw_ak5552_apply_settings();
}

/**-----------------------------------------------------------------------------
 * Perform AK4113 reset
 */
void input_adc_hw_ak5552_reset(bool on)
{
    logger_log_debugf(LOGGER_CODE_INPUT_ADC, "INPUT ADC AK5552 Reset - Switching %s ", on == true ? "On" : "Off");
    
    PLIB_PORTS_PinWrite(INPUT_ADC_AK5552_RESET_PORT, 
                    INPUT_ADC_AK5552_RESET_CHANNEL, 
                    INPUT_ADC_AK5552_RESET_PIN,
                    on);    
    
    if (on)
    {
        // Setup for a resend of the settings
        input_adc_data.settings_applied = false;
    }
}

/* *****************************************************************************
 End of File
 */
