/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    input_control.c

  @Summary
    Manages the amplifiers inputs.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/input_control.h"
#include "utilities/analogue_direct_feature.h"
#include "utilities/avatar.h"
#include "utilities/avatar_browsing.h"
#include "utilities/events.h"
#include "utilities/hdmi_arc_data.h"
#include "utilities/i2c.h"
#include "utilities/input_adc/input_adc_feature.h"
#include "utilities/logger/logger.h"
#include "utilities/mutes.h"
#include "utilities/power.h"
#include "utilities/settings.h"
#include "utilities/spdif_rx/spdif_rx_defs.h"
#include "app_config.h"
#include "utilities/timers.h"
#include "utilities/volume.h"
#include <stdio.h>
#include <math.h>

#ifdef DISPLAY_MESSAGE_ACTIVE
#include "utilities/menu.h"
#endif

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Info use to define a switch action
 */
typedef struct
{
    INPUT_CONTROL_TYPE input_type;
    
    bool analogue_direct;
    VOLUME_INPUT volume_input;
    int8_t volume_boost;
    SPDIF_RX_ADC_MODE adc_mode;
    bool adc_reset;
    bool dac_reset;
    bool setup_filter;
    
    INPUT_CONTROL_DEFINITION* spdif_rx_input;
    
    bool sample_rate_refresh;
    AVATAR_SAMPLE_RATE sample_rate;
    
    HDMI_ARC_STREAM_STATE hdmi_state;
    
    AVATAR_INPUT avatar_input;
} INPUT_CONTROL_SWITCH_INFO;

/**
 * Hold the state of the settings
 */
    
typedef enum 
{
    INPUT_CONTROL_STATE_INITIALISING,
    INPUT_CONTROL_STATE_CONFIGURING,
    INPUT_CONTROL_START_DIGITAL,
    INPUT_CONTROL_CONFIGURE_DIGITAL,
    INPUT_CONTROL_START_ADC,
    INPUT_CONTROL_START_DAC,
    INPUT_CONTROL_CONFIGURE_DAC,
    INPUT_CONTROL_WAIT_INITIALISE_COMPLETE,
    INPUT_CONTROL_STATE_MUTE_PRE_SWITCH,
    INPUT_CONTROL_STATE_SWITCH_INPUTS,
    INPUT_CONTROL_STATE_WAIT_DAC_SETTINGS,
    INPUT_CONTROL_STATE_UNMUTE_POST_SWITCH,
    INPUT_CONTROL_STATE_RUNNING,

    //  We have either tried to read or write from the settings byte there has 
    // been errors
    INPUT_CONTROL_STATE_ERROR,
} INPUT_CONTROL_STATE;

/**
 * Global data
 */
typedef struct
{
    // The current state of the settings re load/save
    INPUT_CONTROL_STATE state;

    //! stores the number of registered inputs
    int num_inputs;
    
    //! Stores the mute state during a switch operation
    bool mute_state_pre_switch;
    
    //! The volume prior to a switch
    uint8_t volume_pre_switch;
    
    //! Then index of the input currently acting as processor mode
    uint8_t procesor_mode_input;

    //! the Volume to be used in processor mode
    uint8_t procesor_mode_volume;
    
    //! the volume prior to entering processor mode
    uint8_t pre_procesor_mode_volume;
    
    //! the volume prior to entering processor mode
    uint8_t pre_procesor_mode_balance;    
    
    //! A call back to be invoked when switching inputs
    INPUT_CONTROL_SWITCH_CALLBACK switch_callback;

     
} INPUT_CONTROL_DATA;

static INPUT_CONTROL_DATA input_control_data = {
    .switch_callback = NULL,
};

//static uint8_t buffer[2];

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Event raised when there is an Error detected at the Hardware level
 */
static void input_control_spdif_rx_error_event_callback(EVENT* event)
{
	// Do something event related...
    SPDIF_RX_HW_ERROR_EVENT_DATA* event_data = (SPDIF_RX_HW_ERROR_EVENT_DATA*)event->context_data;
    
    //logger_log_debugf(0, "EVENTS DATA ERROR %d %d --", event_data->error_identified, event_data->sample_rate_change);            
    //logger_process();
   
    // First check that its not just a sample rate change (shouldn't be but...)
    if (event_data->error_identified)
    {
        if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
        {
            input_control_mute_inputs();

            // BUG 3825 Changed to output_control_mute_digital_outputs as Pete did not want th analogue outputs muting
            output_control_mute_digital_outputs();
            input_control_mute_dac(true);
        }
    }
    else if (event_data->sample_rate_change)
    {
        // No need to do anything - the spdif_rx module will try and get the sample rate
    }
    else
    {
        // In theory there are no errors , sometimes the inputs have been muted
        /// by the interrupt even when it reports no erros so unmute
        if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
        {
            //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX Callback NO ERRORS BUT UNMUTING INPUTS ANYWAY --");

            // Check for an invalid sample rate and retry if found TODO - Test
            SAMPLE_RATE sample_rate = spdif_rx_get_sample_rate();
            if (SAMPLE_RATE_VALUES[sample_rate] != 0)
            {
                // Unmute the inputs
                input_control_unmute_inputs(INPUT_CONTROL_TYPE_DIGITAL); 
            }
        }                
    }
}

/**-----------------------------------------------------------------------------
 * Event raised when there is an Sample Rate change detected at the Hardware level
 * This will only hapen if there is a sample raet change
 */
static void input_control_spdif_rx_sample_rate_event_callback(EVENT* event)
{
	// Do something event related...
    SPDIF_RX_SAMPLE_RATE_EVENT_DATA* event_data = (SPDIF_RX_SAMPLE_RATE_EVENT_DATA*)event->context_data;
    
    logger_log_debugf(0, "INPUT CONTROL SPDIF RX SAMPLE RATE CALLBACK %d %d --", event_data->error_identified, event_data->sample_rate);            
    logger_process();

    // Check for errors and trigger an error check if found
    if (event_data->error_identified)
    {
         // Check if the digital input is currently active
        if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
        {
            logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate errors MUTING OUTPUTS ");

            // Mute all inputs
            input_control_mute_inputs();
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate Un-muting Inputs --");

        input_control_unmute_inputs(input_control_get_current_input_type()); // TODO Moved to Avatar Set States Response callback
        // BUG 3825 
        output_control_unmute_outputs();
        input_control_mute_dac(false);
    }

    // Inform the Avatar of the new sample rate
    if (event_data->is_valid)
    {
        logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL AVATAR Setting Sample Rate %d --", SAMPLE_RATE_VALUES[event_data->sample_rate]);
        avatar_set_sample_rate2(SAMPLE_RATE_VALUES[event_data->sample_rate]);
    }
}

/**-----------------------------------------------------------------------------
 * Event raised when there is an Sample Rate change detected at the Hardware level
 * This will only hapen if there is a sample raet change
 */
static void input_control_spdif_rx_switch_event_callback(EVENT* event)
{
	// Do something event related...
    SPDIF_RX_SWITCH_EVENT_DATA* event_data = (SPDIF_RX_SWITCH_EVENT_DATA*)event->context_data;
    
    logger_log_debugf(0, "INPUT CONTROL SPDIF RX SWITCH CALLBACK %d %d --", event_data->input);            
    logger_process();
}

/**-----------------------------------------------------------------------------
 * 
 */
static void input_control_update_input_settings()
{
    // First store the current input as the previous input
    uint8_t previous_setting;
    settings_get(SETTING_ACTIVE_INPUT, &previous_setting);
    previous_setting = previous_setting < INPUT_CONTROL_NUM_SETTINGS ? previous_setting : AUDIO_INPUT_NONE; // Attempt to handle crash out if settings not properly initialised

    logger_log_debugf(0, "INPUT CONTROL Previous Input is %s %d --", AUDIO_INPUT_STRINGS[previous_setting], previous_setting);
    
    // Check that the input is enabled (note: this also applied to the Avatar input
    // that is always disabled)
    INPUT_CONTROL_DEFINITION* previous_definition = input_control_get_definition(previous_setting);
    if ((previous_definition != NULL) && (previous_definition->enabled))
    {
        logger_log_debugf(0, "INPUT CONTROL Setting Previous Input %s %d --", AUDIO_INPUT_STRINGS[previous_setting], previous_setting);
        settings_set(SETTING_PREVIOUS_INPUT, &previous_setting);
    }

    // Update the active input
    settings_set(SETTING_ACTIVE_INPUT, &input_control_get_current_definition()->index);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Add a definition of an input control to the module
 * @param index
 * @param definition
 */
void input_control_add(INPUT_CONTROL_DEFINITION* definition)
{
    if (definition->index < INPUT_CONTROL_NUM_SETTINGS)
    {
        memcpy(input_control_get_definition(definition->index), definition, sizeof(INPUT_CONTROL_DEFINITION));
        input_control_data.num_inputs++;
    }
}
        
/**-----------------------------------------------------------------------------
 * Used to determine whether an input signal is present..
 */
bool input_control_audio_input_present()
{
    bool result = false;
    
#if INPUT_CONTROL_PINS_ENABLED
    
    if (INPUT_CONTROL_AUDIO_PRESENT_PIN_STATE == INPUT_CONTROL_PIN_STATE_HIGH)
    {
        result = PLIB_PORTS_PinGet(INPUT_CONTROL_AUDIO_PRESENT_ID, 
                                   INPUT_CONTROL_AUDIO_PRESENT_PORT_CHANNEL, 
                                   INPUT_CONTROL_AUDIO_PRESENT_PIN);
    }
    else
    {
        result = !PLIB_PORTS_PinGet(INPUT_CONTROL_AUDIO_PRESENT_ID, 
                                   INPUT_CONTROL_AUDIO_PRESENT_PORT_CHANNEL, 
                                   INPUT_CONTROL_AUDIO_PRESENT_PIN);
    }

#endif        
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the display sample rate of the current digital input 19200 if analog
 * @return 
 */
SAMPLE_RATE input_control_get_display_sample_rate()
{
    SAMPLE_RATE result = SAMPLE_RATE_192000;
    
    if (input_control_get_current_definition() != NULL)
    {
        switch(input_control_get_current_definition()->type)
        {
            case INPUT_CONTROL_TYPE_ANALOGUE:
            {
                // Do nothing
                break;
            }

            case INPUT_CONTROL_TYPE_DIGITAL:
            {
                result = spdif_rx_get_sample_rate();
                break;
            }

            case INPUT_CONTROL_TYPE_AVATAR:
            {
                //NOTE:  This currently relies on the SAMPLE_RATE and AVATAR_SAMPLE_RATE 
                //       enums having the same entries
                result = (SAMPLE_RATE)avatar_message_get_display_sample_rate();
                break;
            }
        }
    }
    
    return result; //input_control_get_dac_sample_rate();
}
   
/**-----------------------------------------------------------------------------
 * Gets the next enabled input relative to the input supplied
 * @param forwards - Indicates the direction to move through the inputs
 * @return The Id of the next input
 */
uint8_t input_control_get_next_input(uint8_t input, bool forwards)
{
    uint8_t result = 0;
    INPUT_CONTROL_DEFINITION* definition = input_control_get_next_definition(input, forwards);
    if (definition != NULL)
    {
        result = definition->index;
    }
    else
    {
        logger_log_debugf(0, "INPUT CONTROL get next input %d forwards %d def idx %d--", input, forwards, definition->index);
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns the number of registered inputs
 * @return 
 */
uint8_t input_control_get_num_inputs()
{
    return input_control_data.num_inputs;
}

/**-----------------------------------------------------------------------------
 * Gets the previous input.  If that is not an enable input then it looks for the
 * first enabled input
 * @return 
 */
uint8_t input_control_get_previous_input()
{
    uint8_t previous_input;
    settings_get(SETTING_PREVIOUS_INPUT, &previous_input);

    // Now check that the previous input is still enabled
    INPUT_CONTROL_DEFINITION* input_definition = input_control_get_definition(previous_input);
    if ((input_definition == NULL) || (!input_definition->enabled))
    {
        // If not default to the first enabled input (miss 0 as this is input None))
        int x;
        for (x = 1; x < input_control_get_num_inputs(); x++)
        {
            input_definition = input_control_get_definition(x);
            if (input_definition->enabled)
            {
                previous_input = input_definition->index;
                break;
            }
        }
    }

    logger_log_debugf(0, "INPUT CONTROL Restoring to Previous Input %s %d ", AUDIO_INPUT_STRINGS[previous_input], previous_input);
    return previous_input;
}

/**-----------------------------------------------------------------------------
 * Gets the index of the current input designated for Processor mode
 * @return 
 */
uint8_t input_control_get_processor_mode_input()
{
    return input_control_data.procesor_mode_input;
}

/**-----------------------------------------------------------------------------
 * Gets current volume for Processor mode
 * @return 
 */
uint8_t input_control_get_processor_mode_volume()
{
    return input_control_data.procesor_mode_volume;
}
     
/**-----------------------------------------------------------------------------
 * Gets the sample rate of the current digital input 0 if analog
 * @return 
 */
SAMPLE_RATE input_control_get_sample_rate()
{
    SAMPLE_RATE result = SAMPLE_RATE_0;
    
    if (input_control_get_current_definition() != NULL)
    {
        switch(input_control_get_current_definition()->type)
        {
            case INPUT_CONTROL_TYPE_ANALOGUE:
            {
                // Do nothing
                break;
            }

            case INPUT_CONTROL_TYPE_DIGITAL:
            {
                result = spdif_rx_get_sample_rate();
                break;
            }

            case INPUT_CONTROL_TYPE_AVATAR:
            {
                //NOTE:  This currently relies on the SAMPLE_RATE and AVATAR_SAMPLE_RATE 
                //       enums having the same entries
                result = (SAMPLE_RATE)avatar_message_get_sample_rate();
                break;
            }
        }
    }
    
    return result; //input_control_get_dac_sample_rate();
}

/**-----------------------------------------------------------------------------
 * Gets the index of the currently selected input
 * @return 
 */
uint8_t input_control_get_selected_input()
{
    uint8_t result = INPUT_CONTROL_NUM_SETTINGS; // Out of range value
    
    if (input_control_get_current_definition() != NULL)
    {
        result = input_control_get_current_definition()->index;
        
        // Check if we are in a transition
        if (input_control_data.state != INPUT_CONTROL_STATE_RUNNING)
        {
            if (input_control_get_new_definition() != NULL)
            {
                result = input_control_get_new_definition()->index;
            }
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the current type of input control (Analog/Digital)
 * @return 
 */
INPUT_CONTROL_TYPE input_control_get_type()
{
    INPUT_CONTROL_TYPE result = INPUT_CONTROL_TYPE_NONE;
    
    if (input_control_get_current_definition() != NULL)
    {
        result = input_control_get_current_definition()->type;
       
        // Check if we are in a transition
        if (input_control_data.state != INPUT_CONTROL_STATE_RUNNING)
        {
            if (input_control_get_new_definition() != NULL)
            {
                result = input_control_get_new_definition()->type;
            }
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Determines whether the input control believes that the volume should be muted
 * This is not the same as volume_is_muted as its possible to have the volume
 * muted but input control say its not (e.g. when there is an error on the digital signal)
 * @return 
 */
bool input_control_in_mute_state()
{
    bool result = volume_is_muted();
     
    return result;
}    

/**-----------------------------------------------------------------------------
 * Handles interrupt e.g.on the DAC GPIO1 which is configured via Register8 to change
 * When the dpll_lock status changes
 * @param channel - The channel the interrupt is being raised on
 */
void input_control_handle_interrupt(uint8_t channel)
{
    if (channel == INPUT_CONTROL_DAC_GPIO_PORT_CHANNEL)
    {
        input_control_handle_dac_interrupt();
    }
    
    // Note this check which channel the interrupt has been raised on
    spdif_rx_handle_interrupt(channel);
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether the selected input is the processor mode input
 * @return 
 */
bool input_control_in_processor_mode()
{
    bool result = false;
    
    INPUT_CONTROL_DEFINITION* current_def = input_control_get_current_definition();
    
    if (current_def != NULL)
    {
        result = current_def->index == input_control_data.procesor_mode_input;
            
        // Check if we are in a transition
        if (input_control_data.state != INPUT_CONTROL_STATE_RUNNING)
        {
            if (input_control_get_new_definition() != NULL)
            {
                result = input_control_get_new_definition()->index == input_control_data.procesor_mode_input;
            }
        }
    }
    
    return result;
}
     
/**-----------------------------------------------------------------------------
 * Can be used to determine whether the input control is currently switching inputs
 * @return 
 */
bool input_control_in_switch()
{
//    bool result =  (input_control_data.state == INPUT_CONTROL_STATE_MUTE_PRE_SWITCH) ||
//                   (input_control_data.state == INPUT_CONTROL_STATE_SWITCH_INPUTS) ||
//                   (input_control_data.state == INPUT_CONTROL_STATE_UNMUTE_POST_SWITCH);
    bool result =  ((input_control_data.state >= INPUT_CONTROL_STATE_MUTE_PRE_SWITCH) &&
                   (input_control_data.state <= INPUT_CONTROL_STATE_UNMUTE_POST_SWITCH));
    
    return result;
}

/**-----------------------------------------------------------------------------
 * 
 * @return 
 */
INPUT_CONTROL_STATE input_control_state()
{
    return input_control_data.state;
}

/**-----------------------------------------------------------------------------
 * Initialise the input control module
 */
void input_control_initialise()
{
    input_control_data.num_inputs = 0;
    
    input_control_data.procesor_mode_input = 0; // Assumes 0 is no input or default input
    input_control_data.procesor_mode_volume = VOLUME_DEFAULT_VOLUME;
    input_control_data.pre_procesor_mode_volume = volume_get_stored();
    input_control_data.pre_procesor_mode_balance = volume_get_balance();

    input_control_data.mute_state_pre_switch = false;

    if (input_control_dac_initialise())
    {
#if APP_FEATURE_ANALOGUE_INPUTS == APP_FEATURE_ON
        if (input_control_analogue_initialise())
        {
#endif            
//            if (input_control_digital_initialise())
            if (spdif_rx_initialise())
            {
                // Subscribe to the events raise in the HW layer.  Note the Event Module ID should
                // work for either HW type 
                events_subscribe(EVENT_MODULE_SPDIF_RX, 
                                 SPDIF_RX_EVENT_ERROR, 
                                 input_control_spdif_rx_error_event_callback, 
                                 (void*)NULL, 
                                true);
                
                events_subscribe(EVENT_MODULE_SPDIF_RX, 
                                 SPDIF_RX_EVENT_SAMPLE_RATE, 
                                 input_control_spdif_rx_sample_rate_event_callback, 
                                 (void*)NULL, 
                                 true);
                
                events_subscribe(EVENT_MODULE_SPDIF_RX, 
                                 SPDIF_RX_EVENT_SWITCH, 
                                 input_control_spdif_rx_switch_event_callback, 
                                 (void*)NULL, 
                                 true);
                
                input_control_data.state = INPUT_CONTROL_STATE_INITIALISING;
            }
            else
            {
                input_control_data.state = INPUT_CONTROL_STATE_ERROR;
            }
#if APP_FEATURE_ANALOGUE_INPUTS == APP_FEATURE_ON
        }
        else
        {
            input_control_data.state = INPUT_CONTROL_STATE_ERROR;
        }
#endif
    }
    else
    {
        input_control_data.state = INPUT_CONTROL_STATE_ERROR;
    }
}
    
/**-----------------------------------------------------------------------------
 * This actually switches the SPDIF MUX to a channel with no inputs.
 */
void input_control_mute_inputs()
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX MUTING OUTPUT ");
    
    // Create a dummy definition of type INPUT_CONTROL_TYPE_NONE
    INPUT_CONTROL_DEFINITION def;
    def.type = INPUT_CONTROL_TYPE_NONE;
    input_control_select_mux_input(&def);
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether the input was muted prior to a current switch action
 * @return 
 */
bool input_control_mute_state_pre_switch()
{
    return input_control_data.mute_state_pre_switch;
}
    
/**-----------------------------------------------------------------------------
 * Used to display the new volume during a switch operation
 * @return 
 */
uint8_t input_control_volume_in_switch()
{
    uint8_t result = input_control_data.volume_pre_switch;
            
    // Check if we are in a transition
    if (input_control_data.state != INPUT_CONTROL_STATE_RUNNING)
    {
        if (input_control_get_new_definition() != NULL)
        {
            if (input_control_get_new_definition()->index == input_control_data.procesor_mode_input)
            {
                result = input_control_data.procesor_mode_volume;
            }
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Perform input control processing
 * Note: See app_commands_input_configure() for the rest of the configuration settings
 */

void input_control_process()
{
    //logger_log_debugf(0, "INPUT CONTROL Process %d ", input_control_data.state);
    
//    preview_info = input_control_get_preview_info();
//    current_definition = input_control_get_current_definition();
//    new_definition = input_control_get_definition(preview_info->input);

    // Handle input switching
    switch(input_control_data.state)
    {
        case INPUT_CONTROL_STATE_INITIALISING:
        {
            // Sit in the initialising state until input_control_start_configuration is called
            // We only move out of the initialising state once power is up 
            break;
        }
        
        case INPUT_CONTROL_STATE_CONFIGURING:
        {
            logger_log_debugf(0, "INPUT CONTROL CONFIGURING");
            logger_process();
            // Start the configuration process
            input_control_data.state = INPUT_CONTROL_START_DIGITAL; 
        }
        
        case INPUT_CONTROL_START_DIGITAL:
        {
            logger_log_debugf(0, "INPUT CONTROL START_DIGITAL --");
            logger_process();
            
             // Get SPDIF MUX out of Reset
            spdif_rx_reset(INPUT_CONTROL_SWITCH_ON);
            logger_log_debugf(0, "INPUT CONTROL START_DIGITAL AA --");
            logger_process();

            timers_reset(TIMER_INDEX_INPUT_CONTROL_WAIT);

            input_control_data.state = INPUT_CONTROL_CONFIGURE_DIGITAL;
              
            break;
        }
        
        case INPUT_CONTROL_CONFIGURE_DIGITAL:
        {
            if (timers_has_expired(TIMER_INDEX_INPUT_CONTROL_WAIT))
            {
                logger_log_debugf(0, "INPUT CONTROL CONFIGURE_DIGITAL");
                logger_process();
                
//                input_control_digital_configure();
                spdif_rx_configure();
                //TODO:  Consider changing this to use an event handler OnConfigured...rather than a wait command?
                
                timers_reset(TIMER_INDEX_INPUT_CONTROL_WAIT);
    
                if (input_adc_feature_is_present())
                {
                    input_control_data.state = INPUT_CONTROL_START_ADC;
                }
                else
                {
                    input_control_data.state = INPUT_CONTROL_START_DAC;
                }
            }
            
            break;
        }
        
#if APP_FEATURE_ANALOGUE_INPUTS == APP_FEATURE_ON
        case INPUT_CONTROL_START_ADC:
        {
            if (timers_has_expired(TIMER_INDEX_INPUT_CONTROL_WAIT))
            {
                input_adc_feature_reset(INPUT_CONTROL_SWITCH_ON);
                
                input_control_data.state = INPUT_CONTROL_START_DAC;
               
                timers_reset(TIMER_INDEX_INPUT_CONTROL_WAIT);
            }
            
            break;
        }
#endif
        
        case INPUT_CONTROL_START_DAC:
        {
            if (timers_has_expired(TIMER_INDEX_INPUT_CONTROL_WAIT))
            {
                // Release DAC reset mode
                input_control_dac_reset(INPUT_CONTROL_SWITCH_ON);
               
                timers_reset(TIMER_INDEX_INPUT_CONTROL_WAIT);
    
                input_control_data.state = INPUT_CONTROL_CONFIGURE_DAC;
            }
 
            break;
        }
        
        case INPUT_CONTROL_CONFIGURE_DAC:
        {
            if (timers_has_expired(TIMER_INDEX_INPUT_CONTROL_WAIT))
            {
                logger_log_debugf(0, "INPUT CONTROL INPUT_CONTROL_CONFIGURE_DAC");
                logger_process();
                //logger_log_debugf(0, "INPUT CONTROL Status %d %d %d %d %d %d  --", I2C1STATbits.BCL, I2C1STATbits.IWCOL, I2C1STATbits.I2COV, I2C1STATbits.S, I2C1STATbits.RBF, I2C1STATbits.TBF);
                //logger_log_debugf(0, "INPUT CONTROL Status %d %d %d %d  --", I2C1STAT, I2C1CON, I2C1TRN, I2C1RCV);
                
//                input_control_analogue_configure();
                input_adc_feature_configure();
                input_control_dac_configure();
              
                // Note must be run before input_control_start()
                input_control_data.state = INPUT_CONTROL_WAIT_INITIALISE_COMPLETE;
            }
            
            break;
        }
        
        case INPUT_CONTROL_WAIT_INITIALISE_COMPLETE:
        {
            bool configured = false;
            if (input_adc_feature_is_present())
            {
                configured = input_adc_feature_is_configured() && input_control_dac_configured();
            }
            else
            {
                configured = input_control_dac_configured();
            }
            
            if (configured)
            {
                logger_log_debugf(0, "INPUT CONTROL CONFIGURATION_COMPLETE --");
                input_control_data.state = INPUT_CONTROL_STATE_RUNNING;  // This is required to get the input selected in input control start
                input_control_setup_filter();
                input_control_start();  
                
                // Start SPDIF RX Polling
                timers_reset(TIMER_INDEX_SPDIF_RX_POLL);
            }
            
            break;
        }
        
        case INPUT_CONTROL_STATE_MUTE_PRE_SWITCH:
        {
            logger_log_debugf(0, "INPUT CONTROL INPUT_CONTROL_STATE_MUTE_PRE_SWITCH --");
            //logger_log_debugf(0, "INPUT CONTROL: Switch - Muting Inputs ");

//TODO Not sure if this is requres as it should be OK with SA30/SA70 as well  needs testing
#ifdef INPUT_CONTROL_DAC_CONTROLLED_VOLUME    
            // BUG 3356 - Muting and unmuting when switching inputs
            volume_set_immediate(0, false);
            // Note - we call the mutes function rather than volume_mute() to
            // stop the display showing "MUTE"
            mutes_mute(MUTE_INDEX_VOLUME);
#else            
            // Note - we call the mutes function rather than volume_mute() to
            // stop the display showing "MUTE"
            mutes_mute(MUTE_INDEX_VOLUME);
#endif
            
            // This is an attempt to stop any I2S audio being sent to the avatar during 
            // the switch - it should be un-muted by the call in INPUT_CONTROL_STATE_UNMUTE_POST_SWITCH
            input_control_mute_inputs();
            
            // Now pause while the mutes take effect
            timers_reset(TIMER_INDEX_INPUT_SWITCH);
            
            if (input_control_data.switch_callback != NULL)
            {
                input_control_data.switch_callback(INPUT_CONTROL_SWITCH_STATE_PRE);
            }

            input_control_data.state = INPUT_CONTROL_STATE_SWITCH_INPUTS;
   
            break;
        }

        case INPUT_CONTROL_STATE_SWITCH_INPUTS:
        {
            // If we turn the input encoder just after the preview has expired,
            // we can still update the input provided the mute timer hasn't expired
            // in this case the difference between the current and new definitions
            // will be 2
            if (!timers_has_expired(TIMER_INDEX_INPUT_SWITCH))
            {
                // we're muting before changing the input, but has the input has just changed?
                INPUT_CONTROL_PREVIEW_INFO* preview_info = input_control_get_preview_info();
                INPUT_CONTROL_DEFINITION* current_definition = input_control_get_current_definition();
                INPUT_CONTROL_DEFINITION* new_definition = input_control_get_definition(preview_info->input);
                
                if (current_definition > new_definition)
                {
                    if ((current_definition - new_definition) == 2)
                    {
                        input_control_set_new_definition(new_definition->index);
                    }
                }
                else
                {
                    if ((new_definition - current_definition) == 2)
                    {
                        input_control_set_new_definition(new_definition->index);
                    }
                }
            }
           
            // only test once for timer expired
            //if ((timers_has_expired(TIMER_INDEX_INPUT_SWITCH)) && (avatar_get_state() == AVATAR_STATE_RUNNING))
            
            else if (avatar_get_state() == AVATAR_STATE_RUNNING)
            {
                logger_log_debugf(0, "INPUT CONTROL INPUT_CONTROL_STATE_SWITCH_INPUTS --");
                logger_log_debugf(0, "INPUT CONTROL Switch - Switching Inputs %d --", volume_get_current());

//                // Define the steps to be taken in the switch
//                INPUT_CONTROL_SWITCH_INFO switch_info;
//                switch_info.input_type = input_control_get_new_definition()->type;
//                switch_info.analogue_direct = input_analogue_direct_get(input_control_get_new_definition()->index);
//                
//                switch(switch_info.input_type)
//                {
//                    case INPUT_CONTROL_TYPE_DIGITAL:
//                    {
//                        switch_info.volume_input = VOLUME_INPUT_DEFAULT;
//                        switch_info.volume_boost = INPUT_CONTROL_VOLUME_BOOST_DIGITAL;
//                        switch_info.adc_mode = SPDIF_RX_ADC_MODE_POWER_ON_MASTER;
//                        switch_info.adc_reset = INPUT_CONTROL_SWITCH_OFF;
//                        switch_info.dac_reset = INPUT_CONTROL_SWITCH_ON;
//                        switch_info.setup_filter = true;
//                        switch_info.spdif_rx_input = input_control_get_new_definition();
//                                
//                        if (input_control_get_new_definition()->index == AUDIO_INPUT_HDMI_ARC)
//                        {
//                            switch_info.hdmi_state = HDMI_ARC_STREAM_STATE_ACTIVE;
//                        }
//                        else
//                        {
//                            switch_info.hdmi_state = HDMI_ARC_STREAM_STATE_INACTIVE;
//                        }
//                        
//                        switch_info.sample_rate_refresh = true;
//                        switch_info.sample_rate = AVATAR_SAMPLE_RATE_0;
//                        
//                        switch_info.avatar_input = AVATAR_INPUT_EXTERNAL;
//                        break;
//                    }
//
//                    case INPUT_CONTROL_TYPE_ANALOGUE:
//                    {
//                        switch_info.volume_input = VOLUME_INPUT_DEFAULT;
//                        if (switch_info.analogue_direct)
//                        {
//                            switch_info.volume_boost = INPUT_CONTROL_VOLUME_BOOST_ANALOGUE_DIRECT;
//                            switch_info.adc_mode = SPDIF_RX_ADC_MODE_POWER_OFF_MASTER;
//                            switch_info.adc_reset = INPUT_CONTROL_SWITCH_OFF;
//                            switch_info.dac_reset = INPUT_CONTROL_SWITCH_OFF;
//                        }
//                        else
//                        {
//                            switch_info.volume_boost = INPUT_CONTROL_VOLUME_BOOST_ANALOGUE;
//                            switch_info.adc_mode = SPDIF_RX_ADC_MODE_POWER_ON_SLAVE;
//                            switch_info.adc_reset = INPUT_CONTROL_SWITCH_ON;
//                            switch_info.dac_reset = INPUT_CONTROL_SWITCH_ON;
//                        }
//                        switch_info.setup_filter = true;
//                        switch_info.spdif_rx_input = input_control_get_definition(AUDIO_INPUT_DIGITAL_MUTE);
//                        switch_info.hdmi_state = HDMI_ARC_STREAM_STATE_INACTIVE;
//
//                        switch_info.sample_rate_refresh = false;
//                        switch_info.sample_rate = AVATAR_SAMPLE_RATE_192000;
//
//                        switch_info.avatar_input = AVATAR_INPUT_EXTERNAL;
//                        break;
//                    }
//
//                    case INPUT_CONTROL_TYPE_AVATAR:
//                    {
//                        switch_info.volume_input = VOLUME_INPUT_STREAMER;
//                        switch_info.volume_boost = INPUT_CONTROL_VOLUME_BOOST_AVATAR;
//                        switch_info.adc_mode = SPDIF_RX_ADC_MODE_POWER_OFF_SLAVE;
//                        switch_info.adc_reset = INPUT_CONTROL_SWITCH_OFF;
//                        switch_info.dac_reset = INPUT_CONTROL_SWITCH_ON;
//                        switch_info.setup_filter = false;
//
//                        switch_info.spdif_rx_input = input_control_get_definition(AUDIO_INPUT_DIGITAL_MUTE);
//                        switch_info.hdmi_state = HDMI_ARC_STREAM_STATE_INACTIVE;
//
//                        switch_info.sample_rate_refresh = false;
//                        switch_info.sample_rate = AVATAR_SAMPLE_RATE_0;
//                        
//                        switch_info.avatar_input = AVATAR_INPUT_INTERNAL;
//                        break;
//                    }
//                }
                
                // Turn Analogue Direct Off
                //! BUG 3356 - Muting and unmuting when switching inputs   
                // Switch analogue direct off (reviewed when setting to analogue))
                analogue_direct_feature_set_active(false);
                
#if APP_FEATURE_DIRAC == APP_FEATURE_ON
                // Set Dirac Curve
                // Set the Dirac curve for the selected input
                // Check that this is the right place to do it and that it doesn't need
                // to be done after the avatar has been told to change input
                // BUG 3483 - Dirac active when headphones connected - added check for headphone connection
                if (output_control_headphones_connected() && !output_control_get_headphone_override())
                {
                    avatar_set_dirac_curve(DIRAC_NO_CURVE);
                }
                else
                {
                    avatar_set_dirac_curve(dirac_get_curve_for_input(input_control_get_new_definition()->index));
                }
#endif

#if APP_FEATURE_LIP_SYNC == APP_FEATURE_ON                
                // Send Lip Sync Delay
                uint8_t lip_sync_settings[INPUT_CONTROL_NUM_SETTINGS];
                settings_get(SETTING_LIP_SYNC, lip_sync_settings);

                uint8_t lip_sync_value_idx = lip_sync_settings[input_control_get_new_definition()->index];
                avatar_send_lip_sync_delay(LIP_SYNC_VALUES[lip_sync_value_idx]);
#endif               
                
//                // Setup max volume
//                volume_set_current_input(switch_info.volume_input);
//                avatar_send_maximum_volume(volume_get_maximum_volume());
//                
//                // Boost/Attenuate the volume
//                if (switch_info.volume_boost > 0)
//                {
//                    volume_boost(switch_info.volume_boost);
//                }
//                else
//                {
//                    volume_attenuate(switch_info.volume_boost * -1);
//                }
//                
//                // Set SPDIF RX ADC Mode
//                spdif_rx_set_adc_mode(switch_info.adc_mode);
//                
//                // ADC Reset
//                input_adc_feature_reset(switch_info.adc_reset);
//                
//                // DAC Reset
//                input_control_dac_reset(switch_info.dac_reset);
//                
//                // Reset the filter in case we are moving away from the Avatar playing MQA with the custom filter 
//                if (switch_info.setup_filter)
//                {
//                    input_control_setup_filter();
//                }
//                
//                // SPDIF RX Switch
//                 spdif_rx_switch(switch_info.spdif_rx_input);
// 
//                 // Analogue switch 
//                 if (switch_info.input_type == INPUT_CONTROL_TYPE_ANALOGUE)
//                 {
//                    // Tell analogue MUX to switch
//                    input_control_switch_analog(input_control_get_new_definition());
//                 }
//                 
//#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON                
//                // HDMI ARC Stream State
//                hdmi_arc_send_arc_stream_message(switch_info.hdmi_state);
//#endif
//                
//                // Sample Rate
//                if (switch_info.sample_rate_refresh)
//                {
//                    // attempt to trigger the sample rate mssage - not sure this is needed see BUG 3776
//                    avatar_message_set_sample_rate(AVATAR_SAMPLE_RATE_0);
//
//                    // Trigger a sample rate calculation as the interrupts don't appear to fire 
//                    spdif_rx_refresh_sample_rate();     
//                }
//                else if (switch_info.sample_rate != AVATAR_SAMPLE_RATE_0)
//                {
//                    avatar_set_sample_rate(switch_info.sample_rate, false);
//                }
//                
//                // Avatar Input
//                avatar_set_input(switch_info.avatar_input);
                
                
                
                
                
                // Set default input - changed for Avatar below
                volume_set_current_input(VOLUME_INPUT_DEFAULT);
                
                switch(input_control_get_new_definition()->type)
                {
                    case INPUT_CONTROL_TYPE_DIGITAL:
                    {
                        // Send maximum volume to the avatar
                        avatar_send_maximum_volume(volume_get_maximum_volume());
                        
                        // ISSUE-20171211-1 compensate for a louder digital signal 
                        // we attenuate the volume of the digital signal by 6dB
                        // Changed to 2dB at Pete/Johns request - note change to 
                        // DAC config so that the other 4dB is attenuated by the DAC itself.
#ifdef INPUT_CONTROL_VOLUME_ATTENUATE_DIGITAL                            
                        volume_attenuate(INPUT_CONTROL_VOLUME_ATTENUATE_DIGITAL);    // BUG 3632
#else
                        volume_attenuate(8);    // BUG3506 was 0x08
#endif
                        
                        // See Petes email 28-05-2019 10:52
                        //input_control_digital_set_master_slave(true);
                        //input_control_digital_set_power(true);
                        spdif_rx_set_adc_mode(SPDIF_RX_ADC_MODE_POWER_ON_MASTER);
                        
                        input_control_dac_reset(INPUT_CONTROL_SWITCH_ON);
                 
                        // Reset the filter in case we are moving away from the Avatar playing MQA with the custom filter 
                        input_control_setup_filter();
                       
                        // Switch the SPDIR Rx mux
                        spdif_rx_switch(input_control_get_new_definition());
                           
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                        if (input_control_get_new_definition()->index == AUDIO_INPUT_HDMI_ARC)
                        {
                            hdmi_arc_send_arc_stream_message(HDMI_ARC_STREAM_STATE_ACTIVE);
                        }
                        else
                        {
                            hdmi_arc_send_arc_stream_message(HDMI_ARC_STREAM_STATE_INACTIVE);
                        }
#endif
                        // attempt to trigger the sample rate mssage - not sure this is needed see BUG 3776
                        avatar_message_set_sample_rate(AVATAR_SAMPLE_RATE_0);
                        
                        // Trigger a sample rate calculation as the interrupts don't appear to fire 
                        spdif_rx_refresh_sample_rate();     
                        
                        // Tell the Avatar that were using the I2S Mux output as input
                        avatar_set_input(AVATAR_INPUT_EXTERNAL);
                        
                        break;
                    }

#if APP_FEATURE_ANALOGUE_INPUTS == APP_FEATURE_ON
                    case INPUT_CONTROL_TYPE_ANALOGUE:
                    {
                        // Send maximum volume to the avatar
                        avatar_send_maximum_volume(volume_get_maximum_volume());
                        
                        if (input_analogue_direct_get(input_control_get_new_definition()->index))    //input_control_get_new_definition()->analogue_direct)
                        {
#ifdef INPUT_CONTROL_VOLUME_ATTENUATE_ANALOGUE_DIRECT                            
                            volume_attenuate(INPUT_CONTROL_VOLUME_ATTENUATE_ANALOGUE_DIRECT);    // BUG 3632
#else
                            volume_attenuate(7);    // BUG3506 was 0x08
#endif
                            //input_control_digital_set_master_slave(true);
                            //input_control_digital_set_power(false);
                            spdif_rx_set_adc_mode(SPDIF_RX_ADC_MODE_POWER_OFF_MASTER);
                            
                            input_adc_feature_reset(INPUT_CONTROL_SWITCH_OFF);
                            input_control_dac_reset(INPUT_CONTROL_SWITCH_OFF);
                        }
                        else
                        {
#ifdef INPUT_CONTROL_VOLUME_ATTENUATE_ANALOGUE                            
                            volume_attenuate(INPUT_CONTROL_VOLUME_ATTENUATE_ANALOGUE);    // BUG 3632
#else
                            volume_boost(5);    // BUG3506 was 0x08
#endif
                            //input_control_digital_set_master_slave(false);
                            //input_control_digital_set_power(true);
                            spdif_rx_set_adc_mode(SPDIF_RX_ADC_MODE_POWER_ON_SLAVE);
                            input_adc_feature_reset(INPUT_CONTROL_SWITCH_ON);
                            input_control_dac_reset(INPUT_CONTROL_SWITCH_ON);
                        }
                        
                        // Switch off the SPDIF RX  otherwise it makes a noise....(See Pete email 21/08/2019 09:25))
                        spdif_rx_switch(input_control_get_definition(AUDIO_INPUT_DIGITAL_MUTE));

                        // Reset the filter in case we are moving away from the Avatar playing MQA with the custom filter 
                        input_control_setup_filter();
    
                        // Reapply the settings (just in case); - should this be moved to only when analogue direct is off?
                        input_adc_feature_reapply_settings();

                        // Tell analogue MUX to switch
                        input_control_switch_analog(input_control_get_new_definition());

                        analogue_direct_feature_set_active(input_analogue_direct_get(input_control_get_new_definition()->index));   
   
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                        hdmi_arc_send_arc_stream_message(HDMI_ARC_STREAM_STATE_INACTIVE);
#endif
                        
                        // Tell the Avatar that were using the I2S Mux output as input
                        avatar_set_input(AVATAR_INPUT_EXTERNAL);
                    
                        // Set the sample rate
                        avatar_set_sample_rate(AVATAR_SAMPLE_RATE_192000, false);

                        break;
                    }
#endif
                    
                    case INPUT_CONTROL_TYPE_AVATAR:
                    {
                        // Send maximum volume to the avatar
                        avatar_send_maximum_volume(volume_get_maximum_net_volume());
                        
                        volume_set_current_input(VOLUME_INPUT_STREAMER);
                        
#ifdef INPUT_CONTROL_VOLUME_ATTENUATE_AVATAR                           
                        volume_attenuate(INPUT_CONTROL_VOLUME_ATTENUATE_AVATAR);    // BUG 3632
#else
                        volume_attenuate(8);    // BUG3506 was 0x08
#endif
                        
                        //input_control_digital_set_master_slave(false);
                        //input_control_digital_set_power(false);
                        spdif_rx_set_adc_mode(SPDIF_RX_ADC_MODE_POWER_OFF_SLAVE);
                        
                        input_control_dac_reset(INPUT_CONTROL_SWITCH_ON);
                        
                        // Switch off the SPDIF RX  
                        spdif_rx_switch(input_control_get_definition(AUDIO_INPUT_DIGITAL_MUTE));
                        
                        // Tell the Avatar that were using its internal source as input
                        avatar_set_input(AVATAR_INPUT_INTERNAL);

                        //input_control_switch_avatar(input_control_get_new_definition());
                        
   
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                        hdmi_arc_send_arc_stream_message(HDMI_ARC_STREAM_STATE_INACTIVE);
#endif

                        break;
                    }
                }
                
//#define APP_FUNCTION_PRESENT(t) if (t == APP_FEATURE_ON) 
//                
//                APP_FUNCTION_PRESENT(APP_FEATURE_ANALOGUE_INPUTS)
//                {
//                    logger_log_debugf(0, "APP_FUNCTION_TEST --");
//                }
                
                
                //timers_delay_ms(10);
#if APP_FEATURE_ANALOGUE_INPUTS == APP_FEATURE_ON
                input_adc_feature_reapply_settings();
                
                // Applying DAC Settings kicks off a whole load of callbacks and interrupts
                // Note: Below we wait for all this to finish before moving to the 
                // default running state.  The reason this was added was because
                // when switching between ANalogue Direct modes using the Direct button
                // the whole unit could keel over, probably because it was getting 
                // in a state over and interrupt
                // BUG 3593 - No output + "lock up" when switching between ADC and DIRECT analogue paths - When in analogue direct DAc is in reset so applying DAC Settings doesn't work
                if (!analogue_direct_get_active())
                {
                    input_control_apply_dac_settings();
                }
                else
                {
                    logger_log_debugf(0, "INPUT CONTROL Omitting DAC Settings as on Analogue Direct input --");
                }
#else
                input_control_apply_dac_settings();                
#endif

               // Update current definition
                input_control_set_current_definition(input_control_get_new_definition()->index);

#if APP_FEATURE_DIRAC == APP_FEATURE_ON
                // Check if we are calibration - if so leave the screen alone
                // otherwise display the default screen
                if (avatar_message_get_dirac_client_status() == AVATAR_DIRAC_CLIENT_STATUS_DISCONNECTED)
                {   
                    app_commands_display_default_screen(NULL);
                }
                else
                {
                    logger_log_debugf(0, "INPUT CONTROL Suppressing display of default screen --");
                }
#else
                logger_log_debugf(0, "DIRAC CLIENT STATUS D %d --", avatar_message_get_dirac_client_status());
                // At this point we need to check which menu to display.
                // Not sure at the moment whether to raise a callback here (likely), or to put it into input_control_set_current_definition )
                app_commands_display_default_screen(NULL);
#endif 
                
#ifdef DISPLAY_MESSAGE_ACTIVE
                MENU_ITEM* menu_item = menu_get_current_item();
                if (menu_item->menu_id == APP_COMMAND_MENU_CURRENTLY_PLAYING)
                {
                    display_message_build_screen_data(APP_COMMAND_MENU_CURRENTLY_PLAYING);
                }
                else if (menu_item->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
                {
                    display_message_build_screen_data(APP_COMMAND_MENU_MEDIA_BROWSING);
                }
#endif        
                
                input_control_data.state = INPUT_CONTROL_STATE_WAIT_DAC_SETTINGS;
            }
            
            break;
        }
        
        case INPUT_CONTROL_STATE_WAIT_DAC_SETTINGS:
        {
            bool dac_ready = false;
#if APP_FEATURE_ANALOGUE_INPUTS == APP_FEATURE_ON
            dac_ready = input_control_dac_configured() || !analogue_direct_get_active();
#else
            dac_ready = input_control_dac_configured();
#endif
            // Wait for all the DAC configuration to complete.  
            // BUG 3593 - No output + "lock up" when switching between ADC and DIRECT analogue paths - When in analogue direct DAc is in reset so applying DAC Settings doesn't work
            if (dac_ready  && !spdif_rx_in_switch())
            {    
                logger_log_debugf(0, "INPUT CONTROL DAC Settings are now applied - moving to post-switch");
                
                // Because we've been waiting for the DAC config we've missed the 
                // trigger for sample rate checking so kick that off here..
                input_control_check_source_type();

                // Pause to allow th DAC to sort itself out  - when this was not here it causes a CLUNK on the XLR outputs of th ST60
                timers_reset(TIMER_INDEX_INPUT_SWITCH);
                
                // Go to the running state
                input_control_data.state = INPUT_CONTROL_STATE_UNMUTE_POST_SWITCH;
            }
            
            break;
        }

        case INPUT_CONTROL_STATE_UNMUTE_POST_SWITCH:
        {    
            if (timers_has_expired(TIMER_INDEX_INPUT_SWITCH)) 
            {
                logger_log_debugf(0, "INPUT CONTROL VOLUME INPUT_CONTROL_STATE_UNMUTE_POST_SWITCH --");
                logger_log_debugf(0, "INPUT CONTROL: Switch - UnMuting Inputs INPUT_CONTROL_STATE_UNMUTE_POST_SWITCH --");
                
                // Switch the I2S Mux
                input_control_select_mux_input(input_control_get_current_definition());
                
                // BUG 3762 
                // Check whether we are in startup and the mute state os on
                bool mute_on_startup = false;
                if (!app_context_is_running())
                {
                    // Get the MUTE_ON_STARTUP setting
                    settings_get(SETTING_MUTE_ON_STARTUP, &mute_on_startup);
                    logger_log_infof(LOGGER_CODE_INITIALISATION, "APP: Checking MUTE_ON_STARTUP %d --", mute_on_startup);
                }
                
               // Unmute inputs
                if (!input_control_data.mute_state_pre_switch && !mute_on_startup)
                {
                    uint8_t volume;
                    
                    volume_set(0, false);
                    
                    if (input_control_get_new_definition()->index != input_control_data.procesor_mode_input)
                    {
                        volume = input_control_data.volume_pre_switch;
                        volume_set(volume, true);
                        volume_set_balance(input_control_data.pre_procesor_mode_balance, true);
                    }
                    else
                    {
                        volume = input_control_data.procesor_mode_volume;
                        volume_set(volume, false);
                        logger_log_debugf(0, "VOLUME BALANCE DEFAULT A %d --", input_control_data.pre_procesor_mode_balance);
                        //volume_set_balance(input_control_data.pre_procesor_mode_balance, true);
                        volume_set_balance(VOLUME_DEFAULT_BALANCE, false);
                    }
                    
                    // Make sure the avatar is aware of the current volume
                    input_control_set_avatar_volume(volume);
   
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                    input_control_set_hdmi_arc_volume(volume);
#endif
                    
                    // For some reason this appears to be acting in a muted state
                    // when first switching so mute/unmute to clear
                    output_control_unmute_outputs();
                    
                    logger_log_debugf(0, "INPUT CONTROL: SETTING UP VOLUME %d %d  --", volume_is_muted(), volume_get());
                }
                else
                {
                    logger_log_debugf(0, "INPUT CONTROL: Setting back to pre-switch mute --");
                    volume_mute();
                }

                logger_log_debugf(0, "INPUT CONTROL Setting Active Input %s %d --", AUDIO_INPUT_STRINGS[input_control_get_current_definition()->index], input_control_get_current_definition()->index);
                
                input_control_update_input_settings();
                
                logger_log_debugf(0, "INPUT CONTROL DAC now applied - moving to running state --");
                input_control_data.state = INPUT_CONTROL_STATE_RUNNING;
                
                // This is to allow the system to detect that Input control has completed its setup
                app_context_module_running(APP_MODULE_INPUT_CONTROL);
            
                if (input_control_data.switch_callback != NULL)
                {
                    input_control_data.switch_callback(INPUT_CONTROL_SWITCH_STATE_POST);
                }
            }
 
            break;
        }
        
        default:
        {
            if (timers_has_expired(TIMER_INDEX_DPLL_LOCK))
            {
                // Note we're using the DPLL lock timer here just as a convenient 
                // way to check that the dac has not reset itself - it could be another timer
                // independent of the DPLL lock.
                if (avatar_handshake())
                {
                    // Don't check while the avatar is handshaking
                    input_control_check_dac_reset();
                }
                
                // Get the DAC input source then calculate the sample rate (if applicable)
                input_control_check_source_type();
        
                timers_reset(TIMER_INDEX_DPLL_LOCK);
            }
            
            INPUT_CONTROL_PREVIEW_INFO* preview_info = input_control_get_preview_info();
            if (preview_info->active)
            {
                if (timers_has_expired(TIMER_INDEX_INPUT_PREVIEW))
                {
                    input_control_select_input(preview_info->input);
                }
            }
            
            break;
        }
    }
    
//    input_control_digital_process();
    spdif_rx_process();
}

/**-----------------------------------------------------------------------------
* Reset the input control
* @param index
*/
void input_control_reset_input()
{
    logger_log_debugf(0, "INPUT CONTROL attempt input_control_reset_input %s --", AUDIO_INPUT_STRINGS[input_control_get_current_definition()->index]);

    if (input_control_data.state == INPUT_CONTROL_STATE_RUNNING)
    {
        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, 
                "INPUT CONTROL Switch reset %s %s --", 
                AUDIO_INPUT_STRINGS[input_control_get_current_definition()->index],
                INPUT_CONTROL_TYPE_STRINGS[input_control_get_current_definition()->type]);
        
        // Mute state indicates whether the device has been manually muted.  There
        // are cases where it has been muted by the software because of an invalid signal
        input_control_data.mute_state_pre_switch = input_control_in_mute_state(); //volume_is_muted();
        input_control_data.volume_pre_switch = volume_get_stored();
        
        // Set the pre-processor mode balance
        INPUT_CONTROL_DEFINITION* current_definition = input_control_get_current_definition();
        if ((current_definition != NULL) && (current_definition->index != input_control_data.procesor_mode_input))
        {
            // Store the pre processor mode volume and balance
            input_control_data.pre_procesor_mode_volume = volume_get_stored();
            input_control_data.pre_procesor_mode_balance = volume_get_balance();
        }     
        
        // Reset preview info - this is done for when the device wakes up
        input_control_get_preview_info()->active = false;
    
        // Update current definition
        input_control_set_new_definition(input_control_get_current_definition()->index);
        
        input_control_data.state = INPUT_CONTROL_STATE_MUTE_PRE_SWITCH;
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Incorrect state %d to select input %d ", input_control_data.state, input_control_get_current_definition()->index);
    }
}
    
/**-----------------------------------------------------------------------------
 * Select the input control identified by the index
 * Note: This should not check for whether the input is enabled, as we are using
 * a disabled channel for HDMI_ARC so that it doesn't appear in menus/selection
 * @param index
*/
void input_control_select_input(uint8_t index)
{
    logger_log_debugf(0, "INPUT CONTROL attempt input_control_select_input %d %s --", index, AUDIO_INPUT_STRINGS[index]);

    // Stop any preview
    input_control_deactivate_preview();
    
    // Get ref to new definition
    INPUT_CONTROL_DEFINITION* new_definition = input_control_get_definition(index);
    INPUT_CONTROL_DEFINITION* old_definition = input_control_get_current_definition();
        
    if ((input_control_data.state == INPUT_CONTROL_STATE_RUNNING) && 
        (old_definition != new_definition))
    {
        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, 
                "INPUT CONTROL Switching to %s %s --", 
                AUDIO_INPUT_STRINGS[index],
                INPUT_CONTROL_TYPE_STRINGS[new_definition->type]);
        
        input_control_data.mute_state_pre_switch = input_control_in_mute_state();//volume_is_muted();
        input_control_data.volume_pre_switch = volume_get_stored();
        
        // Set the pre-processor mode balance
        if ((old_definition != NULL) && (old_definition->index != input_control_data.procesor_mode_input))
        {
            // Store the pre processor mode volume and balance
            input_control_data.pre_procesor_mode_volume = volume_get_stored();
            input_control_data.pre_procesor_mode_balance = volume_get_balance();
        }     
    
        // Update new definition
        input_control_set_new_definition(new_definition->index);
        
        input_control_data.state = INPUT_CONTROL_STATE_MUTE_PRE_SWITCH;
    }
    else
    {
        if (old_definition != new_definition)
        {
            logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Incorrect state %d to select input %d ", input_control_data.state, index);
        }
        else
        {
            app_commands_display_default_screen(NULL);
        }
    }
}
   
/**-----------------------------------------------------------------------------
 * Sets the avatar volume
 */
void input_control_set_avatar_balance(uint8_t balance)
{
    // WE need to map the balance from +/-12 to +127/-128
    float mapped_balance = (((float)balance - (float)VOLUME_DEFAULT_BALANCE) * 256.0) / 24.0;
    mapped_balance = (int8_t)mapped_balance != -128 ? mapped_balance : 127;
    mapped_balance = balance == 38 ? -128 : mapped_balance;
    
//    float mapped_balance = (((float)balance - (float)VOLUME_DEFAULT_BALANCE) * 100.0) / 25.0;
//    mapped_balance = mapped_balance > 0.0 ? ceil(mapped_balance) : ceil(mapped_balance);
//    mapped_balance *= 2.56;

    logger_log_debugf(0, "AVATAR Setting (Mapped) BALANCE %d %d --", balance, (int8_t)mapped_balance);

    avatar_send_balance((uint8_t)mapped_balance);
}
   
/**-----------------------------------------------------------------------------
 * Sets the avatar volume
 */
void input_control_set_avatar_volume(uint8_t volume)
{
    logger_log_debugf(0, "AVATAR Setting Volume %d ", volume);
    //avatar_set_volume(volume);  
    
// BUG 3737 - Removed code below
//    // Google Cast spec 5.5 implies that we need to re-map the range sent to 
//    // the avatar based on the MAXIMUM_VOLUME setting
// 20201022 - re-added following email from Pete to Matjaz 22/10/2020 10:53
    uint16_t mapped_volume = volume * VOLUME_MAX;
    mapped_volume = min(mapped_volume / volume_get_maximum_net_volume(), VOLUME_MAX);

    logger_log_debugf(0, "AVATAR Setting (Mapped) VOLUME %d %d --", volume, mapped_volume);

    avatar_set_volume((uint8_t)mapped_volume);
}
   
/**-----------------------------------------------------------------------------
 * Sets the HDMI_ARC volume
 */
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  

void input_control_set_hdmi_arc_volume(uint8_t volume)
{
    //logger_log_debugf(0, "AVATAR Setting Volume %d ", volume);
    if (app_commands_hdmi_arc_is_active())
    {
//        // Google Cast spec 5.5 implies that we need to re-map the range sent to 
//        // the avatar based on the MAXIMUM_VOLUME setting
//        uint16_t mapped_volume = volume * VOLUME_MAX;
//        mapped_volume = min(mapped_volume / volume_get_maximum_volume(), VOLUME_MAX);
//
//        logger_log_debugf(0, "HDMI ARC Setting (Mapped) Volume %d %d ", volume, mapped_volume);

        hdmi_arc_send_volume_message(volume);
    }
}

#endif // HDMI_ARC_ACTIVE

/**-----------------------------------------------------------------------------
 * Sets the index of the current input designated for Processor mode
 * @param index - The index of the input to set for processor mode 
 */
void input_control_set_processor_mode_input(uint8_t index)
{
    if ((index >= 0) && (index < INPUT_CONTROL_NUM_SETTINGS))
    {
        input_control_data.procesor_mode_input = index;
        
        //logger_log_command(APP_COMMANDS_PROCESSOR_MODE_INPUT, &input_control_data.procesor_mode_input, 1);
        
        settings_set(SETTING_PROCESSOR_MODE_INPUT, &input_control_data.procesor_mode_input);
    }
}

/**-----------------------------------------------------------------------------
 * Sets current volume for Processor mode
 * @volume - The new processsor mode volume 
 */
void input_control_set_processor_mode_volume(uint8_t volume)
{
    if ((volume >= 0) && (volume < 100))
    {
        if (input_control_data.procesor_mode_volume != volume)
        {
            input_control_data.procesor_mode_volume = volume;
        
            // logger_log_command(APP_COMMANDS_PROCESSOR_MODE_VOLUME, &input_control_data.processor_mode_volume, 1);
        
            // Check if we are currently in processor mode and adjust the volume if necessary
            if ((input_control_get_current_definition() != NULL) &&
                (input_control_get_current_definition()->index == input_control_data.procesor_mode_input))
            {
                volume_set(input_control_data.procesor_mode_volume, false);
            }
        
            settings_set(SETTING_PROCESSOR_MODE_VOLUME, &input_control_data.procesor_mode_volume);
        }
    }
}
   
/**-----------------------------------------------------------------------------
 * Sets a callback that will be used to inform when the input is switching
 */
void input_control_set_switch_callback(INPUT_CONTROL_SWITCH_CALLBACK callback)
{
    input_control_data.switch_callback = callback;
}
    
/**-----------------------------------------------------------------------------
 * Starts input control processing - performs initialisation based on settings
 */
void input_control_start()
{
    uint8_t input;
    settings_get(SETTING_ACTIVE_INPUT, &input);

    // Check the input is still active
    INPUT_CONTROL_DEFINITION* current_definition = input_control_get_definition(input);
    if ((current_definition == NULL) || (!current_definition->enabled))
    {
        input = input_control_get_previous_input();
    }   
   
    logger_log_debugf(0, "INPUT CONTROL Start %d --", input);
    
    settings_get(SETTING_PROCESSOR_MODE_INPUT, &input_control_data.procesor_mode_input);
    settings_get(SETTING_PROCESSOR_MODE_VOLUME, &input_control_data.procesor_mode_volume);
    
    // Set the current definition to one thats not the selected input to force a change in input_control_select_input
    input_control_set_current_definition(input == AUDIO_INPUT_ALTERNATIVE ? AUDIO_INPUT_DEFAULT : AUDIO_INPUT_ALTERNATIVE);
    input_control_select_input(input);
}
    
/**-----------------------------------------------------------------------------
 * Starts input control configuration process - performs initialisation based on settings
 */
void input_control_start_configuration()
{
    input_control_data.state = INPUT_CONTROL_STATE_CONFIGURING;
}
    
/**-----------------------------------------------------------------------------
 * This actually switches the SPDIF MUX to a channel with the supplied input type
 */
void input_control_unmute_inputs(INPUT_CONTROL_TYPE type)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF MUX UNMUTING INPUT %d --", type);
    
    INPUT_CONTROL_DEFINITION def;
    def.type = type;
    input_control_select_mux_input(&def);
}

/* *****************************************************************************
 End of File
 */
