/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/input_control_dac.h"
#include "app_config.h"
#include "utilities/i2c.h"
#include "utilities/input_control_defs.h"
#include "utilities/logger/logger.h"
#include "utilities/mutes.h"
#include "utilities/settings.h"
#include "utilities/timers.h"
//#include "utilities/volume.h"
#include <math.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! The maximum number of registers used in the sample rate calculation
#define SAMPLERATE_MAX_REGISTERS    4

//! Audio DAC Master clock frequency (for sample rate calculation))
// Note: MCLK is defined by a crystal that is on the DAC board
// In this case its set to 50MHz
#define AUDIO_MCLK_FREQUENCY        50000000 

#define SAMPLE_RATE_DIVIDER         4294967296 // 2^32

#define VOLUME_REGISTER_COUNT       2

//! Define the number of filter coefficients
#define INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_STAGE1_COUNT   128
#define INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_STAGE2_COUNT   16
#define INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_COUNT   INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_STAGE1_COUNT + INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_STAGE2_COUNT

/**
 * Global data
 */
typedef struct
{
    
    //! I2C Client Id for identifying the module on the I2C bus
    I2C_CLIENT_ID i2c_client_id;
    
    // Indicates whether the dpll lock is on or not
    bool dpll_lock;
    
    //! The last calculated digital sample rate
    SAMPLE_RATE sample_rate;
    
    //! Defines the address of the register used to setup the custom filter.  
    //  It assumes that 4 registers will be used.
    INPUT_CONTROL_I2C_REQUEST custom_filter_request[4];
    
    //! The request structure used to enable/disable use of the custom filter and also editing of the filter
    INPUT_CONTROL_I2C_REQUEST custom_filter_enable;
    
    //! Used during the setup of a filter coefficients
    uint8_t applied_coefficient_count;
    
    //! The coefficient s used for the fiter - the high byte indicates a first or second stage filter
    uint32_t custom_filter_coefficients[INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_COUNT];
    
    //! The request structure used to query the dac volume - used to decide whether the 
    // dac_has reset itself.
    INPUT_CONTROL_I2C_REQUEST dac_reset;
    
    //! An array of setting values 
    INPUT_CONTROL_I2C_REQUEST dac_settings[INPUT_CONTROL_NUM_DAC_SETTING];
    
    //! Keeps track of the number of settings that have been added 
    //  will be < INPUT_CONTROL_NUM_DAC_SETTING
    uint8_t dac_setting_count;
    
    //! Keeps track of the number of settings that have been applied
    //  will be <= dac_setting_count
    uint8_t dac_applied_setting_count;
    
    //! switch to determine whether settings have been applied (actually the process started))
    bool dac_settings_applied;
    
    //! Keeps track of the state of the DAC reset.  On the SA20 this was done by
    //  reading the Reset pin, but it looks like on the PICMZ the pin (an output)
    //  always reads as low (in reset)
    bool in_reset;
    
    //! The request structure used to switch inputs
    INPUT_CONTROL_I2C_REQUEST source_type;
    
    //! The request structure used to switch inputs
    //INPUT_CONTROL_I2C_REQUEST switch_request;
    
    //! The request structure used to gather data for the sample rate calculation
    INPUT_CONTROL_I2C_REQUEST sample_rate_request[SAMPLERATE_MAX_REGISTERS];
    
    //! The actual number of sample rate request registers used.
    uint8_t sample_rate_request_count;
    
    //! The buffer used to hold ther dpll_num returned from the sample rate register queries
    uint32_t sample_rate_buffer;
    
    //! The request structure used to switch filters
    INPUT_CONTROL_I2C_REQUEST filter;
   
    //! The currently active filter type
    FILTER_TYPE filter_type;
    
    //! The number of filter types in use
    int num_filter_types;
    
    //! Descriptions for the filter types
    char* filter_type_descs[MAXIMUM_FILTER_TYPES];
    
    //! Values of bytes 7-4 of the filter type register (7)
    int filter_type_register[MAXIMUM_FILTER_TYPES];
    
    //! Callback to apply configuration for the SPDIF MUX
    INPUT_CONTROL_DAC_CONFIGURE_CALLBACK configure_callback;
    
// TODO REMOVE If using the standard register request works OK..    
    //! The request structure used to gather data for the sample rate calculation
   // INPUT_CONTROL_I2C_REQUEST volume_request[VOLUME_REGISTER_COUNT];
    
    bool checking_source_type;
    
} INPUT_CONTROL_DAC_DATA;

static INPUT_CONTROL_DAC_DATA input_control_dac_data =
{
    .in_reset = false,
    .configure_callback = NULL,
    .applied_coefficient_count = INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_COUNT,
    .checking_source_type = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward Definitions
static bool input_control_dac_is_reset();
static INPUT_CONTROL_I2C_REQUEST* input_control_find_definition_for_address(uint8_t address);
static void input_control_process_dpll_lock_change(bool dpll_lock);
static void input_control_select_filter();

/**-----------------------------------------------------------------------------
 * Initiates call to DAC to get sample rate
 */
static void input_control_calculate_dac_sample_rate()
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL input_control_calculate_dac_sample_rate ");
    if (!input_control_dac_is_reset())
    {
        int x;
        for (x = 0; x < input_control_dac_data.sample_rate_request_count; x++)
        {
            // Note: address should have been defined
            input_control_dac_data.sample_rate_request[x].actioned = false;
            input_control_dac_data.sample_rate_request[x].value = 0x00;
            input_control_dac_data.sample_rate_request[x].read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;

#if INPUT_CONTROL_I2C_DAC_ENABLED 

            // Trigger a request for the dpll num value
            if (i2c_read(input_control_dac_data.i2c_client_id, 
                         &input_control_dac_data.sample_rate_request[x].address,  
                         sizeof(uint8_t), 
                         &input_control_dac_data.sample_rate_request[x].value,  
                         sizeof(uint8_t), 
                         &input_control_dac_data.sample_rate_request[x],
                         true))
            {
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL sample rate error --");
            }

#endif  

        }
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * A separate call is made for each register read for the address
 * @param request
 */
static void input_control_calculate_dac_sample_rate_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL input_control_calculate_dac_sample_rate_callback %d ", request->index);
     
    // We can assume that the registers are being read in the ascending order??
    switch (request->read_state)
    {
        case INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS:
        {
            //logger_log_info(LOGGER_CODE_INPUT_CONTROL, "Input Control INITIATE RECEIVE");
            request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA;
            
            break;
        }
        
        case INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA:
        {
            uint32_t value = request->value;
            value = (value << (8 * request->index));
            
            switch (request->index)
            {
                case 0:
                    input_control_dac_data.sample_rate_buffer = value;
                    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL %d %d %d %d ", request->index, input_control_dac_data.sample_rate_buffer, value, request->value);
                    break;
                    
                default:
                {
                    input_control_dac_data.sample_rate_buffer += value;
                    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL %d %d %d %d ", request->index, input_control_dac_data.sample_rate_buffer, value, request->value);
                    //logger_process();
                    
                    if (request->index == input_control_dac_data.sample_rate_request_count - 1)
                    {
                        // The sample rate calculation for the ESS9038 is
                        // fsr = dpll_num * MCLK / 2^32
                        // Note: MCLK is defined by a crystal that is on the DAC board
                        // In this case its set to 100MHz
                        double divider = (double)AUDIO_MCLK_FREQUENCY / (double)SAMPLE_RATE_DIVIDER;
                        uint32_t sample_rate = (input_control_dac_data.sample_rate_buffer * divider); 
                        //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL new sample rate %d --", sample_rate);
                        
                        SAMPLE_RATE old_sample_rate = input_control_dac_data.sample_rate;
                        input_control_dac_data.sample_rate = SAMPLE_RATE_0;
                        
                        // Identify the enumerated sample rate from the rate calculated
                        int x;
                        for (x = 0; x < SAMPLE_RATE_MAX; x++)
                        {
                            if ((sample_rate > SAMPLE_RATE_VALUES[x] * 0.95) && (sample_rate < SAMPLE_RATE_VALUES[x] * 1.05))
                            {
                                input_control_dac_data.sample_rate = (SAMPLE_RATE)x;
                                
                                break;
                            }
                        }              
                                
                        // If the sample rate has changed send a sample rate change log message
                        if (old_sample_rate != input_control_dac_data.sample_rate)
                        {
                            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Sample Rate Comparison %d %d", old_sample_rate, input_control_dac_data.sample_rate);
                            logger_process();
                            uint8_t value = SAMPLE_RATE_GETSTATUS_VALUES[input_control_dac_data.sample_rate];
                            logger_log_command(APP_COMMANDS_SAMPLE_RATE, &value, 1);

                            //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Sample Rate set to %s (reg 68: %d)", SAMPLE_RATE_STRINGS[input_control_data.sample_rate], request->value);
                            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Sample Rate set to %d (detected: %d) --", SAMPLE_RATE_VALUES[input_control_dac_data.sample_rate], sample_rate);
                            logger_process();
                        }
                    }
                    break;
                }
            }
            
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * A separate call is made for each register read for the address
 * @param request
 */
static void input_control_check_dac_reset_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC input_control_check_dac_reset_callback %d ", request->index);
     
    switch (request->read_state)
    {
        case INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS:
        {
            //logger_log_info(LOGGER_CODE_INPUT_CONTROL, "Input Control INITIATE RECEIVE");
            request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA;
            
            break;
        }
        
        case INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA:
        {
            // We are reading register 15 to determine whether the volume has been reset back to 
            // 0.  This depends on the volume being adjusted for the dac (which it is in both
            // the SA10 and SA20)
            //logger_log_debugf(223, "INPUT CONTROL Checking DAC reset");
            
            // Find dac setting corresponding to the dac_reset address
            // If null we're not using that setting so cant be used for comparison
            INPUT_CONTROL_I2C_REQUEST* setting = input_control_find_definition_for_address(input_control_dac_data.dac_reset.address);
            if (setting != NULL)
            {
                // Compare the appropriate part of the register values to see if they've changed
                // Get the current value of the register
                uint8_t current_value = request->value;
                
                // adjust for the bit values
                current_value = (current_value << (7 - input_control_dac_data.dac_reset.bit_high));
                current_value = (current_value >> (7 - input_control_dac_data.dac_reset.bit_high + input_control_dac_data.dac_reset.bit_low));

                // Get the current value of the register
                uint8_t setting_value = setting->value;
                
                // adjust for the bit values
                setting_value = (setting_value << (7 - input_control_dac_data.dac_reset.bit_high));
                setting_value = (setting_value >> (7 - input_control_dac_data.dac_reset.bit_high + input_control_dac_data.dac_reset.bit_low));

                //logger_log_debugf(223, "INPUT CONTROL DAC Reset Comparing current with default %d %d --", current_value, setting_value);
                
                // Note - the default for the DAC is 0x80 - this test only works if our setting is something else (which is 0x08 (-4dB) at the moment))
                if (current_value != setting_value)
                {
                    logger_log_debugf(223, "INPUT CONTROL DAC Reset Comparing current with default %d %d --", current_value, setting_value);
                    
                    //logger_log_debugf(223, "INPUT CONTROL check DAC Reset - reset identified ");
                    
                    
                    // Check were not in the process of applying settings
                    if (input_control_dac_data.dac_applied_setting_count == input_control_dac_data.dac_setting_count)
                    {
                        if (current_value != 0)
                        {
                            logger_log_debugf(223, "INPUT CONTROL DAC Reset RESETTING %d %d  --", current_value, input_control_dac_data.dac_applied_setting_count);
                            input_control_apply_dac_settings(); 
                            input_control_select_filter();
                            input_control_reset_input();
                        }
                        else
                        {
                            // This should be the scenario where the DAC has gone AWOL and we're trying to bring it back
                            // At the moment this is not working and it just loops on that
                            logger_log_debugf(223, "INPUT CONTROL DAC Reset TRYING RESET %d %d  --", current_value, input_control_dac_data.dac_applied_setting_count);
                            
                            //input_control_calculate_spdif_rxx_sample_rate();
                            
                            input_control_dac_reset(INPUT_CONTROL_SWITCH_OFF);
                            timers_delay_ms(1);
                            input_control_dac_reset(INPUT_CONTROL_SWITCH_ON);
                        }
                    }
                    else
                    {
                        logger_log_debugf(223, "INPUT CONTROL DAC Reset IGNORED %d %d --", input_control_dac_data.dac_applied_setting_count, input_control_dac_data.dac_setting_count);

                    }
                }
            }
            else
            {
                logger_log_warningf(0, "INPUT CONTROL Attempting to use register %d for DAC Reset check when setting always default ", input_control_dac_data.dac_reset.address);
            }
 
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * A separate call is made for each register read for the address
 * @param request
 */
static void input_control_check_source_type_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL input_control_check_source_type_callback %d ", request->index);
     
    switch (request->read_state)
    {
        case INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS:
        {
            //logger_log_info(LOGGER_CODE_INPUT_CONTROL, "Input Control INITIATE RECEIVE");
            request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA;
            
            break;
        }
        
        case INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA:
        {
            // We are reading register 70 and comparing bit 1 which contains information
            // on whether the signal is audio or data (1 = data))
            uint32_t value = request->value;
           
            if ((value & 0x02) == 0x02)
            {
                input_control_dac_data.sample_rate = SAMPLE_RATE_NOTSUPPORTED;
                //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC check sample rate input NOT supported --");
            }
            else
            {
                input_control_calculate_dac_sample_rate();
            }

            input_control_dac_data.checking_source_type = false;
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Enables / Disables the custom filter
 * This is used to do 2 things:
 * - Enable use of filter - enable_write = false
 * - Enable writing filter - enable_write = false
 * @param on
 */
static void input_control_custom_filter_enable(bool switch_on, bool enable_write)
{
    INPUT_CONTROL_I2C_REQUEST* filter = &input_control_dac_data.custom_filter_enable;
    filter->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;
    
    // Try and find a default setting with this address
    uint8_t new_value = INPUT_CONTROL_CUSTOM_FILTER_ENABLE_REGISTER_DEFAULT;  // This is the default specified by pete and should be overwritten by the value set in app_initialise....

    // Handle enabling the filter
    if (enable_write)
    {
        // Setting up filter for writing
        if (switch_on)
        {
            new_value |= 0x02;
        }
    }
    else
    {
        // Setting up filter for enabling
        if (switch_on)
        {
            new_value |= 0x01;
        }
    }
    
    filter->value = new_value;
 
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL enabling custom filter %d %d --", switch_on, enable_write);

#if INPUT_CONTROL_I2C_DAC_ENABLED   
 
    //Trigger a register update
    // Note makes the assumption that the address/value are the first two
    // elements in the INPUT_CONTROL_DAC_SETTING structure
    // Note DAC Settings MUST be applied before calling this! Otherwise it kills the processor for some reason!
    if (i2c_transmit(input_control_dac_data.i2c_client_id, 
                     filter,  
                     2 * sizeof(uint8_t), 
                     filter))
    {
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL custom filter enable error --");
    }
#endif  

}

/**-----------------------------------------------------------------------------
 * Used to get the state of the DAC reset switch
 * @param on
 */
static bool input_control_dac_is_reset()
{
    return input_control_dac_data.in_reset;
}

/**-----------------------------------------------------------------------------
 * Called when a setting has been applied
 * @param request
 */
static void input_control_dac_setting_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    input_control_dac_data.dac_applied_setting_count++;

//    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Reset Setting Callback %d/%d ", 
//            input_control_dac_data.dac_applied_setting_count,
//            input_control_dac_data.dac_setting_count);
    
    // Check if all settings have been applied, and if so set input (actually left in
    // the applications initialise_input_control as its done after these and so will be queued)
    if (input_control_dac_data.dac_applied_setting_count == input_control_dac_data.dac_setting_count)
    {
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Initialised --");
        
        // Start monitoring dpll lock status
        // We're doing this about every 1 sec
        input_control_check_source_type();
        
        // Set the timer off to stop input_control_process from initiating more of these
        // before the first one is complete
        timers_reset(TIMER_INDEX_DPLL_LOCK);
    }
}

/**-----------------------------------------------------------------------------
 * Called when a setting has been applied
 * @param request
 */
static void input_control_dac_volume_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC VOLUME Callback %X %d --", 
            request->address, request->value);
    
    request->callback = NULL;
//    if (request->address == 0x10)
//    {
//                mutes_unmute(MUTE_INDEX_VOLUME);
//
//    }
}
    
/**-----------------------------------------------------------------------------
 * 
 * @param event
 * @param context
 */
static void input_control_dac_callback(DRV_I2C_BUFFER_EVENT event,
                                           I2C_REQUEST_TYPE request_type,
                                           uintptr_t context)
{
    //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Callback %d -- ", event);

    switch(event)
    {
        case DRV_I2C_BUFFER_EVENT_COMPLETE:
        {
            if (context != NULL)
            {
                // Look at the context
                INPUT_CONTROL_I2C_REQUEST* request = (INPUT_CONTROL_I2C_REQUEST*)context;
                request->actioned = true;

                //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Callback addr:0x%02x val:0x%02x ", request->address, request->value);

                // Invoke a callback proc
                if (request->callback != NULL)
                {
                    INPUT_CONTROL_I2C_REQUEST_CALLBACK callback = (INPUT_CONTROL_I2C_REQUEST_CALLBACK)request->callback;
                    callback(request);
                }
            }
            
            break;
        }
        
        case DRV_I2C_BUFFER_EVENT_ERROR:
        {
            if (context != NULL)
            {
                INPUT_CONTROL_I2C_REQUEST* request = (INPUT_CONTROL_I2C_REQUEST*)context;
                logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Callback Error addr:0x%02x val:0x%02x --", request->address, request->value);
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Callback Error unknown context --");
            }
            break;
        }
        
        default:
        {
            //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "Input Control Callback Other %d %d", (int)event, settings_data.state);
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Searches the dac settings to see if we have one that matches the supplied address
 * @return 
 */
static INPUT_CONTROL_I2C_REQUEST* input_control_find_definition_for_address(uint8_t address)
{
    INPUT_CONTROL_I2C_REQUEST* result = NULL;
            
    // Find dac setting corresponding to the dac_reset address
    int x;
    for (x = 0; x < input_control_dac_data.dac_setting_count; x++)
    {
        INPUT_CONTROL_I2C_REQUEST* setting = &input_control_dac_data.dac_settings[x];
        if (setting->address == address)
        {
            result = setting;
            break;
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Handles a change in the dpll_lock status
 * @param request
 */
static void input_control_process_dpll_lock_change(bool dpll_lock)
{
    // Monitor DPLL on DAC to see if the lock has changed
    // Get current lock value
    if (dpll_lock != input_control_dac_data.dpll_lock)
    {
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC_dpll_changed --");

        settings_suppress_save(true);

        if (dpll_lock)
        {
            // recalculate sample rate..
            input_control_check_source_type();
        }
        else
        {
            input_control_dac_data.sample_rate = SAMPLE_RATE_0;
        }

        settings_suppress_save(false);

        // Store the current value
        input_control_dac_data.dpll_lock = dpll_lock;
    }  
}

/**-----------------------------------------------------------------------------
 * Initiates call to DAC to check on the current source type (Audio/Non Audio)
 */
static void input_control_select_filter()
{
    INPUT_CONTROL_I2C_REQUEST* filter = &input_control_dac_data.filter;
    filter->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;
    
    // Try and find a setting with this address
    uint8_t new_value = input_control_dac_data.filter_type_register[input_control_dac_data.filter_type] << input_control_dac_data.filter.bit_low;
       
    // Check there isnt a default for this register, if there is then OR is 
    // (assumes default is only in the lower bits)..
    INPUT_CONTROL_I2C_REQUEST* setting = input_control_find_definition_for_address(input_control_dac_data.filter.address);
    if (setting != NULL)
    {
        new_value |= (setting->value & ((uint8_t)pow(2, input_control_dac_data.filter.bit_low) - 1));
    }

    filter->value = new_value;
 
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL selecting filter %s %d --", input_control_dac_data.filter_type_descs[input_control_dac_data.filter_type], filter->value);

#if INPUT_CONTROL_I2C_DAC_ENABLED   
 
    //Trigger a register update
    // Note makes the assumption that the address/value are the first two
    // elements in the INPUT_CONTROL_DAC_SETTING structure
    // Note DAC Settings MUST be applied before calling this! Otherwise it kills the processor for some reason!
    if (i2c_transmit(input_control_dac_data.i2c_client_id, 
                     filter,  
                     2 * sizeof(uint8_t), 
                     filter))
    {
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL select filter error --");
    }
#endif  
    
    // Switch off the custom filter
    // 20200928  - Removed as custom filters are not being used an suspect that this might
    // be why the thing keels over occasionally
    //input_control_custom_filter_enable(false, false);
}
   
/**-----------------------------------------------------------------------------
 * Sets the Data volume registers
 * @param volume1
 * @param volume2
 */
static void input_control_set_dac_volume_register(uint8_t address, uint8_t volume)
{
    if (!input_control_dac_is_reset())
    {

        // Check there isnt a default for this register, if there is then OR is 
        // (assumes default is only in the lower bits)..
        INPUT_CONTROL_I2C_REQUEST* request = input_control_find_definition_for_address(address);
        if (request != NULL)
        {
            request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;
            request->value = volume;
//            request->callback = input_control_dac_volume_callback;
        }

        //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Setting Volume %x %d --", request->address, request->value);
        //logger_log_infof(0, "INPUT CONTROL Setting Volume %x %d --", request->address, request->value);

#if INPUT_CONTROL_I2C_DAC_ENABLED   

    //Trigger a register update
    // Note makes the assumption that the address/value are the first two
    // elements in the INPUT_CONTROL_DAC_SETTING structure
    // Note DAC Settings MUST be applied before calling this! Otherwise it kills the processor for some reason!
    if (i2c_transmit(input_control_dac_data.i2c_client_id, 
                     request,  
                     2 * sizeof(uint8_t), 
                     request))
    {
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Set Volume error %x --", request->address);
    }
#endif 
    }
}

/**-----------------------------------------------------------------------------
 * Used to setup the filter coefficient based on input_control_dac_data.applied_coefficient_count 
 */
static void input_control_setup_next_filter_coefficient()
{
    if (input_control_dac_data.applied_coefficient_count < INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_COUNT)
    {
        int x;
        for (x = 0; x < 4; x++)
        {
            // 4 Filter registers used to set one coefficient - 0-Address, 1-3-Coefficient values
            INPUT_CONTROL_I2C_REQUEST* filter = &input_control_dac_data.custom_filter_request[x];
            filter->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;

            // get the coefficient value as a uint32_t.  The coefficient is Address + 24bit value
            uint32_t coefficient = input_control_dac_data.custom_filter_coefficients[input_control_dac_data.applied_coefficient_count];

            //get the value from the appropriate part of the coefficient
            filter->value = (coefficient >> (8 * (3 - x))) & 0xFF;

            //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Setting Filter coefficient %d %X %d %X --", input_control_dac_data.applied_coefficient_count, coefficient, x, filter->value);

#if INPUT_CONTROL_I2C_DAC_ENABLED   
 
            //Trigger a register update
            // Note makes the assumption that the address/value are the first two
            // elements in the INPUT_CONTROL_DAC_SETTING structure
            // Note DAC Settings MUST be applied before calling this! Otherwise it kills the processor for some reason!
            if (i2c_transmit(input_control_dac_data.i2c_client_id, 
                             filter,  
                             2 * sizeof(uint8_t), 
                             filter))
            {
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL filter coefficient error --");
            }
#endif  
        
        }    
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * A separate call is made for each register read for the address
 * @param request
 */
static void input_control_setup_next_filter_coefficient_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL input_control_setup_next_filter_coefficient_callback %d %d --", request->read_state, input_control_dac_data.applied_coefficient_count);
    input_control_dac_data.applied_coefficient_count++;
    if (input_control_dac_data.applied_coefficient_count < INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_COUNT)
    {
        input_control_setup_next_filter_coefficient();
    }
    else
    {
        // Turn Editing off
        // 20200928  - Removed as custom filters are not being used an suspect that this might
        // be why the thing keels over occasionally
        //input_control_custom_filter_enable(false, false);
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used to define DAC configuration values.
 * Note: This works at a register level rather than the individual settings
 * @param register_ident - the register identifier
 * @param value - th value to be written
 */
void input_control_add_dac_setting(uint8_t register_ident, uint8_t value)
{
    INPUT_CONTROL_I2C_REQUEST* setting = &input_control_dac_data.dac_settings[input_control_dac_data.dac_setting_count];
    setting->address = register_ident;
    setting->value = value;
    setting->callback = input_control_dac_setting_callback;
    
    input_control_dac_data.dac_setting_count++;
    
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Add DAC Setting addr: 0x%02X value: 0x%02X ", setting->address, setting->value);
}

/**-----------------------------------------------------------------------------
 * Starts the process of applying settings to the dac
 * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
 */
void input_control_apply_dac_settings()
{
    if ((!input_control_dac_data.dac_settings_applied) && (input_control_dac_data.dac_setting_count > 0))
    {
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Applying DAC Settings --");

        input_control_dac_data.dac_applied_setting_count = 0;

        // Make all settings unapplied
        int x;
        for (x = 0; x < input_control_dac_data.dac_setting_count; x++)
        {
            INPUT_CONTROL_I2C_REQUEST* setting = &input_control_dac_data.dac_settings[x];

#if INPUT_CONTROL_I2C_DAC_ENABLED   

            //Trigger a register update
            // Note makes the assumption that the address/value are the first two
            // elements in the INPUT_CONTROL_DAC_SETTING structure
            //logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL i2c_transmit ");
            if (i2c_transmit(input_control_dac_data.i2c_client_id, 
                             setting,  
                             2 * sizeof(uint8_t), 
                             setting))
            {
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL apply settings error --");
            }
#endif      
        }  
        
        input_control_dac_data.dac_settings_applied = true;
    }
}

/**-----------------------------------------------------------------------------
 * Initiates call to DAC to check on the current source type (Audio/Non Audio)
 */
void input_control_check_dac_reset()
{
    // 20190913 - Added a check for settings_applied as this was causing the unit to
    //            keel over when waking from sleep.
    if (!input_control_dac_is_reset() && input_control_dac_data.dac_settings_applied)
    {
        // Note: address should have been defined
        input_control_dac_data.dac_reset.actioned = false;
        input_control_dac_data.dac_reset.value = 0x00;
        input_control_dac_data.dac_reset.read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;

#if INPUT_CONTROL_I2C_DAC_ENABLED 

        // Trigger a request for the source type (Audio/Non Audio))
        if (i2c_read(input_control_dac_data.i2c_client_id, 
                     &input_control_dac_data.dac_reset.address,  
                     sizeof(uint8_t), 
                     &input_control_dac_data.dac_reset.value,  
                     sizeof(uint8_t), 
                     &input_control_dac_data.dac_reset,
                     true))
        {
        }
        else
        {
            logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL check dac reset error --");
        }
#endif  
    }
}

/**-----------------------------------------------------------------------------
 * Initiates call to DAC to check on the current source type (Audio/Non Audio)
 */
void input_control_check_source_type()
{
    //logger_log_debugf(0, "INPUT CONTROL CHECK SOURCE TYPE %d -- ", input_control_dac_is_reset());
    //logger_process();
    if (!input_control_dac_is_reset())
    {
        if (!input_control_dac_data.checking_source_type)
        {
            input_control_dac_data.checking_source_type = true;

            // Note: address should have been defined
            input_control_dac_data.source_type.actioned = false;
            input_control_dac_data.source_type.value = 0x00;
            input_control_dac_data.source_type.read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;

#if INPUT_CONTROL_I2C_DAC_ENABLED 

            // Trigger a request for the source type (Audio/Non Audio))
            if (i2c_read(input_control_dac_data.i2c_client_id, 
                         &input_control_dac_data.source_type.address,  
                         sizeof(uint8_t), 
                         &input_control_dac_data.source_type.value,  
                         sizeof(uint8_t), 
                         &input_control_dac_data.source_type,
                         true))
            {
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL source type error --");
            }
#endif  
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Configures the DAC module
 * @return 
 */
bool input_control_dac_configure()
{
    if (input_control_dac_data.configure_callback != NULL)
    {
        input_control_dac_data.configure_callback();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Configuration callback not specified --");
    }
}
    
/**-----------------------------------------------------------------------------
 * True when the DAC has been configured
 * @return 
 */
bool input_control_dac_configured()
{
    bool result = false;
    
    if (input_control_dac_data.dac_setting_count > 0)
    {
        result = input_control_dac_data.dac_applied_setting_count == input_control_dac_data.dac_setting_count;
    }
    
    // 20200928 - Removed as we ar no longer applying the custon filter
    //if (result)
    //{
    //    result = input_control_dac_data.applied_coefficient_count == INPUT_CONTROL_CUSTOM_FILTER_COEFFICIENT_COUNT;
    //}
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to change the state of the DAC reset switch
 * Switched ON during power up, OFF the rest of the time
 * @param on c- true - take the DAC out of reset, false - put it into reset
 */
void input_control_dac_reset(bool on)
{
#if INPUT_CONTROL_PINS_ENABLED
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC Reset - Switching %s %d --", on == true ? "On" : "Off", on);
    
    PLIB_PORTS_PinWrite(INPUT_CONTROL_DAC_RESET_PORTS_ID, 
                        INPUT_CONTROL_DAC_RESET_PORT_CHANNEL, 
                        INPUT_CONTROL_DAC_RESET_PIN,
                        on); 

#endif        
 
    // Its in reset when the pin is low.
    input_control_dac_data.in_reset = !on;
    
    if (on)
    {
        // Setup for a resend of the settings
        input_control_dac_data.dac_settings_applied = false;
    }
    
    logger_log_debugf(0, "INPUT CONTROL Reset %d --", input_control_dac_is_reset());
}
    
/**-----------------------------------------------------------------------------
 * Sets callbacks for the input_control_digital module
 * @param channel
 */
void input_control_dac_set_callbacks(INPUT_CONTROL_DAC_CONFIGURE_CALLBACK configure_callback)
{
    input_control_dac_data.configure_callback = configure_callback;
}
    
/**-----------------------------------------------------------------------------
 * Defines the address of the register used to setup the custom filter.  It assumes that
 * 4 registers will be used.
 * @param address
 */
void input_control_define_custom_filter_address(REGISTER_BITS_DEFINITION addresses[])
{
    int x;
    for (x = 0; x < 4; x++)
    {
        input_control_dac_data.custom_filter_request[x].address = addresses[x].address;
        input_control_dac_data.custom_filter_request[x].bit_low = addresses[x].bit_low;
        input_control_dac_data.custom_filter_request[x].bit_high = addresses[x].bit_high;
        input_control_dac_data.custom_filter_request[x].callback = NULL;
    }
    
    // Setup the callback on the last register
    input_control_dac_data.custom_filter_request[3].callback = input_control_setup_next_filter_coefficient_callback;
}

/**-----------------------------------------------------------------------------
 * Defines the address of the register used to manage the sample_rate
 * @param address
 */
void input_control_define_custom_filter_enable_address(REGISTER_BITS_DEFINITION address)
{
    input_control_dac_data.custom_filter_enable.address = address.address;
    
    // Note: bit_low/bit_high not used in input_control_custom_filter_enable
    input_control_dac_data.custom_filter_enable.bit_low = address.bit_low;
    input_control_dac_data.custom_filter_enable.bit_high = address.bit_high;
}

/**-----------------------------------------------------------------------------
 * Defines the address of the register used to manage the sample_rate
 * @param address
 */
void input_control_define_dac_reset_check_address(REGISTER_BITS_DEFINITION address)
{
    input_control_dac_data.dac_reset.address = address.address;
    input_control_dac_data.dac_reset.bit_low = address.bit_low;
    input_control_dac_data.dac_reset.bit_high = address.bit_high;
}
    
/**-----------------------------------------------------------------------------
 * Defines the address of the register used to manage the sample_rate calculation
 * based on the dpll_num
 * @param address
 */
void input_control_define_dac_sample_rate_address(REGISTER_BITS_DEFINITION addresses[], size_t num)
{
    input_control_dac_data.sample_rate_request_count = num;
    
    int x;
    for (x = 0; x < input_control_dac_data.sample_rate_request_count; x++)
    {
        input_control_dac_data.sample_rate_request[x].address = addresses[x].address;
        input_control_dac_data.sample_rate_request[x].bit_low = addresses[x].bit_low;
        input_control_dac_data.sample_rate_request[x].bit_high = addresses[x].bit_high;
    }
}
    
/**-----------------------------------------------------------------------------
 * Defines the address of the register used to manage the volume
 * @param address - expects a 2 element array
 */
void input_control_define_dac_volume_address(REGISTER_BITS_DEFINITION addresses[])
{
    int x;
    for (x = 0; x < VOLUME_REGISTER_COUNT; x++)
    {
//        input_control_dac_data.volume_request[x].address = addresses[x].address;
//        input_control_dac_data.volume_request[x].bit_low = addresses[x].bit_low;
//        input_control_dac_data.volume_request[x].bit_high = addresses[x].bit_high;
    }
}
    
/**-----------------------------------------------------------------------------
 * Defines the address of the register used to manage the current input
 * @param address
 */
void input_control_define_filter_address(REGISTER_BITS_DEFINITION address)
{
    input_control_dac_data.filter.address = address.address;
    input_control_dac_data.filter.bit_low = address.bit_low;
    input_control_dac_data.filter.bit_high = address.bit_high;
    input_control_dac_data.filter.callback = NULL;
}
     
/**-----------------------------------------------------------------------------
 * Defines the address of the register used to read the input source type (Audio / Non Audio)
 * @param address
 */
void input_control_define_source_type_address(REGISTER_BITS_DEFINITION address)
{
    input_control_dac_data.source_type.address = address.address;
    input_control_dac_data.source_type.bit_low = address.bit_low;
    input_control_dac_data.source_type.bit_high = address.bit_high;
}

/**-----------------------------------------------------------------------------
 * Initialise the input control module
 */
bool input_control_dac_initialise()
{
    bool result = true;
    
    input_control_dac_data.dpll_lock = false;
    input_control_dac_data.sample_rate = SAMPLE_RATE_0;
    input_control_dac_data.dac_setting_count = 0;
    input_control_dac_data.dac_settings_applied = false;
    
    input_control_dac_data.dac_reset.callback = input_control_check_dac_reset_callback;
    input_control_dac_data.source_type.callback = input_control_check_source_type_callback;
    input_control_dac_data.custom_filter_enable.callback = NULL;
    
    settings_get(SETTING_FILTER_TYPE, &input_control_dac_data.filter_type);
    //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Initial Filter Type will be %d --", input_control_dac_data.filter_type);
    
    // Sort out filters
    input_control_dac_data.filter_type_descs[0] = DAC_FILTER_STRINGS[0]; //"Linear Phase Fast";
    input_control_dac_data.filter_type_register[0] = 0x00;
    input_control_dac_data.filter_type_descs[1] = DAC_FILTER_STRINGS[1]; //"Linear Phase Slow";
    input_control_dac_data.filter_type_register[1] = 0x01;
    input_control_dac_data.filter_type_descs[2] = DAC_FILTER_STRINGS[2]; //"Minimum Phase Fast";
    input_control_dac_data.filter_type_register[2] = 0x02;
    input_control_dac_data.filter_type_descs[3] = DAC_FILTER_STRINGS[3]; //"Minimum Phase Slow";
    input_control_dac_data.filter_type_register[3] = 0x03;
    input_control_dac_data.filter_type_descs[4] = DAC_FILTER_STRINGS[4]; //"Brick Wall";
    input_control_dac_data.filter_type_register[4] = 0x07;
    input_control_dac_data.filter_type_descs[5] = DAC_FILTER_STRINGS[5]; //"Correlated Minimum Phase";
    input_control_dac_data.filter_type_register[5] = 0x06;
    input_control_dac_data.filter_type_descs[6] = DAC_FILTER_STRINGS[6]; //"Apodizing"; // Default
    input_control_dac_data.filter_type_register[6] = 0x04;
    input_control_dac_data.num_filter_types = 7;
    input_control_dac_data.filter_type = 6;
    
    int x;
    for (x = 0; x < SAMPLERATE_MAX_REGISTERS; x++)
    {
        input_control_dac_data.sample_rate_request[x].callback = input_control_calculate_dac_sample_rate_callback;
        input_control_dac_data.sample_rate_request[x].index = x;
    }
    input_control_dac_data.sample_rate_request_count = 0;
    
    // Custom MQA Filter Coefficients
    static const uint32_t filter_coefficients[] = {
        
        // Stage 1
        0x00000000,0x01000000,0x02000000,0x03000000,0x04000000,0x05000000,0x06000000,0x07000000,
        0x08000000,0x09000000,0x0A000000,0x0B000000,0x0C000000,0x0D000000,0x0E000000,0x0F000000,
        0x10000000,0x11000000,0x12000000,0x13000000,0x14000000,0x15000000,0x16000000,0x17000000,
        0x18000000,0x191B1C4E,0x1A537843,0x1B5B174C,0x1C2AF75D,0x1D0A6462,0x1E019060,0x1FFF6804,
        0x20000000,0x21000000,0x22000000,0x23000000,0x24000000,0x25000000,0x26000000,0x27000000,
        0x28000000,0x29000000,0x2A000000,0x2B000000,0x2C000000,0x2D000000,0x2E000000,0x2F000000,
        0x30000000,0x31000000,0x32000000,0x33000000,0x34000000,0x35000000,0x36000000,0x37000000,
        0x38000000,0x39000000,0x3A000000,0x3B000000,0x3C000000,0x3D000000,0x3E000000,0x3F000000,
        0x40000000,0x41000000,0x42000000,0x43000000,0x44000000,0x45000000,0x46000000,0x47000000,
        0x48000000,0x49000000,0x4A000000,0x4B000000,0x4C000000,0x4D000000,0x4E000000,0x4F000000,
        0x50000000,0x51000000,0x52000000,0x53000000,0x54000000,0x55000000,0x56000000,0x57000000,
        0x58000000,0x59000000,0x5A000000,0x5B000000,0x5C000000,0x5D000000,0x5E000000,0x5F000000,
        0x60000000,0x61000000,0x62000000,0x63000000,0x64000000,0x65000000,0x66000000,0x67000000,
        0x68000000,0x69000000,0x6A000000,0x6B000000,0x6C000000,0x6D000000,0x6E000000,0x6F000000,
        0x70000000,0x71000000,0x72000000,0x73000000,0x74000000,0x75000000,0x76000000,0x77000000,
        0x78000000,0x79000000,0x7A000000,0x7B000000,0x7C000000,0x7D000000,0x7E000000,0x7F000000,
        
        // Stage 2
        0x80000000,0x81000000,0x82000000,0x83000000,0x84000000,0x85000000,0x86000000,0x87000000,
        0x88000000,0x89000000,0x8A000000,0x8B000000,0x8C77CA8D,0x8D77CA8D,0x8E000000,0x8F000000,
    };
    
    memcpy(input_control_dac_data.custom_filter_coefficients, filter_coefficients, sizeof(filter_coefficients));
    
    i2c_initialise();

#if INPUT_CONTROL_I2C_DAC_ENABLED   
     
    input_control_dac_data.i2c_client_id = i2c_open(INPUT_CONTROL_DAC_I2C_MODULE, 
            INPUT_CONTROL_DAC_I2C_READ_ADDRESS,
            INPUT_CONTROL_DAC_I2C_WRITE_ADDRESS,
            input_control_dac_callback);
    
    if (I2C_INVALID_CLIENT_ID != input_control_dac_data.i2c_client_id)
    {
        // Initialise the DAC
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC I2C Open --");
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL DAC I2C Open Error --");
        result = false;
    }    
   
#endif
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Sets the current filter
 */
FILTER_TYPE input_control_get_filter()
{
    return input_control_dac_data.filter_type;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the logger level
 * @return 
 */
void input_control_get_filter_string(char* buffer, size_t size)
{
    char filter_name[30];
    
    sprintf(filter_name, "%s", input_control_dac_data.filter_type_descs[(int)input_control_get_filter()]);
    
    memset(buffer, 0, size);
    memcpy(buffer, filter_name, min(size, strlen(filter_name)));
}
    
/**-----------------------------------------------------------------------------
 * Returns the number of registered filters
 * @return 
 */
uint8_t input_control_get_num_filters()
{
    return input_control_dac_data.num_filter_types;
}
    
/**-----------------------------------------------------------------------------
 * Gets the sample rate of the current digital input 0 if analog
 * @return 
 */
SAMPLE_RATE input_control_get_dac_sample_rate()
{
    return input_control_dac_data.sample_rate; // Maybe should be an enumeration?
}
 

/**-----------------------------------------------------------------------------
 * Handles interrupt on the DAC GPIO1 which is configured via Register 8 to change
 * When the dpll_lock status changes
 * @param definition
 */
void input_control_handle_dac_interrupt()
{
#if INPUT_CONTROL_PINS_ENABLED
    
    // Not using this at the moment because the DAC does not appear to be raising the interrupts
    // If we do manage to adopt this method then the automatic polling can be removed.
    bool dpll_lock = PLIB_PORTS_PinGet(INPUT_CONTROL_DAC_GPIO_PORTS_ID, 
                        INPUT_CONTROL_DAC_GPIO_PORT_CHANNEL, 
                        INPUT_CONTROL_DAC_GPIO_PIN);
    
//    if (dpll_lock)
//    {
//        logger_log_debugf(0, "INTERRUPT: %d --", dpll_lock);
//    }
    input_control_process_dpll_lock_change(dpll_lock);
    
#endif  
}

/**-----------------------------------------------------------------------------
 * Mutes / Unmutes the DAC
 * @param on
 */
void input_control_mute_dac(bool on)
{
    // Note the MUTE register is the same as the Filter
    INPUT_CONTROL_I2C_REQUEST* filter = &input_control_dac_data.filter;
    filter->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;
 
    if (on)
    {
        //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Muting DAC --");
        filter->value |= 0x01; // Set the first bit
    }
    else
    {
        //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Un-muting  DAC --");
        filter->value &= ~1;
    }
 

#if INPUT_CONTROL_I2C_DAC_ENABLED   
 
    //Trigger a register update
    // Note makes the assumption that the address/value are the first two
    // elements in the INPUT_CONTROL_DAC_SETTING structure
    // Note DAC Settings MUST be applied before calling this! Otherwise it kills the processor for some reason!
    if (i2c_transmit(input_control_dac_data.i2c_client_id, 
                     filter,  
                     2 * sizeof(uint8_t), 
                     filter))
    {
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Un/mute DAC error --");
    }
#endif  
}

/**-----------------------------------------------------------------------------
 * Sets the Data volume registers
 * @param volume1
 * @param volume2
 */
void input_control_set_dac_volume(uint8_t volume1, uint8_t volume2)
{
    input_control_set_dac_volume_register(0x0F, volume1);
    input_control_set_dac_volume_register(0x10, volume2);
}

/**-----------------------------------------------------------------------------
 * Sets the current filter
 */
void input_control_set_custom_filter()
{
    // 20200515 - Not using the MQA Custom Filter so use the reccommended 
    logger_log_debugf(LOGGER_CODE_AVATAR, "INPUT CONTROL MQA Setting Filter Minimum Phase Slow --");
    input_control_set_filter(3); // Filter type Minimum Phase Slow

//TODO Put back if we decide to use Custom Filters - Search On CUSTOM_FILTER for other places where it has been removed    
//    // It appears that the custom filter uses a 2 stage filter so we need to set the 
//    // first stage to Linear Phase Fast according to MQA
//    logger_log_debugf(LOGGER_CODE_AVATAR, "INPUT CONTROL MQA Setting Filter Linear Phase Fast --");
//    input_control_set_filter(0); // Filter type Linear Phase Fast
//    
//    logger_log_debugf(LOGGER_CODE_AVATAR, "INPUT CONTROL MQA Enabling Custom Filter --");
//    input_control_custom_filter_enable(true, false);
}

/**-----------------------------------------------------------------------------
 * Sets the current filter
 */
void input_control_set_filter(FILTER_TYPE filter)
{
    input_control_dac_data.filter_type = min(filter, input_control_get_num_filters());
    
    input_control_select_filter();
}

/**-----------------------------------------------------------------------------
 * Loads the custom filter coefficients.  Once coefficient is loaded at a time, and 
 * completion kicks off the next coefficient
 */
void input_control_setup_custom_filter()
{
    logger_log_debugf(0, "INPUT CONTROL Setting up Custom Filter --");
    // Turn editing on
    // 20200928  - Removed as custom filters are not being used an suspect that this might
    // be why the thing keels over occasionally
    //input_control_custom_filter_enable(false, true);
    
    // Set the applied coefficient count to 0;
    input_control_dac_data.applied_coefficient_count = 0;
    
    // Kick off with first coefficient 
    // 20200928  - Removed as custom filters are not being used an suspect that this might
    // be why the thing keels over occasionally
    //input_control_setup_next_filter_coefficient();
 }

/**-----------------------------------------------------------------------------
 * Called after settings initialisation to setup the filter stored in the settings
 */
void input_control_setup_filter()
{
    FILTER_TYPE type;
    settings_get(SETTING_FILTER_TYPE, &type);

    if (type >= MAXIMUM_FILTER_TYPES)
    {
        settings_set(SETTING_FILTER_TYPE, &input_control_dac_data.filter_type);
        logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Setting Filter type %d --", input_control_dac_data.filter_type);
    }
    else
    {
        input_control_dac_data.filter_type = type;
    }
    
    input_control_select_filter();
}

/* *****************************************************************************
 End of File
 */
