/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/input_control_defs.h"
#include "app_config.h"
#include "app_context.h"
#include "utilities/timers.h"

#ifdef DISPLAY_MESSAGE_ACTIVE
#include "utilities/display_message.h"
#include "utilities/menu.h"
#endif

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Global data
 */
typedef struct
{
    //! Pointer to the currently active definition in the definitions array
    INPUT_CONTROL_DEFINITION* current_definition;
    
    //! Pointer to the definition in the definitions array to be switched to in a switch operation
    INPUT_CONTROL_DEFINITION* new_definition;
    
    //! Pointer to the previously active definition in the definitions array
    INPUT_CONTROL_DEFINITION* previous_definition;
    
    //! Array of input definitions
    //  Assumes that the input definition index can also be used as the ordinal 
    //  pos in the definitions array
    INPUT_CONTROL_DEFINITION definitions[INPUT_CONTROL_NUM_SETTINGS];
        
    //! Preview info
    INPUT_CONTROL_PREVIEW_INFO preview_info;

} INPUT_CONTROL_DEF_DATA;

static INPUT_CONTROL_DEF_DATA input_control_def_data =
{
    .current_definition = NULL,
    .new_definition = NULL,
    .previous_definition = NULL,
    .preview_info.active = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Deactivate the preview mode
 */
void input_control_deactivate_preview()
{
    // Set the PREVIEW timer to non-active
    input_control_def_data.preview_info.active = false;
    timers_suspend(TIMER_INDEX_INPUT_PREVIEW);
}
    
/**-----------------------------------------------------------------------------
 * Gets a pointer to the current definition
 * @param index
 * @return 
 */
INPUT_CONTROL_DEFINITION* input_control_get_current_definition()
{
    return input_control_def_data.current_definition;
}

/**-----------------------------------------------------------------------------
 * Gets the type of the currently selected input (Analogue, Digital, Avatar etc..)
 * @return 
 */
INPUT_CONTROL_TYPE input_control_get_current_input_type()
{
    INPUT_CONTROL_TYPE result = INPUT_CONTROL_TYPE_NONE;
    
    INPUT_CONTROL_DEFINITION* definition = input_control_get_current_definition();
    if (definition != NULL)
    {
        result = definition->type;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets a pointer to the supplied definition
 * @param index
 * @return 
 */
INPUT_CONTROL_DEFINITION* input_control_get_definition(uint8_t index)
{
    INPUT_CONTROL_DEFINITION* result = NULL;
    
    if (index < INPUT_CONTROL_NUM_SETTINGS)
    {
        result = &input_control_def_data.definitions[index];
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets a pointer to the current definition
 * @param index
 * @return 
 */
INPUT_CONTROL_DEFINITION* input_control_get_new_definition()
{
    return input_control_def_data.new_definition;
}

/**-----------------------------------------------------------------------------
 * Gets a pointer to the next definition relative to the index of the one supplied
 * @param index
 * @return 
 */
INPUT_CONTROL_DEFINITION* input_control_get_next_definition(uint8_t index, bool forwards)
{
    int pos = index;
    INPUT_CONTROL_DEFINITION* result = NULL;
    
    int x;
    for (x = 0; x < INPUT_CONTROL_NUM_SETTINGS; x++)
    {
        // Find next input to look at
        if (forwards)
        {
            pos++;
            if (pos == INPUT_CONTROL_NUM_SETTINGS)
            {
                pos = 1;    // Miss out on Input NONE
            }
        }
        else
        {
            pos--;
            if (pos == 0)
            {
                pos = INPUT_CONTROL_NUM_SETTINGS - 1;
            }
        }
        
        INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(pos);
        
        if (definition->enabled)
        {
            result = definition;
            logger_log_debugf(0, "INPUT input_control_get_next_definition FOUND %d %s --", pos, AUDIO_INPUT_STRINGS[definition->index]);
            break;
        }
        else
        {
            logger_log_debugf(0, "INPUT input_control_get_next_definition DISABLED %d %s --", pos, AUDIO_INPUT_STRINGS[definition->index]);
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the index of the currently selected input
 * @return 
 */
INPUT_CONTROL_PREVIEW_INFO* input_control_get_preview_info()
{
    return &input_control_def_data.preview_info;
}
    
/**-----------------------------------------------------------------------------
 * Gets a pointer to the current definition
 * @param index
 * @return 
 */
INPUT_CONTROL_DEFINITION* input_control_get_previous_definition()
{
    return input_control_def_data.previous_definition;
}

/**-----------------------------------------------------------------------------
 * Sets the current definition by index
 * @param index
 * @return 
 */
void input_control_set_current_definition(uint8_t index)
{
    // This was here just incase it ever gets set to 0 (None) the push it on otherwise the device wont load...
    // However It was being used  in inputr_control_startup to ensure that the 
    // input changes which messsed it up when the startup input was phono...
//    if (index == AUDIO_INPUT_NONE)
//    {
//        index = AUDIO_INPUT_PHONO;
//    }
    
    // Store the current definition as the previous definition
    input_control_def_data.previous_definition = input_control_def_data.current_definition;
    
    // Change the current definition
    input_control_def_data.current_definition = input_control_get_definition(index);

    logger_log_debugf(0, "INPUT CONTROL SET CURRENT %d %s --", index, AUDIO_INPUT_STRINGS[input_control_def_data.current_definition->index]);

    bool handle_preview = true;
#if APP_FEATURE_DIRAC == APP_FEATURE_ON
    // Check if we are calibration - if so leave the screen alone
    // otherwise display the default screen
    if (avatar_message_get_dirac_client_status() == AVATAR_DIRAC_CLIENT_STATUS_CONNECTED)
    {   
        logger_log_debugf(0, "INPUT CONTROL Suppressing display of preview screen --");
        handle_preview = false;
    }
#endif
    
    // BUG 3783 - Currently Playing screen appears before Powering Up screen when waking up from standby
    if (app_context_get_app_state() == APP_STATE_POWEREDDOWN)
    {
        handle_preview = false;
    }

    if (handle_preview)
    {
        // Update the preview info
        input_control_set_preview(index);

        // Set the PREVIEW timer to non-active
        input_control_deactivate_preview();
    }
}

/**-----------------------------------------------------------------------------
 * Sets the current definition by index
 * @param index
 * @return 
 */
void input_control_set_new_definition(uint8_t index)
{
    //logger_log_debugf(0, "INPUT CONTROL SET NEW %d %s --", index, AUDIO_INPUT_STRINGS[index]);
    input_control_def_data.new_definition = input_control_get_definition(index);
}
    
/**-----------------------------------------------------------------------------
 * Updates preview info with the supplied input
 */
void input_control_set_preview(uint8_t input)
{
    INPUT_CONTROL_PREVIEW_INFO* preview_info = &input_control_def_data.preview_info;
    preview_info->active = true;
    preview_info->input = input;
    preview_info->type = input_control_get_definition(input)->type;
    
    app_commands_display_default_screen(NULL);
    
    // Activate the timer if its not already active
    timers_activate(TIMER_INDEX_INPUT_PREVIEW);
    
    // Reset the timer
    timers_reset(TIMER_INDEX_INPUT_PREVIEW);
                
#ifdef DISPLAY_MESSAGE_ACTIVE
    if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_CURRENTLY_PLAYING)
    {
        display_message_build_screen_data(APP_COMMAND_MENU_CURRENTLY_PLAYING);
    }
    else if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_INPUT_SELECTION)
    {
        display_message_build_screen_data(APP_COMMAND_MENU_INPUT_SELECTION);
    }
#endif
}

/* *****************************************************************************
 End of File
 */
