/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/input_control_i2c.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**-----------------------------------------------------------------------------
 * Used to define an Input Control I2C setting
 * Note: This works at a register level rather than the individual settings
 * @param register_ident - the register identifier
 * @param value - th value to be written
 */
void input_control_i2c_add_setting(INPUT_CONTROL_I2C_REQUEST* setting, uint8_t address, uint8_t value, void* callback)
{
    setting->address = address;
    setting->value = value;
    setting->callback = callback;
}

/**-----------------------------------------------------------------------------
 * Starts the process of applying settings to the SPDIF Mux
 * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
 */
void input_control_i2c_apply_setting(I2C_CLIENT_ID i2c_client_id, INPUT_CONTROL_I2C_REQUEST* setting)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Applying I2C Setting 0x%2.2X 0x%2.2X --", setting->address, setting->value);
    logger_process();

    //Trigger a register update
    // Note makes the assumption that the address/value are the first two
    // elements in the INPUT_CONTROL_DAC_SETTING structure
    if (i2c_transmit(i2c_client_id, 
                     setting,  
                     2 * sizeof(uint8_t), 
                     setting))
    {
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Apply I2C settings error");
    }
}

/**-----------------------------------------------------------------------------
 * Starts the process of applying settings to an Inpt Control settings array
 * Note the define I2C_REQUEST_QUEUE_SIZE must be at least as big as this
 * @param setting - a pointer to the first setting in the array
 * @param count - the number of settings in the aray
 */
void input_control_i2c_apply_settings(I2C_CLIENT_ID i2c_client_id, INPUT_CONTROL_I2C_REQUEST* setting, uint8_t count)
{
    logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL Applying I2C Settings --");
    logger_process();
    int x;
    for (x = 0; x < count; x++)
    {
        input_control_i2c_apply_setting(i2c_client_id, setting);
        
        // Move to the next setting
        setting++;
    }  
}

/**-----------------------------------------------------------------------------
 * Generic handler for Input Control I2C request callbacks
 * @param event
 * @param context
 */
void input_control_i2c_callback(DRV_I2C_BUFFER_EVENT event,
                                I2C_REQUEST_TYPE request_type,
                                uintptr_t context)
{
    //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL I2C Callback %d -- ", event);

    switch(event)
    {
        case DRV_I2C_BUFFER_EVENT_COMPLETE:
        {
            if (context != NULL)
            {
                // Look at the context
                INPUT_CONTROL_I2C_REQUEST* request = (INPUT_CONTROL_I2C_REQUEST*)context;
                request->actioned = true;

                //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL I2C Callback addr:%02x val:0x%02x", request->address, request->value);

                // Invoke a callback proc
                if (request->callback != NULL)
                {
                    INPUT_CONTROL_I2C_REQUEST_CALLBACK callback = (INPUT_CONTROL_I2C_REQUEST_CALLBACK)request->callback;
                    callback(request);
                }               
            }
            
            break;
        }
        
        case DRV_I2C_BUFFER_EVENT_ERROR:
        {
            if (context != NULL)
            {
                INPUT_CONTROL_I2C_REQUEST* request = (INPUT_CONTROL_I2C_REQUEST*)context;
                logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL I2C Callback Error addr:%02x val:0x%02x", request->address, request->value);
            }
            else
            {
                logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL I2C Callback Error unknown context");
            }
            break;
        }
        
        default:
        {
            //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL I2C Callback Other %d %d", (int)event, settings_data.state);
            break;
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * For use in I2C read callbacks.  This checks the request state, setting the state to 
 * INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA if it is INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS
 * (returning false) an returning true if state is INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA
 * The idea of this is to remove the need for a switch statement in the callback -
 * with this you just need an if (input_control_i2c_callback_has_data()) then 
 * handle the data.
 */
bool input_control_i2c_callback_has_data(INPUT_CONTROL_I2C_REQUEST* request)
{
    bool result = false;
    
    switch (request->read_state)
    {
        case INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS:
        {
            // This is called when the address has been sent.  We now need to request
            // the data
            request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA;
            
            break;
        }
        
        case INPUT_CONTROL_I2C_REQUEST_STATE_RECEIVE_DATA:
        {
            // We have data - set the state to inactive.  Normally we would do 
            // this after processing but it should be OK here. 
            //request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE;
            
            result = true;
            
            break;
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * MUST be called at the end of the callback processing where input_control_i2c_callback_has_data
 * is in use..
 */
bool input_control_i2c_callback_release(INPUT_CONTROL_I2C_REQUEST* request)
{
    request->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE;
}
    
/**-----------------------------------------------------------------------------
 * Defines an I2C register used for a specific task
 * @param address
 */
void input_control_i2c_define_register(INPUT_CONTROL_I2C_REQUEST* register_request, REGISTER_BITS_DEFINITION definition)
{
    register_request->address = definition.address;
    register_request->bit_low = definition.bit_low;
    register_request->bit_high = definition.bit_high;
}

/**-----------------------------------------------------------------------------
 * Searches the settings to see if we have one that matches the supplied address
 * @param definition - pointer to the first definition in the array
 * @param count - the number of definitions in the array
 * @param address - the address to match
 * @return 
 */
INPUT_CONTROL_I2C_REQUEST* input_control_i2c_definition_for_address(INPUT_CONTROL_I2C_REQUEST* definitions, uint8_t count, uint8_t address)
{
    INPUT_CONTROL_I2C_REQUEST* result = NULL;
            
    // Find definition corresponding to the definitions address
    int x;
    for (x = 0; x < count; x++)
    {
        if (definitions->address == address)
        {
            result = definitions;
            break;
        }
        
        definitions++;
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Starts the process of reading a setting from the SPDIF Mux
 */
void input_control_i2c_read_setting(I2C_CLIENT_ID i2c_client_id, INPUT_CONTROL_I2C_REQUEST* setting)
{
    if (setting->read_state == INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE)
    {
        // Note: address should have been defined
        setting->actioned = false;
        setting->value = 0x00;
        setting->read_state = INPUT_CONTROL_I2C_REQUEST_STATE_TRANSMIT_ADDRESS;

        // Trigger a request for the source type (Audio/Non Audio))
        if (i2c_read(i2c_client_id, 
                     &setting->address,  
                     sizeof(uint8_t), 
                     &setting->value,  
                     sizeof(uint8_t), 
                     setting,
                     true))
        {
        }
        else
        {
            logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL I2C Read Setting error");
        }
    }
}

/* *****************************************************************************
 End of File
 */
