/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    logger.c

  @Summary
 Handles simple application logging for the utility libraries

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/logger/logger.h"
#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include "app_config.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define LOGGER_BUFFER_LEN   80

#ifdef LOGGER_PROCESS_ASYNC

#ifndef LOGGER_MESSAGE_QUEUE_SIZE
#define LOGGER_MESSAGE_QUEUE_SIZE   100
#endif // LOGGER_MESSAGE_QUEUE_SIZE

#endif // ifdef LOGGER_PROCESS_ASYNC

typedef struct
{
    //! The callback to be invoked by the application when a logger message is raised
    LOGGER_CALLBACK logger_callback;

    //! The maximum log level
    LOGGER_TYPE logger_level;

    //! The log level before logging was suppressed
    LOGGER_TYPE prev_logger_level;

    //! determine whether the logger is currently active to stop recursive method calls
    bool logger_active;

    bool suppress_command_logging;

#ifdef LOGGER_PROCESS_ASYNC

    volatile uint8_t logger_queue_read_pos;
    
    volatile uint8_t logger_queue_write_pos;
    
    LOGGER_MESSAGE logger_queue[LOGGER_MESSAGE_QUEUE_SIZE];

    char logger_queue_buffers[LOGGER_MESSAGE_QUEUE_SIZE][LOGGER_BUFFER_LEN];
    
    //! Used to override async message processing when in ASYNC mode for the debug logger message
    bool override_async;
    
#endif // ifdef LOGGER_PROCESS_ASYNC
} LOGGER_DATA;

static LOGGER_DATA logger_data =
{
#ifdef LOGGER_PROCESS_ASYNC

    .logger_queue_read_pos = 0,
    .logger_queue_write_pos = 0,
    .override_async = false,

#endif // ifdef LOGGER_PROCESS_ASYNC
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Returns the level of logging
 * e.g LOGGER_TYPE_WARNING include LOGGER_TYPE_WARNING, and LOGGER_TYPE_ERROR
 */
LOGGER_TYPE logger_get_level()
{
    return logger_data.logger_level;
}

/**-----------------------------------------------------------------------------
 * Initialises the Logger
 */
void logger_initialise()
{
    logger_data.logger_callback = NULL;
    logger_data.logger_level = LOGGER_TYPE_ALL; //Note - I was going to change this to NONE but having this means that ALL messages are sent during startup for diagnosis issues
    logger_data.logger_active = false;
}

/**-----------------------------------------------------------------------------
 * Used to generate a RS232 command message - note callback needs to handle LOGGER_TYPE_COMMAND
 * and generate the message
 * @param command - The associated command command code from app_config.g APP_COMMANDS
 * @param data - The data to be put into the message
 * @param data_length - The length of the data
 */
void logger_log_command(unsigned char command, unsigned char* data, size_t data_length)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && 
        ((uint8_t)logger_data.logger_level != (uint8_t)LOGGER_TYPE_NONE) && 
        (!logger_data.suppress_command_logging))
    {
        // Note: LOGGER_CODE is used for the command value
#ifdef LOGGER_PROCESS_ASYNC
        logger_data.logger_queue[logger_data.logger_queue_write_pos].code = command;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_COMMAND;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];
        logger_data.logger_queue[logger_data.logger_queue_write_pos].length = data_length;

        memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], data, LOGGER_BUFFER_LEN); 

        logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
        LOGGER_MESSAGE message_info = { command, LOGGER_TYPE_COMMAND, data, data_length };
        logger_data.logger_callback(&message_info);
#endif      
    }
#endif
}
    
/**-----------------------------------------------------------------------------
 * Log an debug message
 * @param message
 */
void logger_log_debug(LOGGER_CODE code, char* message)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_DEBUG))
    {
#ifdef LOGGER_PROCESS_ASYNC
        logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_DEBUG;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];

        memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], message, LOGGER_BUFFER_LEN); 

        logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
        LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_DEBUG, message, };
        logger_data.logger_callback(&message_info);
#endif    
    }
#endif
}

/**-----------------------------------------------------------------------------
 * Formatter logger debug message
 * @param code The logger code to be sent with the message
 * @param fmt - the sprintf like format string
 * @param ... A list of variables used to populate the string;
 */
void logger_log_debugf(LOGGER_CODE code, char* fmt, ...)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_DEBUG))
    {
        char buffer[LOGGER_BUFFER_LEN];
        memset(buffer, 0, LOGGER_BUFFER_LEN);

        va_list args;
        va_start(args, fmt);
        vsnprintf(buffer, LOGGER_BUFFER_LEN, fmt, args);
        va_end(args);
        
#ifdef LOGGER_PROCESS_ASYNC
        if (!logger_data.override_async)
        {
            logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
            logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_DEBUG;
            logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];

            memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], &buffer, LOGGER_BUFFER_LEN); 

            logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
        }
        else
        {
            LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_DEBUG, buffer, };
            logger_data.logger_callback(&message_info);
        }
#else
        LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_DEBUG, buffer, };
        logger_data.logger_callback(&message_info);
#endif    
    }
#endif
}
    
/**-----------------------------------------------------------------------------
 * Log an error message
 * @param message
 */
void logger_log_error(LOGGER_CODE code, char* message)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_ERROR))
    {
#ifdef LOGGER_PROCESS_ASYNC
        logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_ERROR;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];

        memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], message, LOGGER_BUFFER_LEN); 

        logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
        LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_ERROR, message, };
        logger_data.logger_callback(&message_info);
#endif
    }
#endif
}
 
/**-----------------------------------------------------------------------------
 * Formatter logger error message
 * @param code The logger code to be sent with the message
 * @param fmt - the sprintf like format string
 * @param ... A list of variables used to populate the string;
 */
void logger_log_errorf(LOGGER_CODE code, char* fmt, ...)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_ERROR))
    {
        char buffer[LOGGER_BUFFER_LEN];
        memset(buffer, 0, LOGGER_BUFFER_LEN);

        va_list args;
        va_start(args, fmt);
        vsnprintf(buffer, LOGGER_BUFFER_LEN, fmt, args);
        va_end(args);
        
#ifdef LOGGER_PROCESS_ASYNC
        logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_ERROR;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];

        memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], &buffer, LOGGER_BUFFER_LEN); 

        logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
        LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_ERROR, buffer, };
        logger_data.logger_callback(&message_info);
#endif        
    }
#endif
}
   
/**-----------------------------------------------------------------------------
 * Log an information message
 * @param message
 */
void logger_log_info(LOGGER_CODE code, char* message)
{
#ifdef LOGGER_ENABLED
    if (!logger_data.logger_active)
    {
        logger_data.logger_active = true;

        if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_INFO))
        {
        
#ifdef LOGGER_PROCESS_ASYNC
            logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
            logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_INFO;
            logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];

            memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], message, LOGGER_BUFFER_LEN); 

            logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
            LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_INFO, message, };
            logger_data.logger_callback(&message_info);
#endif        
        }
        
        logger_data.logger_active = false;
    }
#endif
}

/**-----------------------------------------------------------------------------
 * Formatter logger info message
 * @param code The logger code to be sent with the message
 * @param fmt - the sprintf like format string
 * @param ... A list of variables used to populate the string;
 */
void logger_log_infof(LOGGER_CODE code, char* fmt, ...)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_INFO))
    {
        char buffer[LOGGER_BUFFER_LEN];
        memset(buffer, 0, LOGGER_BUFFER_LEN);

        va_list args;
        va_start(args, fmt);
        vsnprintf(buffer, LOGGER_BUFFER_LEN, fmt, args);
        va_end(args);
        
        
#ifdef LOGGER_PROCESS_ASYNC
        logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_INFO;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];
        
        memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], &buffer, LOGGER_BUFFER_LEN); 
        
        logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
        LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_INFO, buffer, };
        logger_data.logger_callback(&message_info);
#endif        
    }
#endif
}

/**-----------------------------------------------------------------------------
 * Formatter logger info message
 * @param code The logger code to be sent with the message
 * @param fmt - the sprintf like format string
 * @param ... A list of variables used to populate the string;
 */
void logger_log_testf(LOGGER_CODE code, char* fmt, ...)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_TEST))
    {
        char buffer[LOGGER_BUFFER_LEN];
        memset(buffer, 0, LOGGER_BUFFER_LEN);

        va_list args;
        va_start(args, fmt);
        vsnprintf(buffer, LOGGER_BUFFER_LEN, fmt, args);
        va_end(args);
        
        
#ifdef LOGGER_PROCESS_ASYNC
        logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_TEST;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];
        
        memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], &buffer, LOGGER_BUFFER_LEN); 
        
        logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
        LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_TEST, buffer, };
        logger_data.logger_callback(&message_info);
#endif        
    }
#endif
}
   
/**-----------------------------------------------------------------------------
 * Log a warning message
 * @param message
 */
void logger_log_warning(LOGGER_CODE code, char* message)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_WARNING))
    {
        
#ifdef LOGGER_PROCESS_ASYNC
        logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_WARNING;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];
        
        memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], message, LOGGER_BUFFER_LEN); 
        
        logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
        LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_WARNING, message, };
        logger_data.logger_callback(&message_info);
#endif        
    }
#endif
}

/**-----------------------------------------------------------------------------
 * Formatter logger info message
 * @param code The logger code to be sent with the message
 * @param fmt - the sprintf like format string
 * @param ... A list of variables used to populate the string;
 */
void logger_log_warningf(LOGGER_CODE code, char* fmt, ...)
{
#ifdef LOGGER_ENABLED
    if ((logger_data.logger_callback != NULL) && (logger_data.logger_level >= LOGGER_TYPE_WARNING))
    {
        char buffer[LOGGER_BUFFER_LEN];
        memset(buffer, 0, LOGGER_BUFFER_LEN);

        va_list args;
        va_start(args, fmt);
        vsnprintf(buffer, LOGGER_BUFFER_LEN, fmt, args);
        va_end(args);
        
#ifdef LOGGER_PROCESS_ASYNC
        logger_data.logger_queue[logger_data.logger_queue_write_pos].code = code;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].type = LOGGER_TYPE_WARNING;
        logger_data.logger_queue[logger_data.logger_queue_write_pos].message = &logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0];
        
        memcpy(&logger_data.logger_queue_buffers[logger_data.logger_queue_write_pos][0], &buffer, LOGGER_BUFFER_LEN); 
        
        logger_data.logger_queue_write_pos = ++logger_data.logger_queue_write_pos % LOGGER_MESSAGE_QUEUE_SIZE;
#else
        LOGGER_MESSAGE message_info = { code, LOGGER_TYPE_WARNING, buffer, };
        logger_data.logger_callback(&message_info);
#endif        
    }
#endif
}
   
/**-----------------------------------------------------------------------------
 * Used for debug purposes
 */
void logger_log_nothing()
{
    // Do nothing
}

/**-----------------------------------------------------------------------------
 * Overrides async mode for logger_log_debug
 * @param override
 */
void logger_override_async(bool override)
{
#ifdef LOGGER_PROCESS_ASYNC
    logger_data.override_async = override;
#endif
}
    
/**-----------------------------------------------------------------------------
 * Processes the logger queue when LOGGER_PROCESS_ASYNC is defined
 */
void logger_process()
{
#ifdef LOGGER_PROCESS_ASYNC

    if (logger_data.logger_callback != NULL)
    {
        // Make it slightly more robust by limiting the number of message processes in any cycle
        //uint8_t count = 0;
        while ((logger_data.logger_queue_read_pos != logger_data.logger_queue_write_pos))// && (count < 3))
        {
            logger_data.logger_callback(&logger_data.logger_queue[logger_data.logger_queue_read_pos]);

            logger_data.logger_queue_read_pos = ++logger_data.logger_queue_read_pos % LOGGER_MESSAGE_QUEUE_SIZE;
            
            //count++;
        }
    }
#endif    
}

/**-----------------------------------------------------------------------------
 * Sets the application callback for logger messages
 * @param callback
 */
void logger_set_callback(LOGGER_CALLBACK callback)
{
    logger_data.logger_callback = callback;
}

/**-----------------------------------------------------------------------------
 * Sets the level of logging
 * e.g LOGGER_TYPE_WARNING include LOGGER_TYPE_WARNING, and LOGGER_TYPE_ERROR
 * @param level The maximum logging level
 */
void logger_set_level(LOGGER_TYPE level)
{
    logger_data.logger_level  = min(level, LOGGER_TYPE_ALL);
    
    logger_log_infof(LOGGER_CODE_SET_LOGGER_LEVEL, "LOGGER: Log level %d %s --", logger_data.logger_level , LOGGER_TYPE_STRINGS[logger_data.logger_level]);
}
    
/**-----------------------------------------------------------------------------
 * Temporarily suppresses logging
 * @param suppress
 */
void logger_suppress_logging(bool suppress)
{
    if (suppress)
    {
        if (logger_data.logger_level > LOGGER_TYPE_NONE)
        {
            logger_data.prev_logger_level = logger_data.logger_level;
            logger_set_level((LOGGER_TYPE_NONE));
        }
    }
    else
    {
        logger_set_level(logger_data.prev_logger_level);
    }
}

/**-----------------------------------------------------------------------------
 * Suppresses the logging of Command logs messages
 * @param suppress
 */
void logger_suppress_command_logging(bool suppress)
{
    logger_data.suppress_command_logging = suppress;
}

/* *****************************************************************************
 End of File
 */
