/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Application level commands that are used to manage menus
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/menu_app.h"
#include "app_config.h"
#include "utilities/app_commands_utilities.h"
#include "utilities/menu.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Makes the supplied item the current menu item
 * @param item
 */
static void menu_app_set_current_item(MENU_ITEM* item)
{
        // We have moved to a child menu
    app_context_set_menu(item->menu_id);
    app_context_set_mode(APP_CONTEXT_MODE_MENU);

    // Check if we want to immediately enter edit mode 
    // We are assuming mode == APP_CONTEXT_MODE_MENU
    if ((item->edit_function != NULL) && item->edit_on_entry)
    {
        // Are we entering edit mode?
        app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
    }        
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Handle menu navigation - goes to a specific menu
 */
void menu_app_handle_switch(STRUCTURED_MESSAGE* message)
{
    APP_COMMAND_MENU menu = message->data[0];
    app_context_set_menu(menu);
}

/**-----------------------------------------------------------------------------
 * Handle menu navigation - goes to a relative menu
 */
void menu_app_handle_navigation(STRUCTURED_MESSAGE* message)
{
    MENU_ITEM* current_item = menu_get_current_item();
   
    // First check whether we are in edit mode - if so we will give the edit function
    // the option of processing the message
    bool processed = false;
    if (app_context_get_mode() == APP_CONTEXT_MODE_MENU_EDIT)
    {
        if ((current_item->edit_function != NULL) && (current_item->edit_wants_navigation))
        {
            processed = current_item->edit_function((void*) message);
        }
    }
    
    if (!processed)
    {
        MENU_NAVIGATION option = message->data[0];

        switch (option)
        {
            case MENU_NAVIGATION_UP:
            {
                MENU_ITEM* item = menu_move_up(current_item);
                if (item != NULL)
                {
                    menu_app_set_current_item(item);
                }
                break;
            }

            case MENU_NAVIGATION_OK:
            {
                switch(app_context_get_mode())
                {
                    case APP_CONTEXT_MODE_MENU:
                    {
                        MENU_ITEM* item = menu_move_down(current_item);
                        if (item != NULL)
                        {
                            menu_app_set_current_item(item);
                        }
                        // This item is at the bottom of the tree
                        else if (current_item->edit_function != NULL)
                        {
                            // Are we entering edit mode?
                            app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                        }
                        else 
                        {
                            if (current_item->process_function != NULL)
                            {
                                // Invoke a process if set.
                                logger_log_debugf(0, "PROCESS Start %s --", menu_get_title(current_item->menu_id));

                                app_commands_start_process(current_item->process_function, NULL);
                            }
                            else
                            {
                                // Removed after review meeting 20170928
                                //MENU_ITEM* item = menu_move_next(current_item);
                                //if (item != NULL)
                                //{
                                //    app_commands_menu_set_current_item(item);
                                //}
                            }
                        }

                        break;
                    }

                    case APP_CONTEXT_MODE_MENU_EDIT:
                    {
                        // Implies we have finished editing
                        // so either go back menu mode or return to default mode
                        // This version allows for a bespoke process to decide what to do next
                        bool processed = false;
                        if (current_item->exit_edit_function != NULL)
                        {
                            processed = current_item->exit_edit_function(current_item->menu_id);
                        }

                        if (!processed)
                        {
                            // By default this version will take you back to the default volume display
                            //app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
                            //app_context_set_menu(APP_COMMAND_MENU_VOLUME);

                            // This version will leave you in the same menu but not in edit mode
                            //app_context_set_mode(APP_CONTEXT_MODE_MENU);
                            
                            // Removed after review meeting 20170928
                            //MENU_ITEM* item = menu_move_next(current_item);
                            //if (item != NULL)
                            //{
                            //    app_commands_menu_set_current_item(item);
                            //}
                        }

                        break;
                    }
                }

                break;
            }

            case MENU_NAVIGATION_LEFT:
            {
                MENU_ITEM* item = menu_move_left(current_item);
                if (item != NULL)
                {
                    menu_app_set_current_item(item);
                }
                break;
            }

            case MENU_NAVIGATION_RIGHT:
            {
                MENU_ITEM* item = menu_move_right(current_item);
                if (item != NULL)
                {
                    menu_app_set_current_item(item);
                }
                break;
            }    
        }
    }
}

/**-----------------------------------------------------------------------------
 * Can be used to reset the inactivity timeout in cases where its been adjusted (DIRAC/ANALOGUE_DIRECT_NO_NAVIGTION)
 * @param menu
 */
MENU_ID menu_app_reset_inactivity_timeout(MENU_ID menu)
{
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;
    
    // Make sure Inactive Menu timeout is reset
    timers_update_timeout(TIMER_INDEX_MENU_INACTIVE, TIMER_MENU_INACTIVE_TIMEOUT_MS, TIMERS_RESET_IMMEDIATE);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Handle a change to a value during menu edit mode
 * Note: This invokes a context specific method for a given menu item
 */
void menu_app_value_change(STRUCTURED_MESSAGE* message)
{
    // Get the current menu
    MENU_ITEM* current_item = menu_get_current_item();

    // if the menu item has a value_changed function associated with it then invoke 
    // it to make the change - Note: generally a bespoke function is required for each 
    // different value
    if (current_item->edit_function != NULL)
    {
        current_item->edit_function(message);
    }
}

/* *****************************************************************************
 End of File
 */
