/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    menu_display.c

  @Summary
     Handles the display of the menu system on the screen

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/menu_display.h"
#include "utilities/display.h"
#include "app_context.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! The ordinal position of the top level menu - e.g. SYSTEM SETTINGS
    int top_menu_index;
} MENU_DISPLAY_DATA;

static MENU_DISPLAY_DATA menu_display_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Draw the standard navigation buttons for the given menu item
 * @param item
 */
void menu_display_draw_generic_navigation(MENU_ITEM* item, MENU_SOFTKEYS edit_navigation_keys, bool enter_branch)
{
    display_clear_line(DISPLAY_LINE_2);
    
    // Do we need a home key
//    if (item != menu_get_item(menu_display_data.top_menu_index))
//    {
        display_special_character(DISPLAY_LINE_2, 2, 
                DISPLAY_SPECIAL_CHARACTERS_HOME);
//    }
    
    // Get the parent menu item
    MENU_ITEM* parent = menu_get_parent(item);
    bool wrap = false;   // top menu does not wrap..
    if (((item->previous_sibling != NULL) || (item->next_sibling != NULL)) && 
         (parent != NULL))
    {
        wrap = parent->loop_back_children;
    }
    
    // Are we in menu or edit mode
    switch(app_context_get_mode())
    {
        case APP_CONTEXT_MODE_MENU:
        {
            if (wrap || (item->previous_sibling != NULL))
            {
                display_special_character(DISPLAY_LINE_2, 5, 
                        DISPLAY_SPECIAL_CHARACTERS_UP);
            }
            
            if (wrap || (item->next_sibling != NULL))
            {
                display_special_character(DISPLAY_LINE_2, 8, 
                        DISPLAY_SPECIAL_CHARACTERS_DOWN);
            }
            
            break;
        }
        
        case APP_CONTEXT_MODE_MENU_EDIT:
        {
            switch (edit_navigation_keys)
            {
                case MENU_SOFTKEYS_LEFT_RIGHT:
                {
                    display_special_character(DISPLAY_LINE_2, 5, 
                            DISPLAY_SPECIAL_CHARACTERS_LEFT);

                    display_special_character(DISPLAY_LINE_2, 8, 
                            DISPLAY_SPECIAL_CHARACTERS_RIGHT);
                    break;
                }
                
                case MENU_SOFTKEYS_UP_DOWN:
                {
                    // Model of up/down navigation
                    if (wrap || (item->previous_sibling != NULL))
                    {
                        display_special_character(DISPLAY_LINE_2, 5, 
                                DISPLAY_SPECIAL_CHARACTERS_UP);
                    }

                    if (wrap || (item->next_sibling != NULL))
                    {
                        display_special_character(DISPLAY_LINE_2, 8, 
                                DISPLAY_SPECIAL_CHARACTERS_DOWN);
                    }
                    
                    break;
                }
                
                default:
                    break;
            }
            
            break;
        }
    }
    
    // Check if we are in a leaf node
    if (item->first_child == NULL)
    {
        if ((item->exit_edit_function != NULL) || (item->process_function != NULL))
        {
            display_special_character(DISPLAY_LINE_2, 11, 
                    DISPLAY_SPECIAL_CHARACTERS_OK);  
        }
    }
    else
    {
        if (enter_branch)
        {
            display_special_character(DISPLAY_LINE_2, 11, 
                    DISPLAY_SPECIAL_CHARACTERS_RIGHT);
        }
    }

#ifdef NO_BACK_BUTTON

#else    
    if ((item != menu_get_item(APP_COMMAND_MENU_SYSTEM_SETTINGS)) &&
        (item != menu_get_item(APP_COMMAND_MENU_STANDBY_SETTINGS)) &&
        (item != menu_get_item(APP_COMMAND_MENU_AUDIO_SETTINGS)) &&
        (item != menu_get_item(APP_COMMAND_MENU_NETWORK_SETTINGS)) &&
        (item != menu_get_item(APP_COMMAND_MENU_ENGINEERING_SETTINGS)))
    
    {
        display_special_character(DISPLAY_LINE_2, 14, 
                DISPLAY_SPECIAL_CHARACTERS_BACK); // OR UP???
    }
#endif
    
    display_write_line(DISPLAY_LINE_2, false);
}

/**-----------------------------------------------------------------------------
 * Draws a standard navigation menu with title on top line and navigation buttons 
 * on the bottom
 */
void menu_display_draw_generic_menu()
{
    MENU_ITEM* item = menu_get_current_item();

    // Draw the top line
        // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);
    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    
    // Check if there is a value to be displayed
    if (item->display_function != NULL)
    {
        size_t len = display_get_line_length();
        char value[len];
        
        item->display_function((char*)&value[0], len);
        
        // Check whether we're displaying the edited value or not
        if (menu_is_editable_and_original(item))
        {
            //display_special_character(DISPLAY_LINE_1, len - strlen(value) - 1, DISPLAY_SPECIAL_CHARACTERS_EDIT_ORIGINAL);
            display_append_text_right(DISPLAY_LINE_1, "*%s", value);
        }
        else
        {
            display_append_text_right(DISPLAY_LINE_1, "%s", value);
        }

        // Following a review meeting 02/06/2017 it was decided that we would not 
        // flash fields while editing by default
        //if ((app_context.mode == APP_CONTEXT_MODE_MENU_EDIT) &&
        //    (item->edit_function != NULL))
        //{
        //    int x;
        //    for (x = max(strlen(value) - 2, 0); x >= 0; x--)
        //    {
        //        display_highlight_character_right(DISPLAY_LINE_1, x, false);
        //    }
        //}
    }
    
    //display_write_line(DISPLAY_LINE_1, false);
    // Make sure fields such as friendly name, which may have user-specified
    // characters, can display utf-8 characters
    display_write_line_inc_utf(DISPLAY_LINE_1, false);

    // Draw the navigation buttons 
    menu_display_draw_generic_navigation(item, MENU_SOFTKEYS_UP_DOWN, true);
}
    
/**-----------------------------------------------------------------------------
 * Initialises the menu display module
 * @param top_menu_index - the index of the top menu - e.g. SYSTEM SETTINGS
 */
void menu_display_initialise(int top_menu_index)
{
    menu_display_data.top_menu_index = top_menu_index;
}

MENU_ITEM* previous_item = NULL;

/**-----------------------------------------------------------------------------
 * Process the menu display
 */
void menu_display_process()
{
    MENU_ITEM* item = menu_get_current_item();
    
    if (item != previous_item)
    {
        logger_log_debugf(0, "DISPLAY TEST Displaying MENU %s --", menu_get_title(item->menu_id));
        previous_item = item;
    }    
    
    if (item->draw_function != NULL)
    {
        item->draw_function(item->menu_id);
    }
    else
    {
        menu_display_draw_generic_menu();    
    }
}

/* *****************************************************************************
 End of File
 */
