/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    mutes.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/mutes.h"
#include "utilities/logger/logger.h"
#include <stdio.h>
#include <string.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */
   
//! Use MUTE_ITEMS defined in app_config.h

// Now use the input data to define the enum
#define X(a, b, c, d, e, f) b,
    static int MUTE_PORT[] = { MUTE_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e, f) c,
    static int MUTE_CHANNEL[] = { MUTE_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e, f) d,
    static int MUTE_BIT_POS[] = { MUTE_ITEMS };
#undef X   

// Now use the input data to define the enum
#define X(a, b, c, d, e, f) e,
    static bool MUTE_ACTIVE_HIGH[] = { MUTE_ITEMS };
#undef X
    
// Redefine the macro and use to create an array of strings
#define X(a, b, c, d, e, f) #f,
	static char *MUTE_STRINGS[] = { MUTE_ITEMS };
#undef X
    
/**
 * Information describing a Mute
 */
typedef struct 
{
    // The ordinal position of the mute in the mutes array - used to 
    // reference the mute
    uint8_t             mute_index;

    // Port definition
    PORTS_MODULE_ID     module_id;
    PORTS_CHANNEL       channel;
    PORTS_BIT_POS       bit_pos;

    // The current state of the mute on/off - also use to set the initial state
    MUTE_STATE          state;

    //! Determines whether enabling the mute sets the pin high or low (the majority will be low))
    bool                active_high;
} MUTES_MUTE_INFO;
    
typedef struct
{
    MUTES_MUTE_INFO mutes[MUTE_INDEX_MAX];
} MUTES_DATA;

static MUTES_DATA mutes_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Perform mute initialisation
 */
void mutes_initialise()
{
    // Initialise mute indexes and ports etc...
    int x;
    for (x = 0; x < MUTE_INDEX_MAX; x++)
    {
        MUTES_MUTE_INFO* info = &mutes_data.mutes[x];
        
        info->mute_index = x;
        info->module_id = MUTE_PORT[x];
        info->channel = MUTE_CHANNEL[x];
        info->bit_pos = MUTE_BIT_POS[x];
        info->state =  MUTE_STATE_OFF;
        info->active_high = MUTE_ACTIVE_HIGH[x];
        
        // Start off in a muted state
        mutes_mute(x);
    }
}
   
/**-----------------------------------------------------------------------------
 * Mutes the specified mute index
 */
void mutes_mute(uint8_t mute_index)
{
    if (mute_index < MUTE_INDEX_MAX)
    {
        MUTES_MUTE_INFO* info = &mutes_data.mutes[mute_index];
        
#if MUTES_PINS_ENABLED   
        
        if (info->active_high)
        {
            PLIB_PORTS_PinSet(info->module_id, 
                              info->channel, 
                              info->bit_pos);
        }
        else
        {
            // Pins are Low to mute
            PLIB_PORTS_PinClear(info->module_id, 
                              info->channel, 
                              info->bit_pos);
        }
#endif
        
        info->state = MUTE_STATE_ON;
        
        logger_log_infof(LOGGER_CODE_MUTES_MUTE, "MUTE %s  state %s --", MUTE_STRINGS[info->mute_index], MUTE_STATE_STRINGS[info->state]);
    }
}
    
/**-----------------------------------------------------------------------------
 * Returns whether the mute is currently muted
 * 
 * @param mute_index - The index of the mute
 */
MUTE_STATE mutes_state(uint8_t mute_index)
{
    MUTE_STATE result = MUTE_STATE_OFF;
    
    if (mute_index < MUTE_INDEX_MAX)
    {
        MUTES_MUTE_INFO* info = &mutes_data.mutes[mute_index];       
        result = info->state;
    }    
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Unmutes the specified mute index
 */
void mutes_unmute(uint8_t mute_index)
{
    if (mute_index < MUTE_INDEX_MAX)
    {
        MUTES_MUTE_INFO* info = &mutes_data.mutes[mute_index];
        
#if MUTES_PINS_ENABLED   
       
        if (info->active_high)
        {
            PLIB_PORTS_PinClear(info->module_id, 
                                info->channel, 
                                info->bit_pos);
        }
        else
        {
            PLIB_PORTS_PinSet(info->module_id, 
                                info->channel, 
                                info->bit_pos);
        }
#endif
        
        info->state = MUTE_STATE_OFF;
    
        logger_log_infof(LOGGER_CODE_MUTES_UNMUTE, "MUTE VOLUME Unmute %s --", MUTE_STRINGS[info->mute_index]);
    }
}


/* *****************************************************************************
 End of File
 */
