/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    pins.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/pins.h"
#include <stdbool.h>
#include "app_config.h"
#include "utilities/logger/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#ifndef PINS_MAX
#define PINS_MAX 39
#endif

#define PIN_LOW     false;
#define PIN_HIGH    true;

/**
 * Pin definition
 * Note: Its primary use is to store the state of a pin at a point in time rather than the current state
 * which can be obtained dynamically
 */
typedef struct
{
    uint8_t port;
    uint8_t channel;
    uint8_t pin;
    bool state;
    uint8_t mask;
    bool set_as_input;
} PIN_DEFINITION;

typedef struct
{
    uint8_t count;
    
    PIN_DEFINITION pins[PINS_MAX];
} PINS_DATA;

static PINS_DATA pins_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**
 * Add a pin definition
 * @param index
 * @param port
 * @param channel
 * @param pin
 */
void pins_add_pin(uint8_t port, uint8_t channel, uint8_t pin, uint8_t mask, bool set_as_input)
{
    logger_log_debugf(LOGGER_CODE_PINS, "PINS Adding %d %d-%d-%d %d --", pins_data.count, port, channel, pin, mask);
    if (pins_data.count < PINS_MAX)
    {
        pins_data.pins[pins_data.count].port = port;
        pins_data.pins[pins_data.count].channel = channel;
        pins_data.pins[pins_data.count].pin = pin;
        pins_data.pins[pins_data.count].state = PIN_LOW;
        pins_data.pins[pins_data.count].mask = mask;
        pins_data.pins[pins_data.count].set_as_input = set_as_input;
        
        pins_data.count++;
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_PINS, "PINS Unable to add pin %d-%d-%d as it would exceed PINS_MAX (%d) ", port, channel, pin, PINS_MAX);
    }
}

/**
 * initialises the pins module
 */
void pins_initialise()
{
    // Nothing to do here yet!
}

/**
 * Clears all the pins (saves state first)
 */
void pins_set_all_low(uint8_t mask)
{
    logger_log_debugf(LOGGER_CODE_PINS, "PINS Setting %d low %d --", pins_data.count, mask);
    
    int x;
    for (x = 0; x < pins_data.count; x++)
    {
        if ((pins_data.pins[x].mask & mask) == mask)
        {
            logger_log_debugf(LOGGER_CODE_PINS, "PINS Setting pin low %d-%d-%d --", pins_data.pins[x].port, pins_data.pins[x].channel, pins_data.pins[x].pin, PINS_MAX);
            PLIB_PORTS_PinClear(pins_data.pins[x].port, pins_data.pins[x].channel, pins_data.pins[x].pin); 
            
            if (pins_data.pins[x].set_as_input)
            {
                PLIB_PORTS_PinDirectionInputSet(pins_data.pins[x].port, pins_data.pins[x].channel, pins_data.pins[x].pin); 
            }
        }
    }
}

/**
 * Restores the pins to the most recent saved state
 */
void pins_restore_state()
{
    logger_log_debugf(LOGGER_CODE_PINS, "PINS Restoring State");
    
    int x;
    for (x = 0; x < pins_data.count; x++)
    {
        if (pins_data.pins[x].set_as_input)
        {
            PLIB_PORTS_PinDirectionOutputSet(pins_data.pins[x].port, pins_data.pins[x].channel, pins_data.pins[x].pin); 
        }

        PLIB_PORTS_PinWrite(pins_data.pins[x].port, pins_data.pins[x].channel, pins_data.pins[x].pin, pins_data.pins[x].state); 
        logger_log_debugf(LOGGER_CODE_PINS, "PINS Restoring %d %d %d --", x, pins_data.pins[x].state, PLIB_PORTS_PinGet(pins_data.pins[x].port, pins_data.pins[x].channel, pins_data.pins[x].pin));           
    }
}

/**
 * Saves the current state of all the pins
 */
void pins_save_state()
{
    logger_log_debugf(LOGGER_CODE_PINS, "PINS Saving State");
    
    int x;
    for (x = 0; x < pins_data.count; x++)
    {
        pins_data.pins[x].state = PLIB_PORTS_PinGet(pins_data.pins[x].port, pins_data.pins[x].channel, pins_data.pins[x].pin); 
        logger_log_debugf(LOGGER_CODE_PINS, "PINS Saving %d %d --", x, pins_data.pins[x].state);
    }
}

/* *****************************************************************************
 End of File
 */
