/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    power.c

  @Summary
 Manages simple power control functions

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/power.h"
#include "utilities/logger/logger.h"
#include "utilities/settings.h"
#include "utilities/timers.h"
#include <stdio.h>
#include <strings.h>

#ifdef POWER_AC_DETECT_ANALOGUE
#include "utilities/adc/adc.h"
#endif

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

// Ensure defines exist - to override add to app_config,h
#ifndef POWER_PINS_ENABLED
#define POWER_PINS_ENABLED          1 
#endif

#ifndef POWER_SOFT_START_ENABLED
#define POWER_SOFT_START_ENABLED    0
#endif

//! Uses POWER_ITEMS defined in app_config.h   
// Now use the input data to define the enum
#define X(a, b, c, d, e) b,
    static int POWER_PORT[] = { POWER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e) c,
    static int POWER_CHANNEL[] = { POWER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e) d,
    static int POWER_BIT_POS[] = { POWER_ITEMS };
#undef X
   
// Now use the input data to define the enum
#define X(a, b, c, d, e) e,
    static int POWER_FLAGS[] = { POWER_ITEMS };
#undef X

typedef struct
{
    //! The current power state of the application (based upon calls to power_enable/disable))
    POWER_STATE state;
    
    // The auto shutdown timeout in minutes
    // To convert to milliseconds * 3600000
    uint16_t auto_shutdown_timeout;
    
    // Determines whether Network PSU is enabled or not on shutdown
    bool network_standby;
    
    // Determines whether standby mode is sleep or idle
    bool rs232_standby;
    
    //! True if the ac_detect pin should be high (SA20), false if low (SA10))
    bool ac_detect_pin_high;
} POWER_DATA;

static POWER_DATA power_data =
{
   .network_standby = false,
   .rs232_standby = false, 
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Sets/Clears pin based on PSU index
 * @param index
 * @param on
 */
static void power_set_pins(POWER_INDEX index, bool on)
{
#if POWER_PINS_ENABLED   
       
    if (on)
    {
//        logger_log_debugf(LOGGER_CODE_POWER, "POWER %d Set Pin ON %d %d %d --", index, POWER_PORT[index], 
//                          POWER_CHANNEL[index], 
//                          POWER_BIT_POS[index]);
        
        PLIB_PORTS_PinSet(POWER_PORT[index], 
                          POWER_CHANNEL[index], 
                          POWER_BIT_POS[index]);    
    }
    else
    {
        logger_log_debugf(LOGGER_CODE_POWER, "POWER %d Set Pin OFF --", index);
        PLIB_PORTS_PinClear(POWER_PORT[index], 
                            POWER_CHANNEL[index], 
                            POWER_BIT_POS[index]);
    }
#endif
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Disables both the PSU and Digital PSU
 */
void power_disable()
{
    logger_log_debugf(0, "POWER Disable --");
    int x;
    for (x = POWER_INDEX_MAX - 1; x >= 0; x--)
    {
        logger_log_debugf(0, "POWER Disable %d %d %d  --", x, POWER_FLAGS[x], POWER_FLAGS[x] & POWER_FLAGS_NETWORK, POWER_FLAGS_NETWORK);
        if ((POWER_FLAGS[x] & (int)POWER_FLAGS_ALWAYS_ON) == (int)POWER_FLAGS_ALWAYS_ON)
        {
            // Do nothing
            logger_log_debugf(0, "POWER Disable ALWAYS_ON IGNORED %d %d %d %d  --", x, POWER_FLAGS[x], POWER_FLAGS[x] & POWER_FLAGS_NETWORK, POWER_FLAGS_NETWORK);
        }
        else if ((POWER_FLAGS[x] & (int)POWER_FLAGS_NETWORK) == (int)POWER_FLAGS_NETWORK)
        {
            if (!power_get_network_active_in_standby())
            {
                logger_log_debugf(0, "POWER Disable NETWORK %d %d %d %d  --", x, POWER_FLAGS[x], POWER_FLAGS[x] & POWER_FLAGS_NETWORK, POWER_FLAGS_NETWORK);
                power_disable_by_index(x);
            }
            else
            {
                logger_log_debugf(0, "POWER Disable NETWORK IGNORED %d %d %d %d  --", x, POWER_FLAGS[x], POWER_FLAGS[x] & POWER_FLAGS_NETWORK, POWER_FLAGS_NETWORK);
            }
        }
        else
        {
            power_disable_by_index(x);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Disables an individual PSU by Index
 */
void power_disable_by_index(POWER_INDEX index)
{
    logger_log_debugf(0, "POWER Disabling %d --", index);
    power_set_pins(index, false);
    
    if (POWER_FLAGS[index] & POWER_FLAGS_MAIN == POWER_FLAGS_MAIN)
    {
        power_data.state = POWER_STATE_OFF;

        uint8_t value = power_data.state;
        logger_log_command(APP_COMMANDS_POWER, &value, 1);

        logger_log_info(LOGGER_CODE_POWER, "POWER PSU Disabled --");
    }
}

/**-----------------------------------------------------------------------------
 * Enables all PSUs
 */
void power_enable()
{
    int x;
    for (x = 0; x < POWER_INDEX_MAX; x++)
    {
        logger_log_debugf(0, "POWER power_enable %d %d %d --", x, POWER_FLAGS[x], POWER_FLAGS[x] & POWER_FLAGS_DELAYED);
        // Check were not delaying the start of this item
        if ((POWER_FLAGS[x] & POWER_FLAGS_DELAYED) != POWER_FLAGS_DELAYED)
        {
            power_enable_by_index(x);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Enables an individual PSU by Index
 */
void power_enable_by_index(POWER_INDEX index)
{    
    //logger_log_debugf(LOGGER_CODE_POWER, "POWER %d Enabled --", index);
    
    power_set_pins(index, true);

    if (POWER_FLAGS[index] & POWER_FLAGS_MAIN == POWER_FLAGS_MAIN)
    {
        power_data.state = POWER_STATE_ON;

        uint8_t value = power_data.state;
        logger_log_command(APP_COMMANDS_POWER, &value, 1);
        logger_log_info(LOGGER_CODE_POWER, "POWER PSU Enabled --");
    }
}

/**-----------------------------------------------------------------------------
 * Used to determine whether AC is present
 * @return 
 */
bool power_get_ac_present()
{
    bool result = true;

#if POWER_PINS_ENABLED   
    
#ifdef POWER_AC_DETECT_ANALOGUE  

    // We read the analogue ADC input - if it is < 10% then we consider it to have no AC
    // (Maximum value for a 10bit ADC is 1023 so 102 = 1023/10)
    uint32_t value =  adc_get_raw_value(POWER_AC_DETECT_INPUT);
    if (value < 102)
    {
        logger_log_warningf(LOGGER_CODE_POWER, "POWER No AC Detected value:%d -- ", value);
        result = false;
    }

#else
    
    result = PLIB_PORTS_PinGet(POWER_AC_DETECT_PORTS_ID, 
                               POWER_AC_DETECT_PORT_CHANNEL, 
                               POWER_AC_DETECT_PIN);
    
    
    result = (result == power_data.ac_detect_pin_high);

#endif
    
#endif    
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Sets the auto shutdown setting
 * @return 
 */
bool power_get_auto_shutdown()
{
    return power_get_auto_shutdown_timeout() != 0;
}
    
/**-----------------------------------------------------------------------------
 * Sets the auto shutdown timeout setting (in mins)
 * @return 
 */
uint16_t power_get_auto_shutdown_timeout()
{
    return power_data.auto_shutdown_timeout;
}

/**-----------------------------------------------------------------------------
 * Gets the On/Off state of network standby
 * @return 
 */
bool power_get_network_active_in_standby()
{
    return power_data.network_standby;
}

/**-----------------------------------------------------------------------------
 * Gets the On/Off state of rs232 standby
 * @return 
 */
bool power_get_rs232_active_in_standby()
{
    return power_data.rs232_standby;
}

/**-----------------------------------------------------------------------------
 * Returns the current Power state
 * @return 
 */
POWER_STATE power_get_state()
{
    return power_data.state;
}

/**-----------------------------------------------------------------------------
 * Initialises power options from settings
 * @param ac_detect_pin_high - true if high for ac detect, false if low
 */
void power_initialise(bool ac_detect_pin_high)
{
    power_data.ac_detect_pin_high = ac_detect_pin_high;
    
    // Get setting values
    settings_get(SETTING_AUTOSHUTDOWN_TIMEOUT, &power_data.auto_shutdown_timeout);
    settings_get(SETTING_NETWORK_STANDBY, &power_data.network_standby);
    settings_get(SETTING_RS232_STANDBY, &power_data.rs232_standby);
    
    // ISSUE-20171215-1 - Removed this option
    // Put a minimum value on the timer
    //if (power_data.auto_shutdown_timeout == 0)
    //{
    //    power_data.auto_shutdown_timeout = POWER_AUTO_TIMEOUT_INCREMENT;
    //    settings_set(SETTING_AUTOSHUTDOWN_TIMEOUT, &power_data.auto_shutdown_timeout);
    //}

    power_data.network_standby = (power_data.network_standby > 0 ? true : false);
    power_data.rs232_standby = (power_data.rs232_standby > 0 ? true : false);
    
    timers_update_timeout(TIMER_INDEX_AUTO_SHUTDOWN, power_data.auto_shutdown_timeout * 60000, TIMERS_RESET_IMMEDIATE);
    
    logger_log_infof(LOGGER_CODE_POWER, "POWER (Init) Auto shutdown timeout set %d --", timers_get_timeout(TIMER_INDEX_AUTO_SHUTDOWN));
}
         
/**-----------------------------------------------------------------------------
 * Resets the auto shutdown timer - used by heartbeat
 */
void power_reset_auto_shutdown_timer()
{
    if (power_get_auto_shutdown())
    {
        //logger_log_infof(LOGGER_CODE_POWER, "POWER Reset timeout %d --", timers_get_timeout(TIMER_INDEX_AUTO_SHUTDOWN));
        timers_reset(TIMER_INDEX_AUTO_SHUTDOWN);
    }
}
    
/**-----------------------------------------------------------------------------
 * Gets the auto shutdown timeout setting
 * Note: This is either a multiple of 30 mins upto 24 hours or 1min
 * @return 
 */
void power_set_auto_shutdown_timeout(uint16_t timeout)
{
    if (power_data.auto_shutdown_timeout != timeout)
    {
        // Make sure timeout fits to one of the fixed values except when its set to override
        if (timeout != 1)
        {
            int x;
            for (x = 0; x < APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_MAX; x++)
            {
                if (timeout <= APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_VALUES[x])
                {
                    timeout = APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_VALUES[x];
                    break;
                }
            }
        }
        
        // Adjust timeout to allow cycling through the menu option in an ascending way
        if ((power_data.auto_shutdown_timeout == POWER_AUTO_TIMEOUT_MAX) && (timeout > POWER_AUTO_TIMEOUT_MAX))
        {
            timeout = 0;
        }
        power_data.auto_shutdown_timeout = min(max(timeout, 0), POWER_AUTO_TIMEOUT_MAX); // 4 hours

        // Update settings?
        settings_set(SETTING_AUTOSHUTDOWN_TIMEOUT, &power_data.auto_shutdown_timeout);

        timers_update_timeout(TIMER_INDEX_AUTO_SHUTDOWN, power_data.auto_shutdown_timeout * 60000, TIMERS_RESET_IMMEDIATE);

        logger_log_infof(LOGGER_CODE_POWER, "POWER Auto Shutdown Timeout %d --", power_data.auto_shutdown_timeout);
    }
}

/**-----------------------------------------------------------------------------
 * Sets the On/Off stane of network standby
 * @param on_off - true - On , false - off 
 */
void power_set_network_active_in_standby(bool on_off)
{
    if (power_data.network_standby != on_off)
    {
        power_data.network_standby = on_off;

        // Update settings?
        settings_set(SETTING_NETWORK_STANDBY, &power_data.network_standby);
        
        logger_log_infof(LOGGER_CODE_POWER, "POWER Network standby %s --", power_data.network_standby ? APP_STRINGS[APP_STRINGS_ENUM_ON] : APP_STRINGS[APP_STRINGS_ENUM_OFF]);
    }
}

/**-----------------------------------------------------------------------------
 * Sets the On/Off stane of network standby
 * @param on_off - true - On , false - off 
 */
void power_set_rs232_active_in_standby(bool on_off)
{
    if (power_data.rs232_standby != on_off)
    {
        power_data.rs232_standby = on_off;

        // Update settings?
        settings_set(SETTING_RS232_STANDBY, &power_data.rs232_standby);
        
        logger_log_infof(LOGGER_CODE_POWER, "POWER RS232 standby %s --", power_data.rs232_standby ? APP_STRINGS[APP_STRINGS_ENUM_ON] : APP_STRINGS[APP_STRINGS_ENUM_OFF]);
    }
}
     
/**-----------------------------------------------------------------------------
 * Clears the soft start enable switch - needs to be called 0.5s+ after 
 * power is applied
 */
void power_soft_start_clear()
{ 
#if POWER_PINS_ENABLED && POWER_SOFT_START_ENABLED

    PLIB_PORTS_PinClear(POWER_SOFT_START_PORTS_ID, 
                        POWER_SOFT_START_PORT_CHANNEL, 
                        POWER_SOFT_START_PIN);
    
#endif
}
   
/**-----------------------------------------------------------------------------
 * Enables the soft start pin
 */
void power_soft_start_enable()
{
#if POWER_PINS_ENABLED && POWER_SOFT_START_ENABLED  
       
    PLIB_PORTS_PinSet(POWER_SOFT_START_PORTS_ID, 
                      POWER_SOFT_START_PORT_CHANNEL, 
                      POWER_SOFT_START_PIN);

    logger_log_info(LOGGER_CODE_POWER, "POWER Soft Start Enabled --");    

#endif
}

/* *****************************************************************************
 End of File
 */
