/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Arcam

  @File Name
    rs232.c

  @Summary
 Handles communication with the RS232 port

  @Description
 Maintains an RS232 connection and parses standard messages
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/rs232.h"
#include "utilities/logger/logger.h"
#include "app_config.h"
#include "utilities/timers.h"
#include "driver/usart/drv_usart.h"

#include <xc.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//-----------------------------------
// These will need moving to a common messages file
//---------------------------------------------------

/**
 * Holds global data
 */
typedef struct
{
    // I2C handle
    DRV_HANDLE          usart_handle;
    
    // Holds the current message read from the USART
    STRUCTURED_MESSAGE  current_message;
    
    // Callback function supplied to rs232_initialise - called when a message is 
    // received
    STRUCTURED_MESSAGE_CALLBACK callback;

} RS232_DATA;

static RS232_DATA rs232_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Initialises a structured message
 * @param message
 */
static void initialise_structured_message(STRUCTURED_MESSAGE* message)
{
    structured_message_initialise(message, STRUCTURED_MESSAGE_SOURCE_RS232);
    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * RS232 module initialisation
 */
void rs232_initialise(const SYS_MODULE_INDEX module, STRUCTURED_MESSAGE_CALLBACK callback)
{
    rs232_data.callback = callback;
    
    initialise_structured_message(&rs232_data.current_message);
    
    rs232_data.usart_handle = DRV_USART_Open(module, 
                                             DRV_IO_INTENT_NONBLOCKING | DRV_IO_INTENT_READWRITE);
    if (DRV_HANDLE_INVALID == rs232_data.usart_handle)
    {
        //settings_data.state = SETTINGS_STATE_ERROR;
        logger_log_error(LOGGER_CODE_USART_OPEN_ERROR, "USART Open");
    }    
}

/**
 * RS232 processing
 */
void rs232_process()
{
    unsigned char data_byte;
        
    char buffer[STRUCTURED_MESSAGE_MAX_DATA + 6];
    size_t count  = DRV_USART_Read(rs232_data.usart_handle, &buffer, STRUCTURED_MESSAGE_MAX_DATA + 5);
    
    if(count > 0)// != DRV_USART_READ_ERROR)
    {
        // Added a brief pause as occasionally it does not get the data in non-blocking mode
        timers_delay_us(10);
        
        // So we've got some data - its possible its not a whole message, or a whole message of bits of more than one message
        int x;
        for (x = 0; x < count; x++)
        {
            structured_message_process_byte(&rs232_data.current_message, buffer[x]);
            
            // If the message is complete
            if (rs232_data.current_message.processing_state == STRUCTURED_MESSAGE_STATE_PERFORM_ACTION)
            {
                // Pass the message onto a callback method for validation and processing
                if (rs232_data.callback != NULL)
                {
                    rs232_data.callback(&rs232_data.current_message);
                }

                // Re initialise the message
                initialise_structured_message(&rs232_data.current_message);
            }

            // If we need to send a reply then lets get that started
            if (rs232_data.current_message.processing_state == STRUCTURED_MESSAGE_STATE_SEND_REPLY)
            {
                rs232_send_message(&rs232_data.current_message);
                initialise_structured_message(&rs232_data.current_message);
            }
        }
    }
    else if (count == DRV_USART_READ_ERROR)
    {
       // There was an error. The DRV_USART_ErrorGet() function
       // can be called to find the exact error.        
       logger_log_info(LOGGER_CODE_USART_READ_ERROR, "RS232: Read Error");
    }
}

/**
 * Sends a buffer of data over the RS232 connection
 * @param buffer
 * @param len
 */
void rs232_send_buffer(char* buffer, size_t len)
{
    // Send the buffer
    int total = 0;
    do
    {
        int count  = DRV_USART_Write(rs232_data.usart_handle, 
                                     &buffer[total], 
                                     len - total);
        if (count == DRV_USART_WRITE_ERROR)
        {
            logger_log_errorf(LOGGER_CODE_USART_WRITE_ERROR, "RS232 Write Error %d --", len);
            break;
        }
        
        total += count;
    } while(total < len);
}

/**
 * Send a structured message to the RS232 port
 * Note: This is an asynchronous implementation, so the message may not have been 
 * completely sent by the time it completes
 * @param message
 */
void rs232_send_message(STRUCTURED_MESSAGE* message)
{
    // Allocate a buffer to hold the message
    char buffer[STRUCTURED_MESSAGE_MAX_SIZE];
    
    // Fill the buffer with the message
    int len = structured_message_write_response_buffer(message, buffer, STRUCTURED_MESSAGE_MAX_SIZE);

    rs232_send_buffer(buffer, len);
}

/* *****************************************************************************
 End of File
 */
