/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Arcam

  @File Name
    rs232.c

  @Summary
 Handles communication with the RS232 port

  @Description
 Maintains an RS232 connection and parses standard messages
 * 
 * v2.1 - Includes display_message processing
 *        This is a DYNAMIC implementation - to work the number of Driver Clients needs to be > 1 in Harmony (I used 4)
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/rs232_display.h"
#include "app_config.h"
#include "utilities/display_message_data.h"
#include "utilities/logger.h"
#include "utilities/timers.h"
#include "driver/usart/drv_usart.h"

#include <xc.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Holds global data
 */
typedef struct
{
    // I2C handle
    DRV_HANDLE          usart_handle;
    
    //! Callback for display messages
    DISPLAY_MESSAGE_CALLBACK display_callback;
    
    //! Display message to hold incoming Display message data
    DISPLAY_MESSAGE display_message;
    
    //! Buffers used for transferring large images - these need to be at least big enough
    //  to store an entire image
    uint8_t write_buffer[RS232_DISPLAY_TX_QUEUE_SIZE][RS232_DISPLAY_TX_BUFFER_SIZE];
    uint8_t write_buffer_len[RS232_DISPLAY_TX_QUEUE_SIZE];
    
    //! Simple circular buffer index
    uint8_t write_buffer_idx;
    
    //! Simple circular buffer index
    uint8_t write_buffer_read_idx;
     
    //! Buffers used for transferring large images - these need to be at least big enough
    //  to store an entire image
    uint8_t read_buffer[RS232_DISPLAY_RX_QUEUE_SIZE][RS232_DISPLAY_RX_BUFFER_SIZE];
    bool read_buffer_complete[RS232_DISPLAY_RX_QUEUE_SIZE];
    DRV_USART_BUFFER_HANDLE read_buffer_handle[RS232_DISPLAY_RX_QUEUE_SIZE];
     
    //! Simple circular buffer index
    //uint8_t read_buffer_idx;
    
    //! Simple circular buffer index
    uint8_t read_buffer_read_idx;
   
    //! Check to monitor buffer usage
    uint8_t write_buffer_used_count;
} RS232_DATA;

static RS232_DATA rs232_data = {
//    .write_buffer_idx = 0,
//    .write_buffer_read_idx = 0,
//    .read_buffer_idx = 0,
    .read_buffer_read_idx = 0,
    .write_buffer_used_count = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * 
 * @param event
 * @param bufferHandle
 * @param contextHandle
 */
static void rs232_display_callback(DRV_USART_BUFFER_EVENT event,
        DRV_USART_BUFFER_HANDLE bufferHandle, 
        uintptr_t contextHandle)
{
    // contextHandle points to myAppObj.
    
    rs232_data.write_buffer_used_count--;

    switch(event)
    {
        case DRV_USART_BUFFER_EVENT_COMPLETE:
        {
            // This means the data was transferred.
            //logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 Data Written %d --", bufferHandle);
            int x;
            for (x = 0; x < RS232_DISPLAY_RX_QUEUE_SIZE; x++)
            {
                if (rs232_data.read_buffer_handle[x] == bufferHandle)
                {
                    rs232_data.read_buffer_complete[x] = true;
//                    rs232_data.read_buffer_idx = ++rs232_data.read_buffer_idx % RS232_DISPLAY_RX_QUEUE_SIZE; 
                    break;
                }
            }
            break;
        }

        case DRV_USART_BUFFER_EVENT_ERROR:
        {
            //logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Data Error %d --", bufferHandle);
            // Error handling here.
            int x;
            for (x = 0; x < RS232_DISPLAY_RX_QUEUE_SIZE; x++)
            {
                if (rs232_data.read_buffer_handle[x] == bufferHandle)
                {
                    rs232_data.read_buffer_handle[x] = DRV_USART_BUFFER_HANDLE_INVALID;
//                    rs232_data.read_buffer_idx = ++rs232_data.read_buffer_idx % RS232_DISPLAY_RX_QUEUE_SIZE; 
                    break;
                }
            }

            break;
        }
        
        default:
            break;
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param idx
*/
static bool rs232_display_check_read_buffer(uint8_t idx)
{
    bool result = false;
    if (rs232_data.read_buffer_complete[idx])
    {
        //logger_log_debugf(0, "DISPLAY Processing Buffer %d %d ********************** --", idx, rs232_data.read_buffer_handle[idx]);
        
        // Action the buffer
        int y;
        for (y = 0; y < RS232_DISPLAY_RX_BUFFER_SIZE; y++)
        {
            if (rs232_data.read_buffer[idx][y] != 0xFD)
            {
                //logger_log_debugf(0, "DISPLAY Processing Byte %d %d --", idx, rs232_data.read_buffer[idx][y]);
                display_message_process_byte(&rs232_data.display_message, (unsigned char)rs232_data.read_buffer[idx][y]);

                if (display_message_is_complete(&rs232_data.display_message))
                {
                    //logger_log_debugf(0, "DISPLAY Processing Complete %d %d --", idx, rs232_data.read_buffer[idx][y]);
                    if (rs232_data.display_callback != NULL)
                    {
                        //logger_log_debugf(0, "DISPLAY Processing Callback --");
                        rs232_data.display_callback(&rs232_data.display_message);
                    }

                    display_message_initialise(&rs232_data.display_message);
                }                        
            }
            else
            {
                break;
            }
        }

        // Re add the buffer
        rs232_data.read_buffer_complete[idx] = false;
        memset(rs232_data.read_buffer, 0xFD, RS232_DISPLAY_RX_BUFFER_SIZE);

        DRV_USART_BUFFER_HANDLE buffer_handle;
        DRV_USART_BufferAddRead(rs232_data.usart_handle, 
                &buffer_handle,
                &rs232_data.read_buffer[idx][0], 
                RS232_DISPLAY_RX_BUFFER_SIZE);

        if (DRV_USART_BUFFER_HANDLE_INVALID != buffer_handle)
        {
            rs232_data.read_buffer_handle[idx] = buffer_handle;
        }
        
        result = true;
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * RS232 module initialisation
 */
void rs232_display_initialise(const SYS_MODULE_INDEX module, 
        DISPLAY_MESSAGE_CALLBACK display_callback)
{
    rs232_data.display_callback = display_callback;
    display_message_initialise(&rs232_data.display_message);

    rs232_data.usart_handle = DRV_USART_Open(module, 
                                             DRV_IO_INTENT_NONBLOCKING | DRV_IO_INTENT_READWRITE);
    if (DRV_HANDLE_INVALID == rs232_data.usart_handle)
    {
        //settings_data.state = SETTINGS_STATE_ERROR;
        logger_log_error(LOGGER_CODE_RS232_DISPLAY, "USART Open");
    }    
    
//    DRV_USART_BufferEventHandlerSet(rs232_data.usart_handle,
//                rs232_display_callback, (uintptr_t)NULL);
//    
//    // Add read buffers
//    int x;
//    for (x = 0; x < RS232_DISPLAY_RX_QUEUE_SIZE; x++)
//    {
//        rs232_data.read_buffer_complete[x] = false;
//        memset(rs232_data.read_buffer, 0xFD, RS232_DISPLAY_RX_BUFFER_SIZE);
//        
//        DRV_USART_BUFFER_HANDLE buffer_handle;
//        DRV_USART_BufferAddRead(rs232_data.usart_handle, 
//                &buffer_handle,
//                &rs232_data.read_buffer[x][0], 
//                RS232_DISPLAY_RX_BUFFER_SIZE);
//
//        if (DRV_USART_BUFFER_HANDLE_INVALID != buffer_handle)
//        {
//            rs232_data.read_buffer_handle[x] = buffer_handle;
//        }
//    }
}

/**-----------------------------------------------------------------------------
 * RS232 processing
 */
void rs232_display_process()
{
    // Look through then read buffers from the point that they were last read
//    int x;
//    for (x = 0; x < RS232_DISPLAY_RX_QUEUE_SIZE; x++)
//    {
//        // Check if we have read the buffer, if so move the read index on by 1
//        if (rs232_display_check_read_buffer(rs232_data.read_buffer_read_idx))
//        {
//            // Increment the read index
//            rs232_data.read_buffer_read_idx = ++rs232_data.read_buffer_read_idx % RS232_DISPLAY_RX_QUEUE_SIZE; 
//        }
//        else
//        {
//            // Otherwise break out;
//            break;
//        }
//    }

    unsigned char data_byte;
        
    char buffer[DISPLAY_MESSAGE_MAX_DATA_LENGTH];
    size_t count  = DRV_USART_Read(rs232_data.usart_handle, &buffer, DISPLAY_MESSAGE_MAX_DATA_LENGTH);

    if(count > 0)// != DRV_USART_READ_ERROR)
    {
        // Added a brief pause as occasionally it does not get the data in non-blocking mode
        timers_delay_us(10);
        
        // So we've got some data - its possible its not a whole message, or a whole message of bits of more than one message
        //logger_log_debugf(0, "DISPLAY RS %d %d %d %d %d %d %d --", 
        //        buffer[0], buffer[1], buffer[2],buffer[3], buffer[4], buffer[5],buffer[6] );
        int x;
        for (x = 0; x < count; x++)
        {
            //logger_log_debugf(0, "DISPLAY RS %d %d --", x, buffer[x]);
            display_message_process_byte(&rs232_data.display_message, buffer[x]);
            
            if (display_message_is_complete(&rs232_data.display_message))
            {
                if (rs232_data.display_callback != NULL)
                {
                    rs232_data.display_callback(&rs232_data.display_message);
                }

                display_message_initialise(&rs232_data.display_message);
            }
        }
    }
    else if (count == DRV_USART_READ_ERROR)
    {
       // There was an error. The DRV_USART_ErrorGet() function
       // can be called to find the exact error.        
       logger_log_info(LOGGER_CODE_RS232_DISPLAY, "RS232: Read Error");
    }
}

/**
 * Sends a buffer of data over the RS232 connection
 * @param buffer
 * @param len
 */
void rs232_display_send_buffer(char* buffer, size_t len)
{
    // This uses the asynchronous method of transferring the data as we were having 
    // trouble sending album art - it was keeling over after it send a few images 
    // Note the number of buffer items needs to match the number of Transmit buffers
    // defined in Harmony
    // Also the Harmony config needs Interrupt Mode On, and Buffer Queue support enabled
    // Send the buffer
     
    // Find a buffer
    uint8_t* write_buffer = &rs232_data.write_buffer[rs232_data.write_buffer_idx][0];
    
    memcpy(write_buffer, buffer, len);
    rs232_data.write_buffer_len[rs232_data.write_buffer_idx] = len;
    
    rs232_data.write_buffer_idx = ++rs232_data.write_buffer_idx % RS232_DISPLAY_TX_QUEUE_SIZE;
    
    //logger_log_debugf(0, "DISPLAY RS %d %d %d %d %d %d %d --", 
    //        buffer[0], buffer[1], buffer[2],buffer[3], buffer[4], buffer[5],buffer[6] );

    // Check buffer usage
    rs232_data.write_buffer_used_count++;    
    if ((uint16_t)rs232_data.write_buffer_used_count > ((uint16_t)RS232_DISPLAY_TX_QUEUE_SIZE * 9 / 10))
    {
        // If this happens then we need to increase RS232_DISPLAY_TX_QUEUE_SIZE (in Harmony))
        logger_log_warningf(LOGGER_CODE_RS232_DISPLAY, "RS232 Display Buffer filling up %d/%d --", rs232_data.write_buffer_used_count, RS232_DISPLAY_TX_QUEUE_SIZE);
    }
    
    // Place a write request
    DRV_USART_BUFFER_HANDLE buffer_handle;
    DRV_USART_BufferAddWrite(rs232_data.usart_handle, 
            &buffer_handle,
            write_buffer, 
            RS232_DISPLAY_TX_BUFFER_SIZE);//????????????????len); 
    
    if (buffer_handle == DRV_USART_BUFFER_HANDLE_INVALID)
    {
        logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Display unable to write data --");
    }
    
    //logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 Data Queued %d %d --", buffer_handle, len);
}

/* *****************************************************************************
 End of File
 */
