/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Arcam

  @File Name
    rs232.c

  @Summary
 Handles communication with the RS232 port

  @Description
 Maintains an RS232 connection and parses standard messages
 * 
 * v2.1 - Includes display_message processing
 *        This is a DYNAMIC implementation - to work the number of Driver Clients needs to be > 1 in Harmony (I used 4)
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/rs232_display.h"
#include "app_config.h"
#include "utilities/display_message_data.h"
#include "utilities/logger/logger.h"
#include "utilities/timers.h"
#include "driver/usart/drv_usart.h"

#include <xc.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Holds global data
 */
typedef struct
{
    // I2C handle
    DRV_HANDLE          usart_handle;
    
    SYS_MODULE_INDEX module;
    
    //! Callback for display messages
    DISPLAY_MESSAGE_CALLBACK display_callback;
    
    //! Display message to hold incoming Display message data
    DISPLAY_MESSAGE display_message;
    
    //! Buffers used for transferring large images - these need to be at least big enough
    //  to store an entire image
    uint8_t write_buffer[RS232_DISPLAY_TX_QUEUE_SIZE][RS232_DISPLAY_TX_BUFFER_SIZE];
    uint8_t write_buffer_len[RS232_DISPLAY_TX_QUEUE_SIZE];
    
    //! Simple circular buffer index
    uint8_t write_buffer_idx;
    bool write_buffer_complete[RS232_DISPLAY_TX_QUEUE_SIZE];
    DRV_USART_BUFFER_HANDLE write_buffer_handle[RS232_DISPLAY_TX_QUEUE_SIZE];
     
    //! Buffers used for transferring large images - these need to be at least big enough
    //  to store an entire image
    uint8_t read_buffer[RS232_DISPLAY_RX_QUEUE_SIZE][RS232_DISPLAY_RX_BUFFER_SIZE];
    bool read_buffer_complete[RS232_DISPLAY_RX_QUEUE_SIZE];
    DRV_USART_BUFFER_HANDLE read_buffer_handle[RS232_DISPLAY_RX_QUEUE_SIZE];
    
    //! Simple circular buffer index
    uint8_t read_buffer_read_idx;
} RS232_DATA;

static RS232_DATA rs232_data = {
    .read_buffer_read_idx = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * 
 * @param event
 * @param bufferHandle
 * @param contextHandle
 */
static void rs232_display_callback(DRV_USART_BUFFER_EVENT event,
        DRV_USART_BUFFER_HANDLE bufferHandle, 
        uintptr_t contextHandle)
{
    bool found = false;
    
    switch(event)
    {
        case DRV_USART_BUFFER_EVENT_COMPLETE:
        {
            // This means the data was transferred.
            //logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 Data Written %d --", bufferHandle);
            //logger_process();
            int x;
            for (x = 0; x < RS232_DISPLAY_RX_QUEUE_SIZE; x++)
            {
                if (rs232_data.read_buffer_handle[x] == bufferHandle)
                {
                    //logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 F/P Reading Data %d %d --", bufferHandle, rs232_data.read_buffer[x][0]);
                    rs232_data.read_buffer_complete[x] = true;
                    
                    found = true;
                    break;
                }
            }
            
            if (!found)
            {
                for (x = 0; x < RS232_DISPLAY_TX_QUEUE_SIZE; x++)
                {
                    if (rs232_data.write_buffer_handle[x] == bufferHandle)
                    {
                        rs232_data.write_buffer_complete[x] = true;

                        found = true;
                        break;
                    }
                }
            }
            
            break;
        }

        case DRV_USART_BUFFER_EVENT_ERROR:
        {
            DRV_USART_ERROR error = DRV_USART_ErrorGet(rs232_data.usart_handle);            

            logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Data Error %d %d --", bufferHandle, error);
            
            // Tran and get the error status note this is for the ST60 DRV_UART1 = U3
            logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Error STATE ST60 %d %d %d --", U3STAbits.PERR, U3STAbits.FERR, U3STAbits.OERR);    
            if (U3STAbits.OERR)
            {
                // Clear the error
                U3STAbits.OERR = 0;
            }
            
            // Tran and get the error status note this is for the ST60.Display DRV_UART0 = U2
            logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Error STATE ST60.Display %d %d %d --", U2STAbits.PERR, U2STAbits.FERR, U2STAbits.OERR);    
            if (U2STAbits.OERR)
            {
                // Clear the error
                U2STAbits.OERR = 0;
            }
            
            // Error handling here.
            int x;
            for (x = 0; x < RS232_DISPLAY_RX_QUEUE_SIZE; x++)
            {
                if (rs232_data.read_buffer_handle[x] == bufferHandle)
                {
                    logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Data Error Resetting %d --", bufferHandle);
                    rs232_data.read_buffer_complete[x] = false;
                    memset(rs232_data.read_buffer, 0xFD, RS232_DISPLAY_RX_BUFFER_SIZE);

                    DRV_USART_BUFFER_HANDLE buffer_handle;
                    DRV_USART_BufferAddRead(rs232_data.usart_handle, 
                            &buffer_handle,
                            &rs232_data.read_buffer[x][0], 
                            RS232_DISPLAY_RX_BUFFER_SIZE);

                    if (DRV_USART_BUFFER_HANDLE_INVALID != buffer_handle)
                    {
                        rs232_data.read_buffer_handle[x] = buffer_handle;
                    }
                    
                    if (x == rs232_data.read_buffer_read_idx)
                    {
                        // Increment the read index
                        logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Data Error Incr read index %d --", bufferHandle);
                        rs232_data.read_buffer_read_idx = ++rs232_data.read_buffer_read_idx % RS232_DISPLAY_RX_QUEUE_SIZE; 
                    }

                    //rs232_data.read_buffer_handle[x] = DRV_USART_BUFFER_HANDLE_INVALID;
                    found = true;
                    
                    display_message_initialise(&rs232_data.display_message);
                    break;
                }
            }
            
            if (!found)
            {
                for (x = 0; x < RS232_DISPLAY_TX_QUEUE_SIZE; x++)
                {
                    if (rs232_data.write_buffer_handle[x] == bufferHandle)
                    {
                        rs232_data.write_buffer_handle[x] = DRV_USART_BUFFER_HANDLE_INVALID;
                        rs232_data.write_buffer_complete[x] = true;

                        found = true;
                        break;
                    }
                }
            }

            break;
        }
        
        default:
            break;
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param idx
*/
static bool rs232_display_check_read_buffer(uint8_t idx)
{
    bool result = false;
    if (rs232_data.read_buffer_complete[idx])
    {
        // Added a brief pause as occasionally it does not get the data in non-blocking mode - done for upgrade but wasnt the issue
        //timers_delay_us(40);
        
        uint8_t data_length = rs232_data.read_buffer[idx][0];
        //logger_log_debugf(0, "DISPLAY Reading F/P %d %d %d --", rs232_data.read_buffer_handle[idx], idx, data_length);    
        if (data_length > RS232_DISPLAY_RX_BUFFER_SIZE - 1)
        {
            data_length = RS232_DISPLAY_RX_BUFFER_SIZE - 1;
            logger_log_debugf(0, "DISPLAY Reading F/P - Adjusting data length %d --", data_length);
        }
        
        int y;
        for (y = 1; y <= data_length; y++)
        {
            display_message_process_byte(&rs232_data.display_message, (unsigned char)rs232_data.read_buffer[idx][y]);

            if (display_message_is_complete(&rs232_data.display_message))
            {
                //logger_log_debugf(0, "DISPLAY Processing Complete %d %d --", idx, rs232_data.read_buffer[idx][y]);
                if (rs232_data.display_callback != NULL)
                {
                    //logger_log_debugf(0, "DISPLAY Processing Callback --");
                    rs232_data.display_callback(&rs232_data.display_message, true);
                }

                display_message_initialise(&rs232_data.display_message);
            }                        
            else if (display_message_has_errors(&rs232_data.display_message))
            {
                //logger_log_debugf(0, "DISPLAY Processing Complete %d %d --", idx, rs232_data.read_buffer[idx][y]);
                if (rs232_data.display_callback != NULL)
                {
                    //logger_log_debugf(0, "DISPLAY Processing Callback --");
                    rs232_data.display_callback(&rs232_data.display_message, false);
                }

                display_message_initialise(&rs232_data.display_message);
            }                        
        }

        //DRV_USART_BufferRemove(rs232_data.read_buffer_handle[idx]);

        //USART_ReceiverOverrunErrorClear_Default(USART_ID_1);    
        
        if(DRV_USART_ClientStatus(rs232_data.usart_handle) != DRV_USART_CLIENT_STATUS_READY)
        {
            logger_log_debugf(0, "DISPLAY Processing DRV_USART_ClientStatus %d --", DRV_USART_ClientStatus(rs232_data.usart_handle));
        }
        
        //_DRV_USART_ErrorConditionClear(rs232_data.module);

        // Re add the buffer
        rs232_data.read_buffer_complete[idx] = false;
        memset(&rs232_data.read_buffer[idx][0], 0xFD, RS232_DISPLAY_RX_BUFFER_SIZE);

        DRV_USART_BUFFER_HANDLE buffer_handle;
        DRV_USART_BufferAddRead(rs232_data.usart_handle, 
                &buffer_handle,
                &rs232_data.read_buffer[idx][0], 
                RS232_DISPLAY_RX_BUFFER_SIZE);

        if (DRV_USART_BUFFER_HANDLE_INVALID != buffer_handle)
        {
            rs232_data.read_buffer_handle[idx] = buffer_handle;
        }
        
        result = true;
    }
    
    logger_process();
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * RS232 module initialisation
 */
void rs232_display_initialise(const SYS_MODULE_INDEX module, 
        DISPLAY_MESSAGE_CALLBACK display_callback)
{
    rs232_data.display_callback = display_callback;
    display_message_initialise(&rs232_data.display_message);
    
    rs232_data.module = module;

    rs232_data.usart_handle = DRV_USART_Open(module, 
                                             DRV_IO_INTENT_NONBLOCKING | DRV_IO_INTENT_READWRITE);
    if (DRV_HANDLE_INVALID == rs232_data.usart_handle)
    {
        //settings_data.state = SETTINGS_STATE_ERROR;
        logger_log_error(LOGGER_CODE_RS232_DISPLAY, "USART Open");
    }    
    
    DRV_USART_BufferEventHandlerSet(rs232_data.usart_handle,
                rs232_display_callback, (uintptr_t)NULL);
    
    // Add read buffers
    int x;
    for (x = 0; x < RS232_DISPLAY_RX_QUEUE_SIZE; x++)
    {
        rs232_data.read_buffer_complete[x] = false;
        memset(rs232_data.read_buffer, 0xFD, RS232_DISPLAY_RX_BUFFER_SIZE);
        
        DRV_USART_BUFFER_HANDLE buffer_handle;
        DRV_USART_BufferAddRead(rs232_data.usart_handle, 
                &buffer_handle,
                &rs232_data.read_buffer[x][0], 
                RS232_DISPLAY_RX_BUFFER_SIZE);

        if (DRV_USART_BUFFER_HANDLE_INVALID != buffer_handle)
        {
            rs232_data.read_buffer_handle[x] = buffer_handle;
        }
    }
    
    // Initialise write buffers
    for (x = 0; x < RS232_DISPLAY_TX_QUEUE_SIZE; x++)
    {
        rs232_data.write_buffer_complete[x] = true;
    }
}

/**-----------------------------------------------------------------------------
 * RS232 processing
 */
void rs232_display_process()
{
    // Look through then read buffers from the point that they were last read
    int x;
    for (x = 0; x < RS232_DISPLAY_RX_QUEUE_SIZE; x++)
    {
        // Check if we have read the buffer, if so move the read index on by 1
        if (rs232_display_check_read_buffer(rs232_data.read_buffer_read_idx))
        {
            // Increment the read index
            rs232_data.read_buffer_read_idx = ++rs232_data.read_buffer_read_idx % RS232_DISPLAY_RX_QUEUE_SIZE; 
        }
        else
        {
            // Otherwise break out;
            break;
        }
    }
}
   
/**-----------------------------------------------------------------------------
 * Resets the messsge, drops all outstanding buffers...
 */
void rs232_display_reset()
{
            
    //_DRV_USART_ErrorConditionClear(rs232_data.module);

    logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 Restarting UART --");
    DRV_USART_Close(rs232_data.usart_handle);
   
//    uint8_t repeat;
//
//    for (repeat = 0; repeat < 100; repeat++)
//    {
//        PLIB_USART_Disable(rs232_data.usart_handle);
//        Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); 
//        PLIB_USART_Enable(rs232_data.usart_handle);
//        Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); 
//    }
//    
    rs232_display_initialise(rs232_data.module, rs232_data.display_callback);
    
//   logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 Resetting Display Message --");
//    display_message_initialise(&rs232_data.display_message);
    
    timers_delay_ms(1);
}

/**-----------------------------------------------------------------------------
 * Sends a buffer of data over the RS232 connection
 * @param buffer
 * @param len
 */
void rs232_display_send_buffer(char* buffer, size_t len)
{
    // This uses the asynchronous method of transferring the data as we were having 
    // trouble sending album art - it was keeling over after it send a few images 
    // Note the number of buffer items needs to match the number of Transmit buffers
    // defined in Harmony
    // Also the Harmony config needs Interrupt Mode On, and Buffer Queue support enabled
    // Send the buffer
    
    char* buffer_ptr = buffer;
    size_t len_remaining = len;
    
    // Work out how many buffers we need
    uint8_t buffer_count = len % (RS232_DISPLAY_TX_BUFFER_SIZE - 1) == 0 ? (len / (RS232_DISPLAY_TX_BUFFER_SIZE - 1)) : (len / (RS232_DISPLAY_TX_BUFFER_SIZE - 1)) + 1;
    int x;
    
    //logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 SENDING DATA F/P %d/%d %d %d --", (uint8_t)buffer[4], (uint8_t)buffer[5], len, buffer_count);
    
    // Write to the buffers
    for (x = 0; x < buffer_count; x++)
    {
        // Check if the buffer has completed writing
        if (rs232_data.write_buffer_complete[rs232_data.write_buffer_idx])
        {
            //logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 WRITING DATA %d %d --", buffer_count, rs232_data.write_buffer_idx);
            // Find a buffer
            uint8_t* write_buffer = &rs232_data.write_buffer[rs232_data.write_buffer_idx][0];

            size_t bytes_to_write = min(len_remaining, RS232_DISPLAY_TX_BUFFER_SIZE - 1);

            // Clear the buffer as Tim has reported leftover stuff being transferred.
            memset(write_buffer, 0, RS232_DISPLAY_TX_BUFFER_SIZE);
            
            // Store the number of bytes to be written (note this is a uint8 so max buffee size if 255)
            write_buffer[0] = bytes_to_write;
            
            memcpy(&write_buffer[1], buffer_ptr, bytes_to_write);
            rs232_data.write_buffer_len[rs232_data.write_buffer_idx] = bytes_to_write;

            buffer_ptr += bytes_to_write;
            len_remaining -= bytes_to_write;

            // Place a write request
            DRV_USART_BUFFER_HANDLE buffer_handle;
            DRV_USART_BufferAddWrite(rs232_data.usart_handle, 
                    &buffer_handle,
                    write_buffer, 
                    //TODDO Try this....rs232_data.write_buffer_len[rs232_data.write_buffer_idx]);
                    RS232_DISPLAY_TX_BUFFER_SIZE); 

            if (buffer_handle != DRV_USART_BUFFER_HANDLE_INVALID)
            {
                rs232_data.write_buffer_complete[rs232_data.write_buffer_idx] = false;
            }
            else
            {
                logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Display unable to write data --");
            }
            
            rs232_data.write_buffer_handle[rs232_data.write_buffer_idx] = buffer_handle;

            //logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 Data Queued %d %d --", buffer_handle, len);
        }
        else
        {
            logger_log_errorf(LOGGER_CODE_RS232_DISPLAY, "RS232 Unable to write to buffer %d as its in use", rs232_data.write_buffer_idx);
        }

        // Move onto the next buffer
        rs232_data.write_buffer_idx = ++rs232_data.write_buffer_idx % RS232_DISPLAY_TX_QUEUE_SIZE;
    }
}

/* *****************************************************************************
 End of File
 */
