/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/spdif_rx/spdif_rx.h"
#include "app_config.h"
#include "utilities/events.h"
#include "utilities/spdif_rx/private/spdif_rx_data.h"
#include "utilities/spdif_rx/spdif_rx_defs.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Event raised when there is an Error detected at the Hardware level
 */
static void spdif_rx_error_event_callback(EVENT* event)
{
	// Do something event related...
    SPDIF_RX_HW_ERROR_EVENT_DATA* event_data = (SPDIF_RX_HW_ERROR_EVENT_DATA*)event->context_data;
    
    // First check that its not just a sample rate change (shouldn't be but...)
    if (event_data->error_identified)
    {
        if (event_data->pll_locked)
        {
            spdif_rx_data_set_sample_rate(SAMPLE_RATE_NOTVALID);
        }
        else
        {
            spdif_rx_data_set_sample_rate(SAMPLE_RATE_NOINPUT);
        }
    }
    else if (event_data->sample_rate_change)
    {
        // No need to do anything - the spdif_rx module will try and get the sample rate
        logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR Sample Rate ONLY --");

        // If were active then unmute outputs
        if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
        {
            logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Callback Refreshing Sample Rate ");

            spdif_rx_refresh_sample_rate();
        }                
    }
    else
    {
        if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
        {
            // Check for an invalid sample rate and retry if found TODO - Test
            SAMPLE_RATE sample_rate = spdif_rx_data_get_sample_rate();
            if (SAMPLE_RATE_VALUES[sample_rate] == 0)
            {
                spdif_rx_refresh_sample_rate();
            }
            else
            {
                // Unmute the inputs
            }
        }                
    }
        
    // Re-Raise the event
    events_raise(EVENT_MODULE_SPDIF_RX, SPDIF_RX_EVENT_ERROR, (void*)event->context_data);
}

/**-----------------------------------------------------------------------------
 * Event raised when there is an Sample Rate change detected at the Hardware level
 */
static void spdif_rx_sample_rate_event_callback(EVENT* event)
{
	// Do something event related...
    SPDIF_RX_SAMPLE_RATE_EVENT_DATA* event_data = (SPDIF_RX_SAMPLE_RATE_EVENT_DATA*)event->context_data;
    
    logger_log_debugf(0, "SPDIF RX SAMPLE RATE CALLBACK %d %d --", event_data->error_identified, event_data->sample_rate);            
    logger_process();
            
    spdif_rx_data_set_sample_rate(event_data->sample_rate);
            
    // Re-Raise the event
    events_raise(EVENT_MODULE_SPDIF_RX, SPDIF_RX_EVENT_SAMPLE_RATE, (void*)event->context_data);
}

/**-----------------------------------------------------------------------------
 * Event raised when there is an Sample Rate change detected at the Hardware level
 */
static void spdif_rx_switch_event_callback(EVENT* event)
{
	// Do something event related...
    SPDIF_RX_SWITCH_EVENT_DATA* event_data = (SPDIF_RX_SWITCH_EVENT_DATA*)event->context_data;
    
    logger_log_debugf(0, "SPDIF RX SWITCH CALLBACK %d --", event_data->input);            
    logger_process();
            
    // Re-Raise the event
    events_raise(EVENT_MODULE_SPDIF_RX, SPDIF_RX_EVENT_SWITCH, (void*)event->context_data);
    
    spdif_rx_data_set_in_switch(false);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Configures the SPDIF RX module.  Note this includes a callback to the 
 * application to define which hardware implementation will be used
 */
void spdif_rx_configure()
{
    // Invoke the hardware layer configuration method
    SPDIF_RX_DATA* data = spdif_rx_data_get();
    if (data->functions.configure_function != NULL)
    {
        data->functions.configure_function();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Configuration function not specified --");
    }
}

/**-----------------------------------------------------------------------------
 * Used to define the callbacks for the particular spdif receiver.  Called in app_initialise
 * and is passed the hw version of define callbacks
 * @param callback_data
 */
void spdif_rx_define_functions(SPDIF_RX_FUNCTION_DATA* function_data)
{
    SPDIF_RX_DATA* data = spdif_rx_data_get();
    
    // Copy the function pointers to the data module
    memcpy(&(data->functions), function_data, sizeof(SPDIF_RX_FUNCTION_DATA));
}

/**-----------------------------------------------------------------------------
 * Gets the current smoothed sample rate
 * @return 
 */
SAMPLE_RATE spdif_rx_get_sample_rate()
{
    return spdif_rx_data_get_sample_rate();
}
    
/**
 * Handles interrupt on the SPDIF RX Interrupt pins
 * @param definition
 */
void spdif_rx_handle_interrupt(uint8_t channel)
{
    // Invoke the hardware layer configuration method
    SPDIF_RX_DATA* data = spdif_rx_data_get();
    if (data->functions.interrupt_function != NULL)
    {
        data->functions.interrupt_function(channel);
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Interrupt function not specified --");
    }    
}

/**-----------------------------------------------------------------------------
 * Returns true if the spdif_rx module is in a switch
 * @return 
 */
bool spdif_rx_in_switch()
{
    return spdif_rx_data_get_in_switch();
}

/**-----------------------------------------------------------------------------
 * Performs cyclic SPDIF_RX processing
 */
bool spdif_rx_initialise()
{
    bool result = false;
    
    // Invoke the hardware layer configuration method
    SPDIF_RX_DATA* data = spdif_rx_data_get();

	// Register the events that the spdif_rx module supports in the EVENTS structure, 
	events_register(EVENT_MODULE_SPDIF_RX, data->event_subscriptions, SPDIF_RX_MAX_EVENT_SUBSCRIPIONS);
    
    if (data->functions.initialise_function != NULL)
    {
        result = data->functions.initialise_function();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Initialise callback not specified --");
    }    
    
    // Subscribe to the events raise in the HW layer.  Note the Event Module ID should
    // work for either HW type 
	events_subscribe(EVENT_MODULE_SPDIF_RX_HW, 
			 		 SPDIF_RX_EVENT_ERROR, 
					 spdif_rx_error_event_callback, 
					 (void*)NULL, 
					 true);
    
	events_subscribe(EVENT_MODULE_SPDIF_RX_HW, 
                     SPDIF_RX_EVENT_SAMPLE_RATE, 
					 spdif_rx_sample_rate_event_callback, 
					 (void*)NULL, 
					 true);
    
	events_subscribe(EVENT_MODULE_SPDIF_RX_HW, 
                     SPDIF_RX_EVENT_SWITCH, 
					 spdif_rx_switch_event_callback, 
					 (void*)NULL, 
					 true);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs cyclic SPDIF_RX processing
 */
void spdif_rx_process()
{
    if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_DIGITAL)
    {
        if (timers_has_expired(TIMER_INDEX_SPDIF_RX_POLL))
        {
            // Invoke the hardware layer configuration method
            SPDIF_RX_DATA* data = spdif_rx_data_get();
            if (data->functions.process_function != NULL)
            {
                data->functions.process_function();
            }
            else
            {
                logger_log_warningf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Process callback not specified --");
            }    

            timers_reset(TIMER_INDEX_SPDIF_RX_POLL);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Requests that th SPDIF_RX module refresh its sample rate
 */
void spdif_rx_refresh_sample_rate(/* May be needs a callback when complete */)
{
    logger_log_debugf(0, "SPDIF RX spdif_rx_refresh_sample_rate --");
    logger_process();
    
    // Invoke the hardware layer configuration method
    SPDIF_RX_DATA* data = spdif_rx_data_get();
    if (data->functions.sample_rate_function != NULL)
    {
        if (!data->functions.sample_rate_function())
        {
        }
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Sample Rate function not specified --");
    }    
}

/**-----------------------------------------------------------------------------
 * Put SPDIF RX in or out of reset
 * @param on
 */
void spdif_rx_reset(bool on)
{
    // Invoke the hardware layer configuration method
    SPDIF_RX_DATA* data = spdif_rx_data_get();
    if (data->functions.reset_function != NULL)
    {
logger_log_debugf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Reset B %d --", data->functions.reset_function);
logger_process();
        data->functions.reset_function(on);
logger_log_debugf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Reset C --");
logger_process();
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Reset callback not specified --");
    }    
}

/**-----------------------------------------------------------------------------
 * Sets the SPDIF Rx ADC Mode for Master/Slave and Power
 * @param mode
 */
void spdif_rx_set_adc_mode(SPDIF_RX_ADC_MODE mode)
{
    // Invoke the hardware layer configuration method
    SPDIF_RX_DATA* data = spdif_rx_data_get();
    if (data->functions.adc_mode_function != NULL)
    {
        data->functions.adc_mode_function(mode);
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_SPDIF_RX, "SPDIF RX ADC Mode callback not specified --");
    }    
}

/**-----------------------------------------------------------------------------
 * Switch to a specific input on the digital input control
 * @param definition - the definition of the control to switch to
 */
void spdif_rx_switch(INPUT_CONTROL_DEFINITION* definition)
{
    // See input_control_switch_digital - does this wan to tak an INPUT_CONTROL_DEF or something more specific....
    // Invoke the hardware layer configuration method
    SPDIF_RX_DATA* data = spdif_rx_data_get();
    if (data->functions.switch_function != NULL)
    {
        spdif_rx_data_set_in_switch(true);
        
        data->functions.switch_function(definition);
    }
    else
    {
        logger_log_warningf(LOGGER_CODE_SPDIF_RX, "SPDIF RX Switch callback not specified --");
    }    
}


/* *****************************************************************************
 End of File
 */
