/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/spdif_rx/spdif_rx_hw_ak4113.h"
#include "app_config.h"
#include "utilities/events.h"
#include "utilities/i2c.h"
#include "utilities/input_control_i2c.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define SPDIF_RX_AK4113_NUM_SETTINGS            7

//! Defines the maximum number o event subscriptions that can be made to the events 
// in this module - this is set to 3 one for each event as its only the spdif_rx module 
// accesses them
#define SPDIF_RX_AK4113_HW_MAX_EVENT_SUBSCRIPIONS  3

    
/**
 * Converts AKMN4113 Sample rates (Register 0x08 bits 4-7) to our internal sample rates
 */
#define AK4113_SAMPLE_RATE_CONVERSIONS \
	X(0, SAMPLE_RATE_44100)\
	X(1, SAMPLE_RATE_NOTSUPPORTED)\
    X(2, SAMPLE_RATE_48000)\
   	X(3, SAMPLE_RATE_32000)\
    X(4, SAMPLE_RATE_NOTSUPPORTED)\
   	X(5, SAMPLE_RATE_NOTSUPPORTED)\
	X(6, SAMPLE_RATE_NOTSUPPORTED)\
    X(7, SAMPLE_RATE_NOTSUPPORTED)\
    X(8, SAMPLE_RATE_88200)\
    X(9, SAMPLE_RATE_NOTSUPPORTED)\
    X(10, SAMPLE_RATE_96000)\
    X(11, SAMPLE_RATE_NOTSUPPORTED)\
    X(12, SAMPLE_RATE_176400)\
    X(13, SAMPLE_RATE_NOTSUPPORTED)\
    X(14, SAMPLE_RATE_192000)\
    X(15, SAMPLE_RATE_NOTSUPPORTED)\
    
// Redefine the macro and use to create an array of int
#define X(a, b) b,
	static AK4113_SAMPLE_RATE_CONVERSION_VALUES[] = { AK4113_SAMPLE_RATE_CONVERSIONS };
#undef X

typedef struct
{
    //! I2C Client Id for identifying the module on the I2C bus
    I2C_CLIENT_ID i2c_client_id;

    //! Register setting definitions
    INPUT_CONTROL_I2C_REQUEST settings[SPDIF_RX_AK4113_NUM_SETTINGS];
    
    SPDIF_RX_FUNCTION_DATA functions;
    
    //! used to monitor the number of settings that have been applied in apply_settings
    uint8_t active_setting_count;
    
    //! Indicats whether the settings have been applied
    bool settings_applied;
   
    //! Attempts t prevent interrupt handling being stacked up
    volatile bool in_interrupt;
    
    //! Keeps track of the state of the reset.  
    bool in_reset;
    
    //! True if the signal is currently invalid
    bool invalid_signal;
    
    // Set to true if the SPDIF RX post switch callback has just occurred.  
    // It appears that the SPDIF RX
    // returns a NOT SUPPORTED sample rate when queried just after a switch
    bool post_switch;
    
    //! The request structure used to gather data for any errors
    INPUT_CONTROL_I2C_REQUEST error_request;
    
    //! The request structure used to gather data for the sample rate calculation
    INPUT_CONTROL_I2C_REQUEST sample_rate_request;
    
    //! The request structure used to switch inputs
    INPUT_CONTROL_I2C_REQUEST switch_request;
    
	//! An Array of event subscription structures.  These ara managed by the events module when
	// invoking events.  There MUST be enough declared to handle all the concurrent event 
	// subscriptions for the module.  
    // In this case it will be 1 for each event as its only th spdif_rx module subscribing
	EVENT_SUBSCRIPTION event_subscriptions[SPDIF_RX_AK4113_HW_MAX_EVENT_SUBSCRIPIONS];    
    
} SPDIF_RX_AK4113_DATA;

static SPDIF_RX_AK4113_DATA spdif_rx_data = {
    .active_setting_count = 0,
    .settings_applied = false,
    .in_reset = false,
    .post_switch = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward Declarations
static bool spdif_rx_hw_ak4113_is_reset();

/**-----------------------------------------------------------------------------
 * Applies the Register settings
 */
static void spdif_rx_hw_ak4113_apply_settings()
{
    if (!spdif_rx_data.settings_applied)
    {
        logger_log_info(LOGGER_CODE_INITIALISATION, "INPUT CONTROL SPDIF RX AK4113 Applying Settings --");
        logger_process();
        
        // Suspend change notifications
        PLIB_PORTS_ChannelChangeNoticeDisable(SPDIF_RX_AK4113_CHANGE_NOTICE_PORT, 
                SPDIF_RX_AK4113_CHANGE_NOTICE_CHANNEL, 
                SPDIF_RX_AK4113_CHANGE_NOTICE_MASK);
        
        spdif_rx_data.active_setting_count += SPDIF_RX_AK4113_NUM_SETTINGS;
        
        input_control_i2c_apply_settings(spdif_rx_data.i2c_client_id, spdif_rx_data.settings, SPDIF_RX_AK4113_NUM_SETTINGS);
        //Moved to callback spdif_rx_data.settings_applied = true;
    }
}
 
/**-----------------------------------------------------------------------------
 * Initiates call to the AK4113 to get any errors
 */
static void spdif_rx_hw_ak4113_get_errors()
{
    if (!input_control_in_switch()) // Added to try and lower the number of errors during a switch
    {
        if (!spdif_rx_hw_ak4113_is_reset())
        {
            input_control_i2c_read_setting(spdif_rx_data.i2c_client_id,
                    &spdif_rx_data.error_request);
        }
        else if (spdif_rx_data.error_request.read_state != INPUT_CONTROL_I2C_REQUEST_STATE_INACTIVE)
        {
            //logger_log_debugf(0, "INPUT CONTROL SPDIF Rx AK4113 Get Error suppressed --");
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Used to get the state of the SPDIF Rx reset switch
 * @param on
 */
static bool spdif_rx_hw_ak4113_is_reset()
{
    return spdif_rx_data.in_reset;
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * @param request
 */
static void spdif_rx_hw_ak4113_error_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF Rx AK4113 Errorcallback %d %d --", request->index, request->read_state);
    
    if (input_control_i2c_callback_has_data(request))
    {
        // Unpack the event data then raise the event
        // First bdefin the data type for the error
        SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
        event_data.error_identified = false;
        event_data.sample_rate_change = false;
        event_data.pll_locked = false;
        
        // Split up request.value which is the state of register 0x07 to determine 
        // the type of error..
        //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR %d --", request->value);

        // Report on the error
        bool par = false;
        if ((request->value & 0x01) == 0x01)
        {
            par = true;
        }

        bool audion = false;
        if ((request->value & 0x02) == 0x02)
        {
            audion = true;
        }

        bool stc = false;
        if ((request->value & 0x04) == 0x04)
        {
            stc = true;
        }

        bool v = false;
        if ((request->value & 0x08) == 0x08)
        {
            //v = true;
            // See if we have channel status checking switch on (engineering menu))
            settings_get(SETTING_CHECK_CHANNEL_STATUS, &v);
            if (!v)
            {
                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR Ignoring Channel Status error --");
            }
        }

        bool unlock = false;
        if ((request->value & 0x10) == 0x10)
        {
            unlock = true;
        }

        //CInt
        //if ((request->value & 0x20) == 0x20)
        //{
        //}

        // Auto
        bool auto_detect = false;
        if ((request->value & 0x40) == 0x40)
        {
            auto_detect = true;
        }

        //QINT
        //if ((request->value & 0x80) == 0x80)
        //{
        //}

        //logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR 0x%02X [par:%d aud:%d stc:%d v:%d, unlk:%d auto:%d] --",
        //        request->value, par, audion, stc, v, unlock, auto_detect);
        //logger_process();

        // First check that its not just a sample rate change (shouldn't be but...)
        if (par || audion || v || unlock || auto_detect)
        {
            event_data.error_identified = true;
            
            if (par || audion || v || auto_detect)
            {
                // Unlock errors can be ignored fo the moment - this is when thre is no signal
                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR 0x%02X [par:%d aud:%d stc:%d v:%d, unlk:%d auto:%d] --",
                    request->value, par, audion, stc, v, unlock, auto_detect);
            }

            if (!unlock)
            {
                event_data.pll_locked = true;
                spdif_rx_data_set_sample_rate(SAMPLE_RATE_NOTVALID);
            }
            else
            {
                spdif_rx_data_set_sample_rate(SAMPLE_RATE_NOINPUT);
            }

            // Set the internal invalid signal flag
            spdif_rx_data.invalid_signal = true;
        }
        else if (stc)
        {
            logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR Sample Rate ONLY --");

            event_data.sample_rate_change = true;

            // No errors so check if we currently think were muted and unmute
            spdif_rx_data.invalid_signal = false;
        }
       
        // Raise the event
        events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);
        
        input_control_i2c_callback_release(request);
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * A separate call is made for each register read for the address
 * @param request
 */
static void spdif_rx_hw_ak4113_sample_rate_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX sample_rate_callback %d %d ---", request->index, request->read_state);
    logger_process();
     
    if (input_control_i2c_callback_has_data(request))
    {
        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX sample_rate %d %d ---", spdif_rx_data_get_sample_rate(), AK4113_SAMPLE_RATE_CONVERSION_VALUES[request->value >> 4]);
        // Get the current value of the register
        if (spdif_rx_data_get_sample_rate() != AK4113_SAMPLE_RATE_CONVERSION_VALUES[request->value >> 4])
        {
            SAMPLE_RATE sample_rate = AK4113_SAMPLE_RATE_CONVERSION_VALUES[request->value >> 4];
            
            SPDIF_RX_SAMPLE_RATE_EVENT_DATA event_data;
            event_data.sample_rate = sample_rate;
            event_data.error_identified = false;
            event_data.is_valid = true;
            
            spdif_rx_data_set_sample_rate(sample_rate);

            logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate 0x%02X 0x%02X %s --", request->value, request->value >> 4, SAMPLE_RATE_STRINGS[sample_rate]);

            spdif_rx_data.invalid_signal = false;

            // Check for errors and trigger an error check if found
            if ((request->value & 0x0F) > 0x00)
            {
                logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate errors found 0x%02X --", request->value - (request->value >> 4));

                event_data.error_identified = true;
                
                // Set internal mute flag
                spdif_rx_data.invalid_signal = true;

                spdif_rx_hw_ak4113_get_errors();
            }

            // Set the post_switch flag Inform the Avatar of the new sample rate
            if (spdif_rx_data.post_switch && (sample_rate == SAMPLE_RATE_NOTSUPPORTED))
            {
                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL AVATAR Sample Rate suppressed --");
                event_data.is_valid = false;
            }
            else
            {
//TODO Remove                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL AVATAR Setting Sample Rate %d --", SAMPLE_RATE_VALUES[sample_rate]);
//                avatar_set_sample_rate2(SAMPLE_RATE_VALUES[sample_rate]);
//
                spdif_rx_data.post_switch = false;
            }
            
            // Raise the event
//logger_log_debugf(0, "SPFDIF RX HW SAMPLE RATE CALLBACK %d %d --", event_data.error_identified, event_data.sample_rate);            
//logger_process();
            events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_SAMPLE_RATE, (void*)&event_data);
        }
           
        spdif_rx_data_set_calculating_sample_rate(false);
        
        input_control_i2c_callback_release(request);
    }
}

/**-----------------------------------------------------------------------------
 * Used to set the Spdif Rx Master/Slave mode
 * @param reset
 * @param master
 */
static void spdif_rx_hw_ak4113_set_master_slave(bool master)
{
    logger_log_debugf(0, "INPUT CONTROL Set mode %s --", master ? "Master" : "Slave");
    INPUT_CONTROL_I2C_REQUEST* setting = input_control_i2c_definition_for_address(spdif_rx_data.settings, 
        SPDIF_RX_AK4113_NUM_SETTINGS, 
        0X01);
    
    if (setting != NULL)
    {
        if (!master)
        {
            // Slave mode used for Analogue and Avatar 0x6A 0110 1010
            setting->value &= ~(1 << 4);    // Set bit 4 OFF
            setting->value |= 1 << 5;       // Set bit 5 ON
        }
        else
        {
            // Normal mode 0x5A  0101 1010
            setting->value |= 1 << 4;       // Set bit 4 ON
            setting->value &= ~(1 << 5);    // Switch bit 5 OFF
        }

        logger_log_debugf(0, "SPDIF RX Disabling Interrupts A %d --", spdif_rx_data.active_setting_count);
        logger_process();
        
        PLIB_PORTS_ChannelChangeNoticeDisable(SPDIF_RX_AK4113_CHANGE_NOTICE_PORT, 
                SPDIF_RX_AK4113_CHANGE_NOTICE_CHANNEL, 
                SPDIF_RX_AK4113_CHANGE_NOTICE_MASK);
        
        spdif_rx_data.active_setting_count++;
        
        input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                    setting);
    }
}

/**-----------------------------------------------------------------------------
 * Used to set the SPdif Rx Power state
 * @param on
 */
static void spdif_rx_hw_ak4113_set_power(bool on)
{
    INPUT_CONTROL_I2C_REQUEST* setting = input_control_i2c_definition_for_address(spdif_rx_data.settings, 
        SPDIF_RX_AK4113_NUM_SETTINGS, 
        0X00);
    
    if (setting != NULL)
    {
        if (on)
        {
            setting->value = 0x23;

            spdif_rx_data.active_setting_count++;

            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                        setting);

            // Reapply the settings (just in case);
            spdif_rx_data.settings_applied = false;
            spdif_rx_hw_ak4113_apply_settings();

        }
        else
        {
            setting->value = 0x00;

            logger_log_debugf(0, "SPDIF RX Disabling Interrupts B %d --", spdif_rx_data.active_setting_count);
            logger_process();
            PLIB_PORTS_ChannelChangeNoticeDisable(SPDIF_RX_AK4113_CHANGE_NOTICE_PORT, 
                SPDIF_RX_AK4113_CHANGE_NOTICE_CHANNEL, 
                SPDIF_RX_AK4113_CHANGE_NOTICE_MASK);

            spdif_rx_data.active_setting_count++;
            
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                        setting);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Called when a setting has been applied
 * @param request
 */
static void spdif_rx_hw_ak4113_setting_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    spdif_rx_data.active_setting_count--;
    
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF Rx Settings applied %d %d %d --", 
            request->address,
            request->value,
            spdif_rx_data.active_setting_count);

    if (spdif_rx_data.active_setting_count == 0)
    {
        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX All Settings applied --");
        logger_process();
        
        // Switch on SPDIF Interrupts now that the unit is configured
        // If they are not set on before then we get a load of interrupts while initialising
        // for Ports RF0/RF1
        PLIB_PORTS_ChannelChangeNoticeEnable(SPDIF_RX_AK4113_CHANGE_NOTICE_PORT, 
                SPDIF_RX_AK4113_CHANGE_NOTICE_CHANNEL, 
                SPDIF_RX_AK4113_CHANGE_NOTICE_MASK);
        
        spdif_rx_data.settings_applied = true;
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * @param request
 */
static void spdif_rx_hw_ak4113_switch_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX AK4113 Switch Callback ");
    
    // Set the post switch flag to suppress reporting of NOT SUPPORTED sample rate
    // in input_control_calculate_digital_sample_rate if we have just
    // switched the SPDIX RX Rx
    spdif_rx_data.post_switch = true;
    
    SPDIF_RX_SWITCH_EVENT_DATA event_data;
    event_data.input = request->value;  // Note this is the whol registr valu - neds changing to just the bits representing the input (if needed))
    events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_SWITCH, (void*)&event_data);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Configure the ak4113 ADC Mode
 */
void spdif_rx_hw_ak4113_adc_mode(SPDIF_RX_ADC_MODE mode)
{
    switch(mode)
    {
        case SPDIF_RX_ADC_MODE_POWER_ON_MASTER:
        {
            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX SPDIF_RX_ADC_MODE_POWER_ON_MASTER ---");
            spdif_rx_hw_ak4113_set_master_slave(true);
            spdif_rx_hw_ak4113_set_power(true);
            break;
        }
        
        case SPDIF_RX_ADC_MODE_POWER_OFF_MASTER:
        {
            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX SPDIF_RX_ADC_MODE_POWER_OFF_MASTER ---");
            spdif_rx_hw_ak4113_set_master_slave(true);
            spdif_rx_hw_ak4113_set_power(false);
            break;
        }
        
        case SPDIF_RX_ADC_MODE_POWER_ON_SLAVE:
        {
            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX SPDIF_RX_ADC_MODE_POWER_ON_SLAVE ---");
            spdif_rx_hw_ak4113_set_master_slave(false);
            spdif_rx_hw_ak4113_set_power(true);
            break;
        }
        
        case SPDIF_RX_ADC_MODE_POWER_OFF_SLAVE:
        {
             logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX SPDIF_RX_ADC_MODE_POWER_OFF_SLAVE ---");
           spdif_rx_hw_ak4113_set_master_slave(false);
            spdif_rx_hw_ak4113_set_power(false);
            break;
        }
        
        default:
            break;

    }
}

/**-----------------------------------------------------------------------------
 * Configure the ak4113
 */
void spdif_rx_hw_ak4113_configure()
{
    logger_log_info(LOGGER_CODE_INITIALISATION, "INPUT CONTROL SPDIF RX AK4113 Configuration --");
    logger_process();

    input_control_i2c_add_setting(&spdif_rx_data.settings[0], 0X00, 0x23, (void*)spdif_rx_hw_ak4113_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[1], 0X01, 0x5A, (void*)spdif_rx_hw_ak4113_setting_callback);  // 00000000
    input_control_i2c_add_setting(&spdif_rx_data.settings[2], 0X02, 0x40, (void*)spdif_rx_hw_ak4113_setting_callback);  // 00000000
    input_control_i2c_add_setting(&spdif_rx_data.settings[3], 0X03, 0x50, (void*)spdif_rx_hw_ak4113_setting_callback);  // IP Select
    input_control_i2c_add_setting(&spdif_rx_data.settings[4], 0X04, 0xFB, (void*)spdif_rx_hw_ak4113_setting_callback);  // Sample Rate Interrupt Mask
    input_control_i2c_add_setting(&spdif_rx_data.settings[5], 0X05, 0xAD, (void*)spdif_rx_hw_ak4113_setting_callback);  // BUG 3558 - See Petes Email 29/07/2020 13:27 0xA4);  // Error Interrupt Mask
    input_control_i2c_add_setting(&spdif_rx_data.settings[6], 0X06, 0x1F, (void*)spdif_rx_hw_ak4113_setting_callback);  // 00000000

    input_control_i2c_define_register(&spdif_rx_data.error_request, (REGISTER_BITS_DEFINITION){ 0x07, 0, 7 }); 
    input_control_i2c_define_register(&spdif_rx_data.switch_request, (REGISTER_BITS_DEFINITION){ 0x03, 0, 2 }); 
    input_control_i2c_define_register(&spdif_rx_data.sample_rate_request, (REGISTER_BITS_DEFINITION){ 0x08, 4, 7 });
    logger_log_info(LOGGER_CODE_INITIALISATION, "INPUT CONTROL SPDIF RX AK4113 Configuration A --");
    logger_process();

    spdif_rx_hw_ak4113_apply_settings();    
    logger_log_info(LOGGER_CODE_INITIALISATION, "INPUT CONTROL SPDIF RX AK4113 Configuration B --");
    logger_process();
}

/**-----------------------------------------------------------------------------
 * Defines the SPDIF_RX callbacks used when using the AK4113.  To be used as
 * the parameter to spdif_rx_define_callbacks() function
 * @return 
 */    
SPDIF_RX_FUNCTION_DATA* spdif_rx_hw_ak4113_define_functions()
{
    // Set up the functions here
    spdif_rx_data.functions.adc_mode_function = spdif_rx_hw_ak4113_adc_mode;
    spdif_rx_data.functions.configure_function = spdif_rx_hw_ak4113_configure;
    spdif_rx_data.functions.initialise_function = spdif_rx_hw_ak4113_initialise;
    spdif_rx_data.functions.interrupt_function = spdif_rx_hw_ak4113_interrupt;
    spdif_rx_data.functions.process_function = spdif_rx_hw_ak4113_process;
    spdif_rx_data.functions.reset_function = spdif_rx_hw_ak4113_reset;
    spdif_rx_data.functions.sample_rate_function = spdif_rx_hw_ak4113_sample_rate;
    spdif_rx_data.functions.switch_function = spdif_rx_hw_ak4113_switch;
            
    return &spdif_rx_data.functions;
}
 
/**-----------------------------------------------------------------------------
 * Initialise the AK4113
 */
bool spdif_rx_hw_ak4113_initialise()
{
    bool result = true;

    spdif_rx_data.error_request.callback = spdif_rx_hw_ak4113_error_callback;
    spdif_rx_data.switch_request.callback = spdif_rx_hw_ak4113_switch_callback;
    spdif_rx_data.sample_rate_request.callback = spdif_rx_hw_ak4113_sample_rate_callback;    

	// Register the events in the EVENTS structure, 
	events_register(EVENT_MODULE_SPDIF_RX_HW, spdif_rx_data.event_subscriptions, SPDIF_RX_AK4113_HW_MAX_EVENT_SUBSCRIPIONS);
    
    i2c_initialise();
     
    spdif_rx_data.i2c_client_id = i2c_open(SPDIF_RX_AK4113_I2C_MODULE, 
            SPDIF_RX_AK4113_I2C_READ_ADDRESS,
            SPDIF_RX_AK4113_I2C_WRITE_ADDRESS,
            input_control_i2c_callback);
    
    if (I2C_INVALID_CLIENT_ID != spdif_rx_data.i2c_client_id)
    {
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX AK4113 I2C Open");
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX AK4113 I2C Open Error");
        result = false;
    }    
    
    return result;
}
     
/**-----------------------------------------------------------------------------
 * Handles interrupt on the SPDIF RX Interrupt pins
 * @param definition
 */
void spdif_rx_hw_ak4113_interrupt(uint8_t channel)
{
    if (!spdif_rx_data.in_interrupt)
    {
        spdif_rx_data.in_interrupt = true;

        bool error_check_reqd = true;

        if (channel == SPDIF_RX_AK4113_INT0_PORT_CHANNEL)
        {
            //logger_log_debugf(0, "INPUT CONTROL INTERRUPT --");
            // Read the port to clear the interrupt
            uint32_t value = PLIB_PORTS_Read(SPDIF_RX_AK4113_INT0_ID,
                                             SPDIF_RX_AK4113_INT0_PORT_CHANNEL);

            // Check it was the sample rate pin that triggered the interrupt
            bool sample_rate_change = (value & (1 << SPDIF_RX_AK4113_INT0_PIN)) != 0;
            if (sample_rate_change)
            {
                SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
                event_data.error_identified = true;
                event_data.sample_rate_change = true;
                event_data.pll_locked = true;
        
                // Raise the event
                events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);

                spdif_rx_refresh_sample_rate();

                // Note the sample rate calculation will trigger a check for errors
                error_check_reqd = false;
            }

            bool error_change = (value & (1 << SPDIF_RX_AK4113_INT1_PIN)) != 0;
            if (error_change)
            {
                SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
                event_data.error_identified = true;
                event_data.sample_rate_change = false;
                event_data.pll_locked = true;
        
                // Raise the event
                events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);
            }
            
            // Disable Interrupts and move to polling as this goes into a loop
            //PLIB_PORTS_ChannelChangeNoticeDisable(PORTS_ID_0, PORT_CHANNEL_F, SYS_PORT_F_CNEN || 0x03);

        }
        else if (channel == SPDIF_RX_AK4113_INT1_PORT_CHANNEL)
        {
            //logger_log_debugf(0, "INPUT CONTROL INTERRUPT ERROR --");
            // Read the other interrupt port to clear it.
            // Note this would clear the interrupts for any other peripherals 
            // monitoring on this port.
            bool error_detected = PLIB_PORTS_PinGet(SPDIF_RX_AK4113_INT1_ID, 
                                                    SPDIF_RX_AK4113_INT1_PORT_CHANNEL, 
                                                    SPDIF_RX_AK4113_INT1_PIN);

            bool error_change = (error_detected & (1 << SPDIF_RX_AK4113_INT1_PIN)) != 0;
            if (error_change)
            {
                SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
                event_data.error_identified = true;
                event_data.sample_rate_change = false;
                event_data.pll_locked = true;
        
                // Raise the event
                events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);
            }
        }

        spdif_rx_data.in_interrupt = false;
    }
}

/**-----------------------------------------------------------------------------
 * Perform AK4113 processing
 */
void spdif_rx_hw_ak4113_process()
{
    spdif_rx_hw_ak4113_get_errors();   
}

/**-----------------------------------------------------------------------------
 * Perform AK4113 reset
 */
void spdif_rx_hw_ak4113_reset(bool on)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Reset - Switching %s --", on == true ? "On" : "Off");
    logger_process();
    
    PLIB_PORTS_PinWrite(SPDIF_RX_AK4113_RESET_PORTS_ID, 
                        SPDIF_RX_AK4113_RESET_PORT_CHANNEL, 
                        SPDIF_RX_AK4113_RESET_PIN,
                        on); // Note pin is low

    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Reset - Switching A %s --", on == true ? "On" : "Off");
    logger_process();
    spdif_rx_data.in_reset = !on;
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Reset - Switching %s B --", on == true ? "On" : "Off");
    logger_process();
    
    if (on)
    {
        // Setup for a resend of the settings
        spdif_rx_data.settings_applied = false;
    }
}

/**-----------------------------------------------------------------------------
 * Trigger AK4113 sample rate calculation 
 */
bool spdif_rx_hw_ak4113_sample_rate()
{
    bool result = false;
    
    if (!spdif_rx_data_get_calculating_sample_rate())
    {
        logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate --");

        // Shouldn't need to check whether this is already running as that is done in input_control_i2c_read_setting
        if (!spdif_rx_hw_ak4113_is_reset())
        {
            // Set before running in-case the sample rate calculation is completd
            // before the end of the function call.
            spdif_rx_data_set_calculating_sample_rate(true);
            
            input_control_i2c_read_setting(spdif_rx_data.i2c_client_id,
                    &spdif_rx_data.sample_rate_request);

            result = true;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Perform AK4113 switch 
 */
void spdif_rx_hw_ak4113_switch(INPUT_CONTROL_DEFINITION* definition)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Switch --");

    uint8_t new_value = definition->value << spdif_rx_data.switch_request.bit_low;
    
    // Find the setting Request that matches the switch request as that also has the current value
    // to get the default value.
    // Check there isnt a default for this register, if there is then OR is 
    // (assumes default is only in the lower bits)..
    INPUT_CONTROL_I2C_REQUEST* setting = input_control_i2c_definition_for_address(spdif_rx_data.settings, 
            SPDIF_RX_AK4113_NUM_SETTINGS, 
            spdif_rx_data.switch_request.address);
    if (setting != NULL)
    {
        new_value |= setting->value;
    }

    // Check if we have had an invalid signal and if so mute outputs before switching
    // to stop noise..
    if (spdif_rx_data.invalid_signal)
    {
        logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Switch Digital errors MUTING OUTPUTS ");

        SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
        event_data.error_identified = true;
        event_data.sample_rate_change = false;
        event_data.pll_locked = true;

        // Raise the event
        events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);
    }

    // Reapply the settings (just in case);
//    spdif_rx_hw_ak4113_apply_settings();
    
    // Removed as was not catching analogue to digital switches
    //if (new_value != input_control_data.switch_request.value)
    //{
    // Note: address should have been defined
    spdif_rx_data.switch_request.actioned = false;
    spdif_rx_data.switch_request.value = new_value;
    
    input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
            &spdif_rx_data.switch_request);
}

/* *****************************************************************************
 End of File
 */
