/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/spdif_rx/spdif_rx_hw_tipcm9211.h"
#include "app_config.h"
#include "utilities/events.h"
#include "utilities/i2c.h"
#include "utilities/input_control_i2c.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define SPDIF_RX_TIPCM9211_NUM_SETTINGS             13

//! Defines the maximum number o event subscriptions that can be made to the events 
// in this module - this is set to 3 one for each event as its only the spdif_rx module 
// accesses them
#define SPDIF_RX_TIPCM9211_HW_MAX_EVENT_SUBSCRIPIONS  3

    
/**
 * Converts AKMN4113 Sample rates (Register 0x08 bits 4-7) to our internal sample rates
 */
#define TIPCM9211_SAMPLE_RATE_CONVERSIONS \
	X(0, SAMPLE_RATE_NOTSUPPORTED)\
	X(1, SAMPLE_RATE_NOTSUPPORTED)\
    X(2, SAMPLE_RATE_NOTSUPPORTED)\
   	X(3, SAMPLE_RATE_NOTSUPPORTED)\
    X(4, SAMPLE_RATE_NOTSUPPORTED)\
   	X(5, SAMPLE_RATE_NOTSUPPORTED)\
	X(6, SAMPLE_RATE_NOTSUPPORTED)\
    X(7, SAMPLE_RATE_32000)\
    X(8, SAMPLE_RATE_44100)\
    X(9, SAMPLE_RATE_48000)\
    X(10, SAMPLE_RATE_NOTSUPPORTED)\
    X(11, SAMPLE_RATE_88200)\
    X(12, SAMPLE_RATE_96000)\
    X(13, SAMPLE_RATE_NOTSUPPORTED)\
    X(14, SAMPLE_RATE_176400)\
    X(15, SAMPLE_RATE_192000)\
    
// Redefine the macro and use to create an array of int
#define X(a, b) b,
	static TIPCM9211_SAMPLE_RATE_CONVERSION_VALUES[] = { TIPCM9211_SAMPLE_RATE_CONVERSIONS };
#undef X

typedef struct
{
    //! I2C Client Id for identifying the module on the I2C bus
    I2C_CLIENT_ID i2c_client_id;

    //! Register setting definitions
    INPUT_CONTROL_I2C_REQUEST settings[SPDIF_RX_TIPCM9211_NUM_SETTINGS];
    
    SPDIF_RX_FUNCTION_DATA functions;
    
    //! used to monitor the number of settings that have ben aplied in apply_settings
    uint8_t active_setting_count;
    
    //! Indicats whether the settings have been applied
    bool settings_applied;
   
    //! Attempts t prevent interrupt handling being stacked up
    bool in_interrupt;
    
    //! Keeps track of the state of the reset.  
    bool in_reset;
    
    //! True if the signal is currently invalid
    bool invalid_signal;
    
    // Set to true if the SPDIF RX post switch callback has just occurred.  
    // It appears that the SPDIF RX
    // returns a NOT SUPPORTED sample rate when queried just after a switch
    bool post_switch;
    
    //! The request structure used to set the adc mode
    INPUT_CONTROL_I2C_REQUEST adc_mode_request[2];
    
    //! The request structure used to gather data for any errors
    INPUT_CONTROL_I2C_REQUEST error_request[2];
    
    //! The request structure used to gather data for the sample rate calculation
    INPUT_CONTROL_I2C_REQUEST sample_rate_request;
    
    //! The request structure used to switch inputs
    INPUT_CONTROL_I2C_REQUEST switch_request;
    
	//! An Array of event subscription structures.  These ara managed by the events module when
	// invoking events.  There MUST be enough declared to handle all the concurrent event 
	// subscriptions for the module.  
    // In this case it will be 1 for each event as its only th spdif_rx module subscribing
	EVENT_SUBSCRIPTION event_subscriptions[SPDIF_RX_TIPCM9211_HW_MAX_EVENT_SUBSCRIPIONS];    
    
} SPDIF_RX_TIPCM9211_DATA;

static SPDIF_RX_TIPCM9211_DATA spdif_rx_data = {
    .active_setting_count = 0,
    .settings_applied = false,
    .in_reset = false,
    .post_switch = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward Declarations
static bool spdif_rx_hw_tipcm9211_is_reset();

/**-----------------------------------------------------------------------------
 * Applies the Register settings
 */
static void spdif_rx_hw_tipcm9211_apply_settings()
{
    if (!spdif_rx_data.settings_applied)
    {
        logger_log_info(LOGGER_CODE_INITIALISATION, "INPUT CONTROL SPDIF RX TIPCM9211 Applying Settings --");
        logger_process();
        
        // Suspend change notifications
        PLIB_PORTS_ChannelChangeNoticeDisable(SPDIF_RX_TIPCM9211_CHANGE_NOTICE_PORT, 
                SPDIF_RX_TIPCM9211_CHANGE_NOTICE_CHANNEL, 
                SPDIF_RX_TIPCM9211_CHANGE_NOTICE_MASK);
        
        spdif_rx_data.active_setting_count += SPDIF_RX_TIPCM9211_NUM_SETTINGS;
        
        input_control_i2c_apply_settings(spdif_rx_data.i2c_client_id, spdif_rx_data.settings, SPDIF_RX_TIPCM9211_NUM_SETTINGS);
        spdif_rx_data.settings_applied = true;
    }
}
 
/**-----------------------------------------------------------------------------
 * Initiates call to the TIPCM9211 to get any errors
 */
static void spdif_rx_hw_tipcm9211_get_errors()
{
    if (!input_control_in_switch()) // Added to try and lower the number of errors during a switch
    {
        if (!spdif_rx_hw_tipcm9211_is_reset())
        {
            input_control_i2c_read_setting(spdif_rx_data.i2c_client_id,
                    &spdif_rx_data.error_request[0]);
            
            input_control_i2c_read_setting(spdif_rx_data.i2c_client_id,
                    &spdif_rx_data.error_request[1]);
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Used to get the state of the SPDIF Rx reset switch
 * @param on
 */
static bool spdif_rx_hw_tipcm9211_is_reset()
{
    return spdif_rx_data.in_reset;
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * @param request
 */
static void spdif_rx_hw_tipcm9211_adc_mode_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF Rx TIPCM9211 Errorcallback %d %d --", request->index, request->read_state);
    
    if (input_control_i2c_callback_has_data(request))
    {
        logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF Rx TIPCM9211 ADC Mode callback %d %d --", request->index, request->read_state);
        input_control_i2c_callback_release(request);
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * @param request
 */
static void spdif_rx_hw_tipcm9211_error_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    //logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF Rx TIPCM9211 Errorcallback %d %d --", request->index, request->read_state);
    
    if (input_control_i2c_callback_has_data(request))
    {
        // Unpack the event data then raise the event
        // First bdefin the data type for the error
        SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
        event_data.error_identified = false;
        event_data.sample_rate_change = false;
        event_data.pll_locked = false;
        
        // Check which error_request it is
        if (request == &spdif_rx_data.error_request[0])
        {
            // This is a sample change request
            logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR SAMPLE_RATE CHANGE --");

            event_data.sample_rate_change = true;
            event_data.pll_locked = false;  // Assume that this is a Unlocked error until h sample rate has been calculated
            
            // No errors so check if we currently think were muted and unmute
            spdif_rx_data.invalid_signal = false;

            spdif_rx_hw_tipcm9211_sample_rate();
        }
        else if (request == &spdif_rx_data.error_request[1])
        {
            // This is an error
            logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX ERROR --");
            
            event_data.error_identified = true;
            
            spdif_rx_data_set_sample_rate(SAMPLE_RATE_NOTVALID);
            
            spdif_rx_data.invalid_signal = true;
        }
       
        // Raise the event
        events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);
        
        input_control_i2c_callback_release(request);
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * A separate call is made for each register read for the address
 * @param request
 */
static void spdif_rx_hw_tipcm9211_sample_rate_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX sample_rate_callback %d %d ---", request->index, request->read_state);
    logger_process();
     
    if (input_control_i2c_callback_has_data(request))
    {
        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX sample_rate %d %d ---", spdif_rx_data_get_sample_rate(), TIPCM9211_SAMPLE_RATE_CONVERSION_VALUES[request->value >> 4]);

        uint8_t sample_rate_idx = request->value & 0x0F;
        
        // Get the current value of the register
        if (spdif_rx_data_get_sample_rate() != TIPCM9211_SAMPLE_RATE_CONVERSION_VALUES[sample_rate_idx])
        {
            SAMPLE_RATE sample_rate = TIPCM9211_SAMPLE_RATE_CONVERSION_VALUES[sample_rate_idx];
            
            SPDIF_RX_SAMPLE_RATE_EVENT_DATA event_data;
            event_data.sample_rate = sample_rate;
            event_data.error_identified = false;
            event_data.is_valid = true;
            
            spdif_rx_data_set_sample_rate(sample_rate);

            // Check for errors and trigger an error check if found
            if ((request->value & 0x80) > 0x00)
            {
                logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate errors found 0x%02X --", request->value - (request->value >> 4));

                event_data.error_identified = true;
                
                // Set internal mute flag
                spdif_rx_data.invalid_signal = true;

                // Do what we do if its unlocked - thats what the 0x80 state indicates
                SPDIF_RX_HW_ERROR_EVENT_DATA error_event_data;
                error_event_data.error_identified = true;
                error_event_data.sample_rate_change = true;
                error_event_data.pll_locked = true;
                
                spdif_rx_data_set_sample_rate(SAMPLE_RATE_NOINPUT);

                // Raise the event
                events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&error_event_data);
            }

            // Set the post_switch flag Inform the Avatar of the new sample rate
            if (spdif_rx_data.post_switch && (sample_rate == SAMPLE_RATE_NOTSUPPORTED))
            {
                logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL AVATAR Sample Rate suppressed --");
                event_data.is_valid = false;
            }
            else
            {
                spdif_rx_data.post_switch = false;
            }
            
            // Raise the event
//logger_log_debugf(0, "SPFDIF RX HW SAMPLE RATE CALLBACK %d %d --", event_data.error_identified, event_data.sample_rate);            
//logger_process();
            events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_SAMPLE_RATE, (void*)&event_data);
        }
           
        spdif_rx_data_set_calculating_sample_rate(false);
        
        input_control_i2c_callback_release(request);
    }
}

/**-----------------------------------------------------------------------------
 * Called when a setting has been applied
 * @param request
 */
static void spdif_rx_hw_tipcm9211_setting_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    spdif_rx_data.active_setting_count--;
    
    logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF Rx Settings applied %d %d %d --", 
            request->address,
            request->value,
            spdif_rx_data.active_setting_count);

    if (spdif_rx_data.active_setting_count == 0)
    {
        logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX All Settings applied --");
        logger_process();
        
        // Switch on SPDIF Interrupts now that the unit is configured
        // If they are not set on before then we get a load of interrupts while initialising
        // for Ports RF0/RF1
        PLIB_PORTS_ChannelChangeNoticeEnable(SPDIF_RX_TIPCM9211_CHANGE_NOTICE_PORT, 
                SPDIF_RX_TIPCM9211_CHANGE_NOTICE_CHANNEL, 
                SPDIF_RX_TIPCM9211_CHANGE_NOTICE_MASK);
    }
}

/**-----------------------------------------------------------------------------
 * Called when the input has been changed
 * @param request
 */
static void spdif_rx_hw_tipcm9211_switch_callback(INPUT_CONTROL_I2C_REQUEST* request)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX TIPCM9211 Switch Callback ");
    
    // Set the post switch flag to suppress reporting of NOT SUPPORTED sample rate
    // in input_control_calculate_digital_sample_rate if we have just
    // switched the SPDIX RX Rx
    spdif_rx_data.post_switch = true;
    
    SPDIF_RX_SWITCH_EVENT_DATA event_data;
    event_data.input = request->value;  // Note this is the whole registr valu - neds changing to just the bits representing the input (if needed))
    events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_SWITCH, (void*)&event_data);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Configure the tipcm9211 ADC Mode
 */
void spdif_rx_hw_tipcm9211_adc_mode(SPDIF_RX_ADC_MODE mode)
{
    switch(mode)
    {
        case SPDIF_RX_ADC_MODE_POWER_ON_MASTER:
        {
            // Digital Input
            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX SPDIF_RX_ADC_MODE_POWER_ON_MASTER ---");
            
            spdif_rx_data.adc_mode_request[0].value = 0xF0;
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                &spdif_rx_data.adc_mode_request[0]);

            spdif_rx_data.adc_mode_request[0].value = 0xFF;
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                &spdif_rx_data.adc_mode_request[1]);

            break;
        }
        
        case SPDIF_RX_ADC_MODE_POWER_OFF_MASTER:
        {
            // Analogue direct
            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX SPDIF_RX_ADC_MODE_POWER_OFF_MASTER ---");
            
            spdif_rx_data.adc_mode_request[0].value = 0xFF;
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                &spdif_rx_data.adc_mode_request[0]);

            spdif_rx_data.adc_mode_request[0].value = 0xFF;
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                &spdif_rx_data.adc_mode_request[1]);
            
            break;
        }
        
        case SPDIF_RX_ADC_MODE_POWER_ON_SLAVE:
        {
            // ADC
            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX SPDIF_RX_ADC_MODE_POWER_ON_SLAVE ---");
           
            spdif_rx_data.adc_mode_request[0].value = 0xFF;
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                &spdif_rx_data.adc_mode_request[0]);

            spdif_rx_data.adc_mode_request[0].value = 0xFB;
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                &spdif_rx_data.adc_mode_request[1]);
            
            break;
        }
        
        case SPDIF_RX_ADC_MODE_POWER_OFF_SLAVE:
        {
            // Avatar
            logger_log_infof(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX SPDIF_RX_ADC_MODE_POWER_OFF_SLAVE ---");
            
            spdif_rx_data.adc_mode_request[0].value = 0xFF;
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                &spdif_rx_data.adc_mode_request[0]);

            spdif_rx_data.adc_mode_request[0].value = 0xFF;
            input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
                &spdif_rx_data.adc_mode_request[1]);

            break;
        }
        
        default:
            break;

    }
}

/**-----------------------------------------------------------------------------
 * Configure the TIPCM9211
 */
void spdif_rx_hw_tipcm9211_configure()
{
    logger_log_info(LOGGER_CODE_INITIALISATION, "INPUT CONTROL SPDIF RX TIPCM9211 Configuration --");
    logger_process();

    input_control_i2c_add_setting(&spdif_rx_data.settings[0], 0X20, 0x05, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[1], 0X24, 0x10, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[2], 0X25, 0x1E, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[3], 0X26, 0x01, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[4], 0X27, 0xDF, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[5], 0X2A, 0xFD, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[6], 0X2B, 0x7F, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[7], 0X2E, 0x44, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[8], 0X35, 0x12, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[9], 0X36, 0x12, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[10], 0X40, 0xE0, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[11], 0X6A, 0x0C, (void*)spdif_rx_hw_tipcm9211_setting_callback);
    input_control_i2c_add_setting(&spdif_rx_data.settings[12], 0X78, 0x00, (void*)spdif_rx_hw_tipcm9211_setting_callback);


    input_control_i2c_add_setting(&spdif_rx_data.error_request[0], 0X2C, 0x00, (void*)spdif_rx_hw_tipcm9211_error_callback);
    input_control_i2c_add_setting(&spdif_rx_data.error_request[1], 0X2D, 0x00, (void*)spdif_rx_hw_tipcm9211_error_callback);
    
    input_control_i2c_add_setting(&spdif_rx_data.switch_request, 0X34, 0x00, (void*)spdif_rx_hw_tipcm9211_switch_callback);
    
    input_control_i2c_add_setting(&spdif_rx_data.adc_mode_request[0], 0X6D, 0x00, (void*)spdif_rx_hw_tipcm9211_adc_mode_callback);
    input_control_i2c_add_setting(&spdif_rx_data.adc_mode_request[1], 0X6E, 0x00, (void*)spdif_rx_hw_tipcm9211_adc_mode_callback);

    input_control_i2c_add_setting(&spdif_rx_data.sample_rate_request, 0X39, 0x00, (void*)spdif_rx_hw_tipcm9211_sample_rate_callback);

//    input_control_i2c_define_register(&spdif_rx_data.error_request[0], (REGISTER_BITS_DEFINITION){ 0x2C, 0, 7 }); // INT0 Error register
//    input_control_i2c_define_register(&spdif_rx_data.error_request[1], (REGISTER_BITS_DEFINITION){ 0x2D, 0, 7 }); // INT1 Error register
//    input_control_i2c_define_register(&spdif_rx_data.switch_request, (REGISTER_BITS_DEFINITION){ 0x34, 0, 7 }); // Switch request    
//    input_control_i2c_define_register(&spdif_rx_data.adc_mode_request[0], (REGISTER_BITS_DEFINITION){ 0x6D, 0, 7 }); // ADC Mode request 
//    input_control_i2c_define_register(&spdif_rx_data.adc_mode_request[1], (REGISTER_BITS_DEFINITION){ 0x6E, 0, 7 }); // ADC Mode request 
//    input_control_i2c_define_register(&spdif_rx_data.sample_rate_request[1], (REGISTER_BITS_DEFINITION){ 0x39, 0, 7 }); // Sample rate
//    spdif_rx_data.error_request.callback = spdif_rx_hw_tipcm9211_error_callback;
//    spdif_rx_data.switch_request.callback = spdif_rx_hw_tipcm9211_switch_callback;
//    spdif_rx_data.sample_rate_request.callback = spdif_rx_hw_tipcm9211_sample_rate_callback;    
    
    spdif_rx_hw_tipcm9211_apply_settings();    

    logger_log_info(LOGGER_CODE_INITIALISATION, "INPUT CONTROL SPDIF RX TIPCM9211 Configuration B --");
    logger_process();
}

/**-----------------------------------------------------------------------------
 * Defines the SPDIF_RX callbacks used when using the TIPCM9211.  To be used as
 * the parameter to spdif_rx_define_callbacks() function
 * @return 
 */    
SPDIF_RX_FUNCTION_DATA* spdif_rx_hw_tipcm9211_define_functions()
{
    // Set up the functions here
    spdif_rx_data.functions.adc_mode_function = spdif_rx_hw_tipcm9211_adc_mode;
    spdif_rx_data.functions.configure_function = spdif_rx_hw_tipcm9211_configure;
    spdif_rx_data.functions.initialise_function = spdif_rx_hw_tipcm9211_initialise;
    spdif_rx_data.functions.interrupt_function = spdif_rx_hw_tipcm9211_interrupt;
    spdif_rx_data.functions.process_function = spdif_rx_hw_tipcm9211_process;
    spdif_rx_data.functions.reset_function = spdif_rx_hw_tipcm9211_reset;
    spdif_rx_data.functions.sample_rate_function = spdif_rx_hw_tipcm9211_sample_rate;
    spdif_rx_data.functions.switch_function = spdif_rx_hw_tipcm9211_switch;
            
    return &spdif_rx_data.functions;
}
 
/**-----------------------------------------------------------------------------
 * Initialise the TIPCM9211
 */
bool spdif_rx_hw_tipcm9211_initialise()
{
    bool result = true;

	// Register the events in the EVENTS structure, 
	events_register(EVENT_MODULE_SPDIF_RX_HW, spdif_rx_data.event_subscriptions, SPDIF_RX_TIPCM9211_HW_MAX_EVENT_SUBSCRIPIONS);
    
    i2c_initialise();
     
    spdif_rx_data.i2c_client_id = i2c_open(SPDIF_RX_TIPCM9211_I2C_MODULE, 
            SPDIF_RX_TIPCM9211_I2C_READ_ADDRESS,
            SPDIF_RX_TIPCM9211_I2C_WRITE_ADDRESS,
            input_control_i2c_callback);
    
    if (I2C_INVALID_CLIENT_ID != spdif_rx_data.i2c_client_id)
    {
        logger_log_info(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX TIPCM9211 I2C Open");
    }
    else
    {
        logger_log_error(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX TIPCM9211 I2C Open Error");
        result = false;
    }    
    
    return result;
}
     
/**-----------------------------------------------------------------------------
 * Handles interrupt on the SPDIF RX Interrupt pins
 * @param definition
 */
void spdif_rx_hw_tipcm9211_interrupt(uint8_t channel)
{
    if (!spdif_rx_data.in_interrupt)
    {
        spdif_rx_data.in_interrupt = true;

        bool error_check_reqd = true;

        if (channel == SPDIF_RX_TIPCM9211_INT0_PORT_CHANNEL)
        {
            //logger_log_debugf(0, "INPUT CONTROL INTERRUPT --");
            // Read the port to clear the interrupt
            uint32_t value = PLIB_PORTS_Read(SPDIF_RX_TIPCM9211_INT0_ID,
                                             SPDIF_RX_TIPCM9211_INT0_PORT_CHANNEL);

            // Check it was the sample rate pin that triggered the interrupt
            bool sample_rate_change = (value & (1 << SPDIF_RX_TIPCM9211_INT0_PIN)) != 0;
            if (sample_rate_change)
            {
                SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
                event_data.error_identified = true;
                event_data.sample_rate_change = true;
                event_data.pll_locked = true;
        
                // Raise the event
                events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);

                spdif_rx_refresh_sample_rate();

                // Note the sample rate calculation will trigger a check for errors
                error_check_reqd = false;
            }

            bool error_change = (value & (1 << SPDIF_RX_TIPCM9211_INT1_PIN)) != 0;
            if (error_change)
            {
                SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
                event_data.error_identified = true;
                event_data.sample_rate_change = false;
                event_data.pll_locked = true;
        
                // Raise the event
                events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);
            }
            
            // Disable Interrupts and move to polling as this goes into a loop
            //PLIB_PORTS_ChannelChangeNoticeDisable(PORTS_ID_0, PORT_CHANNEL_F, SYS_PORT_F_CNEN || 0x03);

        }
        else if (channel == SPDIF_RX_TIPCM9211_INT1_PORT_CHANNEL)
        {
            //logger_log_debugf(0, "INPUT CONTROL INTERRUPT ERROR --");
            // Read the other interrupt port to clear it.
            // Note this would clear the interrupts for any other peripherals 
            // monitoring on this port.
            bool error_detected = PLIB_PORTS_PinGet(SPDIF_RX_TIPCM9211_INT1_ID, 
                                                    SPDIF_RX_TIPCM9211_INT1_PORT_CHANNEL, 
                                                    SPDIF_RX_TIPCM9211_INT1_PIN);

            bool error_change = (error_detected & (1 << SPDIF_RX_TIPCM9211_INT1_PIN)) != 0;
            if (error_change)
            {
                SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
                event_data.error_identified = true;
                event_data.sample_rate_change = false;
                event_data.pll_locked = true;
        
                // Raise the event
                events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);
            }
        }

        spdif_rx_data.in_interrupt = false;
    }
}

/**-----------------------------------------------------------------------------
 * Perform TIPCM9211 processing
 */
void spdif_rx_hw_tipcm9211_process()
{
    spdif_rx_hw_tipcm9211_get_errors();   
}

/**-----------------------------------------------------------------------------
 * Perform TIPCM9211 reset
 */
void spdif_rx_hw_tipcm9211_reset(bool on)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Reset - Switching %s ", on == true ? "On" : "Off");
    
    PLIB_PORTS_PinWrite(SPDIF_RX_TIPCM9211_RESET_PORTS_ID, 
                        SPDIF_RX_TIPCM9211_RESET_PORT_CHANNEL, 
                        SPDIF_RX_TIPCM9211_RESET_PIN,
                        on); // Note pin is low

    spdif_rx_data.in_reset = !on;
    
    if (on)
    {
        // Setup for a resend of the settings
        spdif_rx_data.settings_applied = false;
    }
}

/**-----------------------------------------------------------------------------
 * Trigger TIPCM9211 sample rate calculation 
 */
bool spdif_rx_hw_tipcm9211_sample_rate()
{
    bool result = false;
    
    if (!spdif_rx_data_get_calculating_sample_rate())
    {
        logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Sample Rate --");

        // Shouldn't need to check whether this is already running as that is done in input_control_i2c_read_setting
        if (!spdif_rx_hw_tipcm9211_is_reset())
        {
            // Set before running in-case the sample rate calculation is completd
            // before the end of the function call.
            spdif_rx_data_set_calculating_sample_rate(true);
            
            input_control_i2c_read_setting(spdif_rx_data.i2c_client_id,
                    &spdif_rx_data.sample_rate_request);

            result = true;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Perform TIPCM9211 switch 
 */
void spdif_rx_hw_tipcm9211_switch(INPUT_CONTROL_DEFINITION* definition)
{
    logger_log_debugf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Switch --");

    // Definitions are currently set up in app_commands_input/app_commands_input_configure
    //TODO these are still set for the AK4113 - need a way of being able to define them for th TI
    uint8_t new_value = definition->value;

    // Check if we have had an invalid signal and if so mute outputs before switching
    // to stop noise..
    if (spdif_rx_data.invalid_signal)
    {
        logger_log_errorf(LOGGER_CODE_INPUT_CONTROL, "INPUT CONTROL SPDIF RX Switch Digital errors MUTING OUTPUTS ");

        SPDIF_RX_HW_ERROR_EVENT_DATA event_data;
        event_data.error_identified = true;
        event_data.sample_rate_change = false;
        event_data.pll_locked = true;

        // Raise the event
        events_raise(EVENT_MODULE_SPDIF_RX_HW, SPDIF_RX_EVENT_ERROR, (void*)&event_data);
    }

    // Reapply the settings (just in case);
//    spdif_rx_hw_tipcm9211_apply_settings();
    
    // Removed as was not catching analogue to digital switches
    // Note: address should have been defined
    spdif_rx_data.switch_request.actioned = false;
    spdif_rx_data.switch_request.value = new_value;
    
    input_control_i2c_apply_setting(spdif_rx_data.i2c_client_id,
            &spdif_rx_data.switch_request);
}

/* *****************************************************************************
 End of File
 */
