/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    standby.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/standby.h"
#include <stdint.h>
#include "app_config.h"
#include "utilities/logger/logger.h"
#include "utilities/watchdog.h"

#if ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MZ

#include <xc.h>
#include "system_config.h"
#include "system_definitions.h"

#endif

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    // True if the standby can be in idle state
    bool can_idle;
    
    // True if the standby can be in sleep stare
    bool can_sleep;
    
    STANDBY_STATE state;
   
    uint16_t active_sleep_counter;    
} STANDBY_DATA;

static STANDBY_DATA standby_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Original for PICMX in..
 * D:\microchip\xc32\v2.10\pic32-libs\peripheral\power\source\power_sleep_lib.c
 * 
 * See
 * https://www.microchip.com/forums/m991184.aspx
 * 
 * #ifdef _DMAC
 * #define	mSysUnlockOpLock(op)	do{int intStat, dmaSusp; intStat=INTDisableInterrupts(); dmaSusp=DmaSuspend(); \
 * 				SYSKEY = 0, SYSKEY = 0xAA996655, SYSKEY = 0x556699AA; \
 * 				(op); \
 * 				SYSKEY = 0x33333333; \
 * 				DmaResume(dmaSusp); INTRestoreInterrupts(intStat);}while(0)
 * #else
 * #define	mSysUnlockOpLock(op)	do{int intStat; intStat=INTDisableInterrupts(); \
 * 				SYSKEY = 0, SYSKEY = 0xAA996655, SYSKEY = 0x556699AA; \
 * 				(op); \
 * 				SYSKEY = 0x33333333; \
 * 				INTRestoreInterrupts(intStat);}while(0)
 * #endif	// _DMA
 */
inline void __attribute__((nomips16)) standby_idle(void)
{
#if ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MX    
    
    PowerSaveIdle();
    
#elif ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MZ    
    
    SYS_DEVCON_SystemUnlock ( );
    PLIB_OSC_OnWaitActionSet(OSC_ID_0, OSC_ON_WAIT_IDLE);
    SYS_DEVCON_SystemLock ( );
    SYS_INT_PROCESSOR_STATUS status = SYS_INT_StatusGetAndDisable();
    asm("WAIT");
    SYS_INT_StatusRestore(status);
#else
    logger_log_warningf(LOGGER_CODE_STANDBY, "STANDBY No standby_idle function defined ");
#endif    
}

/**
 * 
 */
inline void __attribute__((nomips16)) standby_sleep(void)
{
#if ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MX    
    
    PowerSaveSleep();
    
#elif ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MZ    
    
    SYS_DEVCON_SystemUnlock ( );
    PLIB_OSC_OnWaitActionSet(OSC_ID_0, OSC_ON_WAIT_SLEEP);
    SYS_DEVCON_SystemLock ( );
    SYS_INT_PROCESSOR_STATUS status = SYS_INT_StatusGetAndDisable();
    asm("WAIT");
    SYS_INT_StatusRestore(status);
 
    
#else
    logger_log_warningf(LOGGER_CODE_STANDBY, "STANDBY No standby_sleep function defined ");
#endif    
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Gets the current standby state
 * @return 
 */
STANDBY_STATE standby_get_state()
{
    return standby_data.state;
}

/**
 * Use when leaving standby mode to reset standby state to STANDBY_STATE_ENTRY
 */
void standby_reset()
{
    logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY Resetting ENTRY State");
                
    standby_data.state = STANDBY_STATE_ENTRY;
}

/**
 * Use with STANDBY_STATE_ENTRY to determine which default standby state to move to
 * @param can_sleep -  generally supplied by !power_network_standby()
 */
void standby_set_initial_state(bool can_idle, bool can_sleep, STANDBY_STATE default_state)
{
    standby_data.can_idle = can_idle;
    standby_data.can_sleep = can_sleep;
    
    // Default to Sleep state
    standby_data.state = default_state;

    if (!can_sleep && !can_idle)
    {
        // Put in an active awake state
        standby_data.state = STANDBY_STATE_ACTIVE;
        logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY Setting State ACTIVE");
    }   
    else
    {
        logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY Setting State SLEEP");
        
    }
}

/**
 * Directly sets the Standby state
 * @return 
 */
void standby_set_state(STANDBY_STATE state)
{
    standby_data.state = state;
}

/**
 * Set the standby state into an active high power state
 * @param state
 */
void standby_set_state_active()
{
    standby_data.state = STANDBY_STATE_ACTIVE;
    standby_data.active_sleep_counter = 0;
                        
    logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY Exiting SLEEP State %d --", standby_data.state );
}

/**
 * Puts the unit into sleep mode if set in the appropriate standby mode
 */
void standby_try_sleep()
{
    // Clear the watch dog
    watchdog_reset();
 
    if (standby_data.state == STANDBY_STATE_SLEEP)
    {
        if (standby_data.can_sleep)
        {
            logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY Entering SLEEP State --");
            standby_sleep();
        }
        else
        {
            logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY Entering IDLE State --");
            standby_idle();
        }
    }
}

/**
 * Checks to see whether the state can be put back into sleep state
 */
void standby_update_state(bool allow_idle, bool allow_sleep)
{
    standby_data.can_idle = allow_idle;
    standby_data.can_sleep = allow_sleep;

    if (standby_data.state != STANDBY_STATE_ENTER_SLEEP)
    {
        if (allow_idle || allow_sleep)
        {
            standby_data.active_sleep_counter++;
            //logger_log_debugf(230, "POWERDOWN: sleep_counter %d %d --", app_data.powereddown_active_sleep_counter, app_data.powereddown_state);

            // If we're in network standby mode and there has been some activity
            // which has moved us into the active state we'll stay there for 1000 cycles to 
            // make sure we're processing messages before giving up and
            // going back to sleep
            if (standby_data.active_sleep_counter > 1000)
            {
                // Put in an active awake state
                standby_data.state = STANDBY_STATE_SLEEP;
            }
        }
    }
}

/* *****************************************************************************
 End of File
 */
