/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    structured_message.c

  @Summary
    common structured message used to handle rs232, IR and Ethernet

  @Description
    Describe the purpose of this file.
 * 
 * v2.1 - Changes to incorporate display_message 
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/structured_message.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Message constants
#define STX                     0x21
#define ET                      0x0D

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Initialises a structured message
 * @param message
 */
void structured_message_initialise(STRUCTURED_MESSAGE* message, STRUCTURED_MESSAGE_SOURCE source)
{
    message->source = source;
    message->action = STRUCTURED_MESSAGE_ACTION_UNDEFINED;
    message->data_only = false;
    message->zone = STRUCTURED_MESSAGE_ZONE_1;
    message->command = STRUCTURED_MESSAGE_COMMAND_UNDEFINED; 
    
    message->response = STATUS_UPDATE;
    message->data_length = 0;
    
    message->processing_state = STRUCTURED_MESSAGE_STATE_INITIALISED;
    message->processing_data_index = 0;
    
    // Set is _activity to true by default
    message->is_activity = true;
    
    //message->data_index = 0;
	message->state_data[0] = 0;

    message->repeat = false;
     
    memset(&message->data[0], 0, STRUCTURED_MESSAGE_MAX_DATA);
};

/**-----------------------------------------------------------------------------
 * Copies the details from one structured message into another
 * @param from - the message to copy
 * @param to - the message whose data will be overwritten by the first message
 */
void structured_message_copy(STRUCTURED_MESSAGE* from, STRUCTURED_MESSAGE* to)
{
    memcpy(to, from, sizeof(STRUCTURED_MESSAGE));
}

/**-----------------------------------------------------------------------------
 * Returns the length of the structured message including headers, commands, footer
 */
int structured_message_get_length(STRUCTURED_MESSAGE* message)
{
    return message->data_length + 6;
}

/**-----------------------------------------------------------------------------
 * Can be used to determine whether the structured message is currently being populated
 * @param message
 * @return 
 */
bool structured_message_is_processing(STRUCTURED_MESSAGE* message)
{
    return message->processing_state != STRUCTURED_MESSAGE_STATE_INITIALISED;
}

/**-----------------------------------------------------------------------------
 * Process a buffer of data - used to construct structured messages from streams of data
 * This takes a structured message a and depending on its current state adds data 
 * into the message.  It is intended for use with controllers that produce streams of data
 * in the standard format for our internal messaging structure.
 * @param message   - The message the data is to be added to 
 * @param buffer - The message that the data is added to 
 * @param size  - The size of the buffer (amount of data to be processed)
 */
void structured_message_process_buffer(STRUCTURED_MESSAGE* message, unsigned char* buffer, size_t size)
{
    int x;
    for (x = 0; x < size; x++)
    {
        structured_message_process_byte(message, buffer[x]);
    }
}

/**-----------------------------------------------------------------------------
 * Process a byte of data read from the RX channel
 * @param data_byte
 */
bool structured_message_process_byte(STRUCTURED_MESSAGE* message, unsigned char data_byte)
{
    bool result = true;
    
    switch (message->processing_state)
    {
        case STRUCTURED_MESSAGE_STATE_INITIALISED:
        {
            // Assumes the message has been pre initialised using
            // structured_message_initialise
            // Start the message off - it expects either STX or an 'A'
            if (data_byte == STX)
            {
                message->processing_state = STRUCTURED_MESSAGE_STATE_WAIT_FOR_ZONE;
            }
            else if (data_byte == 'A')
            {
                message->processing_state = STRUCTURED_MESSAGE_STATE_WAIT_FOR_M;
                message->data_only = true;
            }
            else
            {
                result = false;
            }
            
            break;
        }
        
        case STRUCTURED_MESSAGE_STATE_WAIT_FOR_ZONE:
        {
            message->zone = data_byte;
            
            if ((data_byte != STRUCTURED_MESSAGE_ZONE_1) && (data_byte != STRUCTURED_MESSAGE_ZONE_2))
            {
                message->response = ZONE_INVALID;
            }
            
            message->processing_state = STRUCTURED_MESSAGE_STATE_WAIT_FOR_CMD;
            
            break;
        }
        
        case STRUCTURED_MESSAGE_STATE_WAIT_FOR_CMD:
        {
            // Set the command
            message->command = data_byte;

            message->processing_state = STRUCTURED_MESSAGE_STATE_WAIT_FOR_LENGTH;
            
            break;
        }
        
        case STRUCTURED_MESSAGE_STATE_WAIT_FOR_LENGTH:
        {
            message->data_length = data_byte;
            
            if (data_byte >= STRUCTURED_MESSAGE_MAX_DATA)
            {
                message->response = COMMAND_NOT_RECOGNISED;
            }

            message->processing_state = STRUCTURED_MESSAGE_STATE_WAIT_DATA;
            
            break;
        }
        
        case STRUCTURED_MESSAGE_STATE_WAIT_DATA:
        {
            // Check whether we have parsed all the data yet
            if (message->processing_data_index < message->data_length)
            {
                message->data[message->processing_data_index] = data_byte;
                message->processing_data_index++;
                if (message->processing_data_index >= message->data_length)
                {
                    message->processing_state = STRUCTURED_MESSAGE_STATE_WAIT_FOR_END;
                }
            }
            else if (message->data_length == 0) // e.g. no data
            {
                // There is no associated data with the command, so we expect an end message
                if (data_byte == ET)
                {
                    message->processing_state = STRUCTURED_MESSAGE_STATE_PERFORM_ACTION;
                }
                else
                {
                    message->response = INVALID_DATA_LENGTH;
                    message->processing_state = STRUCTURED_MESSAGE_STATE_SEND_REPLY;
                }
            }
            else
            {
                message->response = INVALID_DATA_LENGTH;
                message->processing_state = STRUCTURED_MESSAGE_STATE_SEND_REPLY;
            }
            
            break;
        }
        
        case STRUCTURED_MESSAGE_STATE_WAIT_FOR_END:
        {
            if (data_byte == ET)
            {
                message->processing_state = STRUCTURED_MESSAGE_STATE_PERFORM_ACTION;
            }
            else
            {
                message->response = INVALID_DATA_LENGTH;
                message->processing_state = STRUCTURED_MESSAGE_STATE_SEND_REPLY;
            }
            
            break;
        }
             
        case STRUCTURED_MESSAGE_STATE_WAIT_FOR_M:
        {
            if (data_byte == 'M')
            {
                message->processing_state = STRUCTURED_MESSAGE_STATE_WAIT_FOR_X;
            }
            else
            {
                message->response = COMMAND_NOT_RECOGNISED;
                message->processing_state = STRUCTURED_MESSAGE_STATE_SEND_REPLY;
            }
            
            break;
        }
        
        case STRUCTURED_MESSAGE_STATE_WAIT_FOR_X:
        {
            if (data_byte == 'X')
            {
                message->processing_state = STRUCTURED_MESSAGE_STATE_WAIT_FOR_CR;
            }
            else
            {
                message->response = COMMAND_NOT_RECOGNISED;
                message->processing_state = STRUCTURED_MESSAGE_STATE_SEND_REPLY;
            }
            
            
            break;
        }
        
        case STRUCTURED_MESSAGE_STATE_WAIT_FOR_CR:
        {
            if (data_byte == '\r')
            {
                message->command = STRUCTURED_MESSAGE_COMMAND_AMX_DUET;
                message->processing_state = STRUCTURED_MESSAGE_STATE_PERFORM_ACTION;
            }
            else
            {
                message->data[0] = data_byte;
                message->response = COMMAND_NOT_RECOGNISED;
                message->processing_state = STRUCTURED_MESSAGE_STATE_SEND_REPLY;
            }
            
            break;
        }
        
        default:
            break;
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Set the data on a structured message (saves memcpys etc...) 
 */
void structured_message_set_message_data(STRUCTURED_MESSAGE* message, char* data, size_t data_length)
{
    char* data_ptr = (char*)&message->data;
    memset(data_ptr, 0, STRUCTURED_MESSAGE_MAX_DATA);
    
    if (data_length > 0)
    {
        message->data_length = data_length;
        memcpy(data_ptr, data, min(data_length, STRUCTURED_MESSAGE_MAX_DATA));
    }
}

/**-----------------------------------------------------------------------------
 * Fills the supplied buffer with a structured message ready for transmission 
 * (including STX and ET start/stop bytes).
 * Note: The buffer MUST be sufficiently large enough to hold the length of the data + 6
 * @param message - the message to be written
 * @param buffer - the buffer to be written to
 * @param size - the actual size of the buffer
 * @return - the number of bytes written to the buffer
 */
int structured_message_write_response_buffer(STRUCTURED_MESSAGE* message, unsigned char* buffer, size_t size)
{
    int result = 0;
    
    if (!message->data_only)
    {
        if (size >= message->data_length + 6)
        {
            buffer[0] = STX;
            buffer[1] = message->zone;
            buffer[2] = message->command;
            buffer[3] = message->response;
            buffer[4] = message->data_length;
            memcpy(&buffer[5], &message->data[0], message->data_length);
            buffer[5 + message->data_length] = ET;

            result = message->data_length + 6;
        }
    }
    else
    {
            memcpy(buffer, message->data, message->data_length);
            buffer[message->data_length] = 0;
            
            result = message->data_length;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Fills the supplied buffer with a structured message ready for transmission 
 * (including STX and ET start/stop bytes).
 * Note: The buffer MUST be sufficiently large enough to hold the length of the data + 6
 * @param message - the message to be written
 * @param buffer - the buffer to be written to
 * @param size - the actual size of the buffer
 * @return - the number of bytes written to the buffer
 */
int structured_message_write_request_buffer(STRUCTURED_MESSAGE* message, unsigned char* buffer, size_t size)
{
    int result = 0;
    
    if (size >= message->data_length + 6)
    {
        buffer[0] = STX;
        buffer[1] = message->zone;
        buffer[2] = message->command;
        buffer[3] = message->data_length;
        memcpy(&buffer[4], &message->data[0], message->data_length);
        buffer[4 + message->data_length] = ET;

        result = message->data_length + 5;
    }
    
    return result;
}

/* *****************************************************************************
 End of File
 */
