/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    system_fault.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/system_fault.h"
#include "utilities/logger/logger.h"
#include "utilities/settings.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! Determines whether to save the fault in settings
    bool fault_settings_save;
    
    //! Used to store the fault state causing a power down
    SYSTEM_FAULT system_fault;

    //! Stores the last shutdown fault code
    SYSTEM_FAULT_CODE last_shutdown_code;

} SYSTEM_FAULT_DATA;

static SYSTEM_FAULT_DATA system_fault_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the current System Fault 
 */
SYSTEM_FAULT system_fault_get()
{
    return system_fault_data.system_fault;
}

/**-----------------------------------------------------------------------------
 * Gets the current System Fault 
 */
SYSTEM_FAULT_CODE system_fault_get_last_shutdown_code()
{
    return system_fault_data.last_shutdown_code;
}
    
/**-----------------------------------------------------------------------------
 * Initialise the system fault context
 */
void system_fault_initialise()
{
    system_fault_data.system_fault.code = SYSTEM_FAULT_CODE_NO_FAULT;
    memset(system_fault_data.system_fault.description, 0, SYSTEM_FAULT_MAX_DESCRIPTION_LEN);
    system_fault_data.fault_settings_save = false;
    
    // Log out current history
    SYSTEM_FAULT_CODE history[SHUTDOWN_CODE_HISTORY_LENGTH];
    settings_get(SETTING_SHUTDOWN_CODE_HISTORY, &history);

    int x;
    for (x = 0; x < SHUTDOWN_CODE_HISTORY_LENGTH; x++)
    {
        if ((history[x] >= SYSTEM_FAULT_CODE_NO_FAULT) && (history[x] < SYSTEM_FAULT_CODE_MAX))
        {
            logger_log_infof(LOGGER_CODE_LAST_SHUTDOWN_CODE, "LAST SHUTDOWN CODE HISTORY: %d %d %s --", x, history[x], system_fault_codes[history[x]]);
        }
    }

    // Copy current shutdown code into history
    memcpy(&history[1], &history[0], (SHUTDOWN_CODE_HISTORY_LENGTH - 1) * sizeof(SYSTEM_FAULT_CODE));
    history[0] = system_fault_get_last_shutdown_code();

    // By default set the system shutdown code - this will be updated later if there are 
    // any faults.  We need to set it here as the user may just power the unit down..
    SYSTEM_FAULT_CODE code = SYSTEM_FAULT_CODE_NO_FAULT;
    settings_suppress_save(true);
    settings_set(SETTING_SHUTDOWN_CODE, &code);
    settings_set(SETTING_SHUTDOWN_CODE_HISTORY, &history[0]);
    settings_suppress_save(false);
}
    
/**-----------------------------------------------------------------------------
 * Saves the current system fault into settings.  Returns true once the settings have 
 * been successfully saved
 */
bool system_fault_save_to_settings()
{
    bool result = false;
    SYSTEM_FAULT fault = system_fault_get();
    
    if (!system_fault_data.fault_settings_save)
    {
        // Log the fault
        logger_log_errorf(LOGGER_CODE_SYSTEM_FAULT, "SYSTEM FAULT: %d %s --", fault.code, fault.description);
        logger_process();

        SYSTEM_FAULT_CODE history[SHUTDOWN_CODE_HISTORY_LENGTH];
        settings_get(SETTING_SHUTDOWN_CODE_HISTORY, &history);
        memcpy(&history[1], &history[0], (SHUTDOWN_CODE_HISTORY_LENGTH - 1) * sizeof(SYSTEM_FAULT_CODE));
        history[0] = fault.code;
        settings_set(SETTING_SHUTDOWN_CODE_HISTORY, &history[0]);

        // Store shutdown reason in settings and force a save - note this is also stored 
        // when testing via RS232 commands
        settings_set(SETTING_SHUTDOWN_CODE, &fault.code);
        settings_save();

        system_fault_data.fault_settings_save = true;
            
        system_fault_set_last_shutdown_code(fault.code);
    }
    else
    {
        // If fault code is saved them move on
        if (settings_available())
        {
            system_fault_data.fault_settings_save = false;

            SYSTEM_FAULT_CODE code;
            settings_get(SETTING_SHUTDOWN_CODE, &code);
            logger_log_errorf(LOGGER_CODE_SYSTEM_FAULT, "SYSTEM FAULT: %d --", code);

            result = true;
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Sets the current system fault
 */
void system_fault_set(SYSTEM_FAULT* fault)
{
        // Store the fault
    memcpy(&system_fault_data.system_fault, fault, sizeof(SYSTEM_FAULT));
}

/**-----------------------------------------------------------------------------
 * Sets the current System Fault 
 */
void system_fault_set_last_shutdown_code(SYSTEM_FAULT_CODE code)
{
    system_fault_data.last_shutdown_code = code;
}
   
/* *****************************************************************************
 End of File
 */
