/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    tcpip.c

  @Summary
 TCP/IP management

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/tcpip.h"
#include "utilities/avatar_control.h"
#include "utilities/logger/logger.h"
#include "app_config.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Maximum number of client connections note to work with the UDP stack this number 
//  needs to be a maximum of the Harmony TCP config "Maximum Number Of Sockets Supported" - 3
#ifndef TCPIP_MAX_CLIENTS
#ifdef MAX_BSD_SOCKETS
#define TCPIP_MAX_CLIENTS       MAX_BSD_SOCKETS - 3    
#else
#define TCPIP_MAX_CLIENTS       4   
#endif
#endif

//! Indicated a disconnected connection.
#define TCP_CLIENT_ID_INVALID   255 // Assuming we wont have 255 connections!

typedef struct
{
    //! The client_id used for the connection TCP_CLIENT_ID_INVALID if not in use.
    uint8_t client_id;
    
     // Holds the current message read from the client socket
    STRUCTURED_MESSAGE  current_message;
    
    // The ordinal position of the connectioin in the list
    uint8_t index;
   
} TCPIP_CLIENT_CONNECTION;

/**
 * Global TCP/IP data
 */
typedef struct
{
    // Client sockets
    TCPIP_CLIENT_CONNECTION client_connections[TCPIP_MAX_CLIENTS];
    
    // The MAC address of the ethernet card
    TCPIP_MAC_ADDR mac_address;
    
    // The MAC address of the network card
    TCPIP_MAC_ADDR wifi_mac_address;
            
    // A friendly name for the unit on the network
    char network_name[TCPIP_NETWORK_NAME_MAX_LEN];

    // The IP Address currently used.
    IPV4_ADDR runtime_address;
            
    // A friendly name for the unit on the network
    char ssid[TCPIP_SSID_MAX_LEN];
        
    // Callback function supplied to tcpip_initialise - called when a message is 
    // received
    STRUCTURED_MESSAGE_CALLBACK callback;

    //! Callback invoked when a new client connection is established
    TCP_CONNECTED_CALLBACK connected_callback;
    
    //! Indicates whether the module has been initialised as this sets TCP_CLIENT_ID_INVALID 
    bool initialised;

} TCPIP_DATA;

static TCPIP_DATA tcpip_data = 
{
    .callback = NULL,
    .connected_callback = NULL,
    .initialised = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

// Forward declarations
static TCPIP_CLIENT_CONNECTION* tcpip_get_client_connection(uint8_t client_id);
static void tcpip_initialise_structured_message(STRUCTURED_MESSAGE* message);
static void tcpip_receive(int client_idx);

/**-----------------------------------------------------------------------------
 * Sets the client as connected or not
 * Note: Client Id equates to the ordinal position in the client_connections array
 * Assumes that client_id starts from 0
 * @param client_id
 * @param connected
 */
void tcpip_avatar_set_client_connected(uint8_t client_id, bool connected)
{
    // Get a connection (note this will be allocated if not already connected)
    TCPIP_CLIENT_CONNECTION* client = tcpip_get_client_connection(client_id);
    //logger_log_debugf(0, "TCPIP tcpip_avatar_set_client_connected %d %s --", client_id, client != NULL ? "FOUND" : "NOT FOUND");
    
    if ((client != NULL) && connected)
    {
        if (tcpip_data.connected_callback != NULL)
        {
            tcpip_data.connected_callback(true, NULL, NULL);
        }
    }
    else
    {
        if (tcpip_data.connected_callback != NULL)
        {
            tcpip_data.connected_callback(false, NULL, NULL);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Sets the IP Address
 */
void tcpip_avatar_set_ip_address(uint32_t* ip_address)
{
    tcpip_data.runtime_address.v[0] = (*ip_address >> (3 * 8)) & 0xFF;
    tcpip_data.runtime_address.v[1] = (*ip_address >> (2 * 8)) & 0xFF;
    tcpip_data.runtime_address.v[2] = (*ip_address >> (1 * 8)) & 0xFF;
    tcpip_data.runtime_address.v[3] = (*ip_address >> (0 * 8)) & 0xFF;
    //memcpy(&tcpip_data.runtime_address, ip_address, sizeof(uint32_t));
}

/**-----------------------------------------------------------------------------
 * Sets the MAC Address
 * Expects a uint8_t[6]
 */
void tcpip_avatar_set_mac_address(uint8_t* mac_address)
{
    logger_log_debugf(0, "MAC ADDRESS C %d %d %d %d %d %d --", mac_address[0], mac_address[1], mac_address[2], mac_address[3], mac_address[4], mac_address[5]);
    memcpy(&tcpip_data.mac_address, mac_address, 6);
}

/**-----------------------------------------------------------------------------
 * Sets the MAC Address
 * Expects a uint8_t[6]
 */
void tcpip_avatar_set_wifi_mac_address(uint8_t* mac_address)
{
    memcpy(&tcpip_data.wifi_mac_address, mac_address, 6);
}

/**-----------------------------------------------------------------------------
 * Callback for Avatar TCP messages
 * @param client_idx - The TCP client identifier
 * @param data - A pointer to the TCP data
 * @param offset - For multi-part messages this is the offset of this data in the whole message
 * @param data_len - The length of the data pointed at by the data pointer.
 */
void tcpip_avatar_tcp_callback(uint8_t client_id, char* data, uint16_t offset, size_t data_len)
{
    TCPIP_CLIENT_CONNECTION* client = tcpip_get_client_connection(client_id);
    
    if (client != NULL)
    {
        logger_log_debugf(0, "TCPIP  FOUND CLIENT %d --", client_id);
        
        if (client->client_id != TCP_CLIENT_ID_INVALID)
        {
            if(data_len > 0)
            {
                structured_message_process_buffer(&client->current_message, data, data_len);
            }

            // If we've received a message  then lets process it
            if (client->current_message.processing_state == STRUCTURED_MESSAGE_STATE_PERFORM_ACTION)
            {

                // Pass the message onto a callback method for validation and processing
                if (tcpip_data.callback != NULL)
                {
                    client->current_message.state_data[0] = client_id;

                    logger_log_debugf(0, "TCPIP Received Message");
                    tcpip_data.callback(&client->current_message);
                }

                // Re initialise the message
                tcpip_initialise_structured_message(&client->current_message);
            }

            // If we need to send a reply then lets get that started
            if (client->current_message.processing_state == STRUCTURED_MESSAGE_STATE_SEND_REPLY)
            {
                client->current_message.state_data[0] = client->index;

                tcpip_send_structured_message(&client->current_message);
            }
        }
    }
}
 
/**-----------------------------------------------------------------------------
 * Finds the client connection representing the client id supplied
 * @param client_id
 * @return 
 */
static TCPIP_CLIENT_CONNECTION* tcpip_get_client_connection(uint8_t client_id)
{
    TCPIP_CLIENT_CONNECTION* result = NULL;
    
    int x;
    for (x = 0; x < TCPIP_MAX_CLIENTS; x++)
    {
        if (tcpip_data.client_connections[x].client_id == client_id)
        {
            result = &tcpip_data.client_connections[x];
            break;
        }
        else if ((tcpip_data.client_connections[x].client_id == TCP_CLIENT_ID_INVALID) && (result == NULL))
        {
            // Allocate the first free connection just incase we don not find one connected
            result = &tcpip_data.client_connections[x];
        }
    }
    
    // Makes sure the result is now liked to the client id
    if (result != NULL)
    {
        result->client_id = client_id;
    }
    
    return result;
}
   
/**-----------------------------------------------------------------------------
 * Initialises a structured message
 * @param message
 */
static void tcpip_initialise_structured_message(STRUCTURED_MESSAGE* message)
{
    structured_message_initialise(message, STRUCTURED_MESSAGE_SOURCE_TCPIP);
    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**----------------------------------------------------------------------------- 
 * Close all open TCPIP connections
 */
void tcpip_close()
{
    int x;
    for(x = 0; x < TCPIP_MAX_CLIENTS; x++)
    {
        if (tcpip_data.client_connections[x].client_id != TCP_CLIENT_ID_INVALID)
        {
            logger_log_debugf(LOGGER_CODE_TCPIP, "TCPIP Closing client socket %d --", x);
            avatar_tcp_client_control(x, AVATAR_TCP_CLIENT_CONTROL_STATE_DISCONNECT);
		}
    }
}

/**----------------------------------------------------------------------------- 
 * Performs steps to connect TCP/IP to a network
 */
void tcpip_connect()
{
    logger_log_warningf(LOGGER_CODE_TCPIP, "TCPIP tcpip_connect called but not implemented ");
}
    
/**-----------------------------------------------------------------------------
 * Returns whether DHCP is enabled or not
 * @param enable
 */
bool tcpip_get_dhcp_enabled()
{
    logger_log_warningf(LOGGER_CODE_TCPIP, "TCPIP tcpip_get_dhcp_enabled called but not implemented ");
    return true;
}
    
/**-----------------------------------------------------------------------------
 * Gets a displayable version if the IP Address
 * @param buffer
 */
void tcpip_get_ip_address_string(char* buffer, size_t size)
{
    char ip_addr[16];
    sprintf(ip_addr, "%d.%d.%d.%d",
            tcpip_data.runtime_address.v[0],
            tcpip_data.runtime_address.v[1],
            tcpip_data.runtime_address.v[2],
            tcpip_data.runtime_address.v[3]);
    
    memset(buffer, 0, size);
    memcpy(buffer, ip_addr, min(16, size));
 }
     
/**-----------------------------------------------------------------------------
 * Returns a the current MAC address
 * @return 
 */
TCPIP_MAC_ADDR tcpip_get_mac_address()
{
    return tcpip_data.mac_address;
}
    
/**-----------------------------------------------------------------------------
 * Returns a string representation of the current MAC address
 * @return 
 */
void tcpip_get_mac_address_string(char* buffer, size_t size)
{
    char mac_addr[18];
    sprintf(mac_addr, "%02X%02X%02X%02X%02X%02X",
            tcpip_data.mac_address.v[0],
            tcpip_data.mac_address.v[1],
            tcpip_data.mac_address.v[2],
            tcpip_data.mac_address.v[3],
            tcpip_data.mac_address.v[4],
            tcpip_data.mac_address.v[5]);
    
    memset(buffer, 0, size);
    memcpy(buffer, mac_addr, min(17, size));
 }
    
/**-----------------------------------------------------------------------------
 * Returns a string representation of the current MAC address
 * @return 
 */
void tcpip_get_network_name_string(char* buffer, size_t size)
{
    snprintf(buffer, min(size, TCPIP_NETWORK_NAME_MAX_LEN), "%s", tcpip_data.network_name);
}

/**-----------------------------------------------------------------------------
 * Gets the static IP Address
 */
IP_ADDR tcpip_get_runtime_ip_address()
{
    return tcpip_data.runtime_address;
}
   
/**-----------------------------------------------------------------------------
 * Returns a string representation of the current MAC address
 * @return 
 */
void tcpip_get_ssid_string(char* buffer, size_t size)
{
    snprintf(buffer, min(size, TCPIP_SSID_MAX_LEN), "%s", tcpip_data.ssid);
}

/**-----------------------------------------------------------------------------
 * Gets the static IP Address - actually returns the runtime address obtained from the Avatar
 */
IP_ADDR tcpip_get_static_ip_address()
{
    return tcpip_data.runtime_address;
}
    
/**-----------------------------------------------------------------------------
 * Returns a the current MAC address
 * @return 
 */
TCPIP_MAC_ADDR tcpip_get_wifi_mac_address()
{
    return tcpip_data.wifi_mac_address;
}
    
/**-----------------------------------------------------------------------------
 * Returns a string representation of the current MAC address
 * @return 
 */
void tcpip_get_wifi_mac_address_string(char* buffer, size_t size)
{
    char mac_addr[18];
    sprintf(mac_addr, "%02X:%02X:%02X:%02X:%02X:%02X",
            tcpip_data.wifi_mac_address.v[0],
            tcpip_data.wifi_mac_address.v[1],
            tcpip_data.wifi_mac_address.v[2],
            tcpip_data.wifi_mac_address.v[3],
            tcpip_data.wifi_mac_address.v[4],
            tcpip_data.wifi_mac_address.v[5]);
    
    memset(buffer, 0, size);
    memcpy(buffer, mac_addr, min(17, size));
 }

/**-----------------------------------------------------------------------------
 * Initialise the tcpip module
 */
void tcpip_initialise(SYS_MODULE_OBJ* tcpip_module, 
        STRUCTURED_MESSAGE_CALLBACK callback,
        TCP_CONNECTED_CALLBACK connected_callback)
{
    tcpip_data.callback = callback;
    tcpip_data.connected_callback = connected_callback;
    memset(tcpip_data.network_name, 0, TCPIP_NETWORK_NAME_MAX_LEN);
    memset(tcpip_data.ssid, 0, TCPIP_SSID_MAX_LEN);
    
    avatar_message_set_tcp_callback(tcpip_avatar_tcp_callback);
    
    //logger_log_infof(LOGGER_CODE_TCPIP_SOCKET_INITILISATION, "TCPIP Name %s", tcpip_data.network_name);
    
    int x;
    for (x = 0; x < TCPIP_MAX_CLIENTS; x++)
    {
        tcpip_data.client_connections[x].index = x;
        
        // Initialise the client id
        tcpip_data.client_connections[x].client_id = TCP_CLIENT_ID_INVALID;
        
        // Initialise the message
        tcpip_initialise_structured_message(&tcpip_data.client_connections[x].current_message);
    }
    
    tcpip_data.initialised = true;
}

/**----------------------------------------------------------------------------- 
 * Determines whether TCP/IP is connected to a network
 */
bool tcpip_is_connected()
{
    bool result = false;

    if (tcpip_data.initialised)
    {
        int x;
        for (x = 0; x < TCPIP_MAX_CLIENTS; x++)
        {
            if (tcpip_data.client_connections[x].client_id != TCP_CLIENT_ID_INVALID)
            {
                logger_log_debugf(0, "TCP IS ACTIVE FOR CLIENT %d --", x);
                result = true;
                break;
            }
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Perform TCP/IP processing
 */
void tcpip_process()
{
    logger_log_warningf(LOGGER_CODE_TCPIP, "TCPIP tcpip_process called but not implemented ");
}

/**-----------------------------------------------------------------------------
 * Reviews the TCPIP connection process
 */
void tcpip_review_connection()
{
    logger_log_warningf(LOGGER_CODE_TCPIP, "TCPIP tcpip_review_connection called but not implemented ");
}

/**-----------------------------------------------------------------------------
 * Send a bufer of data to the specified client
 * @param client_idx - The index of the client socket
 * @param buffer - A pointer to the buffer of data to be sent
 * @param len - The length of the buffer
 */
void tcpip_send_buffer(uint8_t client_idx, char* buffer, int len)
{
    if (tcpip_data.client_connections[client_idx].client_id != TCP_CLIENT_ID_INVALID)
    {
        avatar_send_tcp_data(tcpip_data.client_connections[client_idx].client_id, buffer, len);
    }
}

/**-----------------------------------------------------------------------------
 * Send a bufer of data to the specified client
 * @param client_idx - The index of the client socket
 * @param buffer - A pointer to the buffer of data to be sent
 * @param len - The length of the buffer
 */
void tcpip_send_buffer_to(uint8_t client_idx, char* buffer, int len)
{
    if (tcpip_data.client_connections[client_idx].client_id == TCP_CLIENT_ID_INVALID)
    {
        avatar_send_tcp_data(tcpip_data.client_connections[client_idx].client_id, buffer, len);
    }
}

/**-----------------------------------------------------------------------------
 * Send a structured message over TCPIP
 * @param message
 */
void tcpip_send_structured_message(STRUCTURED_MESSAGE* message)
{
    if (tcpip_data.initialised)
    {
        char buffer[STRUCTURED_MESSAGE_MAX_SIZE];

        // Fill the buffer with the message
        int len = structured_message_write_response_buffer(message, buffer, STRUCTURED_MESSAGE_MAX_SIZE);

        if (message->state_data[0] != TCPIP_UNSOLICITED_CLIENT_CONNECTION)
        {
            tcpip_send_buffer(message->state_data[0], buffer, len);
        }
        else
        {
            // Unsolicited messages are broadcast to all active clients
            int x;
            for (x = 0; x < TCPIP_MAX_CLIENTS; x++)
            {
                tcpip_send_buffer(x, buffer, len);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Control whether DHCP is enabled or not
 * @param enable
 */
void tcpip_set_dhcp_enabled(bool enable)
{
    // Not Implemented - handled by the Avatar
    logger_log_warningf(LOGGER_CODE_TCPIP, "TCPIP tcpip_set_dhcp_enabled called but not implemented ");
}
 
/**-----------------------------------------------------------------------------
 * Sets the network name for the device - This will be obtained from the Avatar
 * @param enable
 */
void tcpip_set_network_name(char* name, size_t size)
{
    memcpy(&tcpip_data.network_name, name, min(size, TCPIP_NETWORK_NAME_MAX_LEN));
    
    logger_log_infof(LOGGER_CODE_TCPIP, "TCPIP Network Name %s --", tcpip_data.network_name);
}
 
/**-----------------------------------------------------------------------------
 * Sets the network name for the device - This will be obtained from the Avatar
 * @param enable
 */
void tcpip_set_ssid(char* name, size_t size)
{
    memcpy(&tcpip_data.ssid, name, min(size, TCPIP_SSID_MAX_LEN));
    
    logger_log_infof(LOGGER_CODE_TCPIP, "TCPIP SSID %s --", tcpip_data.ssid);
}

/**-----------------------------------------------------------------------------
 * Sets the static IP Address
 */
void tcpip_set_static_ip_address(IP_ADDR* ip_address)
{
    // Not Implemented - handled by the Avatar
    logger_log_warningf(LOGGER_CODE_TCPIP, "TCPIP tcpip_set_static_ip_address called but not implemented ");
}

/* *****************************************************************************
 End of File
 */
