/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Arcam

  @File Name
    timers.c

  @Summary
    Manages a set of timers whose duration is a multiple of 1ms.

  @Description
    The number of timers is defined in the TIMERS_NUM_TIMERS declaration in the timers.h
 * header
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/timers.h"
#include "driver/driver_common.h"
#include "system_config.h"
#include <sys/time.h>
#include <time.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define READ_CORE_TIMER()               _CP0_GET_COUNT() // Read the MIPS Core Timer

// Defines the maximum number of system timers that this can manage
#define TIMERS_MAX_CONTEXTS             2

// Error returned from timers_add_timer if the number of timer modules defined in
// TIMERS_MAX_CONTEXTS has been exceeded
#define TIMERS_MAX_CONTEXTS_EXCEEDED    -1

typedef struct
{
    // Holds references to context indexes for use in the callback
    DRV_HANDLE module_handle[TIMERS_MAX_CONTEXTS];
    
    // Holds references to context indexes for use in the callback
    short module_context[TIMERS_MAX_CONTEXTS];
    
    // Maintains the number of assigned timer modules
    unsigned short module_count;
    
    //! The array of timer active values. When a timer is active it counts down to expiry
    // When not active it does not expire but is reset by a timers_reset.
    bool timer_active[TIMERS_NUM_TIMERS];
    
    //! The array of timer suspended values. When a timer is suspended it will never expire
    bool timer_suspended[TIMERS_NUM_TIMERS];
    
    //! The array of timer countdown values
    unsigned int timer_counter[TIMERS_NUM_TIMERS];
    
    //! The array of timeout values in ms
    unsigned int timer_timeout[TIMERS_NUM_TIMERS];
    
    //! Identifies which module the timer is associated with
    unsigned int timer_module_index[TIMERS_NUM_TIMERS];
   
    //! The array of timer cycles (always increments)
    unsigned long int timer_cycle_counter[TIMERS_NUM_TIMERS];
    
    //! Array of timer events
    TIMERS_EXPIRED_EVENT timer_expired_events[TIMERS_NUM_TIMERS];
    
    //! Used for <time.h> functions.
    // Note this is incremented by the cycle of the most active time (currently 100us)
    // so should not be used for accurate time stuff.
    volatile struct timeval current_time;
    
    // The module context of the shortest time period timer
    short time_context;
    
    // The amount of time change when the time_context call back is invoked
    int time_delta;

} TIMERS_DATA;

static TIMERS_DATA timers_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! timers_callback
/**
 * Called when initialise_managed_timers is used.
 * @param context
 * @param alarmCount
 */
static void timers_callback(uintptr_t context, uint32_t alarmCount)
{
    int x;
    short module_context = context;
    
    // Try and increment time (assumes timers_data.time_delta <= 1000)
    if (module_context == timers_data.time_context)
    {
        timers_data.current_time.tv_usec += timers_data.time_delta;  
        timers_data.current_time.tv_sec += (timers_data.current_time.tv_usec / 1000);
        timers_data.current_time.tv_sec %= 1000;
    }
    
    for (x = 0; x < TIMERS_NUM_TIMERS; x++)
    {
        // Check each timer
        if (module_context == timers_data.timer_module_index[x])
        {
             if ((timers_data.timer_active[x]) &&
                (timers_data.timer_timeout[x] > 0) && 
                (timers_data.timer_counter[x] > 0))
            {
                timers_data.timer_counter[x]--;

                // Raise an interrupt if we reach 0
                if (timers_data.timer_counter[x] == 0)
                {
                     if (timers_data.timer_expired_events[x] != NULL)
                    {
                        // Raise interrupt
                        timers_data.timer_expired_events[x](x);
                    }
                }
            }
            
            timers_data.timer_cycle_counter[x]++;
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
/*
 * Section: <time.h> functions   
 * Code to handle the warning "c:/microchip/harmony/v1_10/framework/crypto/src/asn.c:2866: 
 * warning: Linking code with default time() stub. Hint: Write an app-specific implementation." 
 * when using the Crypto API
 * http://www.microchip.com/forums/m712292.aspx
 * */
/* ************************************************************************** */
/* ************************************************************************** */
  
int gettimeofday(struct timeval *tv , void *tz)
{ 
    unsigned int intStatus;
    intStatus = INTDisableInterrupts(); 
    tv->tv_usec = timers_data.current_time.tv_usec; 
    tv->tv_sec = timers_data.current_time.tv_sec; 
    INTRestoreInterrupts(intStatus); 
    return 0;
}
 
int settimeofday(const struct timeval *tv , void *tz) 
{ 
    unsigned int intStatus;
    intStatus = INTDisableInterrupts();        
    timers_data.current_time.tv_usec = tv->tv_usec; 
    timers_data.current_time.tv_sec= tv->tv_sec; 
    INTRestoreInterrupts(intStatus); 
    return 0;
}

time_t time(time_t *tod)
{ 
    time_t t;
    t = (time_t)timers_data.current_time.tv_sec; 
    if (tod != NULL) *tod = t;  
    return t;
}

/**
 * Activates.Deactivates a timer
 * By default a timer is active if added with TIMERS_TIMEOUT_RESET_IMMEDIATE
 * and deactivates with TIMERS_TIMEOUT_RESET_NONE
 * @param timer_idx
 * @param active
 */
static void timers_set_active(unsigned int timer_idx, bool active)
{
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        // Store the timeout value
        timers_data.timer_active[timer_idx] = active;
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**
 * Activates a suspended timer - when suspended it will never expire
 * @param timer_idx
 */
void timers_activate(unsigned int timer_idx)
{
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        timers_data.timer_suspended[timer_idx] = false;
        timers_data.timer_active[timer_idx] = true;
    }
}
    
/**
 * Adds an event to the specified timer tat is triggered whenever the timer 
 * expires this is done as part of the interrupt event so should be kept as 
 * brief as possible
 * @param timer_idx
 * @param event
 */
void timers_add_expired_event(unsigned int timer_idx, TIMERS_EXPIRED_EVENT event)
{
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        timers_data.timer_expired_events[timer_idx] = event;
    }
}

/**
 * Add a timer to the timers module
 * @param module_index - The module index e.g. DRV_TMR_INDEX_0
 * @param frequency - the frequency of the time (in Hz)
 * @return - An index to reference the timer by when adding timer instances (see
 * timers_set_timeout)
 */
short timers_add_timer(const SYS_MODULE_INDEX module_index, int frequency)
{
    short result = TIMERS_MAX_CONTEXTS_EXCEEDED;
    
    if (TIMERS_MAX_CONTEXTS > timers_data.module_count)
    {
        // Use the current module_count as the index for the module
        result = timers_data.module_count;
        
        // Store the index in the context array for comparison in the callback
        // This is needed so it can be passed as a context
        timers_data.module_context[result] = result;
        
        // Increment for next time
        timers_data.module_count++;

        // Initialise the 1ms Timer 
        timers_data.module_handle[result] = DRV_TMR_Open(module_index,
                                                         DRV_IO_INTENT_EXCLUSIVE);   
        
        // See if this should be used to calculate time <time.h> 
        if ((1000000 /frequency) < timers_data.time_delta)
        {
            timers_data.time_context = result;
            timers_data.time_delta = (1000000 /frequency); // > 1000
        }

        // Set up timer 
        uint32_t desiredFrequency = frequency; 
        uint32_t actualFrequency = DRV_TMR_CounterFrequencyGet(timers_data.module_handle[result]);;
        uint32_t divider = actualFrequency / desiredFrequency; // calculate divider value

        DRV_TMR_AlarmRegister(timers_data.module_handle[result], 
                              divider, 
                              true, 
                              timers_data.module_context[result], 
                              timers_callback);

        // Starting the Timer   
        DRV_TMR_Start(timers_data.module_handle[result]);
    }
    
    return result;
}
    
/**
 * Delay a process by a given number of milliseconds
 * Note: Uses the core timer
 */
void timers_delay_ms(uint16_t milliseconds)
{
    timers_delay_us(milliseconds * 1000);
}

/**
 * Delay a process by a given number of 100 nanoseconds
 * http://www.microchip.com/forums/m954661.aspx
 * Note: Uses the core timer
 */
void timers_delay_100ns(uint16_t nanoseconds)
{
    uint32_t time;
    
    time = READ_CORE_TIMER(); // Read Core Timer    
    time += (SYS_CLK_FREQ / 2 / 10000000) * nanoseconds; // calc the Stop Time    
    while ((int32_t)(time - READ_CORE_TIMER()) > 0){};    
}

/**
 * Delay a process by a given number of microseconds
 * http://www.microchip.com/forums/m954661.aspx
 * Note: Uses the core timer
 */
void timers_delay_us(uint16_t microseconds)
{
    uint32_t time;
    
    time = READ_CORE_TIMER(); // Read Core Timer    
    time += (SYS_CLK_FREQ / 2 / 1000000) * microseconds; // calc the Stop Time    
    while ((int32_t)(time - READ_CORE_TIMER()) > 0){};    
}

/**
 * Forces the current timer to put itself into the expired state
 * @param timer_idx - The timer index
 */
void timers_expire(unsigned int timer_idx)
{
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        timers_data.timer_counter[timer_idx] = 0;
    }
}
     
/**
 * Gets the number of cycles since the last reset
 * Note : there is no range checking on the index supplied
 * @param timer_idx - The timer index
 * @return - the number of complete cycles
 */
unsigned int timers_get_elapsed_count(unsigned int timer_idx)
{
   return timers_data.timer_timeout[timer_idx] - timers_data.timer_counter[timer_idx];
}

/**
 * Gets the number of cycles since the last reset
 * Note : there is no range checking on the index supplied
 * @param timer_idx - The timer index
 * @return - the number of complete cycles
 */
unsigned int timers_get_remaining_count(unsigned int timer_idx)
{
    unsigned int result = 0;
    
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        result = timers_data.timer_counter[timer_idx];
    }
    
    return result;
}
     
/**
 * Gets the current timeout value for a given timer
 * @param timer_idx
 * @return 
 */
unsigned int timers_get_timeout(unsigned int timer_idx)
{
    unsigned int result = 0;
    
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        result = timers_data.timer_timeout[timer_idx];
    }
    
    return result;
}

/**
 * Checks if the timer has expired
 * 
 * @param timer_idx
 * @return - true if the timer has expired, false if not
 */
bool timers_has_expired(unsigned int timer_idx)
{
    bool result = false;
    
    if ((timers_data.timer_active[timer_idx]) &&
        (timers_data.timer_timeout[timer_idx] > 0) && 
        (timers_data.timer_counter[timer_idx] == 0))
    {
        result = true;
    }
   
    return result;
}

/**
 * Initialise the timer functionality 
 */
void timers_initialise(void)
{
    // <time.h> data
    timers_data.current_time.tv_sec = 0;    
    timers_data.current_time.tv_usec = 0;    

    // Initialise all the timer counts to 0
    memset(timers_data.timer_counter, 0, sizeof(int) * TIMERS_NUM_TIMERS);
    memset(timers_data.timer_timeout, 0, sizeof(int) * TIMERS_NUM_TIMERS);
    memset(timers_data.timer_active, 0, sizeof(bool) * TIMERS_NUM_TIMERS);
    memset(timers_data.timer_suspended, 0, sizeof(bool) * TIMERS_NUM_TIMERS);
    memset(timers_data.timer_expired_events, 0, sizeof(TIMERS_EXPIRED_EVENT) * TIMERS_NUM_TIMERS);

    timers_data.module_count = 0;
    
    timers_data.time_context = -1;
    timers_data.time_delta = 10000; // > 1000
}
    
/**
 * Used to test whether a timer is currently active.  
 * @param timer_idx
 * @return 
 */
bool timers_is_active(unsigned int timer_idx)
{
    return timers_data.timer_active[timer_idx];
}

/**
  * resets the timeout of a timer in ms
 * Note: Will cause a reset
 * Note: timer_idx must be < MAX_TIMERS
 * 
 * @param timer_idx - The index of timer to use
 */
void timers_reset(unsigned int timer_idx)
{
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        timers_data.timer_counter[timer_idx] = timers_data.timer_timeout[timer_idx];
       
        if (!timers_data.timer_suspended[timer_idx] == true)
        {
            timers_data.timer_active[timer_idx] = true;
        }
    }
}

/**
  * Set the timeout of a timer in ms
 * Note: Will cause a reset
 * Note: timer_idx must be < MAX_TIMERS
 * 
 * @param timer_idx - The index of timer to use
 * @param timeout_ms - The timeout in ms
 * @param reset - if TIMERS_TIMEOUT_RESET_IMMEDIATE the timer is reset to its 
 * , otherwise it is left until it is left to expire and not reset
 */
void timers_set_timeout(short module_context, unsigned int timer_idx, unsigned int timeout, TIMERS_TIMEOUT_RESET reset)
{
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        // Store the timeout value
        timers_data.timer_module_index[timer_idx] = module_context;
        timers_update_timeout(timer_idx, timeout, reset);
    }
}

/**
 * Suspends the timer - when suspended it cannot be reset
 * @param timer_idx
 */
void timers_suspend(unsigned int timer_idx)
{
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        timers_data.timer_suspended[timer_idx] = true;
        timers_data.timer_active[timer_idx] = false;
    }
}
    
/**
  * Set the timeout of a timer in ms
 * Note: Will cause a reset
 * Note: timer_idx must be < MAX_TIMERS
 * 
 * @param timer_idx - The index of timer to use
 * @param timeout_ms - The timeout in ms
 * @param reset - if TIMERS_TIMEOUT_RESET_IMMEDIATE the timer is reset to its 
 * , otherwise it is left until it is left to expire and not reset
 */
void timers_update_timeout(unsigned int timer_idx, unsigned int timeout, TIMERS_TIMEOUT_RESET reset)
{
    if (timer_idx < TIMERS_NUM_TIMERS)
    {
        // Store the timeout value
        timers_data.timer_timeout[timer_idx] = timeout;
        timers_data.timer_counter[timer_idx] = 0;
        
        switch(reset)
        {
            case TIMERS_TIMEOUT_RESET_NONE:
                timers_data.timer_active[timer_idx] = true;
                break;
                
            case TIMERS_TIMEOUT_RESET_INACTIVE:
                timers_data.timer_active[timer_idx] = false;
                timers_data.timer_counter[timer_idx] = timeout;
                break;
                
            default:
                timers_data.timer_active[timer_idx] = true;
                timers_reset(timer_idx);
                break;
        }
    }
}

/* *****************************************************************************
 End of File
 */
