/* ************************************************************************** */
/** Descriptive File Name

  @Company
    ARCAM

  @File Name
    udp.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/udp.h"
#include <stdint.h>
#include "tcpip/tcpip.h"
#include "utilities/logger.h"
#include "app_config.h"
#include <sys/errno.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Maximum number of ports to monitor
#ifndef UDP_MAX_SERVER_PORTS
#define UDP_MAX_SERVER_PORTS        2
#endif

//! Maximum number of callbacks
#ifndef UDP_MESSAGE_CALLBACK_MAX
#define UDP_MESSAGE_CALLBACK_MAX    2
#endif

typedef enum
{
    UDP_STATE_WAIT_INIT,
    UDP_STATE_BSD_BIND,
    UDP_STATE_CREATE_CLIENT,
    UDP_STATE_BSD_OPERATION,
    UDP_STATE_ERROR,
} UDP_STATE;

/**
 */
typedef struct
{
    // The current state of the UDP connection
    UDP_STATE state;
    
    UDP_MESSAGE_CALLBACK callbacks[UDP_MESSAGE_CALLBACK_MAX];
    
    // Server socket
    SOCKET client_socket;
    
    //! An array containing port numbers
    uint16_t ports[UDP_MAX_SERVER_PORTS];
    
    //! The number of ports in use (MUST be <= UDP_MAX_SERVER_PORTS)
    uint8_t num_ports;
    
    // Server socket
    SOCKET server_socket[UDP_MAX_SERVER_PORTS];
    
    // A buffer used to receive data from a socket - will be overwritten regularly
    unsigned char receive_buffer[UDP_MESSAGE_MAX_SIZE];
    
} UDP_DATA;

static UDP_DATA udp_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

static bool udp_bind()
{
    bool result = true;
logger_log_debugf(LOGGER_CODE_UDP, "UDP BIND");    
    int x;
    for (x = 0; x < udp_data.num_ports; x++)
    {
        struct sockaddr_in address;
        int addrlen = sizeof(struct sockaddr_in);
        address.sin_family = AF_INET;
        address.sin_addr.s_addr = INADDR_ANY;
        address.sin_port = udp_data.ports[x];

        if (bind(udp_data.server_socket[x], (struct sockaddr*)&address, addrlen) == SOCKET_ERROR)
        {
            logger_log_errorf(LOGGER_CODE_UDP_SOCKET_INITILISATION, "UDP Unable to bind server socket on port %d ", udp_data.ports[x]);

            result = false;       
        }
        else
        {
            logger_log_infof(LOGGER_CODE_UDP_SOCKET_INITILISATION, "UDP Server Socket bound on port %d ", udp_data.ports[x]);
        }
    }
    
    return result;
}

/**
 * Setup client sockets
 */
static bool udp_create_client_socket()
{
    bool result = true;
    
    // Create a socket for this server to send data 
    udp_data.client_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    
    if(udp_data.client_socket == INVALID_SOCKET)
    {
        logger_log_error(LOGGER_CODE_UDP_SOCKET_ERROR, "UDP Unable to create client socket");
        
        result = false;
    }
    else
    {
        logger_log_info(LOGGER_CODE_UDP_SOCKET_INITILISATION, "UDP client socket created");
    }
    
    return result;
}

/**
 * Setup client sockets
 */
static bool udp_create_server_socket()
{
    bool result = true;
    
    int x;
    for (x = 0; x < udp_data.num_ports; x++)
    {
        // Create a socket for this server to receive data 
        udp_data.server_socket[x] = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

        if(udp_data.server_socket[x] == INVALID_SOCKET)
        {
            logger_log_errorf(LOGGER_CODE_UDP_SOCKET_ERROR, "UDP Unable to create server socket for port %d ", udp_data.ports[x]);

            result = false;
        }
        else
        {
            logger_log_infof(LOGGER_CODE_UDP_SOCKET_INITILISATION, "UDP Server Socket created for port %d ", udp_data.ports[x]);
        }
    }
    
    return result;
}

/**
 * Receive incoming data and turn into messages
 */
static void udp_receive()
{
    int x;
    for (x = 0; x < udp_data.num_ports; x++)
    {
        if (udp_data.server_socket[x] != INVALID_SOCKET)
        {
            struct sockaddr_in address;
            int address_len = sizeof(struct sockaddr_in);

            // See if there is any data to be retrieved into the global work buffer
            int count = recvfrom(udp_data.server_socket[x], 
                    udp_data.receive_buffer, 
                    sizeof(udp_data.receive_buffer), 
                    0,
                    (struct sockaddr*)&address, 
                    &address_len);

            if(count > 0)
            {
                logger_log_debugf(LOGGER_CODE_UDP_SENDTO_ERROR, "UDP %d bytes recvfrom %d.%d.%d.%d:%d on port %d  --", 
                    count,
                    address.sin_addr.S_un.S_un_b.s_b1,
                    address.sin_addr.S_un.S_un_b.s_b2,
                    address.sin_addr.S_un.S_un_b.s_b3,
                    address.sin_addr.S_un.S_un_b.s_b4,
                    address.sin_port,
                    udp_data.ports[x]);

                int x;
                for (x = 0; x < UDP_MESSAGE_CALLBACK_MAX; x++)
                {
                    UDP_MESSAGE_CALLBACK callback = (UDP_MESSAGE_CALLBACK)udp_data.callbacks[x];
                    if (callback != NULL)
                    {
                        if (callback(udp_data.ports[x], &address, &address_len, udp_data.receive_buffer, count))
                        {
                            //break; Removed as all callbacks need a go at the message
                        }
                    }
    //                else
    //                {
    //                    break;
    //                }
                }
            }
            else if (count == SOCKET_ERROR)
            {
                // Clear the socket down
                if (errno == ENOTCONN)
                {
                    logger_log_infof(LOGGER_CODE_UDP_RECV_ERROR, "UDP Server Socket Disconnected");
                    udp_data.server_socket[x] = INVALID_SOCKET;

                    // Attempt to reconnect
                    udp_data.state = UDP_STATE_WAIT_INIT;
                }
            }
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Adds a callback to the UDP module - can have upto a maximum of UDP_MESSAGE_CALLBACK_MAX
 * @param callback
 */
void udp_add_callback(UDP_MESSAGE_CALLBACK callback)
{
    int x; 
    bool added = false;
    
    for (x = 0; x < UDP_MESSAGE_CALLBACK_MAX; x++)
    {
        if (udp_data.callbacks[x] == NULL)
        {
            udp_data.callbacks[x] = callback;
            added = true;
            break;
        }
    }
    
    if (!added)
    {
        logger_log_errorf(LOGGER_CODE_UDP_SENDTO_ERROR, "UDP Unable to add callback - max callbacks (%d) exceeded", UDP_MESSAGE_CALLBACK_MAX);
    }
}

/**
 * Setup and connect UDP socket
 */
void udp_connect()
{
    switch(udp_data.state)
    {
        case UDP_STATE_WAIT_INIT:
        {
            if (udp_create_server_socket())
            {
                // This will loop if there is an error! 
                udp_data.state = UDP_STATE_BSD_BIND;
            }
            
            break;
        }
        
        case UDP_STATE_BSD_BIND:
        {
            if (udp_bind())
            {
                udp_data.state = UDP_STATE_CREATE_CLIENT;   
            }
            
            break;
        }
        
        case UDP_STATE_CREATE_CLIENT:
        {
            if (udp_create_client_socket())
            {
                udp_data.state = UDP_STATE_BSD_OPERATION;   
            }
        }
        
        default:
            break;
    }
}

/**
 * Test whether UDP connect has completed
 * @return 
 */
bool udp_is_connected()
{
    return udp_data.state == UDP_STATE_BSD_OPERATION;
}
    
/**
 * Initialise the UDP module
 */
void udp_initialise(uint16_t* port_numbers, uint8_t num_ports)
{
    logger_log_debugf(LOGGER_CODE_UDP, "UDP Initialise");    

    udp_data.state = UDP_STATE_WAIT_INIT;
    
    // Sort out ports
    if (num_ports > UDP_MAX_SERVER_PORTS)
    {
        logger_log_errorf(LOGGER_CODE_UDP, "UDP Number of ports supplied (%d) exceeds the maximum allowed (%d)", num_ports, UDP_MAX_SERVER_PORTS);
    }
    udp_data.num_ports = min(num_ports, UDP_MAX_SERVER_PORTS);
    
    memset(udp_data.ports, 0, sizeof(uint16_t) * UDP_MAX_SERVER_PORTS);
    memcpy(udp_data.ports, port_numbers, sizeof(uint16_t) * udp_data.num_ports);
    
    int x;
    for (x = 0; x < UDP_MESSAGE_CALLBACK_MAX; x++)
    {
        udp_data.callbacks[x] = NULL;
    }
}

/**
 * Process UDP messages
 */
void udp_process()
{
    switch(udp_data.state)
    {
        case UDP_STATE_BSD_OPERATION:
        {
            udp_receive();
            break;
        }

        case UDP_STATE_WAIT_INIT:
        {
            // Attempt to reconnect
            udp_connect();
            break;
        }
        
        default:
            break;
    }        
}
   
/**
 * Send a UDP message
 */
void udp_sendto(IP_ADDR* ip_address, uint16_t port, char* buffer, int len)
{
    struct sockaddr_in address;
    
      /* The address is IPv4 */
    address.sin_family = AF_INET;
    address.sin_addr.s_addr = ip_address->Val;

    /* sockets are unsigned shorts, htons(x) ensures x is in network byte order, set the port to 7654 */
    address.sin_port = port;
 
    int count = sendto(udp_data.client_socket, 
            buffer, 
            len, 
            0,
            (struct sockaddr*)&address, 
            sizeof address);
    
    if (count < 0) 
    {
        logger_log_errorf(LOGGER_CODE_UDP_SENDTO_ERROR, "UDP Unable to send buffer to %d.%d.%d.%d:%d", 
                address.sin_addr.S_un.S_un_b.s_b1,
                address.sin_addr.S_un.S_un_b.s_b2,
                address.sin_addr.S_un.S_un_b.s_b3,
                address.sin_addr.S_un.S_un_b.s_b4,
                address.sin_port);
    }
    else
    {
        logger_log_infof(LOGGER_CODE_UDP_SENDTO_ERROR, "UDP buffer sent to to %d.%d.%d.%d:%d", 
                address.sin_addr.S_un.S_un_b.s_b1,
                address.sin_addr.S_un.S_un_b.s_b2,
                address.sin_addr.S_un.S_un_b.s_b3,
                address.sin_addr.S_un.S_un_b.s_b4,
                address.sin_port);
    }
}

/* *****************************************************************************
 End of File
 */
