/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/upgrade.h"
#include "sys/kmem.h"
#include "app_config.h"
#include "utilities/logger/logger.h"
#include "bootloader/src/nvm.h"
#include "peripheral/nvm/plib_nvm.h"
#include "peripheral/int/plib_int.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define END_OF_FILE_RECORD          1
#define EXT_SEG_ADRS_RECORD         2
#define EXT_LIN_ADRS_RECORD         4
#define REC_FLASHED                 0
#define REC_NOT_FOUND               1
#define REC_FOUND_BUT_NOT_FLASHED   2

//! Copied from nvm.c
typedef struct
{
    uint8_t RecDataLen;
    uint32_t Address;
    uint8_t RecType;
    uint8_t* Data;
    uint8_t CheckSum;
    uint32_t ExtSegAddress;
    uint32_t ExtLinAddress;
}T_HEX_RECORD;

// Copied from GenericTypeDefs.h
typedef unsigned char		BYTE;				// 8-bit unsigned

typedef struct _DWORD_VAL
{
    BYTE LB;
    BYTE HB;
    BYTE UB;
    BYTE MB;
} DWORD_VAL;

/**
 * Used to keep track of upload progress
 */
//typedef struct
//{
//    //UINT8 *start;
//    UINT8 len;
//    UINT8 status;
//} PROGRESS_RECORD;

//! The size of the buffer used to store data between upgrade_write_buffer calls
//#define UPGRADE_BUFFER_SIZE     512
#define HEX_RECORD_BUFFER_SIZE  100

typedef struct
{
    UINT8 len;
    UINT8 status;
    uint8_t hex_record[HEX_RECORD_BUFFER_SIZE];
    uint32_t processed_records;
    
} UPGRADE_DATA;

static UPGRADE_DATA upgrade_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Convert an ascii record into a hex representation
 * @param ascii_rec
 * @param hex_rec
 */
static void upgrade_ascii_to_hex(UINT8* ascii_rec, UINT8* hex_rec)
{
	UINT8 i = 0;
	UINT8 k = 0;
	UINT8 hex;
	
    //logger_log_debugf(23, "UPGRADE CONVERT A %d %d %d --", i, ascii_rec[i], hex_rec[i/2]);
    
	while((ascii_rec[i] >= 0x30) && (ascii_rec[i] <= 0x66))
	{
		// Check if the ascii values are in alpha numeric range.
		if(ascii_rec[i] < 0x3A)
		{
			// Numerical representation in ASCII found.
			hex = ascii_rec[i] & 0x0F;
		}
		else
		{
			// Alphabetical value.
			hex = 0x09 + (ascii_rec[i] & 0x0F);						
		}
	
		// Following logic converts 2 bytes of ASCII to 1 byte of hex.
		k = i%2;
		
		if(k)
		{
			hex_rec[i / 2] |= hex;
			
		}
		else
		{
			hex_rec[i / 2] = (hex << 4) & 0xF0;
		}	
	
        //logger_log_debugf(23, "UPGRADE CONVERT %d %c %d --", i, ascii_rec[i], hex_rec[i/2]);
        
		i++;		
	}		
}

/**-----------------------------------------------------------------------------
 * Clears progress indicators ready to start processing of a new record
 */
static void upgrade_clear_progress()
{
    memset(upgrade_data.hex_record, 0, HEX_RECORD_BUFFER_SIZE);
    upgrade_data.len = 0;
    upgrade_data.status = REC_FLASHED;
}

/**-----------------------------------------------------------------------------
 * Write the hex buffer to flash
 * @param hex_rec
 */
static UINT upgrade_write_flash(UINT8* hex_rec)
{
    UINT result = 0;
	static T_HEX_RECORD hex_data;
	UINT WrData;
	void* prog_address;
		
	hex_data.RecDataLen = hex_rec[0];
	hex_data.RecType = hex_rec[3];	
	hex_data.Data = &hex_rec[4];	

    //logger_log_debugf(23, "UPGRADE Loading hex data %d ", hex_data.RecDataLen);
    result = APP_ProgramHexRecord(hex_rec, hex_data.RecDataLen + 5);	
        
    return result;
}	

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Use to determine whether an upgrade can proceed
 * @return 
 */
bool upgrade_can_proceed()
{
    return PLIB_NVM_FlashWriteCycleHasCompleted(NVM_ID_0);
}

/**-----------------------------------------------------------------------------
 * Erase the current flash memory
 */
void upgrade_erase_flash()
{
    logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE Erase Flash ");

    PLIB_NVM_MemoryModifyInhibit(NVM_ID_0);

    //upgrade_erase_flash();
    APP_FlashErase();

    while (!PLIB_NVM_FlashWriteCycleHasCompleted(NVM_ID_0));

    APP_NVMClearError();

    PLIB_NVM_MemoryModifyInhibit(NVM_ID_0);         
}

/**-----------------------------------------------------------------------------
 * Initialises the upgrade process - must be called at the start of any upgrade
 */
void upgrade_initialise()
{
    logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE Initialise ");
    
    upgrade_erase_flash();
    
    upgrade_data.processed_records = 0;
    
    upgrade_clear_progress();
}

/**-----------------------------------------------------------------------------
 * Writes a buffer from a hex file into flash
 * @param buffer
 * @param num_bytes
 * @return 
 */
UINT upgrade_write_buffer(UINT8* buffer, size_t num_bytes)
{
    UINT result = 0;
    volatile UINT x;
    
    for (x = 0; x < num_bytes; x++)
    {
        //logger_log_debugf(23, "UPGRADE Reviewing Buffer %d %d %d ", upgrade_data.progress.status, x, buffer[x]);
        // This state machine separates-out the valid hex records from the read 512 bytes.
        switch (upgrade_data.status)
        {
            case REC_FLASHED:
            case REC_NOT_FOUND:
            {
                if (buffer[x] == ':')
                {
                    // We have a record found in the 512 bytes of data in the buffer.
                    //logger_log_debugf(23, "UPGRADE Found record start at %d ", x);

                    upgrade_clear_progress();
                    upgrade_data.status = REC_FOUND_BUT_NOT_FLASHED;
                }

                break;
            }

            case REC_FOUND_BUT_NOT_FLASHED:
            {
                upgrade_data.hex_record[upgrade_data.len - 1] = buffer[x];

                //logger_log_debugf(23, "UPGRADE REC_FOUND_BUT_NOT_FLASHED %d %d %d --", upgrade_data.len - 1, upgrade_data.hex_record[upgrade_data.len - 1], buffer[x]);

                if ((buffer[x] == 0x0A) || (buffer[x] == 0xFF))
                {
                    // We have got a complete record. (0x0A is new line feed and 0xFF is End of file)
                    // Start the hex conversion from element
                    // 1. This will discard the ':' which is
                    // the start of the hex record.
                    UINT8 hex_record[UPGRADE_MAX_HEX_RECORD_SIZE];
                    upgrade_ascii_to_hex(upgrade_data.hex_record, hex_record);
                    result = upgrade_write_flash(hex_record);

                    upgrade_data.processed_records++;
                    
                    if (result != HEX_REC_NORMAL)
                    {
                        upgrade_data.hex_record[upgrade_data.len] = 0;
                        logger_log_errorf(0, "UPGRADE Invalid HEX record (record number %d) error %d %s --", upgrade_data.processed_records, result, upgrade_data.hex_record);
                        result = UPGRADE_WRITE_ERROR;
                    }
                        
                    //logger_log_debugf(23, "UPGRADE Writing HEX record 3 --");

                    upgrade_clear_progress();
                    upgrade_data.status = REC_FLASHED;
                }

                break;
            }
        }

        if (result == UPGRADE_WRITE_COMPLETE)
        {
            break;
        }
        else if (result == UPGRADE_WRITE_ERROR)
        {
            break;
        }
        else
        {
            // Move to next byte in the buffer.
            upgrade_data.len++;
            
            // Check were not overflowing records
            if (upgrade_data.len >= UPGRADE_MAX_HEX_RECORD_SIZE)
            {
                result = UPGRADE_WRITE_ERROR;
                break;
            }
        }
    }

    if (result == 0)
    {
        if (upgrade_data.status == REC_FOUND_BUT_NOT_FLASHED)
        {
            // We still have a half read record in the buffer. The next half part of the record is read 
            // when we read 512 bytes of data from the next file read. 
            //memmove(upgrade_data.write_buffer, &upgrade_data.write_buffer[upgrade_data.write_buffer_pos], UPGRADE_BUFFER_LEN - upgrade_data.write_buffer_pos);

            //upgrade_data.status = REC_NOT_FOUND;
            //upgrade_data.write_buffer_pos = upgrade_data.progress.len;
        }
    }
    
    return result;
}

/* *****************************************************************************
 End of File
 */
