/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\usb_audio.h"
#include "system_definitions.h"
#include "app_config.h"
#include "utilities\logger.h"
#include "utilities\output_control.h"
#include "utilities\usb_audio_buffer.h"
#include "utilities\usb_audio_defs.h"
#include "utilities\usb_audio_codec.h"
#include "utilities\usb_audio_feedback.h"
#include "utilities\usb_audio_port.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

// Entity IDs
#define USB_AUDIO_ID_INPUT_TERMINAL             0x02
#define USB_AUDIO_ID_FEATURE_UNIT               0x0A
#define USB_AUDIO_ID_OUTPUT_TERMINAL            0x14
#define USB_AUDIO_ID_MIXER_UNIT                 0x04
#define USB_AUDIO_ID_SELECTOR_UNIT              0x05
#define USB_AUDIO_ID_PROCESSING_UNIT            0x06
#define USB_AUDIO_ID_EXTENSION_UNIT             0x07
#define USB_AUDIO_ID_CLOCK_SOURCE               0x28
#define USB_AUDIO_ID_CLOCK_SELECT               0x11

/**
 * # of times we let the DMA ISR call syncLRClock, to make sure it is sync'ed with incoming USB packets
 * the number is somewhat arbitrary; 100 is too low and a much higher number will result in too long a delay before audio begins playing
 */
#define USB_AUDIO_LRCLOCKSYNC                   2000        

/**
 */
typedef enum
{
    USB_SPEAKER_PLAYBACK_NONE  = 0,
    USB_SPEAKER_PLAYBACK_STEREO = 1
} USB_ALTERNATE_SETTINGS;

/**
 */
typedef enum
{
    USB_AUDIO_STATE_INITIALISE,
    USB_AUDIO_STATE_OPEN_CODEC,
    USB_AUDIO_STATE_OPEN_PORT,
    USB_AUDIO_STATE_WAIT_CONFIGURATION,
    USB_AUDIO_STATE_IDLE,
    USB_AUDIO_STATE_INTERFACE_ALTERNATE_SETTING_RCVD,
    USB_AUDIO_STATE_MUTE_AUDIO_PLAYBACK,
    USB_AUDIO_STATE_DAC_MUTE,
    USB_AUDIO_STATE_CLOCKSOURCE_SET,
    USB_AUDIO_STATE_SUBMIT_READ_REQUEST,
    USB_AUDIO_STATE_PROCESS_DATA,
} USB_AUDIO_STATE;

/* 
 * Defines a list of audio controls supported by this application.
 * This enumeration defines audio controls supported by this application. This
 * could be used for checking which Control Transfer is currently active.
*/
typedef enum
{
    USB_CONTROL_NONE,
    USB_AUDIO_MUTE_CONTROL,
    USB_AUDIO_CLOCKSOURCE_CONTROL,
    USB_AUDIO_CLOCKSELECT_CONTROL
} USB_AUDIO_CONTROLS;


/**
 * The range of available sample rates.
 */
typedef struct __attribute__((packed))
{
    unsigned short  numSampleRate;          // Number of supported sampling rates
    USB_AUDIO_V2_LAYOUT_3_RANGE_PARAMETER_BLOCK sampleRate32;
    USB_AUDIO_V2_LAYOUT_3_RANGE_PARAMETER_BLOCK sampleRate44;
    USB_AUDIO_V2_LAYOUT_3_RANGE_PARAMETER_BLOCK sampleRate48;
    USB_AUDIO_V2_LAYOUT_3_RANGE_PARAMETER_BLOCK sampleRate88;
    USB_AUDIO_V2_LAYOUT_3_RANGE_PARAMETER_BLOCK sampleRate96;
    USB_AUDIO_V2_LAYOUT_3_RANGE_PARAMETER_BLOCK sampleRate176;
    USB_AUDIO_V2_LAYOUT_3_RANGE_PARAMETER_BLOCK sampleRate192;
} USB_AUDIO_V2_CLOCKSOURCE_RANGE;

/**
 * Main data structure
 */
typedef struct
{
    // Allows the code to effectively turn USB processing On/Off.  Because it is 
    // processor intensive other activity can cause packets to be dropped and for 
    bool active;
    
    //! 
    USB_ALTERNATE_SETTINGS active_interface_alternate_setting;
    
    //! True when configuration is complete
    bool configured;
    
    //!
    USB_AUDIO_CONTROLS current_audio_control;
     
    //! Determines whether to initialise a read request
    bool initialise_read_request;

    //! The previous sampling rate
    uint32_t last_sample_rate;    
    
    //! Used to sync the L/R clocks
    volatile uint32_t lr_clock_sync;
    
    //! DAC muted state
    bool is_muted;

    //! The current sampling rate
    uint32_t sample_rate;   
    
    USB_AUDIO_V2_CLOCKSOURCE_RANGE sample_rate_range;

    //! Note this replaces clockValid , and is never changed from its initial value = 1
    uint8_t sample_rate_valid;    

    //! Current processing state
    USB_AUDIO_STATE state;
    
} USB_AUDIO_DATA;

static USB_AUDIO_DATA usb_audio_data =
{
    .active = false,
    //.codec_data.next_buffer = 0,
    
    .configured = false,
    .initialise_read_request = true,
    .is_muted = false,
    .lr_clock_sync = USB_AUDIO_LRCLOCKSYNC,
    //.read_queue_idx = 0,
    
    .sample_rate_range.numSampleRate = 7,
    .sample_rate_range.sampleRate32.dMin = USB_AUDIO_SAMPLING_RATE_32KHZ,
    .sample_rate_range.sampleRate32.dMax = USB_AUDIO_SAMPLING_RATE_32KHZ,
    .sample_rate_range.sampleRate32.dRes = 0,
    .sample_rate_range.sampleRate44.dMin = USB_AUDIO_SAMPLING_RATE_44_1KHZ,
    .sample_rate_range.sampleRate44.dMax = USB_AUDIO_SAMPLING_RATE_44_1KHZ,
    .sample_rate_range.sampleRate44.dRes = 0,
    .sample_rate_range.sampleRate48.dMin = USB_AUDIO_SAMPLING_RATE_48KHZ,
    .sample_rate_range.sampleRate48.dMax = USB_AUDIO_SAMPLING_RATE_48KHZ,
    .sample_rate_range.sampleRate48.dRes = 0,
    .sample_rate_range.sampleRate88.dMin = USB_AUDIO_SAMPLING_RATE_88_2KHZ,
    .sample_rate_range.sampleRate88.dMax = USB_AUDIO_SAMPLING_RATE_88_2KHZ,
    .sample_rate_range.sampleRate88.dRes = 0,
    .sample_rate_range.sampleRate96.dMin = USB_AUDIO_SAMPLING_RATE_96KHZ,
    .sample_rate_range.sampleRate96.dMax = USB_AUDIO_SAMPLING_RATE_96KHZ,
    .sample_rate_range.sampleRate96.dRes = 0,
    .sample_rate_range.sampleRate176.dMin = USB_AUDIO_SAMPLING_RATE_176_4KHZ,
    .sample_rate_range.sampleRate176.dMax = USB_AUDIO_SAMPLING_RATE_176_4KHZ,
    .sample_rate_range.sampleRate176.dRes = 0,
    .sample_rate_range.sampleRate192.dMin = USB_AUDIO_SAMPLING_RATE_192KHZ,
    .sample_rate_range.sampleRate192.dMax = USB_AUDIO_SAMPLING_RATE_192KHZ,
    .sample_rate_range.sampleRate192.dRes = 0,
    
    .sample_rate = USB_AUDIO_SAMPLING_RATE_192KHZ,
    .sample_rate_valid = 1,
    .state = USB_AUDIO_STATE_INITIALISE,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations
//!
static void usb_audio_configure_sample_rate(uint32_t sample_rate);
//!
static int usb_audio_current_entity_settings_handler
(
    USB_DEVICE_AUDIO_V2_INDEX index,
    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST *setup_packet
);
//!
static int usb_audio_feature_unit_request_handler
(
    USB_DEVICE_AUDIO_V2_INDEX index ,
    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST* control_request
);
//!
static USB_AUDIO_CONTROLS usb_audio_get_current_audio_control();
//!
static void usb_audio_initialise_audio_buffer();
//!
static void usb_audio_initialise_data();
//!
static int usb_audio_range_entity_settings_handler
(
    USB_DEVICE_AUDIO_V2_INDEX index ,
    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST *setup_packet
);
//!
static void usb_audio_set_configured(bool value);
//!
static void usb_audio_set_current_audio_control(USB_AUDIO_CONTROLS control);

/**-----------------------------------------------------------------------------
 * Audio event callback 
 */
static void usb_audio_event_handler
(
    USB_DEVICE_AUDIO_V2_INDEX index,
    USB_DEVICE_AUDIO_V2_EVENT event,
    void * data,
    uintptr_t context
)
{
//    logger_log_debugf(0, "USB AUDIO usb_audio_event_handler %d %d  --", index, event);
    
    if (index == 0)
    {
        switch (event)
        {
            case USB_DEVICE_AUDIO_V2_EVENT_READ_COMPLETE:
            {
                if (usb_audio_get_active())
                {
                    //logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_EVENT_READ_COMPLETE %d %d  --", index, event);
                    //We have received an audio frame from the Host.
                    //Now send this audio frame to Audio Codec for Playback.
                    USB_DEVICE_AUDIO_V2_EVENT_DATA_READ_COMPLETE* read_event_data;
                    read_event_data = (USB_DEVICE_AUDIO_V2_EVENT_DATA_READ_COMPLETE *)data;
        //logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_EVENT_READ_COMPLETE %d %d %d  --", index, event, read_event_data->handle);

                    if (read_event_data->status == USB_DEVICE_AUDIO_V2_RESULT_OK)
                    {
                        uint32_t x;
                        for (x = 0; x < USB_AUDIO_BUFFER_QUEUE_MAX_COUNT; x++)
                        {
                            if (usb_audio_buffer_is_handle(x, read_event_data->handle))
                            {
        //logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_EVENT_READ_COMPLETE %d %d %d %d  --", index, event, read_event_data->status, x);
                                usb_audio_buffer_set_data_available(x, read_event_data->length);
                                break;
                            }
                        }
                    }
                    else
                    {
//logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_EVENT_READ_COMPLETE FAIL %d %d %d  --", index, event, read_event_data->status);
                    }
                }
                
                break;
            }

            case USB_DEVICE_AUDIO_V2_EVENT_WRITE_COMPLETE:
            {
logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_EVENT_WRITE_COMPLETE %d %d  --", index, event);
                usb_audio_feedback_complete((USB_DEVICE_AUDIO_V2_EVENT_DATA_READ_COMPLETE*)data);
                break;
            }

            case USB_DEVICE_AUDIO_V2_EVENT_INTERFACE_SETTING_CHANGED:
            {
//logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_EVENT_INTERFACE_SETTING_CHANGED %d %d  --", index, event);
                // We have received a request from USB host to change the Interface-
                // Alternate setting.
                USB_DEVICE_AUDIO_V2_EVENT_DATA_SET_ALTERNATE_INTERFACE* interface_info;
                interface_info = (USB_DEVICE_AUDIO_V2_EVENT_DATA_SET_ALTERNATE_INTERFACE *)data;
                
                usb_audio_data.active_interface_alternate_setting = interface_info->interfaceAlternateSetting;
                
                usb_audio_data.state = USB_AUDIO_STATE_INTERFACE_ALTERNATE_SETTING_RCVD;

                break;
            }
            
            case USB_DEVICE_AUDIO_V2_CUR_ENTITY_SETTINGS_RECEIVED:
            {
//logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_CUR_ENTITY_SETTINGS_RECEIVED %d %d  --", index, event);
                usb_audio_current_entity_settings_handler(index, data);
                break;
            }
            
            case USB_DEVICE_AUDIO_V2_RANGE_ENTITY_SETTINGS_RECEIVED:
            {
//logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_RANGE_ENTITY_SETTINGS_RECEIVED %d %d  --", index, event);
                usb_audio_range_entity_settings_handler(index, data);
                break;
            }
            
            case USB_DEVICE_AUDIO_V2_EVENT_CONTROL_TRANSFER_DATA_RECEIVED:
            {
logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_EVENT_CONTROL_TRANSFER_DATA_RECEIVED %d %d  --", index, event);
                usb_audio_port_send_control_status_ok();
                
                switch (usb_audio_get_current_audio_control())
                {
                    case USB_AUDIO_MUTE_CONTROL:
                    {
                        usb_audio_set_current_audio_control(USB_CONTROL_NONE);
                        usb_audio_data.state = USB_AUDIO_STATE_MUTE_AUDIO_PLAYBACK;
                        break;
                    }
                    
                    case USB_AUDIO_CLOCKSOURCE_CONTROL:
                    {
                        usb_audio_configure_sample_rate(usb_audio_data.sample_rate);
                        
                        usb_audio_set_current_audio_control(USB_CONTROL_NONE);
                        
                        usb_audio_data.state = USB_AUDIO_STATE_CLOCKSOURCE_SET;     
                        
                        break;
                    }
                    
                    case USB_AUDIO_CLOCKSELECT_CONTROL:
                    {
                        usb_audio_set_current_audio_control(USB_CONTROL_NONE);
                        break;
                    }
                    
                    case USB_CONTROL_NONE:
                    {
                        break;                    
                    }
                }

                break;
            }
            
            case  USB_DEVICE_AUDIO_V2_EVENT_CONTROL_TRANSFER_DATA_SENT:
            {
logger_log_debugf(0, "USB AUDIO - USB_DEVICE_AUDIO_V2_EVENT_CONTROL_TRANSFER_DATA_SENT %d %d  --", index, event);
                break;
            }
            
            default:
            {
                logger_log_errorf(0, "USB AUDIO Audio Event Handler Invalid callback --");
                break;
            }
        } 
    }
    
    //logger_log_debugf(0, "USB AUDIO Audio Event Handler End --");
}

/**-----------------------------------------------------------------------------
 */
static int usb_audio_clock_source_request_handler
(
    USB_DEVICE_AUDIO_V2_INDEX index ,
    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST* control_request
)
{
    USB_AUDIO_V2_CLOCKSOURCE_CONTROL_REQUEST* clock_source_request;
    clock_source_request = (USB_AUDIO_V2_CLOCKSOURCE_CONTROL_REQUEST*) control_request;

    int8_t result = 0;
   
    //logger_log_debugf(0, "USB AUDIO usb_audio_clock_source_request_handler %d %d --", clock_source_request->bRequest, clock_source_request->controlSelector);

    if (clock_source_request->bRequest == AUDIO_V2_CUR)
    {
        switch(clock_source_request->controlSelector)
        {
            case AUDIO_V2_CS_SAM_FREQ_CONTROL:
            {
                if ((control_request->bmRequestType & 0x80) == 0)
                {
                    //A control write transfer received from Host. Now receive data from Host.
                    usb_audio_port_control_receive((void *) &usb_audio_data.sample_rate, sizeof(uint32_t));
                    logger_log_debugf(0, "USB AUDIO Sample Rate Receive %d --", usb_audio_data.sample_rate);
                    usb_audio_set_current_audio_control(USB_AUDIO_CLOCKSOURCE_CONTROL);
                    
                    // Setting the sample rate was added to ensure the correct 
                    // clock/baud rate is set whenever the sample rate changes,
                    // otherwise this is not necessarily called.
                    usb_audio_codec_set_sample_rate(usb_audio_data.sample_rate);  
                }
                else
                {
                    /*Handle Get request*/
                    usb_audio_port_control_send((void *)&(usb_audio_data.sample_rate), sizeof(uint32_t) );
                    logger_log_debugf(0, "USB AUDIO Sample Rate Send %d --", usb_audio_data.sample_rate);
                    usb_audio_set_current_audio_control(USB_CONTROL_NONE);
                }
                
                break;
            }

            case AUDIO_V2_CS_CLOCK_VALID_CONTROL:
            {
                if ((control_request->bmRequestType & 0x80) == 0x80)
                {
                    /*Handle Get request*/
                    //logger_log_debugf(0, "USB AUDIO usb_audio_clock_source_request_handler %d %d --", clock_source_request->bRequest, clock_source_request->controlSelector);
                    usb_audio_port_control_send((void *)&(usb_audio_data.sample_rate_valid), sizeof(bool));
                }
                else
                {
                    usb_audio_port_send_control_status_error();
                }
                
                break;
            }

            default:
            {
                //This USB Audio Speaker application does not support any other feature unit request
                // from Host. So Stall if any other feature unit request received from Host.
                usb_audio_port_send_control_status_error();
                break;
            }

        } 
    }
    else if (clock_source_request->bRequest == AUDIO_V2_RANGE)
    {
        // This is a request for the frequency range to be sent to the USB port
        switch(clock_source_request->controlSelector)
        {
            case AUDIO_V2_CS_SAM_FREQ_CONTROL:
            {
                if ((control_request->bmRequestType & 0x80) == 0x80)
                {
                    //logger_log_debugf(0, "USB AUDIO usb_audio_clock_source_request_handler %d %d %d --", clock_source_request->bRequest, clock_source_request->controlSelector, clock_source_request->wLength);
                    //A control read transfer received from Host. Now send data to Host.
                    usb_audio_port_control_send((void *) &(usb_audio_data.sample_rate_range), clock_source_request->wLength);
                }
                else
                {
                    /*Handle Get request*/
                    usb_audio_port_send_control_status_error();
                }
            }
            break;

            default:
            {
                //This USB Audio Speaker application does not support any other feature unit request
                // from Host. So Stall if any other feature unit request received from Host.
                usb_audio_port_send_control_status_error();
                break;
            }

        } 
    }

    return (int) result;
} 

/**----------------------------------------------------------------------------
 * Configures stuff for a specific sample rate
 */
static void usb_audio_configure_sample_rate(uint32_t sample_rate)
{
    // Note this should be done before using max_audio_samples
    usb_audio_buffer_configure(sample_rate);
    usb_audio_feedback_setup_rates(sample_rate, usb_audio_buffer_get_max_samples());
}

/**-----------------------------------------------------------------------------
 */
static int usb_audio_current_entity_settings_handler
(
    USB_DEVICE_AUDIO_V2_INDEX index,
    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST *setup_packet
)
{
    int8_t result = -1;

    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST* control_request;
    control_request = (USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST*) setup_packet;
    
    switch(control_request->entityID)
    {
        case USB_AUDIO_ID_CLOCK_SOURCE:
        {
            result = usb_audio_clock_source_request_handler(index, control_request);
            break;
        }
        
        //case APP_ID_CLOCK_SELECT:
        //    err = APP_ClockSelectRequestHandler (iAudio, controlRequest);
        //    break;
        
        case USB_AUDIO_ID_INPUT_TERMINAL:
        {
            /* we need to support get CUR request */
            break;
        }
        
        case USB_AUDIO_ID_OUTPUT_TERMINAL:
        {
            /*we need to support set CUR request */
            break;
        }

        case USB_AUDIO_ID_FEATURE_UNIT:
        {
            result = usb_audio_feature_unit_request_handler(index, control_request);
            break;
        }
        
        case USB_AUDIO_ID_CLOCK_SELECT:        
        case USB_AUDIO_ID_MIXER_UNIT:
        case USB_AUDIO_ID_SELECTOR_UNIT:
        case USB_AUDIO_ID_PROCESSING_UNIT:
        case USB_AUDIO_ID_EXTENSION_UNIT:
        default:
        {
            //This USB Audio Speaker application does not support any other control request
            // received for other unit from Host. So Stall if any other request received from Host.
            usb_audio_port_send_control_status_error();
            break;
        }
    }
    
    return (int) result;
}

/**-----------------------------------------------------------------------------
 * Callback for device events
 * @param event
 * @param event_data
 * @param context
 */
static void usb_audio_device_handler(USB_DEVICE_EVENT event, void* event_data, uintptr_t context)
{
    logger_log_debugf(0, "USB AUDIO usb_audio_device_handler %d --", event);
    
    switch( event )
    {
        case USB_DEVICE_EVENT_RESET:    
        {
            logger_log_debugf(0, "USB AUDIO USB_DEVICE_EVENT_RESET %d --", event);
            // Reset the read error count
            usb_audio_port_reset();
            
            // Mute the Audio
            volume_mute();
            //output_control_mute_outputs(true);
            break;
        }
        
        case USB_DEVICE_EVENT_DECONFIGURED:
        {
            logger_log_debugf(0, "USB AUDIO USB_DEVICE_EVENT_DECONFIGURED %d --", event);
            // USB device is reset or device is de-configured.
            // This means that USB device layer is about to de-initialise
            // all function drivers. So close handles to previously opened
            // function drivers.            
            break;
        }

        case USB_DEVICE_EVENT_CONFIGURED:
        {
            logger_log_debugf(0, "USB AUDIO - USB_DEVICE_EVENT_CONFIGURED %d --", event);

            /* check the configuration */
            /* Initialise the Application */
            if(*(uint8_t*)event_data == 1)
            {
                USB_DEVICE_AUDIO_V2_EventHandlerSet
                (
                    0,
                    usb_audio_event_handler,
                    (uintptr_t)NULL
                );

                /* Set configuration is complete */
                usb_audio_set_configured(true);
            }

            // Un-mute the CODEC
            volume_unmute();
            //output_control_unmute_outputs();

            break;
        }

        case USB_DEVICE_EVENT_SUSPENDED:
        {
            logger_log_debugf(0, "USB AUDIO USB_DEVICE_EVENT_SUSPENDED %d --", event);
            // called when USB cable is unplugged
            if(usb_audio_get_configured())
            {
                usb_audio_set_configured(false);                    
            
                // called when USB cable is unplugged
                volume_mute();
                //output_control_mute_outputs(true);
            
                usb_audio_initialise_data();            

                usb_audio_data.state = USB_AUDIO_STATE_IDLE;
            }
            
            break;
        }
        
        case USB_DEVICE_EVENT_POWER_DETECTED:
        {
            //Called when the USB cable is plugged in
            logger_log_debugf(0, "USB AUDIO USB_DEVICE_EVENT_POWER_DETECTED %d --", event);
            
            /* Attach the device */
            usb_audio_port_attach();
            
            break;
        }

        case USB_DEVICE_EVENT_POWER_REMOVED:
        {
            logger_log_debugf(0, "USB AUDIO USB_DEVICE_EVENT_POWER_REMOVED %d --", event);
            /* VBUS is not available. We can detach the device */
            usb_audio_port_detach();
            usb_audio_set_configured(false);  
            break;
        }

        case USB_DEVICE_EVENT_RESUMED:
            logger_log_debugf(0, "USB AUDIO USB_DEVICE_EVENT_RESUMED %d --", event);
            break;
            
        case USB_DEVICE_EVENT_ERROR:
            logger_log_debugf(0, "USB AUDIO USB_DEVICE_EVENT_ERROR %d --", event);
            break;
            
        default:
            //logger_log_debugf(0, "USB AUDIO DEFAULT %d --", event);
            break;
    }
    
    logger_log_debugf(0, "usb_audio_device_handler %d END --", event);
}

/**-----------------------------------------------------------------------------
 * General error recovery
 */
static void usb_audio_error_recovery(uint8_t level)
{
    //logger_log_errorf(0, "USB AUDIO Error Recovery %d ", level);
    
    volume_mute();
    //output_control_mute_outputs(true);

    usb_audio_initialise_data();
    
    if (level > USB_AUDIO_LEVEL1_ERROR)
    {
        usb_audio_port_detach();                       
        usb_audio_port_attach();                            
        usb_audio_port_reset();
        usb_audio_set_configured(false);        
        
        usb_audio_data.state = USB_AUDIO_STATE_WAIT_CONFIGURATION;
    }
    else
    {
        usb_audio_data.state = USB_AUDIO_STATE_SUBMIT_READ_REQUEST;
    }
}

/**-----------------------------------------------------------------------------
 */
static int usb_audio_feature_unit_request_handler
(
    USB_DEVICE_AUDIO_V2_INDEX index ,
    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST* control_request
)
{
    USB_AUDIO_V2_FEATURE_CONTROL_REQUEST* feature_request;
    feature_request = (USB_AUDIO_V2_FEATURE_CONTROL_REQUEST*) control_request;

    int8_t result = 0;

    switch(feature_request->controlSelector)
    {
        case USB_AUDIO_V2_FCS_MUTE_CONTROL:
        {
            if ((control_request->bmRequestType & 0x80) == 0)
            {
                //A control write transfer received from Host. Now receive data from Host.
                usb_audio_port_control_receive((void *) &(usb_audio_data.is_muted), sizeof(bool));
                usb_audio_set_current_audio_control(USB_AUDIO_MUTE_CONTROL);
            }
            else
            {
                /*Handle Get request*/
                usb_audio_port_control_send((void *)&(usb_audio_data.is_muted), sizeof(bool));
            }
            break;
        }

        case USB_AUDIO_V2_FCS_VOLUME_CONTROL:
        case USB_AUDIO_V2_FCS_BASS_CONTROL:
        case USB_AUDIO_V2_FCS_MID_CONTROL:
        case USB_AUDIO_V2_FCS_TREBLE_CONTROL:
        case USB_AUDIO_V2_FCS_GRAPHIC_EQUALIZER_CONTROL:
        case USB_AUDIO_V2_FCS_AUTOMATIC_GAIN_CONTROL:
        case USB_AUDIO_V2_FCS_DELAY_CONTROL:
        case USB_AUDIO_V2_FCS_BASS_BOOST_CONTROL:
        case USB_AUDIO_V2_FCS_LOUDNESS_CONTROL :
        case USB_AUDIO_V2_FCS_FU_CONTROL_UNDEFINED:
        default:
        {
            //This USB Audio Speaker application does not support any other feature unit request
            // from Host. So Stall if any other feature unit request received from Host.
            usb_audio_port_send_control_status_error();
            break;
        }
    } 
    
    return result;
} 

/**-----------------------------------------------------------------------------
 * 
 * @return 
 */
static USB_AUDIO_CONTROLS usb_audio_get_current_audio_control()
{
    return usb_audio_data.current_audio_control;
}

/**-----------------------------------------------------------------------------
 * Initialise data
 */
static void usb_audio_initialise_data()
{
//TODO                DRV_AK4384_BufferQueueFlush(usb_audio_data.codec_data.handle);                     

    usb_audio_data.initialise_read_request = true;
    //usb_audio_data.codec_data.next_buffer = 0;           
    usb_audio_data.lr_clock_sync = USB_AUDIO_LRCLOCKSYNC;

    /* Initializing appData members*/
    //TODOcodec_initialise_write_queue(AUDIO_FRAME_SIZE_UAC2);
    
    usb_audio_initialise_audio_buffer();
}

/**-----------------------------------------------------------------------------
 * Initialises all the buffers in the read queue
 */
static void usb_audio_initialise_audio_buffer()
{
    usb_audio_buffer_clear();
    
    //usb_audio_data.read_queue_idx = 0;
    
    usb_audio_port_reset();
    usb_audio_codec_reset();
}

/**-----------------------------------------------------------------------------
 */
static int usb_audio_range_entity_settings_handler
(
    USB_DEVICE_AUDIO_V2_INDEX index ,
    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST *setup_packet
)
{
    int8_t result = -1;
    USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST* control_request;
    control_request = (USB_AUDIO_V2_CONTROL_INTERFACE_REQUEST*) setup_packet;
    
    //logger_log_debugf(0, "USB AUDIO usb_audio_range_entity_settings_handler %d --", control_request->entityID);
    
    switch(control_request->entityID)
    {
        case USB_AUDIO_ID_CLOCK_SOURCE:
        {
            result = usb_audio_clock_source_request_handler(index, control_request);
            break;
        }
        
        //case APP_ID_CLOCK_SELECT:
        //    err = usb_audio_clock_source_request_handler (index, control_request);
        //    break;
        
        case USB_AUDIO_ID_INPUT_TERMINAL:
        {
            /* we need to support get CUR request */
            break;
        }

        case USB_AUDIO_ID_OUTPUT_TERMINAL:
        {
            /*we need to support set CUR request */
            break;
        }

        case USB_AUDIO_ID_FEATURE_UNIT:
        {
            result = usb_audio_feature_unit_request_handler(index, control_request);
            break;
        }
        
        case USB_AUDIO_ID_CLOCK_SELECT:        
        case USB_AUDIO_ID_MIXER_UNIT:
        case USB_AUDIO_ID_SELECTOR_UNIT:
        case USB_AUDIO_ID_PROCESSING_UNIT:
        case USB_AUDIO_ID_EXTENSION_UNIT:
        default:
        {
            //This USB Audio Speaker application does not support any other control request
            // received for other unit from Host. So Stall if any other request received from Host.
            usb_audio_port_send_control_status_error();
            break;
        }
    }
    
    return (int) result;
}

/**-----------------------------------------------------------------------------
 * Moves the read_queue index on..
 */
//static void usb_audio_select_next_read_buffer()
//{
//    usb_audio_data.read_queue_idx += 1;
//    usb_audio_data.read_queue_idx %= USB_AUDIO_BUFFER_QUEUE_MAX_COUNT;
//}

/**-----------------------------------------------------------------------------
 * Moves the read_queue index on..
 */
//static void usb_audio_select_next_write_buffer()
//{
//    usb_audio_data.codec_data.next_buffer++;
//    usb_audio_data.codec_data.next_buffer %= USB_AUDIO_BUFFER_QUEUE_MAX_COUNT;
//}

/**-----------------------------------------------------------------------------
 * Sets USB Audio configured state
 */
static void usb_audio_set_configured(bool value)
{
    usb_audio_data.configured = value;
}

/**-----------------------------------------------------------------------------
 * 
 * @param 
 */
static void usb_audio_set_current_audio_control(USB_AUDIO_CONTROLS control)
{
    usb_audio_data.current_audio_control = control;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Returns a value identifying whether USB Audio is active or not;
 * @return 
 */
bool usb_audio_get_active()
{
    //return true;
    return usb_audio_data.active;
}

/**-----------------------------------------------------------------------------
 * Gets USB Audio configured state
 */
bool usb_audio_get_configured()
{
    return usb_audio_data.configured;
}

/**-----------------------------------------------------------------------------
 * Initialise the usb_audio module
 */
bool usb_audio_initialise()
{
    logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO Initialise ");
    bool result = false;
    
    if (usb_audio_port_initialise(usb_audio_device_handler, 
            usb_audio_event_handler))
    {
        result = usb_audio_codec_initialise();
    }
    
    if (result)
    {
        usb_audio_data.state = USB_AUDIO_STATE_OPEN_CODEC;
    
        logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO Initialise END ");
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_DEBUG, "USB AUDIO Initialise ERRORS ");
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Perform USB Audio processing
 */
void usb_audio_process()
{
    switch (usb_audio_data.state)
    {
        case USB_AUDIO_STATE_INITIALISE:
        {
            break;
        }
        
        case USB_AUDIO_STATE_OPEN_CODEC:
        {
            // Note: We Open the Codec /I2S before the USB Port as setting the 
            // event handler on the port kicks off a load of interrupts which
            // cause the CODEC initialisation to be missed..
            if (usb_audio_codec_open())
            {
                logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_OPEN_CODEC --");
                usb_audio_data.state =  USB_AUDIO_STATE_OPEN_PORT;
            }
            break;
        }
        
        case USB_AUDIO_STATE_OPEN_PORT:
        {
            if (usb_audio_port_open())
            {
                logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_OPEN_PORT --");
                usb_audio_data.state =  USB_AUDIO_STATE_WAIT_CONFIGURATION;
            }
            break;
        }
        
        case USB_AUDIO_STATE_WAIT_CONFIGURATION:
        {
            // Holding state until configuration is complete.  That is triggered by
            // usb_audio_initialise.
            break;
        }
        
        case USB_AUDIO_STATE_IDLE:
        {
            // Holding state while there is no Audio.
            break;
        }
        
        case USB_AUDIO_STATE_INTERFACE_ALTERNATE_SETTING_RCVD:
        {
            logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_INTERFACE_ALTERNATE_SETTING_RCVD %d --", usb_audio_data.active_interface_alternate_setting);
            if (usb_audio_data.active_interface_alternate_setting == USB_SPEAKER_PLAYBACK_NONE)
            {
                usb_audio_initialise_data();
                
                usb_audio_data.state = USB_AUDIO_STATE_IDLE;
            }
            else if (usb_audio_data.active_interface_alternate_setting == USB_SPEAKER_PLAYBACK_STEREO)
            {                 
                // Un-mute output
                volume_unmute();
                //output_control_unmute_outputs();
                
                usb_audio_data.state =  USB_AUDIO_STATE_SUBMIT_READ_REQUEST;
            }
            
            break;
        }
        
        case USB_AUDIO_STATE_MUTE_AUDIO_PLAYBACK:
        {
            // Note this is a toggle
            if (usb_audio_data.active_interface_alternate_setting == USB_SPEAKER_PLAYBACK_NONE) 
            {
                usb_audio_data.state = USB_AUDIO_STATE_IDLE;
            } 
            else if (usb_audio_data.active_interface_alternate_setting == USB_SPEAKER_PLAYBACK_STEREO) 
            {
                usb_audio_data.state = USB_AUDIO_STATE_DAC_MUTE;
            }                 
            break;
        }
        
        case USB_AUDIO_STATE_DAC_MUTE:
        {
            if (usb_audio_get_configured()) 
            {
                if (!usb_audio_data.is_muted) 
                {
                    usb_audio_data.is_muted = true;

                    volume_mute();
                    //output_control_mute_outputs(true);
                    
                    usb_audio_initialise_data();
                    
                    usb_audio_data.state = USB_AUDIO_STATE_IDLE;                                                         
                } 
                else 
                {
                    usb_audio_data.is_muted = false;        
                    usb_audio_data.lr_clock_sync = USB_AUDIO_LRCLOCKSYNC;
                }
                
                // Work out where to go next
                if ((usb_audio_data.active_interface_alternate_setting == USB_SPEAKER_PLAYBACK_NONE)||
                    (usb_audio_data.is_muted))
                {
                    usb_audio_data.state = USB_AUDIO_STATE_IDLE;
                }
                else
                {
                    usb_audio_data.state = USB_AUDIO_STATE_PROCESS_DATA;                     
                }              
            }   
            else
            {
                usb_audio_data.state = USB_AUDIO_STATE_WAIT_CONFIGURATION;
            }
            
            break;
        }
        
        case USB_AUDIO_STATE_CLOCKSOURCE_SET:
        {
        logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_CLOCKSOURCE_SET ");
            // The sample rate has changed
            if (usb_audio_data.active_interface_alternate_setting == USB_SPEAKER_PLAYBACK_NONE)
            {
                usb_audio_codec_set_sample_rate(usb_audio_data.sample_rate);

                if(usb_audio_data.last_sample_rate!= usb_audio_data.sample_rate)
                {
                    usb_audio_data.last_sample_rate = usb_audio_data.sample_rate;  // changed
                    
                    // these two lines commented out because we assume for now user 
                    // has to pause the audio before changing sample rates
                    //muteAudio();
                    usb_audio_data.lr_clock_sync = USB_AUDIO_LRCLOCKSYNC;      // reset LRCLK
                }
            }

            usb_audio_data.state = USB_AUDIO_STATE_IDLE;            
            
            break;
        }
        
        case USB_AUDIO_STATE_SUBMIT_READ_REQUEST:
        {
            if (usb_audio_get_active())
            {
                //logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_SUBMIT_READ_REQUEST --");
                if (usb_audio_data.initialise_read_request)
                {    
                logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_SUBMIT_READ_REQUEST --");
                    usb_audio_data.initialise_read_request = false;
                    //usb_audio_initialise_data();
                    
                    volume_mute();
                    //output_control_mute_outputs(true);

                    // Initiate reading data from the USB Host
                    //usb_audio_port_queue_read();
                    usb_audio_port_read_buffers();
                    if (usb_audio_port_error_limit_exceeded())
                    {
                        usb_audio_error_recovery(USB_AUDIO_LEVEL2_ERROR);
                    }

                    usb_audio_feedback_set_rate(USB_AUDIO_FEEDBACK_STEADY);
                }

                // Set the speed of processing
                usb_audio_feedback_send_rate_request();

                // For some reason this enables then disables interrupts
                uint32_t unprocessed_reads = 0;
                bool int_status = SYS_INT_Disable();
                {
                    // Get the number of unprocessed read buffers
                    unprocessed_reads = usb_audio_buffer_get_unprocessed_reads();
                }

                if (int_status)
                {
                    SYS_INT_Enable();
                }

                // Process the unprocessed reads
                if (unprocessed_reads > 0)
                {
                    // Unmute the outputs
                    volume_unmute();
                    //output_control_unmute_outputs();

                    // Start sending the read buffers to the codec, send any that are ready
    //                usb_audio_codec_try_write();
                    int x;
                    for (x = 0; x < USB_AUDIO_BUFFER_QUEUE_MAX_COUNT - 100; x++)
                    {
                        if (!usb_audio_codec_try_write())
                        {
                            break;
                        }
                    }

                    usb_audio_data.state = USB_AUDIO_STATE_PROCESS_DATA;
                }
            }
            
            break;
        }
        
        case USB_AUDIO_STATE_PROCESS_DATA:
        {
            if (usb_audio_get_active())
            {
                ///logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_PROCESS_DATA ");

                // Check how were doing re read/write speeds
                // Get the size of the available buffers
                size_t buffer_size = usb_audio_codec_get_buffer_size();
                //logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_PROCESS_DATA %d %d --", buffer_size, usb_audio_buffer_get_max_size());
                if (buffer_size >= usb_audio_buffer_get_max_size())           
                {
                    /*Slow down the host data rate*/      
                    usb_audio_feedback_set_rate(USB_AUDIO_FEEDBACK_SLOWDOWN);
                }                
                else if (buffer_size < usb_audio_buffer_get_min_size())                
                {
                    /*Speed up the host data rate*/
                    usb_audio_feedback_set_rate(USB_AUDIO_FEEDBACK_SPEEDUP);
                }
                else
                {
                    usb_audio_feedback_set_rate(USB_AUDIO_FEEDBACK_STEADY);
                }

                // Write to CODEC
                usb_audio_codec_try_write();
//                int x;
//                for (x = 0; x < USB_AUDIO_BUFFER_QUEUE_MAX_COUNT - 16; x++)
//                {
//                    if (!usb_audio_codec_try_write())
//                    {
//                        break;
//                    }
//                }

//                if (usb_audio_port_get_write_index() == usb_audio_codec_get_read_index())
//                {
//                    logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO MATCH %d %d %d  --", usb_audio_port_get_write_index() , usb_audio_codec_get_read_index(), buffer_size);
//                    if (buffer_size == 0)
//                    {
//                        /* Initialising appData members*/
//                        usb_audio_initialise_data();
//                    }
//                }

    //            // Check if buffer is currently in use - if not read data into it
    //logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO USB_AUDIO_STATE_PROCESS_DATA - Read Buffer --");
                USB_DEVICE_AUDIO_V2_RESULT audioErr = usb_audio_port_queue_read();
                if (audioErr != USB_DEVICE_AUDIO_V2_RESULT_OK)
                {
                    if (usb_audio_port_error_limit_exceeded())
                    {
                        usb_audio_error_recovery(USB_AUDIO_LEVEL2_ERROR);
                    }
                }

                usb_audio_feedback_send_rate_request();
            }
            
            break;
        }
        
        default:
            break;
    }
}
     
/**-----------------------------------------------------------------------------
 * Allows the code to effectively turn USB processing On/Off.  Because it is 
 * processor intensive other activity can cause packets to be dropped and for 
 * the USB Audio module to become stuck.
 * @param active
 */
void usb_audio_set_active(bool active)
{
    logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO Setting ACTIVE %s --", active ? "On" : "Off");
    usb_audio_data.active = active;
}
    
/**-----------------------------------------------------------------------------
 * This function is called from the DMA Channel 0 interrupt handler in system_interrupt.c
 * to synchronise up the I2S LRCLK when the audio stream is restarted.  Otherwise, the left/right
 * channels may become swapped.
 */
void usb_audio_sync_LR_clock()
{
    if (usb_audio_data.lr_clock_sync > 0)
    {       
        if (usb_audio_data.lr_clock_sync == 1) // = 1, therefore will be decremented to 0 before leaving function
        {
            // turn off SPI interface
            PLIB_SPI_Disable(DRV_I2S_PERIPHERAL_ID_IDX0);       
            
            if (usb_audio_data.sample_rate <= USB_AUDIO_SAMPLING_RATE_48KHZ)
            {
                PLIB_SPI_FrameSyncPulsePolaritySelect(DRV_I2S_PERIPHERAL_ID_IDX0, 1);   // invert polarity, otherwise L/R are always backwards           
            }
            else
            {
                PLIB_SPI_FrameSyncPulsePolaritySelect(DRV_I2S_PERIPHERAL_ID_IDX0, 0);   // regular polarity for higher bitrates
            }
            
            // turn SPI interface back on again (restarts LRCLK)
            PLIB_SPI_Enable(DRV_I2S_PERIPHERAL_ID_IDX0);                 
            
            volume_mute();
            //output_control_mute_outputs(true);
        }

        usb_audio_data.lr_clock_sync--; // seems to work better if we do this more than once, # is defined as LRCLOCKSYNC
    }
}

/* *****************************************************************************
 End of File
 */
