/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\usb_audio_buffer.h"
#include "utilities\usb_audio_defs.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Holds the data that is transferred between the USB port and the CODEC
 *  Application USB transfer client
 *  This object holds the USB queued handles handle and state of its respective buffer
 *  created and the context
*/
typedef struct
{
    //! A handle to identify the read instruction
    USB_DEVICE_AUDIO_V2_TRANSFER_HANDLE handle;

    //! The actual data read
    uint8_t __attribute__((coherent)) __attribute__((aligned(16))) buffer[NO_OF_SAMPLES_IN_A_USB_FRAME];        

    //! True if this buffer is being used to hold a current read request
    volatile bool data_available;
    
    //! True if this buffer has a request in progress
    volatile bool data_queued;

    //! The length of the data read ( < NO_OF_SAMPLES_IN_A_USB_FRAME)
    uint16_t data_length;

} USB_AUDIO_BUFFER_QUEUE;

/**
 * An Audio buffer to hold data either read from the USB port or from the I2S port
 */
typedef struct
{
    // A collection of queues
    USB_AUDIO_BUFFER_QUEUE queue[USB_AUDIO_BUFFER_QUEUE_MAX_COUNT];
} USB_AUDIO_BUFFER;
    
/**
 * Main data structure
 */
typedef struct
{
    // Buffer size limits
    uint32_t buffer_max_size;
    uint32_t buffer_min_size;

    //! Maximum number of sample that can be stored
    uint32_t max_audio_samples;
    
    //! Queue to hold read data.
    USB_AUDIO_BUFFER audio_buffer;
    
    //! The number of USB reads that have not been processed
    volatile uint32_t unprocessed_reads;
    
    uint32_t count;
    
} USB_AUDIO_DATA;

static USB_AUDIO_DATA usb_audio_data =
{
    .unprocessed_reads = 0,
    .count = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Clears down the buffer
 */
void usb_audio_buffer_clear()
{
    int x;
    for (x = 0; x < USB_AUDIO_BUFFER_QUEUE_MAX_COUNT; x++)
    {
//logger_log_debugf(0, "USB AUDIO - usb_audio_buffer_clear %d --", x);
        memset(&usb_audio_data.audio_buffer.queue[x].buffer, 0, sizeof(usb_audio_data.audio_buffer.queue[x].buffer));
        usb_audio_data.audio_buffer.queue[x].data_available = false;
        usb_audio_data.audio_buffer.queue[x].data_queued = false;
        usb_audio_data.audio_buffer.queue[x].data_length = 0;
        usb_audio_data.audio_buffer.queue[x].handle = USB_DEVICE_AUDIO_V2_TRANSFER_HANDLE_INVALID;
    } 
    
    usb_audio_data.unprocessed_reads = 0;
}

/**-----------------------------------------------------------------------------
 * Used to indicate that USB read queue data is available
 */
void usb_audio_buffer_clear_data_available(uint32_t index)
{
    usb_audio_data.audio_buffer.queue[index].data_available = false;
    usb_audio_data.audio_buffer.queue[index].data_queued = false;
    usb_audio_data.audio_buffer.queue[index].data_length = 0;
    usb_audio_data.unprocessed_reads--;  
   
    if ((usb_audio_data.count % 1000) == 0)
    {
        logger_log_debugf(0, "USB AUDIO - usb_audio_buffer_clear_data_available %d %d %d --", index, usb_audio_data.unprocessed_reads, usb_audio_data.count);
    }
}

/**----------------------------------------------------------------------------
 * Configures the buffer
 */
void usb_audio_buffer_configure(uint32_t sample_rate)
{
    usb_audio_data.max_audio_samples = sample_rate / 1000;
    
    /*Generating Feedback Speedup data rate */                           
    if (USB_AUDIO_SAMPLING_RATE_32KHZ == sample_rate)
    {
        usb_audio_data.buffer_max_size = USB_AUDIO_BUFFER_QUEUE_TUNE_UPPER_LIMIT_32;
        usb_audio_data.buffer_min_size = USB_AUDIO_BUFFER_QUEUE_TUNE_LOWER_LIMIT_32;
    }
    else if (USB_AUDIO_SAMPLING_RATE_44_1KHZ == sample_rate)
    {
        usb_audio_data.buffer_max_size = USB_AUDIO_BUFFER_QUEUE_TUNE_UPPER_LIMIT_44_1;
        usb_audio_data.buffer_min_size = USB_AUDIO_BUFFER_QUEUE_TUNE_LOWER_LIMIT_44_1;                                                                
    }
    else if (USB_AUDIO_SAMPLING_RATE_48KHZ == sample_rate)
    {
        usb_audio_data.buffer_max_size = USB_AUDIO_BUFFER_QUEUE_TUNE_UPPER_LIMIT_48;
        usb_audio_data.buffer_min_size = USB_AUDIO_BUFFER_QUEUE_TUNE_LOWER_LIMIT_48;
    }                        
    else if (USB_AUDIO_SAMPLING_RATE_88_2KHZ == sample_rate)
    {
        usb_audio_data.buffer_max_size = USB_AUDIO_BUFFER_QUEUE_TUNE_UPPER_LIMIT_88_2;
        usb_audio_data.buffer_min_size = USB_AUDIO_BUFFER_QUEUE_TUNE_LOWER_LIMIT_88_2;
    }   
    else if (USB_AUDIO_SAMPLING_RATE_96KHZ == sample_rate)
    {
        usb_audio_data.buffer_max_size = USB_AUDIO_BUFFER_QUEUE_TUNE_UPPER_LIMIT_96;
        usb_audio_data.buffer_min_size = USB_AUDIO_BUFFER_QUEUE_TUNE_LOWER_LIMIT_96;
    }  
    else if (USB_AUDIO_SAMPLING_RATE_176_4KHZ == sample_rate)
    {
        usb_audio_data.buffer_max_size = USB_AUDIO_BUFFER_QUEUE_TUNE_UPPER_LIMIT_176_4;
        usb_audio_data.buffer_min_size = USB_AUDIO_BUFFER_QUEUE_TUNE_LOWER_LIMIT_176_4;
    }                          
    else if (USB_AUDIO_SAMPLING_RATE_192KHZ == sample_rate)
    {
        usb_audio_data.buffer_max_size = USB_AUDIO_BUFFER_QUEUE_TUNE_UPPER_LIMIT_192;
        usb_audio_data.buffer_min_size = USB_AUDIO_BUFFER_QUEUE_TUNE_LOWER_LIMIT_192;                                                             
    }
    else
    {
        /* Unsupported Sampling rate tuning */
    }
}

/**-----------------------------------------------------------------------------
 * Determines whether the read queue has unread data available
 */
bool usb_audio_buffer_data_available(uint32_t index)
{
    return usb_audio_data.audio_buffer.queue[index].data_available;
    //return usb_audio_data.audio_buffer.queue[index].data_length != 0;
}

/**-----------------------------------------------------------------------------
 * Determines whether the read queue has unread data available
 */
bool usb_audio_buffer_data_clear(uint32_t index)
{
    return (!usb_audio_data.audio_buffer.queue[index].data_queued && 
            !usb_audio_data.audio_buffer.queue[index].data_available);
    //return usb_audio_data.audio_buffer.queue[index].data_length != 0;
}

/**-----------------------------------------------------------------------------
 * Gets the length of the available data in the buffer
 */
uint16_t usb_audio_buffer_get_data_length(uint32_t index)
{
    return usb_audio_data.audio_buffer.queue[index].data_length;
}

/**-----------------------------------------------------------------------------
 * Returns the maximum number of samples that can be used.  Initialised by the
 * usb_audio_buffer_configure function
 */
uint32_t usb_audio_buffer_get_max_samples()
{
    return usb_audio_data.max_audio_samples;
}

/**-----------------------------------------------------------------------------
 * Gets the maximum buffer size
 * @return 
 */
uint32_t usb_audio_buffer_get_max_size()
{
    return usb_audio_data.buffer_max_size * usb_audio_data.max_audio_samples;
}

/**-----------------------------------------------------------------------------
 * Gets the minimum buffer size
 * @return 
 */
uint32_t usb_audio_buffer_get_min_size()
{
    return usb_audio_data.buffer_min_size * usb_audio_data.max_audio_samples;
}

/**-----------------------------------------------------------------------------
 * Gets a pointer to the read handle
 */
USB_DEVICE_AUDIO_V2_TRANSFER_HANDLE* usb_audio_buffer_get_read_handle(uint32_t index)
{
    return &usb_audio_data.audio_buffer.queue[index].handle;
}

/**-----------------------------------------------------------------------------
 * Gets a pointer to the read buffer
 */
uint8_t* usb_audio_buffer_get_read_buffer(uint32_t index)
{
    return &usb_audio_data.audio_buffer.queue[index].buffer[0];
}

/**-----------------------------------------------------------------------------
 * Gets the number of unprocessed reads
 * @return 
 */
uint32_t usb_audio_buffer_get_unprocessed_reads()
{
    return usb_audio_data.unprocessed_reads;
}

/**-----------------------------------------------------------------------------
 * Checks whether the supplied handle is the same as the one for the supplied 
 * buffer index
 */
bool usb_audio_buffer_is_handle(uint32_t index, USB_DEVICE_AUDIO_V2_TRANSFER_HANDLE handle)
{
//if (handle == usb_audio_data.audio_buffer.queue[index].handle)
//{    
 logger_log_debugf(0, "USB AUDIO - usb_audio_buffer_is_handle %d %d --", handle, usb_audio_data.audio_buffer.queue[index].handle);
//}
    return handle == usb_audio_data.audio_buffer.queue[index].handle;
}

/**-----------------------------------------------------------------------------
 * Used to indicate that USB read queue data is available
 */
void usb_audio_buffer_set_data_available(uint32_t index, uint16_t length)
{
    usb_audio_data.audio_buffer.queue[index].data_available = true;
    usb_audio_data.audio_buffer.queue[index].data_queued = false;
    usb_audio_data.audio_buffer.queue[index].data_length = length;
    usb_audio_data.unprocessed_reads++; 
    
    usb_audio_data.count++;
    
    //timers_delay_ms(10);//
    
//    if ((usb_audio_data.count % 100) == 0)
//    {
//        logger_log_debugf(0, "USB AUDIO - usb_audio_buffer_set_data_available %d %d %d --", index, usb_audio_data.unprocessed_reads, usb_audio_data.count);
//    }
}

/**-----------------------------------------------------------------------------
 * Used to indicate that USB read queue data is available
 */
void usb_audio_buffer_set_data_queued(uint32_t index)
{
    usb_audio_data.audio_buffer.queue[index].data_queued = true;
}
    
/* *****************************************************************************
 End of File
 */
