/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/usb_audio_feedback.h"
#include "system_definitions.h"
#include "usb/usb_audio_v2_0.h"
#include "usb/usb_device_audio_v2_0.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */
    
#define USB_DATA_PACKET_SIZE                                8       

#define USB_FEEDBACK_32_NORMAL                              0x00   
#define USB_FEEDBACK_32_SPEED_UP                            0x00001500
#define USB_FEEDBACK_32_SLOW_DOWN                           0x00

#define USB_FEEDBACK_44_1_NORMAL                            0x00   
#define USB_FEEDBACK_44_1_SPEED_UP                          0x00009450
#define USB_FEEDBACK_44_1_SLOW_DOWN                         0x00    

#define USB_FEEDBACK_48_NORMAL                              0x00   
#define USB_FEEDBACK_48_SPEED_UP                            0x00002500
#define USB_FEEDBACK_48_SLOW_DOWN                           0x00   

#define USB_FEEDBACK_88_2_NORMAL                            0x00   
#define USB_FEEDBACK_88_2_SPEED_UP                          0x00002500
#define USB_FEEDBACK_88_2_SLOW_DOWN                         0x00        

#define USB_FEEDBACK_96_NORMAL                              0x00   
#define USB_FEEDBACK_96_SPEED_UP                            0x00002500
#define USB_FEEDBACK_96_SLOW_DOWN                           0x00     

#define USB_FEEDBACK_176_4_NORMAL                           0x00   
#define USB_FEEDBACK_176_4_SPEED_UP                         0x00008450
#define USB_FEEDBACK_176_4_SLOW_DOWN                        0x00            

#define USB_FEEDBACK_192_NORMAL                             0x00   
#define USB_FEEDBACK_192_SPEED_UP                           0x00008450
#define USB_FEEDBACK_192_SLOW_DOWN                          0x00    

/**
 */
typedef struct
{
    //! Values used to change USB read rate
    uint8_t feedback_steady[4];
    uint8_t feedback_speedup[4];
    uint8_t feedback_slowdown[4];  

    uint8_t __attribute__((coherent)) __attribute__((aligned(16))) feedback_buffer[4];

    //! Tests whether transfer is complete
    bool write_complete;
    
    // USB Write handle
    USB_DEVICE_AUDIO_V2_TRANSFER_HANDLE write_handle;

} USB_AUDIO_FEEDBACK_DATA;

static USB_AUDIO_FEEDBACK_DATA usb_audio_data = 
{
    .write_complete = false,
    .write_handle = USB_DEVICE_AUDIO_V2_TRANSFER_HANDLE_INVALID,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Called in response to a write complete event
 * @param data
 */
void usb_audio_feedback_complete(USB_DEVICE_AUDIO_V2_EVENT_DATA_READ_COMPLETE* data)
{
    if (data->handle == usb_audio_data.write_handle)
    {
        usb_audio_data.write_complete = true;
    }
}

/**-----------------------------------------------------------------------------
 * Sends a request to the USB host to adjust the transfer rate
 */
void usb_audio_feedback_send_rate_request()
{
    if (usb_audio_data.write_complete == true)
    {                      
        usb_audio_data.write_complete = false;
        USB_DEVICE_AUDIO_V2_Write(USB_DEVICE_INDEX_0 , &usb_audio_data.write_handle, 1 , usb_audio_data.feedback_buffer, 4 );
    }
}

/**-----------------------------------------------------------------------------
 * Sets up feedback rates and buffer limits
 */
void usb_audio_feedback_setup_rates(uint32_t sample_rate, uint32_t max_audio_samples)
{
    uint32_t usb_feedback_normal;
    uint32_t usb_feedback_min;  
    
    usb_feedback_normal = ((max_audio_samples / USB_DATA_PACKET_SIZE) << 16);
    usb_feedback_min = (((max_audio_samples / USB_DATA_PACKET_SIZE) - 1) << 16);    

    /*Generating Feedback normal data rate */
    
    usb_audio_data.feedback_steady[0] = usb_feedback_normal & 0xFF;
    usb_audio_data.feedback_steady[1] = (usb_feedback_normal >> 8) & 0xFF; 
    usb_audio_data.feedback_steady[2] = (usb_feedback_normal >> 16) & 0xFF; 
    usb_audio_data.feedback_steady[3] = (usb_feedback_normal >> 24) & 0xFF;                                  

    /*Generating Feedback Slow down data rate */                        
    usb_feedback_min = usb_feedback_min | 0x00008000;       
    usb_audio_data.feedback_slowdown[0] = usb_feedback_min & 0xFF;
    usb_audio_data.feedback_slowdown[1] = (usb_feedback_min >> 8) & 0xFF;
    usb_audio_data.feedback_slowdown[2] = (usb_feedback_min >> 16) & 0xFF; 
    usb_audio_data.feedback_slowdown[3] = (usb_feedback_min >> 24) & 0xFF;                  

    /*Generating Feedback Speedup data rate */                           
    if (USB_AUDIO_SAMPLING_RATE_32KHZ == sample_rate)
    {
        usb_feedback_normal = usb_feedback_normal | USB_FEEDBACK_32_SPEED_UP; 
    }
    else if (USB_AUDIO_SAMPLING_RATE_44_1KHZ == sample_rate)
    {
        usb_feedback_normal = usb_feedback_normal | USB_FEEDBACK_44_1_SPEED_UP;                                                                                                                
    }
    else if (USB_AUDIO_SAMPLING_RATE_48KHZ == sample_rate)
    {
        usb_feedback_normal = usb_feedback_normal | USB_FEEDBACK_48_SPEED_UP;
    }                        
    else if (USB_AUDIO_SAMPLING_RATE_88_2KHZ == sample_rate)
    {
        usb_feedback_normal = usb_feedback_normal | USB_FEEDBACK_88_2_SPEED_UP;
    }   
    else if (USB_AUDIO_SAMPLING_RATE_96KHZ == sample_rate)
    {
        usb_feedback_normal = usb_feedback_normal | USB_FEEDBACK_96_SPEED_UP;
    }  
    else if (USB_AUDIO_SAMPLING_RATE_176_4KHZ == sample_rate)
    {
        usb_feedback_normal = usb_feedback_normal | USB_FEEDBACK_176_4_SPEED_UP;
    }                          
    else if (USB_AUDIO_SAMPLING_RATE_192KHZ == sample_rate)
    {
        usb_feedback_normal = usb_feedback_normal | USB_FEEDBACK_192_SPEED_UP;
    }
    else
    {
        /* Unsupported Sampling rate tuning */
    }
    
    usb_audio_data.feedback_speedup[0] = usb_feedback_normal & 0xFF;
    usb_audio_data.feedback_speedup[1] = (usb_feedback_normal >> 8) & 0xFF; 
    usb_audio_data.feedback_speedup[2] = (usb_feedback_normal >> 16) & 0xFF; 
    usb_audio_data.feedback_speedup[3] = (usb_feedback_normal >> 24) & 0xFF;                                                                                           
}

/**-----------------------------------------------------------------------------
 * Populates the supplied buffer (uint8_t[4]) with he appropriate feedback rate
 * @param copy_to
 * @param rate
 */
void usb_audio_feedback_set_rate(USB_AUDIO_FEEDBACK_RATE rate)
{
    //logger_log_debugf(LOGGER_CODE_DEBUG, "usb_audio_feedback_set_rate %d --", rate);
    memset(&usb_audio_data.feedback_buffer, 0, sizeof(uint32_t)); 

    switch(rate)
    {
        case USB_AUDIO_FEEDBACK_STEADY:
            memcpy(usb_audio_data.feedback_buffer, usb_audio_data.feedback_steady, sizeof(uint32_t));
            break;
            
        case USB_AUDIO_FEEDBACK_SPEEDUP:
            memcpy(usb_audio_data.feedback_buffer, usb_audio_data.feedback_speedup, sizeof(uint32_t));
            break;
            
        case USB_AUDIO_FEEDBACK_SLOWDOWN:
            memcpy(usb_audio_data.feedback_buffer, usb_audio_data.feedback_slowdown, sizeof(uint32_t));
            break;
            
        default:
            break;
    }
}

/* *****************************************************************************
 End of File
 */
