/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/usb_audio_port.h"
#include "app_config.h"
#include "utilities/logger.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! Handler for USB Device events
    USB_AUDIO_DEVICE_EVENT_HANDLER device_handler;
    
    //! Handler for USB Audio events
    USB_AUDIO_EVENT_HANDLER audio_handler;
    
    //! Count of write errors
    uint32_t write_errors;
    
    //! device layer handle returned by device layer open function 
    USB_DEVICE_HANDLE usb_handle;    
 
    //! The index of the next buffer to be written to with data from the USB Host
    volatile uint8_t write_queue_idx;
   
} USB_AUDIO_PORT_DATA;

static USB_AUDIO_PORT_DATA usb_audio_data =
{
    .write_queue_idx = 0,
    .write_errors = 0,
    .usb_handle = -1,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**-----------------------------------------------------------------------------
 * Attache the USB port
 */
void usb_audio_port_attach()
{
    USB_DEVICE_Attach(usb_audio_data.usb_handle);
}
    
/**-----------------------------------------------------------------------------
 * Wrap control receive message
 */
USB_DEVICE_CONTROL_TRANSFER_RESULT usb_audio_port_control_receive(void * data, size_t length)
{
    USB_DEVICE_ControlReceive(usb_audio_data.usb_handle, (void *)data, length);
}

/**-----------------------------------------------------------------------------
 * Wrap control send message
 * @param data
 * @param length
 * @return 
 */
USB_DEVICE_CONTROL_TRANSFER_RESULT usb_audio_port_control_send(void * data, size_t length)
{
    USB_DEVICE_ControlSend(usb_audio_data.usb_handle, (void *)data, length);
}
    
/**-----------------------------------------------------------------------------
 * Attache the USB port
 */
void usb_audio_port_detach()
{
    USB_DEVICE_Detach(usb_audio_data.usb_handle);
}

/**-----------------------------------------------------------------------------
 * tests whether the read error limit is exceeded
 */
bool usb_audio_port_error_limit_exceeded()
{
    return usb_audio_data.write_errors > USB_MAX_BAD_PACKETS;
}

/**-----------------------------------------------------------------------------
 * Gets the index of the next buffer to be written into with data from the USB Host
 * @return 
 */
uint32_t usb_audio_port_get_write_index()
{
    return usb_audio_data.write_queue_idx;
}
    
/**-----------------------------------------------------------------------------
 * Initialise the USB port
 */
bool usb_audio_port_initialise(USB_AUDIO_DEVICE_EVENT_HANDLER device_handler, 
        USB_AUDIO_EVENT_HANDLER audio_handler)
{
    logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO PORT Initialise ");
    bool result = true;
    
    usb_audio_data.device_handler = device_handler;
    usb_audio_data.audio_handler = audio_handler;
    
    logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO PORT Initialise END ");
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Opens the port for reading
 */
bool usb_audio_port_open()
{
    logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO PORT Open ");
    bool result = true;
    
    /* Open the device layer */
    usb_audio_data.usb_handle = USB_DEVICE_Open(USB_DEVICE_INDEX_0,
                                                DRV_IO_INTENT_READWRITE);

    if(usb_audio_data.usb_handle != USB_DEVICE_HANDLE_INVALID)
    {
        /* Register a callback with device layer to get event notification (for end point 0) */
        USB_DEVICE_EventHandlerSet(usb_audio_data.usb_handle, usb_audio_data.device_handler, 0);

        /* Attach the device */
        USB_DEVICE_Attach(usb_audio_data.usb_handle);
    }
    else
    {
        /* The Device Layer is not ready to be opened. We should try
         * again later. */
        logger_log_errorf(LOGGER_CODE_DEBUG, "USB AUDIO PORT Open error %d  ", usb_audio_data.usb_handle);
        result = false;
    }    
    
    logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO PORT Open END ");
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Reads the next audio buffer
 */
USB_DEVICE_AUDIO_V2_RESULT usb_audio_port_queue_read()
{
    USB_DEVICE_AUDIO_V2_RESULT result = USB_DEVICE_AUDIO_V2_RESULT_OK;
    
    if (usb_audio_buffer_data_clear(usb_audio_data.write_queue_idx))
    {
//logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO usb_audio_port_queue_read %d --", usb_audio_data.write_queue_idx);

        result = USB_DEVICE_AUDIO_V2_Read(USB_DEVICE_INDEX_0, 
                    (USB_DEVICE_AUDIO_V2_TRANSFER_HANDLE*)usb_audio_buffer_get_read_handle(usb_audio_data.write_queue_idx), 
                    1, 
                    (void*)usb_audio_buffer_get_read_buffer(usb_audio_data.write_queue_idx), 
                    NO_OF_SAMPLES_IN_A_USB_FRAME);//1024);

        if (result == USB_DEVICE_AUDIO_V2_RESULT_OK)
        {
            usb_audio_buffer_set_data_queued(usb_audio_data.write_queue_idx);
            
            usb_audio_data.write_queue_idx++;
            usb_audio_data.write_queue_idx %= USB_AUDIO_BUFFER_QUEUE_MAX_COUNT; 
        }
        else
        {
            usb_audio_data.write_errors++;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Reads all the audio buffers
 */
USB_DEVICE_AUDIO_V2_RESULT usb_audio_port_read_buffers()
{
    USB_DEVICE_AUDIO_V2_RESULT result = USB_DEVICE_AUDIO_V2_RESULT_OK;
    
    int x;
//    for (x = 0; x < (USB_AUDIO_BUFFER_QUEUE_MAX_COUNT - 4); x++)
    for (x = 0; x < (USB_AUDIO_BUFFER_QUEUE_MAX_COUNT); x++)
    {
        result = usb_audio_port_queue_read();
        if (usb_audio_port_error_limit_exceeded())
        {
            usb_audio_port_detach();                       
            usb_audio_port_attach();                            
        }
    }  
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 *  Resets the port
 */
void usb_audio_port_reset()
{
//logger_log_debugf(LOGGER_CODE_DEBUG, "USB AUDIO usb_audio_port_reset %d --", usb_audio_data.write_queue_idx);
    usb_audio_data.write_errors = 0;
    usb_audio_data.write_queue_idx = 0;
}
    
/**-----------------------------------------------------------------------------
 * Sends the USB host an error status
 */
void usb_audio_port_send_control_status_error()
{
    logger_log_debugf(0, "USB AUDIO - usb_audio_port_send_control_status_error");
    USB_DEVICE_ControlStatus(usb_audio_data.usb_handle, USB_DEVICE_CONTROL_STATUS_ERROR);
}

/**-----------------------------------------------------------------------------
 * Sends the USB host an OK status
 */
void usb_audio_port_send_control_status_ok()
{
    logger_log_debugf(0, "USB AUDIO - usb_audio_port_send_control_status_ok");
    USB_DEVICE_ControlStatus(usb_audio_data.usb_handle, USB_DEVICE_CONTROL_STATUS_OK);
}

/* *****************************************************************************
 End of File
 */
