/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    vfd.c

  @Summary
 Manages communication with a VFD

  @Description
 *  06/07/2018 - Added concept of VFD_SPECIAL_CHARACTER_SLOT_0 to enable 16
 *               user defined slots (rather than 15)
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/vfd.h"
#include "app_config.h"
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include "utilities/logger/logger.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define VFD_DISPLAY_LINE_COUNT      2

// The max length of a display line
#define VFD_DISPLAY_LINE_LEN        16

// The number of slots available to hold inverted characters
#define VFD_SPECIAL_CHARACTER_SLOTS    16

// Special characters are represented by character codes 0-15.  Our approach stores 
// these codes in the character buffer representing a line on the screen.  Because
// character \0 is the string terminator we can't use this in a string otherwise
// we will think the string is terminated at that character, sooooo we are using
// character 0xFF (DEL) to represent character slot 0
#define VFD_SPECIAL_CHARACTER_SLOT_0   0xFF

typedef enum
{
    VFD_COMMAND_DCRAM_DATA_WRITE    = 0x10,
    VFD_COMMAND_CGRAM_DATA_WRITE    = 0x20,
    VFD_COMMAND_ADRAM_DATA_WRITE    = 0x30,
    VFD_COMMAND_DUTY_SET            = 0x50,
    VFD_COMMAND_NO_OF_DIGITS_SET	= 0x60,
    VFD_COMMAND_ALL_LIGHTS_ON_OFF	= 0x70,
    VFD_COMMAND_TEST_MODE			= 0x80,
} VFD_COMMAND;

typedef enum 
{
    VFD_ALL_LIGHTS_NORMAL           = 0x00,
    VFD_ALL_LIGHTS_LOW              = 0x01,
    VFD_ALL_LIGHTS_HIGH             = 0x03,
} VFD_ALL_LIGHTS;

/**
 * Highlight mode behaviours
 */
typedef struct
{
    // Indicates whether the highlight flashes or not
    bool flash;
    
    // Set to tru if the highlight is a static image, false if dynamic
    bool cache;
    
    // The time for a flash cycle
    int flash_speed;
    
} VFD_HIGHLIGT_BEHAVIOUR;

/**
 * Structure representing an character created in CGRAM
 */
typedef struct
{
    // The character code to be used to access the special character in CGRAM
    char index;
    
    // The normal character code
    char character;
    
    //! There are only 16 (15) slots available - if we wanted to use the same 
    //  inverted character on both lines we would need 2 entries as they are addressed 
    //  differently
    DISPLAY_LINE line;
    
    //! The character offset if supplied
    int offset;
    
    // Set to true if this has been used since the last refresh
    // Used to find slots to use
    bool in_use;
    
    // A value indicating when this slot was last used - used to work out the stalest slots
    int last_used;
} VFD_SPECIAL_CHARACTER_SLOT;

/**
 * Global data structure containing data for the VFD
 */
typedef struct
{
    //! Keep track of brightness
    DISPLAY_BRIGHTNESS brightness;
    
    // Buffer to hold the display lines
    char display_lines[VFD_DISPLAY_LINE_COUNT][VFD_DISPLAY_LINE_LEN];
    
    // Array of structs containing info on the current state of special characters
    VFD_SPECIAL_CHARACTER_SLOT special_characters[VFD_SPECIAL_CHARACTER_SLOTS];
    
    // Set to judge whether to redisplay the line
    char refresh_line[VFD_DISPLAY_LINE_COUNT];
    
    VFD_HIGHLIGHT_MODE highlight_mode;
    
    // Toggle for flashing cursors
    bool flash_toggle;
    
    // Counter for flashing line
    uint8_t flash_counter;
} VFD_DATA;

static VFD_DATA vfd_data;

// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX

static volatile unsigned char cgram_data[16][6];
static volatile unsigned char cgram_index = 0;
static volatile unsigned char cgram_data_index = 0;

//static volatile unsigned char buffer2[6]; // 20200918 - Removed as it appears its not used and is causing compiler warnings

//static volatile int16_t idx;
//static volatile char c;

//static volatile char characters[6];

// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX

/**
 * Array of character definitions
 * Note: The first element is the character to replace, followed by 5 elements
 * each representing a column of data 1-5 containing 7 values.
 */
static char vfd_characters[][6] =
{
    {' ', 0x00, 0x00, 0x00,0x00, 0x00},	// 	space
    {'A', 0x7C, 0x12, 0x11,0x12, 0x7C},	// 	A
    {'B', 0x41, 0x7F, 0x49,0x49, 0x36},	// 	B
    {'C', 0x3E, 0x41, 0x41,0x41, 0x22},	// 	C
    {'D', 0x41, 0x7F, 0x41,0x41, 0x3E},	// 	D
    {'E', 0x7F, 0x49, 0x49,0x49, 0x41},	// 	E
    {'F', 0x7F, 0x09, 0x09,0x09, 0x01},	// 	F
    {'G', 0x3E, 0x41, 0x41,0x49, 0x7A},	// 	G
    {'H', 0x7F, 0x08, 0x08,0x08, 0x7F},	// 	H
    {'I', 0x00, 0x41, 0x7F,0x41, 0x00},	// 	I
    {'J', 0x20, 0x40, 0x41,0x3F, 0x01},	// 	J
    {'K', 0x7F, 0x08, 0x14,0x22, 0x41},	// 	K
    {'L', 0x7F, 0x40, 0x40,0x40, 0x40},	// 	L
    {'M', 0x7F, 0x02, 0x0C,0x02, 0x7F},	// 	M
    {'N', 0x7F, 0x04, 0x08,0x10, 0x7F},	// 	N
    {'O', 0x3E, 0x41, 0x41,0x41, 0x3E},	// 	O
    {'P', 0x7F, 0x09, 0x09,0x09, 0x06},	// 	P
    {'Q', 0x3E, 0x41, 0x51,0x21, 0x5E},	// 	Q
    {'R', 0x7F, 0x09, 0x19,0x29, 0x46},	// 	R
    {'S', 0x26, 0x49, 0x49,0x49, 0x32},	// 	S
    {'T', 0x01, 0x01, 0x7F,0x01, 0x01},	// 	T
    {'U', 0x3F, 0x40, 0x40,0x40, 0x3F},	// 	U
    {'V', 0x0F, 0x30, 0x40,0x30, 0x0F},	// 	V
    {'W', 0x7F, 0x20, 0x18,0x20, 0x7F},	// 	W
    {'X', 0x63, 0x14, 0x08,0x14, 0x63},	// 	X
    {'Y', 0x07, 0x08, 0x70,0x08, 0x07},	// 	Y
    {'Z', 0x61, 0x11, 0x49,0x45, 0x43},	// 	Z
    {'1', 0x00, 0x02, 0x02,0x7F, 0x00},	// 	1
    {'2', 0x62, 0x51, 0x49,0x49, 0x46},	// 	2
    {'3', 0x22, 0x41, 0x49,0x49, 0x36},	// 	3
    {'4', 0x18, 0x14, 0x12,0x7F, 0x10},	// 	4
    {'5', 0x27, 0x45, 0x45,0x45, 0x39},	// 	5
    {'6', 0x3E, 0x49, 0x49,0x49, 0x30},	// 	6
    {'7', 0x01, 0x01, 0x71,0x0D, 0x03},	// 	7
    {'8', 0x36, 0x49, 0x49,0x49, 0x36},	// 	8
    {'9', 0x06, 0x49, 0x49,0x49, 0x3E},	// 	9
    {'0', 0x3E, 0x51, 0x49,0x45, 0x3E},	// 	0
    {':', 0x00, 0x36, 0x36,0x00, 0x00},	// 	:
    {'.', 0x00, 0x60, 0x60,0x00, 0x00},	// 	.
    {'!', 0x00, 0xCF, 0xCF,0x00, 0x00},	// 	!
    
    {'a', 0x20, 0x54, 0x54,0x54, 0x78},
    {'b', 0x7F, 0x44, 0x44,0x44, 0x38},
    {'c', 0x38, 0x44, 0x44,0x44, 0x44},
    {'d', 0x38, 0x44, 0x44,0x44, 0x7F},
    {'e', 0x38, 0x54, 0x54,0x54, 0x58},
    {'f', 0x04, 0x04, 0x7E,0x05, 0x05},
    {'g', 0x0C, 0x54, 0x54,0x54, 0x3C},
    {'h', 0x7F, 0x08, 0x04,0x04, 0x78},
    {'i', 0x00, 0x44, 0x7D,0x40, 0x00},
    {'j', 0x20, 0x40, 0x40,0x3D, 0x00},
    {'k', 0x00, 0x7F, 0x10,0x28, 0x44},
    //{'l', 0x00, 0x41, 0x3F,0x40, 0x00}, // missing dot at bottom of long line
    {'l', 0x00, 0x41, 0x7F,0x40, 0x00},
    {'m', 0x7C, 0x04, 0x78,0x04, 0x78},
    {'n', 0x7C, 0x08, 0x04,0x04, 0x78},
    {'o', 0x38, 0x44, 0x44,0x44, 0x38},
    {(char)0xD7, 0x38, 0x45, 0x44,0x45, 0x38},
    {'p', 0x7C, 0x24, 0x24,0x24, 0x18},
    {'q', 0x18, 0x24, 0x24,0x24, 0x7C},
    {'r', 0x00, 0x7C, 0x08,0x04, 0x04},
    {'s', 0x48, 0x54, 0x54,0x54, 0x24},
    {'t', 0x00, 0x04, 0x3F,0x44, 0x44},
    {'u', 0x3C, 0x40, 0x40,0x20, 0x7C},
    {'v', 0x0C, 0x30, 0x40,0x30, 0x0C},
    {'w', 0x7C, 0x20, 0x18,0x20, 0x7C},
    {'x', 0x44, 0x28, 0x10,0x28, 0x44},
    {'y', 0x0C, 0x50, 0x50,0x50, 0x3C},
    {'z', 0x44, 0x64, 0x54,0x4C, 0x44},

    //{',', 0x00, 0x40, 0x20,0x00, 0x00},
    {',', 0x00, 0x50, 0x30,0x00, 0x00}, // VFD style
    {'"', 0x00, 0x07, 0x00,0x07, 0x00},
    {'+', 0x08, 0x08, 0x3E,0x08, 0x08},
    {'-', 0x08, 0x08, 0x08,0x08, 0x08},
    
    //{(char)0x27, 0x00, 0x02, 0x01,0x00, 0x00}, // apostrophe
    {(char)0x27, 0x00, 0x05, 0x03,0x00, 0x00}, // apostrophe VFD style
    {';', 0x00, 0x56, 0x30,0x00, 0x00},
    
    {'&', 0x36, 0x49, 0x55,0x22, 0x50},
    {'(', 0x00, 0x1C, 0x22,0x41, 0x00},
    {')', 0x00, 0x41, 0x22,0x1C, 0x00},
    {'/', 0x20, 0x10, 0x08,0x04, 0x02},
    {DISPLAY_SPECIAL_CHARACTERS_HOME, 0x7C, 0x42, 0x41,0x42, 0x7C}, // HOME
    {DISPLAY_SPECIAL_CHARACTERS_EMPTY_BLOCK, 0x00, 0x00, 0x00,0x00, 0x00}, // EMPTY
//    {DISPLAY_SPECIAL_CHARACTERS_BLOCK_20_L, 0x7F, 0x00, 0x00,0x00, 0x00}, // BLOCK_20_L
//    {DISPLAY_SPECIAL_CHARACTERS_BLOCK_40_L, 0x7F, 0x7F, 0x00,0x00, 0x00}, // BLOCK_40_L
//    {DISPLAY_SPECIAL_CHARACTERS_BLOCK_60_L, 0x7F, 0x7F, 0x7F,0x00, 0x00}, // BLOCK_60_L
//    {DISPLAY_SPECIAL_CHARACTERS_BLOCK_80_L, 0x7F, 0x7F, 0x7F,0x7F, 0x00}, // BLOCK_80_L
    {DISPLAY_SPECIAL_CHARACTERS_BLOCK, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F}, // BLOCK
    {DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_1, 0x55, 0x2A, 0x55,0x2A, 0x55}, // BLOCK
    {DISPLAY_SPECIAL_CHARACTERS_BLOCK_GREY_2, 0x2A, 0x55, 0x2A,0x55, 0x2A}, // BLOCK
    {DISPLAY_SPECIAL_CHARACTERS_HEADPHONES, 0x3E, 0x79, 0x01,0x79, 0x3E}, // HEADPHONES
    {DISPLAY_SPECIAL_CHARACTERS_DEGREES, 0x00, 0x00, 0x02,0x05, 0x02}, // DEGREES
    //{DISPLAY_SPECIAL_CHARACTERS_EDIT_ORIGINAL, 0x00, 0x00, 0x00,0x55, 0x55}, // EDIT_ORIGINAL
    //{DISPLAY_SPECIAL_CHARACTERS_EDIT_ORIGINAL, 0x00, 0x00, 0x00,0x2A, 0x2A}, // EDIT_ORIGINAL
    {DISPLAY_SPECIAL_CHARACTERS_EDIT_ORIGINAL, 0x00, 0x00, 0x00,0x14, 0x14}, // EDIT_ORIGINAL
    {DISPLAY_SPECIAL_CHARACTERS_DIRAC_LOGO, 0x36, 0x2A, 0x22,0x22, 0x1C},
    {DISPLAY_SPECIAL_CHARACTERS_ANALOGUE, 0x08, 0x06, 0x08,0x30, 0x08},
    {DISPLAY_SPECIAL_CHARACTERS_DIGITAL, 0x02, 0x02, 0x3E,0x20, 0x20},
    {DISPLAY_SPECIAL_CHARACTERS_USB_RECORD, 0x08, 0x0C, 0x3A,0x2A, 0x08},
    {DISPLAY_SPECIAL_CHARACTERS_UPGRADE_AVAILABLE, 0x12, 0x24, 0x48,0x24, 0x12},
    
    {0xE2, 0xFE, 0x09, 0x7F, 0x49, 0x41}, // AE
    {0xCE, 0x08, 0x7F, 0x49, 0x41, 0x3E}, // ETH
};

/**
 * Array of inverted character definitions
 * Note: The first element is the character to replace, followed by 5 elements
 * each representing a column of data 1-5 containing 7 values.
 */
//static char vfd_inverted_characters[][6] = 
//{
//    {'A', 0x03, 0x6D, 0x6E, 0x6D, 0x03},	
//    {'B', 0x3E, 0x00, 0x36, 0x36, 0x49},
//    {'C', 0x41, 0x3E, 0x3E, 0x3E, 0x5D},
//    {'D', 0x3E, 0x00, 0x3E, 0x3E, 0x41},
//    {'E', 0x00, 0x36, 0x36, 0x36, 0x3E},
//    {'F', 0x00 ,0x76, 0x76, 0x76, 0x7E},
//    {'G', 0x41, 0x3E, 0x3E, 0x36, 0x05},
//    {'H', 0x00, 0x77, 0x77, 0x77, 0x00},
//    {'I', 0x7F, 0x3E, 0x00, 0x3E, 0x7F},
//    {'J', 0x5F, 0x3F, 0x3E, 0x40, 0x7E},
//    {'K', 0x00, 0x77, 0x6B, 0x5D, 0x3E},
//    {'L', 0x00, 0x3F, 0x3F, 0x3F, 0x3F},
//    {'M', 0x00, 0x7D, 0x73, 0x7D, 0x00},
//    {'N', 0x00, 0x7B, 0x77, 0x6F, 0x00},
//    {'O', 0x41, 0x3E, 0x3E, 0x3E, 0x41},
//    {'P', 0x00, 0x76, 0x76, 0x76, 0x79},
//    {'Q', 0x41, 0x3E, 0x2E, 0x5E, 0x21},
//    {'R', 0x00, 0x76, 0x66, 0x56, 0x39},
//    {'S', 0x59, 0x36, 0x36, 0x36, 0x4D},
//    {'T', 0x7E, 0x7E, 0x00, 0x7E, 0x7E},
//    {'U', 0x40, 0x3F, 0x3F, 0x3F, 0x40},
//    {'V', 0x70, 0x4F, 0x3F, 0x4F, 0x70},
//    {'W', 0x00 ,0x5F, 0x67, 0x5F, 0x00},
//    {'X', 0x1C, 0x6B, 0x77, 0x6B, 0x1C},
//    {'Y', 0x78, 0x77, 0x0F, 0x77, 0x78},
//    {'Z', 0x1E, 0x2E, 0x36, 0x3A, 0x3C},
//    {'0', 0x41, 0x2E, 0x36, 0x3A, 0x41},
//    {'1', 0x7F, 0x7D, 0x7D, 0x00, 0x7F},
//    {'2', 0x1D, 0x2E, 0x36, 0x36, 0x39},
//    {'3', 0x5D, 0x3E, 0x36, 0x36, 0x49},
//    {'4', 0x67, 0x6B, 0x6D, 0x00 ,0x6F},
//    {'5', 0x58, 0x3A, 0x3A, 0x3A, 0x46},
//    {'6', 0x41, 0x36, 0x36, 0x36, 0x4F},
//    {'7', 0x7E, 0x7E, 0x0E, 0x72, 0x7C},
//    {'8', 0x49, 0x36, 0x36, 0x36, 0x49},
//    {'9', 0x79, 0x36, 0x36, 0x36, 0x41},
//    {' ', 0x7F, 0x7F, 0x7F, 0x7F, 0x7F},
//    {'-', 0x77, 0x77, 0x77, 0x77, 0x77},
//    {'.', 0x7F, 0x1F, 0x1F, 0x7F, 0x7F}
//};

static VFD_HIGHLIGT_BEHAVIOUR vfd_highlight_behaviours[] =
{
    { false, true, 500 },    //VFD_HIGHLIGHT_MODE_INVERT
    { true, true, 500 },     //VFD_HIGHLIGHT_MODE_INVERT_FLASH
    { true, true, 500 },     //VFD_HIGHLIGHT_MODE_TOGGLE_BOTTOM_LINE,
    { true, false, 500 },    //VFD_HIGHLIGHT_MODE_TOGGLE_CYCLE_LINE,
    { true, false, 200 },    //VFD_HIGHLIGHT_MODE_CYCLE_LINE,
    { true, false, 250 },    //VFD_HIGHLIGHT_MODE_FLASH_BLOCK,
    { false, false, 100 },    //VFD_HIGHLIGHT_MODE_CYCLE_FILL_DROP
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations
static int vfd_get_special_character_index(const char* character);
static VFD_SPECIAL_CHARACTER_SLOT* vfd_get_special_character_slot(DISPLAY_LINE line, const char* character, int offset, bool reuse, int* next_free_slot);
static void vfd_offset_special_character(char* buffer, int character_idx, int offset_h, int offset_v);
static void vfd_send_command(uint8_t* command, size_t size);

/**
 * Writes a character into CG RAM
 * @param address - the address in CGRAM to write to
 */
static void vfd_cgram_write_character(DISPLAY_LINE line, uint8_t slot_index, const char* character)
{
    size_t len = 6;
    uint8_t command[len];
    int x;
    
    uint8_t address = slot_index;
    if (address == VFD_SPECIAL_CHARACTER_SLOT_0)
    {
        address = 0;
    }

    command[0] = VFD_COMMAND_CGRAM_DATA_WRITE | (address & 0x0F);
    if (line == DISPLAY_LINE_1)
    {
        command[0] |= 0x80;
    }
    
    for (x = 1; x < len; x++)
    {
        command[x] = *(character + x);
        
        // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX
        cgram_data[address][x] = command[x];
        // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX
    }

    vfd_send_command(command, len);
}

/**
 * Highlights the character 
 * How the character is highlighted depends on the underlying display
 * @param line - the line to copy
 * @param character -  the character to be inverted
 * @return  - The character code to replace the current character - note a value = 0
 * should be ignored
 */
static char vfd_create_highlight_character(DISPLAY_LINE line, const char* character, bool highlight_spaces)
{
    char result = *character;
    int next_free_address = 0;
    bool proceed = false;
    
    VFD_HIGHLIGT_BEHAVIOUR behaviour = vfd_highlight_behaviours[vfd_data.highlight_mode];
    
    // Check for VFD_HIGHLIGHT_MODE_FLASH_BLOCK mode
    if (vfd_data.highlight_mode == VFD_HIGHLIGHT_MODE_FLASH_BLOCK)
    {
        if (vfd_data.flash_toggle)
        {
            // Flash a block
            result = 0x20;
            if (highlight_spaces && (*character == ' '))
            {
                result = 0x7F; // Block
            }
        }
    }
    // Work out whether we need to continue based on flashing option
    else if (!behaviour.flash || (vfd_data.flash_toggle && behaviour.flash))
    {
        proceed = true;
    }
    
    if (proceed)
    {
        // Do we already have the inverted version loaded ?
        VFD_SPECIAL_CHARACTER_SLOT* slot = NULL;

        if (behaviour.cache)
        {
            slot = vfd_get_special_character_slot(line, character, 0, true, &next_free_address);
        }
        
        // otherwise find next free position and
        // create inverted character
        if (slot != NULL)
        {
            result = slot->index;
        }
        else
        {
            // we should have the next free slot
            // so find the character array that defines the inverted character and create it at the 
            // specified address (next_free_idx + 1);
            int character_idx = vfd_get_special_character_index(character);

            if (character_idx >= 0)
            {
                // Update the inverted character
                VFD_SPECIAL_CHARACTER_SLOT* slot = &vfd_data.special_characters[next_free_address];
                slot->character = character[0];
                slot->line = line;
                slot->offset = 0;
                slot->in_use = true;
                slot->last_used = 0;

                size_t len = 6;
                uint8_t command[len];
                int x;

                command[0] = VFD_COMMAND_CGRAM_DATA_WRITE | (slot->index & 0x0F) | 0x80;
                for (x = 1; x < len; x++)
                {
                    // Useful bitwise stuff
                    // https://stackoverflow.com/questions/47981/how-do-you-set-clear-and-toggle-a-single-bit
                    switch(vfd_data.highlight_mode)
                    {
                        case VFD_HIGHLIGHT_MODE_INVERT:
                        case VFD_HIGHLIGHT_MODE_INVERT_FLASH:
                        {
                            // Invert
                            command[x] = ~vfd_characters[character_idx][x];
                            break;
                        }
                        
                        case VFD_HIGHLIGHT_MODE_TOGGLE_BOTTOM_LINE:
                        {
                            command[x] = vfd_characters[character_idx][x];
                            command[x] ^= 1 << 6;
                            break;
                        }
                            
                        case VFD_HIGHLIGHT_MODE_TOGGLE_CYCLE_LINE:
                        {
                            command[x] = vfd_characters[character_idx][x];
                            command[x] ^= 1 << vfd_data.flash_counter;
                            command[x] ^= 1 << 6;
                            break;
                        }
                        
                        case VFD_HIGHLIGHT_MODE_CYCLE_LINE:
                        {
                            command[x] = vfd_characters[character_idx][x];
                            command[x] |= 1 << vfd_data.flash_counter;
                            break;
                        }
                        
                        case VFD_HIGHLIGHT_MODE_CYCLE_FILL_DROP:
                        {
                            command[x] = vfd_characters[character_idx][x];
                            
                            int y;
                            for (y = vfd_data.flash_counter + 1; y < 7; y++)
                            {
                                command[x] |= 1 << y;
                            }
                            
                            break;
                            
                        }
                        
                        default:
                        {
                    // flash bottom line
                            command[x] = vfd_characters[character_idx][x];
                            //command[x] |= vfd_data.flash_toggle << 6;
                            if (vfd_data.flash_toggle)
                            {
                                command[x] ^= (-vfd_data.flash_toggle ^ command[x]) & (1 << 6);
                            }
                            else
                            {
                                // If the bottom line has more than 4 lit pixels then switch off
                                // otherwise leave
                                //command[x] ^= 1 << 6;
                            }
                            break;
                        }
                    }
                }

                vfd_send_command(command, len);

                result = slot->index;
            }
        }
    }
    
    return result;
}

/**
 * 
 * @param characters
 * @param character_idx
 * @param offset_h
 * @param offset_v
 */
static void vfd_merge_special_characters(char* characters, int character_idx, int offset_h, int offset_v)
{
    if (character_idx >= 0)
    {
        char buffer[6];
        vfd_offset_special_character(buffer, 
                character_idx,
                offset_h, 
                offset_v);

        //logger_log_debugf(90, "characters  %d %d %d %d %d %d -- ", characters[0], characters[1], characters[2], characters[3], characters[4], characters[5], characters[6] );
        //logger_log_debugf(90, "buffer      %d %d %d %d %d %d -- ", buffer[0], buffer[1], buffer[2], buffer[3], buffer[4], buffer[5], buffer[6] );
        int x;
        for (x = 1; x <= 6; x++)
        {
            characters[x] |= buffer[x];
        }
        //logger_log_debugf(90, "characters2 %d %d %d %d %d %d -- ", characters[0], characters[1], characters[2], characters[3], characters[4], characters[5], characters[6] );
    }
}

/**
 * 
 * @param buffer - buffer should be a pointer to a character[6] array containing the 
 *                special character CGRAM info
 * @param character_idx
 * @param offset_h
 * @param offset_v
 * @return 
 */
static void vfd_offset_special_character(char* buffer, int character_idx, int offset_h, int offset_v)
{
    if (character_idx >= 0)
    {
        // Copy the character into the buffer - note index 0 is the character code
        // 1-5 contain the data
        memcpy(buffer, vfd_characters[character_idx], 6);
        //memcpy(buffer2, vfd_characters[character_idx], 6);
        
        //logger_log_debugf(90, "buffer1 - %c %d %d %d %d %d %d %d -- ", buffer[0], offset_h, buffer[0], buffer[1], buffer[2], buffer[3], buffer[4], buffer[5], buffer[6] );


        if (offset_h != 0)
        {
            int x;
            for (x = 0; x < 5; x++)
            {
                if ((x - offset_h >= 0) && (x - offset_h < 5))
                {
                    buffer[x + 1] = vfd_characters[character_idx][x - offset_h + 1];
                }
                else
                {
                    buffer[x + 1] = 0x00; //Blank column
                }
            }
        }
        //logger_log_debugf(90, "buffer2 - %c %d %d %d %d %d %d %d -- ", buffer[0], offset_h, buffer[0], buffer[1], buffer[2], buffer[3], buffer[4], buffer[5], buffer[6] );

        // Apply vertical offset
        if (offset_v != 0)
        {
            int x;
            for (x = 0; x < 5; x++)
            {
                if (offset_v > 0)
                {
                    buffer[x + 1] = (buffer[x + 1] << offset_v);
                    //buffer2[x + 1] = (buffer2[x + 1] << offset_v);
                }
                else
                {
                    buffer[x + 1] = (buffer[x + 1] >> (offset_v * -1));
                    //buffer2[x + 1] = (buffer2[x + 1] >> (offset_v * -1));
                }
            }
        }
        //logger_log_debugf(90, "buffer3 - %c %d %d %d %d %d %d %d -- ", buffer[0], offset_h, buffer[0], buffer[1], buffer[2], buffer[3], buffer[4], buffer[5], buffer[6] );
    }
}

/**
 * Works out the position in a nXn array of the next vertically placed character
 * including an off set e.g. +1 that ensures the offset character is on the line
 * above/below the source character
 * @param index
 * @param buffer_height
 * @param buffer_width
 * @param offset_up
 * @return 
 */
static uint8_t vfd_offset_horizontal_buffer_index(uint8_t index, int offset, uint8_t buffer_height, uint8_t buffer_width)
{
    int16_t result = index;

    // Which line are we on
    uint8_t line = result / buffer_width;

    // Add the offset
    result += offset;

    // Make sure its on the same line
    result = (result % buffer_width) + (line * buffer_width);

    return (uint8_t)result;
}

/**
 * Works out the position in a nXn array of the next vertically placed character
 * including an off set e.g. +1 that ensures the offset character is on the line
 * above/below the source character
 * @param index
 * @param buffer_height
 * @param buffer_width
 * @param offset_up
 * @return 
 */
static uint16_t vfd_offset_vertical_buffer_index(uint16_t index, uint8_t offset, uint8_t buffer_height, uint8_t buffer_width, bool offset_up)
{
    int16_t result;
     
    if (offset_up)
    {
        // Offset is negative (meaning we will only be looking at the bottom of the main character)
        // so get the index of the line below)
        result = index + buffer_width;
    }
    else
    {
        // Get the index of the character above
        result = index - buffer_width;
    }

    // Which line are we on (Note: Ordinal value for line))
    int16_t line = result / buffer_width;
    if (line >= buffer_height)
    {
        line = 0;
    }
    else if (line < 0)
    {
        line = buffer_height - 1;
    }

    // Add the offset
    result += offset;

    // Make sure its on the same line
    result = (result % buffer_width) + (line * buffer_width);

    return (uint16_t)result;
}

/**
 * Creates a special character
 * @param line - the line on the display
 * @param buffer -  a pointer to a buffer containing buffer_height X buffer_width characters, note if the buffer is height = 1 the text will not wrap around vertically, if width -= 1 it will not wrap horizontally
 * @param buffer_height - the height of the buffer (i.e. number of rows)
 * @param buffer_width - the width of the buffer (i.e. number of columns)
 * @param index - the index of the character in the buffer scanning from left to right, top row first e.g for a 10X3 buffer index 11 is the second character from the left on the second row
 * @param offset_h - the horizontal offset a negative number moves it to the left, a positive to the right
 * @param offset_v - the vertical offset, a negative number moves it down, a positive number moves it up.
 * @return  - The character code to replace the current character - note a value = 0
 * should be ignored, slot 0 is represented by VFD_SPECIAL_CHARACTER_SLOT_0
 */
static char vfd_create_special_character(DISPLAY_LINE line, char* buffer, uint8_t buffer_height, uint8_t buffer_width, uint16_t index, int offset_h, int offset_v)
{
    char result = (char)buffer[index];
    int next_free_address = -1;
    int offset = (offset_h << 8) | offset_v;
    bool reuse = true;
    
    // See if we can find a slot thats already doing what we want, if theres no
    // display of other characters so we can do this, other wise we get 
    // issues where a block of text contains the same character twice but their
    // surrounding characters differ
    if ((buffer_height > 1) || (buffer_width > 1))
    {
        reuse = false;
    }
    
    VFD_SPECIAL_CHARACTER_SLOT* slot = vfd_get_special_character_slot(line, (const char*)&buffer[index], offset, reuse, &next_free_address);;
    
    // So if we found a slot and were not in a text block use it
    if (slot != NULL)
    {
        result = slot->index;
    }
    else
    {
        // we should have the next free slot
        // so find the character array that defines the inverted character and create it at the 
        // specified address (next_free_idx + 1);
        int character_idx = vfd_get_special_character_index((const char*)&buffer[index]);
        if (character_idx >= 0)
        {
            slot = &vfd_data.special_characters[next_free_address];
            slot->character = buffer[index];
            slot->line = line;
            slot->offset = offset;
            slot->in_use = true;
            slot->last_used = 0;
            
            // No offset
            if (offset == 0)
            {
                vfd_cgram_write_character(line, slot->index, vfd_characters[character_idx]);
            }
            else
            {
                // Apply offset to the main character
                char characters[6];
                vfd_offset_special_character(characters, character_idx, offset_h, offset_v);

                uint16_t buffer_size = buffer_height * buffer_width;
                int16_t idx;
                int char_offset_h = offset_h; 
                int char_offset_v = offset_v;
                
                int char_height = vfd_get_character_height();
                int char_width = vfd_get_character_width();
                
                uint8_t character_spacing = 1;
                
                // Do the top line and bottom line of the surrounding characters
                if (buffer_height > 1)
                {
                    idx = vfd_offset_vertical_buffer_index(index, 0, buffer_height, buffer_width, offset_v < 0 ? true : false);

                    character_idx = vfd_get_special_character_index((const char*)&buffer[idx]);
                    if (character_idx >= 0)
                    {
                        if (offset_v > 0)
                        {
                            // top center
                            char_offset_h = offset_h;
                            char_offset_v = -1 * (vfd_get_character_height() - offset_v) - character_spacing;
                            
                            //logger_log_debugf(89, "IDX 1 - %c %c %d %d %d %d --", buffer[index], buffer[idx], offset_h, offset_v, char_offset_h, char_offset_v);
                        }
                        else if (offset_v < 0)
                        {
                            // bottom center
                            char_offset_h = offset_h;//(offset_h + char_width) % char_width;
                            char_offset_v = vfd_get_character_height() + offset_v + character_spacing;
                            
                            //logger_log_debugf(89, "IDX 2 - %c %c %d %d %d %d --", buffer[index], buffer[idx], offset_h, offset_v, char_offset_h, char_offset_v);
                        }

                        vfd_merge_special_characters(characters, character_idx, char_offset_h, char_offset_v);
                    }
                    
                    if (buffer_width > 1)
                    {
                        // Only do left side if offset_h > 0
                        if (offset_h > 0)
                        {
                            idx = vfd_offset_vertical_buffer_index(index, -1, buffer_height, buffer_width, offset_v < 0 ? true : false);
                            
                            character_idx = vfd_get_special_character_index((const char*)&buffer[idx]);
                            if (character_idx >= 0)
                            {
                                if (offset_v > 0)
                                {
                                    // top left
                                    char_offset_h = -1 * (vfd_get_character_width() - offset_h + character_spacing);
                                    char_offset_v = -1 * (vfd_get_character_height() - offset_v) - character_spacing;
                             
                                    //logger_log_debugf(89, "IDX 3 - %c %c %d %d %d %d --", buffer[index], buffer[idx], offset_h, offset_v, char_offset_h, char_offset_v);
                               }
                                else if (offset_v < 0)
                                {
                                    // bottom left
                                    char_offset_h = -1 * (vfd_get_character_width() - offset_h + character_spacing);
                                    char_offset_v = vfd_get_character_height() + offset_v + character_spacing;
                             
                                    //logger_log_debugf(89, "IDX 4 - %c %c %d %d %d %d --", buffer[index], buffer[idx], offset_h, offset_v, char_offset_h, char_offset_v);
                                }
                            
                                vfd_merge_special_characters(characters, character_idx, char_offset_h, char_offset_v);
                            }
                        } 
                        // Only do right side if offset_h < 0
                        else if (offset_h < 0)
                        {
                            idx = vfd_offset_vertical_buffer_index(index, 1, buffer_height, buffer_width, offset_v < 0 ? true : false);

                            character_idx = vfd_get_special_character_index((const char*)&buffer[idx]);
                            if (character_idx >= 0)
                            {
                                if (offset_v > 0)
                                {
                                    // top right
                                    char_offset_h = vfd_get_character_width() + offset_h + character_spacing;
                                    char_offset_v = -1 * (vfd_get_character_height() - offset_v) - character_spacing;
                             
                                    //logger_log_debugf(89, "IDX 5 - %c %c %d %d %d %d --", buffer[index], buffer[idx], offset_h, offset_v, char_offset_h, char_offset_v);
                                }
                                else if (offset_v < 0)
                                {
                                    // bottom right
                                    char_offset_h = vfd_get_character_width() + offset_h + character_spacing;
                                    char_offset_v = vfd_get_character_height() + offset_v + character_spacing;
                             
                                    //logger_log_debugf(89, "IDX 6 - %c %c %d %d %d %d --", buffer[index], buffer[idx], offset_h, offset_v, char_offset_h, char_offset_v);
                                }

                                vfd_merge_special_characters(characters, character_idx, char_offset_h, char_offset_v);
                            }
                        }
                   }
                } 
                
                if (buffer_width > 1)
                {
                    if (offset_h > 0)
                    {
                        // center left
                        idx = vfd_offset_horizontal_buffer_index(index, -1, buffer_height, buffer_width);
                                                                
                        character_idx = vfd_get_special_character_index((const char*)&buffer[idx]);
                        if (character_idx >= 0)
                        {
                            char_offset_h = -1 * (vfd_get_character_width() - offset_h + character_spacing);
                            char_offset_v = offset_v;
                             
                            //logger_log_debugf(89, "IDX 7 - %c %c %d %d %d %d --", buffer[index], buffer[idx], offset_h, offset_v, char_offset_h, char_offset_v);
                        }
                   }
                    else if (offset_h < 0)
                    {
                        // center right
                       idx = vfd_offset_horizontal_buffer_index(index, 1, buffer_height, buffer_width);
                      
                        character_idx = vfd_get_special_character_index((const char*)&buffer[idx]);
                        if (character_idx >= 0)
                        {
                            char_offset_h = vfd_get_character_width() + offset_h + character_spacing;
                            char_offset_v = offset_v;
                             
                            //logger_log_debugf(89, "IDX 8 - %c %c %d %d %d %d --", buffer[index], buffer[idx], offset_h, offset_v, char_offset_h, char_offset_v);
                        }
                    }

                     vfd_merge_special_characters(characters, character_idx, char_offset_h, char_offset_v);
                }

                vfd_cgram_write_character(line, slot->index, characters);
            }
            
            result = slot->index;
        }
    }

    return result;
}

/**
 * Disable Chip Select pin
 */
static void vfd_disable_transfer()
{
#ifdef VFD_PINS_ENABLED
    
    PLIB_PORTS_PinSet(VFD_CS_PORTS_ID, 
                      VFD_CS_PORT_CHANNEL, 
                      VFD_CS_PIN);

#endif
}

/**
 * Disable CS pin
 */
static void vfd_enable_transfer()
{
#ifdef VFD_PINS_ENABLED
    
    PLIB_PORTS_PinClear(VFD_CS_PORTS_ID, 
                        VFD_CS_PORT_CHANNEL, 
                        VFD_CS_PIN);
    
#endif
}

/**
 * Used to make inverted character slots available for reuse
 */
static void vfd_free_special_character_slots()
{
    int x;
    for (x = 0; x < VFD_SPECIAL_CHARACTER_SLOTS; x++)
    {
        VFD_SPECIAL_CHARACTER_SLOT* slot = &vfd_data.special_characters[x];
        slot->in_use = false;
        slot->last_used++;
    }    
}

/**
 * Returns the index of the supplied character from the vfd_characters array
 * @param character
 * @return 
 */
static int vfd_get_special_character_index(const char* character)
{
    int result = -1;

    int x;
    
    for (x = 0; x < sizeof(vfd_characters) / (6 * sizeof(char)); x++)
    {
        if (vfd_characters[x][0] == character[0])
        {
            result = x;
            break;
        }
    }

    return result;
}

/**
 * Gets the special character slot that matches the line/character or returns the index of the 
 * next free slot
 * @param line - The line on which the character will be displayed
 * @param character - The special character code
 * @param next_free_slot - if no slot is found this will contain the index of the next free slot
 * @return - a pointer to the VFD_SPECIAL_CHARACTER slot representing the character
 */
static VFD_SPECIAL_CHARACTER_SLOT* vfd_get_special_character_slot(DISPLAY_LINE line, const char* character, int offset, bool reuse, int* next_free_slot)
{
    VFD_SPECIAL_CHARACTER_SLOT* result = NULL;
    int last_used_min = 0;
    int candidate_free_slot = -1;
    
    int x;
    for (x = 0; x < VFD_SPECIAL_CHARACTER_SLOTS; x++)
    {
        VFD_SPECIAL_CHARACTER_SLOT* slot = &vfd_data.special_characters[x];
        if ((reuse) &&
            (slot->line == line) &&
            (slot->character == character[0] &&
             slot->offset == offset))
        {
            // If so return char code
            slot->in_use = true;
            slot->last_used = 0;
            result = slot;  // +1 is offset so we dont use char 0.... 
            break;
        }
        else if (next_free_slot != NULL)
        {
            if (!slot->in_use)
            {
                if (last_used_min < slot->last_used)
                {
                    last_used_min = slot->last_used;
                    candidate_free_slot = x;
                }
            }
            
//            // Find next CGRAM position to write inverted character to
//            if ((*next_free_slot == -1) &&
//                !slot->in_use)  
//            {
//                *next_free_slot = x; 
//            }
        }
    }
    
    if (next_free_slot != NULL)
    {
        *next_free_slot = candidate_free_slot; 
    }

    return result;
}

/**
 * Used to make inverted character slots available for reuse
 */
static void vfd_initialise_special_character_slots()
{
    // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX
    for (cgram_index = 0; cgram_index < 16; cgram_index++)
    {
        for (cgram_data_index = 0; cgram_data_index < 6; cgram_data_index++)
        {
            cgram_data[cgram_index][cgram_data_index] = 0;
        }
    }
    
    cgram_index = 0;
    cgram_data_index = 0;
    
    
    
    
    int x;
    for (x = 0; x < VFD_SPECIAL_CHARACTER_SLOTS; x++)
    {
        VFD_SPECIAL_CHARACTER_SLOT* slot = &vfd_data.special_characters[x];
        slot->character = 0;
        slot->in_use = false;
        slot->last_used = -1;
        
        //slot->index = x + 1;  // We do not use slot 0 as it can be confused with a null string terminator
        if (x == 0)
        {
            // Slot zero has a special index so it does not represent a null character
            slot->index = VFD_SPECIAL_CHARACTER_SLOT_0;
        }
        else
        {
            slot->index = x;
        }
    }    
}

/**
 * Send a command to the VFD
 * A VFD command can be multiple bytes that are sent in sequence with then first
 * byte containing the start address to write to, followed by multiple lines of data
 * which address automatically increments
 * @param command - An array of uint8_t, each element contaning a command
 * @param size - the number of elements in the array
 */
static void vfd_send_command(uint8_t* command, size_t size)
{
#if VFD_PINS_ENABLED
 	
	// enable transfer
    vfd_enable_transfer();

    int x;
    for (x = 0; x < size; x++)
    {
        uint8_t pos = 0;
        for (pos = 0; pos < 8; pos++)
        {
            // set clock low
            PLIB_PORTS_PinClear(VFD_SCLK_PORTS_ID, 
                                VFD_SCLK_PORT_CHANNEL, 
                                VFD_SCLK_PIN);

            // set data from msb
            PLIB_PORTS_PinWrite(VFD_MOSI_PORTS_ID, 
                                VFD_MOSI_PORT_CHANNEL, 
                                VFD_MOSI_PIN,
                                ((command[x] >> pos) & 1) == 1);


            timers_delay_us(1);

            // set clock high
            PLIB_PORTS_PinSet(VFD_SCLK_PORTS_ID, 
                              VFD_SCLK_PORT_CHANNEL, 
                              VFD_SCLK_PIN);

            timers_delay_us(1);
        }

        // Delay between command lines
        timers_delay_us(2);
    }
	
	// terminate transfer
	vfd_disable_transfer();
    
    // Delay before next transfer
    timers_delay_us(2);
    
#endif

}

/**
 * Set the all lights on/off settings
 * @param value
 */
static void vfd_set_all_lights(VFD_ALL_LIGHTS value)
{
    uint8_t command = VFD_COMMAND_ALL_LIGHTS_ON_OFF | value;
    vfd_send_command(&command, 1);
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Clear the VFD display
 * Actually does nothing as the buffer is used to compare with any new display to
 * decide whether an update is needed or not - The clear is actually done in the
 * vfd_write_line, thoygh this means that there is no way of clearing the vfd without
 * writing something which may need addressing but not done at the moment..
 */
void vfd_clear()
{
    // Clear the buffer
    //memset(vfd_data.display_lines, ' ', VFD_DISPLAY_LINE_COUNT * VFD_DISPLAY_LINE_LEN);    
}

/**
 * Clears the VFD into reset mode - 
 */
void vfd_clear_reset()
{
#ifdef VFD_PINS_ENABLED
    
    PLIB_PORTS_PinSet(VFD_CS_PORTS_ID, 
                      VFD_CS_PORT_CHANNEL, 
                      VFD_CS_PIN);

    // set clock high
    PLIB_PORTS_PinSet(VFD_SCLK_PORTS_ID, 
                      VFD_SCLK_PORT_CHANNEL, 
                      VFD_SCLK_PIN);
    
    PLIB_PORTS_PinClear(VFD_MOSI_PORTS_ID, 
                        VFD_MOSI_PORT_CHANNEL, 
                        VFD_MOSI_PIN);
    
    PLIB_PORTS_PinSet(VFD_RESET_PORTS_ID, 
                      VFD_RESET_PORT_CHANNEL, 
                      VFD_RESET_PIN);
   
#endif    
}

/**
 * Gets the height of a character in pixels
 * @return 
 */
size_t vfd_get_character_height()
{
    return 7;
}

/**
 * Gets the width of a character in pixels
 * @return 
 */
size_t vfd_get_character_width()
{
    return 5;
}

/**
 * Highlights the character 
 * How the character is highlighted depends on the underlying display
 * @param line - the line to copy
 * @param character -  the character to be inverted
 * @return  - The character code to replace the current character - note a value = 0
 * should be ignored
 */
char vfd_get_highlight_character(DISPLAY_LINE line, const char* character, bool highlight_spaces)
{
    char result = *character;

    if (timers_has_expired(TIMER_INDEX_VFD_FLASH))
    {
        vfd_data.flash_toggle = !vfd_data.flash_toggle;

        if (vfd_data.flash_toggle)
        {
            vfd_data.flash_counter++;
            vfd_data.flash_counter = (vfd_data.flash_counter + 6) % 6;
        }

        timers_reset(TIMER_INDEX_VFD_FLASH);
    }
    
    result = vfd_create_highlight_character(line, character, highlight_spaces);
    
    return result;
}
    
/**
 * Returns the length of the display in characters
 * @return 
 */
size_t vfd_get_line_length()
{
    return VFD_DISPLAY_LINE_LEN;
}
    
/**
 * Returns a character code 
 * @param line - the line to add the special character to 
 * @param character - a character selected from the DISPLAY_SPECIAL_CHARACTERS enum
 * @return 
 */
char vfd_get_special_character(DISPLAY_LINE line, char* buffer, uint8_t buffer_height, uint8_t buffer_width,  uint16_t index, int offset_h, int offset_v)
{
    return vfd_create_special_character(line, buffer, buffer_height, buffer_width, index, offset_h, offset_v);
}
    
/**
 * Returns the character code of the special character in the given slot
 * @param slot - the special character slot number 
 * @return 
 */
char vfd_get_special_character_in_slot(uint8_t slot)
{
    char result = slot;
        
    if (slot < VFD_SPECIAL_CHARACTER_SLOTS)
    {
        //result = vfd_data.special_characters[slot - 1].character;
        result = vfd_data.special_characters[slot].character;
    }
    else if (slot == VFD_SPECIAL_CHARACTER_SLOT_0)
    {
        // The character VFD_SPECIAL_CHARACTER_SLOT_0 (0xFF) represents the first
        // slot
        result = vfd_data.special_characters[0].character; 
    }
    
    return result;
}
    
/**
 * Puts the VFD into reset mode - used during power up/down to prevent spurious 
 * characters being displayed
 */
void vfd_hold_in_reset()
{
#ifdef VFD_PINS_ENABLED
    
//    PLIB_PORTS_PinSet(VFD_CS_PORTS_ID, 
//                      VFD_CS_PORT_CHANNEL, 
//                      VFD_CS_PIN);
//
//    // set clock high
//    PLIB_PORTS_PinSet(VFD_SCLK_PORTS_ID, 
//                      VFD_SCLK_PORT_CHANNEL, 
//                      VFD_SCLK_PIN);
    
    PLIB_PORTS_PinClear(VFD_CS_PORTS_ID, 
                      VFD_CS_PORT_CHANNEL, 
                      VFD_CS_PIN);

    // set clock high
    PLIB_PORTS_PinClear(VFD_SCLK_PORTS_ID, 
                      VFD_SCLK_PORT_CHANNEL, 
                      VFD_SCLK_PIN);
    
    PLIB_PORTS_PinClear(VFD_MOSI_PORTS_ID, 
                        VFD_MOSI_PORT_CHANNEL, 
                        VFD_MOSI_PIN);
    
    PLIB_PORTS_PinClear(VFD_RESET_PORTS_ID, 
                        VFD_RESET_PORT_CHANNEL, 
                        VFD_RESET_PIN);
   
#endif    
}

/**
 * Called by display_initialise()
 */
void vfd_initialise()
{
    vfd_set_highlight_mode(VFD_HIGHLIGHT_MODE_FLASH_BLOCK);
    vfd_data.flash_toggle = true;
    vfd_data.flash_counter = 0;

    vfd_initialise_special_character_slots();
    
    vfd_clear_reset();
    
    timers_delay_us(200);
    
    // Clear the buffer
    memset(vfd_data.display_lines, ' ', VFD_DISPLAY_LINE_COUNT * VFD_DISPLAY_LINE_LEN);    
    
    int x;
    for (x = 0; x < VFD_DISPLAY_LINE_COUNT; x++)
    {
        vfd_data.refresh_line[x] = true;
    }
    
    vfd_set_all_lights(VFD_ALL_LIGHTS_LOW);
    
    //Set number of digits to all COM1-20
    uint8_t command = VFD_COMMAND_NO_OF_DIGITS_SET | 0x00;
    vfd_send_command(&command, 1);
    
    vfd_set_brightness(DISPLAY_BRIGHTNESS_OFF);
    
    vfd_refresh();
}

/**
 * Refresh the VFD display
 */
void vfd_refresh()
{
    size_t len = VFD_DISPLAY_LINE_LEN;
    uint8_t command[len + 1];
    int x;
   
    // Draw top line
    if (vfd_data.refresh_line[DISPLAY_LINE_1])
    {
        command[0] = VFD_COMMAND_DCRAM_DATA_WRITE | (0 & 0x0F) | 0x80;
        for (x = 0; x < len; x++)
        {
            uint8_t character = *(vfd_data.display_lines[DISPLAY_LINE_1] + x);

// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX            
            cgram_index = 0; // use as a flag for unset
            
            if (character < 16)
            {
                // special character // check for data
                
                for (cgram_data_index = 0; cgram_data_index < 6; cgram_data_index++)
                {
                    if (cgram_data[character][cgram_data_index] != 0)
                    {
                        //pixel displayed
                        cgram_index = 1; // use as a flag for set;
                    }
                }
                
                if (cgram_index > 0)
                {
                    cgram_index = 0;
                    volatile int q = 1;
                }
            }
            
// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX            
            if (character == VFD_SPECIAL_CHARACTER_SLOT_0)
            {
                character = 0x00; // CGRAM slot 0
            }
            command[x+1] = character;
        }
    
        vfd_send_command(command, len + 1);
        
        vfd_data.refresh_line[DISPLAY_LINE_1] = false;
    }
    
    // Draw bottom line
    if (vfd_data.refresh_line[DISPLAY_LINE_2])
    {
        command[0] = VFD_COMMAND_DCRAM_DATA_WRITE | (0 & 0x0F);
        for (x = 0; x < len; x++)
        {
            uint8_t character = *(vfd_data.display_lines[DISPLAY_LINE_2] + x);
            
            
            // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX            
            cgram_index = 0; // use as a flag for unset
            
            if (character < 16)
            {
                // special character // check for data
                
                for (cgram_data_index = 0; cgram_data_index < 6; cgram_data_index++)
                {
                    if (cgram_data[character][cgram_data_index] != 0)
                    {
                        //pixel displayed
                        cgram_index = 1; // use as a flag for set;
                    }
                }
                
                if (cgram_index > 0)
                {
                    cgram_index = 0;
                    volatile int q = 1;
                }
            }
            
// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX DEBUG XXXXXXXXXXXXXXXXXX
            
            
            
            
            if (character == VFD_SPECIAL_CHARACTER_SLOT_0)
            {
                character = 0x00; // CGRAM slot 0
            }
            command[x+1] = character;
        }

        vfd_send_command(command, len + 1);
        
        vfd_data.refresh_line[DISPLAY_LINE_2] = false;
    }
    
    vfd_free_special_character_slots();
}

/**
 * Set the VFD display brightness
 */
void vfd_set_brightness(DISPLAY_BRIGHTNESS brightness)
{
    uint8_t command[2];
    
    vfd_data.brightness = brightness;
    
    switch (brightness)
    {
        case DISPLAY_BRIGHTNESS_OFF:
        {
            command[0] = VFD_COMMAND_DUTY_SET |  0x00;
            command[1] = 0x00;
            vfd_send_command(command, 2);

            vfd_set_all_lights(VFD_ALL_LIGHTS_LOW);

            break;
        }
        
        case DISPLAY_BRIGHTNESS_DIMMED:
        {
            // Sets brightness to 1/2
            command[0] = VFD_COMMAND_DUTY_SET |  0x03;
            command[1] = 0x7F;
            vfd_send_command(command, 2);

            vfd_set_all_lights(VFD_ALL_LIGHTS_NORMAL);
            break;
        }
            
        case DISPLAY_BRIGHTNESS_FULL:
        {
            command[0] = VFD_COMMAND_DUTY_SET |  0x03;
            command[1] = 0xFF;
            vfd_send_command(command, 2);

            vfd_set_all_lights(VFD_ALL_LIGHTS_NORMAL);
            break;
        }
        
        default:
            break;
    }
}

/**
 * Sets the highlight mode
 * @param mode- a value from the VFD_HIGHLIGHT_MODE enum
 */
void vfd_set_highlight_mode(VFD_HIGHLIGHT_MODE mode)
{
    vfd_data.highlight_mode = mode;

    VFD_HIGHLIGT_BEHAVIOUR behaviour = vfd_highlight_behaviours[vfd_data.highlight_mode];
    timers_update_timeout(TIMER_INDEX_VFD_FLASH, behaviour.flash_speed, TIMERS_TIMEOUT_RESET_IMMEDIATE);
}
    
/**
 * Simple write line to get started - no special characters.....
 */
void vfd_write_line(DISPLAY_LINE line, const char* string)
{    
    if (strncmp(vfd_data.display_lines[line], string, VFD_DISPLAY_LINE_LEN))
    {
        vfd_data.refresh_line[line] = true;
    
        size_t len = min(VFD_DISPLAY_LINE_LEN, strlen(string));

        memset(vfd_data.display_lines[line], ' ', VFD_DISPLAY_LINE_LEN);
        
        // Added in an attempt to stop spurious characters being displayed randomly
        if (vfd_data.brightness != DISPLAY_BRIGHTNESS_OFF)
        {
            memcpy(vfd_data.display_lines[line], string, len);
        }
    }
}


char* get_vfd_display_lines(void)
{
    return &vfd_data.display_lines[0][0];
}


/* *****************************************************************************
 End of File
 */
