/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Arcam

  @File Name
    adc.h

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

#ifndef _ADC_H    /* Guard against multiple inclusion */
#define _ADC_H

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdint.h>
#include <stdbool.h>
#include "app_config.h"
#include "system/ports/sys_ports.h"
#include "driver/adc/drv_adc_static.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Constants                                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * ADC_MAX_SCAN_INPUTS
 * This should be >= to the number on scan inputs supplied to adc_initialise
 * as it determines how many array elements are created. This should be defined 
 * in system_setup.h
 */
//#ifndef ADC_MAX_SCAN_INPUTS
//#define ADC_MAX_SCAN_INPUTS  1
//#endif
    
/**
 * The maximum number of switches on each scan input
 */
#define ADC_MAX_SCAN_INPUT_SWITCHES  4
    
/**
 * Value returned when a request to one of the get messages is made when the 
 * input type does not support the request
 */
#define ADC_INVALID_REQUEST -1
    
/** 
 * Values returned from a request to get temperature has been made but the data is not yet available
 */    
#define ADC_INVALID_INPUT   -2    
    

/**
 * scale ADC reading to degrees * 10
 * (actual relationship is a quadratic, but straight line approximation
 * over -10 to +110 is within 1 degree)
 */
#define ADC_MAX_READING     65536L //4294967296L     //2^32
#define ADC_TEMP_RANGE      185L            // LMT89 130 - (-55)
#define ADC_TEMP_OFFSET     130L            // LMT89 - The temperature increases as the voltage decreases.    
//#define ADC_TEMP_SF         (-27135L)/(10000L)
//#define ADC_TEMP_OFFSET     1565

/**
 * Used to define active/inactive ADC input
 */    
#define ADC_INPUT_ACTIVE    true
#define ADC_INPUT_INACTIVE  false
    
    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
    
    /*
     * Enum of available switches - Currently supports upto 9 but can be extended
     */
    typedef enum 
    {
        ADC_SWITCH_NONE,
        ADC_SWITCH_1,
        ADC_SWITCH_2,
        ADC_SWITCH_3,
        ADC_SWITCH_4,
        ADC_SWITCH_5,
        ADC_SWITCH_6,
        ADC_SWITCH_7,
        ADC_SWITCH_8,
        ADC_SWITCH_9,
        ADC_SWITCH_ERROR = -1,
    } ADC_SWITCH;
    
    /**
     * Defines the mode of the scan input e.g. whether it is to be treated as an 
     * array of buttons, or as a temperature sensor
     */
    typedef enum
    {
        ADC_SCAN_INPUT_MODE_BUTTONS,
        ADC_SCAN_INPUT_MODE_TEMPERATURE,
        ADC_SCAN_INPUT_MODE_VOLTAGE,
        ADC_SCAN_INPUT_MODE_POT,
        ADC_SCAN_INPUT_MODE_DISCRETE,
    } ADC_SCAN_INPUT_MODE;
    
    typedef struct
    {
        //! The lower value (in mV) that defines the range
        uint32_t range_low;
        
        //! The upper value (in mV) that defines the range
        uint32_t range_high;
        
        //! The switch identifier to be returned
        ADC_SWITCH adc_switch;
         
    } ADC_SCAN_INPUT_SWITCH_DEFINITION;
    
    /**
     * Scan input definition
     * An array of these are passed to adc_initialise()
     */
    typedef struct
    {
        //! Determines whether the ADC input is currently active.  If not then 
        //  it is not read or updated
        bool active;
        
        /**
         * The input to scan e.g. ADC_INPUT_SCAN_AN14
         */
        ADCHS_AN_INPUT_ID input; 
        
        /**
         * The mode determining how the input is to be treated
         */
        ADC_SCAN_INPUT_MODE mode;
        
         /**
         * The upper value (in mV) that defines the limit of the range
          * Used to check for errors / no input
         */
        uint32_t input_upper;
            
        union input_data
        {
            struct input_switch
            {
                /**
                 * An array of button definitions
                 */
                ADC_SCAN_INPUT_SWITCH_DEFINITION switches[ADC_MAX_SCAN_INPUT_SWITCHES];
            } input_switch;
            
            struct pot
            {
                // Balance data
                // Balance values: see header for calculation of initial value
                // For normal processing, we don't want the balance to be hopping between two values, if
                // the analog reading is on the boarder between two zones
                // We therefore calculate the limits for each of the zones. Note that for the top and bottom,
                // there is no adjacent value, so we can use 0 and 4095 in these cases.
                bool initialised;

                uint8_t value;
            } pot;
        } input_data;

        // Indicates whether the input has been processed
        bool processed;
        
    } ADC_SCAN_INPUT_DEFINITION;

    /**
     * Store data on the currently selected switch
     */
    typedef struct
    {
        // The switch whose details are being recorded
        ADC_SWITCH adc_switch;

        // The time the switch was depressed for
        uint32_t start_ticks;

        // The time the button was/has been depressed
        uint32_t duration_ticks;

    } ADC_SWITCH_INFO;

    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************
//    
//    /**
//     * ADC Calibration
//     * @return 
//     */
//    void adc_calibrate();
    
    /**
     * Returns a pointer to an array of ADC_SWITCH_INFO structs that contain information on the 
     * current state of a switch for each ADC channel. Where an ADC input is not 
     * a switch then the adc_switch member of the struct is ADC_SWITCH_ERROR.
     * This is intended to enable the detection of multiple concurrent switch depressions,
     * combined with "sticky" switch functionality meaning a switch remains active for
     * a second after being pressed.
     * @return 
     */
    ADC_SWITCH_INFO* adc_get_switches();
    
    /**
     * Gets the currently selected switch
     * @return 
     */
    ADC_SWITCH adc_get_current_switch();
    
    /**
     * Gets Whether the input is currently active or not 
     * @return - Whether the input is currently active or not 
     */
    //bool adc_get_input_active(unsigned int input_index);
    
    /**
     * Gets the type of ADC input 
     * @return - Whether ADC input mode (Buttons, Temp, etc..)
     */
    ADC_SCAN_INPUT_MODE adc_get_input_mode(unsigned int input_index);
     
    /**
     * Gets Whether the input is currently active or not 
     * @return - Whether the input is currently active or not 
     */
    bool adc_get_input_processed(unsigned int input_index);
    
    /**
     * Gets the calculated temperature on the supplied input.
     * @param input_index - The index of the input to be reported on
     * @return - The calculated temperature.  If the input != ADC_SCAN_INPUT_MODE_TEMPERATURE
     * it returns ADC_INVALID_REQUEST
     */
    int16_t adc_get_input_temperature(unsigned int input_index);

    /**
     * Gets the index of the Scan input that the button is associated with
     * @param button
     */
    int adc_get_input_for_switch(ADC_SWITCH button);
    
    /**
     * Gets the calculated voltage on the supplied input.
     * @param input_index - The index of the input to be reported on
     * @return - The calculated voltage in mV.  
     */
    uint32_t adc_get_input_mV(unsigned int input_index);
    
    /**
     * Gets the previously selected switch
     * Normally this will be ADC_SWITCH_NONE when a new switch is pressed, 
     * and ADC_SWITCH_??? when the new switch is ADC_SWITCH_NONE
     * @return 
     */
    ADC_SWITCH adc_get_previous_switch();

    /** 
     * Initialise the ADC
     * This MUST be called before any other of the ADC commands are used.
     * Note: ADC_MAX_SCAN_INPUTS should be set to >= the number of scan inputs defined 
     * here
     */
    void adc_initialise(ADC_SCAN_INPUT_DEFINITION scanInputs[], int num_inputs);
  
    /**
    * Perform ADC processing
     * 
    * @return - true if samples processed, false if not
    */
    bool adc_process(void);
    
    /**
     * Gets Whether the input is currently active or not 
     * @return - Whether the input is currently active or not 
     */
    void adc_set_input_active(unsigned int input_index, bool active);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _ADC_H */

/* *****************************************************************************
 End of File
 */
