/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    input_context.h

  @Summary
    Maintains the state of the input controls in a given cycle

  @Description
 * Will be updated by the application during the processing of inputs in a single cycle
 * and is then used to perform any appropriate actions.
 * This adds monitoring of input change state
 */
/* ************************************************************************** */

#ifndef _INPUT_CONTEXT_H    /* Guard against multiple inclusion */
#define _INPUT_CONTEXT_H


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include <stdint.h>
#include <stdbool.h>
#include "utilities/adc/adc_module.h"
#include "utilities/encoder_multi.h"
#include "utilities/structured_message.h"

/* Provide C++ Compatibility */
#ifdef __cplusplus
extern "C" {
#endif


    /* ************************************************************************** */
    /* ************************************************************************** */
    /* Section: Constants                                                         */
    /* ************************************************************************** */
    /* ************************************************************************** */

    // *****************************************************************************
    // *****************************************************************************
    // Section: Data Types
    // *****************************************************************************
    // *****************************************************************************
     
    // *****************************************************************************
    // *****************************************************************************
    // Section: Interface Functions
    // *****************************************************************************
    // *****************************************************************************

    /**
     * Gets the current button
     */
    ADC_BUTTON input_context_get_button();

    /**
     * Gets the current encoder state
     */
    ENCODER_STATE input_context_get_encoder(ENCODER_INDEX index);
    
    /**
     * Gets the last processed external message (an initialised message if a reset has been done)
     */
    STRUCTURED_MESSAGE* input_context_get_external_message();
    
    /**
     * Gets the current remote state
     */
    STRUCTURED_MESSAGE* input_context_get_remote_message();

    /**
     * Used to indicate that an external message has been received (used for redrawing
     */
    void input_context_external_message_processed(STRUCTURED_MESSAGE* message);
    
    /**
     * Indicates whether the balance pot has been moved
     */
    bool input_context_has_balance_changed();
    
    /**
     * Indicates whether the context button has changed assuming an attempt is 
     * made to set it each cycle
     * @return 
     */
    bool input_context_has_button_changed();

    /**
     * Indicates whether the context has changed since the last reset
     * @return 
     */
    bool input_context_has_changed();

    /**
     * Indicates whether the remote has changed since the last reset
     * @return 
     */
    bool input_context_has_remote_changed();

    /**
     * Resets the changed state of the context - should be called at the start of a cycle
     * Will not change current button/encoder state (see input_context_clear())
     */
    void input_context_reset();

    /**
     * Sets whether th (SA750) Balance pot has changed
     */
    void input_context_set_balance_changed(bool changed);
    
    /**
     * Sets the current button
     * @param button
     */
    void input_context_set_button(ADC_BUTTON button);

    /**
     * Sets the current encoder state
     * @param state
     * @param moved 
     */
    void input_context_set_encoder(ENCODER_STATE* state, ENCODER_INDEX index);
    
    /**
     * Sets the remote structured message
     * @param mode - A structured message containing information from the remote
     */
    void input_context_set_remote_message(STRUCTURED_MESSAGE* message);

    /**
     * Suspends/Unsuspends button processing 
     * @param suspend
     */
    void input_context_suspend_button(bool suspend);

    /**
     * Suspends/Unsuspends remote processing 
     * @param suspend
     */
    void input_context_suspend_remote(bool suspend);
    
    /* Provide C++ Compatibility */
#ifdef __cplusplus
}
#endif

#endif /* _EXAMPLE_FILE_NAME_H */

/* *****************************************************************************
 End of File
 */
