/*******************************************************************************
  MPLAB Harmony Application Source File
  
  Company:
    Microchip Technology Inc.
  
  File Name:
    app.c

  Summary:
    This file contains the source code for the MPLAB Harmony application.

  Description:
    This file contains the source code for the MPLAB Harmony application.  It 
    implements the logic of the application's state machine and it may call 
    API routines of other MPLAB Harmony modules in the system, such as drivers,
    system services, and middleware.  However, it does not call any of the
    system interfaces (such as the "Initialize" and "Tasks" functions) of any of
    the modules in the system or make any assumptions about when those functions
    are called.  That is the responsibility of the configuration-specific system
    files.
 *******************************************************************************/

// DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright (c) 2013-2014 released Microchip Technology Inc.  All rights reserved.

Microchip licenses to you the right to use, modify, copy and distribute
Software only when embedded on a Microchip microcontroller or digital signal
controller that is integrated into your product or third party product
(pursuant to the sublicense terms in the accompanying license agreement).

You should refer to the license agreement accompanying this Software for
additional information regarding your rights and obligations.

SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER
CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR
OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
(INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 *******************************************************************************/
// DOM-IGNORE-END


// *****************************************************************************
// *****************************************************************************
// Section: Included Files 
// *****************************************************************************
// *****************************************************************************

#include "app.h"
#include "system_definitions.h"
#include "app_commands.h"
#include "app_commands_avatar.h"
#include "app_commands_control4.h"
#include "app_commands_crestron.h"
#include "app_commands_hdmi_arc.h"
#include "app_commands_power.h"
#include "app_commands_spdif_rx.h"
#include "app_commands_standby.h"
#include "app_commands_system_fault.h"
#include "app_commands_upgrade.h"
#include "app_config.h"
#include "app_context.h"
#include "app_initialise.h"
#include "utilities/adc/adc.h"
#include "utilities/avatar.h"
#include "utilities/display.h"
#include "utilities/display_test.h"
#include "utilities/hdmi_arc_processing.h"
#include "utilities/i2c_slave.h"
#include "utilities/input_control.h"
#include "utilities/logger/logger.h"
#include "utilities/machine_uuid.h"
#include "utilities/power.h"
#include "utilities/rgb_led.h"
#include "utilities/rs232.h"
#include "utilities/settings.h"
#include "utilities/structured_message.h"
#include "utilities/system_fault.h"
#include "utilities/timers.h"
#include "utilities/upgrade_metadata.h"
#include "utilities/volume.h"
#include "utilities/watchdog.h"

//#include "framework/driver/spi/static/drv_spi_static.h"


// *****************************************************************************
// *****************************************************************************
// Section: Global Data Definitions
// *****************************************************************************
// *****************************************************************************


static bool log_app_running = true;

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Functions
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
// *****************************************************************************
// Section: Application Local Functions
// *****************************************************************************
// *****************************************************************************

// Forward declaration
static void message_logged(LOGGER_MESSAGE* message);
static void structured_message_callback(STRUCTURED_MESSAGE* message);

/**-----------------------------------------------------------------------------
 * Ensures the Logger Type is within bounds
 */
static void check_logger_type()
{
    // Set the logger level - Note before this all logger messages will be 
    // output - this is so that if there are initialisation issues they will 
    // be logged irrespective of the previous logging level
    uint8_t logger_type;
    settings_get(SETTING_LOGGER_TYPE, &logger_type);
    if (logger_type > LOGGER_TYPE_ALL)
    {
        logger_type = LOGGER_TYPE_DEBUG;
    }
    logger_set_level(logger_type);
    logger_set_level(LOGGER_TYPE_DEBUG);  //TODO Remove when up and running
}

/**-----------------------------------------------------------------------------
 * Handles messages generated by the logger
 * @param message
 */
static void message_logged(LOGGER_MESSAGE* message)
{
    // Convert the logger message to a structured message
//    STRUCTURED_MESSAGE output_message;
//    output_message.command = APP_COMMANDS_LOGGER_MESSAGE;
//    output_message.zone = STRUCTURED_MESSAGE_ZONE_1;
//    output_message.response = STATUS_UPDATE;
//    output_message.data_only = false;
//    
//    // wrap up in logger_pack_message()?)
//    int len = sizeof(LOGGER_MESSAGE) + strlen(message->message);
//    char buffer[len];
//    memcpy(&buffer, &message->code, sizeof(LOGGER_CODE));
//    memcpy(&buffer[2], &message->type, sizeof(LOGGER_TYPE));
//    memcpy(&buffer[4], &message->message[0], min(strlen(message->message), STRUCTURED_MESSAGE_MAX_DATA));
//    buffer[len - 1] = 0;
//    
//    structured_message_set_message_data(&output_message, buffer, len);
    
    // Convert the logger message to a structured message
    int len = 0;
    STRUCTURED_MESSAGE output_message;
    output_message.zone = STRUCTURED_MESSAGE_ZONE_1;
    output_message.response = STATUS_UPDATE;
    output_message.data_only = false;
     
    if (message->type != LOGGER_TYPE_COMMAND)
    {
        output_message.command = APP_COMMANDS_LOGGER_MESSAGE;

        // wrap up in logger_pack_message()?)
        len = sizeof(LOGGER_MESSAGE) + strlen(message->message);// - 1; // -1 = size of buffer pointer in message
        char buffer[len];
        memcpy(&buffer, &message->code, sizeof(LOGGER_CODE));
        memcpy(&buffer[2], &message->type, sizeof(LOGGER_TYPE));
        memcpy(&buffer[4], &message->message[0], min(strlen(message->message), STRUCTURED_MESSAGE_MAX_DATA));
        buffer[len - 1] = 0;
    
        structured_message_set_message_data(&output_message, buffer, len);
    }
    else
    {
        // Construct a command message
        output_message.command = message->code;     // Using code element to hold the command
    
        // wrap up in logger_pack_message()?)
        len = message->length;
        char buffer[len];
        memcpy(&buffer, message->message, message->length);
    
        structured_message_set_message_data(&output_message, buffer, len);
    }

    rs232_send_message(&output_message);
    
    // Send to the first TCP/IP client
    output_message.state_data[0] = 0;
    //TODO put back - todo with avatar tcp send tcpip_send_structured_message(&output_message);
}


/**-----------------------------------------------------------------------------
 * Sets up the flashing lights for the fault display
 */
void set_fault_display()
{
    switch(system_fault_get().code)
    {
        case SYSTEM_FAULT_CODE_DC_OFFSET:
        {
            rgb_led_set(RGB_LED_COLOUR_ORANGE);
            rgb_led_flash();
            break;
        }
        
        case SYSTEM_FAULT_CODE_SHORT_CIRCUIT:
        {
            rgb_led_set(RGB_LED_COLOUR_RED);
            rgb_led_flash();
            break;
        }
        
        case SYSTEM_FAULT_CODE_OUTPUT_OVER_TEMP:
        {
            rgb_led_set(RGB_LED_COLOUR_BLUE);
            rgb_led_flash();
            break;
        }
        
        case SYSTEM_FAULT_CODE_LIFTER_OVER_TEMP:
        {
            rgb_led_set(RGB_LED_COLOUR_BLUE);
            rgb_led_flash();
            break;
        }
        
        default:
            break;
    }
}

/**-----------------------------------------------------------------------------
 * A number of the app states have some standard processing which is wrapped up here
 */
static void standard_processing()
{
    // Reset the input context
    input_context_reset();

    // Perform I2C Request checking
    i2c_process();

    // RGB LEd
    rgb_led_process();

    rs232_process();


#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
    rs232_hdmi_arc_process();
#endif

    flash_memory_process();

    // check whether we need to save the current state of the settings
    settings_process();

    input_control_process();

    // Output control processing
    output_control_process();
    
    // Avatar processing
    avatar_process();
    
    app_commands_avatar_process();
    
    // HDMI ARC processing to ensure handshaking
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
    app_commands_hdmi_arc_process();
#endif
    
    app_commands_process_lip_sync();
    
    // Clear the watch dog
    watchdog_reset();
    
    logger_process();
}

/**-----------------------------------------------------------------------------
 * Updates the splash screen during power up
 */
static void update_splash_screen()
{
    // Update the Splash screen display
    stateful_process_invoke();
           
    if (display_test_needs_redraw())
    {
        menu_display_process();
        display_refresh();
    }
}

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/**-----------------------------------------------------------------------------
 * 
 * @param event
 * @param context
 */
//void APP_I2C_SlaveTransfer_Callback(DRV_I2C_BUFFER_EVENT event, void * context)
//{
//    //logger_log_debugf(0, "AVATAR CALLBACK");
//    avatar_i2c_slave_callback(event, context);
//}
	

/*******************************************************************************
  Function:
    void APP_Initialize ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Initialize ( void )
{
    // Check which type of spdif receiver we are using and call the appropriate 
    // define_callback methods;
    // Note this MUST be done before th initialise is called
    if (app_commands_spdif_model_select())
    {
        spdif_rx_define_functions(spdif_rx_hw_ak4113_define_functions());
        input_adc_define_functions(input_adc_hw_ak5552_define_functions());
    }
    else
    {
        spdif_rx_define_functions(spdif_rx_hw_tipcm9211_define_functions());
        input_adc_define_functions(input_adc_hw_tipcm4202_define_functions());
    }
                
    // Ensure the VFD does not show spurious characters
    vfd_hold_in_reset();               
    
    // Put SPDIF MUX/ADC/DAC in Reset
    input_adc_reset(INPUT_CONTROL_SWITCH_OFF);
    input_control_dac_reset(INPUT_CONTROL_SWITCH_OFF);
    spdif_rx_reset(INPUT_CONTROL_SWITCH_OFF);    
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
    hdmi_arc_reset(false);
#endif
 
    // Enable the Avatar PSU up front to allow avatar to get going asap
    power_enable_by_index(POWER_INDEX_STREAMER_PSU);
   
    /* Place the App state machine in its initial state. */
    app_context_initialise();
    
    input_context_set_button(ADC_BUTTON_NONE);
    
    app_commands_standby_initialise();
}

static bool test = false;
/******************************************************************************
  Function:
    void APP_Tasks ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Tasks ( void )
{
    /* Check the application's current state. */
    switch ( app_context_get_app_state() )
    {
        /* Application's initial state. */
        case APP_STATE_INIT:
        {
            initialise_rgb_led();
                
            // Initialise the logger
            initialise_logger(message_logged);
            
            // Initialise the mutes
            initialise_mutes();
            
            // Initialise the timers
            initialise_timers();
            
            // Initialise RS232 port
            initialise_rs232();
            
            // Initialise flash_memory
            initialise_flash_memory();
            
            // Initialise the settings
            initialise_settings(&initialise_settings_defaults);  
            
            // Initialise I2C
            initialise_i2c();

            // Initialise the menu system
            app_menu_initialise();
            
            // Avatar initi
            initialise_avatar();
            
            // Initialise upgrade info
            upgrade_metadata_initialise(app_commands_upgrade_write_callback, NULL);
            
            rgb_led_process();
            
            app_context_set_app_state(APP_STATE_LOADING_SETTINGS);
            
            logger_log_debugf(0, "APP: Loading Settings ");
    
            logger_process();
            
            break;
        }

        case APP_STATE_LOADING_SETTINGS:
        {         
            standard_processing();
             
            if (settings_initialised())
            {
                logger_log_info(LOGGER_CODE_INITIALISATION, "APP: Settings initialised --");

                app_context_set_app_state(APP_STATE_READ_SETTINGS);
            }
            
            break;
        }
        
        case APP_STATE_READ_SETTINGS:
        {
            logger_log_debugf(LOGGER_CODE_INITIALISATION, "APP: Reading Settings --");
            
            standard_processing();
            
            // Software Version checking
            initialise_software_version();

            system_fault_initialise();
            
            //check_logger_type(); Moved to APP_STATE_READY so that we get all the startup message output

            // Enabling power
            // This resets the TIMER_INDEX_POWER_CYCLE and so ensures that we 
            // wait for the power to come up (see APP_STATE_WAIT_FOR_POWER below))
            initialise_power();
   
            // Initialise input control
            initialise_input_control();

            // Setup inputs - MUST Be done after settings are loaded
            app_commands_input_configure();
            
            app_commands_process_controls();
            
            // Wait for VFD to sort itself out
            initialise_display();
            app_commands_splash_screen_start();

            app_context_set_app_state(APP_STATE_WAIT_FOR_POWER);
            
            break;
        }
        
        case APP_STATE_WAIT_FOR_POWER:
        {
            standard_processing();
            
            app_commands_process_controls();
            
            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
            {
                timers_reset(TIMER_INDEX_POWER_CYCLE);
                
                app_context_set_app_state(APP_STATE_POWERUP);
            }

            // Update the Splash screen display
            update_splash_screen();
            
            break;
        }
        
        case APP_STATE_POWERUP:
        {
            logger_log_debugf(LOGGER_CODE_INITIALISATION, "APP: Power-up --");
            
            standard_processing();
            
            app_commands_process_controls();

            app_context_module_running(APP_MODULE_POWER);
            
            app_context_set_app_state(APP_STATE_COMPLETE_INITIALISE);
            
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
            hdmi_arc_reset(true);
#endif

            //app_commands_hdmi_arc_system_ready();

            // Update the Splash screen display
            update_splash_screen();
            
            break;
        }
        
        case APP_STATE_COMPLETE_INITIALISE:
        {
            logger_log_debugf(LOGGER_CODE_INITIALISATION, "APP: Complete initialise --");
            
            standard_processing();
            
            machine_uuid_load();
            
            // Initialise output control
            output_control_initialise(&system_fault_set);
            
            initialise_network(app_commands_structured_message_callback);
            
            app_commands_process_controls();
            
            // Initialise the ADC
            initialise_adc();
            
#if APP_FEATURE_SPECIAL_SWITCHES == APP_FEATURE_ON    
            initialise_special_switches();
#endif

            logger_log_info(LOGGER_CODE_INITIALISATION, "Initialising Volume --");
            volume_initialise();
                        
            // Initialise Pins for standby
            initialise_pins();
            
            // Reset standby mode
            standby_reset();
            
            // Initialise the encoder
            initialise_encoder();

            // Initialise the remote
            logger_log_info(LOGGER_CODE_INITIALISATION, "Initialising Remote --");
            remote_initialise(&app_commands_structured_message_callback);

            // Now I2C is up and running start input control configuration
            input_control_start_configuration();
            
            // Initialise the watchdog
            watchdog_initialise();

            app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_ORANGE);

            ///////app_context_set_app_state(APP_STATE_WAIT_FOR_AVATAR);
            app_context_set_app_state(APP_STATE_WAIT_READY);

            // Update the Splash screen display
            update_splash_screen();

            break;
        }
        
        case APP_STATE_WAIT_READY:
        {
            standard_processing();
            
            // Now perform any processing based on the current state of the input
            // in either the buttons or the rotary, or something via the IR or rs232
            app_commands_process_controls();

            // We are waiting for Power/Avatar and Input Control to be ready
            if (app_context_is_running())
            {
                app_context_set_app_state(APP_STATE_READY);
            }
            
            app_commands_check_display(); 

            // Update the Splash screen display
            update_splash_screen();
            
            break;
        }
        
        case APP_STATE_READY:
        {
            logger_log_debugf(LOGGER_CODE_INITIALISATION, "APP: Initialise complete --");
            
            standard_processing();

            // Required when coming out of sleep for some reason!
            app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_WHITE);

            // Ensure the Avatar has the correct Balance setting
            input_control_set_avatar_balance(volume_get_balance());

            // Now perform any processing based on the current state of the input
            // in either the buttons or the rotary, or something via the IR or rs232
            app_commands_process_controls();

            input_control_dac_reset(INPUT_CONTROL_SWITCH_ON);

            // Reset the TIMER_INDEX_HDMI_ARC_HANDSHAKE timer to trigger a handshake
            // This needs doing when it is initially started and when it comes out of standby if network standby is not on..
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
            if (((hdmi_arc_get_firmware_version_major() == 0) && (hdmi_arc_get_firmware_version_minor() == 0)) ||
                  !power_get_network_active_in_standby())
            {
                hdmi_arc_initiate_handshake();
            }
#endif
            
            //  Check whether we need to return to sleep after the upgrade.
            bool sleep = false;
            settings_get(SETTING_UPGRADE_SLEEP_ON_RESTART, &sleep);

            if (!sleep)
            {
                app_context_set_app_state(APP_STATE_RUNNING);
            }
            else
            {
                app_context_set_app_state(APP_STATE_POWERDOWN);
            }
            
            sleep = false;
            settings_set(SETTING_UPGRADE_SLEEP_ON_RESTART, &sleep);
            settings_save();

            // Update the Splash screen display
            update_splash_screen();

            // Start the watchdog timer service
            watchdog_enable();
            
            // Attempt to get the Watchdog prescalar out - appears to be a factor of 2 out..
            uint32_t test = DEVCFG1bits.WDTPS;
            logger_log_debugf(LOGGER_CODE_INITIALISATION, "APP: WDT Pre-scaler %d  --", test);
            
            check_logger_type();

            break;
        }
        
        case APP_STATE_RUNNING:
        {
            if (log_app_running)
            {
                logger_log_debugf(LOGGER_CODE_INITIALISATION, "APP RUNNING --");
                log_app_running = false;
            }
            
            standard_processing();    
            
            // Now perform any processing based on the current state of the input
            // in either the buttons or the rotary, or something via the IR or rs232
            app_commands_process_controls();

            // Process volume ramping
            volume_process();
             
            avatar_browsing_process();
             
            // Do checks for power down, auto shutdown etc...
            app_commands_check_power_state();

            // Check for any faults
            app_commands_check_for_fault();
            
            app_commands_check_display(); 
            
            // Finally refresh the display in whatever state its in
            // Note if this or any of the refresh immediate parameters are called
            // earlier then the screen will flash...
            // Update the display (basically menu handling))
            menu_display_process();
            display_refresh();
            
            break;
        }
        
        case APP_STATE_POWERDOWN:
        {
            logger_log_infof(LOGGER_CODE_POWER, "POWER Informing Avatar of Power down --");

            app_commands_set_power_state_transition();
            
            // BUG 3243
            avatar_send_player_control(AVATAR_PLAYER_CONTROL_STOP);

            avatar_send_power_state(AVATAR_POWER_STATE_STANDBY);
            
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
            if (!power_get_network_active_in_standby())
            {
                // Put the avatar into the standby state
                //avatar_send_power_state(AVATAR_POWER_STATE_STANDBY);
                hdmi_arc_send_power_message(HDMI_ARC_POWER_STATE_OFF);
            }
            else
            {
                hdmi_arc_send_power_message(HDMI_ARC_POWER_STATE_IDLE);
            }
#endif
                       
            // Set the LED to red if no faults to report
            SYSTEM_FAULT fault = system_fault_get();
            if (fault.code == SYSTEM_FAULT_CODE_NO_FAULT)
            {
                rgb_led_set(RGB_LED_COLOUR_RED);
                rgb_led_flash_stop();
            }
            else
            {
                rgb_led_set(RGB_LED_COLOUR_OFF);
                rgb_led_flash_stop();
            }

            // BUG 3762 
            bool mute_on_startup = volume_is_muted();
            logger_log_infof(LOGGER_CODE_INITIALISATION, "APP: Storing MUTE_ON_STARTUP %d --", mute_on_startup);
            settings_set(SETTING_MUTE_ON_STARTUP, &mute_on_startup);
            settings_save();

            timers_reset(TIMER_INDEX_AVATAR_SHUTDOWN);
            
            app_context_set_app_state(APP_STATE_WAIT_FOR_AVATAR_ROON_POWERDDOWN);
            
            break;
        }
        
        case APP_STATE_WAIT_FOR_AVATAR_ROON_POWERDDOWN:
        {
            //logger_log_info(LOGGER_CODE_POWER, "STANDBY Waiting for Avatar to complete processing --");
            
            standard_processing();
            
            // Hide the display
            app_commands_display_clear();
            display_refresh();

            if (timers_has_expired(TIMER_INDEX_AVATAR_SHUTDOWN))
            {
                logger_log_info(LOGGER_CODE_POWER, "STANDBY Moving to APP_STATE_WAIT_FOR_AVATAR_POWERDDOWN --");

                timers_reset(TIMER_INDEX_AVATAR_SHUTDOWN);
                app_context_set_app_state(APP_STATE_WAIT_FOR_AVATAR_POWERDDOWN);
            }
            
            logger_process();

            break;
        }
        
        case APP_STATE_WAIT_FOR_AVATAR_POWERDDOWN:
        {
            logger_log_info(LOGGER_CODE_POWER, "STANDBY Waiting for Avatar to complete processing --");
            
            // Reset the input context
            input_context_reset();

            // Perform I2C Request checking
            i2c_process();

            // Avatar processing
            avatar_process();

            // Clear the watch dog
            watchdog_reset();

            logger_process();
            
            // Hide the display
            app_commands_display_clear();
            display_refresh();
            
            if (!timers_has_expired(TIMER_INDEX_AVATAR_SHUTDOWN))
            {
                if (!avatar_is_active() && (app_commands_get_player_status() == AVATAR_PLAYER_STATUS_STOPPED))
                {
                    logger_log_info(LOGGER_CODE_POWER, "STANDBY Moving to POWEREDDOWN --");

                    app_context_set_app_state(APP_STATE_PRE_POWEREDDOWN);
                }
            }
            else
            {
                logger_log_info(LOGGER_CODE_POWER, "STANDBY: Timedout waiting for Avatar moving to POWEREDDOWN --");

                app_commands_force_player_status_stopped();
                
                app_context_set_app_state(APP_STATE_PRE_POWEREDDOWN);
            }
    
            logger_process();

            break;
        }
        
        case APP_STATE_PRE_POWEREDDOWN:
        {
            logger_log_info(LOGGER_CODE_POWER, "STANDBY: Moving to POWEREDDOWN --");

            rgb_led_process();

            // Set the dac into reset mode during power up
            input_control_dac_reset(INPUT_CONTROL_SWITCH_OFF);
            spdif_rx_reset(INPUT_CONTROL_SWITCH_OFF);
            input_adc_reset(INPUT_CONTROL_SWITCH_OFF);

            avatar_reset_queues();
            
            // Perform the power down
            app_commands_power_down();

            // Hide the display
            //app_commands_display_clear();
            display_refresh();

            timers_delay_ms(1);

            // Ensure the VFD does not show spurious characters 
            vfd_hold_in_reset();

//            // Set the LED to red if no faults to report
//            SYSTEM_FAULT fault = system_fault_get();
//            if (fault.code == SYSTEM_FAULT_CODE_NO_FAULT)
//            {
//                rgb_led_set(RGB_LED_COLOUR_RED);
//                rgb_led_flash_stop();
//            }
//            else
//            {
//                rgb_led_set(RGB_LED_COLOUR_OFF);
//                rgb_led_flash_stop();
//            }

            // Use power timer to prevent waking up too quickly
            timers_update_timeout(TIMER_INDEX_POWER_CYCLE, 5000, TIMERS_TIMEOUT_RESET_IMMEDIATE);

            // Set the state to powered down
            app_context_set_app_state(APP_STATE_POWEREDDOWN);

            pins_save_state();
            
            // Disabling the watchdog during sleep.  The reason for this was that 
            // the watchdog postscalar was set to its largest value (17mins) (not
            // by design) and that caused the resting voltage to increase when 
            // the 17 min mark was reached..
            watchdog_disable();
            
            logger_process();
            
            break;
        }
        
        case APP_STATE_POWEREDDOWN:
        {
            app_commands_clear_power_state_transition();
            
            app_commands_standby_powereddown();
    
            logger_process();

            break;
        }
        
        case APP_STATE_WAKEUP:
        {
            logger_log_info(LOGGER_CODE_POWER, "STANDBY: Waking up --");

            app_commands_set_power_state_transition();            
            
            pins_restore_state();

            // Following discussion with Pete - there is no point in trying to do a controlled restart
            // might as well just reboot.....Note - this will invalidate some of the other code in the wakeup process..
            if (!power_get_network_active_in_standby())
            {
                logger_log_debugf(0, "RESTARTING...................... --");
                logger_process();
                SYS_RESET_SoftwareReset();
            }
                        
            // Clear any fault codes
            SYSTEM_FAULT fault =
            {
                .code = SYSTEM_FAULT_CODE_NO_FAULT,
                .description = "",
            };
            system_fault_set(&fault);

            app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_ORANGE);
            rgb_led_flash_stop();
            
            rgb_led_process();

            input_context_reset();
            
            // Power up
            power_enable();
            
            // Send unsolicited power message
            app_commands_request_status(APP_COMMANDS_POWER, app_commands_power);
            
            // To pause before resetting the VFD
            timers_delay_ms(1);
            vfd_clear_reset();
            
            // Now perform any processing based on the current state of the input
            // in either the buttons or the rotary, or something via the IR or rs232
            app_commands_process_controls();

            // Start wait timer
            timers_reset(TIMER_INDEX_POWER_CYCLE);
            
            watchdog_enable();

            // Clear the watch dog
            watchdog_reset();
           
            app_context_set_app_state(APP_STATE_WOKENUP);
    
            logger_process();
            
            break;
        }
        
        case APP_STATE_WOKENUP:
        {
            // Perform I2C Request checking
            i2c_process();
            
            // Wait 3 seconds for proper power up
            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
            {
                logger_log_info(LOGGER_CODE_POWER, "STANDBY: Woken up --");
                
                input_context_reset();
                    
                // Now perform any processing based on the current state of the input
                // in either the buttons or the rotary, or something via the IR or rs232
                app_commands_process_controls();

                display_initialise();
                
                display_reset_brightness();

                app_commands_adjust_brightness(APP_COMMAND_BRIGHTNESS_FULL);
                app_commands_check_display();

                app_commands_display_clear();
                
                // Ensure max on volume is checked when setting volume
                volume_check_max_on_volume();
                
                // reset the power down timeout
                timers_update_timeout(TIMER_INDEX_POWER_CYCLE, 3000, TIMERS_RESET_NONE);
 
                // restart the auto shutdown timer
                power_reset_auto_shutdown_timer();
                
                // Set the display to the default screen
                app_commands_display_default_screen(NULL);

                watchdog_enable();
                  
                // Initialise input control and reset the DAC
//****                input_control_dac_reset(INPUT_CONTROL_SWITCH_ON);
                spdif_rx_reset(INPUT_CONTROL_SWITCH_ON);            
                input_adc_reset(INPUT_CONTROL_SWITCH_ON);
                        
                // Get the Avatar going again
                app_comands_avatar_invalidate_info();
                
                if (!power_get_network_active_in_standby())
                {
                    avatar_send_power_state(AVATAR_POWER_STATE_ONLINE);
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                    hdmi_arc_send_power_message(HDMI_ARC_POWER_STATE_ON);
#endif
                    avatar_wakeup();

                    //TODO 20191001 - Check this is OK when Network is active in standby and move out if this if statement - suspect ir won't wake up the avatar
                    app_context_clear_running();
                    app_context_module_running(APP_MODULE_POWER);

                    //Done by the above app_---------------
                    app_context_module_reset(APP_MODULE_AVATAR);
                    app_context_set_app_state(APP_STATE_WAIT_READY);
                    
                    logger_log_debugf(0, "WAKEUP to APP_STATE_WAIT_READY --");
                }
                else
                {
                    app_context_set_app_state(APP_STATE_READY);
                    
//                    input_control_reset_input();
                    
                    logger_log_debugf(0, "WAKEUP to APP_STATE_READY --");
                }
                    
                //logger_log_debugf(0, "WAKEUP SELECTED INPUT %d %d --", input_control_get_selected_input(), input_control_get_previous_input());
                if (input_control_get_selected_input() != AUDIO_INPUT_AVATAR)
                {
                    input_control_reset_input();
                }
                else
                {
                    // It was woken up by the aar and the input has been set in STANDBY_STATE_ACTIVE
                    if (app_commands_get_player_status() != AVATAR_PLAYER_STATUS_STOPPED)
                    {
                        input_control_reset_input();
                    }
                    else
                    {
                        input_control_select_input(input_control_get_previous_input()); 
                    }
                }
 
                app_commands_clear_power_state_transition();
                
                //input_control_reset_input();
                
                // Display splash logo screen
                app_commands_splash_screen_start();
            }

            menu_display_process();
            display_refresh();
  
            // Clear the watch dog
            watchdog_reset();
    
            logger_process();
           
            break;
        }
        
        case APP_STATE_ENTERING_FAULT:
        {
            // Fault detected -0 we're trying ton save the fault status so
            // keep other processing to a minimum.
            // Stop the watchdog timer service
            watchdog_disable();

            // Put ARC in reset and powerdown
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
            hdmi_arc_reset(false);
#endif
            power_disable_by_index(POWER_INDEX_ARC_IC_PSU);
            
            // RGB LEd
            rgb_led_process();

            rs232_process();

            flash_memory_process();

            // check whether we need to save the current state of the settings
            settings_process();
            
            app_commands_check_for_fault();
            
            logger_process();
            
            break;
        }
        
        case APP_STATE_FAULT:
        {
            // Stop the watchdog timer service
            watchdog_disable();
            
            // Mute speakers/pre-amp and headphones
            output_control_mute_outputs(true);

            // Set the current fault
            set_fault_display();
            
            // Display the current fault
            app_commands_display_system_fault();
            
            // Kick off the timer controlling the pause while we shut down
            timers_reset(TIMER_INDEX_SYSTEM_FAULT);
            
            //appData.state = APP_STATE_FAULT_WAIT;
            app_context_set_app_state(APP_STATE_FAULT_WAIT);
    
            logger_process();

            break;
        }
        
        case APP_STATE_FAULT_WAIT:
        {
            rgb_led_process();
            
            if (timers_has_expired(TIMER_INDEX_SYSTEM_FAULT))
            {
                logger_log_info(LOGGER_CODE_POWER, "STANDBY: Moving to POWEREDDOWN from FAULT --");
                
                // Power down
                app_context_set_app_state(APP_STATE_PRE_POWEREDDOWN);
            }
    
            logger_process();
            
            break;
        }
        
        case APP_STATE_ENTER_BIAS_CHECK:
        {
            logger_log_info(LOGGER_CODE_STATE_CHANGE, "APP_STATE_ENTER_BIAS_CHECK --");
            
            // Perform standard processing
            standard_processing();
            
            // Mute outputs pause for 1 second
            app_commands_enter_bias_check(0);

            app_context_set_mode(APP_CONTEXT_MODE_BIAS_TEST);
            app_context_set_menu(APP_COMMAND_MENU_BIAS_CHECK);
            
            timers_reset(TIMER_INDEX_BIAS_TEST);
            
            // Update the display (basically menu handling))
            menu_display_process();
            display_refresh();
        
            app_context_set_app_state(APP_STATE_BIAS_CHECK);
            
            break;
        }
        
        case APP_STATE_BIAS_CHECK:
        {
             // Perform standard processing
            standard_processing();
            
            app_commands_process_controls();

            app_commands_perform_bias_check();            
            
            // Update the display (basically menu handling))
            menu_display_process();
            display_refresh();
  
            // Clear the watch dog
            watchdog_reset();

            break;
        }
        
        case APP_STATE_EXIT_BIAS_CHECK:
        {
            logger_log_info(LOGGER_CODE_STATE_CHANGE, "APP_STATE_EXIT_BIAS_CHECK --");
            
            // Perform standard processing
            standard_processing();
            
            app_commands_exit_bias_check(0);
            
            app_commands_display_default_screen(NULL);
             
            // Update the display (basically menu handling))
            menu_display_process();
            display_refresh();
    
            app_context_set_app_state(APP_STATE_RUNNING);
            break;
        }

        case APP_STATE_PRE_RESTART:
        {
            // Restart after an avatar upgrade
            standard_processing();
            
            if (timers_has_expired(TIMER_INDEX_RESTART_PAUSE))
            {
                avatar_stop();
                power_disable(); 
                app_context_set_app_state(APP_STATE_RESTART);
            }

            break;
        }
        
        case APP_STATE_RESTART:
        {
            
            SYS_RESET_SoftwareReset();
            break;
        }

        /* The default state should never be executed. */
        default:
        {
            break;
        }
    }
}

/*******************************************************************************
 End of File
 */
