/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    app_commands.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands.h"
#include "app_commands_display.h"
#include "app_commands_engineering.h"
#include "app_commands_input.h"
//#include "app_commands_menu.h"
#include "app_commands_power.h"
#include "app_commands_processor_mode.h"
#include "app_commands_system.h"
#include "app_commands_volume.h"
#include "app_config.h"
#include "app_context.h"
#include "device_configuration.h"
#include "utilities/app_commands_utilities.h"
#include "utilities/avatar_browsing.h"
#include "utilities/dirac.h"
#include "utilities/display_test.h"
#include "utilities/encoder_multi.h"
#include "utilities/input_context.h"
#include "utilities/input_control.h"
#include "utilities/menu_app.h"
#include "utilities/remote.h"
#include "utilities/structured_message.h"
#include "utilities/timers.h"
#include "utilities/volume.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    bool wants_reset;
} APP_COMMANDS_DATA;

static APP_COMMANDS_DATA app_commands_data = {
    .wants_reset = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Use to determine if we are displaying one of the balance menus
 * @param menu
 * @return 
 */
static bool is_balance_menu(MENU_ITEM* menu)
{
    bool result = false;
    if ((menu != NULL) && 
       ((menu->menu_id == APP_COMMAND_MENU_BALANCE) || (menu->menu_id == APP_COMMAND_MENU_BALANCE_NO_NAVIGATION)))
    {
        result = true;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs default balance pot processing
 */
#if APP_FEATURE_BALANCE_POT == APP_FEATURE_ON
static bool process_balance_default(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    if (input_context_has_balance_changed())
    {
        // Added so that pressing the current input button will light up the display
        // when brightness = off - it did not before
        timers_reset(TIMER_INDEX_BRIGHTNESS);
        
        // initialise balance at volume
        volume_set_balance(VOLUME_DEFAULT_BALANCE + adc_pot_get_offset_from_center(APP_ADC_INPUTS_BALANCE), false);
        
        // Show balance menu in edit mode
        app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
        message->command = APP_COMMANDS_MENU;
        message->data[0] = APP_COMMAND_MENU_BALANCE_NO_NAVIGATION;
        message->data_length = 1;
    }
    else
    {
        result = false;
    }
    
    return result;
}
#endif

/**-----------------------------------------------------------------------------
 * Performs default balance pot processing
 */
#if APP_FEATURE_BALANCE_POT == APP_FEATURE_ON
static bool process_balance_menu(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    if (input_context_has_balance_changed())
    {
        // Added so that pressing the current input button will light up the display
        // when brightness = off - it did not before
        timers_reset(TIMER_INDEX_BRIGHTNESS);
        
        // initialise balance at volume
        volume_set_balance(VOLUME_DEFAULT_BALANCE + adc_pot_get_offset_from_center(APP_ADC_INPUTS_BALANCE), false);
    }
    else
    {
        result = false;
    }
    
    return result;
}
#endif

/**-----------------------------------------------------------------------------
 * Performs default balance pot processing
 */
#if APP_FEATURE_BALANCE_POT == APP_FEATURE_ON
static bool process_balance_menu_edit(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    if (input_context_has_balance_changed())
    {
        // Added so that pressing the current input button will light up the display
        // when brightness = off - it did not before
        timers_reset(TIMER_INDEX_BRIGHTNESS);
        
        // initialise balance at volume
        volume_set_balance(VOLUME_DEFAULT_BALANCE + adc_pot_get_offset_from_center(APP_ADC_INPUTS_BALANCE), false);
    }
    else
    {
        result = false;
    }
    
    return result;
}
#endif

/**-----------------------------------------------------------------------------
 * Performs default button/switch processing
 */
static bool process_buttons_default(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    ADC_BUTTON button = input_context_get_button();
    
    bool proceed = true;
    
#if APP_FEATURE_SPECIAL_SWITCHES == APP_FEATURE_ON    
    if (special_switches_active_button_count() > 1)
    {
        proceed = false;
    }
#endif    
    //logger_log_debugf(0, "PROCESSING BUTTONS %d %d %d %d --", button, result, special_switches_active_button_count(), input_context_has_button_changed());
    
    if (input_context_has_button_changed() && proceed)
    {
        // Added so that pressing the current input button will light up the display
        // when brightness = off - it did not before
        timers_reset(TIMER_INDEX_BRIGHTNESS);
        
        // See if there is any bespoke processing
        result = app_device_process_buttons_default(button, message);
    
        if (!result)
        {
            // Assume we will be processing the message below
            result = true;

            switch (button)
            {
                case APP_BUTTONS_DIRAC:
                {
                    // DIRAC
                    //logger_log_debugf(0, "BUTTON Dirac");
                    // Display the DIRAC menu
                    app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                    message->command = APP_COMMANDS_MENU;
                    message->data[0] = APP_COMMAND_MENU_DIRAC_NO_NAVIGATION;
                    message->data_length = 1;

                    break;
                }

                case APP_BUTTONS_MUTE:
                {
                   // Process MUTE command 
                    logger_log_debugf(0, "BUTTON Mute --");
                    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                    message->command = APP_COMMANDS_MUTE;
                    message->data[0] = APP_COMMAND_MUTE_TOGGLE;
                    message->data_length = 1;
                    logger_log_debugf(0, "APP: Toggling Mute --");
                    break;
                }

                case APP_BUTTONS_DIRECT:
                {
                    logger_log_debugf(0, "BUTTON Direct --");
    //                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
    //                message->command = APP_COMMANDS_ANALOGUE_DIRECT; // Will actually Toggle
    //                message->data[0] = input_control_get_selected_input();
    //                message->data[1] = !input_control_get_analogue_direct();
    //                message->data_length = 2;

                    app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                    message->command = APP_COMMANDS_MENU;
                    message->data[0] = APP_COMMAND_MENU_ANALOGUE_DIRECT_NO_NAVIGATION;
                    message->data_length = 1;

                    break;
                }

                default:
                    result = false;
                    break;
            }
        }
    }
    else
    {
        result = false;
    }
    
#if APP_FEATURE_SPECIAL_SWITCHES == APP_FEATURE_ON    
    // Just before leaving lets see if we have the buttons depressed to launch the 
    // Engineering menu
    //if (!result)
    //{
        result = special_switches_test(message);
    //}
#endif
    
    return result;
}

/**
 * Performs default button/switch processing when in menu mode
 */
static bool process_buttons_menu(STRUCTURED_MESSAGE* message)
{
    bool result = true;

    // Check whether it has changed
    // so if its changed the we are either changing the menu item t be displayed
    // this will be based on the current button, or rotary value

    if (input_context_has_button_changed())
    {
        ADC_BUTTON button = input_context_get_button();
        
        result = app_device_process_buttons_menu(button, message);
        
        if (!result)
        {
            switch(button)
            {

                default:
                    result = false;
                    break;
            }
        }
    }
    else
    {
        result = false;
    }
    
    return result;
}

/**
 * Performs default button/switch processing when in menu mode
 */
static bool process_buttons_menu_edit(STRUCTURED_MESSAGE* message)
{
    bool result = true;

    // Check whether it has changed
    // so if its changed the we are either changing the menu item t be displayed
    // this will be based on the current button, or rotary value

    ADC_BUTTON button = input_context_get_button();
    
    result = app_device_process_buttons_menu_edit(button, message);
    
    if (!result)
    {
        result = true;
        
        switch(button)
        {
            default:
                result = false;
                break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs button checking for system reset during startup
 */
static bool process_buttons_reset(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    ADC_BUTTON button = input_context_get_button();
    
    bool proceed = false;
    
#if APP_FEATURE_SPECIAL_SWITCHES == APP_FEATURE_ON    
    if (special_switches_active_button_count() == 1)
    {
        proceed = true;
    }
#endif    
    
    if (input_context_has_button_changed() && proceed)
    {
        logger_log_debugf(0, "BUTTON ACTIVE %d %d %d --", button, input_context_has_button_changed(), special_switches_active_button_count());
        
        // Added so that pressing the current input button will light up the display
        // when brightness = off - it did not before
        timers_reset(TIMER_INDEX_BRIGHTNESS);
        
        if (button > APP_BUTTONS_NONE)// != APP_BUTTONS_NONE)
        {
            app_commands_data.wants_reset = false;
        }
        
        switch (button)
        {
            case APP_BUTTONS_NONE:
            case ADC_BUTTON_ERROR:  //TODO This is a bodge with the ADC as it is reporting an error (probably due to voltages) - Waiting for Tims fixes before reviewing
            {
                logger_log_debugf(0, "BUTTON NONE --");
                if (timers_has_expired(TIMER_INDEX_MENU_SELECT))
                {
                    if (app_commands_data.wants_reset)
                    {
                        logger_log_debugf(0, "BUTTON NONE Factory Reset --");
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_FACTORY_RESET;
                        message->data[0] = 0xAA;
                        message->data[1] = 0xAA;
                        message->data_length = 2;
                        app_commands_data.wants_reset = false;
                    }
                }
                break;
            }
            
            case APP_BUTTONS_DIRAC: // Changed from BALANCE so it works for both SA30 and SA750
            {
                logger_log_debugf(0, "BUTTON BALANCE --");
                timers_reset(TIMER_INDEX_MENU_SELECT);
                app_commands_data.wants_reset = true;
                
                break;
            }
            
            default:
                result = false;
                break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs rotary input in default
 */
static bool process_encoder_default(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    int x;
    for (x = 0; x < ENCODER_INDEX_MAX; x++)
    {
        ENCODER_STATE state = input_context_get_encoder(x);
        if (state.delta != 0)
        {
            switch (x)
            {
                case ENCODER_INDEX_MAIN:
                {
                    // Short delay to fix issue with a jump of 2 volume when Logging switched off
                    //timers_delay_us(400);
                    //logger_log_debugf(221, "ENCODER: DELTA %d --", state.delta);
                    message->action = state.delta > 0 ? STRUCTURED_MESSAGE_ACTION_SET_INCREMENT : STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                    message->command = APP_COMMANDS_VOLUME;
                    message->data[0] = state.delta;
                    message->data_length = 1;
                    result = true;
                    break;
                }
                
                case ENCODER_INDEX_INPUT_SELECT:
                {
                    if (!input_control_in_switch())
                    {
                        INPUT_CONTROL_PREVIEW_INFO* info = input_control_get_preview_info();
                        message->command = APP_COMMANDS_PREVIEW_SOURCE;
                        message->data[0] = info->input;
                        message->data_length = 1;
                        
                        if (state.delta > 0)
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                        }
                        else
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                        }
                        
                        result = true;
                    }
                }
                
                break;
            }
        }
    }
     
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs default button/switch processing when in menu mode
 */
static bool process_encoder_menu(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    ENCODER_STATE state = input_context_get_encoder(ENCODER_INDEX_MAIN);
    if (state.delta != 0)
    {
        // Short delay to fix issue with a jump of 2 when Logging switched off
        //timers_delay_us(400);
        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
        message->command = APP_COMMANDS_MENU_NAVIGATION;
        message->data[0] = state.delta > 0 ? MENU_NAVIGATION_RIGHT : MENU_NAVIGATION_LEFT;
    }
    else
    {
        result = false;
    }
     
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs rotary input in default
 */
static bool process_encoder_menu_edit(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    ENCODER_STATE state = input_context_get_encoder(ENCODER_INDEX_MAIN);
    if (state.delta != 0)
    {
        // Short delay to fix issue with a jump of 2 when Logging switched off
        //timers_delay_us(400);
        message->action = state.delta > 0 ? STRUCTURED_MESSAGE_ACTION_SET_INCREMENT : STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
        message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
        message->data[0] = state.delta;
        message->repeat = false;
        result = true;
    }
     
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs rotary input in default
 */
static bool process_remote_default(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    // The structured message passed in here will have its command set to the command from 
    // the remote rather than one of the standard commands - in this method we need to
    // translate from one to the other.
    if (input_context_has_remote_changed())
    {
        STRUCTURED_MESSAGE* remote = input_context_get_remote_message();
        
        logger_log_infof(LOGGER_CODE_APP_CONTEXT_REMOTE_MESSAGE, "CONTEXT Remote (default mode) src: %d  cmd: %d rpt: %d --", remote->source, remote->command, remote->repeat);
        
        if (remote_use_message(remote))
        {
            // Is the message a repeat and if so do we need to process it?
            if (!remote->repeat)
            {
                // Added so that pressing the remote will light up the display
                // when brightness = off - it did not before
                timers_reset(TIMER_INDEX_BRIGHTNESS);
                
                switch (remote->command)
                {
                    case REMOTE_CMD_INPUT_POWER:
                    {
                        // Toggle the current power state
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_POWER;
                        message->data[0] = APP_COMMAND_POWER_TOGGLE;   
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_POWER_ON:
                    {
                        // Toggle the current power state
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_POWER;
                        message->data[0] = APP_COMMAND_POWER_ON;   
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_POWER_OFF:
                    {
                        // Toggle the current power state
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_POWER;
                        message->data[0] = APP_COMMAND_POWER_OFF;   
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_DISPLAY:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_BRIGHTNESS;
                        message->data[0] = APP_COMMAND_BRIGHTNESS_TOGGLE;   
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_DISPLAY_OFF:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_BRIGHTNESS;
                        message->data[0] = APP_COMMAND_BRIGHTNESS_OFF;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_DISPLAY_L1:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_BRIGHTNESS;
                        message->data[0] = APP_COMMAND_BRIGHTNESS_DIMMED;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_DISPLAY_L2:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_BRIGHTNESS;
                        message->data[0] = APP_COMMAND_BRIGHTNESS_FULL;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_MUTE:
                    {
                        // Toggle muting
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MUTE;
                        message->data[0] = APP_COMMAND_MUTE_TOGGLE;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_MUTE_ON:
                    {
                        // Mute On
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MUTE;
                        message->data[0] = APP_COMMAND_MUTE_ON;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_MUTE_OFF:
                    {
                        // Mute Off
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MUTE;
                        message->data[0] = APP_COMMAND_MUTE_OFF;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_DIRAC:
                    {
                        app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU;
                        message->data[0] = APP_COMMAND_MENU_DIRAC_NO_NAVIGATION;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_INFO:
                    {
                        // Show Info menu in edit mode
                        if ((menu_get_current_item()->menu_id == APP_COMMAND_MENU_INFO) ||
                            (input_control_get_type() != INPUT_CONTROL_TYPE_AVATAR))
                        {
                            AUDIO_INPUT selected_input = input_control_get_selected_input();
                            if (selected_input == AUDIO_INPUT_PHONO)
                            {
                                logger_log_debugf(0, "REMOTE Info->PHONO --");
                                app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                                message->command = APP_COMMANDS_PHONO_INPUT_MM_MC;
                                message->data[0] = 0x00;
                                message->data_length = 1;

                            }
                            else
                            {
                                logger_log_debugf(0, "REMOTE Info->Default --");
                                app_commands_display_default_screen(message);
                            }
                        }
                        else
                        {
                            logger_log_debugf(0, "REMOTE Info --");
                            display_scroll_line();
                            result = false;
                        }

                        break;
                    }

                    case REMOTE_CMD_INPUT_ANALOG_DIRECT:
                    {
                        app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU;
                        message->data[0] = APP_COMMAND_MENU_ANALOGUE_DIRECT_NO_NAVIGATION;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_ANALOG_DIRECT_ON:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_ANALOGUE_DIRECT;
                        message->data[0] = input_control_get_selected_input();
                        message->data[1] = true;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_ANALOG_DIRECT_OFF:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_ANALOGUE_DIRECT;
                        message->data[0] = input_control_get_selected_input();
                        message->data[1] = false;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_VOLUME_UP:
                    {
                        // Volume up
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                        message->command = APP_COMMANDS_VOLUME;
                        message->data[0] = 1;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_VOLUME_DOWN:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                        message->command = APP_COMMANDS_VOLUME;
                        message->data[0] = -1;
                        message->data_length = 1;
                        break;
                    }            

                    case REMOTE_CMD_INPUT_BALANCE:
                    {
                        // Show balance menu in edit mode
                        app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU;
                        message->data[0] = APP_COMMAND_MENU_BALANCE_NO_NAVIGATION;
                        message->data_length = 1;

                        break;
                    }

                    case REMOTE_CMD_INPUT_BALANCE_LEFT:
                    {
                        // Decrement the balance
                        volume_set_balance(volume_get_balance() - 1, true);

                        // Show balance menu in edit mode
                        app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU;
                        message->data[0] = APP_COMMAND_MENU_BALANCE_NO_NAVIGATION;
                        message->data_length = 1;

                        break;
                    }

                    case REMOTE_CMD_INPUT_BALANCE_RIGHT:
                    {
                        // Increment the balance
                        volume_set_balance(volume_get_balance() + 1, true);

                        // Show balance menu in edit mode
                        app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU;
                        message->data[0] = APP_COMMAND_MENU_BALANCE_NO_NAVIGATION;
                        message->data_length = 1;

                        break;
                    }            

                    case REMOTE_CMD_INPUT_PHONO:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_PHONO;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_AUX:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_AUX;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_NET:
                    case REMOTE_CMD_INPUT_USB:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_NET;
                        message->data_length = 1;
                        break;
                    }

//                    case REMOTE_CMD_INPUT_USB:
//                    {
//                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
//                        message->command = APP_COMMANDS_CURRENT_SOURCE;
//                        message->data[0] = AUDIO_INPUT_USB;
//                        message->data_length = 1;
//                        break;
//                    }

                    case REMOTE_CMD_INPUT_AV:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_AV;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_SAT:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_SAT;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_PVR:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_PVR;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_GAME:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_GAME;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_BD:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_BD;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_CD:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_CD;
                        message->data_length = 1;
                        break;
                    }
                    
                    case REMOTE_CMD_INPUT_STB:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_STB;
                        message->data_length = 1;
                        break;
                    }
                    
                    case REMOTE_CMD_INPUT_PCUSB:
                    {
                        // |Select the ARC Input
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_CURRENT_SOURCE;
                        message->data[0] = AUDIO_INPUT_HDMI_ARC;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_MENU:
                    {
                        app_context_set_mode(APP_CONTEXT_MODE_MENU);

                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU;
                        message->data[0] = APP_COMMAND_MENU_SYSTEM_SETTINGS;
                        message->data_length = 1;
                        break;
                    }

                                        
                    // Assume these are passed to Avatar
                    case REMOTE_CMD_INPUT_ZERO:
                    case REMOTE_CMD_INPUT_ONE:
                    case REMOTE_CMD_INPUT_TWO:
                    case REMOTE_CMD_INPUT_THREE:
                    case REMOTE_CMD_INPUT_FOUR:
                    case REMOTE_CMD_INPUT_FIVE:
                    case REMOTE_CMD_INPUT_SIX:
                    case REMOTE_CMD_INPUT_SEVEN:
                    case REMOTE_CMD_INPUT_EIGHT:
                    case REMOTE_CMD_INPUT_NINE:
                    case REMOTE_CMD_INPUT_RANDOM:
                    case REMOTE_CMD_INPUT_REPEAT:
                    case REMOTE_CMD_INPUT_RANDOM_REPEAT:
                    case REMOTE_CMD_INPUT_RANDOM_REPEAT_MODE:
                    case REMOTE_CMD_INPUT_REWIND:
                    case REMOTE_CMD_INPUT_FAST_FORWARD:
                    case REMOTE_CMD_INPUT_SKIP_BACK:
                    case REMOTE_CMD_INPUT_SKIP_FORWARD:
                    case REMOTE_CMD_INPUT_STOP:
                    case REMOTE_CMD_INPUT_PLAY:
                    case REMOTE_CMD_INPUT_PAUSE:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_REMOTE_AVATAR_MESSAGE;
                        message->data[0] = remote->command; 
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_LEFT:
                    {
                        logger_log_debugf(0, "REMOTE UP --");
                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
                        {
                            if (avatar_browsing_get_skip_control() == AVATAR_BROWSING_SKIP_CONTROL_NONE)
                            {
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                                message->command = APP_COMMANDS_BROWSING;
                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_PREVIOUS;
                            }
                         }
                        else
                        {
                            // Volume up
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_VOLUME;
                            message->data[0] = 1;
                            message->data_length = 1;
                        }
                        break;
                    }

                    case REMOTE_CMD_INPUT_RIGHT:
                    {
                        logger_log_debugf(0, "REMOTE DOWN --");
                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
                        {
                            if (avatar_browsing_get_skip_control() == AVATAR_BROWSING_SKIP_CONTROL_NONE)
                            {
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                                message->command = APP_COMMANDS_BROWSING;
                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_NEXT;
                            }
                        }
                        else
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                            message->command = APP_COMMANDS_VOLUME;
                            message->data[0] = -1;
                            message->data_length = 1;
                        }
                        break;
                    }           
                    
                    case REMOTE_CMD_INPUT_OK:
                    {
                        logger_log_debugf(0, "REMOTE OK --");
                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message->command = APP_COMMANDS_BROWSING;
                            message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_SELECT;
                        }
                        else
                        {
                        }
                        break;
                    }
                    
                    case REMOTE_CMD_INPUT_BACK:
                    {
                        logger_log_debugf(0, "REMOTE BACK %d --", avatar_browsing_get_exit_state());
                        if ((menu_get_current_item()->menu_id == APP_COMMAND_MENU_VOLUME) &&
                            (avatar_browsing_get_exit_state() == AVATAR_BROWSING_STATE_PLAYING))
                        {
                            // If we came from the Media Browsing menu then go back to it..
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message->command = APP_COMMANDS_MENU;
                            message->data[0] = APP_COMMAND_MENU_MEDIA_BROWSING;    // Top menu
                        }
                        else if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
                        {
                            if (avatar_browsing_get_skip_control() != AVATAR_BROWSING_SKIP_CONTROL_NONE)
                            {
                                avatar_browsing_set_skip_control(AVATAR_BROWSING_SKIP_CONTROL_EXIT);
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                                message->command = APP_COMMANDS_BROWSING;
                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_SELECT;
                            }
                            else
                            {
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                                message->command = APP_COMMANDS_BROWSING;
                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_BACK;
                            }
                        }
                        else
                        {
                        }
                        break;
                    }

                    case REMOTE_CMD_INPUT_UP:
                    {
                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
                        {
                            if (avatar_browsing_get_skip_control() != AVATAR_BROWSING_SKIP_CONTROL_NONE)
                            {
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                                message->command = APP_COMMANDS_BROWSING;
                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_PREVIOUS;
                            }
                            else
                            {
//                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
//                                message->command = APP_COMMANDS_BROWSING;
//                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_SKIP_UP;
                            }
                        }
                        
                        break;
                    }
                    
                    case REMOTE_CMD_INPUT_DOWN:
                    {
                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
                        {
                            if (avatar_browsing_get_skip_control() != AVATAR_BROWSING_SKIP_CONTROL_NONE)
                            {
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                                message->command = APP_COMMANDS_BROWSING;
                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_NEXT;
                            }
                            else
                            {
//                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
//                                message->command = APP_COMMANDS_BROWSING;
//                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_SKIP_DOWN;
                            }
                        }
                        
                        break;
                    }
                    
                    case REMOTE_CMD_INPUT_HOME:
                    {
                        //TODO  IR Navigation in default mode - do we need it/
                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_MEDIA_BROWSING)
                        {
                            if (avatar_browsing_get_skip_control() == AVATAR_BROWSING_SKIP_CONTROL_NONE)
                            {
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                                message->command = APP_COMMANDS_BROWSING;
                                message->data[0] = BROWSE_COMMAND_OPTION_BROWSE_TOP;
                            }
                        }
                        result = false;
                        break;
                    }

                    default:
                        result = false;
                        break;
                }
            }
            else
            {
                switch (remote->command)
                {
                    case REMOTE_CMD_INPUT_POWER:
                    {
                        // Toggle the current power state
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_POWER;
                        message->data[0] = APP_COMMAND_POWER_ON;   
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_VOLUME_UP:
                    {
                        // value up
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                        message->command = APP_COMMANDS_VOLUME;
                        message->data[0] = 1;
                        message->data_length = 1;
                        break;
                    }
                    
                    case REMOTE_CMD_INPUT_VOLUME_DOWN:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                        message->command = APP_COMMANDS_VOLUME;
                        message->data[0] = -1;
                        message->data_length = 1;
                        break;
                    }   
                    
                    case REMOTE_CMD_INPUT_REWIND:
                    case REMOTE_CMD_INPUT_FAST_FORWARD:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_REMOTE_AVATAR_MESSAGE;
                        message->data[0] = remote->command; 
                        message->data_length = 1;
                        break;
                    }

                    default:
                        result = false;
                        break;
                }
            }
        }
        else
        {
            result = false;
        }
    }
    else
    {
        result = false;
    }
   
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs remote input in default
 */
static bool process_remote_menu(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    if (input_context_has_remote_changed())
    {
        STRUCTURED_MESSAGE* remote = input_context_get_remote_message();

        logger_log_infof(LOGGER_CODE_APP_CONTEXT_REMOTE_MESSAGE, "CONTEXT Remote (menu mode) src: %d  cmd: %d rpt: %d --", remote->source, remote->command, remote->repeat);

        if (remote_use_message(remote))
        {
            if (!remote->repeat)
            {
                switch (remote->command)
                {
                    case REMOTE_CMD_INPUT_HOME:
                    case REMOTE_CMD_INPUT_MENU:
                    {
                        app_context_set_mode(APP_CONTEXT_MODE_MENU);

                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_SYSTEM_SETTINGS)
                        {
                            // Return to default state
                            app_commands_display_default_screen(message);
                        }
                        else
                        {
                            // Return to HOME menu
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message->command = APP_COMMANDS_MENU;
                            message->data[0] = APP_COMMAND_MENU_SYSTEM_SETTINGS;
                        }

                        result = true;
                        break;
                    }

                    case REMOTE_CMD_INPUT_BALANCE_LEFT:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Decrement the balance
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = -1; // Decrement
                       }

                        break;
                    }

                    case REMOTE_CMD_INPUT_BALANCE_RIGHT:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Increment the balance
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = 1; // Increment
                        }

                        break;
                    }            

                    case REMOTE_CMD_INPUT_LEFT:
                    {
                        // Map left key
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU_NAVIGATION;
                        message->data[0] = MENU_NAVIGATION_LEFT;

                        result = true;
                        break;
                    }

                    case REMOTE_CMD_INPUT_RIGHT:
                    {
                        // Map right key
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU_NAVIGATION;
                        message->data[0] = MENU_NAVIGATION_RIGHT;

                        result = true;
                        break;
                    }

                    case REMOTE_CMD_INPUT_UP:
                    {
                        // Map left key
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU_NAVIGATION;
                        message->data[0] = MENU_NAVIGATION_UP;

                        result = true;
                        break;
                    }

                    case REMOTE_CMD_INPUT_DOWN:
                    {
                        // Map right key
                        if (menu_has_children(menu_get_current_item()))
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message->command = APP_COMMANDS_MENU_NAVIGATION;
                            message->data[0] = MENU_NAVIGATION_OK;
                        }
                        else
                        {
                            app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                        }

                        result = true;
                        break;
                    }

                    case REMOTE_CMD_INPUT_OK:
                    {
                        // Map OK key
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU_NAVIGATION;
                        message->data[0] = MENU_NAVIGATION_OK;

                        result = true;
                        break;
                    }

                    case REMOTE_CMD_INPUT_BACK:
                    {
                        // Map back key
                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_SYSTEM_SETTINGS)
                        {
                            // Return to default state
                            app_commands_display_default_screen(message);
                        }
                        else
                        {
                            // Go Up a level
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message->command = APP_COMMANDS_MENU_NAVIGATION;
                            message->data[0] = MENU_NAVIGATION_UP;
                        }
                        break;
                    }
                }
            }
        }
        else
        {
            result = false;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs rotary input in default
 */
static bool process_remote_menu_edit(STRUCTURED_MESSAGE* message)
{
    bool result = true;
     
    if (input_context_has_remote_changed())
    {
        STRUCTURED_MESSAGE* remote = input_context_get_remote_message();

        logger_log_infof(LOGGER_CODE_APP_CONTEXT_REMOTE_MESSAGE, "CONTEXT Remote (edit mode) src: %d  cmd: %d rpt: %d --", remote->source, remote->command, remote->repeat);

        if (remote_use_message(remote))
        {
            message->repeat = remote->repeat;

            if (!remote->repeat)
            {
                switch (remote->command)
                {
                    case REMOTE_CMD_INPUT_POWER:
                    {
                        // Toggle the current power state
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_POWER;
                        message->data[0] = APP_COMMAND_POWER_TOGGLE;   
                        break;
                    }

                    case REMOTE_CMD_INPUT_POWER_ON:
                    {
                        // Toggle the current power state
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_POWER;
                        message->data[0] = APP_COMMAND_POWER_ON;   
                        break;
                    }

                    case REMOTE_CMD_INPUT_POWER_OFF:
                    {
                        // Toggle the current power state
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_POWER;
                        message->data[0] = APP_COMMAND_POWER_OFF;   
                        break;
                    }

                    case REMOTE_CMD_INPUT_BALANCE_LEFT:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Decrement the balance
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = -1; // Decrement
                        }
                        
                        break;
                    }

                    case REMOTE_CMD_INPUT_BALANCE:
                    {
                        // Return to default state
                        app_commands_display_default_screen(message);
                        break;
                    }

                    case REMOTE_CMD_INPUT_BALANCE_RIGHT:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Increment the balance
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = 1; // Decrement
                        }

                        break;
                    }       
                    

                    case REMOTE_CMD_INPUT_DOWN:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Increment the balance
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = -1; // Decrement
                        }
                        else
                        {
                            // Decrement the value
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                        }
                        
                        break;
                    }

                    case REMOTE_CMD_INPUT_UP:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Increment the balance
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = 1; // Decrement
                        }
                        else
                        {
                            // Decrement the value
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                        }

                        break;
                    }       
 
                    case REMOTE_CMD_INPUT_HOME:
                    case REMOTE_CMD_INPUT_MENU:
                    {
                        app_context_set_mode(APP_CONTEXT_MODE_MENU);

                        if ((menu_get_current_item()->menu_id == APP_COMMAND_MENU_SYSTEM_SETTINGS) ||
                            (menu_get_current_item()->menu_id == APP_COMMAND_MENU_BALANCE_NO_NAVIGATION) ||
                            (menu_get_current_item()->menu_id == APP_COMMAND_MENU_DIRAC_NO_NAVIGATION) ||
                            (menu_get_current_item()->menu_id == APP_COMMAND_MENU_ANALOGUE_DIRECT_NO_NAVIGATION) ||   
                            (menu_get_current_item()->menu_id == APP_COMMAND_MENU_INFO))    
                        {
                            // Return to default state
                            app_commands_display_default_screen(message);
                        }
                        else
                        {
                            // Return to HOME menu
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message->command = APP_COMMANDS_MENU;
                            message->data[0] = APP_COMMAND_MENU_SYSTEM_SETTINGS;
                        }

                        break;
                    }

                    case REMOTE_CMD_INPUT_VOLUME_UP:
                    {
                        // The volume up/down can be used to adjust any value in
                        // any menu except for the balance at the moment...
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Volume up
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            message->data[0] = 1; // Decrement
                        }
                        break;
                    }

                    case REMOTE_CMD_INPUT_VOLUME_DOWN:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            message->data[0] = -1; // Increment
                        }
                        break;
                    }       

                    case REMOTE_CMD_INPUT_LEFT:
                    {
                        //app_context_set_mode(APP_CONTEXT_MODE_MENU);

                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU_NAVIGATION;
                        message->data[0] = MENU_NAVIGATION_LEFT;

                        break;
                    }

                    case REMOTE_CMD_INPUT_RIGHT:
                    {
                        //app_context_set_mode(APP_CONTEXT_MODE_MENU);

                        // Move to the previous sibling
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_MENU_NAVIGATION;
                        message->data[0] = MENU_NAVIGATION_RIGHT;
                        break;
                    }

                    case REMOTE_CMD_INPUT_OK:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (menu->menu_id == APP_COMMAND_MENU_ANALOGUE_DIRECT_NO_NAVIGATION)
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_ANALOGUE_DIRECT; // Will actually Toggle
                            message->data[0] = input_control_get_selected_input();
                            //message->data[1] = !input_control_get_analogue_direct();
                            message->data_length = 1;
                        }
                        else
                        {
                            // Move to the first child in the menu
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message->command = APP_COMMANDS_MENU_NAVIGATION;
                            message->data[0] = MENU_NAVIGATION_OK;
                        }
                        break;
                    }

                    case REMOTE_CMD_INPUT_BACK:
                    {
                        // Map home key
                        if (menu_get_current_item()->menu_id == APP_COMMAND_MENU_SYSTEM_SETTINGS)
                        {
                            // Return to default state
                            app_commands_display_default_screen(message);
                        }
                        else
                        {
                            // Go Up a level
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message->command = APP_COMMANDS_MENU_NAVIGATION;
                            message->data[0] = MENU_NAVIGATION_UP;
                        }

                        break;
                    }

                    case REMOTE_CMD_INPUT_DIRAC:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                        message->command = APP_COMMANDS_ROOM_EQUALISATION;
                        message->data[0] = DIRAC_NO_CURVE; //NOT USEDdirac_get_selected_curve_index();
                        message->data_length = 1;            
                        break;
                    }

                    case REMOTE_CMD_INPUT_ANALOG_DIRECT:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                        message->command = APP_COMMANDS_ANALOGUE_DIRECT;
                        message->data[0] = input_control_get_selected_input();
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_ANALOG_DIRECT_ON:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_ANALOGUE_DIRECT;
                        message->data[0] = input_control_get_selected_input();
                        message->data[1] = true;
                        message->data_length = 1;
                        break;
                    }

                    case REMOTE_CMD_INPUT_ANALOG_DIRECT_OFF:
                    {
                        message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                        message->command = APP_COMMANDS_ANALOGUE_DIRECT;
                        message->data[0] = input_control_get_selected_input();
                        message->data[1] = false;
                        message->data_length = 1;
                        break;
                    }

                    default:
                        result = false;
                        break;
                }
            }
            else
            {
                switch (remote->command)
                {
                    case REMOTE_CMD_INPUT_VOLUME_UP:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = 1; // Increment
                        }
                        // value up
                        break;
                    }
                    case REMOTE_CMD_INPUT_VOLUME_DOWN:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = -1; // Decrement
                        }
                        break;
                    }            

                    case REMOTE_CMD_INPUT_BALANCE_LEFT:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            uint8_t balance = volume_get_balance();
                            if (balance != VOLUME_DEFAULT_BALANCE)
                            {
                                // Decrement the balance
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                                message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                                //message->data[0] = -1; // Decrement
                             }
                        }
                        
                        break;
                    }

                    case REMOTE_CMD_INPUT_BALANCE_RIGHT:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            uint8_t balance = volume_get_balance();
                            if (balance != VOLUME_DEFAULT_BALANCE)
                            {
                                // Increment the balance
                                message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                                message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
//                                message->data[0] = 1; // Decrement
                            }
                        }

                        break;
                    }  

                    case REMOTE_CMD_INPUT_DOWN:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Decrement the balance
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = -1; // Decrement
                        }
                        
                        break;
                    }

                    case REMOTE_CMD_INPUT_UP:
                    {
                        MENU_ITEM* menu = menu_get_current_item();
                        if (is_balance_menu(menu))
                        {
                            // Increment the balance
                            message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                            message->command = APP_COMMANDS_MENU_VALUE_CHANGE;
                            //message->data[0] = 1; // Decrement
                        }

                        break;
                    }       
                    
                    default:
                        result = false;
                        break;
                }
            }
        }
        else
        {
            result = false;
        }
    }
    else
    {
        result = false;
    }
    
    return result;
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Send the AMX duet response
 * @param message
 */
void app_commands_process_amx_duet_message(STRUCTURED_MESSAGE* message)
{
    // Construct the response
    int len = strlen(STRUCTURED_MESSAGE_AMX_DUET_RESPONSE) + 10; // +10 to accommodate model name 
    char buffer[len];
    
    memset(buffer, 0, len);
    sprintf(buffer, STRUCTURED_MESSAGE_AMX_DUET_RESPONSE, SOFTWARE_MODEL_STRING, SOFTWARE_VERSION_MAJOR, SOFTWARE_VERSION_MINOR);
    
    structured_message_set_message_data(message, buffer, strlen(buffer));
    message->data_only = true;
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Processes the current input context to setup a structured message that can then
 * be processed by the app_commands
 * @param - a callback function to handle the generated structured message
 */
void app_commands_process_input_controls(STRUCTURED_MESSAGE_CALLBACK callback)
{
    if (callback != NULL)
    {
        // Create a structured message to contain the command
        STRUCTURED_MESSAGE message;
        structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
        
        //logger_log_debugf(0, "IP mode %d ", app_context_get_mode());
        switch (app_context_get_mode())
        {
            case APP_CONTEXT_MODE_DEFAULT:
            {
                // First see if any buttons have been processed
                if (process_buttons_default(&message))
                {
                    //logger_log_debugf(0, "PROCESSING BUTTONS");
                    message.source = STRUCTURED_MESSAGE_SOURCE_BUTTONS;
                }
#if APP_FEATURE_BALANCE_POT == APP_FEATURE_ON
                else if (process_balance_default(&message))
                {
                    message.source = STRUCTURED_MESSAGE_SOURCE_BALANCE_POT;
                }
#endif                
                // Check encoder
                else if (process_encoder_default(&message))
                {
                    message.source = STRUCTURED_MESSAGE_SOURCE_ENCODER;
                }
                // Check remote
                else if (process_remote_default(&message))
                {
                    //logger_log_debugf(0, "PROCESSING REMOTE");
                    message.source = STRUCTURED_MESSAGE_SOURCE_REMOTE;
                }
                 
                break;
            }

            case APP_CONTEXT_MODE_MENU:
            {
                if (!timers_has_expired(TIMER_INDEX_MENU_INACTIVE))
                {
                    if (input_context_has_changed())
                    {
                        timers_reset(TIMER_INDEX_MENU_INACTIVE);

                        // First see if any buttons have been processed
                        if (process_buttons_menu(&message))
                        {
                            message.source = STRUCTURED_MESSAGE_SOURCE_BUTTONS;
                        }
#if APP_FEATURE_BALANCE_POT == APP_FEATURE_ON
                        else if (process_balance_menu(&message))
                        {
                            message.source = STRUCTURED_MESSAGE_SOURCE_BALANCE_POT;
                        }
#endif
                        // Check encoder
                        else if (process_encoder_menu(&message))
                        {
                            message.source = STRUCTURED_MESSAGE_SOURCE_ENCODER;
                        }
                        // Check remote
                        else if (process_remote_menu(&message))
                        {
                            message.source = STRUCTURED_MESSAGE_SOURCE_REMOTE;
                        }
                        else
                        {
                           // no action
                        }
                    }
                }
                else
                {
                    // Go back to volume display
                    //logger_log_debugf(0, "MENU 3");
                    app_commands_display_default_screen(&message);
                }
                
                break;
            }
            
            case APP_CONTEXT_MODE_MENU_EDIT:
            {
                if (!timers_has_expired(TIMER_INDEX_MENU_INACTIVE))
                {
                    if (input_context_has_changed())
                    {
                        timers_reset(TIMER_INDEX_MENU_INACTIVE);

                        // First see if any buttons have been processed
                        if (process_buttons_menu_edit(&message))
                        {
                            message.source = STRUCTURED_MESSAGE_SOURCE_BUTTONS;
                        }
#if APP_FEATURE_BALANCE_POT == APP_FEATURE_ON
                        else if (process_balance_menu_edit(&message))
                        {
                            message.source = STRUCTURED_MESSAGE_SOURCE_BALANCE_POT;
                        }
#endif                        
                        // Check encoder
                        else if (process_encoder_menu_edit(&message))
                        {
                            message.source = STRUCTURED_MESSAGE_SOURCE_ENCODER;
                        }
                        // Check remote
                        else if (process_remote_menu_edit(&message))
                        {
                           message.source = STRUCTURED_MESSAGE_SOURCE_REMOTE;
                        }
                    }
                }
                else
                {
                    //logger_log_debugf(0, "MENU 4");
                    // Go back to volume display
                    app_commands_display_default_screen(&message);
                }
                
                break;
            }
            
            case APP_CONTEXT_MODE_PROCESS:
            {
                // A process is in progress
                stateful_process_invoke();
                
                // BUG 3701 - Check MENU button to see if we need to reset the system settings
                if (process_buttons_reset(&message))
                {
                   // logger_log_debugf(0, "PROCESSING BUTTONS");
                    message.source = STRUCTURED_MESSAGE_SOURCE_BUTTONS;
                }
                break;
            }
            
            case APP_CONTEXT_MODE_STANDBY:
            {
//                if (input_context_has_changed())
//                {
//                    app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
//                }
                
                break;
            }
        }
        
        if (message.command != STRUCTURED_MESSAGE_COMMAND_UNDEFINED)
        {
            callback(&message);
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Process a structured message
 * This command will map the messages command and data to a function
 */
void app_commands_process_message(STRUCTURED_MESSAGE* message)
{
    app_context_pre_process_message(message);
    logger_suppress_command_logging(true);
    
    // Execute the command
    switch(message->command)
    {
        // Power
        case APP_COMMANDS_POWER:
        {
            app_commands_power(message);
            break;
        }
        
        // Display Brightness
        case APP_COMMANDS_BRIGHTNESS:
        {
            app_commands_edit_display_brightness(message);
            break;
        }
        
        // Headphones
        case APP_COMMANDS_HEADPHONES:
        {
            app_commands_headphones(message);
            break;
        }
        
        // Software Version
        case APP_COMMANDS_SOFTWARE_VERSION:
        {
            app_commands_software_version(message);
            break;
        }

        // Factory Reset
        case APP_COMMANDS_FACTORY_RESET:
        {
            app_commands_factory_reset(message);
            break;
        }
        
        // Simulate RC5
        case APP_COMMANDS_SIMULATE_RC5:
        {
            app_commands_simulate_remote(message);
            break;
        }
        
        // Volume
        case APP_COMMANDS_VOLUME:
        {
            app_commands_edit_volume(message);
            break;
        }
        
        // Mute
        case APP_COMMANDS_MUTE:
        {
            if (app_context_get_mode() != APP_CONTEXT_MODE_BIAS_TEST)
            {
                app_commands_mute_user(message);
            }
            break;
        }
        
        // Direct Mode Status
        case APP_COMMANDS_ANALOGUE_DIRECT:
        {
            app_commands_edit_analogue_direct(message);
            break;
        }
        
        // Network Playback status
        case APP_COMMANDS_PLAYBACK_STATUS:
        {
            app_commands_player_status(message);
            break;
        }
        
        case APP_COMMANDS_CURRENT_SOURCE:
        {
            app_commands_current_source(message);
            break;
        }

        // Headphone Override
        case APP_COMMANDS_HEADPHONE_OVERRIDE:
        {
            app_commands_edit_headphone_override(message);
            break;
        }
        
        // Heartbeat
        case APP_COMMANDS_HEARTBEAT:
        {
            app_commands_heartbeat(message);
            break;
        }
        
        // Reboot
        case APP_COMMANDS_REBOOT:
        {
            app_commands_reboot(message);
            break;
        }
        
        // Network - IP Address, MAC Address, Friendly Name
        case APP_COMMANDS_NETWORK:
        {
            app_commands_network(message);
            break;
        }
        
        // Balance
        case APP_COMMANDS_BALANCE:
        {
            app_commands_edit_balance(message);

            if (((message->source == STRUCTURED_MESSAGE_SOURCE_RS232) ||
                 (message->source == STRUCTURED_MESSAGE_SOURCE_TCPIP)) &&
                 (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS))
            {
                // Show balance menu in edit mode
                app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
                message->source = STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT;
                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                message->command = APP_COMMANDS_MENU;
                message->data[0] = APP_COMMAND_MENU_BALANCE_NO_NAVIGATION;
                
                app_commands_process_message(message);
            }

            break;
        }
        
        case APP_COMMANDS_ROOM_EQ_NAMES:
        {
            app_commands_room_eq_names(message);
            break;
        }
        
        // Room Equalisation
        case APP_COMMANDS_ROOM_EQUALISATION:
        {
            app_commands_edit_dirac_selected(message);
            break;
        }
        
        // Incoming audio sample rate
        case APP_COMMANDS_SAMPLE_RATE:
        {
            app_commands_get_sample_rate(message);
            break;
        }
        
        // DC Offset
        case APP_COMMANDS_DC_OFFSET_STATUS:
        {
            app_commands_get_dc_offset_status(message);
            break;
        }
        
        // Short Circuit status
        case APP_COMMANDS_SHORT_CIRCUIT_STATUS:
        {
            app_commands_get_short_circuit_status(message);
            break;
        }
        
        // Timeout Counter
        case APP_COMMANDS_TIMEOUT_COUNTER:
        {
            app_commands_get_timeout_counter(message);
            break;
        }
        
        case APP_COMMANDS_LIFTER_TEMP:
        {
            app_commands_get_lifter_temperature(message);
            break;
        }

        case APP_COMMANDS_OUTPUT_TEMP:
        {
            app_commands_get_output_temperature(message);
            break;
        }
        
        // Auto Shutdown Control
         case APP_COMMANDS_AUTO_SHUTDOWN:
        {
            app_commands_edit_auto_shutdown_timeout(message);
            break;
        }
        
        case APP_COMMANDS_PHONO_INPUT_MM_MC:
        {
            app_commands_edit_mm_mc(message);
            break;
        }
       
        // Input Detect
        case APP_COMMANDS_INPUT_DETECT:
        {
            app_commands_get_input_detection_status(message);
            break;
        }
        
        // Processor Mode Input
        case APP_COMMANDS_PROCESSOR_MODE_INPUT:
        {
            app_commands_edit_processor_mode_input_command(message);
            break;
        }
        
        // Processor Mode Volume
        case APP_COMMANDS_PROCESSOR_MODE_VOLUME:
        {
            app_commands_edit_processor_mode_volume(message);
            break;
        }

        case APP_COMMANDS_AMX_DUET:
        {
            app_commands_process_amx_duet_message(message);
            break;
        }
       
        // System Status
        case APP_COMMANDS_GET_STATUS_SUMMARY:
        {
            app_commands_system_status(message);
            break;
        }
       
        // System Model
        case APP_COMMANDS_MODEL_NUMBER:
        {
            app_commands_get_model_number(message);
            break;
        }
        
        case APP_COMMANDS_BIAS_CHECK:
        {
            app_commands_bias_test(message);
            break;
        }
        
        // DAC Filter
        case APP_COMMANDS_SET_FILTER:
        {
            app_commands_edit_filter(message);
            break;
        }
        
        case APP_COMMANDS_NOW_PLAYING:
        {
            app_commands_now_playing(message);
            break;
        }
        
        case APP_COMMANDS_MAX_ON_VOLUME:
        {
            app_commands_edit_max_on_volume(message);
            break;
        }
        
        case APP_COMMANDS_MAXIMUM_VOLUME:
        {
            app_commands_edit_maximum_volume(message);
            break;
        }
        
        case APP_COMMANDS_MAXIMUM_STREAMING_VOLUME:
        {
            app_commands_edit_maximum_streaming_volume(message);
            break;
        }
        
        case APP_COMMANDS_REMOTE_AVATAR_MESSAGE:
        {
            app_commands_remote_avatar_message(message);
            break;
        }

        case APP_COMMANDS_TEST_RGB_COLOURS:
        {
            rgb_led_set_colour(message->data[0], message->data[1], message->data[2]);
            break;
        }

        // Menu Commands
        //----------------------------------------------------------------------
        case APP_COMMANDS_MENU:
        {
            if ((message->source != STRUCTURED_MESSAGE_SOURCE_RS232) &&
                (message->source != STRUCTURED_MESSAGE_SOURCE_TCPIP))
            {
                menu_app_handle_switch(message);
            }
            break;
        }
        
        case APP_COMMANDS_MENU_NAVIGATION:
        {  
            if ((message->source != STRUCTURED_MESSAGE_SOURCE_RS232) &&
                (message->source != STRUCTURED_MESSAGE_SOURCE_TCPIP))
            {
                menu_app_handle_navigation(message);
            }
            break;
        }
        
        case APP_COMMANDS_MENU_VALUE_CHANGE:
        {
            if ((message->source != STRUCTURED_MESSAGE_SOURCE_RS232) &&
                (message->source != STRUCTURED_MESSAGE_SOURCE_TCPIP))
            {
                menu_app_value_change(message);
            }
            break;
        }
        
        // Input Preview
        //----------------------------------------------------------------------
        case APP_COMMANDS_PREVIEW_SOURCE:
        {
            app_commands_preview_source(message);
            break;
        }

        // Upgrade Commands
        //----------------------------------------------------------------------
        case APP_COMMANDS_UPGRADE_LOAD_START:
        {
            app_commands_upgrade_upload_start(message);
            break;
        }

        case APP_COMMANDS_UPGRADE_LOAD_DATA:
        {
            app_commands_upgrade_upload_data(message);
            break;
        }

        case APP_COMMANDS_UPGRADE_LOAD_END:
        {
            app_commands_upgrade_upload_end(message);
            break;
        }
        
        case APP_COMMANDS_BROWSING:
        {
            app_commands_handle_browsing(message);
            break;
        }

        case APP_COMMANDS_ERASE_FLASH:
        {
            logger_log_debugf(0, "PRE Flash --");
            logger_process();
            flash_memory_erase();
            logger_log_debugf(0, "POST Flash --");
            logger_process();
            //SYS_RESET_SoftwareReset();
            break;
        }

        case APP_COMMANDS_PRINT_UPGRADE_INFO:
        {
            app_commands_upgrade_print_info(message);
            break;
        }
        
        // Engineering stuff
        case APP_COMMANDS_ADC_RAW_VALUES:
        {
            //TODO APP_COMMANDS_ADC_RAW_VALUES
            break;
        }
        
        case APP_COMMANDS_INJECT_FAULT:
        {
            app_commands_inject_fault(message);
            break;
        }

        case APP_COMMANDS_WATCHDOG_STATUS:
        {
            app_commands_watchdog_status(message);
            break;
        }
        
        case APP_COMMANDS_SET_LOGGER_LEVEL:
        {
            app_commands_edit_logger_level(message);
            break;
        }
                    
        default:
        {
            message->response = COMMAND_NOT_RECOGNISED;
            
            app_commands_send_message(message);

            break;
        }
    }
    
    logger_suppress_command_logging(false);
}

/**-----------------------------------------------------------------------------
 * Invoked when the splash screen display is complete
 */
void app_commands_splash_screen_end()
{
    logger_log_debugf(LOGGER_CODE_DEBUG, "APP: Splash screen complete START --");
    //output_control_unmute_outputs();
    //volume_fade_in(); // To get ramp
            
    app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_WHITE);
    rgb_led_flash_stop();

    bool upgrade_status = false; 
    settings_get(SETTING_UPGRADE_STATUS, &upgrade_status);
    if (upgrade_status)
    {
        app_context_set_mode(APP_CONTEXT_MODE_MENU);
        app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_COMPLETE);
        
        avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_UPGRADE_COMPLETE);
        
        upgrade_status = false;
        settings_set(SETTING_UPGRADE_STATUS, &upgrade_status);
        settings_save();
    }
    
    // Check the Upgrade Option - If set to On Power Up check if there is
    // an upgrade available
    if (avatar_upgrade_available())
    {
        logger_log_debugf(LOGGER_CODE_INITIALISATION, "APP: Performing POWER UP Upgrade --");

        app_commands_avatar_upgrade();
    }

    logger_log_debugf(LOGGER_CODE_DEBUG, "APP: Splash screen complete END --");
}
    
/**-----------------------------------------------------------------------------
 * Initiates the splash screen
 */
void app_commands_splash_screen_start()
{
    display_test_set_mode(DISPLAY_TEST_MODE_LOGO);
    app_commands_start_process(display_test_run, app_commands_splash_screen_end);
}

/**-----------------------------------------------------------------------------
 * Handle structured messages received via rs232
 * @param message - A pointer to a structured message - do NOT store this pointer
 */
void app_commands_structured_message_callback(STRUCTURED_MESSAGE* message)
{
    switch (message->source)
    {
        case STRUCTURED_MESSAGE_SOURCE_REMOTE_RAW:
        {
            // Add to input context - this will be processed in a short while as 
            // an STRUCTURED_MESSAGE_SOURCE_INPUTS message
            input_context_set_remote_message(message);
            break;
        }
        
        default:
        {
            // Now process the message
            // Translate command based on message source
            // Process message? or store for later processing
            app_commands_process_message(message);
            
            // Check if the external message indicates activity
            if ((message->source == STRUCTURED_MESSAGE_SOURCE_RS232) ||
                (message->source == STRUCTURED_MESSAGE_SOURCE_TCPIP))
            {
                if (message->is_activity)
                {
                    input_context_external_message_processed(message);
                }
            }
        }
    }
}

/* *****************************************************************************
 End of File
 */
