/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    app_commands_auto_shutdown.c

  @Summary
 This file gathers the functions used for auto shutdown in the menu system.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_auto_shutdown.h"
#include "app_context.h"
#include "utilities/app_commands_utilities.h"
#include "utilities/avatar_message_processing.h"
#include "utilities/power.h"
#include "utilities/timers.h"
#include <stdbool.h>
#include <stdio.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**-----------------------------------------------------------------------------
 * Check the Auto Shutdown timer
 */
bool app_commands_check_auto_shutdown()
{
    bool result = false;
    
    // Check whether there has been any activity
    if (input_context_has_changed() || input_control_audio_input_present() || avatar_message_in_upgrade())
    {
        //logger_log_debugf(0, "SHUTDOWN RESET %d %d %d %d --", power_get_auto_shutdown_timeout(), input_context_has_changed(), input_control_audio_input_present(), avatar_message_in_upgrade());
        // Reset timer
        power_reset_auto_shutdown_timer();
    }
    else
    {
        // Check if timer has expired - if so then shutdown
        if ((power_get_state() == POWER_STATE_ON) &&
             power_get_auto_shutdown() && 
             timers_has_expired(TIMER_INDEX_AUTO_SHUTDOWN))
        {
            logger_log_infof(LOGGER_CODE_POWER, "POWER: Power Down Required --");
            result = true;
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Used to edit the auto_shutdown timeout based on the current inputs
 * @param message
 */
bool app_commands_edit_auto_shutdown_timeout(STRUCTURED_MESSAGE* message)
{
    bool result = false;

    int16_t value = (int16_t)message->data[0];
    
    bool proceed = true;
    
    // Try and find in the APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT enum
    if ((message->source == STRUCTURED_MESSAGE_SOURCE_RS232) ||
        (message->source == STRUCTURED_MESSAGE_SOURCE_TCPIP))
    {
        if (value < APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_MAX)
        {
            value = APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_VALUES[value];
        }
        else if (value == APP_COMMAND_VALUE_REQUEST_STATUS)
        {
            // Just carry on
        }
        else
        {
            proceed = false;
        }
    }
     
    app_commands_review_message_action(message);
   
    message->command = APP_COMMANDS_AUTO_SHUTDOWN;
    
    if (proceed)
    {
         uint16_t current_value = power_get_auto_shutdown_timeout();
    
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(uint16_t));
        }

        // Increment by 30 mins
        app_commands_process_message_value_array(message, &current_value, &value, APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_VALUES, APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_MAX);
        
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            power_set_auto_shutdown_timeout(current_value);

            menu_update_change_state(&current_value, sizeof(uint16_t));

            message->data[0] = (uint8_t) power_get_auto_shutdown_timeout();
        }
            
        // Convert back into enum
        int x;
        for (x = 0; x < APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_MAX; x++)
        {
            if (message->data[0] == APP_COMMAND_AUTO_SHUTDOWN_TIMEOUT_VALUES[x])
            {
                message->data[0] = x;
                break;
            }
        }

        message->response = STATUS_UPDATE;
        message->data_length = 1;
    }
    else
    {
        message->response = PARAM_NOT_RECOGNISED;
    }

    app_commands_send_message(message);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the auto shutdown value
 * @return 
 */
void app_commands_get_auto_shutdown_timeout_string(char* buffer, size_t size)
{
    // This will be in minutes
    uint16_t timeout = power_get_auto_shutdown_timeout();
    
    if (timeout >= 60)
    {
        if (timeout % 60 != 0)
        {
            sprintf(buffer, "%dh %dm", timeout / 60, timeout % 60);
        }
        else
        {
            sprintf(buffer, "%dh", timeout / 60);
        }
    }
    else if (timeout > 0)
    {
        sprintf(buffer, "%dm", timeout);
    }
    else
    {
        sprintf(buffer, "%s", APP_STRINGS[APP_STRINGS_ENUM_OFF]);
    }
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the auto shutdown value
 * @return 
 */
void app_commands_get_auto_shutdown_timeout_remaining_string(char* buffer, size_t size)
{
    if (power_get_auto_shutdown())
    {
        // This will be in milliseconds
        unsigned int ms = timers_get_remaining_count(TIMER_INDEX_AUTO_SHUTDOWN);
        unsigned int seconds = ms / 1000;
        
        unsigned int timeout = timers_get_timeout(TIMER_INDEX_AUTO_SHUTDOWN); //power_get_auto_shutdown_timeout();
        
        // Where there is audio input the display flickers between timeout/timeout-1
        // so display Active instead..
        if ((ms + 1000) > (timeout))
        {
            sprintf(buffer, APP_STRINGS[APP_STRINGS_ENUM_ACTIVE]); //"Active");
        }
        else if (seconds < 60)
        {
            sprintf(buffer, "%ds", seconds);
        }
        else
        {
            uint16_t minutes = seconds / 60;
            seconds = seconds % 60;

            if (minutes < 60)
            {
                sprintf(buffer, "%dm %ds", minutes, seconds);
            }
            else
            {
                uint16_t hours = minutes / 60;
                minutes = minutes % 60;
                sprintf(buffer, "%dh %dm %ds", hours, minutes, seconds);
            }
        }
    }
    else
    {
        sprintf(buffer, "%s", APP_STRINGS[APP_STRINGS_ENUM_OFF]);
    }
}

/**-----------------------------------------------------------------------------
 * Gets the number of minutes until the auto shutdown timeout occurs
 * @param message
 */
void app_commands_get_timeout_counter(STRUCTURED_MESSAGE* message)
{
    // This message should not be regarded as making the system active
    message->is_activity = false;
    
    if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
    {
        // Reset the Auto shutdown timer
        unsigned int timeout = timers_get_remaining_count(TIMER_INDEX_AUTO_SHUTDOWN);
        timeout = timeout / (1000 * 60);
        
        uint16_t minutes = (uint16_t)timeout;

        // Update the message
        message->command = APP_COMMANDS_TIMEOUT_COUNTER;
        message->response = STATUS_UPDATE;
        message->data_length = sizeof(uint16_t);
        message->data[0] = minutes / 256;
        message->data[1] = minutes % 256;
    }
    else
    {
        message->response = PARAM_NOT_RECOGNISED;
    }
    
    app_commands_send_message(message);
}

/* *****************************************************************************
 End of File
 */
