/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_avatar.h"
#include <stdio.h>
#include "app_commands.h"
#include "app_commands_input.h"
#include "app_config.h"
#include "app_context.h"
#include "utilities/analogue_direct.h"
#include "utilities/avatar.h"
#include "utilities/avatar_browsing.h"
#include "utilities/avatar_message_processing.h"
#include "utilities/dirac.h"
#include "utilities/display.h"
#include "utilities/display_test.h"
#include "utilities/input_control.h"
#include "utilities/input_control_app.h"
#include "utilities/menu_display.h"
#include "utilities/mutes.h"
#include "utilities/tcpip.h"
#include "utilities/volume.h"
#include "tcpip/tcpip.h"
#include "utilities/timers.h"
#include "utilities/upgrade_metadata.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define AVATAR_PLAYING_INFO_MAX_LENGTH  40
#define AVATAR_FAST_FWD_TIME_MS         10000

typedef struct 
{
    char album_name[AVATAR_PLAYING_INFO_MAX_LENGTH];
    
    char artist_name[AVATAR_PLAYING_INFO_MAX_LENGTH];
    
    // The avatar firmware version
    char firmware_version[40];
    
    AVATAR_PLAYER_STATUS player_status;
    
    char current_playing_info[DISPLAY_MAX_SCROLL_TEXT_LEN];
    
    // Indicates that the current_playing_info has changed
    bool current_playing_info_refresh;
    
    // Stores the last sample rate when redrawing scrolling info
    uint8_t current_sample_rate;
    
    char track_name[AVATAR_PLAYING_INFO_MAX_LENGTH];
    
    uint64_t track_length;
    
    uint64_t track_time;
    
    //! The offset of the data being written in the upgrade
    uint32_t upgrade_offset;
    
    //! The total size of the upgrade data
    uint32_t upgrade_size;
    
    //! Set to true when dirac calibration is on
    bool dirac_calibration;
    
    //! Stores analogue direct state during dirac calibration
    bool dirac_analogue_direct;
    
    // Stores the volume when the device is muted when in airplay, as we set the volume
    // to 0 becaiuse airplay deos not send a volume 0 but goes from 5 to MUTE....
    uint8_t airplay_volume_pre_mute;
    
    // Set t0 true when a Dirac connected message is received - used to stop the
    // dirac disconnected message triggering an input change during startup when
    // the subscription state change message is received
    bool dirac_connected;
    
} AVATAR_DATA;

static AVATAR_DATA avatar_data =
{
    .player_status = AVATAR_PLAYER_STATUS_STOPPED,
    .current_playing_info_refresh = false,
    .album_name[0] = 0,
    .artist_name[0] = 0,
    .track_name[0] = 0,
    .track_length = 0,
    .track_time = 0,
    .upgrade_offset = 0,
    .upgrade_size = 0,
    .dirac_calibration = false,
    .dirac_analogue_direct = false,
    .dirac_connected = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward Declarations
static void app_commands_process_upgrade_status(AVATAR_MESSAGE* message);
static void app_commands_update_player_metadata(AVATAR_MESSAGE* message);

/**-----------------------------------------------------------------------------
 * Tests whether the current Track/Album/Artist have valid values or are empty
 * @return 
 */
static bool app_commands_has_player_metadata()
{
    bool result = true;
    
    // Note the chromecast app name should have been processed (if available)
    // before this is called
    char chromecast_app_name[DISPLAY_LINE_LEN + 1];
    memset(chromecast_app_name, 0, DISPLAY_LINE_LEN + 1);
    avatar_message_get_chromecast_app_name_string(chromecast_app_name, DISPLAY_LINE_LEN);

    
    if (((strlen(avatar_data.track_name) == 0) || 
         (strncmp(avatar_data.track_name, "Chromecast built-in", 19) == 0)) && 
        (strlen(avatar_data.album_name) == 0) &&
        (strlen(avatar_data.artist_name) == 0) &&
        (strlen(chromecast_app_name) == 0))  
    {
        result = false;
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Callback processing for  Audio Input message
 * @param message
*/
static void app_commands_process_audio_input(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint8_t audio_input;
    uint16_t len;
    
    //------------------------------------------------------------
    // Player status processing
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_AUDIO_INPUT, 
            &encoding, 
            &audio_input, 
            sizeof(uint8_t)))
    {
        if ((app_context_get_app_state() == APP_STATE_RUNNING) && (avatar_get_state() == AVATAR_STATE_RUNNING) && (audio_input == AVATAR_INPUT_INTERNAL))
        {
            // 20190920
            // This message is being sent after we have woken up from sleep as part of
            // the response to us re-subscribing.  When it does that it causes the
            // input to be set to the Avatar when it shouldn't.  This check, checks for
            // the presence of the SSID message in the packet - this is only sent 
            // as part of the subscription response, not when normally changing to
            // casting.
            if (!avatar_message_has_state(message, AVATAR_STATE_ID_SSID))
            {
                // Tell the display to refresh 
                avatar_data.current_playing_info_refresh = true;
                input_control_select_input(AUDIO_INPUT_AVATAR);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for The Citation Select Audio Source message. 
 * @param message
*/
static void app_commands_process_audio_source(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint8_t audio_source;
    uint16_t len;
    
    //------------------------------------------------------------
    // Player status processing
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_AUDIO_SOURCE, 
            &encoding, 
            &audio_source, 
            sizeof(uint8_t)))
    {
        // This if statement is a fix for when we bring the device out of sleep to stop the Avatar becoming the
        // input if it was'nt when it went to sleep
        if ((app_context_get_app_state() == APP_STATE_RUNNING) && (avatar_get_state() == AVATAR_STATE_RUNNING))
        {
            if ((audio_source != AVATAR_AUDIO_SOURCE_UNKNOWN) &&
                (audio_source != AVATAR_AUDIO_SOURCE_AUX))
            {
                // Clear down browsing state 
                if ((audio_source != AVATAR_AUDIO_SOURCE_UPNP) &&
                    (audio_source != AVATAR_AUDIO_SOURCE_USB))
                {
                    app_commands_browsing_reset();
                }

                input_control_select_input(AUDIO_INPUT_AVATAR);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Crestron config
 * @param message
*/
static void app_commands_process_crestron(AVATAR_MESSAGE* message)
{
    // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    uint8_t crestron_config;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_CRESTRON_CONFIGURE, 
            &encoding, 
            (uint8_t*)&crestron_config, 
            sizeof(uint8_t)))
    {
        
        switch((AVATAR_CRESTRON_CONFIGURATION_STATUS)crestron_config)
        {
            case AVATAR_CRESTRON_CONFIGURATION_STATUS_START:
            {
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR CRESTRON CONFIGURATION Start --");
                // Switch to the Crestron Configure menu
                app_context_set_mode(APP_CONTEXT_MODE_MENU);
                app_context_set_menu(APP_COMMAND_MENU_CRESTRON_CONFIG); 
                break;
            }
            
            case AVATAR_CRESTRON_CONFIGURATION_STATUS_END:
            {
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR CRESTRON CONFIGURATION End --");
                // Switch to the Default menu
                app_commands_display_default_screen(NULL);
                break;
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Dirac Calibration
 * @param message
*/
static void app_commands_process_dirac_calibration(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint16_t len;
    uint8_t dirac_calibration;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_DIRAC_CALIBRATION, 
            &encoding, 
            (uint8_t*)&dirac_calibration, 
            sizeof(uint8_t)))
    {
        switch((AVATAR_DIRAC_CALIBRATION_STATUS)dirac_calibration)
        {
            case AVATAR_DIRAC_CALIBRATION_STATUS_STARTED:
            {
                logger_log_debugf(0, "AVATAR DIRAC_CALIBRATION Started %d %d --", input_control_get_current_definition()->type, analogue_direct_get_active());
                
                input_control_mute_inputs();
                
                avatar_data.dirac_calibration = true;
                
                if (analogue_direct_get_active())
                {
                    avatar_data.dirac_analogue_direct = true;
                    analogue_direct_set_active(false);
                
                    input_control_reset_input();
                }

                // Switch to the Dirac Calibration menu
                app_context_set_mode(APP_CONTEXT_MODE_MENU);
                app_context_set_menu(APP_COMMAND_MENU_DIRAC_CALIBRATION); 
                break;
            }

            case AVATAR_DIRAC_CALIBRATION_STATUS_STOPPED:
            {
                logger_log_debugf(0, "AVATAR DIRAC_CALIBRATION Stopped --");

                // Check we've got a curve name returned too.
                // Check whether this is the first dirac calibration
                // Enable EQ curve
                //dirac_set_curve_calibrated(curve, curve_name);

                // Switch to the Default menu 
                // We sometimes get a Calibration Stopped when starting up so
                //  make sure the system is running..
                if (app_context_is_running() && avatar_data.dirac_calibration)
                {
logger_log_debugf(0, "AVATAR DIRAC_CALIBRATION Stopped %d --", input_control_get_current_definition()->type);
                
                    if (avatar_data.dirac_analogue_direct)
                    {
                        analogue_direct_set_active(true);
                    }

//input_control_unmute_inputs(input_control_get_current_definition()->type);
                    input_control_reset_input();
                    
                    // TODO REMOVE once happy with Dirac Client Status
//                    app_context_set_mode(APP_CONTEXT_MODE_MENU);
//                    app_context_set_menu(APP_COMMAND_MENU_DIRAC_CALIBRATION); 
                }

                avatar_data.dirac_calibration = false;
                
                break;
            }

            case AVATAR_DIRAC_CALIBRATION_STATUS_UPDATED:
            {
                // This message should in include a new list of updated
                // curves that will be processed by avatar_message_process_dirac_curve_names()
                //dirac_remove_curve_calibration(curve);
                break;
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Dirac Client Status
 * @param message
*/
static void app_commands_process_dirac_client_status(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint16_t len;
    uint8_t dirac_client_status;
            
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_DIRAC_CLIENT_STATUS, 
            &encoding, 
            (uint8_t*)&dirac_client_status, 
            sizeof(uint8_t)))
    {
        if ((app_context_is_running()) && (avatar_get_state() == AVATAR_STATE_RUNNING))
        {
            switch((AVATAR_DIRAC_CLIENT_STATUS)dirac_client_status)
            {
                case AVATAR_DIRAC_CLIENT_STATUS_DISCONNECTED:
                {
                    logger_log_debugf(0, "AVATAR DIRAC_CLIENT_STATUS Disconnected --");

                    if (avatar_data.dirac_connected)
                    {
                        if (avatar_data.dirac_analogue_direct)
                        {
                            input_control_set_input_analogue_direct(input_control_get_selected_input(), true);
                            avatar_data.dirac_analogue_direct = false;  // Reset it
                        }
                        
                        //BUG 3630 - no audio after dirac app is quit - Matjaz requested we reset the input at the end of calibration
                        if (input_control_get_selected_input() != AUDIO_INPUT_AVATAR)
                        {
                            input_control_reset_input();
                        }
                        else
                        {
                            // It was woken up by the avatar and the input has been set in STANDBY_STATE_ACTIVE
                            if (app_commands_get_player_status() != AVATAR_PLAYER_STATUS_STOPPED)
                            {
                                input_control_reset_input();
                            }
                            else
                            {
                                input_control_select_input(input_control_get_previous_input()); 
                            }
                        }
                    }
                    
                    avatar_data.dirac_connected = false;

                    // The default screen will be displayed as part of the process of switching inputs.
                    // If done here is appears to cause issues on the ST60 restart, though not sure why...
                    //app_commands_display_default_screen(NULL);
                    break;
                }

                case AVATAR_DIRAC_CLIENT_STATUS_CONNECTED:
                {
                    logger_log_debugf(0, "AVATAR DIRAC_CLIENT_STATUS Connected --");

                    avatar_data.dirac_connected = true;

                    // Switch to the Dirac Calibration menu
                    app_context_set_mode(APP_CONTEXT_MODE_MENU);
                    app_context_set_menu(APP_COMMAND_MENU_DIRAC_CALIBRATION); 
                
                    if (analogue_direct_get_active())
                    {
                        // Rather than set analogue directly we need to change the input to have analogue direct off.....
                        avatar_data.dirac_analogue_direct = true;
                        
                        input_control_set_input_analogue_direct(input_control_get_selected_input(), false);

                        input_control_reset_input();
                    }

                    break;
                }
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Dirac Curve names request
 * @param message
*/
static void app_commands_process_dirac_curve_names(AVATAR_MESSAGE* message)
{
    if (avatar_message_has_state(message, AVATAR_STATE_ID_DIRAC_CURVE_NAMES))
    {
        // Only send the unsolicited message if the number of curves is greater than 1 (No Curve))
        if (dirac_get_num_curves() > 1)
        {
            app_commands_request_status(APP_COMMANDS_ROOM_EQ_NAMES, app_commands_room_eq_names);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Mute (0x0007)
 * @param message
*/
static void app_commands_process_mute(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint8_t mute_state;
    
    // Don't unmute while the avatar is loading
    if (app_context_is_running() &&
        (avatar_get_state() == AVATAR_STATE_RUNNING))
    {
        if (timers_has_expired(TIMER_INDEX_AVATAR_VOLUME_CHANGE))
        {
            if (avatar_message_get_state_data(message, 
                    AVATAR_STATE_ID_MUTE, 
                    &encoding, 
                    (uint8_t*)&mute_state, 
                    sizeof(uint8_t)))
            {
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Mute %d --", mute_state);

                switch(mute_state)
                {
                    case 0:
                    {
                        avatar_set_processing_mute(true);
                        output_control_unmute_outputs();
                        volume_unmute(); // To get the display
                        
                        // BUG 3379
                        if (avatar_message_get_audio_source() == AVATAR_AUDIO_SOURCE_AIRPLAY) //(volume_get() <= 5)
                        {
                            STRUCTURED_MESSAGE message;
                            structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
                            message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message.command = APP_COMMANDS_VOLUME;
                            message.data[0] = avatar_data.airplay_volume_pre_mute;
                            message.data_length = 1;

                            avatar_set_processing_volume(true);
                            app_commands_structured_message_callback(&message);
                            avatar_set_processing_volume(false);
                        }
                        
                        input_control_set_avatar_volume(volume_get());
                        avatar_set_processing_mute(false);

//                        if (avatar_message_get_audio_source() == AVATAR_AUDIO_SOURCE_ROON)
//                        {
                            // BUG 3380
                            //input_control_set_avatar_volume(volume_get());
                            //avatar_set_volume(volume_get());   // BUG 
//                        }
                        break;
                    }

                    case 1:
                    {
                        // Set the volume to 0
                        // This is required because on some casting apps (AIRPLAY) when you adjust the volume
                        // the final step to 0 is replaced by a MUTE.  It looks like
                        // volume commands are pretty consistent  in ramping down to 0
                        // so check if the volume is currently <= 5 and only do it then
                        if (avatar_message_get_audio_source() == AVATAR_AUDIO_SOURCE_AIRPLAY) //(volume_get() <= 5)
                        {
                            // BUG 3379
                            avatar_data.airplay_volume_pre_mute = volume_get();
                            
                            STRUCTURED_MESSAGE message;
                            structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
                            message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                            message.command = APP_COMMANDS_VOLUME;
                            message.data[0] = 0;
                            message.data_length = 1;

                            avatar_set_processing_volume(true);
                            app_commands_structured_message_callback(&message);
                            avatar_set_processing_volume(false);
                        }
                        else if (avatar_message_get_audio_source() == AVATAR_AUDIO_SOURCE_ROON)
                        {
                            // BUG 3380
                            input_control_set_avatar_volume(0);
                        }
                        
                        // Mute the outputs
                        avatar_set_processing_mute(true);
                        output_control_mute_outputs(true);
                        volume_mute();  // To get the display
                        input_control_set_avatar_volume(0);
                        avatar_set_processing_mute(false);
                        break;
                    }
                }
            }
            else
            {
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Suppressing Mute %d --", mute_state);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Player Status
 * @param message
*/
static void app_commands_process_player_status(AVATAR_MESSAGE* message)
{
    // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    void* data;
    uint16_t len;
    
    if (avatar_data.dirac_calibration)
    {
        logger_log_debugf(0, "AVATAR Ignoring Player Status during Dirac Calibration --");
        return;
    }
    
    //------------------------------------------------------------
    // Player status processing
    if (avatar_message_get_state_value(message, 
            AVATAR_STATE_ID_PLAYER_STATUS, 
            &encoding, 
            &data, 
            &len))
    {
        // Stop the stopped check timer as we are actually doing something
        timers_update_timeout(TIMER_INDEX_AVATAR_STOPPED_CHECK, TIMER_AVATAR_STOPPED_CHECK_TIMEOUT_MS, TIMERS_TIMEOUT_RESET_INACTIVE);

        // Store previous player status so it can be compared
        AVATAR_PLAYER_STATUS prev_player_status = avatar_data.player_status;
        
        memcpy(&avatar_data.player_status, data, len);
        
        logger_log_debugf(0, "AVATAR PLAYER STATUS Status %s (%d %d) --", 
                AVATAR_PLAYER_STATUS_STRINGS[avatar_data.player_status], len, 
                encoding);
        
        bool set_mute = false;  // Determines whether we need try and mute/unmute at all
        bool mute = false;      // Default to unmute
        
        if (avatar_data.player_status == AVATAR_PLAYER_STATUS_PLAYING)
        {
            // Check that we're on an AVATAR input as the AVatar is sending 
            // player_status PLAYING even when its switched to the external audio source
            if (avatar_message_get_audio_source() != AVATAR_AUDIO_SOURCE_AUX)
            {
                app_commands_update_player_metadata(message);
    //            
    //            sprintf(avatar_data.current_playing_info, "%s%c%s%c%s ", avatar_data.track_name, DISPLAY_NEWLINE_CHARACTER, avatar_data.artist_name, DISPLAY_NEWLINE_CHARACTER, avatar_data.album_name);
    //            
    //            if (strncmp(current_playing_info, avatar_data.current_playing_info, DISPLAY_MAX_SCROLL_TEXT_LEN) != 0)
    //            {
    //                avatar_data.current_playing_info_refresh = true;
    //            }

                //input_control_select_input(AUDIO_INPUT_AVATAR);
                //TODO:  MIMETYPE not being responded to at the moment..or rarely, and then its audio/unknown!!! avatar_request_mime_type();

                set_mute = true;
                mute = false;   // Unmute

                // Check that were actually playing something
                if (!app_commands_has_player_metadata())
                {
//                    logger_log_debugf(0, "AVATAR PLAYER STATUS STOPPED A --");
//                    logger_process();
                    avatar_data.player_status = AVATAR_PLAYER_STATUS_STOPPED;
                }
                else
                {
                    // Bodge to switch to NET input - will be replaced by Audio Input command
                    if (input_control_get_selected_input() != AUDIO_INPUT_AVATAR)
                    {
                        input_control_select_input(AUDIO_INPUT_AVATAR);
                    }

                    //BUG 3228 - Attempt o prevent Avatar sending 52 when casting
                    //TODO - Note this does not solve the problem see JIRA bug 101 https://extern.streamunlimited.com/jira/projects/SSDKARCAM/issues/SSDKARCAM-101
                    input_control_set_avatar_volume(volume_get());
                    //avatar_set_volume(volume_get());

                }
            }
            else
            {
                // Were not
//                logger_log_debugf(0, "AVATAR PLAYER STATUS STOPPED B --");
//                logger_process();
                avatar_data.player_status = AVATAR_PLAYER_STATUS_STOPPED;
            }
        }
        else if (avatar_data.player_status == AVATAR_PLAYER_STATUS_PAUSED)
        {
            app_commands_update_player_metadata(message);
                
            // Check that were actually playing something
            if (!app_commands_has_player_metadata())
            {
//                logger_log_debugf(0, "AVATAR PLAYER STATUS STOPPED C --");
//                logger_process();
                avatar_data.player_status = AVATAR_PLAYER_STATUS_STOPPED;
            }
        }
        else if (avatar_data.player_status == AVATAR_PLAYER_STATUS_CHANGING)
        {
            // This attempts to mute the output while the streamer changes tracks as we
            // have had it reported that there is some digital noise in between tracks.
            // Normally the Avatar sends a Changing status when it leaves a track 
            // and another when it starts the next track, though it is only about
            // 200ms
            
            // All paths require a mute change
            set_mute = true;
            
            if (prev_player_status == AVATAR_PLAYER_STATUS_PLAYING)
            {
//logger_log_debugf(0, "AVATAR PLAYER STATUS CHANGING A --");
//logger_process();
                mute = true;
            }
            else if ((prev_player_status == AVATAR_PLAYER_STATUS_CHANGING) ||
                     (prev_player_status == AVATAR_PLAYER_STATUS_STOPPED) ||
                     (prev_player_status == AVATAR_PLAYER_STATUS_PAUSED))   // BUG 3743
            {
                /*
                 * If the current Track/Artist/Album is not blank and is not the same 
                 * as the previous values then we will unmute, otherwise we will mute
                 */

                // Copy the current values
                char album_name[AVATAR_PLAYING_INFO_MAX_LENGTH];
                char artist_name[AVATAR_PLAYING_INFO_MAX_LENGTH];
                char track_name[AVATAR_PLAYING_INFO_MAX_LENGTH];
                memcpy(album_name, avatar_data.album_name, AVATAR_PLAYING_INFO_MAX_LENGTH);
                memcpy(artist_name, avatar_data.artist_name, AVATAR_PLAYING_INFO_MAX_LENGTH);
                memcpy(track_name, avatar_data.track_name, AVATAR_PLAYING_INFO_MAX_LENGTH);

                // Get the new values
                app_commands_update_player_metadata(message);

                // See what sort of change we have and whether we need to mute/unmute
                mute = true;
//logger_log_debugf(0, "AVATAR PLAYER STATUS CHANGING B --");
//logger_process();

                if ((strcmp(avatar_data.track_name, track_name) != 0) || 
                    (strcmp(avatar_data.album_name, album_name) != 0) ||
                    (strcmp(avatar_data.artist_name, artist_name) != 0))
                {
                    // Then Unmute if the album and artist are not blank
                    // Is it an empty track (e.g. Album/Artist blank, Track blank or "Chromecast built-in")
                    // Just check Album/Artist for the moment
                    if (app_commands_has_player_metadata())
                    {
//logger_log_debugf(0, "AVATAR PLAYER STATUS CHANGING B1 --");
//logger_process();
                        // This is NOT a blank track so UnMute
                        mute = false;
                    }
                }
                else
                {
//logger_log_debugf(0, "AVATAR PLAYER STATUS CHANGING B21 %d --", prev_player_status);
//logger_process();
                    // This happens when it first connects
                    mute = false;
                }
            }
            else
            {
//logger_log_debugf(0, "AVATAR PLAYER STATUS CHANGING C --");
//logger_process();
                // Just to be sure
                mute = false;
            }
        }
        
        // Now check stopped status
        if (avatar_data.player_status == AVATAR_PLAYER_STATUS_STOPPED)
        {
            //logger_log_debugf(0, "AVATAR avatar_data.player_status == AVATAR_PLAYER_STATUS_STOPPED input %d --", avatar_message_get_audio_input());
                
            // BUG 3679 Removed check for ausio_input = internal as we now check the current menu in
            // app_commands_avatar_process - this allows the check to go back to browsing menu
            // if (avatar_message_get_audio_input() == AVATAR_INPUT_INTERNAL)
            // {
                // Start a timer after which we go back to the previous input
                // We cant do this immediatly because the avatar sends a stop message
                // when scrolling backwards - ughhhhh!
                // see app_commands_avatar_process();
                timers_reset(TIMER_INDEX_AVATAR_STOPPED_CHECK);
                //input_control_select_input(input_control_get_previous_input()); 
            //}
        }
        
        // Now perform the Mute/Unmute if we are not in an input change
        //TODO Remove logger_log_debugf(0, "AVATAR VOLUME %d %d %d %d --", set_mute, input_control_in_switch(), volume_in_ramp(), input_control_state());
        if (set_mute && !input_control_in_switch() && !volume_in_ramp())
        {
            if (mute)
            {
//logger_log_debugf(0, "AVATAR PLAYER STATUS CHANGING MUTE --");
//logger_process();
                mutes_mute(MUTE_INDEX_VOLUME);
            }
            else
            {
//logger_log_debugf(0, "AVATAR PLAYER STATUS CHANGING UNMUTE --");
//logger_process();
                mutes_unmute(MUTE_INDEX_VOLUME);
            }  
        }
        
        // Send unsolicited message
        if (avatar_data.player_status != prev_player_status)
        {
            app_commands_request_status(APP_COMMANDS_PLAYBACK_STATUS, app_commands_player_status);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Set Sample Rate
 * @param message
*/
static void app_commands_process_set_sample_rate_response(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint8_t sample_rate;

    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_SET_SAMPLE_RATE, 
            &encoding, 
            (uint8_t*)&sample_rate, 
            sizeof(uint8_t)))
    {
        // We have told the Avatar that the sample rate has changed and it 
        // has now responded so we can un-mute the inputs.
        // This has only been tested with input change - not yet sure if it 
        //      will affect Sample rate changes.
        //logger_log_debugf(0, "VOLUME Sample Rate change Fade In --");
        //input_control_unmute_inputs(input_control_get_current_input_type()); 
        //output_control_unmute_outputs();
        //volume_fade_in();
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Track Length
 * @param message
*/
static void app_commands_process_track_length(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;

    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_TRACK_LENGTH, 
            &encoding, 
            (uint8_t*)&avatar_data.track_length, 
            sizeof(uint64_t)))
    {
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Track Time
 * @param message
*/
static void app_commands_process_track_time(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
     
    if (avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_TRACK_TIME, 
            &encoding, 
            (uint8_t*)&avatar_data.track_time, 
            sizeof(uint64_t)))
    {
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Upgrade Data request
 * @param message
*/
static void app_commands_process_process_upgrade_data(AVATAR_MESSAGE* message)
{
    // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    uint16_t actual_len;
    
    uint32_t upgrade_data_size;
    uint32_t offset = UINT32_MAX;
    void* data;
            
    // Get the total data size to be transferred
    actual_len = avatar_message_get_state_data(message, 
                    AVATAR_STATE_ID_UPGRADE_DATA_SIZE, 
                    &encoding, 
                    (uint8_t*)&avatar_data.upgrade_size, 
                    sizeof(uint32_t));
    
    if (actual_len > 0)
    {
        // This should give us the total size of the upgrade data.  Note the 
        // size of the data in this message is inside the BIN16 field.
        // We don't actually need this (I think..))
            
        // Get the Offset of this data packet in the total data
        actual_len = avatar_message_get_state_data(message, 
                        AVATAR_STATE_ID_UPGRADE_DATA_OFFSET, 
                        &encoding, 
                        (uint8_t*)&avatar_data.upgrade_offset, 
                        sizeof(uint32_t));
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Volume
 * @param message
*/
static void app_commands_process_upgrade_status(AVATAR_MESSAGE* message)
{
    // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    uint16_t actual_len;
    uint8_t data;
    
    actual_len = avatar_message_get_state_data(message, 
                    AVATAR_STATE_ID_UPGRADE_STATUS, 
                    &encoding, 
                    (uint8_t*)&data, 
                    sizeof(uint8_t));
    
    if (actual_len > 0)
    {
// Does the message include a Player Status change
        switch ((AVATAR_UPGRADE_STATUS)data)
        {
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_AVAILABLE:
            case AVATAR_UPGRADE_STATUS_DUAL_UPGRADE_AVAILABLE:
            case AVATAR_UPGRADE_STATUS_DEVICE_UPGRADE_AVAILABLE:
            {
                // Indicate that an upgrade is available
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_START:
            {
                avatar_data.upgrade_offset = 0;
                avatar_data.upgrade_size = 0;

                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Device Firmware Upgrade Starting --");
                
                // Set the UPGRADE_STATUS Setting
                //bool upgrade_status = true; // In Upgrade
                //settings_set(SETTING_UPGRADE_STATUS, &upgrade_status);
                //settings_save();        
                
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                app_commands_hdmi_arc_suppress_handshake(true);
#endif
                
                // Put into process mode to stop menu timing out
                app_context_set_mode(APP_CONTEXT_MODE_PROCESS);
                app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_IN_PROGRESS);
                
                // Stop the Splash screen being abandoned while the upgrade is taking place.
                timers_suspend(TIMER_INDEX_DISPLAY_TEST_TIMEOUT);
                
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_IN_PROGRESS:
            {
                //This indicates that the Avatar is updating
                // Set the UPGRADE_STATUS Setting - NOT NEEDED HERE AS THE AVATAR UPGRADE DOES NOT INITIATED A RESTART
                //bool upgrade_status = true; // In Upgrade
                //settings_set(SETTING_UPGRADE_STATUS, &upgrade_status);
                //settings_save();
                
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                app_commands_hdmi_arc_suppress_handshake(true);
#endif
                
                //app_context_set_mode(APP_CONTEXT_MODE_PROCESS);
                // Setting to PROCESS mode means the menu never changes when updating the avatar-> updating the MCU - try setting the timeout to 2 mins
                timers_update_timeout(TIMER_INDEX_MENU_INACTIVE, 120000, TIMERS_RESET_IMMEDIATE);
                app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_IN_PROGRESS);
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_COMPLETE:
            {
                // Check whether we are expecting an Upgrade Complete.  This is 
                // because there's a bug in the Avatar (rel 104) that sends it every time.
                bool upgrade_status = true; 
                settings_get(SETTING_UPGRADE_STATUS, &upgrade_status);
                if (!upgrade_status)
                {
                    // This means the avatar firmware upgrade is complete and there
                    // hasnt been a MCU upgrade
                    logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Avatar Firmware Upgrade Complete --");
                    app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_COMPLETE);
                    app_context_set_mode(APP_CONTEXT_MODE_MENU);
                
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                    app_commands_hdmi_arc_suppress_handshake(false);
#endif                    
                }
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_AVATAR_UPGRADE_FAIL:
            {
                logger_log_errorf(LOGGER_CODE_AVATAR, "AVATAR Avatar Firmware Upgrade Fail --");
                app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_FAIL);
                app_context_set_mode(APP_CONTEXT_MODE_MENU);
                
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                app_commands_hdmi_arc_suppress_handshake(false);
#endif
                break;
            }
            
            case AVATAR_UPGRADE_STATUS_DEVICE_UPGRADE_TRANSFER_COMPLETE:
            {
                // Note this is NOT sent by the AVatar - it is generated internally by 
                // avatar_message_process_upload_end_callback
                // Check the upload was successful and start the upgrade
                // This is just checking for a main MCU upgrade - its not working if
                // were just updating the HDMI ARC for instance
                if ((upgrade_metadata_get_upgrade_state() == UPGRADE_METADATA_STATE_AVAILABLE_INSTALL) ||
                    (upgrade_metadata_get_upgrade_state() == UPGRADE_METADATA_STATE_NONE))
                {
                    //This is just before the Device reboots
                    app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_COMPLETE);

                    avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_UPGRADE_STARTING);

                    // Set the UPGRADE_STATUS Setting
                    bool upgrade_status = true; // In Upgrade
                    settings_set(SETTING_UPGRADE_STATUS, &upgrade_status);
                    settings_save();

                    logger_log_debugf(0, "AVATAR Upgrade Transfer Complete - Rebooting --");
                    if (app_context_get_app_state() == APP_STATE_POWEREDDOWN)
                    {
                        logger_log_debugf(0, "AVATAR Upgrade Setting Sleep On Restart --");
                        // Set a flag so that after the restart we go back into 
                        // sleep mode
                        bool sleep = true;
                        settings_set(SETTING_UPGRADE_SLEEP_ON_RESTART, &sleep);
                        settings_save();
                    }
                    
                    // Reset the Reset Pause - This timer has been set to 5 seconds 
                    // to give time for the HDMI ARC upgrade to complete when in 
                    // recovery mode - If this needs shortening then we will need
                    // to move the avatar_message_generate_transfer_complete call
                    // in to the upgrade end callback rather than when processing
                    // the data packet.
                    timers_reset(TIMER_INDEX_RESTART_PAUSE);
                    app_context_set_app_state(APP_STATE_PRE_RESTART);
                }
                else
                {
                    // Display Upgrade failed
                    app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_FAIL);
                    app_context_set_mode(APP_CONTEXT_MODE_MENU);
                }
                
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                app_commands_hdmi_arc_suppress_handshake(false);
#endif
                
                break;
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Callback processing for Volume
 * @param message
*/
static void app_commands_process_volume(AVATAR_MESSAGE* message)
{
    // Does the message include a Player Status change
    AVATAR_ENCODING encoding;
    void* data;
    uint16_t len;

    // Check its running as the Avatar sends out a Volume 0 during initialisation
    // This may only be when the Avatar is first flashed and that it stores its 
    // value from then on - needs checking but we need . Basically we only need 
    // to respond to volume commands when the Avatar is on its Internal input
    // Note: The check for SSID is because we want to ignore VOLUME message during startup
    //  however a volume message is sent as a state change after subscribing to 
    //  message - this is the only time it occurs in a message with the SSID so
    // SSID can be used to ignore the volume change.
    if (app_context_is_running() &&
        (avatar_get_state() == AVATAR_STATE_RUNNING) &&
        !avatar_message_has_state(message, AVATAR_STATE_ID_SSID) &&
        !input_control_in_switch() /* BUG 3722 */)
    {
        //if (timers_has_expired(TIMER_INDEX_AVATAR_VOLUME_CHANGE) && timers_has_expired(TIMER_INDEX_AVATAR_VOLUME_CHANGE_INCOMING))
        //{
        // Note: If these messages are received too fast and logging is on it can cause
        // the unit to die!!!
        if (timers_has_expired(TIMER_INDEX_AVATAR_VOLUME_CHANGE))
        {
            if (avatar_message_get_state_value(message, 
                AVATAR_STATE_ID_VOLUME, 
                &encoding, 
                &data, 
                &len))
            {
                uint8_t volume = 0;
                memcpy(&volume, data, len);

// BUG 3737                
// Build 772 testing      
// 20201022 - re-added following email from Pete to Matjaz 22/10/2020 10:53
                // Google Cast spec 5.5 implies that we need to re-map the range sent to 
                // the avatar based on the MAXIMUM_VOLUME setting
                uint16_t mapped_volume = volume * volume_get_maximum_net_volume();
                volume = min(mapped_volume / (uint16_t)VOLUME_MAX, (uint16_t)volume_get_maximum_net_volume());

                logger_log_debugf(0, "VOLUME Avatar %d %d %d --", mapped_volume, volume, volume_get_maximum_net_volume());
                
                STRUCTURED_MESSAGE message;
                structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
                message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                message.command = APP_COMMANDS_VOLUME;
                message.data[0] = volume;
                message.data_length = 1;

                avatar_set_processing_volume(true);
                app_commands_structured_message_callback(&message);
                avatar_set_processing_volume(false);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Updates the Track/Album/Artist info
 * @param message
 */
static void app_commands_update_player_metadata(AVATAR_MESSAGE* message)
{
    AVATAR_ENCODING encoding;
    uint16_t len;

    len = AVATAR_PLAYING_INFO_MAX_LENGTH;
    
    avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_TRACK_NAME, 
            &encoding, 
            (uint8_t*)avatar_data.track_name, 
            len);

    avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_ALBUM_NAME, 
            &encoding, 
            (uint8_t*)avatar_data.album_name, 
            len);

    avatar_message_get_state_data(message, 
            AVATAR_STATE_ID_ARTIST_NAME, 
            &encoding, 
            (uint8_t*)avatar_data.artist_name, 
            len);
    
    // Get a pre change copy to see if it changes
    char current_playing_info[DISPLAY_MAX_SCROLL_TEXT_LEN];
    memcpy(current_playing_info, avatar_data.current_playing_info, DISPLAY_MAX_SCROLL_TEXT_LEN);

    // Build the track info (note only adding new lines where necessary)
    uint16_t idx = 0;
    
    if ((strlen(avatar_data.track_name) > 0) && (strncmp(avatar_data.track_name, "Chromecast built-in", 19) != 0))
    {
        sprintf(&avatar_data.current_playing_info[idx], "%s%c ", avatar_data.track_name, DISPLAY_NEWLINE_CHARACTER);
        idx += strlen(avatar_data.track_name) + 1;
    }
    
    if (strlen(avatar_data.artist_name) > 0)
    {
        sprintf(&avatar_data.current_playing_info[idx], "%s%c ", avatar_data.artist_name, DISPLAY_NEWLINE_CHARACTER);
        idx += strlen(avatar_data.artist_name) + 1;
    }
   
    if (strlen(avatar_data.album_name) > 0)
    {
        sprintf(&avatar_data.current_playing_info[idx], "%s%c ", avatar_data.album_name, DISPLAY_NEWLINE_CHARACTER);
        idx += strlen(avatar_data.album_name) + 1;
    }
    
    // Moved to top line of info
//    if (avatar_message_get_mqa_state() == AVATAR_MQA_STATE_NO_MQA)
//    {
//        uint8_t mime_type_idx = avatar_message_get_mime_type_index();
//        if (mime_type_idx != 0)
//        {
//            sprintf(&avatar_data.current_playing_info[idx], "%s%c ", AVATAR_MIME_TYPE_DISPLAY_STRINGS[avatar_message_get_mime_type_index()], DISPLAY_NEWLINE_CHARACTER);
//            idx += strlen(AVATAR_MIME_TYPE_DISPLAY_STRINGS[avatar_message_get_mime_type_index()]) + 1;
//        }
//    }
//    else
//    {
//        sprintf(&avatar_data.current_playing_info[idx], "%s%c ", AVATAR_MQA_STATE_STRINGS[avatar_message_get_mqa_state()], DISPLAY_NEWLINE_CHARACTER);
//        idx += strlen(AVATAR_MQA_STATE_STRINGS[avatar_message_get_mqa_state()]) + 1;
//    }
    
    if (idx > 0)
    {
        // Remove the last newline
        avatar_data.current_playing_info[idx - 1] = 0;
    }
    
    //sprintf(avatar_data.current_playing_info, "%s%c%s%c%s ", avatar_data.track_name, DISPLAY_NEWLINE_CHARACTER, avatar_data.artist_name, DISPLAY_NEWLINE_CHARACTER, avatar_data.album_name);

    // Checking to see if we need a refresh
    if (strncmp(current_playing_info, avatar_data.current_playing_info, DISPLAY_MAX_SCROLL_TEXT_LEN) != 0)
    {
        avatar_data.current_playing_info_refresh = true;
        
        // Send unsolicited messages
        app_commands_invoke_message(APP_COMMANDS_NOW_PLAYING, app_commands_now_playing, APP_COMMAND_NOW_PLAYING_TRACK);
        app_commands_invoke_message(APP_COMMANDS_NOW_PLAYING, app_commands_now_playing, APP_COMMAND_NOW_PLAYING_ARTIST);
        app_commands_invoke_message(APP_COMMANDS_NOW_PLAYING, app_commands_now_playing, APP_COMMAND_NOW_PLAYING_ALBUM);
        app_commands_invoke_message(APP_COMMANDS_NOW_PLAYING, app_commands_now_playing, APP_COMMAND_NOW_PLAYING_APPLICATION);
        app_commands_invoke_message(APP_COMMANDS_NOW_PLAYING, app_commands_now_playing, APP_COMMAND_NOW_PLAYING_SAMPLE_RATE);
        app_commands_invoke_message(APP_COMMANDS_NOW_PLAYING, app_commands_now_playing, APP_COMMAND_NOW_PLAYING_ENCODER);
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Performs a factory reset - run as a stateful_process
 */
void app_commands_avatar_arc_upgrade()
{
    // State 1 - start
    logger_log_info(LOGGER_CODE_FACTORY_RESET, "AVATAR ARC Upgrade Start --");
    
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
    app_commands_hdmi_arc_override_upgrade_version_check();
#endif
    app_commands_suppress_upgrade();
    
    avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_READY_TO_RECEIVE);
    
    app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
    app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_IN_PROGRESS);
}
   
/**-----------------------------------------------------------------------------
 * Draw the bespoke Info display
 */
void app_commands_avatar_draw_info(APP_COMMAND_MENU menu, char* top_line)
{
    if (menu == APP_COMMAND_MENU_INFO)
    {
        AUDIO_INPUT selected_input = input_control_get_selected_input();
        uint8_t volume = volume_get();
        bool analogue_direct = analogue_direct_get_active();
        INPUT_CONTROL_TYPE type = input_control_get_type();

        if (selected_input == INPUT_CONTROL_NUM_SETTINGS)
        {
            selected_input = AUDIO_INPUT_NONE;
        }

        // Clear the display
        display_clear(false);

        // Sort out Input name if we are in the Avatar
        char input_name[16];
        memset(input_name, 0, 16);
        if (selected_input == AUDIO_INPUT_AVATAR)
        {
            AVATAR_AUDIO_SOURCE avatar_source = avatar_message_get_audio_source();
            memcpy(input_name, AVATAR_AUDIO_SOURCE_STRINGS[avatar_source], strlen(AVATAR_AUDIO_SOURCE_STRINGS[avatar_source]));
        }
        else
        {
            memcpy(input_name, AUDIO_INPUT_STRINGS[selected_input], strlen(AUDIO_INPUT_STRINGS[selected_input]));
        }

        display_append_text(DISPLAY_LINE_1, 0, "%s", input_name);

        display_write_line(DISPLAY_LINE_1, false);
    }

    display_clear_line(DISPLAY_LINE_2);
    
    if (avatar_data.current_playing_info_refresh)
    {
        display_scroll_cancel();

        uint8_t scroll_focus_line = 0;
        
        if (top_line != NULL)
        {
            char current_playing_info[DISPLAY_MAX_SCROLL_TEXT_LEN];
            
            // Add the top line contents
            
            memcpy(current_playing_info, top_line, strlen(top_line));
            
            // pad out to the line len 
            memset(&current_playing_info[strlen(top_line)], DISPLAY_WHITESPACE_CHARACTER, DISPLAY_LINE_LEN); // DISPLAY_LINE_LEN is lazy but there will be the space
                    
            AVATAR_AUDIO_SOURCE audio_source = avatar_message_get_audio_source();
            char chromecast_app_name[DISPLAY_LINE_LEN];
            avatar_message_get_chromecast_app_name_string(chromecast_app_name, DISPLAY_LINE_LEN);
            
            // Draw the sample rate or the Casting app name if known
//            if ((audio_source == AVATAR_AUDIO_SOURCE_GOOGLECAST) && (strlen(chromecast_app_name) != 0))
            if (audio_source == AVATAR_AUDIO_SOURCE_GOOGLECAST)
            {
                uint8_t len = strlen(chromecast_app_name);
                memcpy(current_playing_info + DISPLAY_LINE_LEN - len, chromecast_app_name, len);
            }
            else if (audio_source == AVATAR_AUDIO_SOURCE_AIRPLAY)
            {
                // Don't add the sample rate
                scroll_focus_line = 1;
            }
            else
            {
                // Add the sample rate to the buffer (right aligned))
                //uint8_t len = strlen(SAMPLE_RATE_STRINGS[input_control_get_sample_rate()]);
                //strncpy(current_playing_info + DISPLAY_LINE_LEN - len, SAMPLE_RATE_STRINGS[input_control_get_sample_rate()], len);
                uint8_t len = strlen(SAMPLE_RATE_STRINGS[avatar_message_get_display_sample_rate()]);
                strncpy(current_playing_info + DISPLAY_LINE_LEN - len, SAMPLE_RATE_STRINGS[avatar_message_get_display_sample_rate()], len);
                
                //avatar_data.current_sample_rate = input_control_get_sample_rate();
                avatar_data.current_sample_rate = avatar_message_get_display_sample_rate();
            }

            // Force a newline
            current_playing_info[DISPLAY_LINE_LEN] = DISPLAY_NEWLINE_CHARACTER;
            
            // Initialise the rest of the data
            memset(&current_playing_info[DISPLAY_LINE_LEN + 1], 0, DISPLAY_MAX_SCROLL_TEXT_LEN - DISPLAY_LINE_LEN);
            
            // Add the currently playing info
            memcpy(&current_playing_info[DISPLAY_LINE_LEN + 1], avatar_data.current_playing_info, DISPLAY_MAX_SCROLL_TEXT_LEN - DISPLAY_LINE_LEN);
            
            display_scroll_text_block(DISPLAY_LINE_2, 0, 16, current_playing_info, 1000, 1000, true, scroll_focus_line);
        }
        else
        {
            display_scroll_text_block(DISPLAY_LINE_2, 0, 16, avatar_data.current_playing_info, 1000, 1000, true, scroll_focus_line);
        }
        
        avatar_data.current_playing_info_refresh = false;
    }
    else
    {
        bool suppress_sample_rate_correction = false;
        
        // Attempt to replace 
        char* text = (char*)display_get_scroll_text();
        if ((top_line != NULL) && (memcmp(text, top_line, DISPLAY_LINE_LEN) != 0))
        {
            // Replace the top line in the scroll text
            memcpy(text, top_line, DISPLAY_LINE_LEN);  // To make sure the icons are updated as well..
            
            // pad out to the line len 
            memset(&text[strlen(top_line)], ' ', DISPLAY_LINE_LEN - strlen(top_line)); // DISPLAY_LINE_LEN is lazy but there will be the space
                    
            AVATAR_AUDIO_SOURCE audio_source = avatar_message_get_audio_source();
            char chromecast_app_name[DISPLAY_LINE_LEN + 1];
            memset(chromecast_app_name, 0, DISPLAY_LINE_LEN + 1);
            avatar_message_get_chromecast_app_name_string(chromecast_app_name, DISPLAY_LINE_LEN);
            
            // Draw the sample rate or the Casting app name if known
//            if ((audio_source == AVATAR_AUDIO_SOURCE_GOOGLECAST) && (strlen(chromecast_app_name) != 0))
            if (audio_source == AVATAR_AUDIO_SOURCE_GOOGLECAST)
            {
                uint8_t len = strlen(chromecast_app_name);
                memcpy(text + DISPLAY_LINE_LEN - len, chromecast_app_name, len);
                
                suppress_sample_rate_correction = true;
            }
            else if (audio_source == AVATAR_AUDIO_SOURCE_AIRPLAY)
            {
                // Don't add the sample rate
                //TODO - Review if we want to put the sample rate back after airplay 
                //certification
                suppress_sample_rate_correction = true;
            }
            else
            {
                // Add the sample rate to the buffer (right aligned))
                //uint8_t len = strlen(SAMPLE_RATE_STRINGS[input_control_get_sample_rate()]);
                //strncpy(text + DISPLAY_LINE_LEN - len, SAMPLE_RATE_STRINGS[input_control_get_sample_rate()], len);
                uint8_t len = strlen(SAMPLE_RATE_STRINGS[avatar_message_get_display_sample_rate()]);
                strncpy(text + DISPLAY_LINE_LEN - len, SAMPLE_RATE_STRINGS[avatar_message_get_display_sample_rate()], len);
            }
        }
        
        //if (!suppress_sample_rate_correction && (avatar_data.current_sample_rate != input_control_get_sample_rate()))
        //{
        //   uint8_t len = strlen(SAMPLE_RATE_STRINGS[input_control_get_sample_rate()]);
        //    memcpy(text + DISPLAY_LINE_LEN - len, SAMPLE_RATE_STRINGS[input_control_get_sample_rate()], len);
        //    avatar_data.current_sample_rate = input_control_get_sample_rate();
        //}
        
        if (!suppress_sample_rate_correction && (avatar_data.current_sample_rate != avatar_message_get_display_sample_rate()))
        {
            uint8_t len = strlen(SAMPLE_RATE_STRINGS[avatar_message_get_display_sample_rate()]);
            memcpy(text + DISPLAY_LINE_LEN - len, SAMPLE_RATE_STRINGS[avatar_message_get_display_sample_rate()], len);
            avatar_data.current_sample_rate = avatar_message_get_display_sample_rate();
        }
    }
     
    display_scroll_process();
}
 
/**-----------------------------------------------------------------------------
 * Draw the avatar reset menu
 */
void app_commands_avatar_draw_reset(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();
 
    // Draw the top line
        // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);

    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    display_write_line(DISPLAY_LINE_1, false);

    // Draw bottom line
    display_clear_line(DISPLAY_LINE_2);
    
    display_write_line(DISPLAY_LINE_2, false);
}
 
/**-----------------------------------------------------------------------------
 * Draw the avatar reset menu
 */
void app_commands_avatar_draw_serial_number(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();
 
    // Draw the top line
        // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);

    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    display_write_line(DISPLAY_LINE_1, false);

    // Draw bottom line
    display_clear_line(DISPLAY_LINE_2);
    
    char serial_number[DISPLAY_LINE_LEN + 1];
    avatar_message_get_serial_number_string(serial_number, DISPLAY_LINE_LEN + 1);
    serial_number[DISPLAY_LINE_LEN] = 0;
    if (strlen(serial_number) > 0)
    {
        display_start_text(DISPLAY_LINE_2, serial_number);
    }
    else
    {
        display_start_text(DISPLAY_LINE_2, "Not Available");
    }
    
    display_write_line(DISPLAY_LINE_2, false);
}
 
/**-----------------------------------------------------------------------------
 * Draw the avatar upgrade menu
 */
void app_commands_avatar_draw_upgrade(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();
 
    // Draw the top line
    // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);

    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    display_write_line(DISPLAY_LINE_1, false);

    // Draw bottom line
    display_clear_line(DISPLAY_LINE_2);
    
    if (avatar_data.upgrade_size > 0)
    {
//        char progress[16];
//        sprintf(progress, "%d%", (avatar_data.upgrade_offset * 100) / avatar_data.upgrade_size);
//        display_append_text_right(DISPLAY_LINE_2, progress);
//        
       
        uint32_t pixels = (avatar_data.upgrade_offset * 80) / avatar_data.upgrade_size;
        
        int full_blocks = pixels / 5;
        int partial_pixels = pixels % 5;
        int x;
        for (x = 0; x < full_blocks; x++)
        {
            display_special_character(DISPLAY_LINE_2, x, DISPLAY_SPECIAL_CHARACTERS_BLOCK);
        }
        
        if ((full_blocks < DISPLAY_LINE_LEN) && (partial_pixels > 0))
        {
            DISPLAY_SPECIAL_CHARACTERS character = DISPLAY_SPECIAL_CHARACTERS_BLOCK;
            display_special_character_offset(DISPLAY_LINE_2, full_blocks, (char*)&character, 1, 1, 0, partial_pixels - 5, 0);
        }
    }
    
    display_write_line(DISPLAY_LINE_2, false);
}
 
/**-----------------------------------------------------------------------------
 * Draw the avatar upgrade menu
 */
void app_commands_avatar_draw_upgrade_available(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();
 
    // Draw the top line
        // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);

    // Check to see if an upgrade is available
    if (avatar_upgrade_available())
    {
        display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    }
    else
    {
        display_start_text(DISPLAY_LINE_1, "No Upgrade Reqd");
    }
    
    display_write_line(DISPLAY_LINE_1, false);

    // Draw bottom line
    display_clear_line(DISPLAY_LINE_2);
    
   // Check to see if an upgrade is available
    menu_display_draw_generic_navigation(menu_get_item(menu), MENU_SOFTKEYS_UP_DOWN, avatar_upgrade_available());
}
 
/**-----------------------------------------------------------------------------
 * Draw the avatar upgrade menu
 */
void app_commands_avatar_draw_upgrade_complete(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();
 
    // Draw the top line
    // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);

    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    display_write_line(DISPLAY_LINE_1, false);

    // Draw bottom line
    display_clear_line(DISPLAY_LINE_2);
    
    if (app_context_get_app_state() == APP_STATE_PRE_RESTART)
    {
        display_start_text(DISPLAY_LINE_2, "Restarting");
    }
    
    display_write_line(DISPLAY_LINE_2, false);
}

/**-----------------------------------------------------------------------------
 * Used to change the region
 */
bool app_commands_avatar_edit_region(STRUCTURED_MESSAGE* message)
{
    bool result = false;

    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = (int8_t)message->data[0];
        
        AVATAR_REGION region = avatar_message_get_region();

        int8_t current_value = (int8_t)region;

        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                current_value = value;
                region = (AVATAR_REGION)current_value;
                avatar_message_set_region(region);
                break;
            }
            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            {
                current_value++;
                if (current_value >= AVATAR_REGION_MAX)
                {
                    current_value = 0;
                }
                region = (AVATAR_REGION)current_value;
                avatar_message_set_region(region);
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                if (current_value > 0)
                {
                    current_value--;
                }
                else
                {
                    current_value = AVATAR_REGION_MAX - 1;
                }
                region = (AVATAR_REGION)current_value;
                avatar_message_set_region(region);
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
            {
                message->data[0] = (uint8_t)current_value;
                message->response = STATUS_UPDATE;

                app_commands_send_message(message);
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                app_commands_send_message(message);
                break;
            }
        }
                
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_update_change_state(&current_value, sizeof(int8_t));
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Manages the avatr reset - this is intended to be called repeatedly while in the 
 * AVATAR_FACTORY_RESET menu
 */
void app_commands_avatar_factory_reset()
{
    // Well probably need to put something up and wait for the Avatar to go through its reboot/handshaking again????
    //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR FACTORY RESET");
    
    switch (avatar_get_state())
    {
        case AVATAR_STATE_RUNNING:
        {
            avatar_factory_reset();
            
            // Show the resetting menu
            app_context_set_menu(APP_COMMAND_MENU_SYSTEM_INFO_AVATAR_FACTORY_RESETTING);

            break;
        }
        
        case AVATAR_STATE_HANDSHAKE:
        {
            // Wait
            break;
        }
        
        case AVATAR_STATE_CONFIGURATION:
        {
            avatar_configure();
            
            // go to main screen
            app_commands_display_default_screen(NULL);
            break;
        }
    }
}
     
/**-----------------------------------------------------------------------------
 * Gets the current status of the region as a string
 * @return 
 */
void app_commands_avatar_get_region_string(char* buffer, size_t size)
{
    AVATAR_REGION region = avatar_message_get_region();

    sprintf(buffer, "%s", AVATAR_REGION_STRINGS[min(region, AVATAR_REGION_MAX - 1)]);
}
    
/**-----------------------------------------------------------------------------
 * Trigger a refresh of the Avatar playing info
 */
void app_comands_avatar_invalidate_info()
{
    avatar_data.current_playing_info_refresh = true;
}

/**-----------------------------------------------------------------------------
 * Callback for avatar messages
 * @param message
 */
void app_commands_avatar_message_callback(AVATAR_MESSAGE* message)
{
    if (message->message_id != AVATAR_MESSAGE_ID_SET_STATES_RESPONSE)
    {
        // Player status processing
       app_commands_process_dirac_calibration(message);
       
       app_commands_process_dirac_client_status(message);

       // Player status processing
       app_commands_process_player_status(message);

       app_commands_process_audio_input(message);

       app_commands_process_audio_source(message);    

       // Volume processing
       app_commands_process_volume(message);

       // Mute processing
       app_commands_process_mute(message);

       app_commands_process_track_length(message);

       app_commands_process_track_time(message);

       // Crestron Config processing
       app_commands_process_crestron(message);

       app_commands_process_process_upgrade_data(message);

       app_commands_process_upgrade_status(message);
       
       app_commands_process_dirac_curve_names(message);
    }
    else
    {
       app_commands_process_set_sample_rate_response(message);
    }
}
   
/**-----------------------------------------------------------------------------
 * Command level avatar processing
 */
void app_commands_avatar_process()
{
    if (timers_has_expired(TIMER_INDEX_AVATAR_STOPPED_CHECK))
    {
        //logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR Check STOPPED Status");
        //logger_process();
        
        // Check its still stopped
        if (avatar_data.player_status == AVATAR_PLAYER_STATUS_STOPPED)
        {
            // Not media browsing...
            if ((avatar_message_get_audio_input() == AVATAR_INPUT_INTERNAL) && (menu_get_current_item()->menu_id == APP_COMMAND_MENU_DEFAULT))
            {
                logger_log_debugf(LOGGER_CODE_AVATAR, "AVATAR PLAYER STATUS STOPPED Switching to previous input");
                logger_process();
                input_control_select_input(input_control_get_previous_input()); 
            }
        }
        
        // Stop the stopped check timer as we are actually doing something
        timers_update_timeout(TIMER_INDEX_AVATAR_STOPPED_CHECK, TIMER_AVATAR_STOPPED_CHECK_TIMEOUT_MS, TIMERS_TIMEOUT_RESET_INACTIVE);
    }
}

/**-----------------------------------------------------------------------------
 * Used to start the Upgrade process
 */
void app_commands_avatar_upgrade()
{
    if (avatar_upgrade_available())
    {
        // Tell the avatar to start the upgrade
        avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_READY_TO_RECEIVE);

        // Switch to the Upgrade in progress menu
        avatar_data.upgrade_offset = 0;
        avatar_data.upgrade_size = 0;
        
        logger_log_debugf(0, "AVATAR Start Upgrade --");

        app_context_set_mode(APP_CONTEXT_MODE_PROCESS);
        app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_IN_PROGRESS);
    }
    else
    {
        // Stay where we are..
    }
}

/**-----------------------------------------------------------------------------
 * Performs a factory reset - run as a stateful_process
 */
void app_commands_avatar_usb_upgrade()
{
    // State 1 - start
    logger_log_info(LOGGER_CODE_FACTORY_RESET, "AVATAR USB Upgrade Start --");
    
    avatar_send_device_upgrade_status(DEVICE_UPGRADE_STATUS_START_USB_UPGRADE);
    
    app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
    app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE_IN_PROGRESS);
}

/**-----------------------------------------------------------------------------
 * Draws a screen displayed during Dirac calibration
 */
void app_commands_draw_dirac_calibration(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();

    // Draw the top line
    // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);
    display_start_text(DISPLAY_LINE_1, menu_get_title(item->menu_id));
    display_write_line(DISPLAY_LINE_1, false);

    display_clear_line(DISPLAY_LINE_2);
    
    display_start_text(DISPLAY_LINE_2, "Calibrating");
    
    display_write_line(DISPLAY_LINE_2, false);
}

/**-----------------------------------------------------------------------------
 * Draws a standard navigation menu with title on top line and navigation buttons 
 * on the bottom
 */
void app_commands_draw_dirac_no_navigation(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();

    // Draw the top line
    // Clear the display
    display_clear(false);

    size_t len = display_get_line_length();
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);
    
    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    
    // Check if there is a value to be displayed
    char value[len];
    
    INPUT_CONTROL_DEFINITION* definition = input_control_get_current_definition();
    if (!input_analogue_direct_get(definition->index)) //definition->analogue_direct)
    {
        app_commands_get_dirac_selected_string((char*)&value[0], len);

        // Check whether we're displaying the edited value or not
        // MENU_ITEM* current_item = menu_get_current_item(); 
        if (menu_is_editable_and_original(item))
        {
            display_append_text_right(DISPLAY_LINE_1, "*%s", value);
        }
        else
        {
            display_append_text_right(DISPLAY_LINE_1, "%s", value);
        }
     
        display_write_line(DISPLAY_LINE_1, false);

        display_clear_line(DISPLAY_LINE_2);
        display_start_text(DISPLAY_LINE_2, AUDIO_INPUT_STRINGS[input_control_app_get_edit_index()]);
        display_write_line(DISPLAY_LINE_2, false);
    }
    else
    {
        display_write_line(DISPLAY_LINE_1, false);

        display_clear_line(DISPLAY_LINE_2);
        display_append_text_right(DISPLAY_LINE_2, "Not Available");
        display_write_line(DISPLAY_LINE_2, false);
    }
}
 
/**-----------------------------------------------------------------------------
 * Used to edit the whether dirac is enabled or not
 * @param message
 */
bool app_commands_edit_dirac_selected(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    logger_log_debugf(0, "INPUT app_commands_edit_dirac_selected %d ", input_control_app_get_edit_index());
        
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        uint8_t input = input_control_app_get_edit_index();
        int8_t value = (int8_t)message->data[0];
        int8_t current_value = (int8_t)dirac_get_curve_for_input(input); 
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }
    
        message->command = APP_COMMANDS_ROOM_EQUALISATION;
        message->response = STATUS_UPDATE;
        
        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                current_value = value;
                message->data[0] = dirac_set_curve_for_input(input, (uint8_t)current_value);
                if (message->data[0] != (uint8_t)current_value)
                {
                    // Assume that we haven't been able to set the curve
                    message->response = PARAM_NOT_RECOGNISED;
                }
                break;
            }
            
            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            {
                DIRAC_EQ_CURVE* curve = dirac_get_next_curve(current_value);
                if (curve != NULL)
                {
                    current_value = curve->index;
                }
                message->data[0] = dirac_set_curve_for_input(input, (uint8_t)current_value);
                if (message->data[0] != (uint8_t)current_value)
                {
                    // Assume that we haven't been able to set the curve
                    message->response = PARAM_NOT_RECOGNISED;
                }
                break;
            }
            
            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                DIRAC_EQ_CURVE* curve = dirac_get_previous_curve(current_value);
                if (curve != NULL)
                {
                    current_value = curve->index;
                }
                message->data[0] = dirac_set_curve_for_input(input, (uint8_t)current_value);
                if (message->data[0] != (uint8_t)current_value)
                {
                    // Assume that we haven't been able to set the curve
                    message->response = PARAM_NOT_RECOGNISED;
                }
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
            {
                message->data[0] = (uint8_t)current_value;
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                break;
            }
        }

        // Just incase its not been defined
        message->data_length = 1;
    
        app_commands_send_message(message);
                
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_update_change_state(&current_value, sizeof(int16_t));
        }
    }

    return result;
}
 
/**-----------------------------------------------------------------------------
 * Called when the Region menus is exited.  Used to set the Avatar Region
 * @param message
 * @return 
 */
MENU_ID app_commands_edit_exit_region(MENU_ID menu)
{
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;
    
    // Check if the region has changed (convoluted mechanism but it should work)
    if (!menu_is_editable_and_original(menu_get_item(menu)))
    {
        logger_log_debugf(0, "REGION has changed");
        avatar_set_region(avatar_message_get_region());
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to change the mqa decoder level
 */
bool app_commands_edit_mqa_decoder_level(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = message->data[0];
        int8_t current_value = (int8_t)input_control_get_filter();
        
        uint8_t mqa_value;
        settings_get(SETTING_MQA_DECODER_LEVEL, &mqa_value);
        current_value = mqa_value;
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        message->command = APP_COMMANDS_AVATAR_MQA_DECODER_LEVEL;
        message->response = STATUS_UPDATE;
        
        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                current_value = value;
                
                if (value < AVATAR_MQA_DECODER_LEVEL_MAX)
                {
                    settings_set(SETTING_MQA_DECODER_LEVEL, (uint8_t*)&current_value);
                    avatar_send_mqa_decoder_level((AVATAR_MQA_DECODER_LEVEL) current_value);
                }
                else
                {
                    message->response = PARAM_NOT_RECOGNISED;
                }
                
                message->data[0] = (uint8_t)current_value;
                break;
            }
            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            {
                current_value++;
                if (current_value >= AVATAR_MQA_DECODER_LEVEL_MAX)
                {
                    current_value--; // Stop loopback
//                    current_value = (FILTER_TYPE)0;
                }

                settings_set(SETTING_MQA_DECODER_LEVEL, (uint8_t*)&current_value);
                avatar_send_mqa_decoder_level((AVATAR_MQA_DECODER_LEVEL) current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                current_value--;
                if (current_value < 0)
                {
                    current_value = 0; // Stop loopback
                    //current_value = (FILTER_TYPE)((uint8_t)input_control_get_num_filters() - 1);
                }

                settings_set(SETTING_MQA_DECODER_LEVEL, (uint8_t*)&current_value);
                avatar_send_mqa_decoder_level((AVATAR_MQA_DECODER_LEVEL) current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
            {
                message->data[0] = (uint8_t)current_value;
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                break;
            }
        }

        message->data_length = 1;
        app_commands_send_message(message);
                
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_update_change_state(&current_value, sizeof(int8_t));
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Used to edit the auto_shutdown mode based on the current inputs
 * @param message
 */
bool app_commands_edit_upgrade_option(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int16_t value = (int16_t)message->data[0];
        int16_t current_value = (int16_t)avatar_get_upgrade_option();
         
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int16_t));
        }
    
        message->command = APP_COMMANDS_AVATAR_UPGRADE_OPTION;
        message->response = STATUS_UPDATE;
        
        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                current_value = value;
                avatar_set_upgrade_option(current_value);

                message->data[0] = (uint8_t)current_value;

                break;
            }
            
            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            {
                current_value++;
                
                // Note: when using input_control_get_num_inputs as an index we 
                // take into account that the None options is there but is not 
                // added when initialising the module hence we compare to the 
                // actual number rather than the number -1
                if (current_value >= AVATAR_UPGRADE_OPTION_MAX)
                {
                    current_value = 0;
                }

                avatar_set_upgrade_option(current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }
            
            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                current_value--;
                
                // Note: when using input_control_get_num_inputs as an index we 
                // take into account that the None options is there but is not 
                // added when initialising the module hence we compare to the 
                // actual number rather than the number -1
                if (current_value < 0)
                {
                    current_value = AVATAR_UPGRADE_OPTION_MAX - 1;
                }

                avatar_set_upgrade_option(current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
            {
                message->data[0] = (uint8_t)current_value;
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                break;
            }
        }

        // Just incase its not been defined
        message->data_length = 1;
    
        app_commands_send_message(message);
                
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_update_change_state(&current_value, sizeof(int16_t));
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Called when entering the avatar upgrade menu
 */
void app_commands_enter_avatar_upgrade_menu(MENU_ID menu)
{
    // Ensure the menu is only visible for a while
    logger_log_debugf(0, "AVATAR Setting Upgrade menu timeout to %d --", TIMER_MENU_AVATAR_UPGRADE_INACTIVE_TIMEOUT_MS);
    timers_update_timeout(TIMER_INDEX_MENU_INACTIVE, TIMER_MENU_AVATAR_UPGRADE_INACTIVE_TIMEOUT_MS, TIMERS_RESET_IMMEDIATE);
}

/**-----------------------------------------------------------------------------
 * Called when entering the avatar upgrade menu
 */
void app_commands_enter_avatar_upgrade_continue_menu(MENU_ID menu)
{
    if (!avatar_upgrade_available())
    {
        // Stay on the upgrade menu
        app_context_set_menu(APP_COMMAND_MENU_AVATAR_UPGRADE);
    }
}

/**-----------------------------------------------------------------------------
 * Called when leaving the Balance menu
 */
void app_commands_enter_dirac_calibration(MENU_ID menu)
{
    // Ensure the Dirac menu is visible for as long as the application is alive
    timers_update_timeout(TIMER_INDEX_MENU_INACTIVE, 0, TIMERS_RESET_IMMEDIATE);
}

/**-----------------------------------------------------------------------------
 * Called when leaving the Balance menu
 */
void app_commands_enter_dirac_no_navigation(MENU_ID menu)
{
    app_commands_set_input_for_edit(*menu_get_item(menu));
    
    // Ensure the Dirac menu is only visible for a while
    timers_update_timeout(TIMER_INDEX_MENU_INACTIVE, TIMER_MENU_DIRAC_INACTIVE_TIMEOUT_MS, TIMERS_RESET_IMMEDIATE);
}

/**-----------------------------------------------------------------------------
 * Called when entering the INFO menu
 */
void app_commands_enter_info(MENU_ID menu)
{
    avatar_data.current_playing_info_refresh = true;
}

/**-----------------------------------------------------------------------------
 * Used to simulate player status stopped when entering standby
 */
void app_commands_force_player_status_stopped()
{
    logger_log_debugf(0, "AVATAR PLAYER STATUS STOPPED D --");
    logger_process();
    avatar_data.player_status = AVATAR_PLAYER_STATUS_STOPPED;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the currently playing album
 * @param buffer
 * @param size
 */
void app_commands_get_album_name_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", avatar_data.album_name);
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the currently playing artist
 * @param buffer
 * @param size
 */
void app_commands_get_artist_name_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", avatar_data.artist_name);
}

/**-----------------------------------------------------------------------------
* Returns a string representation of the current audio source (no adjustment for chromecast app name
* @return 
*/
void app_commands_get_audio_source_string(char* buffer, size_t len)
{
    AVATAR_AUDIO_SOURCE audio_source = avatar_message_get_audio_source();
    memcpy(buffer, AVATAR_AUDIO_SOURCE_STRINGS[audio_source], min(strlen(AVATAR_AUDIO_SOURCE_STRINGS[audio_source]), len));
}
  
/**-----------------------------------------------------------------------------
 * Returns a the avatars firmware version
 * @return 
 */
uint16_t app_commands_get_avatar_firmware_version()
{
    char out_buffer[50];

    avatar_message_get_firmware_version(out_buffer, 50);

    // The Avatar Version is in the format 0.nnn.vvv.0xnnnnnn where vvv is the actual release number
    // Lets just extract the 0.nnn.vvv for the minute
    uint8_t start_idx = 0;
    uint8_t end_idx = 0;
    uint8_t dot_count = 0;
    uint8_t x;
    for (x = 0; x < 50; x++)
    {
        if (out_buffer[x] == '.')
        {
            dot_count++;
        }

        if ((dot_count == 2) && (start_idx == 0))
        {
            start_idx = x + 1;
        }
        else if (dot_count == 3)
        {
            end_idx = x - 1;
            break;
        }
    }
    
    char version_buffer[5];
    memset(version_buffer, 0, 5);
    memcpy(version_buffer, &out_buffer[start_idx], min(end_idx - start_idx + 1, 5));

    return (uint16_t)atoi(version_buffer);
}
  
/**-----------------------------------------------------------------------------
 * Returns a string representation of the avatars firmware version
 * @return 
 */
void app_commands_get_avatar_firmware_version_string(char* buffer, size_t len)
{
    sprintf(buffer, "v%d", app_commands_get_avatar_firmware_version());
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of whether dirac is enabled or not
 * @param buffer
 * @param size
 */
void app_commands_get_dirac_selected_string(char* buffer, size_t size)
{
    uint8_t curve_idx = dirac_get_curve_for_input(input_control_app_get_edit_index());
    DIRAC_EQ_CURVE* curve = dirac_get_curve(curve_idx);
    
    //logger_log_debugf(0, "DIRAC %d %d %d --", input_control_app_get_edit_index(), curve_idx, curve->index);
    if (curve != NULL)
    {
        char* name = dirac_get_curve_name(curve->index);
        
        memset(buffer, 0, size);
        memcpy(buffer, name, min(strlen(name), 11));
    }
    else
    {
        sprintf(buffer, "%s", "Not Calc");
    }
}

/**-----------------------------------------------------------------------------
 * Gets the last known player status.
 */
AVATAR_PLAYER_STATUS app_commands_get_player_status()
{
    return avatar_data.player_status;
}

/**-----------------------------------------------------------------------------
 * Returns the current track length
  */
uint64_t app_commands_get_track_length()
{
    return avatar_data.track_length;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the currently playing track
 * @param buffer
 * @param size
 */
void app_commands_get_track_name_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", avatar_data.track_name);   
}

/**-----------------------------------------------------------------------------
 * Returns the current track time (position)
 */
uint64_t app_commands_get_track_time()
{
    return avatar_data.track_time;
}

/**-----------------------------------------------------------------------------
 * Can be used to reset the inactivity timeout in cases where its been adjusted 
 * by displaying the Avatar Upgrade Complete menu
 * @param menu
 */
MENU_ID app_commands_menu_reset_upgrade_complete(MENU_ID menu)
{
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;
    
    logger_log_debugf(0, "APP app_commands_menu_reset_upgrade_complete --");
    // Make sure Inactive Menu timeout is reset
    // Note while this is set to 3 seconds it actually takes longer as nothing
    // is processing the TIMER_INDEX_MENU_INACTIVE while the avatar is loading - 
    // it  has to wait until it enters the running state...
    timers_update_timeout(TIMER_INDEX_MENU_INACTIVE, TIMER_MENU_INACTIVE_TIMEOUT_MS, TIMERS_RESET_IMMEDIATE);
 
    // Are we still in the splash screen
    // If so..Make sure the Display Testing screen is displayed
    if (!app_context_is_running() && (menu != APP_COMMAND_MENU_AVATAR_UPGRADE_IN_PROGRESS))
    {
        // Get the splash screen process going again
        // We put it into MENU mode because otherwise we cant cancel the Update 
        // Complete screen
        app_context_set_mode(APP_CONTEXT_MODE_MENU); 
        result = APP_COMMAND_MENU_DISPLAY_TESTING;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to report on IPAddress, MAC Address, Friendly Name
 * @param message
 */
bool app_commands_network(STRUCTURED_MESSAGE* message)
{
    bool result = false;

    message->command = APP_COMMANDS_NETWORK;
    message->response = STATUS_UPDATE;

    switch ((APP_COMMAND_NETWORK)message->data[0])
    {
        case APP_COMMAND_NETWORK_IP_ADDRESS:
        {
            IP_ADDR ip_address = tcpip_get_runtime_ip_address();
            message->data_length = 4;
            memcpy(message->data, &ip_address.v[0], message->data_length);
            break;
        }

        case APP_COMMAND_NETWORK_MAC_ADDRESS:
        {
            TCPIP_MAC_ADDR mac_address = tcpip_get_mac_address();
            message->data_length = 6;
            memcpy(message->data, &mac_address.v[0], message->data_length);
            break;
        }

        case APP_COMMAND_NETWORK_WIFI_MAC_ADDRESS:
        {
            TCPIP_MAC_ADDR mac_address = tcpip_get_wifi_mac_address();
            message->data_length = 6;
            memcpy(message->data, &mac_address.v[0], message->data_length);
            
            logger_log_debugf(0, "MAC ADDRESS A %d %d %d %d %d %d --", mac_address.v[0], mac_address.v[1], mac_address.v[2], mac_address.v[3], mac_address.v[4], mac_address.v[5]);
            logger_log_debugf(0, "MAC ADDRESS B %d %d %d %d %d %d --", message->data[0], message->data[1], message->data[2], message->data[3], message->data[4], message->data[5]);
            break;
        }

        case APP_COMMAND_NETWORK_HOST_NAME:
        {
            tcpip_get_network_name_string(message->data, TCPIP_NETWORK_NAME_MAX_LEN);
            message->data_length = strlen(message->data);
            break;
        }

        case APP_COMMAND_NETWORK_FRIENDLY_NAME:
        {
            avatar_message_get_friendly_name_string(message->data, TCPIP_NETWORK_NAME_MAX_LEN);
            message->data_length = strlen(message->data);
            break;
        }

        case APP_COMMAND_NETWORK_SSID:
        {
            tcpip_get_ssid_string(message->data, TCPIP_SSID_MAX_LEN);
            message->data_length = strlen(message->data);
            break;
        }
        
        default:
        {
            message->response = COMMAND_NOT_RECOGNISED;
            message->data_length = 1;

            break;
        }
    }

    app_commands_send_message(message);
                
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to report on Track/Artist/Album info
 * @param message
 */
bool app_commands_now_playing(STRUCTURED_MESSAGE* message)
{
    bool result = false;

    message->command = APP_COMMANDS_NOW_PLAYING;
    message->response = STATUS_UPDATE;

    switch ((APP_COMMAND_NOW_PLAYING)message->data[0])
    {
        case APP_COMMAND_NOW_PLAYING_TRACK:
        {
            message->data_length = strlen(avatar_data.track_name);
            memcpy(message->data, avatar_data.track_name, message->data_length);
            break;
        }

        case APP_COMMAND_NOW_PLAYING_ARTIST:
        {
            message->data_length = strlen(avatar_data.artist_name);
            memcpy(message->data, avatar_data.artist_name, message->data_length);
            break;
        }

        case APP_COMMAND_NOW_PLAYING_ALBUM:
        {
            message->data_length = strlen(avatar_data.album_name);
            memcpy(message->data, avatar_data.album_name, message->data_length);
            break;
        }

        case APP_COMMAND_NOW_PLAYING_APPLICATION:
        {
            avatar_message_get_audio_source_string(message->data, 40);  // Arbitrary max length for maximum app name - may change
            message->data_length = strlen(message->data);
            break;
        }

        case APP_COMMAND_NOW_PLAYING_SAMPLE_RATE:
        {
            //0x00 ? 32 kHz
            //0x01 ? 44.1 kHz
            //0x02 ? 48 kHz
            //0x03 ? 88.2 kHz
            //0x04 ? 96 kHz
            //0x05 ? 176.4 kHz
            //0x06 ? 192 kHz
            //0x07 ? Unknown
            //0x08 ? Undetected
            AVATAR_SAMPLE_RATE sample_rate = avatar_message_get_sample_rate();

            message->data[0] =  AVATAR_SAMPLE_RATE_RS232[sample_rate];
            message->data_length = 1;
            break;
        }

        case APP_COMMAND_NOW_PLAYING_ENCODER:
        {
            if (avatar_message_get_mqa_state() == AVATAR_MQA_STATE_NO_MQA)
            {
                message->data[0] =  AVATAR_MIME_TYPE_VALUES[avatar_message_get_mime_type_index()];
                message->data_length = 1;
            }
            else
            {
                message->data[0] =  0x0F;   // MQA
                message->data_length = 1;
            }
            break;
        }

        default:
        {
            message->response = COMMAND_NOT_RECOGNISED;
            message->data_length = 1;

            break;
        }
    }

    app_commands_send_message(message);
                
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to report on the current player status
 * @param message
 */
bool app_commands_player_status(STRUCTURED_MESSAGE* message)
{
    bool result = false;

    message->command = APP_COMMANDS_PLAYBACK_STATUS;
    message->response = STATUS_UPDATE;

    switch(message->data[0])
    {
        case APP_COMMAND_VALUE_REQUEST_STATUS:
        {
            // 20190514 - Changed after discussion with Pete Tim - Pete to change doc to use Hostlink values
            message->data[0] = (uint8_t)app_commands_get_player_status();
 
            break;
        }

        default:
        {
            message->response = COMMAND_NOT_RECOGNISED;
            break;
        }
    }

    // Just incase its not been defined
    message->data_length = 1;

    app_commands_send_message(message);
                
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to process avatar messages triggered by the remote, or as a simulated RC5 command
 * The message->command is the RC5 code
 * @param message
 */
void app_commands_remote_avatar_message(STRUCTURED_MESSAGE* message)
{
    switch (message->data[0])
    {
        case REMOTE_CMD_INPUT_SKIP_BACK:
        {
            avatar_send_player_control(AVATAR_PLAYER_CONTROL_PREVIOUS);
            break;
        }
        
        case REMOTE_CMD_INPUT_SKIP_FORWARD:
        {
            avatar_send_player_control(AVATAR_PLAYER_CONTROL_NEXT);
            break;
        }

        case REMOTE_CMD_INPUT_STOP:
        {
            // Note Pause toggles Pause/Play
            if (app_commands_get_player_status() == AVATAR_PLAYER_STATUS_PLAYING)
            {
                avatar_send_player_control(AVATAR_PLAYER_CONTROL_PAUSE);
            }
            break;
        }
        
        case REMOTE_CMD_INPUT_PLAY:
        {
            // Note Pause toggles Pause/Play
            if ((app_commands_get_player_status() == AVATAR_PLAYER_STATUS_PAUSED) ||
                (app_commands_get_player_status() == AVATAR_PLAYER_STATUS_STOPPED))     
            {
                avatar_send_player_control(AVATAR_PLAYER_CONTROL_PAUSE);
            }
            break;
        }
        
        case REMOTE_CMD_INPUT_PAUSE:
        {
            // Note Pause toggles Pause/Play
            if (app_commands_get_player_status() == AVATAR_PLAYER_STATUS_PLAYING)
            {
                avatar_send_player_control(AVATAR_PLAYER_CONTROL_PAUSE);
            }
            break;
        }
        
        case REMOTE_CMD_INPUT_RANDOM:
        //case REMOTE_CMD_INPUT_SHUFFLE:
        {
            if ((app_commands_get_player_status() == AVATAR_PLAYER_STATUS_PLAYING) ||
                (app_commands_get_player_status() == AVATAR_PLAYER_STATUS_PAUSED))
            {
                app_commands_browsing_handle_shuffle();
            }
            break;
        }
       
        case REMOTE_CMD_INPUT_REPEAT:
        case REMOTE_CMD_INPUT_RANDOM_REPEAT:
        case REMOTE_CMD_INPUT_RANDOM_REPEAT_MODE:
        {
            if ((app_commands_get_player_status() == AVATAR_PLAYER_STATUS_PLAYING) ||
                (app_commands_get_player_status() == AVATAR_PLAYER_STATUS_PAUSED))
            {
                app_commands_browsing_handle_repeat();
            }
            
            break;
        }

        case REMOTE_CMD_INPUT_REWIND:
        {
            logger_log_debugf(0, "REMOTE_CMD_INPUT_REWIND %d %d %d --", avatar_data.track_length, avatar_data.track_time, min(avatar_data.track_length, avatar_data.track_time - AVATAR_FAST_FWD_TIME_MS));
            if (avatar_data.track_length > 0)
            {
                if (avatar_data.track_time > AVATAR_FAST_FWD_TIME_MS)
                {
                    avatar_send_track_seek(avatar_data.track_time - AVATAR_FAST_FWD_TIME_MS);
                }
                else
                {
                    avatar_send_track_seek(0);
                }
            }
            break;
        }

        case REMOTE_CMD_INPUT_FAST_FORWARD:
        {
            // TODO Avatar Messages
            logger_log_debugf(0, "REMOTE_CMD_INPUT_FAST_FORWARD %d %d %d --", avatar_data.track_length, avatar_data.track_time, min(avatar_data.track_length, avatar_data.track_time + AVATAR_FAST_FWD_TIME_MS));
            if (avatar_data.track_length > 0)
            {
                avatar_send_track_seek(min(avatar_data.track_length, avatar_data.track_time + AVATAR_FAST_FWD_TIME_MS));
            }
            break;
        }
        
        case REMOTE_CMD_INPUT_ZERO:
        case REMOTE_CMD_INPUT_ONE:
        case REMOTE_CMD_INPUT_TWO:
        case REMOTE_CMD_INPUT_THREE:
        case REMOTE_CMD_INPUT_FOUR:
        case REMOTE_CMD_INPUT_FIVE:
        case REMOTE_CMD_INPUT_SIX:
        case REMOTE_CMD_INPUT_SEVEN:
        case REMOTE_CMD_INPUT_EIGHT:
        case REMOTE_CMD_INPUT_NINE:
        {
            // TODO Avatar Messages
            break;
        }
        
        default:
        {
            logger_log_warningf(LOGGER_CODE_AVATAR, "AVATAR Attempt to process un-handled Avatar RC5 command %d ", message->data[0]);
            break;
        }
        
    }
}

/**-----------------------------------------------------------------------------
 * Used to report on the names of the Dirac EQ settings
 * @param message
 */
bool app_commands_room_eq_names(STRUCTURED_MESSAGE* message)
{
    bool result = false;

    message->command = APP_COMMANDS_ROOM_EQ_NAMES;
    message->response = STATUS_UPDATE;

    switch(message->data[0])
    {
        case APP_COMMAND_VALUE_REQUEST_STATUS:
        {
            memset(message->data, 0, (DIRAC_MAX_EQ_CURVES - 1) * 20);
            
            // Loop upto the number of curves as the No Curve is not included in the number
            int x;
            for (x = 1; x <= dirac_get_num_curves(); x++)
            {
                char* name = dirac_get_curve_name(x);
                if (name != NULL)
                {
                    memcpy(&message->data[(x - 1) * 20], name, min(strlen(name), 20));
                }
            }
            
            message->data_length = (DIRAC_MAX_EQ_CURVES - 1) * 20;
            
            break;
        }

        default:
        {
            message->response = COMMAND_NOT_RECOGNISED;
            message->data_length = 1;
            
            break;
        }
    }

    app_commands_send_message(message);
                
    return result;
}

/* *****************************************************************************
 End of File
 */
