/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_browsing.h"
#include "utilities/avatar_browsing.h"
#include "utilities/display/display_defs.h"
#include "utilities/display_message.h"
#include "utilities/input_control_defs.h"
#include "utilities/menu_app.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct 
{
    AVATAR_REPEAT_STATE repeat_state;
    AVATAR_SHUFFLE_STATE shuffle_state;
    
} AVATAR_DATA;

static AVATAR_DATA avatar_data =
{
    .repeat_state = AVATAR_REPEAT_STATE_NORMAL,
    .shuffle_state = AVATAR_SHUFFLE_STATE_OFF,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */


/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the current value of the repeat state
 * @return 
 */
AVATAR_REPEAT_STATE app_commands_browsing_get_repeat_state()
{
    return avatar_data.repeat_state;
}

/**-----------------------------------------------------------------------------
 * Gets the current value of the shuffle state
 * @return 
 */
AVATAR_SHUFFLE_STATE app_commands_browsing_get_shuffle_state()
{
    return avatar_data.shuffle_state;
}

/**-----------------------------------------------------------------------------
 * Handles a repeat commands from a remote
 */
void app_commands_browsing_handle_repeat()
{
    avatar_data.repeat_state++;
    avatar_data.repeat_state %= AVATAR_REPEAT_STATE_MAX;

    switch (avatar_data.repeat_state)
    {
        case AVATAR_REPEAT_STATE_ONE:
        {
            avatar_send_shuffle_state(AVATAR_SHUFFLE_STATE_OFF);
            avatar_send_repeat_state(AVATAR_REPEAT_STATE_ONE);
            break;
        }

        case AVATAR_REPEAT_STATE_ALL:
        {
            avatar_send_shuffle_state(AVATAR_SHUFFLE_STATE_OFF);
            avatar_send_repeat_state(AVATAR_REPEAT_STATE_ALL);
            break;
        }

        default:
        {
            avatar_send_shuffle_state(AVATAR_SHUFFLE_STATE_OFF);
            avatar_send_repeat_state(AVATAR_REPEAT_STATE_NORMAL);
            break;
        }
    }

    avatar_data.shuffle_state = AVATAR_SHUFFLE_STATE_OFF;
}

/**-----------------------------------------------------------------------------
 * Handles a shuffle or random commands from a remote
 */
void app_commands_browsing_handle_shuffle()
{
    avatar_data.shuffle_state = !avatar_data.shuffle_state;

    if (avatar_data.shuffle_state)
    {
        avatar_send_shuffle_state(AVATAR_SHUFFLE_STATE_ON);
        avatar_send_repeat_state(AVATAR_REPEAT_STATE_NORMAL);
    }
    else
    {
        avatar_send_shuffle_state(AVATAR_SHUFFLE_STATE_OFF);
        avatar_send_repeat_state(AVATAR_REPEAT_STATE_NORMAL);
    }

    avatar_data.repeat_state = AVATAR_REPEAT_STATE_NORMAL;    
}
    
/**-----------------------------------------------------------------------------
 * Resets browsing shuffl/repeat state
 */
void app_commands_browsing_reset()
{
    avatar_data.repeat_state = AVATAR_REPEAT_STATE_NORMAL;
    avatar_data.shuffle_state = AVATAR_SHUFFLE_STATE_OFF;
    
    avatar_browsing_set_running_state(AVATAR_BROWSING_STATE_INACTIVE);
}

/**-----------------------------------------------------------------------------
 * Draws a screen displayed while browsing
 */
void app_commands_draw_browsing(APP_COMMAND_MENU menu)
{
    // Clear the display
    display_clear(false);
    
    switch (avatar_browsing_get_state())
    {
        case AVATAR_BROWSING_STATE_STARTING:
        {
            ////-----------------------------------------------------------------------------
            // Check for a preview
            INPUT_CONTROL_PREVIEW_INFO* preview_info = input_control_get_preview_info();
            if (preview_info->active)
            {
                app_commands_draw_volume(menu);
            }
            else
            {
                AUDIO_INPUT selected_input = input_control_get_selected_input();
                uint8_t volume = volume_get();
                uint8_t input_start_pos = 0;

                if (selected_input == INPUT_CONTROL_NUM_SETTINGS)
                {
                    selected_input = AUDIO_INPUT_NONE;
                }

                char input_name[14];
                memset(input_name, 0, 14);
                memcpy(input_name, AUDIO_INPUT_STRINGS[selected_input], strlen(AUDIO_INPUT_STRINGS[selected_input]));
                display_append_text(DISPLAY_LINE_1, input_start_pos, "%s", input_name);

                // display volume or Mute
                if (input_control_in_mute_state())
                {
                    if (!input_control_in_switch())
                    {
                        display_append_text_right(DISPLAY_LINE_1, "%s", APP_STRINGS[APP_STRINGS_ENUM_MUTE]);
                    }
                    else
                    {
                        if (input_control_mute_state_pre_switch())
                        {
                            display_append_text_right(DISPLAY_LINE_1, "%s", APP_STRINGS[APP_STRINGS_ENUM_MUTE]);
                        }
                        else
                        {
                            //display_append_text_right(DISPLAY_LINE_1, "%s", APP_STRINGS[APP_STRINGS_ENUM_SWITCHING]);
                            display_append_text_right(DISPLAY_LINE_1, "%02d", volume);
                        }
                    }
                }
                else if (selected_input != AUDIO_INPUT_NONE)
                {
                    if (!input_control_in_switch())
                    {
                        display_append_text_right(DISPLAY_LINE_1, "%02d", volume);
                    }
                    else
                    {
                        if (input_control_mute_state_pre_switch())
                        {
                            display_append_text_right(DISPLAY_LINE_1, "%s", APP_STRINGS[APP_STRINGS_ENUM_MUTE]);
                        }
                        else
                        {
                            display_append_text_right(DISPLAY_LINE_1, "%02d", input_control_volume_in_switch());
                        }
                    }
                }

                display_special_character(DISPLAY_LINE_2, 0, DISPLAY_SPECIAL_CHARACTERS_RIGHT);
                display_append_text(DISPLAY_LINE_2, 1, "OK TO BROWSE");            

                display_write_line(DISPLAY_LINE_1, false);
                display_write_line(DISPLAY_LINE_2, false);
            }
            
            break;
        }
        
        case AVATAR_BROWSING_STATE_BROWSING:
        {
            AVATAR_BROWSE_ITEM* parent = avatar_browsing_get_parent_item();
            display_append_text(DISPLAY_LINE_1, 0, parent->name);

            // Bottom Line
            display_clear_line(DISPLAY_LINE_2);
            AVATAR_BROWSE_ITEM* current = avatar_browsing_get_current_item();
            if (current != NULL)
            {
                char buffer[16];
                memcpy(buffer, current->name, 16);
                buffer[15] = 0;
                display_append_text(DISPLAY_LINE_2, 0, buffer);//current->name);
            }
            else
            {
                display_append_text(DISPLAY_LINE_2, 0, "{empty}");
            }

            //display_write_line(DISPLAY_LINE_2, false);

            display_write_line(DISPLAY_LINE_1, false);

            display_write_line(DISPLAY_LINE_2, false);
            
            break;
        }
        
        case AVATAR_BROWSING_STATE_PLAYING:
        {
            app_commands_draw_volume(menu);
            break;
        }
        
        default:
            break;
        
    }
}

/**-----------------------------------------------------------------------------
 * Called when entering the Analogue Direct no navigation menu - setup the timeout
 */
void app_commands_enter_browsing(MENU_ID menu)
{
    avatar_data.repeat_state = AVATAR_REPEAT_STATE_NORMAL;
    
    avatar_browsing_start_browsing(NULL);
}

/**-----------------------------------------------------------------------------
 * Called when leaving the Balance menu
 * @param menu
 */
MENU_ID app_commands_exit_browsing(MENU_ID menu)
{
    // Check to see if analogue direct needs applying to active input
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;
    
    avatar_browsing_stop();
    
    // Make sure Inactive Menu timeout is reset
    menu_app_reset_inactivity_timeout(menu);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Handles internal browsing messages
 * @param message
 */
void app_commands_handle_browsing(STRUCTURED_MESSAGE* message)
{
    BROWSE_COMMAND_OPTION option = (BROWSE_COMMAND_OPTION)message->data[0];
    
    logger_log_debugf(0, "BROWSING %s --", BROWSE_COMMAND_OPTION_STRINGS[option]);
    switch (option)
    {
        case BROWSE_COMMAND_OPTION_BROWSE_SELECT:
        {
            avatar_browsing_browse_current(false);
            break;
        }
        
        case BROWSE_COMMAND_OPTION_BROWSE_NEXT:
        {
            avatar_browsing_select_next();
            break;
        }
        
        case BROWSE_COMMAND_OPTION_BROWSE_PREVIOUS:
        {
            avatar_browsing_select_previous();
            break;
        }
        
        case BROWSE_COMMAND_OPTION_BROWSE_BACK:
        {
            avatar_browsing_send_browse_back(AVATAR_BROWSE_BACK_LEVEL_UP_ONE);
            break;
        }
        
        case BROWSE_COMMAND_OPTION_BROWSE_TOP:
        {
            avatar_browsing_send_browse_back(AVATAR_BROWSE_BACK_LEVEL_ROOT);
            break;
        }
        
        case BROWSE_COMMAND_OPTION_BROWSE_SKIP_UP:
        {
            avatar_browsing_display_skip(AVATAR_BROWSING_SKIP_CONTROL_BACK);
            break;
        }
        
        case BROWSE_COMMAND_OPTION_BROWSE_SKIP_DOWN:
        {
            avatar_browsing_display_skip(AVATAR_BROWSING_SKIP_CONTROL_FORWARD);
            break;
        }
        
        case BROWSE_COMMAND_OPTION_BROWSE_PLAY:
        {
            logger_log_debugf(0, "BROWSING avatar_browsing_has_played_in_session %d --", avatar_browsing_has_played_in_session());
            avatar_browsing_restart_playing();
            if ((app_commands_get_player_status() == AVATAR_PLAYER_STATUS_PAUSED) ||
                (app_commands_get_player_status() == AVATAR_PLAYER_STATUS_STOPPED))     
            {
                avatar_send_player_control(AVATAR_PLAYER_CONTROL_PAUSE);
            }
            break;
        }
        
        default:
            break;
    }
}

/* *****************************************************************************
 End of File
 */
