/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_control.h"
#include "app_context.h"
#include "app_commands.h"
#include "utilities/adc/adc.h"
#include "utilities/display.h"
#include "utilities/encoder_multi.h"
#include "utilities/input_context.h"
#include "utilities/logger/logger.h"
#include "utilities/remote.h"
#include "utilities/standby.h"
#include "utilities/structured_message.h"
#include "utilities/system_fault.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Process ADC Input
 */
static void app_commands_process_adc()
{
    if (timers_has_expired(TIMER_INDEX_ADC))
    {
        if (adc_process())
        {
            // Only reset the timer if we have successfully processed some samples
            // Other wise retry immediately
            timers_reset(TIMER_INDEX_ADC); 
            
            
#if APP_FEATURE_BALANCE_POT == APP_FEATURE_ON
            if (adc_pot_has_changed(APP_ADC_INPUTS_BALANCE))
            {
                input_context_set_balance_changed(true);
            }
#endif

            // Update the current input context
            ADC_BUTTON current_button;
            ADC_BUTTON_INFO* button_info = adc_button_get_current();
            if (button_info != NULL)
            {
//logger_log_debugf(0, "ADC BUTTON %d --", button_info->button);                
                current_button = button_info->button;
            }
            else
            {
                current_button = ADC_BUTTON_NONE;
            }
    
            //TODO Review
            if (current_button == ADC_BUTTON_NONE)
            {
                // No button detected - are we in standby mode?
                if (standby_get_state() == STANDBY_STATE_SLEEP)
                {
                    // Can we detect the standby button being pressed in digital mode?
                    // make sure we're in digital mode
                     SYS_PORTS_PinModeSelect(PORTS_ID_0, PORTS_ANALOG_PIN_2, PORTS_PIN_MODE_DIGITAL);

                    if (!SYS_PORTS_PinRead(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_2))
                    {
                        current_button = ADC_BUTTON_5; // power//TODO Sort out!!!!!??????????????????????????????????????????????????
                    }
                }
            }
            
            input_context_set_button(current_button);
            
            // Get the balance offset
            
            long temp = 0;
            
            // Check for Temperature based faults
            temp = adc_temperature_get(APP_ADC_INPUTS_LIFTER_TEMPERATURE);
temp = 0;  //TODO We have temporarily removed lifter temperature checking as for
           // for some reason it keep going over MAX, and random code changes seem tomfix it - not happy!!
            if ((temp != ADC_INVALID_INPUT) && (temp / 10 >= APP_MAX_WORKING_TEMPERATURE))
            {
                SYSTEM_FAULT fault;
                fault.code = SYSTEM_FAULT_CODE_LIFTER_OVER_TEMP;
                sprintf(fault.description, "Lifter Temp %2d.%d%cC --", temp / 10, temp % 10, display_get_special_character(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_DEGREES));
                system_fault_set(&fault);
            }
            else
            {
            }
            
            temp = adc_temperature_get(APP_ADC_INPUTS_OUTPUT_TEMPERATURE);
            if ((temp != ADC_INVALID_INPUT) && (temp / 10 >= APP_MAX_WORKING_TEMPERATURE))
            {
                SYSTEM_FAULT fault;
                fault.code = SYSTEM_FAULT_CODE_OUTPUT_OVER_TEMP;
                sprintf(fault.description, "Output Temp %2d.%d%cC --", temp / 10, temp % 10, display_get_special_character(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_DEGREES));

                system_fault_set(&fault);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Process input from the rotary encoder
 */
static void app_commands_process_encoder()
{
    // Update the current input context - Get the current encoder values
    ENCODER_STATE state[ENCODER_INDEX_MAX];
    if (encoder_process_state(state))
    {
        // pass the results to the relevant encoder context
        int x;
        for (x = 0; x < ENCODER_INDEX_MAX; x++)
        {
            input_context_set_encoder(&state[x], x);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Report on input changes when in test mode
 */
static void app_commands_process_input_tests()
{
    if ((logger_get_level() >= LOGGER_TYPE_TEST) &&
        (input_context_has_changed()))
    {
        ENCODER_STATE encoder = input_context_get_encoder(ENCODER_INDEX_MAIN);
        if (encoder.delta > 0)
        {
            logger_log_testf(LOGGER_CODE_TEST_REMOTE_INCREMENT, "");
        }
        else if (encoder.delta < 0)
        {
            logger_log_testf(LOGGER_CODE_TEST_REMOTE_DECREMENT, "");
        }
        
        if (input_context_has_button_changed())
        {
            ADC_BUTTON button = input_context_get_button();
            logger_log_testf(LOGGER_CODE_TEST_BUTTON + (int)button, "");
        }
     }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Processes input controls (Buttons, Encoder, IR)
 */
void app_commands_process_controls()
{
    // Gather the current settings from the buttons, encode, and remote
    // Handle any any changes if the adc timer has expired
    app_commands_process_adc();
 
    // Handle rotary input
    app_commands_process_encoder();

    // Process the remote - this will have generated a structured message that
    // will be added to the input context
    remote_process();
            
    // Construct a structured message then call structured_message_callback
    // Review the state of the buttons, encoder, and remote
    // command will be the type of action edit, menu_navigation
    // It will use the input_context to gather the data
    app_commands_process_input_controls(app_commands_structured_message_callback);
    
    // If in test mode report on logger level
    app_commands_process_input_tests();
}
    
/* *****************************************************************************
 End of File
 */
