/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_display.h"
#include <stdio.h>
#include "app_config.h"
#include "app_commands_rgb_led.h"
#include "utilities\display.h"
#include "utilities/display/display_defs.h"
#include "utilities\menu.h"
#include "utilities\power.h"
#include "utilities\rgb_led.h"
#include "utilities\system_fault.h"
#include "utilities/timers.h"
#include "utilities\volume.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    // The display brightness when in full display - adjusted through the
    // menus etc..
    APP_COMMAND_BRIGHTNESS full_brightness;
    
    //! Indicates that the brightness is being controlled by the remote
    bool remote_change;
    
} BRIGHTNESS_DATA;

static BRIGHTNESS_DATA brightness_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**----------------------------------------------------------------------------- 
 * Changes the display brightness dependent on both the setting and the currently 
 * selected value
 */
void app_commands_adjust_brightness(APP_COMMAND_BRIGHTNESS brightness)
{
    //logger_log_debugf(0, "BRIGHTNESS Adjust %d %d %d --", brightness, app_commands_get_dark_mode(), app_commands_get_brightness());
    // Take the current brightness from the app_context (read from settings) and 
    // decide what is meant by the request to change the brightness)
    DISPLAY_BRIGHTNESS adjusted_brightness = DISPLAY_BRIGHTNESS_FULL;
    
    switch (brightness)
    {
        case APP_COMMAND_BRIGHTNESS_OFF:
            adjusted_brightness = DISPLAY_BRIGHTNESS_OFF;
            break;
            
        case APP_COMMAND_BRIGHTNESS_DIMMED:
        {
            switch (app_commands_get_brightness())
            {
                case APP_COMMAND_BRIGHTNESS_OFF:
                    adjusted_brightness = DISPLAY_BRIGHTNESS_OFF;
                    break;
            
                case APP_COMMAND_BRIGHTNESS_DIMMED:
                    adjusted_brightness = DISPLAY_BRIGHTNESS_DIMMED;
                    if (timers_has_expired(TIMER_INDEX_BRIGHTNESS) && (app_commands_get_dark_mode() == APP_COMMAND_DARK_MODE_ON))
                    {
                        adjusted_brightness = DISPLAY_BRIGHTNESS_OFF;
                    }
                    break;
                
                default:
                    adjusted_brightness = DISPLAY_BRIGHTNESS_FULL;
                    if (timers_has_expired(TIMER_INDEX_BRIGHTNESS) && (app_commands_get_dark_mode() == APP_COMMAND_DARK_MODE_ON))
                    {
                        adjusted_brightness = DISPLAY_BRIGHTNESS_OFF;
                    }
                    break;
            }
        
            break;
        }
            
        case APP_COMMAND_BRIGHTNESS_FULL:
        {
            switch (app_commands_get_brightness())
            {
                case APP_COMMAND_BRIGHTNESS_OFF:
                    if (!app_commands_brightness_get_remote_change())
                    {
                        adjusted_brightness = DISPLAY_BRIGHTNESS_FULL;
                    }
                    else
                    {
                        adjusted_brightness = DISPLAY_BRIGHTNESS_OFF;
                    }
                    break;
            
                case APP_COMMAND_BRIGHTNESS_DIMMED:
                    adjusted_brightness = DISPLAY_BRIGHTNESS_DIMMED;
                    if (timers_has_expired(TIMER_INDEX_BRIGHTNESS) && (app_commands_get_dark_mode() == APP_COMMAND_DARK_MODE_ON))
                    {
                        adjusted_brightness = DISPLAY_BRIGHTNESS_OFF;
                    }
                    break;
                
                default:
                    adjusted_brightness = DISPLAY_BRIGHTNESS_FULL;
                    if (timers_has_expired(TIMER_INDEX_BRIGHTNESS) && (app_commands_get_dark_mode() == APP_COMMAND_DARK_MODE_ON))
                    {
                        adjusted_brightness = DISPLAY_BRIGHTNESS_OFF;
                    }
                    break;
            }
        }
    }
    
    display_set_brightness(adjusted_brightness);
}
 
/**-----------------------------------------------------------------------------
 * Manages current display and led state dependent on power state etc...
 */
void app_commands_check_display()
{
    POWER_STATE power = power_get_state();
    
    //if (timers_has_expired(TIMER_INDEX_BRIGHTNESS))
    //{
    //logger_log_debugf(0, "BRIGHTNESS AA %d %d %d --", power, timers_has_expired(TIMER_INDEX_BRIGHTNESS), app_commands_get_dark_mode());
    //}
    
    switch(power)
    {
        case POWER_STATE_ON:
        {
            // Adjust for dark mode
            // Note this comes first other wise you get a PWM ing of the LED..
            if ((timers_has_expired(TIMER_INDEX_BRIGHTNESS)) && (app_commands_get_dark_mode() == APP_COMMAND_DARK_MODE_ON))
            {
                rgb_led_set(RGB_LED_COLOUR_OFF);
            }
            else
            {
                if (volume_is_muted())
                {
                    if (!input_control_in_switch())
                    {
                        rgb_led_set(RGB_LED_COLOUR_ORANGE);
                    }
                    else
                    {
                        if (input_control_mute_state_pre_switch())
                        {
                            rgb_led_set(RGB_LED_COLOUR_ORANGE);
                        }
                        else
                        {
                            rgb_led_set(app_commands_rgb_led_get_running_colour());
                        }
                    }
                }
                else
                {
                    rgb_led_set(app_commands_rgb_led_get_running_colour());
                }
            }
            break;
        }
            
        case POWER_STATE_OFF:
        {
            break;
        }
    }
    
    
    // Check brightness setting when in default mode
    if (app_context_get_mode() == APP_CONTEXT_MODE_DEFAULT)
    {
        if (!input_context_has_changed())
        {
            if (timers_has_expired(TIMER_INDEX_BRIGHTNESS))
            {
                if (power == POWER_STATE_ON)
                {
                    app_commands_adjust_brightness(APP_COMMAND_BRIGHTNESS_DIMMED);
                }
                else
                {
                    app_commands_adjust_brightness(APP_COMMAND_BRIGHTNESS_OFF);
                }
            }
            else
            {
                // BUG-20170928-6 Just reset the brightness - this is to handle the issue where 
                // Off is displayed for 5 seconds except when it is set via the remote
                //app_commands_adjust_brightness(app_commands_get_brightness());
            }
        }
        else
        {
            if (power == POWER_STATE_ON)
            {
                app_commands_adjust_brightness(APP_COMMAND_BRIGHTNESS_FULL);
                //app_commands_adjust_brightness(app_commands_get_brightness());
            }
            else
            {
                app_commands_adjust_brightness(APP_COMMAND_BRIGHTNESS_DIMMED);
            }
            
            // BUG-20170928-6 Don't reset the brightness - this is to handle the issue where 
            // Off is displayed for 5 seconds except when it is set via the remote
            // timers_reset(APP_TIMER_INDEX_BRIGHTNESS);
        } 
    }
    else
    {
        // Menu, Menu Edit and Processing mode display is FULL
        display_set_brightness(DISPLAY_BRIGHTNESS_FULL);
    }
}

/**-----------------------------------------------------------------------------
 * Draws a standard navigation menu with title on top line and navigation buttons 
 * on the bottom
 */
void app_commands_draw_display_no_navigation(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();

    // Draw the top line
    // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);
    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    
    // Check if there is a value to be displayed
    size_t len = display_get_line_length();
    char value[len];

    app_commands_get_brightness_string((char*)&value[0], len);

    // Check whether we're displaying the edited value or not
    // MENU_ITEM* current_item = menu_get_current_item(); 
    if (menu_is_editable_and_original(item))
    {
        display_append_text_right(DISPLAY_LINE_1, "*%s", value);
    }
    else
    {
        display_append_text_right(DISPLAY_LINE_1, "%s", value);
    }
     
    display_write_line(DISPLAY_LINE_1, false);

    display_clear_line(DISPLAY_LINE_2);
    display_write_line(DISPLAY_LINE_2, false);
}
  
/**-----------------------------------------------------------------------------
 * Used to edit the balance setting based on the current inputs
 * @param message
 */
bool app_commands_edit_display_brightness(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = (int8_t)message->data[0];
        APP_COMMAND_BRIGHTNESS current_value = app_commands_get_brightness();
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(APP_COMMAND_BRIGHTNESS));
        }
   
        message->command = APP_COMMANDS_BRIGHTNESS;
        message->response = STATUS_UPDATE;

        if (value == APP_COMMAND_BRIGHTNESS_TOGGLE)
        {
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                
            switch(app_commands_get_brightness())
            {
                case APP_COMMAND_BRIGHTNESS_OFF:
                    value = APP_COMMAND_BRIGHTNESS_FULL;
                    break;
                case APP_COMMAND_BRIGHTNESS_DIMMED:
                    value = APP_COMMAND_BRIGHTNESS_OFF;
                    break;
                case APP_COMMAND_BRIGHTNESS_FULL:
                    value = APP_COMMAND_BRIGHTNESS_DIMMED;
                    break;
            }
        }
        else if (value != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            app_commands_brightness_set_remote_change(false);
            if (message->source == STRUCTURED_MESSAGE_SOURCE_REMOTE)
            {
                app_commands_brightness_set_remote_change(true);
            }
        }

        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                // Check we have a valid command
                if (value <= APP_COMMAND_BRIGHTNESS_FULL)
                {
                    current_value = max(min(value, APP_COMMAND_BRIGHTNESS_FULL), APP_COMMAND_BRIGHTNESS_OFF);
                    app_commands_set_brightness(current_value);

                    message->data[0] = current_value;
                }
                else
                {
                    message->response = PARAM_NOT_RECOGNISED;
                }
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                switch(current_value)
                {
                    case APP_COMMAND_BRIGHTNESS_OFF:
                        current_value = value < 0 ? APP_COMMAND_BRIGHTNESS_FULL : APP_COMMAND_BRIGHTNESS_DIMMED;
                        break;
                    case APP_COMMAND_BRIGHTNESS_DIMMED:
                        current_value = value < 0 ? APP_COMMAND_BRIGHTNESS_OFF : APP_COMMAND_BRIGHTNESS_FULL;
                        break;
                    case APP_COMMAND_BRIGHTNESS_FULL:
                        current_value = value < 0 ? APP_COMMAND_BRIGHTNESS_DIMMED : APP_COMMAND_BRIGHTNESS_OFF;
                        break;
                    default:
                        break;
                }

                app_commands_set_brightness(current_value);

                message->data[0] = current_value;

                break;
            }
            
            default:
                message->data[0] = current_value;
                break;
        }
        
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            APP_COMMAND_BRIGHTNESS brightness = app_commands_get_brightness();
            menu_update_change_state(&brightness, sizeof(APP_COMMAND_BRIGHTNESS));
        }
        
        message->data_length = 1;
        app_commands_send_message(message);
    }
     
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to edit the dark mode
 * @param message
 */
bool app_commands_edit_dark_mode(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = (int8_t)message->data[0];
        int8_t current_value = (int8_t)app_commands_get_dark_mode();
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        if (app_commands_process_message_value_bool(message, &current_value, &value, true))
        {
            app_commands_set_dark_mode(current_value);
            
            if (current_value == APP_COMMAND_DARK_MODE_ON)
            {
                app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_OFF);
                brightness_data.full_brightness = APP_COMMAND_BRIGHTNESS_OFF;
            }
            else
            {
                app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_WHITE);
                uint8_t brightness;
                settings_get(SETTING_BRIGHTNESS, &brightness);
                app_commands_set_brightness((APP_COMMAND_BRIGHTNESS)brightness);
            }

            menu_update_change_state(&current_value, sizeof(int8_t));
        }             
    }    
     
    return result;
}

/**-----------------------------------------------------------------------------
 * Display a blank screen
 */
void app_commands_display_clear()
{
    display_clear(false);
    display_start_text(DISPLAY_LINE_1, "                ");
    display_write_line(DISPLAY_LINE_1, false);
    display_start_text(DISPLAY_LINE_2, "                ");
    display_write_line(DISPLAY_LINE_2, false);
    display_refresh();
}
    
/**-----------------------------------------------------------------------------
 * Display the current system fault on the screen
 */
void app_commands_display_system_fault()
{
    display_clear(false);
    display_start_text(DISPLAY_LINE_1, APP_STRINGS[APP_STRINGS_ENUM_SYSTEM_FAULT]);
    display_write_line(DISPLAY_LINE_1, false);
    SYSTEM_FAULT fault = system_fault_get();
    display_start_text(DISPLAY_LINE_2, fault.description);
    display_write_line(DISPLAY_LINE_2, false);
    display_refresh();
}
    
/**-----------------------------------------------------------------------------
 * 
 * @return 
 */
APP_COMMAND_BRIGHTNESS app_commands_get_brightness()
{
    return brightness_data.full_brightness;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the current balance value
 * @return 
 */
void app_commands_get_brightness_string(char* buffer, size_t size)
{
    APP_COMMAND_BRIGHTNESS brightness = max(min(app_commands_get_brightness(), 
                                                APP_COMMAND_BRIGHTNESS_FULL), 
                                            APP_COMMAND_BRIGHTNESS_OFF);
    
    sprintf(buffer, "%s", APP_COMMAND_BRIGHTNESS_STRINGS[brightness]);
}
    
/**-----------------------------------------------------------------------------
 * 
 * @return 
 */
APP_COMMAND_DARK_MODE app_commands_get_dark_mode()
{
    bool result;
    // Update settings?
    settings_get(SETTING_DARK_MODE, &result); 
    
    return (APP_COMMAND_DARK_MODE)result;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the current balance value
 * @return 
 */
void app_commands_get_dark_mode_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", APP_COMMAND_DARK_MODE_STRINGS[(uint8_t)app_commands_get_dark_mode()]);
}

/**-----------------------------------------------------------------------------
 * 
 * @return 
 */
bool app_commands_brightness_get_remote_change()
{
    return brightness_data.remote_change;
}

/**-----------------------------------------------------------------------------
 * Sets the current full brightness setting
 */
void app_commands_set_brightness(APP_COMMAND_BRIGHTNESS brightness)
{
    if (app_commands_get_brightness() != brightness)
    {
        brightness_data.full_brightness = max(min(brightness, APP_COMMAND_BRIGHTNESS_FULL), APP_COMMAND_BRIGHTNESS_OFF);

        // Because this is changed in so many places we do it with a logger_log_command message
        uint8_t value = (uint8_t)app_commands_get_brightness();
        logger_log_command(APP_COMMANDS_BRIGHTNESS, &value, 1);
        
        // Update settings?
        settings_set(SETTING_BRIGHTNESS, &value); //app_context.full_brightness); use value as setting is defined as uint8_t whereas APP_COMMAND_BRIGHTNESS is uint32_t

        logger_log_infof(LOGGER_CODE_BRIGHTNESS, "BRIGHTNESS: %s", APP_COMMAND_BRIGHTNESS_STRINGS[app_commands_get_brightness()]);
    }
}

/**-----------------------------------------------------------------------------
 * Sets the current dark mode setting
 */
void app_commands_set_dark_mode(APP_COMMAND_DARK_MODE dark_mode)
{
    if (app_commands_get_dark_mode() != dark_mode)
    {
        bool value = (bool)dark_mode;
        // Update settings?
        settings_set(SETTING_DARK_MODE, &value); 

        logger_log_infof(LOGGER_CODE_DARK_MODE, "DARK MODE %s", APP_COMMAND_DARK_MODE_STRINGS[app_commands_get_dark_mode()]);
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param change
 */
void app_commands_brightness_set_remote_change(bool change)
{
    brightness_data.remote_change = change;
}

/* *****************************************************************************
 End of File
 */
