/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Manages commands related to the engineering menu

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_engineering.h"
#include <stdio.h>
#include "app_config.h"
#include "utilities/input_context.h"
#include "utilities/display.h"
#include "utilities/menu_display.h"
#include "utilities/remote.h"
#include "utilities/structured_message.h"
#include "utilities/system_fault.h"
#include "utilities/timers.h"
#include "utilities/watchdog.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */
//--------------------------------------------------------------------------
// Application Commands - Fault simulation
//--------------------------------------------------------------------------
typedef enum
{
    APP_COMMAND_INJECT_FAULT_SHORT_CIRCUIT      = 0x00,
    APP_COMMAND_INJECT_FAULT_DC_OFFSET_FAULT    = 0x01,
    APP_COMMAND_INJECT_FAULT_OUTPUT_OVER_TEMP   = 0x02,
    APP_COMMAND_INJECT_FAULT_LIFTER_OVER_TEMP   = 0x03,
    APP_COMMAND_INJECT_FAULT_SHUTDOWN_TIMER     = 0x04,
} APP_COMMAND_INJECT_FAULT;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used to enter/exit the BIAS test
 */
bool app_commands_bias_test(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    // Check the message command (switch on/off)
    // Check state and see if command is valid)

    if (!message->repeat)
    {
        message->command = APP_COMMANDS_BIAS_CHECK;
        message->response = STATUS_UPDATE;
        message->data_length = 0;
        
        switch (message->data[0])
        {
            case APP_COMMAND_BIAS_CHECK_START:
            {
                if (app_context_get_app_state() == APP_STATE_RUNNING)
                {
                    app_context_set_app_state(APP_STATE_ENTER_BIAS_CHECK);
                }
                else
                {
                    message->response = COMMAND_INVALID_NOW;
                }
                break;
            }

            case APP_COMMAND_BIAS_CHECK_STOP:
            {
                if (app_context_get_app_state() == APP_STATE_BIAS_CHECK)
                {
                    app_context_set_app_state(APP_STATE_EXIT_BIAS_CHECK);
                }
                else
                {
                    message->response = COMMAND_INVALID_NOW;
                }
                break;
            }

            case APP_COMMAND_BIAS_CHECK_REQUEST_STATE:
            {
                // 0x00 - bias test not running, 0x01 - running
                message->data[0] = app_context_get_app_state() == APP_STATE_BIAS_CHECK;
                message->data_length = 1;
                break;
            }
                
            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                break;
            }
        }
    }
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Draw the ir_debug menu
 */
void app_commands_draw_bias_check(APP_COMMAND_MENU menu)
{
   // Clear the display
    display_clear(false);
    
    char buffer[16];
    char temp_buffer[8];

    sprintf(buffer, "BIAS");
    display_start_text(DISPLAY_LINE_1, buffer);
    
    // Using menu_item.edit_info to store aggregate values for temp and bias readings
    //MENU_ITEM* menu_item = menu_get_item(APP_COMMAND_MENU_BIAS_CHECK);
       
    app_commands_get_amplifier_temperature_string(temp_buffer, 8);

    display_append_text_right(DISPLAY_LINE_1, "%s", temp_buffer);
 
    display_write_line(DISPLAY_LINE_1, false);

    uint32_t mVL = adc_voltage_get(APP_ADC_INPUTS_BIAS_TEST_L);
    uint32_t mVR = adc_voltage_get(APP_ADC_INPUTS_BIAS_TEST_R);

    // Format the output voltage and convert to original mV by / 50
    sprintf(buffer, "L%2.1fmV  R%2.1fmV", (float) mVL / 50, (float) mVR / 50);

    display_start_text(DISPLAY_LINE_2, buffer);
    display_write_line(DISPLAY_LINE_2, false);
}

/**-----------------------------------------------------------------------------
 * Draw the ir_debug menu
 */
void app_commands_draw_ir_debug(APP_COMMAND_MENU menu)
{
   // Clear the display
    display_clear(false);

    STRUCTURED_MESSAGE* message = input_context_get_remote_message();
    
    display_start_text(DISPLAY_LINE_1, "IR");
    
    // Try and find the value in the command enumeration that has this value
    int idx = app_commands_get_array_index_of((int)message->command, REMOTE_CMD_VALUES, sizeof(REMOTE_CMD_VALUES) / sizeof(int));
    
    REMOTE_SYSTEM_CODE code;
    memcpy(&code, message->data, sizeof(REMOTE_SYSTEM_CODE));
    
    if (idx > -1)
    {
        display_append_text(DISPLAY_LINE_1, 3, "%s", REMOTE_CMD_INPUT_STRINGS[idx]);
        display_append_text_right(DISPLAY_LINE_1, " %d/%d", code, message->command);
    }
    else
    {
        display_append_text(DISPLAY_LINE_1, 3, "Unused");
        display_append_text_right(DISPLAY_LINE_1, " %d/%d", code, message->command);
    }
  
    display_write_line(DISPLAY_LINE_1, false);

    // Draw navigation buttons
    menu_display_draw_generic_navigation(menu_get_item(menu), MENU_SOFTKEYS_UP_DOWN, true);    
}


/**-----------------------------------------------------------------------------
 * Used to edit the check channel status setting
 */
bool app_commands_edit_check_channel_status(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = (int8_t)message->data[0];
        
        bool check_channel_status = false;
        settings_get(SETTING_CHECK_CHANNEL_STATUS, &check_channel_status);
        
        int8_t current_value = (int8_t)check_channel_status;
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        if (app_commands_process_message_value_bool(message, &current_value, &value, true))
        {
            check_channel_status = (bool)current_value;
            settings_set(SETTING_CHECK_CHANNEL_STATUS, &check_channel_status);

            menu_update_change_state(&current_value, sizeof(int8_t));
        }             
    }    
     
    return result;
}

//! Called when entering the IR Debug menu
void app_commands_enter_bias_check(MENU_ID menu)
{
    // Mute outputs pause for 1 second
    output_control_mute_outputs(true);

    // Make the ADC BIAS Test channels active
    adc_set_input_active(APP_ADC_INPUTS_BIAS_TEST_L, true);
    adc_set_input_active(APP_ADC_INPUTS_BIAS_TEST_R, true);
    
    //------------------------------------
    app_context_set_mode(APP_CONTEXT_MODE_BIAS_TEST);

    timers_reset(TIMER_INDEX_BIAS_TEST);
    
    app_context_set_app_state(APP_STATE_BIAS_CHECK);
}

/**-----------------------------------------------------------------------------
 * 
 * @param menu
 * @return 
 */
//! Called when leaving the IR Debug menu
MENU_ID app_commands_exit_bias_check(MENU_ID menu)
{
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;
    
    adc_set_input_active(APP_ADC_INPUTS_BIAS_TEST_L, false);
    adc_set_input_active(APP_ADC_INPUTS_BIAS_TEST_R, false);

    // Unmute outputs
    output_control_unmute_outputs();
    
    //-------------------------------------------
    power_reset_auto_shutdown_timer();
    
    app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
    
    app_context_set_app_state(APP_STATE_RUNNING);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Called when entering the IR Debug menu
 */
void app_commands_enter_ir_debug(MENU_ID menu)
{
    input_context_suspend_remote(true);
}

/**-----------------------------------------------------------------------------
 * Called when leaving the IR Debug menu
 */
MENU_ID app_commands_exit_ir_debug(MENU_ID menu)
{
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;
    
    input_context_suspend_remote(false);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Called when entering the Switch Voltage menu
 */
void app_commands_enter_switch_voltages(MENU_ID menu)
{
    input_context_suspend_button(true);
}

/**-----------------------------------------------------------------------------
 * Called when leaving the Switch Voltage menu
 */
MENU_ID app_commands_exit_switch_voltages(MENU_ID menu)
{
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;
    
    input_context_suspend_button(false);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the on/off check channel status value
 * @return 
 */
void app_commands_get_check_channel_status_string(char* buffer, size_t size)
{
    bool check_channel_status = false;
    settings_get(SETTING_CHECK_CHANNEL_STATUS, &check_channel_status);
    sprintf(buffer, "%s", APP_COMMAND_OFF_ON_STRINGS[check_channel_status ? 1 : 0]);
}

/**-----------------------------------------------------------------------------
 * Used to simulate fault conditions via external commands
 * @param message
 */
void app_commands_inject_fault(STRUCTURED_MESSAGE* message)
{
    message->command = APP_COMMANDS_INJECT_FAULT;
    message->response = STATUS_UPDATE;
    
    SYSTEM_FAULT fault;
    fault.code = SYSTEM_FAULT_CODE_NO_FAULT;
    
    APP_COMMAND_INJECT_FAULT fault_type = message->data[0];
    
    logger_log_debugf(0, "APP COMMANDS: Injecting fault %d ", fault_type);
    
    switch(fault_type)
    {
        case APP_COMMAND_INJECT_FAULT_SHORT_CIRCUIT:
        {
            fault.code = SYSTEM_FAULT_CODE_SHORT_CIRCUIT;
            sprintf(fault.description, "Short (SIM)");

            // Store the fault
            system_fault_set(&fault);

            message->data[0] = 0x01;
            break;
        }
        case APP_COMMAND_INJECT_FAULT_DC_OFFSET_FAULT:
        {
            fault.code = SYSTEM_FAULT_CODE_DC_OFFSET;
            sprintf(fault.description, "DC Offset (SIM)");

            // Store the fault
            system_fault_set(&fault);

            message->data[0] = 0x01;
            break;
        }
        case APP_COMMAND_INJECT_FAULT_OUTPUT_OVER_TEMP:
        {
            fault.code = SYSTEM_FAULT_CODE_OUTPUT_OVER_TEMP;
            sprintf(fault.description, "Out Temp (SIM)");

            // Store the fault
            system_fault_set(&fault);

            message->data[0] = 0x01;
            break;
        }
        case APP_COMMAND_INJECT_FAULT_LIFTER_OVER_TEMP:
        {
            fault.code = SYSTEM_FAULT_CODE_LIFTER_OVER_TEMP;
            sprintf(fault.description, "Lifter (SIM)");

            // Store the fault
            system_fault_set(&fault);

            message->data[0] = 0x01;
            break;
        }
        case APP_COMMAND_INJECT_FAULT_SHUTDOWN_TIMER:
        {
            // Sets the timer to 1 minute, but does not update the setting which
            // will be restored on reboot
            timers_update_timeout(TIMER_INDEX_AUTO_SHUTDOWN, 60000, TIMERS_RESET_IMMEDIATE);
            
            app_context_set_menu(APP_COMMAND_MENU_STANDBY_TIMER);
            
            message->data[0] = 0x01;
            break;
        }
        default:
        {
            message->response = PARAM_NOT_RECOGNISED;         
        }
    }

    message->is_activity = false;
    message->data_length = 1;
    
    strcpy(fault.description, system_fault_codes[fault.code]);
    system_fault_set(&fault);
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Launch the engineering menu
 */
void app_commands_launch_engineering_menu(STRUCTURED_MESSAGE* message)
{
    app_context_set_mode(APP_CONTEXT_MODE_MENU);

    // Suspend the timeout while in engineering mode
    timers_suspend(TIMER_INDEX_MENU_INACTIVE);

    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
    message->command = APP_COMMANDS_MENU;
    message->data[0] = APP_COMMAND_MENU_ENGINEERING_SETTINGS;     
}

/**-----------------------------------------------------------------------------
 * Launch the software version menu
 */
void app_commands_launch_reset_menu(STRUCTURED_MESSAGE* message)
{
    app_context_set_mode(APP_CONTEXT_MODE_MENU);

    // Suspend the timeout while in engineering mode
    //timers_suspend(APP_TIMER_INDEX_MENU_INACTIVE);

    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
    message->command = APP_COMMANDS_MENU;
    message->data[0] = APP_COMMAND_MENU_SYSTEM_RESET;     
}

/**-----------------------------------------------------------------------------
 * Launch the software version menu
 */
void app_commands_launch_software_version_menu(STRUCTURED_MESSAGE* message)
{
    app_context_set_mode(APP_CONTEXT_MODE_MENU);

    // Suspend the timeout while in engineering mode
    //timers_suspend(APP_TIMER_INDEX_MENU_INACTIVE);

    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
    message->command = APP_COMMANDS_MENU;
    message->data[0] = APP_COMMAND_MENU_SYSTEM_INFO_HOST_VERSION;     
}

/**-----------------------------------------------------------------------------
 * Performs the BIAS check as part of the main message loop
 */
void app_commands_perform_bias_check()
{
    // Special state for performing the Bias checks
    // Repeat test every n seconds until told to stop
    if (timers_has_expired(TIMER_INDEX_BIAS_TEST))
    {
        STRUCTURED_MESSAGE message;
        structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_RS232);
        message.command = APP_COMMANDS_BIAS_CHECK;
        message.data[0] = APP_COMMAND_BIAS_CHECK_DATA;
        message.data[1] = 2; // Number of channels

        // Get voltages 
        uint32_t temp;
        temp = adc_voltage_get(APP_ADC_INPUTS_BIAS_TEST_L);
        memcpy(&message.data[2], &temp, sizeof(uint32_t));

        temp = adc_voltage_get(APP_ADC_INPUTS_BIAS_TEST_R);
        memcpy(&message.data[6], &temp, sizeof(uint32_t));

        message.data_length = 2 + (message.data[1] * 4);

        app_commands_send_message(&message);

        // Start the timer
        timers_reset(TIMER_INDEX_BIAS_TEST);
    }
    
}
 
/**-----------------------------------------------------------------------------
 * Returns the current watchdog status
 */
void app_commands_watchdog_status(STRUCTURED_MESSAGE* message)
{
    if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
    {
        message->data[0] = watchdog_get_status();
        message->data_length = 1;
        message->response = STATUS_UPDATE;
    }
    else
    {
        message->response = COMMAND_INVALID_NOW;
    }

    app_commands_send_message(message);
}
   
/* *****************************************************************************
 End of File
 */
