/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_hdmi_arc.h"
#include "app_config.h"
#include "app_context.h"
#include "crypto/crypto.h"
#include "crypto/src/md5.h"
#include "utilities\hdmi_arc.h"
#include "utilities\hdmi_arc_control.h"
#include "utilities\hdmi_arc_data.h"
#include "utilities\input_control.h"
#include "utilities/input_phono_input.h"
#include "utilities\mutes.h"
#include "utilities\power.h"
#include "utilities\structured_message.h"
#include "utilities\tcpip.h"
#include "utilities\timers.h"
#include "utilities\volume.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef enum
{
    HDMI_ARC_UPGRADE_STATUS_INACTIVE,
    HDMI_ARC_UPGRADE_STATUS_PREPARED,
    HDMI_ARC_UPGRADE_STATUS_STARTED,
    HDMI_ARC_UPGRADE_STATUS_DATA,
} HDMI_ARC_UPGRADE_STATUS;

typedef enum
{
    HDMI_ARC_PROCESSING_STAGE_WAIT,
    HDMI_ARC_PROCESSING_STAGE_RESET_OFF,
    HDMI_ARC_PROCESSING_STAGE_RESET_ON,
    HDMI_ARC_PROCESSING_STAGE_IN_UPDATE,
} HDMI_ARC_PROCESSING_STAGE;

typedef struct
{
    // Keeps track of the status messages received from the ARC module
    HDMI_ARC_UPGRADE_STATUS status;
    
    //! The load ID associated with the current upload (set to 0x80)
    UPGRADE_METADATA_LOAD_ID upload_load_id;
    
    //! The number of bytes written in the current upload
    size_t upload_bytes_written;    
    
    //! Allows for a 1-time override of the version checking for upgrade
    bool upload_override_version_check;
    
    //! The size of the current upload
    size_t upload_upload_size;
    
    //! Major/Minor version of current upgrade
    uint8_t major_version; 
    uint8_t minor_version; 
            
    //! Size of curret upgrade
    size_t size;
            
    // The checksum provided at the start of the upgrade    
    uint8_t checksum[MD5_DIGEST_SIZE];
    
    //! MD5 context structure used in MD5 checksum generation
    CRYPT_MD5_CTX md5_context;
    
    //! A cache to hold the data sent by the avatar while it is being transferred 
    // to the HDMI Arc in 512 packets.
    uint8_t data_cache[5000];//4096];
    
    //! The actual length of the data stored in the cache.
    uint16_t data_cache_length;
    
    //! The offset of data in the cache that is to be sent to the HDMI ARC
    uint16_t data_cache_offset;
    
    //! The number of retries for the current packet
    uint8_t data_cache_retry_count;
    
    //! The size of the data sent in the current packet
    uint16_t data_cache_packet_data_size;
    
    //! The HDMI ARC only accepts packets with a whole number of bytes.  We can 
    // receive data with odd numbers of bytes so we need to save the last byte and 
    // add it to the cache for the next lump..To do this we set the position in
    // the cache where we start copying data to 
    uint8_t cache_start_pos;
    
    UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback;

} HDMI_ARC_UPGRADE_DATA;

typedef struct
{
    HDMI_ARC_PROCESSING_STAGE processing_stage;
    
    //! Handshaking during an upgrade messes up the upgrade (even if its the SA30 not th ARC module)
    //  for some reason!
    bool suppress_handshake;
    
    //! Information about the current display upgrade
    HDMI_ARC_UPGRADE_DATA upgrade_data;
} HDMI_ARC_DATA;

static HDMI_ARC_DATA hdmi_arc_data = {
    .processing_stage = HDMI_ARC_PROCESSING_STAGE_WAIT,
    .suppress_handshake = false,
    .upgrade_data.local_callback = NULL,
    .upgrade_data.upload_override_version_check = false,
    .upgrade_data.cache_start_pos = 0,
    .upgrade_data.status = HDMI_ARC_UPGRADE_STATUS_INACTIVE,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Overrides the version check for the upgrade one time only
 */
void app_commands_hdmi_arc_override_upgrade_version_check()
{
    hdmi_arc_data.upgrade_data.upload_override_version_check = true;
}

/**-----------------------------------------------------------------------------
 * Version check used to decide whether the upgrade can continue based on the 
 * version number read from the packed upgrade image.
 * @param major_version
 * @param minor_version
 * @return 
 */
bool app_commands_hdmi_arc_check_upgrade_version(uint8_t major_version, uint8_t minor_version)
{
    bool result = true;
    uint8_t current_version_major = hdmi_arc_get_firmware_version_major();
    uint8_t current_version_minor = hdmi_arc_get_firmware_version_minor();;
    
    if (hdmi_arc_data.upgrade_data.upload_override_version_check)
    {
        result = true;
    }
    else if ((current_version_major == 0) && (current_version_minor == 0))
    {
        // Assume there is no HDMI ARC board
        result = false;
    }
    else if (major_version < current_version_major)
    {
        result = false;
    }
    else if (major_version == current_version_major)
    {
        if (minor_version <= current_version_minor)
        {
            result = false;
        }
    }
    
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Checking Version Numbers %d/%d with upgrade %d/%d (%d/%d) --", 
            current_version_major, 
            current_version_minor, 
            major_version, 
            minor_version, 
            result, 
            hdmi_arc_data.upgrade_data.upload_override_version_check);
   
    return result;
}

/**-----------------------------------------------------------------------------
 * Edits the HDMI ARC Audio control setting
 */
bool app_commands_hdmi_arc_edit_audio_control(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    logger_log_debugf(0, "HDMI AUDIO CONTROL %d --", message->data[0]);
    logger_process();
    
    app_commands_review_message_action(message);

    int16_t value = (int16_t)message->data[0];
    HDMI_ARC_CONTROL_STATE current_value = hdmi_arc_get_audio_control_setting();

    // Set changed state
    if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        menu_try_initialise_change_state(&current_value, sizeof(HDMI_ARC_CONTROL_STATE));
    }

    app_commands_process_message_value_bool(message, &current_value, &value, true);

    //message->command = APP_COMMANDS_HDMI_ARC_AUDIO_CONTROL;
    message->response = STATUS_UPDATE;

    if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        hdmi_arc_set_audio_control_setting(current_value);

        menu_update_change_state(&current_value, sizeof(PHONO_INPUT));

        message->data[0] = (uint8_t)current_value;
 
        message->data_length = 1;
    }
        
    // Send the response
    app_commands_send_message(message);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Edits the HDMI ARC Power control setting
 */
bool app_commands_hdmi_arc_edit_power_control(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    logger_log_debugf(0, "HDMI POWER CONTROL %d --", message->data[0]);
    logger_process();
    
    app_commands_review_message_action(message);

    int16_t value = (int16_t)message->data[0];
    HDMI_ARC_CONTROL_STATE current_value = hdmi_arc_get_power_control_setting();

    // Set changed state
    if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        menu_try_initialise_change_state(&current_value, sizeof(HDMI_ARC_CONTROL_STATE));
    }

    app_commands_process_message_value_bool(message, &current_value, &value, true);

    //message->command = APP_COMMANDS_HDMI_ARC_AUDIO_CONTROL;
    message->response = STATUS_UPDATE;

    if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        hdmi_arc_set_power_control_setting(current_value);

        menu_update_change_state(&current_value, sizeof(PHONO_INPUT));

        message->data[0] = (uint8_t)current_value;
 
        message->data_length = 1;
    }
        
    // Send the response
    app_commands_send_message(message);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the HDMI ARC Audio Control
 * @return 
 */
void app_commands_hdmi_arc_get_hdmi_audio_control_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", HDMI_ARC_CONTROL_STATE_STRINGS[hdmi_arc_get_audio_control_setting()]);
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the HDMI ARC Power Control
 * @return 
 */
void app_commands_hdmi_arc_get_hdmi_power_control_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", HDMI_ARC_CONTROL_STATE_STRINGS[hdmi_arc_get_power_control_setting()]);
}

/**-----------------------------------------------------------------------------
 * Processes the Mute message from the HDMI ARC
 * @param message
 */
static void app_commands_hdmi_arc_process_arc_stream(HDMI_ARC_MESSAGE* message)
{
    // We check for APp state because the RC module was sending an ARC Stream 0 
    // after a power down message which caused the input to revert to PHONO
    // When the ARC wakes the module up it then was going to the wrong input
    if (message->command_type == HDMI_ARC_COMMAND_ARC_STREAM)
    {
        //(app_context_get_app_state() != APP_STATE_POWEREDDOWN))
        if (app_context_get_app_state() == APP_STATE_RUNNING)
        {
            bool process = true;
            HDMI_ARC_STREAM_STATE state = (HDMI_ARC_STREAM_STATE)message->data[0];

            logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Processing ARC Stream %d --", state);
            logger_process();

            STRUCTURED_MESSAGE message;
            structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
            message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message.command = APP_COMMANDS_CURRENT_SOURCE;
            message.data_length = 1;   

            switch (state)
            {
                case HDMI_ARC_STREAM_STATE_INACTIVE:
                {
                    // logger_log_debugf(0, "HDMI ARC app_commands_hdmi_arc_process_arc_stream HDMI_ARC_STREAM_STATE_INACTIVE --");
                    // logger_process();
                    INPUT_CONTROL_DEFINITION* prev_input = input_control_get_previous_definition();

                    if (prev_input != NULL)
                    {
                        message.data[0] = prev_input->index;
                        if (message.data[0] == AUDIO_INPUT_HDMI_ARC)
                        {
                            process = false;
                        }
                    }
                    else
                    {
                        message.data[0] = AUDIO_INPUT_CD;
                    }

                    break;
                }

                case HDMI_ARC_STREAM_STATE_ACTIVE:
                {
                    message.data[0] = AUDIO_INPUT_HDMI_ARC;

                    //Check whether were in sleep mode - if we are then we need to come out of it...
                    break;
                }
            }

            if (process)
            {
                // Translate into AUDIO_INPUT_VALUES as its translated back in app_commands_current_source
                message.data[0] = AUDIO_INPUT_VALUES[message.data[0]];

                app_commands_process_message(&message);
            }
        }
        else
        {
            logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Ignoring ARC Stream in non RUNNING STATE %d --", app_context_get_app_state());
            logger_process();
        }
    }
}

/**-----------------------------------------------------------------------------
 * Processes the Mute message from the HDMI ARC
 * @param message
 */
static void app_commands_hdmi_arc_process_mute(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_MUTE)
    {
        HDMI_ARC_MUTE_STATE state = message->data[0];
        
        STRUCTURED_MESSAGE message;
        structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
        message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
        message.command = APP_COMMANDS_MUTE;
        message.data_length = 1;   
        
        switch (state)
        {
            case HDMI_ARC_MUTE_STATE_MUTED:
            {
                message.data[0] = MUTE_STATE_ON;
                break;
            }

            case HDMI_ARC_MUTE_STATE_UNMUTED:
            {
                message.data[0] = MUTE_STATE_OFF;
                break;
            }
        }
        
        app_commands_process_message(&message);
    }
}

/**-----------------------------------------------------------------------------
 * Processes the Power message from the HDMI ARC
 * @param message
 */
static void app_commands_hdmi_arc_process_power(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_POWER)
    {
        if (hdmi_arc_get_power_control_setting() == HDMI_ARC_CONTROL_STATE_AUTO)
        {
            bool process = true;

            HDMI_ARC_POWER_STATE state = message->data[0];

            STRUCTURED_MESSAGE message;
            structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);
            message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message.command = APP_COMMANDS_POWER;
            message.data_length = 1;   

            switch (state)
            {
                case HDMI_ARC_POWER_STATE_OFF:
                case HDMI_ARC_POWER_STATE_IDLE:
                {
                    message.data[0] = POWER_STATE_OFF;

                    // Check the current input is ARC 27/09/2019*16:36
                    if (input_control_get_selected_input() != AUDIO_INPUT_HDMI_ARC)
                    {
                        process = false;
                    }

                    break;
                }

                case HDMI_ARC_POWER_STATE_ON:
                {
                    message.data[0] = POWER_STATE_ON;
                    break;
                }
            }

            if (process)
            {
                app_commands_process_message(&message);
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Processes the Upgrade Data message sent from the ARC MCU when it is requesting more data
 * @param message
 */
static void app_commands_hdmi_arc_process_upgrade_data(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_UPGRADE_DATA)
    {
        HDMI_ARC_UPGRADE_DATA_STATE state = message->data[0];

        hdmi_arc_data.upgrade_data.status = HDMI_ARC_UPGRADE_STATUS_DATA;

        switch (state)
        {
            case HDMI_ARC_UPGRADE_DATA_STATE_TRANSFER_SUCCESS:
            {
                logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC UPGRADE STATUS Data Received - invoking local callback --");
                logger_process();
                
                // Update the number of bytes written
                hdmi_arc_data.upgrade_data.upload_bytes_written += hdmi_arc_data.upgrade_data.data_cache_packet_data_size;
                
                // Increment the offset in the current data
                hdmi_arc_data.upgrade_data.data_cache_offset += hdmi_arc_data.upgrade_data.data_cache_packet_data_size;
                
                // Check if we've sent all the data over
                if ((hdmi_arc_data.upgrade_data.data_cache_offset >= hdmi_arc_data.upgrade_data.data_cache_length) ||
                    (hdmi_arc_data.upgrade_data.cache_start_pos > 0))
                {
                    // We've sent the whole data cache so tell the Avatar to send the next packet
                    if (hdmi_arc_data.upgrade_data.local_callback != NULL)
                    {
                        hdmi_arc_data.upgrade_data.local_callback(hdmi_arc_data.upgrade_data.upload_load_id, 
                                hdmi_arc_data.upgrade_data.upload_upload_size, 
                                hdmi_arc_data.upgrade_data.upload_bytes_written, 
                                UPGRADE_METADATA_UPGRADE_STATUS_OK);
                    }
                }
                else
                {
                    hdmi_arc_data.upgrade_data.data_cache_packet_data_size = 
                            min(hdmi_arc_data.upgrade_data.data_cache_length - hdmi_arc_data.upgrade_data.data_cache_offset,
                                HDMI_ARC_UPLOAD_DATA_MAX_PACKET_SIZE);
                    
                    //logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Upgrade Dat B %d %d %d  --", 
                    //        hdmi_arc_data.upgrade_data.data_cache_offset, 
                    //        hdmi_arc_data.upgrade_data.data_cache_packet_data_size, 
                    //        hdmi_arc_data.upgrade_data.data_cache_offset);
                    //logger_process();

                    // The HDMI ARC only accepts packets with an even number of bytes - ughhhhhhhhhhhhhhhhhh!
                    // So take the last byte and add to the cache for next time
                    // then set the cache_start_pos to 1
                    if ((hdmi_arc_data.upgrade_data.data_cache_packet_data_size < HDMI_ARC_UPLOAD_DATA_MAX_PACKET_SIZE) &&
                        (hdmi_arc_data.upgrade_data.data_cache_packet_data_size % 2 != 0))
                    {
                        // Copy the last byte to the start of the cache 
                        hdmi_arc_data.upgrade_data.data_cache[0] = hdmi_arc_data.upgrade_data.data_cache[hdmi_arc_data.upgrade_data.data_cache_length - 1];
                        
                        //logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Upgrade Dat BB %x %x %d %d --", hdmi_arc_data.upgrade_data.data_cache[0],
                        //        hdmi_arc_data.upgrade_data.data_cache[hdmi_arc_data.upgrade_data.data_cache_length - 1], 
                        //        hdmi_arc_data.upgrade_data.data_cache_offset + hdmi_arc_data.upgrade_data.data_cache_packet_data_size - 1,
                        //        hdmi_arc_data.upgrade_data.data_cache_length);
                        
                        // Truncate the leftover data 
                        hdmi_arc_data.upgrade_data.data_cache_packet_data_size = ((hdmi_arc_data.upgrade_data.data_cache_packet_data_size / 2) * 2);
                        
                        // Copy the last byte to the start of the cache 
                        //hdmi_arc_data.upgrade_data.data_cache[0] = hdmi_arc_data.upgrade_data.data_cache[hdmi_arc_data.upgrade_data.data_cache_offset + hdmi_arc_data.upgrade_data.data_cache_packet_data_size];
                        
                        // set the cache start pos
                        hdmi_arc_data.upgrade_data.cache_start_pos = 1;
                     
                        //logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Upgrade Dat CC %d %d %d  --", 
                        //        hdmi_arc_data.upgrade_data.data_cache_offset, 
                        //        hdmi_arc_data.upgrade_data.data_cache_packet_data_size, 
                        //        hdmi_arc_data.upgrade_data.data_cache_offset + hdmi_arc_data.upgrade_data.data_cache_packet_data_size);
                        //logger_process();
                   }
                    else
                    {
                        // Reset the cache start pos
                        hdmi_arc_data.upgrade_data.cache_start_pos = 0;
                    }
    
                    // Send the next packet of data
                    hdmi_arc_send_upgrade_data_message(
                            hdmi_arc_data.upgrade_data.upload_upload_size,
                            hdmi_arc_data.upgrade_data.upload_bytes_written,
                            (uint8_t*)(hdmi_arc_data.upgrade_data.data_cache + hdmi_arc_data.upgrade_data.data_cache_offset),
                            hdmi_arc_data.upgrade_data.data_cache_packet_data_size);
                }
                
                break;
            }
            
            case HDMI_ARC_UPGRADE_DATA_STATE_TRANSFER_FAILED:
            {
                // Prepare to resend upto 3 times
                hdmi_arc_data.upgrade_data.data_cache_retry_count++;
                

                if (hdmi_arc_data.upgrade_data.data_cache_retry_count > 3)
                {
                    // Give up
                    // 
                    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC UPGRADE STATUS Transfer failed on attempt %d --", hdmi_arc_data.upgrade_data.data_cache_retry_count);
                    if (hdmi_arc_data.upgrade_data.local_callback != NULL)
                    {
                        hdmi_arc_data.upgrade_data.local_callback(hdmi_arc_data.upgrade_data.upload_load_id, 
                                hdmi_arc_data.upgrade_data.upload_upload_size, 
                                hdmi_arc_data.upgrade_data.upload_bytes_written, 
                                UPGRADE_METADATA_UPGRADE_STATUS_ERROR);
                    }
                    
                }
                else
                {
                    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC UPGRADE STATUS Transfer failed on attempt %d Resending --", hdmi_arc_data.upgrade_data.data_cache_retry_count);
                    
                    // Send the last packet again
                    hdmi_arc_send_upgrade_data_message(
                            hdmi_arc_data.upgrade_data.upload_upload_size,
                            hdmi_arc_data.upgrade_data.upload_bytes_written,
                            (uint8_t*)(hdmi_arc_data.upgrade_data.data_cache + hdmi_arc_data.upgrade_data.data_cache_offset),
                            hdmi_arc_data.upgrade_data.data_cache_packet_data_size);
                }
                break;
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Processes the Upgrade Data message sent from the ARC MCU when it is requesting more data
 * @param message
 */
static void app_commands_hdmi_arc_process_upgrade_end(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_UPGRADE_END)
    {
        HDMI_ARC_UPGRADE_END_STATE state = message->data[0];

        hdmi_arc_data.upgrade_data.status = HDMI_ARC_UPGRADE_STATUS_INACTIVE;
        
        // At the moment we do the same thing whether it worked or failed
        switch (state)
        {
            case HDMI_ARC_UPGRADE_END_STATE_UPGRADE_COMPLETE:
            {
                // Complete and compare checksums and data_length?
                uint8_t checksum[MD5_DIGEST_SIZE];
                CRYPT_MD5_Finalize(&hdmi_arc_data.upgrade_data.md5_context, checksum);
                
                logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE HDMI ARC Checksum - Orig %02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x --", 
                    hdmi_arc_data.upgrade_data.checksum[0],
                    hdmi_arc_data.upgrade_data.checksum[1],
                    hdmi_arc_data.upgrade_data.checksum[2],
                    hdmi_arc_data.upgrade_data.checksum[3],
                    hdmi_arc_data.upgrade_data.checksum[4],
                    hdmi_arc_data.upgrade_data.checksum[5],
                    hdmi_arc_data.upgrade_data.checksum[6],
                    hdmi_arc_data.upgrade_data.checksum[7],
                    hdmi_arc_data.upgrade_data.checksum[8],
                    hdmi_arc_data.upgrade_data.checksum[9],
                    hdmi_arc_data.upgrade_data.checksum[10],
                    hdmi_arc_data.upgrade_data.checksum[11],
                    hdmi_arc_data.upgrade_data.checksum[12],
                    hdmi_arc_data.upgrade_data.checksum[13],
                    hdmi_arc_data.upgrade_data.checksum[14],
                    hdmi_arc_data.upgrade_data.checksum[15]
                    );

                logger_log_debugf(LOGGER_CODE_UPGRADE, "UPGRADE HDMI ARC Checksum - Tran %02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x --", 
                    checksum[0],
                    checksum[1],
                    checksum[2],
                    checksum[3],
                    checksum[4],
                    checksum[5],
                    checksum[6],
                    checksum[7],
                    checksum[8],
                    checksum[9],
                    checksum[10],
                    checksum[11],
                    checksum[12],
                    checksum[13],
                    checksum[14],
                    checksum[15]
                    );

                if (memcmp(hdmi_arc_data.upgrade_data.checksum, checksum, MD5_DIGEST_SIZE) == 0)
                {
                    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC UPGRADE STATUS Transfer successful --");
                }
                else
                {
                    logger_log_errorf(LOGGER_CODE_HDMI_ARC, "HDMI ARC UPGRADE STATUS Transfer failed --");
                }
                
                logger_process();
// These have to be take out otherwise the end is called twice and the upgrade is abandoned half way through                
//                if (hdmi_arc_data.upgrade_data.local_callback != NULL)
//                {
//                    hdmi_arc_data.upgrade_data.local_callback(hdmi_arc_data.upgrade_data.upload_load_id, 
//                            hdmi_arc_data.upgrade_data.upload_upload_size, 
//                            hdmi_arc_data.upgrade_data.upload_bytes_written, 
//                            UPGRADE_METADATA_UPGRADE_STATUS_OK);
//                }
                break;
            }
            
            case HDMI_ARC_UPGRADE_END_STATE_UPGRADE_FAILED:
            {
                logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC UPGRADE STATUS Transfer errors --");
//                if (hdmi_arc_data.upgrade_data.local_callback != NULL)
//                {
//                    hdmi_arc_data.upgrade_data.local_callback(hdmi_arc_data.upgrade_data.upload_load_id, 
//                            hdmi_arc_data.upgrade_data.upload_upload_size, 
//                            hdmi_arc_data.upgrade_data.upload_bytes_written, 
//                            UPGRADE_METADATA_UPGRADE_STATUS_ERROR);
//                }
                break;
            }
            
            default:
            {
                logger_log_errorf(LOGGER_CODE_HDMI_ARC, "HDMI ARC UPGRADE STATUS Invalid status %d --");
                break;
            }
        }
    }
}

/**-----------------------------------------------------------------------------
 * Processes the Upgrade Data message sent from the ARC MCU when it is requesting more data
 * @param message
 */
static void app_commands_hdmi_arc_process_upgrade_start(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_UPGRADE_START)
    {
        if (hdmi_arc_data.upgrade_data.status == HDMI_ARC_UPGRADE_STATUS_PREPARED)
        {
            // Send the first packet of data
            hdmi_arc_send_upgrade_data_message(
                    hdmi_arc_data.upgrade_data.upload_upload_size,
                    hdmi_arc_data.upgrade_data.upload_bytes_written,
                    hdmi_arc_data.upgrade_data.data_cache,
                    hdmi_arc_data.upgrade_data.data_cache_packet_data_size);
        }

        hdmi_arc_data.upgrade_data.status = HDMI_ARC_UPGRADE_STATUS_STARTED;
    }
}

/**-----------------------------------------------------------------------------
 * Processes the Volume message from the HDMI ARC
 * @param message
 */
static void app_commands_hdmi_arc_process_volume(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_VOLUME)
    {
        if (hdmi_arc_get_audio_control_setting() == HDMI_ARC_CONTROL_STATE_AUTO)
        {
            uint8_t volume = message->data[0];

            STRUCTURED_MESSAGE message_out;
            structured_message_initialise(&message_out, STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT);

            // Map volume
            uint8_t max_volume;
            settings_get(SETTING_MAXIMUM_VOLUME, &max_volume);
        
            uint16_t mapped_volume = volume * (uint16_t)max_volume;
            volume = min(mapped_volume / (uint16_t)VOLUME_MAX, (uint16_t)max_volume);

            logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Volume recv:%d map:%d --", message->data[0], volume);

            message_out.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message_out.command = APP_COMMANDS_VOLUME;
            message_out.data[0] = volume;
            message_out.data_length = 1;

#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
            hdmi_arc_suppress_messages(true);
#endif
            app_commands_process_message(&message_out);
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
            hdmi_arc_suppress_messages(false);
#endif
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Callback for RS232 connection
 * @param message
 */
void app_commands_hdmi_arc_callback(HDMI_ARC_MESSAGE* message)
{
    app_commands_hdmi_arc_process_mute(message);
    
    app_commands_hdmi_arc_process_power(message);

    app_commands_hdmi_arc_process_volume(message);
    
    app_commands_hdmi_arc_process_arc_stream(message);
    
    app_commands_hdmi_arc_process_upgrade_start(message);
    
    app_commands_hdmi_arc_process_upgrade_data(message);
    
    app_commands_hdmi_arc_process_upgrade_end(message);
}

/**-----------------------------------------------------------------------------
 * Sends a request for the HDMI ARC Transmitter version when entering the version number menu
 * @param 
 */
void app_commands_enter_hdmi_arc_rx_version_menu(MENU_ID menu_id)
{
    hdmi_arc_send_transmitter_version_request();
}

/**-----------------------------------------------------------------------------
 * Determines whether the HDMI ARC Input is currently active
 * @return 
 */
bool app_commands_hdmi_arc_is_active()
{
    AUDIO_INPUT selected_input = input_control_get_selected_input();
    return selected_input == AUDIO_INPUT_HDMI_ARC;
}
    
/**-----------------------------------------------------------------------------
 * Regular HDMI ARC processing to ensure that handshaking has occurred
 */
void app_commands_hdmi_arc_process()
{
//    if ((hdmi_arc_get_firmware_version_major() == 0) && (hdmi_arc_get_firmware_version_minor() == 0) &&
//            !hdmi_arc_data.suppress_handshake)
    if (!hdmi_arc_data.suppress_handshake)
    {
        if (timers_has_expired(TIMER_INDEX_HDMI_ARC_HANDSHAKE))
        {
            switch (hdmi_arc_data.processing_stage)
            {
                case HDMI_ARC_PROCESSING_STAGE_WAIT:
                {
                    if (hdmi_arc_has_firmware_version())
                    {
                        // Miss the reset
                        hdmi_arc_data.processing_stage = HDMI_ARC_PROCESSING_STAGE_RESET_ON;
                    }
                    else
                    {
                        // Resetting HDMI_ARC Module
                        logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Resetting HDMI ARC Module OFF --");

                        hdmi_arc_reset(false);
                        timers_update_timeout(TIMER_INDEX_HDMI_ARC_HANDSHAKE, 1000, TIMERS_TIMEOUT_RESET_IMMEDIATE);

                        hdmi_arc_data.processing_stage = HDMI_ARC_PROCESSING_STAGE_RESET_OFF;
                    }
                    break;
                }

                case HDMI_ARC_PROCESSING_STAGE_RESET_OFF:
                {
                    // Resetting HDMI_ARC Module
                    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Resetting HDMI ARC Module ON --");
                    
                    hdmi_arc_reset(true);
                    timers_update_timeout(TIMER_INDEX_HDMI_ARC_HANDSHAKE, 1000, TIMERS_TIMEOUT_RESET_IMMEDIATE);
                    
                    hdmi_arc_data.processing_stage = HDMI_ARC_PROCESSING_STAGE_RESET_ON;
                    break;
                }

                case HDMI_ARC_PROCESSING_STAGE_RESET_ON:
                {
                    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Attempting Handshake --");
                    
                    app_commands_hdmi_arc_system_ready();
                    timers_update_timeout(TIMER_INDEX_HDMI_ARC_HANDSHAKE, 60000, TIMERS_TIMEOUT_RESET_IMMEDIATE);
                   
                    hdmi_arc_data.processing_stage = HDMI_ARC_PROCESSING_STAGE_WAIT;
                    break;
                }
            }
        }
    }
}
    
/**-----------------------------------------------------------------------------
 * Used to suppress/unsuppress handshaking (during update where it can mess up the
 * SA30 update
 */
void app_commands_hdmi_arc_suppress_handshake(bool suppress)
{
    hdmi_arc_data.suppress_handshake = suppress;
}
    
/**-----------------------------------------------------------------------------
 * Indicates to the ARC MCU that the system is ready to receive ARC commands
 */
void app_commands_hdmi_arc_system_ready()
{
    if (hdmi_arc_data.upgrade_data.status == HDMI_ARC_UPGRADE_STATUS_INACTIVE)
    {
        logger_log_infof(LOGGER_CODE_HDMI_ARC, "HDMI ARC Handshake Start --");

        hdmi_arc_send_system_ready_message(HDMI_ARC_SYSTEM_READY_STATE_HANDSHAKE_START);

        hdmi_arc_send_volume_message(volume_get());

        hdmi_arc_send_mute_message(mutes_state(MUTE_INDEX_SPEAKER) == MUTE_STATE_ON ? HDMI_ARC_MUTE_STATE_MUTED : HDMI_ARC_MUTE_STATE_UNMUTED);

        hdmi_arc_send_arc_stream_message(app_commands_hdmi_arc_is_active() ? HDMI_ARC_STREAM_STATE_ACTIVE : HDMI_ARC_STREAM_STATE_INACTIVE);

        char name[20];
        tcpip_get_network_name_string(name, 20);
        hdmi_arc_send_cec_name_message(name, strlen(name));

        hdmi_arc_send_system_ready_message(HDMI_ARC_SYSTEM_READY_STATE_HANDSHAKE_END);

        logger_log_infof(LOGGER_CODE_HDMI_ARC, "HDMI ARC Handshake End --");
        logger_process();
    }
    else
    {
        logger_log_infof(LOGGER_CODE_HDMI_ARC, "HDMI ARC Handshake Skipped --");
        logger_process();
    }
    
    timers_reset(TIMER_INDEX_HDMI_ARC_HANDSHAKE);
}

/**-----------------------------------------------------------------------------
 * Prepare the HDMI Arc upgrade
 */
UPGRADE_METADATA_LOAD_ID app_commands_hdmi_arc_upload_prepare(uint8_t major_version, uint8_t minor_version, size_t size, uint8_t* checksum, bool override)
{
    logger_log_debugf(0, "HDMI ARC UPLOAD Prepare --");
    
    hdmi_arc_data.processing_stage = HDMI_ARC_PROCESSING_STAGE_IN_UPDATE;
    
    // Store the upgrade information and initialise write variables
    hdmi_arc_data.upgrade_data.upload_load_id = 0x81;
    hdmi_arc_data.upgrade_data.upload_upload_size = size;
    hdmi_arc_data.upgrade_data.upload_bytes_written = 0;
    hdmi_arc_data.upgrade_data.major_version = major_version;
    hdmi_arc_data.upgrade_data.minor_version = minor_version;
    hdmi_arc_data.upgrade_data.size = size;
    memcpy(hdmi_arc_data.upgrade_data.checksum, checksum, MD5_DIGEST_SIZE);
        
    CRYPT_MD5_Initialize(&hdmi_arc_data.upgrade_data.md5_context);

    if (hdmi_arc_data.upgrade_data.status == HDMI_ARC_UPGRADE_STATUS_INACTIVE)
    {
        hdmi_arc_send_upgrade_start_message();
        
        hdmi_arc_data.upgrade_data.status = HDMI_ARC_UPGRADE_STATUS_PREPARED;
    }

    
    return hdmi_arc_data.upgrade_data.upload_load_id;
}

/**-----------------------------------------------------------------------------
 * Upload HDMI Arc upgrade data to the display
 */
UPGRADE_METADATA_UPLOAD_STATUS app_commands_hdmi_arc_upload_data(UPGRADE_METADATA_LOAD_ID load_id, 
        uint8_t* data, 
        size_t data_length,
        UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback)
{
    logger_log_debugf(0, "HDMI ARC app_commands_hdmi_arc_upload_data %d --", data_length);
       
    hdmi_arc_data.processing_stage = HDMI_ARC_PROCESSING_STAGE_IN_UPDATE;

    // Because the HDMI ARC can only accept packets upto 1024 bytes and the Avatar
    // sends packets in 4096 chunks we need to split the messages
    // So...
    // Copy the data into a buffer, send the first 512 bytes
    // in the app_commands_hdmi_arc_process_upgrade_data check how much of the 
    // data has been transferred, transfer the next packet until all has been sent
    // then send the DATA_READY_TO_RECEIVE to the Avatar via the callback
    
    // Copy the data into the buffer
    memcpy(hdmi_arc_data.upgrade_data.data_cache + hdmi_arc_data.upgrade_data.cache_start_pos, data, data_length);
    hdmi_arc_data.upgrade_data.data_cache_length = data_length + hdmi_arc_data.upgrade_data.cache_start_pos;
    hdmi_arc_data.upgrade_data.data_cache_offset = 0;
    hdmi_arc_data.upgrade_data.data_cache_retry_count = 0;
    
    // Reset the cache start pos
    hdmi_arc_data.upgrade_data.cache_start_pos = 0;
    
    hdmi_arc_data.upgrade_data.data_cache_packet_data_size = min(hdmi_arc_data.upgrade_data.data_cache_length, HDMI_ARC_UPLOAD_DATA_MAX_PACKET_SIZE);
    
    // Update the checksum
    CRYPT_MD5_DataAdd(&hdmi_arc_data.upgrade_data.md5_context, (char*)data, data_length);

    // Dont write data yet if this is the first data packet - wait for an Upgrade 
    // Start 0x06 from the ARC Module
    if (((hdmi_arc_data.upgrade_data.status != HDMI_ARC_UPGRADE_STATUS_INACTIVE) && 
            (hdmi_arc_data.upgrade_data.status != HDMI_ARC_UPGRADE_STATUS_PREPARED)) || 
            (hdmi_arc_data.upgrade_data.upload_bytes_written > 0))
    {
        // Send the first packet of data
        hdmi_arc_send_upgrade_data_message(
                hdmi_arc_data.upgrade_data.upload_upload_size,
                hdmi_arc_data.upgrade_data.upload_bytes_written,
                hdmi_arc_data.upgrade_data.data_cache,
                hdmi_arc_data.upgrade_data.data_cache_packet_data_size);
    }
    
    // Update the number of bytes written
    //hdmi_arc_data.upgrade_data.upload_bytes_written += data_length;
    
    logger_log_debugf(0, "HDMI ARC UPLOAD Upload Data %d %d --", hdmi_arc_data.upgrade_data.upload_bytes_written, hdmi_arc_data.upgrade_data.upload_upload_size);
    
    hdmi_arc_data.upgrade_data.local_callback = local_callback;
    
    return UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
}

/**-----------------------------------------------------------------------------
 * Complete the HDMI Arc upgrade
 */
UPGRADE_METADATA_UPLOAD_STATUS app_commands_hdmi_arc_upload_end(UPGRADE_METADATA_LOAD_ID load_id,
        UPGRADE_METADATA_UPLOAD_LOCAL_CALLBACK local_callback, UPGRADE_METADATA_UPGRADE_STATUS status)
{
    logger_log_debugf(0, "AVATAR HDMI ARC UPLOAD Upload End %d %d --", hdmi_arc_data.upgrade_data.upload_bytes_written, hdmi_arc_data.upgrade_data.upload_upload_size);
    
    hdmi_arc_send_upgrade_end_message();
    
    // Reset the version check override.
    hdmi_arc_data.upgrade_data.upload_override_version_check = false;
    
    if (local_callback != NULL)
    {
        local_callback(hdmi_arc_data.upgrade_data.upload_load_id, 
                hdmi_arc_data.upgrade_data.upload_upload_size, 
                hdmi_arc_data.upgrade_data.upload_bytes_written, 
                status);
    }

    hdmi_arc_data.processing_stage = HDMI_ARC_PROCESSING_STAGE_WAIT;
    
    return UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
}
    
/* *****************************************************************************
 End of File
 */
