/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_input.h"
#include <stdio.h>
#include "app_config.h"
#include "app_commands_avatar.h"
#include "utilities/analogue_direct.h"
#include "utilities/display.h"
#include "utilities/input_control.h"
#include "utilities/input_control_defs.h"
#include "utilities/input_control_usb.h"
#include "utilities/input_phono_input.h"
#include "utilities/remote.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//typedef struct
//{
//    // The index of the Input definition that is currently being display for 
//    // analogue direct, and enabled input editing 
//    uint8_t input_edit_index;
//} APP_COMMANDS_INPUT_DATA;
//
//static APP_COMMANDS_INPUT_DATA app_commands_input_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Performs an input change
 */
void app_commands_current_source(STRUCTURED_MESSAGE* message)
{
    message->response = STATUS_UPDATE;
    
    switch (message->source)
    {
        case STRUCTURED_MESSAGE_SOURCE_RS232:
        case STRUCTURED_MESSAGE_SOURCE_TCPIP:
        case STRUCTURED_MESSAGE_SOURCE_UNDEFINED_INPUT:
        {
            // If it is coming via RS232/TCPIP the index needs converting
            uint8_t index = message->data[0];
            int index_idx = -1;
            
            if (index != APP_COMMAND_VALUE_REQUEST_STATUS)
            {
                // Note the NET and AVATAR inputs have the same values (so Paul
                // can show it in his APP), so we are relying on the fact that NET comes before AVATAR 
                // for selecting the NET input.
                int x;
                for (x = 0; x < sizeof(AUDIO_INPUT_VALUES)/sizeof(int); x++)
                {
                    if (AUDIO_INPUT_VALUES[x] == index)
                    {
                        index_idx = x;
                        break;
                    }
                }
                
                if (index_idx > -1)
                {
                    input_control_select_input(index_idx);

                    message->response = STATUS_UPDATE;
                }
                else
                {
                    message->response = PARAM_NOT_RECOGNISED;
                }
            }
            else
            {
                // Return the currently selected input
                uint8_t input = input_control_get_selected_input();

                // Translate from input index to the one required for the message
                uint8_t input_idx = AUDIO_INPUT_VALUES[input];
                        
                message->data[0] = input_idx;
            }
            
            break;
        }
        
        default:
        {
            AUDIO_INPUT input = message->data[0];

            switch (message->action)
            {
                case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
                {
                    // First find the next enabled definition
                    input = input_control_get_next_input(input, true);
                    break;
                }
                
                case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
                {
                    input = input_control_get_next_input(input, false);
                    break;
                }
                
                default:
                {
                }
            }

            input_control_select_input(input);

            // Translate into external Input value
            message->data[0] = AUDIO_INPUT_VALUES[input];

            break;
        }
    }
     
    // Sort out whether were in processor mode
    uint8_t processor = input_control_get_processor_mode_input();    

    int x;
    for (x = 0; x < sizeof(AUDIO_INPUT_VALUES)/sizeof(int); x++)
    {
        if (AUDIO_INPUT_VALUES[x] == message->data[0])
        {
            if (x == processor)
            {
                message->data[0] += 0x10;
            }
            break;
        }
    }

    message->data_length = 1;

    // Send the response
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Draws the Analogue Direct menu when initiated from a button press.
 */
void app_commands_draw_analogue_direct_no_navigation(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();

    // Draw the top line
    // Clear the display
    display_clear(false);
    
    // Draw the top line which is the title
    display_clear_line(DISPLAY_LINE_1);
    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
     
    display_write_line(DISPLAY_LINE_1, false);

    display_clear_line(DISPLAY_LINE_2);
        
    AUDIO_INPUT selected_input = input_control_get_selected_input();
    
    INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(input_control_app_get_edit_index());
        
    if (definition->type == INPUT_CONTROL_TYPE_ANALOGUE)
    {
        display_append_text(DISPLAY_LINE_2, 0, "%s", AUDIO_INPUT_STRINGS[selected_input]);
        
        // Check if there is a value to be displayed
        size_t len = display_get_line_length();
        char value[len];
        
        input_analogue_direct_get_string((char*)&value[0], len);

        // Check whether we're displaying the edited value or not
        // MENU_ITEM* current_item = menu_get_current_item(); 
        if (menu_is_editable_and_original(item))
        {
            display_append_text_right(DISPLAY_LINE_2, "*%s", value);
        }
        else
        {
            display_append_text_right(DISPLAY_LINE_2, "%s", value);
        }
    }
    else
    {
        display_append_text_right(DISPLAY_LINE_2, "Not Available");
    }

    display_write_line(DISPLAY_LINE_2, false);
}

/**-----------------------------------------------------------------------------
 * Sets the maximum volume at switch on
 * Note: This is used for all the analogue direct menus and is dependent on
 * the correct value of app_commands_input_data.analogue_direct_edit_index being
 * set before (usually in the menu enter event)
 */
bool app_commands_edit_analogue_direct(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = 0;
        uint8_t input = 0;
        bool live_update = true;

        // If not being called from a menu then we expect the input and the value
        input = (int8_t)message->data[0];
        value = (int8_t)message->data[1];

        switch (message->source)
        {
            case STRUCTURED_MESSAGE_SOURCE_RS232:
            case STRUCTURED_MESSAGE_SOURCE_TCPIP:
            {
                // If the input is 0xF0 then we use the current input
                if (input == APP_COMMAND_VALUE_REQUEST_STATUS)
                {
                    // Change the message action - working out whether this is a get status is dependent on the length of the message
                    input = input_control_get_selected_input();
                    message->data[0] = AUDIO_INPUT_VALUES[input];
                }
                else
                {
                    // If this is an external message then we just process the values
                    // whatever the menu state
                    int x;
                    for (x = 0; x < sizeof(AUDIO_INPUT_VALUES)/sizeof(int); x++)
                    {
                        if (AUDIO_INPUT_VALUES[x] == input)
                        {
                            input = x;
                            break;
                        }
                    }
                }
                
                if (message->data_length == 1)
                {
                    message->action = STRUCTURED_MESSAGE_ACTION_GET_STATUS;                }
                else
                {
                    message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;                    
                }
                
                break;
            }

            default:
            {
                switch (app_context_get_mode())
                {
                    case APP_CONTEXT_MODE_MENU:
                    case APP_CONTEXT_MODE_MENU_EDIT:
                    {
                        // If were calling this from a menu then we are acting on the input selected in the menu system
                        // which is determined by which menu you are in
                        input = input_control_app_get_edit_index();
                        value = (int8_t)message->data[0];

                        if (menu_get_current_item()->menu_id != APP_COMMAND_MENU_ANALOGUE_DIRECT_NO_NAVIGATION)
                        {
                            live_update = false;
                        }

                        break;
                    }

                    default:
                    {
                        break;
                    }
                }
            }
        }

        message->command = APP_COMMANDS_ANALOGUE_DIRECT;
        message->response = STATUS_UPDATE;

        // Check for dodgy input provided
        if (input < AUDIO_INPUT_MAX)
        {
            INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(input);

            if (definition->type == INPUT_CONTROL_TYPE_ANALOGUE)
            {
                // First check the length of the data - this is a bespoke message len = 1 then a Query, len = 2 then an update
                if (message->action == STRUCTURED_MESSAGE_ACTION_GET_STATUS)
                {
                   // Translate into external Input value
                    message->data[0] = AUDIO_INPUT_VALUES[input];
                    message->data[1] = input_analogue_direct_get(input);//definition->analogue_direct;
                }
                else
                {
                    int8_t current_value = (int8_t)input_analogue_direct_get(input);//definition->analogue_direct;

                    // Set changed state
                    menu_try_initialise_change_state(&current_value, sizeof(int8_t));

                    if (app_commands_process_message_value_bool(message, &current_value, &value, false))
                    {
                        input_control_set_input_analogue_direct(input, current_value);

                        message->data[1] = current_value;

                        // Send on unsolicited message
                        //message->data_length = 2;
                        //message->response = STATUS_UPDATE;
                        //app_commands_send_message(message);

                        if (live_update && (input == input_control_get_selected_input()))
                        {
                            input_control_reset_input();
                        }
                        else
                        {
                            // This is reviewed when the menu is exited .
                        }

                        menu_update_change_state(&current_value, sizeof(int8_t));
                    }
                }
            }
            else
            {
                message->response = COMMAND_INVALID_NOW;
            }
        }
        else
        {
            message->response = PARAM_NOT_RECOGNISED;
        }
    }    

    message->data_length = 2;
    app_commands_send_message(message);
     
    return result;
}

/**-----------------------------------------------------------------------------
 * Called when one of the Dirac Curve menus is exited.  Used to select the curve
 * @param message
 * @return 
 */
MENU_ID app_commands_edit_exit_dirac_curve(MENU_ID menu)
{
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;
    
    INPUT_CONTROL_DEFINITION* definition = input_control_get_current_definition();
    logger_log_debugf(0, "DIRAC app_commands_edit_exit_dirac_curve %d %d --", input_control_app_get_edit_index(), definition->index);
    if (input_control_app_get_edit_index() == definition->index)
    {
        // Added check for headphones connected
        if (!output_control_headphones_connected())
        {
            avatar_set_dirac_curve(dirac_get_curve_for_input(definition->index));
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to change the logger level
 */
bool app_commands_edit_filter(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = message->data[0];
        int8_t current_value = (int8_t)input_control_get_filter();
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        message->command = APP_COMMANDS_SET_FILTER;
        message->response = STATUS_UPDATE;
        
        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                current_value = value;
                
                if (value < input_control_get_num_filters())
                {
                    input_control_set_filter((FILTER_TYPE)current_value);
                }
                else
                {
                    message->response = PARAM_NOT_RECOGNISED;
                }
                
                message->data[0] = (uint8_t)current_value;
                break;
            }
            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            {
                current_value++;
                if (current_value >= input_control_get_num_filters())
                {
                    current_value = (FILTER_TYPE)0;
                }

                input_control_set_filter((FILTER_TYPE)current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                current_value--;
                if (current_value < 0)
                {
                    current_value = (FILTER_TYPE)((uint8_t)input_control_get_num_filters() - 1);
                }

                input_control_set_filter((FILTER_TYPE)current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
            {
                message->data[0] = (uint8_t)current_value;
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                break;
            }
        }

        message->data_length = 1;
        app_commands_send_message(message);
                
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            // Store Filter Type setting (Note: Only stored here, and Test type 
            // never stored)
            FILTER_TYPE filter_type = input_control_get_filter();
            settings_set(SETTING_FILTER_TYPE, &filter_type);
            
            menu_update_change_state(&current_value, sizeof(int8_t));
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Used for editing input enabled settings on individual channels
 * Note: This is used for all the analogue direct menus and is dependent on
 * the correct value of app_commands_input_data.input_edit_index being
 * set before (usually in the menu enter event)
 */
bool app_commands_edit_input_enabled(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    logger_log_debugf(0, "INPUT app_commands_edit_input_enabled %d ", input_control_app_get_edit_index());
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(input_control_app_get_edit_index());

        int8_t value = (int8_t)message->data[0];
        int8_t current_value = (int8_t)definition->enabled;
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        if (app_commands_process_message_value_bool(message, &current_value, &value, true))
        {
            input_control_set_input_enabled(input_control_app_get_edit_index(), current_value);

            menu_update_change_state(&current_value, sizeof(int8_t));
        }
    }    
     
    return result;
}

/**-----------------------------------------------------------------------------
 * Sets the maximum volume at switch on
 */
bool app_commands_edit_mm_mc(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    logger_log_debugf(0, "PHONO INPUT --");
    logger_process();
    app_commands_review_message_action(message);

    int16_t value = (int16_t)message->data[0];
    PHONO_INPUT current_value = input_phono_input_get();

    // Set changed state
    if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        menu_try_initialise_change_state(&current_value, sizeof(PHONO_INPUT));
    }

    app_commands_process_message_value_bool(message, &current_value, &value, true);

    message->command = APP_COMMANDS_PHONO_INPUT_MM_MC;
    message->response = STATUS_UPDATE;

    if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        input_phono_input_set(current_value);
 
        menu_update_change_state(&current_value, sizeof(PHONO_INPUT));

        message->data[0] = (uint8_t)current_value;
 
        message->data_length = 1;
    }
        
    // Send the response
    app_commands_send_message(message);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Used to change the USB Record setting
 */
bool app_commands_edit_usb_record(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = (int8_t)message->data[0];
        int8_t current_value = (int8_t)input_control_usb_get_record();
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        if (app_commands_process_message_value_bool(message, &current_value, &value, true))
        {
            input_control_usb_set_record(current_value);

            menu_update_change_state(&current_value, sizeof(int8_t));
        }
    }    
     
    return result;
}
//
///**-----------------------------------------------------------------------------
// * Called when entering the Analogue Direct no navigation menu - setup the timeout
// */
//void app_commands_enter_analogue_direct_no_navigation(MENU_ID menu)
//{
//    // Set the edit index to the current input
//    input_control_app_set_edit_index(input_control_get_selected_input());
//    
//    // Ensure the Analogue Direct menu is only visible for a while
//    timers_update_timeout(TIMER_INDEX_MENU_INACTIVE, TIMER_MENU_ANALOGUE_DIRECT_INACTIVE_TIMEOUT_MS, TIMERS_RESET_IMMEDIATE);
//}

///**-----------------------------------------------------------------------------
// * Called when leaving the Balance menu
// * @param menu
// */
//MENU_ID app_commands_exit_analogue_direct_no_navigation(MENU_ID menu)
//{
//    // Check to see if analogue direct needs applying to active input
//    MENU_ID result = app_commands_check_analogue_direct(*menu_get_item(menu));
//    
//    // Make sure Inactive Menu timeout is reset
//    app_commands_menu_reset_inactivity_timeout(menu);
//    
//    return result;
//}

/**-----------------------------------------------------------------------------
 * Gets the analogue direct setting for the current channel as a string
 * Note: This is used for all the analogue direct menus and is dependent on
 * the correct value of app_commands_input_data.analogue_direct_edit_index being
 * set before (usually in the menu enter event)
  * @return 
// */
//void app_commands_get_analogue_direct_string(char* buffer, size_t size)
//{
//    INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(input_control_app_get_edit_index());
//    
//    if (definition != NULL)
//    {
//        sprintf(buffer, "%s", APP_COMMAND_OFF_ON_STRINGS[(int)definition->analogue_direct]);    
//    }
//}

/**-----------------------------------------------------------------------------
 * 
 * @param message
 * @return 
 */
void app_commands_get_input_detection_status(STRUCTURED_MESSAGE* message)
{
    message->command = APP_COMMANDS_INPUT_DETECT;
    
    if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
    {
        message->data[0] = input_control_audio_input_present();
        message->data_length = 1;
        
        message->response = STATUS_UPDATE;
    }
    else
    {
        message->response = COMMAND_INVALID_NOW;
    }
    
    app_commands_send_message(message);
}

///**-----------------------------------------------------------------------------
// * Gets the index of the Audio Input currently being edited.
// * @return 
// */
//AUDIO_INPUT app_commands_get_input_edit_index()
//{
//    return app_commands_input_data.input_edit_index;
//}

/**-----------------------------------------------------------------------------
 * Gets the input enabled setting for the current channel as a string
 * Note: This is used for all the analogue direct menus and is dependent on
 * the correct value of app_commands_input_data.analogue_direct_edit_index being
 * set before (usually in the menu enter event)
 * NOTE - we may need to set analogue_direct_edit_index when we change channels if this is used anywhere other than in the menus
 * @return 
 */
void app_commands_get_input_enabled_string(char* buffer, size_t size)
{
    INPUT_CONTROL_DEFINITION* definition = input_control_get_definition(input_control_app_get_edit_index());
    
    if (definition != NULL)
    {
        sprintf(buffer, "%s", APP_COMMAND_OFF_ON_STRINGS[(int)definition->enabled]);    
    }
    else
    {
        logger_log_debugf(0, "INPUT app_commands_get_input_enabled_string CANT FIND %d ", input_control_app_get_edit_index());
        sprintf(buffer, "%s", " ");    
    }
}

/**-----------------------------------------------------------------------------
 * Gets the max on volume as a string
 * @return 
 */
//void app_commands_get_mm_mc_string(char* buffer, size_t size)
//{
//    sprintf(buffer, "%s", PHONO_INPUT_STRINGS[input_control_get_phono_input()]);
//}

/**----------------------------------------------------------------------------- 
 * Used to return the sample rate.  This is invoked by an RS232/TCPIP call and will currently
 * only handle the GET_STATUS message
 */
void app_commands_get_sample_rate(STRUCTURED_MESSAGE* message)
{
    message->command = APP_COMMANDS_SAMPLE_RATE;
    
    if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
    {
        if (input_control_get_type() == INPUT_CONTROL_TYPE_DIGITAL)
        {
            SAMPLE_RATE sample_rate = input_control_get_sample_rate();

            // We need to map between the internal sample rate representation and the 
            // version defined in the message
            if (sample_rate != SAMPLE_RATE_NOTSUPPORTED)
            {
                message->data[0] = SAMPLE_RATE_GETSTATUS_VALUES[sample_rate];
                message->data_length = 1;
                message->response = STATUS_UPDATE;
            }
            else
            {
                message->response = COMMAND_INVALID_NOW;
            }
        }
        else
        {
            message->response = COMMAND_INVALID_NOW;
        }
    }
    else
    {
        message->response = PARAM_NOT_RECOGNISED;
    }    
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the USB Record setting
 * @return 
 */
void app_commands_get_usb_record_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", APP_COMMAND_OFF_ON_STRINGS[(int)input_control_usb_get_record()]);
}

/**-----------------------------------------------------------------------------
 * Configures the input control module
 * Note: See input_control_process() for the register configuration settings
 */
void app_commands_input_configure()
{
    logger_log_debugf(0, "INPUT Configure");
    // Note I2S switch select defined in input_control.c
    
    PHONO_INPUT phono_input;
    settings_get(SETTING_MM_MC, &phono_input);
    input_phono_input_set(phono_input);
    
    uint8_t input_enabled_settings[INPUT_CONTROL_NUM_SETTINGS];
    
    settings_get(SETTING_INPUT_ENABLED, input_enabled_settings);
 
    INPUT_CONTROL_DEFINITION def;
    def.index = AUDIO_INPUT_PHONO;
    def.type = INPUT_CONTROL_TYPE_ANALOGUE;
    def.value = 0;
    //def.analogue_direct = input_analogue_direct_get(AUDIO_INPUT_PHONO);
    def.enabled = input_enabled_settings[AUDIO_INPUT_PHONO];
    input_control_add(&def);
    
    def.index = AUDIO_INPUT_AUX;
    //def.analogue_direct = input_analogue_direct_get(AUDIO_INPUT_AUX);
    def.enabled = input_enabled_settings[AUDIO_INPUT_AUX];
    def.value = 1;
    input_control_add(&def);

    def.index = AUDIO_INPUT_CD;
    //def.analogue_direct = input_analogue_direct_get(AUDIO_INPUT_CD);
    def.enabled = input_enabled_settings[AUDIO_INPUT_CD];
    def.value = 3;
    input_control_add(&def);
    
    def.index = AUDIO_INPUT_PVR;
    //def.analogue_direct = input_analogue_direct_get(AUDIO_INPUT_PVR);
    def.enabled = input_enabled_settings[AUDIO_INPUT_PVR];
    def.value = 2;
    input_control_add(&def);
    
    def.index = AUDIO_INPUT_STB;
    //def.analogue_direct = input_analogue_direct_get(AUDIO_INPUT_STB);
    def.enabled = input_enabled_settings[AUDIO_INPUT_STB];
    def.value = 4;
    input_control_add(&def);

    def.type = INPUT_CONTROL_TYPE_DIGITAL;
    //def.analogue_direct = false;
    
    def.index = AUDIO_INPUT_AV;
    def.enabled = input_enabled_settings[AUDIO_INPUT_AV];
    def.value = 0X03;
    input_control_add(&def);
    
    def.index = AUDIO_INPUT_BD;
    def.enabled = input_enabled_settings[AUDIO_INPUT_BD];
    def.value = 0X02;
    input_control_add(&def);
    
    def.index = AUDIO_INPUT_SAT;
    def.enabled = input_enabled_settings[AUDIO_INPUT_SAT];
    def.value = 0X00;
    input_control_add(&def);
    
    def.index = AUDIO_INPUT_GAME;
    def.enabled = input_enabled_settings[AUDIO_INPUT_GAME];
    def.value = 0X01;
    input_control_add(&def);
    
    def.index = AUDIO_INPUT_HDMI_ARC;
    def.enabled = true;        // Note: This is never enabled as the channel is not selectable
    def.value = 0X04;           
    input_control_add(&def);    
    
    def.index = AUDIO_INPUT_DIGITAL_MUTE;   // Used for when device is in Analogue mode to silence spdif rx
    def.enabled = false;        // Note: This is never enabled as the channel is not selectable
    def.value = 0X05;           
    input_control_add(&def);    
    
    def.type = INPUT_CONTROL_TYPE_AVATAR;
    def.index = AUDIO_INPUT_NET;
    def.enabled = input_enabled_settings[AUDIO_INPUT_NET];
    def.value = 0X00; 
    input_control_add(&def);
    
    def.index = AUDIO_INPUT_AVATAR;
    def.enabled = false;
    def.value = 0X00; 
    input_control_add(&def);
    
    PHONO_INPUT type;
    settings_get(SETTING_MM_MC, &type);
    input_phono_input_set(type);
}

/**-----------------------------------------------------------------------------
 * Callback for when inputs are switching
 */
void app_commands_input_switch_callback(INPUT_CONTROL_SWITCH_STATE state)
{
    switch (state)
    {
        case INPUT_CONTROL_SWITCH_STATE_PRE:
            break;
            
        case INPUT_CONTROL_SWITCH_STATE_POST:
        {
            // Send unsolicited messages re input state
            app_commands_request_status(APP_COMMANDS_CURRENT_SOURCE, app_commands_current_source);
            app_commands_request_status(APP_COMMANDS_ROOM_EQUALISATION, app_commands_edit_dirac_selected);
            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Performs a of an preview input change
 * This is to allow us to display the new channel without actually doing the switch
 * Having said that once the INPUT_PREVIEW timeout expires it will switch to the 
 * selected preview channel
 */
void app_commands_preview_source(STRUCTURED_MESSAGE* message)
{
    bool needs_response = true;
    message->response = STATUS_UPDATE;
    
    switch (message->source)
    {
        case STRUCTURED_MESSAGE_SOURCE_RS232:
        case STRUCTURED_MESSAGE_SOURCE_TCPIP:
        {
            // If we get this message from an external source (which we shouldn't))
            // Treat it as if its a change source message
            app_commands_current_source(message);
            needs_response = false;
            break;
        }
        
        default:
        {
            INPUT_CONTROL_PREVIEW_INFO* preview_info = input_control_get_preview_info();
            AUDIO_INPUT input;
            if (preview_info->active)
            {
                // If encoder was turned while preview is active
                // increment the preview, rather than the current selection
                input = preview_info->input;
            }
            else
            {
                // If the preview has expired, but the switch to the previewed input
                // hasn't yet completed, the selected input is the same as the preview input

                input = input_control_get_selected_input();
            }

            switch (message->action)
            {
                case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
                {
                    // First find the next enabled definition
                    input = input_control_get_next_input(input, true);
                    break;
                }
                
                case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
                {
                    input = input_control_get_next_input(input, false);
                    break;
                }
                
                default:
                {
                }
            }
            
            input_control_set_preview(input);

            break;
        }
    }
}

/**-----------------------------------------------------------------------------
 * Prepares editing of the analog direct and input enabled setting for an input 
 * based on the current menu
 * @param menu
 */
void app_commands_set_input_for_edit(MENU_ITEM menu)
{
    switch (menu.menu_id)
    {
        case APP_COMMAND_MENU_ANALOGUE_DIRECT_PHONO:
        case APP_COMMAND_MENU_INPUT_ENABLED_PHONO:
        case APP_COMMAND_MENU_DIRAC_CURVE_PHONO:
        case APP_COMMAND_MENU_LIP_SYNC_PHONO:
            input_control_app_set_edit_index(AUDIO_INPUT_PHONO);
            break;
            
        case APP_COMMAND_MENU_ANALOGUE_DIRECT_AUX:
        case APP_COMMAND_MENU_INPUT_ENABLED_AUX:
        case APP_COMMAND_MENU_DIRAC_CURVE_AUX:
        case APP_COMMAND_MENU_LIP_SYNC_AUX:
            input_control_app_set_edit_index(AUDIO_INPUT_AUX);
            break;
            
        case APP_COMMAND_MENU_ANALOGUE_DIRECT_CD:
        case APP_COMMAND_MENU_INPUT_ENABLED_CD:
        case APP_COMMAND_MENU_DIRAC_CURVE_CD:
        case APP_COMMAND_MENU_LIP_SYNC_CD:
            input_control_app_set_edit_index(AUDIO_INPUT_CD);
            break;
            
        case APP_COMMAND_MENU_ANALOGUE_DIRECT_PVR:
        case APP_COMMAND_MENU_INPUT_ENABLED_PVR:
        case APP_COMMAND_MENU_DIRAC_CURVE_PVR:
        case APP_COMMAND_MENU_LIP_SYNC_PVR:
            input_control_app_set_edit_index(AUDIO_INPUT_PVR);
            break;
            
        case APP_COMMAND_MENU_ANALOGUE_DIRECT_STB:
        case APP_COMMAND_MENU_INPUT_ENABLED_STB:
        case APP_COMMAND_MENU_DIRAC_CURVE_STB:
        case APP_COMMAND_MENU_LIP_SYNC_STB:
            input_control_app_set_edit_index(AUDIO_INPUT_STB);
            break;
            
        case APP_COMMAND_MENU_INPUT_ENABLED_AV:
        case APP_COMMAND_MENU_DIRAC_CURVE_AV:
        case APP_COMMAND_MENU_LIP_SYNC_AV:
            input_control_app_set_edit_index(AUDIO_INPUT_AV);
            break;
            
        case APP_COMMAND_MENU_INPUT_ENABLED_BD:
        case APP_COMMAND_MENU_DIRAC_CURVE_BD:
        case APP_COMMAND_MENU_LIP_SYNC_BD:
            input_control_app_set_edit_index(AUDIO_INPUT_BD);
            break;
            
        case APP_COMMAND_MENU_INPUT_ENABLED_SAT:
        case APP_COMMAND_MENU_DIRAC_CURVE_SAT:
        case APP_COMMAND_MENU_LIP_SYNC_SAT:
            input_control_app_set_edit_index(AUDIO_INPUT_SAT);
            break;
            
        case APP_COMMAND_MENU_INPUT_ENABLED_GAME:
        case APP_COMMAND_MENU_DIRAC_CURVE_GAME:
        case APP_COMMAND_MENU_LIP_SYNC_GAME:
            input_control_app_set_edit_index(AUDIO_INPUT_GAME);
            break;
            
//        case APP_COMMAND_MENU_INPUT_ENABLED_PCUSB:
//        case APP_COMMAND_MENU_DIRAC_CURVE_PCUSB:
//            app_commands_input_data.input_edit_index = AUDIO_INPUT_PCUSB;
//            break;
            
        case APP_COMMAND_MENU_INPUT_ENABLED_NET:
        case APP_COMMAND_MENU_DIRAC_CURVE_NET:
        case APP_COMMAND_MENU_LIP_SYNC_NET:
            input_control_app_set_edit_index(AUDIO_INPUT_NET);
            break;
            
//        case APP_COMMAND_MENU_INPUT_ENABLED_USB:
//        case APP_COMMAND_MENU_DIRAC_CURVE_USB:
//            app_commands_input_data.input_edit_index = AUDIO_INPUT_USB;
//            break;
            
        case APP_COMMAND_MENU_INPUT_ENABLED_ARC:
        case APP_COMMAND_MENU_LIP_SYNC_ARC:
            input_control_app_set_edit_index(AUDIO_INPUT_HDMI_ARC);
            break;
            
        default:
            input_control_app_set_edit_index(input_control_get_selected_input()); 
            break;
    }
    
    //logger_log_debugf(0, "app_commands_set_input_for_edit %d %d --", menu.menu_id, app_commands_input_data.input_edit_index);
}

/**-----------------------------------------------------------------------------
 * Simulate remote control commands via RS232
 * @param 
 */
void app_commands_simulate_remote(STRUCTURED_MESSAGE* message)
{
    if (remote_get_system_code() == message->data[0])
    {
        STRUCTURED_MESSAGE remote_message;
        structured_message_copy(message, &remote_message);
        remote_message.source = STRUCTURED_MESSAGE_SOURCE_REMOTE_RAW;
        remote_message.zone = STRUCTURED_MESSAGE_ZONE_1;
        remote_message.command = message->data[1];
        remote_message.response = STATUS_UPDATE;
        //remote_message.data_length = 0;
        remote_message.repeat = false;

        // Pass the message onto a callback method for validation and processing
        input_context_set_remote_message(&remote_message);
        
        message->response = STATUS_UPDATE;
    }
    else
    {
        message->response = COMMAND_INVALID_NOW;
    }
    
    app_commands_send_message(message);
}

/* *****************************************************************************
 End of File
 */
