/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_lip_sync.h"
#include "app_config.h"
#include "app_commands_input.h"
#include "utilities/avatar_control.h"
#include "utilities/menu_app.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    // True is lip_syc_process need to check whether to send data
    bool needs_check;
} LIP_SYNC_DATA;

static LIP_SYNC_DATA lip_sync_data = {
    .needs_check = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used in menus to review whether a change in lip sync for a channel needs 
 * the actual lip sync value setting on the avatar or not.
 * @param menu
 * @return 
 */
MENU_ID app_commands_check_lip_sync(MENU_ITEM menu)
{
    MENU_ID result = MENU_USE_SUPPLIED_MENU_ID;

    uint8_t idx = input_control_app_get_edit_index();
        
    // Check we're on the current input
    INPUT_CONTROL_DEFINITION* definition = input_control_get_current_definition();
    //logger_log_debugf(0, "LIP SYNC - Change %d %d %d --", menu.menu_id, idx, definition->index);
    if (idx == definition->index)
    {
        uint8_t lip_sync_settings[INPUT_CONTROL_NUM_SETTINGS];
        settings_get(SETTING_LIP_SYNC, lip_sync_settings);

        uint8_t lip_sync_value_idx = lip_sync_settings[idx];

        logger_log_debugf(0, "LIP SYNC - Sending %d %d --", lip_sync_value_idx, LIP_SYNC_VALUES[lip_sync_value_idx]);

        avatar_send_lip_sync_delay(LIP_SYNC_VALUES[lip_sync_value_idx]);
    }
             
    // reset the menu timeout...
    menu_app_reset_inactivity_timeout(menu.menu_id);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Prepares editing of the lip_sync setting for an input 
 * based on the current menu
 * @param menu
 */
void app_commands_enter_lip_sync_menu(MENU_ITEM menu)
{
    // Reset for lip_sync menus
    timers_reset(TIMER_INDEX_LIP_SYNC_SEND);
    
    app_commands_set_input_for_edit(menu);
}

/**-----------------------------------------------------------------------------
 * gets a string representing the Lip Sync setting for the currently selected 
 * audio channel
 * @return 
 */
void app_commands_get_lip_sync_string(char* buffer, size_t size)
{
    uint8_t idx = input_control_app_get_edit_index();
    if (idx < INPUT_CONTROL_NUM_SETTINGS)
    {
        uint8_t lip_sync_settings[INPUT_CONTROL_NUM_SETTINGS];
        settings_get(SETTING_LIP_SYNC, lip_sync_settings);
        
        uint8_t lip_sync_value_idx = lip_sync_settings[idx];

        sprintf(buffer, "%s", LIP_SYNC_STRINGS[lip_sync_value_idx]);    
    }
    else
    {
        sprintf(buffer, "Undefined");    
    }
 }

/**-----------------------------------------------------------------------------
 * Used for editing lip sync settings on individual channels
 * Note: This is used for all the analogue direct menus and is dependent on
 * the correct value of input_control_app_data.input_edit_index being
 * set before (usually in the menu enter event)
 */
bool app_commands_edit_lip_sync(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    logger_log_debugf(0, "LIP SYNC app_commands_edit_lip_sync %d ", input_control_app_get_edit_index());

    int16_t value = (int16_t)message->data[0];

    // Note : If this is changed to handle RS232 message the  look at app_commands_edit_auto_shutdown_timeout
     
    app_commands_review_message_action(message);

    uint8_t idx = input_control_app_get_edit_index();
    if (idx < INPUT_CONTROL_NUM_SETTINGS)
    {
        uint8_t lip_sync_settings[INPUT_CONTROL_NUM_SETTINGS];
        settings_get(SETTING_LIP_SYNC, lip_sync_settings);

        uint8_t lip_sync_value_idx = lip_sync_settings[idx];

        uint16_t current_value = lip_sync_value_idx;

        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(uint16_t));
        }

        logger_log_debugf(0, "LIP SYNC %d %d %d %d --", value, current_value, message->action, idx);
        // 
        app_commands_process_message_value_array(message, &current_value, &value, LIP_SYNC_IDX, LIP_SYNC_OPTIONS_MAX);

        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            lip_sync_settings[idx] = current_value;
            settings_set(SETTING_LIP_SYNC, lip_sync_settings);

            menu_update_change_state(&current_value, sizeof(uint16_t));

            message->data[0] = (uint8_t) current_value;
            
            INPUT_CONTROL_DEFINITION* definition = input_control_get_current_definition();
            //logger_log_debugf(0, "LIP SYNC - Change %d %d %d --", menu.menu_id, idx, definition->index);
            if (idx == definition->index)
            {
                if (timers_has_expired(TIMER_INDEX_LIP_SYNC_SEND))
                {
                    // Send the Avatar a message if we're on the current input
                    logger_log_debugf(0, "LIP SYNC - Sending %d %d --", lip_sync_value_idx, LIP_SYNC_VALUES[lip_sync_value_idx]);

                    avatar_send_lip_sync_delay(LIP_SYNC_VALUES[current_value]);

                    timers_reset(TIMER_INDEX_LIP_SYNC_SEND);
                }
                    
                lip_sync_data.needs_check = true;
            }
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Makes sure that the Avatar Lip Sync message is up-to-date
 */
void app_commands_process_lip_sync()
{
    if (lip_sync_data.needs_check)
    {
        if (app_context_get_mode() == APP_CONTEXT_MODE_MENU_EDIT)
        {
            // Assumes that the lip sync menus follow the main menu in the declaration, and that there are as many menus as inputs
            if ((menu_get_current_item()->menu_id > APP_COMMAND_MENU_LIP_SYNC) && 
                (menu_get_current_item()->menu_id < APP_COMMAND_MENU_LIP_SYNC + AUDIO_INPUT_MAX))
            {
                uint8_t idx = input_control_app_get_edit_index();
                INPUT_CONTROL_DEFINITION* definition = input_control_get_current_definition();
                //logger_log_debugf(0, "LIP SYNC - Change %d %d %d --", menu.menu_id, idx, definition->index);
                if (idx == definition->index)
                {
                    if (timers_has_expired(TIMER_INDEX_LIP_SYNC_SEND))
                    {
                        uint8_t lip_sync_settings[INPUT_CONTROL_NUM_SETTINGS];
                        settings_get(SETTING_LIP_SYNC, lip_sync_settings);

                        uint8_t lip_sync_value_idx = lip_sync_settings[idx];
                        lip_sync_value_idx = lip_sync_value_idx < LIP_SYNC_OPTIONS_MAX ? lip_sync_value_idx : LIP_SYNC_OPTION_0MS; // See Release 1.35

                        logger_log_debugf(0, "LIP SYNC - Sending Process %d %d --", lip_sync_value_idx, LIP_SYNC_VALUES[lip_sync_value_idx]);

                        avatar_send_lip_sync_delay(LIP_SYNC_VALUES[lip_sync_value_idx]);

                        timers_reset(TIMER_INDEX_LIP_SYNC_SEND);
                        
                        lip_sync_data.needs_check = false;
                    }
                }
            }
        }
    }
}
    
/* *****************************************************************************
 End of File
 */
