/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_power.h"
#include "app_commands_auto_shutdown.h"
#include "app_config.h"
#include "app_context.h"
#include "utilities/power.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
     //! Set to true if the reason that STANDBY has been entered was due to an AC NOT PRESECT
    bool ac_not_present_triggered_powerdown;
    
    //! Set to true when either powering down or powering up
    bool in_transition;
} POWER_DATA;

static POWER_DATA power_data =
{
    .ac_not_present_triggered_powerdown = false,
    .in_transition = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Checks the power state during normal processing
 */
void app_commands_check_power_state()
{
    // Check Auto shutdown state
    bool powerdown = app_commands_check_auto_shutdown();
    
    // Check the power state - Note we don't do this if the reset process
    // is underway
    if (app_context_get_mode() != APP_CONTEXT_MODE_PROCESS)
    {
        // Check power state and move to POWERED OFF
        if (powerdown) //2019-03-12 At this stage we are no longer in a state of being powered down power_get_state() == POWER_STATE_OFF)
        {
            logger_log_info(LOGGER_CODE_NO_AC_DETECT, "POWER Power state OFF detected --");

            display_refresh();

            timers_update_timeout(TIMER_INDEX_POWER_CYCLE, 1000, TIMERS_TIMEOUT_RESET_IMMEDIATE);

            app_context_set_app_state(APP_STATE_POWERDOWN);
        }
    }

    // Check AC status
    app_commands_set_ac_not_present_triggered_powerdown(false);
    if (!power_get_ac_present())
    {
        // Mute speakers/preamp and headphones
        output_control_mute_outputs(true);

        logger_log_info(LOGGER_CODE_NO_AC_DETECT, "POWER No AC Present --");

        app_context_set_app_state(APP_STATE_POWERDOWN);

        app_commands_set_ac_not_present_triggered_powerdown(true);
    }
}
 
/**-----------------------------------------------------------------------------
 * Used to notify the app that we have stopped moving between Power states
 */
void app_commands_clear_power_state_transition()
{
    power_data.in_transition = false;
}
 
/**-----------------------------------------------------------------------------
 * Gets the state of the ac_not_present_triggered_powerdown flag
 */
bool app_commands_get_ac_not_present_triggered_powerdown()
{
    return power_data.ac_not_present_triggered_powerdown;
}
 
/**
 * Used to check whether we are currently moving between Power states
 */
bool app_commands_in_power_state_transition()
{
    return power_data.in_transition;
}

/**-----------------------------------------------------------------------------
 * Switches power on/off
 */
void app_commands_power(STRUCTURED_MESSAGE* message)
{
    APP_COMMAND_POWER action = message->data[0];
    
    //Check that we are not currently powering down or powering up
    if (!app_commands_in_power_state_transition())
    {
        // Handle toggle action
        if (action == APP_COMMAND_POWER_TOGGLE)
        {
            switch (power_get_state())
            {
                case POWER_STATE_ON:
                    action = APP_COMMAND_POWER_OFF;
                    break;

                case POWER_STATE_OFF:
                    action = APP_COMMAND_POWER_ON;
                    break;
            }

            message->data[0] = action;
        }

        message->command = APP_COMMANDS_POWER;
        message->response = STATUS_UPDATE;

        switch (action)
        {
            case APP_COMMAND_POWER_ON:
            {
                if (app_context_get_app_state() == APP_STATE_POWEREDDOWN)
                {
                    logger_log_info(LOGGER_CODE_POWER, "Powering Up --");

                    app_context_set_app_state(APP_STATE_WAKEUP);
                }

                break;
            }

            case APP_COMMAND_POWER_OFF:
            {
                if (app_context_get_app_state() == APP_STATE_RUNNING)
                {
                    app_context_set_app_state(APP_STATE_POWERDOWN);
                }
                break;
            }

            case APP_COMMAND_POWER_REQUEST_STATE:
            {
                // put the state into the message
                // Note this reuses the message and alters it
                message->data[0] = power_get_state();
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                break;
            }
        }
    }
    else
    {
        message->response = COMMAND_INVALID_NOW;
    }

    // Just incase its not been defined
    message->data_length = 1;
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Handles the power down operation
 */
void app_commands_power_down()
{
    if (power_get_state() != POWER_STATE_OFF)
    {
        logger_log_info(LOGGER_CODE_POWER, "Powering Down --");

        // Mute speakers/preamp and headphones
        output_control_mute_outputs(true);

        // check whether we need to save the current state of the settings
        settings_process();

        // Disable the power supplies
        power_disable(); 

        app_commands_check_display();

        // Check we're not in reboot mode
        if (app_context_get_mode() != APP_CONTEXT_MODE_PROCESS)
        {
            app_context_set_mode(APP_CONTEXT_MODE_STANDBY);
        }

        logger_log_info(LOGGER_CODE_POWER, "Powered Down --");
    }
}

/**-----------------------------------------------------------------------------
 * Sets the state of the ac_not_present_triggered_powerdown flag
 * @param on
 */
void app_commands_set_ac_not_present_triggered_powerdown(bool on)
{
    power_data.ac_not_present_triggered_powerdown = on;
}
 
/**
 * Used to notify the app that we have started moving between Power states
 */
void app_commands_set_power_state_transition()
{
    power_data.in_transition = true;
}

/* *****************************************************************************
 End of File
 */
