/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
 Handles procesor mode functions

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_processor_mode.h"
#include <stdint.h>
#include <stdio.h>
#include "app_context.h"
#include "utilities/app_commands_utilities.h"
#include "utilities//menu.h"
#include "utilities/volume.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Wrapper for saetting processor mode input
 * @param input
 */
static void app_commands_set_processor_mode_input(int16_t input)
{
    input_control_set_processor_mode_input(input);

    if (input_control_in_processor_mode())
    {
        volume_set(input_control_get_processor_mode_volume(), false);
        volume_set_balance(VOLUME_DEFAULT_BALANCE, false);
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Performs a processor mode input change based upon a delta value
 * Note: The input index is the position in the APP_AUDIO_INPUT_ enum whereas 
 * the output is the corresponding value from the AUDIO_INPUT_VALUES array
 * THESE ARE DIFFERENT!
 */
bool app_commands_edit_processor_mode_input(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int16_t value = (int16_t)message->data[0];
        int16_t current_value = (int16_t)input_control_get_processor_mode_input();
        
        //logger_log_debugf(234, "app_commands_edit_processor_mode_input %d %d --", value, current_value);
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int16_t));
        }
    
        message->command = APP_COMMANDS_PROCESSOR_MODE_INPUT;
        message->response = STATUS_UPDATE;
        
        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                if (value != AUDIO_INPUT_PHONO)
                {
                    current_value = value;
                    app_commands_set_processor_mode_input(current_value);

                    message->data[0] = (uint8_t)current_value;
                }
                else
                {
                    message->response = PARAM_NOT_RECOGNISED;
                }
                break;
            }
            
            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            {
                current_value++;
                
                // It does not make sense to include PHONO as an option here
                while ((current_value == AUDIO_INPUT_PHONO) ||
                       //(current_value == AUDIO_INPUT_PCUSB) ||
                       (current_value == AUDIO_INPUT_AVATAR) ||
                       (current_value == AUDIO_INPUT_NET) ||
                       //(current_value == AUDIO_INPUT_USB) ||
                       (current_value == AUDIO_INPUT_HDMI_ARC) ||
                       (current_value == AUDIO_INPUT_DIGITAL_MUTE))
                {
                    current_value++;
                }
                
                // Note: when using input_control_get_num_inputs as an index we 
                // take into account that the None options is there but is not 
                // added when initialising the module hence we compare to the 
                // actual number rather than the number -1
                if (current_value > input_control_get_num_inputs())
                {
                    current_value = 0;
                }

                app_commands_set_processor_mode_input(current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }
            
            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                current_value--;
                // It does not make sense to include PHONO as an option here
                while ((current_value == AUDIO_INPUT_PHONO) ||
                       //(current_value == AUDIO_INPUT_PCUSB) ||
                       (current_value == AUDIO_INPUT_AVATAR) ||
                       (current_value == AUDIO_INPUT_NET) ||
                       //(current_value == AUDIO_INPUT_USB) ||
                       (current_value == AUDIO_INPUT_HDMI_ARC) ||
                       (current_value == AUDIO_INPUT_DIGITAL_MUTE))
                {
                    current_value--;
                }
                
                // Note: when using input_control_get_num_inputs as an index we 
                // take into account that the None options is there but is not 
                // added when initialising the module hence we compare to the 
                // actual number rather than the number -1
                if (current_value < 0)
                {
                    current_value = input_control_get_num_inputs(); //- 1;
                    while ((current_value == AUDIO_INPUT_PHONO) ||
                            //(current_value == AUDIO_INPUT_PCUSB) ||
                            (current_value == AUDIO_INPUT_AVATAR) ||
                            (current_value == AUDIO_INPUT_NET) ||
                            //(current_value == AUDIO_INPUT_USB) ||
                            (current_value == AUDIO_INPUT_HDMI_ARC) ||
                            (current_value == AUDIO_INPUT_DIGITAL_MUTE))
                     {
                         current_value--;
                     }
                }

                app_commands_set_processor_mode_input(current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
            {
                message->data[0] = (uint8_t)current_value;
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                break;
            }
        }

        // Convert the input code to the external value rather then the internal
        message->data[0] = AUDIO_INPUT_VALUES[message->data[0]];

        // Just incase its not been defined
        message->data_length = 1;
    
        app_commands_send_message(message);
                
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_update_change_state(&current_value, sizeof(int16_t));
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * The version used for processing messages rather than menus - includes translation between the different
 * AUDIO_INPUT representations
 */
bool app_commands_edit_processor_mode_input_command(STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    int x;
    bool found = false;
    for (x = 0; x < sizeof(AUDIO_INPUT_VALUES)/sizeof(int); x++)
    {
        if (AUDIO_INPUT_VALUES[x] == message->data[0])
        {
            message->data[0] = x;
            found = true;
            break;
        }
    }

    if ((message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS) || found)
    {
        app_commands_edit_processor_mode_input(message);
    }
    else
    {
        message->response = PARAM_NOT_RECOGNISED;
        app_commands_send_message(message);
        
        result = false;
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs a procesor mode volume change based upon a delta value
 */
bool app_commands_edit_processor_mode_volume(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    int16_t value = (int16_t)message->data[0];
    int16_t current_value = input_control_get_processor_mode_volume();

    // Set changed state
    if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        menu_try_initialise_change_state(&current_value, sizeof(int16_t));
    }
    
    app_commands_process_message_value(message, &current_value, &value, 1);

    if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        input_control_set_processor_mode_volume(current_value);

        message->data[0] = input_control_get_processor_mode_volume();

        // If the current input is the processor mode input then adjust the volume
        if (input_control_get_selected_input() == input_control_get_processor_mode_input())
        {
            volume_set(message->data[0], true);
        }

        menu_update_change_state(&current_value, sizeof(int16_t));
    }

    message->command = APP_COMMANDS_PROCESSOR_MODE_VOLUME;
    message->response = STATUS_UPDATE;

    // Just incase its not been defined
    message->data_length = 1;

    app_commands_send_message(message);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the processor mode input as a string
 * @return 
 */
void app_commands_get_processor_mode_input_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", AUDIO_INPUT_STRINGS[input_control_get_processor_mode_input()]);
}

/**-----------------------------------------------------------------------------
 * Gets the processor mode input as a string
 * @return 
 */
void app_commands_get_processor_mode_volume_string(char* buffer, size_t size)
{
    sprintf(buffer, "%02d", input_control_get_processor_mode_volume());
}

/* *****************************************************************************
 End of File
 */
