/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    app_commands_rgb_led.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_rgb_led.h"
#include "app_config.h"
#include "app_context.h"
#include "utilities/app_commands_utilities.h"
#include "utilities/rgb_led.h"
#include "utilities/logger/logger.h"
#include "utilities/stateful_process.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef enum
{
    RGB_LED_TEST_STATE_START,
    RGB_LED_TEST_STATE_START_FLASH,
    RGB_LED_TEST_STATE_START_SOLID,
    RGB_LED_TEST_STATE_RESET,
    RGB_LED_TEST_STATE_END,
} RGB_LED_TEST_STATE;

typedef struct
{
    //! A copy of the message that started the process - if 
    STRUCTURED_MESSAGE initiating_message;
    
    //! Colour state pre process so it can be reset at the end
    RGB_LED_COLOUR prerun_colour;
    
    //! Flash state pre processing
    bool prerun_flash;
        
    //! The current rgb led colour
    RGB_LED_COLOUR running_colour;
    
    //! Colour state stored pre test process so it can be reset at the end
    RGB_LED_COLOUR test_colour;
        
    //! Determines whether we are sending progress messages
    bool progress_messages;

} RGB_LED_DATA;

static RGB_LED_DATA rgb_led_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 */
static void app_commands_rgb_led_send_message(RGB_LED_COLOUR colour)
{
    if (rgb_led_data.progress_messages == true)
    {
        rgb_led_data.initiating_message.data[0] = stateful_process_get_info()->state;
        rgb_led_data.initiating_message.data[1] = colour;
        rgb_led_data.initiating_message.data_length = 2;
        rgb_led_data.initiating_message.response = STATUS_UPDATE;
        
        app_commands_send_message(&rgb_led_data.initiating_message);
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**----------------------------------------------------------------------------- 
 * Gets the current running colour
 */
RGB_LED_COLOUR app_commands_rgb_led_get_running_colour()
{
    return rgb_led_data.running_colour;
}
     
/**-----------------------------------------------------------------------------
 * Sets the apps running colour
 * @param colour
 */
void app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR colour)
{
    rgb_led_data.running_colour = colour;
}

/**-----------------------------------------------------------------------------
 * Run the RGB LED test
 */
void app_commands_rgb_led_test()
{
    // Get the state from the Menu items PROCESS_INFO
    switch (stateful_process_get_info()->state)
    {
        case RGB_LED_TEST_STATE_START:
        {
            logger_log_infof(LOGGER_CODE_RGB_LED_TEST, "RGB LED TEST: Start");
            app_commands_rgb_led_send_message(RGB_LED_COLOUR_OFF);
            
            // Store current LED state
            rgb_led_data.prerun_colour = app_commands_rgb_led_get_running_colour();
            rgb_led_data.prerun_flash = rgb_led_get_flash();
            
            logger_log_infof(LOGGER_CODE_RGB_LED_TEST, "RGB LED TEST: Start %d --", rgb_led_data.prerun_colour);
            rgb_led_data.test_colour = RGB_LED_COLOUR_OFF;
            
            timers_expire(TIMER_INDEX_RGB_LED_TEST);
            
            stateful_process_get_info()->state = RGB_LED_TEST_STATE_START_FLASH;
            break;
        }
        
        case RGB_LED_TEST_STATE_START_FLASH:
        {
            if (timers_has_expired(TIMER_INDEX_RGB_LED_TEST))
            {
                logger_log_infof(LOGGER_CODE_RGB_LED_TEST, "RGB LED TEST: Start Flash");

                // Get next colour
                if (rgb_led_data.test_colour != RGB_LED_COLOUR_WHITE)
                {
                    rgb_led_data.test_colour++;
                    
                    app_commands_rgb_led_send_message(rgb_led_data.test_colour);

                    // Set the LED to Red
                    app_commands_rgb_led_set_running_colour(rgb_led_data.test_colour);
                    rgb_led_set(rgb_led_data.test_colour);
                    
                    rgb_led_flash();

                    timers_reset(TIMER_INDEX_RGB_LED_TEST);

                    stateful_process_get_info()->state = RGB_LED_TEST_STATE_START_SOLID;
                }
                else
                {
                    stateful_process_get_info()->state = RGB_LED_TEST_STATE_RESET;
                }
            }
            
            break;
        }
        
        case RGB_LED_TEST_STATE_START_SOLID:
        {
            if (timers_has_expired(TIMER_INDEX_RGB_LED_TEST))
            {
                logger_log_infof(LOGGER_CODE_RGB_LED_TEST, "RGB LED TEST: Start Solid");
                app_commands_rgb_led_send_message(rgb_led_data.test_colour);
                
                rgb_led_flash_stop();
                
                timers_reset(TIMER_INDEX_RGB_LED_TEST);
                
                stateful_process_get_info()->state = RGB_LED_TEST_STATE_START_FLASH;
            }
            break;
        }
        
        case RGB_LED_TEST_STATE_RESET:
        {
            if (timers_has_expired(TIMER_INDEX_RGB_LED_TEST))
            {
                logger_log_infof(LOGGER_CODE_RGB_LED_TEST, "RGB LED TEST: Reset %d --", rgb_led_data.prerun_colour);
                app_commands_rgb_led_send_message(RGB_LED_COLOUR_OFF);
                
                // Reset the colour - note this may immediatly get reset to the runtime colour
                // but at least we've had a go...
                rgb_led_set(rgb_led_data.prerun_colour);
                
                logger_log_infof(LOGGER_CODE_RGB_LED_TEST, "RGB LED TEST: Reset2 %d --", rgb_led_data.prerun_colour);

                stateful_process_get_info()->state = RGB_LED_TEST_STATE_END;
            }
            
            break;
        }
        
        case RGB_LED_TEST_STATE_END:
        {
            if (timers_has_expired(TIMER_INDEX_RGB_LED_TEST))
            {
                logger_log_infof(LOGGER_CODE_RGB_LED_TEST, "RGB LED TEST: End %d --", rgb_led_data.prerun_colour);
                app_commands_rgb_led_send_message(RGB_LED_COLOUR_OFF);

                // We've waited for the power up so now going back to the splash screen
                // Changing the mode finishes the process
                app_commands_display_default_screen(NULL);

                rgb_led_data.progress_messages = false;
                
                stateful_process_end();
            }
            
            break;
        }
        
        default:
            break;
    }
}
    
/**-----------------------------------------------------------------------------
 * Where the process is called from a structured message this stores the original message
 * which is then used to send response messages
 * @param message
 */
void app_commands_rgb_led_test_set_initiating_message(STRUCTURED_MESSAGE* message)
{
    if (message != NULL)
    {
        structured_message_copy(message, &rgb_led_data.initiating_message);
        rgb_led_data.progress_messages = true;
    }
    else
    {
        memset(&rgb_led_data.initiating_message, 0, sizeof(STRUCTURED_MESSAGE));
        rgb_led_data.progress_messages = false;
    }
}

/* *****************************************************************************
 End of File
 */
