/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    app_commands_standby.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_standby.h"
#include "app_config.h"
#include "app_context.h"
#include "utilities\cn_detect.h"
#include "utilities\i2c.h"
#include "utilities\input_context.h"
#include "utilities\input_control_defs.h"
#include "utilities/logger/logger.h"
#include "utilities\power.h"
#include "utilities\rgb_led.h"
#include "utilities\rs232.h"
#include "utilities\standby.h"
#include "utilities\system_fault.h"
#include "utilities\timers.h"
#include "utilities\watchdog.h"


/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! If true then this overrides sleep mode checks and puts standby state into Sleep
    bool force_sleep;
    
    // The number sleep cycles triggerd by the WDT allowed to run before a temprary wake-up.  This is to 
    // avoid the issue where afterv 17mins of sleep the device goes nuts with some 
    // interrupts that we cant explain
    uint32_t wake_up_cycles;
    
 } APP_COMMANDS_STANDBY_DATA;

static APP_COMMANDS_STANDBY_DATA standby_data = 
{
    .force_sleep = false,
    .wake_up_cycles = 0, // 30 * 1.025s to start with
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Determines whether the device is in a state to Idle
 * @return 
 */
static bool app_commands_standby_can_idle()
{
    //logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY IDLE %d %d --", !tcpip_is_connected() || !power_get_network_active_in_standby());
    // Removed test for tcp connection as Avatar and HDMI ARC do not need tcp...
    //return !tcpip_is_connected() || !power_get_network_active_in_standby();
    return !power_get_network_active_in_standby();
}

/**-----------------------------------------------------------------------------
 * Determines whether the device is in a state to Sleep
 * @return 
 */
static bool app_commands_standby_can_sleep()
{
    //logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY SLEEP %d %d %d ", standby_data.force_sleep, power_get_network_active_in_standby(), power_get_rs232_active_in_standby());
    return standby_data.force_sleep || (app_commands_standby_can_idle() && !power_get_rs232_active_in_standby());
}

/**-----------------------------------------------------------------------------
 * Used in POWERDOWN to check if there is power up activity on the inputs
 */
static bool check_inputs_for_powerup()
{
    bool result = false;
    
    if (input_context_has_changed())
    {
        if (input_context_has_remote_changed())
        {
            // If its a remote message check it a POWER UP message
            STRUCTURED_MESSAGE* message = input_context_get_remote_message();
            if ((message->command == REMOTE_CMD_INPUT_POWER) ||
                (message->command == REMOTE_CMD_INPUT_POWER_ON))
            {
                logger_log_debugf(230, "STANDBY Waking up from REMOTE message START --");
                result = true;
            }
        }
        else 
        {
            // It wasnt a remote message so is a button or encoder so just wake up
            logger_log_debugf(230, "STANDBY Waking up from INTERNAL activity START --");
            result = true;
        }
        
        // Reset the input context.  We have put in a 5second wait when it first gos to sleep 
        // to ensure that it does not immediatemly wake up - I suspect without this
        // the remote message was stored so just waited the 5 seconds before automatically restarting
        input_context_reset();
    }    
    
    return result;
}

/**
 * Disables the change notification ports used for detecting wake-up during power down
 * @return 
 */
static void disable_change_notice_ports()
{
    // Setup change notice Ports for change notification in Idle/Sleep mode - 
    // enabled in Harmony config but disabled until entering Idle/Sleep mode
    // Note this is only needed where Change Notifications are not used for the
    // peripheral for any other purpose (i.e. not the Remote that uses CN to work out
    // which RC5 code is being sent))

    // Encoder
    encoder_change_notice_disable();
    
    // RS232
    PLIB_PORTS_PinChangeNoticePerPortDisable(RS232_CHANGE_NOTICE_PORTS_ID,
                                             RS232_CHANGE_NOTICE_PORT_CHANNEL, 
                                             RS232_CHANGE_NOTICE_PIN);
}

/**
 * Disables the change notification ports used for detecting wake-up during power down
 * @return 
 */
static void enable_change_notice_ports()
{
    // Setup change notice Ports for change notification in Idle/Sleep mode - 
    // enabled in Harmony config but disabled until entering Idle/Sleep mode
    // Note this is only needed where Change Notifications are not used for the
    // peripheral for any other purpose (i.e. not the Remote that uses CN to work out
    // which RC5 code is being sent))

    // Encoder
    encoder_change_notice_enable();
    
    // RS232
    if (power_get_rs232_active_in_standby())
    {
        PLIB_PORTS_PinChangeNoticePerPortEnable(RS232_CHANGE_NOTICE_PORTS_ID,
                                                RS232_CHANGE_NOTICE_PORT_CHANNEL, 
                                                RS232_CHANGE_NOTICE_PIN);
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Sets up registers for standby settings
 */
void app_commands_standby_initialise()
{
    disable_change_notice_ports();

    // The registers to set are worked out by searching the data-sheet for "SIDL"
    // (Stop In Idle) and adding an entry for each Register that supports SIDL.
    // They can then be commented out of they are needed during idle mode.
    
    // Reference Oscillator Configuration
    REFO1CONbits.SIDL = 1;
    REFO2CONbits.SIDL = 1;
    REFO3CONbits.SIDL = 1;
    REFO4CONbits.SIDL = 1;
    
    // Timer Configuration
    T1CONbits.SIDL = 1;
    T2CONbits.SIDL = 1; 
    T3CONbits.SIDL = 1;
    T4CONbits.SIDL = 1;
    T5CONbits.SIDL = 1;
    T6CONbits.SIDL = 1;
    T7CONbits.SIDL = 1;
    T8CONbits.SIDL = 1;
    T9CONbits.SIDL = 1;
    
    // Input Capture
    IC1CONbits.SIDL = 1;
    IC2CONbits.SIDL = 1;
    IC3CONbits.SIDL = 1;
    IC4CONbits.SIDL = 1;
    IC5CONbits.SIDL = 1;
    IC6CONbits.SIDL = 1;
    IC7CONbits.SIDL = 1;
    IC8CONbits.SIDL = 1;
    IC9CONbits.SIDL = 1;
    
    // Output Compare
    OC1CONbits.SIDL = 1;
    OC2CONbits.SIDL = 1;
    OC3CONbits.SIDL = 1;
    OC4CONbits.SIDL = 1;
    OC5CONbits.SIDL = 1;
    OC6CONbits.SIDL = 1;
    OC7CONbits.SIDL = 1;
    OC8CONbits.SIDL = 1;
    OC9CONbits.SIDL = 1;
    
    // SPI
    SPI1CONbits.SIDL = 1;
    SPI2CONbits.SIDL = 1;
    SPI3CONbits.SIDL = 1;
    SPI4CONbits.SIDL = 1;
    SPI5CONbits.SIDL = 1;
    SPI6CONbits.SIDL = 1;
    
    // I2C
    I2C1CONbits.SIDL = 1;
    I2C2CONbits.SIDL = 1;
    I2C3CONbits.SIDL = 1;
    I2C4CONbits.SIDL = 1;
    I2C5CONbits.SIDL = 1;
    
    // UART
    U1MODEbits.SIDL = 0; // Keep UART 1 Alive in Idle state
    U2MODEbits.SIDL = 1;
    U3MODEbits.SIDL = 1;
    U4MODEbits.SIDL = 1;
    U5MODEbits.SIDL = 1;
    U6MODEbits.SIDL = 1;
    
    // PMP Configuration
    PMCONbits.SIDL = 1;
    
    // RTCC Configuration
    RTCCONbits.SIDL = 1;
    
    // ADC Configuration
    ADCCON1bits.SIDL = 1;
    
    // Ethernet
    ETHCON1bits.SIDL = 1;
    
    // Comparator
    CMSTATbits.SIDL = 1;
}

/**-----------------------------------------------------------------------------
 * Manages standby in the Powered down state
 */
void app_commands_standby_powereddown()
{
    switch(standby_get_state())
    {
        case STANDBY_STATE_ENTRY:
        {
            // Reset the input context first
            input_context_reset();

            SYSTEM_FAULT fault = system_fault_get();
            if (fault.code == SYSTEM_FAULT_CODE_NO_FAULT)
            {
                logger_log_debugf(LOGGER_CODE_STANDBY, "STANDBY Entering Standby Idle:%s Sleep:%s ", 
                                  app_commands_standby_can_idle() ? "Yes" : "No",
                                  app_commands_standby_can_sleep() ? "Yes" : "No");
                
                standby_set_initial_state(true, //NETWORK STANDBY POWERDOWNapp_commands_standby_can_idle(),
                                          app_commands_standby_can_sleep(),
                                          STANDBY_STATE_SLEEP);
                
                rgb_led_flash_stop();
                rgb_led_set(RGB_LED_COLOUR_RED);
            }
            else
            {
                standby_set_initial_state(true, true, STANDBY_STATE_SLEEP);
                rgb_led_process();
            }

            // Sort-out display 
            app_commands_adjust_brightness(APP_COMMAND_BRIGHTNESS_OFF);
            
            //20191113 - Removed because Pete wants the LED on when in standby
            // If in dark mode turn of the LED
            //if (app_commands_get_dark_mode() == APP_COMMAND_DARK_MODE_ON)
            //{
            //    rgb_led_set(RGB_LED_COLOUR_OFF);
            //}
            
            //set_ports_low();
            if (app_commands_standby_can_sleep())
            {
                pins_set_all_low(STANDBY_SLEEP_STATE_SLEEP);
            }
            else if (app_commands_standby_can_idle())
            {
                pins_set_all_low(STANDBY_SLEEP_STATE_IDLE);
            }
            else
            {
                pins_set_all_low(STANDBY_SLEEP_STATE_ACTIVE);
            }

            // Set wake_up_cycles
            standby_data.wake_up_cycles = 0;

            input_context_reset();
            remote_clear();
            //SA30v22beta1-power_disable_by_index(POWER_INDEX_STREAMER_PSU);
            break;
        }

        case STANDBY_STATE_SLEEP:
        {
            // Add a small pause otherwise the Watchdog appears to get messed up
            // and is not coming out of sleep on interrupts.
            // 1.3 - 24/12/2019 -  Changed delay from 5ms to 25ms (although 15 
            // appeared to work) because the remote would not wake up the device
            // when logging was switched off..
            timers_delay_ms(25);

            logger_log_debugf(230, "POWERDOWN:: Entering SLEEP State --");
            logger_process();

            // In Low power mode we put the processor to sleep and use 
            // either the watchdog timer to wake it up periodically to check if 
            // we need to wake up or change notifications.
            // Note: Change notifications need to be enabled - you would expect 
            // Harmony to do this when you define a port as a CN port but it 
            // only appears to have done this for the RC5 port (unless i've 
            // forgotten something) so am adding in PLIB_PORTS_PinChangeNoticeEnable(PORTS_ID_0, CN13);
            // commands to APP_Init ..
            //logger_log_debugf(230, "POWERDOWN %d --", timers_get_remaining_count(APP_TIMER_INDEX_CN_DETECT));
            cn_detect_enable(40);  // 40ms is enough to handle a reasonable press of the power off button

            // Check if we've had a change notification - note this also clears the flag
            if (cn_detect_get())
            {
                logger_log_debugf(230, "POWERDOWN DETECTED CN --");
                cn_detect_disable();

                // As we move out of sleep state clear the power cycle timer
                // so it doesn't stop us waking up if we need to , as it 
                // will not have been incrementing while the unit was in sleep
                // mode
                timers_expire(TIMER_INDEX_POWER_CYCLE);

                // Try processing any inputs before we move on to ensure
                // it will be picked up when in the ACTIVE state
                app_commands_process_controls();

                // Set the state to active
                standby_set_state_active();
            }
            else
            {
                logger_log_debugf(230, "POWERDOWN DETECTED NO CN --");

                encoder_initialise();
                remote_pause();

                // We have the RS232 Change Notification CN9 on port RG7
                // It appears that this needs enabling before generating change
                // notice interrupts, however it also has a side effect of 
                // causing a unit reset if the client connects or disconnects 
                // over RS232, so we are only enabling it if:
                // a: We have RS232 enabled in standby mode and
                // b: If we are actually trying to idle/sleep rather than have 
                // it on the whole time.
                enable_change_notice_ports();

                // Note this working is dependent upon the settings in setup_standby_registers()
                standby_try_sleep();

                disable_change_notice_ports();
            }

            break;
        }

        case STANDBY_STATE_ACTIVE:
        {
            //logger_log_debugf(230, "POWERDOWN:: Entering ACTIVE State --");

            // Handle upgrade messages 
            if (avatar_message_in_upgrade())
            {
                // Perform I2C Request checking
                i2c_process();

                // RGB LEd
                rgb_led_process();

                rs232_process();

                flash_memory_process();
    
                avatar_process();
            }
            
            // Test for activity that occurred in sleep state before clearing and starting again
            bool wake_up = check_inputs_for_powerup();

            if (!wake_up)
            {
                rgb_led_process();

                input_context_reset();

                // Perform I2C Request checking
                i2c_process();

                // External message processing
                // Handle TCP/IP processing
                if (power_get_network_active_in_standby())
                {
                    //TODOprocess_tcpip();
                        
                    // Avatar processing
                    if (avatar_process())
                    {
                        if (avatar_message_in_upgrade())
                        {
                            rgb_led_flash();
                        }
                        else if (app_commands_get_player_status() != AVATAR_PLAYER_STATUS_STOPPED) // BUG 3824 - Put this back to stop device being woken up by TCPIP connection - appears to still be OK waking up from Spotify or web client
                        {
                            // BUG 3413 - Change from forcing a wakeup to checking the TIMER_INDEX_POWER_CYCLE before waking up
                            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
                            {
                                logger_log_debugf(230, "POWERDOWN Waking up from AVATAR message --");
                                // Expire the timeout so that we force a wakeup
                                //timers_expire(TIMER_INDEX_POWER_CYCLE);

                                //app_context_set_mode(APP_CONTEXT_MODE_MENU);
                                //app_context_set_menu(APP_COMMAND_MENU_POWERING_UP, DISPLAY_MESSAGE_SCREEN_NAVIGATION_DIRECT);

                                // Set the current definition omits starts as the Avatar input
                                input_control_set_current_definition(AUDIO_INPUT_AVATAR);

                                wake_up = true;
                            }
                        }
                    }
                }

                if (!wake_up)
                {
                    // Handle rs232 processing
                    rs232_process();


#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
                    //logger_log_debugf(230, "POWERDOWN:: Entering ACTIVE State 5 --");
                    rs232_hdmi_arc_process();
#endif

                    // Check if we have received an external message
                    STRUCTURED_MESSAGE* message = input_context_get_external_message();
                    
                    //logger_log_debugf(230, "POWERDOWN:: Entering ACTIVE State 6 %d %d --", message->command, input_context_has_external_changed());
                    
                    if ((input_context_has_external_changed()) &&
                        ((message->command == REMOTE_CMD_INPUT_POWER) ||
                         (message->command == REMOTE_CMD_INPUT_POWER_ON) ||
                         ((message->command == APP_COMMANDS_SIMULATE_RC5) && (message->data[1] == REMOTE_CMD_INPUT_POWER))))
                    {
                        logger_log_debugf(230, "POWERDOWN Waking up from EXTERNAL message --");
                        wake_up = true;
                    }
                    else
                    {
                        // reset the context to clear the external message
                        input_context_reset();

                        // Now perform any processing based on the current state of the input
                        // in either the buttons or the rotary, or something via the IR or rs232
                        app_commands_process_controls();

                        // Check to see if application mode has been changed from APP_CONTEXT_MODE_STANDBY
                        // If so then wake up.  The wait is there because sometimes there was overspill from the POWER button
                        // which meant it either powered don the straight back up or powered up the straight back off
                        wake_up = check_inputs_for_powerup();
                    }
                }
            }
            
            if ((!wake_up) && app_commands_get_ac_not_present_triggered_powerdown())
            {
                logger_log_infof(230, "STANDBY Waking up from AC not present powerdown --");
                wake_up = true;
            }
            
            if ((wake_up) &&
                (timers_has_expired(TIMER_INDEX_POWER_CYCLE)))
            {
                rgb_led_set(RGB_LED_COLOUR_ORANGE);
                
                // Set the ap state to Wake Up
                app_context_set_app_state(APP_STATE_WAKEUP);
            }

            break;
        }

        default:
            break;
    }

    // Clear the watch dog
    watchdog_reset();

    // Review the powered down state
    if (app_context_get_app_state() == APP_STATE_POWEREDDOWN)
    {
        SYSTEM_FAULT fault = system_fault_get();
        if (fault.code == SYSTEM_FAULT_CODE_NO_FAULT)
        {
            standby_update_state(true, //NETWORK STANDBY POWERDOWNapp_commands_standby_can_idle(),
                                 app_commands_standby_can_sleep());
        }
        else
        {
            standby_update_state(false, false);
        }
    }
    else
    {
        // Reset the stand-by state
        standby_reset();
    }
}

/* *****************************************************************************
 End of File
 */
