/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_system.h"
#include "app_commands_power.h"
#include "app_commands_rgb_led.h"
#include "app_config.h"
#include "app_context.h"
#include "app_initialise.h"
#include "utilities/adc/adc_module.h"
#include "utilities/app_commands_utilities.h"
#include "utilities/display/display_defs.h"
#include "utilities/input_control.h"
#include "utilities/logger/logger.h"
#include "utilities\power.h"
#include "utilities\stateful_process.h"
#include "utilities\system_fault.h"
#include "utilities\timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct 
{
    //! Stores the last shutdown fault code
    SYSTEM_FAULT_CODE last_shutdown_code;
    
} SYSTEM_DATA;

static SYSTEM_DATA system_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Performs a reboot
 */
static void app_commands_process_reboot(MENU_ID menu_id)
{
    // Get the state from the Menu items PROCESS_INFO
    switch (stateful_process_get_info()->state)
    {
        case 0:
        {
            // State 1 - start
            logger_log_info(LOGGER_CODE_FACTORY_RESET,"Reboot Start");

            // Display the 
            //NOTE: Dont reset the mode during a process....until you've finished
            //app_context_set_mode(APP_CONTEXT_MODE_MENU);
            app_context_set_menu(APP_COMMAND_MENU_SYSTEM_RESET_REBOOT);
            
            // Save the settings
            settings_save();

            timers_update_timeout(TIMER_INDEX_POWER_CYCLE, 3000, TIMERS_TIMEOUT_RESET_IMMEDIATE);
            
            stateful_process_get_info()->state = 1;
            break;
        }
        
        case 1:
        {
            flash_memory_process();

            // check whether we need to save the current state of the settings
            settings_process();

            // Wait 1 second to display the message "Powering Down"
            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
            {
                 // Set state
                stateful_process_get_info()->state = 2;
            }
            break;
        }
        
        case 2:
        {
            // State 2 - Wait for settings to be saved
            if (settings_available())
            {
                logger_log_info(LOGGER_CODE_FACTORY_RESET, "Power Down");

                // Set the dac into reset mode during power up
                input_control_dac_reset(INPUT_CONTROL_SWITCH_OFF);
                spdif_rx_reset(INPUT_CONTROL_SWITCH_OFF);                

                // Stop the Avatar
                avatar_stop();

                // Power down
                app_commands_power_down();

                // 06/09/2019
                // Do a Full blown reset - this then gives the chance for the 
                // system and Avatar to go through the full handshaking process again
                // NOTE: All the code following this is never run!!!
                SYS_RESET_SoftwareReset();
                
                //-------------------------------------------------------
                
                app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_RED);
                rgb_led_set(app_commands_rgb_led_get_running_colour());
                
                vfd_hold_in_reset();
                
                // Start wait timer
                //timers_reset(APP_TIMER_INDEX_POWER_CYCLE);
                timers_update_timeout(TIMER_INDEX_POWER_CYCLE, 3000, TIMERS_TIMEOUT_RESET_IMMEDIATE);
                
                // Set state
                stateful_process_get_info()->state = 3;
            }
            break;
        }
        
        case 3:
        {
            // Wait 3 seconds for proper power down
            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
            {
                 // Set state
                stateful_process_get_info()->state = 4;
            }
            break;
        }
        
        case 4:
        {
            // Wait 3 seconds for proper power down
            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
            {
                logger_log_info(LOGGER_CODE_FACTORY_RESET,"Power Up");
                
                // Power up
                power_enable();
            
                // Send unsolicited power message
                app_commands_request_status(APP_COMMANDS_POWER, app_commands_power);
              
                // To pause before resetting the VFD
                timers_delay_ms(1);
                vfd_clear_reset();

                // Start wait timer
                timers_update_timeout(TIMER_INDEX_POWER_CYCLE, 500, TIMERS_TIMEOUT_RESET_IMMEDIATE);
               
                // Set state
                stateful_process_get_info()->state = 5;
            }
            
            break;
        }
        
        case 5:
        {
            // Wait 3 seconds for proper power down
            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
            {
                logger_log_info(LOGGER_CODE_FACTORY_RESET,"Power Up");
                
                display_initialise();
                 
                display_reset_brightness();
            
                app_commands_adjust_brightness(APP_COMMAND_BRIGHTNESS_FULL);
                app_commands_check_display();

                // Start wait timer
                timers_update_timeout(TIMER_INDEX_POWER_CYCLE, 2000, TIMERS_TIMEOUT_RESET_IMMEDIATE);
                
                // Set state
                stateful_process_get_info()->state = 6;
            }
            break;
        }
        
        case 6:
        {
            // Wait 3 seconds for proper power down
            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
            {
                logger_log_info(LOGGER_CODE_FACTORY_RESET,"Resetting DAC");
                
                app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_ORANGE);
               
                // Initialise input control and reset the DAC
                spdif_rx_reset(INPUT_CONTROL_SWITCH_ON);
                input_control_dac_reset(INPUT_CONTROL_SWITCH_ON);
                
                input_control_setup_filter();
                input_control_reset_input();
                
                // Start wait timer
                timers_reset(TIMER_INDEX_POWER_CYCLE);
                
                // Set state
                stateful_process_get_info()->state = 7;
            }
            break;
        }
       
        case 7:
        {
            // Wait 3 seconds for proper power down
            if (timers_has_expired(TIMER_INDEX_POWER_CYCLE))
            {
                logger_log_info(LOGGER_CODE_FACTORY_RESET, "Reboot Complete");

                app_commands_rgb_led_set_running_colour(RGB_LED_COLOUR_WHITE);
                   
                output_control_unmute_outputs();
                
                // Mute/Unmute volume to get ramp
                volume_fade_in();
               
                // We've waited for the power up so now going back to the splash screen
                // Changing the mode finishes the process
                app_commands_display_default_screen(NULL);
                
                // Send a message to any potential External client
                STRUCTURED_MESSAGE message;
                structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_RS232);
                message.command = APP_COMMANDS_REBOOT;
                message.response = STATUS_UPDATE;
                message.data_length = 1;
                message.data[0] = 0x00;
                
                app_commands_send_message(&message);
            }
            
            break;
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Draw the bespoke Balance display
 */
void app_commands_draw_reboot(APP_COMMAND_MENU menu)
{
    MENU_ITEM* item = menu_get_current_item();
 
    // Draw the top line
        // Clear the display
    display_clear(false);
    
    // Draw the top line
    display_clear_line(DISPLAY_LINE_1);

    display_start_text(DISPLAY_LINE_1, menu_get_title(menu_get_current_item()->menu_id));
    display_write_line(DISPLAY_LINE_1, false);

    // Draw bottom line
    display_clear_line(DISPLAY_LINE_2);
    switch(stateful_process_get_info()->state)
    {
        case 0:
            display_start_text(DISPLAY_LINE_2, "Save settings");
            break;
        case 1:
        case 2:
            display_start_text(DISPLAY_LINE_2, "Powering down");
            break;
        case 3:
        case 4:
        case 5:
        case 6:
            display_start_text(DISPLAY_LINE_2, "Powering up");
            break;
        case 7:
            display_start_text(DISPLAY_LINE_2, "Complete");
            break;
        default:
            break;
    }
    
    display_write_line(DISPLAY_LINE_2, false);
}

/**-----------------------------------------------------------------------------
 * Used to change the logger level
 */
bool app_commands_edit_logger_level(STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = message->data[0];
        int8_t current_value = (int8_t)logger_get_level();
        
        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        message->command = APP_COMMANDS_SET_LOGGER_LEVEL;
        message->response = STATUS_UPDATE;
        
        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                current_value = value;
                logger_set_level((LOGGER_TYPE)current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }
            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            {
                current_value++;
                if (current_value >= LOGGER_TYPE_ALL)
                {
                    current_value = LOGGER_TYPE_NONE;
                }

                logger_set_level((LOGGER_TYPE)current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                current_value--;
                if (current_value < 0)
                {
                    current_value = LOGGER_TYPE_DEBUG;
                }

                logger_set_level((LOGGER_TYPE)current_value);
                
                message->data[0] = (uint8_t)current_value;
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
            {
                message->data[0] = (uint8_t)current_value;
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                break;
            }
        }

        app_commands_send_message(message);
                
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            // Store Logger Type setting (Note: Only stored here, and Test type 
            // never stored)
            LOGGER_TYPE logger_type = logger_get_level();
            if (logger_type != LOGGER_TYPE_TEST)
            {
                settings_set(SETTING_LOGGER_TYPE, &logger_type);
            }
            
            menu_update_change_state(&current_value, sizeof(int8_t));
        }
    }

    return result;
}

/**-----------------------------------------------------------------------------
 * Used to change the remote system code
 */
bool app_commands_edit_system_code(STRUCTURED_MESSAGE* message)
{
    bool result = false;

    app_commands_review_message_action(message);

    if (!message->repeat)
    {
        int8_t value = (int8_t)message->data[0];
        int8_t current_value = (int8_t)remote_get_system_code();

        // Set changed state
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_try_initialise_change_state(&current_value, sizeof(int8_t));
        }

        switch(message->action)
        {
            case STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE:
            {
                current_value = value;
                remote_set_system_code(current_value);
                break;
            }
            case STRUCTURED_MESSAGE_ACTION_SET_INCREMENT:
            case STRUCTURED_MESSAGE_ACTION_SET_DECREMENT:
            {
                current_value = current_value == 16 ? 19 : 16;
                remote_set_system_code(current_value);
                break;
            }

            case STRUCTURED_MESSAGE_ACTION_GET_STATUS:
            {
                message->data[0] = (uint8_t)current_value;
                message->response = STATUS_UPDATE;

                app_commands_send_message(message);
                break;
            }

            default:
            {
                message->response = COMMAND_NOT_RECOGNISED;
                app_commands_send_message(message);
                break;
            }
        }
                
        if (message->action != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
        {
            menu_update_change_state(&current_value, sizeof(int8_t));
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Initiates a factory reset from an external command
 * Includes checking a code before proceeding
 */
void app_commands_factory_reset(STRUCTURED_MESSAGE* message)
{
    if (message->data_length == 2)
    {
        if ((message->data[0] == 0xAA) && (message->data[1] == 0xAA))
        {
            // Invoke a process if set.
            app_commands_start_process(app_commands_process_factory_reset, NULL);
        }
        else
        {
            message->response = PARAM_NOT_RECOGNISED;
            app_commands_send_message(message);
        }
    }
    else
    {
        message->response = INVALID_DATA_LENGTH;
        app_commands_send_message(message);
    }
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the currently pressed button voltage
 * @param buffer
 * @param size
 */
void app_commands_get_active_switch_voltage_string(char* buffer, size_t size)
{
    ADC_BUTTON_INFO* button_info = adc_button_get_current();
    
    if (button_info != NULL)
    {
        logger_log_debugf(0, "ACTIVE BUTTON %d %d --", button_info->button, button_info->input);
    
        // Get its voltage
        uint32_t mV = adc_voltage_get(button_info->input);
        if (mV != ADC_INVALID_REQUEST)
        {
            sprintf(buffer, "  Input: %d %d.%3.3dV", button_info->button, mV / 1000, mV % 1000);
        }
        else
        {
            sprintf(buffer, " N/A");
        }
    }
    else
    {
        sprintf(buffer, " N/A");
    }
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the current amplifier temperature
 * @param buffer
 * @param size
 */
void app_commands_get_amplifier_temperature_string(char* buffer, size_t size)
{
    //#define DEGREES 22
    long temp = adc_temperature_get(APP_ADC_INPUTS_OUTPUT_TEMPERATURE);
    
    if (temp >= 0)
    {
        // Note display assumes its on line 1...
        sprintf(buffer, "%2d.%d%cC", temp / 10, temp % 10, display_get_special_character(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_DEGREES));
    }
    else
    {
        sprintf(buffer, "N/A");
    }
}

/**-----------------------------------------------------------------------------
 * Returns the last shutdown reason.
 * @param buffer
 * @param size
 */
void app_commands_get_last_shutdown_code_string(char* buffer, size_t size)
{
    SYSTEM_FAULT_CODE history[SHUTDOWN_CODE_HISTORY_LENGTH];
    settings_get(SETTING_SHUTDOWN_CODE_HISTORY, &history);
    
    // Note: This is assuming that SHUTDOWN_CODE_HISTORY_LENGTH >= 10, note it
    // also assumes the first entry in the history is the current last shutdown 
    // code as this is copied up-front once settings are initialised.
    sprintf(buffer, "%d %d%d%d%d%d%d%d%d%d", 
            system_data.last_shutdown_code, 
            history[1],
            history[2],
            history[3],
            history[4],
            history[5],
            history[6],
            history[7],
            history[8],
            history[9]);
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the current amplifier temperature
 * @param buffer
 * @param size
 */
void app_commands_get_lifter_temperature(STRUCTURED_MESSAGE* message)
{
    message->command = APP_COMMANDS_LIFTER_TEMP;

    if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
    {
        long temp = adc_temperature_get(APP_ADC_INPUTS_LIFTER_TEMPERATURE);

        if (temp >= 0)
        {
            message->data[0] = temp / 10;
            message->data_length = 1;
            message->response = STATUS_UPDATE;
        }
        else
        {
            message->response = COMMAND_INVALID_NOW;
        }
    }
    else
    {
        message->response = PARAM_NOT_RECOGNISED;
    }
   
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the current amplifier temperature
 * @param buffer
 * @param size
 */
void app_commands_get_lifter_temperature_string(char* buffer, size_t size)
{
    //#define DEGREES 22
    long temp = adc_temperature_get(APP_ADC_INPUTS_LIFTER_TEMPERATURE);

    if (temp >= 0)
    {
        // Note display assumes its on line 1...
        sprintf(buffer, "%2d.%d%cC", temp / 10, temp % 10, display_get_special_character(DISPLAY_LINE_1, DISPLAY_SPECIAL_CHARACTERS_DEGREES));
    }
    else
    {
        sprintf(buffer, "N/A");
    }
}   

/**-----------------------------------------------------------------------------
 * Returns a string representation of the logger level
 * @return 
 */
void app_commands_get_logger_level_string(char* buffer, size_t size)
{
    sprintf(buffer, "%s", LOGGER_TYPE_STRINGS[(int)logger_get_level()]);
}

/**-----------------------------------------------------------------------------
 * Returns a string representation of the current model number e.g. SA20
 * @param buffer
 * @param size
 */
void app_commands_get_model_number(STRUCTURED_MESSAGE* message)
{
    message->command = APP_COMMANDS_MODEL_NUMBER;
    
    if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
    {
        strncpy(message->data, SOFTWARE_MODEL_STRING, strlen(SOFTWARE_MODEL_STRING));
        message->data_length = strlen(SOFTWARE_MODEL_STRING);
        message->response = STATUS_UPDATE;
    }
    else
    {
        message->response = COMMAND_INVALID_NOW;
    }

    app_commands_send_message(message);
}
 
/**-----------------------------------------------------------------------------
 * Returns a string representation of the current amplifier temperature
 * @param buffer
 * @param size
 */
void app_commands_get_output_temperature(STRUCTURED_MESSAGE* message)
{
    message->command = APP_COMMANDS_OUTPUT_TEMP;
    
    if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
    {
        long temp = adc_temperature_get(APP_ADC_INPUTS_OUTPUT_TEMPERATURE);

        if (temp >= 0)
        {
            message->data[0] = temp / 10;
            message->data_length = 1;
            message->response = STATUS_UPDATE;
        }
        else
        {
            message->response = COMMAND_INVALID_NOW;
        }
    }
    else
    {
        message->response = PARAM_NOT_RECOGNISED;
    }
    
    app_commands_send_message(message);
}
   
/**-----------------------------------------------------------------------------
 * Returns a string representation of the software version
 * @return 
 */
void app_commands_get_software_version(char* buffer)
{
    sprintf(buffer, "v%d.%2.2d", SOFTWARE_VERSION_MAJOR, SOFTWARE_VERSION_MINOR);
}

/**-----------------------------------------------------------------------------
 * Heartbeat command to check unit is still connected and communicating
 * also resets the shutdown timer
 * @param message
 */
void app_commands_heartbeat(STRUCTURED_MESSAGE* message)
{
    if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
    {
        // Reset the Auto shutdown timer
        power_reset_auto_shutdown_timer();
        
        // Update the message
        message->command = APP_COMMANDS_HEARTBEAT;
        message->response = STATUS_UPDATE;
        message->data_length = 1;
        message->data[0] = 0x00;
    }
    else
    {
        message->response = PARAM_NOT_RECOGNISED;
    }
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Performs a factory reset - run as a stateful_process
 */
void app_commands_process_factory_reset(MENU_ID menu_id)
{
    // State 1 - start
    logger_log_info(LOGGER_CODE_FACTORY_RESET, "Factory Reset Start");

    // Display the 
    //NOTE: Dont reset the mode during a process....until you've finished
    //app_context_set_mode(APP_CONTEXT_MODE_MENU);
    app_context_set_menu(APP_COMMAND_MENU_SYSTEM_RESET_RESETTING);

    // Mute speakers/preamp and headphones
    output_control_mute_outputs(true, false);
        
    // Reset the settings
    initialise_settings_defaults();

    logger_log_info(LOGGER_CODE_FACTORY_RESET, "Defaults restored");

    // 06/09/2019
    // Reset the avatar
    avatar_factory_reset();

    app_commands_start_process(app_commands_process_reboot, NULL);
}

/**-----------------------------------------------------------------------------
 * Initiates a reboot from an external command
 * Includes checking a code before proceeding
 */
void app_commands_reboot(STRUCTURED_MESSAGE* message)
{
    if (message->data_length == 6)
    {
        if ((message->data[0] == 0x52) && 
            (message->data[1] == 0x45) &&
            (message->data[2] == 0x42) && 
            (message->data[3] == 0x4F) &&
            (message->data[4] == 0x4F) && 
            (message->data[5] == 0x54))
        {
            // Invoke a process if set.
            app_commands_start_process(app_commands_process_reboot, NULL);
        }
        else
        {
            message->response = PARAM_NOT_RECOGNISED;
            app_commands_send_message(message);
        }
    }
    else
    {
        message->response = INVALID_DATA_LENGTH;
        app_commands_send_message(message);
    }
}

/**-----------------------------------------------------------------------------
 * Gets the software version
 */
void app_commands_software_version(STRUCTURED_MESSAGE* message)
{
    if (APP_COMMAND_VALUE_REQUEST_STATUS == (int16_t)message->data[0])
    {
        message->data_length = 2;
        message->data[0] = SOFTWARE_VERSION_MAJOR;
        message->data[1] = SOFTWARE_VERSION_MINOR;
        message->response = STATUS_UPDATE;
    }
// Avatar version removed as it is > 256.....    
//    else if (APP_COMMAND_VALUE_REQUEST_STATUS + 1 == (int16_t)message->data[0])
//    {
//        uint16_t
//        message->data_length = 2;
//        message->data[0] = app_commands_get_avatar_firmware_version();
//        message->data[1] = 0;
//        message->response = STATUS_UPDATE;
//    }
    else if (APP_COMMAND_VALUE_REQUEST_STATUS + 2 == (int16_t)message->data[0])
    {
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
        message->data_length = 2;
        message->data[0] = hdmi_arc_get_firmware_version_major();
        message->data[1] = hdmi_arc_get_firmware_version_minor();
        message->response = STATUS_UPDATE;
#endif
    }
    else if (APP_COMMAND_VALUE_REQUEST_STATUS + 3 == (int16_t)message->data[0])
    {
#if APP_FEATURE_HDMI_ARC == APP_FEATURE_ON  
        message->data_length = 2;
        message->data[0] = hdmi_arc_get_rx_firmware_version_major();
        message->data[1] = hdmi_arc_get_rx_firmware_version_minor();
        message->response = STATUS_UPDATE;
#endif
    }
    else
    {
        message->response = COMMAND_NOT_RECOGNISED;
    }

    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * System Status triggers multiple status requests to give an overall picture of the
 * system state
 * @param message
 */
void app_commands_system_status(STRUCTURED_MESSAGE* message)
{
    message->is_activity = false;

    if (app_context_get_app_state() >= APP_STATE_RUNNING)
    {
        if (message->data[0] == APP_COMMAND_VALUE_REQUEST_STATUS)
        {
            STRUCTURED_MESSAGE message_copy;

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_POWER;
            message_copy.is_activity = false;
            app_commands_power(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_BRIGHTNESS;
            message_copy.is_activity = false;
            app_commands_edit_display_brightness(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_HEADPHONES;
            message_copy.is_activity = false;
            app_commands_headphones(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_SOFTWARE_VERSION;
            message_copy.is_activity = false;
            app_commands_software_version(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_MODEL_NUMBER;
            message_copy.is_activity = false;
            app_commands_get_model_number(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_VOLUME;
            message_copy.is_activity = false;
            app_commands_edit_volume(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_MUTE;
            message_copy.is_activity = false;
            app_commands_mute_user(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_CURRENT_SOURCE;
            message_copy.is_activity = false;
            app_commands_current_source(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_HEADPHONE_OVERRIDE;
            message_copy.is_activity = false;
            app_commands_edit_headphone_override(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_BALANCE;
            message_copy.is_activity = false;
            app_commands_edit_balance(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_SAMPLE_RATE;
            message_copy.is_activity = false;
            app_commands_get_sample_rate(&message_copy);

//TODO            structured_message_copy(message, &message_copy);
//            message_copy.command = APP_COMMANDS_NETWORK_NAME;
//            message_copy.is_activity = false;
//            app_commands_edit_network_name(&message_copy);

//TODO            structured_message_copy(message, &message_copy);
//            message_copy.command = APP_COMMANDS_IP_ADDRESS;
//            message_copy.is_activity = false;
//            app_commands_get_runtime_ip_address(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_TIMEOUT_COUNTER;
            message_copy.is_activity = false;
            app_commands_get_timeout_counter(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_LIFTER_TEMP;
            message_copy.is_activity = false;
            app_commands_get_lifter_temperature(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_OUTPUT_TEMP;
            message_copy.is_activity = false;
            app_commands_get_output_temperature(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_AUTO_SHUTDOWN;
            message_copy.is_activity = false;
            app_commands_edit_auto_shutdown_timeout(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_INPUT_DETECT;
            message_copy.is_activity = false;
            app_commands_get_input_detection_status(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_PROCESSOR_MODE_INPUT;
            message_copy.is_activity = false;
            app_commands_edit_processor_mode_input(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_PROCESSOR_MODE_VOLUME;
            message_copy.is_activity = false;
            app_commands_edit_processor_mode_volume(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_DC_OFFSET_STATUS;
            message_copy.is_activity = false;
            app_commands_get_dc_offset_status(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_SHORT_CIRCUIT_STATUS;
            message_copy.is_activity = false;
            app_commands_get_short_circuit_status(&message_copy);

            structured_message_copy(message, &message_copy);
            message_copy.command = APP_COMMANDS_SET_FILTER;
            message_copy.is_activity = false;
            app_commands_edit_filter(&message_copy);

            message->response = STATUS_UPDATE;
        }
        else
        {
            message->response = COMMAND_INVALID_NOW;
        }

        app_commands_send_message(message);
    }
}
    
/* *****************************************************************************
 End of File
 */
