/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Manages app level upgrade process
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_commands_upgrade.h"
#include "crypto/crypto.h"
#include "crypto/src/md5.h"
#include "utilities/upgrade_metadata.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! Used for a 1-time suppression of an upgrade
    bool suppress_upgrade;
} UPGRADE_DATA;

static UPGRADE_DATA upgrade_data = {
    .suppress_upgrade = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Version check used to decide whether the upgrade can continue based on the 
 * version number read from the packed upgrade image.
 * @param major_version
 * @param minor_version
 * @return 
 */
bool app_commands_check_upgrade_version(uint8_t major_version, uint8_t minor_version)
{
    bool result = true;
    uint8_t current_version_major = SOFTWARE_VERSION_MAJOR;
    uint8_t current_version_minor = SOFTWARE_VERSION_MINOR;

    if (upgrade_data.suppress_upgrade)
    {
        upgrade_data.suppress_upgrade = false;
        result = false;
    }
    //TODO Removing check as we will always need to upgrade the SA30 to whatever
    //     version the Avatar decides to upgrade us to, otherwise it will just get 
    //     into an upgrade loop..
    //if (major_version < current_version_major)
    //{
    //    result = false;
    //}
    //else if (major_version == current_version_major)
    //{
    //    if (minor_version <= current_version_minor)
    //    {
    //        result = false;
    //    }
    //}

    logger_log_debugf(123, "UPGRADE METADATA Checking Version Numbers %d/%d with upgrade %d/%d (%d) --", current_version_major, current_version_minor, major_version, minor_version, result);
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Performs a 1-time suppression of the upgrade in the version check
 */
void app_commands_suppress_upgrade()
{
    upgrade_data.suppress_upgrade = true;
}

/**-----------------------------------------------------------------------------
 * Prints Upgrade Metadata
 */
void app_commands_upgrade_print_info(STRUCTURED_MESSAGE* message)
{
    upgrade_metadata_print_metadata();
    message->response = STATUS_UPDATE;    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * uploads upgrade data
 * @param message
 */
void app_commands_upgrade_upload_data(STRUCTURED_MESSAGE* message)
{
    UPGRADE_METADATA_LOAD_ID load_id = message->data[0];
    
    //logger_log_debugf(23, "UPGRADE METADATA DATA %d %s --", message->data_length, message->data[1]);
    UPGRADE_METADATA_UPLOAD_STATUS result = upgrade_metadata_upload_data(load_id, &message->data[1], message->data_length - 1, NULL);

    // Add the Id to the message
    if (result != UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS)
    {
        memcpy(message->data, &load_id, sizeof(UPGRADE_METADATA_LOAD_ID));
        memcpy(message->data + sizeof(UPGRADE_METADATA_LOAD_ID), &result, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
        message->data_length = sizeof(UPGRADE_METADATA_LOAD_ID) + sizeof(UPGRADE_METADATA_UPLOAD_STATUS);
        message->response = STATUS_UPDATE;      

        app_commands_send_message(message);
    }
}

/**-----------------------------------------------------------------------------
 * Starts an upgrade upload
 * @param message
 */
void app_commands_upgrade_upload_end(STRUCTURED_MESSAGE* message)
{
    logger_log_debugf(123, "UPGRADE METADATA - Upload completing");

    UPGRADE_METADATA_LOAD_ID load_id = message->data[0];
    
    UPGRADE_METADATA_UPLOAD_STATUS result = upgrade_metadata_upload_end(load_id, NULL, UPGRADE_METADATA_UPGRADE_STATUS_OK);

    memcpy(message->data, &load_id, sizeof(UPGRADE_METADATA_LOAD_ID));
    memcpy(message->data + sizeof(UPGRADE_METADATA_LOAD_ID), &result, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
    message->data_length = sizeof(UPGRADE_METADATA_LOAD_ID) + sizeof(UPGRADE_METADATA_UPLOAD_STATUS);
    message->response = STATUS_UPDATE;      
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Starts an upgrade upload
 * @param message
 */
void app_commands_upgrade_upload_start(STRUCTURED_MESSAGE* message)
{
    logger_log_debugf(123, "UPGRADE START");

    uint32_t size;
    uint8_t checksum[MD5_DIGEST_SIZE];
    
    uint8_t major_version = message->data[0];
    uint8_t minor_version = message->data[1];
    memcpy(&size, &message->data[2], 4);
    
    // Extract the checksum from the message
    memcpy(checksum, &message->data[6], MD5_DIGEST_SIZE);
    bool override = (bool)message->data[6 + MD5_DIGEST_SIZE];
    
    logger_log_debugf(123, "UPGRADE INFO ver:%d.%d size:%d ovr:%s  --", 
            major_version, minor_version,
            size, 
            override == 0 ? "FALSE" : "TRUE");
    logger_log_debugf(123, "UPGRADE Checksum %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d --", 
            checksum[0],
            checksum[1],
            checksum[2],
            checksum[3],
            checksum[4],
            checksum[5],
            checksum[6],
            checksum[7],
            checksum[8],
            checksum[9],
            checksum[10],
            checksum[11],
            checksum[12],
            checksum[13],
            checksum[14],
            checksum[15]
            );

    UPGRADE_METADATA_LOAD_ID result = upgrade_metadata_prepare_upload(major_version, 
            minor_version, 
            size, 
            checksum, 
            override);
    
    message->data_length = sizeof(UPGRADE_METADATA_LOAD_ID);
    memcpy(message->data, &result, sizeof(UPGRADE_METADATA_LOAD_ID));
    
    UPGRADE_METADATA_UPLOAD_STATUS status;
    
    if (result > 0)
    {
        status = UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
        memcpy(message->data + message->data_length, &status, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
        message->response = STATUS_UPDATE;      
    }
    else
    {
        status = UPGRADE_METADATA_UPLOAD_STATUS_START_INVALID;
        memcpy(message->data + message->data_length, &status, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
        message->response = COMMAND_NOT_RECOGNISED;
    }
    
    message->data_length += sizeof(UPGRADE_METADATA_UPLOAD_STATUS);
    
    app_commands_send_message(message);
}

/**-----------------------------------------------------------------------------
 * Callback invoked when a packet of data has been uploaded into the image data 
 * @param request
 */
void app_commands_upgrade_write_callback(UPGRADE_METADATA_LOAD_ID load_id, FLASH_MEMORY_CLIENT_REQUEST* request)
{
    logger_suppress_command_logging(true);
    
    //logger_log_debugf(0, "app_commands_upgrade_write_callback %d --", load_id);
    STRUCTURED_MESSAGE message;
    structured_message_initialise(&message, STRUCTURED_MESSAGE_SOURCE_RS232);
    message.command = APP_COMMANDS_UPGRADE_LOAD_DATA;
    message.action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
    
    // Add the Id to the message
    UPGRADE_METADATA_UPLOAD_STATUS status = UPGRADE_METADATA_UPLOAD_STATUS_SUCCESS;
    memcpy(message.data, &load_id, sizeof(UPGRADE_METADATA_LOAD_ID));
    memcpy(message.data + sizeof(UPGRADE_METADATA_LOAD_ID), &status, sizeof(UPGRADE_METADATA_UPLOAD_STATUS));
    message.data_length = sizeof(UPGRADE_METADATA_LOAD_ID) + sizeof(UPGRADE_METADATA_UPLOAD_STATUS);
    message.response = STATUS_UPDATE;      
    
    app_commands_send_message(&message);   
    
    logger_suppress_command_logging(false);
}

/* *****************************************************************************
 End of File
 */
