/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_context.h"
#include "app_config.h"
#include "app_commands_display.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    //! Indicated that the module has said its up and running app_context_module_running())
    bool running;
} APP_MODULE_RUNTIME;

typedef struct
{
    // The current state active state of the app - decides whether were in a menu or not
    APP_CONTEXT_MODE mode;
    
    // Runtime module information
    APP_MODULE_RUNTIME module[APP_MODULE_MAX];
    
    //! Main application state machine state
    APP_STATE state;

} APP_CONTEXT_DATA;

static APP_CONTEXT_DATA app_context_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Resets the running state on Wake-Up
 */
void app_context_clear_running()
{
    logger_log_debugf(0, "APP CONTEXT Clear Running State --");
    int x;
    for (x = 0; x < APP_MODULE_MAX; x++)
    {
        app_context_data.module[x].running = false;
    }
}

/**-----------------------------------------------------------------------------
 * Returns the current application state
 * @return 
 */
APP_STATE app_context_get_app_state()
{
    return app_context_data.state;
}
    
/**-----------------------------------------------------------------------------
 * Gets the current mode - modes are used during the running state to identify 
 * what type of processing is underway (normal, tests, processes)
 */
APP_CONTEXT_MODE app_context_get_mode()
{
    return app_context_data.mode;
}    

/**-----------------------------------------------------------------------------
 * Initialise the application context
 */
void app_context_initialise()
{
    app_context_set_app_state(APP_STATE_INIT);
    app_context_clear_running();
}
    
/**-----------------------------------------------------------------------------
 * Can be used to determine whether all the initialisation is complete.  This
 * is different to checking the runtime state as its possible for the state 
 * to be RUNNING while initialisation is still finishing off (INPUT_CONTROL)
 * @return 
 */
bool app_context_is_running()
{
    bool result = true;
    
    int x;
    for (x = 0; x < APP_MODULE_MAX; x++)
    {
        // If the module has a startup dependency and its not running return false
        if (APP_MODULE_STARTUP_DEPENDENCY[x] &&
            !app_context_data.module[x].running)
        {
            result = false;
            break;
        }
    }
    
    return result;
}   
    
/**-----------------------------------------------------------------------------
 * Used to set the given module into a stopped state - used by app_context_is_running()
 * @param module
 */
void app_context_module_reset(APP_MODULE module)
{
    app_context_data.module[module].running = false;
}
    
/**-----------------------------------------------------------------------------
 * Used to determine that the given module is running - used by app_context_is_running()
 * @param module
 */
void app_context_module_running(APP_MODULE module)
{
    logger_log_debugf(0, "APP CONTEXT Module running %d --", module);
    app_context_data.module[module].running = true;
}
    
/**-----------------------------------------------------------------------------
 * Called prior to message processing
 * @param messaage - The message to be processed
 */
void app_context_pre_process_message(STRUCTURED_MESSAGE* message)
{
    // Reset the remote triggered brightness flag
    if (message->command != STRUCTURED_MESSAGE_ACTION_GET_STATUS)
    {
        app_commands_brightness_set_remote_change(false);
    }
    
    if (message->source == STRUCTURED_MESSAGE_SOURCE_REMOTE)
    {
        // BUG-20171214-1
        if (message->command == APP_COMMANDS_BRIGHTNESS)
        {
            app_commands_brightness_set_remote_change(true);
        }
        
        // Deviation List Issue 15
        if (message->command == APP_COMMANDS_POWER)
        {
            app_commands_brightness_set_remote_change(true);
        }
    }
}

/**-----------------------------------------------------------------------------
 * Sets the current application state
 * @param state - The application state to change to 
 */
APP_STATE app_context_set_app_state(APP_STATE state)
{
    app_context_data.state = state;
}

/**-----------------------------------------------------------------------------
 * This should be used to directly jump to a menu when not using the standard 
 * menu navigation functions (left/right/up/down)
 * @param menu_id
 */
void app_context_set_menu(MENU_ID menu_id)
{
    logger_log_debugf(0, "MENU Displaying Menu %d %s --", menu_id, app_menu_titles[menu_id]);
    menu_set_current_item(menu_id);
}
    
/**-----------------------------------------------------------------------------
 * Sets the current mode
 */
 void app_context_set_mode(APP_CONTEXT_MODE mode)
{
    if (app_context_data.mode != mode)
    {
        app_context_data.mode = mode;

        logger_log_infof(0, "APP: Entering mode %d --", mode);

        if (app_context_data.mode != APP_CONTEXT_MODE_DEFAULT)
        {
            timers_reset(TIMER_INDEX_MENU_INACTIVE);
        }
        else
        {
            // Make sure its active as we suspend it when in Engineering Menu
            timers_activate(TIMER_INDEX_MENU_INACTIVE);
        }
    }
}   

/* *****************************************************************************
 End of File
 */
