/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_device_configuration.h"
#include "app_config.h"
#include "app_commands_menu.h"
#include "app_commands_power.h"
#include "utilities/app_commands_utilities.h"
#include "utilities/dirac.h"
#include "utilities/events.h"
#include "utilities/input_control.h"
#include "utilities/menu.h"
#include "utilities/menu_app.h"
#include "utilities/volume.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
* Used to initialise ADC inputs.  
*/
void app_device_initialise_adc()
{
    ADC_BUTTON input_buttons[ADC_MAX_BUTTON_COUNT];
    input_buttons[0] = APP_BUTTONS_MENU;
    input_buttons[1] = APP_BUTTONS_NONE;
    input_buttons[2] = APP_BUTTONS_NONE;
    input_buttons[3] = APP_BUTTONS_NONE;
    adc_button_add_input(APP_ADC_INPUTS_BUTTONS_1, input_buttons, ADC_INPUT_STATE_ACTIVE);

    input_buttons[0] = APP_BUTTONS_INPUT_DOWN;
    input_buttons[1] = APP_BUTTONS_INPUT_UP;
    input_buttons[2] = APP_BUTTONS_DIRAC;
    input_buttons[3] = APP_BUTTONS_MUTE;
    adc_button_add_input(APP_ADC_INPUTS_BUTTONS_2, input_buttons, ADC_INPUT_STATE_ACTIVE);

    input_buttons[0] = APP_BUTTONS_INFO;
    input_buttons[1] = APP_BUTTONS_DIRECT;
    input_buttons[2] = APP_BUTTONS_DISPLAY;
    input_buttons[3] = APP_BUTTONS_BALANCE;
    adc_button_add_input(APP_ADC_INPUTS_BUTTONS_3, input_buttons, ADC_INPUT_STATE_ACTIVE);
    
    adc_voltage_add_input(APP_ADC_INPUTS_BIAS_TEST_L, ADC_INPUT_STATE_INACTIVE);
    adc_voltage_add_input(APP_ADC_INPUTS_BIAS_TEST_R, ADC_INPUT_STATE_INACTIVE);
    adc_temperature_add_input(APP_ADC_INPUTS_OUTPUT_TEMPERATURE, ADC_INPUT_STATE_ACTIVE);
    adc_temperature_add_input(APP_ADC_INPUTS_LIFTER_TEMPERATURE, ADC_INPUT_STATE_ACTIVE);
}
    
/**-----------------------------------------------------------------------------
 * Returns true if the message has been processed
 */
bool app_device_process_buttons_default(APP_BUTTONS button, STRUCTURED_MESSAGE* message)
{
    bool result = true;

    switch (button)
    {
        case APP_BUTTONS_MENU:
        {
            logger_log_debugf(0, "BUTTON Menu --");
            app_context_set_mode(APP_CONTEXT_MODE_MENU);

            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU;
            message->data[0] = APP_COMMAND_MENU_SYSTEM_SETTINGS;
            break;
        }

        case APP_BUTTONS_INPUT_DOWN:
        {
            if (!input_control_in_switch())
            {
                INPUT_CONTROL_PREVIEW_INFO* info = input_control_get_preview_info();
                message->action = STRUCTURED_MESSAGE_ACTION_SET_DECREMENT;
                message->command = APP_COMMANDS_PREVIEW_SOURCE;
                message->data[0] = info->input;
                message->data_length = 1;
            }
            else
            {
                result = false;
            }

            break;
        }

        case APP_BUTTONS_INPUT_UP:
        {
            if (!input_control_in_switch())
            {
                INPUT_CONTROL_PREVIEW_INFO* info = input_control_get_preview_info();
                message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                message->command = APP_COMMANDS_PREVIEW_SOURCE;
                message->data[0] = info->input;
                message->data_length = 1;
            }
            else
            {
                result = false;
            }

            break;
        }
            
        case APP_BUTTONS_INFO:
        {
            // Show Info menu in edit mode
            if ((menu_get_current_item()->menu_id == APP_COMMAND_MENU_INFO) ||
                (input_control_get_type() != INPUT_CONTROL_TYPE_AVATAR))
            {
                AUDIO_INPUT selected_input = input_control_get_selected_input();
                if (selected_input == AUDIO_INPUT_PHONO)
                {
                    logger_log_debugf(0, "BUTTON Info->PHONO --");
                    app_context_set_mode(APP_CONTEXT_MODE_DEFAULT);
                    message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                    message->command = APP_COMMANDS_PHONO_INPUT_MM_MC;
                    message->data[0] = 0x00;
                    message->data_length = 1;

                }
                else
                {
                    logger_log_debugf(0, "BUTTON Info->Default --");
                    app_commands_display_default_screen(message);
                }
            }
            else
            {
                logger_log_debugf(0, "BUTTON Info --");
                display_scroll_line();
                result = false;
            }
            
            break;
        }

        case APP_BUTTONS_DISPLAY:
        {
            logger_log_debugf(0, "BUTTON Display --");
            app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU;
            message->data[0] = APP_COMMAND_MENU_DISPLAY_NO_NAVIGATION;
            message->data_length = 1;
            break;
        }

        case APP_BUTTONS_BALANCE:
        {
            // Show balance menu in edit mode
            logger_log_debugf(0, "BUTTON Balance --");
            app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU;
            message->data[0] = APP_COMMAND_MENU_BALANCE_NO_NAVIGATION;
            message->data_length = 1;
            break;
        }

        default:
        {
            result = false;
            break;
        }
    }    
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Returns true if the message has been processed.  This assumes that the button has been identified,
 * there is a change
 */
bool app_device_process_buttons_menu(APP_BUTTONS button, STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    switch(button)
    {
        case APP_BUTTONS_MENU:
        case APP_BUTTONS_HOME:
        {
            // Return to default state
            app_commands_display_default_screen(message);

            break;
        }

        case APP_BUTTONS_UP_LEFT:  // And APP_BUTTONS_LEFT
        {
            // Move to the next sibling
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU_NAVIGATION;
            message->data[0] = MENU_NAVIGATION_LEFT;
            break;
        }

        case APP_BUTTONS_DOWN_RIGHT:  // And APP_BUTTONS_RIGHT
        {
            // Move to the previous sibling
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU_NAVIGATION;
            message->data[0] = MENU_NAVIGATION_RIGHT;
            break;
        }

        case APP_BUTTONS_OK:
        {
            // Move to the first child in the menu
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU_NAVIGATION;
            message->data[0] = MENU_NAVIGATION_OK;
            break;
        }

        case APP_BUTTONS_BACK:
        {
            // Return to the parent menu
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU_NAVIGATION;
            message->data[0] = MENU_NAVIGATION_UP;
            break;
        }
                
        default:
        {
            result = false;
            break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns true if the message has been processed.  This assumes that the button has been identified,
 * there is a change
 */
bool app_device_process_buttons_menu_edit(APP_BUTTONS button, STRUCTURED_MESSAGE* message)
{
    bool result = true;
    
    switch(button)
    {
        case APP_BUTTONS_MENU:
        {
            // Return to default state
            app_commands_display_default_screen(message);
            break;
        }

        case APP_BUTTONS_HOME:
        {
            // Increment the Dirac curve if the DIRAC_NO_NAVIGATION menu is being displayed
            MENU_ITEM* menu = menu_get_current_item();
            if (menu->menu_id == APP_COMMAND_MENU_DIRAC_NO_NAVIGATION)
            {
                message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                message->command = APP_COMMANDS_ROOM_EQUALISATION;
                message->data[0] = DIRAC_NO_CURVE; //NOT USEDdirac_get_selected_curve_index();
                message->data_length = 1;            
            }
            else
            {
                // Return to default state
                app_commands_display_default_screen(message);
            }
            break;
        }

        case APP_BUTTONS_UP_LEFT:  // And APP_BUTTONS_LEFT
        {
            // Navigate Left
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU_NAVIGATION;
            message->data[0] = MENU_NAVIGATION_LEFT;
            break;
        }

        case APP_BUTTONS_DOWN_RIGHT:  // And APP_BUTTONS_RIGHT
        {
            // Navigate Right
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU_NAVIGATION;
            message->data[0] = MENU_NAVIGATION_RIGHT;
            break;
        }

        case APP_BUTTONS_OK:
        {
            MENU_ITEM* menu = menu_get_current_item();
            if (menu->menu_id == APP_COMMAND_MENU_ANALOGUE_DIRECT_NO_NAVIGATION)
            {
                if (input_control_get_current_input_type() == INPUT_CONTROL_TYPE_ANALOGUE)
                {
                    message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
                    message->command = APP_COMMANDS_ANALOGUE_DIRECT; // Will actually Toggle
                    message->data[0] = input_control_get_selected_input();
                    //message->data[1] = !input_control_get_analogue_direct();
                    message->data_length = 1;
                }
                else
                {
                    app_commands_display_default_screen(message);
                }
            }
            else
            {
                // Move to the first child in the menu
                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                message->command = APP_COMMANDS_MENU_NAVIGATION;
                message->data[0] = MENU_NAVIGATION_OK;
            }
            break;
        }

        case APP_BUTTONS_BACK:
        {
            MENU_ITEM* menu = menu_get_current_item();
            if (menu->menu_id == APP_COMMAND_MENU_DISPLAY_NO_NAVIGATION) //is_balance_menu(menu))
            {
                app_commands_display_default_screen(message);
            }
            else
            {
                // Return to the parent menu
                message->command = APP_COMMANDS_MENU_NAVIGATION;
                message->data[0] = MENU_NAVIGATION_UP;
            }
            break;
        }

        case APP_BUTTONS_BALANCE:
        {
            MENU_ITEM* menu = menu_get_current_item();
            if (menu->menu_id == APP_COMMAND_MENU_BALANCE_NO_NAVIGATION) //is_balance_menu(menu))
            {
                app_commands_display_default_screen(message);
            }
        }

        default:
        {
            result = false;
            break;
        }
    }
    
    return result;
}

/* *****************************************************************************
 End of File
 */
