/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "app_device_configuration.h"
#include "app_config.h"
#include "app_commands_menu.h"
#include "app_commands_power.h"
#include "utilities/app_commands_utilities.h"
#include "utilities/events.h"
#include "utilities/menu.h"
#include "utilities/volume.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Event raised when there is an Error detected at the Hardware level
 */
static void app_device_pot_data_available_callback(EVENT* event)
{
	// Do something event related...
    ADC_POT_DATA_AVAILABLE_EVENT_DATA* event_data = (ADC_POT_DATA_AVAILABLE_EVENT_DATA*)event->context_data;
    
    if (event_data->input == APP_ADC_INPUTS_BALANCE)
    {
        if (!event_data->initialised)
        {
            // initialise balance at volume
            //volume_set_balance(VOLUME_DEFAULT_BALANCE + (event_data->current_step - (event_data->step_count / 2)), false);
            volume_set_balance(VOLUME_DEFAULT_BALANCE + adc_pot_get_offset_from_center(APP_ADC_INPUTS_BALANCE), false);
 
            // initialise balance at input_context
            //TODO REMOVE- DONT THINK THIS IS NEEDED input_context_set_balance_changed(event_data->has_changed);
        }
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
* Used to initialise ADC inputs.  
*/
void app_device_initialise_adc()
{
    // Subscribe to the events raise in the HW layer.  Note the Event Module ID should
    // work for either HW type 
    events_subscribe(EVENT_MODULE_ADC_POT, 
                 ADC_EVENT_POT_DATA_AVAILABLE, 
                 app_device_pot_data_available_callback, 
                 (void*)NULL, 
                 true);

    adc_pot_add_input(APP_ADC_INPUTS_BALANCE, 25, ADC_INPUT_STATE_ACTIVE);
 
    ADC_BUTTON input_buttons[ADC_MAX_BUTTON_COUNT];
    input_buttons[0] = APP_BUTTONS_DIRAC;
    input_buttons[1] = APP_BUTTONS_PHONO_TYPE;
    input_buttons[2] = APP_BUTTONS_DIRECT;
    input_buttons[3] = APP_BUTTONS_MUTE;
    adc_button_add_input(APP_ADC_INPUTS_BUTTONS_2, input_buttons, ADC_INPUT_STATE_ACTIVE);

    input_buttons[0] = APP_BUTTONS_STANDBY;
    input_buttons[1] = APP_BUTTONS_NONE;
    input_buttons[2] = APP_BUTTONS_NONE;
    input_buttons[3] = APP_BUTTONS_NONE;
    adc_button_add_input(APP_ADC_INPUTS_BUTTONS_3, input_buttons, ADC_INPUT_STATE_ACTIVE);
    
    adc_voltage_add_input(APP_ADC_INPUTS_BIAS_TEST_L, ADC_INPUT_STATE_INACTIVE);
    adc_voltage_add_input(APP_ADC_INPUTS_BIAS_TEST_R, ADC_INPUT_STATE_INACTIVE);
    adc_temperature_add_input(APP_ADC_INPUTS_OUTPUT_TEMPERATURE, ADC_INPUT_STATE_ACTIVE);
    adc_temperature_add_input(APP_ADC_INPUTS_LIFTER_TEMPERATURE, ADC_INPUT_STATE_ACTIVE);
}
    
/**-----------------------------------------------------------------------------
 * Returns true if the message has been processed
 */
bool app_device_process_buttons_default(APP_BUTTONS button, STRUCTURED_MESSAGE* message)
{
    bool result = true;

    switch (button)
    {
        case APP_BUTTONS_PHONO_TYPE:
        {
            logger_log_debugf(0, "BUTTON Phono Type --");
            app_context_set_mode(APP_CONTEXT_MODE_MENU_EDIT);
            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_MENU;
            message->data[0] = APP_COMMAND_MENU_PHONO_INPUT_NO_NAVIGATION;
            message->data_length = 1;
            break;
        }

        case APP_BUTTONS_STANDBY:
        {
            logger_log_debugf(0, "BUTTON Standby --");

            message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
            message->command = APP_COMMANDS_POWER;
            message->data[0] = APP_COMMAND_POWER_TOGGLE;   
            message->data_length = 1;
            break;
        }
            
        default:
        {
            result = false;
            break;
        }
    }    
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Returns true if the message has been processed.  This assumes that the button has been identified,
 * there is a change
 */
bool app_device_process_buttons_menu(APP_BUTTONS button, STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    switch(button)
    {
        case APP_BUTTONS_PHONO_TYPE:
        {
            MENU_ID menu_id = menu_get_current_item()->menu_id;

            if ((menu_id == APP_COMMAND_MENU_SYSTEM_RESET) || (menu_id == APP_COMMAND_MENU_SYSTEM_RESET_CONFIRM))
            {
                message->action = STRUCTURED_MESSAGE_ACTION_SET_ABSOLUTE;
                message->command = APP_COMMANDS_MENU_NAVIGATION;
                message->data[0] = APP_COMMAND_MENU_NAVIGATION_OK;

                result = true;
            }

            break;
        }

        case APP_BUTTONS_DIRECT:
        {
            MENU_ID menu_id = menu_get_current_item()->menu_id;

            if (menu_id == APP_COMMAND_MENU_SYSTEM_SETTINGS)
            {
                app_commands_launch_engineering_menu(message);
            }
            else
            {
                app_commands_display_default_screen(message);
                result = true;
            }
        }
                
        default:
        {
            result = false;
            break;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Returns true if the message has been processed.  This assumes that the button has been identified,
 * there is a change
 */
bool app_device_process_buttons_menu_edit(APP_BUTTONS button, STRUCTURED_MESSAGE* message)
{
    bool result = false;
    
    switch(button)
    {
        case APP_BUTTONS_PHONO_TYPE:
        {
            logger_log_debugf(0, "BUTTON PHONO type --");
                
            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
            message->command = APP_COMMANDS_PHONO_INPUT_MM_MC;
            message->data[0] = input_control_get_selected_input();
            message->data_length = 1;
            
            result = true;
            break;
        }
        
        case APP_BUTTONS_DIRECT:
        {
            logger_log_debugf(0, "BUTTON Direct --");
            message->action = STRUCTURED_MESSAGE_ACTION_SET_INCREMENT;
            message->command = APP_COMMANDS_ANALOGUE_DIRECT;
            message->data[0] = input_control_get_selected_input();
            message->data_length = 1;
            
            result = true;
            break;
        }

        default:
        {
            result = false;
            break;
        }
    }
    
    return result;
}

/* *****************************************************************************
 End of File
 */
