/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    adc.c

  @Summary
    ADC Implementation - PICMZ version.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/adc.h"
#include "system_definitions.h"
#include "utilities/volume.h"
#include <math.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

// Center value for balance
#define BALANCE_CENTER  12

// Structure to hold adc values for the Balance control 
typedef struct
{
    uint16_t low_limit;
    uint16_t high_limit;
} ADC_BALANCE_LIMIT;

// These are un-scaled values, calculated from balance.xlsx for 25% tolerance from nominal values
static ADC_BALANCE_LIMIT balance_limits[] = 
{
    {0, 102}, // 0
    {225, 266}, // 1
    {389, 430}, // 2
    {553, 594}, // 3
    {717, 758}, // 4
    {880, 921}, // 5
    {1044, 1085}, // 6
    {1208, 1249}, // 7
    {1372, 1413}, // 8
    {1536, 1577}, // 9
    {1699, 1740}, // 10
    {1863, 1904}, // 11
    {2027, 2068}, // 12
    {2191, 2232}, // 13
    {2355, 2396}, // 14
    {2518, 2559}, // 15
    {2682, 2723}, // 16
    {2846, 2887}, // 17
    {3010, 3051}, // 18
    {3174, 3215}, // 19
    {3337, 3378}, // 20
    {3501, 3542}, // 21
    {3665, 3706}, // 22
    {3829, 3870}, // 23
    {3993, 4095} // 24
};
    
// If we scale by 10, we can avoid fractional values
#define ADC_BALANCE_MAX             (4095) // 12 bit resolution 
    
// There are 25 different zones; -12...0...+12
// The centre of the first zone is 1/50 of the full travel
// The zero-based index of the zone therefore, can be found from:
// (where centre = 12)
#define ADC_BALANCE_ZONE(RDG)               (((RDG) * 25) / (ADC_BALANCE_MAX + 1))

/**
 * Enum to show whether calibration mode is on/off
 * Note the enum values correspond to the AD1CON2bits.OFFCAL register bit
 */
typedef enum
{
    ADC_CALIBRATION_MODE_OFF = 0,           
    ADC_CALIBRATION_MODE_ON = 1,
} ADC_CALIBRATION_MODE;

typedef struct 
{
    /**
     * Indicates whether the calibration mode is on/off
     */
    ADC_CALIBRATION_MODE adc_calibration_mode[ADC_MAX_SCAN_INPUTS];
    
    /**
     * The number of scan inputs - this is initially set to ADC_MAX_SCAN_INPUTS
     * but could be made dynamic if necessary?
     */
    unsigned int adc_input_count;
    
    /**
     * An array of offsets on a scan buffer calculated during calibration
     */
    uint32_t adc_offsets[ADC_MAX_SCAN_INPUTS];
    
    /**
     * An array of the values read from each scan buffer
     */
    unsigned int adc_values[ADC_MAX_SCAN_INPUTS];
    
    /**
     * An array containing input definitions
     */
    ADC_SCAN_INPUT_DEFINITION adc_input_defs[ADC_MAX_SCAN_INPUTS];
    
    /**
     * Set once the adc has been started (i.e. read successfully once)
     */
    bool adc_started;

    /**
     * The currently pressed switch on any ADC Channel
     */
    ADC_SWITCH_INFO last_switch[ADC_MAX_SCAN_INPUTS];

    /**
     * Info on the currently selected switch
     */
    ADC_SWITCH_INFO current_switch;

    /**
     * Info on the previously selected switch
     * This stores the switch that was depressed before the current switch - 
     * will normally be set to ADC_SWITCH_NONE when a switch is being pressed, 
     * or ADC_SWITCH_??? when the current switch is ADC_SWITCH_NONE.
     */
    ADC_SWITCH_INFO previous_switch;
    
} ADC_DATA;

static ADC_DATA adc_data = {
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations

/**-----------------------------------------------------------------------------
 * Calibrate the ADC
 */
static void adc_calibrate(int input_index)
{
    adc_data.adc_offsets[input_index] = DRV_ADC_SamplesRead(adc_data.adc_input_defs[input_index].input);
            
    logger_log_infof(LOGGER_CODE_ADC, "ADC: Calibrating input %d %d --", input_index, adc_data.adc_offsets[input_index]);
    
    if (ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MX)
    {
#ifdef __PIC32MX__        
        // disable calibration mode
        AD1CON2bit.OFFCAL = 0;
#endif
    }

    adc_data.adc_calibration_mode[input_index] = ADC_CALIBRATION_MODE_OFF;
}

/**-----------------------------------------------------------------------------
 * Gets the raw value of the supplied input.
 * @param input_index - The index of the input to be reported on
 * @return - The raw value read from the pin
 * it returns 0
 */
static uint32_t adc_get_raw_value(unsigned int input_index)
{
    uint32_t result = ADC_INVALID_REQUEST;
   
    if (input_index < adc_data.adc_input_count)
    {
        result = adc_data.adc_values[input_index]; 
        logger_log_debugf(LOGGER_CODE_DEBUG, "ADC: %d --", adc_data.adc_values[input_index]);
    }
    
    return result;    
}

/**-----------------------------------------------------------------------------
 * 
 */
static void adc_process_pot(int input_index)
{
    uint32_t raw_value = adc_get_raw_value(input_index); 

    ADC_SCAN_INPUT_DEFINITION* def = &adc_data.adc_input_defs[input_index];
    
    if (!def->input_data.pot.initialised)
    {
        def->input_data.pot.initialised = true;
        
        def->input_data.pot.value = ADC_BALANCE_ZONE(raw_value);
        
        // initialise balance at volume
        volume_set_balance(VOLUME_DEFAULT_BALANCE + (def->input_data.pot.value - BALANCE_CENTER), false);
        
        // initialise balance at input_context
//TODO?????????????????????????????????        initialise_balance_port_offset(def->input_data.pot.value - BALANCE_CENTER);
    }
    else // Normal reading
    {
        // calculate the new balance in case we've moved more than one zone
        uint8_t new_balance = ADC_BALANCE_ZONE(raw_value);
        if (new_balance != def->input_data.pot.value)
        {
            // balance changed
            if (new_balance > def->input_data.pot.value)
            {
                if ((new_balance - def->input_data.pot.value) <= 1)
                {
                     // one or zero zone change
                    new_balance = def->input_data.pot.value; // use the current value by default
                }
            }
            else // new_balance <= zone
            {
                if ((def->input_data.pot.value - new_balance) <= 1)
                {
                    // one or zero zone change
                    new_balance = def->input_data.pot.value; // use the current value by default
                }
            }
        }
        
        if (new_balance == def->input_data.pot.value) // no change or +- 1 zone change
        {
            // have we strayed into another zone?
            // is there a higher (more right) zone?
            if (def->input_data.pot.value < 24)
            {
                if (raw_value >= balance_limits[def->input_data.pot.value + 1].low_limit)                
                {
                    // we've moved into a higher zone
                    new_balance = def->input_data.pot.value + 1;
                }
                else // we haven't moved into a higher zone
                {
                    // is there a lower (more left) zone?
                    if (def->input_data.pot.value > 0) 
                    {
                        if (raw_value <= balance_limits[def->input_data.pot.value - 1].high_limit)
                        {
                            // we've moved into a lower (more left) zone
                            new_balance = def->input_data.pot.value - 1;
                        }
                    }
                }
            } // zone < 24
        } // no change or +- 1 zone change
        
        def->input_data.pot.value = new_balance;
    }   
}

/**-----------------------------------------------------------------------------
 * Process switch mode for an input and determine if switch is pressed
 * 
 * @param input_index - The index of the scan input
 */
static void adc_process_switches(int input_index)
{
    // Convert to mV
    // 10 bits = 1024 steps (0 to 1023) 
    // Result_mV = sample_value * (Reference voltage in mV / 1023)
    uint64_t mV = (adc_data.adc_values[input_index] * 3300) / ADC_DATA_RESOLUTION;            

    ADC_SCAN_INPUT_DEFINITION def = adc_data.adc_input_defs[input_index];
    
    uint32_t ticks = SYS_TMR_TickCountGet();
    bool active = false;
    
    ADC_SWITCH active_scan_switch = ADC_SWITCH_NONE;
    
    uint16_t mV16 = mV;
    int x;
    for (x = 0; x < ADC_MAX_SCAN_INPUT_SWITCHES; x++)
    {
        if ((mV16 >= def.input_data.input_switch.switches[x].range_low) && (mV16 <= def.input_data.input_switch.switches[x].range_high))
        {
            adc_data.current_switch.adc_switch = def.input_data.input_switch.switches[x].adc_switch;
            active = true;
            //logger_log_debugf(0, "ADC Switch active %d %d %d %d %d ", input_index, x, def.switches[x].range_low, mV16, def.switches[x].range_high);
            
            // keep track of the button being selected
            active_scan_switch = adc_data.current_switch.adc_switch;
            break;
        }
    }

    if ((adc_data.current_switch.adc_switch == ADC_SWITCH_NONE) ||
        (adc_data.last_switch[input_index].adc_switch == adc_data.current_switch.adc_switch) ||
        (active_scan_switch == ADC_SWITCH_NONE))
    {
        //logger_log_debugf(123, "ADC REVIEWING SWITCH: %d --", adc_data.last_switch[input_index].adc_switch);
        // Update the duration
        adc_data.last_switch[input_index].duration_ticks = ticks - adc_data.last_switch[input_index].start_ticks;
        
        // Now review the last_switch - we only keep it open for 3 seconds
        if ((adc_data.last_switch[input_index].adc_switch != ADC_SWITCH_NONE) &&
            ((adc_data.last_switch[input_index].duration_ticks > (SYS_TMR_TickCounterFrequencyGet() * 1))))
        {
            //logger_log_debugf(123, "ADC CLEARING SWITCH: %d --", adc_data.last_switch[input_index].adc_switch);
            adc_data.last_switch[input_index].adc_switch = ADC_SWITCH_NONE;
            adc_data.last_switch[input_index].start_ticks = 0;
            adc_data.last_switch[input_index].duration_ticks = 0;
        }
    }
    else if ((adc_data.current_switch.adc_switch != ADC_SWITCH_NONE) && active)
    {
        // Change to the new switch
        adc_data.last_switch[input_index].adc_switch = adc_data.current_switch.adc_switch;
        adc_data.last_switch[input_index].start_ticks = ticks;
        adc_data.last_switch[input_index].duration_ticks = 0;
        //logger_log_debugf(123, "ADC SETTING SWITCH: %d --", adc_data.last_switch[input_index].adc_switch);
    }
    
    // Now check for errors
    // Note this is only set if a scan of a previous input has returned ADC_SWITCH_SW_NONE
    if ((adc_data.current_switch.adc_switch == ADC_SWITCH_NONE) && 
        (mV < def.input_upper)) 
    {
        adc_data.current_switch.adc_switch = ADC_SWITCH_ERROR;
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Returns a pointer to an array of ADC_SWITCH_INFO structs that contain information on the 
 * current state of a switch for each ADC channel. Where an ADC input is not 
 * a switch then the adc_switch member of the struct is ADC_SWITCH_ERROR.
 * This is intended to enable the detection of multiple concurrent switch depressions,
 * combined with "sticky" switch functionality meaning a switch remains active for
 * a second after being pressed.
 * @return 
 */
ADC_SWITCH_INFO* adc_get_switches()
{
    return adc_data.last_switch;
}   
    
/**-----------------------------------------------------------------------------
 * Gets the currently selected switch
 * @return 
 */
ADC_SWITCH adc_get_current_switch()
{
    return adc_data.current_switch.adc_switch;
}
     
/**-----------------------------------------------------------------------------
 * Gets Whether the input is currently active or not 
 * @return - Whether the input is currently active or not 
 */
//bool adc_get_input_active(unsigned int input_index)
//{
//    bool result = false;
//   
//    if (input_index < adc_data.adc_input_count)
//    {
//        result = adc_data.adc_input_defs[input_index].active;
//    }
//    
//    return result;    
//}
    
/**-----------------------------------------------------------------------------
 * Gets the value associated with the voltage on a specific channel.
 * @return - The current value based on the channel voltage.  If the input is 
 * not within the ranges identified it returns
 * it returns ADC_SWITCH_ERROR
 */
ADC_SWITCH adc_get_input_discrete(unsigned int input_index)
{
    ADC_SWITCH result = ADC_SWITCH_ERROR;
    
    if ((input_index < adc_data.adc_input_count) &&
        (adc_data.adc_input_defs[input_index].mode == ADC_SCAN_INPUT_MODE_DISCRETE))
    {
        uint64_t mV = (adc_data.adc_values[input_index] * 3300) / ADC_DATA_RESOLUTION;            
//logger_log_debugf(0, "ADC DISCREET %d mV:%d ", input_index, mV);    

        ADC_SCAN_INPUT_DEFINITION def = adc_data.adc_input_defs[input_index];
        int x;
        for (x = 0; x < ADC_MAX_SCAN_INPUT_SWITCHES; x++)
        {
            // Ignore entries where low and high are equal
            if ((def.input_data.input_switch.switches[x].range_low != def.input_data.input_switch.switches[x].range_high) &&
                (mV >= def.input_data.input_switch.switches[x].range_low) && (mV <= def.input_data.input_switch.switches[x].range_high))
            {
                result = def.input_data.input_switch.switches[x].adc_switch;
                break;
            }
        }
    }
    
    return result;
}
    
/**-----------------------------------------------------------------------------
 * Gets the type of ADC input 
 * @return - Whether ADC input mode (Buttons, Temp, etc..)
 */
ADC_SCAN_INPUT_MODE adc_get_input_mode(unsigned int input_index)
{
    bool result = false;
   
    if (input_index < adc_data.adc_input_count)
    {
        result = adc_data.adc_input_defs[input_index].mode;
    }
    
    return result;    
}
     
/**-----------------------------------------------------------------------------
 * Gets Whether the input is currently active or not 
 * @return - Whether the input is currently active or not 
 */
bool adc_get_input_processed(unsigned int input_index)
{
    bool result = false;
   
    if (input_index < adc_data.adc_input_count)
    {
        result = adc_data.adc_input_defs[input_index].processed;
    }
    
    return result;    
}
   
/**-----------------------------------------------------------------------------
 * Gets the calculated temperature on the supplied input.
 * @param input_index - The index of the input to be reported on
 * @return - The calculated temperature to 1 decimal place e.g. 45.1degrees will be returned as 451.  
 *           If the input != ADC_SCAN_INPUT_MODE_TEMPERATURE it returns 0
 */
int16_t adc_get_input_temperature(unsigned int input_index)
{
    int16_t result = ADC_INVALID_REQUEST;
    
    if ((input_index < adc_data.adc_input_count) &&
        (adc_data.adc_input_defs[input_index].mode == ADC_SCAN_INPUT_MODE_TEMPERATURE))
    {
        // Get the latest value - not we are not subtracting the offset, as it 
        // appears that by the time we do the calibration its already a full-running mode.
        if (DRV_ADC_SamplesAvailable(adc_data.adc_input_defs[input_index].input))
        {
            adc_data.adc_values[input_index] = DRV_ADC_SamplesRead(adc_data.adc_input_defs[input_index].input);
        }
        
        if (adc_data.adc_values[input_index] != 0)
        {
            // Calculate the delta reading
            float delta = (float)adc_data.adc_values[input_index] / (float)ADC_DATA_RESOLUTION;
            
            // Check the delta is within sensor ranges - we were getting overtemps
            // because the delta was in the order of 0.00005
            if ((delta < (float)0.9) && (delta > (float)0.1))
            {
                // Adjust the temperature for a linear temperature reading 
                // LMT89 datasheet Figure 1.
                //float temperature = (delta * (float)ADC_TEMP_RANGE * -1.0f) + (float)ADC_TEMP_OFFSET;

                // Calculate the temperature
                // LMT89 datasheet transfer function http://www.ti.com/lit/ds/symlink/lmt89.pdf
                float voltage = delta * (float)3.3;
                float temperature = (float)-1481.96 + (float)sqrt(2196200.0 + (((float)1.8639 - voltage) / (float)0.00000388));

                // Convert by factor of 10
                result = (int16_t)(temperature * 10.0);

                if (result > (APP_MAX_WORKING_TEMPERATURE * 10.0))
                {
                    logger_log_debugf(0, "TEMP MAX EXCEEDED %d %d %2.4f %2.4f --", input_index, result, temperature, delta);
                }
            }
            else
            {
                result = ADC_INVALID_INPUT;
            }
        }
        else
        {
            result = ADC_INVALID_INPUT;
        }
    }
    
    return result;
}

/**-----------------------------------------------------------------------------
 * Gets the calculated voltage on the supplied input.
 * @param input_index - The index of the input to be reported on
 * @return - The calculated voltage in mV.  
 * 
 * Note : This will return values for all input types..
 */
uint32_t adc_get_input_mV(unsigned int input_index)
{
    uint32_t result = ADC_INVALID_REQUEST;
   
    if (input_index < adc_data.adc_input_count)
    {
        // Get the latest value - not we are not subtracting the offset, as it 
        // appears that by the time we do the calibration its already a full-running mode.
        if (DRV_ADC_SamplesAvailable(adc_data.adc_input_defs[input_index].input))
        {
            adc_data.adc_values[input_index] = DRV_ADC_SamplesRead(adc_data.adc_input_defs[input_index].input);
        }
        
        // Convert to mV
        // 10 bits = 1024 steps (0 to 1023) 
        // Result_mV = sample_value * (Reference voltage in mV / 1023)
        //logger_log_debugf(12, "ADC: %d %d --", adc_data.adc_values[input_index], (adc_data.adc_values[input_index] * 3300) / 1023);
        result = ((float)adc_data.adc_values[input_index] * (float)3300.0) / (float)ADC_DATA_RESOLUTION;    
    }

    return result;    
}
    
/**-----------------------------------------------------------------------------
 * Gets the previously selected switch
 * Normally this will be ADC_SWITCH_NONE when a new switch is pressed, 
 * and ADC_SWITCH_??? when the new switch is ADC_SWITCH_NONE
 * @return 
 */
ADC_SWITCH adc_get_previous_switch()
{
    return adc_data.previous_switch.adc_switch;
}

/**-----------------------------------------------------------------------------
 * Gets the index of the Scan input that the button is associated with
 * @param button
 */
int adc_get_input_for_switch(ADC_SWITCH button)
{
    int result = -1;
    
    int x;
    int y;
    for (x = 0; x < ADC_MAX_SCAN_INPUTS; x++)
    {
        if (adc_data.adc_input_defs[x].mode == ADC_SCAN_INPUT_MODE_BUTTONS)
        {
            for (y = 0; y < ADC_MAX_SCAN_INPUT_SWITCHES; y++)
            {
                ADC_SCAN_INPUT_SWITCH_DEFINITION def = adc_data.adc_input_defs[x].input_data.input_switch.switches[y];
                if (def.adc_switch == button)
                {
                    result = x;
                    break;
                }
            }
        }
    }
    
    return  result;
}

/**-----------------------------------------------------------------------------
 * Initialise the ADC module
 * 
 * @param scanInputs - An Array of ADC_SCAN_INPUT_DEFINITION containing 
 * ADC_MAX_SCAN_INPUTS elements
 * @param num_inputs - The number of inputs passes into scanInputs
 * 
 * Harmony Configuration) i.e. e.g ADC_INPUT_SCAN_AN13 | ADC_INPUT_SCAN_AN14
 * 
 * Note: Assumes DRV_ADC_Initialize() is called in system_init.c
 */
void adc_initialise(ADC_SCAN_INPUT_DEFINITION scanInputs[], int num_inputs)
{
    // Initialise data structures and setup scanInputs value
    int scanInput = 0x0000;
    int x;
    for (x = 0; x < ADC_MAX_SCAN_INPUTS; x++)
    {
        adc_data.adc_calibration_mode[x] = ADC_CALIBRATION_MODE_OFF; // default
        
        if (ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MX)
        {
            adc_data.adc_calibration_mode[x] = ADC_CALIBRATION_MODE_ON;
            
            // enable calibration mode
#ifdef __PIC32MX__        
            AD1CON2bit.OFFCAL = 1;
#endif
        }
        else if (ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MZ)
        {
            adc_data.adc_calibration_mode[x] = ADC_CALIBRATION_MODE_OFF;
        }
        
        adc_data.adc_offsets[x] = 0;
        adc_data.adc_values[x] = 0;
        
        // Copy scan inputs into the internal array
        if (x < num_inputs)
        {
            adc_data.adc_input_defs[x] = scanInputs[x];
            scanInput |= adc_data.adc_input_defs[x].input;
            
            // Initialise last switch
            if (scanInputs[x].mode == ADC_SCAN_INPUT_MODE_BUTTONS)
            {
                adc_data.last_switch[x].adc_switch = ADC_SWITCH_NONE;
            }
            else if (scanInputs[x].mode == ADC_SCAN_INPUT_MODE_POT)
            {
                adc_data.adc_input_defs[x].input_data.pot.initialised = false;
                adc_data.last_switch[x].adc_switch = ADC_SWITCH_NONE;
            }
            else
            {
                adc_data.last_switch[x].adc_switch = ADC_SWITCH_ERROR;
            }
            adc_data.last_switch[x].start_ticks = 0;
            adc_data.last_switch[x].duration_ticks = 0;
            
            logger_log_infof(LOGGER_CODE_ADC, "ADC: Adding Input %d --", adc_data.adc_input_defs[x].input);
        }
    }
    
    adc_data.adc_started = false;
    
    // Opens the static ADC Driver
    // Note we need to open ALL the channel instances that are being used.
    // This is setup in the Harmony ADC configuration
    DRV_ADC0_Open();
    DRV_ADC1_Open();
    DRV_ADC2_Open();
    DRV_ADC3_Open();
    DRV_ADC4_Open();
    DRV_ADC5_Open();

    // Input selection
    //DRV_ADC_NegativeInputSelect(DRV_ADC_MUX_A, DRV_ADC_INPUT_NEGATIVE_VREF_MINUS);
    //DRV_ADC_PositiveInputSelect(DRV_ADC_MUX_A, DRV_ADC_POSITIVE_INPUT_NUM);

    // Setup channels to scan e.g ADC_INPUT_SCAN_AN14
    // Think this is now done in Harmony DRV_ADC_ChannelScanInputsAdd(scanInput);
    
    adc_data.adc_input_count = num_inputs; 

    // Enable the scan input on MUX A
    PLIB_ADC_MuxAInputScanEnable(DRV_ADC_ID_1);
    
    // Start the ADC
    DRV_ADC_Start();
}

/**-----------------------------------------------------------------------------
 * Perform ADC processing
 * 
 * @return - true if samples processed, false if not
 */
bool adc_process(void)
{
    bool result = false;
    bool has_stopped = false;

    // Sort out the previous switch (button))
    ADC_SWITCH previous_switch = adc_data.current_switch.adc_switch;

    // Initialise current switch to NONE
    adc_data.current_switch.adc_switch = ADC_SWITCH_NONE;
    
    int x;
    for (x = 0; x < adc_data.adc_input_count; x++)
    {
        adc_data.adc_input_defs[x].processed = false;
        
        if (DRV_ADC_SamplesAvailable(adc_data.adc_input_defs[x].input))
        {
            //logger_log_debugf(0, "ADC Samples available %d ", x);
            // Stop the ADC while we process it
            if (!has_stopped)
            {
                DRV_ADC_Stop();
                has_stopped = true;
            }

            // Check which mode we're in
            if (adc_data.adc_calibration_mode[x] == ADC_CALIBRATION_MODE_ON)
            {
                adc_calibrate(x);
            } 
            else 
            {
                if (adc_data.adc_input_defs[x].active)
                {
                    if (ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MX)
                    {
                        // subtract the calibration offset
                        adc_data.adc_values[x] = (adc_data.adc_offsets[x] - DRV_ADC_SamplesRead(adc_data.adc_input_defs[x].input));
                    }
                    else if (ARCAM_PIC_DEVICE_FAMILY == ARCAM_PIC_DEVICE_FAMILY_MZ)
                    {
                        // No need subtract the calibration offset for PIC32MZ it's all done
                        // in Harmony during initialisation
                        adc_data.adc_values[x] = DRV_ADC_SamplesRead(adc_data.adc_input_defs[x].input);
                    }

                    // Process the value that has been read
                    switch(adc_data.adc_input_defs[x].mode)
                    {
                        case ADC_SCAN_INPUT_MODE_BUTTONS:
                        {
                            // Process the switch definition
                            adc_process_switches(x);
                            break;
                        }

                        case ADC_SCAN_INPUT_MODE_TEMPERATURE:
                            // No action if adc_get_input_temperature() is used)
                            break;

                        case ADC_SCAN_INPUT_MODE_VOLTAGE:
                            // No action if adc_get_input_mV is used
                            break;

                        case ADC_SCAN_INPUT_MODE_POT:
                            adc_process_pot(x);
                            break;

                        default:
                            break;
                    }
                    
                    adc_data.adc_input_defs[x].processed = true;
                }
            }
        }
    }
        
    if (has_stopped)
    {
        // Now check whether the previous button is different to the current one after review
        // If so we'll store it as the previous button
        if (previous_switch != adc_data.current_switch.adc_switch)
        {
            uint32_t ticks = SYS_TMR_TickCountGet();
            adc_data.previous_switch.adc_switch = previous_switch;
            adc_data.previous_switch.start_ticks = adc_data.current_switch.start_ticks;
            adc_data.previous_switch.duration_ticks = (ticks + UINT32_MAX + adc_data.previous_switch.start_ticks) % UINT32_MAX;

            adc_data.current_switch.start_ticks = ticks;
            adc_data.current_switch.duration_ticks = 0;
        }

        // calibration mode false, so if we get here, we have made at least
        // one set of proper readings 
        adc_data.adc_started = true;
    
        // Restart the ADC
        DRV_ADC_Start();

        // Set result
        result = true;
    }
    else
    {
        // reset current switch
        adc_data.current_switch.adc_switch = previous_switch;
    }

    return result;
}
    
/**-----------------------------------------------------------------------------
 * Sets whether the input is currently active or not 
 */
void adc_set_input_active(unsigned int input_index, bool active)
{
    if (input_index < adc_data.adc_input_count)
    {
        adc_data.adc_input_defs[input_index].active = active;
        logger_log_infof(LOGGER_CODE_ADC, "ADC: Input %d set to %s ", input_index, adc_data.adc_input_defs[input_index].active ? "Active" : "Inactive");
    }
}


int8_t balance_pot_offset(void)
{
    return 0;//TODO????????????????????????????????????????????????? (adc_databalance - 12); // offset so centre is 0
}



void reset_current_switch(void)
{
    adc_data.current_switch.adc_switch = ADC_SWITCH_NONE; // clear up after go to standby
}

/* *****************************************************************************
 End of File
 */
