/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\hdmi_arc_control.h"
#include "app_config.h"
#include "utilities/logger/logger.h"
#include "utilities\rs232_hdmi_arc.h"
#include "utilities/volume.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */
typedef struct
{
    //! used to supprss the sending of message when processing messages (Volume)
    bool messages_suppressed;
} HDMI_ARC_CONTROL_DATA;

static HDMI_ARC_CONTROL_DATA hdmi_arc_control_data = 
{
    .messages_suppressed = false,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Builds a standard message buffer for simple 1 byte data messages
 * @param buffer
 * @return 
 */
static uint8_t hdmi_arc_build_simple_message(uint8_t* buffer, HDMI_ARC_COMMAND_TYPE type, uint8_t value)
{
    buffer[0] = HDMI_ARC_MESSAGE_START_OF_PACKET;
    buffer[1] = 6;
    buffer[2] = 0;
    buffer[3] = type;
    buffer[4] = value;
    buffer[5] = HDMI_ARC_MESSAGE_END_OF_PACKET;    
}

/**
 * Checksum supplied by Hualu
 * @param data
 * @param len
 * @return 
 */
static uint8_t hdmi_arc_generate_simple_checksum(const uint8_t* data, uint32_t len)
{
    uint8_t sum1 = 0;
    uint8_t sum2 = 0;

    while( len-- )
    {
        sum1 += *data++;
        sum2 += sum1;
    }

    return sum1 ^ sum2 ;  
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Informs the hdmi arc that the devices ARC Stream state has changed
 * @param state
 */
void hdmi_arc_send_arc_stream_message(HDMI_ARC_STREAM_STATE state)
{
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending ARC Stream %d --", state);
    
    uint8_t buffer[6];
    hdmi_arc_build_simple_message(buffer, (uint8_t)HDMI_ARC_COMMAND_ARC_STREAM, (uint8_t)state);
    rs232_hdmi_arc_send_buffer(buffer, 6);
}

/**-----------------------------------------------------------------------------
 * Informs the hdmi arc that the devices cec name has changed
 * @param state
 */
void hdmi_arc_send_cec_name_message(char* cec_name, uint8_t len)
{
    // Note given were using a s256 buffer len should always be below 256 - 7
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending CEC Name %s --", cec_name);
    
    
    uint8_t buffer[256];
    buffer[0] = HDMI_ARC_MESSAGE_START_OF_PACKET;
    buffer[1] = len + 5 + 2;    // Length + 5 for header/tail and 2 for BIN16
    buffer[2] = 0;
    buffer[3] = HDMI_ARC_COMMAND_CEC_NAME;

    // BIN16 length
    buffer[4] = len;
    buffer[5] = 0;
    
    // BIN 16 data
    memcpy(&buffer[6], (uint8_t*)cec_name, len);

    buffer[len + 4 + 2] = HDMI_ARC_MESSAGE_END_OF_PACKET;    

    rs232_hdmi_arc_send_buffer(buffer, buffer[1]);
}

/**-----------------------------------------------------------------------------
 * Informs the hdmi arc that the devices mute state has changed
 * @param state
 */
void hdmi_arc_send_mute_message(HDMI_ARC_MUTE_STATE state)
{
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Mute %d --", state);
    
    uint8_t buffer[6];
    hdmi_arc_build_simple_message(buffer, (uint8_t)HDMI_ARC_COMMAND_MUTE, (uint8_t)state);
    rs232_hdmi_arc_send_buffer(buffer, 6);
}

/**-----------------------------------------------------------------------------
 * Informs the hdmi arc that the devices power state has changed
 * @param state
 */
void hdmi_arc_send_power_message(HDMI_ARC_POWER_STATE state)
{
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Power %d --", state);
    
    uint8_t buffer[6];
    hdmi_arc_build_simple_message(buffer, (uint8_t)HDMI_ARC_COMMAND_POWER, (uint8_t)state);
    rs232_hdmi_arc_send_buffer(buffer, 6);
}

/**-----------------------------------------------------------------------------
 * Informs the hdmi arc that the devices power state has changed
 * @param state
 */
void hdmi_arc_send_system_ready_message(HDMI_ARC_SYSTEM_READY_STATE state)
{
    uint8_t buffer[6];
    hdmi_arc_build_simple_message(buffer, (uint8_t)HDMI_ARC_COMMAND_SYSTEM_READY, (uint8_t)state);
    rs232_hdmi_arc_send_buffer(buffer, 6);
}

/**-----------------------------------------------------------------------------
 * Sends a request for the HDMI ARC Transmitter version
 */
void hdmi_arc_send_transmitter_version_request()
{
    uint8_t buffer[5];
    buffer[0] = HDMI_ARC_MESSAGE_START_OF_PACKET;
    buffer[1] = 5;
    buffer[2] = 0;
    buffer[3] = HDMI_ARC_COMMAND_ARC_TRANSMITTER_FIRMWARE_VERSION;
    buffer[4] = HDMI_ARC_MESSAGE_END_OF_PACKET;    
    rs232_hdmi_arc_send_buffer(buffer, 5);   
}

/**-----------------------------------------------------------------------------
 * Sends and Upgrade data packet to the ARC MCU
 */
void hdmi_arc_send_upgrade_data_message(uint32_t total_size, uint32_t offset, uint8_t* data, uint16_t len)
{
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Upgrade Data %d %d %d  --", total_size, offset, len);
    logger_process();
    
    uint16_t message_len = len + 2 + 14;
      
    uint8_t buffer[HDMI_ARC_UPLOAD_DATA_MAX_PACKET_SIZE + 20]; // (+20 is for contingency should only need 16))
    memset(buffer, 0, HDMI_ARC_UPLOAD_DATA_MAX_PACKET_SIZE + 20);
    buffer[0] = HDMI_ARC_MESSAGE_START_OF_PACKET;
    buffer[1] = message_len % 256;//len + 5 + 2 + 2 + 8;    // Length + 5 for header/tail and 2 for BIN16 + 4 + 4 for uint32 fiels + 1 for checksum
    buffer[2] = message_len / 256;
    buffer[3] = HDMI_ARC_COMMAND_UPGRADE_DATA;
    
    // Total Size
    memcpy(&buffer[4], &total_size, sizeof(uint32_t));

    // Offset
    memcpy(&buffer[8], &offset, sizeof(uint32_t));

    buffer[13] = len % 256;//len + 5 + 2 + 1 + 8;    // Length + 5 for header/tail and 2 for BIN16 + 4 + 4 for uint32 fiels + 1 for checksum
    buffer[14] = len / 256;

    // Data
    memcpy(&buffer[15], data, len);
    
    // Checksum
    //uint8_t checksum = hdmi_arc_generate_simple_checksum((uint8_t*)data, len);
    uint8_t checksum = hdmi_arc_generate_simple_checksum((uint8_t*)&buffer[13], len + 2);  // Generate checksum from BIN16 field

    buffer[12] = checksum;
    
//    int x;
//    for (x = 0; x < len; x++)
//    {
//        buffer[x + 15] = data[x];
//    }
   
    buffer[message_len - 1] = HDMI_ARC_MESSAGE_END_OF_PACKET;    

    rs232_hdmi_arc_send_buffer(buffer, message_len);
}

/**-----------------------------------------------------------------------------
 * Informs the hdmi arc that upgrade data is available
 * @param md5_checksum - A pointer to an MD5 checksum for the whole upgrade file
 */
void hdmi_arc_send_upgrade_start_message()
{
   // Note given were using a s256 buffer len should always be below 256 - 7
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Upgrade Start --");
      
    uint8_t buffer[25];
    buffer[0] = HDMI_ARC_MESSAGE_START_OF_PACKET;
    buffer[1] = 6;
    buffer[2] = 0;
    buffer[3] = HDMI_ARC_COMMAND_UPGRADE_START;
    buffer[4] = HDMI_ARC_UPGRADE_START_STATE_UPGRADE_AVAILABLE;    
    buffer[5] = HDMI_ARC_MESSAGE_END_OF_PACKET;    

    rs232_hdmi_arc_send_buffer(buffer, buffer[1]);
}

/**-----------------------------------------------------------------------------
 * Informs the hdmi arc that upgrade data has been successfully transferred
  */
void hdmi_arc_send_upgrade_end_message()
{
   // Note given were using a s256 buffer len should always be below 256 - 7
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Upgrade End --");
      
    uint8_t buffer[25];
    buffer[0] = HDMI_ARC_MESSAGE_START_OF_PACKET;
    buffer[1] = 6;
    buffer[2] = 0;
    buffer[3] = HDMI_ARC_COMMAND_UPGRADE_END;
    buffer[4] = HDMI_ARC_UPGRADE_END_STATE_TRANSFER_COMPLETE;    
    buffer[5] = HDMI_ARC_MESSAGE_END_OF_PACKET;    

    rs232_hdmi_arc_send_buffer(buffer, buffer[1]);
}

/**-----------------------------------------------------------------------------
 * Informs the hdmi arc that the devices power state has changed
 * @param state
 */
void hdmi_arc_send_volume_message(uint8_t volume)
{
    if (!hdmi_arc_control_data.messages_suppressed)
    {
        //logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Volume %d --", volume);
        
        // Map the volume 
        uint8_t max_volume;
        settings_get(SETTING_MAXIMUM_VOLUME, &max_volume);
        
        uint16_t mapped_volume = volume * VOLUME_MAX;
        mapped_volume = min(mapped_volume / max_volume, VOLUME_MAX);

        logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Setting (Mapped) Volume %d %d ", volume, mapped_volume);

        uint8_t buffer[6];
        hdmi_arc_build_simple_message(buffer, (uint8_t)HDMI_ARC_COMMAND_VOLUME, mapped_volume);
        rs232_hdmi_arc_send_buffer(buffer, 6);
    }
    else
    {
        logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Sending Volume %d Suppressed --", volume);        
    }
}
    
/**
 * Used to suppress ARC messages being sent while a message is being processed 
 * by the arc module (set to solve problem where volume mapping caused rounding issues) 
 * @param on
 */
void hdmi_arc_suppress_messages(bool on)
{
    hdmi_arc_control_data.messages_suppressed = on;
}

/* *****************************************************************************
 End of File
 */
