/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities\hdmi_arc_processing.h"
#include <stdio.h>
#include "app_config.h"
#include "utilities/timers.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

typedef struct
{
    // Callback for app_commands layer
    HDMI_ARC_MESSAGE_CALLBACK callback;
    
    //! The current ARC format
    HDMI_ARC_FORMAT format;
    
    uint8_t firmware_version_major;
    uint8_t firmware_version_minor;
    uint8_t rx_firmware_version_major;
    uint8_t rx_firmware_version_minor;
    uint8_t rx_firmware_version_build;
    
} HDMI_ARC_DATA;

static HDMI_ARC_DATA hdmi_arc_data = {
    .callback = NULL,
    .format = HDMI_ARC_FORMAT_ARC,
    .firmware_version_major = 0,
    .firmware_version_minor = 0,
    .rx_firmware_version_major = 0,
    .rx_firmware_version_minor = 0,
    .rx_firmware_version_build = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Processes the ARC Format message from the HDMI ARC
 * @param message
 */
static void hdmi_arc_process_arc_format(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_ARC_FORMAT)
    {
        hdmi_arc_data.format = (HDMI_ARC_FORMAT)message->data[0];
        logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Setting Format %d %s --", hdmi_arc_data.format, HDMI_ARC_FORMAT_STRINGS[hdmi_arc_data.format] );
    }
}

/**-----------------------------------------------------------------------------
 * Processes the ARC Format message from the HDMI ARC
 * @param message
 */
static void hdmi_arc_process_firmware_version(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_ARC_FIRMWARE_VERSION)
    {
        hdmi_arc_data.firmware_version_major = message->data[0];
        hdmi_arc_data.firmware_version_minor = message->data[1];
        logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Setting Version %d.%d --", 
                hdmi_arc_data.firmware_version_major,
                hdmi_arc_data.firmware_version_minor);
   }
}

/**-----------------------------------------------------------------------------
 * Processes the ARC Format message from the HDMI ARC
 * @param message
 */
static void hdmi_arc_process_transmitter_firmware_version(HDMI_ARC_MESSAGE* message)
{
    if (message->command_type == HDMI_ARC_COMMAND_ARC_TRANSMITTER_FIRMWARE_VERSION)
    {
        hdmi_arc_data.rx_firmware_version_build = message->data[0];
        hdmi_arc_data.rx_firmware_version_minor = message->data[1];
        hdmi_arc_data.rx_firmware_version_major = message->data[2];
        logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Setting Transmitter Version %d.%d.%d --", 
                hdmi_arc_data.rx_firmware_version_major,
                hdmi_arc_data.rx_firmware_version_minor,
                hdmi_arc_data.rx_firmware_version_build);
   }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Gets the current HDMI ARC Format as a string
 * @param buffer
 * @param len
 */
void hdmi_arc_get_arc_format_string(char* buffer, uint8_t len)
{
    sprintf(buffer, "%s", HDMI_ARC_FORMAT_STRINGS[hdmi_arc_data.format]);
}

/**-----------------------------------------------------------------------------
 * Gets the Major software version
 */
uint8_t hdmi_arc_get_firmware_version_major()
{
    return hdmi_arc_data.firmware_version_major;
}

/**-----------------------------------------------------------------------------
 * Gets the Minor software version
 */
uint8_t hdmi_arc_get_firmware_version_minor()
{
    return hdmi_arc_data.firmware_version_minor;
}
    
/**-----------------------------------------------------------------------------
 * Gets the current HDMI Firmware Versiont as a string
 * @param buffer
 * @param len
 */
void hdmi_arc_get_firmware_version_string(char* buffer, uint8_t len)
{
    sprintf(buffer, "v%d.%d", hdmi_arc_data.firmware_version_major, hdmi_arc_data.firmware_version_minor);    
}

/**-----------------------------------------------------------------------------
 * Gets the Major software version
 */
uint8_t hdmi_arc_get_rx_firmware_version_major()
{
    return hdmi_arc_data.rx_firmware_version_major;
}

/**-----------------------------------------------------------------------------
 * Gets the Minor software version
 */
uint8_t hdmi_arc_get_rx_firmware_version_minor()
{
    return hdmi_arc_data.rx_firmware_version_minor;
}

/**-----------------------------------------------------------------------------
 * Gets the current HDMI Firmware Versiont as a string
 * @param buffer
 * @param len
 */
void hdmi_arc_get_rx_firmware_version_string(char* buffer, uint8_t len)
{
    sprintf(buffer, "v%d.%d.%d", hdmi_arc_data.rx_firmware_version_major, hdmi_arc_data.rx_firmware_version_minor, hdmi_arc_data.rx_firmware_version_build);    
}
    
/**-----------------------------------------------------------------------------
 * Used to determine whether the firmware version has been retrieved from the ARC module 
 * @return 
 */
bool hdmi_arc_has_firmware_version()
{
    return ((hdmi_arc_get_firmware_version_major() != 0) || (hdmi_arc_get_firmware_version_minor() != 0));
}
    
/**-----------------------------------------------------------------------------
 * Initiates a fresh ARC handshake
 */
void hdmi_arc_initiate_handshake()
{
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Initiating Handshake");
            
    // Set version to 0.0 
    hdmi_arc_data.firmware_version_major = 0;
    hdmi_arc_data.firmware_version_minor = 0;
    
    // Reset and expire timer
    timers_activate(TIMER_INDEX_HDMI_ARC_HANDSHAKE);
    timers_expire(TIMER_INDEX_HDMI_ARC_HANDSHAKE);
}
    
/**-----------------------------------------------------------------------------
 * Callback for RS232 connection
 * @param message
 */
void hdmi_arc_processing_callback(HDMI_ARC_MESSAGE* message)
{
    logger_log_debugf(LOGGER_CODE_HDMI_ARC, "HDMI ARC Processing message %d %s --", message->command_type, HDMI_ARC_COMMAND_STRINGS[message->command_type]);
    
    hdmi_arc_process_arc_format(message);
    
    hdmi_arc_process_firmware_version(message);
    
    hdmi_arc_process_transmitter_firmware_version(message);    

    // Invoke the app_command layer callback this will process the other messages
    // Mute, Power, ARC Stream, Volume
    if (hdmi_arc_data.callback != NULL)
    {
        hdmi_arc_data.callback(message);
    }
    
    logger_process();
}
    
/**-----------------------------------------------------------------------------
 * HDMI ARC Message processing initialisation
 */  
void hdmi_arc_processing_initialise(HDMI_ARC_MESSAGE_CALLBACK callback)
{
    hdmi_arc_data.callback = callback;
}
    
/* *****************************************************************************
 End of File
 */
