/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/input_context.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

/**
 * Will be updated by the application during the processing of inputs in a single cycle
 * and is then used to perform any appropriate actions
 */
typedef struct
{
    // The currently active button
    ADC_BUTTON current_button;
    
    // Indicates whether the current button has changed in this cycle
    bool current_button_changed;
    
    //! When set to true any changes in the buttons are ignored (but still recorded))
    bool button_change_suspended;

    // The current state of the encoder
    ENCODER_STATE encoder_state[ENCODER_INDEX_MAX];
    
    // Indicates whether an external message has been processed that causes a change
    // By default external messages will trigger this flag, however there are certain
    // circumstances when we don't want to (mainly around checking auto shutdown 
    // features)
    bool external_changed;
    
    //
    bool balance_changed;
    
    // The last message (that defined activity) processed that was generated by rs232)
    STRUCTURED_MESSAGE external_message;
    
    // The current message from the remote
    STRUCTURED_MESSAGE remote_message;
    
    // Indicates whether the remote has been changed during this cycle
    bool remote_changed;
    
    //! When set to true any changes in the remote are ignored (but still recorded))
    bool remote_change_suspended;
    
    // Indicates whether the context has been changed since the last reset
    bool changed;

} INPUT_CONTEXT_DATA;

static INPUT_CONTEXT_DATA input_context_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * Used to indicate that an external message has been received (used for redrawing
 */
void input_context_external_message_processed(STRUCTURED_MESSAGE* message)
{
    structured_message_copy(message, &input_context_data.external_message);
    input_context_data.external_changed = true;
    input_context_data.changed = true;
}

/**-----------------------------------------------------------------------------
 * Gets the current button
 */
ADC_BUTTON input_context_get_button()
{
    return input_context_data.current_button;
}

/**-----------------------------------------------------------------------------
 * Gets the current encoder state
 */
ENCODER_STATE input_context_get_encoder(ENCODER_INDEX index)
{
    return input_context_data.encoder_state[index];
}

/**-----------------------------------------------------------------------------
 * Gets the last processed external message (an initialised message if a reset has been done)
 * Should be used in conjunction with input_context_has_external_changed
 */
STRUCTURED_MESSAGE* input_context_get_external_message()
{
    return &input_context_data.external_message;
}
    
/**-----------------------------------------------------------------------------
 * Gets the current remote state
 */
STRUCTURED_MESSAGE* input_context_get_remote_message()
{
    return &input_context_data.remote_message;
}
    
/**-----------------------------------------------------------------------------
 * Indicates whether the balance pot has been moved
 */
bool input_context_has_balance_changed()
{
    return input_context_data.balance_changed;
}

/**-----------------------------------------------------------------------------
 * Indicates whether the context button has changed assuming an attempt is 
 * made to set it each cycle
 * @return 
 */
bool input_context_has_button_changed()
{
    return input_context_data.current_button_changed && !input_context_data.button_change_suspended;
}

/**-----------------------------------------------------------------------------
 * Indicates whether the context has changed since the last reset
 * @return 
 */
bool input_context_has_changed()
{
    return input_context_data.changed;
}

/**-----------------------------------------------------------------------------
 * Indicates whether the context button has changed assuming an attempt is 
 * made to set it each cycle
 * @return 
 */
bool input_context_has_external_changed()
{
    return input_context_data.external_changed;
}

/**-----------------------------------------------------------------------------
 * Indicates whether the remote has changed since the last reset
 * @return 
 */
bool input_context_has_remote_changed()
{
    return input_context_data.remote_changed && !input_context_data.remote_change_suspended;
}

/**-----------------------------------------------------------------------------
 * Resets the changed state of the context - should be called at the start of a cycle
 * Will not change current button/encoder state (see input_context_clear())
 */
void input_context_reset()
{
    input_context_data.remote_changed = false;
    input_context_data.changed = false;
    input_context_data.current_button_changed = false;
    input_context_data.remote_changed = false;
    input_context_data.external_changed = false;
    input_context_data.balance_changed = false;
    structured_message_initialise(&input_context_data.external_message, STRUCTURED_MESSAGE_SOURCE_RS232);
}

/**-----------------------------------------------------------------------------
 * Sets whether th (SA750) Balance pot has changed
 */
void input_context_set_balance_changed(bool changed)
{
    input_context_data.balance_changed = changed;
    if (changed)
    {
        input_context_data.changed = true;
    }
}

/**-----------------------------------------------------------------------------
 * Sets the current button - this assumes it will be called at least once per cycle
 * @param current_button
 */
void input_context_set_button(ADC_BUTTON button)
{
    if (input_context_data.current_button != button)
    {
        //logger_log_debugf(0, "BUTTON Change %d %d --", input_context_data.current_button, button);
        input_context_data.current_button = button;
        input_context_data.current_button_changed = true;
        input_context_data.changed = true;
    }
    else
    {
        input_context_data.current_button_changed = false;
    }
}

/**-----------------------------------------------------------------------------
 * Sets the current encoder state
 * @param state
 * @param moved 
 */
void input_context_set_encoder(ENCODER_STATE* state, ENCODER_INDEX index)
{
    if (state->delta != 0)
    {
        //logger_log_debugf(0, "ENCODER[%d] Setting State ", index);
        input_context_data.encoder_state[index] = *state;
        input_context_data.changed = true;
    }
    else 
   {
        // Set the state to unchanged
        input_context_data.encoder_state[index].delta = 0;
        input_context_data.encoder_state[index].moved = 0;
    }
}
    
/**-----------------------------------------------------------------------------
 * Sets the remote structured message
 * @param mode - A structured message containing information from the remote
 */
void input_context_set_remote_message(STRUCTURED_MESSAGE* message)
{
    structured_message_copy(message, &input_context_data.remote_message);
    
    if (remote_use_message(message))
    {
        //logger_log_debugf(241, "INPUT CONTEXT Using Remote Message %d %d --", remote_get_system_code_from_message(message), message->command);
        input_context_data.remote_changed = true;
        input_context_data.changed = true;
    }
}

/**-----------------------------------------------------------------------------
 * Suspends/Unsuspends button processing 
 * @param suspend
 */
void input_context_suspend_button(bool suspend)
{
    input_context_data.button_change_suspended = suspend;
}

/**-----------------------------------------------------------------------------
 * Suspends/Unsuspends remote processing 
 * @param suspend
 */
void input_context_suspend_remote(bool suspend)
{
    input_context_data.remote_change_suspended = suspend;
}

/* *****************************************************************************
 End of File
 */
