/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Arcam

  @File Name
    rs232.c

  @Summary
 Handles communication with the RS232 port

  @Description
 Maintains an RS232 connection and parses standard messages
 * 
 * v2.1 - Includes display_message processing
 *        This is a DYNAMIC implementation - to work the number of Driver Clients needs to be > 1 in Harmony (I used 4)
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/rs232_hdmi_arc.h"
#include "app_config.h"
#include "utilities/hdmi_arc_data.h"
#include "utilities/logger/logger.h"
#include "utilities/timers.h"
#include "driver/usart/drv_usart.h"

#include <xc.h>

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

#define RS232_HDMI_ARC_MESSAGE_POOL 4
/**
 * Holds global data
 */
typedef struct
{
    // I2C handle
    DRV_HANDLE          usart_handle;
    
    //! Callback for display messages
    HDMI_ARC_MESSAGE_CALLBACK hdmi_arc_callback;
    
    //! Display message to hold incoming Display message data
    //HDMI_ARC_MESSAGE hdmi_arc_message;
    
    //! Display message to hold incoming Display message data
    HDMI_ARC_MESSAGE hdmi_arc_messages[RS232_HDMI_ARC_MESSAGE_POOL];
    
    // The position of the Message currently being populated
    uint8_t current_read_write_idx;
    
    // The position of the Message to be read
    uint8_t current_read_read_idx;
    
    //! Buffers used for transferring large images - these need to be at least big enough
    //  to store an entire image
    uint8_t write_buffer[RS232_HDMI_ARC_TX_QUEUE_SIZE][RS232_HDMI_ARC_TX_BUFFER_SIZE];
    uint8_t write_buffer_len[RS232_HDMI_ARC_TX_QUEUE_SIZE];
    
    //! Simple circular buffer index
    uint8_t write_buffer_idx;
    bool write_buffer_complete[RS232_HDMI_ARC_TX_QUEUE_SIZE];
    DRV_USART_BUFFER_HANDLE write_buffer_handle[RS232_HDMI_ARC_TX_QUEUE_SIZE];
     
    //! Buffers used for transferring large images - these need to be at least big enough
    //  to store an entire image
    uint8_t read_buffer[RS232_HDMI_ARC_RX_QUEUE_SIZE][RS232_HDMI_ARC_RX_BUFFER_SIZE];
    bool read_buffer_complete[RS232_HDMI_ARC_RX_QUEUE_SIZE];
    DRV_USART_BUFFER_HANDLE read_buffer_handle[RS232_HDMI_ARC_RX_QUEUE_SIZE];
    
    //! Simple circular buffer index
    uint8_t read_buffer_read_idx;
} RS232_DATA;

static RS232_DATA rs232_data = {
//    .write_buffer_idx = 0,
//    .write_buffer_read_idx = 0,
//    .read_buffer_idx = 0,
    .read_buffer_read_idx = 0,
    .current_read_write_idx = 0,
};

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

static void rs232_hdmi_arc_initialise_read_buffer(uint8_t idx);

/**-----------------------------------------------------------------------------
 * 
 * @param event
 * @param bufferHandle
 * @param contextHandle
 */
static void rs232_hdmi_arc_callback(DRV_USART_BUFFER_EVENT event,
        DRV_USART_BUFFER_HANDLE bufferHandle, 
        uintptr_t contextHandle)
{
    bool found = false;
    
    switch(event)
    {
        case DRV_USART_BUFFER_EVENT_COMPLETE:
        {
            // This means the data was transferred.
            int x;
            for (x = 0; x < RS232_HDMI_ARC_RX_QUEUE_SIZE; x++)
            {
                if (rs232_data.read_buffer_handle[x] == bufferHandle)
                {
                    HDMI_ARC_MESSAGE* message = &rs232_data.hdmi_arc_messages[rs232_data.current_read_write_idx];
                    
                    // Add to the current message
                    int y;
                    for (y = 0; y < RS232_HDMI_ARC_RX_BUFFER_SIZE; y++)
                    {
                        uint8_t data_byte = rs232_data.read_buffer[x][y];
                        if (data_byte != 0xFD)
                        {
//                            if (data_byte == 0x23)
//                            {
//                                logger_log_debugf(LOGGER_CODE_RS232_HDMI_ARC, "HDMI RS232 Data Received +++ 0x23 %d %d %X %d --", bufferHandle, x, data_byte, rs232_data.current_read_write_idx);
//                            }
//                            else if (data_byte == 0x0D)
//                            {
//                                logger_log_debugf(LOGGER_CODE_RS232_HDMI_ARC, "HDMI RS232 Data Received +++ 0x0D %d %d %X %d --", bufferHandle, x, data_byte, rs232_data.current_read_write_idx);
//                            }
//                            else
//                            {
//                                logger_log_debugf(LOGGER_CODE_RS232_HDMI_ARC, "HDMI RS232 Data Received A %d %d %X %d --", bufferHandle, x, data_byte, rs232_data.current_read_write_idx);
//                            }
                    
                            hdmi_arc_message_process_byte(message, (unsigned char)data_byte);

                            if (hdmi_arc_message_is_complete(message))
                            {
                                //logger_log_debugf(LOGGER_CODE_RS232_HDMI_ARC, "HDMI RS232 Data Complete +++ 0x0D %d %d %X %d --", bufferHandle, x, data_byte, rs232_data.current_read_write_idx);
                                rs232_data.current_read_write_idx = ++rs232_data.current_read_write_idx % RS232_HDMI_ARC_MESSAGE_POOL;
                            }                        
                        }
                        else
                        {
                            break;
                        }
                        
                        rs232_hdmi_arc_initialise_read_buffer(x);
                    }
                     
                    //logger_log_debugf(LOGGER_CODE_RS232_DISPLAY, "RS232 Reading Data %d --", bufferHandle);
                    rs232_data.read_buffer_complete[x] = true;

                    found = true;
                    break;
                }
            }
            
            if (!found)
            {
                for (x = 0; x < RS232_HDMI_ARC_TX_QUEUE_SIZE; x++)
                {
                    if (rs232_data.write_buffer_handle[x] == bufferHandle)
                    {
                        rs232_data.write_buffer_complete[x] = true;
                        //logger_log_debugf(LOGGER_CODE_RS232_HDMI_ARC, "HDMI RS232 Data Written A %d %d %X --", bufferHandle, x, rs232_data.write_buffer[x][0]);

                        found = true;
                        break;
                    }
                }
            }
            
            break;
        }

        case DRV_USART_BUFFER_EVENT_ERROR:
        {
            DRV_USART_ERROR error = DRV_USART_ErrorGet(rs232_data.usart_handle);            

            logger_log_errorf(LOGGER_CODE_RS232_HDMI_ARC, "RS232 Data Error %d %d --", bufferHandle, error);
            
            // Error handling here.
            int x;
            for (x = 0; x < RS232_HDMI_ARC_RX_QUEUE_SIZE; x++)
            {
                if (rs232_data.read_buffer_handle[x] == bufferHandle)
                {
                    logger_log_errorf(LOGGER_CODE_RS232_HDMI_ARC, "RS232 Data Error Resetting %d --", bufferHandle);
                    
                    rs232_hdmi_arc_initialise_read_buffer(x);
                    
                    if (x == rs232_data.read_buffer_read_idx)
                    {
                        // Increment the read index
                        logger_log_errorf(LOGGER_CODE_RS232_HDMI_ARC, "RS232 Data Error Incr read index %d --", bufferHandle);
                        rs232_data.read_buffer_read_idx = ++rs232_data.read_buffer_read_idx % RS232_HDMI_ARC_RX_QUEUE_SIZE; 
                    }

                    //rs232_data.read_buffer_handle[x] = DRV_USART_BUFFER_HANDLE_INVALID;
                    found = true;
                    
                    HDMI_ARC_MESSAGE* message = &rs232_data.hdmi_arc_messages[rs232_data.current_read_write_idx];
                    hdmi_arc_message_initialise(message);
                    break;
                }
            }
            
            if (!found)
            {
                for (x = 0; x < RS232_HDMI_ARC_TX_QUEUE_SIZE; x++)
                {
                    if (rs232_data.write_buffer_handle[x] == bufferHandle)
                    {
                        rs232_data.write_buffer_handle[x] = DRV_USART_BUFFER_HANDLE_INVALID;
                        rs232_data.write_buffer_complete[x] = true;

                        found = true;
                        break;
                    }
                }
            }

            break;
        }
        
        default:
            break;
    }
}

/**-----------------------------------------------------------------------------
 * 
 * @param index
 */
static void rs232_hdmi_arc_initialise_read_buffer(uint8_t idx)
{
    rs232_data.read_buffer_complete[idx] = false;
    memset(&rs232_data.read_buffer[idx][0], 0xFD, RS232_HDMI_ARC_RX_BUFFER_SIZE);

    DRV_USART_BUFFER_HANDLE buffer_handle;
    DRV_USART_BufferAddRead(rs232_data.usart_handle, 
            &buffer_handle,
            &rs232_data.read_buffer[idx][0], 
            RS232_HDMI_ARC_RX_BUFFER_SIZE);

    if (DRV_USART_BUFFER_HANDLE_INVALID != buffer_handle)
    {
        rs232_data.read_buffer_handle[idx] = buffer_handle;
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */

/**-----------------------------------------------------------------------------
 * RS232 module initialisation
 */
void rs232_hdmi_arc_initialise(const SYS_MODULE_INDEX module, 
        HDMI_ARC_MESSAGE_CALLBACK hdmi_arc_callback)
{
    rs232_data.hdmi_arc_callback = hdmi_arc_callback;

    rs232_data.usart_handle = DRV_USART_Open(module, 
                                             DRV_IO_INTENT_NONBLOCKING | DRV_IO_INTENT_READWRITE);
    if (DRV_HANDLE_INVALID == rs232_data.usart_handle)
    {
        //settings_data.state = SETTINGS_STATE_ERROR;
        logger_log_error(LOGGER_CODE_RS232_HDMI_ARC, "USART Open");
    }    
    
    DRV_USART_BufferEventHandlerSet(rs232_data.usart_handle,
                rs232_hdmi_arc_callback, (uintptr_t)NULL);
    
    // Add read buffers
    int x;
    for (x = 0; x < RS232_HDMI_ARC_RX_QUEUE_SIZE; x++)
    {
        rs232_hdmi_arc_initialise_read_buffer(x);
    }
    
    // Initialise write buffers
    for (x = 0; x < RS232_HDMI_ARC_TX_QUEUE_SIZE; x++)
    {
        rs232_data.write_buffer_complete[x] = true;
    }
    
    // Initialise write buffers
    for (x = 0; x < RS232_HDMI_ARC_MESSAGE_POOL; x++)
    {
        hdmi_arc_message_initialise(&rs232_data.hdmi_arc_messages[x]);
    }
}

/**-----------------------------------------------------------------------------
 * RS232 processing
 */
void rs232_hdmi_arc_process()
{
    int x;
    for (x = 0; x < RS232_HDMI_ARC_MESSAGE_POOL; x++)
    {
        HDMI_ARC_MESSAGE* message = &rs232_data.hdmi_arc_messages[rs232_data.current_read_read_idx];
        if (hdmi_arc_message_is_complete(message))
        {
            //logger_log_debugf(LOGGER_CODE_RS232_HDMI_ARC, "HDMI RS232 Processing message %d %d --", rs232_data.current_read_read_idx);
            if (rs232_data.hdmi_arc_callback != NULL)
            {
                rs232_data.hdmi_arc_callback(message);
            }
           
           hdmi_arc_message_initialise(message);
        } 
        else if (rs232_data.current_read_read_idx == rs232_data.current_read_write_idx)
        {
            // We are upto the message currently being written
            //logger_log_debugf(0, "HDMI Processing Stopping processing %d %d --", rs232_data.current_read_read_idx, rs232_data.current_read_write_idx);
            break;
        }
        
        rs232_data.current_read_read_idx = ++rs232_data.current_read_read_idx % RS232_HDMI_ARC_MESSAGE_POOL;
    }
}

/**-----------------------------------------------------------------------------
 * Sends a buffer of data over the RS232 connection
 * @param buffer
 * @param len
 */
void rs232_hdmi_arc_send_buffer(char* buffer, size_t len)
{
    int total = 0;
    do
    {
        int count  = DRV_USART_Write(rs232_data.usart_handle, 
                                     &buffer[total], 
                                     len - total);
        if (count ==  DRV_USART_WRITE_ERROR)
        {
            logger_log_info(LOGGER_CODE_USART_WRITE_ERROR, "RS232: Write Error");
            break;
        }
        
        total += count;
    } while(total < len);
}

/* *****************************************************************************
 End of File
 */
