/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    settings.c

  @Summary
    Manages reading and writing settings to flash memory

  @Description
 * Reads and writes setting to an eeprom
 * The initial version is setup to write to a 24LC64.  This requires the memory 
 * address to be written to preceed the read/write operations.  This appears to 
 * be the same as the AT24C02 which will be used in the SA20
 * 
 * It also relies on a change to drv_i2c.c (see // ARCAM comments) where invalid 
 * messages were being sent which lock out the eeprom
 */
/* ************************************************************************** */

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: Included Files                                                    */
/* ************************************************************************** */
/* ************************************************************************** */

#include "utilities/settings.h"
#include "utilities/logger/logger.h"
#include "utilities/timers.h"
#include "system/common/sys_module.h"   
#include <string.h>

#include "utilities/flash_memory.h"

/* ************************************************************************** */
/* ************************************************************************** */
/* Section: File Scope or Global Data                                         */
/* ************************************************************************** */
/* ************************************************************************** */

//! Define the length required to hold the model name - this is used to identify 
//  whether the settings being read are for the correct model
//  Note: This should include a null at the end. e.g. SA20\0
#ifndef SETTINGS_MODEL_STRING_LENGTH
#define SETTINGS_MODEL_STRING_LENGTH    5
#endif

/**
 * Settings are stored in a single block of memory.  This value must be large enough
 * to hold all the settings values plus the model_string.
 * It will also be passed to the flash_memory module to advise it how much space 
 * the settings will take up.
 */
#ifndef SETTINGS_STORAGE_SIZE
#define SETTINGS_STORAGE_SIZE           256
#endif

/**
 * A generic Id that can be used to identify the settings module when a number of
 * modules may be using the same storage
 */
#ifndef SETTINGS_CLIENT_ID
#define SETTINGS_CLIENT_ID              0
#endif

/**
 * Hold the state of the settings
 */
typedef enum 
{
    SETTINGS_STATE_INITIALISING,
    SETTINGS_STATE_LOADING,
    SETTINGS_STATE_LOADED,
    SETTINGS_STATE_SAVING,

    //  We have either tried to read or write from the settings byte there has 
    // been errors
    SETTINGS_STATE_ERROR,
} SETTINGS_STATE;

/**
 * The data stored in the settings
 */
typedef struct
{
    // Contains the model number - this is used to determine whether the settings 
    // on the eeprom are valid 
    char model[SETTINGS_MODEL_STRING_LENGTH];
    
    // Array storing settings values
    // Note this should be the last member in the structure as the read only
    // partially reads into it
    char settings[SETTINGS_STORAGE_SIZE - SETTINGS_MODEL_STRING_LENGTH];
 
} SETTINGS_STORAGE;

/**
 * Settings global state data
 */
typedef struct
{
    // The current state of the settings re load/save
    SETTINGS_STATE state;
    
    // A pointer to a callback (specified during initialise) that can be used to 
    // setup default settings for the application)
    SETTINGS_INITIALISE_CALLBACK init_callback;

    //! Stores the flash memory client id for use in calls to flash memory
    //FLASH_MEMORY_CLIENT_ID client_id;
    
    //! The actual length of the data used by the combined settings.
    uint16_t data_length;
    
    // The highest index used when adding settings
    uint8_t highest_setting_idx;
    
    //! True when the settings are being loaded
    bool loading;
    
    // Indicates whether calls to setting_set should trigger a delayed save
    bool suppress_delayed_save;

    // The data stored in the settings flash
    SETTINGS_STORAGE storage;
    
    // Holds the size of the setting
    size_t storage_size[SETTINGS_NUM_SETTINGS];
    
    // Holds the ordinal position of the setting in the storage
    uint8_t storage_pos[SETTINGS_NUM_SETTINGS];
     
    // Indicated whether the settings have been updated since the last read
    bool updated;
     
} SETTINGS_DATA;

static SETTINGS_DATA settings_data;

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Local Functions                                                   */
/* ************************************************************************** */
/* ************************************************************************** */

//! Forward declarations
static void settings_loaded_from_storage();
static void settings_load();

/**-----------------------------------------------------------------------------
 * A Callback used for flash_memory functionality
 * @param page_data
 */
void settings_callback(FLASH_MEMORY_CLIENT_REQUEST* request)
{
    logger_log_debugf(0, "SETTINGS Flash Memory Callback client: %d --", request->client_id);
    
    switch(settings_data.state)
    {
        case SETTINGS_STATE_LOADING:
        {
            settings_loaded_from_storage();
            settings_data.loading = false;
            break;
        }
        
        case SETTINGS_STATE_SAVING:
        {
            settings_data.updated = false;
            settings_data.state = SETTINGS_STATE_LOADED;
            logger_log_info(LOGGER_CODE_SETTINGS_SETTINGS_SAVED, "SETTINGS: Saved --");
            break;
        }
    }
}

/**
 * Initialise the storage
 */
static void settings_initialise_storage()
{
    memcpy(&settings_data.storage.model, SOFTWARE_MODEL_STRING, SETTINGS_MODEL_STRING_LENGTH - 1);
    memset(settings_data.storage.settings, 0, (sizeof(uint8_t) * SETTINGS_STORAGE_SIZE) - SETTINGS_MODEL_STRING_LENGTH);
}

/**
 * Initiates reading the settings from the Flash memory
 * Note: If a setting has been updated it will be overwritten (open to review!!)
 */
static void settings_load()
{
    if (!settings_data.loading)
    {
        logger_log_debugf(0, "SETTINGS Loading --");
        settings_data.loading = true;
        settings_data.updated = false;

        // Initialise the storage - note this will erase any updated settings, though
        // this shouldn't be a problem as we aim to read only once.  If we do then we 
        // need a check for settings_data.updated...
        //settings_initialise_storage();

        // Put blanks into the model so that the comparison will fail if nothing is read
        memcpy(&settings_data.storage.model, "    ", 4);

        // MAke a flash request that can be passed back in when its done????
        FLASH_MEMORY_CLIENT_REQUEST request;
        request.client_id = SETTINGS_CLIENT_ID;
        request.file_index = 0;
        request.data = (void*)&settings_data.storage;
        request.offset = 0;
        request.length = sizeof(SETTINGS_STORAGE);

        flash_memory_read(&request);
    }
}

/**
 * Processes the newly loaded settings, initiates a save if they have not been 
 * persisted
 */
static void settings_loaded_from_storage()
{
    char* model = SOFTWARE_MODEL_STRING;

    // Load of faff for a debug message but when we have a new eeprom the debug 
    // message is failing when the storage is set to all 0xFF..
    if ((uint8_t)settings_data.storage.model[0] == 0xFF)
    {
        logger_log_debugf(0, "SETTINGS: check model expected: %s found: Uninitialised EEPROM --", model);
    }
    else
    {
        logger_log_debugf(0, "SETTINGS: check model expected: %s found: %s --", model, settings_data.storage.model);
    }

    if (strncmp(settings_data.storage.model, model, SETTINGS_MODEL_STRING_LENGTH - 1) != 0)
    {
        // The settings have not been stored before
        settings_data.state = SETTINGS_STATE_LOADED;
        logger_log_info(LOGGER_CODE_SETTINGS_SAVE_SETTINGS, "SETTINGS: Saving Defaults --");

        // Reinitialise all the settings
        settings_initialise_storage();

        // Use a callback to set defaults
        if (settings_data.init_callback != NULL)
        {
            settings_data.init_callback();
        }

        // Save the settings
        settings_data.updated = true;

        settings_save();
    }
    else
    {
        settings_data.updated = false;

        logger_log_info(LOGGER_CODE_SETTINGS_SETTINGS_LOADED, "SETTINGS: Loaded --");

        settings_data.state = SETTINGS_STATE_LOADED;
    }
}

/* ************************************************************************** */
/* ************************************************************************** */
// Section: Interface Functions                                               */
/* ************************************************************************** */
/* ************************************************************************** */
    
/**
 * Adds a setting for the given index.
 * Note all the settings should be added before any are written
 * @param index - the index to refer to for the setting
 * @param size - the size (in bytes) of the setting
 */
void settings_add(uint8_t index, size_t size)
{
    if (index < SETTINGS_NUM_SETTINGS)
    {
        settings_data.highest_setting_idx = max(settings_data.highest_setting_idx, index);
        settings_data.storage_size[index] = size;
        
        // Sort out where in the settings buffer the particular setting will be stored
        int x = max(index, 1);
        for (; x < SETTINGS_NUM_SETTINGS; x++)
        {
            settings_data.storage_pos[x] = settings_data.storage_pos[x - 1] +
                    settings_data.storage_size[x - 1];
        }
        
        // Update page count
        settings_data.data_length = SETTINGS_MODEL_STRING_LENGTH +
             settings_data.storage_pos[settings_data.highest_setting_idx] +
             settings_data.storage_size[settings_data.highest_setting_idx];
    }
}

/**
 * Used to see if the settings are being saved/loaded
 * Note: Only returns true during normal running
 */
bool settings_available()
{
    bool result = false;
    
    if (settings_data.state == SETTINGS_STATE_LOADED)
    {
        result = true;
    }
    
    return result;
}

/**
 * Get the specified setting
 * @param index - The index of the setting
 * @param output - A pointer to storage to receive the setting value
  */
void settings_get(uint8_t index, void* output)
{
    int result = SETTINGS_INVALID_INDEX;
    
    if (index < SETTINGS_NUM_SETTINGS)
    {
        if (settings_data.state <= SETTINGS_STATE_LOADING)
        {
            logger_log_warningf(LOGGER_CODE_DEBUG, "SETTINGS Attempt to read %s when SETTINGS NOT LOADED! --", SETTING_STRINGS[index]);
        }
    
        char* setting = &settings_data.storage.settings[settings_data.storage_pos[index]];
        size_t len = settings_data.storage_size[index];

        memcpy(output, setting, len);
        
        //logger_log_debugf(LOGGER_CODE_DEBUG, "SETTINGS: get %s %d %d %d --", SETTING_STRINGS[index], settings_data.storage_pos[index], len, *(char*)output);
        //logger_process();
    }
}

/**
 * Initialises the resources used for the settings
 */
void settings_initialise(SETTINGS_INITIALISE_CALLBACK callback)
{
    settings_data.state = SETTINGS_STATE_INITIALISING;
    settings_data.init_callback = callback;
    
    settings_data.loading = false;
    settings_data.updated = false;
    settings_data.suppress_delayed_save = false;
    settings_data.highest_setting_idx = 0;
    
    //This is set by settings_add - settings_data.page_count = 0;
    memset(&settings_data.storage_pos, 0, sizeof(uint8_t) * SETTINGS_NUM_SETTINGS);
    memset(&settings_data.storage_size, 0, sizeof(uint8_t) * SETTINGS_NUM_SETTINGS);
    
    // Initialise settings
    settings_initialise_storage();
    
    // Add the storage to the flash memory
    flash_memory_add_client(SETTINGS_CLIENT_ID, SETTINGS_STORAGE_NAME, &settings_callback);

    int x;
    for (x = 0; x < SETTING_MAX; x++)
    {
        settings_add(x, SETTINGS_LENGTH[x]);
    }
    
    settings_data.state = SETTINGS_STATE_LOADING;
}
    
/**
 * Used to check whether the settings have been initialised.
 * Note during setup this need to be called repeatedly in the main application 
 * loop until settings have been loaded
 * @return - True if the settings have been initialised, false if not.
 */
bool settings_initialised()
{
    bool result = false;
    
    settings_process();
    Nop();
   
    if (settings_data.state == SETTINGS_STATE_LOADED)
    {
        result = true;
    }
    
    return result;
}

/**
 * Checks whether settings are ready to save and performs save and reset
 */
void settings_process()
{
    switch (settings_data.state)
    {
        case SETTINGS_STATE_LOADING:
        {
            settings_load();
            break;
        }

        default:
            break;
    }
            
    // Check the timer
    // If it has expired and there are unsaved changes then save the current state
    if (timers_has_expired(TIMER_INDEX_SETTINGS))
    {
        if (settings_data.updated)
        {
            settings_save();
        }
    }
}

/**
 * Saves the current state of the settings (if changed)
 */
void settings_save()
{
    if (settings_data.updated)
    {
        logger_log_infof(LOGGER_CODE_SETTINGS_SAVING_SETTINGS, "SETTINGS: Saving --");
        settings_data.state = SETTINGS_STATE_SAVING;

        // MAke a flash request that can be passed back in when its done????
        FLASH_MEMORY_CLIENT_REQUEST request;
        request.client_id = SETTINGS_CLIENT_ID;
        request.file_index = 0;
        request.data = (void*)&settings_data.storage;
        request.offset = 0;
        request.length = sizeof(SETTINGS_STORAGE);

        flash_memory_write(&request);
    }
}

/**
 * Set the specified setting
 * @param index - The index of the setting
 * @return 
 */
int settings_set(uint8_t index, void* value)
{
    if (index < SETTINGS_NUM_SETTINGS)
    {
        char* setting = &settings_data.storage.settings[settings_data.storage_pos[index]];
        size_t len = settings_data.storage_size[index];

        if (memcmp(setting, value, len) != 0)
        {
            memcpy(setting, value, len);
           
            settings_data.updated = true;
            
            // Check whether timer is active
            // If not reset the timer to start a countdown to save
            if (timers_has_expired(TIMER_INDEX_SETTINGS) && 
                !settings_data.suppress_delayed_save)
            {
                timers_reset(TIMER_INDEX_SETTINGS);
            }

            logger_log_infof(LOGGER_CODE_SETTINGS_SETTINGS_SET, "SETTINGS set %s  len %d value %d dly_save %d --", SETTING_STRINGS[index], len, *setting, settings_data.suppress_delayed_save);
        }
    }
    else
    {
        logger_log_errorf(LOGGER_CODE_SETTINGS_SETTINGS_SET, "SETTINGS Invalid Index %d --", index);
    }
}

/**
 * Use to temporarily suppress/un-suppress the automatic triggering of a save.
 * Note if a save has already been triggered then it will still go ahead, and any
 * changes made if this is subsequently set before the save is complete will 
 * also be saved
 * This was intended to prevent a save during the loading of settings where 
 * the loaded settings may be used to set properties such as volume/balance etc
 * using methods that may inturn usually trigger another save.
 * @param suppress - true - an automatic delayed save will not be triggered
 * by any calls to settings_set(), false - calls to settings_set(0 will trigger 
 * an automatic delayed save
 */
void settings_suppress_save(bool suppress)
{
    settings_data.suppress_delayed_save = suppress;
}

/* *****************************************************************************
 End of File
 */
